/* Generated by Cython 0.29.33 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/arrayobject.h",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/arrayscalars.h",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/ndarrayobject.h",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/ndarraytypes.h",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/ufuncobject.h"
        ],
        "include_dirs": [
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/ms_peak_picker/_c",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/brainpy/_c"
        ],
        "name": "ms_deisotope._c.feature_map.feature_map",
        "sources": [
            "src/ms_deisotope/_c/feature_map/feature_map.pyx"
        ]
    },
    "module_name": "ms_deisotope._c.feature_map.feature_map"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_33"
#define CYTHON_HEX_VERSION 0x001D21F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 0
  #elif !defined(CYTHON_FAST_THREAD_STATE)
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030A0000)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__ms_deisotope___c__feature_map__feature_map
#define __PYX_HAVE_API__ms_deisotope___c__feature_map__feature_map
/* Early includes */
#include <string.h>
#include <stdio.h>
#include "numpy/arrayobject.h"
#include "numpy/ndarrayobject.h"
#include "numpy/ndarraytypes.h"
#include "numpy/arrayscalars.h"
#include "numpy/ufuncobject.h"

    /* NumPy API declarations from "numpy/__init__.pxd" */
    
#include "pythread.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "src/ms_deisotope/_c/feature_map/feature_map.pyx",
  "src/ms_deisotope/_c/feature_map/feature_map.pxd",
  "stringsource",
  "__init__.pxd",
  "array.pxd",
  "peak_set.pxd",
  "src/ms_deisotope/_c/feature_map/lcms_feature.pxd",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
};

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":689
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */

#line 689 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":690
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */

#line 690 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":691
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */

#line 691 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":692
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */

#line 692 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":696
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */

#line 696 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":697
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */

#line 697 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":698
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */

#line 698 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":699
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */

#line 699 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":703
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */

#line 703 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":704
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */

#line 704 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":713
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */

#line 713 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":714
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */

#line 714 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":715
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */

#line 715 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":717
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */

#line 717 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":718
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */

#line 718 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":719
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */

#line 719 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":721
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */

#line 721 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":722
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */

#line 722 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":724
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */

#line 724 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":725
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */

#line 725 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":726
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */

#line 726 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/
#ifndef _ARRAYARRAY_H
struct arrayobject;
typedef struct arrayobject arrayobject;
#endif
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap;
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak;
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for;
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell;
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list;
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted;

/* "ms_peak_picker/_c/peak_set.pxd":42
 *     cdef FittedPeak _get_nearest_peak(self, double mz, double* errout)
 * 
 *     cdef FittedPeak _has_peak(self, double mz, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef PeakSet clone(self)
 */

#line 42 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak {

#line 42 "ms_peak_picker/_c/peak_set.pxd"
  int __pyx_n;

#line 42 "ms_peak_picker/_c/peak_set.pxd"
  double tolerance;

#line 42 "ms_peak_picker/_c/peak_set.pxd"
};

/* "ms_peak_picker/_c/peak_set.pxd":50
 *     cdef int _between_bounds(self, double m1, double m2, size_t* startp, size_t* endp)
 * 
 *     cpdef tuple all_peaks_for(self, double mz, double error_tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef FittedPeak getitem(self, size_t i)
 */

#line 50 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for {

#line 50 "ms_peak_picker/_c/peak_set.pxd"
  int __pyx_n;

#line 50 "ms_peak_picker/_c/peak_set.pxd"
  double error_tolerance;

#line 50 "ms_peak_picker/_c/peak_set.pxd"
};

/* "ms_peak_picker/_c/peak_set.pxd":64
 * 
 * 
 * cdef struct index_cell:             # <<<<<<<<<<<<<<
 *     double center_value
 *     size_t start
 */

#line 64 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell {

#line 64 "ms_peak_picker/_c/peak_set.pxd"
  double center_value;

#line 64 "ms_peak_picker/_c/peak_set.pxd"
  size_t start;

#line 64 "ms_peak_picker/_c/peak_set.pxd"
  size_t end;

#line 64 "ms_peak_picker/_c/peak_set.pxd"
};

/* "ms_peak_picker/_c/peak_set.pxd":70
 * 
 * 
 * cdef struct index_list:             # <<<<<<<<<<<<<<
 *     index_cell* index
 *     size_t size
 */

#line 70 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list {

#line 70 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell *index;

#line 70 "ms_peak_picker/_c/peak_set.pxd"
  size_t size;

#line 70 "ms_peak_picker/_c/peak_set.pxd"
  double low;

#line 70 "ms_peak_picker/_c/peak_set.pxd"
  double high;

#line 70 "ms_peak_picker/_c/peak_set.pxd"
};

/* "ms_peak_picker/_c/peak_set.pxd":90
 *     cpdef _allocate_index(self, size_t interval_index_size)
 * 
 *     cpdef FittedPeak has_peak_hinted(self, double mz, double tolerance=*, size_t hint_start=*, size_t hint_end=*)             # <<<<<<<<<<<<<<
 * 
 * cdef size_t double_binary_search_ppm(double* array, double value, double tolerance, size_t n)
 */

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted {

#line 90 "ms_peak_picker/_c/peak_set.pxd"
  int __pyx_n;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
  double tolerance;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
  size_t hint_start;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
  size_t hint_end;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
};

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__invalidate;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_split_sparse;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_clone;

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":98
 *     cpdef insert_node(self, LCMSFeatureTreeNode node)
 *     cpdef insert(self, PeakBase peak, double time)
 *     cpdef _invalidate(self, bint reaverage=*)             # <<<<<<<<<<<<<<
 *     cpdef list split_sparse(self, double delta_rt=*)
 *     cpdef LCMSFeature clone(self, deep=*, cls=*)
 */

#line 98 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__invalidate {

#line 98 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_n;

#line 98 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int reaverage;

#line 98 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":99
 *     cpdef insert(self, PeakBase peak, double time)
 *     cpdef _invalidate(self, bint reaverage=*)
 *     cpdef list split_sparse(self, double delta_rt=*)             # <<<<<<<<<<<<<<
 *     cpdef LCMSFeature clone(self, deep=*, cls=*)
 *     cpdef double max_intensity(self)
 */

#line 99 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_split_sparse {

#line 99 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_n;

#line 99 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double delta_rt;

#line 99 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":100
 *     cpdef _invalidate(self, bint reaverage=*)
 *     cpdef list split_sparse(self, double delta_rt=*)
 *     cpdef LCMSFeature clone(self, deep=*, cls=*)             # <<<<<<<<<<<<<<
 *     cpdef double max_intensity(self)
 * 
 */

#line 100 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_clone {

#line 100 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_n;

#line 100 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *deep;

#line 100 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *cls;

#line 100 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":728
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */

#line 728 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":729
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */

#line 729 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":730
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */

#line 730 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":732
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */

#line 732 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_cdouble __pyx_t_5numpy_complex_t;

#line 732 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_find_all;

#line 732 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_search;

#line 732 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_between;

#line 732 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_clone;

#line 732 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_split_sparse;

/* "ms_deisotope/_c/feature_map/feature_map.pxd":11
 *     cdef Py_ssize_t get_size(self)
 *     cdef LCMSFeature get(self, size_t i)
 *     cpdef list find_all(self, double mz, double error_tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef LCMSFeature search(self, double mz, double error_tolerance=*)
 *     cpdef list between(self, double lo, double hi, double error_tolerance=*)
 */

#line 11 "ms_deisotope/_c/feature_map/feature_map.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_find_all {

#line 11 "ms_deisotope/_c/feature_map/feature_map.pxd"
  int __pyx_n;

#line 11 "ms_deisotope/_c/feature_map/feature_map.pxd"
  double error_tolerance;

#line 11 "ms_deisotope/_c/feature_map/feature_map.pxd"
};

/* "ms_deisotope/_c/feature_map/feature_map.pxd":12
 *     cdef LCMSFeature get(self, size_t i)
 *     cpdef list find_all(self, double mz, double error_tolerance=*)
 *     cpdef LCMSFeature search(self, double mz, double error_tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef list between(self, double lo, double hi, double error_tolerance=*)
 *     cpdef list spanning_time(self, double time_point)
 */

#line 12 "ms_deisotope/_c/feature_map/feature_map.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_search {

#line 12 "ms_deisotope/_c/feature_map/feature_map.pxd"
  int __pyx_n;

#line 12 "ms_deisotope/_c/feature_map/feature_map.pxd"
  double error_tolerance;

#line 12 "ms_deisotope/_c/feature_map/feature_map.pxd"
};

/* "ms_deisotope/_c/feature_map/feature_map.pxd":13
 *     cpdef list find_all(self, double mz, double error_tolerance=*)
 *     cpdef LCMSFeature search(self, double mz, double error_tolerance=*)
 *     cpdef list between(self, double lo, double hi, double error_tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef list spanning_time(self, double time_point)
 *     cpdef LCMSFeatureMap clone(self, bint deep=*)
 */

#line 13 "ms_deisotope/_c/feature_map/feature_map.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_between {

#line 13 "ms_deisotope/_c/feature_map/feature_map.pxd"
  int __pyx_n;

#line 13 "ms_deisotope/_c/feature_map/feature_map.pxd"
  double error_tolerance;

#line 13 "ms_deisotope/_c/feature_map/feature_map.pxd"
};

/* "ms_deisotope/_c/feature_map/feature_map.pxd":15
 *     cpdef list between(self, double lo, double hi, double error_tolerance=*)
 *     cpdef list spanning_time(self, double time_point)
 *     cpdef LCMSFeatureMap clone(self, bint deep=*)             # <<<<<<<<<<<<<<
 * 
 * cpdef tuple binary_search_with_flag(list array, double mz, double error_tolerance)
 */

#line 15 "ms_deisotope/_c/feature_map/feature_map.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_clone {

#line 15 "ms_deisotope/_c/feature_map/feature_map.pxd"
  int __pyx_n;

#line 15 "ms_deisotope/_c/feature_map/feature_map.pxd"
  int deep;

#line 15 "ms_deisotope/_c/feature_map/feature_map.pxd"
};

/* "ms_deisotope/_c/feature_map/feature_map.pxd":21
 * cdef void search_sweep(list array, double mz, double error_tolerance, long* loout, long* hiout)
 * 
 * cpdef list split_sparse(LCMSFeatureMap self, double delta_rt=*, size_t min_size=*)             # <<<<<<<<<<<<<<
 */

#line 21 "ms_deisotope/_c/feature_map/feature_map.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_split_sparse {

#line 21 "ms_deisotope/_c/feature_map/feature_map.pxd"
  int __pyx_n;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pxd"
  double delta_rt;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pxd"
  size_t min_size;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pxd"
};

/* "ms_peak_picker/_c/peak_set.pxd":2
 * 
 * cdef class PeakBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 2 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase {

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  PyObject_HEAD

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_vtab;

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  double mz;

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  double intensity;

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  double area;

#line 2 "ms_peak_picker/_c/peak_set.pxd"
};

#line 2 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":12
 * 
 * 
 * cdef class FittedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double signal_to_noise
 */

#line 12 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak {

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  double signal_to_noise;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  double full_width_at_half_max;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  double left_width;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  double right_width;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  long peak_count;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  long index;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
};

#line 12 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":29
 *                             double area)
 * 
 * cdef class PeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */

#line 29 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet {

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  PyObject_HEAD

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_vtab;

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  PyObject *peaks;

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  int indexed;

#line 29 "ms_peak_picker/_c/peak_set.pxd"
};

#line 29 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":83
 * 
 * 
 * cdef class PeakSetIndexed(PeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* mz_index
 */

#line 83 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed {

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet __pyx_base;

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  double *mz_index;

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list *interval_index;

#line 83 "ms_peak_picker/_c/peak_set.pxd"
};

#line 83 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":4
 * 
 * 
 * cdef class LCMSFeatureTreeList(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list roots
 */

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList {

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject_HEAD

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_vtab;

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *roots;

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *_node_id_hash;

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":23
 * 
 * 
 * cdef class LCMSFeatureTreeNodeBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double time
 */

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase {

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject_HEAD

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_vtab;

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double time;

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *members;

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *_most_abundant_member;

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *node_id;

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":44
 * 
 * 
 * cdef class LCMSFeatureTreeNode(LCMSFeatureTreeNodeBase):             # <<<<<<<<<<<<<<
 *     cdef:
 * 
 */

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode {

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase __pyx_base;

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double _mz;

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":56
 * 
 * 
 * cdef class FeatureBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public LCMSFeatureTreeList nodes
 */

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase {

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject_HEAD

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_vtab;

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *nodes;

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double _mz;

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double _start_time;

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double _end_time;

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":73
 * 
 * 
 * cdef class LCMSFeature(FeatureBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double _total_intensity
 */

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature {

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase __pyx_base;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double _total_intensity;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double _last_mz;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *_times;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *_peaks;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *adducts;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *used_as_adduct;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *created_at;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *feature_id;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *_peak_averager;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":104
 * 
 * 
 * cdef class EmptyFeature(FeatureBase):             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef EmptyFeature _create(double mz)
 */

#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature {

#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase __pyx_base;

#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":109
 * 
 * 
 * cdef class FeatureSetIterator(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list features
 */

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator {

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject_HEAD

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_vtab;

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *features;

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *real_features;

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double start_time;

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double end_time;

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double last_time_seen;

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  size_t *index_list;

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":137
 * 
 * 
 * cdef class RunningWeightedAverage(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double current_mean
 */

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage {

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject_HEAD

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_vtab;

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double current_mean;

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  size_t current_count;

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double total_weight;

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":154
 * 
 * 
 * cdef class RunningWeightedAverageNeutralMass(RunningWeightedAverage):             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass {

#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage __pyx_base;

#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/feature_map.pxd":5
 * 
 * 
 * cdef class LCMSFeatureMap(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list features
 */

#line 5 "ms_deisotope/_c/feature_map/feature_map.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap {

#line 5 "ms_deisotope/_c/feature_map/feature_map.pxd"
  PyObject_HEAD

#line 5 "ms_deisotope/_c/feature_map/feature_map.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_vtab;

#line 5 "ms_deisotope/_c/feature_map/feature_map.pxd"
  PyObject *features;

#line 5 "ms_deisotope/_c/feature_map/feature_map.pxd"
};

#line 5 "ms_deisotope/_c/feature_map/feature_map.pxd"


#line 5 "ms_deisotope/_c/feature_map/feature_map.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":2
 * 
 * cdef class PeakBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 2 "ms_peak_picker/_c/peak_set.pxd"


#line 2 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase {

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*clone)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch);

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch);

#line 2 "ms_peak_picker/_c/peak_set.pxd"
};

#line 2 "ms_peak_picker/_c/peak_set.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase;

#line 2 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":12
 * 
 * 
 * cdef class FittedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double signal_to_noise
 */

#line 12 "ms_peak_picker/_c/peak_set.pxd"


#line 12 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak {

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  int (*_eq)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int __pyx_skip_dispatch);

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_create)(double, double, double, double, double, double, long, long, double);

#line 12 "ms_peak_picker/_c/peak_set.pxd"
};

#line 12 "ms_peak_picker/_c/peak_set.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak;

#line 12 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":29
 *                             double area)
 * 
 * cdef class PeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */

#line 29 "ms_peak_picker/_c/peak_set.pxd"


#line 29 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet {

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*_create)(PyObject *);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  PyObject *(*reindex)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  size_t (*_index)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_get_nearest_peak)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double *);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_has_peak)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak *__pyx_optional_args);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*clone)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*_between)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  int (*_between_bounds)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double, size_t *, size_t *);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  PyObject *(*all_peaks_for)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for *__pyx_optional_args);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*getitem)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, size_t);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  size_t (*get_size)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  int (*_eq)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
};

#line 29 "ms_peak_picker/_c/peak_set.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet;

#line 29 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":83
 * 
 * 
 * cdef class PeakSetIndexed(PeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* mz_index
 */

#line 83 "ms_peak_picker/_c/peak_set.pxd"


#line 83 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed {

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet __pyx_base;

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  PyObject *(*_allocate_index)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *, size_t, int __pyx_skip_dispatch);

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*has_peak_hinted)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted *__pyx_optional_args);

#line 83 "ms_peak_picker/_c/peak_set.pxd"
};

#line 83 "ms_peak_picker/_c/peak_set.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed;

#line 83 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":4
 * 
 * 
 * cdef class LCMSFeatureTreeList(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list roots
 */

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList {

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *(*_create)(PyObject *);

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  void (*_invalidate)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *);

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *(*_make_node)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *, double, PyObject *, int __pyx_skip_dispatch);

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*find_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *, double, int __pyx_skip_dispatch);

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *(*_find_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *, double, size_t *);

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *, size_t);

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *);

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList;

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":23
 * 
 * 
 * cdef class LCMSFeatureTreeNodeBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double time
 */

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase {

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double (*_total_intensity_members)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double (*max_intensity)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double (*total_intensity)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*_find_most_abundant_member)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *);

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*_calculate_most_abundant_member)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*_recalculate)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, size_t);

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  size_t (*get_members_size)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *);

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase;

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":44
 * 
 * 
 * cdef class LCMSFeatureTreeNode(LCMSFeatureTreeNodeBase):             # <<<<<<<<<<<<<<
 *     cdef:
 * 
 */

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode {

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase __pyx_base;

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, int __pyx_skip_dispatch);

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int (*_ne)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, int __pyx_skip_dispatch);

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*getpeak)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, size_t);

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double (*get_mz)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *);

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode;

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":56
 * 
 * 
 * cdef class FeatureBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public LCMSFeatureTreeList nodes
 */

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase {

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double (*get_mz)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *);

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double (*get_neutral_mass)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *);

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double (*get_start_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *);

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double (*get_end_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *);

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *);

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*find_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *, double, int __pyx_skip_dispatch);

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *(*_find_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *, double, size_t *);

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase;

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":73
 * 
 * 
 * cdef class LCMSFeature(FeatureBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double _total_intensity
 */

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature {

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase __pyx_base;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, PyObject *, int __pyx_skip_dispatch);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int (*_ne)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, PyObject *, int __pyx_skip_dispatch);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int (*overlaps_in_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *, int __pyx_skip_dispatch);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int (*spans_in_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, double, int __pyx_skip_dispatch);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, size_t);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*_initialize_averager)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*_update_from_averager)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*_reaverage_and_update)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*insert_node)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, int __pyx_skip_dispatch);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*insert)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, double, int __pyx_skip_dispatch);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*_invalidate)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__invalidate *__pyx_optional_args);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*split_sparse)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_split_sparse *__pyx_optional_args);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *(*clone)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_clone *__pyx_optional_args);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double (*max_intensity)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":104
 * 
 * 
 * cdef class EmptyFeature(FeatureBase):             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef EmptyFeature _create(double mz)
 */

#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature {

#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase __pyx_base;

#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *(*_create)(double);

#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature;

#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":109
 * 
 * 
 * cdef class FeatureSetIterator(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list features
 */

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator {

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *(*_create)(PyObject *);

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *(*_create_with_threshold)(PyObject *, PyObject *, double);

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int (*_initialize)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, PyObject *);

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*init_indices)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, int __pyx_skip_dispatch);

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double (*get_next_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, int __pyx_skip_dispatch);

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double (*get_current_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, int __pyx_skip_dispatch);

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int (*has_more)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, int __pyx_skip_dispatch);

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*get_peaks_for_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, double, int __pyx_skip_dispatch);

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*get_next_value)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, int __pyx_skip_dispatch);

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *);

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, size_t);

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator;

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":137
 * 
 * 
 * cdef class RunningWeightedAverage(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double current_mean
 */

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage {

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*reset)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, int __pyx_skip_dispatch);

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*_initialize)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, int __pyx_skip_dispatch);

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *(*add)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch);

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *(*update)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, PyObject *, int __pyx_skip_dispatch);

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int (*feed_from_feature)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch);

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *(*_create)(PyObject *);

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  void (*_update)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, PyObject *);

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage;

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":154
 * 
 * 
 * cdef class RunningWeightedAverageNeutralMass(RunningWeightedAverage):             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass {

#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage __pyx_base;

#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *(*_create)(PyObject *);

#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass;

#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/feature_map.pyx":19
 * 
 * 
 * cdef class LCMSFeatureMap(object):             # <<<<<<<<<<<<<<
 *     def __init__(self, features):
 *         self.features = sorted(features, key=lambda x: x.mz)
 */

#line 19 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 19 "ms_deisotope/_c/feature_map/feature_map.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap {

#line 19 "ms_deisotope/_c/feature_map/feature_map.pyx"
  Py_ssize_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *);

#line 19 "ms_deisotope/_c/feature_map/feature_map.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *(*get)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *, size_t);

#line 19 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *(*find_all)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_find_all *__pyx_optional_args);

#line 19 "ms_deisotope/_c/feature_map/feature_map.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *(*search)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_search *__pyx_optional_args);

#line 19 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *(*between)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *, double, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_between *__pyx_optional_args);

#line 19 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *(*spanning_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *, double, int __pyx_skip_dispatch);

#line 19 "ms_deisotope/_c/feature_map/feature_map.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *(*clone)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_clone *__pyx_optional_args);

#line 19 "ms_deisotope/_c/feature_map/feature_map.pyx"
};

#line 19 "ms_deisotope/_c/feature_map/feature_map.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;  // used by FusedFunction for copying defaults
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_Check(obj)  (__Pyx_TypeCheck(obj, __pyx_CyFunctionType))
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if CYTHON_FAST_PYCALL
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif // CYTHON_FAST_PYCALL
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* DivInt[long].proto */
static CYTHON_INLINE long __Pyx_div_long(long, long);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* ArrayAPI.proto */
#ifndef _ARRAYARRAY_H
#define _ARRAYARRAY_H
typedef struct arraydescr {
    int typecode;
    int itemsize;
    PyObject * (*getitem)(struct arrayobject *, Py_ssize_t);
    int (*setitem)(struct arrayobject *, Py_ssize_t, PyObject *);
#if PY_MAJOR_VERSION >= 3
    char *formats;
#endif
} arraydescr;
struct arrayobject {
    PyObject_HEAD
    Py_ssize_t ob_size;
    union {
        char *ob_item;
        float *as_floats;
        double *as_doubles;
        int *as_ints;
        unsigned int *as_uints;
        unsigned char *as_uchars;
        signed char *as_schars;
        char *as_chars;
        unsigned long *as_ulongs;
        long *as_longs;
#if PY_MAJOR_VERSION >= 3
        unsigned long long *as_ulonglongs;
        long long *as_longlongs;
#endif
        short *as_shorts;
        unsigned short *as_ushorts;
        Py_UNICODE *as_pyunicodes;
        void *as_voidptr;
    } data;
    Py_ssize_t allocated;
    struct arraydescr *ob_descr;
    PyObject *weakreflist;
#if PY_MAJOR_VERSION >= 3
        int ob_exports;
#endif
};
#ifndef NO_NEWARRAY_INLINE
static CYTHON_INLINE PyObject * newarrayobject(PyTypeObject *type, Py_ssize_t size,
    struct arraydescr *descr) {
    arrayobject *op;
    size_t nbytes;
    if (size < 0) {
        PyErr_BadInternalCall();
        return NULL;
    }
    nbytes = size * descr->itemsize;
    if (nbytes / descr->itemsize != (size_t)size) {
        return PyErr_NoMemory();
    }
    op = (arrayobject *) type->tp_alloc(type, 0);
    if (op == NULL) {
        return NULL;
    }
    op->ob_descr = descr;
    op->allocated = size;
    op->weakreflist = NULL;
    __Pyx_SET_SIZE(op, size);
    if (size <= 0) {
        op->data.ob_item = NULL;
    }
    else {
        op->data.ob_item = PyMem_NEW(char, nbytes);
        if (op->data.ob_item == NULL) {
            Py_DECREF(op);
            return PyErr_NoMemory();
        }
    }
    return (PyObject *) op;
}
#else
PyObject* newarrayobject(PyTypeObject *type, Py_ssize_t size,
    struct arraydescr *descr);
#endif
static CYTHON_INLINE int resize(arrayobject *self, Py_ssize_t n) {
    void *items = (void*) self->data.ob_item;
    PyMem_Resize(items, char, (size_t)(n * self->ob_descr->itemsize));
    if (items == NULL) {
        PyErr_NoMemory();
        return -1;
    }
    self->data.ob_item = (char*) items;
    __Pyx_SET_SIZE(self, n);
    self->allocated = n;
    return 0;
}
static CYTHON_INLINE int resize_smart(arrayobject *self, Py_ssize_t n) {
    void *items = (void*) self->data.ob_item;
    Py_ssize_t newsize;
    if (n < self->allocated && n*4 > self->allocated) {
        __Pyx_SET_SIZE(self, n);
        return 0;
    }
    newsize = n + (n / 2) + 1;
    if (newsize <= n) {
        PyErr_NoMemory();
        return -1;
    }
    PyMem_Resize(items, char, (size_t)(newsize * self->ob_descr->itemsize));
    if (items == NULL) {
        PyErr_NoMemory();
        return -1;
    }
    self->data.ob_item = (char*) items;
    __Pyx_SET_SIZE(self, n);
    self->allocated = newsize;
    return 0;
}
#endif

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* VoidPtrImport.proto */
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static Py_ssize_t __pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_get_size(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_get(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, size_t __pyx_v_i); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_search(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, double __pyx_v_mz, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_search *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_find_all(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, double __pyx_v_mz, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_find_all *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_spanning_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, double __pyx_v_time_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_between(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, double __pyx_v_lo, double __pyx_v_hi, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_between *__pyx_optional_args); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_clone(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_clone *__pyx_optional_args); /* proto*/

/* Module declarations from 'ms_peak_picker._c.peak_set' */
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase = 0;
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak = 0;
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet = 0;
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = 0;
static size_t *__pyx_vp_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE = 0;
#define __pyx_v_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE (*__pyx_vp_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE)

/* Module declarations from 'ms_deisotope._c.feature_map.lcms_feature' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass = 0;

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython.version' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'cpython.module' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cpython.tuple' */

/* Module declarations from 'cpython.sequence' */

/* Module declarations from 'cpython.mapping' */

/* Module declarations from 'cpython.iterator' */

/* Module declarations from 'cpython.number' */

/* Module declarations from 'cpython.int' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.bool' */
static PyTypeObject *__pyx_ptype_7cpython_4bool_bool = 0;

/* Module declarations from 'cpython.long' */

/* Module declarations from 'cpython.float' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.complex' */
static PyTypeObject *__pyx_ptype_7cpython_7complex_complex = 0;

/* Module declarations from 'cpython.string' */

/* Module declarations from 'cpython.unicode' */

/* Module declarations from 'cpython.dict' */

/* Module declarations from 'cpython.instance' */

/* Module declarations from 'cpython.function' */

/* Module declarations from 'cpython.method' */

/* Module declarations from 'cpython.weakref' */

/* Module declarations from 'cpython.getargs' */

/* Module declarations from 'cpython.pythread' */

/* Module declarations from 'cpython.pystate' */

/* Module declarations from 'cpython.cobject' */

/* Module declarations from 'cpython.oldbuffer' */

/* Module declarations from 'cpython.set' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'cpython.pycapsule' */

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cython' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_generic = 0;
static PyTypeObject *__pyx_ptype_5numpy_number = 0;
static PyTypeObject *__pyx_ptype_5numpy_integer = 0;
static PyTypeObject *__pyx_ptype_5numpy_signedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_unsignedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_inexact = 0;
static PyTypeObject *__pyx_ptype_5numpy_floating = 0;
static PyTypeObject *__pyx_ptype_5numpy_complexfloating = 0;
static PyTypeObject *__pyx_ptype_5numpy_flexible = 0;
static PyTypeObject *__pyx_ptype_5numpy_character = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE int __pyx_f_5numpy_import_array(void); /*proto*/

/* Module declarations from 'array' */

/* Module declarations from 'cpython.array' */
static PyTypeObject *__pyx_ptype_7cpython_5array_array = 0;
static CYTHON_INLINE int __pyx_f_7cpython_5array_extend_buffer(arrayobject *, char *, Py_ssize_t); /*proto*/

/* Module declarations from 'ms_deisotope._c.feature_map.feature_map' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap = 0;
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_binary_search_with_flag(PyObject *, double, double, int __pyx_skip_dispatch); /*proto*/
static long __pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_binary_search(PyObject *, double, double); /*proto*/
static void __pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_search_sweep(PyObject *, double, double, long *, long *); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_split_sparse(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_split_sparse *__pyx_optional_args); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map___pyx_unpickle_LCMSFeatureMap__set_state(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *, PyObject *); /*proto*/
#define __Pyx_MODULE_NAME "ms_deisotope._c.feature_map.feature_map"
extern int __pyx_module_is_main_ms_deisotope___c__feature_map__feature_map;
int __pyx_module_is_main_ms_deisotope___c__feature_map__feature_map = 0;

/* Implementation of 'ms_deisotope._c.feature_map.feature_map' */
static PyObject *__pyx_builtin_sorted;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_ImportError;
static PyObject *__pyx_builtin_MemoryError;
static const char __pyx_k_L[] = "L";
static const char __pyx_k_d[] = "d";
static const char __pyx_k_hi[] = "hi";
static const char __pyx_k_lo[] = "lo";
static const char __pyx_k_mz[] = "mz";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_copy[] = "copy";
static const char __pyx_k_deep[] = "deep";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_array[] = "array";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_clone[] = "clone";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_stack[] = "stack";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_search[] = "search";
static const char __pyx_k_sorted[] = "sorted";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_between[] = "between";
static const char __pyx_k_lexsort[] = "lexsort";
static const char __pyx_k_delta_rt[] = "delta_rt";
static const char __pyx_k_features[] = "features";
static const char __pyx_k_find_all[] = "find_all";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_min_size[] = "min_size";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_split_sparse[] = "split_sparse";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_spanning_time[] = "spanning_time";
static const char __pyx_k_LCMSFeatureMap[] = "LCMSFeatureMap";
static const char __pyx_k_error_tolerance[] = "error_tolerance";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_init___locals_lambda[] = "__init__.<locals>.<lambda>";
static const char __pyx_k_pyx_unpickle_LCMSFeatureMap[] = "__pyx_unpickle_LCMSFeatureMap";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_self___class_____name___size_fe[] = "{self.__class__.__name__}(<{size} features>)";
static const char __pyx_k_src_ms_deisotope__c_feature_map[] = "src/ms_deisotope/_c/feature_map/feature_map.pyx";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0[] = "Incompatible checksums (0x%x vs (0x9f034b6, 0x5b8a8b5, 0x4f94454) = (features))";
static const char __pyx_k_ms_deisotope__c_feature_map_feat[] = "ms_deisotope._c.feature_map.feature_map";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0;
static PyObject *__pyx_n_s_L;
static PyObject *__pyx_n_s_LCMSFeatureMap;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_append;
static PyObject *__pyx_n_s_array;
static PyObject *__pyx_n_s_between;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_clone;
static PyObject *__pyx_n_s_copy;
static PyObject *__pyx_n_s_d;
static PyObject *__pyx_n_s_deep;
static PyObject *__pyx_n_s_delta_rt;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_error_tolerance;
static PyObject *__pyx_n_s_features;
static PyObject *__pyx_n_s_find_all;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_hi;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_init___locals_lambda;
static PyObject *__pyx_n_s_key;
static PyObject *__pyx_n_s_lexsort;
static PyObject *__pyx_n_s_lo;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_min_size;
static PyObject *__pyx_n_s_ms_deisotope__c_feature_map_feat;
static PyObject *__pyx_n_s_mz;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_s_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_s_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_LCMSFeatureMap;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_search;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_kp_s_self___class_____name___size_fe;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_size;
static PyObject *__pyx_n_s_sorted;
static PyObject *__pyx_n_s_spanning_time;
static PyObject *__pyx_n_s_split_sparse;
static PyObject *__pyx_kp_s_src_ms_deisotope__c_feature_map;
static PyObject *__pyx_n_s_stack;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_lambda_funcdef_lambda(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap___init__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, PyObject *__pyx_v_features); /* proto */
static Py_ssize_t __pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_2__len__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_4__iter__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_6__getitem__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_8__eq__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_10__ne__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_12search(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_14find_all(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_16index_range(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, PyObject *__pyx_v_lo, PyObject *__pyx_v_hi, PyObject *__pyx_v_error_tolerance); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_18spanning_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, double __pyx_v_time_point); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_20between(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, double __pyx_v_lo, double __pyx_v_hi, double __pyx_v_error_tolerance); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_22__repr__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_24clone(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, int __pyx_v_deep); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_26as_arrays(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_8features___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_8features_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_8features_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_28__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_30__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_split_sparse(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, double __pyx_v_delta_rt, size_t __pyx_v_min_size); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_2binary_search_with_flag(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_array, double __pyx_v_mz, double __pyx_v_error_tolerance); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_4__pyx_unpickle_LCMSFeatureMap(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7cpython_5array_5array___getbuffer__(arrayobject *__pyx_v_self, Py_buffer *__pyx_v_info, CYTHON_UNUSED int __pyx_v_flags); /* proto */
static void __pyx_pf_7cpython_5array_5array_2__releasebuffer__(CYTHON_UNUSED arrayobject *__pyx_v_self, Py_buffer *__pyx_v_info); /* proto */
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_float_2eneg_5;
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_83444820;
static PyObject *__pyx_int_95987893;
static PyObject *__pyx_int_166737078;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_codeobj__7;
static PyObject *__pyx_codeobj__9;
/* Late includes */

/* "ms_deisotope/_c/feature_map/feature_map.pyx":20
 * 
 * cdef class LCMSFeatureMap(object):
 *     def __init__(self, features):             # <<<<<<<<<<<<<<
 *         self.features = sorted(features, key=lambda x: x.mz)
 * 
 */

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
/* Python wrapper */

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_features = 0
#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
;

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_r;

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  {

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_features,0};

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
    PyObject* values[1] = {0};

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (unlikely(__pyx_kwds)) {

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
      Py_ssize_t kw_args;

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
      switch (pos_args) {
        case  1: 
#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
        case  0: break;
        default: 
#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
goto __pyx_L5_argtuple_error;

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
      switch (pos_args) {

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
        case  0:

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_features)) != 0)) kw_args--;
        else 
#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
goto __pyx_L5_argtuple_error;

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
      if (unlikely(kw_args > 0)) {

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 20, __pyx_L3_error)

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
      goto __pyx_L5_argtuple_error;

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
    } else {

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_features = values[0];

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); 
#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
__PYX_ERR(0, 20, __pyx_L3_error)

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L3_error:;

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return -1;

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap___init__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self), __pyx_v_features);

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_map.pyx":21
 * cdef class LCMSFeatureMap(object):
 *     def __init__(self, features):
 *         self.features = sorted(features, key=lambda x: x.mz)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
/* Python wrapper */

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_8__init___lambda(PyObject *__pyx_self, PyObject *__pyx_v_x); /*proto*/
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_8__init___lambda = 
#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
{"lambda", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_8__init___lambda, METH_O, 0};

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_8__init___lambda(PyObject *__pyx_self, PyObject *__pyx_v_x) {

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = 0;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("lambda (wrapper)", 0);
  __pyx_r = 
#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
__pyx_lambda_funcdef_lambda(__pyx_self, ((PyObject *)__pyx_v_x));

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_lambda_funcdef_lambda(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x) {

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = NULL;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("lambda", 0);

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_x, __pyx_n_s_mz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 21, __pyx_L1_error)

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = __pyx_t_1;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = 0;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L0;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L1_error:;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.__init__.lambda", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = NULL;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_map.pyx":20
 * 
 * cdef class LCMSFeatureMap(object):
 *     def __init__(self, features):             # <<<<<<<<<<<<<<
 *         self.features = sorted(features, key=lambda x: x.mz)
 * 
 */

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap___init__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, PyObject *__pyx_v_features) {

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_r;

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":21
 * cdef class LCMSFeatureMap(object):
 *     def __init__(self, features):
 *         self.features = sorted(features, key=lambda x: x.mz)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 21, __pyx_L1_error)

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_INCREF(__pyx_v_features);

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GIVEREF(__pyx_v_features);

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_features);

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 21, __pyx_L1_error)

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_8__init___lambda, 0, __pyx_n_s_init___locals_lambda, NULL, __pyx_n_s_ms_deisotope__c_feature_map_feat, __pyx_d, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 21, __pyx_L1_error)

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_key, __pyx_t_3) < 0) __PYX_ERR(0, 21, __pyx_L1_error)

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_sorted, __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 21, __pyx_L1_error)

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (!(likely(PyList_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 21, __pyx_L1_error)

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_v_self->features);

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_v_self->features);

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_v_self->features = ((PyObject*)__pyx_t_3);

#line 21 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":20
 * 
 * cdef class LCMSFeatureMap(object):
 *     def __init__(self, features):             # <<<<<<<<<<<<<<
 *         self.features = sorted(features, key=lambda x: x.mz)
 * 
 */

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = 0;

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L0;

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L1_error:;

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = -1;

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 20 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_map.pyx":23
 *         self.features = sorted(features, key=lambda x: x.mz)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return PyList_GET_SIZE(self.features)
 * 
 */

#line 23 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 23 "ms_deisotope/_c/feature_map/feature_map.pyx"
/* Python wrapper */

#line 23 "ms_deisotope/_c/feature_map/feature_map.pyx"
static Py_ssize_t __pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_3__len__(PyObject *__pyx_v_self); /*proto*/

#line 23 "ms_deisotope/_c/feature_map/feature_map.pyx"
static Py_ssize_t __pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_3__len__(PyObject *__pyx_v_self) {

#line 23 "ms_deisotope/_c/feature_map/feature_map.pyx"
  Py_ssize_t __pyx_r;

#line 23 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 23 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = 
#line 23 "ms_deisotope/_c/feature_map/feature_map.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_2__len__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self));

#line 23 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 23 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 23 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 23 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 23 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

#line 23 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 23 "ms_deisotope/_c/feature_map/feature_map.pyx"
static Py_ssize_t __pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_2__len__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self) {

#line 23 "ms_deisotope/_c/feature_map/feature_map.pyx"
  Py_ssize_t __pyx_r;

#line 23 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 23 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 23 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":24
 * 
 *     def __len__(self):
 *         return PyList_GET_SIZE(self.features)             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */

#line 24 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __pyx_v_self->features;

#line 24 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 24 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = PyList_GET_SIZE(__pyx_t_1);

#line 24 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 24 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":23
 *         self.features = sorted(features, key=lambda x: x.mz)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return PyList_GET_SIZE(self.features)
 * 
 */

#line 23 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 23 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 23 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 23 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 23 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 23 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_map.pyx":26
 *         return PyList_GET_SIZE(self.features)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.features)
 * 
 */

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
/* Python wrapper */

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_5__iter__(PyObject *__pyx_v_self); /*proto*/

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_5__iter__(PyObject *__pyx_v_self) {

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = 0;

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = 
#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_4__iter__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self));

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_4__iter__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self) {

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = NULL;

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":27
 * 
 *     def __iter__(self):
 *         return iter(self.features)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, i):
 */

#line 27 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 27 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __pyx_v_self->features;

#line 27 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 27 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 27, __pyx_L1_error)

#line 27 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 27 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 27 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = __pyx_t_2;

#line 27 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2 = 0;

#line 27 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":26
 *         return PyList_GET_SIZE(self.features)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.features)
 * 
 */

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L1_error:;

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = NULL;

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 26 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_map.pyx":29
 *         return iter(self.features)
 * 
 *     def __getitem__(self, i):             # <<<<<<<<<<<<<<
 *         if isinstance(i, (int, slice)):
 *             return self.features[i]
 */

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
/* Python wrapper */

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_7__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_i); /*proto*/

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_7__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_i) {

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = 0;

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = 
#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_6__getitem__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self), ((PyObject *)__pyx_v_i));

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_6__getitem__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_v_j = NULL
#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
;

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = NULL;

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_t_1;

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_t_2;

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_t_3;

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  Py_ssize_t __pyx_t_6;

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *(*__pyx_t_7)(PyObject *);

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_8 = NULL;

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":30
 * 
 *     def __getitem__(self, i):
 *         if isinstance(i, (int, slice)):             # <<<<<<<<<<<<<<
 *             return self.features[i]
 *         else:
 */

#line 30 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2 = PyInt_Check(__pyx_v_i); 

#line 30 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 30 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (!__pyx_t_3) {

#line 30 "ms_deisotope/_c/feature_map/feature_map.pyx"
  } else {

#line 30 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_1 = __pyx_t_3;

#line 30 "ms_deisotope/_c/feature_map/feature_map.pyx"
    goto __pyx_L4_bool_binop_done;

#line 30 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

#line 30 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_3 = PySlice_Check(__pyx_v_i); 

#line 30 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2 = (__pyx_t_3 != 0);

#line 30 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __pyx_t_2;

#line 30 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L4_bool_binop_done:;

#line 30 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 30 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":31
 *     def __getitem__(self, i):
 *         if isinstance(i, (int, slice)):
 *             return self.features[i]             # <<<<<<<<<<<<<<
 *         else:
 *             return [self.features[j] for j in i]
 */

#line 31 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 31 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (unlikely(__pyx_v_self->features == Py_None)) {

#line 31 "ms_deisotope/_c/feature_map/feature_map.pyx"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 31 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __PYX_ERR(0, 31, __pyx_L1_error)

#line 31 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 31 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_self->features, __pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 31, __pyx_L1_error)

#line 31 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 31 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_r = __pyx_t_4;

#line 31 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_4 = 0;

#line 31 "ms_deisotope/_c/feature_map/feature_map.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":30
 * 
 *     def __getitem__(self, i):
 *         if isinstance(i, (int, slice)):             # <<<<<<<<<<<<<<
 *             return self.features[i]
 *         else:
 */

#line 30 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":33
 *             return self.features[i]
 *         else:
 *             return [self.features[j] for j in i]             # <<<<<<<<<<<<<<
 * 
 *     def __eq__(self, other):
 */

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /*else*/ {

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 33, __pyx_L1_error)

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (likely(PyList_CheckExact(__pyx_v_i)) || PyTuple_CheckExact(__pyx_v_i)) {

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_5 = __pyx_v_i; __Pyx_INCREF(__pyx_t_5); __pyx_t_6 = 0;

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_7 = NULL;

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
    } else {
      __pyx_t_6 = -1; 
#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
__pyx_t_5 = PyObject_GetIter(__pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 33, __pyx_L1_error)

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_GOTREF(__pyx_t_5);

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_7 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 33, __pyx_L1_error)

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
    for (;;) {

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
      if (likely(!__pyx_t_7)) {

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (likely(PyList_CheckExact(__pyx_t_5))) {

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_5)) break;

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_t_8 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_8); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 33, __pyx_L1_error)

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
          #else

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_t_8 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 33, __pyx_L1_error)

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_GOTREF(__pyx_t_8);

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
          #endif

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
        } else {

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_5)) break;

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_8); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 33, __pyx_L1_error)

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
          #else

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_t_8 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 33, __pyx_L1_error)

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_GOTREF(__pyx_t_8);

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
          #endif

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }
      } else 
#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
{

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_8 = __pyx_t_7(__pyx_t_5);

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (unlikely(!__pyx_t_8)) {

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
          PyObject* exc_type = PyErr_Occurred();

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
          if (exc_type) {

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
            else __PYX_ERR(0, 33, __pyx_L1_error)

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
          }

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
          break;

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_GOTREF(__pyx_t_8);

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_XDECREF_SET(__pyx_v_j, __pyx_t_8);

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_8 = 0;

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
      if (unlikely(__pyx_v_self->features == Py_None)) {

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __PYX_ERR(0, 33, __pyx_L1_error)

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_8 = __Pyx_PyObject_GetItem(__pyx_v_self->features, __pyx_v_j); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 33, __pyx_L1_error)

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_GOTREF(__pyx_t_8);

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_8))) __PYX_ERR(0, 33, __pyx_L1_error)

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_r = __pyx_t_4;

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_4 = 0;

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
    goto __pyx_L0;

#line 33 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":29
 *         return iter(self.features)
 * 
 *     def __getitem__(self, i):             # <<<<<<<<<<<<<<
 *         if isinstance(i, (int, slice)):
 *             return self.features[i]
 */

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L1_error:;

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_8);

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = NULL;

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_v_j);

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 29 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_map.pyx":35
 *             return [self.features[j] for j in i]
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         if other is None:
 *             return False
 */

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
/* Python wrapper */

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_9__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_9__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = 0;

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("__eq__ (wrapper)", 0);
  __pyx_r = 
#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_8__eq__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self), ((PyObject *)__pyx_v_other));

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_8__eq__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, PyObject *__pyx_v_other) {

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = NULL;

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_t_1;

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_t_2;

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("__eq__", 0);

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":36
 * 
 *     def __eq__(self, other):
 *         if other is None:             # <<<<<<<<<<<<<<
 *             return False
 *         return self.features == other.features
 */

#line 36 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = (__pyx_v_other == Py_None);

#line 36 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 36 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":37
 *     def __eq__(self, other):
 *         if other is None:
 *             return False             # <<<<<<<<<<<<<<
 *         return self.features == other.features
 * 
 */

#line 37 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 37 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_INCREF(Py_False);

#line 37 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_r = Py_False;

#line 37 "ms_deisotope/_c/feature_map/feature_map.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":36
 * 
 *     def __eq__(self, other):
 *         if other is None:             # <<<<<<<<<<<<<<
 *             return False
 *         return self.features == other.features
 */

#line 36 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":38
 *         if other is None:
 *             return False
 *         return self.features == other.features             # <<<<<<<<<<<<<<
 * 
 *     def __ne__(self, other):
 */

#line 38 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 38 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_features); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 38, __pyx_L1_error)

#line 38 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 38 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_self->features, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 38, __pyx_L1_error)

#line 38 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 38 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = __pyx_t_4;

#line 38 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_4 = 0;

#line 38 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":35
 *             return [self.features[j] for j in i]
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         if other is None:
 *             return False
 */

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L1_error:;

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = NULL;

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 35 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_map.pyx":40
 *         return self.features == other.features
 * 
 *     def __ne__(self, other):             # <<<<<<<<<<<<<<
 *         return not self == other
 * 
 */

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
/* Python wrapper */

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_11__ne__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_11__ne__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = 0;

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("__ne__ (wrapper)", 0);
  __pyx_r = 
#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_10__ne__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self), ((PyObject *)__pyx_v_other));

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_10__ne__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, PyObject *__pyx_v_other) {

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = NULL;

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_t_2;

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("__ne__", 0);

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":41
 * 
 *     def __ne__(self, other):
 *         return not self == other             # <<<<<<<<<<<<<<
 * 
 *     cdef Py_ssize_t get_size(self):
 */

#line 41 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 41 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = PyObject_RichCompare(((PyObject *)__pyx_v_self), __pyx_v_other, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 41, __pyx_L1_error)

#line 41 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 41, __pyx_L1_error)

#line 41 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 41 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 41, __pyx_L1_error)

#line 41 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 41 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = __pyx_t_1;

#line 41 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = 0;

#line 41 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":40
 *         return self.features == other.features
 * 
 *     def __ne__(self, other):             # <<<<<<<<<<<<<<
 *         return not self == other
 * 
 */

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L1_error:;

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.__ne__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = NULL;

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 40 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_map.pyx":43
 *         return not self == other
 * 
 *     cdef Py_ssize_t get_size(self):             # <<<<<<<<<<<<<<
 *         return PyList_GET_SIZE(self.features)
 * 
 */

#line 43 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 43 "ms_deisotope/_c/feature_map/feature_map.pyx"
static Py_ssize_t __pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_get_size(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self) {

#line 43 "ms_deisotope/_c/feature_map/feature_map.pyx"
  Py_ssize_t __pyx_r;

#line 43 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 43 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 43 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("get_size", 0);

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":44
 * 
 *     cdef Py_ssize_t get_size(self):
 *         return PyList_GET_SIZE(self.features)             # <<<<<<<<<<<<<<
 * 
 *     cdef LCMSFeature get(self, size_t i):
 */

#line 44 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __pyx_v_self->features;

#line 44 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 44 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = PyList_GET_SIZE(__pyx_t_1);

#line 44 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 44 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":43
 *         return not self == other
 * 
 *     cdef Py_ssize_t get_size(self):             # <<<<<<<<<<<<<<
 *         return PyList_GET_SIZE(self.features)
 * 
 */

#line 43 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 43 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 43 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 43 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 43 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 43 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_map.pyx":46
 *         return PyList_GET_SIZE(self.features)
 * 
 *     cdef LCMSFeature get(self, size_t i):             # <<<<<<<<<<<<<<
 *         return <LCMSFeature>PyList_GetItem(self.features, i)
 * 
 */

#line 46 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 46 "ms_deisotope/_c/feature_map/feature_map.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_get(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, size_t __pyx_v_i) {

#line 46 "ms_deisotope/_c/feature_map/feature_map.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_r = NULL;

#line 46 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 46 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 46 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_2;

#line 46 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 46 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 46 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 46 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("get", 0);

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":47
 * 
 *     cdef LCMSFeature get(self, size_t i):
 *         return <LCMSFeature>PyList_GetItem(self.features, i)             # <<<<<<<<<<<<<<
 * 
 *     cpdef LCMSFeature search(self, double mz, double error_tolerance=2e-5):
 */

#line 47 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 47 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __pyx_v_self->features;

#line 47 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 47 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2 = PyList_GetItem(__pyx_t_1, __pyx_v_i); if (unlikely(__pyx_t_2 == ((PyObject *)NULL))) __PYX_ERR(0, 47, __pyx_L1_error)

#line 47 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 47 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_INCREF(((PyObject *)((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_2)));

#line 47 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_2);

#line 47 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":46
 *         return PyList_GET_SIZE(self.features)
 * 
 *     cdef LCMSFeature get(self, size_t i):             # <<<<<<<<<<<<<<
 *         return <LCMSFeature>PyList_GetItem(self.features, i)
 * 
 */

#line 46 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 46 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 46 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L1_error:;

#line 46 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 46 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.get", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 46 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = 0;

#line 46 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 46 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 46 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 46 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 46 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_map.pyx":49
 *         return <LCMSFeature>PyList_GetItem(self.features, i)
 * 
 *     cpdef LCMSFeature search(self, double mz, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *         """Search for a single feature within `error_tolerance` of `mz`.
 * 
 */

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_13search(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_search(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, double __pyx_v_mz, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_search *__pyx_optional_args) {

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  double __pyx_v_error_tolerance = ((double)2e-5);
  long __pyx_v_i
#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_out = 0
#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_r = NULL;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_t_7;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_8 = NULL;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_t_9;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_10;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("search", 0);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (__pyx_optional_args) {

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_error_tolerance = __pyx_optional_args->error_tolerance;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* Check if called by wrapper */

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* Check if overridden in Python */

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
      #endif

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_search); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 49, __pyx_L1_error)

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_13search)) {

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_mz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 49, __pyx_L1_error)

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 49, __pyx_L1_error)

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; 
#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
__pyx_t_6 = NULL;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_7 = 0;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          if (likely(__pyx_t_6)) {

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
            __Pyx_INCREF(__pyx_t_6);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
            __Pyx_INCREF(function);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
            __Pyx_DECREF_SET(__pyx_t_5, function);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
            __pyx_t_7 = 1;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          }

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        #if CYTHON_FAST_PYCALL

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (PyFunction_Check(__pyx_t_5)) {

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 49, __pyx_L1_error)

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        } else

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        #endif

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        #if CYTHON_FAST_PYCCALL

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 49, __pyx_L1_error)

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        } else

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        #endif

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        {

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 49, __pyx_L1_error)

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_GOTREF(__pyx_t_8);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          if (__pyx_t_6) {

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          }

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_GIVEREF(__pyx_t_3);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_GIVEREF(__pyx_t_4);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_t_3 = 0;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_t_4 = 0;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 49, __pyx_L1_error)

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature))))) __PYX_ERR(0, 49, __pyx_L1_error)

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_2);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_2 = 0;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        goto __pyx_L0;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
      #endif

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
    #endif

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":73
 *             long i
 *             LCMSFeature out
 *         i = binary_search(self.features, mz, error_tolerance)             # <<<<<<<<<<<<<<
 *         if i == -1:
 *             return None
 */

#line 73 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __pyx_v_self->features;

#line 73 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 73 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_v_i = __pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_binary_search(((PyObject*)__pyx_t_1), __pyx_v_mz, __pyx_v_error_tolerance);

#line 73 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":74
 *             LCMSFeature out
 *         i = binary_search(self.features, mz, error_tolerance)
 *         if i == -1:             # <<<<<<<<<<<<<<
 *             return None
 *         else:
 */

#line 74 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_9 = ((__pyx_v_i == -1L) != 0);

#line 74 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (__pyx_t_9) {

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":75
 *         i = binary_search(self.features, mz, error_tolerance)
 *         if i == -1:
 *             return None             # <<<<<<<<<<<<<<
 *         else:
 *             out = <LCMSFeature>PyList_GET_ITEM(self.features, i)
 */

#line 75 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 75 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)Py_None); __Pyx_INCREF(Py_None);

#line 75 "ms_deisotope/_c/feature_map/feature_map.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":74
 *             LCMSFeature out
 *         i = binary_search(self.features, mz, error_tolerance)
 *         if i == -1:             # <<<<<<<<<<<<<<
 *             return None
 *         else:
 */

#line 74 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":77
 *             return None
 *         else:
 *             out = <LCMSFeature>PyList_GET_ITEM(self.features, i)             # <<<<<<<<<<<<<<
 *             return out
 * 
 */

#line 77 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /*else*/ {

#line 77 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_1 = __pyx_v_self->features;

#line 77 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 77 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_10 = PyList_GET_ITEM(__pyx_t_1, __pyx_v_i);

#line 77 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 77 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_10);

#line 77 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 77 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_out = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_1);

#line 77 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":78
 *         else:
 *             out = <LCMSFeature>PyList_GET_ITEM(self.features, i)
 *             return out             # <<<<<<<<<<<<<<
 * 
 *     cpdef list find_all(self, double mz, double error_tolerance=2e-5):
 */

#line 78 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 78 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_INCREF(((PyObject *)__pyx_v_out));

#line 78 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_r = __pyx_v_out;

#line 78 "ms_deisotope/_c/feature_map/feature_map.pyx"
    goto __pyx_L0;

#line 78 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":49
 *         return <LCMSFeature>PyList_GetItem(self.features, i)
 * 
 *     cpdef LCMSFeature search(self, double mz, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *         """Search for a single feature within `error_tolerance` of `mz`.
 * 
 */

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L1_error:;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_8);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.search", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = 0;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_out);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
/* Python wrapper */

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_13search(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_12search[] = "LCMSFeatureMap.search(self, double mz, double error_tolerance=2e-5) -> LCMSFeature\nSearch for a single feature within `error_tolerance` of `mz`.\n\n        Parameters\n        ----------\n        mz : float\n            The m/z to search for\n        error_tolerance : float, optional\n            The error tolerance to search with, defaults to 2e-5, 20 PPM\n\n        Returns\n        -------\n        :class:`~.LCMSFeature`\n            The a feature matching the query m/z, or :const:`None`.\n\n        See Also\n        --------\n        find_all\n        between\n        ";

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_13search(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_mz
#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  double __pyx_v_error_tolerance
#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = 0;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("search (wrapper)", 0);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  {

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mz,&__pyx_n_s_error_tolerance,0};

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
    PyObject* values[2] = {0,0};

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (unlikely(__pyx_kwds)) {

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
      Py_ssize_t kw_args;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
      switch (pos_args) {
        case  2: 
#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        case  0: break;
        default: 
#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
goto __pyx_L5_argtuple_error;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
      switch (pos_args) {

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        case  0:

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mz)) != 0)) kw_args--;
        else 
#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
goto __pyx_L5_argtuple_error;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        case  1:

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (kw_args > 0) {

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
          if (value) { values[1] = value; kw_args--; }

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
      if (unlikely(kw_args > 0)) {

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "search") < 0)) __PYX_ERR(0, 49, __pyx_L3_error)

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
    } else {

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: 
#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
        break;
        default: 
#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
goto __pyx_L5_argtuple_error;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_mz = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_mz == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 49, __pyx_L3_error)

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (values[1]) {

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 49, __pyx_L3_error)

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
    } else {

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_error_tolerance = ((double)2e-5);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("search", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
__PYX_ERR(0, 49, __pyx_L3_error)

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L3_error:;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.search", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return NULL;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_12search(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self), __pyx_v_mz, __pyx_v_error_tolerance);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_12search(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance) {

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = NULL;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_search __pyx_t_2;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("search", 0);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap->search(__pyx_v_self, __pyx_v_mz, 1, &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 49, __pyx_L1_error)

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = __pyx_t_1;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = 0;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L0;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L1_error:;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.search", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = NULL;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 49 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_map.pyx":80
 *             return out
 * 
 *     cpdef list find_all(self, double mz, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *         """Search for all features within `error_tolerance` of `mz`.
 * 
 */

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_15find_all(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_find_all(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, double __pyx_v_mz, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_find_all *__pyx_optional_args) {

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  double __pyx_v_error_tolerance = ((double)2e-5);
  long __pyx_v_loout
#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  long __pyx_v_hiout
#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  PyObject *__pyx_v_result = 0
#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = NULL;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_t_7;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_8 = NULL;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("find_all", 0);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (__pyx_optional_args) {

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_error_tolerance = __pyx_optional_args->error_tolerance;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* Check if called by wrapper */

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* Check if overridden in Python */

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
      #endif

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_find_all); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 80, __pyx_L1_error)

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_15find_all)) {

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_mz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 80, __pyx_L1_error)

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 80, __pyx_L1_error)

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; 
#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
__pyx_t_6 = NULL;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_7 = 0;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          if (likely(__pyx_t_6)) {

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
            __Pyx_INCREF(__pyx_t_6);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
            __Pyx_INCREF(function);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
            __Pyx_DECREF_SET(__pyx_t_5, function);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
            __pyx_t_7 = 1;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          }

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        #if CYTHON_FAST_PYCALL

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (PyFunction_Check(__pyx_t_5)) {

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 80, __pyx_L1_error)

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        } else

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        #endif

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        #if CYTHON_FAST_PYCCALL

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 80, __pyx_L1_error)

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        } else

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        #endif

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        {

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 80, __pyx_L1_error)

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_GOTREF(__pyx_t_8);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          if (__pyx_t_6) {

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          }

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_GIVEREF(__pyx_t_3);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_GIVEREF(__pyx_t_4);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_t_3 = 0;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_t_4 = 0;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 80, __pyx_L1_error)

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 80, __pyx_L1_error)

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_r = ((PyObject*)__pyx_t_2);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_2 = 0;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        goto __pyx_L0;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
      #endif

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
    #endif

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":102
 *             long loout, hiout
 *             list result
 *         search_sweep(self.features, mz, error_tolerance, &loout, &hiout)             # <<<<<<<<<<<<<<
 *         result = <list>PyList_GetSlice(self.features, loout, hiout)
 *         return result
 */

#line 102 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __pyx_v_self->features;

#line 102 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 102 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_search_sweep(((PyObject*)__pyx_t_1), __pyx_v_mz, __pyx_v_error_tolerance, (&__pyx_v_loout), (&__pyx_v_hiout));

#line 102 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":103
 *             list result
 *         search_sweep(self.features, mz, error_tolerance, &loout, &hiout)
 *         result = <list>PyList_GetSlice(self.features, loout, hiout)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */

#line 103 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __pyx_v_self->features;

#line 103 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 103 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2 = PyList_GetSlice(__pyx_t_1, __pyx_v_loout, __pyx_v_hiout); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 103, __pyx_L1_error)

#line 103 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 103 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 103 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __pyx_t_2;

#line 103 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 103 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 103 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_v_result = ((PyObject*)__pyx_t_1);

#line 103 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":104
 *         search_sweep(self.features, mz, error_tolerance, &loout, &hiout)
 *         result = <list>PyList_GetSlice(self.features, loout, hiout)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def index_range(self, lo, hi, error_tolerance=2e-5):
 */

#line 104 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 104 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_INCREF(__pyx_v_result);

#line 104 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = __pyx_v_result;

#line 104 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":80
 *             return out
 * 
 *     cpdef list find_all(self, double mz, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *         """Search for all features within `error_tolerance` of `mz`.
 * 
 */

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L1_error:;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_8);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.find_all", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = 0;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_v_result);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
/* Python wrapper */

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_15find_all(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_14find_all[] = "LCMSFeatureMap.find_all(self, double mz, double error_tolerance=2e-5) -> list\nSearch for all features within `error_tolerance` of `mz`.\n\n        Parameters\n        ----------\n        mz : float\n            The m/z to search for\n        error_tolerance : float, optional\n            The error tolerance to search with, defaults to 2e-5, 20 PPM\n\n        Returns\n        -------\n        list of :class:`~.LCMSFeature`\n            A list of all features matching the query m/z, or the empty list.\n\n        See Also\n        --------\n        between\n        ";

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_15find_all(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_mz
#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  double __pyx_v_error_tolerance
#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = 0;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("find_all (wrapper)", 0);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  {

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mz,&__pyx_n_s_error_tolerance,0};

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
    PyObject* values[2] = {0,0};

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (unlikely(__pyx_kwds)) {

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
      Py_ssize_t kw_args;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
      switch (pos_args) {
        case  2: 
#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        case  0: break;
        default: 
#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
goto __pyx_L5_argtuple_error;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
      switch (pos_args) {

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        case  0:

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mz)) != 0)) kw_args--;
        else 
#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
goto __pyx_L5_argtuple_error;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        case  1:

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (kw_args > 0) {

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
          if (value) { values[1] = value; kw_args--; }

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
      if (unlikely(kw_args > 0)) {

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "find_all") < 0)) __PYX_ERR(0, 80, __pyx_L3_error)

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
    } else {

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: 
#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
        break;
        default: 
#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
goto __pyx_L5_argtuple_error;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_mz = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_mz == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 80, __pyx_L3_error)

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (values[1]) {

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 80, __pyx_L3_error)

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
    } else {

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_error_tolerance = ((double)2e-5);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("find_all", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
__PYX_ERR(0, 80, __pyx_L3_error)

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L3_error:;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.find_all", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return NULL;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_14find_all(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self), __pyx_v_mz, __pyx_v_error_tolerance);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_14find_all(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, double __pyx_v_mz, double __pyx_v_error_tolerance) {

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = NULL;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_find_all __pyx_t_2;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("find_all", 0);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap->find_all(__pyx_v_self, __pyx_v_mz, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 80, __pyx_L1_error)

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = __pyx_t_1;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = 0;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L0;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L1_error:;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.find_all", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = NULL;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 80 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_map.pyx":106
 *         return result
 * 
 *     def index_range(self, lo, hi, error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *         return (
 *             binary_search_with_flag(
 */

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
/* Python wrapper */

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_17index_range(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_16index_range[] = "LCMSFeatureMap.index_range(self, lo, hi, error_tolerance=2e-5)";

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_17index_range(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_lo = 0
#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  PyObject *__pyx_v_hi = 0
#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  PyObject *__pyx_v_error_tolerance = 0
#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = 0;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("index_range (wrapper)", 0);

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  {

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lo,&__pyx_n_s_hi,&__pyx_n_s_error_tolerance,0};

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
    PyObject* values[3] = {0,0,0};

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
    values[2] = ((PyObject *)__pyx_float_2eneg_5);

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (unlikely(__pyx_kwds)) {

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
      Py_ssize_t kw_args;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
      switch (pos_args) {
        case  3: 
#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
        case  0: break;
        default: 
#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
goto __pyx_L5_argtuple_error;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
      switch (pos_args) {

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
        case  0:

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lo)) != 0)) kw_args--;
        else 
#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
goto __pyx_L5_argtuple_error;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
        case  1:

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hi)) != 0)) kw_args--;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("index_range", 0, 2, 3, 1); 
#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
__PYX_ERR(0, 106, __pyx_L3_error)

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
        case  2:

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (kw_args > 0) {

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance);

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
      if (unlikely(kw_args > 0)) {

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "index_range") < 0)) __PYX_ERR(0, 106, __pyx_L3_error)

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
    } else {

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
        break;
        default: 
#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
goto __pyx_L5_argtuple_error;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_lo = values[0];

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_hi = values[1];

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_error_tolerance = values[2];

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("index_range", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
__PYX_ERR(0, 106, __pyx_L3_error)

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L3_error:;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.index_range", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return NULL;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_16index_range(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self), __pyx_v_lo, __pyx_v_hi, __pyx_v_error_tolerance);

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_16index_range(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, PyObject *__pyx_v_lo, PyObject *__pyx_v_hi, PyObject *__pyx_v_error_tolerance) {

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = NULL;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  double __pyx_t_2;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  double __pyx_t_3;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("index_range", 0);

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":107
 * 
 *     def index_range(self, lo, hi, error_tolerance=2e-5):
 *         return (             # <<<<<<<<<<<<<<
 *             binary_search_with_flag(
 *                 self.features, lo, error_tolerance)[0][0],
 */

#line 107 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_r);

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":109
 *         return (
 *             binary_search_with_flag(
 *                 self.features, lo, error_tolerance)[0][0],             # <<<<<<<<<<<<<<
 *             binary_search_with_flag(
 *                 self.features, hi, error_tolerance)[0][0])
 */

#line 109 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __pyx_v_self->features;

#line 109 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 109 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_v_lo); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 109, __pyx_L1_error)

#line 109 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_v_error_tolerance); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 109, __pyx_L1_error)

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":108
 *     def index_range(self, lo, hi, error_tolerance=2e-5):
 *         return (
 *             binary_search_with_flag(             # <<<<<<<<<<<<<<
 *                 self.features, lo, error_tolerance)[0][0],
 *             binary_search_with_flag(
 */

#line 108 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_4 = __pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_binary_search_with_flag(((PyObject*)__pyx_t_1), __pyx_t_2, __pyx_t_3, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 108, __pyx_L1_error)

#line 108 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 108 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 108 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (unlikely(__pyx_t_4 == Py_None)) {

#line 108 "ms_deisotope/_c/feature_map/feature_map.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 108 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __PYX_ERR(0, 108, __pyx_L1_error)

#line 108 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":109
 *         return (
 *             binary_search_with_flag(
 *                 self.features, lo, error_tolerance)[0][0],             # <<<<<<<<<<<<<<
 *             binary_search_with_flag(
 *                 self.features, hi, error_tolerance)[0][0])
 */

#line 109 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_t_4, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 109, __pyx_L1_error)

#line 109 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 109 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 109 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 109, __pyx_L1_error)

#line 109 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 109 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":111
 *                 self.features, lo, error_tolerance)[0][0],
 *             binary_search_with_flag(
 *                 self.features, hi, error_tolerance)[0][0])             # <<<<<<<<<<<<<<
 * 
 *     cpdef list spanning_time(self, double time_point):
 */

#line 111 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __pyx_v_self->features;

#line 111 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 111 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_v_hi); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 111, __pyx_L1_error)

#line 111 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_v_error_tolerance); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 111, __pyx_L1_error)

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":110
 *             binary_search_with_flag(
 *                 self.features, lo, error_tolerance)[0][0],
 *             binary_search_with_flag(             # <<<<<<<<<<<<<<
 *                 self.features, hi, error_tolerance)[0][0])
 * 
 */

#line 110 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_5 = __pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_binary_search_with_flag(((PyObject*)__pyx_t_1), __pyx_t_3, __pyx_t_2, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 110, __pyx_L1_error)

#line 110 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 110 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 110 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (unlikely(__pyx_t_5 == Py_None)) {

#line 110 "ms_deisotope/_c/feature_map/feature_map.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 110 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __PYX_ERR(0, 110, __pyx_L1_error)

#line 110 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":111
 *                 self.features, lo, error_tolerance)[0][0],
 *             binary_search_with_flag(
 *                 self.features, hi, error_tolerance)[0][0])             # <<<<<<<<<<<<<<
 * 
 *     cpdef list spanning_time(self, double time_point):
 */

#line 111 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 111, __pyx_L1_error)

#line 111 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 111 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 111 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 111, __pyx_L1_error)

#line 111 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 111 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":108
 *     def index_range(self, lo, hi, error_tolerance=2e-5):
 *         return (
 *             binary_search_with_flag(             # <<<<<<<<<<<<<<
 *                 self.features, lo, error_tolerance)[0][0],
 *             binary_search_with_flag(
 */

#line 108 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 108, __pyx_L1_error)

#line 108 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 108 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GIVEREF(__pyx_t_4);

#line 108 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);

#line 108 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GIVEREF(__pyx_t_5);

#line 108 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_5);

#line 108 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_4 = 0;

#line 108 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_5 = 0;

#line 108 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = __pyx_t_1;

#line 108 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = 0;

#line 108 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":106
 *         return result
 * 
 *     def index_range(self, lo, hi, error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *         return (
 *             binary_search_with_flag(
 */

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L1_error:;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.index_range", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = NULL;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 106 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_map.pyx":113
 *                 self.features, hi, error_tolerance)[0][0])
 * 
 *     cpdef list spanning_time(self, double time_point):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_19spanning_time(PyObject *__pyx_v_self, PyObject *__pyx_arg_time_point); /*proto*/

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_spanning_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, double __pyx_v_time_point, int __pyx_skip_dispatch) {
  size_t __pyx_v_i
#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  size_t __pyx_v_n
#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_feature = 0
#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  PyObject *__pyx_v_result = 0
#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = NULL;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  size_t __pyx_t_6;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  size_t __pyx_t_7;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  size_t __pyx_t_8;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_t_9;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_t_10;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("spanning_time", 0);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* Check if called by wrapper */

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* Check if overridden in Python */

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
      #endif

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_spanning_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 113, __pyx_L1_error)

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_19spanning_time)) {

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_time_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 113, __pyx_L1_error)

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
__pyx_t_5 = NULL;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
          if (likely(__pyx_t_5)) {

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
            __Pyx_INCREF(__pyx_t_5);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
            __Pyx_INCREF(function);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
            __Pyx_DECREF_SET(__pyx_t_4, function);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
          }

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 113, __pyx_L1_error)

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 113, __pyx_L1_error)

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_r = ((PyObject*)__pyx_t_2);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_2 = 0;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
        goto __pyx_L0;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
      #endif

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
    #endif

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":118
 *             LCMSFeature feature
 *             list result
 *         result = []             # <<<<<<<<<<<<<<
 *         n = self.get_size()
 *         for i in range(n):
 */

#line 118 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L1_error)

#line 118 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 118 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_v_result = ((PyObject*)__pyx_t_1);

#line 118 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":119
 *             list result
 *         result = []
 *         n = self.get_size()             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             feature = self.get(i)
 */

#line 119 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_v_n = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self->__pyx_vtab)->get_size(__pyx_v_self);

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":120
 *         result = []
 *         n = self.get_size()
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             feature = self.get(i)
 *             if feature.spans_in_time(time_point):
 */

#line 120 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_6 = __pyx_v_n;

#line 120 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_7 = __pyx_t_6;

#line 120 "ms_deisotope/_c/feature_map/feature_map.pyx"
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {

#line 120 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_i = __pyx_t_8;

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":121
 *         n = self.get_size()
 *         for i in range(n):
 *             feature = self.get(i)             # <<<<<<<<<<<<<<
 *             if feature.spans_in_time(time_point):
 *                 result.append(feature)
 */

#line 121 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self->__pyx_vtab)->get(__pyx_v_self, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 121, __pyx_L1_error)

#line 121 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 121 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_XDECREF_SET(__pyx_v_feature, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_1));

#line 121 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":122
 *         for i in range(n):
 *             feature = self.get(i)
 *             if feature.spans_in_time(time_point):             # <<<<<<<<<<<<<<
 *                 result.append(feature)
 *         return result
 */

#line 122 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_9 = (((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_feature->__pyx_base.__pyx_vtab)->spans_in_time(__pyx_v_feature, __pyx_v_time_point, 0) != 0);

#line 122 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (__pyx_t_9) {

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":123
 *             feature = self.get(i)
 *             if feature.spans_in_time(time_point):
 *                 result.append(feature)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */

#line 123 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_result, ((PyObject *)__pyx_v_feature)); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 123, __pyx_L1_error)

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":122
 *         for i in range(n):
 *             feature = self.get(i)
 *             if feature.spans_in_time(time_point):             # <<<<<<<<<<<<<<
 *                 result.append(feature)
 *         return result
 */

#line 122 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 122 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":124
 *             if feature.spans_in_time(time_point):
 *                 result.append(feature)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     cpdef list between(self, double lo, double hi, double error_tolerance=2e-5):
 */

#line 124 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 124 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_INCREF(__pyx_v_result);

#line 124 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = __pyx_v_result;

#line 124 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":113
 *                 self.features, hi, error_tolerance)[0][0])
 * 
 *     cpdef list spanning_time(self, double time_point):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L1_error:;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.spanning_time", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = 0;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_feature);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_v_result);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
/* Python wrapper */

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_19spanning_time(PyObject *__pyx_v_self, PyObject *__pyx_arg_time_point); /*proto*/

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_18spanning_time[] = "LCMSFeatureMap.spanning_time(self, double time_point) -> list";

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_19spanning_time(PyObject *__pyx_v_self, PyObject *__pyx_arg_time_point) {
  double __pyx_v_time_point
#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = 0;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("spanning_time (wrapper)", 0);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  assert(__pyx_arg_time_point); {

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_time_point = __pyx_PyFloat_AsDouble(__pyx_arg_time_point); if (unlikely((__pyx_v_time_point == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 113, __pyx_L3_error)

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L3_error:;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.spanning_time", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return NULL;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_18spanning_time(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self), ((double)__pyx_v_time_point));

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_18spanning_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, double __pyx_v_time_point) {

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = NULL;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("spanning_time", 0);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_spanning_time(__pyx_v_self, __pyx_v_time_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 113, __pyx_L1_error)

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = __pyx_t_1;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = 0;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L0;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L1_error:;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.spanning_time", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = NULL;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 113 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_map.pyx":126
 *         return result
 * 
 *     cpdef list between(self, double lo, double hi, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *         """Search for all features between `lo` and `hi`, allowing `error_tolerance`
 *         around the edges.
 */

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_21between(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_between(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, double __pyx_v_lo, double __pyx_v_hi, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_between *__pyx_optional_args) {

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  double __pyx_v_error_tolerance = ((double)2e-5);
  Py_ssize_t __pyx_v_n
#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  Py_ssize_t __pyx_v_i
#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  Py_ssize_t __pyx_v_lo_ix
#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  Py_ssize_t __pyx_v_hi_ix
#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_f = 0
#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = NULL;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_7 = NULL;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_t_8;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_9 = NULL;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_t_10;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  Py_ssize_t __pyx_t_11;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("between", 0);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (__pyx_optional_args) {

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_error_tolerance = __pyx_optional_args->error_tolerance;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* Check if called by wrapper */

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* Check if overridden in Python */

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
      #endif

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_between); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 126, __pyx_L1_error)

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_21between)) {

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_lo); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 126, __pyx_L1_error)

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_hi); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 126, __pyx_L1_error)

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_5 = PyFloat_FromDouble(__pyx_v_error_tolerance); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 126, __pyx_L1_error)

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_GOTREF(__pyx_t_5);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_6 = __pyx_t_1; 
#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
__pyx_t_7 = NULL;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_8 = 0;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          if (likely(__pyx_t_7)) {

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
            __Pyx_INCREF(__pyx_t_7);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
            __Pyx_INCREF(function);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
            __Pyx_DECREF_SET(__pyx_t_6, function);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
            __pyx_t_8 = 1;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          }

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        #if CYTHON_FAST_PYCALL

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (PyFunction_Check(__pyx_t_6)) {

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_t_3, __pyx_t_4, __pyx_t_5};

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_8, 3+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 126, __pyx_L1_error)

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        } else

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        #endif

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        #if CYTHON_FAST_PYCCALL

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_t_3, __pyx_t_4, __pyx_t_5};

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_8, 3+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 126, __pyx_L1_error)

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        } else

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        #endif

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        {

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_t_9 = PyTuple_New(3+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 126, __pyx_L1_error)

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_GOTREF(__pyx_t_9);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          if (__pyx_t_7) {

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          }

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_GIVEREF(__pyx_t_3);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_t_3);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_GIVEREF(__pyx_t_4);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_t_4);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_GIVEREF(__pyx_t_5);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_8, __pyx_t_5);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_t_3 = 0;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_t_4 = 0;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_t_5 = 0;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_9, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 126, __pyx_L1_error)

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 126, __pyx_L1_error)

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_r = ((PyObject*)__pyx_t_2);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_2 = 0;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        goto __pyx_L0;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
      #endif

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
    #endif

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":152
 *             LCMSFeature f
 * 
 *         n = self.get_size()             # <<<<<<<<<<<<<<
 *         if n == 0:
 *             return []
 */

#line 152 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_v_n = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self->__pyx_vtab)->get_size(__pyx_v_self);

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":153
 * 
 *         n = self.get_size()
 *         if n == 0:             # <<<<<<<<<<<<<<
 *             return []
 *         lo_ix = binary_search_with_flag(
 */

#line 153 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_10 = ((__pyx_v_n == 0) != 0);

#line 153 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (__pyx_t_10) {

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":154
 *         n = self.get_size()
 *         if n == 0:
 *             return []             # <<<<<<<<<<<<<<
 *         lo_ix = binary_search_with_flag(
 *             self.features, lo, error_tolerance)[0][0]
 */

#line 154 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 154 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 154, __pyx_L1_error)

#line 154 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 154 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_r = ((PyObject*)__pyx_t_1);

#line 154 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_1 = 0;

#line 154 "ms_deisotope/_c/feature_map/feature_map.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":153
 * 
 *         n = self.get_size()
 *         if n == 0:             # <<<<<<<<<<<<<<
 *             return []
 *         lo_ix = binary_search_with_flag(
 */

#line 153 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":156
 *             return []
 *         lo_ix = binary_search_with_flag(
 *             self.features, lo, error_tolerance)[0][0]             # <<<<<<<<<<<<<<
 *         if self.get(lo_ix).get_mz() < lo:
 *             lo_ix += 1
 */

#line 156 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __pyx_v_self->features;

#line 156 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_INCREF(__pyx_t_1);

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":155
 *         if n == 0:
 *             return []
 *         lo_ix = binary_search_with_flag(             # <<<<<<<<<<<<<<
 *             self.features, lo, error_tolerance)[0][0]
 *         if self.get(lo_ix).get_mz() < lo:
 */

#line 155 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2 = __pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_binary_search_with_flag(((PyObject*)__pyx_t_1), __pyx_v_lo, __pyx_v_error_tolerance, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 155, __pyx_L1_error)

#line 155 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 155 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 155 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (unlikely(__pyx_t_2 == Py_None)) {

#line 155 "ms_deisotope/_c/feature_map/feature_map.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 155 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __PYX_ERR(0, 155, __pyx_L1_error)

#line 155 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":156
 *             return []
 *         lo_ix = binary_search_with_flag(
 *             self.features, lo, error_tolerance)[0][0]             # <<<<<<<<<<<<<<
 *         if self.get(lo_ix).get_mz() < lo:
 *             lo_ix += 1
 */

#line 156 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_t_2, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 156, __pyx_L1_error)

#line 156 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 156 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 156 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 156, __pyx_L1_error)

#line 156 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 156 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 156 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_11 = __Pyx_PyIndex_AsSsize_t(__pyx_t_2); if (unlikely((__pyx_t_11 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 156, __pyx_L1_error)

#line 156 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 156 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_v_lo_ix = __pyx_t_11;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":157
 *         lo_ix = binary_search_with_flag(
 *             self.features, lo, error_tolerance)[0][0]
 *         if self.get(lo_ix).get_mz() < lo:             # <<<<<<<<<<<<<<
 *             lo_ix += 1
 *         if lo_ix > n:
 */

#line 157 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self->__pyx_vtab)->get(__pyx_v_self, __pyx_v_lo_ix)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 157, __pyx_L1_error)

#line 157 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 157 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_10 = ((((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_2)->__pyx_base.__pyx_vtab)->__pyx_base.get_mz(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_t_2)) < __pyx_v_lo) != 0);

#line 157 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 157 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (__pyx_t_10) {

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":158
 *             self.features, lo, error_tolerance)[0][0]
 *         if self.get(lo_ix).get_mz() < lo:
 *             lo_ix += 1             # <<<<<<<<<<<<<<
 *         if lo_ix > n:
 *             lo_ix = n - 1
 */

#line 158 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_lo_ix = (__pyx_v_lo_ix + 1);

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":157
 *         lo_ix = binary_search_with_flag(
 *             self.features, lo, error_tolerance)[0][0]
 *         if self.get(lo_ix).get_mz() < lo:             # <<<<<<<<<<<<<<
 *             lo_ix += 1
 *         if lo_ix > n:
 */

#line 157 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":159
 *         if self.get(lo_ix).get_mz() < lo:
 *             lo_ix += 1
 *         if lo_ix > n:             # <<<<<<<<<<<<<<
 *             lo_ix = n - 1
 *         i = lo_ix
 */

#line 159 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_10 = ((__pyx_v_lo_ix > __pyx_v_n) != 0);

#line 159 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (__pyx_t_10) {

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":160
 *             lo_ix += 1
 *         if lo_ix > n:
 *             lo_ix = n - 1             # <<<<<<<<<<<<<<
 *         i = lo_ix
 *         while i < n:
 */

#line 160 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_lo_ix = (__pyx_v_n - 1);

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":159
 *         if self.get(lo_ix).get_mz() < lo:
 *             lo_ix += 1
 *         if lo_ix > n:             # <<<<<<<<<<<<<<
 *             lo_ix = n - 1
 *         i = lo_ix
 */

#line 159 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":161
 *         if lo_ix > n:
 *             lo_ix = n - 1
 *         i = lo_ix             # <<<<<<<<<<<<<<
 *         while i < n:
 *             f = self.get(i)
 */

#line 161 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_v_i = __pyx_v_lo_ix;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":162
 *             lo_ix = n - 1
 *         i = lo_ix
 *         while i < n:             # <<<<<<<<<<<<<<
 *             f = self.get(i)
 *             if f.get_mz() > hi:
 */

#line 162 "ms_deisotope/_c/feature_map/feature_map.pyx"
  while (1) {

#line 162 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_10 = ((__pyx_v_i < __pyx_v_n) != 0);

#line 162 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (!__pyx_t_10) break;

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":163
 *         i = lo_ix
 *         while i < n:
 *             f = self.get(i)             # <<<<<<<<<<<<<<
 *             if f.get_mz() > hi:
 *                 break
 */

#line 163 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self->__pyx_vtab)->get(__pyx_v_self, __pyx_v_i)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 163, __pyx_L1_error)

#line 163 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 163 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_XDECREF_SET(__pyx_v_f, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_2));

#line 163 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_2 = 0;

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":164
 *         while i < n:
 *             f = self.get(i)
 *             if f.get_mz() > hi:             # <<<<<<<<<<<<<<
 *                 break
 *             i += 1
 */

#line 164 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_10 = ((((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_f->__pyx_base.__pyx_vtab)->__pyx_base.get_mz(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_f)) > __pyx_v_hi) != 0);

#line 164 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (__pyx_t_10) {

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":165
 *             f = self.get(i)
 *             if f.get_mz() > hi:
 *                 break             # <<<<<<<<<<<<<<
 *             i += 1
 *         hi_ix = i
 */

#line 165 "ms_deisotope/_c/feature_map/feature_map.pyx"
      goto __pyx_L7_break;

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":164
 *         while i < n:
 *             f = self.get(i)
 *             if f.get_mz() > hi:             # <<<<<<<<<<<<<<
 *                 break
 *             i += 1
 */

#line 164 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":166
 *             if f.get_mz() > hi:
 *                 break
 *             i += 1             # <<<<<<<<<<<<<<
 *         hi_ix = i
 *         return self[lo_ix:hi_ix]
 */

#line 166 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_i = (__pyx_v_i + 1);

#line 166 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

#line 166 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L7_break:;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":167
 *                 break
 *             i += 1
 *         hi_ix = i             # <<<<<<<<<<<<<<
 *         return self[lo_ix:hi_ix]
 * 
 */

#line 167 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_v_hi_ix = __pyx_v_i;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":168
 *             i += 1
 *         hi_ix = i
 *         return self[lo_ix:hi_ix]             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */

#line 168 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 168 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetSlice(((PyObject *)__pyx_v_self), __pyx_v_lo_ix, __pyx_v_hi_ix, NULL, NULL, NULL, 1, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 168, __pyx_L1_error)

#line 168 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 168 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 168, __pyx_L1_error)

#line 168 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = ((PyObject*)__pyx_t_2);

#line 168 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2 = 0;

#line 168 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":126
 *         return result
 * 
 *     cpdef list between(self, double lo, double hi, double error_tolerance=2e-5):             # <<<<<<<<<<<<<<
 *         """Search for all features between `lo` and `hi`, allowing `error_tolerance`
 *         around the edges.
 */

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L1_error:;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_7);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_9);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.between", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = 0;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_f);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
/* Python wrapper */

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_21between(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_20between[] = "LCMSFeatureMap.between(self, double lo, double hi, double error_tolerance=2e-5) -> list\nSearch for all features between `lo` and `hi`, allowing `error_tolerance`\n        around the edges.\n\n        Parameters\n        ----------\n        lo : float\n            The m/z to search for the lower bound\n        hi : float\n            The m/z to search for the upper bound\n        error_tolerance : float, optional\n            The error tolerance to search with, defaults to 2e-5, 20 PPM\n\n        Returns\n        -------\n        list of :class:`~.LCMSFeature`\n            The features between the query bounds with the permitted error, or an empty list.\n\n        See Also\n        --------\n        find_all\n        ";

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_21between(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_lo
#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  double __pyx_v_hi
#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  double __pyx_v_error_tolerance
#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = 0;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("between (wrapper)", 0);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  {

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lo,&__pyx_n_s_hi,&__pyx_n_s_error_tolerance,0};

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
    PyObject* values[3] = {0,0,0};

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (unlikely(__pyx_kwds)) {

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
      Py_ssize_t kw_args;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
      switch (pos_args) {
        case  3: 
#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        case  0: break;
        default: 
#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
goto __pyx_L5_argtuple_error;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
      switch (pos_args) {

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        case  0:

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lo)) != 0)) kw_args--;
        else 
#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
goto __pyx_L5_argtuple_error;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        case  1:

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hi)) != 0)) kw_args--;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("between", 0, 2, 3, 1); 
#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
__PYX_ERR(0, 126, __pyx_L3_error)

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        case  2:

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (kw_args > 0) {

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
      if (unlikely(kw_args > 0)) {

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "between") < 0)) __PYX_ERR(0, 126, __pyx_L3_error)

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
    } else {

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
        break;
        default: 
#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
goto __pyx_L5_argtuple_error;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_lo = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_lo == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 126, __pyx_L3_error)

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_hi = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_hi == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 126, __pyx_L3_error)

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (values[2]) {

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 126, __pyx_L3_error)

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
    } else {

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_error_tolerance = ((double)2e-5);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("between", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
__PYX_ERR(0, 126, __pyx_L3_error)

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L3_error:;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.between", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return NULL;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_20between(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self), __pyx_v_lo, __pyx_v_hi, __pyx_v_error_tolerance);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_20between(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, double __pyx_v_lo, double __pyx_v_hi, double __pyx_v_error_tolerance) {

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = NULL;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_between __pyx_t_2;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("between", 0);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2.error_tolerance = __pyx_v_error_tolerance;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap->between(__pyx_v_self, __pyx_v_lo, __pyx_v_hi, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 126, __pyx_L1_error)

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = __pyx_t_1;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = 0;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L0;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L1_error:;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.between", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = NULL;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 126 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_map.pyx":170
 *         return self[lo_ix:hi_ix]
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "{self.__class__.__name__}(<{size} features>)".format(self=self, size=len(self))
 * 
 */

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
/* Python wrapper */

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_23__repr__(PyObject *__pyx_v_self); /*proto*/

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_23__repr__(PyObject *__pyx_v_self) {

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = 0;

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = 
#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_22__repr__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self));

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_22__repr__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self) {

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = NULL;

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
  Py_ssize_t __pyx_t_3;

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":171
 * 
 *     def __repr__(self):
 *         return "{self.__class__.__name__}(<{size} features>)".format(self=self, size=len(self))             # <<<<<<<<<<<<<<
 * 
 *     cpdef LCMSFeatureMap clone(self, bint deep=True):
 */

#line 171 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 171 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_self___class_____name___size_fe, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 171, __pyx_L1_error)

#line 171 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 171 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 171, __pyx_L1_error)

#line 171 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 171 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_self, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 171, __pyx_L1_error)

#line 171 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_3 = PyObject_Length(((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 171, __pyx_L1_error)

#line 171 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_4 = PyInt_FromSsize_t(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 171, __pyx_L1_error)

#line 171 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 171 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_size, __pyx_t_4) < 0) __PYX_ERR(0, 171, __pyx_L1_error)

#line 171 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 171 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 171, __pyx_L1_error)

#line 171 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 171 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 171 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 171 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = __pyx_t_4;

#line 171 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_4 = 0;

#line 171 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":170
 *         return self[lo_ix:hi_ix]
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "{self.__class__.__name__}(<{size} features>)".format(self=self, size=len(self))
 * 
 */

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L1_error:;

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = NULL;

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 170 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_map.pyx":173
 *         return "{self.__class__.__name__}(<{size} features>)".format(self=self, size=len(self))
 * 
 *     cpdef LCMSFeatureMap clone(self, bint deep=True):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_25clone(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_clone(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_clone *__pyx_optional_args) {

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_v_deep = ((int)1);
  size_t __pyx_v_i
#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  size_t __pyx_v_n
#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  PyObject *__pyx_v_result = 0
#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_feature = NULL
#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_r = NULL;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  size_t __pyx_t_6;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  size_t __pyx_t_7;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  size_t __pyx_t_8;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_clone __pyx_t_9;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_t_10;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("clone", 0);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (__pyx_optional_args) {

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_deep = __pyx_optional_args->deep;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* Check if called by wrapper */

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* Check if overridden in Python */

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
      #endif

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_clone); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 173, __pyx_L1_error)

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_25clone)) {

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_deep); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 173, __pyx_L1_error)

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
__pyx_t_5 = NULL;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
          if (likely(__pyx_t_5)) {

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
            __Pyx_INCREF(__pyx_t_5);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
            __Pyx_INCREF(function);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
            __Pyx_DECREF_SET(__pyx_t_4, function);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
          }

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 173, __pyx_L1_error)

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap))))) __PYX_ERR(0, 173, __pyx_L1_error)

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_t_2);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_2 = 0;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
        goto __pyx_L0;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
      #endif

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
    #endif

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":177
 *             size_t i, n
 *             list result
 *         result = []             # <<<<<<<<<<<<<<
 *         n = self.get_size()
 *         for i in range(n):
 */

#line 177 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 177, __pyx_L1_error)

#line 177 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 177 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_v_result = ((PyObject*)__pyx_t_1);

#line 177 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":178
 *             list result
 *         result = []
 *         n = self.get_size()             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             feature = self.get(i).clone(deep)
 */

#line 178 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_v_n = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self->__pyx_vtab)->get_size(__pyx_v_self);

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":179
 *         result = []
 *         n = self.get_size()
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             feature = self.get(i).clone(deep)
 *             result.append(feature)
 */

#line 179 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_6 = __pyx_v_n;

#line 179 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_7 = __pyx_t_6;

#line 179 "ms_deisotope/_c/feature_map/feature_map.pyx"
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {

#line 179 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_i = __pyx_t_8;

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":180
 *         n = self.get_size()
 *         for i in range(n):
 *             feature = self.get(i).clone(deep)             # <<<<<<<<<<<<<<
 *             result.append(feature)
 *         return self.__class__(result)
 */

#line 180 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self->__pyx_vtab)->get(__pyx_v_self, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 180, __pyx_L1_error)

#line 180 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 180 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_deep); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 180, __pyx_L1_error)

#line 180 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 180 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_9.__pyx_n = 1;

#line 180 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_9.deep = __pyx_t_2;

#line 180 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_1)->__pyx_base.__pyx_vtab)->clone(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_1), 0, &__pyx_t_9)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 180, __pyx_L1_error)

#line 180 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 180 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 180 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 180 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_XDECREF_SET(__pyx_v_feature, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_4));

#line 180 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":181
 *         for i in range(n):
 *             feature = self.get(i).clone(deep)
 *             result.append(feature)             # <<<<<<<<<<<<<<
 *         return self.__class__(result)
 * 
 */

#line 181 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_result, ((PyObject *)__pyx_v_feature)); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 181, __pyx_L1_error)

#line 181 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":182
 *             feature = self.get(i).clone(deep)
 *             result.append(feature)
 *         return self.__class__(result)             # <<<<<<<<<<<<<<
 * 
 *     def as_arrays(self):
 */

#line 182 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 182 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 182, __pyx_L1_error)

#line 182 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 182 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = NULL;

#line 182 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 182 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);

#line 182 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (likely(__pyx_t_1)) {

#line 182 "ms_deisotope/_c/feature_map/feature_map.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 182 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_INCREF(__pyx_t_1);

#line 182 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_INCREF(function);

#line 182 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_DECREF_SET(__pyx_t_2, function);

#line 182 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 182 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

#line 182 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, __pyx_v_result) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_result);

#line 182 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 182 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 182, __pyx_L1_error)

#line 182 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 182 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 182 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap))))) __PYX_ERR(0, 182, __pyx_L1_error)

#line 182 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_t_4);

#line 182 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_4 = 0;

#line 182 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":173
 *         return "{self.__class__.__name__}(<{size} features>)".format(self=self, size=len(self))
 * 
 *     cpdef LCMSFeatureMap clone(self, bint deep=True):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L1_error:;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = 0;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_v_result);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_feature);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
/* Python wrapper */

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_25clone(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_24clone[] = "LCMSFeatureMap.clone(self, bool deep=True) -> LCMSFeatureMap";

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_25clone(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_deep
#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = 0;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("clone (wrapper)", 0);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  {

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_deep,0};

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
    PyObject* values[1] = {0};

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (unlikely(__pyx_kwds)) {

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
      Py_ssize_t kw_args;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
      switch (pos_args) {
        case  1: 
#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
        case  0: break;
        default: 
#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
goto __pyx_L5_argtuple_error;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
      switch (pos_args) {

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
        case  0:

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (kw_args > 0) {

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_deep);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
          if (value) { values[0] = value; kw_args--; }

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
      if (unlikely(kw_args > 0)) {

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "clone") < 0)) __PYX_ERR(0, 173, __pyx_L3_error)

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
    } else {

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: 
#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;
        case  0: 
#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
break;
        default: 
#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
goto __pyx_L5_argtuple_error;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (values[0]) {

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_deep = __Pyx_PyObject_IsTrue(values[0]); if (unlikely((__pyx_v_deep == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 173, __pyx_L3_error)

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
    } else {

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_deep = ((int)1);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("clone", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); 
#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
__PYX_ERR(0, 173, __pyx_L3_error)

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L3_error:;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return NULL;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_24clone(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self), __pyx_v_deep);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_24clone(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, int __pyx_v_deep) {

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = NULL;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_clone __pyx_t_2;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("clone", 0);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2.deep = __pyx_v_deep;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap->clone(__pyx_v_self, 1, &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 173, __pyx_L1_error)

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = __pyx_t_1;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = 0;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L0;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L1_error:;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = NULL;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 173 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_map.pyx":184
 *         return self.__class__(result)
 * 
 *     def as_arrays(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n, j, m, k, q
 */

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
/* Python wrapper */

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_27as_arrays(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_26as_arrays[] = "LCMSFeatureMap.as_arrays(self)";

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_27as_arrays(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = 0;

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("as_arrays (wrapper)", 0);
  __pyx_r = 
#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_26as_arrays(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self));

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_26as_arrays(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self) {
  size_t __pyx_v_i
#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  size_t __pyx_v_n
#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  size_t __pyx_v_j
#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  size_t __pyx_v_m
#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  size_t __pyx_v_k
#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  size_t __pyx_v_q
#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  double __pyx_v_time
#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_feature = 0
#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_node = 0
#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak = 0
#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  PyObject *__pyx_v_mz_array = NULL
#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  PyObject *__pyx_v_intensity_array = NULL
#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  PyObject *__pyx_v_ion_mobility_array = NULL
#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  PyObject *__pyx_v_feature_id_array = NULL
#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  PyObject *__pyx_v_mask = NULL
#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
;

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = NULL;

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  size_t __pyx_t_2;

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  size_t __pyx_t_3;

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  size_t __pyx_t_4;

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  size_t __pyx_t_6;

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  size_t __pyx_t_7;

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  size_t __pyx_t_8;

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  double __pyx_t_9;

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  size_t __pyx_t_10;

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  size_t __pyx_t_11;

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  size_t __pyx_t_12;

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_t_13;

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_14 = NULL;

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_15 = NULL;

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_16 = NULL;

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_17 = NULL;

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("as_arrays", 0);

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":192
 *             PeakBase peak
 * 
 *         mz_array = array.array('d')             # <<<<<<<<<<<<<<
 *         intensity_array = array.array('d')
 *         ion_mobility_array = array.array('d')
 */

#line 192 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7cpython_5array_array), __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 192, __pyx_L1_error)

#line 192 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 192 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_v_mz_array = __pyx_t_1;

#line 192 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":193
 * 
 *         mz_array = array.array('d')
 *         intensity_array = array.array('d')             # <<<<<<<<<<<<<<
 *         ion_mobility_array = array.array('d')
 *         feature_id_array = array.array('L')
 */

#line 193 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7cpython_5array_array), __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 193, __pyx_L1_error)

#line 193 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 193 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_v_intensity_array = __pyx_t_1;

#line 193 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":194
 *         mz_array = array.array('d')
 *         intensity_array = array.array('d')
 *         ion_mobility_array = array.array('d')             # <<<<<<<<<<<<<<
 *         feature_id_array = array.array('L')
 *         n = self.get_size()
 */

#line 194 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7cpython_5array_array), __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 194, __pyx_L1_error)

#line 194 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 194 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_v_ion_mobility_array = __pyx_t_1;

#line 194 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":195
 *         intensity_array = array.array('d')
 *         ion_mobility_array = array.array('d')
 *         feature_id_array = array.array('L')             # <<<<<<<<<<<<<<
 *         n = self.get_size()
 *         for i in range(n):
 */

#line 195 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7cpython_5array_array), __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 195, __pyx_L1_error)

#line 195 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 195 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_v_feature_id_array = __pyx_t_1;

#line 195 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":196
 *         ion_mobility_array = array.array('d')
 *         feature_id_array = array.array('L')
 *         n = self.get_size()             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             feature = <LCMSFeature>self.get(i)
 */

#line 196 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_v_n = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self->__pyx_vtab)->get_size(__pyx_v_self);

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":197
 *         feature_id_array = array.array('L')
 *         n = self.get_size()
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             feature = <LCMSFeature>self.get(i)
 *             m = feature.get_size()
 */

#line 197 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2 = __pyx_v_n;

#line 197 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_3 = __pyx_t_2;

#line 197 "ms_deisotope/_c/feature_map/feature_map.pyx"
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 197 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_i = __pyx_t_4;

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":198
 *         n = self.get_size()
 *         for i in range(n):
 *             feature = <LCMSFeature>self.get(i)             # <<<<<<<<<<<<<<
 *             m = feature.get_size()
 *             for j in range(m):
 */

#line 198 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self->__pyx_vtab)->get(__pyx_v_self, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 198, __pyx_L1_error)

#line 198 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 198 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_5 = __pyx_t_1;

#line 198 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_INCREF(__pyx_t_5);

#line 198 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 198 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_XDECREF_SET(__pyx_v_feature, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_5));

#line 198 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_5 = 0;

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":199
 *         for i in range(n):
 *             feature = <LCMSFeature>self.get(i)
 *             m = feature.get_size()             # <<<<<<<<<<<<<<
 *             for j in range(m):
 *                 node = <LCMSFeatureTreeNode>feature.getitem(j)
 */

#line 199 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_m = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_feature->__pyx_base.__pyx_vtab)->__pyx_base.get_size(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_feature));

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":200
 *             feature = <LCMSFeature>self.get(i)
 *             m = feature.get_size()
 *             for j in range(m):             # <<<<<<<<<<<<<<
 *                 node = <LCMSFeatureTreeNode>feature.getitem(j)
 *                 time = node.time
 */

#line 200 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_6 = __pyx_v_m;

#line 200 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_7 = __pyx_t_6;

#line 200 "ms_deisotope/_c/feature_map/feature_map.pyx"
    for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {

#line 200 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_j = __pyx_t_8;

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":201
 *             m = feature.get_size()
 *             for j in range(m):
 *                 node = <LCMSFeatureTreeNode>feature.getitem(j)             # <<<<<<<<<<<<<<
 *                 time = node.time
 *                 q = node.get_members_size()
 */

#line 201 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_feature->__pyx_base.__pyx_vtab)->getitem(__pyx_v_feature, __pyx_v_j)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 201, __pyx_L1_error)

#line 201 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_GOTREF(__pyx_t_5);

#line 201 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_1 = __pyx_t_5;

#line 201 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_INCREF(__pyx_t_1);

#line 201 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 201 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_t_1));

#line 201 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_1 = 0;

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":202
 *             for j in range(m):
 *                 node = <LCMSFeatureTreeNode>feature.getitem(j)
 *                 time = node.time             # <<<<<<<<<<<<<<
 *                 q = node.get_members_size()
 *                 for k in range(q):
 */

#line 202 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_9 = __pyx_v_node->__pyx_base.time;

#line 202 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_time = __pyx_t_9;

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":203
 *                 node = <LCMSFeatureTreeNode>feature.getitem(j)
 *                 time = node.time
 *                 q = node.get_members_size()             # <<<<<<<<<<<<<<
 *                 for k in range(q):
 *                     peak = <PeakBase>node.getitem(k)
 */

#line 203 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_q = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_v_node->__pyx_base.__pyx_vtab)->__pyx_base.get_members_size(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_node));

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":204
 *                 time = node.time
 *                 q = node.get_members_size()
 *                 for k in range(q):             # <<<<<<<<<<<<<<
 *                     peak = <PeakBase>node.getitem(k)
 *                     ion_mobility_array.append(time)
 */

#line 204 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_10 = __pyx_v_q;

#line 204 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_11 = __pyx_t_10;

#line 204 "ms_deisotope/_c/feature_map/feature_map.pyx"
      for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {

#line 204 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_v_k = __pyx_t_12;

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":205
 *                 q = node.get_members_size()
 *                 for k in range(q):
 *                     peak = <PeakBase>node.getitem(k)             # <<<<<<<<<<<<<<
 *                     ion_mobility_array.append(time)
 *                     mz_array.append(peak.mz)
 */

#line 205 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_v_node->__pyx_base.__pyx_vtab)->__pyx_base.getitem(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_node), __pyx_v_k)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 205, __pyx_L1_error)

#line 205 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_GOTREF(__pyx_t_1);

#line 205 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_5 = __pyx_t_1;

#line 205 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_INCREF(__pyx_t_5);

#line 205 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 205 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_5));

#line 205 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_5 = 0;

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":206
 *                 for k in range(q):
 *                     peak = <PeakBase>node.getitem(k)
 *                     ion_mobility_array.append(time)             # <<<<<<<<<<<<<<
 *                     mz_array.append(peak.mz)
 *                     intensity_array.append(peak.intensity)
 */

#line 206 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_5 = PyFloat_FromDouble(__pyx_v_time); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 206, __pyx_L1_error)

#line 206 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_GOTREF(__pyx_t_5);

#line 206 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_13 = __Pyx_PyObject_Append(__pyx_v_ion_mobility_array, __pyx_t_5); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 206, __pyx_L1_error)

#line 206 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":207
 *                     peak = <PeakBase>node.getitem(k)
 *                     ion_mobility_array.append(time)
 *                     mz_array.append(peak.mz)             # <<<<<<<<<<<<<<
 *                     intensity_array.append(peak.intensity)
 *                     feature_id_array.append(i)
 */

#line 207 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_5 = PyFloat_FromDouble(__pyx_v_peak->mz); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 207, __pyx_L1_error)

#line 207 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_GOTREF(__pyx_t_5);

#line 207 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_13 = __Pyx_PyObject_Append(__pyx_v_mz_array, __pyx_t_5); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 207, __pyx_L1_error)

#line 207 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":208
 *                     ion_mobility_array.append(time)
 *                     mz_array.append(peak.mz)
 *                     intensity_array.append(peak.intensity)             # <<<<<<<<<<<<<<
 *                     feature_id_array.append(i)
 * 
 */

#line 208 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_5 = PyFloat_FromDouble(__pyx_v_peak->intensity); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 208, __pyx_L1_error)

#line 208 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_GOTREF(__pyx_t_5);

#line 208 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_13 = __Pyx_PyObject_Append(__pyx_v_intensity_array, __pyx_t_5); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 208, __pyx_L1_error)

#line 208 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":209
 *                     mz_array.append(peak.mz)
 *                     intensity_array.append(peak.intensity)
 *                     feature_id_array.append(i)             # <<<<<<<<<<<<<<
 * 
 *         mz_array = np.array(mz_array, copy=False)
 */

#line 209 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_5 = __Pyx_PyInt_FromSize_t(__pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 209, __pyx_L1_error)

#line 209 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_GOTREF(__pyx_t_5);

#line 209 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_13 = __Pyx_PyObject_Append(__pyx_v_feature_id_array, __pyx_t_5); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 209, __pyx_L1_error)

#line 209 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 209 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 209 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 209 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":211
 *                     feature_id_array.append(i)
 * 
 *         mz_array = np.array(mz_array, copy=False)             # <<<<<<<<<<<<<<
 *         intensity_array = np.array(intensity_array, copy=False)
 *         ion_mobility_array = np.array(ion_mobility_array, copy=False)
 */

#line 211 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 211, __pyx_L1_error)

#line 211 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 211 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 211, __pyx_L1_error)

#line 211 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 211 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 211 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 211, __pyx_L1_error)

#line 211 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 211 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_INCREF(__pyx_v_mz_array);

#line 211 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GIVEREF(__pyx_v_mz_array);

#line 211 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_mz_array);

#line 211 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_14 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 211, __pyx_L1_error)

#line 211 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_14);

#line 211 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (PyDict_SetItem(__pyx_t_14, __pyx_n_s_copy, Py_False) < 0) __PYX_ERR(0, 211, __pyx_L1_error)

#line 211 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_5, __pyx_t_14); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 211, __pyx_L1_error)

#line 211 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_15);

#line 211 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 211 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 211 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

#line 211 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF_SET(__pyx_v_mz_array, __pyx_t_15);

#line 211 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_15 = 0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":212
 * 
 *         mz_array = np.array(mz_array, copy=False)
 *         intensity_array = np.array(intensity_array, copy=False)             # <<<<<<<<<<<<<<
 *         ion_mobility_array = np.array(ion_mobility_array, copy=False)
 *         feature_id_array = np.array(feature_id_array, copy=False)
 */

#line 212 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_np); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 212, __pyx_L1_error)

#line 212 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_15);

#line 212 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_array); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 212, __pyx_L1_error)

#line 212 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_14);

#line 212 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

#line 212 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 212, __pyx_L1_error)

#line 212 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_15);

#line 212 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_INCREF(__pyx_v_intensity_array);

#line 212 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GIVEREF(__pyx_v_intensity_array);

#line 212 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_v_intensity_array);

#line 212 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 212, __pyx_L1_error)

#line 212 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 212 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_copy, Py_False) < 0) __PYX_ERR(0, 212, __pyx_L1_error)

#line 212 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_t_15, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 212, __pyx_L1_error)

#line 212 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 212 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

#line 212 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

#line 212 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 212 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF_SET(__pyx_v_intensity_array, __pyx_t_1);

#line 212 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":213
 *         mz_array = np.array(mz_array, copy=False)
 *         intensity_array = np.array(intensity_array, copy=False)
 *         ion_mobility_array = np.array(ion_mobility_array, copy=False)             # <<<<<<<<<<<<<<
 *         feature_id_array = np.array(feature_id_array, copy=False)
 *         mask = np.lexsort(np.stack((ion_mobility_array, mz_array)))
 */

#line 213 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 213, __pyx_L1_error)

#line 213 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 213 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_array); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 213, __pyx_L1_error)

#line 213 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 213 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 213 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 213, __pyx_L1_error)

#line 213 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 213 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_INCREF(__pyx_v_ion_mobility_array);

#line 213 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GIVEREF(__pyx_v_ion_mobility_array);

#line 213 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_ion_mobility_array);

#line 213 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_15 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 213, __pyx_L1_error)

#line 213 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_15);

#line 213 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (PyDict_SetItem(__pyx_t_15, __pyx_n_s_copy, Py_False) < 0) __PYX_ERR(0, 213, __pyx_L1_error)

#line 213 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_1, __pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 213, __pyx_L1_error)

#line 213 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_14);

#line 213 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 213 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 213 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

#line 213 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF_SET(__pyx_v_ion_mobility_array, __pyx_t_14);

#line 213 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_14 = 0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":214
 *         intensity_array = np.array(intensity_array, copy=False)
 *         ion_mobility_array = np.array(ion_mobility_array, copy=False)
 *         feature_id_array = np.array(feature_id_array, copy=False)             # <<<<<<<<<<<<<<
 *         mask = np.lexsort(np.stack((ion_mobility_array, mz_array)))
 *         return (mz_array[mask], intensity_array[mask], ion_mobility_array[mask], feature_id_array[mask])
 */

#line 214 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_np); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 214, __pyx_L1_error)

#line 214 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_14);

#line 214 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_array); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 214, __pyx_L1_error)

#line 214 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_15);

#line 214 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

#line 214 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_14 = PyTuple_New(1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 214, __pyx_L1_error)

#line 214 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_14);

#line 214 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_INCREF(__pyx_v_feature_id_array);

#line 214 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GIVEREF(__pyx_v_feature_id_array);

#line 214 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_v_feature_id_array);

#line 214 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 214, __pyx_L1_error)

#line 214 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 214 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_copy, Py_False) < 0) __PYX_ERR(0, 214, __pyx_L1_error)

#line 214 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_15, __pyx_t_14, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 214, __pyx_L1_error)

#line 214 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 214 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

#line 214 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

#line 214 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 214 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF_SET(__pyx_v_feature_id_array, __pyx_t_5);

#line 214 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_5 = 0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":215
 *         ion_mobility_array = np.array(ion_mobility_array, copy=False)
 *         feature_id_array = np.array(feature_id_array, copy=False)
 *         mask = np.lexsort(np.stack((ion_mobility_array, mz_array)))             # <<<<<<<<<<<<<<
 *         return (mz_array[mask], intensity_array[mask], ion_mobility_array[mask], feature_id_array[mask])
 * 
 */

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 215, __pyx_L1_error)

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_lexsort); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 215, __pyx_L1_error)

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_14);

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_np); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 215, __pyx_L1_error)

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_15);

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_stack); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 215, __pyx_L1_error)

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_16);

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 215, __pyx_L1_error)

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_15);

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_INCREF(__pyx_v_ion_mobility_array);

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GIVEREF(__pyx_v_ion_mobility_array);

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_v_ion_mobility_array);

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_INCREF(__pyx_v_mz_array);

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GIVEREF(__pyx_v_mz_array);

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_v_mz_array);

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_17 = NULL;

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_16))) {

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_16);

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (likely(__pyx_t_17)) {

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_INCREF(__pyx_t_17);

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_INCREF(function);

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_DECREF_SET(__pyx_t_16, function);

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = (__pyx_t_17) ? __Pyx_PyObject_Call2Args(__pyx_t_16, __pyx_t_17, __pyx_t_15) : __Pyx_PyObject_CallOneArg(__pyx_t_16, __pyx_t_15);

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 215, __pyx_L1_error)

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_16 = NULL;

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_14))) {

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_14);

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (likely(__pyx_t_16)) {

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_INCREF(__pyx_t_16);

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_INCREF(function);

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_DECREF_SET(__pyx_t_14, function);

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_5 = (__pyx_t_16) ? __Pyx_PyObject_Call2Args(__pyx_t_14, __pyx_t_16, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_t_1);

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 215, __pyx_L1_error)

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_v_mask = __pyx_t_5;

#line 215 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_5 = 0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":216
 *         feature_id_array = np.array(feature_id_array, copy=False)
 *         mask = np.lexsort(np.stack((ion_mobility_array, mz_array)))
 *         return (mz_array[mask], intensity_array[mask], ion_mobility_array[mask], feature_id_array[mask])             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 216 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 216 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_mz_array, __pyx_v_mask); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 216 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_14 = __Pyx_PyObject_GetItem(__pyx_v_intensity_array, __pyx_v_mask); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_14);

#line 216 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_ion_mobility_array, __pyx_v_mask); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_16 = __Pyx_PyObject_GetItem(__pyx_v_feature_id_array, __pyx_v_mask); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_16);

#line 216 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_15 = PyTuple_New(4); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_15);

#line 216 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GIVEREF(__pyx_t_5);

#line 216 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_5);

#line 216 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GIVEREF(__pyx_t_14);

#line 216 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_14);

#line 216 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_t_1);

#line 216 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GIVEREF(__pyx_t_16);

#line 216 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyTuple_SET_ITEM(__pyx_t_15, 3, __pyx_t_16);

#line 216 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_5 = 0;

#line 216 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_14 = 0;

#line 216 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = 0;

#line 216 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_16 = 0;

#line 216 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = __pyx_t_15;

#line 216 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_15 = 0;

#line 216 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":184
 *         return self.__class__(result)
 * 
 *     def as_arrays(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n, j, m, k, q
 */

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L1_error:;

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_14);

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_15);

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_16);

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_17);

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.as_arrays", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = NULL;

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_feature);

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_node);

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_v_mz_array);

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_v_intensity_array);

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_v_ion_mobility_array);

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_v_feature_id_array);

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_v_mask);

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 184 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_map.pxd":7
 * cdef class LCMSFeatureMap(object):
 *     cdef:
 *         public list features             # <<<<<<<<<<<<<<
 * 
 *     cdef Py_ssize_t get_size(self)
 */

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"


#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
/* Python wrapper */

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_8features_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_8features_1__get__(PyObject *__pyx_v_self) {

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  PyObject *__pyx_r = 0;

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_RefNannyDeclarations

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_8features___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self));

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"


#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  /* function exit code */

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_RefNannyFinishContext();

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  return __pyx_r;

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
}

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"


#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_8features___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self) {

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  PyObject *__pyx_r = NULL;

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_RefNannyDeclarations

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_INCREF(__pyx_v_self->features);

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __pyx_r = __pyx_v_self->features;

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  goto __pyx_L0;

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"


#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  /* function exit code */

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __pyx_L0:;

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_RefNannyFinishContext();

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  return __pyx_r;

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
}

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"


#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
/* Python wrapper */

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_8features_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_8features_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  int __pyx_r;

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_RefNannyDeclarations

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_8features_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"


#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  /* function exit code */

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_RefNannyFinishContext();

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  return __pyx_r;

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
}

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"


#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_8features_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, PyObject *__pyx_v_value) {

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  int __pyx_r;

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_RefNannyDeclarations

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  int __pyx_lineno = 0;

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  const char *__pyx_filename = NULL;

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  int __pyx_clineno = 0;

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  if (!(likely(PyList_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 7, __pyx_L1_error)

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __pyx_t_1 = __pyx_v_value;

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_INCREF(__pyx_t_1);

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_GIVEREF(__pyx_t_1);

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_GOTREF(__pyx_v_self->features);

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_DECREF(__pyx_v_self->features);

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __pyx_v_self->features = ((PyObject*)__pyx_t_1);

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __pyx_t_1 = 0;

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"


#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  /* function exit code */

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __pyx_r = 0;

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  goto __pyx_L0;

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __pyx_L1_error:;

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.features.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __pyx_r = -1;

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __pyx_L0:;

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_RefNannyFinishContext();

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  return __pyx_r;

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
}

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"


#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
/* Python wrapper */

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_8features_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_8features_5__del__(PyObject *__pyx_v_self) {

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  int __pyx_r;

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_RefNannyDeclarations

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_8features_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self));

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"


#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  /* function exit code */

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_RefNannyFinishContext();

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  return __pyx_r;

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
}

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"


#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_8features_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self) {

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  int __pyx_r;

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_RefNannyDeclarations

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_INCREF(Py_None);

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_GIVEREF(Py_None);

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_GOTREF(__pyx_v_self->features);

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_DECREF(__pyx_v_self->features);

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __pyx_v_self->features = ((PyObject*)Py_None);

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"


#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  /* function exit code */

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __pyx_r = 0;

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  __Pyx_RefNannyFinishContext();

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
  return __pyx_r;

#line 7 "ms_deisotope/_c/feature_map/feature_map.pxd"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_29__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_28__reduce_cython__[] = "LCMSFeatureMap.__reduce_cython__(self)";

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_29__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_28__reduce_cython__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_28__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self) {
  PyObject *__pyx_v_state = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v__dict = 0
#line 1 "(tree fragment)"
;
  int __pyx_v_use_setstate
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_2;

#line 1 "(tree fragment)"
  int __pyx_t_3;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.features,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */

#line 5 "(tree fragment)"
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
  __Pyx_INCREF(__pyx_v_self->features);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_v_self->features);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->features);

#line 5 "(tree fragment)"
  __pyx_v_state = ((PyObject*)__pyx_t_1);

#line 5 "(tree fragment)"
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.features,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */

#line 6 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 6 "(tree fragment)"
  __pyx_v__dict = __pyx_t_1;

#line 6 "(tree fragment)"
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.features,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
  __pyx_t_2 = (__pyx_v__dict != Py_None);

#line 7 "(tree fragment)"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 7 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */

#line 8 "(tree fragment)"
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 8 "(tree fragment)"
    __Pyx_INCREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);

#line 8 "(tree fragment)"
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 8 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 8 "(tree fragment)"
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));

#line 8 "(tree fragment)"
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.features is not None
 */

#line 9 "(tree fragment)"
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.features,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
    goto __pyx_L3;

#line 7 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.features is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_LCMSFeatureMap, (type(self), 0x9f034b6, None), state
 */

#line 11 "(tree fragment)"
  /*else*/ {

#line 11 "(tree fragment)"
    __pyx_t_3 = (__pyx_v_self->features != ((PyObject*)Py_None));

#line 11 "(tree fragment)"
    __pyx_v_use_setstate = __pyx_t_3;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.features is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_LCMSFeatureMap, (type(self), 0x9f034b6, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  __pyx_t_3 = (__pyx_v_use_setstate != 0);

#line 12 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = self.features is not None
 *     if use_setstate:
 *         return __pyx_unpickle_LCMSFeatureMap, (type(self), 0x9f034b6, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_LCMSFeatureMap, (type(self), 0x9f034b6, state)
 */

#line 13 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 13 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_LCMSFeatureMap); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 13 "(tree fragment)"
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 13 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_166737078);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_166737078);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_166737078);

#line 13 "(tree fragment)"
    __Pyx_INCREF(Py_None);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(Py_None);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);

#line 13 "(tree fragment)"
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_4);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_1);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);

#line 13 "(tree fragment)"
    __pyx_t_4 = 0;

#line 13 "(tree fragment)"
    __pyx_t_1 = 0;

#line 13 "(tree fragment)"
    __pyx_r = __pyx_t_5;

#line 13 "(tree fragment)"
    __pyx_t_5 = 0;

#line 13 "(tree fragment)"
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.features is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_LCMSFeatureMap, (type(self), 0x9f034b6, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_LCMSFeatureMap, (type(self), 0x9f034b6, None), state
 *     else:
 *         return __pyx_unpickle_LCMSFeatureMap, (type(self), 0x9f034b6, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_LCMSFeatureMap__set_state(self, __pyx_state)
 */

#line 15 "(tree fragment)"
  /*else*/ {

#line 15 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 15 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_LCMSFeatureMap); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 15 "(tree fragment)"
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 15 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_166737078);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_166737078);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_166737078);

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);

#line 15 "(tree fragment)"
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_5);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_1);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);

#line 15 "(tree fragment)"
    __pyx_t_5 = 0;

#line 15 "(tree fragment)"
    __pyx_t_1 = 0;

#line 15 "(tree fragment)"
    __pyx_r = __pyx_t_4;

#line 15 "(tree fragment)"
    __pyx_t_4 = 0;

#line 15 "(tree fragment)"
    goto __pyx_L0;

#line 15 "(tree fragment)"
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v_state);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v__dict);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LCMSFeatureMap, (type(self), 0x9f034b6, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LCMSFeatureMap__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
/* Python wrapper */

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_31__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 16 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_30__setstate_cython__[] = "LCMSFeatureMap.__setstate_cython__(self, __pyx_state)";

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_31__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 16 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_30__setstate_cython__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_30__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 16 "(tree fragment)"
  int __pyx_lineno = 0;

#line 16 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 16 "(tree fragment)"
  int __pyx_clineno = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_LCMSFeatureMap, (type(self), 0x9f034b6, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_LCMSFeatureMap__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */

#line 17 "(tree fragment)"
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map___pyx_unpickle_LCMSFeatureMap__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 17 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LCMSFeatureMap, (type(self), 0x9f034b6, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LCMSFeatureMap__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 16 "(tree fragment)"
  goto __pyx_L0;

#line 16 "(tree fragment)"
  __pyx_L1_error:;

#line 16 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 16 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 16 "(tree fragment)"
  __pyx_r = NULL;

#line 16 "(tree fragment)"
  __pyx_L0:;

#line 16 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

/* "ms_deisotope/_c/feature_map/feature_map.pyx":220
 * 
 * @cython.binding(True)
 * cpdef list split_sparse(LCMSFeatureMap self, double delta_rt=1.0, size_t min_size=2):             # <<<<<<<<<<<<<<
 *     cdef:
 *         list result, chunks
 */

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_1split_sparse(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_split_sparse(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_split_sparse *__pyx_optional_args) {

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  double __pyx_v_delta_rt = ((double)1.0);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  size_t __pyx_v_min_size = ((size_t)2);
  PyObject *__pyx_v_result = 0
#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  PyObject *__pyx_v_chunks = 0
#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  size_t __pyx_v_i
#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  size_t __pyx_v_n
#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  size_t __pyx_v_j
#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  size_t __pyx_v_m
#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_feature = 0
#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_chunk = 0
#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = NULL;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  size_t __pyx_t_2;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  size_t __pyx_t_3;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  size_t __pyx_t_4;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_split_sparse __pyx_t_5;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  size_t __pyx_t_6;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  size_t __pyx_t_7;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  size_t __pyx_t_8;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_9;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_t_10;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_t_11;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("split_sparse", 0);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (__pyx_optional_args) {

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_delta_rt = __pyx_optional_args->delta_rt;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
      if (__pyx_optional_args->__pyx_n > 1) {

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_v_min_size = __pyx_optional_args->min_size;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":225
 *         size_t i, n, j, m
 *         LCMSFeature feature, chunk
 *     result = []             # <<<<<<<<<<<<<<
 *     n = self.get_size()
 *     for i in range(n):
 */

#line 225 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 225, __pyx_L1_error)

#line 225 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 225 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_v_result = ((PyObject*)__pyx_t_1);

#line 225 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":226
 *         LCMSFeature feature, chunk
 *     result = []
 *     n = self.get_size()             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         feature = self.get(i)
 */

#line 226 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_v_n = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self->__pyx_vtab)->get_size(__pyx_v_self);

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":227
 *     result = []
 *     n = self.get_size()
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         feature = self.get(i)
 *         chunks = feature.split_sparse(delta_rt)
 */

#line 227 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2 = __pyx_v_n;

#line 227 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_3 = __pyx_t_2;

#line 227 "ms_deisotope/_c/feature_map/feature_map.pyx"
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 227 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_i = __pyx_t_4;

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":228
 *     n = self.get_size()
 *     for i in range(n):
 *         feature = self.get(i)             # <<<<<<<<<<<<<<
 *         chunks = feature.split_sparse(delta_rt)
 *         m = PyList_GET_SIZE(chunks)
 */

#line 228 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v_self->__pyx_vtab)->get(__pyx_v_self, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 228, __pyx_L1_error)

#line 228 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 228 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_XDECREF_SET(__pyx_v_feature, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_1));

#line 228 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":229
 *     for i in range(n):
 *         feature = self.get(i)
 *         chunks = feature.split_sparse(delta_rt)             # <<<<<<<<<<<<<<
 *         m = PyList_GET_SIZE(chunks)
 *         for j in range(m):
 */

#line 229 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_5.__pyx_n = 1;

#line 229 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_5.delta_rt = __pyx_v_delta_rt;

#line 229 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_feature->__pyx_base.__pyx_vtab)->split_sparse(__pyx_v_feature, 0, &__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 229, __pyx_L1_error)

#line 229 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 229 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_XDECREF_SET(__pyx_v_chunks, ((PyObject*)__pyx_t_1));

#line 229 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":230
 *         feature = self.get(i)
 *         chunks = feature.split_sparse(delta_rt)
 *         m = PyList_GET_SIZE(chunks)             # <<<<<<<<<<<<<<
 *         for j in range(m):
 *             chunk = <LCMSFeature>PyList_GET_ITEM(chunks, j)
 */

#line 230 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_m = PyList_GET_SIZE(__pyx_v_chunks);

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":231
 *         chunks = feature.split_sparse(delta_rt)
 *         m = PyList_GET_SIZE(chunks)
 *         for j in range(m):             # <<<<<<<<<<<<<<
 *             chunk = <LCMSFeature>PyList_GET_ITEM(chunks, j)
 *             if chunk.get_size() >= min_size:
 */

#line 231 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_6 = __pyx_v_m;

#line 231 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_7 = __pyx_t_6;

#line 231 "ms_deisotope/_c/feature_map/feature_map.pyx"
    for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {

#line 231 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_j = __pyx_t_8;

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":232
 *         m = PyList_GET_SIZE(chunks)
 *         for j in range(m):
 *             chunk = <LCMSFeature>PyList_GET_ITEM(chunks, j)             # <<<<<<<<<<<<<<
 *             if chunk.get_size() >= min_size:
 *                 result.append(chunk)
 */

#line 232 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_9 = PyList_GET_ITEM(__pyx_v_chunks, __pyx_v_j);

#line 232 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_1 = ((PyObject *)__pyx_t_9);

#line 232 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_INCREF(__pyx_t_1);

#line 232 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_XDECREF_SET(__pyx_v_chunk, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_1));

#line 232 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_1 = 0;

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":233
 *         for j in range(m):
 *             chunk = <LCMSFeature>PyList_GET_ITEM(chunks, j)
 *             if chunk.get_size() >= min_size:             # <<<<<<<<<<<<<<
 *                 result.append(chunk)
 *     return result
 */

#line 233 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_10 = ((((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_chunk->__pyx_base.__pyx_vtab)->__pyx_base.get_size(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_chunk)) >= __pyx_v_min_size) != 0);

#line 233 "ms_deisotope/_c/feature_map/feature_map.pyx"
      if (__pyx_t_10) {

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":234
 *             chunk = <LCMSFeature>PyList_GET_ITEM(chunks, j)
 *             if chunk.get_size() >= min_size:
 *                 result.append(chunk)             # <<<<<<<<<<<<<<
 *     return result
 * 
 */

#line 234 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_result, ((PyObject *)__pyx_v_chunk)); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 234, __pyx_L1_error)

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":233
 *         for j in range(m):
 *             chunk = <LCMSFeature>PyList_GET_ITEM(chunks, j)
 *             if chunk.get_size() >= min_size:             # <<<<<<<<<<<<<<
 *                 result.append(chunk)
 *     return result
 */

#line 233 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 233 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 233 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":235
 *             if chunk.get_size() >= min_size:
 *                 result.append(chunk)
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 235 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 235 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_INCREF(__pyx_v_result);

#line 235 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = __pyx_v_result;

#line 235 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":220
 * 
 * @cython.binding(True)
 * cpdef list split_sparse(LCMSFeatureMap self, double delta_rt=1.0, size_t min_size=2):             # <<<<<<<<<<<<<<
 *     cdef:
 *         list result, chunks
 */

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L1_error:;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.split_sparse", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = 0;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_v_result);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_v_chunks);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_feature);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_chunk);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
/* Python wrapper */

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_1split_sparse(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_11feature_map_split_sparse[] = "split_sparse(LCMSFeatureMap self, double delta_rt=1.0, size_t min_size=2) -> list";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_11feature_map_11feature_map_1split_sparse = 
#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
{"split_sparse", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_1split_sparse, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_11feature_map_split_sparse};

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_1split_sparse(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self = 0
#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  double __pyx_v_delta_rt
#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  size_t __pyx_v_min_size
#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = 0;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("split_sparse (wrapper)", 0);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  {

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_delta_rt,&__pyx_n_s_min_size,0};

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
    PyObject* values[3] = {0,0,0};

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (unlikely(__pyx_kwds)) {

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
      Py_ssize_t kw_args;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
      switch (pos_args) {
        case  3: 
#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
        case  0: break;
        default: 
#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
goto __pyx_L5_argtuple_error;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
      switch (pos_args) {

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
        case  0:

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else 
#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
goto __pyx_L5_argtuple_error;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
        case  1:

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (kw_args > 0) {

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_delta_rt);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
          if (value) { values[1] = value; kw_args--; }

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
        case  2:

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (kw_args > 0) {

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_size);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
      if (unlikely(kw_args > 0)) {

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "split_sparse") < 0)) __PYX_ERR(0, 220, __pyx_L3_error)

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
    } else {

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
        break;
        default: 
#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
goto __pyx_L5_argtuple_error;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_self = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)values[0]);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (values[1]) {

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_delta_rt = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_delta_rt == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 220, __pyx_L3_error)

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
    } else {

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_delta_rt = ((double)1.0);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (values[2]) {

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_min_size = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_min_size == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 220, __pyx_L3_error)

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
    } else {

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_min_size = ((size_t)2);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("split_sparse", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
__PYX_ERR(0, 220, __pyx_L3_error)

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L3_error:;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.split_sparse", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return NULL;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap, 1, "self", 0))) __PYX_ERR(0, 220, __pyx_L1_error)
  __pyx_r = 
#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_split_sparse(__pyx_self, __pyx_v_self, __pyx_v_delta_rt, __pyx_v_min_size);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L0;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L1_error:;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = NULL;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_split_sparse(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v_self, double __pyx_v_delta_rt, size_t __pyx_v_min_size) {

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = NULL;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_split_sparse __pyx_t_2;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("split_sparse", 0);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2.__pyx_n = 2;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2.delta_rt = __pyx_v_delta_rt;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2.min_size = __pyx_v_min_size;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_split_sparse(__pyx_v_self, 0, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 220, __pyx_L1_error)

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = __pyx_t_1;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = 0;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L0;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L1_error:;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.split_sparse", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = NULL;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_map.pyx":238
 * 
 * 
 * cpdef tuple binary_search_with_flag(list array, double mz, double error_tolerance):             # <<<<<<<<<<<<<<
 *     cdef:
 *         int i, n, lo, hi, mid, low_end, high_end
 */

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_3binary_search_with_flag(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_binary_search_with_flag(PyObject *__pyx_v_array, double __pyx_v_mz, double __pyx_v_error_tolerance, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_i
#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  int __pyx_v_n
#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  int __pyx_v_lo
#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  int __pyx_v_hi
#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  int __pyx_v_mid
#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  int __pyx_v_low_end
#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  int __pyx_v_high_end
#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  double __pyx_v_err
#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_x = 0
#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = NULL;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  Py_ssize_t __pyx_t_1;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_t_2;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_3;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  double __pyx_t_5;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_7 = NULL;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("binary_search_with_flag", 0);

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":243
 *         double err
 *         LCMSFeature x
 *     lo = 0             # <<<<<<<<<<<<<<
 *     n = hi = len(array)
 *     while hi != lo:
 */

#line 243 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_v_lo = 0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":244
 *         LCMSFeature x
 *     lo = 0
 *     n = hi = len(array)             # <<<<<<<<<<<<<<
 *     while hi != lo:
 *         mid = (hi + lo) // 2
 */

#line 244 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (unlikely(__pyx_v_array == Py_None)) {

#line 244 "ms_deisotope/_c/feature_map/feature_map.pyx"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 244 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

#line 244 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = PyList_GET_SIZE(__pyx_v_array); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 244, __pyx_L1_error)

#line 244 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_v_n = __pyx_t_1;

#line 244 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_v_hi = __pyx_t_1;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":245
 *     lo = 0
 *     n = hi = len(array)
 *     while hi != lo:             # <<<<<<<<<<<<<<
 *         mid = (hi + lo) // 2
 *         x = <LCMSFeature>PyList_GET_ITEM(array, mid)
 */

#line 245 "ms_deisotope/_c/feature_map/feature_map.pyx"
  while (1) {

#line 245 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_2 = ((__pyx_v_hi != __pyx_v_lo) != 0);

#line 245 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (!__pyx_t_2) break;

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":246
 *     n = hi = len(array)
 *     while hi != lo:
 *         mid = (hi + lo) // 2             # <<<<<<<<<<<<<<
 *         x = <LCMSFeature>PyList_GET_ITEM(array, mid)
 *         err = (x.get_mz() - mz) / mz
 */

#line 246 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_mid = __Pyx_div_long((__pyx_v_hi + __pyx_v_lo), 2);

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":247
 *     while hi != lo:
 *         mid = (hi + lo) // 2
 *         x = <LCMSFeature>PyList_GET_ITEM(array, mid)             # <<<<<<<<<<<<<<
 *         err = (x.get_mz() - mz) / mz
 *         if abs(err) <= error_tolerance:
 */

#line 247 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_3 = PyList_GET_ITEM(__pyx_v_array, __pyx_v_mid);

#line 247 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_4 = ((PyObject *)__pyx_t_3);

#line 247 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_INCREF(__pyx_t_4);

#line 247 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_XDECREF_SET(__pyx_v_x, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_4));

#line 247 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":248
 *         mid = (hi + lo) // 2
 *         x = <LCMSFeature>PyList_GET_ITEM(array, mid)
 *         err = (x.get_mz() - mz) / mz             # <<<<<<<<<<<<<<
 *         if abs(err) <= error_tolerance:
 *             i = mid - 1
 */

#line 248 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_5 = (((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_x->__pyx_base.__pyx_vtab)->__pyx_base.get_mz(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_x)) - __pyx_v_mz);

#line 248 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (unlikely(__pyx_v_mz == 0)) {

#line 248 "ms_deisotope/_c/feature_map/feature_map.pyx"
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");

#line 248 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __PYX_ERR(0, 248, __pyx_L1_error)

#line 248 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 248 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_err = (__pyx_t_5 / __pyx_v_mz);

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":249
 *         x = <LCMSFeature>PyList_GET_ITEM(array, mid)
 *         err = (x.get_mz() - mz) / mz
 *         if abs(err) <= error_tolerance:             # <<<<<<<<<<<<<<
 *             i = mid - 1
 *             # Begin Sweep forward
 */

#line 249 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_2 = ((fabs(__pyx_v_err) <= __pyx_v_error_tolerance) != 0);

#line 249 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":250
 *         err = (x.get_mz() - mz) / mz
 *         if abs(err) <= error_tolerance:
 *             i = mid - 1             # <<<<<<<<<<<<<<
 *             # Begin Sweep forward
 *             while i > 0:
 */

#line 250 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_i = (__pyx_v_mid - 1);

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":252
 *             i = mid - 1
 *             # Begin Sweep forward
 *             while i > 0:             # <<<<<<<<<<<<<<
 *                 x = array[i]
 *                 err = (x.get_mz() - mz) / mz
 */

#line 252 "ms_deisotope/_c/feature_map/feature_map.pyx"
      while (1) {

#line 252 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_2 = ((__pyx_v_i > 0) != 0);

#line 252 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (!__pyx_t_2) break;

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":253
 *             # Begin Sweep forward
 *             while i > 0:
 *                 x = array[i]             # <<<<<<<<<<<<<<
 *                 err = (x.get_mz() - mz) / mz
 *                 if abs(err) <= error_tolerance:
 */

#line 253 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (unlikely(__pyx_v_array == Py_None)) {

#line 253 "ms_deisotope/_c/feature_map/feature_map.pyx"
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 253 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __PYX_ERR(0, 253, __pyx_L1_error)

#line 253 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

#line 253 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_4 = __Pyx_GetItemInt_List(__pyx_v_array, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 253, __pyx_L1_error)

#line 253 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 253 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature))))) __PYX_ERR(0, 253, __pyx_L1_error)

#line 253 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_DECREF_SET(__pyx_v_x, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_4));

#line 253 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_4 = 0;

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":254
 *             while i > 0:
 *                 x = array[i]
 *                 err = (x.get_mz() - mz) / mz             # <<<<<<<<<<<<<<
 *                 if abs(err) <= error_tolerance:
 *                     i -= 1
 */

#line 254 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_5 = (((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_x->__pyx_base.__pyx_vtab)->__pyx_base.get_mz(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_x)) - __pyx_v_mz);

#line 254 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (unlikely(__pyx_v_mz == 0)) {

#line 254 "ms_deisotope/_c/feature_map/feature_map.pyx"
          PyErr_SetString(PyExc_ZeroDivisionError, "float division");

#line 254 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __PYX_ERR(0, 254, __pyx_L1_error)

#line 254 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

#line 254 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_v_err = (__pyx_t_5 / __pyx_v_mz);

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":255
 *                 x = array[i]
 *                 err = (x.get_mz() - mz) / mz
 *                 if abs(err) <= error_tolerance:             # <<<<<<<<<<<<<<
 *                     i -= 1
 *                     continue
 */

#line 255 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_2 = ((fabs(__pyx_v_err) <= __pyx_v_error_tolerance) != 0);

#line 255 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (__pyx_t_2) {

          /* "ms_deisotope/_c/feature_map/feature_map.pyx":256
 *                 err = (x.get_mz() - mz) / mz
 *                 if abs(err) <= error_tolerance:
 *                     i -= 1             # <<<<<<<<<<<<<<
 *                     continue
 *                 else:
 */

#line 256 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_v_i = (__pyx_v_i - 1);

          /* "ms_deisotope/_c/feature_map/feature_map.pyx":257
 *                 if abs(err) <= error_tolerance:
 *                     i -= 1
 *                     continue             # <<<<<<<<<<<<<<
 *                 else:
 *                     break
 */

#line 257 "ms_deisotope/_c/feature_map/feature_map.pyx"
          goto __pyx_L6_continue;

          /* "ms_deisotope/_c/feature_map/feature_map.pyx":255
 *                 x = array[i]
 *                 err = (x.get_mz() - mz) / mz
 *                 if abs(err) <= error_tolerance:             # <<<<<<<<<<<<<<
 *                     i -= 1
 *                     continue
 */

#line 255 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":259
 *                     continue
 *                 else:
 *                     break             # <<<<<<<<<<<<<<
 *             low_end = i
 *             i = mid + 1
 */

#line 259 "ms_deisotope/_c/feature_map/feature_map.pyx"
        /*else*/ {

#line 259 "ms_deisotope/_c/feature_map/feature_map.pyx"
          goto __pyx_L7_break;

#line 259 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

#line 259 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_L6_continue:;

#line 259 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 259 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_L7_break:;

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":260
 *                 else:
 *                     break
 *             low_end = i             # <<<<<<<<<<<<<<
 *             i = mid + 1
 * 
 */

#line 260 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_low_end = __pyx_v_i;

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":261
 *                     break
 *             low_end = i
 *             i = mid + 1             # <<<<<<<<<<<<<<
 * 
 *             # Begin Sweep backward
 */

#line 261 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_i = (__pyx_v_mid + 1);

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":264
 * 
 *             # Begin Sweep backward
 *             while i < n:             # <<<<<<<<<<<<<<
 *                 x = array[i]
 *                 err = (x.get_mz() - mz) / mz
 */

#line 264 "ms_deisotope/_c/feature_map/feature_map.pyx"
      while (1) {

#line 264 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_2 = ((__pyx_v_i < __pyx_v_n) != 0);

#line 264 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (!__pyx_t_2) break;

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":265
 *             # Begin Sweep backward
 *             while i < n:
 *                 x = array[i]             # <<<<<<<<<<<<<<
 *                 err = (x.get_mz() - mz) / mz
 *                 if abs(err) <= error_tolerance:
 */

#line 265 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (unlikely(__pyx_v_array == Py_None)) {

#line 265 "ms_deisotope/_c/feature_map/feature_map.pyx"
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 265 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __PYX_ERR(0, 265, __pyx_L1_error)

#line 265 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

#line 265 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_4 = __Pyx_GetItemInt_List(__pyx_v_array, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 265, __pyx_L1_error)

#line 265 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 265 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature))))) __PYX_ERR(0, 265, __pyx_L1_error)

#line 265 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_DECREF_SET(__pyx_v_x, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_4));

#line 265 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_4 = 0;

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":266
 *             while i < n:
 *                 x = array[i]
 *                 err = (x.get_mz() - mz) / mz             # <<<<<<<<<<<<<<
 *                 if abs(err) <= error_tolerance:
 *                     i += 1
 */

#line 266 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_5 = (((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_x->__pyx_base.__pyx_vtab)->__pyx_base.get_mz(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_x)) - __pyx_v_mz);

#line 266 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (unlikely(__pyx_v_mz == 0)) {

#line 266 "ms_deisotope/_c/feature_map/feature_map.pyx"
          PyErr_SetString(PyExc_ZeroDivisionError, "float division");

#line 266 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __PYX_ERR(0, 266, __pyx_L1_error)

#line 266 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

#line 266 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_v_err = (__pyx_t_5 / __pyx_v_mz);

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":267
 *                 x = array[i]
 *                 err = (x.get_mz() - mz) / mz
 *                 if abs(err) <= error_tolerance:             # <<<<<<<<<<<<<<
 *                     i += 1
 *                     continue
 */

#line 267 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_2 = ((fabs(__pyx_v_err) <= __pyx_v_error_tolerance) != 0);

#line 267 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (__pyx_t_2) {

          /* "ms_deisotope/_c/feature_map/feature_map.pyx":268
 *                 err = (x.get_mz() - mz) / mz
 *                 if abs(err) <= error_tolerance:
 *                     i += 1             # <<<<<<<<<<<<<<
 *                     continue
 *                 else:
 */

#line 268 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_v_i = (__pyx_v_i + 1);

          /* "ms_deisotope/_c/feature_map/feature_map.pyx":269
 *                 if abs(err) <= error_tolerance:
 *                     i += 1
 *                     continue             # <<<<<<<<<<<<<<
 *                 else:
 *                     break
 */

#line 269 "ms_deisotope/_c/feature_map/feature_map.pyx"
          goto __pyx_L9_continue;

          /* "ms_deisotope/_c/feature_map/feature_map.pyx":267
 *                 x = array[i]
 *                 err = (x.get_mz() - mz) / mz
 *                 if abs(err) <= error_tolerance:             # <<<<<<<<<<<<<<
 *                     i += 1
 *                     continue
 */

#line 267 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":271
 *                     continue
 *                 else:
 *                     break             # <<<<<<<<<<<<<<
 *             high_end = i
 *             return list(range(low_end, high_end)), True
 */

#line 271 "ms_deisotope/_c/feature_map/feature_map.pyx"
        /*else*/ {

#line 271 "ms_deisotope/_c/feature_map/feature_map.pyx"
          goto __pyx_L10_break;

#line 271 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

#line 271 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_L9_continue:;

#line 271 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 271 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_L10_break:;

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":272
 *                 else:
 *                     break
 *             high_end = i             # <<<<<<<<<<<<<<
 *             return list(range(low_end, high_end)), True
 *         elif (hi - lo) == 1:
 */

#line 272 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_high_end = __pyx_v_i;

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":273
 *                     break
 *             high_end = i
 *             return list(range(low_end, high_end)), True             # <<<<<<<<<<<<<<
 *         elif (hi - lo) == 1:
 *             return [mid], False
 */

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_XDECREF(__pyx_r);

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_low_end); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 273, __pyx_L1_error)

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_GOTREF(__pyx_t_4);

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_high_end); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 273, __pyx_L1_error)

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 273, __pyx_L1_error)

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_GOTREF(__pyx_t_7);

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_GIVEREF(__pyx_t_4);

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4);

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_GIVEREF(__pyx_t_6);

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_6);

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_4 = 0;

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_6 = 0;

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_7, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 273, __pyx_L1_error)

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_7 = PySequence_List(__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 273, __pyx_L1_error)

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_GOTREF(__pyx_t_7);

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 273, __pyx_L1_error)

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_GIVEREF(__pyx_t_7);

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_7);

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_INCREF(Py_True);

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_GIVEREF(Py_True);

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      PyTuple_SET_ITEM(__pyx_t_6, 1, Py_True);

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_7 = 0;

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_r = ((PyObject*)__pyx_t_6);

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_6 = 0;

#line 273 "ms_deisotope/_c/feature_map/feature_map.pyx"
      goto __pyx_L0;

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":249
 *         x = <LCMSFeature>PyList_GET_ITEM(array, mid)
 *         err = (x.get_mz() - mz) / mz
 *         if abs(err) <= error_tolerance:             # <<<<<<<<<<<<<<
 *             i = mid - 1
 *             # Begin Sweep forward
 */

#line 249 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":274
 *             high_end = i
 *             return list(range(low_end, high_end)), True
 *         elif (hi - lo) == 1:             # <<<<<<<<<<<<<<
 *             return [mid], False
 *         elif err > 0:
 */

#line 274 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_2 = (((__pyx_v_hi - __pyx_v_lo) == 1) != 0);

#line 274 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":275
 *             return list(range(low_end, high_end)), True
 *         elif (hi - lo) == 1:
 *             return [mid], False             # <<<<<<<<<<<<<<
 *         elif err > 0:
 *             hi = mid
 */

#line 275 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_XDECREF(__pyx_r);

#line 275 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_mid); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 275, __pyx_L1_error)

#line 275 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 275 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_7 = PyList_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 275, __pyx_L1_error)

#line 275 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_GOTREF(__pyx_t_7);

#line 275 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_GIVEREF(__pyx_t_6);

#line 275 "ms_deisotope/_c/feature_map/feature_map.pyx"
      PyList_SET_ITEM(__pyx_t_7, 0, __pyx_t_6);

#line 275 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_6 = 0;

#line 275 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 275, __pyx_L1_error)

#line 275 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 275 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_GIVEREF(__pyx_t_7);

#line 275 "ms_deisotope/_c/feature_map/feature_map.pyx"
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_7);

#line 275 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_INCREF(Py_False);

#line 275 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __Pyx_GIVEREF(Py_False);

#line 275 "ms_deisotope/_c/feature_map/feature_map.pyx"
      PyTuple_SET_ITEM(__pyx_t_6, 1, Py_False);

#line 275 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_7 = 0;

#line 275 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_r = ((PyObject*)__pyx_t_6);

#line 275 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_t_6 = 0;

#line 275 "ms_deisotope/_c/feature_map/feature_map.pyx"
      goto __pyx_L0;

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":274
 *             high_end = i
 *             return list(range(low_end, high_end)), True
 *         elif (hi - lo) == 1:             # <<<<<<<<<<<<<<
 *             return [mid], False
 *         elif err > 0:
 */

#line 274 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":276
 *         elif (hi - lo) == 1:
 *             return [mid], False
 *         elif err > 0:             # <<<<<<<<<<<<<<
 *             hi = mid
 *         elif err < 0:
 */

#line 276 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_2 = ((__pyx_v_err > 0.0) != 0);

#line 276 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":277
 *             return [mid], False
 *         elif err > 0:
 *             hi = mid             # <<<<<<<<<<<<<<
 *         elif err < 0:
 *             lo = mid
 */

#line 277 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_hi = __pyx_v_mid;

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":276
 *         elif (hi - lo) == 1:
 *             return [mid], False
 *         elif err > 0:             # <<<<<<<<<<<<<<
 *             hi = mid
 *         elif err < 0:
 */

#line 276 "ms_deisotope/_c/feature_map/feature_map.pyx"
      goto __pyx_L5;

#line 276 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":278
 *         elif err > 0:
 *             hi = mid
 *         elif err < 0:             # <<<<<<<<<<<<<<
 *             lo = mid
 *     return [0], False
 */

#line 278 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_2 = ((__pyx_v_err < 0.0) != 0);

#line 278 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":279
 *             hi = mid
 *         elif err < 0:
 *             lo = mid             # <<<<<<<<<<<<<<
 *     return [0], False
 * 
 */

#line 279 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_lo = __pyx_v_mid;

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":278
 *         elif err > 0:
 *             hi = mid
 *         elif err < 0:             # <<<<<<<<<<<<<<
 *             lo = mid
 *     return [0], False
 */

#line 278 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 278 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_L5:;

#line 278 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":280
 *         elif err < 0:
 *             lo = mid
 *     return [0], False             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 280 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 280 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 280, __pyx_L1_error)

#line 280 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_6);

#line 280 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_INCREF(__pyx_int_0);

#line 280 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GIVEREF(__pyx_int_0);

#line 280 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyList_SET_ITEM(__pyx_t_6, 0, __pyx_int_0);

#line 280 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 280, __pyx_L1_error)

#line 280 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_7);

#line 280 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GIVEREF(__pyx_t_6);

#line 280 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_6);

#line 280 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_INCREF(Py_False);

#line 280 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GIVEREF(Py_False);

#line 280 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyTuple_SET_ITEM(__pyx_t_7, 1, Py_False);

#line 280 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_6 = 0;

#line 280 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = ((PyObject*)__pyx_t_7);

#line 280 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_7 = 0;

#line 280 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":238
 * 
 * 
 * cpdef tuple binary_search_with_flag(list array, double mz, double error_tolerance):             # <<<<<<<<<<<<<<
 *     cdef:
 *         int i, n, lo, hi, mid, low_end, high_end
 */

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L1_error:;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_7);

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.binary_search_with_flag", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = 0;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_x);

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
/* Python wrapper */

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_3binary_search_with_flag(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_11feature_map_2binary_search_with_flag[] = "binary_search_with_flag(list array, double mz, double error_tolerance) -> tuple";

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_3binary_search_with_flag(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_array = 0
#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  double __pyx_v_mz
#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  double __pyx_v_error_tolerance
#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = 0;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("binary_search_with_flag (wrapper)", 0);

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  {

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_array,&__pyx_n_s_mz,&__pyx_n_s_error_tolerance,0};

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
    PyObject* values[3] = {0,0,0};

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (unlikely(__pyx_kwds)) {

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
      Py_ssize_t kw_args;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
      switch (pos_args) {
        case  3: 
#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
        case  0: break;
        default: 
#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
goto __pyx_L5_argtuple_error;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
      switch (pos_args) {

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
        case  0:

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_array)) != 0)) kw_args--;
        else 
#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
goto __pyx_L5_argtuple_error;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
        case  1:

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mz)) != 0)) kw_args--;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("binary_search_with_flag", 1, 3, 3, 1); 
#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
__PYX_ERR(0, 238, __pyx_L3_error)

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
        CYTHON_FALLTHROUGH;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
        case  2:

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_error_tolerance)) != 0)) kw_args--;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("binary_search_with_flag", 1, 3, 3, 2); 
#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
__PYX_ERR(0, 238, __pyx_L3_error)

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
      if (unlikely(kw_args > 0)) {

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "binary_search_with_flag") < 0)) __PYX_ERR(0, 238, __pyx_L3_error)

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
      goto __pyx_L5_argtuple_error;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
    } else {

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_array = ((PyObject*)values[0]);

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_mz = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_mz == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 238, __pyx_L3_error)

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_error_tolerance = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_error_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 238, __pyx_L3_error)

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("binary_search_with_flag", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
__PYX_ERR(0, 238, __pyx_L3_error)

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L3_error:;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.binary_search_with_flag", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return NULL;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_array), (&PyList_Type), 1, "array", 1))) __PYX_ERR(0, 238, __pyx_L1_error)
  __pyx_r = 
#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_2binary_search_with_flag(__pyx_self, __pyx_v_array, __pyx_v_mz, __pyx_v_error_tolerance);

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L0;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L1_error:;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = NULL;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_2binary_search_with_flag(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_array, double __pyx_v_mz, double __pyx_v_error_tolerance) {

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_r = NULL;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_lineno = 0;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  const char *__pyx_filename = NULL;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_clineno = 0;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("binary_search_with_flag", 0);

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_binary_search_with_flag(__pyx_v_array, __pyx_v_mz, __pyx_v_error_tolerance, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 238, __pyx_L1_error)

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = __pyx_t_1;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = 0;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L0;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L1_error:;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.binary_search_with_flag", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = NULL;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 238 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_map.pyx":284
 * 
 * @cython.cdivision(True)
 * cdef long binary_search(list array, double mz, double error_tolerance):             # <<<<<<<<<<<<<<
 *     cdef:
 *         int lo, hi, n, mid, i
 */

#line 284 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 284 "ms_deisotope/_c/feature_map/feature_map.pyx"
static long __pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_binary_search(PyObject *__pyx_v_array, double __pyx_v_mz, double __pyx_v_error_tolerance) {
  int __pyx_v_lo
#line 284 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  int __pyx_v_hi
#line 284 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  int __pyx_v_n
#line 284 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  int __pyx_v_mid
#line 284 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  int __pyx_v_i
#line 284 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  double __pyx_v_err
#line 284 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_x = 0
#line 284 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  int __pyx_v_best_index
#line 284 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  double __pyx_v_best_error
#line 284 "ms_deisotope/_c/feature_map/feature_map.pyx"
;

#line 284 "ms_deisotope/_c/feature_map/feature_map.pyx"
  long __pyx_r;

#line 284 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 284 "ms_deisotope/_c/feature_map/feature_map.pyx"
  Py_ssize_t __pyx_t_1;

#line 284 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_t_2;

#line 284 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_3;

#line 284 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 284 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("binary_search", 0);

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":289
 *         double err
 *         LCMSFeature x
 *     lo = 0             # <<<<<<<<<<<<<<
 *     n = hi = PyList_GET_SIZE(array)
 *     while hi != lo:
 */

#line 289 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_v_lo = 0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":290
 *         LCMSFeature x
 *     lo = 0
 *     n = hi = PyList_GET_SIZE(array)             # <<<<<<<<<<<<<<
 *     while hi != lo:
 *         mid = (hi + lo) / 2
 */

#line 290 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = PyList_GET_SIZE(__pyx_v_array);

#line 290 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_v_n = __pyx_t_1;

#line 290 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_v_hi = __pyx_t_1;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":291
 *     lo = 0
 *     n = hi = PyList_GET_SIZE(array)
 *     while hi != lo:             # <<<<<<<<<<<<<<
 *         mid = (hi + lo) / 2
 *         x = <LCMSFeature>PyList_GET_ITEM(array, mid)
 */

#line 291 "ms_deisotope/_c/feature_map/feature_map.pyx"
  while (1) {

#line 291 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_2 = ((__pyx_v_hi != __pyx_v_lo) != 0);

#line 291 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (!__pyx_t_2) break;

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":292
 *     n = hi = PyList_GET_SIZE(array)
 *     while hi != lo:
 *         mid = (hi + lo) / 2             # <<<<<<<<<<<<<<
 *         x = <LCMSFeature>PyList_GET_ITEM(array, mid)
 *         err = (x.get_mz() - mz) / mz
 */

#line 292 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_mid = ((__pyx_v_hi + __pyx_v_lo) / 2);

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":293
 *     while hi != lo:
 *         mid = (hi + lo) / 2
 *         x = <LCMSFeature>PyList_GET_ITEM(array, mid)             # <<<<<<<<<<<<<<
 *         err = (x.get_mz() - mz) / mz
 *         if abs(err) <= error_tolerance:
 */

#line 293 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_3 = PyList_GET_ITEM(__pyx_v_array, __pyx_v_mid);

#line 293 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_4 = ((PyObject *)__pyx_t_3);

#line 293 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_INCREF(__pyx_t_4);

#line 293 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_XDECREF_SET(__pyx_v_x, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_4));

#line 293 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":294
 *         mid = (hi + lo) / 2
 *         x = <LCMSFeature>PyList_GET_ITEM(array, mid)
 *         err = (x.get_mz() - mz) / mz             # <<<<<<<<<<<<<<
 *         if abs(err) <= error_tolerance:
 *             best_index = mid
 */

#line 294 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_err = ((((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_x->__pyx_base.__pyx_vtab)->__pyx_base.get_mz(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_x)) - __pyx_v_mz) / __pyx_v_mz);

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":295
 *         x = <LCMSFeature>PyList_GET_ITEM(array, mid)
 *         err = (x.get_mz() - mz) / mz
 *         if abs(err) <= error_tolerance:             # <<<<<<<<<<<<<<
 *             best_index = mid
 *             best_error = abs(err)
 */

#line 295 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_2 = ((fabs(__pyx_v_err) <= __pyx_v_error_tolerance) != 0);

#line 295 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":296
 *         err = (x.get_mz() - mz) / mz
 *         if abs(err) <= error_tolerance:
 *             best_index = mid             # <<<<<<<<<<<<<<
 *             best_error = abs(err)
 *             i = mid - 1
 */

#line 296 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_best_index = __pyx_v_mid;

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":297
 *         if abs(err) <= error_tolerance:
 *             best_index = mid
 *             best_error = abs(err)             # <<<<<<<<<<<<<<
 *             i = mid - 1
 *             while i > 0:
 */

#line 297 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_best_error = fabs(__pyx_v_err);

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":298
 *             best_index = mid
 *             best_error = abs(err)
 *             i = mid - 1             # <<<<<<<<<<<<<<
 *             while i > 0:
 *                 x = <LCMSFeature>PyList_GET_ITEM(array, i)
 */

#line 298 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_i = (__pyx_v_mid - 1);

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":299
 *             best_error = abs(err)
 *             i = mid - 1
 *             while i > 0:             # <<<<<<<<<<<<<<
 *                 x = <LCMSFeature>PyList_GET_ITEM(array, i)
 *                 err = abs((x.get_mz() - mz) / mz)
 */

#line 299 "ms_deisotope/_c/feature_map/feature_map.pyx"
      while (1) {

#line 299 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_2 = ((__pyx_v_i > 0) != 0);

#line 299 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (!__pyx_t_2) break;

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":300
 *             i = mid - 1
 *             while i > 0:
 *                 x = <LCMSFeature>PyList_GET_ITEM(array, i)             # <<<<<<<<<<<<<<
 *                 err = abs((x.get_mz() - mz) / mz)
 *                 if err < best_error:
 */

#line 300 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_3 = PyList_GET_ITEM(__pyx_v_array, __pyx_v_i);

#line 300 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_4 = ((PyObject *)__pyx_t_3);

#line 300 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_INCREF(__pyx_t_4);

#line 300 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_DECREF_SET(__pyx_v_x, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_4));

#line 300 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_4 = 0;

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":301
 *             while i > 0:
 *                 x = <LCMSFeature>PyList_GET_ITEM(array, i)
 *                 err = abs((x.get_mz() - mz) / mz)             # <<<<<<<<<<<<<<
 *                 if err < best_error:
 *                     best_error = err
 */

#line 301 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_v_err = fabs(((((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_x->__pyx_base.__pyx_vtab)->__pyx_base.get_mz(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_x)) - __pyx_v_mz) / __pyx_v_mz));

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":302
 *                 x = <LCMSFeature>PyList_GET_ITEM(array, i)
 *                 err = abs((x.get_mz() - mz) / mz)
 *                 if err < best_error:             # <<<<<<<<<<<<<<
 *                     best_error = err
 *                     best_index = i
 */

#line 302 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_2 = ((__pyx_v_err < __pyx_v_best_error) != 0);

#line 302 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (__pyx_t_2) {

          /* "ms_deisotope/_c/feature_map/feature_map.pyx":303
 *                 err = abs((x.get_mz() - mz) / mz)
 *                 if err < best_error:
 *                     best_error = err             # <<<<<<<<<<<<<<
 *                     best_index = i
 *                 i -= 1
 */

#line 303 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_v_best_error = __pyx_v_err;

          /* "ms_deisotope/_c/feature_map/feature_map.pyx":304
 *                 if err < best_error:
 *                     best_error = err
 *                     best_index = i             # <<<<<<<<<<<<<<
 *                 i -= 1
 * 
 */

#line 304 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_v_best_index = __pyx_v_i;

          /* "ms_deisotope/_c/feature_map/feature_map.pyx":302
 *                 x = <LCMSFeature>PyList_GET_ITEM(array, i)
 *                 err = abs((x.get_mz() - mz) / mz)
 *                 if err < best_error:             # <<<<<<<<<<<<<<
 *                     best_error = err
 *                     best_index = i
 */

#line 302 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":305
 *                     best_error = err
 *                     best_index = i
 *                 i -= 1             # <<<<<<<<<<<<<<
 * 
 *             i = mid + 1
 */

#line 305 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_v_i = (__pyx_v_i - 1);

#line 305 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":307
 *                 i -= 1
 * 
 *             i = mid + 1             # <<<<<<<<<<<<<<
 *             while i < n:
 *                 x = <LCMSFeature>PyList_GET_ITEM(array, i)
 */

#line 307 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_i = (__pyx_v_mid + 1);

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":308
 * 
 *             i = mid + 1
 *             while i < n:             # <<<<<<<<<<<<<<
 *                 x = <LCMSFeature>PyList_GET_ITEM(array, i)
 *                 err = abs((x.get_mz() - mz) / mz)
 */

#line 308 "ms_deisotope/_c/feature_map/feature_map.pyx"
      while (1) {

#line 308 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_2 = ((__pyx_v_i < __pyx_v_n) != 0);

#line 308 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (!__pyx_t_2) break;

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":309
 *             i = mid + 1
 *             while i < n:
 *                 x = <LCMSFeature>PyList_GET_ITEM(array, i)             # <<<<<<<<<<<<<<
 *                 err = abs((x.get_mz() - mz) / mz)
 *                 if err < best_error:
 */

#line 309 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_3 = PyList_GET_ITEM(__pyx_v_array, __pyx_v_i);

#line 309 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_4 = ((PyObject *)__pyx_t_3);

#line 309 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_INCREF(__pyx_t_4);

#line 309 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_DECREF_SET(__pyx_v_x, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_4));

#line 309 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_4 = 0;

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":310
 *             while i < n:
 *                 x = <LCMSFeature>PyList_GET_ITEM(array, i)
 *                 err = abs((x.get_mz() - mz) / mz)             # <<<<<<<<<<<<<<
 *                 if err < best_error:
 *                     best_error = err
 */

#line 310 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_v_err = fabs(((((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_x->__pyx_base.__pyx_vtab)->__pyx_base.get_mz(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_x)) - __pyx_v_mz) / __pyx_v_mz));

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":311
 *                 x = <LCMSFeature>PyList_GET_ITEM(array, i)
 *                 err = abs((x.get_mz() - mz) / mz)
 *                 if err < best_error:             # <<<<<<<<<<<<<<
 *                     best_error = err
 *                     best_index = i
 */

#line 311 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_2 = ((__pyx_v_err < __pyx_v_best_error) != 0);

#line 311 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (__pyx_t_2) {

          /* "ms_deisotope/_c/feature_map/feature_map.pyx":312
 *                 err = abs((x.get_mz() - mz) / mz)
 *                 if err < best_error:
 *                     best_error = err             # <<<<<<<<<<<<<<
 *                     best_index = i
 *                 i += 1
 */

#line 312 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_v_best_error = __pyx_v_err;

          /* "ms_deisotope/_c/feature_map/feature_map.pyx":313
 *                 if err < best_error:
 *                     best_error = err
 *                     best_index = i             # <<<<<<<<<<<<<<
 *                 i += 1
 *             return best_index
 */

#line 313 "ms_deisotope/_c/feature_map/feature_map.pyx"
          __pyx_v_best_index = __pyx_v_i;

          /* "ms_deisotope/_c/feature_map/feature_map.pyx":311
 *                 x = <LCMSFeature>PyList_GET_ITEM(array, i)
 *                 err = abs((x.get_mz() - mz) / mz)
 *                 if err < best_error:             # <<<<<<<<<<<<<<
 *                     best_error = err
 *                     best_index = i
 */

#line 311 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":314
 *                     best_error = err
 *                     best_index = i
 *                 i += 1             # <<<<<<<<<<<<<<
 *             return best_index
 *         elif (hi - lo) == 1:
 */

#line 314 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_v_i = (__pyx_v_i + 1);

#line 314 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":315
 *                     best_index = i
 *                 i += 1
 *             return best_index             # <<<<<<<<<<<<<<
 *         elif (hi - lo) == 1:
 *             return -1
 */

#line 315 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_r = __pyx_v_best_index;

#line 315 "ms_deisotope/_c/feature_map/feature_map.pyx"
      goto __pyx_L0;

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":295
 *         x = <LCMSFeature>PyList_GET_ITEM(array, mid)
 *         err = (x.get_mz() - mz) / mz
 *         if abs(err) <= error_tolerance:             # <<<<<<<<<<<<<<
 *             best_index = mid
 *             best_error = abs(err)
 */

#line 295 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":316
 *                 i += 1
 *             return best_index
 *         elif (hi - lo) == 1:             # <<<<<<<<<<<<<<
 *             return -1
 *         elif err > 0:
 */

#line 316 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_2 = (((__pyx_v_hi - __pyx_v_lo) == 1) != 0);

#line 316 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":317
 *             return best_index
 *         elif (hi - lo) == 1:
 *             return -1             # <<<<<<<<<<<<<<
 *         elif err > 0:
 *             hi = mid
 */

#line 317 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_r = -1L;

#line 317 "ms_deisotope/_c/feature_map/feature_map.pyx"
      goto __pyx_L0;

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":316
 *                 i += 1
 *             return best_index
 *         elif (hi - lo) == 1:             # <<<<<<<<<<<<<<
 *             return -1
 *         elif err > 0:
 */

#line 316 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":318
 *         elif (hi - lo) == 1:
 *             return -1
 *         elif err > 0:             # <<<<<<<<<<<<<<
 *             hi = mid
 *         elif err < 0:
 */

#line 318 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_2 = ((__pyx_v_err > 0.0) != 0);

#line 318 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":319
 *             return -1
 *         elif err > 0:
 *             hi = mid             # <<<<<<<<<<<<<<
 *         elif err < 0:
 *             lo = mid
 */

#line 319 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_hi = __pyx_v_mid;

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":318
 *         elif (hi - lo) == 1:
 *             return -1
 *         elif err > 0:             # <<<<<<<<<<<<<<
 *             hi = mid
 *         elif err < 0:
 */

#line 318 "ms_deisotope/_c/feature_map/feature_map.pyx"
      goto __pyx_L5;

#line 318 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":320
 *         elif err > 0:
 *             hi = mid
 *         elif err < 0:             # <<<<<<<<<<<<<<
 *             lo = mid
 *     return 0
 */

#line 320 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_2 = ((__pyx_v_err < 0.0) != 0);

#line 320 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":321
 *             hi = mid
 *         elif err < 0:
 *             lo = mid             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */

#line 321 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_lo = __pyx_v_mid;

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":320
 *         elif err > 0:
 *             hi = mid
 *         elif err < 0:             # <<<<<<<<<<<<<<
 *             lo = mid
 *     return 0
 */

#line 320 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 320 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_L5:;

#line 320 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":322
 *         elif err < 0:
 *             lo = mid
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 322 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_r = 0;

#line 322 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":284
 * 
 * @cython.cdivision(True)
 * cdef long binary_search(list array, double mz, double error_tolerance):             # <<<<<<<<<<<<<<
 *     cdef:
 *         int lo, hi, n, mid, i
 */

#line 284 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 284 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 284 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 284 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_x);

#line 284 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 284 "ms_deisotope/_c/feature_map/feature_map.pyx"
  return __pyx_r;

#line 284 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

/* "ms_deisotope/_c/feature_map/feature_map.pyx":326
 * 
 * @cython.cdivision(True)
 * cdef void search_sweep(list array, double mz, double error_tolerance, long* loout, long* hiout):             # <<<<<<<<<<<<<<
 *     cdef:
 *         int i, n, lo, hi, mid, start, end
 */

#line 326 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 326 "ms_deisotope/_c/feature_map/feature_map.pyx"
static void __pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_search_sweep(PyObject *__pyx_v_array, double __pyx_v_mz, double __pyx_v_error_tolerance, long *__pyx_v_loout, long *__pyx_v_hiout) {
  int __pyx_v_i
#line 326 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  int __pyx_v_n
#line 326 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  int __pyx_v_lo
#line 326 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  int __pyx_v_hi
#line 326 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  int __pyx_v_mid
#line 326 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  int __pyx_v_start
#line 326 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  int __pyx_v_end
#line 326 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  double __pyx_v_err
#line 326 "ms_deisotope/_c/feature_map/feature_map.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_x = 0
#line 326 "ms_deisotope/_c/feature_map/feature_map.pyx"
;

#line 326 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyDeclarations

#line 326 "ms_deisotope/_c/feature_map/feature_map.pyx"
  Py_ssize_t __pyx_t_1;

#line 326 "ms_deisotope/_c/feature_map/feature_map.pyx"
  int __pyx_t_2;

#line 326 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_3;

#line 326 "ms_deisotope/_c/feature_map/feature_map.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 326 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannySetupContext("search_sweep", 0);

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":331
 *         double err
 *         LCMSFeature x
 *     lo = 0             # <<<<<<<<<<<<<<
 *     n = hi = PyList_GET_SIZE(array)
 *     loout[0] = 0
 */

#line 331 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_v_lo = 0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":332
 *         LCMSFeature x
 *     lo = 0
 *     n = hi = PyList_GET_SIZE(array)             # <<<<<<<<<<<<<<
 *     loout[0] = 0
 *     hiout[0] = 0
 */

#line 332 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = PyList_GET_SIZE(__pyx_v_array);

#line 332 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_v_n = __pyx_t_1;

#line 332 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_v_hi = __pyx_t_1;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":333
 *     lo = 0
 *     n = hi = PyList_GET_SIZE(array)
 *     loout[0] = 0             # <<<<<<<<<<<<<<
 *     hiout[0] = 0
 *     while hi != lo:
 */

#line 333 "ms_deisotope/_c/feature_map/feature_map.pyx"
  (__pyx_v_loout[0]) = 0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":334
 *     n = hi = PyList_GET_SIZE(array)
 *     loout[0] = 0
 *     hiout[0] = 0             # <<<<<<<<<<<<<<
 *     while hi != lo:
 *         mid = (hi + lo) / 2
 */

#line 334 "ms_deisotope/_c/feature_map/feature_map.pyx"
  (__pyx_v_hiout[0]) = 0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":335
 *     loout[0] = 0
 *     hiout[0] = 0
 *     while hi != lo:             # <<<<<<<<<<<<<<
 *         mid = (hi + lo) / 2
 *         x = <LCMSFeature>PyList_GET_ITEM(array, mid)
 */

#line 335 "ms_deisotope/_c/feature_map/feature_map.pyx"
  while (1) {

#line 335 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_2 = ((__pyx_v_hi != __pyx_v_lo) != 0);

#line 335 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (!__pyx_t_2) break;

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":336
 *     hiout[0] = 0
 *     while hi != lo:
 *         mid = (hi + lo) / 2             # <<<<<<<<<<<<<<
 *         x = <LCMSFeature>PyList_GET_ITEM(array, mid)
 *         err = (x.get_mz() - mz) / mz
 */

#line 336 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_mid = ((__pyx_v_hi + __pyx_v_lo) / 2);

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":337
 *     while hi != lo:
 *         mid = (hi + lo) / 2
 *         x = <LCMSFeature>PyList_GET_ITEM(array, mid)             # <<<<<<<<<<<<<<
 *         err = (x.get_mz() - mz) / mz
 *         if abs(err) <= error_tolerance:
 */

#line 337 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_3 = PyList_GET_ITEM(__pyx_v_array, __pyx_v_mid);

#line 337 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_4 = ((PyObject *)__pyx_t_3);

#line 337 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_INCREF(__pyx_t_4);

#line 337 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __Pyx_XDECREF_SET(__pyx_v_x, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_4));

#line 337 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":338
 *         mid = (hi + lo) / 2
 *         x = <LCMSFeature>PyList_GET_ITEM(array, mid)
 *         err = (x.get_mz() - mz) / mz             # <<<<<<<<<<<<<<
 *         if abs(err) <= error_tolerance:
 *             i = mid - 1
 */

#line 338 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_v_err = ((((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_x->__pyx_base.__pyx_vtab)->__pyx_base.get_mz(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_x)) - __pyx_v_mz) / __pyx_v_mz);

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":339
 *         x = <LCMSFeature>PyList_GET_ITEM(array, mid)
 *         err = (x.get_mz() - mz) / mz
 *         if abs(err) <= error_tolerance:             # <<<<<<<<<<<<<<
 *             i = mid - 1
 *             while i > 0:
 */

#line 339 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_2 = ((fabs(__pyx_v_err) <= __pyx_v_error_tolerance) != 0);

#line 339 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":340
 *         err = (x.get_mz() - mz) / mz
 *         if abs(err) <= error_tolerance:
 *             i = mid - 1             # <<<<<<<<<<<<<<
 *             while i > 0:
 *                 x = <LCMSFeature>PyList_GET_ITEM(array, i)
 */

#line 340 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_i = (__pyx_v_mid - 1);

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":341
 *         if abs(err) <= error_tolerance:
 *             i = mid - 1
 *             while i > 0:             # <<<<<<<<<<<<<<
 *                 x = <LCMSFeature>PyList_GET_ITEM(array, i)
 *                 err = abs((x.get_mz() - mz) / mz)
 */

#line 341 "ms_deisotope/_c/feature_map/feature_map.pyx"
      while (1) {

#line 341 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_2 = ((__pyx_v_i > 0) != 0);

#line 341 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (!__pyx_t_2) break;

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":342
 *             i = mid - 1
 *             while i > 0:
 *                 x = <LCMSFeature>PyList_GET_ITEM(array, i)             # <<<<<<<<<<<<<<
 *                 err = abs((x.get_mz() - mz) / mz)
 *                 if err > error_tolerance:
 */

#line 342 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_3 = PyList_GET_ITEM(__pyx_v_array, __pyx_v_i);

#line 342 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_4 = ((PyObject *)__pyx_t_3);

#line 342 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_INCREF(__pyx_t_4);

#line 342 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_DECREF_SET(__pyx_v_x, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_4));

#line 342 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_4 = 0;

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":343
 *             while i > 0:
 *                 x = <LCMSFeature>PyList_GET_ITEM(array, i)
 *                 err = abs((x.get_mz() - mz) / mz)             # <<<<<<<<<<<<<<
 *                 if err > error_tolerance:
 *                     break
 */

#line 343 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_v_err = fabs(((((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_x->__pyx_base.__pyx_vtab)->__pyx_base.get_mz(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_x)) - __pyx_v_mz) / __pyx_v_mz));

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":344
 *                 x = <LCMSFeature>PyList_GET_ITEM(array, i)
 *                 err = abs((x.get_mz() - mz) / mz)
 *                 if err > error_tolerance:             # <<<<<<<<<<<<<<
 *                     break
 *                 i -= 1
 */

#line 344 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_2 = ((__pyx_v_err > __pyx_v_error_tolerance) != 0);

#line 344 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (__pyx_t_2) {

          /* "ms_deisotope/_c/feature_map/feature_map.pyx":345
 *                 err = abs((x.get_mz() - mz) / mz)
 *                 if err > error_tolerance:
 *                     break             # <<<<<<<<<<<<<<
 *                 i -= 1
 *             start = i + 1
 */

#line 345 "ms_deisotope/_c/feature_map/feature_map.pyx"
          goto __pyx_L7_break;

          /* "ms_deisotope/_c/feature_map/feature_map.pyx":344
 *                 x = <LCMSFeature>PyList_GET_ITEM(array, i)
 *                 err = abs((x.get_mz() - mz) / mz)
 *                 if err > error_tolerance:             # <<<<<<<<<<<<<<
 *                     break
 *                 i -= 1
 */

#line 344 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":346
 *                 if err > error_tolerance:
 *                     break
 *                 i -= 1             # <<<<<<<<<<<<<<
 *             start = i + 1
 *             i = mid + 1
 */

#line 346 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_v_i = (__pyx_v_i - 1);

#line 346 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 346 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_L7_break:;

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":347
 *                     break
 *                 i -= 1
 *             start = i + 1             # <<<<<<<<<<<<<<
 *             i = mid + 1
 *             while i < n:
 */

#line 347 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_start = (__pyx_v_i + 1);

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":348
 *                 i -= 1
 *             start = i + 1
 *             i = mid + 1             # <<<<<<<<<<<<<<
 *             while i < n:
 *                 x = <LCMSFeature>PyList_GET_ITEM(array, i)
 */

#line 348 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_i = (__pyx_v_mid + 1);

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":349
 *             start = i + 1
 *             i = mid + 1
 *             while i < n:             # <<<<<<<<<<<<<<
 *                 x = <LCMSFeature>PyList_GET_ITEM(array, i)
 *                 err = abs((x.get_mz() - mz) / mz)
 */

#line 349 "ms_deisotope/_c/feature_map/feature_map.pyx"
      while (1) {

#line 349 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_2 = ((__pyx_v_i < __pyx_v_n) != 0);

#line 349 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (!__pyx_t_2) break;

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":350
 *             i = mid + 1
 *             while i < n:
 *                 x = <LCMSFeature>PyList_GET_ITEM(array, i)             # <<<<<<<<<<<<<<
 *                 err = abs((x.get_mz() - mz) / mz)
 *                 if err > error_tolerance:
 */

#line 350 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_3 = PyList_GET_ITEM(__pyx_v_array, __pyx_v_i);

#line 350 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_4 = ((PyObject *)__pyx_t_3);

#line 350 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_INCREF(__pyx_t_4);

#line 350 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __Pyx_DECREF_SET(__pyx_v_x, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_4));

#line 350 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_4 = 0;

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":351
 *             while i < n:
 *                 x = <LCMSFeature>PyList_GET_ITEM(array, i)
 *                 err = abs((x.get_mz() - mz) / mz)             # <<<<<<<<<<<<<<
 *                 if err > error_tolerance:
 *                     break
 */

#line 351 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_v_err = fabs(((((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_x->__pyx_base.__pyx_vtab)->__pyx_base.get_mz(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_x)) - __pyx_v_mz) / __pyx_v_mz));

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":352
 *                 x = <LCMSFeature>PyList_GET_ITEM(array, i)
 *                 err = abs((x.get_mz() - mz) / mz)
 *                 if err > error_tolerance:             # <<<<<<<<<<<<<<
 *                     break
 *                 i += 1
 */

#line 352 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_t_2 = ((__pyx_v_err > __pyx_v_error_tolerance) != 0);

#line 352 "ms_deisotope/_c/feature_map/feature_map.pyx"
        if (__pyx_t_2) {

          /* "ms_deisotope/_c/feature_map/feature_map.pyx":353
 *                 err = abs((x.get_mz() - mz) / mz)
 *                 if err > error_tolerance:
 *                     break             # <<<<<<<<<<<<<<
 *                 i += 1
 *             end = i
 */

#line 353 "ms_deisotope/_c/feature_map/feature_map.pyx"
          goto __pyx_L10_break;

          /* "ms_deisotope/_c/feature_map/feature_map.pyx":352
 *                 x = <LCMSFeature>PyList_GET_ITEM(array, i)
 *                 err = abs((x.get_mz() - mz) / mz)
 *                 if err > error_tolerance:             # <<<<<<<<<<<<<<
 *                     break
 *                 i += 1
 */

#line 352 "ms_deisotope/_c/feature_map/feature_map.pyx"
        }

        /* "ms_deisotope/_c/feature_map/feature_map.pyx":354
 *                 if err > error_tolerance:
 *                     break
 *                 i += 1             # <<<<<<<<<<<<<<
 *             end = i
 *             loout[0] = start
 */

#line 354 "ms_deisotope/_c/feature_map/feature_map.pyx"
        __pyx_v_i = (__pyx_v_i + 1);

#line 354 "ms_deisotope/_c/feature_map/feature_map.pyx"
      }

#line 354 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_L10_break:;

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":355
 *                     break
 *                 i += 1
 *             end = i             # <<<<<<<<<<<<<<
 *             loout[0] = start
 *             hiout[0] = end
 */

#line 355 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_end = __pyx_v_i;

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":356
 *                 i += 1
 *             end = i
 *             loout[0] = start             # <<<<<<<<<<<<<<
 *             hiout[0] = end
 *             return
 */

#line 356 "ms_deisotope/_c/feature_map/feature_map.pyx"
      (__pyx_v_loout[0]) = __pyx_v_start;

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":357
 *             end = i
 *             loout[0] = start
 *             hiout[0] = end             # <<<<<<<<<<<<<<
 *             return
 *         elif (hi - lo) == 1:
 */

#line 357 "ms_deisotope/_c/feature_map/feature_map.pyx"
      (__pyx_v_hiout[0]) = __pyx_v_end;

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":358
 *             loout[0] = start
 *             hiout[0] = end
 *             return             # <<<<<<<<<<<<<<
 *         elif (hi - lo) == 1:
 *             return
 */

#line 358 "ms_deisotope/_c/feature_map/feature_map.pyx"
      goto __pyx_L0;

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":339
 *         x = <LCMSFeature>PyList_GET_ITEM(array, mid)
 *         err = (x.get_mz() - mz) / mz
 *         if abs(err) <= error_tolerance:             # <<<<<<<<<<<<<<
 *             i = mid - 1
 *             while i > 0:
 */

#line 339 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":359
 *             hiout[0] = end
 *             return
 *         elif (hi - lo) == 1:             # <<<<<<<<<<<<<<
 *             return
 *         elif err > 0:
 */

#line 359 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_2 = (((__pyx_v_hi - __pyx_v_lo) == 1) != 0);

#line 359 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":360
 *             return
 *         elif (hi - lo) == 1:
 *             return             # <<<<<<<<<<<<<<
 *         elif err > 0:
 *             hi = mid
 */

#line 360 "ms_deisotope/_c/feature_map/feature_map.pyx"
      goto __pyx_L0;

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":359
 *             hiout[0] = end
 *             return
 *         elif (hi - lo) == 1:             # <<<<<<<<<<<<<<
 *             return
 *         elif err > 0:
 */

#line 359 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":361
 *         elif (hi - lo) == 1:
 *             return
 *         elif err > 0:             # <<<<<<<<<<<<<<
 *             hi = mid
 *         elif err < 0:
 */

#line 361 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_2 = ((__pyx_v_err > 0.0) != 0);

#line 361 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":362
 *             return
 *         elif err > 0:
 *             hi = mid             # <<<<<<<<<<<<<<
 *         elif err < 0:
 *             lo = mid
 */

#line 362 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_hi = __pyx_v_mid;

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":361
 *         elif (hi - lo) == 1:
 *             return
 *         elif err > 0:             # <<<<<<<<<<<<<<
 *             hi = mid
 *         elif err < 0:
 */

#line 361 "ms_deisotope/_c/feature_map/feature_map.pyx"
      goto __pyx_L5;

#line 361 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

    /* "ms_deisotope/_c/feature_map/feature_map.pyx":363
 *         elif err > 0:
 *             hi = mid
 *         elif err < 0:             # <<<<<<<<<<<<<<
 *             lo = mid
 *     return
 */

#line 363 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_t_2 = ((__pyx_v_err < 0.0) != 0);

#line 363 "ms_deisotope/_c/feature_map/feature_map.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":364
 *             hi = mid
 *         elif err < 0:
 *             lo = mid             # <<<<<<<<<<<<<<
 *     return
 * 
 */

#line 364 "ms_deisotope/_c/feature_map/feature_map.pyx"
      __pyx_v_lo = __pyx_v_mid;

      /* "ms_deisotope/_c/feature_map/feature_map.pyx":363
 *         elif err > 0:
 *             hi = mid
 *         elif err < 0:             # <<<<<<<<<<<<<<
 *             lo = mid
 *     return
 */

#line 363 "ms_deisotope/_c/feature_map/feature_map.pyx"
    }

#line 363 "ms_deisotope/_c/feature_map/feature_map.pyx"
    __pyx_L5:;

#line 363 "ms_deisotope/_c/feature_map/feature_map.pyx"
  }

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":365
 *         elif err < 0:
 *             lo = mid
 *     return             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 365 "ms_deisotope/_c/feature_map/feature_map.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":326
 * 
 * @cython.cdivision(True)
 * cdef void search_sweep(list array, double mz, double error_tolerance, long* loout, long* hiout):             # <<<<<<<<<<<<<<
 *     cdef:
 *         int i, n, lo, hi, mid, start, end
 */

#line 326 "ms_deisotope/_c/feature_map/feature_map.pyx"


#line 326 "ms_deisotope/_c/feature_map/feature_map.pyx"
  /* function exit code */

#line 326 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_L0:;

#line 326 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_x);

#line 326 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_RefNannyFinishContext();

#line 326 "ms_deisotope/_c/feature_map/feature_map.pyx"
}

/* "(tree fragment)":1
 * def __pyx_unpickle_LCMSFeatureMap(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_5__pyx_unpickle_LCMSFeatureMap(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_11feature_map_4__pyx_unpickle_LCMSFeatureMap[] = "__pyx_unpickle_LCMSFeatureMap(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_11feature_map_11feature_map_5__pyx_unpickle_LCMSFeatureMap = 
#line 1 "(tree fragment)"
{"__pyx_unpickle_LCMSFeatureMap", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_5__pyx_unpickle_LCMSFeatureMap, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_11feature_map_4__pyx_unpickle_LCMSFeatureMap};

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_5__pyx_unpickle_LCMSFeatureMap(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0
#line 1 "(tree fragment)"
;
  long __pyx_v___pyx_checksum
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_state = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_LCMSFeatureMap (wrapper)", 0);

#line 1 "(tree fragment)"
  {

#line 1 "(tree fragment)"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};

#line 1 "(tree fragment)"
    PyObject* values[3] = {0,0,0};

#line 1 "(tree fragment)"
    if (unlikely(__pyx_kwds)) {

#line 1 "(tree fragment)"
      Py_ssize_t kw_args;

#line 1 "(tree fragment)"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 1 "(tree fragment)"
      switch (pos_args) {
        case  3: 
#line 1 "(tree fragment)"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 1 "(tree fragment)"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 1 "(tree fragment)"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  0: break;
        default: 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      kw_args = PyDict_Size(__pyx_kwds);

#line 1 "(tree fragment)"
      switch (pos_args) {

#line 1 "(tree fragment)"
        case  0:

#line 1 "(tree fragment)"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  1:

#line 1 "(tree fragment)"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_LCMSFeatureMap", 1, 3, 3, 1); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  2:

#line 1 "(tree fragment)"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_LCMSFeatureMap", 1, 3, 3, 2); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      if (unlikely(kw_args > 0)) {

#line 1 "(tree fragment)"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_LCMSFeatureMap") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 1 "(tree fragment)"
      goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
    } else {

#line 1 "(tree fragment)"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
    }

#line 1 "(tree fragment)"
    __pyx_v___pyx_type = values[0];

#line 1 "(tree fragment)"
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
    __pyx_v___pyx_state = values[2];

#line 1 "(tree fragment)"
  }

#line 1 "(tree fragment)"
  goto __pyx_L4_argument_unpacking_done;

#line 1 "(tree fragment)"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_LCMSFeatureMap", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 1 "(tree fragment)"
__PYX_ERR(2, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
  __pyx_L3_error:;

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.__pyx_unpickle_LCMSFeatureMap", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return NULL;

#line 1 "(tree fragment)"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_4__pyx_unpickle_LCMSFeatureMap(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_11feature_map_4__pyx_unpickle_LCMSFeatureMap(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_result = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_2;

#line 1 "(tree fragment)"
  int __pyx_t_3;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_LCMSFeatureMap", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9f034b6, 0x5b8a8b5, 0x4f94454):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x9f034b6, 0x5b8a8b5, 0x4f94454) = (features))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__3, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 4 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9f034b6, 0x5b8a8b5, 0x4f94454):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x9f034b6, 0x5b8a8b5, 0x4f94454) = (features))" % __pyx_checksum)
 *     __pyx_result = LCMSFeatureMap.__new__(__pyx_type)
 */

#line 5 "(tree fragment)"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __pyx_v___pyx_PickleError = __pyx_t_1;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x9f034b6, 0x5b8a8b5, 0x4f94454):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x9f034b6, 0x5b8a8b5, 0x4f94454) = (features))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = LCMSFeatureMap.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */

#line 6 "(tree fragment)"
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 6 "(tree fragment)"
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; 
#line 6 "(tree fragment)"
__pyx_t_6 = NULL;

#line 6 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {

#line 6 "(tree fragment)"
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);

#line 6 "(tree fragment)"
      if (likely(__pyx_t_6)) {

#line 6 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 6 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_6);

#line 6 "(tree fragment)"
        __Pyx_INCREF(function);

#line 6 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_1, function);

#line 6 "(tree fragment)"
      }

#line 6 "(tree fragment)"
    }

#line 6 "(tree fragment)"
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 6 "(tree fragment)"
    if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 6 "(tree fragment)"
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9f034b6, 0x5b8a8b5, 0x4f94454):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x9f034b6, 0x5b8a8b5, 0x4f94454) = (features))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x9f034b6, 0x5b8a8b5, 0x4f94454) = (features))" % __pyx_checksum)
 *     __pyx_result = LCMSFeatureMap.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_LCMSFeatureMap__set_state(<LCMSFeatureMap> __pyx_result, __pyx_state)
 */

#line 7 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 7 "(tree fragment)"
  __pyx_t_5 = NULL;

#line 7 "(tree fragment)"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {

#line 7 "(tree fragment)"
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);

#line 7 "(tree fragment)"
    if (likely(__pyx_t_5)) {

#line 7 "(tree fragment)"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 7 "(tree fragment)"
      __Pyx_INCREF(__pyx_t_5);

#line 7 "(tree fragment)"
      __Pyx_INCREF(function);

#line 7 "(tree fragment)"
      __Pyx_DECREF_SET(__pyx_t_1, function);

#line 7 "(tree fragment)"
    }

#line 7 "(tree fragment)"
  }

#line 7 "(tree fragment)"
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);

#line 7 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 7 "(tree fragment)"
  if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 7 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 7 "(tree fragment)"
  __pyx_v___pyx_result = __pyx_t_4;

#line 7 "(tree fragment)"
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x9f034b6, 0x5b8a8b5, 0x4f94454) = (features))" % __pyx_checksum)
 *     __pyx_result = LCMSFeatureMap.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_LCMSFeatureMap__set_state(<LCMSFeatureMap> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);

#line 8 "(tree fragment)"
  __pyx_t_2 = (__pyx_t_3 != 0);

#line 8 "(tree fragment)"
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = LCMSFeatureMap.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_LCMSFeatureMap__set_state(<LCMSFeatureMap> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_LCMSFeatureMap__set_state(LCMSFeatureMap __pyx_result, tuple __pyx_state):
 */

#line 9 "(tree fragment)"
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __pyx_t_4 = __pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map___pyx_unpickle_LCMSFeatureMap__set_state(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 9 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x9f034b6, 0x5b8a8b5, 0x4f94454) = (features))" % __pyx_checksum)
 *     __pyx_result = LCMSFeatureMap.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_LCMSFeatureMap__set_state(<LCMSFeatureMap> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_LCMSFeatureMap__set_state(<LCMSFeatureMap> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_LCMSFeatureMap__set_state(LCMSFeatureMap __pyx_result, tuple __pyx_state):
 *     __pyx_result.features = __pyx_state[0]
 */

#line 10 "(tree fragment)"
  __Pyx_XDECREF(__pyx_r);

#line 10 "(tree fragment)"
  __Pyx_INCREF(__pyx_v___pyx_result);

#line 10 "(tree fragment)"
  __pyx_r = __pyx_v___pyx_result;

#line 10 "(tree fragment)"
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_LCMSFeatureMap(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.__pyx_unpickle_LCMSFeatureMap", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_result);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":11
 *         __pyx_unpickle_LCMSFeatureMap__set_state(<LCMSFeatureMap> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_LCMSFeatureMap__set_state(LCMSFeatureMap __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.features = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map___pyx_unpickle_LCMSFeatureMap__set_state(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {

#line 11 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 11 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 11 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 11 "(tree fragment)"
  int __pyx_t_2;

#line 11 "(tree fragment)"
  Py_ssize_t __pyx_t_3;

#line 11 "(tree fragment)"
  int __pyx_t_4;

#line 11 "(tree fragment)"
  int __pyx_t_5;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_7 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_8 = NULL;

#line 11 "(tree fragment)"
  int __pyx_lineno = 0;

#line 11 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 11 "(tree fragment)"
  int __pyx_clineno = 0;

#line 11 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_LCMSFeatureMap__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_LCMSFeatureMap__set_state(LCMSFeatureMap __pyx_result, tuple __pyx_state):
 *     __pyx_result.features = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->features);

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_v___pyx_result->features);

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->features = ((PyObject*)__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_LCMSFeatureMap__set_state(LCMSFeatureMap __pyx_result, tuple __pyx_state):
 *     __pyx_result.features = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */

#line 13 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 13 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 13 "(tree fragment)"
    __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_4 = ((__pyx_t_3 > 1) != 0);

#line 13 "(tree fragment)"
  if (__pyx_t_4) {

#line 13 "(tree fragment)"
  } else {

#line 13 "(tree fragment)"
    __pyx_t_2 = __pyx_t_4;

#line 13 "(tree fragment)"
    goto __pyx_L4_bool_binop_done;

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_5 = (__pyx_t_4 != 0);

#line 13 "(tree fragment)"
  __pyx_t_2 = __pyx_t_5;

#line 13 "(tree fragment)"
  __pyx_L4_bool_binop_done:;

#line 13 "(tree fragment)"
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.features = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */

#line 14 "(tree fragment)"
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_6);

#line 14 "(tree fragment)"
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_7);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 14 "(tree fragment)"
    if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 14 "(tree fragment)"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 14 "(tree fragment)"
      __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_6);

#line 14 "(tree fragment)"
    __pyx_t_8 = NULL;

#line 14 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {

#line 14 "(tree fragment)"
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);

#line 14 "(tree fragment)"
      if (likely(__pyx_t_8)) {

#line 14 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);

#line 14 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_8);

#line 14 "(tree fragment)"
        __Pyx_INCREF(function);

#line 14 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_7, function);

#line 14 "(tree fragment)"
      }

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);

#line 14 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 14 "(tree fragment)"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_LCMSFeatureMap__set_state(LCMSFeatureMap __pyx_result, tuple __pyx_state):
 *     __pyx_result.features = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */

#line 13 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_LCMSFeatureMap__set_state(<LCMSFeatureMap> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_LCMSFeatureMap__set_state(LCMSFeatureMap __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.features = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
  /* function exit code */

#line 11 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 11 "(tree fragment)"
  goto __pyx_L0;

#line 11 "(tree fragment)"
  __pyx_L1_error:;

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_7);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_8);

#line 11 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.feature_map.__pyx_unpickle_LCMSFeatureMap__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 11 "(tree fragment)"
  __pyx_r = 0;

#line 11 "(tree fragment)"
  __pyx_L0:;

#line 11 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 11 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 11 "(tree fragment)"
  return __pyx_r;

#line 11 "(tree fragment)"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":734
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":735
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 735, __pyx_L1_error)

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":734
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":737
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":738
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 738, __pyx_L1_error)

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":737
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":740
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":741
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 741, __pyx_L1_error)

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":740
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":743
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":744
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 744, __pyx_L1_error)

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":743
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":746
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":747
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 747, __pyx_L1_error)

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":746
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":749
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_1;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":750
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */

#line 750 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);

#line 750 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (__pyx_t_1) {

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":751
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */

#line 751 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_r);

#line 751 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));

#line 751 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);

#line 751 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L0;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":750
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */

#line 750 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":753
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /*else*/ {

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_r);

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_INCREF(__pyx_empty_tuple);

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_r = __pyx_empty_tuple;

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L0;

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":749
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":928
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":929
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */

#line 929 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_INCREF(__pyx_v_base);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":930
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */

#line 930 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":928
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":932
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base
#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_1;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":933
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */

#line 933 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":934
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */

#line 934 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);

#line 934 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (__pyx_t_1) {

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":935
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */

#line 935 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_r);

#line 935 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 935 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L0;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":934
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */

#line 934 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":936
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */

#line 936 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 936 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_v_base));

#line 936 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = ((PyObject *)__pyx_v_base);

#line 936 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":932
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":940
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_2 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_3 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_4;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_5 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_6 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_7 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_8 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":941
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_declare

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_assign

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_1);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_2);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_3);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    /*try:*/ {

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":942
 * cdef inline int import_array() except -1:
 *     try:
 *         __pyx_import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */

#line 942 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(3, 942, __pyx_L3_error)

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":941
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L8_try_end;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L3_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":943
 *     try:
 *         __pyx_import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (__pyx_t_4) {

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(3, 943, __pyx_L5_except_error)

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_5);

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_6);

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":944
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 944, __pyx_L5_except_error)

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_8);

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __PYX_ERR(3, 944, __pyx_L5_except_error)

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L5_except_error;

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L5_except_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":941
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_1);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L1_error;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L8_try_end:;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":940
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_5);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_6);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_7);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_8);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = -1;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":946
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_2 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_3 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_4;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_5 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_6 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_7 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_8 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":947
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_declare

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_assign

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_1);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_2);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_3);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    /*try:*/ {

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":948
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */

#line 948 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(3, 948, __pyx_L3_error)

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":947
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L8_try_end;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L3_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":949
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (__pyx_t_4) {

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(3, 949, __pyx_L5_except_error)

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_5);

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_6);

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":950
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 950, __pyx_L5_except_error)

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_8);

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __PYX_ERR(3, 950, __pyx_L5_except_error)

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L5_except_error;

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L5_except_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":947
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_1);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L1_error;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L8_try_end:;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":946
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_5);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_6);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_7);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_8);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = -1;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":952
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_2 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_3 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_4;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_5 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_6 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_7 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_8 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":953
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_declare

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_assign

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_1);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_2);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_3);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    /*try:*/ {

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":954
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */

#line 954 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(3, 954, __pyx_L3_error)

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":953
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L8_try_end;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L3_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":955
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (__pyx_t_4) {

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(3, 955, __pyx_L5_except_error)

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_5);

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_6);

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":956
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef extern from *:
 */

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 956, __pyx_L5_except_error)

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_8);

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __PYX_ERR(3, 956, __pyx_L5_except_error)

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L5_except_error;

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L5_except_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":953
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_1);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L1_error;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L8_try_end:;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":952
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_5);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_6);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_7);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_8);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = -1;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":966
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_is_timedelta64_object(PyObject *__pyx_v_obj) {

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("is_timedelta64_object", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":978
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyTimedeltaArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 978 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyTimedeltaArrType_Type));

#line 978 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":966
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":981
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_is_datetime64_object(PyObject *__pyx_v_obj) {

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("is_datetime64_object", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":993
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyDatetimeArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 993 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyDatetimeArrType_Type));

#line 993 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":981
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":996
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE npy_datetime __pyx_f_5numpy_get_datetime64_value(PyObject *__pyx_v_obj) {

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  npy_datetime __pyx_r;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1003
 *     also needed.  That can be found using `get_datetime64_unit`.
 *     """
 *     return (<PyDatetimeScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1003 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = ((PyDatetimeScalarObject *)__pyx_v_obj)->obval;

#line 1003 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":996
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1006
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE npy_timedelta __pyx_f_5numpy_get_timedelta64_value(PyObject *__pyx_v_obj) {

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  npy_timedelta __pyx_r;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1010
 *     returns the int64 value underlying scalar numpy timedelta64 object
 *     """
 *     return (<PyTimedeltaScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1010 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = ((PyTimedeltaScalarObject *)__pyx_v_obj)->obval;

#line 1010 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1006
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1013
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE NPY_DATETIMEUNIT __pyx_f_5numpy_get_datetime64_unit(PyObject *__pyx_v_obj) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  NPY_DATETIMEUNIT __pyx_r;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1017
 *     returns the unit part of the dtype for a numpy datetime64 object.
 *     """
 *     return <NPY_DATETIMEUNIT>(<PyDatetimeScalarObject*>obj).obmeta.base             # <<<<<<<<<<<<<<
 */

#line 1017 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = ((NPY_DATETIMEUNIT)((PyDatetimeScalarObject *)__pyx_v_obj)->obmeta.base);

#line 1017 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1013
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "array.pxd":93
 *             __data_union data
 * 
 *         def __getbuffer__(self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

#line 93 "array.pxd"


#line 93 "array.pxd"
/* Python wrapper */

#line 93 "array.pxd"
static CYTHON_UNUSED int __pyx_pw_7cpython_5array_5array_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/

#line 93 "array.pxd"
static CYTHON_UNUSED int __pyx_pw_7cpython_5array_5array_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {

#line 93 "array.pxd"
  int __pyx_r;

#line 93 "array.pxd"
  __Pyx_RefNannyDeclarations

#line 93 "array.pxd"
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = 
#line 93 "array.pxd"
__pyx_pf_7cpython_5array_5array___getbuffer__(((arrayobject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

#line 93 "array.pxd"


#line 93 "array.pxd"
  /* function exit code */

#line 93 "array.pxd"
  __Pyx_RefNannyFinishContext();

#line 93 "array.pxd"
  return __pyx_r;

#line 93 "array.pxd"
}

#line 93 "array.pxd"


#line 93 "array.pxd"
static int __pyx_pf_7cpython_5array_5array___getbuffer__(arrayobject *__pyx_v_self, Py_buffer *__pyx_v_info, CYTHON_UNUSED int __pyx_v_flags) {
  PyObject *__pyx_v_item_count = NULL
#line 93 "array.pxd"
;

#line 93 "array.pxd"
  int __pyx_r;

#line 93 "array.pxd"
  __Pyx_RefNannyDeclarations

#line 93 "array.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 93 "array.pxd"
  char *__pyx_t_2;

#line 93 "array.pxd"
  int __pyx_t_3;

#line 93 "array.pxd"
  PyObject *__pyx_t_4 = NULL;

#line 93 "array.pxd"
  Py_ssize_t __pyx_t_5;

#line 93 "array.pxd"
  int __pyx_t_6;

#line 93 "array.pxd"
  char __pyx_t_7;

#line 93 "array.pxd"
  int __pyx_lineno = 0;

#line 93 "array.pxd"
  const char *__pyx_filename = NULL;

#line 93 "array.pxd"
  int __pyx_clineno = 0;

#line 93 "array.pxd"
  if (__pyx_v_info == NULL) {

#line 93 "array.pxd"
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");

#line 93 "array.pxd"
    return -1;

#line 93 "array.pxd"
  }

#line 93 "array.pxd"
  __Pyx_RefNannySetupContext("__getbuffer__", 0);

#line 93 "array.pxd"
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);

#line 93 "array.pxd"
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "array.pxd":98
 *             # In particular strided access is always provided regardless
 *             # of flags
 *             item_count = Py_SIZE(self)             # <<<<<<<<<<<<<<
 * 
 *             info.suboffsets = NULL
 */

#line 98 "array.pxd"
  __pyx_t_1 = PyInt_FromSsize_t(Py_SIZE(((PyObject *)__pyx_v_self))); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 98, __pyx_L1_error)

#line 98 "array.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 98 "array.pxd"
  __pyx_v_item_count = __pyx_t_1;

#line 98 "array.pxd"
  __pyx_t_1 = 0;

  /* "array.pxd":100
 *             item_count = Py_SIZE(self)
 * 
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *             info.buf = self.data.as_chars
 *             info.readonly = 0
 */

#line 100 "array.pxd"
  __pyx_v_info->suboffsets = NULL;

  /* "array.pxd":101
 * 
 *             info.suboffsets = NULL
 *             info.buf = self.data.as_chars             # <<<<<<<<<<<<<<
 *             info.readonly = 0
 *             info.ndim = 1
 */

#line 101 "array.pxd"
  __pyx_t_2 = __pyx_v_self->data.as_chars;

#line 101 "array.pxd"
  __pyx_v_info->buf = __pyx_t_2;

  /* "array.pxd":102
 *             info.suboffsets = NULL
 *             info.buf = self.data.as_chars
 *             info.readonly = 0             # <<<<<<<<<<<<<<
 *             info.ndim = 1
 *             info.itemsize = self.ob_descr.itemsize   # e.g. sizeof(float)
 */

#line 102 "array.pxd"
  __pyx_v_info->readonly = 0;

  /* "array.pxd":103
 *             info.buf = self.data.as_chars
 *             info.readonly = 0
 *             info.ndim = 1             # <<<<<<<<<<<<<<
 *             info.itemsize = self.ob_descr.itemsize   # e.g. sizeof(float)
 *             info.len = info.itemsize * item_count
 */

#line 103 "array.pxd"
  __pyx_v_info->ndim = 1;

  /* "array.pxd":104
 *             info.readonly = 0
 *             info.ndim = 1
 *             info.itemsize = self.ob_descr.itemsize   # e.g. sizeof(float)             # <<<<<<<<<<<<<<
 *             info.len = info.itemsize * item_count
 * 
 */

#line 104 "array.pxd"
  __pyx_t_3 = __pyx_v_self->ob_descr->itemsize;

#line 104 "array.pxd"
  __pyx_v_info->itemsize = __pyx_t_3;

  /* "array.pxd":105
 *             info.ndim = 1
 *             info.itemsize = self.ob_descr.itemsize   # e.g. sizeof(float)
 *             info.len = info.itemsize * item_count             # <<<<<<<<<<<<<<
 * 
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)
 */

#line 105 "array.pxd"
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_info->itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 105, __pyx_L1_error)

#line 105 "array.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 105 "array.pxd"
  __pyx_t_4 = PyNumber_Multiply(__pyx_t_1, __pyx_v_item_count); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 105, __pyx_L1_error)

#line 105 "array.pxd"
  __Pyx_GOTREF(__pyx_t_4);

#line 105 "array.pxd"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 105 "array.pxd"
  __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_4); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(4, 105, __pyx_L1_error)

#line 105 "array.pxd"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 105 "array.pxd"
  __pyx_v_info->len = __pyx_t_5;

  /* "array.pxd":107
 *             info.len = info.itemsize * item_count
 * 
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)             # <<<<<<<<<<<<<<
 *             if not info.shape:
 *                 raise MemoryError()
 */

#line 107 "array.pxd"
  __pyx_v_info->shape = ((Py_ssize_t *)PyObject_Malloc(((sizeof(Py_ssize_t)) + 2)));

  /* "array.pxd":108
 * 
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)
 *             if not info.shape:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 *             info.shape[0] = item_count      # constant regardless of resizing
 */

#line 108 "array.pxd"
  __pyx_t_6 = ((!(__pyx_v_info->shape != 0)) != 0);

#line 108 "array.pxd"
  if (unlikely(__pyx_t_6)) {

    /* "array.pxd":109
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)
 *             if not info.shape:
 *                 raise MemoryError()             # <<<<<<<<<<<<<<
 *             info.shape[0] = item_count      # constant regardless of resizing
 *             info.strides = &info.itemsize
 */

#line 109 "array.pxd"
    PyErr_NoMemory(); __PYX_ERR(4, 109, __pyx_L1_error)

    /* "array.pxd":108
 * 
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)
 *             if not info.shape:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 *             info.shape[0] = item_count      # constant regardless of resizing
 */

#line 108 "array.pxd"
  }

  /* "array.pxd":110
 *             if not info.shape:
 *                 raise MemoryError()
 *             info.shape[0] = item_count      # constant regardless of resizing             # <<<<<<<<<<<<<<
 *             info.strides = &info.itemsize
 * 
 */

#line 110 "array.pxd"
  __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_v_item_count); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(4, 110, __pyx_L1_error)

#line 110 "array.pxd"
  (__pyx_v_info->shape[0]) = __pyx_t_5;

  /* "array.pxd":111
 *                 raise MemoryError()
 *             info.shape[0] = item_count      # constant regardless of resizing
 *             info.strides = &info.itemsize             # <<<<<<<<<<<<<<
 * 
 *             info.format = <char*> (info.shape + 1)
 */

#line 111 "array.pxd"
  __pyx_v_info->strides = (&__pyx_v_info->itemsize);

  /* "array.pxd":113
 *             info.strides = &info.itemsize
 * 
 *             info.format = <char*> (info.shape + 1)             # <<<<<<<<<<<<<<
 *             info.format[0] = self.ob_descr.typecode
 *             info.format[1] = 0
 */

#line 113 "array.pxd"
  __pyx_v_info->format = ((char *)(__pyx_v_info->shape + 1));

  /* "array.pxd":114
 * 
 *             info.format = <char*> (info.shape + 1)
 *             info.format[0] = self.ob_descr.typecode             # <<<<<<<<<<<<<<
 *             info.format[1] = 0
 *             info.obj = self
 */

#line 114 "array.pxd"
  __pyx_t_7 = __pyx_v_self->ob_descr->typecode;

#line 114 "array.pxd"
  (__pyx_v_info->format[0]) = __pyx_t_7;

  /* "array.pxd":115
 *             info.format = <char*> (info.shape + 1)
 *             info.format[0] = self.ob_descr.typecode
 *             info.format[1] = 0             # <<<<<<<<<<<<<<
 *             info.obj = self
 * 
 */

#line 115 "array.pxd"
  (__pyx_v_info->format[1]) = 0;

  /* "array.pxd":116
 *             info.format[0] = self.ob_descr.typecode
 *             info.format[1] = 0
 *             info.obj = self             # <<<<<<<<<<<<<<
 * 
 *         def __releasebuffer__(self, Py_buffer* info):
 */

#line 116 "array.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 116 "array.pxd"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));

#line 116 "array.pxd"
  __Pyx_GOTREF(__pyx_v_info->obj);

#line 116 "array.pxd"
  __Pyx_DECREF(__pyx_v_info->obj);

#line 116 "array.pxd"
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "array.pxd":93
 *             __data_union data
 * 
 *         def __getbuffer__(self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

#line 93 "array.pxd"


#line 93 "array.pxd"
  /* function exit code */

#line 93 "array.pxd"
  __pyx_r = 0;

#line 93 "array.pxd"
  goto __pyx_L0;

#line 93 "array.pxd"
  __pyx_L1_error:;

#line 93 "array.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 93 "array.pxd"
  __Pyx_XDECREF(__pyx_t_4);

#line 93 "array.pxd"
  __Pyx_AddTraceback("cpython.array.array.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 93 "array.pxd"
  __pyx_r = -1;

#line 93 "array.pxd"
  if (__pyx_v_info->obj != NULL) {

#line 93 "array.pxd"
    __Pyx_GOTREF(__pyx_v_info->obj);

#line 93 "array.pxd"
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;

#line 93 "array.pxd"
  }

#line 93 "array.pxd"
  goto __pyx_L2;

#line 93 "array.pxd"
  __pyx_L0:;

#line 93 "array.pxd"
  if (__pyx_v_info->obj == Py_None) {

#line 93 "array.pxd"
    __Pyx_GOTREF(__pyx_v_info->obj);

#line 93 "array.pxd"
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;

#line 93 "array.pxd"
  }

#line 93 "array.pxd"
  __pyx_L2:;

#line 93 "array.pxd"
  __Pyx_XDECREF(__pyx_v_item_count);

#line 93 "array.pxd"
  __Pyx_RefNannyFinishContext();

#line 93 "array.pxd"
  return __pyx_r;

#line 93 "array.pxd"
}

/* "array.pxd":118
 *             info.obj = self
 * 
 *         def __releasebuffer__(self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             PyObject_Free(info.shape)
 * 
 */

#line 118 "array.pxd"


#line 118 "array.pxd"
/* Python wrapper */

#line 118 "array.pxd"
static CYTHON_UNUSED void __pyx_pw_7cpython_5array_5array_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info); /*proto*/

#line 118 "array.pxd"
static CYTHON_UNUSED void __pyx_pw_7cpython_5array_5array_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info) {

#line 118 "array.pxd"
  __Pyx_RefNannyDeclarations

#line 118 "array.pxd"
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);

#line 118 "array.pxd"
  __pyx_pf_7cpython_5array_5array_2__releasebuffer__(((arrayobject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info));

#line 118 "array.pxd"


#line 118 "array.pxd"
  /* function exit code */

#line 118 "array.pxd"
  __Pyx_RefNannyFinishContext();

#line 118 "array.pxd"
}

#line 118 "array.pxd"


#line 118 "array.pxd"
static void __pyx_pf_7cpython_5array_5array_2__releasebuffer__(CYTHON_UNUSED arrayobject *__pyx_v_self, Py_buffer *__pyx_v_info) {

#line 118 "array.pxd"
  __Pyx_RefNannyDeclarations

#line 118 "array.pxd"
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* "array.pxd":119
 * 
 *         def __releasebuffer__(self, Py_buffer* info):
 *             PyObject_Free(info.shape)             # <<<<<<<<<<<<<<
 * 
 *     array newarrayobject(PyTypeObject* type, Py_ssize_t size, arraydescr *descr)
 */

#line 119 "array.pxd"
  PyObject_Free(__pyx_v_info->shape);

  /* "array.pxd":118
 *             info.obj = self
 * 
 *         def __releasebuffer__(self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             PyObject_Free(info.shape)
 * 
 */

#line 118 "array.pxd"


#line 118 "array.pxd"
  /* function exit code */

#line 118 "array.pxd"
  __Pyx_RefNannyFinishContext();

#line 118 "array.pxd"
}

/* "array.pxd":130
 * 
 * 
 * cdef inline array clone(array template, Py_ssize_t length, bint zero):             # <<<<<<<<<<<<<<
 *     """ fast creation of a new array, given a template array.
 *     type will be same as template.
 */

#line 130 "array.pxd"


#line 130 "array.pxd"
static CYTHON_INLINE arrayobject *__pyx_f_7cpython_5array_clone(arrayobject *__pyx_v_template, Py_ssize_t __pyx_v_length, int __pyx_v_zero) {
  arrayobject *__pyx_v_op = 0
#line 130 "array.pxd"
;

#line 130 "array.pxd"
  arrayobject *__pyx_r = NULL;

#line 130 "array.pxd"
  __Pyx_RefNannyDeclarations

#line 130 "array.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 130 "array.pxd"
  int __pyx_t_2;

#line 130 "array.pxd"
  int __pyx_t_3;

#line 130 "array.pxd"
  int __pyx_t_4;

#line 130 "array.pxd"
  int __pyx_lineno = 0;

#line 130 "array.pxd"
  const char *__pyx_filename = NULL;

#line 130 "array.pxd"
  int __pyx_clineno = 0;

#line 130 "array.pxd"
  __Pyx_RefNannySetupContext("clone", 0);

  /* "array.pxd":134
 *     type will be same as template.
 *     if zero is true, new array will be initialized with zeroes."""
 *     cdef array op = newarrayobject(Py_TYPE(template), length, template.ob_descr)             # <<<<<<<<<<<<<<
 *     if zero and op is not None:
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)
 */

#line 134 "array.pxd"
  __pyx_t_1 = ((PyObject *)newarrayobject(Py_TYPE(((PyObject *)__pyx_v_template)), __pyx_v_length, __pyx_v_template->ob_descr)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 134, __pyx_L1_error)

#line 134 "array.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 134 "array.pxd"
  __pyx_v_op = ((arrayobject *)__pyx_t_1);

#line 134 "array.pxd"
  __pyx_t_1 = 0;

  /* "array.pxd":135
 *     if zero is true, new array will be initialized with zeroes."""
 *     cdef array op = newarrayobject(Py_TYPE(template), length, template.ob_descr)
 *     if zero and op is not None:             # <<<<<<<<<<<<<<
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)
 *     return op
 */

#line 135 "array.pxd"
  __pyx_t_3 = (__pyx_v_zero != 0);

#line 135 "array.pxd"
  if (__pyx_t_3) {

#line 135 "array.pxd"
  } else {

#line 135 "array.pxd"
    __pyx_t_2 = __pyx_t_3;

#line 135 "array.pxd"
    goto __pyx_L4_bool_binop_done;

#line 135 "array.pxd"
  }

#line 135 "array.pxd"
  __pyx_t_3 = (((PyObject *)__pyx_v_op) != Py_None);

#line 135 "array.pxd"
  __pyx_t_4 = (__pyx_t_3 != 0);

#line 135 "array.pxd"
  __pyx_t_2 = __pyx_t_4;

#line 135 "array.pxd"
  __pyx_L4_bool_binop_done:;

#line 135 "array.pxd"
  if (__pyx_t_2) {

    /* "array.pxd":136
 *     cdef array op = newarrayobject(Py_TYPE(template), length, template.ob_descr)
 *     if zero and op is not None:
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)             # <<<<<<<<<<<<<<
 *     return op
 * 
 */

#line 136 "array.pxd"
    (void)(memset(__pyx_v_op->data.as_chars, 0, (__pyx_v_length * __pyx_v_op->ob_descr->itemsize)));

    /* "array.pxd":135
 *     if zero is true, new array will be initialized with zeroes."""
 *     cdef array op = newarrayobject(Py_TYPE(template), length, template.ob_descr)
 *     if zero and op is not None:             # <<<<<<<<<<<<<<
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)
 *     return op
 */

#line 135 "array.pxd"
  }

  /* "array.pxd":137
 *     if zero and op is not None:
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)
 *     return op             # <<<<<<<<<<<<<<
 * 
 * cdef inline array copy(array self):
 */

#line 137 "array.pxd"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 137 "array.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_v_op));

#line 137 "array.pxd"
  __pyx_r = __pyx_v_op;

#line 137 "array.pxd"
  goto __pyx_L0;

  /* "array.pxd":130
 * 
 * 
 * cdef inline array clone(array template, Py_ssize_t length, bint zero):             # <<<<<<<<<<<<<<
 *     """ fast creation of a new array, given a template array.
 *     type will be same as template.
 */

#line 130 "array.pxd"


#line 130 "array.pxd"
  /* function exit code */

#line 130 "array.pxd"
  __pyx_L1_error:;

#line 130 "array.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 130 "array.pxd"
  __Pyx_AddTraceback("cpython.array.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 130 "array.pxd"
  __pyx_r = 0;

#line 130 "array.pxd"
  __pyx_L0:;

#line 130 "array.pxd"
  __Pyx_XDECREF((PyObject *)__pyx_v_op);

#line 130 "array.pxd"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 130 "array.pxd"
  __Pyx_RefNannyFinishContext();

#line 130 "array.pxd"
  return __pyx_r;

#line 130 "array.pxd"
}

/* "array.pxd":139
 *     return op
 * 
 * cdef inline array copy(array self):             # <<<<<<<<<<<<<<
 *     """ make a copy of an array. """
 *     cdef array op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)
 */

#line 139 "array.pxd"


#line 139 "array.pxd"
static CYTHON_INLINE arrayobject *__pyx_f_7cpython_5array_copy(arrayobject *__pyx_v_self) {
  arrayobject *__pyx_v_op = 0
#line 139 "array.pxd"
;

#line 139 "array.pxd"
  arrayobject *__pyx_r = NULL;

#line 139 "array.pxd"
  __Pyx_RefNannyDeclarations

#line 139 "array.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 139 "array.pxd"
  int __pyx_lineno = 0;

#line 139 "array.pxd"
  const char *__pyx_filename = NULL;

#line 139 "array.pxd"
  int __pyx_clineno = 0;

#line 139 "array.pxd"
  __Pyx_RefNannySetupContext("copy", 0);

  /* "array.pxd":141
 * cdef inline array copy(array self):
 *     """ make a copy of an array. """
 *     cdef array op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)             # <<<<<<<<<<<<<<
 *     memcpy(op.data.as_chars, self.data.as_chars, Py_SIZE(op) * op.ob_descr.itemsize)
 *     return op
 */

#line 141 "array.pxd"
  __pyx_t_1 = ((PyObject *)newarrayobject(Py_TYPE(((PyObject *)__pyx_v_self)), Py_SIZE(((PyObject *)__pyx_v_self)), __pyx_v_self->ob_descr)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 141, __pyx_L1_error)

#line 141 "array.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 141 "array.pxd"
  __pyx_v_op = ((arrayobject *)__pyx_t_1);

#line 141 "array.pxd"
  __pyx_t_1 = 0;

  /* "array.pxd":142
 *     """ make a copy of an array. """
 *     cdef array op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)
 *     memcpy(op.data.as_chars, self.data.as_chars, Py_SIZE(op) * op.ob_descr.itemsize)             # <<<<<<<<<<<<<<
 *     return op
 * 
 */

#line 142 "array.pxd"
  (void)(memcpy(__pyx_v_op->data.as_chars, __pyx_v_self->data.as_chars, (Py_SIZE(((PyObject *)__pyx_v_op)) * __pyx_v_op->ob_descr->itemsize)));

  /* "array.pxd":143
 *     cdef array op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)
 *     memcpy(op.data.as_chars, self.data.as_chars, Py_SIZE(op) * op.ob_descr.itemsize)
 *     return op             # <<<<<<<<<<<<<<
 * 
 * cdef inline int extend_buffer(array self, char* stuff, Py_ssize_t n) except -1:
 */

#line 143 "array.pxd"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 143 "array.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_v_op));

#line 143 "array.pxd"
  __pyx_r = __pyx_v_op;

#line 143 "array.pxd"
  goto __pyx_L0;

  /* "array.pxd":139
 *     return op
 * 
 * cdef inline array copy(array self):             # <<<<<<<<<<<<<<
 *     """ make a copy of an array. """
 *     cdef array op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)
 */

#line 139 "array.pxd"


#line 139 "array.pxd"
  /* function exit code */

#line 139 "array.pxd"
  __pyx_L1_error:;

#line 139 "array.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 139 "array.pxd"
  __Pyx_AddTraceback("cpython.array.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 139 "array.pxd"
  __pyx_r = 0;

#line 139 "array.pxd"
  __pyx_L0:;

#line 139 "array.pxd"
  __Pyx_XDECREF((PyObject *)__pyx_v_op);

#line 139 "array.pxd"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 139 "array.pxd"
  __Pyx_RefNannyFinishContext();

#line 139 "array.pxd"
  return __pyx_r;

#line 139 "array.pxd"
}

/* "array.pxd":145
 *     return op
 * 
 * cdef inline int extend_buffer(array self, char* stuff, Py_ssize_t n) except -1:             # <<<<<<<<<<<<<<
 *     """ efficient appending of new stuff of same type
 *     (e.g. of same array type)
 */

#line 145 "array.pxd"


#line 145 "array.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_5array_extend_buffer(arrayobject *__pyx_v_self, char *__pyx_v_stuff, Py_ssize_t __pyx_v_n) {
  Py_ssize_t __pyx_v_itemsize
#line 145 "array.pxd"
;
  Py_ssize_t __pyx_v_origsize
#line 145 "array.pxd"
;

#line 145 "array.pxd"
  int __pyx_r;

#line 145 "array.pxd"
  __Pyx_RefNannyDeclarations

#line 145 "array.pxd"
  int __pyx_t_1;

#line 145 "array.pxd"
  int __pyx_lineno = 0;

#line 145 "array.pxd"
  const char *__pyx_filename = NULL;

#line 145 "array.pxd"
  int __pyx_clineno = 0;

#line 145 "array.pxd"
  __Pyx_RefNannySetupContext("extend_buffer", 0);

  /* "array.pxd":149
 *     (e.g. of same array type)
 *     n: number of elements (not number of bytes!) """
 *     cdef Py_ssize_t itemsize = self.ob_descr.itemsize             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t origsize = Py_SIZE(self)
 *     resize_smart(self, origsize + n)
 */

#line 149 "array.pxd"
  __pyx_t_1 = __pyx_v_self->ob_descr->itemsize;

#line 149 "array.pxd"
  __pyx_v_itemsize = __pyx_t_1;

  /* "array.pxd":150
 *     n: number of elements (not number of bytes!) """
 *     cdef Py_ssize_t itemsize = self.ob_descr.itemsize
 *     cdef Py_ssize_t origsize = Py_SIZE(self)             # <<<<<<<<<<<<<<
 *     resize_smart(self, origsize + n)
 *     memcpy(self.data.as_chars + origsize * itemsize, stuff, n * itemsize)
 */

#line 150 "array.pxd"
  __pyx_v_origsize = Py_SIZE(((PyObject *)__pyx_v_self));

  /* "array.pxd":151
 *     cdef Py_ssize_t itemsize = self.ob_descr.itemsize
 *     cdef Py_ssize_t origsize = Py_SIZE(self)
 *     resize_smart(self, origsize + n)             # <<<<<<<<<<<<<<
 *     memcpy(self.data.as_chars + origsize * itemsize, stuff, n * itemsize)
 *     return 0
 */

#line 151 "array.pxd"
  __pyx_t_1 = resize_smart(__pyx_v_self, (__pyx_v_origsize + __pyx_v_n)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(4, 151, __pyx_L1_error)

  /* "array.pxd":152
 *     cdef Py_ssize_t origsize = Py_SIZE(self)
 *     resize_smart(self, origsize + n)
 *     memcpy(self.data.as_chars + origsize * itemsize, stuff, n * itemsize)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */

#line 152 "array.pxd"
  (void)(memcpy((__pyx_v_self->data.as_chars + (__pyx_v_origsize * __pyx_v_itemsize)), __pyx_v_stuff, (__pyx_v_n * __pyx_v_itemsize)));

  /* "array.pxd":153
 *     resize_smart(self, origsize + n)
 *     memcpy(self.data.as_chars + origsize * itemsize, stuff, n * itemsize)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * cdef inline int extend(array self, array other) except -1:
 */

#line 153 "array.pxd"
  __pyx_r = 0;

#line 153 "array.pxd"
  goto __pyx_L0;

  /* "array.pxd":145
 *     return op
 * 
 * cdef inline int extend_buffer(array self, char* stuff, Py_ssize_t n) except -1:             # <<<<<<<<<<<<<<
 *     """ efficient appending of new stuff of same type
 *     (e.g. of same array type)
 */

#line 145 "array.pxd"


#line 145 "array.pxd"
  /* function exit code */

#line 145 "array.pxd"
  __pyx_L1_error:;

#line 145 "array.pxd"
  __Pyx_AddTraceback("cpython.array.extend_buffer", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 145 "array.pxd"
  __pyx_r = -1;

#line 145 "array.pxd"
  __pyx_L0:;

#line 145 "array.pxd"
  __Pyx_RefNannyFinishContext();

#line 145 "array.pxd"
  return __pyx_r;

#line 145 "array.pxd"
}

/* "array.pxd":155
 *     return 0
 * 
 * cdef inline int extend(array self, array other) except -1:             # <<<<<<<<<<<<<<
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:
 */

#line 155 "array.pxd"


#line 155 "array.pxd"
static CYTHON_INLINE int __pyx_f_7cpython_5array_extend(arrayobject *__pyx_v_self, arrayobject *__pyx_v_other) {

#line 155 "array.pxd"
  int __pyx_r;

#line 155 "array.pxd"
  __Pyx_RefNannyDeclarations

#line 155 "array.pxd"
  int __pyx_t_1;

#line 155 "array.pxd"
  int __pyx_t_2;

#line 155 "array.pxd"
  int __pyx_lineno = 0;

#line 155 "array.pxd"
  const char *__pyx_filename = NULL;

#line 155 "array.pxd"
  int __pyx_clineno = 0;

#line 155 "array.pxd"
  __Pyx_RefNannySetupContext("extend", 0);

  /* "array.pxd":157
 * cdef inline int extend(array self, array other) except -1:
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:             # <<<<<<<<<<<<<<
 *         PyErr_BadArgument()
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 */

#line 157 "array.pxd"
  __pyx_t_1 = ((__pyx_v_self->ob_descr->typecode != __pyx_v_other->ob_descr->typecode) != 0);

#line 157 "array.pxd"
  if (__pyx_t_1) {

    /* "array.pxd":158
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:
 *         PyErr_BadArgument()             # <<<<<<<<<<<<<<
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 * 
 */

#line 158 "array.pxd"
    __pyx_t_2 = PyErr_BadArgument(); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(4, 158, __pyx_L1_error)

    /* "array.pxd":157
 * cdef inline int extend(array self, array other) except -1:
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:             # <<<<<<<<<<<<<<
 *         PyErr_BadArgument()
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 */

#line 157 "array.pxd"
  }

  /* "array.pxd":159
 *     if self.ob_descr.typecode != other.ob_descr.typecode:
 *         PyErr_BadArgument()
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))             # <<<<<<<<<<<<<<
 * 
 * cdef inline void zero(array self):
 */

#line 159 "array.pxd"
  __pyx_t_2 = __pyx_f_7cpython_5array_extend_buffer(__pyx_v_self, __pyx_v_other->data.as_chars, Py_SIZE(((PyObject *)__pyx_v_other))); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(4, 159, __pyx_L1_error)

#line 159 "array.pxd"
  __pyx_r = __pyx_t_2;

#line 159 "array.pxd"
  goto __pyx_L0;

  /* "array.pxd":155
 *     return 0
 * 
 * cdef inline int extend(array self, array other) except -1:             # <<<<<<<<<<<<<<
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:
 */

#line 155 "array.pxd"


#line 155 "array.pxd"
  /* function exit code */

#line 155 "array.pxd"
  __pyx_L1_error:;

#line 155 "array.pxd"
  __Pyx_AddTraceback("cpython.array.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 155 "array.pxd"
  __pyx_r = -1;

#line 155 "array.pxd"
  __pyx_L0:;

#line 155 "array.pxd"
  __Pyx_RefNannyFinishContext();

#line 155 "array.pxd"
  return __pyx_r;

#line 155 "array.pxd"
}

/* "array.pxd":161
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 * 
 * cdef inline void zero(array self):             # <<<<<<<<<<<<<<
 *     """ set all elements of array to zero. """
 *     memset(self.data.as_chars, 0, Py_SIZE(self) * self.ob_descr.itemsize)
 */

#line 161 "array.pxd"


#line 161 "array.pxd"
static CYTHON_INLINE void __pyx_f_7cpython_5array_zero(arrayobject *__pyx_v_self) {

#line 161 "array.pxd"
  __Pyx_RefNannyDeclarations

#line 161 "array.pxd"
  __Pyx_RefNannySetupContext("zero", 0);

  /* "array.pxd":163
 * cdef inline void zero(array self):
 *     """ set all elements of array to zero. """
 *     memset(self.data.as_chars, 0, Py_SIZE(self) * self.ob_descr.itemsize)             # <<<<<<<<<<<<<<
 */

#line 163 "array.pxd"
  (void)(memset(__pyx_v_self->data.as_chars, 0, (Py_SIZE(((PyObject *)__pyx_v_self)) * __pyx_v_self->ob_descr->itemsize)));

  /* "array.pxd":161
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 * 
 * cdef inline void zero(array self):             # <<<<<<<<<<<<<<
 *     """ set all elements of array to zero. """
 *     memset(self.data.as_chars, 0, Py_SIZE(self) * self.ob_descr.itemsize)
 */

#line 161 "array.pxd"


#line 161 "array.pxd"
  /* function exit code */

#line 161 "array.pxd"
  __Pyx_RefNannyFinishContext();

#line 161 "array.pxd"
}

#line 161 "array.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap __pyx_vtable_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap;

#line 161 "array.pxd"


#line 161 "array.pxd"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {

#line 161 "array.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *p;

#line 161 "array.pxd"
  PyObject *o;

#line 161 "array.pxd"
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {

#line 161 "array.pxd"
    o = (*t->tp_alloc)(t, 0);

#line 161 "array.pxd"
  } else {

#line 161 "array.pxd"
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);

#line 161 "array.pxd"
  }

#line 161 "array.pxd"
  if (unlikely(!o)) return 0;

#line 161 "array.pxd"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)o);

#line 161 "array.pxd"
  p->__pyx_vtab = __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap;

#line 161 "array.pxd"
  p->features = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 161 "array.pxd"
  return o;

#line 161 "array.pxd"
}

#line 161 "array.pxd"


#line 161 "array.pxd"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap(PyObject *o) {

#line 161 "array.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)o;

#line 161 "array.pxd"
  #if CYTHON_USE_TP_FINALIZE

#line 161 "array.pxd"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {

#line 161 "array.pxd"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 161 "array.pxd"
  }

#line 161 "array.pxd"
  #endif

#line 161 "array.pxd"
  PyObject_GC_UnTrack(o);

#line 161 "array.pxd"
  Py_CLEAR(p->features);

#line 161 "array.pxd"
  (*Py_TYPE(o)->tp_free)(o);

#line 161 "array.pxd"
}

#line 161 "array.pxd"


#line 161 "array.pxd"
static int __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap(PyObject *o, visitproc v, void *a) {

#line 161 "array.pxd"
  int e;

#line 161 "array.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)o;

#line 161 "array.pxd"
  if (p->features) {

#line 161 "array.pxd"
    e = (*v)(p->features, a); if (e) return e;

#line 161 "array.pxd"
  }

#line 161 "array.pxd"
  return 0;

#line 161 "array.pxd"
}

#line 161 "array.pxd"


#line 161 "array.pxd"
static int __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap(PyObject *o) {

#line 161 "array.pxd"
  PyObject* tmp;

#line 161 "array.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *)o;

#line 161 "array.pxd"
  tmp = ((PyObject*)p->features);

#line 161 "array.pxd"
  p->features = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 161 "array.pxd"
  Py_XDECREF(tmp);

#line 161 "array.pxd"
  return 0;

#line 161 "array.pxd"
}

#line 161 "array.pxd"
static PyObject *__pyx_sq_item_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap(PyObject *o, Py_ssize_t i) {

#line 161 "array.pxd"
  PyObject *r;

#line 161 "array.pxd"
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;

#line 161 "array.pxd"
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);

#line 161 "array.pxd"
  Py_DECREF(x);

#line 161 "array.pxd"
  return r;

#line 161 "array.pxd"
}

#line 161 "array.pxd"


#line 161 "array.pxd"
static PyObject *__pyx_tp_richcompare_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap(PyObject *o1, PyObject *o2, int op) {

#line 161 "array.pxd"
  switch (op) {

#line 161 "array.pxd"
    case Py_EQ: {

#line 161 "array.pxd"
      return __pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_9__eq__(o1, o2);

#line 161 "array.pxd"
    }

#line 161 "array.pxd"
    case Py_NE: {

#line 161 "array.pxd"
      return __pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_11__ne__(o1, o2);

#line 161 "array.pxd"
    }

#line 161 "array.pxd"
    default: {

#line 161 "array.pxd"
      return __Pyx_NewRef(Py_NotImplemented);

#line 161 "array.pxd"
    }

#line 161 "array.pxd"
  }

#line 161 "array.pxd"
}

#line 161 "array.pxd"


#line 161 "array.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_features(PyObject *o, CYTHON_UNUSED void *x) {

#line 161 "array.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_8features_1__get__(o);

#line 161 "array.pxd"
}

#line 161 "array.pxd"


#line 161 "array.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_features(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 161 "array.pxd"
  if (v) {

#line 161 "array.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_8features_3__set__(o, v);

#line 161 "array.pxd"
  }

#line 161 "array.pxd"
  else {

#line 161 "array.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_8features_5__del__(o);

#line 161 "array.pxd"
  }

#line 161 "array.pxd"
}

#line 161 "array.pxd"


#line 161 "array.pxd"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap[] = {

#line 161 "array.pxd"
  {"search", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_13search, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_12search},

#line 161 "array.pxd"
  {"find_all", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_15find_all, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_14find_all},

#line 161 "array.pxd"
  {"index_range", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_17index_range, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_16index_range},

#line 161 "array.pxd"
  {"spanning_time", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_19spanning_time, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_18spanning_time},

#line 161 "array.pxd"
  {"between", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_21between, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_20between},

#line 161 "array.pxd"
  {"clone", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_25clone, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_24clone},

#line 161 "array.pxd"
  {"as_arrays", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_27as_arrays, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_26as_arrays},

#line 161 "array.pxd"
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_29__reduce_cython__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_28__reduce_cython__},

#line 161 "array.pxd"
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_31__setstate_cython__, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_30__setstate_cython__},

#line 161 "array.pxd"
  {0, 0, 0, 0}

#line 161 "array.pxd"
};

#line 161 "array.pxd"


#line 161 "array.pxd"
static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap[] = {

#line 161 "array.pxd"
  {(char *)"features", __pyx_getprop_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_features, __pyx_setprop_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_features, (char *)"features: list", 0},

#line 161 "array.pxd"
  {0, 0, 0, 0, 0}

#line 161 "array.pxd"
};

#line 161 "array.pxd"


#line 161 "array.pxd"
static PySequenceMethods __pyx_tp_as_sequence_LCMSFeatureMap = {

#line 161 "array.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_3__len__, /*sq_length*/

#line 161 "array.pxd"
  0, /*sq_concat*/

#line 161 "array.pxd"
  0, /*sq_repeat*/

#line 161 "array.pxd"
  __pyx_sq_item_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap, /*sq_item*/

#line 161 "array.pxd"
  0, /*sq_slice*/

#line 161 "array.pxd"
  0, /*sq_ass_item*/

#line 161 "array.pxd"
  0, /*sq_ass_slice*/

#line 161 "array.pxd"
  0, /*sq_contains*/

#line 161 "array.pxd"
  0, /*sq_inplace_concat*/

#line 161 "array.pxd"
  0, /*sq_inplace_repeat*/

#line 161 "array.pxd"
};

#line 161 "array.pxd"


#line 161 "array.pxd"
static PyMappingMethods __pyx_tp_as_mapping_LCMSFeatureMap = {

#line 161 "array.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_3__len__, /*mp_length*/

#line 161 "array.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_7__getitem__, /*mp_subscript*/

#line 161 "array.pxd"
  0, /*mp_ass_subscript*/

#line 161 "array.pxd"
};

#line 161 "array.pxd"


#line 161 "array.pxd"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap = {

#line 161 "array.pxd"
  PyVarObject_HEAD_INIT(0, 0)

#line 161 "array.pxd"
  "ms_deisotope._c.feature_map.feature_map.LCMSFeatureMap", /*tp_name*/

#line 161 "array.pxd"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap), /*tp_basicsize*/

#line 161 "array.pxd"
  0, /*tp_itemsize*/

#line 161 "array.pxd"
  __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap, /*tp_dealloc*/

#line 161 "array.pxd"
  #if PY_VERSION_HEX < 0x030800b4

#line 161 "array.pxd"
  0, /*tp_print*/

#line 161 "array.pxd"
  #endif

#line 161 "array.pxd"
  #if PY_VERSION_HEX >= 0x030800b4

#line 161 "array.pxd"
  0, /*tp_vectorcall_offset*/

#line 161 "array.pxd"
  #endif

#line 161 "array.pxd"
  0, /*tp_getattr*/

#line 161 "array.pxd"
  0, /*tp_setattr*/

#line 161 "array.pxd"
  #if PY_MAJOR_VERSION < 3

#line 161 "array.pxd"
  0, /*tp_compare*/

#line 161 "array.pxd"
  #endif

#line 161 "array.pxd"
  #if PY_MAJOR_VERSION >= 3

#line 161 "array.pxd"
  0, /*tp_as_async*/

#line 161 "array.pxd"
  #endif

#line 161 "array.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_23__repr__, /*tp_repr*/

#line 161 "array.pxd"
  0, /*tp_as_number*/

#line 161 "array.pxd"
  &__pyx_tp_as_sequence_LCMSFeatureMap, /*tp_as_sequence*/

#line 161 "array.pxd"
  &__pyx_tp_as_mapping_LCMSFeatureMap, /*tp_as_mapping*/

#line 161 "array.pxd"
  0, /*tp_hash*/

#line 161 "array.pxd"
  0, /*tp_call*/

#line 161 "array.pxd"
  0, /*tp_str*/

#line 161 "array.pxd"
  0, /*tp_getattro*/

#line 161 "array.pxd"
  0, /*tp_setattro*/

#line 161 "array.pxd"
  0, /*tp_as_buffer*/

#line 161 "array.pxd"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 161 "array.pxd"
  "LCMSFeatureMap(features)", /*tp_doc*/

#line 161 "array.pxd"
  __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap, /*tp_traverse*/

#line 161 "array.pxd"
  __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap, /*tp_clear*/

#line 161 "array.pxd"
  __pyx_tp_richcompare_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap, /*tp_richcompare*/

#line 161 "array.pxd"
  0, /*tp_weaklistoffset*/

#line 161 "array.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_5__iter__, /*tp_iter*/

#line 161 "array.pxd"
  0, /*tp_iternext*/

#line 161 "array.pxd"
  __pyx_methods_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap, /*tp_methods*/

#line 161 "array.pxd"
  0, /*tp_members*/

#line 161 "array.pxd"
  __pyx_getsets_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap, /*tp_getset*/

#line 161 "array.pxd"
  0, /*tp_base*/

#line 161 "array.pxd"
  0, /*tp_dict*/

#line 161 "array.pxd"
  0, /*tp_descr_get*/

#line 161 "array.pxd"
  0, /*tp_descr_set*/

#line 161 "array.pxd"
  0, /*tp_dictoffset*/

#line 161 "array.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_1__init__, /*tp_init*/

#line 161 "array.pxd"
  0, /*tp_alloc*/

#line 161 "array.pxd"
  __pyx_tp_new_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap, /*tp_new*/

#line 161 "array.pxd"
  0, /*tp_free*/

#line 161 "array.pxd"
  0, /*tp_is_gc*/

#line 161 "array.pxd"
  0, /*tp_bases*/

#line 161 "array.pxd"
  0, /*tp_mro*/

#line 161 "array.pxd"
  0, /*tp_cache*/

#line 161 "array.pxd"
  0, /*tp_subclasses*/

#line 161 "array.pxd"
  0, /*tp_weaklist*/

#line 161 "array.pxd"
  0, /*tp_del*/

#line 161 "array.pxd"
  0, /*tp_version_tag*/

#line 161 "array.pxd"
  #if PY_VERSION_HEX >= 0x030400a1

#line 161 "array.pxd"
  0, /*tp_finalize*/

#line 161 "array.pxd"
  #endif

#line 161 "array.pxd"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 161 "array.pxd"
  0, /*tp_vectorcall*/

#line 161 "array.pxd"
  #endif

#line 161 "array.pxd"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 161 "array.pxd"
  0, /*tp_print*/

#line 161 "array.pxd"
  #endif

#line 161 "array.pxd"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 161 "array.pxd"
  0, /*tp_pypy_flags*/

#line 161 "array.pxd"
  #endif

#line 161 "array.pxd"
};

#line 161 "array.pxd"


#line 161 "array.pxd"
static PyMethodDef __pyx_methods[] = {

#line 161 "array.pxd"
  {"split_sparse", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_1split_sparse, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_11feature_map_split_sparse},

#line 161 "array.pxd"
  {"binary_search_with_flag", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_11feature_map_3binary_search_with_flag, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_11feature_map_2binary_search_with_flag},

#line 161 "array.pxd"
  {0, 0, 0, 0}

#line 161 "array.pxd"
};

#line 161 "array.pxd"


#line 161 "array.pxd"
#if PY_MAJOR_VERSION >= 3

#line 161 "array.pxd"
#if CYTHON_PEP489_MULTI_PHASE_INIT

#line 161 "array.pxd"
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/

#line 161 "array.pxd"
static int __pyx_pymod_exec_feature_map(PyObject* module); /*proto*/

#line 161 "array.pxd"
static PyModuleDef_Slot __pyx_moduledef_slots[] = {

#line 161 "array.pxd"
  {Py_mod_create, (void*)__pyx_pymod_create},

#line 161 "array.pxd"
  {Py_mod_exec, (void*)__pyx_pymod_exec_feature_map},

#line 161 "array.pxd"
  {0, NULL}

#line 161 "array.pxd"
};

#line 161 "array.pxd"
#endif

#line 161 "array.pxd"


#line 161 "array.pxd"
static struct PyModuleDef __pyx_moduledef = {

#line 161 "array.pxd"
    PyModuleDef_HEAD_INIT,

#line 161 "array.pxd"
    "feature_map",

#line 161 "array.pxd"
    0, /* m_doc */

#line 161 "array.pxd"
  #if CYTHON_PEP489_MULTI_PHASE_INIT

#line 161 "array.pxd"
    0, /* m_size */

#line 161 "array.pxd"
  #else

#line 161 "array.pxd"
    -1, /* m_size */

#line 161 "array.pxd"
  #endif

#line 161 "array.pxd"
    __pyx_methods /* m_methods */,

#line 161 "array.pxd"
  #if CYTHON_PEP489_MULTI_PHASE_INIT

#line 161 "array.pxd"
    __pyx_moduledef_slots, /* m_slots */

#line 161 "array.pxd"
  #else

#line 161 "array.pxd"
    NULL, /* m_reload */

#line 161 "array.pxd"
  #endif

#line 161 "array.pxd"
    NULL, /* m_traverse */

#line 161 "array.pxd"
    NULL, /* m_clear */

#line 161 "array.pxd"
    NULL /* m_free */

#line 161 "array.pxd"
};

#line 161 "array.pxd"
#endif

#line 161 "array.pxd"
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_k_Incompatible_checksums_0x_x_vs_0, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0), 0, 0, 1, 0},
  {&__pyx_n_s_L, __pyx_k_L, sizeof(__pyx_k_L), 0, 0, 1, 1},
  {&__pyx_n_s_LCMSFeatureMap, __pyx_k_LCMSFeatureMap, sizeof(__pyx_k_LCMSFeatureMap), 0, 0, 1, 1},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
  {&__pyx_n_s_array, __pyx_k_array, sizeof(__pyx_k_array), 0, 0, 1, 1},
  {&__pyx_n_s_between, __pyx_k_between, sizeof(__pyx_k_between), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_clone, __pyx_k_clone, sizeof(__pyx_k_clone), 0, 0, 1, 1},
  {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
  {&__pyx_n_s_d, __pyx_k_d, sizeof(__pyx_k_d), 0, 0, 1, 1},
  {&__pyx_n_s_deep, __pyx_k_deep, sizeof(__pyx_k_deep), 0, 0, 1, 1},
  {&__pyx_n_s_delta_rt, __pyx_k_delta_rt, sizeof(__pyx_k_delta_rt), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_error_tolerance, __pyx_k_error_tolerance, sizeof(__pyx_k_error_tolerance), 0, 0, 1, 1},
  {&__pyx_n_s_features, __pyx_k_features, sizeof(__pyx_k_features), 0, 0, 1, 1},
  {&__pyx_n_s_find_all, __pyx_k_find_all, sizeof(__pyx_k_find_all), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_hi, __pyx_k_hi, sizeof(__pyx_k_hi), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_init___locals_lambda, __pyx_k_init___locals_lambda, sizeof(__pyx_k_init___locals_lambda), 0, 0, 1, 1},
  {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
  {&__pyx_n_s_lexsort, __pyx_k_lexsort, sizeof(__pyx_k_lexsort), 0, 0, 1, 1},
  {&__pyx_n_s_lo, __pyx_k_lo, sizeof(__pyx_k_lo), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_min_size, __pyx_k_min_size, sizeof(__pyx_k_min_size), 0, 0, 1, 1},
  {&__pyx_n_s_ms_deisotope__c_feature_map_feat, __pyx_k_ms_deisotope__c_feature_map_feat, sizeof(__pyx_k_ms_deisotope__c_feature_map_feat), 0, 0, 1, 1},
  {&__pyx_n_s_mz, __pyx_k_mz, sizeof(__pyx_k_mz), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_s_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 0, 1, 0},
  {&__pyx_kp_s_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 0, 1, 0},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_LCMSFeatureMap, __pyx_k_pyx_unpickle_LCMSFeatureMap, sizeof(__pyx_k_pyx_unpickle_LCMSFeatureMap), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_search, __pyx_k_search, sizeof(__pyx_k_search), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_kp_s_self___class_____name___size_fe, __pyx_k_self___class_____name___size_fe, sizeof(__pyx_k_self___class_____name___size_fe), 0, 0, 1, 0},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_n_s_sorted, __pyx_k_sorted, sizeof(__pyx_k_sorted), 0, 0, 1, 1},
  {&__pyx_n_s_spanning_time, __pyx_k_spanning_time, sizeof(__pyx_k_spanning_time), 0, 0, 1, 1},
  {&__pyx_n_s_split_sparse, __pyx_k_split_sparse, sizeof(__pyx_k_split_sparse), 0, 0, 1, 1},
  {&__pyx_kp_s_src_ms_deisotope__c_feature_map, __pyx_k_src_ms_deisotope__c_feature_map, sizeof(__pyx_k_src_ms_deisotope__c_feature_map), 0, 0, 1, 0},
  {&__pyx_n_s_stack, __pyx_k_stack, sizeof(__pyx_k_stack), 0, 0, 1, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_sorted = __Pyx_GetBuiltinName(__pyx_n_s_sorted); if (!__pyx_builtin_sorted) __PYX_ERR(0, 21, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 120, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(3, 944, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(4, 109, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":192
 *             PeakBase peak
 * 
 *         mz_array = array.array('d')             # <<<<<<<<<<<<<<
 *         intensity_array = array.array('d')
 *         ion_mobility_array = array.array('d')
 */

#line 192 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_n_s_d); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 192, __pyx_L1_error)

#line 192 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_tuple_);

#line 192 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":195
 *         intensity_array = array.array('d')
 *         ion_mobility_array = array.array('d')
 *         feature_id_array = array.array('L')             # <<<<<<<<<<<<<<
 *         n = self.get_size()
 *         for i in range(n):
 */

#line 195 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_n_s_L); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 195, __pyx_L1_error)

#line 195 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_tuple__2);

#line 195 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9f034b6, 0x5b8a8b5, 0x4f94454):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x9f034b6, 0x5b8a8b5, 0x4f94454) = (features))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  __pyx_tuple__3 = PyTuple_Pack(3, __pyx_int_166737078, __pyx_int_95987893, __pyx_int_83444820); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(2, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__3);

#line 4 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":944
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(3, 944, __pyx_L1_error)

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_tuple__4);

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":950
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(3, 950, __pyx_L1_error)

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_tuple__5);

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":220
 * 
 * @cython.binding(True)
 * cpdef list split_sparse(LCMSFeatureMap self, double delta_rt=1.0, size_t min_size=2):             # <<<<<<<<<<<<<<
 *     cdef:
 *         list result, chunks
 */

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_tuple__6 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_delta_rt, __pyx_n_s_min_size); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 220, __pyx_L1_error)

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_tuple__6);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GIVEREF(__pyx_tuple__6);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_codeobj__7 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_ms_deisotope__c_feature_map, __pyx_n_s_split_sparse, 220, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__7)) __PYX_ERR(0, 220, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle_LCMSFeatureMap(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"
  __pyx_tuple__8 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__8);

#line 1 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__8);

#line 1 "(tree fragment)"
  __pyx_codeobj__9 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_LCMSFeatureMap, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__9)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return 0;

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return -1;

#line 1 "(tree fragment)"
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_2eneg_5 = PyFloat_FromDouble(2e-5); if (unlikely(!__pyx_float_2eneg_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_83444820 = PyInt_FromLong(83444820L); if (unlikely(!__pyx_int_83444820)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_95987893 = PyInt_FromLong(95987893L); if (unlikely(!__pyx_int_95987893)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_166737078 = PyInt_FromLong(166737078L); if (unlikely(!__pyx_int_166737078)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("binary_search_with_flag", (void (*)(void))__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_binary_search_with_flag, "PyObject *(PyObject *, double, double, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("binary_search", (void (*)(void))__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_binary_search, "long (PyObject *, double, double)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("search_sweep", (void (*)(void))__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_search_sweep, "void (PyObject *, double, double, long *, long *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("split_sparse", (void (*)(void))__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_split_sparse, "PyObject *(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_split_sparse *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap = &__pyx_vtable_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap.get_size = (Py_ssize_t (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *))__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_get_size;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap.get = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *, size_t))__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_get;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap.find_all = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_find_all *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_find_all;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap.search = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_search *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_search;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap.between = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *, double, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_between *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_between;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap.spanning_time = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *, double, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_spanning_time;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap.clone = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_clone *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_11feature_map_11feature_map_14LCMSFeatureMap_clone;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_LCMSFeatureMap, (PyObject *)&__pyx_type_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap = &__pyx_type_12ms_deisotope_2_c_11feature_map_11feature_map_LCMSFeatureMap;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("ms_peak_picker._c.peak_set"); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "PeakBase", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase) __PYX_ERR(5, 2, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase)) __PYX_ERR(5, 2, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "FittedPeak", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak) __PYX_ERR(5, 12, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak)) __PYX_ERR(5, 12, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "PeakSet", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet) __PYX_ERR(5, 29, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet)) __PYX_ERR(5, 29, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "PeakSetIndexed", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed) __PYX_ERR(5, 83, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed)) __PYX_ERR(5, 83, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.feature_map.lcms_feature"); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.lcms_feature", "LCMSFeatureTreeList", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList) __PYX_ERR(6, 4, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList)) __PYX_ERR(6, 4, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.lcms_feature", "LCMSFeatureTreeNodeBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase) __PYX_ERR(6, 23, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase)) __PYX_ERR(6, 23, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.lcms_feature", "LCMSFeatureTreeNode", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode) __PYX_ERR(6, 44, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode)) __PYX_ERR(6, 44, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.lcms_feature", "FeatureBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase) __PYX_ERR(6, 56, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase)) __PYX_ERR(6, 56, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.lcms_feature", "LCMSFeature", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature) __PYX_ERR(6, 73, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature)) __PYX_ERR(6, 73, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.lcms_feature", "EmptyFeature", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature) __PYX_ERR(6, 104, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature)) __PYX_ERR(6, 104, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.lcms_feature", "FeatureSetIterator", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator) __PYX_ERR(6, 109, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator)) __PYX_ERR(6, 109, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.lcms_feature", "RunningWeightedAverage", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage) __PYX_ERR(6, 137, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage)) __PYX_ERR(6, 137, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.feature_map.lcms_feature", "RunningWeightedAverageNeutralMass", sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass) __PYX_ERR(6, 154, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass)) __PYX_ERR(6, 154, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(7, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(8, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_7complex_complex) __PYX_ERR(9, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(3, 199, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(3, 222, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(3, 226, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(3, 238, __pyx_L1_error)
  __pyx_ptype_5numpy_generic = __Pyx_ImportType(__pyx_t_1, "numpy", "generic", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_generic) __PYX_ERR(3, 770, __pyx_L1_error)
  __pyx_ptype_5numpy_number = __Pyx_ImportType(__pyx_t_1, "numpy", "number", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_number) __PYX_ERR(3, 772, __pyx_L1_error)
  __pyx_ptype_5numpy_integer = __Pyx_ImportType(__pyx_t_1, "numpy", "integer", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_integer) __PYX_ERR(3, 774, __pyx_L1_error)
  __pyx_ptype_5numpy_signedinteger = __Pyx_ImportType(__pyx_t_1, "numpy", "signedinteger", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_signedinteger) __PYX_ERR(3, 776, __pyx_L1_error)
  __pyx_ptype_5numpy_unsignedinteger = __Pyx_ImportType(__pyx_t_1, "numpy", "unsignedinteger", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_unsignedinteger) __PYX_ERR(3, 778, __pyx_L1_error)
  __pyx_ptype_5numpy_inexact = __Pyx_ImportType(__pyx_t_1, "numpy", "inexact", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_inexact) __PYX_ERR(3, 780, __pyx_L1_error)
  __pyx_ptype_5numpy_floating = __Pyx_ImportType(__pyx_t_1, "numpy", "floating", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_floating) __PYX_ERR(3, 782, __pyx_L1_error)
  __pyx_ptype_5numpy_complexfloating = __Pyx_ImportType(__pyx_t_1, "numpy", "complexfloating", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_complexfloating) __PYX_ERR(3, 784, __pyx_L1_error)
  __pyx_ptype_5numpy_flexible = __Pyx_ImportType(__pyx_t_1, "numpy", "flexible", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_flexible) __PYX_ERR(3, 786, __pyx_L1_error)
  __pyx_ptype_5numpy_character = __Pyx_ImportType(__pyx_t_1, "numpy", "character", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_character) __PYX_ERR(3, 788, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(3, 826, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("array"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_5array_array = __Pyx_ImportType(__pyx_t_1, "array", "array", sizeof(arrayobject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_5array_array) __PYX_ERR(4, 58, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __pyx_t_1 = PyImport_ImportModule("ms_peak_picker._c.peak_set"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "INTERVAL_INDEX_SIZE", (void **)&__pyx_vp_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE, "size_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initfeature_map(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initfeature_map(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_feature_map(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_feature_map(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_feature_map(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'feature_map' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_feature_map(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("feature_map", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_ms_deisotope___c__feature_map__feature_map) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "ms_deisotope._c.feature_map.feature_map")) {
      if (unlikely(PyDict_SetItemString(modules, "ms_deisotope._c.feature_map.feature_map", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  if (unlikely(__Pyx_modinit_function_export_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_variable_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":10
 * 
 * cimport numpy as np
 * import numpy as np             # <<<<<<<<<<<<<<
 * 
 * np.import_array()
 */

#line 10 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 10, __pyx_L1_error)

#line 10 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 10 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_1) < 0) __PYX_ERR(0, 10, __pyx_L1_error)

#line 10 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":12
 * import numpy as np
 * 
 * np.import_array()             # <<<<<<<<<<<<<<
 * 
 * from cpython cimport array
 */

#line 12 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_2 = __pyx_f_5numpy_import_array(); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 12, __pyx_L1_error)

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":15
 * 
 * from cpython cimport array
 * import array             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 15 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_array, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error)

#line 15 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 15 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_array, __pyx_t_1) < 0) __PYX_ERR(0, 15, __pyx_L1_error)

#line 15 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":220
 * 
 * @cython.binding(True)
 * cpdef list split_sparse(LCMSFeatureMap self, double delta_rt=1.0, size_t min_size=2):             # <<<<<<<<<<<<<<
 *     cdef:
 *         list result, chunks
 */

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_12ms_deisotope_2_c_11feature_map_11feature_map_1split_sparse, 0, __pyx_n_s_split_sparse, NULL, __pyx_n_s_ms_deisotope__c_feature_map_feat, __pyx_d, ((PyObject *)__pyx_codeobj__7)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 220, __pyx_L1_error)

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_split_sparse, __pyx_t_1) < 0) __PYX_ERR(0, 220, __pyx_L1_error)

#line 220 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_LCMSFeatureMap(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_11feature_map_11feature_map_5__pyx_unpickle_LCMSFeatureMap, NULL, __pyx_n_s_ms_deisotope__c_feature_map_feat); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 1 "(tree fragment)"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_LCMSFeatureMap, __pyx_t_1) < 0) __PYX_ERR(2, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/feature_map.pyx":1
 * # cython: embedsignature =True             # <<<<<<<<<<<<<<
 * 
 * cimport cython
 */

#line 1 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)

#line 1 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 1 "ms_deisotope/_c/feature_map/feature_map.pyx"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)

#line 1 "ms_deisotope/_c/feature_map/feature_map.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "array.pxd":161
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 * 
 * cdef inline void zero(array self):             # <<<<<<<<<<<<<<
 *     """ set all elements of array to zero. """
 *     memset(self.data.as_chars, 0, Py_SIZE(self) * self.ob_descr.itemsize)
 */

#line 161 "array.pxd"


#line 161 "array.pxd"
  /*--- Wrapped vars code ---*/

#line 161 "array.pxd"


#line 161 "array.pxd"
  goto __pyx_L0;

#line 161 "array.pxd"
  __pyx_L1_error:;

#line 161 "array.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 161 "array.pxd"
  if (__pyx_m) {

#line 161 "array.pxd"
    if (__pyx_d) {

#line 161 "array.pxd"
      __Pyx_AddTraceback("init ms_deisotope._c.feature_map.feature_map", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 161 "array.pxd"
    }

#line 161 "array.pxd"
    Py_CLEAR(__pyx_m);

#line 161 "array.pxd"
  } else if (!PyErr_Occurred()) {

#line 161 "array.pxd"
    PyErr_SetString(PyExc_ImportError, "init ms_deisotope._c.feature_map.feature_map");

#line 161 "array.pxd"
  }

#line 161 "array.pxd"
  __pyx_L0:;

#line 161 "array.pxd"
  __Pyx_RefNannyFinishContext();

#line 161 "array.pxd"
  #if CYTHON_PEP489_MULTI_PHASE_INIT

#line 161 "array.pxd"
  return (__pyx_m != NULL) ? 0 : -1;

#line 161 "array.pxd"
  #elif PY_MAJOR_VERSION >= 3

#line 161 "array.pxd"
  return __pyx_m;

#line 161 "array.pxd"
  #else

#line 161 "array.pxd"
  return;

#line 161 "array.pxd"
  #endif

#line 161 "array.pxd"
}

#line 1 "cython_utility"

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* FetchCommonType */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunctionShared */
#include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    if (unlikely(op == NULL))
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
#if PY_MAJOR_VERSION < 3
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
#endif
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
#if PY_MAJOR_VERSION > 2
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
#else
            PyErr_SetString(PyExc_TypeError,
                            "unbound method needs an argument");
#endif
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* append */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* DivInt[long] */
static CYTHON_INLINE long __Pyx_div_long(long a, long b) {
    long q = a / b;
    long r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* Declarations */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionExport */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* VoidPtrImport */
#ifndef __PYX_HAVE_RT_ImportVoidPtr
#define __PYX_HAVE_RT_ImportVoidPtr
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, name);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C variable %.200s",
                PyModule_GetName(module), name);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, desc);
        goto bad;
    }
    *p = PyCObject_AsVoidPtr(cobj);}
#endif
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
