/* Generated by Cython 0.29.33 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/arrayobject.h",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/arrayscalars.h",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/ndarrayobject.h",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/ndarraytypes.h",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/ufuncobject.h"
        ],
        "include_dirs": [
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/brainpy/_c",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/ms_peak_picker/_c",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include"
        ],
        "name": "ms_deisotope._c.scoring",
        "sources": [
            "src/ms_deisotope/_c/scoring.pyx"
        ]
    },
    "module_name": "ms_deisotope._c.scoring"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_33"
#define CYTHON_HEX_VERSION 0x001D21F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 0
  #elif !defined(CYTHON_FAST_THREAD_STATE)
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030A0000)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__ms_deisotope___c__scoring
#define __PYX_HAVE_API__ms_deisotope___c__scoring
/* Early includes */
#include <string.h>
#include <stdio.h>
#include "numpy/arrayobject.h"
#include "numpy/ndarrayobject.h"
#include "numpy/ndarraytypes.h"
#include "numpy/arrayscalars.h"
#include "numpy/ufuncobject.h"

    /* NumPy API declarations from "numpy/__init__.pxd" */
    
#include <math.h>
#include <stdlib.h>
#include "pythread.h"
#include "pystate.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "src/ms_deisotope/_c/scoring.pyx",
  "src/ms_deisotope/_c/scoring.pxd",
  "__init__.pxd",
  "stringsource",
  "composition.pxd",
  "isotopic_distribution.pxd",
  "peak_set.pxd",
  "type.pxd",
  "peak_index.pxd",
  "src/ms_deisotope/_c/averagine.pxd",
  "src/ms_deisotope/_c/deconvoluter_base.pxd",
};
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* BufferFormatStructs.proto */
#define IS_UNSIGNED(type) (((type) -1) > 0)
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;

/* Atomics.proto */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __PYX_CYTHON_ATOMICS_ENABLED() CYTHON_ATOMICS
#define __pyx_atomic_int_type int
#if CYTHON_ATOMICS && (__GNUC__ >= 5 || (__GNUC__ == 4 &&\
                    (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ >= 2))))
    #define __pyx_atomic_incr_aligned(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_aligned(value) __sync_fetch_and_sub(value, 1)
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER) && CYTHON_COMPILING_IN_NOGIL
    #include <intrin.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type long
    #pragma intrinsic (_InterlockedExchangeAdd)
    #define __pyx_atomic_incr_aligned(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_decr_aligned(value) _InterlockedExchangeAdd(value, -1)
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif
typedef volatile __pyx_atomic_int_type __pyx_atomic_int;
#if CYTHON_ATOMICS
    #define __pyx_add_acquisition_count(memview)\
             __pyx_atomic_incr_aligned(__pyx_get_slice_count_pointer(memview))
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_atomic_decr_aligned(__pyx_get_slice_count_pointer(memview))
#else
    #define __pyx_add_acquisition_count(memview)\
            __pyx_add_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_sub_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
#endif

/* MemviewSliceStruct.proto */
struct __pyx_memoryview_obj;
typedef struct {
  struct __pyx_memoryview_obj *memview;
  char *data;
  Py_ssize_t shape[8];
  Py_ssize_t strides[8];
  Py_ssize_t suboffsets[8];
} __Pyx_memviewslice;
#define __Pyx_MemoryView_Len(m)  (m.shape[0])


/* "brainpy/_c/composition.pxd":3
 * from brainpy.mass_dict import nist_mass as __nist_mass
 * 
 * ctypedef long count_type             # <<<<<<<<<<<<<<
 * 
 * cdef dict nist_mass
 */

#line 3 "brainpy/_c/composition.pxd"
typedef long __pyx_t_7brainpy_2_c_11composition_count_type;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":689
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */

#line 689 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":690
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */

#line 690 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":691
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */

#line 691 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":692
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */

#line 692 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":696
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */

#line 696 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":697
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */

#line 697 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":698
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */

#line 698 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":699
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */

#line 699 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":703
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */

#line 703 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":704
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */

#line 704 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":713
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */

#line 713 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":714
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */

#line 714 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":715
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */

#line 715 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":717
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */

#line 717 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":718
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */

#line 718 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":719
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */

#line 719 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":721
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */

#line 721 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":722
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */

#line 722 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":724
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */

#line 724 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":725
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */

#line 725 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":726
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */

#line 726 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/
struct __pyx_obj_7brainpy_2_c_11composition_PyComposition;
struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed;
struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex;
struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine;
struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern;
struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_MinimizeFitSelector;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_MaximizeFitSelector;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ChiSquareFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__;
struct __pyx_array_obj;
struct __pyx_MemviewEnum_obj;
struct __pyx_memoryview_obj;
struct __pyx_memoryviewslice_obj;
struct __pyx_opt_args_7brainpy_2_c_11composition_neutral_mass;
struct __pyx_opt_args_7brainpy_2_c_11composition_mass_charge_ratio;
struct __pyx_t_7brainpy_2_c_11composition_Isotope;
struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap;
struct __pyx_t_7brainpy_2_c_11composition_Element;
struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket;
struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable;
struct __pyx_t_7brainpy_2_c_11composition_Composition;
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl;
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition_pop;

/* "brainpy/_c/composition.pxd":10
 * cdef double PROTON = nist_mass["H+"][0][0]
 * 
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*) nogil             # <<<<<<<<<<<<<<
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*) nogil
 * cdef char* _parse_isotope_string(char* label, int* isotope_num, char* element_name) nogil
 */

#line 10 "brainpy/_c/composition.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_neutral_mass {

#line 10 "brainpy/_c/composition.pxd"
  int __pyx_n;

#line 10 "brainpy/_c/composition.pxd"
  double charge_carrier;

#line 10 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":11
 * 
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*) nogil
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*) nogil             # <<<<<<<<<<<<<<
 * cdef char* _parse_isotope_string(char* label, int* isotope_num, char* element_name) nogil
 * 
 */

#line 11 "brainpy/_c/composition.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_mass_charge_ratio {

#line 11 "brainpy/_c/composition.pxd"
  int __pyx_n;

#line 11 "brainpy/_c/composition.pxd"
  double charge_carrier;

#line 11 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":17
 * # Isotope and IsotopeMap Declarations
 * 
 * cdef struct Isotope:             # <<<<<<<<<<<<<<
 *     double mass
 *     double abundance
 */

#line 17 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_Isotope {

#line 17 "brainpy/_c/composition.pxd"
  double mass;

#line 17 "brainpy/_c/composition.pxd"
  double abundance;

#line 17 "brainpy/_c/composition.pxd"
  int neutrons;

#line 17 "brainpy/_c/composition.pxd"
  int neutron_shift;

#line 17 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":23
 *     int neutron_shift
 * 
 * cdef struct IsotopeMap:             # <<<<<<<<<<<<<<
 *     Isotope* bins
 *     size_t size
 */

#line 23 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap {

#line 23 "brainpy/_c/composition.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Isotope *bins;

#line 23 "brainpy/_c/composition.pxd"
  size_t size;

#line 23 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":36
 * # Element Declarations
 * 
 * cdef struct Element:             # <<<<<<<<<<<<<<
 *     char* symbol
 *     IsotopeMap* isotopes
 */

#line 36 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_Element {

#line 36 "brainpy/_c/composition.pxd"
  char *symbol;

#line 36 "brainpy/_c/composition.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *isotopes;

#line 36 "brainpy/_c/composition.pxd"
  int monoisotopic_isotope_index;

#line 36 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":56
 * # ElementHashTable and ElementHashBucket Declarations
 * 
 * cdef struct ElementHashBucket:             # <<<<<<<<<<<<<<
 *     Element** elements
 *     size_t used
 */

#line 56 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket {

#line 56 "brainpy/_c/composition.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Element **elements;

#line 56 "brainpy/_c/composition.pxd"
  size_t used;

#line 56 "brainpy/_c/composition.pxd"
  size_t size;

#line 56 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":65
 * 
 * 
 * cdef struct ElementHashTable:             # <<<<<<<<<<<<<<
 *     ElementHashBucket* buckets
 *     size_t size
 */

#line 65 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable {

#line 65 "brainpy/_c/composition.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket *buckets;

#line 65 "brainpy/_c/composition.pxd"
  size_t size;

#line 65 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":94
 * # Composition Declarations
 * 
 * cdef struct Composition:             # <<<<<<<<<<<<<<
 *     char** elements
 *     count_type* counts
 */

#line 94 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_Composition {

#line 94 "brainpy/_c/composition.pxd"
  char **elements;

#line 94 "brainpy/_c/composition.pxd"
  __pyx_t_7brainpy_2_c_11composition_count_type *counts;

#line 94 "brainpy/_c/composition.pxd"
  size_t size;

#line 94 "brainpy/_c/composition.pxd"
  size_t used;

#line 94 "brainpy/_c/composition.pxd"
  int max_variants;

#line 94 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":130
 *     @staticmethod
 *     cdef PyComposition _create(Composition* base)
 *     cdef void _set_impl(self, Composition* composition, bint free_existing=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _initialize_from_formula(self, str formula)
 */

#line 130 "brainpy/_c/composition.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl {

#line 130 "brainpy/_c/composition.pxd"
  int __pyx_n;

#line 130 "brainpy/_c/composition.pxd"
  int free_existing;

#line 130 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":144
 *     cpdef list items(self)
 * 
 *     cpdef pop(self, str key, object default=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef count_type getitem(self, str key)
 */

#line 144 "brainpy/_c/composition.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition_pop {

#line 144 "brainpy/_c/composition.pxd"
  int __pyx_n;

#line 144 "brainpy/_c/composition.pxd"
  PyObject *__pyx_default;

#line 144 "brainpy/_c/composition.pxd"
};

#line 144 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector;

/* "brainpy/_c/double_vector.pxd":1
 * cdef struct DoubleVector:             # <<<<<<<<<<<<<<
 *     double* v
 *     size_t used
 */

#line 1 "brainpy/_c/double_vector.pxd"
struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector {

#line 1 "brainpy/_c/double_vector.pxd"
  double *v;

#line 1 "brainpy/_c/double_vector.pxd"
  size_t used;

#line 1 "brainpy/_c/double_vector.pxd"
  size_t size;

#line 1 "brainpy/_c/double_vector.pxd"
};

#line 1 "brainpy/_c/double_vector.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters;

#line 1 "brainpy/_c/double_vector.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants;

#line 1 "brainpy/_c/double_vector.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants;

/* "brainpy/_c/isotopic_constants.pxd":6
 * from brainpy._c.double_vector cimport DoubleVector
 * 
 * ctypedef DoubleVector dvec             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 6 "brainpy/_c/isotopic_constants.pxd"
typedef struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector __pyx_t_7brainpy_2_c_18isotopic_constants_dvec;

/* "brainpy/_c/isotopic_constants.pxd":9
 * 
 * 
 * cdef struct PolynomialParameters:             # <<<<<<<<<<<<<<
 *     dvec* elementary_symmetric_polynomial
 *     dvec* power_sum
 */

#line 9 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters {

#line 9 "brainpy/_c/isotopic_constants.pxd"
  __pyx_t_7brainpy_2_c_18isotopic_constants_dvec *elementary_symmetric_polynomial;

#line 9 "brainpy/_c/isotopic_constants.pxd"
  __pyx_t_7brainpy_2_c_18isotopic_constants_dvec *power_sum;

#line 9 "brainpy/_c/isotopic_constants.pxd"
};

/* "brainpy/_c/isotopic_constants.pxd":14
 * 
 * 
 * cdef struct PhiConstants:             # <<<<<<<<<<<<<<
 *     int order
 *     Element* element
 */

#line 14 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants {

#line 14 "brainpy/_c/isotopic_constants.pxd"
  int order;

#line 14 "brainpy/_c/isotopic_constants.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Element *element;

#line 14 "brainpy/_c/isotopic_constants.pxd"
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters *element_coefficients;

#line 14 "brainpy/_c/isotopic_constants.pxd"
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters *mass_coefficients;

#line 14 "brainpy/_c/isotopic_constants.pxd"
};

/* "brainpy/_c/isotopic_constants.pxd":21
 * 
 * 
 * cdef struct IsotopicConstants:             # <<<<<<<<<<<<<<
 *     int order
 *     PhiConstants** constants
 */

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants {

#line 21 "brainpy/_c/isotopic_constants.pxd"
  int order;

#line 21 "brainpy/_c/isotopic_constants.pxd"
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants **constants;

#line 21 "brainpy/_c/isotopic_constants.pxd"
  size_t size;

#line 21 "brainpy/_c/isotopic_constants.pxd"
  size_t used;

#line 21 "brainpy/_c/isotopic_constants.pxd"
};

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_guess_npeaks;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_isotopic_variants;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_max_variants_approx;

/* "brainpy/_c/isotopic_distribution.pxd":7
 * 
 * 
 * cdef struct Peak:             # <<<<<<<<<<<<<<
 *     double mz
 *     double intensity
 */

#line 7 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak {

#line 7 "brainpy/_c/isotopic_distribution.pxd"
  double mz;

#line 7 "brainpy/_c/isotopic_distribution.pxd"
  double intensity;

#line 7 "brainpy/_c/isotopic_distribution.pxd"
  int charge;

#line 7 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":13
 * 
 * 
 * cdef struct PeakList:             # <<<<<<<<<<<<<<
 *     Peak* peaks
 *     size_t used
 */

#line 13 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList {

#line 13 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *peaks;

#line 13 "brainpy/_c/isotopic_distribution.pxd"
  size_t used;

#line 13 "brainpy/_c/isotopic_distribution.pxd"
  size_t size;

#line 13 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":35
 * 
 * 
 * cdef struct ElementCache:             # <<<<<<<<<<<<<<
 *     Element** elements
 *     size_t used
 */

#line 35 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache {

#line 35 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Element **elements;

#line 35 "brainpy/_c/isotopic_distribution.pxd"
  size_t used;

#line 35 "brainpy/_c/isotopic_distribution.pxd"
  size_t size;

#line 35 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *source;

#line 35 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":49
 * 
 * 
 * cdef struct IsotopicDistribution:             # <<<<<<<<<<<<<<
 *     Composition* composition
 *     IsotopicConstants* _isotopic_constants
 */

#line 49 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution {

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Composition *composition;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *_isotopic_constants;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  int order;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  double average_mass;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *monoisotopic_peak;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *cache;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":58
 * 
 * 
 * cdef IsotopicDistribution* make_isotopic_distribution(Composition* composition, int order, ElementCache* cache=*) nogil             # <<<<<<<<<<<<<<
 * cdef void free_isotopic_distribution(IsotopicDistribution* distribution)  nogil
 * 
 */

#line 58 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution {

#line 58 "brainpy/_c/isotopic_distribution.pxd"
  int __pyx_n;

#line 58 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *cache;

#line 58 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":77
 * 
 * cdef int max_variants(Composition* composition, ElementCache* cache) nogil
 * cdef int guess_npeaks(Composition* composition_struct, size_t max_npeaks, ElementCache* cache=*) nogil             # <<<<<<<<<<<<<<
 * 
 * cpdef list _isotopic_variants(object composition, object npeaks=*, int charge=*, double charge_carrier=*)
 */

#line 77 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_guess_npeaks {

#line 77 "brainpy/_c/isotopic_distribution.pxd"
  int __pyx_n;

#line 77 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *cache;

#line 77 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":79
 * cdef int guess_npeaks(Composition* composition_struct, size_t max_npeaks, ElementCache* cache=*) nogil
 * 
 * cpdef list _isotopic_variants(object composition, object npeaks=*, int charge=*, double charge_carrier=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 79 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants {

#line 79 "brainpy/_c/isotopic_distribution.pxd"
  int __pyx_n;

#line 79 "brainpy/_c/isotopic_distribution.pxd"
  PyObject *npeaks;

#line 79 "brainpy/_c/isotopic_distribution.pxd"
  int charge;

#line 79 "brainpy/_c/isotopic_distribution.pxd"
  double charge_carrier;

#line 79 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":82
 * 
 * 
 * cdef PeakList* isotopic_variants(Composition* composition, int npeaks, int charge=*, double charge_carrier=*) nogil             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 82 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_isotopic_variants {

#line 82 "brainpy/_c/isotopic_distribution.pxd"
  int __pyx_n;

#line 82 "brainpy/_c/isotopic_distribution.pxd"
  int charge;

#line 82 "brainpy/_c/isotopic_distribution.pxd"
  double charge_carrier;

#line 82 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":85
 * 
 * 
 * cdef size_t max_variants_approx(double mass, double lambda_factor=*, size_t maxiter=*, double threshold=*) nogil             # <<<<<<<<<<<<<<
 */

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_max_variants_approx {

#line 85 "brainpy/_c/isotopic_distribution.pxd"
  int __pyx_n;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
  double lambda_factor;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
  size_t maxiter;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
  double threshold;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
};

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted;

/* "ms_peak_picker/_c/peak_set.pxd":42
 *     cdef FittedPeak _get_nearest_peak(self, double mz, double* errout)
 * 
 *     cdef FittedPeak _has_peak(self, double mz, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef PeakSet clone(self)
 */

#line 42 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak {

#line 42 "ms_peak_picker/_c/peak_set.pxd"
  int __pyx_n;

#line 42 "ms_peak_picker/_c/peak_set.pxd"
  double tolerance;

#line 42 "ms_peak_picker/_c/peak_set.pxd"
};

/* "ms_peak_picker/_c/peak_set.pxd":50
 *     cdef int _between_bounds(self, double m1, double m2, size_t* startp, size_t* endp)
 * 
 *     cpdef tuple all_peaks_for(self, double mz, double error_tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef FittedPeak getitem(self, size_t i)
 */

#line 50 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for {

#line 50 "ms_peak_picker/_c/peak_set.pxd"
  int __pyx_n;

#line 50 "ms_peak_picker/_c/peak_set.pxd"
  double error_tolerance;

#line 50 "ms_peak_picker/_c/peak_set.pxd"
};

/* "ms_peak_picker/_c/peak_set.pxd":64
 * 
 * 
 * cdef struct index_cell:             # <<<<<<<<<<<<<<
 *     double center_value
 *     size_t start
 */

#line 64 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell {

#line 64 "ms_peak_picker/_c/peak_set.pxd"
  double center_value;

#line 64 "ms_peak_picker/_c/peak_set.pxd"
  size_t start;

#line 64 "ms_peak_picker/_c/peak_set.pxd"
  size_t end;

#line 64 "ms_peak_picker/_c/peak_set.pxd"
};

/* "ms_peak_picker/_c/peak_set.pxd":70
 * 
 * 
 * cdef struct index_list:             # <<<<<<<<<<<<<<
 *     index_cell* index
 *     size_t size
 */

#line 70 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list {

#line 70 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell *index;

#line 70 "ms_peak_picker/_c/peak_set.pxd"
  size_t size;

#line 70 "ms_peak_picker/_c/peak_set.pxd"
  double low;

#line 70 "ms_peak_picker/_c/peak_set.pxd"
  double high;

#line 70 "ms_peak_picker/_c/peak_set.pxd"
};

/* "ms_peak_picker/_c/peak_set.pxd":90
 *     cpdef _allocate_index(self, size_t interval_index_size)
 * 
 *     cpdef FittedPeak has_peak_hinted(self, double mz, double tolerance=*, size_t hint_start=*, size_t hint_end=*)             # <<<<<<<<<<<<<<
 * 
 * cdef size_t double_binary_search_ppm(double* array, double value, double tolerance, size_t n)
 */

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted {

#line 90 "ms_peak_picker/_c/peak_set.pxd"
  int __pyx_n;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
  double tolerance;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
  size_t hint_start;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
  size_t hint_end;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
};

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":728
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */

#line 728 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":729
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */

#line 729 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":730
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */

#line 730 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":732
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */

#line 732 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_cdouble __pyx_t_5numpy_complex_t;

#line 732 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex__has_peak;

#line 732 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex_all_peaks_for;

/* "ms_peak_picker/_c/peak_index.pxd":11
 *         public PeakSet peaks
 * 
 *     cdef FittedPeak _has_peak(self, double mz, double tolerance=*)             # <<<<<<<<<<<<<<
 *     cdef PeakSet _between(self, double start, double stop)
 * 
 */

#line 11 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex__has_peak {

#line 11 "ms_peak_picker/_c/peak_index.pxd"
  int __pyx_n;

#line 11 "ms_peak_picker/_c/peak_index.pxd"
  double tolerance;

#line 11 "ms_peak_picker/_c/peak_index.pxd"
};

/* "ms_peak_picker/_c/peak_index.pxd":14
 *     cdef PeakSet _between(self, double start, double stop)
 * 
 *     cpdef tuple all_peaks_for(self, double mz, double error_tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef size_t get_size(self)
 */

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex_all_peaks_for {

#line 14 "ms_peak_picker/_c/peak_index.pxd"
  int __pyx_n;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
  double error_tolerance;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
};

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_mass_charge_ratio;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_scale;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_isotopic_cluster;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine__isotopic_cluster;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_ignore_below;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_truncate_after;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_scale;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern__scale;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_has_mz_charge_pair;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_isotopic_cluster;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_isotopic_shift;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_poisson_approximate;

/* "ms_deisotope/_c/averagine.pxd":7
 * cdef double PROTON
 * 
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*)             # <<<<<<<<<<<<<<
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*)
 * cdef void slide(double mz, list peaklist)
 */

#line 7 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_mass_charge_ratio {

#line 7 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 7 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 7 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":8
 * 
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*)
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*)             # <<<<<<<<<<<<<<
 * cdef void slide(double mz, list peaklist)
 * cdef dict scale_dict(dict data, double factor)
 */

#line 8 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass {

#line 8 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 8 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 8 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":17
 *         public dict base_composition
 * 
 *     cpdef dict scale(self, double mz, int charge=*, double charge_carrier=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 *     cdef TheoreticalIsotopicPattern _isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 */

#line 17 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_scale {

#line 17 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 17 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 17 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 17 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":18
 * 
 *     cpdef dict scale(self, double mz, int charge=*, double charge_carrier=*)
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 *     cdef TheoreticalIsotopicPattern _isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 * 
 */

#line 18 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_isotopic_cluster {

#line 18 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 18 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 18 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 18 "ms_deisotope/_c/averagine.pxd"
  double truncate_after;

#line 18 "ms_deisotope/_c/averagine.pxd"
  double ignore_below;

#line 18 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":19
 *     cpdef dict scale(self, double mz, int charge=*, double charge_carrier=*)
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 *     cdef TheoreticalIsotopicPattern _isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 19 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine__isotopic_cluster {

#line 19 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 19 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 19 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 19 "ms_deisotope/_c/averagine.pxd"
  double truncate_after;

#line 19 "ms_deisotope/_c/averagine.pxd"
  double ignore_below;

#line 19 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":39
 *     cdef bint _eq_inst(self, TheoreticalIsotopicPattern other)
 * 
 *     cpdef TheoreticalIsotopicPattern ignore_below(self, double ignore_below=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern truncate_after(self, double truncate_after=*)
 *     cpdef TheoreticalIsotopicPattern shift(self, double mz)
 */

#line 39 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_ignore_below {

#line 39 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 39 "ms_deisotope/_c/averagine.pxd"
  double ignore_below;

#line 39 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":40
 * 
 *     cpdef TheoreticalIsotopicPattern ignore_below(self, double ignore_below=*)
 *     cpdef TheoreticalIsotopicPattern truncate_after(self, double truncate_after=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern shift(self, double mz)
 *     cpdef TheoreticalIsotopicPattern scale(self, list experimental_distribution, str method=*)
 */

#line 40 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_truncate_after {

#line 40 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 40 "ms_deisotope/_c/averagine.pxd"
  double truncate_after;

#line 40 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":42
 *     cpdef TheoreticalIsotopicPattern truncate_after(self, double truncate_after=*)
 *     cpdef TheoreticalIsotopicPattern shift(self, double mz)
 *     cpdef TheoreticalIsotopicPattern scale(self, list experimental_distribution, str method=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern scale_raw(self, double scale_factor)
 *     cpdef double drop_last_peak(self)
 */

#line 42 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_scale {

#line 42 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 42 "ms_deisotope/_c/averagine.pxd"
  PyObject *method;

#line 42 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":53
 * 
 *     @cython.final
 *     cdef inline TheoreticalIsotopicPattern _scale(self, list experimental_distribution, str method=*)             # <<<<<<<<<<<<<<
 * 
 *     @cython.final
 */

#line 53 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern__scale {

#line 53 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 53 "ms_deisotope/_c/averagine.pxd"
  PyObject *method;

#line 53 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":88
 *     cdef double _make_cache_key(self, double mz)
 * 
 *     cdef TheoreticalIsotopicPattern has_mz_charge_pair(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 * 
 */

#line 88 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_has_mz_charge_pair {

#line 88 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 88 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 88 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 88 "ms_deisotope/_c/averagine.pxd"
  double truncate_after;

#line 88 "ms_deisotope/_c/averagine.pxd"
  double ignore_below;

#line 88 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":89
 * 
 *     cdef TheoreticalIsotopicPattern has_mz_charge_pair(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 89 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_isotopic_cluster {

#line 89 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 89 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 89 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 89 "ms_deisotope/_c/averagine.pxd"
  double truncate_after;

#line 89 "ms_deisotope/_c/averagine.pxd"
  double ignore_below;

#line 89 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":92
 * 
 * 
 * cpdef double isotopic_shift(int charge=*)             # <<<<<<<<<<<<<<
 * 
 * cpdef TheoreticalIsotopicPattern poisson_approximate(double mass, size_t n_peaks, double lambda_factor=*, int charge=*)
 */

#line 92 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_isotopic_shift {

#line 92 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 92 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 92 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":94
 * cpdef double isotopic_shift(int charge=*)
 * 
 * cpdef TheoreticalIsotopicPattern poisson_approximate(double mass, size_t n_peaks, double lambda_factor=*, int charge=*)             # <<<<<<<<<<<<<<
 */

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_poisson_approximate {

#line 94 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 94 "ms_deisotope/_c/averagine.pxd"
  double lambda_factor;

#line 94 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 94 "ms_deisotope/_c/averagine.pxd"
};

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base__get_all_peak_charge_pairs;

/* "ms_deisotope/_c/deconvoluter_base.pxd":42
 *     cdef FittedPeak _has_peak(self, double mz, double error_tolerance)
 * 
 *     cpdef list match_theoretical_isotopic_distribution(self, list theoretical_distribution, double error_tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef scale_theoretical_distribution(self, TheoreticalIsotopicPattern theoretical_distribution, list experimental_distribution)
 *     cpdef IsotopicFitRecord _evaluate_theoretical_distribution(self, list experimental, TheoreticalIsotopicPattern theoretical, FittedPeak peak, int charge)
 */

#line 42 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution {

#line 42 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 42 "ms_deisotope/_c/deconvoluter_base.pxd"
  double error_tolerance;

#line 42 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":45
 *     cpdef scale_theoretical_distribution(self, TheoreticalIsotopicPattern theoretical_distribution, list experimental_distribution)
 *     cpdef IsotopicFitRecord _evaluate_theoretical_distribution(self, list experimental, TheoreticalIsotopicPattern theoretical, FittedPeak peak, int charge)
 *     cpdef subtraction(self, TheoreticalIsotopicPattern isotopic_cluster, double error_tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef list fit_incremental_truncation(self, IsotopicFitRecord seed_fit, double lower_bound)
 * 
 */

#line 45 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction {

#line 45 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 45 "ms_deisotope/_c/deconvoluter_base.pxd"
  double error_tolerance;

#line 45 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":50
 *     cpdef bint _check_fit(self, IsotopicFitRecord fit)
 * 
 *     cpdef list _find_next_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)
 * 
 */

#line 50 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak {

#line 50 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 50 "ms_deisotope/_c/deconvoluter_base.pxd"
  int step;

#line 50 "ms_deisotope/_c/deconvoluter_base.pxd"
  double tolerance;

#line 50 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":51
 * 
 *     cpdef list _find_next_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)
 *     cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)
 */

#line 51 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak {

#line 51 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 51 "ms_deisotope/_c/deconvoluter_base.pxd"
  int step;

#line 51 "ms_deisotope/_c/deconvoluter_base.pxd"
  double tolerance;

#line 51 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":53
 *     cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)
 * 
 *     cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 *     cdef int _find_previous_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)
 * 
 */

#line 53 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace {

#line 53 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 53 "ms_deisotope/_c/deconvoluter_base.pxd"
  int step;

#line 53 "ms_deisotope/_c/deconvoluter_base.pxd"
  double tolerance;

#line 53 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":54
 * 
 *     cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)
 *     cdef int _find_previous_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 54 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace {

#line 54 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 54 "ms_deisotope/_c/deconvoluter_base.pxd"
  int step;

#line 54 "ms_deisotope/_c/deconvoluter_base.pxd"
  double tolerance;

#line 54 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":61
 *         public AveragineCache averagine
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,             # <<<<<<<<<<<<<<
 *                                                          double charge_carrier=*, double truncate_after=*,
 *                                                          double ignore_below=*)
 */

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution {

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
  double charge_carrier;

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
  double truncate_after;

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
  double ignore_below;

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":65
 *                                                          double ignore_below=*)
 * 
 *     cpdef set _fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=*, double truncate_after=*,             # <<<<<<<<<<<<<<
 *                                     double ignore_below=*)
 * 
 */

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges {

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
  double charge_carrier;

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
  double truncate_after;

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
  double ignore_below;

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":73
 *         public list averagines
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,             # <<<<<<<<<<<<<<
 *                                                          AveragineCache  averagine, double charge_carrier=*, double truncate_after=*,
 *                                                          double ignore_below=*)
 */

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution {

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
  double charge_carrier;

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
  double truncate_after;

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
  double ignore_below;

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":76
 *                                                          AveragineCache  averagine, double charge_carrier=*, double truncate_after=*,
 *                                                          double ignore_below=*)
 *     cpdef set _fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=*, double truncate_after=*,             # <<<<<<<<<<<<<<
 *                                     double ignore_below=*)
 * 
 */

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges {

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
  double charge_carrier;

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
  double truncate_after;

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
  double ignore_below;

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":83
 * 
 * 
 * cpdef set _get_all_peak_charge_pairs(DeconvoluterBase self, FittedPeak peak, double error_tolerance=*,             # <<<<<<<<<<<<<<
 *                                      object charge_range=*,
 *                                      int left_search_limit=*, int right_search_limit=*,
 */

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base__get_all_peak_charge_pairs {

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  double error_tolerance;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *charge_range;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  int left_search_limit;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  int right_search_limit;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  int recalculate_starting_peak;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  int use_quick_charge;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_isclose;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_ms_deconv_score_peak;

/* "ms_deisotope/_c/scoring.pxd":29
 * 
 *     @staticmethod
 *     cdef IsotopicFitRecord _create(FittedPeak seed_peak, double score, int charge, TheoreticalIsotopicPattern theoretical,             # <<<<<<<<<<<<<<
 *                                    list experimental, object data=*, int missed_peaks=*)
 * 
 */

#line 29 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create {

#line 29 "ms_deisotope/_c/scoring.pxd"
  int __pyx_n;

#line 29 "ms_deisotope/_c/scoring.pxd"
  PyObject *data;

#line 29 "ms_deisotope/_c/scoring.pxd"
  int missed_peaks;

#line 29 "ms_deisotope/_c/scoring.pxd"
};

/* "ms_deisotope/_c/scoring.pxd":75
 *     cdef public double mass_error_tolerance
 * 
 *     cdef double score_peak(self, FittedPeak obs, TheoreticalPeak theo, double mass_error_tolerance=*, double minimum_signal_to_noise=*) nogil             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 75 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak {

#line 75 "ms_deisotope/_c/scoring.pxd"
  int __pyx_n;

#line 75 "ms_deisotope/_c/scoring.pxd"
  double mass_error_tolerance;

#line 75 "ms_deisotope/_c/scoring.pxd"
  double minimum_signal_to_noise;

#line 75 "ms_deisotope/_c/scoring.pxd"
};

/* "ms_deisotope/_c/scoring.pxd":92
 *     cdef public PeakSet peaklist
 * 
 *     cdef double detect_interference(self, list experimental_peaks, double lower=*, double upper=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference {

#line 92 "ms_deisotope/_c/scoring.pxd"
  int __pyx_n;

#line 92 "ms_deisotope/_c/scoring.pxd"
  double lower;

#line 92 "ms_deisotope/_c/scoring.pxd"
  double upper;

#line 92 "ms_deisotope/_c/scoring.pxd"
};

/* "ms_deisotope/_c/scoring.pyx":17
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef bint isclose(double x, double y, double rtol=1.e-5, double atol=1.e-8):             # <<<<<<<<<<<<<<
 *     return abs(x-y) <= (atol + rtol * abs(y))
 * 
 */

#line 17 "ms_deisotope/_c/scoring.pyx"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_isclose {

#line 17 "ms_deisotope/_c/scoring.pyx"
  int __pyx_n;

#line 17 "ms_deisotope/_c/scoring.pyx"
  double rtol;

#line 17 "ms_deisotope/_c/scoring.pyx"
  double atol;

#line 17 "ms_deisotope/_c/scoring.pyx"
};

/* "ms_deisotope/_c/scoring.pyx":633
 * 
 * @cython.cdivision
 * cdef double ms_deconv_score_peak(FittedPeak obs, TheoreticalPeak theo, double mass_error_tolerance=0.02, double minimum_signal_to_noise=1) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         double mass_error, abundance_diff
 */

#line 633 "ms_deisotope/_c/scoring.pyx"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_ms_deconv_score_peak {

#line 633 "ms_deisotope/_c/scoring.pyx"
  int __pyx_n;

#line 633 "ms_deisotope/_c/scoring.pyx"
  double mass_error_tolerance;

#line 633 "ms_deisotope/_c/scoring.pyx"
  double minimum_signal_to_noise;

#line 633 "ms_deisotope/_c/scoring.pyx"
};

/* "brainpy/_c/composition.pxd":123
 * 
 * 
 * cdef class PyComposition(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* impl
 */

#line 123 "brainpy/_c/composition.pxd"
struct __pyx_obj_7brainpy_2_c_11composition_PyComposition {

#line 123 "brainpy/_c/composition.pxd"
  PyObject_HEAD

#line 123 "brainpy/_c/composition.pxd"
  struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition *__pyx_vtab;

#line 123 "brainpy/_c/composition.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Composition *impl;

#line 123 "brainpy/_c/composition.pxd"
  double cached_mass;

#line 123 "brainpy/_c/composition.pxd"
  int _clean;

#line 123 "brainpy/_c/composition.pxd"
};

#line 123 "brainpy/_c/composition.pxd"


/* "brainpy/_c/isotopic_distribution.pxd":63
 * 
 * @freelist(100000)
 * cdef class TheoreticalPeak(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 63 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak {

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  PyObject_HEAD

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_vtab;

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  double mz;

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  double intensity;

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  int charge;

#line 63 "brainpy/_c/isotopic_distribution.pxd"
};

#line 63 "brainpy/_c/isotopic_distribution.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":2
 * 
 * cdef class PeakBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 2 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase {

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  PyObject_HEAD

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_vtab;

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  double mz;

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  double intensity;

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  double area;

#line 2 "ms_peak_picker/_c/peak_set.pxd"
};

#line 2 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":12
 * 
 * 
 * cdef class FittedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double signal_to_noise
 */

#line 12 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak {

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  double signal_to_noise;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  double full_width_at_half_max;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  double left_width;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  double right_width;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  long peak_count;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  long index;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
};

#line 12 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":29
 *                             double area)
 * 
 * cdef class PeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */

#line 29 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet {

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  PyObject_HEAD

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_vtab;

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  PyObject *peaks;

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  int indexed;

#line 29 "ms_peak_picker/_c/peak_set.pxd"
};

#line 29 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":83
 * 
 * 
 * cdef class PeakSetIndexed(PeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* mz_index
 */

#line 83 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed {

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet __pyx_base;

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  double *mz_index;

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list *interval_index;

#line 83 "ms_peak_picker/_c/peak_set.pxd"
};

#line 83 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_index.pxd":5
 * 
 * 
 * cdef class PeakIndex(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public np.ndarray mz_array
 */

#line 5 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex {

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  PyObject_HEAD

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_vtab;

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  PyArrayObject *mz_array;

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  PyArrayObject *intensity_array;

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *peaks;

#line 5 "ms_peak_picker/_c/peak_index.pxd"
};

#line 5 "ms_peak_picker/_c/peak_index.pxd"


/* "ms_deisotope/_c/averagine.pxd":12
 * cdef dict scale_dict(dict data, double factor)
 * 
 * cdef class Averagine(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double base_mass
 */

#line 12 "ms_deisotope/_c/averagine.pxd"
struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine {

#line 12 "ms_deisotope/_c/averagine.pxd"
  PyObject_HEAD

#line 12 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine *__pyx_vtab;

#line 12 "ms_deisotope/_c/averagine.pxd"
  double base_mass;

#line 12 "ms_deisotope/_c/averagine.pxd"
  PyObject *base_composition;

#line 12 "ms_deisotope/_c/averagine.pxd"
};

#line 12 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/averagine.pxd":23
 * 
 * @cython.final
 * cdef class TheoreticalIsotopicPattern(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list peaklist
 */

#line 23 "ms_deisotope/_c/averagine.pxd"
struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern {

#line 23 "ms_deisotope/_c/averagine.pxd"
  PyObject_HEAD

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_vtab;

#line 23 "ms_deisotope/_c/averagine.pxd"
  PyObject *peaklist;

#line 23 "ms_deisotope/_c/averagine.pxd"
  double origin;

#line 23 "ms_deisotope/_c/averagine.pxd"
  double offset;

#line 23 "ms_deisotope/_c/averagine.pxd"
};

#line 23 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/averagine.pxd":79
 * 
 * @cython.final
 * cdef class AveragineCache(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public dict backend
 */

#line 79 "ms_deisotope/_c/averagine.pxd"
struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache {

#line 79 "ms_deisotope/_c/averagine.pxd"
  PyObject_HEAD

#line 79 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache *__pyx_vtab;

#line 79 "ms_deisotope/_c/averagine.pxd"
  PyObject *backend;

#line 79 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *averagine;

#line 79 "ms_deisotope/_c/averagine.pxd"
  double cache_truncation;

#line 79 "ms_deisotope/_c/averagine.pxd"
  int enabled;

#line 79 "ms_deisotope/_c/averagine.pxd"
};

#line 79 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":25
 * 
 * 
 * cdef class DeconvoluterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public bint use_subtraction
 */

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase {

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject_HEAD

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_vtab;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int use_subtraction;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *scale_method;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int merge_isobaric_peaks;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  double minimum_intensity;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *peaklist;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *scorer;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int verbose;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *_slice_cache;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":57
 * 
 * 
 * cdef class AveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public AveragineCache averagine
 */

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase {

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *averagine;

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":69
 * 
 * 
 * cdef class MultiAveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list averagines
 */

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase {

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *averagines;

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":91
 * 
 * 
 * cdef class ChargeIterator(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public int lower
 */

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator {

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject_HEAD

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_vtab;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int lower;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int upper;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int sign;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int *values;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  size_t size;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  size_t index;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/scoring.pxd":16
 * 
 * 
 * cdef class IsotopicFitRecord(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FittedPeak seed_peak
 */

#line 16 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord {

#line 16 "ms_deisotope/_c/scoring.pxd"
  PyObject_HEAD

#line 16 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_vtab;

#line 16 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *seed_peak;

#line 16 "ms_deisotope/_c/scoring.pxd"
  double score;

#line 16 "ms_deisotope/_c/scoring.pxd"
  int charge;

#line 16 "ms_deisotope/_c/scoring.pxd"
  PyObject *experimental;

#line 16 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *theoretical;

#line 16 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *monoisotopic_peak;

#line 16 "ms_deisotope/_c/scoring.pxd"
  int missed_peaks;

#line 16 "ms_deisotope/_c/scoring.pxd"
  PyObject *data;

#line 16 "ms_deisotope/_c/scoring.pxd"
  Py_hash_t _hash;

#line 16 "ms_deisotope/_c/scoring.pxd"
};

#line 16 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":38
 * 
 * 
 * cdef class FitSelectorBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double minimum_score
 */

#line 38 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase {

#line 38 "ms_deisotope/_c/scoring.pxd"
  PyObject_HEAD

#line 38 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_vtab;

#line 38 "ms_deisotope/_c/scoring.pxd"
  double minimum_score;

#line 38 "ms_deisotope/_c/scoring.pxd"
};

#line 38 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":50
 * 
 * 
 * cdef class IsotopicFitterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FitSelectorBase select
 */

#line 50 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase {

#line 50 "ms_deisotope/_c/scoring.pxd"
  PyObject_HEAD

#line 50 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_vtab;

#line 50 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *select;

#line 50 "ms_deisotope/_c/scoring.pxd"
};

#line 50 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":62
 * 
 * 
 * cdef class LeastSquaresFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

#line 62 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter {

#line 62 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 62 "ms_deisotope/_c/scoring.pxd"
};

#line 62 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":66
 * 
 * 
 * cdef class GTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

#line 66 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter {

#line 66 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 66 "ms_deisotope/_c/scoring.pxd"
};

#line 66 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":69
 *     pass
 * 
 * cdef class ScaledGTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

#line 69 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter {

#line 69 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 69 "ms_deisotope/_c/scoring.pxd"
};

#line 69 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":72
 *     pass
 * 
 * cdef class MSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef public double mass_error_tolerance
 * 
 */

#line 72 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter {

#line 72 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 72 "ms_deisotope/_c/scoring.pxd"
  double mass_error_tolerance;

#line 72 "ms_deisotope/_c/scoring.pxd"
};

#line 72 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":78
 * 
 * 
 * cdef class PenalizedMSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mass_error_tolerance
 */

#line 78 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter {

#line 78 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 78 "ms_deisotope/_c/scoring.pxd"
  double mass_error_tolerance;

#line 78 "ms_deisotope/_c/scoring.pxd"
  double penalty_factor;

#line 78 "ms_deisotope/_c/scoring.pxd"
};

#line 78 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":84
 * 
 * 
 * cdef class FunctionScorer(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object function
 */

#line 84 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer {

#line 84 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 84 "ms_deisotope/_c/scoring.pxd"
  PyObject *function;

#line 84 "ms_deisotope/_c/scoring.pxd"
};

#line 84 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":89
 * 
 * 
 * cdef class InterferenceDetection(object):             # <<<<<<<<<<<<<<
 *     cdef public PeakSet peaklist
 * 
 */

#line 89 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection {

#line 89 "ms_deisotope/_c/scoring.pxd"
  PyObject_HEAD

#line 89 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_vtab;

#line 89 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *peaklist;

#line 89 "ms_deisotope/_c/scoring.pxd"
};

#line 89 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":95
 * 
 * 
 * cdef class DistinctPatternFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public InterferenceDetection interference_detector
 */

#line 95 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter {

#line 95 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 95 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *interference_detector;

#line 95 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *g_test_scaled;

#line 95 "ms_deisotope/_c/scoring.pxd"
  double peak_count_scale;

#line 95 "ms_deisotope/_c/scoring.pxd"
  double domain_scale;

#line 95 "ms_deisotope/_c/scoring.pxd"
};

#line 95 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":102
 *         public double domain_scale
 * 
 * cdef class ScaledPenalizedMSDeconvFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double scale_factor
 */

#line 102 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter {

#line 102 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 102 "ms_deisotope/_c/scoring.pxd"
  double scale_factor;

#line 102 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *scorer;

#line 102 "ms_deisotope/_c/scoring.pxd"
};

#line 102 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":112
 * 
 * 
 * cdef class DotProductFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 */

#line 112 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter {

#line 112 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 112 "ms_deisotope/_c/scoring.pxd"
};

#line 112 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pyx":199
 * 
 * 
 * cdef class MinimizeFitSelector(FitSelectorBase):             # <<<<<<<<<<<<<<
 *     """A FitSelector which tries to minimize the score of the best fit.
 *     """
 */

#line 199 "ms_deisotope/_c/scoring.pyx"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_MinimizeFitSelector {

#line 199 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase __pyx_base;

#line 199 "ms_deisotope/_c/scoring.pyx"
};

#line 199 "ms_deisotope/_c/scoring.pyx"


/* "ms_deisotope/_c/scoring.pyx":262
 * 
 * 
 * cdef class MaximizeFitSelector(FitSelectorBase):             # <<<<<<<<<<<<<<
 *     """A FitSelector which tries to maximize the score of the best fit.
 *     """
 */

#line 262 "ms_deisotope/_c/scoring.pyx"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_MaximizeFitSelector {

#line 262 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase __pyx_base;

#line 262 "ms_deisotope/_c/scoring.pyx"
};

#line 262 "ms_deisotope/_c/scoring.pyx"


/* "ms_deisotope/_c/scoring.pyx":615
 * 
 * 
 * cdef class ChiSquareFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision
 */

#line 615 "ms_deisotope/_c/scoring.pyx"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ChiSquareFitter {

#line 615 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 615 "ms_deisotope/_c/scoring.pyx"
};

#line 615 "ms_deisotope/_c/scoring.pyx"


/* "ms_deisotope/_c/scoring.pyx":146
 *         return self._hash
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         yield self.score
 *         yield self.charge
 */

#line 146 "ms_deisotope/_c/scoring.pyx"
struct __pyx_obj_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__ {

#line 146 "ms_deisotope/_c/scoring.pyx"
  PyObject_HEAD

#line 146 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self;

#line 146 "ms_deisotope/_c/scoring.pyx"
};

#line 146 "ms_deisotope/_c/scoring.pyx"


/* "View.MemoryView":106
 * 
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

#line 106 "View.MemoryView"
struct __pyx_array_obj {

#line 106 "View.MemoryView"
  PyObject_HEAD

#line 106 "View.MemoryView"
  struct __pyx_vtabstruct_array *__pyx_vtab;

#line 106 "View.MemoryView"
  char *data;

#line 106 "View.MemoryView"
  Py_ssize_t len;

#line 106 "View.MemoryView"
  char *format;

#line 106 "View.MemoryView"
  int ndim;

#line 106 "View.MemoryView"
  Py_ssize_t *_shape;

#line 106 "View.MemoryView"
  Py_ssize_t *_strides;

#line 106 "View.MemoryView"
  Py_ssize_t itemsize;

#line 106 "View.MemoryView"
  PyObject *mode;

#line 106 "View.MemoryView"
  PyObject *_format;

#line 106 "View.MemoryView"
  void (*callback_free_data)(void *);

#line 106 "View.MemoryView"
  int free_data;

#line 106 "View.MemoryView"
  int dtype_is_object;

#line 106 "View.MemoryView"
};

#line 106 "View.MemoryView"


/* "View.MemoryView":280
 * 
 * @cname('__pyx_MemviewEnum')
 * cdef class Enum(object):             # <<<<<<<<<<<<<<
 *     cdef object name
 *     def __init__(self, name):
 */

#line 280 "View.MemoryView"
struct __pyx_MemviewEnum_obj {

#line 280 "View.MemoryView"
  PyObject_HEAD

#line 280 "View.MemoryView"
  PyObject *name;

#line 280 "View.MemoryView"
};

#line 280 "View.MemoryView"


/* "View.MemoryView":331
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */

#line 331 "View.MemoryView"
struct __pyx_memoryview_obj {

#line 331 "View.MemoryView"
  PyObject_HEAD

#line 331 "View.MemoryView"
  struct __pyx_vtabstruct_memoryview *__pyx_vtab;

#line 331 "View.MemoryView"
  PyObject *obj;

#line 331 "View.MemoryView"
  PyObject *_size;

#line 331 "View.MemoryView"
  PyObject *_array_interface;

#line 331 "View.MemoryView"
  PyThread_type_lock lock;

#line 331 "View.MemoryView"
  __pyx_atomic_int acquisition_count[2];

#line 331 "View.MemoryView"
  __pyx_atomic_int *acquisition_count_aligned_p;

#line 331 "View.MemoryView"
  Py_buffer view;

#line 331 "View.MemoryView"
  int flags;

#line 331 "View.MemoryView"
  int dtype_is_object;

#line 331 "View.MemoryView"
  __Pyx_TypeInfo *typeinfo;

#line 331 "View.MemoryView"
};

#line 331 "View.MemoryView"


/* "View.MemoryView":967
 * 
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */

#line 967 "View.MemoryView"
struct __pyx_memoryviewslice_obj {

#line 967 "View.MemoryView"
  struct __pyx_memoryview_obj __pyx_base;

#line 967 "View.MemoryView"
  __Pyx_memviewslice from_slice;

#line 967 "View.MemoryView"
  PyObject *from_object;

#line 967 "View.MemoryView"
  PyObject *(*to_object_func)(char *);

#line 967 "View.MemoryView"
  int (*to_dtype_func)(char *, PyObject *);

#line 967 "View.MemoryView"
};

#line 967 "View.MemoryView"


#line 967 "View.MemoryView"


/* "brainpy/_c/composition.pxd":123
 * 
 * 
 * cdef class PyComposition(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* impl
 */

#line 123 "brainpy/_c/composition.pxd"


#line 123 "brainpy/_c/composition.pxd"
struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition {

#line 123 "brainpy/_c/composition.pxd"
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *(*_create)(struct __pyx_t_7brainpy_2_c_11composition_Composition *);

#line 123 "brainpy/_c/composition.pxd"
  void (*_set_impl)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_t_7brainpy_2_c_11composition_Composition *, struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl *__pyx_optional_args);

#line 123 "brainpy/_c/composition.pxd"
  void (*_initialize_from_formula)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *);

#line 123 "brainpy/_c/composition.pxd"
  double (*mass)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  int (*__pyx___equality_pycomposition)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  int (*__pyx___equality_dict)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *(*copy)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  PyObject *(*update)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  PyObject *(*keys)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  PyObject *(*values)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  PyObject *(*items)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  PyObject *(*pop)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition_pop *__pyx_optional_args);

#line 123 "brainpy/_c/composition.pxd"
  __pyx_t_7brainpy_2_c_11composition_count_type (*getitem)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *);

#line 123 "brainpy/_c/composition.pxd"
  void (*setitem)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, __pyx_t_7brainpy_2_c_11composition_count_type);

#line 123 "brainpy/_c/composition.pxd"
  void (*increment)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, __pyx_t_7brainpy_2_c_11composition_count_type);

#line 123 "brainpy/_c/composition.pxd"
  void (*add_from)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *);

#line 123 "brainpy/_c/composition.pxd"
  void (*subtract_from)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *);

#line 123 "brainpy/_c/composition.pxd"
  void (*scale_by)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, long);

#line 123 "brainpy/_c/composition.pxd"
};

#line 123 "brainpy/_c/composition.pxd"
static struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition *__pyx_vtabptr_7brainpy_2_c_11composition_PyComposition;

#line 123 "brainpy/_c/composition.pxd"


/* "brainpy/_c/isotopic_distribution.pxd":63
 * 
 * @freelist(100000)
 * cdef class TheoreticalPeak(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 63 "brainpy/_c/isotopic_distribution.pxd"


#line 63 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak {

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *(*_create)(double, double, int);

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  int (*_eq)(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, int __pyx_skip_dispatch);

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *(*clone)(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, int __pyx_skip_dispatch);

#line 63 "brainpy/_c/isotopic_distribution.pxd"
};

#line 63 "brainpy/_c/isotopic_distribution.pxd"
static struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_vtabptr_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak;

#line 63 "brainpy/_c/isotopic_distribution.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":2
 * 
 * cdef class PeakBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 2 "ms_peak_picker/_c/peak_set.pxd"


#line 2 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase {

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*clone)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch);

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch);

#line 2 "ms_peak_picker/_c/peak_set.pxd"
};

#line 2 "ms_peak_picker/_c/peak_set.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase;

#line 2 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":12
 * 
 * 
 * cdef class FittedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double signal_to_noise
 */

#line 12 "ms_peak_picker/_c/peak_set.pxd"


#line 12 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak {

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  int (*_eq)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int __pyx_skip_dispatch);

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_create)(double, double, double, double, double, double, long, long, double);

#line 12 "ms_peak_picker/_c/peak_set.pxd"
};

#line 12 "ms_peak_picker/_c/peak_set.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak;

#line 12 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":29
 *                             double area)
 * 
 * cdef class PeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */

#line 29 "ms_peak_picker/_c/peak_set.pxd"


#line 29 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet {

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*_create)(PyObject *);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  PyObject *(*reindex)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  size_t (*_index)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_get_nearest_peak)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double *);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_has_peak)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak *__pyx_optional_args);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*clone)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*_between)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  int (*_between_bounds)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double, size_t *, size_t *);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  PyObject *(*all_peaks_for)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for *__pyx_optional_args);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*getitem)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, size_t);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  size_t (*get_size)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  int (*_eq)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
};

#line 29 "ms_peak_picker/_c/peak_set.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet;

#line 29 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":83
 * 
 * 
 * cdef class PeakSetIndexed(PeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* mz_index
 */

#line 83 "ms_peak_picker/_c/peak_set.pxd"


#line 83 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed {

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet __pyx_base;

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  PyObject *(*_allocate_index)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *, size_t, int __pyx_skip_dispatch);

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*has_peak_hinted)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted *__pyx_optional_args);

#line 83 "ms_peak_picker/_c/peak_set.pxd"
};

#line 83 "ms_peak_picker/_c/peak_set.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed;

#line 83 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_index.pxd":5
 * 
 * 
 * cdef class PeakIndex(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public np.ndarray mz_array
 */

#line 5 "ms_peak_picker/_c/peak_index.pxd"


#line 5 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex {

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_has_peak)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, double, struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex__has_peak *__pyx_optional_args);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*_between)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, double, double);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  PyObject *(*all_peaks_for)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex_all_peaks_for *__pyx_optional_args);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  size_t (*get_size)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *(*clone)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, int __pyx_skip_dispatch);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, int __pyx_skip_dispatch);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
};

#line 5 "ms_peak_picker/_c/peak_index.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_vtabptr_14ms_peak_picker_2_c_10peak_index_PeakIndex;

#line 5 "ms_peak_picker/_c/peak_index.pxd"


/* "ms_deisotope/_c/averagine.pxd":12
 * cdef dict scale_dict(dict data, double factor)
 * 
 * cdef class Averagine(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double base_mass
 */

#line 12 "ms_deisotope/_c/averagine.pxd"


#line 12 "ms_deisotope/_c/averagine.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine {

#line 12 "ms_deisotope/_c/averagine.pxd"
  PyObject *(*scale)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_scale *__pyx_optional_args);

#line 12 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*isotopic_cluster)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_isotopic_cluster *__pyx_optional_args);

#line 12 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*_isotopic_cluster)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *, double, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine__isotopic_cluster *__pyx_optional_args);

#line 12 "ms_deisotope/_c/averagine.pxd"
};

#line 12 "ms_deisotope/_c/averagine.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine *__pyx_vtabptr_12ms_deisotope_2_c_9averagine_Averagine;

#line 12 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/averagine.pxd":23
 * 
 * @cython.final
 * cdef class TheoreticalIsotopicPattern(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list peaklist
 */

#line 23 "ms_deisotope/_c/averagine.pxd"


#line 23 "ms_deisotope/_c/averagine.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern {

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*_create)(PyObject *, double, double);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*clone)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  int (*_eq_inst)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*ignore_below)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_ignore_below *__pyx_optional_args);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*truncate_after)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_truncate_after *__pyx_optional_args);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*shift)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*scale)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_scale *__pyx_optional_args);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*scale_raw)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  double (*drop_last_peak)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  double (*total)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*normalize)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*_scale)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern__scale *__pyx_optional_args);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*clone_shift)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *(*get)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, Py_ssize_t);

#line 23 "ms_deisotope/_c/averagine.pxd"
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);

#line 23 "ms_deisotope/_c/averagine.pxd"
  double (*get_monoisotopic_mz)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);

#line 23 "ms_deisotope/_c/averagine.pxd"
  PyObject *(*get_processed_peaks)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);

#line 23 "ms_deisotope/_c/averagine.pxd"
  size_t (*basepeak_index)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  PyObject *(*incremental_truncation)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*clone_drop_last)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);

#line 23 "ms_deisotope/_c/averagine.pxd"
};

#line 23 "ms_deisotope/_c/averagine.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_vtabptr_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern;

#line 23 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/averagine.pxd":79
 * 
 * @cython.final
 * cdef class AveragineCache(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public dict backend
 */

#line 79 "ms_deisotope/_c/averagine.pxd"


#line 79 "ms_deisotope/_c/averagine.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache {

#line 79 "ms_deisotope/_c/averagine.pxd"
  double (*_make_cache_key)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, double);

#line 79 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*has_mz_charge_pair)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, double, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_has_mz_charge_pair *__pyx_optional_args);

#line 79 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*isotopic_cluster)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_isotopic_cluster *__pyx_optional_args);

#line 79 "ms_deisotope/_c/averagine.pxd"
};

#line 79 "ms_deisotope/_c/averagine.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache *__pyx_vtabptr_12ms_deisotope_2_c_9averagine_AveragineCache;

#line 79 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":25
 * 
 * 
 * cdef class DeconvoluterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public bint use_subtraction
 */

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"


#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase {

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*between)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, double, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*has_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, double, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_has_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, double);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*match_theoretical_isotopic_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*scale_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*_evaluate_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*subtraction)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*fit_incremental_truncation)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, double, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int (*_check_fit)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*_find_next_putative_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*_find_previous_putative_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int (*_find_next_putative_peak_inplace)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int (*_find_previous_putative_peak_inplace)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":57
 * 
 * 
 * cdef class AveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public AveragineCache averagine
 */

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"


#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase {

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*fit_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution *__pyx_optional_args);

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*_fit_peaks_at_charges)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *, PyObject *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges *__pyx_optional_args);

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase;

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":69
 * 
 * 
 * cdef class MultiAveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list averagines
 */

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"


#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase {

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*fit_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, double, int, struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution *__pyx_optional_args);

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*_fit_peaks_at_charges)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *, PyObject *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges *__pyx_optional_args);

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase;

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":91
 * 
 * 
 * cdef class ChargeIterator(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public int lower
 */

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"


#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator {

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *(*_create)(int, int);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *(*from_tuple)(PyObject *);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *(*from_quickcharge)(PyObject *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*sequence_from_quickcharge)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int __pyx_skip_dispatch);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*make_sequence)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  void (*release_sequence)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  void (*set_bounds)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int, int);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int (*has_more)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*reset)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int (*get_next_value)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/scoring.pyx":22
 * 
 * @cython.freelist(50000)
 * cdef class IsotopicFitRecord(object):             # <<<<<<<<<<<<<<
 *     """Describes a single isotopic pattern fit, comparing how well an
 *     experimentally observed sequence of peaks matches a theoretical isotopic
 */

#line 22 "ms_deisotope/_c/scoring.pyx"


#line 22 "ms_deisotope/_c/scoring.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord {

#line 22 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*_create)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, double, int, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create *__pyx_optional_args);

#line 22 "ms_deisotope/_c/scoring.pyx"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 22 "ms_deisotope/_c/scoring.pyx"
  int (*_ne)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 22 "ms_deisotope/_c/scoring.pyx"
  int (*_lt)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 22 "ms_deisotope/_c/scoring.pyx"
  int (*_gt)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 22 "ms_deisotope/_c/scoring.pyx"
};

#line 22 "ms_deisotope/_c/scoring.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitRecord;

#line 22 "ms_deisotope/_c/scoring.pyx"


/* "ms_deisotope/_c/scoring.pyx":161
 * 
 * 
 * cdef class FitSelectorBase(object):             # <<<<<<<<<<<<<<
 *     """An object that controls the filtering and
 *     selection of IsotopicFitRecord
 */

#line 161 "ms_deisotope/_c/scoring.pyx"


#line 161 "ms_deisotope/_c/scoring.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase {

#line 161 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*best)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, PyObject *, int __pyx_skip_dispatch);

#line 161 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*_best_from_set)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, PyObject *);

#line 161 "ms_deisotope/_c/scoring.pyx"
  int (*reject)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 161 "ms_deisotope/_c/scoring.pyx"
  int (*reject_score)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, double, int __pyx_skip_dispatch);

#line 161 "ms_deisotope/_c/scoring.pyx"
  int (*is_maximizing)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, int __pyx_skip_dispatch);

#line 161 "ms_deisotope/_c/scoring.pyx"
};

#line 161 "ms_deisotope/_c/scoring.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FitSelectorBase;

#line 161 "ms_deisotope/_c/scoring.pyx"


/* "ms_deisotope/_c/scoring.pyx":323
 * 
 * 
 * cdef class IsotopicFitterBase(object):             # <<<<<<<<<<<<<<
 *     """A base class for Isotopic Pattern Fitters, objects
 *     which given a set of experimental peaks and a set of matching
 */

#line 323 "ms_deisotope/_c/scoring.pyx"


#line 323 "ms_deisotope/_c/scoring.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase {

#line 323 "ms_deisotope/_c/scoring.pyx"
  double (*_evaluate)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, PyObject *, PyObject *, int __pyx_skip_dispatch);

#line 323 "ms_deisotope/_c/scoring.pyx"
  int (*reject)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 323 "ms_deisotope/_c/scoring.pyx"
  int (*reject_score)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, double, int __pyx_skip_dispatch);

#line 323 "ms_deisotope/_c/scoring.pyx"
  int (*is_maximizing)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, int __pyx_skip_dispatch);

#line 323 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *(*_configure)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, int __pyx_skip_dispatch);

#line 323 "ms_deisotope/_c/scoring.pyx"
};

#line 323 "ms_deisotope/_c/scoring.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;

#line 323 "ms_deisotope/_c/scoring.pyx"


/* "ms_deisotope/_c/scoring.pyx":566
 * 
 * 
 * cdef class LeastSquaresFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     r"""Evaluate an isotopic fit using least squares coefficient of
 *     determination :math:`R^2`.
 */

#line 566 "ms_deisotope/_c/scoring.pyx"


#line 566 "ms_deisotope/_c/scoring.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_LeastSquaresFitter {

#line 566 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 566 "ms_deisotope/_c/scoring.pyx"
};

#line 566 "ms_deisotope/_c/scoring.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_LeastSquaresFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_LeastSquaresFitter;

#line 566 "ms_deisotope/_c/scoring.pyx"


/* "ms_deisotope/_c/scoring.pyx":470
 * 
 * 
 * cdef class GTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     r"""Evaluate an isotopic fit using a G-test
 * 
 */

#line 470 "ms_deisotope/_c/scoring.pyx"


#line 470 "ms_deisotope/_c/scoring.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_GTestFitter {

#line 470 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 470 "ms_deisotope/_c/scoring.pyx"
};

#line 470 "ms_deisotope/_c/scoring.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_GTestFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_GTestFitter;

#line 470 "ms_deisotope/_c/scoring.pyx"


/* "ms_deisotope/_c/scoring.pyx":502
 * 
 * 
 * cdef class ScaledGTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     r"""Evaluate an isotopic fit using a G-test after normalizing the
 *     list of experimental and theoretical peaks to both sum to 1.
 */

#line 502 "ms_deisotope/_c/scoring.pyx"


#line 502 "ms_deisotope/_c/scoring.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledGTestFitter {

#line 502 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 502 "ms_deisotope/_c/scoring.pyx"
};

#line 502 "ms_deisotope/_c/scoring.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledGTestFitter;

#line 502 "ms_deisotope/_c/scoring.pyx"


/* "ms_deisotope/_c/scoring.pyx":656
 * 
 * 
 * cdef class MSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     r"""An implementation of the scoring function used in :title-reference:`MSDeconV`
 * 
 */

#line 656 "ms_deisotope/_c/scoring.pyx"


#line 656 "ms_deisotope/_c/scoring.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MSDeconVFitter {

#line 656 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 656 "ms_deisotope/_c/scoring.pyx"
  double (*score_peak)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak *__pyx_optional_args);

#line 656 "ms_deisotope/_c/scoring.pyx"
};

#line 656 "ms_deisotope/_c/scoring.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MSDeconVFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_MSDeconVFitter;

#line 656 "ms_deisotope/_c/scoring.pyx"


/* "ms_deisotope/_c/scoring.pyx":722
 * 
 * 
 * cdef class PenalizedMSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     r"""An Isotopic Fitter which uses the :class:`MSDeconVFitter` score
 *     weighted by 1 - :attr:`penalty_factor` * :class:`ScaledGTestFitter` score
 */

#line 722 "ms_deisotope/_c/scoring.pyx"


#line 722 "ms_deisotope/_c/scoring.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter {

#line 722 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 722 "ms_deisotope/_c/scoring.pyx"
};

#line 722 "ms_deisotope/_c/scoring.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter;

#line 722 "ms_deisotope/_c/scoring.pyx"


/* "ms_deisotope/_c/scoring.pyx":781
 * 
 * 
 * cdef class FunctionScorer(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     """Use a user-provided Python function or callable object to evaluate an isotopic
 *     envelope fit.
 */

#line 781 "ms_deisotope/_c/scoring.pyx"


#line 781 "ms_deisotope/_c/scoring.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FunctionScorer {

#line 781 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 781 "ms_deisotope/_c/scoring.pyx"
};

#line 781 "ms_deisotope/_c/scoring.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FunctionScorer *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FunctionScorer;

#line 781 "ms_deisotope/_c/scoring.pyx"


/* "ms_deisotope/_c/scoring.pyx":812
 * 
 * 
 * cdef class InterferenceDetection(object):             # <<<<<<<<<<<<<<
 *     def __init__(self, PeakSet peaklist):
 *         self.peaklist = peaklist
 */

#line 812 "ms_deisotope/_c/scoring.pyx"


#line 812 "ms_deisotope/_c/scoring.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection {

#line 812 "ms_deisotope/_c/scoring.pyx"
  double (*detect_interference)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference *__pyx_optional_args);

#line 812 "ms_deisotope/_c/scoring.pyx"
};

#line 812 "ms_deisotope/_c/scoring.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_InterferenceDetection;

#line 812 "ms_deisotope/_c/scoring.pyx"


/* "ms_deisotope/_c/scoring.pyx":863
 *         return self.detect_interference(experimental_peaks, lower, upper)
 * 
 * cdef class DistinctPatternFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, minimum_score=0.3, peak_count_scale=1.5, domain_scale=100.):
 */

#line 863 "ms_deisotope/_c/scoring.pyx"


#line 863 "ms_deisotope/_c/scoring.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DistinctPatternFitter {

#line 863 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 863 "ms_deisotope/_c/scoring.pyx"
};

#line 863 "ms_deisotope/_c/scoring.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DistinctPatternFitter;

#line 863 "ms_deisotope/_c/scoring.pyx"


/* "ms_deisotope/_c/scoring.pyx":911
 * 
 * 
 * cdef class ScaledPenalizedMSDeconvFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self,  minimum_score=0.3, penalty_factor=1., mass_error_tolerance=0.02):
 */

#line 911 "ms_deisotope/_c/scoring.pyx"


#line 911 "ms_deisotope/_c/scoring.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter {

#line 911 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 911 "ms_deisotope/_c/scoring.pyx"
  double (*_calculate_scale_factor)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 911 "ms_deisotope/_c/scoring.pyx"
  void (*scale_fitted_peaks)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *, PyObject *, double);

#line 911 "ms_deisotope/_c/scoring.pyx"
  void (*scale_theoretical_peaks)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *, PyObject *, double);

#line 911 "ms_deisotope/_c/scoring.pyx"
};

#line 911 "ms_deisotope/_c/scoring.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter;

#line 911 "ms_deisotope/_c/scoring.pyx"


/* "ms_deisotope/_c/scoring.pyx":981
 * 
 * 
 * cdef class DotProductFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self,  minimum_score=100):
 */

#line 981 "ms_deisotope/_c/scoring.pyx"


#line 981 "ms_deisotope/_c/scoring.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DotProductFitter {

#line 981 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 981 "ms_deisotope/_c/scoring.pyx"
};

#line 981 "ms_deisotope/_c/scoring.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DotProductFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DotProductFitter;

#line 981 "ms_deisotope/_c/scoring.pyx"


/* "ms_deisotope/_c/scoring.pyx":199
 * 
 * 
 * cdef class MinimizeFitSelector(FitSelectorBase):             # <<<<<<<<<<<<<<
 *     """A FitSelector which tries to minimize the score of the best fit.
 *     """
 */

#line 199 "ms_deisotope/_c/scoring.pyx"


#line 199 "ms_deisotope/_c/scoring.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MinimizeFitSelector {

#line 199 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase __pyx_base;

#line 199 "ms_deisotope/_c/scoring.pyx"
};

#line 199 "ms_deisotope/_c/scoring.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MinimizeFitSelector *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_MinimizeFitSelector;

#line 199 "ms_deisotope/_c/scoring.pyx"


/* "ms_deisotope/_c/scoring.pyx":262
 * 
 * 
 * cdef class MaximizeFitSelector(FitSelectorBase):             # <<<<<<<<<<<<<<
 *     """A FitSelector which tries to maximize the score of the best fit.
 *     """
 */

#line 262 "ms_deisotope/_c/scoring.pyx"


#line 262 "ms_deisotope/_c/scoring.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MaximizeFitSelector {

#line 262 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase __pyx_base;

#line 262 "ms_deisotope/_c/scoring.pyx"
};

#line 262 "ms_deisotope/_c/scoring.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MaximizeFitSelector *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_MaximizeFitSelector;

#line 262 "ms_deisotope/_c/scoring.pyx"


/* "ms_deisotope/_c/scoring.pyx":615
 * 
 * 
 * cdef class ChiSquareFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision
 */

#line 615 "ms_deisotope/_c/scoring.pyx"


#line 615 "ms_deisotope/_c/scoring.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ChiSquareFitter {

#line 615 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 615 "ms_deisotope/_c/scoring.pyx"
};

#line 615 "ms_deisotope/_c/scoring.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ChiSquareFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ChiSquareFitter;

#line 615 "ms_deisotope/_c/scoring.pyx"


/* "View.MemoryView":106
 * 
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

#line 106 "View.MemoryView"


#line 106 "View.MemoryView"
struct __pyx_vtabstruct_array {

#line 106 "View.MemoryView"
  PyObject *(*get_memview)(struct __pyx_array_obj *);

#line 106 "View.MemoryView"
};

#line 106 "View.MemoryView"
static struct __pyx_vtabstruct_array *__pyx_vtabptr_array;

#line 106 "View.MemoryView"


/* "View.MemoryView":331
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */

#line 331 "View.MemoryView"


#line 331 "View.MemoryView"
struct __pyx_vtabstruct_memoryview {

#line 331 "View.MemoryView"
  char *(*get_item_pointer)(struct __pyx_memoryview_obj *, PyObject *);

#line 331 "View.MemoryView"
  PyObject *(*is_slice)(struct __pyx_memoryview_obj *, PyObject *);

#line 331 "View.MemoryView"
  PyObject *(*setitem_slice_assignment)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);

#line 331 "View.MemoryView"
  PyObject *(*setitem_slice_assign_scalar)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *);

#line 331 "View.MemoryView"
  PyObject *(*setitem_indexed)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);

#line 331 "View.MemoryView"
  PyObject *(*convert_item_to_object)(struct __pyx_memoryview_obj *, char *);

#line 331 "View.MemoryView"
  PyObject *(*assign_item_from_object)(struct __pyx_memoryview_obj *, char *, PyObject *);

#line 331 "View.MemoryView"
};

#line 331 "View.MemoryView"
static struct __pyx_vtabstruct_memoryview *__pyx_vtabptr_memoryview;

#line 331 "View.MemoryView"


/* "View.MemoryView":967
 * 
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */

#line 967 "View.MemoryView"


#line 967 "View.MemoryView"
struct __pyx_vtabstruct__memoryviewslice {

#line 967 "View.MemoryView"
  struct __pyx_vtabstruct_memoryview __pyx_base;

#line 967 "View.MemoryView"
};

#line 967 "View.MemoryView"
static struct __pyx_vtabstruct__memoryviewslice *__pyx_vtabptr__memoryviewslice;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if CYTHON_FAST_PYCALL
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif // CYTHON_FAST_PYCALL
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed);

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* set_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_set_iterator(PyObject* iterable, int is_set,
                                                  Py_ssize_t* p_orig_length, int* p_source_is_set);
static CYTHON_INLINE int __Pyx_set_iter_next(
        PyObject* iter_obj, Py_ssize_t orig_length,
        Py_ssize_t* ppos, PyObject **value,
        int source_is_set);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* BufferIndexError.proto */
static void __Pyx_RaiseBufferIndexError(int axis);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* DivInt[Py_ssize_t].proto */
static CYTHON_INLINE Py_ssize_t __Pyx_div_Py_ssize_t(Py_ssize_t, Py_ssize_t);

/* UnaryNegOverflows.proto */
#define UNARY_NEG_WOULD_OVERFLOW(x)\
        (((x) < 0) & ((unsigned long)(x) == 0-(unsigned long)(x)))

static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *); /*proto*/
/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* decode_c_string_utf16.proto */
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 0;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16LE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = -1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16BE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}

/* decode_c_string.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* DivInt[long].proto */
static CYTHON_INLINE long __Pyx_div_long(long, long);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


/* BufferStructDeclare.proto */
typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

/* MemviewSliceIsContig.proto */
static int __pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim);

/* OverlappingSlices.proto */
static int __pyx_slices_overlap(__Pyx_memviewslice *slice1,
                                __Pyx_memviewslice *slice2,
                                int ndim, size_t itemsize);

/* Capsule.proto */
static CYTHON_INLINE PyObject *__pyx_capsule_create(void *p, const char *sig);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* MemviewSliceCopyTemplate.proto */
static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object);

/* MemviewSliceInit.proto */
#define __Pyx_BUF_MAX_NDIMS %(BUF_MAX_NDIMS)d
#define __Pyx_MEMVIEW_DIRECT   1
#define __Pyx_MEMVIEW_PTR      2
#define __Pyx_MEMVIEW_FULL     4
#define __Pyx_MEMVIEW_CONTIG   8
#define __Pyx_MEMVIEW_STRIDED  16
#define __Pyx_MEMVIEW_FOLLOW   32
#define __Pyx_IS_C_CONTIG 1
#define __Pyx_IS_F_CONTIG 2
static int __Pyx_init_memviewslice(
                struct __pyx_memoryview_obj *memview,
                int ndim,
                __Pyx_memviewslice *memviewslice,
                int memview_is_new_reference);
static CYTHON_INLINE int __pyx_add_acquisition_count_locked(
    __pyx_atomic_int *acquisition_count, PyThread_type_lock lock);
static CYTHON_INLINE int __pyx_sub_acquisition_count_locked(
    __pyx_atomic_int *acquisition_count, PyThread_type_lock lock);
#define __pyx_get_slice_count_pointer(memview) (memview->acquisition_count_aligned_p)
#define __pyx_get_slice_count(memview) (*__pyx_get_slice_count_pointer(memview))
#define __PYX_INC_MEMVIEW(slice, have_gil) __Pyx_INC_MEMVIEW(slice, have_gil, __LINE__)
#define __PYX_XDEC_MEMVIEW(slice, have_gil) __Pyx_XDEC_MEMVIEW(slice, have_gil, __LINE__)
static CYTHON_INLINE void __Pyx_INC_MEMVIEW(__Pyx_memviewslice *, int, int);
static CYTHON_INLINE void __Pyx_XDEC_MEMVIEW(__Pyx_memviewslice *, int, int);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* VoidPtrExport.proto */
static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig);

/* VoidPtrImport.proto */
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_f_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_seed_peak, double __pyx_v_score, int __pyx_v_charge, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_theoretical, PyObject *__pyx_v_experimental, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create *__pyx_optional_args); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__eq(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__ne(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__lt(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__gt(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_f_12ms_deisotope_2_c_7scoring_15FitSelectorBase_best(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_results, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_f_12ms_deisotope_2_c_7scoring_15FitSelectorBase__best_from_set(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_results); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_7scoring_15FitSelectorBase_reject(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_result, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_7scoring_15FitSelectorBase_reject_score(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self, CYTHON_UNUSED double __pyx_v_score, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_7scoring_15FitSelectorBase_is_maximizing(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_f_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_best(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MinimizeFitSelector *__pyx_v_self, PyObject *__pyx_v_results, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_f_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector__best_from_set(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_MinimizeFitSelector *__pyx_v_self, PyObject *__pyx_v_results); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_reject(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MinimizeFitSelector *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_fit, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_reject_score(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MinimizeFitSelector *__pyx_v_self, double __pyx_v_score, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_is_maximizing(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_MinimizeFitSelector *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_f_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_best(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MaximizeFitSelector *__pyx_v_self, PyObject *__pyx_v_results, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_f_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector__best_from_set(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_MaximizeFitSelector *__pyx_v_self, PyObject *__pyx_v_results); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_reject(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MaximizeFitSelector *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_fit, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_reject_score(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MaximizeFitSelector *__pyx_v_self, double __pyx_v_score, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_is_maximizing(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_MaximizeFitSelector *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase__evaluate(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, CYTHON_UNUSED PyObject *__pyx_v_observed, CYTHON_UNUSED PyObject *__pyx_v_expected, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_reject(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_fit, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_reject_score(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, double __pyx_v_score, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_is_maximizing(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_f_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase__configure(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_deconvoluter, CYTHON_UNUSED PyObject *__pyx_v_kwargs, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_7scoring_11GTestFitter__evaluate(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_7scoring_17ScaledGTestFitter__evaluate(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_7scoring_18LeastSquaresFitter__evaluate(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_7scoring_15ChiSquareFitter__evaluate(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_ChiSquareFitter *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_obs, struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_theo, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak *__pyx_optional_args); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_7scoring_14MSDeconVFitter__evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter__evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_7scoring_14FunctionScorer__evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_v_self, PyObject *__pyx_v_experimental_peaks, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference *__pyx_optional_args); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_f_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter__configure(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_deconvoluter, CYTHON_UNUSED PyObject *__pyx_v_kwargs, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter__evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_experimental, PyObject *__pyx_v_theoretical, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter__calculate_scale_factor(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_scale_fitted_peaks(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self, PyObject *__pyx_v_experimental, double __pyx_v_factor); /* proto*/
static void __pyx_f_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_scale_theoretical_peaks(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self, PyObject *__pyx_v_theoretical, double __pyx_v_factor); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter__evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_experimental, PyObject *__pyx_v_theoretical, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_7scoring_16DotProductFitter__evaluate(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_experimental, PyObject *__pyx_v_theoretical, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self); /* proto*/
static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto*/
static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/

/* Module declarations from 'brainpy._c.composition' */
static PyTypeObject *__pyx_ptype_7brainpy_2_c_11composition_PyComposition = 0;
static PyObject **__pyx_vp_7brainpy_2_c_11composition_nist_mass = 0;
#define __pyx_v_7brainpy_2_c_11composition_nist_mass (*__pyx_vp_7brainpy_2_c_11composition_nist_mass)
static double *__pyx_vp_7brainpy_2_c_11composition_PROTON = 0;
#define __pyx_v_7brainpy_2_c_11composition_PROTON (*__pyx_vp_7brainpy_2_c_11composition_PROTON)
static struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable **__pyx_vp_7brainpy_2_c_11composition__ElementTable = 0;
#define __pyx_v_7brainpy_2_c_11composition__ElementTable (*__pyx_vp_7brainpy_2_c_11composition__ElementTable)

/* Module declarations from 'brainpy._c.double_vector' */

/* Module declarations from 'brainpy._c.isotopic_constants' */
static size_t *__pyx_vp_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE = 0;
#define __pyx_v_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE (*__pyx_vp_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE)

/* Module declarations from 'brainpy._c.isotopic_distribution' */
static PyTypeObject *__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak = 0;

/* Module declarations from 'ms_peak_picker._c.peak_set' */
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase = 0;
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak = 0;
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet = 0;
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = 0;
static size_t *__pyx_vp_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE = 0;
#define __pyx_v_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE (*__pyx_vp_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE)

/* Module declarations from 'cython.view' */

/* Module declarations from 'cython' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_generic = 0;
static PyTypeObject *__pyx_ptype_5numpy_number = 0;
static PyTypeObject *__pyx_ptype_5numpy_integer = 0;
static PyTypeObject *__pyx_ptype_5numpy_signedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_unsignedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_inexact = 0;
static PyTypeObject *__pyx_ptype_5numpy_floating = 0;
static PyTypeObject *__pyx_ptype_5numpy_complexfloating = 0;
static PyTypeObject *__pyx_ptype_5numpy_flexible = 0;
static PyTypeObject *__pyx_ptype_5numpy_character = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;

/* Module declarations from 'ms_peak_picker._c.peak_index' */
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex = 0;

/* Module declarations from 'ms_deisotope._c.averagine' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache = 0;
static double *__pyx_vp_12ms_deisotope_2_c_9averagine_PROTON = 0;
#define __pyx_v_12ms_deisotope_2_c_9averagine_PROTON (*__pyx_vp_12ms_deisotope_2_c_9averagine_PROTON)

/* Module declarations from 'ms_deisotope._c.deconvoluter_base' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator = 0;

/* Module declarations from 'libc.math' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cpython.sequence' */

/* Module declarations from 'ms_deisotope._c.scoring' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_MinimizeFitSelector = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_MaximizeFitSelector = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_ChiSquareFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__ = 0;
static PyTypeObject *__pyx_array_type = 0;
static PyTypeObject *__pyx_MemviewEnum_type = 0;
static PyTypeObject *__pyx_memoryview_type = 0;
static PyTypeObject *__pyx_memoryviewslice_type = 0;
static double __pyx_v_12ms_deisotope_2_c_7scoring_INFINITY;
static struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *__pyx_v_12ms_deisotope_2_c_7scoring_g_test_scaled = 0;
static struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter *__pyx_v_12ms_deisotope_2_c_7scoring_least_squares = 0;
static PyObject *generic = 0;
static PyObject *strided = 0;
static PyObject *indirect = 0;
static PyObject *contiguous = 0;
static PyObject *indirect_contiguous = 0;
static int __pyx_memoryview_thread_locks_used;
static PyThread_type_lock __pyx_memoryview_thread_locks[8];
static double __pyx_f_12ms_deisotope_2_c_7scoring_sum_intensity_theoretical(PyObject *); /*proto*/
static double __pyx_f_12ms_deisotope_2_c_7scoring_sum_intensity_fitted(PyObject *); /*proto*/
static double __pyx_f_12ms_deisotope_2_c_7scoring_max_intensity_theoretical(PyObject *); /*proto*/
static double __pyx_f_12ms_deisotope_2_c_7scoring_max_intensity_fitted(PyObject *); /*proto*/
static double __pyx_f_12ms_deisotope_2_c_7scoring_ms_deconv_score_peak(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_ms_deconv_score_peak *__pyx_optional_args); /*proto*/
static struct __pyx_array_obj *__pyx_array_new(PyObject *, Py_ssize_t, char *, char *, char *); /*proto*/
static void *__pyx_align_pointer(void *, size_t); /*proto*/
static PyObject *__pyx_memoryview_new(PyObject *, int, int, __Pyx_TypeInfo *); /*proto*/
static CYTHON_INLINE int __pyx_memoryview_check(PyObject *); /*proto*/
static PyObject *_unellipsify(PyObject *, int); /*proto*/
static PyObject *assert_direct_dimensions(Py_ssize_t *, int); /*proto*/
static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *, PyObject *); /*proto*/
static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int, int); /*proto*/
static char *__pyx_pybuffer_index(Py_buffer *, char *, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memslice_transpose(__Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice, int, PyObject *(*)(char *), int (*)(char *, PyObject *), int); /*proto*/
static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *); /*proto*/
static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static Py_ssize_t abs_py_ssize_t(Py_ssize_t); /*proto*/
static char __pyx_get_best_slice_order(__Pyx_memviewslice *, int); /*proto*/
static void _copy_strided_to_strided(char *, Py_ssize_t *, char *, Py_ssize_t *, Py_ssize_t *, Py_ssize_t *, int, size_t); /*proto*/
static void copy_strided_to_strided(__Pyx_memviewslice *, __Pyx_memviewslice *, int, size_t); /*proto*/
static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *, int); /*proto*/
static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *, Py_ssize_t *, Py_ssize_t, int, char); /*proto*/
static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *, __Pyx_memviewslice *, char, int); /*proto*/
static int __pyx_memoryview_err_extents(int, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memoryview_err_dim(PyObject *, char *, int); /*proto*/
static int __pyx_memoryview_err(PyObject *, char *); /*proto*/
static int __pyx_memoryview_copy_contents(__Pyx_memviewslice, __Pyx_memviewslice, int, int, int); /*proto*/
static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *, int, int); /*proto*/
static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *, int, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *, int, size_t, void *, int); /*proto*/
static void __pyx_memoryview__slice_assign_scalar(char *, Py_ssize_t *, Py_ssize_t *, int, size_t, void *); /*proto*/
static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *, PyObject *); /*proto*/
#define __Pyx_MODULE_NAME "ms_deisotope._c.scoring"
extern int __pyx_module_is_main_ms_deisotope___c__scoring;
int __pyx_module_is_main_ms_deisotope___c__scoring = 0;

/* Implementation of 'ms_deisotope._c.scoring' */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_NotImplementedError;
static PyObject *__pyx_builtin_min;
static PyObject *__pyx_builtin_max;
static PyObject *__pyx_builtin_ImportError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_Ellipsis;
static PyObject *__pyx_builtin_id;
static PyObject *__pyx_builtin_IndexError;
static const char __pyx_k_O[] = "O";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_eq[] = "_eq";
static const char __pyx_k_gt[] = "_gt";
static const char __pyx_k_id[] = "id";
static const char __pyx_k_lt[] = "_lt";
static const char __pyx_k_ne[] = "_ne";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_max[] = "max";
static const char __pyx_k_min[] = "min";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_obj[] = "obj";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_base[] = "base";
static const char __pyx_k_best[] = "best";
static const char __pyx_k_data[] = "data";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_iter[] = "__iter__";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mode[] = "mode";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_ndim[] = "ndim";
static const char __pyx_k_pack[] = "pack";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_step[] = "step";
static const char __pyx_k_stop[] = "stop";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_ASCII[] = "ASCII";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_error[] = "error";
static const char __pyx_k_flags[] = "flags";
static const char __pyx_k_lower[] = "lower";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_score[] = "score";
static const char __pyx_k_shape[] = "shape";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_upper[] = "upper";
static const char __pyx_k_charge[] = "charge";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_fields[] = "fields";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_npeaks[] = "npeaks";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_reject[] = "reject";
static const char __pyx_k_struct[] = "struct";
static const char __pyx_k_unpack[] = "unpack";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_fortran[] = "fortran";
static const char __pyx_k_memview[] = "memview";
static const char __pyx_k_Ellipsis[] = "Ellipsis";
static const char __pyx_k_INFINITY[] = "INFINITY";
static const char __pyx_k_evaluate[] = "_evaluate";
static const char __pyx_k_expected[] = "expected";
static const char __pyx_k_function[] = "function";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_itemsize[] = "itemsize";
static const char __pyx_k_observed[] = "observed";
static const char __pyx_k_operator[] = "operator";
static const char __pyx_k_peaklist[] = "peaklist";
static const char __pyx_k_pyx_capi[] = "__pyx_capi__";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_configure[] = "_configure";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_seed_peak[] = "seed_peak";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_attrgetter[] = "attrgetter";
static const char __pyx_k_evaluate_2[] = "evaluate";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_GTestFitter[] = "GTestFitter";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_theoretical[] = "theoretical";
static const char __pyx_k_deconvoluter[] = "deconvoluter";
static const char __pyx_k_domain_scale[] = "domain_scale";
static const char __pyx_k_experimental[] = "experimental";
static const char __pyx_k_missed_peaks[] = "missed_peaks";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_reject_score[] = "reject_score";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_is_maximizing[] = "is_maximizing";
static const char __pyx_k_minimum_score[] = "minimum_score";
static const char __pyx_k_pyx_getbuffer[] = "__pyx_getbuffer";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_selector_type[] = "selector_type";
static const char __pyx_k_FunctionScorer[] = "FunctionScorer";
static const char __pyx_k_MSDeconVFitter[] = "MSDeconVFitter";
static const char __pyx_k_penalty_factor[] = "penalty_factor";
static const char __pyx_k_ChiSquareFitter[] = "ChiSquareFitter";
static const char __pyx_k_FitSelectorBase[] = "FitSelectorBase";
static const char __pyx_k_View_MemoryView[] = "View.MemoryView";
static const char __pyx_k_allocate_buffer[] = "allocate_buffer";
static const char __pyx_k_dtype_is_object[] = "dtype_is_object";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_score_threshold[] = "score_threshold";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_DotProductFitter[] = "DotProductFitter";
static const char __pyx_k_peak_count_scale[] = "peak_count_scale";
static const char __pyx_k_IsotopicFitRecord[] = "IsotopicFitRecord";
static const char __pyx_k_ScaledGTestFitter[] = "ScaledGTestFitter";
static const char __pyx_k_pyx_unpickle_Enum[] = "__pyx_unpickle_Enum";
static const char __pyx_k_IsotopicFitterBase[] = "IsotopicFitterBase";
static const char __pyx_k_LeastSquaresFitter[] = "LeastSquaresFitter";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_experimental_peaks[] = "experimental_peaks";
static const char __pyx_k_strided_and_direct[] = "<strided and direct>";
static const char __pyx_k_MaximizeFitSelector[] = "MaximizeFitSelector";
static const char __pyx_k_MinimizeFitSelector[] = "MinimizeFitSelector";
static const char __pyx_k_NotImplementedError[] = "NotImplementedError";
static const char __pyx_k_mass_error_tolerance[] = "mass_error_tolerance";
static const char __pyx_k_strided_and_indirect[] = "<strided and indirect>";
static const char __pyx_k_DistinctPatternFitter[] = "DistinctPatternFitter";
static const char __pyx_k_InterferenceDetection[] = "InterferenceDetection";
static const char __pyx_k_contiguous_and_direct[] = "<contiguous and direct>";
static const char __pyx_k_MemoryView_of_r_object[] = "<MemoryView of %r object>";
static const char __pyx_k_calculate_scale_factor[] = "_calculate_scale_factor";
static const char __pyx_k_MemoryView_of_r_at_0x_x[] = "<MemoryView of %r at 0x%x>";
static const char __pyx_k_PenalizedMSDeconVFitter[] = "PenalizedMSDeconVFitter";
static const char __pyx_k_contiguous_and_indirect[] = "<contiguous and indirect>";
static const char __pyx_k_ms_deisotope__c_scoring[] = "ms_deisotope._c.scoring";
static const char __pyx_k_Cannot_index_with_type_s[] = "Cannot index with type '%s'";
static const char __pyx_k_IsotopicFitRecord___iter[] = "IsotopicFitRecord.__iter__";
static const char __pyx_k_Invalid_shape_in_axis_d_d[] = "Invalid shape in axis %d: %d.";
static const char __pyx_k_itemsize_0_for_cython_array[] = "itemsize <= 0 for cython.array";
static const char __pyx_k_ScaledPenalizedMSDeconvFitter[] = "ScaledPenalizedMSDeconvFitter";
static const char __pyx_k_unable_to_allocate_array_data[] = "unable to allocate array data.";
static const char __pyx_k_self___class_____name___fields[] = "{self.__class__.__name__}({fields})";
static const char __pyx_k_strided_and_direct_or_indirect[] = "<strided and direct or indirect>";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_self___class_____name___minimum[] = "{self.__class__.__name__}(minimum_score={self.minimum_score})";
static const char __pyx_k_Buffer_view_does_not_expose_stri[] = "Buffer view does not expose strides";
static const char __pyx_k_Can_only_create_a_buffer_that_is[] = "Can only create a buffer that is contiguous in memory.";
static const char __pyx_k_Cannot_assign_to_read_only_memor[] = "Cannot assign to read-only memoryview";
static const char __pyx_k_Cannot_create_writable_memory_vi[] = "Cannot create writable memory view from read-only memoryview";
static const char __pyx_k_Empty_shape_tuple_for_cython_arr[] = "Empty shape tuple for cython.array";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0[] = "Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))";
static const char __pyx_k_Indirect_dimensions_not_supporte[] = "Indirect dimensions not supported";
static const char __pyx_k_Invalid_mode_expected_c_or_fortr[] = "Invalid mode, expected 'c' or 'fortran', got %s";
static const char __pyx_k_IsotopicFitRecord_score_0_5f_cha[] = "IsotopicFitRecord(score=%0.5f, charge=%d, npeaks=%d, monoisotopic_mz=%0.5f)";
static const char __pyx_k_Out_of_bounds_on_buffer_access_a[] = "Out of bounds on buffer access (axis %d)";
static const char __pyx_k_Unable_to_convert_item_to_object[] = "Unable to convert item to object";
static const char __pyx_k_got_differing_extents_in_dimensi[] = "got differing extents in dimension %d (got %d and %d)";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_unable_to_allocate_shape_and_str[] = "unable to allocate shape and strides.";
static PyObject *__pyx_n_s_ASCII;
static PyObject *__pyx_kp_s_Buffer_view_does_not_expose_stri;
static PyObject *__pyx_kp_s_Can_only_create_a_buffer_that_is;
static PyObject *__pyx_kp_s_Cannot_assign_to_read_only_memor;
static PyObject *__pyx_kp_s_Cannot_create_writable_memory_vi;
static PyObject *__pyx_kp_s_Cannot_index_with_type_s;
static PyObject *__pyx_n_s_ChiSquareFitter;
static PyObject *__pyx_n_s_DistinctPatternFitter;
static PyObject *__pyx_n_s_DotProductFitter;
static PyObject *__pyx_n_s_Ellipsis;
static PyObject *__pyx_kp_s_Empty_shape_tuple_for_cython_arr;
static PyObject *__pyx_n_s_FitSelectorBase;
static PyObject *__pyx_n_s_FunctionScorer;
static PyObject *__pyx_n_s_GTestFitter;
static PyObject *__pyx_n_s_INFINITY;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0;
static PyObject *__pyx_n_s_IndexError;
static PyObject *__pyx_kp_s_Indirect_dimensions_not_supporte;
static PyObject *__pyx_n_s_InterferenceDetection;
static PyObject *__pyx_kp_s_Invalid_mode_expected_c_or_fortr;
static PyObject *__pyx_kp_s_Invalid_shape_in_axis_d_d;
static PyObject *__pyx_n_s_IsotopicFitRecord;
static PyObject *__pyx_n_s_IsotopicFitRecord___iter;
static PyObject *__pyx_kp_s_IsotopicFitRecord_score_0_5f_cha;
static PyObject *__pyx_n_s_IsotopicFitterBase;
static PyObject *__pyx_n_s_LeastSquaresFitter;
static PyObject *__pyx_n_s_MSDeconVFitter;
static PyObject *__pyx_n_s_MaximizeFitSelector;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_kp_s_MemoryView_of_r_at_0x_x;
static PyObject *__pyx_kp_s_MemoryView_of_r_object;
static PyObject *__pyx_n_s_MinimizeFitSelector;
static PyObject *__pyx_n_s_NotImplementedError;
static PyObject *__pyx_n_b_O;
static PyObject *__pyx_kp_s_Out_of_bounds_on_buffer_access_a;
static PyObject *__pyx_n_s_PenalizedMSDeconVFitter;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_ScaledGTestFitter;
static PyObject *__pyx_n_s_ScaledPenalizedMSDeconvFitter;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_kp_s_Unable_to_convert_item_to_object;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_View_MemoryView;
static PyObject *__pyx_n_s_allocate_buffer;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_attrgetter;
static PyObject *__pyx_n_s_base;
static PyObject *__pyx_n_s_best;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_n_u_c;
static PyObject *__pyx_n_s_calculate_scale_factor;
static PyObject *__pyx_n_s_charge;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_configure;
static PyObject *__pyx_kp_s_contiguous_and_direct;
static PyObject *__pyx_kp_s_contiguous_and_indirect;
static PyObject *__pyx_n_s_data;
static PyObject *__pyx_n_s_deconvoluter;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_domain_scale;
static PyObject *__pyx_n_s_dtype_is_object;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_eq;
static PyObject *__pyx_n_s_error;
static PyObject *__pyx_n_s_evaluate;
static PyObject *__pyx_n_s_evaluate_2;
static PyObject *__pyx_n_s_expected;
static PyObject *__pyx_n_s_experimental;
static PyObject *__pyx_n_s_experimental_peaks;
static PyObject *__pyx_n_s_fields;
static PyObject *__pyx_n_s_flags;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_fortran;
static PyObject *__pyx_n_u_fortran;
static PyObject *__pyx_n_s_function;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_kp_s_got_differing_extents_in_dimensi;
static PyObject *__pyx_n_s_gt;
static PyObject *__pyx_n_s_id;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_is_maximizing;
static PyObject *__pyx_n_s_itemsize;
static PyObject *__pyx_kp_s_itemsize_0_for_cython_array;
static PyObject *__pyx_n_s_iter;
static PyObject *__pyx_n_s_key;
static PyObject *__pyx_n_s_kwargs;
static PyObject *__pyx_n_s_lower;
static PyObject *__pyx_n_s_lt;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_mass_error_tolerance;
static PyObject *__pyx_n_s_max;
static PyObject *__pyx_n_s_memview;
static PyObject *__pyx_n_s_min;
static PyObject *__pyx_n_s_minimum_score;
static PyObject *__pyx_n_s_missed_peaks;
static PyObject *__pyx_n_s_mode;
static PyObject *__pyx_n_s_ms_deisotope__c_scoring;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_s_ndim;
static PyObject *__pyx_n_s_ne;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_npeaks;
static PyObject *__pyx_kp_s_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_s_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_obj;
static PyObject *__pyx_n_s_observed;
static PyObject *__pyx_n_s_operator;
static PyObject *__pyx_n_s_pack;
static PyObject *__pyx_n_s_peak_count_scale;
static PyObject *__pyx_n_s_peaklist;
static PyObject *__pyx_n_s_penalty_factor;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_capi;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_getbuffer;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_Enum;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_reject;
static PyObject *__pyx_n_s_reject_score;
static PyObject *__pyx_n_s_score;
static PyObject *__pyx_n_s_score_threshold;
static PyObject *__pyx_n_s_seed_peak;
static PyObject *__pyx_n_s_selector_type;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_kp_s_self___class_____name___fields;
static PyObject *__pyx_kp_s_self___class_____name___minimum;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_shape;
static PyObject *__pyx_n_s_size;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_s_step;
static PyObject *__pyx_n_s_stop;
static PyObject *__pyx_kp_s_strided_and_direct;
static PyObject *__pyx_kp_s_strided_and_direct_or_indirect;
static PyObject *__pyx_kp_s_strided_and_indirect;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_struct;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_theoretical;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_kp_s_unable_to_allocate_array_data;
static PyObject *__pyx_kp_s_unable_to_allocate_shape_and_str;
static PyObject *__pyx_n_s_unpack;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_upper;
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord___init__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_seed_peak, double __pyx_v_score, int __pyx_v_charge, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_theoretical, PyObject *__pyx_v_experimental, PyObject *__pyx_v_data, int __pyx_v_missed_peaks); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_2clone(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_4__reduce__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_6_eq(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_8_ne(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_10_lt(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12_gt(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_14__richcmp__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_other, int __pyx_v_code); /* proto */
static Py_hash_t __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_16__hash__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_18__iter__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_6npeaks___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_21__repr__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_9seed_peak___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_9seed_peak_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_9seed_peak_4__del__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_5score___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_5score_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_6charge___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_6charge_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12experimental___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12experimental_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12experimental_4__del__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_11theoretical___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_11theoretical_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_11theoretical_4__del__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_17monoisotopic_peak___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_17monoisotopic_peak_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_17monoisotopic_peak_4__del__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12missed_peaks___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12missed_peaks_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_4data___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_4data_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_4data_4__del__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_5_hash___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_5_hash_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase___init__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self, PyObject *__pyx_v_minimum_score); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase_2__reduce__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase_4best(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self, PyObject *__pyx_v_results); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase_6__call__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase_8reject(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_result); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase_10reject_score(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self, double __pyx_v_score); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase_12is_maximizing(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase_14__repr__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase_13minimum_score___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase_13minimum_score_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_best(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MinimizeFitSelector *__pyx_v_self, PyObject *__pyx_v_results); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_2reject(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MinimizeFitSelector *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_fit); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_4reject_score(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MinimizeFitSelector *__pyx_v_self, double __pyx_v_score); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_6is_maximizing(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MinimizeFitSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_best(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MaximizeFitSelector *__pyx_v_self, PyObject *__pyx_v_results); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_2reject(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MaximizeFitSelector *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_fit); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_4reject_score(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MaximizeFitSelector *__pyx_v_self, double __pyx_v_score); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_6is_maximizing(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MaximizeFitSelector *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase___init__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, PyObject *__pyx_v_score_threshold); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_2__reduce__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_4__getstate__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_6__setstate__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_8evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_10_evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_12__call__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_14reject(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_fit); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_16reject_score(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, double __pyx_v_score); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_18is_maximizing(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_20_configure(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_deconvoluter, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_22configure(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_deconvoluter, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_24__repr__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_6select___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_6select_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_6select_4__del__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_11GTestFitter__evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17ScaledGTestFitter__evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_18LeastSquaresFitter__evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_15ChiSquareFitter__evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ChiSquareFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_14MSDeconVFitter___init__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *__pyx_v_self, PyObject *__pyx_v_minimum_score, PyObject *__pyx_v_mass_error_tolerance); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_2evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_4__reduce__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_6__getstate__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_8__setstate__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *__pyx_v_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_10_evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_20mass_error_tolerance___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_20mass_error_tolerance_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter___init__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_v_self, PyObject *__pyx_v_minimum_score, PyObject *__pyx_v_penalty_factor, PyObject *__pyx_v_mass_error_tolerance); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_2__reduce__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_4__getstate__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_6__setstate__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_v_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_8evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_10_evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_20mass_error_tolerance___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_20mass_error_tolerance_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_14penalty_factor___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_14penalty_factor_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_14FunctionScorer___init__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer *__pyx_v_self, PyObject *__pyx_v_function, PyObject *__pyx_v_minimum_score, PyObject *__pyx_v_selector_type); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_14FunctionScorer_2__getstate__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_14FunctionScorer_4__setstate__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer *__pyx_v_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_14FunctionScorer_6__reduce__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_14FunctionScorer_8_evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_21InterferenceDetection___init__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_21InterferenceDetection_2__reduce__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_21InterferenceDetection_4__getstate__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_21InterferenceDetection_6__setstate__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_v_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_21InterferenceDetection_8__call__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_v_self, PyObject *__pyx_v_experimental_peaks, PyObject *__pyx_v_lower, PyObject *__pyx_v_upper); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_21InterferenceDetection_8peaklist___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_21InterferenceDetection_8peaklist_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_21InterferenceDetection_8peaklist_4__del__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter___init__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self, PyObject *__pyx_v_minimum_score, PyObject *__pyx_v_peak_count_scale, PyObject *__pyx_v_domain_scale); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_2__getstate__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_4__setstate__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_6_configure(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_deconvoluter, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_8_evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_experimental, PyObject *__pyx_v_theoretical); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_21interference_detector___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_21interference_detector_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_21interference_detector_4__del__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_13g_test_scaled___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_13g_test_scaled_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_13g_test_scaled_4__del__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_16peak_count_scale___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_16peak_count_scale_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_12domain_scale___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_12domain_scale_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter___init__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self, PyObject *__pyx_v_minimum_score, PyObject *__pyx_v_penalty_factor, PyObject *__pyx_v_mass_error_tolerance); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_2__getstate__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_4__setstate__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_14penalty_factor___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_14penalty_factor_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_6_calculate_scale_factor(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_8evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_10_evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_experimental, PyObject *__pyx_v_theoretical); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_12scale_factor___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_12scale_factor_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_6scorer___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_6scorer_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_6scorer_4__del__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_7scoring_16DotProductFitter___init__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter *__pyx_v_self, PyObject *__pyx_v_minimum_score); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_16DotProductFitter_2_evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_experimental, PyObject *__pyx_v_theoretical); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object); /* proto */
static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_IsotopicFitRecord(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_FitSelectorBase(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_IsotopicFitterBase(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_LeastSquaresFitter(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_GTestFitter(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_ScaledGTestFitter(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_MSDeconVFitter(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_FunctionScorer(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_InterferenceDetection(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_DistinctPatternFitter(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_DotProductFitter(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_MinimizeFitSelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_MaximizeFitSelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_ChiSquareFitter(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_float_1_;
static PyObject *__pyx_float_0_3;
static PyObject *__pyx_float_0_5;
static PyObject *__pyx_float_1_5;
static PyObject *__pyx_float_0_02;
static PyObject *__pyx_float_100_;
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_10;
static PyObject *__pyx_int_100;
static PyObject *__pyx_int_112105877;
static PyObject *__pyx_int_136983863;
static PyObject *__pyx_int_184977713;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_k__2;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_slice__20;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_codeobj__31;
/* Late includes */

/* "ms_deisotope/_c/scoring.pyx":17
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef bint isclose(double x, double y, double rtol=1.e-5, double atol=1.e-8):             # <<<<<<<<<<<<<<
 *     return abs(x-y) <= (atol + rtol * abs(y))
 * 
 */

#line 17 "ms_deisotope/_c/scoring.pyx"


#line 17 "ms_deisotope/_c/scoring.pyx"
static int __pyx_f_12ms_deisotope_2_c_7scoring_isclose(double __pyx_v_x, double __pyx_v_y, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_isclose *__pyx_optional_args) {

#line 17 "ms_deisotope/_c/scoring.pyx"
  double __pyx_v_rtol = ((double)1.e-5);

#line 17 "ms_deisotope/_c/scoring.pyx"
  double __pyx_v_atol = ((double)1.e-8);

#line 17 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 17 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 17 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("isclose", 0);

#line 17 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_optional_args) {

#line 17 "ms_deisotope/_c/scoring.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 17 "ms_deisotope/_c/scoring.pyx"
      __pyx_v_rtol = __pyx_optional_args->rtol;

#line 17 "ms_deisotope/_c/scoring.pyx"
      if (__pyx_optional_args->__pyx_n > 1) {

#line 17 "ms_deisotope/_c/scoring.pyx"
        __pyx_v_atol = __pyx_optional_args->atol;

#line 17 "ms_deisotope/_c/scoring.pyx"
      }

#line 17 "ms_deisotope/_c/scoring.pyx"
    }

#line 17 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":18
 * @cython.cdivision(True)
 * cdef bint isclose(double x, double y, double rtol=1.e-5, double atol=1.e-8):
 *     return abs(x-y) <= (atol + rtol * abs(y))             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 18 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = (fabs((__pyx_v_x - __pyx_v_y)) <= (__pyx_v_atol + (__pyx_v_rtol * fabs(__pyx_v_y))));

#line 18 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":17
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef bint isclose(double x, double y, double rtol=1.e-5, double atol=1.e-8):             # <<<<<<<<<<<<<<
 *     return abs(x-y) <= (atol + rtol * abs(y))
 * 
 */

#line 17 "ms_deisotope/_c/scoring.pyx"


#line 17 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 17 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 17 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 17 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 17 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":50
 *         The theoretical isotopic pattern being fitted on the experimental data
 *     """
 *     def __init__(self, FittedPeak seed_peak, double score, int charge, TheoreticalIsotopicPattern theoretical, list experimental,             # <<<<<<<<<<<<<<
 *                   object data=None, int missed_peaks=0):
 *         self.seed_peak = seed_peak
 */

#line 50 "ms_deisotope/_c/scoring.pyx"


#line 50 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 50 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 50 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_seed_peak = 0
#line 50 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_score
#line 50 "ms_deisotope/_c/scoring.pyx"
;
  int __pyx_v_charge
#line 50 "ms_deisotope/_c/scoring.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_theoretical = 0
#line 50 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_experimental = 0
#line 50 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_data = 0
#line 50 "ms_deisotope/_c/scoring.pyx"
;
  int __pyx_v_missed_peaks
#line 50 "ms_deisotope/_c/scoring.pyx"
;

#line 50 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 50 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 50 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 50 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 50 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 50 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 50 "ms_deisotope/_c/scoring.pyx"
  {

#line 50 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_seed_peak,&__pyx_n_s_score,&__pyx_n_s_charge,&__pyx_n_s_theoretical,&__pyx_n_s_experimental,&__pyx_n_s_data,&__pyx_n_s_missed_peaks,0};

#line 50 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[7] = {0,0,0,0,0,0,0};

    /* "ms_deisotope/_c/scoring.pyx":51
 *     """
 *     def __init__(self, FittedPeak seed_peak, double score, int charge, TheoreticalIsotopicPattern theoretical, list experimental,
 *                   object data=None, int missed_peaks=0):             # <<<<<<<<<<<<<<
 *         self.seed_peak = seed_peak
 *         self.score = score
 */

#line 51 "ms_deisotope/_c/scoring.pyx"
    values[5] = ((PyObject *)Py_None);

#line 51 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 51 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 51 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 51 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  7: 
#line 51 "ms_deisotope/_c/scoring.pyx"
values[6] = PyTuple_GET_ITEM(__pyx_args, 6);

#line 51 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  6: 
#line 51 "ms_deisotope/_c/scoring.pyx"
values[5] = PyTuple_GET_ITEM(__pyx_args, 5);

#line 51 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  5: 
#line 51 "ms_deisotope/_c/scoring.pyx"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 51 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 51 "ms_deisotope/_c/scoring.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 51 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 51 "ms_deisotope/_c/scoring.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 51 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 51 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 51 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 51 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 51 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 51 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 51 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 51 "ms_deisotope/_c/scoring.pyx"
      }

#line 51 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 51 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 51 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 51 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seed_peak)) != 0)) kw_args--;
        else 
#line 51 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 51 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 51 "ms_deisotope/_c/scoring.pyx"
        case  1:

#line 51 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_score)) != 0)) kw_args--;

#line 51 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 5, 7, 1); 
#line 51 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 50, __pyx_L3_error)

#line 51 "ms_deisotope/_c/scoring.pyx"
        }

#line 51 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 51 "ms_deisotope/_c/scoring.pyx"
        case  2:

#line 51 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge)) != 0)) kw_args--;

#line 51 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 5, 7, 2); 
#line 51 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 50, __pyx_L3_error)

#line 51 "ms_deisotope/_c/scoring.pyx"
        }

#line 51 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 51 "ms_deisotope/_c/scoring.pyx"
        case  3:

#line 51 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_theoretical)) != 0)) kw_args--;

#line 51 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 5, 7, 3); 
#line 51 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 50, __pyx_L3_error)

#line 51 "ms_deisotope/_c/scoring.pyx"
        }

#line 51 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 51 "ms_deisotope/_c/scoring.pyx"
        case  4:

#line 51 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_experimental)) != 0)) kw_args--;

#line 51 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 5, 7, 4); 
#line 51 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 50, __pyx_L3_error)

#line 51 "ms_deisotope/_c/scoring.pyx"
        }

#line 51 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 51 "ms_deisotope/_c/scoring.pyx"
        case  5:

#line 51 "ms_deisotope/_c/scoring.pyx"
        if (kw_args > 0) {

#line 51 "ms_deisotope/_c/scoring.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data);

#line 51 "ms_deisotope/_c/scoring.pyx"
          if (value) { values[5] = value; kw_args--; }

#line 51 "ms_deisotope/_c/scoring.pyx"
        }

#line 51 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 51 "ms_deisotope/_c/scoring.pyx"
        case  6:

#line 51 "ms_deisotope/_c/scoring.pyx"
        if (kw_args > 0) {

#line 51 "ms_deisotope/_c/scoring.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_missed_peaks);

#line 51 "ms_deisotope/_c/scoring.pyx"
          if (value) { values[6] = value; kw_args--; }

#line 51 "ms_deisotope/_c/scoring.pyx"
        }

#line 51 "ms_deisotope/_c/scoring.pyx"
      }

#line 51 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 51 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 50, __pyx_L3_error)

#line 51 "ms_deisotope/_c/scoring.pyx"
      }

#line 51 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 51 "ms_deisotope/_c/scoring.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  7: 
#line 51 "ms_deisotope/_c/scoring.pyx"
values[6] = PyTuple_GET_ITEM(__pyx_args, 6);

#line 51 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  6: 
#line 51 "ms_deisotope/_c/scoring.pyx"
values[5] = PyTuple_GET_ITEM(__pyx_args, 5);

#line 51 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  5: 
#line 51 "ms_deisotope/_c/scoring.pyx"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 51 "ms_deisotope/_c/scoring.pyx"
        values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 51 "ms_deisotope/_c/scoring.pyx"
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 51 "ms_deisotope/_c/scoring.pyx"
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 51 "ms_deisotope/_c/scoring.pyx"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 51 "ms_deisotope/_c/scoring.pyx"
        break;
        default: 
#line 51 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 51 "ms_deisotope/_c/scoring.pyx"
      }

#line 51 "ms_deisotope/_c/scoring.pyx"
    }

#line 51 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_seed_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)values[0]);

#line 51 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_score = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_score == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 50, __pyx_L3_error)

#line 51 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_charge = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 50, __pyx_L3_error)

#line 51 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_theoretical = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)values[3]);

#line 51 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_experimental = ((PyObject*)values[4]);

#line 51 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_data = values[5];

#line 51 "ms_deisotope/_c/scoring.pyx"
    if (values[6]) {

#line 51 "ms_deisotope/_c/scoring.pyx"
      __pyx_v_missed_peaks = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_missed_peaks == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 51, __pyx_L3_error)

#line 51 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 51 "ms_deisotope/_c/scoring.pyx"
      __pyx_v_missed_peaks = ((int)0);

#line 51 "ms_deisotope/_c/scoring.pyx"
    }

#line 51 "ms_deisotope/_c/scoring.pyx"
  }

#line 51 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 51 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 5, 7, PyTuple_GET_SIZE(__pyx_args)); 
#line 51 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 50, __pyx_L3_error)

#line 51 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 51 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitRecord.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 51 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 51 "ms_deisotope/_c/scoring.pyx"
  return -1;

#line 51 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 51 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_seed_peak), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak, 1, "seed_peak", 0))) __PYX_ERR(0, 50, __pyx_L1_error)

#line 51 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_theoretical), __pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern, 1, "theoretical", 0))) __PYX_ERR(0, 50, __pyx_L1_error)

#line 51 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_experimental), (&PyList_Type), 1, "experimental", 1))) __PYX_ERR(0, 50, __pyx_L1_error)
  __pyx_r = 
#line 51 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord___init__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self), __pyx_v_seed_peak, __pyx_v_score, __pyx_v_charge, __pyx_v_theoretical, __pyx_v_experimental, __pyx_v_data, __pyx_v_missed_peaks);

  /* "ms_deisotope/_c/scoring.pyx":50
 *         The theoretical isotopic pattern being fitted on the experimental data
 *     """
 *     def __init__(self, FittedPeak seed_peak, double score, int charge, TheoreticalIsotopicPattern theoretical, list experimental,             # <<<<<<<<<<<<<<
 *                   object data=None, int missed_peaks=0):
 *         self.seed_peak = seed_peak
 */

#line 50 "ms_deisotope/_c/scoring.pyx"


#line 50 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 50 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 50 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 50 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = -1;

#line 50 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 50 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 50 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 50 "ms_deisotope/_c/scoring.pyx"
}

#line 50 "ms_deisotope/_c/scoring.pyx"


#line 50 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord___init__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_seed_peak, double __pyx_v_score, int __pyx_v_charge, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_theoretical, PyObject *__pyx_v_experimental, PyObject *__pyx_v_data, int __pyx_v_missed_peaks) {

#line 50 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 50 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 50 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 50 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 50 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 50 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 50 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_deisotope/_c/scoring.pyx":52
 *     def __init__(self, FittedPeak seed_peak, double score, int charge, TheoreticalIsotopicPattern theoretical, list experimental,
 *                   object data=None, int missed_peaks=0):
 *         self.seed_peak = seed_peak             # <<<<<<<<<<<<<<
 *         self.score = score
 *         self.charge = charge
 */

#line 52 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_seed_peak));

#line 52 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_seed_peak));

#line 52 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->seed_peak);

#line 52 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->seed_peak));

#line 52 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->seed_peak = __pyx_v_seed_peak;

  /* "ms_deisotope/_c/scoring.pyx":53
 *                   object data=None, int missed_peaks=0):
 *         self.seed_peak = seed_peak
 *         self.score = score             # <<<<<<<<<<<<<<
 *         self.charge = charge
 *         self.experimental = experimental
 */

#line 53 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->score = __pyx_v_score;

  /* "ms_deisotope/_c/scoring.pyx":54
 *         self.seed_peak = seed_peak
 *         self.score = score
 *         self.charge = charge             # <<<<<<<<<<<<<<
 *         self.experimental = experimental
 *         self.theoretical = theoretical
 */

#line 54 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->charge = __pyx_v_charge;

  /* "ms_deisotope/_c/scoring.pyx":55
 *         self.score = score
 *         self.charge = charge
 *         self.experimental = experimental             # <<<<<<<<<<<<<<
 *         self.theoretical = theoretical
 *         self.monoisotopic_peak = experimental[0]
 */

#line 55 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_v_experimental);

#line 55 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_v_experimental);

#line 55 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->experimental);

#line 55 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_v_self->experimental);

#line 55 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->experimental = __pyx_v_experimental;

  /* "ms_deisotope/_c/scoring.pyx":56
 *         self.charge = charge
 *         self.experimental = experimental
 *         self.theoretical = theoretical             # <<<<<<<<<<<<<<
 *         self.monoisotopic_peak = experimental[0]
 *         self.data = data
 */

#line 56 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_theoretical));

#line 56 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_theoretical));

#line 56 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->theoretical);

#line 56 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->theoretical));

#line 56 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->theoretical = __pyx_v_theoretical;

  /* "ms_deisotope/_c/scoring.pyx":57
 *         self.experimental = experimental
 *         self.theoretical = theoretical
 *         self.monoisotopic_peak = experimental[0]             # <<<<<<<<<<<<<<
 *         self.data = data
 *         self.missed_peaks = missed_peaks
 */

#line 57 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_v_experimental == Py_None)) {

#line 57 "ms_deisotope/_c/scoring.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 57 "ms_deisotope/_c/scoring.pyx"
    __PYX_ERR(0, 57, __pyx_L1_error)

#line 57 "ms_deisotope/_c/scoring.pyx"
  }

#line 57 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_experimental, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 57, __pyx_L1_error)

#line 57 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 57 "ms_deisotope/_c/scoring.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak))))) __PYX_ERR(0, 57, __pyx_L1_error)

#line 57 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 57 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->monoisotopic_peak);

#line 57 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->monoisotopic_peak));

#line 57 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->monoisotopic_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1);

#line 57 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/scoring.pyx":58
 *         self.theoretical = theoretical
 *         self.monoisotopic_peak = experimental[0]
 *         self.data = data             # <<<<<<<<<<<<<<
 *         self.missed_peaks = missed_peaks
 *         self._hash = -1
 */

#line 58 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_v_data);

#line 58 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_v_data);

#line 58 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->data);

#line 58 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_v_self->data);

#line 58 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->data = __pyx_v_data;

  /* "ms_deisotope/_c/scoring.pyx":59
 *         self.monoisotopic_peak = experimental[0]
 *         self.data = data
 *         self.missed_peaks = missed_peaks             # <<<<<<<<<<<<<<
 *         self._hash = -1
 * 
 */

#line 59 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->missed_peaks = __pyx_v_missed_peaks;

  /* "ms_deisotope/_c/scoring.pyx":60
 *         self.data = data
 *         self.missed_peaks = missed_peaks
 *         self._hash = -1             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

#line 60 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->_hash = -1L;

  /* "ms_deisotope/_c/scoring.pyx":50
 *         The theoretical isotopic pattern being fitted on the experimental data
 *     """
 *     def __init__(self, FittedPeak seed_peak, double score, int charge, TheoreticalIsotopicPattern theoretical, list experimental,             # <<<<<<<<<<<<<<
 *                   object data=None, int missed_peaks=0):
 *         self.seed_peak = seed_peak
 */

#line 50 "ms_deisotope/_c/scoring.pyx"


#line 50 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 50 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 50 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 50 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 50 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 50 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitRecord.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 50 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = -1;

#line 50 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 50 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 50 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 50 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":63
 * 
 *     @staticmethod
 *     cdef IsotopicFitRecord _create(FittedPeak seed_peak, double score, int charge, TheoreticalIsotopicPattern theoretical,             # <<<<<<<<<<<<<<
 *                                    list experimental, object data=None, int missed_peaks=0):
 *         cdef IsotopicFitRecord inst
 */

#line 63 "ms_deisotope/_c/scoring.pyx"


#line 63 "ms_deisotope/_c/scoring.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_f_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_seed_peak, double __pyx_v_score, int __pyx_v_charge, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_v_theoretical, PyObject *__pyx_v_experimental, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create *__pyx_optional_args) {

  /* "ms_deisotope/_c/scoring.pyx":64
 *     @staticmethod
 *     cdef IsotopicFitRecord _create(FittedPeak seed_peak, double score, int charge, TheoreticalIsotopicPattern theoretical,
 *                                    list experimental, object data=None, int missed_peaks=0):             # <<<<<<<<<<<<<<
 *         cdef IsotopicFitRecord inst
 *         inst = IsotopicFitRecord.__new__(IsotopicFitRecord)
 */

#line 64 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_v_data = ((PyObject *)Py_None);

#line 64 "ms_deisotope/_c/scoring.pyx"
  int __pyx_v_missed_peaks = ((int)0);
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_inst = 0
#line 64 "ms_deisotope/_c/scoring.pyx"
;

#line 64 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_r = NULL;

#line 64 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 64 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 64 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2;

#line 64 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 64 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 64 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 64 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_create", 0);

#line 64 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_optional_args) {

#line 64 "ms_deisotope/_c/scoring.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 64 "ms_deisotope/_c/scoring.pyx"
      __pyx_v_data = __pyx_optional_args->data;

#line 64 "ms_deisotope/_c/scoring.pyx"
      if (__pyx_optional_args->__pyx_n > 1) {

#line 64 "ms_deisotope/_c/scoring.pyx"
        __pyx_v_missed_peaks = __pyx_optional_args->missed_peaks;

#line 64 "ms_deisotope/_c/scoring.pyx"
      }

#line 64 "ms_deisotope/_c/scoring.pyx"
    }

#line 64 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":66
 *                                    list experimental, object data=None, int missed_peaks=0):
 *         cdef IsotopicFitRecord inst
 *         inst = IsotopicFitRecord.__new__(IsotopicFitRecord)             # <<<<<<<<<<<<<<
 *         inst.seed_peak = seed_peak
 *         inst.score = score
 */

#line 66 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_12ms_deisotope_2_c_7scoring_IsotopicFitRecord(((PyTypeObject *)__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 66, __pyx_L1_error)

#line 66 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));

#line 66 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_inst = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_t_1);

#line 66 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/scoring.pyx":67
 *         cdef IsotopicFitRecord inst
 *         inst = IsotopicFitRecord.__new__(IsotopicFitRecord)
 *         inst.seed_peak = seed_peak             # <<<<<<<<<<<<<<
 *         inst.score = score
 *         inst.charge = charge
 */

#line 67 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_seed_peak));

#line 67 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_seed_peak));

#line 67 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_inst->seed_peak);

#line 67 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_inst->seed_peak));

#line 67 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_inst->seed_peak = __pyx_v_seed_peak;

  /* "ms_deisotope/_c/scoring.pyx":68
 *         inst = IsotopicFitRecord.__new__(IsotopicFitRecord)
 *         inst.seed_peak = seed_peak
 *         inst.score = score             # <<<<<<<<<<<<<<
 *         inst.charge = charge
 *         inst.theoretical = theoretical
 */

#line 68 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_inst->score = __pyx_v_score;

  /* "ms_deisotope/_c/scoring.pyx":69
 *         inst.seed_peak = seed_peak
 *         inst.score = score
 *         inst.charge = charge             # <<<<<<<<<<<<<<
 *         inst.theoretical = theoretical
 *         inst.experimental = experimental
 */

#line 69 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_inst->charge = __pyx_v_charge;

  /* "ms_deisotope/_c/scoring.pyx":70
 *         inst.score = score
 *         inst.charge = charge
 *         inst.theoretical = theoretical             # <<<<<<<<<<<<<<
 *         inst.experimental = experimental
 *         inst.monoisotopic_peak = <FittedPeak>PyList_GetItem(experimental, 0)
 */

#line 70 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_theoretical));

#line 70 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_theoretical));

#line 70 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_inst->theoretical);

#line 70 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_inst->theoretical));

#line 70 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_inst->theoretical = __pyx_v_theoretical;

  /* "ms_deisotope/_c/scoring.pyx":71
 *         inst.charge = charge
 *         inst.theoretical = theoretical
 *         inst.experimental = experimental             # <<<<<<<<<<<<<<
 *         inst.monoisotopic_peak = <FittedPeak>PyList_GetItem(experimental, 0)
 *         inst.data = data
 */

#line 71 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_v_experimental);

#line 71 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_v_experimental);

#line 71 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_inst->experimental);

#line 71 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_v_inst->experimental);

#line 71 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_inst->experimental = __pyx_v_experimental;

  /* "ms_deisotope/_c/scoring.pyx":72
 *         inst.theoretical = theoretical
 *         inst.experimental = experimental
 *         inst.monoisotopic_peak = <FittedPeak>PyList_GetItem(experimental, 0)             # <<<<<<<<<<<<<<
 *         inst.data = data
 *         inst.missed_peaks = missed_peaks
 */

#line 72 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = PyList_GetItem(__pyx_v_experimental, 0); if (unlikely(__pyx_t_2 == ((PyObject *)NULL))) __PYX_ERR(0, 72, __pyx_L1_error)

#line 72 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_t_2);

#line 72 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 72 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 72 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_inst->monoisotopic_peak);

#line 72 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_inst->monoisotopic_peak));

#line 72 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_inst->monoisotopic_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1);

#line 72 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/scoring.pyx":73
 *         inst.experimental = experimental
 *         inst.monoisotopic_peak = <FittedPeak>PyList_GetItem(experimental, 0)
 *         inst.data = data             # <<<<<<<<<<<<<<
 *         inst.missed_peaks = missed_peaks
 *         inst._hash = -1
 */

#line 73 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_v_data);

#line 73 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_v_data);

#line 73 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_inst->data);

#line 73 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_v_inst->data);

#line 73 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_inst->data = __pyx_v_data;

  /* "ms_deisotope/_c/scoring.pyx":74
 *         inst.monoisotopic_peak = <FittedPeak>PyList_GetItem(experimental, 0)
 *         inst.data = data
 *         inst.missed_peaks = missed_peaks             # <<<<<<<<<<<<<<
 *         inst._hash = -1
 *         return inst
 */

#line 74 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_inst->missed_peaks = __pyx_v_missed_peaks;

  /* "ms_deisotope/_c/scoring.pyx":75
 *         inst.data = data
 *         inst.missed_peaks = missed_peaks
 *         inst._hash = -1             # <<<<<<<<<<<<<<
 *         return inst
 * 
 */

#line 75 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_inst->_hash = -1L;

  /* "ms_deisotope/_c/scoring.pyx":76
 *         inst.missed_peaks = missed_peaks
 *         inst._hash = -1
 *         return inst             # <<<<<<<<<<<<<<
 * 
 *     def clone(self):
 */

#line 76 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 76 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_inst));

#line 76 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_v_inst;

#line 76 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":63
 * 
 *     @staticmethod
 *     cdef IsotopicFitRecord _create(FittedPeak seed_peak, double score, int charge, TheoreticalIsotopicPattern theoretical,             # <<<<<<<<<<<<<<
 *                                    list experimental, object data=None, int missed_peaks=0):
 *         cdef IsotopicFitRecord inst
 */

#line 63 "ms_deisotope/_c/scoring.pyx"


#line 63 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 63 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 63 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 63 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitRecord._create", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 63 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 63 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 63 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);

#line 63 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 63 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 63 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 63 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":78
 *         return inst
 * 
 *     def clone(self):             # <<<<<<<<<<<<<<
 *         return IsotopicFitRecord._create(self.seed_peak, self.score, self.charge, self.theoretical,
 *                               self.experimental, self.data, self.missed_peaks)
 */

#line 78 "ms_deisotope/_c/scoring.pyx"


#line 78 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 78 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_3clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 78 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_2clone[] = "IsotopicFitRecord.clone(self)";

#line 78 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_3clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 78 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 78 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 78 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("clone (wrapper)", 0);
  __pyx_r = 
#line 78 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_2clone(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self));

#line 78 "ms_deisotope/_c/scoring.pyx"


#line 78 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 78 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 78 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 78 "ms_deisotope/_c/scoring.pyx"
}

#line 78 "ms_deisotope/_c/scoring.pyx"


#line 78 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_2clone(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self) {

#line 78 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 78 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 78 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 78 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 78 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 78 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 78 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 78 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create __pyx_t_6;

#line 78 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 78 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 78 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 78 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("clone", 0);

  /* "ms_deisotope/_c/scoring.pyx":79
 * 
 *     def clone(self):
 *         return IsotopicFitRecord._create(self.seed_peak, self.score, self.charge, self.theoretical,             # <<<<<<<<<<<<<<
 *                               self.experimental, self.data, self.missed_peaks)
 * 
 */

#line 79 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 79 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_v_self->seed_peak);

#line 79 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 79 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = ((PyObject *)__pyx_v_self->theoretical);

#line 79 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_t_2);

  /* "ms_deisotope/_c/scoring.pyx":80
 *     def clone(self):
 *         return IsotopicFitRecord._create(self.seed_peak, self.score, self.charge, self.theoretical,
 *                               self.experimental, self.data, self.missed_peaks)             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */

#line 80 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = __pyx_v_self->experimental;

#line 80 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_t_3);

#line 80 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_4 = __pyx_v_self->data;

#line 80 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_t_4);

  /* "ms_deisotope/_c/scoring.pyx":79
 * 
 *     def clone(self):
 *         return IsotopicFitRecord._create(self.seed_peak, self.score, self.charge, self.theoretical,             # <<<<<<<<<<<<<<
 *                               self.experimental, self.data, self.missed_peaks)
 * 
 */

#line 79 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_6.__pyx_n = 2;

#line 79 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_6.data = __pyx_t_4;

#line 79 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_6.missed_peaks = __pyx_v_self->missed_peaks;

#line 79 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_5 = ((PyObject *)__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitRecord->_create(((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1), __pyx_v_self->score, __pyx_v_self->charge, ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_t_2), ((PyObject*)__pyx_t_3), &__pyx_t_6)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 79, __pyx_L1_error)

#line 79 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 79 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 79 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 79 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 79 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 79 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_5;

#line 79 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_5 = 0;

#line 79 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":78
 *         return inst
 * 
 *     def clone(self):             # <<<<<<<<<<<<<<
 *         return IsotopicFitRecord._create(self.seed_peak, self.score, self.charge, self.theoretical,
 *                               self.experimental, self.data, self.missed_peaks)
 */

#line 78 "ms_deisotope/_c/scoring.pyx"


#line 78 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 78 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 78 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 78 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 78 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 78 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 78 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 78 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitRecord.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 78 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 78 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 78 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 78 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 78 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 78 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":82
 *                               self.experimental, self.data, self.missed_peaks)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.seed_peak, self.score, self.charge, self.theoretical,
 *                                 self.experimental, self.data, self.missed_peaks)
 */

#line 82 "ms_deisotope/_c/scoring.pyx"


#line 82 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 82 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_5__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 82 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_4__reduce__[] = "IsotopicFitRecord.__reduce__(self)";

#line 82 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_5__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 82 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 82 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 82 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = 
#line 82 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_4__reduce__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self));

#line 82 "ms_deisotope/_c/scoring.pyx"


#line 82 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 82 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 82 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 82 "ms_deisotope/_c/scoring.pyx"
}

#line 82 "ms_deisotope/_c/scoring.pyx"


#line 82 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_4__reduce__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self) {

#line 82 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 82 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 82 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 82 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 82 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 82 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 82 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 82 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 82 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 82 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 82 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "ms_deisotope/_c/scoring.pyx":83
 * 
 *     def __reduce__(self):
 *         return self.__class__, (self.seed_peak, self.score, self.charge, self.theoretical,             # <<<<<<<<<<<<<<
 *                                 self.experimental, self.data, self.missed_peaks)
 * 
 */

#line 83 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 83 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 83, __pyx_L1_error)

#line 83 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 83 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->score); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 83, __pyx_L1_error)

#line 83 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 83 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->charge); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 83, __pyx_L1_error)

#line 83 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_3);

  /* "ms_deisotope/_c/scoring.pyx":84
 *     def __reduce__(self):
 *         return self.__class__, (self.seed_peak, self.score, self.charge, self.theoretical,
 *                                 self.experimental, self.data, self.missed_peaks)             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint _eq(self, IsotopicFitRecord other):
 */

#line 84 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_self->missed_peaks); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 84, __pyx_L1_error)

#line 84 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_4);

  /* "ms_deisotope/_c/scoring.pyx":83
 * 
 *     def __reduce__(self):
 *         return self.__class__, (self.seed_peak, self.score, self.charge, self.theoretical,             # <<<<<<<<<<<<<<
 *                                 self.experimental, self.data, self.missed_peaks)
 * 
 */

#line 83 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_5 = PyTuple_New(7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 83, __pyx_L1_error)

#line 83 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 83 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->seed_peak));

#line 83 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->seed_peak));

#line 83 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_v_self->seed_peak));

#line 83 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 83 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);

#line 83 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 83 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);

#line 83 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->theoretical));

#line 83 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->theoretical));

#line 83 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_5, 3, ((PyObject *)__pyx_v_self->theoretical));

#line 83 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_v_self->experimental);

#line 83 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_v_self->experimental);

#line 83 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_v_self->experimental);

#line 83 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_v_self->data);

#line 83 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_v_self->data);

#line 83 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_5, 5, __pyx_v_self->data);

#line 83 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_4);

#line 83 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_5, 6, __pyx_t_4);

#line 83 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = 0;

#line 83 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = 0;

#line 83 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_4 = 0;

#line 83 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 83, __pyx_L1_error)

#line 83 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 83 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 83 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);

#line 83 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_5);

#line 83 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);

#line 83 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 83 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_5 = 0;

#line 83 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_4;

#line 83 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_4 = 0;

#line 83 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":82
 *                               self.experimental, self.data, self.missed_peaks)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.seed_peak, self.score, self.charge, self.theoretical,
 *                                 self.experimental, self.data, self.missed_peaks)
 */

#line 82 "ms_deisotope/_c/scoring.pyx"


#line 82 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 82 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 82 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 82 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 82 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 82 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 82 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 82 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitRecord.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 82 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 82 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 82 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 82 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 82 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 82 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":86
 *                                 self.experimental, self.data, self.missed_peaks)
 * 
 *     cpdef bint _eq(self, IsotopicFitRecord other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             bint val
 */

#line 86 "ms_deisotope/_c/scoring.pyx"


#line 86 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_7_eq(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/

#line 86 "ms_deisotope/_c/scoring.pyx"
static int __pyx_f_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__eq(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_other, int __pyx_skip_dispatch) {
  int __pyx_v_val
#line 86 "ms_deisotope/_c/scoring.pyx"
;
  size_t __pyx_v_i
#line 86 "ms_deisotope/_c/scoring.pyx"
;
  size_t __pyx_v_n
#line 86 "ms_deisotope/_c/scoring.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_fp1 = 0
#line 86 "ms_deisotope/_c/scoring.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_fp2 = 0
#line 86 "ms_deisotope/_c/scoring.pyx"
;

#line 86 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 86 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 86 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 86 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 86 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 86 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 86 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_5;

#line 86 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_6;

#line 86 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_7;

#line 86 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_8;

#line 86 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_9;

#line 86 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_10;

#line 86 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_11;

#line 86 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 86 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 86 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 86 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_eq", 0);

#line 86 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 86 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 86 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 86 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 86 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 86 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 86 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 86 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 86 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 86 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_eq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 86, __pyx_L1_error)

#line 86 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 86 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_7_eq)) {

#line 86 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 86 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 86 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 86 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 86 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 86 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 86 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 86 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 86 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 86 "ms_deisotope/_c/scoring.pyx"
          }

#line 86 "ms_deisotope/_c/scoring.pyx"
        }

#line 86 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_other)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other));

#line 86 "ms_deisotope/_c/scoring.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 86 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 86, __pyx_L1_error)

#line 86 "ms_deisotope/_c/scoring.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 86 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 86 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 86, __pyx_L1_error)

#line 86 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 86 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = __pyx_t_5;

#line 86 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 86 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 86 "ms_deisotope/_c/scoring.pyx"
      }

#line 86 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 86 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 86 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 86 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 86 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 86 "ms_deisotope/_c/scoring.pyx"
      }

#line 86 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 86 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 86 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 86 "ms_deisotope/_c/scoring.pyx"
    }

#line 86 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 86 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":92
 *             FittedPeak fp1, fp2
 * 
 *         val = self.score == other.score             # <<<<<<<<<<<<<<
 *         if not val:
 *             return val
 */

#line 92 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_val = (__pyx_v_self->score == __pyx_v_other->score);

  /* "ms_deisotope/_c/scoring.pyx":93
 * 
 *         val = self.score == other.score
 *         if not val:             # <<<<<<<<<<<<<<
 *             return val
 *         val = self.charge == other.charge
 */

#line 93 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_5 = ((!(__pyx_v_val != 0)) != 0);

#line 93 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/scoring.pyx":94
 *         val = self.score == other.score
 *         if not val:
 *             return val             # <<<<<<<<<<<<<<
 *         val = self.charge == other.charge
 *         if not val:
 */

#line 94 "ms_deisotope/_c/scoring.pyx"
    __pyx_r = __pyx_v_val;

#line 94 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/scoring.pyx":93
 * 
 *         val = self.score == other.score
 *         if not val:             # <<<<<<<<<<<<<<
 *             return val
 *         val = self.charge == other.charge
 */

#line 93 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":95
 *         if not val:
 *             return val
 *         val = self.charge == other.charge             # <<<<<<<<<<<<<<
 *         if not val:
 *             return val
 */

#line 95 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_val = (__pyx_v_self->charge == __pyx_v_other->charge);

  /* "ms_deisotope/_c/scoring.pyx":96
 *             return val
 *         val = self.charge == other.charge
 *         if not val:             # <<<<<<<<<<<<<<
 *             return val
 *         n = PyList_GET_SIZE(self.experimental)
 */

#line 96 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_5 = ((!(__pyx_v_val != 0)) != 0);

#line 96 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/scoring.pyx":97
 *         val = self.charge == other.charge
 *         if not val:
 *             return val             # <<<<<<<<<<<<<<
 *         n = PyList_GET_SIZE(self.experimental)
 *         val = n == PyList_GET_SIZE(other.experimental)
 */

#line 97 "ms_deisotope/_c/scoring.pyx"
    __pyx_r = __pyx_v_val;

#line 97 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/scoring.pyx":96
 *             return val
 *         val = self.charge == other.charge
 *         if not val:             # <<<<<<<<<<<<<<
 *             return val
 *         n = PyList_GET_SIZE(self.experimental)
 */

#line 96 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":98
 *         if not val:
 *             return val
 *         n = PyList_GET_SIZE(self.experimental)             # <<<<<<<<<<<<<<
 *         val = n == PyList_GET_SIZE(other.experimental)
 *         if not val:
 */

#line 98 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __pyx_v_self->experimental;

#line 98 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 98 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_n = PyList_GET_SIZE(__pyx_t_1);

#line 98 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/scoring.pyx":99
 *             return val
 *         n = PyList_GET_SIZE(self.experimental)
 *         val = n == PyList_GET_SIZE(other.experimental)             # <<<<<<<<<<<<<<
 *         if not val:
 *             return val
 */

#line 99 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __pyx_v_other->experimental;

#line 99 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 99 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_val = (__pyx_v_n == PyList_GET_SIZE(__pyx_t_1));

#line 99 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/scoring.pyx":100
 *         n = PyList_GET_SIZE(self.experimental)
 *         val = n == PyList_GET_SIZE(other.experimental)
 *         if not val:             # <<<<<<<<<<<<<<
 *             return val
 *         for i in range(n):
 */

#line 100 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_5 = ((!(__pyx_v_val != 0)) != 0);

#line 100 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/scoring.pyx":101
 *         val = n == PyList_GET_SIZE(other.experimental)
 *         if not val:
 *             return val             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             fp1 = <FittedPeak>PyList_GET_ITEM(self.experimental, i)
 */

#line 101 "ms_deisotope/_c/scoring.pyx"
    __pyx_r = __pyx_v_val;

#line 101 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/scoring.pyx":100
 *         n = PyList_GET_SIZE(self.experimental)
 *         val = n == PyList_GET_SIZE(other.experimental)
 *         if not val:             # <<<<<<<<<<<<<<
 *             return val
 *         for i in range(n):
 */

#line 100 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":102
 *         if not val:
 *             return val
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             fp1 = <FittedPeak>PyList_GET_ITEM(self.experimental, i)
 *             fp2 = <FittedPeak>PyList_GET_ITEM(other.experimental, i)
 */

#line 102 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_6 = __pyx_v_n;

#line 102 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_7 = __pyx_t_6;

#line 102 "ms_deisotope/_c/scoring.pyx"
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {

#line 102 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_i = __pyx_t_8;

    /* "ms_deisotope/_c/scoring.pyx":103
 *             return val
 *         for i in range(n):
 *             fp1 = <FittedPeak>PyList_GET_ITEM(self.experimental, i)             # <<<<<<<<<<<<<<
 *             fp2 = <FittedPeak>PyList_GET_ITEM(other.experimental, i)
 *             val = fp1._eq(fp2)
 */

#line 103 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = __pyx_v_self->experimental;

#line 103 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 103 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_9 = PyList_GET_ITEM(__pyx_t_1, __pyx_v_i);

#line 103 "ms_deisotope/_c/scoring.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 103 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_9);

#line 103 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 103 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF_SET(__pyx_v_fp1, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1));

#line 103 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/scoring.pyx":104
 *         for i in range(n):
 *             fp1 = <FittedPeak>PyList_GET_ITEM(self.experimental, i)
 *             fp2 = <FittedPeak>PyList_GET_ITEM(other.experimental, i)             # <<<<<<<<<<<<<<
 *             val = fp1._eq(fp2)
 *             if not val:
 */

#line 104 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = __pyx_v_other->experimental;

#line 104 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 104 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_9 = PyList_GET_ITEM(__pyx_t_1, __pyx_v_i);

#line 104 "ms_deisotope/_c/scoring.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 104 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_9);

#line 104 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 104 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF_SET(__pyx_v_fp2, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1));

#line 104 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/scoring.pyx":105
 *             fp1 = <FittedPeak>PyList_GET_ITEM(self.experimental, i)
 *             fp2 = <FittedPeak>PyList_GET_ITEM(other.experimental, i)
 *             val = fp1._eq(fp2)             # <<<<<<<<<<<<<<
 *             if not val:
 *                 return val
 */

#line 105 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_val = ((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_v_fp1->__pyx_base.__pyx_vtab)->_eq(__pyx_v_fp1, __pyx_v_fp2, 0);

    /* "ms_deisotope/_c/scoring.pyx":106
 *             fp2 = <FittedPeak>PyList_GET_ITEM(other.experimental, i)
 *             val = fp1._eq(fp2)
 *             if not val:             # <<<<<<<<<<<<<<
 *                 return val
 *         val = self.theoretical._eq_inst(other.theoretical)
 */

#line 106 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = ((!(__pyx_v_val != 0)) != 0);

#line 106 "ms_deisotope/_c/scoring.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/scoring.pyx":107
 *             val = fp1._eq(fp2)
 *             if not val:
 *                 return val             # <<<<<<<<<<<<<<
 *         val = self.theoretical._eq_inst(other.theoretical)
 *         if not val:
 */

#line 107 "ms_deisotope/_c/scoring.pyx"
      __pyx_r = __pyx_v_val;

#line 107 "ms_deisotope/_c/scoring.pyx"
      goto __pyx_L0;

      /* "ms_deisotope/_c/scoring.pyx":106
 *             fp2 = <FittedPeak>PyList_GET_ITEM(other.experimental, i)
 *             val = fp1._eq(fp2)
 *             if not val:             # <<<<<<<<<<<<<<
 *                 return val
 *         val = self.theoretical._eq_inst(other.theoretical)
 */

#line 106 "ms_deisotope/_c/scoring.pyx"
    }

#line 106 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":108
 *             if not val:
 *                 return val
 *         val = self.theoretical._eq_inst(other.theoretical)             # <<<<<<<<<<<<<<
 *         if not val:
 *             return val
 */

#line 108 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_v_other->theoretical);

#line 108 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 108 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_val = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_v_self->theoretical->__pyx_vtab)->_eq_inst(__pyx_v_self->theoretical, ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_t_1));

#line 108 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/scoring.pyx":109
 *                 return val
 *         val = self.theoretical._eq_inst(other.theoretical)
 *         if not val:             # <<<<<<<<<<<<<<
 *             return val
 *         if self.data is not None or other.data is not None:
 */

#line 109 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_5 = ((!(__pyx_v_val != 0)) != 0);

#line 109 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/scoring.pyx":110
 *         val = self.theoretical._eq_inst(other.theoretical)
 *         if not val:
 *             return val             # <<<<<<<<<<<<<<
 *         if self.data is not None or other.data is not None:
 *             val = (self.data == other.data)
 */

#line 110 "ms_deisotope/_c/scoring.pyx"
    __pyx_r = __pyx_v_val;

#line 110 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/scoring.pyx":109
 *                 return val
 *         val = self.theoretical._eq_inst(other.theoretical)
 *         if not val:             # <<<<<<<<<<<<<<
 *             return val
 *         if self.data is not None or other.data is not None:
 */

#line 109 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":111
 *         if not val:
 *             return val
 *         if self.data is not None or other.data is not None:             # <<<<<<<<<<<<<<
 *             val = (self.data == other.data)
 *         return val
 */

#line 111 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_10 = (__pyx_v_self->data != Py_None);

#line 111 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_11 = (__pyx_t_10 != 0);

#line 111 "ms_deisotope/_c/scoring.pyx"
  if (!__pyx_t_11) {

#line 111 "ms_deisotope/_c/scoring.pyx"
  } else {

#line 111 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = __pyx_t_11;

#line 111 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L11_bool_binop_done;

#line 111 "ms_deisotope/_c/scoring.pyx"
  }

#line 111 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_11 = (__pyx_v_other->data != Py_None);

#line 111 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_10 = (__pyx_t_11 != 0);

#line 111 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_5 = __pyx_t_10;

#line 111 "ms_deisotope/_c/scoring.pyx"
  __pyx_L11_bool_binop_done:;

#line 111 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/scoring.pyx":112
 *             return val
 *         if self.data is not None or other.data is not None:
 *             val = (self.data == other.data)             # <<<<<<<<<<<<<<
 *         return val
 * 
 */

#line 112 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self->data, __pyx_v_other->data, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 112, __pyx_L1_error)

#line 112 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 112, __pyx_L1_error)

#line 112 "ms_deisotope/_c/scoring.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 112 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_val = __pyx_t_5;

    /* "ms_deisotope/_c/scoring.pyx":111
 *         if not val:
 *             return val
 *         if self.data is not None or other.data is not None:             # <<<<<<<<<<<<<<
 *             val = (self.data == other.data)
 *         return val
 */

#line 111 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":113
 *         if self.data is not None or other.data is not None:
 *             val = (self.data == other.data)
 *         return val             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint _ne(self, IsotopicFitRecord other):
 */

#line 113 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_v_val;

#line 113 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":86
 *                                 self.experimental, self.data, self.missed_peaks)
 * 
 *     cpdef bint _eq(self, IsotopicFitRecord other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             bint val
 */

#line 86 "ms_deisotope/_c/scoring.pyx"


#line 86 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 86 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 86 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 86 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 86 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 86 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 86 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.IsotopicFitRecord._eq", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 86 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 86 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 86 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_fp1);

#line 86 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_fp2);

#line 86 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 86 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 86 "ms_deisotope/_c/scoring.pyx"
}

#line 86 "ms_deisotope/_c/scoring.pyx"


#line 86 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 86 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_7_eq(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/

#line 86 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_6_eq[] = "IsotopicFitRecord._eq(self, IsotopicFitRecord other) -> bool";

#line 86 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_7_eq(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {

#line 86 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 86 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 86 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 86 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 86 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 86 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_eq (wrapper)", 0);

#line 86 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord, 1, "other", 0))) __PYX_ERR(0, 86, __pyx_L1_error)
  __pyx_r = 
#line 86 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_6_eq(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_other));

#line 86 "ms_deisotope/_c/scoring.pyx"


#line 86 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 86 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 86 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 86 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 86 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 86 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 86 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 86 "ms_deisotope/_c/scoring.pyx"
}

#line 86 "ms_deisotope/_c/scoring.pyx"


#line 86 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_6_eq(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_other) {

#line 86 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 86 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 86 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 86 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 86 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 86 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 86 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_eq", 0);

#line 86 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 86 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__eq(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 86, __pyx_L1_error)

#line 86 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 86 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 86 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 86 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 86 "ms_deisotope/_c/scoring.pyx"


#line 86 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 86 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 86 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 86 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitRecord._eq", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 86 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 86 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 86 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 86 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 86 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 86 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":115
 *         return val
 * 
 *     cpdef bint _ne(self, IsotopicFitRecord other):             # <<<<<<<<<<<<<<
 *         return not (self._eq(other))
 * 
 */

#line 115 "ms_deisotope/_c/scoring.pyx"


#line 115 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_9_ne(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/

#line 115 "ms_deisotope/_c/scoring.pyx"
static int __pyx_f_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__ne(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_other, int __pyx_skip_dispatch) {

#line 115 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 115 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 115 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 115 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 115 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 115 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 115 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_5;

#line 115 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 115 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 115 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 115 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_ne", 0);

#line 115 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 115 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 115 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 115 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 115 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 115 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 115 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 115 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 115 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 115 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_ne); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 115, __pyx_L1_error)

#line 115 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 115 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_9_ne)) {

#line 115 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 115 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 115 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 115 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 115 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 115 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 115 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 115 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 115 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 115 "ms_deisotope/_c/scoring.pyx"
          }

#line 115 "ms_deisotope/_c/scoring.pyx"
        }

#line 115 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_other)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other));

#line 115 "ms_deisotope/_c/scoring.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 115 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 115, __pyx_L1_error)

#line 115 "ms_deisotope/_c/scoring.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 115 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 115 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 115, __pyx_L1_error)

#line 115 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 115 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = __pyx_t_5;

#line 115 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 115 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 115 "ms_deisotope/_c/scoring.pyx"
      }

#line 115 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 115 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 115 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 115 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 115 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 115 "ms_deisotope/_c/scoring.pyx"
      }

#line 115 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 115 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 115 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 115 "ms_deisotope/_c/scoring.pyx"
    }

#line 115 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 115 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":116
 * 
 *     cpdef bint _ne(self, IsotopicFitRecord other):
 *         return not (self._eq(other))             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint _lt(self, IsotopicFitRecord other):
 */

#line 116 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = (!(((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self->__pyx_vtab)->_eq(__pyx_v_self, __pyx_v_other, 0) != 0));

#line 116 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":115
 *         return val
 * 
 *     cpdef bint _ne(self, IsotopicFitRecord other):             # <<<<<<<<<<<<<<
 *         return not (self._eq(other))
 * 
 */

#line 115 "ms_deisotope/_c/scoring.pyx"


#line 115 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 115 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 115 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 115 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 115 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 115 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 115 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.IsotopicFitRecord._ne", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 115 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 115 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 115 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 115 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 115 "ms_deisotope/_c/scoring.pyx"
}

#line 115 "ms_deisotope/_c/scoring.pyx"


#line 115 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 115 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_9_ne(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/

#line 115 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_8_ne[] = "IsotopicFitRecord._ne(self, IsotopicFitRecord other) -> bool";

#line 115 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_9_ne(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {

#line 115 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 115 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 115 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 115 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 115 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 115 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_ne (wrapper)", 0);

#line 115 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord, 1, "other", 0))) __PYX_ERR(0, 115, __pyx_L1_error)
  __pyx_r = 
#line 115 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_8_ne(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_other));

#line 115 "ms_deisotope/_c/scoring.pyx"


#line 115 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 115 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 115 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 115 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 115 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 115 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 115 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 115 "ms_deisotope/_c/scoring.pyx"
}

#line 115 "ms_deisotope/_c/scoring.pyx"


#line 115 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_8_ne(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_other) {

#line 115 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 115 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 115 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 115 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 115 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 115 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 115 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_ne", 0);

#line 115 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 115 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__ne(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 115, __pyx_L1_error)

#line 115 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 115 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 115 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 115 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 115 "ms_deisotope/_c/scoring.pyx"


#line 115 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 115 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 115 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 115 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitRecord._ne", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 115 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 115 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 115 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 115 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 115 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 115 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":118
 *         return not (self._eq(other))
 * 
 *     cpdef bint _lt(self, IsotopicFitRecord other):             # <<<<<<<<<<<<<<
 *         return self.score < other.score
 * 
 */

#line 118 "ms_deisotope/_c/scoring.pyx"


#line 118 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_11_lt(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/

#line 118 "ms_deisotope/_c/scoring.pyx"
static int __pyx_f_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__lt(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_other, int __pyx_skip_dispatch) {

#line 118 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 118 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 118 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 118 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 118 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 118 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 118 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_5;

#line 118 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 118 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 118 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 118 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_lt", 0);

#line 118 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 118 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 118 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 118 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 118 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 118 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 118 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 118 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 118 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 118 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_lt); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L1_error)

#line 118 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 118 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_11_lt)) {

#line 118 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 118 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 118 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 118 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 118 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 118 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 118 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 118 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 118 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 118 "ms_deisotope/_c/scoring.pyx"
          }

#line 118 "ms_deisotope/_c/scoring.pyx"
        }

#line 118 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_other)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other));

#line 118 "ms_deisotope/_c/scoring.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 118 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 118, __pyx_L1_error)

#line 118 "ms_deisotope/_c/scoring.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 118 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 118 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 118, __pyx_L1_error)

#line 118 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 118 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = __pyx_t_5;

#line 118 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 118 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 118 "ms_deisotope/_c/scoring.pyx"
      }

#line 118 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 118 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 118 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 118 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 118 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 118 "ms_deisotope/_c/scoring.pyx"
      }

#line 118 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 118 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 118 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 118 "ms_deisotope/_c/scoring.pyx"
    }

#line 118 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 118 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":119
 * 
 *     cpdef bint _lt(self, IsotopicFitRecord other):
 *         return self.score < other.score             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint _gt(self, IsotopicFitRecord other):
 */

#line 119 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = (__pyx_v_self->score < __pyx_v_other->score);

#line 119 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":118
 *         return not (self._eq(other))
 * 
 *     cpdef bint _lt(self, IsotopicFitRecord other):             # <<<<<<<<<<<<<<
 *         return self.score < other.score
 * 
 */

#line 118 "ms_deisotope/_c/scoring.pyx"


#line 118 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 118 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 118 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 118 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 118 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 118 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 118 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.IsotopicFitRecord._lt", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 118 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 118 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 118 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 118 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 118 "ms_deisotope/_c/scoring.pyx"
}

#line 118 "ms_deisotope/_c/scoring.pyx"


#line 118 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 118 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_11_lt(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/

#line 118 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_10_lt[] = "IsotopicFitRecord._lt(self, IsotopicFitRecord other) -> bool";

#line 118 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_11_lt(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {

#line 118 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 118 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 118 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 118 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 118 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 118 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_lt (wrapper)", 0);

#line 118 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord, 1, "other", 0))) __PYX_ERR(0, 118, __pyx_L1_error)
  __pyx_r = 
#line 118 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_10_lt(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_other));

#line 118 "ms_deisotope/_c/scoring.pyx"


#line 118 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 118 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 118 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 118 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 118 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 118 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 118 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 118 "ms_deisotope/_c/scoring.pyx"
}

#line 118 "ms_deisotope/_c/scoring.pyx"


#line 118 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_10_lt(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_other) {

#line 118 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 118 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 118 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 118 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 118 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 118 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 118 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_lt", 0);

#line 118 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 118 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__lt(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L1_error)

#line 118 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 118 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 118 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 118 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 118 "ms_deisotope/_c/scoring.pyx"


#line 118 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 118 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 118 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 118 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitRecord._lt", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 118 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 118 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 118 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 118 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 118 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 118 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":121
 *         return self.score < other.score
 * 
 *     cpdef bint _gt(self, IsotopicFitRecord other):             # <<<<<<<<<<<<<<
 *         return self.score > other.score
 * 
 */

#line 121 "ms_deisotope/_c/scoring.pyx"


#line 121 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_13_gt(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/

#line 121 "ms_deisotope/_c/scoring.pyx"
static int __pyx_f_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__gt(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_other, int __pyx_skip_dispatch) {

#line 121 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 121 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 121 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 121 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 121 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 121 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 121 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_5;

#line 121 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 121 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 121 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 121 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_gt", 0);

#line 121 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 121 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 121 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 121 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 121 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 121 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 121 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 121 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 121 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 121 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_gt); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 121, __pyx_L1_error)

#line 121 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 121 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_13_gt)) {

#line 121 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 121 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 121 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 121 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 121 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 121 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 121 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 121 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 121 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 121 "ms_deisotope/_c/scoring.pyx"
          }

#line 121 "ms_deisotope/_c/scoring.pyx"
        }

#line 121 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_other)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other));

#line 121 "ms_deisotope/_c/scoring.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 121 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 121, __pyx_L1_error)

#line 121 "ms_deisotope/_c/scoring.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 121 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 121 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 121, __pyx_L1_error)

#line 121 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 121 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = __pyx_t_5;

#line 121 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 121 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 121 "ms_deisotope/_c/scoring.pyx"
      }

#line 121 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 121 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 121 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 121 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 121 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 121 "ms_deisotope/_c/scoring.pyx"
      }

#line 121 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 121 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 121 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 121 "ms_deisotope/_c/scoring.pyx"
    }

#line 121 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 121 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":122
 * 
 *     cpdef bint _gt(self, IsotopicFitRecord other):
 *         return self.score > other.score             # <<<<<<<<<<<<<<
 * 
 *     def __richcmp__(self, IsotopicFitRecord other, int code):
 */

#line 122 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = (__pyx_v_self->score > __pyx_v_other->score);

#line 122 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":121
 *         return self.score < other.score
 * 
 *     cpdef bint _gt(self, IsotopicFitRecord other):             # <<<<<<<<<<<<<<
 *         return self.score > other.score
 * 
 */

#line 121 "ms_deisotope/_c/scoring.pyx"


#line 121 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 121 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 121 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 121 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 121 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 121 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 121 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.IsotopicFitRecord._gt", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 121 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 121 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 121 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 121 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 121 "ms_deisotope/_c/scoring.pyx"
}

#line 121 "ms_deisotope/_c/scoring.pyx"


#line 121 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 121 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_13_gt(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/

#line 121 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12_gt[] = "IsotopicFitRecord._gt(self, IsotopicFitRecord other) -> bool";

#line 121 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_13_gt(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {

#line 121 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 121 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 121 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 121 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 121 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 121 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_gt (wrapper)", 0);

#line 121 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord, 1, "other", 0))) __PYX_ERR(0, 121, __pyx_L1_error)
  __pyx_r = 
#line 121 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12_gt(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_other));

#line 121 "ms_deisotope/_c/scoring.pyx"


#line 121 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 121 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 121 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 121 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 121 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 121 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 121 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 121 "ms_deisotope/_c/scoring.pyx"
}

#line 121 "ms_deisotope/_c/scoring.pyx"


#line 121 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12_gt(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_other) {

#line 121 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 121 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 121 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 121 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 121 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 121 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 121 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_gt", 0);

#line 121 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 121 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__gt(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 121, __pyx_L1_error)

#line 121 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 121 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 121 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 121 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 121 "ms_deisotope/_c/scoring.pyx"


#line 121 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 121 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 121 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 121 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitRecord._gt", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 121 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 121 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 121 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 121 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 121 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 121 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":124
 *         return self.score > other.score
 * 
 *     def __richcmp__(self, IsotopicFitRecord other, int code):             # <<<<<<<<<<<<<<
 *         if other is None:
 *             if code == 3:
 */

#line 124 "ms_deisotope/_c/scoring.pyx"


#line 124 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 124 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_15__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code); /*proto*/

#line 124 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_15__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code) {

#line 124 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 124 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 124 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 124 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 124 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 124 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);

#line 124 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord, 1, "other", 0))) __PYX_ERR(0, 124, __pyx_L1_error)
  __pyx_r = 
#line 124 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_14__richcmp__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_other), ((int)__pyx_v_code));

#line 124 "ms_deisotope/_c/scoring.pyx"


#line 124 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 124 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 124 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 124 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 124 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 124 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 124 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 124 "ms_deisotope/_c/scoring.pyx"
}

#line 124 "ms_deisotope/_c/scoring.pyx"


#line 124 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_14__richcmp__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_other, int __pyx_v_code) {

#line 124 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 124 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 124 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_1;

#line 124 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_2;

#line 124 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 124 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 124 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 124 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 124 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "ms_deisotope/_c/scoring.pyx":125
 * 
 *     def __richcmp__(self, IsotopicFitRecord other, int code):
 *         if other is None:             # <<<<<<<<<<<<<<
 *             if code == 3:
 *                 return True
 */

#line 125 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = (((PyObject *)__pyx_v_other) == Py_None);

#line 125 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 125 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/scoring.pyx":126
 *     def __richcmp__(self, IsotopicFitRecord other, int code):
 *         if other is None:
 *             if code == 3:             # <<<<<<<<<<<<<<
 *                 return True
 *             else:
 */

#line 126 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_2 = ((__pyx_v_code == 3) != 0);

#line 126 "ms_deisotope/_c/scoring.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/scoring.pyx":127
 *         if other is None:
 *             if code == 3:
 *                 return True             # <<<<<<<<<<<<<<
 *             else:
 *                 return False
 */

#line 127 "ms_deisotope/_c/scoring.pyx"
      __Pyx_XDECREF(__pyx_r);

#line 127 "ms_deisotope/_c/scoring.pyx"
      __Pyx_INCREF(Py_True);

#line 127 "ms_deisotope/_c/scoring.pyx"
      __pyx_r = Py_True;

#line 127 "ms_deisotope/_c/scoring.pyx"
      goto __pyx_L0;

      /* "ms_deisotope/_c/scoring.pyx":126
 *     def __richcmp__(self, IsotopicFitRecord other, int code):
 *         if other is None:
 *             if code == 3:             # <<<<<<<<<<<<<<
 *                 return True
 *             else:
 */

#line 126 "ms_deisotope/_c/scoring.pyx"
    }

    /* "ms_deisotope/_c/scoring.pyx":129
 *                 return True
 *             else:
 *                 return False             # <<<<<<<<<<<<<<
 * 
 *         if code == 0:
 */

#line 129 "ms_deisotope/_c/scoring.pyx"
    /*else*/ {

#line 129 "ms_deisotope/_c/scoring.pyx"
      __Pyx_XDECREF(__pyx_r);

#line 129 "ms_deisotope/_c/scoring.pyx"
      __Pyx_INCREF(Py_False);

#line 129 "ms_deisotope/_c/scoring.pyx"
      __pyx_r = Py_False;

#line 129 "ms_deisotope/_c/scoring.pyx"
      goto __pyx_L0;

#line 129 "ms_deisotope/_c/scoring.pyx"
    }

    /* "ms_deisotope/_c/scoring.pyx":125
 * 
 *     def __richcmp__(self, IsotopicFitRecord other, int code):
 *         if other is None:             # <<<<<<<<<<<<<<
 *             if code == 3:
 *                 return True
 */

#line 125 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":131
 *                 return False
 * 
 *         if code == 0:             # <<<<<<<<<<<<<<
 *             return self._lt(other)
 *         elif code == 2:
 */

#line 131 "ms_deisotope/_c/scoring.pyx"
  switch (__pyx_v_code) {

#line 131 "ms_deisotope/_c/scoring.pyx"
    case 0:

    /* "ms_deisotope/_c/scoring.pyx":132
 * 
 *         if code == 0:
 *             return self._lt(other)             # <<<<<<<<<<<<<<
 *         elif code == 2:
 *             return self._eq(other)
 */

#line 132 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 132 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_3 = __Pyx_PyBool_FromLong(((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self->__pyx_vtab)->_lt(__pyx_v_self, __pyx_v_other, 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 132, __pyx_L1_error)

#line 132 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 132 "ms_deisotope/_c/scoring.pyx"
    __pyx_r = __pyx_t_3;

#line 132 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_3 = 0;

#line 132 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/scoring.pyx":131
 *                 return False
 * 
 *         if code == 0:             # <<<<<<<<<<<<<<
 *             return self._lt(other)
 *         elif code == 2:
 */

#line 131 "ms_deisotope/_c/scoring.pyx"
    break;

#line 131 "ms_deisotope/_c/scoring.pyx"
    case 2:

    /* "ms_deisotope/_c/scoring.pyx":134
 *             return self._lt(other)
 *         elif code == 2:
 *             return self._eq(other)             # <<<<<<<<<<<<<<
 *         elif code == 3:
 *             return not (self._eq(other))
 */

#line 134 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 134 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_3 = __Pyx_PyBool_FromLong(((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self->__pyx_vtab)->_eq(__pyx_v_self, __pyx_v_other, 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 134, __pyx_L1_error)

#line 134 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 134 "ms_deisotope/_c/scoring.pyx"
    __pyx_r = __pyx_t_3;

#line 134 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_3 = 0;

#line 134 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/scoring.pyx":133
 *         if code == 0:
 *             return self._lt(other)
 *         elif code == 2:             # <<<<<<<<<<<<<<
 *             return self._eq(other)
 *         elif code == 3:
 */

#line 133 "ms_deisotope/_c/scoring.pyx"
    break;

#line 133 "ms_deisotope/_c/scoring.pyx"
    case 3:

    /* "ms_deisotope/_c/scoring.pyx":136
 *             return self._eq(other)
 *         elif code == 3:
 *             return not (self._eq(other))             # <<<<<<<<<<<<<<
 *         elif code == 4:
 *             return self._gt(other)
 */

#line 136 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 136 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_3 = __Pyx_PyBool_FromLong((!(((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self->__pyx_vtab)->_eq(__pyx_v_self, __pyx_v_other, 0) != 0))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 136, __pyx_L1_error)

#line 136 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 136 "ms_deisotope/_c/scoring.pyx"
    __pyx_r = __pyx_t_3;

#line 136 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_3 = 0;

#line 136 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/scoring.pyx":135
 *         elif code == 2:
 *             return self._eq(other)
 *         elif code == 3:             # <<<<<<<<<<<<<<
 *             return not (self._eq(other))
 *         elif code == 4:
 */

#line 135 "ms_deisotope/_c/scoring.pyx"
    break;

#line 135 "ms_deisotope/_c/scoring.pyx"
    case 4:

    /* "ms_deisotope/_c/scoring.pyx":138
 *             return not (self._eq(other))
 *         elif code == 4:
 *             return self._gt(other)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 138 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 138 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_3 = __Pyx_PyBool_FromLong(((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self->__pyx_vtab)->_gt(__pyx_v_self, __pyx_v_other, 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 138, __pyx_L1_error)

#line 138 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 138 "ms_deisotope/_c/scoring.pyx"
    __pyx_r = __pyx_t_3;

#line 138 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_3 = 0;

#line 138 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/scoring.pyx":137
 *         elif code == 3:
 *             return not (self._eq(other))
 *         elif code == 4:             # <<<<<<<<<<<<<<
 *             return self._gt(other)
 * 
 */

#line 137 "ms_deisotope/_c/scoring.pyx"
    break;

#line 137 "ms_deisotope/_c/scoring.pyx"
    default: break;

#line 137 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":124
 *         return self.score > other.score
 * 
 *     def __richcmp__(self, IsotopicFitRecord other, int code):             # <<<<<<<<<<<<<<
 *         if other is None:
 *             if code == 3:
 */

#line 124 "ms_deisotope/_c/scoring.pyx"


#line 124 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 124 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 124 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 124 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 124 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 124 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitRecord.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 124 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 124 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 124 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 124 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 124 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 124 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":141
 * 
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         if self._hash == -1:
 *             self._hash = hash((self.monoisotopic_peak.mz, self.charge))
 */

#line 141 "ms_deisotope/_c/scoring.pyx"


#line 141 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 141 "ms_deisotope/_c/scoring.pyx"
static Py_hash_t __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_17__hash__(PyObject *__pyx_v_self); /*proto*/

#line 141 "ms_deisotope/_c/scoring.pyx"
static Py_hash_t __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_17__hash__(PyObject *__pyx_v_self) {

#line 141 "ms_deisotope/_c/scoring.pyx"
  Py_hash_t __pyx_r;

#line 141 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 141 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_r = 
#line 141 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_16__hash__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self));

#line 141 "ms_deisotope/_c/scoring.pyx"


#line 141 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 141 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 141 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 141 "ms_deisotope/_c/scoring.pyx"
}

#line 141 "ms_deisotope/_c/scoring.pyx"


#line 141 "ms_deisotope/_c/scoring.pyx"
static Py_hash_t __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_16__hash__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self) {

#line 141 "ms_deisotope/_c/scoring.pyx"
  Py_hash_t __pyx_r;

#line 141 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 141 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_1;

#line 141 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 141 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 141 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 141 "ms_deisotope/_c/scoring.pyx"
  Py_hash_t __pyx_t_5;

#line 141 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 141 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 141 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 141 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__hash__", 0);

  /* "ms_deisotope/_c/scoring.pyx":142
 * 
 *     def __hash__(self):
 *         if self._hash == -1:             # <<<<<<<<<<<<<<
 *             self._hash = hash((self.monoisotopic_peak.mz, self.charge))
 *         return self._hash
 */

#line 142 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = ((__pyx_v_self->_hash == -1L) != 0);

#line 142 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/scoring.pyx":143
 *     def __hash__(self):
 *         if self._hash == -1:
 *             self._hash = hash((self.monoisotopic_peak.mz, self.charge))             # <<<<<<<<<<<<<<
 *         return self._hash
 * 
 */

#line 143 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->monoisotopic_peak->__pyx_base.mz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 143, __pyx_L1_error)

#line 143 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 143 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->charge); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 143, __pyx_L1_error)

#line 143 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 143 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 143, __pyx_L1_error)

#line 143 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 143 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GIVEREF(__pyx_t_2);

#line 143 "ms_deisotope/_c/scoring.pyx"
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);

#line 143 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GIVEREF(__pyx_t_3);

#line 143 "ms_deisotope/_c/scoring.pyx"
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);

#line 143 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_2 = 0;

#line 143 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_3 = 0;

#line 143 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = PyObject_Hash(__pyx_t_4); if (unlikely(__pyx_t_5 == ((Py_hash_t)-1))) __PYX_ERR(0, 143, __pyx_L1_error)

#line 143 "ms_deisotope/_c/scoring.pyx"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 143 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_self->_hash = __pyx_t_5;

    /* "ms_deisotope/_c/scoring.pyx":142
 * 
 *     def __hash__(self):
 *         if self._hash == -1:             # <<<<<<<<<<<<<<
 *             self._hash = hash((self.monoisotopic_peak.mz, self.charge))
 *         return self._hash
 */

#line 142 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":144
 *         if self._hash == -1:
 *             self._hash = hash((self.monoisotopic_peak.mz, self.charge))
 *         return self._hash             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */

#line 144 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_v_self->_hash;

#line 144 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":141
 * 
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         if self._hash == -1:
 *             self._hash = hash((self.monoisotopic_peak.mz, self.charge))
 */

#line 141 "ms_deisotope/_c/scoring.pyx"


#line 141 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 141 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 141 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 141 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 141 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 141 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitRecord.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 141 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = -1;

#line 141 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 141 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;

#line 141 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 141 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 141 "ms_deisotope/_c/scoring.pyx"
}

#line 141 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_gb_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_20generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "ms_deisotope/_c/scoring.pyx":146
 *         return self._hash
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         yield self.score
 *         yield self.charge
 */

#line 146 "ms_deisotope/_c/scoring.pyx"


#line 146 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 146 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_19__iter__(PyObject *__pyx_v_self); /*proto*/

#line 146 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_19__iter__(PyObject *__pyx_v_self) {

#line 146 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 146 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 146 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = 
#line 146 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_18__iter__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self));

#line 146 "ms_deisotope/_c/scoring.pyx"


#line 146 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 146 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 146 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 146 "ms_deisotope/_c/scoring.pyx"
}

#line 146 "ms_deisotope/_c/scoring.pyx"


#line 146 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_18__iter__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self) {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__ *__pyx_cur_scope
#line 146 "ms_deisotope/_c/scoring.pyx"
;

#line 146 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 146 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 146 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 146 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 146 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 146 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__iter__", 0);

#line 146 "ms_deisotope/_c/scoring.pyx"
  __pyx_cur_scope = (struct __pyx_obj_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__ *)__pyx_tp_new_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__(__pyx_ptype_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__, __pyx_empty_tuple, NULL);

#line 146 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__pyx_cur_scope)) {

#line 146 "ms_deisotope/_c/scoring.pyx"
    __pyx_cur_scope = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__ *)Py_None);

#line 146 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(Py_None);

#line 146 "ms_deisotope/_c/scoring.pyx"
    __PYX_ERR(0, 146, __pyx_L1_error)

#line 146 "ms_deisotope/_c/scoring.pyx"
  } else {

#line 146 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_cur_scope);

#line 146 "ms_deisotope/_c/scoring.pyx"
  }

#line 146 "ms_deisotope/_c/scoring.pyx"
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;

#line 146 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);

#line 146 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);

#line 146 "ms_deisotope/_c/scoring.pyx"
  {

#line 146 "ms_deisotope/_c/scoring.pyx"
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_20generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_iter, __pyx_n_s_IsotopicFitRecord___iter, __pyx_n_s_ms_deisotope__c_scoring); if (unlikely(!gen)) __PYX_ERR(0, 146, __pyx_L1_error)

#line 146 "ms_deisotope/_c/scoring.pyx"
    __Pyx_DECREF(__pyx_cur_scope);

#line 146 "ms_deisotope/_c/scoring.pyx"
    __Pyx_RefNannyFinishContext();

#line 146 "ms_deisotope/_c/scoring.pyx"
    return (PyObject *) gen;

#line 146 "ms_deisotope/_c/scoring.pyx"
  }

#line 146 "ms_deisotope/_c/scoring.pyx"


#line 146 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 146 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 146 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitRecord.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 146 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 146 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));

#line 146 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 146 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 146 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 146 "ms_deisotope/_c/scoring.pyx"
}

#line 146 "ms_deisotope/_c/scoring.pyx"


#line 146 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_gb_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_20generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{

#line 146 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__ *__pyx_cur_scope = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__ *)__pyx_generator->closure);

#line 146 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 146 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 146 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 146 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 146 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 146 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 146 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__iter__", 0);

#line 146 "ms_deisotope/_c/scoring.pyx"
  switch (__pyx_generator->resume_label) {

#line 146 "ms_deisotope/_c/scoring.pyx"
    case 0: goto __pyx_L3_first_run;

#line 146 "ms_deisotope/_c/scoring.pyx"
    case 1: goto __pyx_L4_resume_from_yield;

#line 146 "ms_deisotope/_c/scoring.pyx"
    case 2: goto __pyx_L5_resume_from_yield;

#line 146 "ms_deisotope/_c/scoring.pyx"
    case 3: goto __pyx_L6_resume_from_yield;

#line 146 "ms_deisotope/_c/scoring.pyx"
    case 4: goto __pyx_L7_resume_from_yield;

#line 146 "ms_deisotope/_c/scoring.pyx"
    default: /* CPython raises the right error here */

#line 146 "ms_deisotope/_c/scoring.pyx"
    __Pyx_RefNannyFinishContext();

#line 146 "ms_deisotope/_c/scoring.pyx"
    return NULL;

#line 146 "ms_deisotope/_c/scoring.pyx"
  }

#line 146 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_first_run:;

#line 146 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 146, __pyx_L1_error)

  /* "ms_deisotope/_c/scoring.pyx":147
 * 
 *     def __iter__(self):
 *         yield self.score             # <<<<<<<<<<<<<<
 *         yield self.charge
 *         yield self.experimental
 */

#line 147 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_cur_scope->__pyx_v_self->score); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 147, __pyx_L1_error)

#line 147 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 147 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 147 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 147 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 147 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 147 "ms_deisotope/_c/scoring.pyx"
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);

#line 147 "ms_deisotope/_c/scoring.pyx"
  /* return from generator, yielding value */

#line 147 "ms_deisotope/_c/scoring.pyx"
  __pyx_generator->resume_label = 1;

#line 147 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 147 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_resume_from_yield:;

#line 147 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 147, __pyx_L1_error)

  /* "ms_deisotope/_c/scoring.pyx":148
 *     def __iter__(self):
 *         yield self.score
 *         yield self.charge             # <<<<<<<<<<<<<<
 *         yield self.experimental
 *         yield self.theoretical
 */

#line 148 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_cur_scope->__pyx_v_self->charge); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 148, __pyx_L1_error)

#line 148 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 148 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 148 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 148 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 148 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 148 "ms_deisotope/_c/scoring.pyx"
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);

#line 148 "ms_deisotope/_c/scoring.pyx"
  /* return from generator, yielding value */

#line 148 "ms_deisotope/_c/scoring.pyx"
  __pyx_generator->resume_label = 2;

#line 148 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 148 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_resume_from_yield:;

#line 148 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 148, __pyx_L1_error)

  /* "ms_deisotope/_c/scoring.pyx":149
 *         yield self.score
 *         yield self.charge
 *         yield self.experimental             # <<<<<<<<<<<<<<
 *         yield self.theoretical
 * 
 */

#line 149 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self->experimental);

#line 149 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_cur_scope->__pyx_v_self->experimental;

#line 149 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 149 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 149 "ms_deisotope/_c/scoring.pyx"
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);

#line 149 "ms_deisotope/_c/scoring.pyx"
  /* return from generator, yielding value */

#line 149 "ms_deisotope/_c/scoring.pyx"
  __pyx_generator->resume_label = 3;

#line 149 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 149 "ms_deisotope/_c/scoring.pyx"
  __pyx_L6_resume_from_yield:;

#line 149 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 149, __pyx_L1_error)

  /* "ms_deisotope/_c/scoring.pyx":150
 *         yield self.charge
 *         yield self.experimental
 *         yield self.theoretical             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 150 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_v_self->theoretical));

#line 150 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = ((PyObject *)__pyx_cur_scope->__pyx_v_self->theoretical);

#line 150 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 150 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 150 "ms_deisotope/_c/scoring.pyx"
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);

#line 150 "ms_deisotope/_c/scoring.pyx"
  /* return from generator, yielding value */

#line 150 "ms_deisotope/_c/scoring.pyx"
  __pyx_generator->resume_label = 4;

#line 150 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 150 "ms_deisotope/_c/scoring.pyx"
  __pyx_L7_resume_from_yield:;

#line 150 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 150, __pyx_L1_error)

#line 150 "ms_deisotope/_c/scoring.pyx"
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "ms_deisotope/_c/scoring.pyx":146
 *         return self._hash
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         yield self.score
 *         yield self.charge
 */

#line 146 "ms_deisotope/_c/scoring.pyx"


#line 146 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 146 "ms_deisotope/_c/scoring.pyx"
  PyErr_SetNone(PyExc_StopIteration);

#line 146 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 146 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 146 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 146 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 146 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 146 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;

#line 146 "ms_deisotope/_c/scoring.pyx"
  #if !CYTHON_USE_EXC_INFO_STACK

#line 146 "ms_deisotope/_c/scoring.pyx"
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);

#line 146 "ms_deisotope/_c/scoring.pyx"
  #endif

#line 146 "ms_deisotope/_c/scoring.pyx"
  __pyx_generator->resume_label = -1;

#line 146 "ms_deisotope/_c/scoring.pyx"
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);

#line 146 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 146 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 146 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":153
 * 
 *     @property
 *     def npeaks(self):             # <<<<<<<<<<<<<<
 *         return len(self.experimental)
 * 
 */

#line 153 "ms_deisotope/_c/scoring.pyx"


#line 153 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 153 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_6npeaks_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 153 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_6npeaks_1__get__(PyObject *__pyx_v_self) {

#line 153 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 153 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 153 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 153 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_6npeaks___get__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self));

#line 153 "ms_deisotope/_c/scoring.pyx"


#line 153 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 153 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 153 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 153 "ms_deisotope/_c/scoring.pyx"
}

#line 153 "ms_deisotope/_c/scoring.pyx"


#line 153 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_6npeaks___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self) {

#line 153 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 153 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 153 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 153 "ms_deisotope/_c/scoring.pyx"
  Py_ssize_t __pyx_t_2;

#line 153 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 153 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 153 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 153 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ms_deisotope/_c/scoring.pyx":154
 *     @property
 *     def npeaks(self):
 *         return len(self.experimental)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */

#line 154 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 154 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __pyx_v_self->experimental;

#line 154 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 154 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_t_1 == Py_None)) {

#line 154 "ms_deisotope/_c/scoring.pyx"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 154 "ms_deisotope/_c/scoring.pyx"
    __PYX_ERR(0, 154, __pyx_L1_error)

#line 154 "ms_deisotope/_c/scoring.pyx"
  }

#line 154 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = PyList_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 154, __pyx_L1_error)

#line 154 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 154 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 154, __pyx_L1_error)

#line 154 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 154 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 154 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 154 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":153
 * 
 *     @property
 *     def npeaks(self):             # <<<<<<<<<<<<<<
 *         return len(self.experimental)
 * 
 */

#line 153 "ms_deisotope/_c/scoring.pyx"


#line 153 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 153 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 153 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 153 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitRecord.npeaks.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 153 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 153 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 153 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 153 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 153 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 153 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":156
 *         return len(self.experimental)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "IsotopicFitRecord(score=%0.5f, charge=%d, npeaks=%d, monoisotopic_mz=%0.5f)" % (
 *             self.score, self.charge, self.npeaks, self.monoisotopic_peak.mz)
 */

#line 156 "ms_deisotope/_c/scoring.pyx"


#line 156 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 156 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_22__repr__(PyObject *__pyx_v_self); /*proto*/

#line 156 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_22__repr__(PyObject *__pyx_v_self) {

#line 156 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 156 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 156 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = 
#line 156 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_21__repr__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self));

#line 156 "ms_deisotope/_c/scoring.pyx"


#line 156 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 156 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 156 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 156 "ms_deisotope/_c/scoring.pyx"
}

#line 156 "ms_deisotope/_c/scoring.pyx"


#line 156 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_21__repr__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self) {

#line 156 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 156 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 156 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 156 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 156 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 156 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 156 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 156 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 156 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 156 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 156 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "ms_deisotope/_c/scoring.pyx":157
 * 
 *     def __repr__(self):
 *         return "IsotopicFitRecord(score=%0.5f, charge=%d, npeaks=%d, monoisotopic_mz=%0.5f)" % (             # <<<<<<<<<<<<<<
 *             self.score, self.charge, self.npeaks, self.monoisotopic_peak.mz)
 * 
 */

#line 157 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

  /* "ms_deisotope/_c/scoring.pyx":158
 *     def __repr__(self):
 *         return "IsotopicFitRecord(score=%0.5f, charge=%d, npeaks=%d, monoisotopic_mz=%0.5f)" % (
 *             self.score, self.charge, self.npeaks, self.monoisotopic_peak.mz)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 158 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->score); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 158, __pyx_L1_error)

#line 158 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 158 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->charge); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 158, __pyx_L1_error)

#line 158 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 158 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_npeaks); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 158, __pyx_L1_error)

#line 158 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 158 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->monoisotopic_peak->__pyx_base.mz); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 158, __pyx_L1_error)

#line 158 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 158 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 158, __pyx_L1_error)

#line 158 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 158 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 158 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);

#line 158 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 158 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);

#line 158 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 158 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);

#line 158 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_4);

#line 158 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);

#line 158 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 158 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = 0;

#line 158 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = 0;

#line 158 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_4 = 0;

  /* "ms_deisotope/_c/scoring.pyx":157
 * 
 *     def __repr__(self):
 *         return "IsotopicFitRecord(score=%0.5f, charge=%d, npeaks=%d, monoisotopic_mz=%0.5f)" % (             # <<<<<<<<<<<<<<
 *             self.score, self.charge, self.npeaks, self.monoisotopic_peak.mz)
 * 
 */

#line 157 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_IsotopicFitRecord_score_0_5f_cha, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 157, __pyx_L1_error)

#line 157 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 157 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 157 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_4;

#line 157 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_4 = 0;

#line 157 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":156
 *         return len(self.experimental)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "IsotopicFitRecord(score=%0.5f, charge=%d, npeaks=%d, monoisotopic_mz=%0.5f)" % (
 *             self.score, self.charge, self.npeaks, self.monoisotopic_peak.mz)
 */

#line 156 "ms_deisotope/_c/scoring.pyx"


#line 156 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 156 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 156 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 156 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 156 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 156 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 156 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 156 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitRecord.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 156 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 156 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 156 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 156 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 156 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 156 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pxd":18
 * cdef class IsotopicFitRecord(object):
 *     cdef:
 *         public FittedPeak seed_peak             # <<<<<<<<<<<<<<
 *         public double score
 *         public int charge
 */

#line 18 "ms_deisotope/_c/scoring.pxd"


#line 18 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 18 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_9seed_peak_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 18 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_9seed_peak_1__get__(PyObject *__pyx_v_self) {

#line 18 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = 0;

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 18 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_9seed_peak___get__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self));

#line 18 "ms_deisotope/_c/scoring.pxd"


#line 18 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 18 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 18 "ms_deisotope/_c/scoring.pxd"
}

#line 18 "ms_deisotope/_c/scoring.pxd"


#line 18 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_9seed_peak___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self) {

#line 18 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = NULL;

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->seed_peak));

#line 18 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = ((PyObject *)__pyx_v_self->seed_peak);

#line 18 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 18 "ms_deisotope/_c/scoring.pxd"


#line 18 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 18 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 18 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 18 "ms_deisotope/_c/scoring.pxd"
}

#line 18 "ms_deisotope/_c/scoring.pxd"


#line 18 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 18 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_9seed_peak_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 18 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_9seed_peak_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 18 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 18 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_9seed_peak_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 18 "ms_deisotope/_c/scoring.pxd"


#line 18 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 18 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 18 "ms_deisotope/_c/scoring.pxd"
}

#line 18 "ms_deisotope/_c/scoring.pxd"


#line 18 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_9seed_peak_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, PyObject *__pyx_v_value) {

#line 18 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 18 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 18 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 18 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 18 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 18 "ms_deisotope/_c/scoring.pxd"
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak))))) __PYX_ERR(1, 18, __pyx_L1_error)

#line 18 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = __pyx_v_value;

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(__pyx_t_1);

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GIVEREF(__pyx_t_1);

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_v_self->seed_peak);

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->seed_peak));

#line 18 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->seed_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1);

#line 18 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = 0;

#line 18 "ms_deisotope/_c/scoring.pxd"


#line 18 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 18 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 18 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 18 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitRecord.seed_peak.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 18 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = -1;

#line 18 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 18 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 18 "ms_deisotope/_c/scoring.pxd"
}

#line 18 "ms_deisotope/_c/scoring.pxd"


#line 18 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 18 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_9seed_peak_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 18 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_9seed_peak_5__del__(PyObject *__pyx_v_self) {

#line 18 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 18 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_9seed_peak_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self));

#line 18 "ms_deisotope/_c/scoring.pxd"


#line 18 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 18 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 18 "ms_deisotope/_c/scoring.pxd"
}

#line 18 "ms_deisotope/_c/scoring.pxd"


#line 18 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_9seed_peak_4__del__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self) {

#line 18 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(Py_None);

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GIVEREF(Py_None);

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_v_self->seed_peak);

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->seed_peak));

#line 18 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->seed_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)Py_None);

#line 18 "ms_deisotope/_c/scoring.pxd"


#line 18 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 18 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 18 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 18 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 18 "ms_deisotope/_c/scoring.pxd"
}

/* "ms_deisotope/_c/scoring.pxd":19
 *     cdef:
 *         public FittedPeak seed_peak
 *         public double score             # <<<<<<<<<<<<<<
 *         public int charge
 *         public list experimental
 */

#line 19 "ms_deisotope/_c/scoring.pxd"


#line 19 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 19 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_5score_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 19 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_5score_1__get__(PyObject *__pyx_v_self) {

#line 19 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = 0;

#line 19 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 19 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 19 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_5score___get__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self));

#line 19 "ms_deisotope/_c/scoring.pxd"


#line 19 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 19 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 19 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 19 "ms_deisotope/_c/scoring.pxd"
}

#line 19 "ms_deisotope/_c/scoring.pxd"


#line 19 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_5score___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self) {

#line 19 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = NULL;

#line 19 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 19 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 19 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 19 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 19 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 19 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 19 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 19 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->score); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 19, __pyx_L1_error)

#line 19 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 19 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = __pyx_t_1;

#line 19 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = 0;

#line 19 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 19 "ms_deisotope/_c/scoring.pxd"


#line 19 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 19 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 19 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 19 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitRecord.score.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 19 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = NULL;

#line 19 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 19 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 19 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 19 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 19 "ms_deisotope/_c/scoring.pxd"
}

#line 19 "ms_deisotope/_c/scoring.pxd"


#line 19 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 19 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_5score_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 19 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_5score_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 19 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 19 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 19 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 19 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_5score_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 19 "ms_deisotope/_c/scoring.pxd"


#line 19 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 19 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 19 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 19 "ms_deisotope/_c/scoring.pxd"
}

#line 19 "ms_deisotope/_c/scoring.pxd"


#line 19 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_5score_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, PyObject *__pyx_v_value) {

#line 19 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 19 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 19 "ms_deisotope/_c/scoring.pxd"
  double __pyx_t_1;

#line 19 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 19 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 19 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 19 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 19 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 19, __pyx_L1_error)

#line 19 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->score = __pyx_t_1;

#line 19 "ms_deisotope/_c/scoring.pxd"


#line 19 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 19 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 19 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 19 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 19 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitRecord.score.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 19 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = -1;

#line 19 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 19 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 19 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 19 "ms_deisotope/_c/scoring.pxd"
}

/* "ms_deisotope/_c/scoring.pxd":20
 *         public FittedPeak seed_peak
 *         public double score
 *         public int charge             # <<<<<<<<<<<<<<
 *         public list experimental
 *         public TheoreticalIsotopicPattern theoretical
 */

#line 20 "ms_deisotope/_c/scoring.pxd"


#line 20 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 20 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_6charge_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 20 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_6charge_1__get__(PyObject *__pyx_v_self) {

#line 20 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = 0;

#line 20 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 20 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 20 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_6charge___get__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self));

#line 20 "ms_deisotope/_c/scoring.pxd"


#line 20 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 20 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 20 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 20 "ms_deisotope/_c/scoring.pxd"
}

#line 20 "ms_deisotope/_c/scoring.pxd"


#line 20 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_6charge___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self) {

#line 20 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = NULL;

#line 20 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 20 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 20 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 20 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 20 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 20 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 20 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 20 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->charge); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 20, __pyx_L1_error)

#line 20 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 20 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = __pyx_t_1;

#line 20 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = 0;

#line 20 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 20 "ms_deisotope/_c/scoring.pxd"


#line 20 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 20 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 20 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 20 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitRecord.charge.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 20 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = NULL;

#line 20 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 20 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 20 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 20 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 20 "ms_deisotope/_c/scoring.pxd"
}

#line 20 "ms_deisotope/_c/scoring.pxd"


#line 20 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 20 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_6charge_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 20 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_6charge_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 20 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 20 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 20 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 20 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_6charge_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 20 "ms_deisotope/_c/scoring.pxd"


#line 20 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 20 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 20 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 20 "ms_deisotope/_c/scoring.pxd"
}

#line 20 "ms_deisotope/_c/scoring.pxd"


#line 20 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_6charge_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, PyObject *__pyx_v_value) {

#line 20 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 20 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 20 "ms_deisotope/_c/scoring.pxd"
  int __pyx_t_1;

#line 20 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 20 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 20 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 20 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 20 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 20, __pyx_L1_error)

#line 20 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->charge = __pyx_t_1;

#line 20 "ms_deisotope/_c/scoring.pxd"


#line 20 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 20 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 20 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 20 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 20 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitRecord.charge.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 20 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = -1;

#line 20 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 20 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 20 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 20 "ms_deisotope/_c/scoring.pxd"
}

/* "ms_deisotope/_c/scoring.pxd":21
 *         public double score
 *         public int charge
 *         public list experimental             # <<<<<<<<<<<<<<
 *         public TheoreticalIsotopicPattern theoretical
 *         public FittedPeak monoisotopic_peak
 */

#line 21 "ms_deisotope/_c/scoring.pxd"


#line 21 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 21 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12experimental_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 21 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12experimental_1__get__(PyObject *__pyx_v_self) {

#line 21 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = 0;

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 21 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12experimental___get__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self));

#line 21 "ms_deisotope/_c/scoring.pxd"


#line 21 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 21 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 21 "ms_deisotope/_c/scoring.pxd"
}

#line 21 "ms_deisotope/_c/scoring.pxd"


#line 21 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12experimental___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self) {

#line 21 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = NULL;

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(__pyx_v_self->experimental);

#line 21 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = __pyx_v_self->experimental;

#line 21 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 21 "ms_deisotope/_c/scoring.pxd"


#line 21 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 21 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 21 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 21 "ms_deisotope/_c/scoring.pxd"
}

#line 21 "ms_deisotope/_c/scoring.pxd"


#line 21 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 21 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12experimental_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 21 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12experimental_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 21 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 21 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12experimental_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 21 "ms_deisotope/_c/scoring.pxd"


#line 21 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 21 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 21 "ms_deisotope/_c/scoring.pxd"
}

#line 21 "ms_deisotope/_c/scoring.pxd"


#line 21 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12experimental_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, PyObject *__pyx_v_value) {

#line 21 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 21 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 21 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 21 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 21 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 21 "ms_deisotope/_c/scoring.pxd"
  if (!(likely(PyList_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 21, __pyx_L1_error)

#line 21 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = __pyx_v_value;

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(__pyx_t_1);

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GIVEREF(__pyx_t_1);

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_v_self->experimental);

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_DECREF(__pyx_v_self->experimental);

#line 21 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->experimental = ((PyObject*)__pyx_t_1);

#line 21 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = 0;

#line 21 "ms_deisotope/_c/scoring.pxd"


#line 21 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 21 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 21 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 21 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitRecord.experimental.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 21 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = -1;

#line 21 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 21 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 21 "ms_deisotope/_c/scoring.pxd"
}

#line 21 "ms_deisotope/_c/scoring.pxd"


#line 21 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 21 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12experimental_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 21 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12experimental_5__del__(PyObject *__pyx_v_self) {

#line 21 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 21 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12experimental_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self));

#line 21 "ms_deisotope/_c/scoring.pxd"


#line 21 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 21 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 21 "ms_deisotope/_c/scoring.pxd"
}

#line 21 "ms_deisotope/_c/scoring.pxd"


#line 21 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12experimental_4__del__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self) {

#line 21 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(Py_None);

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GIVEREF(Py_None);

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_v_self->experimental);

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_DECREF(__pyx_v_self->experimental);

#line 21 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->experimental = ((PyObject*)Py_None);

#line 21 "ms_deisotope/_c/scoring.pxd"


#line 21 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 21 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 21 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 21 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 21 "ms_deisotope/_c/scoring.pxd"
}

/* "ms_deisotope/_c/scoring.pxd":22
 *         public int charge
 *         public list experimental
 *         public TheoreticalIsotopicPattern theoretical             # <<<<<<<<<<<<<<
 *         public FittedPeak monoisotopic_peak
 *         public int missed_peaks
 */

#line 22 "ms_deisotope/_c/scoring.pxd"


#line 22 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 22 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_11theoretical_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 22 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_11theoretical_1__get__(PyObject *__pyx_v_self) {

#line 22 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = 0;

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 22 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_11theoretical___get__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self));

#line 22 "ms_deisotope/_c/scoring.pxd"


#line 22 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 22 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 22 "ms_deisotope/_c/scoring.pxd"
}

#line 22 "ms_deisotope/_c/scoring.pxd"


#line 22 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_11theoretical___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self) {

#line 22 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = NULL;

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->theoretical));

#line 22 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = ((PyObject *)__pyx_v_self->theoretical);

#line 22 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 22 "ms_deisotope/_c/scoring.pxd"


#line 22 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 22 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 22 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 22 "ms_deisotope/_c/scoring.pxd"
}

#line 22 "ms_deisotope/_c/scoring.pxd"


#line 22 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 22 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_11theoretical_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 22 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_11theoretical_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 22 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 22 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_11theoretical_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 22 "ms_deisotope/_c/scoring.pxd"


#line 22 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 22 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 22 "ms_deisotope/_c/scoring.pxd"
}

#line 22 "ms_deisotope/_c/scoring.pxd"


#line 22 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_11theoretical_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, PyObject *__pyx_v_value) {

#line 22 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 22 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 22 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 22 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 22 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 22 "ms_deisotope/_c/scoring.pxd"
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern))))) __PYX_ERR(1, 22, __pyx_L1_error)

#line 22 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = __pyx_v_value;

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(__pyx_t_1);

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GIVEREF(__pyx_t_1);

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_v_self->theoretical);

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->theoretical));

#line 22 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->theoretical = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)__pyx_t_1);

#line 22 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = 0;

#line 22 "ms_deisotope/_c/scoring.pxd"


#line 22 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 22 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 22 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 22 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitRecord.theoretical.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 22 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = -1;

#line 22 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 22 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 22 "ms_deisotope/_c/scoring.pxd"
}

#line 22 "ms_deisotope/_c/scoring.pxd"


#line 22 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 22 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_11theoretical_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 22 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_11theoretical_5__del__(PyObject *__pyx_v_self) {

#line 22 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 22 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_11theoretical_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self));

#line 22 "ms_deisotope/_c/scoring.pxd"


#line 22 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 22 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 22 "ms_deisotope/_c/scoring.pxd"
}

#line 22 "ms_deisotope/_c/scoring.pxd"


#line 22 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_11theoretical_4__del__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self) {

#line 22 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(Py_None);

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GIVEREF(Py_None);

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_v_self->theoretical);

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->theoretical));

#line 22 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->theoretical = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)Py_None);

#line 22 "ms_deisotope/_c/scoring.pxd"


#line 22 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 22 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 22 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 22 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 22 "ms_deisotope/_c/scoring.pxd"
}

/* "ms_deisotope/_c/scoring.pxd":23
 *         public list experimental
 *         public TheoreticalIsotopicPattern theoretical
 *         public FittedPeak monoisotopic_peak             # <<<<<<<<<<<<<<
 *         public int missed_peaks
 *         public object data
 */

#line 23 "ms_deisotope/_c/scoring.pxd"


#line 23 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 23 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_17monoisotopic_peak_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 23 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_17monoisotopic_peak_1__get__(PyObject *__pyx_v_self) {

#line 23 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = 0;

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 23 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_17monoisotopic_peak___get__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self));

#line 23 "ms_deisotope/_c/scoring.pxd"


#line 23 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 23 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 23 "ms_deisotope/_c/scoring.pxd"
}

#line 23 "ms_deisotope/_c/scoring.pxd"


#line 23 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_17monoisotopic_peak___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self) {

#line 23 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = NULL;

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->monoisotopic_peak));

#line 23 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = ((PyObject *)__pyx_v_self->monoisotopic_peak);

#line 23 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 23 "ms_deisotope/_c/scoring.pxd"


#line 23 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 23 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 23 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 23 "ms_deisotope/_c/scoring.pxd"
}

#line 23 "ms_deisotope/_c/scoring.pxd"


#line 23 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 23 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_17monoisotopic_peak_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 23 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_17monoisotopic_peak_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 23 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 23 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_17monoisotopic_peak_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 23 "ms_deisotope/_c/scoring.pxd"


#line 23 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 23 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 23 "ms_deisotope/_c/scoring.pxd"
}

#line 23 "ms_deisotope/_c/scoring.pxd"


#line 23 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_17monoisotopic_peak_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, PyObject *__pyx_v_value) {

#line 23 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 23 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 23 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 23 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 23 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 23 "ms_deisotope/_c/scoring.pxd"
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak))))) __PYX_ERR(1, 23, __pyx_L1_error)

#line 23 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = __pyx_v_value;

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(__pyx_t_1);

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GIVEREF(__pyx_t_1);

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_v_self->monoisotopic_peak);

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->monoisotopic_peak));

#line 23 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->monoisotopic_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1);

#line 23 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = 0;

#line 23 "ms_deisotope/_c/scoring.pxd"


#line 23 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 23 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 23 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 23 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitRecord.monoisotopic_peak.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 23 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = -1;

#line 23 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 23 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 23 "ms_deisotope/_c/scoring.pxd"
}

#line 23 "ms_deisotope/_c/scoring.pxd"


#line 23 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 23 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_17monoisotopic_peak_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 23 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_17monoisotopic_peak_5__del__(PyObject *__pyx_v_self) {

#line 23 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 23 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_17monoisotopic_peak_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self));

#line 23 "ms_deisotope/_c/scoring.pxd"


#line 23 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 23 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 23 "ms_deisotope/_c/scoring.pxd"
}

#line 23 "ms_deisotope/_c/scoring.pxd"


#line 23 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_17monoisotopic_peak_4__del__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self) {

#line 23 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(Py_None);

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GIVEREF(Py_None);

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_v_self->monoisotopic_peak);

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->monoisotopic_peak));

#line 23 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->monoisotopic_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)Py_None);

#line 23 "ms_deisotope/_c/scoring.pxd"


#line 23 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 23 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 23 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 23 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 23 "ms_deisotope/_c/scoring.pxd"
}

/* "ms_deisotope/_c/scoring.pxd":24
 *         public TheoreticalIsotopicPattern theoretical
 *         public FittedPeak monoisotopic_peak
 *         public int missed_peaks             # <<<<<<<<<<<<<<
 *         public object data
 *         public Py_hash_t _hash
 */

#line 24 "ms_deisotope/_c/scoring.pxd"


#line 24 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 24 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12missed_peaks_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 24 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12missed_peaks_1__get__(PyObject *__pyx_v_self) {

#line 24 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = 0;

#line 24 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 24 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 24 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12missed_peaks___get__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self));

#line 24 "ms_deisotope/_c/scoring.pxd"


#line 24 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 24 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 24 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 24 "ms_deisotope/_c/scoring.pxd"
}

#line 24 "ms_deisotope/_c/scoring.pxd"


#line 24 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12missed_peaks___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self) {

#line 24 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = NULL;

#line 24 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 24 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 24 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 24 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 24 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 24 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 24 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 24 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->missed_peaks); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 24, __pyx_L1_error)

#line 24 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 24 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = __pyx_t_1;

#line 24 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = 0;

#line 24 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 24 "ms_deisotope/_c/scoring.pxd"


#line 24 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 24 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 24 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 24 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitRecord.missed_peaks.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 24 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = NULL;

#line 24 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 24 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 24 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 24 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 24 "ms_deisotope/_c/scoring.pxd"
}

#line 24 "ms_deisotope/_c/scoring.pxd"


#line 24 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 24 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12missed_peaks_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 24 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12missed_peaks_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 24 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 24 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 24 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 24 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12missed_peaks_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 24 "ms_deisotope/_c/scoring.pxd"


#line 24 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 24 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 24 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 24 "ms_deisotope/_c/scoring.pxd"
}

#line 24 "ms_deisotope/_c/scoring.pxd"


#line 24 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12missed_peaks_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, PyObject *__pyx_v_value) {

#line 24 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 24 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 24 "ms_deisotope/_c/scoring.pxd"
  int __pyx_t_1;

#line 24 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 24 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 24 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 24 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 24 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 24, __pyx_L1_error)

#line 24 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->missed_peaks = __pyx_t_1;

#line 24 "ms_deisotope/_c/scoring.pxd"


#line 24 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 24 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 24 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 24 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 24 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitRecord.missed_peaks.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 24 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = -1;

#line 24 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 24 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 24 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 24 "ms_deisotope/_c/scoring.pxd"
}

/* "ms_deisotope/_c/scoring.pxd":25
 *         public FittedPeak monoisotopic_peak
 *         public int missed_peaks
 *         public object data             # <<<<<<<<<<<<<<
 *         public Py_hash_t _hash
 * 
 */

#line 25 "ms_deisotope/_c/scoring.pxd"


#line 25 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 25 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_4data_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 25 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_4data_1__get__(PyObject *__pyx_v_self) {

#line 25 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = 0;

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 25 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_4data___get__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self));

#line 25 "ms_deisotope/_c/scoring.pxd"


#line 25 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 25 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 25 "ms_deisotope/_c/scoring.pxd"
}

#line 25 "ms_deisotope/_c/scoring.pxd"


#line 25 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_4data___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self) {

#line 25 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = NULL;

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(__pyx_v_self->data);

#line 25 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = __pyx_v_self->data;

#line 25 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 25 "ms_deisotope/_c/scoring.pxd"


#line 25 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 25 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 25 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 25 "ms_deisotope/_c/scoring.pxd"
}

#line 25 "ms_deisotope/_c/scoring.pxd"


#line 25 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 25 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_4data_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 25 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_4data_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 25 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 25 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_4data_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 25 "ms_deisotope/_c/scoring.pxd"


#line 25 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 25 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 25 "ms_deisotope/_c/scoring.pxd"
}

#line 25 "ms_deisotope/_c/scoring.pxd"


#line 25 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_4data_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, PyObject *__pyx_v_value) {

#line 25 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(__pyx_v_value);

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GIVEREF(__pyx_v_value);

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_v_self->data);

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_DECREF(__pyx_v_self->data);

#line 25 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->data = __pyx_v_value;

#line 25 "ms_deisotope/_c/scoring.pxd"


#line 25 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 25 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 25 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 25 "ms_deisotope/_c/scoring.pxd"
}

#line 25 "ms_deisotope/_c/scoring.pxd"


#line 25 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 25 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_4data_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 25 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_4data_5__del__(PyObject *__pyx_v_self) {

#line 25 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 25 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_4data_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self));

#line 25 "ms_deisotope/_c/scoring.pxd"


#line 25 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 25 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 25 "ms_deisotope/_c/scoring.pxd"
}

#line 25 "ms_deisotope/_c/scoring.pxd"


#line 25 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_4data_4__del__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self) {

#line 25 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(Py_None);

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GIVEREF(Py_None);

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_v_self->data);

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_DECREF(__pyx_v_self->data);

#line 25 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->data = Py_None;

#line 25 "ms_deisotope/_c/scoring.pxd"


#line 25 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 25 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 25 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 25 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 25 "ms_deisotope/_c/scoring.pxd"
}

/* "ms_deisotope/_c/scoring.pxd":26
 *         public int missed_peaks
 *         public object data
 *         public Py_hash_t _hash             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

#line 26 "ms_deisotope/_c/scoring.pxd"


#line 26 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 26 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_5_hash_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 26 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_5_hash_1__get__(PyObject *__pyx_v_self) {

#line 26 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = 0;

#line 26 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 26 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 26 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_5_hash___get__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self));

#line 26 "ms_deisotope/_c/scoring.pxd"


#line 26 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 26 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 26 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 26 "ms_deisotope/_c/scoring.pxd"
}

#line 26 "ms_deisotope/_c/scoring.pxd"


#line 26 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_5_hash___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self) {

#line 26 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = NULL;

#line 26 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 26 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 26 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 26 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 26 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 26 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 26 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 26 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = __Pyx_PyInt_FromHash_t(__pyx_v_self->_hash); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 26, __pyx_L1_error)

#line 26 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 26 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = __pyx_t_1;

#line 26 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = 0;

#line 26 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 26 "ms_deisotope/_c/scoring.pxd"


#line 26 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 26 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 26 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 26 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitRecord._hash.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 26 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = NULL;

#line 26 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 26 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 26 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 26 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 26 "ms_deisotope/_c/scoring.pxd"
}

#line 26 "ms_deisotope/_c/scoring.pxd"


#line 26 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 26 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_5_hash_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 26 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_5_hash_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 26 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 26 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 26 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 26 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_5_hash_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 26 "ms_deisotope/_c/scoring.pxd"


#line 26 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 26 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 26 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 26 "ms_deisotope/_c/scoring.pxd"
}

#line 26 "ms_deisotope/_c/scoring.pxd"


#line 26 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_5_hash_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_self, PyObject *__pyx_v_value) {

#line 26 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 26 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 26 "ms_deisotope/_c/scoring.pxd"
  Py_hash_t __pyx_t_1;

#line 26 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 26 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 26 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 26 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 26 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = __Pyx_PyInt_AsHash_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (Py_hash_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 26, __pyx_L1_error)

#line 26 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->_hash = __pyx_t_1;

#line 26 "ms_deisotope/_c/scoring.pxd"


#line 26 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 26 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 26 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 26 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 26 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitRecord._hash.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 26 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = -1;

#line 26 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 26 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 26 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 26 "ms_deisotope/_c/scoring.pxd"
}

/* "ms_deisotope/_c/scoring.pyx":171
 *         FitSelector is `minimizing` it is the maximal score to be a candidate.
 *     """
 *     def __init__(self, minimum_score=0):             # <<<<<<<<<<<<<<
 *         self.minimum_score = minimum_score
 * 
 */

#line 171 "ms_deisotope/_c/scoring.pyx"


#line 171 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 171 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 171 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_minimum_score = 0
#line 171 "ms_deisotope/_c/scoring.pyx"
;

#line 171 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 171 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 171 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 171 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 171 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 171 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 171 "ms_deisotope/_c/scoring.pyx"
  {

#line 171 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_minimum_score,0};

#line 171 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[1] = {0};

#line 171 "ms_deisotope/_c/scoring.pyx"
    values[0] = ((PyObject *)__pyx_int_0);

#line 171 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 171 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 171 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 171 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  1: 
#line 171 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 171 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 171 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 171 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 171 "ms_deisotope/_c/scoring.pyx"
      }

#line 171 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 171 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 171 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 171 "ms_deisotope/_c/scoring.pyx"
        if (kw_args > 0) {

#line 171 "ms_deisotope/_c/scoring.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_minimum_score);

#line 171 "ms_deisotope/_c/scoring.pyx"
          if (value) { values[0] = value; kw_args--; }

#line 171 "ms_deisotope/_c/scoring.pyx"
        }

#line 171 "ms_deisotope/_c/scoring.pyx"
      }

#line 171 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 171 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 171, __pyx_L3_error)

#line 171 "ms_deisotope/_c/scoring.pyx"
      }

#line 171 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 171 "ms_deisotope/_c/scoring.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: 
#line 171 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 171 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  0: 
#line 171 "ms_deisotope/_c/scoring.pyx"
break;
        default: 
#line 171 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 171 "ms_deisotope/_c/scoring.pyx"
      }

#line 171 "ms_deisotope/_c/scoring.pyx"
    }

#line 171 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_minimum_score = values[0];

#line 171 "ms_deisotope/_c/scoring.pyx"
  }

#line 171 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 171 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); 
#line 171 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 171, __pyx_L3_error)

#line 171 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 171 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.FitSelectorBase.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 171 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 171 "ms_deisotope/_c/scoring.pyx"
  return -1;

#line 171 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 171 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase___init__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_v_self), __pyx_v_minimum_score);

#line 171 "ms_deisotope/_c/scoring.pyx"


#line 171 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 171 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 171 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 171 "ms_deisotope/_c/scoring.pyx"
}

#line 171 "ms_deisotope/_c/scoring.pyx"


#line 171 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase___init__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self, PyObject *__pyx_v_minimum_score) {

#line 171 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 171 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 171 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_1;

#line 171 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 171 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 171 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 171 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_deisotope/_c/scoring.pyx":172
 *     """
 *     def __init__(self, minimum_score=0):
 *         self.minimum_score = minimum_score             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */

#line 172 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_minimum_score); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 172, __pyx_L1_error)

#line 172 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->minimum_score = __pyx_t_1;

  /* "ms_deisotope/_c/scoring.pyx":171
 *         FitSelector is `minimizing` it is the maximal score to be a candidate.
 *     """
 *     def __init__(self, minimum_score=0):             # <<<<<<<<<<<<<<
 *         self.minimum_score = minimum_score
 * 
 */

#line 171 "ms_deisotope/_c/scoring.pyx"


#line 171 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 171 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 171 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 171 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 171 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.FitSelectorBase.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 171 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = -1;

#line 171 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 171 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 171 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 171 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":174
 *         self.minimum_score = minimum_score
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.minimum_score,)
 * 
 */

#line 174 "ms_deisotope/_c/scoring.pyx"


#line 174 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 174 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_3__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 174 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_15FitSelectorBase_2__reduce__[] = "FitSelectorBase.__reduce__(self)";

#line 174 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_3__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 174 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 174 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 174 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = 
#line 174 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase_2__reduce__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_v_self));

#line 174 "ms_deisotope/_c/scoring.pyx"


#line 174 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 174 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 174 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 174 "ms_deisotope/_c/scoring.pyx"
}

#line 174 "ms_deisotope/_c/scoring.pyx"


#line 174 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase_2__reduce__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self) {

#line 174 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 174 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 174 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 174 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 174 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 174 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 174 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 174 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 174 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "ms_deisotope/_c/scoring.pyx":175
 * 
 *     def __reduce__(self):
 *         return self.__class__, (self.minimum_score,)             # <<<<<<<<<<<<<<
 * 
 *     cpdef IsotopicFitRecord best(self, object results):
 */

#line 175 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 175 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 175, __pyx_L1_error)

#line 175 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 175 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->minimum_score); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 175, __pyx_L1_error)

#line 175 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 175 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 175, __pyx_L1_error)

#line 175 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 175 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 175 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);

#line 175 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = 0;

#line 175 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 175, __pyx_L1_error)

#line 175 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 175 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 175 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);

#line 175 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 175 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);

#line 175 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 175 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = 0;

#line 175 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_2;

#line 175 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = 0;

#line 175 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":174
 *         self.minimum_score = minimum_score
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.minimum_score,)
 * 
 */

#line 174 "ms_deisotope/_c/scoring.pyx"


#line 174 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 174 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 174 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 174 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 174 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 174 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.FitSelectorBase.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 174 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 174 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 174 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 174 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 174 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 174 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":177
 *         return self.__class__, (self.minimum_score,)
 * 
 *     cpdef IsotopicFitRecord best(self, object results):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

#line 177 "ms_deisotope/_c/scoring.pyx"


#line 177 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_5best(PyObject *__pyx_v_self, PyObject *__pyx_v_results); /*proto*/

#line 177 "ms_deisotope/_c/scoring.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_f_12ms_deisotope_2_c_7scoring_15FitSelectorBase_best(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_results, int __pyx_skip_dispatch) {

#line 177 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_r = NULL;

#line 177 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 177 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 177 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 177 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 177 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 177 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 177 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 177 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 177 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("best", 0);

#line 177 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 177 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 177 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 177 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 177 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 177 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 177 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 177 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 177 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 177 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_best); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 177, __pyx_L1_error)

#line 177 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 177 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_5best)) {

#line 177 "ms_deisotope/_c/scoring.pyx"
        __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 177 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 177 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 177 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 177 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 177 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 177 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 177 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 177 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 177 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 177 "ms_deisotope/_c/scoring.pyx"
          }

#line 177 "ms_deisotope/_c/scoring.pyx"
        }

#line 177 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_results) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_results);

#line 177 "ms_deisotope/_c/scoring.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 177 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 177, __pyx_L1_error)

#line 177 "ms_deisotope/_c/scoring.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 177 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 177 "ms_deisotope/_c/scoring.pyx"
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord))))) __PYX_ERR(0, 177, __pyx_L1_error)

#line 177 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_t_2);

#line 177 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_2 = 0;

#line 177 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 177 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 177 "ms_deisotope/_c/scoring.pyx"
      }

#line 177 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 177 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 177 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 177 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 177 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 177 "ms_deisotope/_c/scoring.pyx"
      }

#line 177 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 177 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 177 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 177 "ms_deisotope/_c/scoring.pyx"
    }

#line 177 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 177 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":178
 * 
 *     cpdef IsotopicFitRecord best(self, object results):
 *         raise NotImplementedError()             # <<<<<<<<<<<<<<
 * 
 *     cdef IsotopicFitRecord _best_from_set(self, set results):
 */

#line 178 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_NotImplementedError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 178, __pyx_L1_error)

#line 178 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 178 "ms_deisotope/_c/scoring.pyx"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 178 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 178 "ms_deisotope/_c/scoring.pyx"
  __PYX_ERR(0, 178, __pyx_L1_error)

  /* "ms_deisotope/_c/scoring.pyx":177
 *         return self.__class__, (self.minimum_score,)
 * 
 *     cpdef IsotopicFitRecord best(self, object results):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

#line 177 "ms_deisotope/_c/scoring.pyx"


#line 177 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 177 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 177 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 177 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 177 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 177 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 177 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.FitSelectorBase.best", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 177 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 177 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 177 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 177 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 177 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 177 "ms_deisotope/_c/scoring.pyx"
}

#line 177 "ms_deisotope/_c/scoring.pyx"


#line 177 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 177 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_5best(PyObject *__pyx_v_self, PyObject *__pyx_v_results); /*proto*/

#line 177 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_15FitSelectorBase_4best[] = "FitSelectorBase.best(self, results) -> IsotopicFitRecord";

#line 177 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_5best(PyObject *__pyx_v_self, PyObject *__pyx_v_results) {

#line 177 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 177 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 177 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("best (wrapper)", 0);
  __pyx_r = 
#line 177 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase_4best(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_v_self), ((PyObject *)__pyx_v_results));

#line 177 "ms_deisotope/_c/scoring.pyx"


#line 177 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 177 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 177 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 177 "ms_deisotope/_c/scoring.pyx"
}

#line 177 "ms_deisotope/_c/scoring.pyx"


#line 177 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase_4best(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self, PyObject *__pyx_v_results) {

#line 177 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 177 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 177 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 177 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 177 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 177 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 177 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("best", 0);

#line 177 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 177 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_7scoring_15FitSelectorBase_best(__pyx_v_self, __pyx_v_results, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 177, __pyx_L1_error)

#line 177 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 177 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 177 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 177 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 177 "ms_deisotope/_c/scoring.pyx"


#line 177 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 177 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 177 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 177 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.FitSelectorBase.best", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 177 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 177 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 177 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 177 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 177 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 177 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":180
 *         raise NotImplementedError()
 * 
 *     cdef IsotopicFitRecord _best_from_set(self, set results):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

#line 180 "ms_deisotope/_c/scoring.pyx"


#line 180 "ms_deisotope/_c/scoring.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_f_12ms_deisotope_2_c_7scoring_15FitSelectorBase__best_from_set(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_results) {

#line 180 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_r = NULL;

#line 180 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 180 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 180 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 180 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 180 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 180 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_best_from_set", 0);

  /* "ms_deisotope/_c/scoring.pyx":181
 * 
 *     cdef IsotopicFitRecord _best_from_set(self, set results):
 *         raise NotImplementedError()             # <<<<<<<<<<<<<<
 * 
 *     def __call__(self, *args, **kwargs):
 */

#line 181 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_NotImplementedError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 181, __pyx_L1_error)

#line 181 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 181 "ms_deisotope/_c/scoring.pyx"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 181 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 181 "ms_deisotope/_c/scoring.pyx"
  __PYX_ERR(0, 181, __pyx_L1_error)

  /* "ms_deisotope/_c/scoring.pyx":180
 *         raise NotImplementedError()
 * 
 *     cdef IsotopicFitRecord _best_from_set(self, set results):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

#line 180 "ms_deisotope/_c/scoring.pyx"


#line 180 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 180 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 180 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 180 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.FitSelectorBase._best_from_set", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 180 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 180 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 180 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 180 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 180 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":183
 *         raise NotImplementedError()
 * 
 *     def __call__(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         return self.best(*args, **kwargs)
 * 
 */

#line 183 "ms_deisotope/_c/scoring.pyx"


#line 183 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 183 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_7__call__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 183 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_7__call__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0
#line 183 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_kwargs = 0
#line 183 "ms_deisotope/_c/scoring.pyx"
;

#line 183 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 183 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 183 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__call__ (wrapper)", 0);

#line 183 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__call__", 1))) return NULL;
  __pyx_v_kwargs = (__pyx_kwds) ? PyDict_Copy(__pyx_kwds) : PyDict_New(); 
#line 183 "ms_deisotope/_c/scoring.pyx"
if (unlikely(!__pyx_v_kwargs)) return NULL;

#line 183 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_kwargs);

#line 183 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_args);

#line 183 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_args = __pyx_args;
  __pyx_r = 
#line 183 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase_6__call__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

#line 183 "ms_deisotope/_c/scoring.pyx"


#line 183 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 183 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_v_args);

#line 183 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_v_kwargs);

#line 183 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 183 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 183 "ms_deisotope/_c/scoring.pyx"
}

#line 183 "ms_deisotope/_c/scoring.pyx"


#line 183 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase_6__call__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {

#line 183 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 183 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 183 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 183 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 183 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 183 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 183 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 183 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 183 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__call__", 0);

  /* "ms_deisotope/_c/scoring.pyx":184
 * 
 *     def __call__(self, *args, **kwargs):
 *         return self.best(*args, **kwargs)             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint reject(self, IsotopicFitRecord result):
 */

#line 184 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 184 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_best); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 184, __pyx_L1_error)

#line 184 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 184 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 184, __pyx_L1_error)

#line 184 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 184 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_v_args, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 184, __pyx_L1_error)

#line 184 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 184 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 184 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 184 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_3;

#line 184 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = 0;

#line 184 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":183
 *         raise NotImplementedError()
 * 
 *     def __call__(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         return self.best(*args, **kwargs)
 * 
 */

#line 183 "ms_deisotope/_c/scoring.pyx"


#line 183 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 183 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 183 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 183 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 183 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 183 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.FitSelectorBase.__call__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 183 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 183 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 183 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 183 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 183 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 183 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":186
 *         return self.best(*args, **kwargs)
 * 
 *     cpdef bint reject(self, IsotopicFitRecord result):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

#line 186 "ms_deisotope/_c/scoring.pyx"


#line 186 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_9reject(PyObject *__pyx_v_self, PyObject *__pyx_v_result); /*proto*/

#line 186 "ms_deisotope/_c/scoring.pyx"
static int __pyx_f_12ms_deisotope_2_c_7scoring_15FitSelectorBase_reject(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_result, int __pyx_skip_dispatch) {

#line 186 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 186 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 186 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 186 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 186 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 186 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 186 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_5;

#line 186 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 186 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 186 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 186 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("reject", 0);

#line 186 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 186 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 186 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 186 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 186 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 186 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 186 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 186 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 186 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 186 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_reject); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 186, __pyx_L1_error)

#line 186 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 186 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_9reject)) {

#line 186 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 186 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 186 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 186 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 186 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 186 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 186 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 186 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 186 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 186 "ms_deisotope/_c/scoring.pyx"
          }

#line 186 "ms_deisotope/_c/scoring.pyx"
        }

#line 186 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_result)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_result));

#line 186 "ms_deisotope/_c/scoring.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 186 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 186, __pyx_L1_error)

#line 186 "ms_deisotope/_c/scoring.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 186 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 186 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 186, __pyx_L1_error)

#line 186 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 186 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = __pyx_t_5;

#line 186 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 186 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 186 "ms_deisotope/_c/scoring.pyx"
      }

#line 186 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 186 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 186 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 186 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 186 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 186 "ms_deisotope/_c/scoring.pyx"
      }

#line 186 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 186 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 186 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 186 "ms_deisotope/_c/scoring.pyx"
    }

#line 186 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 186 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":187
 * 
 *     cpdef bint reject(self, IsotopicFitRecord result):
 *         raise NotImplementedError()             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint reject_score(self, double score):
 */

#line 187 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_NotImplementedError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 187, __pyx_L1_error)

#line 187 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 187 "ms_deisotope/_c/scoring.pyx"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 187 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 187 "ms_deisotope/_c/scoring.pyx"
  __PYX_ERR(0, 187, __pyx_L1_error)

  /* "ms_deisotope/_c/scoring.pyx":186
 *         return self.best(*args, **kwargs)
 * 
 *     cpdef bint reject(self, IsotopicFitRecord result):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

#line 186 "ms_deisotope/_c/scoring.pyx"


#line 186 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 186 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 186 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 186 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 186 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 186 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 186 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.FitSelectorBase.reject", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 186 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 186 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 186 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 186 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 186 "ms_deisotope/_c/scoring.pyx"
}

#line 186 "ms_deisotope/_c/scoring.pyx"


#line 186 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 186 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_9reject(PyObject *__pyx_v_self, PyObject *__pyx_v_result); /*proto*/

#line 186 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_15FitSelectorBase_8reject[] = "FitSelectorBase.reject(self, IsotopicFitRecord result) -> bool";

#line 186 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_9reject(PyObject *__pyx_v_self, PyObject *__pyx_v_result) {

#line 186 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 186 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 186 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 186 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 186 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 186 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("reject (wrapper)", 0);

#line 186 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord, 1, "result", 0))) __PYX_ERR(0, 186, __pyx_L1_error)
  __pyx_r = 
#line 186 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase_8reject(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_result));

#line 186 "ms_deisotope/_c/scoring.pyx"


#line 186 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 186 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 186 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 186 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 186 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 186 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 186 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 186 "ms_deisotope/_c/scoring.pyx"
}

#line 186 "ms_deisotope/_c/scoring.pyx"


#line 186 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase_8reject(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_result) {

#line 186 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 186 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 186 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 186 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 186 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 186 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 186 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("reject", 0);

#line 186 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 186 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_7scoring_15FitSelectorBase_reject(__pyx_v_self, __pyx_v_result, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 186, __pyx_L1_error)

#line 186 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 186 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 186 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 186 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 186 "ms_deisotope/_c/scoring.pyx"


#line 186 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 186 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 186 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 186 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.FitSelectorBase.reject", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 186 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 186 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 186 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 186 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 186 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 186 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":189
 *         raise NotImplementedError()
 * 
 *     cpdef bint reject_score(self, double score):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

#line 189 "ms_deisotope/_c/scoring.pyx"


#line 189 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_11reject_score(PyObject *__pyx_v_self, PyObject *__pyx_arg_score); /*proto*/

#line 189 "ms_deisotope/_c/scoring.pyx"
static int __pyx_f_12ms_deisotope_2_c_7scoring_15FitSelectorBase_reject_score(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self, CYTHON_UNUSED double __pyx_v_score, int __pyx_skip_dispatch) {

#line 189 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 189 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 189 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 189 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 189 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 189 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 189 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 189 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_6;

#line 189 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 189 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 189 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 189 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("reject_score", 0);

#line 189 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 189 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 189 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 189 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 189 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 189 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 189 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 189 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 189 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 189 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_reject_score); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 189, __pyx_L1_error)

#line 189 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 189 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_11reject_score)) {

#line 189 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_score); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 189, __pyx_L1_error)

#line 189 "ms_deisotope/_c/scoring.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 189 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
#line 189 "ms_deisotope/_c/scoring.pyx"
__pyx_t_5 = NULL;

#line 189 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {

#line 189 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 189 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_5)) {

#line 189 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 189 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_5);

#line 189 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 189 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_4, function);

#line 189 "ms_deisotope/_c/scoring.pyx"
          }

#line 189 "ms_deisotope/_c/scoring.pyx"
        }

#line 189 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);

#line 189 "ms_deisotope/_c/scoring.pyx"
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 189 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 189 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 189, __pyx_L1_error)

#line 189 "ms_deisotope/_c/scoring.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 189 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 189 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 189, __pyx_L1_error)

#line 189 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 189 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = __pyx_t_6;

#line 189 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 189 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 189 "ms_deisotope/_c/scoring.pyx"
      }

#line 189 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 189 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 189 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 189 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 189 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 189 "ms_deisotope/_c/scoring.pyx"
      }

#line 189 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 189 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 189 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 189 "ms_deisotope/_c/scoring.pyx"
    }

#line 189 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 189 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":190
 * 
 *     cpdef bint reject_score(self, double score):
 *         raise NotImplementedError()             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint is_maximizing(self):
 */

#line 190 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_NotImplementedError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 190, __pyx_L1_error)

#line 190 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 190 "ms_deisotope/_c/scoring.pyx"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 190 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 190 "ms_deisotope/_c/scoring.pyx"
  __PYX_ERR(0, 190, __pyx_L1_error)

  /* "ms_deisotope/_c/scoring.pyx":189
 *         raise NotImplementedError()
 * 
 *     cpdef bint reject_score(self, double score):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

#line 189 "ms_deisotope/_c/scoring.pyx"


#line 189 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 189 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 189 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 189 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 189 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 189 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 189 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 189 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.FitSelectorBase.reject_score", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 189 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 189 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 189 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 189 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 189 "ms_deisotope/_c/scoring.pyx"
}

#line 189 "ms_deisotope/_c/scoring.pyx"


#line 189 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 189 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_11reject_score(PyObject *__pyx_v_self, PyObject *__pyx_arg_score); /*proto*/

#line 189 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_15FitSelectorBase_10reject_score[] = "FitSelectorBase.reject_score(self, double score) -> bool";

#line 189 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_11reject_score(PyObject *__pyx_v_self, PyObject *__pyx_arg_score) {
  double __pyx_v_score
#line 189 "ms_deisotope/_c/scoring.pyx"
;

#line 189 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 189 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 189 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 189 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 189 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 189 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("reject_score (wrapper)", 0);

#line 189 "ms_deisotope/_c/scoring.pyx"
  assert(__pyx_arg_score); {

#line 189 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_score = __pyx_PyFloat_AsDouble(__pyx_arg_score); if (unlikely((__pyx_v_score == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 189, __pyx_L3_error)

#line 189 "ms_deisotope/_c/scoring.pyx"
  }

#line 189 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 189 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 189 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.FitSelectorBase.reject_score", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 189 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 189 "ms_deisotope/_c/scoring.pyx"
  return NULL;

#line 189 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 189 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase_10reject_score(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_v_self), ((double)__pyx_v_score));

#line 189 "ms_deisotope/_c/scoring.pyx"


#line 189 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 189 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 189 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 189 "ms_deisotope/_c/scoring.pyx"
}

#line 189 "ms_deisotope/_c/scoring.pyx"


#line 189 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase_10reject_score(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self, double __pyx_v_score) {

#line 189 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 189 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 189 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 189 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 189 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 189 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 189 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("reject_score", 0);

#line 189 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 189 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_7scoring_15FitSelectorBase_reject_score(__pyx_v_self, __pyx_v_score, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 189, __pyx_L1_error)

#line 189 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 189 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 189 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 189 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 189 "ms_deisotope/_c/scoring.pyx"


#line 189 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 189 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 189 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 189 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.FitSelectorBase.reject_score", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 189 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 189 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 189 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 189 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 189 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 189 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":192
 *         raise NotImplementedError()
 * 
 *     cpdef bint is_maximizing(self):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */

#line 192 "ms_deisotope/_c/scoring.pyx"


#line 192 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_13is_maximizing(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 192 "ms_deisotope/_c/scoring.pyx"
static int __pyx_f_12ms_deisotope_2_c_7scoring_15FitSelectorBase_is_maximizing(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self, int __pyx_skip_dispatch) {

#line 192 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 192 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 192 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 192 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 192 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 192 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 192 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_5;

#line 192 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 192 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 192 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 192 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("is_maximizing", 0);

#line 192 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 192 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 192 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 192 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 192 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 192 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 192 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 192 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 192 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 192 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_maximizing); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 192, __pyx_L1_error)

#line 192 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 192 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_13is_maximizing)) {

#line 192 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 192 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 192 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 192 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 192 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 192 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 192 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 192 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 192 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 192 "ms_deisotope/_c/scoring.pyx"
          }

#line 192 "ms_deisotope/_c/scoring.pyx"
        }

#line 192 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 192 "ms_deisotope/_c/scoring.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 192 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 192, __pyx_L1_error)

#line 192 "ms_deisotope/_c/scoring.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 192 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 192 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 192, __pyx_L1_error)

#line 192 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 192 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = __pyx_t_5;

#line 192 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 192 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 192 "ms_deisotope/_c/scoring.pyx"
      }

#line 192 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 192 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 192 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 192 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 192 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 192 "ms_deisotope/_c/scoring.pyx"
      }

#line 192 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 192 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 192 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 192 "ms_deisotope/_c/scoring.pyx"
    }

#line 192 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 192 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":193
 * 
 *     cpdef bint is_maximizing(self):
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */

#line 193 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 193 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":192
 *         raise NotImplementedError()
 * 
 *     cpdef bint is_maximizing(self):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */

#line 192 "ms_deisotope/_c/scoring.pyx"


#line 192 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 192 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 192 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 192 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 192 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 192 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 192 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.FitSelectorBase.is_maximizing", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 192 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 192 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 192 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 192 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 192 "ms_deisotope/_c/scoring.pyx"
}

#line 192 "ms_deisotope/_c/scoring.pyx"


#line 192 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 192 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_13is_maximizing(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 192 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_15FitSelectorBase_12is_maximizing[] = "FitSelectorBase.is_maximizing(self) -> bool";

#line 192 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_13is_maximizing(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 192 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 192 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 192 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("is_maximizing (wrapper)", 0);
  __pyx_r = 
#line 192 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase_12is_maximizing(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_v_self));

#line 192 "ms_deisotope/_c/scoring.pyx"


#line 192 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 192 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 192 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 192 "ms_deisotope/_c/scoring.pyx"
}

#line 192 "ms_deisotope/_c/scoring.pyx"


#line 192 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase_12is_maximizing(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self) {

#line 192 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 192 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 192 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 192 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 192 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 192 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 192 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("is_maximizing", 0);

#line 192 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 192 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_7scoring_15FitSelectorBase_is_maximizing(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 192, __pyx_L1_error)

#line 192 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 192 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 192 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 192 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 192 "ms_deisotope/_c/scoring.pyx"


#line 192 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 192 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 192 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 192 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.FitSelectorBase.is_maximizing", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 192 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 192 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 192 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 192 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 192 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 192 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":195
 *         return False
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "{self.__class__.__name__}(minimum_score={self.minimum_score})".format(self=self)
 * 
 */

#line 195 "ms_deisotope/_c/scoring.pyx"


#line 195 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 195 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_15__repr__(PyObject *__pyx_v_self); /*proto*/

#line 195 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_15__repr__(PyObject *__pyx_v_self) {

#line 195 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 195 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 195 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = 
#line 195 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase_14__repr__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_v_self));

#line 195 "ms_deisotope/_c/scoring.pyx"


#line 195 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 195 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 195 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 195 "ms_deisotope/_c/scoring.pyx"
}

#line 195 "ms_deisotope/_c/scoring.pyx"


#line 195 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase_14__repr__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self) {

#line 195 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 195 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 195 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 195 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 195 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 195 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 195 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 195 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 195 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "ms_deisotope/_c/scoring.pyx":196
 * 
 *     def __repr__(self):
 *         return "{self.__class__.__name__}(minimum_score={self.minimum_score})".format(self=self)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 196 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 196 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_self___class_____name___minimum, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 196, __pyx_L1_error)

#line 196 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 196 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 196, __pyx_L1_error)

#line 196 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 196 "ms_deisotope/_c/scoring.pyx"
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_self, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 196, __pyx_L1_error)

#line 196 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 196, __pyx_L1_error)

#line 196 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 196 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 196 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 196 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_3;

#line 196 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = 0;

#line 196 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":195
 *         return False
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "{self.__class__.__name__}(minimum_score={self.minimum_score})".format(self=self)
 * 
 */

#line 195 "ms_deisotope/_c/scoring.pyx"


#line 195 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 195 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 195 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 195 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 195 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 195 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.FitSelectorBase.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 195 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 195 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 195 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 195 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 195 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 195 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pxd":40
 * cdef class FitSelectorBase(object):
 *     cdef:
 *         public double minimum_score             # <<<<<<<<<<<<<<
 * 
 *     cpdef IsotopicFitRecord best(self, object results)
 */

#line 40 "ms_deisotope/_c/scoring.pxd"


#line 40 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 40 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_13minimum_score_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 40 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_13minimum_score_1__get__(PyObject *__pyx_v_self) {

#line 40 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = 0;

#line 40 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 40 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 40 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase_13minimum_score___get__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_v_self));

#line 40 "ms_deisotope/_c/scoring.pxd"


#line 40 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 40 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 40 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 40 "ms_deisotope/_c/scoring.pxd"
}

#line 40 "ms_deisotope/_c/scoring.pxd"


#line 40 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase_13minimum_score___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self) {

#line 40 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = NULL;

#line 40 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 40 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 40 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 40 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 40 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 40 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 40 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 40 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->minimum_score); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 40, __pyx_L1_error)

#line 40 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 40 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = __pyx_t_1;

#line 40 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = 0;

#line 40 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 40 "ms_deisotope/_c/scoring.pxd"


#line 40 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 40 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 40 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 40 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.FitSelectorBase.minimum_score.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 40 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = NULL;

#line 40 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 40 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 40 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 40 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 40 "ms_deisotope/_c/scoring.pxd"
}

#line 40 "ms_deisotope/_c/scoring.pxd"


#line 40 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 40 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_13minimum_score_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 40 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_13minimum_score_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 40 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 40 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 40 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 40 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase_13minimum_score_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 40 "ms_deisotope/_c/scoring.pxd"


#line 40 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 40 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 40 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 40 "ms_deisotope/_c/scoring.pxd"
}

#line 40 "ms_deisotope/_c/scoring.pxd"


#line 40 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_15FitSelectorBase_13minimum_score_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_v_self, PyObject *__pyx_v_value) {

#line 40 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 40 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 40 "ms_deisotope/_c/scoring.pxd"
  double __pyx_t_1;

#line 40 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 40 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 40 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 40 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 40 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 40, __pyx_L1_error)

#line 40 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->minimum_score = __pyx_t_1;

#line 40 "ms_deisotope/_c/scoring.pxd"


#line 40 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 40 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 40 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 40 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 40 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.FitSelectorBase.minimum_score.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 40 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = -1;

#line 40 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 40 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 40 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 40 "ms_deisotope/_c/scoring.pxd"
}

/* "ms_deisotope/_c/scoring.pyx":202
 *     """A FitSelector which tries to minimize the score of the best fit.
 *     """
 *     cpdef IsotopicFitRecord best(self, object results):             # <<<<<<<<<<<<<<
 *         """Returns the IsotopicFitRecord with the smallest score
 * 
 */

#line 202 "ms_deisotope/_c/scoring.pyx"


#line 202 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_1best(PyObject *__pyx_v_self, PyObject *__pyx_v_results); /*proto*/

#line 202 "ms_deisotope/_c/scoring.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_f_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_best(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MinimizeFitSelector *__pyx_v_self, PyObject *__pyx_v_results, int __pyx_skip_dispatch) {

#line 202 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_r = NULL;

#line 202 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 202 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 202 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 202 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 202 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 202 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_5;

#line 202 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_6;

#line 202 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_7 = NULL;

#line 202 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 202 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 202 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 202 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("best", 0);

#line 202 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 202 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 202 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 202 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 202 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 202 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 202 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 202 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 202 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 202 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_best); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 202, __pyx_L1_error)

#line 202 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 202 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_1best)) {

#line 202 "ms_deisotope/_c/scoring.pyx"
        __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 202 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 202 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 202 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 202 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 202 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 202 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 202 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 202 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 202 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 202 "ms_deisotope/_c/scoring.pyx"
          }

#line 202 "ms_deisotope/_c/scoring.pyx"
        }

#line 202 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_results) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_results);

#line 202 "ms_deisotope/_c/scoring.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 202 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 202, __pyx_L1_error)

#line 202 "ms_deisotope/_c/scoring.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 202 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 202 "ms_deisotope/_c/scoring.pyx"
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord))))) __PYX_ERR(0, 202, __pyx_L1_error)

#line 202 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_t_2);

#line 202 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_2 = 0;

#line 202 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 202 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 202 "ms_deisotope/_c/scoring.pyx"
      }

#line 202 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 202 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 202 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 202 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 202 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 202 "ms_deisotope/_c/scoring.pyx"
      }

#line 202 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 202 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 202 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 202 "ms_deisotope/_c/scoring.pyx"
    }

#line 202 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 202 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":215
 *             The most optimal fit
 *         """
 *         if isinstance(results, set):             # <<<<<<<<<<<<<<
 *             return self._best_from_set(<set>results)
 *         return min(results, key=operator.attrgetter("score"))
 */

#line 215 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_5 = PySet_Check(__pyx_v_results); 

#line 215 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_6 = (__pyx_t_5 != 0);

#line 215 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_t_6) {

    /* "ms_deisotope/_c/scoring.pyx":216
 *         """
 *         if isinstance(results, set):
 *             return self._best_from_set(<set>results)             # <<<<<<<<<<<<<<
 *         return min(results, key=operator.attrgetter("score"))
 * 
 */

#line 216 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 216 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MinimizeFitSelector *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._best_from_set(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_v_self), ((PyObject*)__pyx_v_results))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/scoring.pyx"
    __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_t_1);

#line 216 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = 0;

#line 216 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/scoring.pyx":215
 *             The most optimal fit
 *         """
 *         if isinstance(results, set):             # <<<<<<<<<<<<<<
 *             return self._best_from_set(<set>results)
 *         return min(results, key=operator.attrgetter("score"))
 */

#line 215 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":217
 *         if isinstance(results, set):
 *             return self._best_from_set(<set>results)
 *         return min(results, key=operator.attrgetter("score"))             # <<<<<<<<<<<<<<
 * 
 *     cdef IsotopicFitRecord _best_from_set(self, set results):
 */

#line 217 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 217 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 217, __pyx_L1_error)

#line 217 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 217 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_v_results);

#line 217 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_v_results);

#line 217 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_results);

#line 217 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 217, __pyx_L1_error)

#line 217 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 217 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_operator); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 217, __pyx_L1_error)

#line 217 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 217 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_attrgetter); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 217, __pyx_L1_error)

#line 217 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_7);

#line 217 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 217 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_4 = NULL;

#line 217 "ms_deisotope/_c/scoring.pyx"
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {

#line 217 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);

#line 217 "ms_deisotope/_c/scoring.pyx"
    if (likely(__pyx_t_4)) {

#line 217 "ms_deisotope/_c/scoring.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);

#line 217 "ms_deisotope/_c/scoring.pyx"
      __Pyx_INCREF(__pyx_t_4);

#line 217 "ms_deisotope/_c/scoring.pyx"
      __Pyx_INCREF(function);

#line 217 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF_SET(__pyx_t_7, function);

#line 217 "ms_deisotope/_c/scoring.pyx"
    }

#line 217 "ms_deisotope/_c/scoring.pyx"
  }

#line 217 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_4, __pyx_n_s_score) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_n_s_score);

#line 217 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 217 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 217, __pyx_L1_error)

#line 217 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 217 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 217 "ms_deisotope/_c/scoring.pyx"
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_key, __pyx_t_3) < 0) __PYX_ERR(0, 217, __pyx_L1_error)

#line 217 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 217 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_min, __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 217, __pyx_L1_error)

#line 217 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 217 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 217 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 217 "ms_deisotope/_c/scoring.pyx"
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord))))) __PYX_ERR(0, 217, __pyx_L1_error)

#line 217 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_t_3);

#line 217 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = 0;

#line 217 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":202
 *     """A FitSelector which tries to minimize the score of the best fit.
 *     """
 *     cpdef IsotopicFitRecord best(self, object results):             # <<<<<<<<<<<<<<
 *         """Returns the IsotopicFitRecord with the smallest score
 * 
 */

#line 202 "ms_deisotope/_c/scoring.pyx"


#line 202 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 202 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 202 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 202 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 202 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 202 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 202 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_7);

#line 202 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.MinimizeFitSelector.best", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 202 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 202 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 202 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 202 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 202 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 202 "ms_deisotope/_c/scoring.pyx"
}

#line 202 "ms_deisotope/_c/scoring.pyx"


#line 202 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 202 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_1best(PyObject *__pyx_v_self, PyObject *__pyx_v_results); /*proto*/

#line 202 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_best[] = "MinimizeFitSelector.best(self, results) -> IsotopicFitRecord\nReturns the IsotopicFitRecord with the smallest score\n\n        Parameters\n        ----------\n        results : list of IsotopicFitRecord\n            List of isotopic fits to select the most optimal case from\n\n        Returns\n        -------\n        IsotopicFitRecord\n            The most optimal fit\n        ";

#line 202 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_1best(PyObject *__pyx_v_self, PyObject *__pyx_v_results) {

#line 202 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 202 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 202 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("best (wrapper)", 0);
  __pyx_r = 
#line 202 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_best(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_MinimizeFitSelector *)__pyx_v_self), ((PyObject *)__pyx_v_results));

#line 202 "ms_deisotope/_c/scoring.pyx"


#line 202 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 202 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 202 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 202 "ms_deisotope/_c/scoring.pyx"
}

#line 202 "ms_deisotope/_c/scoring.pyx"


#line 202 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_best(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MinimizeFitSelector *__pyx_v_self, PyObject *__pyx_v_results) {

#line 202 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 202 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 202 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 202 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 202 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 202 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 202 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("best", 0);

#line 202 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 202 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_best(__pyx_v_self, __pyx_v_results, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 202, __pyx_L1_error)

#line 202 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 202 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 202 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 202 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 202 "ms_deisotope/_c/scoring.pyx"


#line 202 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 202 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 202 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 202 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.MinimizeFitSelector.best", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 202 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 202 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 202 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 202 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 202 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 202 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":219
 *         return min(results, key=operator.attrgetter("score"))
 * 
 *     cdef IsotopicFitRecord _best_from_set(self, set results):             # <<<<<<<<<<<<<<
 *         cdef:
 *             IsotopicFitRecord fit, best_fit
 */

#line 219 "ms_deisotope/_c/scoring.pyx"


#line 219 "ms_deisotope/_c/scoring.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_f_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector__best_from_set(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_MinimizeFitSelector *__pyx_v_self, PyObject *__pyx_v_results) {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_fit = 0
#line 219 "ms_deisotope/_c/scoring.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_best_fit = 0
#line 219 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_obj = 0
#line 219 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_best_score
#line 219 "ms_deisotope/_c/scoring.pyx"
;

#line 219 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_r = NULL;

#line 219 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 219 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 219 "ms_deisotope/_c/scoring.pyx"
  Py_ssize_t __pyx_t_2;

#line 219 "ms_deisotope/_c/scoring.pyx"
  Py_ssize_t __pyx_t_3;

#line 219 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_4;

#line 219 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 219 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_6;

#line 219 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_7;

#line 219 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_8;

#line 219 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 219 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 219 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 219 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_best_from_set", 0);

  /* "ms_deisotope/_c/scoring.pyx":225
 *             double best_score
 * 
 *         best_fit = None             # <<<<<<<<<<<<<<
 *         best_score = INFINITY
 *         for obj in results:
 */

#line 225 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(Py_None);

#line 225 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_best_fit = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)Py_None);

  /* "ms_deisotope/_c/scoring.pyx":226
 * 
 *         best_fit = None
 *         best_score = INFINITY             # <<<<<<<<<<<<<<
 *         for obj in results:
 *             fit = <IsotopicFitRecord>obj
 */

#line 226 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_best_score = __pyx_v_12ms_deisotope_2_c_7scoring_INFINITY;

  /* "ms_deisotope/_c/scoring.pyx":227
 *         best_fit = None
 *         best_score = INFINITY
 *         for obj in results:             # <<<<<<<<<<<<<<
 *             fit = <IsotopicFitRecord>obj
 *             if fit.score < best_score:
 */

#line 227 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = 0;

#line 227 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_5 = __Pyx_set_iterator(__pyx_v_results, 1, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 227, __pyx_L1_error)

#line 227 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 227 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 227 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __pyx_t_5;

#line 227 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_5 = 0;

#line 227 "ms_deisotope/_c/scoring.pyx"
  while (1) {

#line 227 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_6 = __Pyx_set_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, __pyx_t_4);

#line 227 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_t_6 == 0)) break;

#line 227 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 227, __pyx_L1_error)

#line 227 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_5);

#line 227 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF_SET(__pyx_v_obj, __pyx_t_5);

#line 227 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = 0;

    /* "ms_deisotope/_c/scoring.pyx":228
 *         best_score = INFINITY
 *         for obj in results:
 *             fit = <IsotopicFitRecord>obj             # <<<<<<<<<<<<<<
 *             if fit.score < best_score:
 *                 best_score = fit.score
 */

#line 228 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = __pyx_v_obj;

#line 228 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_5);

#line 228 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF_SET(__pyx_v_fit, ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_t_5));

#line 228 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = 0;

    /* "ms_deisotope/_c/scoring.pyx":229
 *         for obj in results:
 *             fit = <IsotopicFitRecord>obj
 *             if fit.score < best_score:             # <<<<<<<<<<<<<<
 *                 best_score = fit.score
 *                 best_fit = fit
 */

#line 229 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_7 = ((__pyx_v_fit->score < __pyx_v_best_score) != 0);

#line 229 "ms_deisotope/_c/scoring.pyx"
    if (__pyx_t_7) {

      /* "ms_deisotope/_c/scoring.pyx":230
 *             fit = <IsotopicFitRecord>obj
 *             if fit.score < best_score:
 *                 best_score = fit.score             # <<<<<<<<<<<<<<
 *                 best_fit = fit
 *         return best_fit
 */

#line 230 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_8 = __pyx_v_fit->score;

#line 230 "ms_deisotope/_c/scoring.pyx"
      __pyx_v_best_score = __pyx_t_8;

      /* "ms_deisotope/_c/scoring.pyx":231
 *             if fit.score < best_score:
 *                 best_score = fit.score
 *                 best_fit = fit             # <<<<<<<<<<<<<<
 *         return best_fit
 * 
 */

#line 231 "ms_deisotope/_c/scoring.pyx"
      __Pyx_INCREF(((PyObject *)__pyx_v_fit));

#line 231 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF_SET(__pyx_v_best_fit, __pyx_v_fit);

      /* "ms_deisotope/_c/scoring.pyx":229
 *         for obj in results:
 *             fit = <IsotopicFitRecord>obj
 *             if fit.score < best_score:             # <<<<<<<<<<<<<<
 *                 best_score = fit.score
 *                 best_fit = fit
 */

#line 229 "ms_deisotope/_c/scoring.pyx"
    }

#line 229 "ms_deisotope/_c/scoring.pyx"
  }

#line 229 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/scoring.pyx":232
 *                 best_score = fit.score
 *                 best_fit = fit
 *         return best_fit             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 232 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 232 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_best_fit));

#line 232 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_v_best_fit;

#line 232 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":219
 *         return min(results, key=operator.attrgetter("score"))
 * 
 *     cdef IsotopicFitRecord _best_from_set(self, set results):             # <<<<<<<<<<<<<<
 *         cdef:
 *             IsotopicFitRecord fit, best_fit
 */

#line 219 "ms_deisotope/_c/scoring.pyx"


#line 219 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 219 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 219 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 219 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 219 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.MinimizeFitSelector._best_from_set", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 219 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 219 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 219 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_fit);

#line 219 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_best_fit);

#line 219 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_v_obj);

#line 219 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 219 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 219 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 219 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":235
 * 
 * 
 *     cpdef bint reject(self, IsotopicFitRecord fit):             # <<<<<<<<<<<<<<
 *         """Decide whether the fit should be discarded for having too
 *         large a score. Compares against :attr:`minimum_score`
 */

#line 235 "ms_deisotope/_c/scoring.pyx"


#line 235 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_3reject(PyObject *__pyx_v_self, PyObject *__pyx_v_fit); /*proto*/

#line 235 "ms_deisotope/_c/scoring.pyx"
static int __pyx_f_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_reject(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MinimizeFitSelector *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_fit, int __pyx_skip_dispatch) {

#line 235 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 235 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 235 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 235 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 235 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 235 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 235 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_5;

#line 235 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 235 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 235 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 235 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("reject", 0);

#line 235 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 235 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 235 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 235 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 235 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 235 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 235 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 235 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 235 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 235 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_reject); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 235, __pyx_L1_error)

#line 235 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 235 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_3reject)) {

#line 235 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 235 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 235 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 235 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 235 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 235 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 235 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 235 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 235 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 235 "ms_deisotope/_c/scoring.pyx"
          }

#line 235 "ms_deisotope/_c/scoring.pyx"
        }

#line 235 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_fit)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_fit));

#line 235 "ms_deisotope/_c/scoring.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 235 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 235, __pyx_L1_error)

#line 235 "ms_deisotope/_c/scoring.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 235 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 235 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 235, __pyx_L1_error)

#line 235 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 235 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = __pyx_t_5;

#line 235 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 235 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 235 "ms_deisotope/_c/scoring.pyx"
      }

#line 235 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 235 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 235 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 235 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 235 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 235 "ms_deisotope/_c/scoring.pyx"
      }

#line 235 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 235 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 235 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 235 "ms_deisotope/_c/scoring.pyx"
    }

#line 235 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 235 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":247
 *         bool
 *         """
 *         return fit.score > self.minimum_score             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint reject_score(self, double score):
 */

#line 247 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = (__pyx_v_fit->score > __pyx_v_self->__pyx_base.minimum_score);

#line 247 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":235
 * 
 * 
 *     cpdef bint reject(self, IsotopicFitRecord fit):             # <<<<<<<<<<<<<<
 *         """Decide whether the fit should be discarded for having too
 *         large a score. Compares against :attr:`minimum_score`
 */

#line 235 "ms_deisotope/_c/scoring.pyx"


#line 235 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 235 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 235 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 235 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 235 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 235 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 235 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.MinimizeFitSelector.reject", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 235 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 235 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 235 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 235 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 235 "ms_deisotope/_c/scoring.pyx"
}

#line 235 "ms_deisotope/_c/scoring.pyx"


#line 235 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 235 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_3reject(PyObject *__pyx_v_self, PyObject *__pyx_v_fit); /*proto*/

#line 235 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_2reject[] = "MinimizeFitSelector.reject(self, IsotopicFitRecord fit) -> bool\nDecide whether the fit should be discarded for having too\n        large a score. Compares against :attr:`minimum_score`\n\n        Parameters\n        ----------\n        fit : IsotopicFitRecord\n\n        Returns\n        -------\n        bool\n        ";

#line 235 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_3reject(PyObject *__pyx_v_self, PyObject *__pyx_v_fit) {

#line 235 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 235 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 235 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 235 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 235 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 235 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("reject (wrapper)", 0);

#line 235 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_fit), __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord, 1, "fit", 0))) __PYX_ERR(0, 235, __pyx_L1_error)
  __pyx_r = 
#line 235 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_2reject(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_MinimizeFitSelector *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_fit));

#line 235 "ms_deisotope/_c/scoring.pyx"


#line 235 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 235 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 235 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 235 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 235 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 235 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 235 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 235 "ms_deisotope/_c/scoring.pyx"
}

#line 235 "ms_deisotope/_c/scoring.pyx"


#line 235 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_2reject(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MinimizeFitSelector *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_fit) {

#line 235 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 235 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 235 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 235 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 235 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 235 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 235 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("reject", 0);

#line 235 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 235 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_reject(__pyx_v_self, __pyx_v_fit, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 235, __pyx_L1_error)

#line 235 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 235 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 235 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 235 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 235 "ms_deisotope/_c/scoring.pyx"


#line 235 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 235 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 235 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 235 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.MinimizeFitSelector.reject", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 235 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 235 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 235 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 235 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 235 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 235 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":249
 *         return fit.score > self.minimum_score
 * 
 *     cpdef bint reject_score(self, double score):             # <<<<<<<<<<<<<<
 *         return score > self.minimum_score
 * 
 */

#line 249 "ms_deisotope/_c/scoring.pyx"


#line 249 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_5reject_score(PyObject *__pyx_v_self, PyObject *__pyx_arg_score); /*proto*/

#line 249 "ms_deisotope/_c/scoring.pyx"
static int __pyx_f_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_reject_score(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MinimizeFitSelector *__pyx_v_self, double __pyx_v_score, int __pyx_skip_dispatch) {

#line 249 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 249 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 249 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 249 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 249 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 249 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 249 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 249 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_6;

#line 249 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 249 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 249 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 249 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("reject_score", 0);

#line 249 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 249 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 249 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 249 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 249 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 249 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 249 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 249 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 249 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 249 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_reject_score); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 249, __pyx_L1_error)

#line 249 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 249 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_5reject_score)) {

#line 249 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_score); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 249, __pyx_L1_error)

#line 249 "ms_deisotope/_c/scoring.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 249 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
#line 249 "ms_deisotope/_c/scoring.pyx"
__pyx_t_5 = NULL;

#line 249 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {

#line 249 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 249 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_5)) {

#line 249 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 249 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_5);

#line 249 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 249 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_4, function);

#line 249 "ms_deisotope/_c/scoring.pyx"
          }

#line 249 "ms_deisotope/_c/scoring.pyx"
        }

#line 249 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);

#line 249 "ms_deisotope/_c/scoring.pyx"
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 249 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 249 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 249, __pyx_L1_error)

#line 249 "ms_deisotope/_c/scoring.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 249 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 249 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 249, __pyx_L1_error)

#line 249 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 249 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = __pyx_t_6;

#line 249 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 249 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 249 "ms_deisotope/_c/scoring.pyx"
      }

#line 249 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 249 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 249 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 249 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 249 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 249 "ms_deisotope/_c/scoring.pyx"
      }

#line 249 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 249 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 249 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 249 "ms_deisotope/_c/scoring.pyx"
    }

#line 249 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 249 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":250
 * 
 *     cpdef bint reject_score(self, double score):
 *         return score > self.minimum_score             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint is_maximizing(self):
 */

#line 250 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = (__pyx_v_score > __pyx_v_self->__pyx_base.minimum_score);

#line 250 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":249
 *         return fit.score > self.minimum_score
 * 
 *     cpdef bint reject_score(self, double score):             # <<<<<<<<<<<<<<
 *         return score > self.minimum_score
 * 
 */

#line 249 "ms_deisotope/_c/scoring.pyx"


#line 249 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 249 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 249 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 249 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 249 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 249 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 249 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 249 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.MinimizeFitSelector.reject_score", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 249 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 249 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 249 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 249 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 249 "ms_deisotope/_c/scoring.pyx"
}

#line 249 "ms_deisotope/_c/scoring.pyx"


#line 249 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 249 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_5reject_score(PyObject *__pyx_v_self, PyObject *__pyx_arg_score); /*proto*/

#line 249 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_4reject_score[] = "MinimizeFitSelector.reject_score(self, double score) -> bool";

#line 249 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_5reject_score(PyObject *__pyx_v_self, PyObject *__pyx_arg_score) {
  double __pyx_v_score
#line 249 "ms_deisotope/_c/scoring.pyx"
;

#line 249 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 249 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 249 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 249 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 249 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 249 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("reject_score (wrapper)", 0);

#line 249 "ms_deisotope/_c/scoring.pyx"
  assert(__pyx_arg_score); {

#line 249 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_score = __pyx_PyFloat_AsDouble(__pyx_arg_score); if (unlikely((__pyx_v_score == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 249, __pyx_L3_error)

#line 249 "ms_deisotope/_c/scoring.pyx"
  }

#line 249 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 249 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 249 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.MinimizeFitSelector.reject_score", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 249 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 249 "ms_deisotope/_c/scoring.pyx"
  return NULL;

#line 249 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 249 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_4reject_score(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_MinimizeFitSelector *)__pyx_v_self), ((double)__pyx_v_score));

#line 249 "ms_deisotope/_c/scoring.pyx"


#line 249 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 249 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 249 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 249 "ms_deisotope/_c/scoring.pyx"
}

#line 249 "ms_deisotope/_c/scoring.pyx"


#line 249 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_4reject_score(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MinimizeFitSelector *__pyx_v_self, double __pyx_v_score) {

#line 249 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 249 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 249 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 249 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 249 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 249 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 249 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("reject_score", 0);

#line 249 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 249 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_reject_score(__pyx_v_self, __pyx_v_score, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 249, __pyx_L1_error)

#line 249 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 249 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 249 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 249 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 249 "ms_deisotope/_c/scoring.pyx"


#line 249 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 249 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 249 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 249 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.MinimizeFitSelector.reject_score", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 249 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 249 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 249 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 249 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 249 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 249 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":252
 *         return score > self.minimum_score
 * 
 *     cpdef bint is_maximizing(self):             # <<<<<<<<<<<<<<
 *         """Returns that this is *not* a maximizing selector
 * 
 */

#line 252 "ms_deisotope/_c/scoring.pyx"


#line 252 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_7is_maximizing(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 252 "ms_deisotope/_c/scoring.pyx"
static int __pyx_f_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_is_maximizing(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_MinimizeFitSelector *__pyx_v_self, int __pyx_skip_dispatch) {

#line 252 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 252 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 252 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 252 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 252 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 252 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 252 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_5;

#line 252 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 252 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 252 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 252 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("is_maximizing", 0);

#line 252 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 252 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 252 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 252 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 252 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 252 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 252 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 252 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 252 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 252 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_maximizing); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 252, __pyx_L1_error)

#line 252 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 252 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_7is_maximizing)) {

#line 252 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 252 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 252 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 252 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 252 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 252 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 252 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 252 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 252 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 252 "ms_deisotope/_c/scoring.pyx"
          }

#line 252 "ms_deisotope/_c/scoring.pyx"
        }

#line 252 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 252 "ms_deisotope/_c/scoring.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 252 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 252, __pyx_L1_error)

#line 252 "ms_deisotope/_c/scoring.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 252 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 252 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 252, __pyx_L1_error)

#line 252 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 252 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = __pyx_t_5;

#line 252 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 252 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 252 "ms_deisotope/_c/scoring.pyx"
      }

#line 252 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 252 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 252 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 252 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 252 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 252 "ms_deisotope/_c/scoring.pyx"
      }

#line 252 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 252 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 252 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 252 "ms_deisotope/_c/scoring.pyx"
    }

#line 252 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 252 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":259
 *         False
 *         """
 *         return False             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 259 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 259 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":252
 *         return score > self.minimum_score
 * 
 *     cpdef bint is_maximizing(self):             # <<<<<<<<<<<<<<
 *         """Returns that this is *not* a maximizing selector
 * 
 */

#line 252 "ms_deisotope/_c/scoring.pyx"


#line 252 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 252 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 252 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 252 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 252 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 252 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 252 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.MinimizeFitSelector.is_maximizing", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 252 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 252 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 252 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 252 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 252 "ms_deisotope/_c/scoring.pyx"
}

#line 252 "ms_deisotope/_c/scoring.pyx"


#line 252 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 252 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_7is_maximizing(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 252 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_6is_maximizing[] = "MinimizeFitSelector.is_maximizing(self) -> bool\nReturns that this is *not* a maximizing selector\n\n        Returns\n        -------\n        False\n        ";

#line 252 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_7is_maximizing(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 252 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 252 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 252 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("is_maximizing (wrapper)", 0);
  __pyx_r = 
#line 252 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_6is_maximizing(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_MinimizeFitSelector *)__pyx_v_self));

#line 252 "ms_deisotope/_c/scoring.pyx"


#line 252 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 252 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 252 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 252 "ms_deisotope/_c/scoring.pyx"
}

#line 252 "ms_deisotope/_c/scoring.pyx"


#line 252 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_6is_maximizing(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MinimizeFitSelector *__pyx_v_self) {

#line 252 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 252 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 252 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 252 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 252 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 252 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 252 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("is_maximizing", 0);

#line 252 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 252 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_is_maximizing(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 252, __pyx_L1_error)

#line 252 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 252 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 252 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 252 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 252 "ms_deisotope/_c/scoring.pyx"


#line 252 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 252 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 252 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 252 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.MinimizeFitSelector.is_maximizing", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 252 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 252 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 252 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 252 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 252 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 252 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":265
 *     """A FitSelector which tries to maximize the score of the best fit.
 *     """
 *     cpdef IsotopicFitRecord best(self, object results):             # <<<<<<<<<<<<<<
 *         """Returns the IsotopicFitRecord with the largest score
 * 
 */

#line 265 "ms_deisotope/_c/scoring.pyx"


#line 265 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_1best(PyObject *__pyx_v_self, PyObject *__pyx_v_results); /*proto*/

#line 265 "ms_deisotope/_c/scoring.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_f_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_best(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MaximizeFitSelector *__pyx_v_self, PyObject *__pyx_v_results, int __pyx_skip_dispatch) {

#line 265 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_r = NULL;

#line 265 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 265 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 265 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 265 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 265 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 265 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_5;

#line 265 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_6;

#line 265 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_7 = NULL;

#line 265 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 265 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 265 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 265 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("best", 0);

#line 265 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 265 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 265 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 265 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 265 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 265 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 265 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 265 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 265 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 265 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_best); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 265, __pyx_L1_error)

#line 265 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 265 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_1best)) {

#line 265 "ms_deisotope/_c/scoring.pyx"
        __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 265 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 265 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 265 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 265 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 265 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 265 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 265 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 265 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 265 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 265 "ms_deisotope/_c/scoring.pyx"
          }

#line 265 "ms_deisotope/_c/scoring.pyx"
        }

#line 265 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_results) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_results);

#line 265 "ms_deisotope/_c/scoring.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 265 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 265, __pyx_L1_error)

#line 265 "ms_deisotope/_c/scoring.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 265 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 265 "ms_deisotope/_c/scoring.pyx"
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord))))) __PYX_ERR(0, 265, __pyx_L1_error)

#line 265 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_t_2);

#line 265 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_2 = 0;

#line 265 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 265 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 265 "ms_deisotope/_c/scoring.pyx"
      }

#line 265 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 265 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 265 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 265 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 265 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 265 "ms_deisotope/_c/scoring.pyx"
      }

#line 265 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 265 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 265 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 265 "ms_deisotope/_c/scoring.pyx"
    }

#line 265 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 265 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":278
 *             The most optimal fit
 *         """
 *         if isinstance(results, set):             # <<<<<<<<<<<<<<
 *             return self._best_from_set(<set>results)
 *         return max(results, key=operator.attrgetter("score"))
 */

#line 278 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_5 = PySet_Check(__pyx_v_results); 

#line 278 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_6 = (__pyx_t_5 != 0);

#line 278 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_t_6) {

    /* "ms_deisotope/_c/scoring.pyx":279
 *         """
 *         if isinstance(results, set):
 *             return self._best_from_set(<set>results)             # <<<<<<<<<<<<<<
 *         return max(results, key=operator.attrgetter("score"))
 * 
 */

#line 279 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 279 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MaximizeFitSelector *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._best_from_set(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_v_self), ((PyObject*)__pyx_v_results))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 279, __pyx_L1_error)

#line 279 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 279 "ms_deisotope/_c/scoring.pyx"
    __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_t_1);

#line 279 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = 0;

#line 279 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/scoring.pyx":278
 *             The most optimal fit
 *         """
 *         if isinstance(results, set):             # <<<<<<<<<<<<<<
 *             return self._best_from_set(<set>results)
 *         return max(results, key=operator.attrgetter("score"))
 */

#line 278 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":280
 *         if isinstance(results, set):
 *             return self._best_from_set(<set>results)
 *         return max(results, key=operator.attrgetter("score"))             # <<<<<<<<<<<<<<
 * 
 *     cdef IsotopicFitRecord _best_from_set(self, set results):
 */

#line 280 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 280 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 280, __pyx_L1_error)

#line 280 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 280 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_v_results);

#line 280 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_v_results);

#line 280 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_results);

#line 280 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 280, __pyx_L1_error)

#line 280 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 280 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_operator); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 280, __pyx_L1_error)

#line 280 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 280 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_attrgetter); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 280, __pyx_L1_error)

#line 280 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_7);

#line 280 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 280 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_4 = NULL;

#line 280 "ms_deisotope/_c/scoring.pyx"
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {

#line 280 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);

#line 280 "ms_deisotope/_c/scoring.pyx"
    if (likely(__pyx_t_4)) {

#line 280 "ms_deisotope/_c/scoring.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);

#line 280 "ms_deisotope/_c/scoring.pyx"
      __Pyx_INCREF(__pyx_t_4);

#line 280 "ms_deisotope/_c/scoring.pyx"
      __Pyx_INCREF(function);

#line 280 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF_SET(__pyx_t_7, function);

#line 280 "ms_deisotope/_c/scoring.pyx"
    }

#line 280 "ms_deisotope/_c/scoring.pyx"
  }

#line 280 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_4, __pyx_n_s_score) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_n_s_score);

#line 280 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 280 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 280, __pyx_L1_error)

#line 280 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 280 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 280 "ms_deisotope/_c/scoring.pyx"
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_key, __pyx_t_3) < 0) __PYX_ERR(0, 280, __pyx_L1_error)

#line 280 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 280 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_max, __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 280, __pyx_L1_error)

#line 280 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 280 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 280 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 280 "ms_deisotope/_c/scoring.pyx"
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord))))) __PYX_ERR(0, 280, __pyx_L1_error)

#line 280 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_t_3);

#line 280 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = 0;

#line 280 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":265
 *     """A FitSelector which tries to maximize the score of the best fit.
 *     """
 *     cpdef IsotopicFitRecord best(self, object results):             # <<<<<<<<<<<<<<
 *         """Returns the IsotopicFitRecord with the largest score
 * 
 */

#line 265 "ms_deisotope/_c/scoring.pyx"


#line 265 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 265 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 265 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 265 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 265 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 265 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 265 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_7);

#line 265 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.MaximizeFitSelector.best", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 265 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 265 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 265 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 265 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 265 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 265 "ms_deisotope/_c/scoring.pyx"
}

#line 265 "ms_deisotope/_c/scoring.pyx"


#line 265 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 265 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_1best(PyObject *__pyx_v_self, PyObject *__pyx_v_results); /*proto*/

#line 265 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_best[] = "MaximizeFitSelector.best(self, results) -> IsotopicFitRecord\nReturns the IsotopicFitRecord with the largest score\n\n        Parameters\n        ----------\n        results : list of IsotopicFitRecord\n            List of isotopic fits to select the most optimal case from\n\n        Returns\n        -------\n        IsotopicFitRecord\n            The most optimal fit\n        ";

#line 265 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_1best(PyObject *__pyx_v_self, PyObject *__pyx_v_results) {

#line 265 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 265 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 265 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("best (wrapper)", 0);
  __pyx_r = 
#line 265 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_best(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_MaximizeFitSelector *)__pyx_v_self), ((PyObject *)__pyx_v_results));

#line 265 "ms_deisotope/_c/scoring.pyx"


#line 265 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 265 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 265 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 265 "ms_deisotope/_c/scoring.pyx"
}

#line 265 "ms_deisotope/_c/scoring.pyx"


#line 265 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_best(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MaximizeFitSelector *__pyx_v_self, PyObject *__pyx_v_results) {

#line 265 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 265 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 265 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 265 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 265 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 265 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 265 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("best", 0);

#line 265 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 265 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_best(__pyx_v_self, __pyx_v_results, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 265, __pyx_L1_error)

#line 265 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 265 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 265 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 265 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 265 "ms_deisotope/_c/scoring.pyx"


#line 265 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 265 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 265 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 265 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.MaximizeFitSelector.best", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 265 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 265 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 265 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 265 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 265 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 265 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":282
 *         return max(results, key=operator.attrgetter("score"))
 * 
 *     cdef IsotopicFitRecord _best_from_set(self, set results):             # <<<<<<<<<<<<<<
 *         cdef:
 *             IsotopicFitRecord fit, best_fit
 */

#line 282 "ms_deisotope/_c/scoring.pyx"


#line 282 "ms_deisotope/_c/scoring.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_f_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector__best_from_set(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_MaximizeFitSelector *__pyx_v_self, PyObject *__pyx_v_results) {
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_fit = 0
#line 282 "ms_deisotope/_c/scoring.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_best_fit = 0
#line 282 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_best_score
#line 282 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_obj = NULL
#line 282 "ms_deisotope/_c/scoring.pyx"
;

#line 282 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_r = NULL;

#line 282 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 282 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 282 "ms_deisotope/_c/scoring.pyx"
  Py_ssize_t __pyx_t_2;

#line 282 "ms_deisotope/_c/scoring.pyx"
  Py_ssize_t __pyx_t_3;

#line 282 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_4;

#line 282 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 282 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_6;

#line 282 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_7;

#line 282 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_8;

#line 282 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 282 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 282 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 282 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_best_from_set", 0);

  /* "ms_deisotope/_c/scoring.pyx":287
 *             double best_score
 * 
 *         best_fit = None             # <<<<<<<<<<<<<<
 *         best_score = -INFINITY
 *         for obj in results:
 */

#line 287 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(Py_None);

#line 287 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_best_fit = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)Py_None);

  /* "ms_deisotope/_c/scoring.pyx":288
 * 
 *         best_fit = None
 *         best_score = -INFINITY             # <<<<<<<<<<<<<<
 *         for obj in results:
 *             fit = <IsotopicFitRecord>obj
 */

#line 288 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_best_score = (-__pyx_v_12ms_deisotope_2_c_7scoring_INFINITY);

  /* "ms_deisotope/_c/scoring.pyx":289
 *         best_fit = None
 *         best_score = -INFINITY
 *         for obj in results:             # <<<<<<<<<<<<<<
 *             fit = <IsotopicFitRecord>obj
 *             if fit.score > best_score:
 */

#line 289 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = 0;

#line 289 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_5 = __Pyx_set_iterator(__pyx_v_results, 1, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 289, __pyx_L1_error)

#line 289 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 289 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 289 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __pyx_t_5;

#line 289 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_5 = 0;

#line 289 "ms_deisotope/_c/scoring.pyx"
  while (1) {

#line 289 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_6 = __Pyx_set_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, __pyx_t_4);

#line 289 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_t_6 == 0)) break;

#line 289 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 289, __pyx_L1_error)

#line 289 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_5);

#line 289 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF_SET(__pyx_v_obj, __pyx_t_5);

#line 289 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = 0;

    /* "ms_deisotope/_c/scoring.pyx":290
 *         best_score = -INFINITY
 *         for obj in results:
 *             fit = <IsotopicFitRecord>obj             # <<<<<<<<<<<<<<
 *             if fit.score > best_score:
 *                 best_score = fit.score
 */

#line 290 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = __pyx_v_obj;

#line 290 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_5);

#line 290 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF_SET(__pyx_v_fit, ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_t_5));

#line 290 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = 0;

    /* "ms_deisotope/_c/scoring.pyx":291
 *         for obj in results:
 *             fit = <IsotopicFitRecord>obj
 *             if fit.score > best_score:             # <<<<<<<<<<<<<<
 *                 best_score = fit.score
 *                 best_fit = fit
 */

#line 291 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_7 = ((__pyx_v_fit->score > __pyx_v_best_score) != 0);

#line 291 "ms_deisotope/_c/scoring.pyx"
    if (__pyx_t_7) {

      /* "ms_deisotope/_c/scoring.pyx":292
 *             fit = <IsotopicFitRecord>obj
 *             if fit.score > best_score:
 *                 best_score = fit.score             # <<<<<<<<<<<<<<
 *                 best_fit = fit
 *         return best_fit
 */

#line 292 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_8 = __pyx_v_fit->score;

#line 292 "ms_deisotope/_c/scoring.pyx"
      __pyx_v_best_score = __pyx_t_8;

      /* "ms_deisotope/_c/scoring.pyx":293
 *             if fit.score > best_score:
 *                 best_score = fit.score
 *                 best_fit = fit             # <<<<<<<<<<<<<<
 *         return best_fit
 * 
 */

#line 293 "ms_deisotope/_c/scoring.pyx"
      __Pyx_INCREF(((PyObject *)__pyx_v_fit));

#line 293 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF_SET(__pyx_v_best_fit, __pyx_v_fit);

      /* "ms_deisotope/_c/scoring.pyx":291
 *         for obj in results:
 *             fit = <IsotopicFitRecord>obj
 *             if fit.score > best_score:             # <<<<<<<<<<<<<<
 *                 best_score = fit.score
 *                 best_fit = fit
 */

#line 291 "ms_deisotope/_c/scoring.pyx"
    }

#line 291 "ms_deisotope/_c/scoring.pyx"
  }

#line 291 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/scoring.pyx":294
 *                 best_score = fit.score
 *                 best_fit = fit
 *         return best_fit             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint reject(self, IsotopicFitRecord fit):
 */

#line 294 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 294 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_best_fit));

#line 294 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_v_best_fit;

#line 294 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":282
 *         return max(results, key=operator.attrgetter("score"))
 * 
 *     cdef IsotopicFitRecord _best_from_set(self, set results):             # <<<<<<<<<<<<<<
 *         cdef:
 *             IsotopicFitRecord fit, best_fit
 */

#line 282 "ms_deisotope/_c/scoring.pyx"


#line 282 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 282 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 282 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 282 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 282 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.MaximizeFitSelector._best_from_set", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 282 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 282 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 282 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_fit);

#line 282 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_best_fit);

#line 282 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_v_obj);

#line 282 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 282 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 282 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 282 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":296
 *         return best_fit
 * 
 *     cpdef bint reject(self, IsotopicFitRecord fit):             # <<<<<<<<<<<<<<
 *         """Decide whether the fit should be discarded for having too
 *         small a score. Compares against :attr:`minimum_score`
 */

#line 296 "ms_deisotope/_c/scoring.pyx"


#line 296 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_3reject(PyObject *__pyx_v_self, PyObject *__pyx_v_fit); /*proto*/

#line 296 "ms_deisotope/_c/scoring.pyx"
static int __pyx_f_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_reject(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MaximizeFitSelector *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_fit, int __pyx_skip_dispatch) {

#line 296 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 296 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 296 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 296 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 296 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 296 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 296 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_5;

#line 296 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 296 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 296 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 296 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("reject", 0);

#line 296 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 296 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 296 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 296 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 296 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 296 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 296 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 296 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 296 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 296 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_reject); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 296, __pyx_L1_error)

#line 296 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 296 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_3reject)) {

#line 296 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 296 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 296 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 296 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 296 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 296 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 296 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 296 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 296 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 296 "ms_deisotope/_c/scoring.pyx"
          }

#line 296 "ms_deisotope/_c/scoring.pyx"
        }

#line 296 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_fit)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_fit));

#line 296 "ms_deisotope/_c/scoring.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 296 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 296, __pyx_L1_error)

#line 296 "ms_deisotope/_c/scoring.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 296 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 296 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 296, __pyx_L1_error)

#line 296 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 296 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = __pyx_t_5;

#line 296 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 296 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 296 "ms_deisotope/_c/scoring.pyx"
      }

#line 296 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 296 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 296 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 296 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 296 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 296 "ms_deisotope/_c/scoring.pyx"
      }

#line 296 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 296 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 296 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 296 "ms_deisotope/_c/scoring.pyx"
    }

#line 296 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 296 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":308
 *         bool
 *         """
 *         return fit.score < self.minimum_score             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint reject_score(self, double score):
 */

#line 308 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = (__pyx_v_fit->score < __pyx_v_self->__pyx_base.minimum_score);

#line 308 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":296
 *         return best_fit
 * 
 *     cpdef bint reject(self, IsotopicFitRecord fit):             # <<<<<<<<<<<<<<
 *         """Decide whether the fit should be discarded for having too
 *         small a score. Compares against :attr:`minimum_score`
 */

#line 296 "ms_deisotope/_c/scoring.pyx"


#line 296 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 296 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 296 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 296 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 296 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 296 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 296 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.MaximizeFitSelector.reject", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 296 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 296 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 296 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 296 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 296 "ms_deisotope/_c/scoring.pyx"
}

#line 296 "ms_deisotope/_c/scoring.pyx"


#line 296 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 296 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_3reject(PyObject *__pyx_v_self, PyObject *__pyx_v_fit); /*proto*/

#line 296 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_2reject[] = "MaximizeFitSelector.reject(self, IsotopicFitRecord fit) -> bool\nDecide whether the fit should be discarded for having too\n        small a score. Compares against :attr:`minimum_score`\n\n        Parameters\n        ----------\n        fit : IsotopicFitRecord\n\n        Returns\n        -------\n        bool\n        ";

#line 296 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_3reject(PyObject *__pyx_v_self, PyObject *__pyx_v_fit) {

#line 296 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 296 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 296 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 296 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 296 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 296 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("reject (wrapper)", 0);

#line 296 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_fit), __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord, 1, "fit", 0))) __PYX_ERR(0, 296, __pyx_L1_error)
  __pyx_r = 
#line 296 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_2reject(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_MaximizeFitSelector *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_fit));

#line 296 "ms_deisotope/_c/scoring.pyx"


#line 296 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 296 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 296 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 296 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 296 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 296 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 296 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 296 "ms_deisotope/_c/scoring.pyx"
}

#line 296 "ms_deisotope/_c/scoring.pyx"


#line 296 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_2reject(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MaximizeFitSelector *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_fit) {

#line 296 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 296 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 296 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 296 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 296 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 296 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 296 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("reject", 0);

#line 296 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 296 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_reject(__pyx_v_self, __pyx_v_fit, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 296, __pyx_L1_error)

#line 296 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 296 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 296 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 296 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 296 "ms_deisotope/_c/scoring.pyx"


#line 296 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 296 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 296 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 296 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.MaximizeFitSelector.reject", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 296 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 296 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 296 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 296 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 296 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 296 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":310
 *         return fit.score < self.minimum_score
 * 
 *     cpdef bint reject_score(self, double score):             # <<<<<<<<<<<<<<
 *         return score < self.minimum_score
 * 
 */

#line 310 "ms_deisotope/_c/scoring.pyx"


#line 310 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_5reject_score(PyObject *__pyx_v_self, PyObject *__pyx_arg_score); /*proto*/

#line 310 "ms_deisotope/_c/scoring.pyx"
static int __pyx_f_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_reject_score(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MaximizeFitSelector *__pyx_v_self, double __pyx_v_score, int __pyx_skip_dispatch) {

#line 310 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 310 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 310 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 310 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 310 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 310 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 310 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 310 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_6;

#line 310 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 310 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 310 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 310 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("reject_score", 0);

#line 310 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 310 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 310 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 310 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 310 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 310 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 310 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 310 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 310 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 310 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_reject_score); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 310, __pyx_L1_error)

#line 310 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 310 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_5reject_score)) {

#line 310 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_score); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 310, __pyx_L1_error)

#line 310 "ms_deisotope/_c/scoring.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 310 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
#line 310 "ms_deisotope/_c/scoring.pyx"
__pyx_t_5 = NULL;

#line 310 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {

#line 310 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 310 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_5)) {

#line 310 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 310 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_5);

#line 310 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 310 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_4, function);

#line 310 "ms_deisotope/_c/scoring.pyx"
          }

#line 310 "ms_deisotope/_c/scoring.pyx"
        }

#line 310 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);

#line 310 "ms_deisotope/_c/scoring.pyx"
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 310 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 310 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 310, __pyx_L1_error)

#line 310 "ms_deisotope/_c/scoring.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 310 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 310 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 310, __pyx_L1_error)

#line 310 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 310 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = __pyx_t_6;

#line 310 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 310 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 310 "ms_deisotope/_c/scoring.pyx"
      }

#line 310 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 310 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 310 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 310 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 310 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 310 "ms_deisotope/_c/scoring.pyx"
      }

#line 310 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 310 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 310 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 310 "ms_deisotope/_c/scoring.pyx"
    }

#line 310 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 310 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":311
 * 
 *     cpdef bint reject_score(self, double score):
 *         return score < self.minimum_score             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint is_maximizing(self):
 */

#line 311 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = (__pyx_v_score < __pyx_v_self->__pyx_base.minimum_score);

#line 311 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":310
 *         return fit.score < self.minimum_score
 * 
 *     cpdef bint reject_score(self, double score):             # <<<<<<<<<<<<<<
 *         return score < self.minimum_score
 * 
 */

#line 310 "ms_deisotope/_c/scoring.pyx"


#line 310 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 310 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 310 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 310 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 310 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 310 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 310 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 310 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.MaximizeFitSelector.reject_score", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 310 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 310 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 310 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 310 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 310 "ms_deisotope/_c/scoring.pyx"
}

#line 310 "ms_deisotope/_c/scoring.pyx"


#line 310 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 310 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_5reject_score(PyObject *__pyx_v_self, PyObject *__pyx_arg_score); /*proto*/

#line 310 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_4reject_score[] = "MaximizeFitSelector.reject_score(self, double score) -> bool";

#line 310 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_5reject_score(PyObject *__pyx_v_self, PyObject *__pyx_arg_score) {
  double __pyx_v_score
#line 310 "ms_deisotope/_c/scoring.pyx"
;

#line 310 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 310 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 310 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 310 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 310 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 310 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("reject_score (wrapper)", 0);

#line 310 "ms_deisotope/_c/scoring.pyx"
  assert(__pyx_arg_score); {

#line 310 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_score = __pyx_PyFloat_AsDouble(__pyx_arg_score); if (unlikely((__pyx_v_score == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 310, __pyx_L3_error)

#line 310 "ms_deisotope/_c/scoring.pyx"
  }

#line 310 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 310 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 310 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.MaximizeFitSelector.reject_score", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 310 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 310 "ms_deisotope/_c/scoring.pyx"
  return NULL;

#line 310 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 310 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_4reject_score(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_MaximizeFitSelector *)__pyx_v_self), ((double)__pyx_v_score));

#line 310 "ms_deisotope/_c/scoring.pyx"


#line 310 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 310 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 310 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 310 "ms_deisotope/_c/scoring.pyx"
}

#line 310 "ms_deisotope/_c/scoring.pyx"


#line 310 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_4reject_score(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MaximizeFitSelector *__pyx_v_self, double __pyx_v_score) {

#line 310 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 310 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 310 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 310 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 310 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 310 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 310 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("reject_score", 0);

#line 310 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 310 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_reject_score(__pyx_v_self, __pyx_v_score, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 310, __pyx_L1_error)

#line 310 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 310 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 310 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 310 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 310 "ms_deisotope/_c/scoring.pyx"


#line 310 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 310 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 310 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 310 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.MaximizeFitSelector.reject_score", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 310 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 310 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 310 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 310 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 310 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 310 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":313
 *         return score < self.minimum_score
 * 
 *     cpdef bint is_maximizing(self):             # <<<<<<<<<<<<<<
 *         """Returns that this *is* a maximizing selector
 * 
 */

#line 313 "ms_deisotope/_c/scoring.pyx"


#line 313 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_7is_maximizing(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 313 "ms_deisotope/_c/scoring.pyx"
static int __pyx_f_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_is_maximizing(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_MaximizeFitSelector *__pyx_v_self, int __pyx_skip_dispatch) {

#line 313 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 313 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 313 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 313 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 313 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 313 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 313 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_5;

#line 313 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 313 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 313 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 313 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("is_maximizing", 0);

#line 313 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 313 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 313 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 313 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 313 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 313 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 313 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 313 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 313 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 313 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_maximizing); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 313, __pyx_L1_error)

#line 313 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 313 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_7is_maximizing)) {

#line 313 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 313 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 313 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 313 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 313 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 313 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 313 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 313 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 313 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 313 "ms_deisotope/_c/scoring.pyx"
          }

#line 313 "ms_deisotope/_c/scoring.pyx"
        }

#line 313 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 313 "ms_deisotope/_c/scoring.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 313 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 313, __pyx_L1_error)

#line 313 "ms_deisotope/_c/scoring.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 313 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 313 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 313, __pyx_L1_error)

#line 313 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 313 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = __pyx_t_5;

#line 313 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 313 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 313 "ms_deisotope/_c/scoring.pyx"
      }

#line 313 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 313 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 313 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 313 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 313 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 313 "ms_deisotope/_c/scoring.pyx"
      }

#line 313 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 313 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 313 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 313 "ms_deisotope/_c/scoring.pyx"
    }

#line 313 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 313 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":320
 *         False
 *         """
 *         return True             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 320 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 1;

#line 320 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":313
 *         return score < self.minimum_score
 * 
 *     cpdef bint is_maximizing(self):             # <<<<<<<<<<<<<<
 *         """Returns that this *is* a maximizing selector
 * 
 */

#line 313 "ms_deisotope/_c/scoring.pyx"


#line 313 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 313 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 313 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 313 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 313 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 313 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 313 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.MaximizeFitSelector.is_maximizing", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 313 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 313 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 313 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 313 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 313 "ms_deisotope/_c/scoring.pyx"
}

#line 313 "ms_deisotope/_c/scoring.pyx"


#line 313 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 313 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_7is_maximizing(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 313 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_6is_maximizing[] = "MaximizeFitSelector.is_maximizing(self) -> bool\nReturns that this *is* a maximizing selector\n\n        Returns\n        -------\n        False\n        ";

#line 313 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_7is_maximizing(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 313 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 313 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 313 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("is_maximizing (wrapper)", 0);
  __pyx_r = 
#line 313 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_6is_maximizing(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_MaximizeFitSelector *)__pyx_v_self));

#line 313 "ms_deisotope/_c/scoring.pyx"


#line 313 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 313 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 313 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 313 "ms_deisotope/_c/scoring.pyx"
}

#line 313 "ms_deisotope/_c/scoring.pyx"


#line 313 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_6is_maximizing(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MaximizeFitSelector *__pyx_v_self) {

#line 313 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 313 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 313 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 313 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 313 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 313 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 313 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("is_maximizing", 0);

#line 313 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 313 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_is_maximizing(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 313, __pyx_L1_error)

#line 313 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 313 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 313 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 313 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 313 "ms_deisotope/_c/scoring.pyx"


#line 313 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 313 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 313 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 313 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.MaximizeFitSelector.is_maximizing", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 313 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 313 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 313 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 313 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 313 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 313 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":335
 *     """
 * 
 *     def __init__(self, score_threshold=0.5):             # <<<<<<<<<<<<<<
 *         self.select = MinimizeFitSelector(score_threshold)
 * 
 */

#line 335 "ms_deisotope/_c/scoring.pyx"


#line 335 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 335 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 335 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_score_threshold = 0
#line 335 "ms_deisotope/_c/scoring.pyx"
;

#line 335 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 335 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 335 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 335 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 335 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 335 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 335 "ms_deisotope/_c/scoring.pyx"
  {

#line 335 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_score_threshold,0};

#line 335 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[1] = {0};

#line 335 "ms_deisotope/_c/scoring.pyx"
    values[0] = ((PyObject *)__pyx_float_0_5);

#line 335 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 335 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 335 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 335 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  1: 
#line 335 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 335 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 335 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 335 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 335 "ms_deisotope/_c/scoring.pyx"
      }

#line 335 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 335 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 335 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 335 "ms_deisotope/_c/scoring.pyx"
        if (kw_args > 0) {

#line 335 "ms_deisotope/_c/scoring.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_score_threshold);

#line 335 "ms_deisotope/_c/scoring.pyx"
          if (value) { values[0] = value; kw_args--; }

#line 335 "ms_deisotope/_c/scoring.pyx"
        }

#line 335 "ms_deisotope/_c/scoring.pyx"
      }

#line 335 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 335 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 335, __pyx_L3_error)

#line 335 "ms_deisotope/_c/scoring.pyx"
      }

#line 335 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 335 "ms_deisotope/_c/scoring.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: 
#line 335 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 335 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  0: 
#line 335 "ms_deisotope/_c/scoring.pyx"
break;
        default: 
#line 335 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 335 "ms_deisotope/_c/scoring.pyx"
      }

#line 335 "ms_deisotope/_c/scoring.pyx"
    }

#line 335 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_score_threshold = values[0];

#line 335 "ms_deisotope/_c/scoring.pyx"
  }

#line 335 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 335 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); 
#line 335 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 335, __pyx_L3_error)

#line 335 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 335 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitterBase.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 335 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 335 "ms_deisotope/_c/scoring.pyx"
  return -1;

#line 335 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 335 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase___init__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self), __pyx_v_score_threshold);

#line 335 "ms_deisotope/_c/scoring.pyx"


#line 335 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 335 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 335 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 335 "ms_deisotope/_c/scoring.pyx"
}

#line 335 "ms_deisotope/_c/scoring.pyx"


#line 335 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase___init__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, PyObject *__pyx_v_score_threshold) {

#line 335 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 335 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 335 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 335 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 335 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 335 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 335 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_deisotope/_c/scoring.pyx":336
 * 
 *     def __init__(self, score_threshold=0.5):
 *         self.select = MinimizeFitSelector(score_threshold)             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */

#line 336 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_7scoring_MinimizeFitSelector), __pyx_v_score_threshold); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 336, __pyx_L1_error)

#line 336 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 336 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 336 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->select);

#line 336 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->select));

#line 336 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->select = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_t_1);

#line 336 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/scoring.pyx":335
 *     """
 * 
 *     def __init__(self, score_threshold=0.5):             # <<<<<<<<<<<<<<
 *         self.select = MinimizeFitSelector(score_threshold)
 * 
 */

#line 335 "ms_deisotope/_c/scoring.pyx"


#line 335 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 335 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 335 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 335 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 335 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 335 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitterBase.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 335 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = -1;

#line 335 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 335 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 335 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 335 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":338
 *         self.select = MinimizeFitSelector(score_threshold)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (0,), self.__getstate__()
 * 
 */

#line 338 "ms_deisotope/_c/scoring.pyx"


#line 338 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 338 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_3__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 338 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_2__reduce__[] = "IsotopicFitterBase.__reduce__(self)";

#line 338 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_3__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 338 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 338 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 338 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = 
#line 338 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_2__reduce__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self));

#line 338 "ms_deisotope/_c/scoring.pyx"


#line 338 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 338 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 338 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 338 "ms_deisotope/_c/scoring.pyx"
}

#line 338 "ms_deisotope/_c/scoring.pyx"


#line 338 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_2__reduce__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self) {

#line 338 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 338 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 338 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 338 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 338 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 338 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 338 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 338 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 338 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 338 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "ms_deisotope/_c/scoring.pyx":339
 * 
 *     def __reduce__(self):
 *         return self.__class__, (0,), self.__getstate__()             # <<<<<<<<<<<<<<
 * 
 *     def __getstate__(self):
 */

#line 339 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 339 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 339, __pyx_L1_error)

#line 339 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 339 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_getstate); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 339, __pyx_L1_error)

#line 339 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 339 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_4 = NULL;

#line 339 "ms_deisotope/_c/scoring.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {

#line 339 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 339 "ms_deisotope/_c/scoring.pyx"
    if (likely(__pyx_t_4)) {

#line 339 "ms_deisotope/_c/scoring.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 339 "ms_deisotope/_c/scoring.pyx"
      __Pyx_INCREF(__pyx_t_4);

#line 339 "ms_deisotope/_c/scoring.pyx"
      __Pyx_INCREF(function);

#line 339 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF_SET(__pyx_t_3, function);

#line 339 "ms_deisotope/_c/scoring.pyx"
    }

#line 339 "ms_deisotope/_c/scoring.pyx"
  }

#line 339 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 339 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 339 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 339, __pyx_L1_error)

#line 339 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 339 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 339 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 339, __pyx_L1_error)

#line 339 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 339 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 339 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);

#line 339 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_tuple_);

#line 339 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_tuple_);

#line 339 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_tuple_);

#line 339 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 339 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);

#line 339 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 339 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = 0;

#line 339 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_3;

#line 339 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = 0;

#line 339 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":338
 *         self.select = MinimizeFitSelector(score_threshold)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (0,), self.__getstate__()
 * 
 */

#line 338 "ms_deisotope/_c/scoring.pyx"


#line 338 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 338 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 338 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 338 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 338 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 338 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 338 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitterBase.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 338 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 338 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 338 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 338 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 338 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 338 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":341
 *         return self.__class__, (0,), self.__getstate__()
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return (self.select,)
 * 
 */

#line 341 "ms_deisotope/_c/scoring.pyx"


#line 341 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 341 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_5__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 341 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_4__getstate__[] = "IsotopicFitterBase.__getstate__(self)";

#line 341 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_5__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 341 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 341 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 341 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = 
#line 341 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_4__getstate__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self));

#line 341 "ms_deisotope/_c/scoring.pyx"


#line 341 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 341 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 341 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 341 "ms_deisotope/_c/scoring.pyx"
}

#line 341 "ms_deisotope/_c/scoring.pyx"


#line 341 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_4__getstate__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self) {

#line 341 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 341 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 341 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 341 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 341 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 341 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 341 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "ms_deisotope/_c/scoring.pyx":342
 * 
 *     def __getstate__(self):
 *         return (self.select,)             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, state):
 */

#line 342 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 342 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 342, __pyx_L1_error)

#line 342 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 342 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->select));

#line 342 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->select));

#line 342 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->select));

#line 342 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 342 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 342 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":341
 *         return self.__class__, (0,), self.__getstate__()
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return (self.select,)
 * 
 */

#line 341 "ms_deisotope/_c/scoring.pyx"


#line 341 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 341 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 341 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 341 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitterBase.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 341 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 341 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 341 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 341 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 341 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 341 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":344
 *         return (self.select,)
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.select, = state
 * 
 */

#line 344 "ms_deisotope/_c/scoring.pyx"


#line 344 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 344 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_7__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state); /*proto*/

#line 344 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_6__setstate__[] = "IsotopicFitterBase.__setstate__(self, state)";

#line 344 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_7__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state) {

#line 344 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 344 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 344 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  __pyx_r = 
#line 344 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_6__setstate__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self), ((PyObject *)__pyx_v_state));

#line 344 "ms_deisotope/_c/scoring.pyx"


#line 344 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 344 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 344 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 344 "ms_deisotope/_c/scoring.pyx"
}

#line 344 "ms_deisotope/_c/scoring.pyx"


#line 344 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_6__setstate__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, PyObject *__pyx_v_state) {

#line 344 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 344 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 344 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 344 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 344 "ms_deisotope/_c/scoring.pyx"
  PyObject *(*__pyx_t_3)(PyObject *);

#line 344 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 344 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 344 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 344 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* "ms_deisotope/_c/scoring.pyx":345
 * 
 *     def __setstate__(self, state):
 *         self.select, = state             # <<<<<<<<<<<<<<
 * 
 *     def evaluate(self, PeakSet peaklist, list observed, list expected, **kwargs):
 */

#line 345 "ms_deisotope/_c/scoring.pyx"
  if ((likely(PyTuple_CheckExact(__pyx_v_state))) || (PyList_CheckExact(__pyx_v_state))) {

#line 345 "ms_deisotope/_c/scoring.pyx"
    PyObject* sequence = __pyx_v_state;

#line 345 "ms_deisotope/_c/scoring.pyx"
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);

#line 345 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(size != 1)) {

#line 345 "ms_deisotope/_c/scoring.pyx"
      if (size > 1) __Pyx_RaiseTooManyValuesError(1);

#line 345 "ms_deisotope/_c/scoring.pyx"
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);

#line 345 "ms_deisotope/_c/scoring.pyx"
      __PYX_ERR(0, 345, __pyx_L1_error)

#line 345 "ms_deisotope/_c/scoring.pyx"
    }

#line 345 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 345 "ms_deisotope/_c/scoring.pyx"
    if (likely(PyTuple_CheckExact(sequence))) {

#line 345 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 

#line 345 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 345 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 

#line 345 "ms_deisotope/_c/scoring.pyx"
    }

#line 345 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 345 "ms_deisotope/_c/scoring.pyx"
    #else

#line 345 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 345, __pyx_L1_error)

#line 345 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 345 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 345 "ms_deisotope/_c/scoring.pyx"
  } else {

#line 345 "ms_deisotope/_c/scoring.pyx"
    Py_ssize_t index = -1;

#line 345 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_2 = PyObject_GetIter(__pyx_v_state); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 345, __pyx_L1_error)

#line 345 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 345 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_3 = Py_TYPE(__pyx_t_2)->tp_iternext;
    index = 0; __pyx_t_1 = __pyx_t_3(__pyx_t_2); if (unlikely(!__pyx_t_1)) 
#line 345 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L3_unpacking_failed;

#line 345 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 345 "ms_deisotope/_c/scoring.pyx"
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_3(__pyx_t_2), 1) < 0) __PYX_ERR(0, 345, __pyx_L1_error)

#line 345 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_3 = NULL;

#line 345 "ms_deisotope/_c/scoring.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 345 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L4_unpacking_done;

#line 345 "ms_deisotope/_c/scoring.pyx"
    __pyx_L3_unpacking_failed:;

#line 345 "ms_deisotope/_c/scoring.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 345 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_3 = NULL;

#line 345 "ms_deisotope/_c/scoring.pyx"
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);

#line 345 "ms_deisotope/_c/scoring.pyx"
    __PYX_ERR(0, 345, __pyx_L1_error)

#line 345 "ms_deisotope/_c/scoring.pyx"
    __pyx_L4_unpacking_done:;

#line 345 "ms_deisotope/_c/scoring.pyx"
  }

#line 345 "ms_deisotope/_c/scoring.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase))))) __PYX_ERR(0, 345, __pyx_L1_error)

#line 345 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 345 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->select);

#line 345 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->select));

#line 345 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->select = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_t_1);

#line 345 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/scoring.pyx":344
 *         return (self.select,)
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.select, = state
 * 
 */

#line 344 "ms_deisotope/_c/scoring.pyx"


#line 344 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 344 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 344 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 344 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 344 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 344 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 344 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitterBase.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 344 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 344 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 344 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 344 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 344 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 344 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":347
 *         self.select, = state
 * 
 *     def evaluate(self, PeakSet peaklist, list observed, list expected, **kwargs):             # <<<<<<<<<<<<<<
 *         """Evaluate a pair of peak lists for goodness-of-fit.
 * 
 */

#line 347 "ms_deisotope/_c/scoring.pyx"


#line 347 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 347 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_9evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 347 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_8evaluate[] = "IsotopicFitterBase.evaluate(self, PeakSet peaklist, list observed, list expected, **kwargs)\nEvaluate a pair of peak lists for goodness-of-fit.\n\n        Parameters\n        ----------\n        peaklist : :class:`~.PeakSet`\n            The full set of all experimental peaks\n        observed : list\n            The list of experimental peaks that are part of this fit\n        expected : list\n            The list of theoretical peaks that are part of this fit\n        **kwargs\n\n        Returns\n        -------\n        float\n            The score\n        ";

#line 347 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_9evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist = 0
#line 347 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_observed = 0
#line 347 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_expected = 0
#line 347 "ms_deisotope/_c/scoring.pyx"
;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0
#line 347 "ms_deisotope/_c/scoring.pyx"
;

#line 347 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 347 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 347 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 347 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 347 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 347 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("evaluate (wrapper)", 0);

#line 347 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;

#line 347 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_kwargs);

#line 347 "ms_deisotope/_c/scoring.pyx"
  {

#line 347 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peaklist,&__pyx_n_s_observed,&__pyx_n_s_expected,0};

#line 347 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[3] = {0,0,0};

#line 347 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 347 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 347 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 347 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  3: 
#line 347 "ms_deisotope/_c/scoring.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 347 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 347 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 347 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 347 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 347 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 347 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 347 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 347 "ms_deisotope/_c/scoring.pyx"
      }

#line 347 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 347 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 347 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 347 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaklist)) != 0)) kw_args--;
        else 
#line 347 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 347 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 347 "ms_deisotope/_c/scoring.pyx"
        case  1:

#line 347 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_observed)) != 0)) kw_args--;

#line 347 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("evaluate", 1, 3, 3, 1); 
#line 347 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 347, __pyx_L3_error)

#line 347 "ms_deisotope/_c/scoring.pyx"
        }

#line 347 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 347 "ms_deisotope/_c/scoring.pyx"
        case  2:

#line 347 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_expected)) != 0)) kw_args--;

#line 347 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("evaluate", 1, 3, 3, 2); 
#line 347 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 347, __pyx_L3_error)

#line 347 "ms_deisotope/_c/scoring.pyx"
        }

#line 347 "ms_deisotope/_c/scoring.pyx"
      }

#line 347 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 347 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, pos_args, "evaluate") < 0)) __PYX_ERR(0, 347, __pyx_L3_error)

#line 347 "ms_deisotope/_c/scoring.pyx"
      }

#line 347 "ms_deisotope/_c/scoring.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 347 "ms_deisotope/_c/scoring.pyx"
      goto __pyx_L5_argtuple_error;

#line 347 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 347 "ms_deisotope/_c/scoring.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 347 "ms_deisotope/_c/scoring.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 347 "ms_deisotope/_c/scoring.pyx"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 347 "ms_deisotope/_c/scoring.pyx"
    }

#line 347 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_peaklist = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[0]);

#line 347 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_observed = ((PyObject*)values[1]);

#line 347 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_expected = ((PyObject*)values[2]);

#line 347 "ms_deisotope/_c/scoring.pyx"
  }

#line 347 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 347 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("evaluate", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 347 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 347, __pyx_L3_error)

#line 347 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 347 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;

#line 347 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitterBase.evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 347 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 347 "ms_deisotope/_c/scoring.pyx"
  return NULL;

#line 347 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 347 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peaklist), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peaklist", 0))) __PYX_ERR(0, 347, __pyx_L1_error)

#line 347 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_observed), (&PyList_Type), 1, "observed", 1))) __PYX_ERR(0, 347, __pyx_L1_error)

#line 347 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_expected), (&PyList_Type), 1, "expected", 1))) __PYX_ERR(0, 347, __pyx_L1_error)
  __pyx_r = 
#line 347 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_8evaluate(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self), __pyx_v_peaklist, __pyx_v_observed, __pyx_v_expected, __pyx_v_kwargs);

#line 347 "ms_deisotope/_c/scoring.pyx"


#line 347 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 347 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 347 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 347 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 347 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 347 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_v_kwargs);

#line 347 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 347 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 347 "ms_deisotope/_c/scoring.pyx"
}

#line 347 "ms_deisotope/_c/scoring.pyx"


#line 347 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_8evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {

#line 347 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 347 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 347 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 347 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 347 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 347 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 347 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("evaluate", 0);

  /* "ms_deisotope/_c/scoring.pyx":365
 *             The score
 *         """
 *         return self._evaluate(peaklist, observed, expected)             # <<<<<<<<<<<<<<
 * 
 *     cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):
 */

#line 365 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 365 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyFloat_FromDouble(((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self->__pyx_vtab)->_evaluate(__pyx_v_self, __pyx_v_peaklist, __pyx_v_observed, __pyx_v_expected, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 365, __pyx_L1_error)

#line 365 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 365 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 365 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 365 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":347
 *         self.select, = state
 * 
 *     def evaluate(self, PeakSet peaklist, list observed, list expected, **kwargs):             # <<<<<<<<<<<<<<
 *         """Evaluate a pair of peak lists for goodness-of-fit.
 * 
 */

#line 347 "ms_deisotope/_c/scoring.pyx"


#line 347 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 347 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 347 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 347 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitterBase.evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 347 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 347 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 347 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 347 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 347 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 347 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":367
 *         return self._evaluate(peaklist, observed, expected)
 * 
 *     cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

#line 367 "ms_deisotope/_c/scoring.pyx"


#line 367 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_11_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 367 "ms_deisotope/_c/scoring.pyx"
static double __pyx_f_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase__evaluate(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, CYTHON_UNUSED PyObject *__pyx_v_observed, CYTHON_UNUSED PyObject *__pyx_v_expected, int __pyx_skip_dispatch) {

#line 367 "ms_deisotope/_c/scoring.pyx"
  double __pyx_r;

#line 367 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 367 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 367 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 367 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 367 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 367 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_5;

#line 367 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 367 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_7;

#line 367 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 367 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 367 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 367 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate", 0);

#line 367 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 367 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 367 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 367 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 367 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 367 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 367 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 367 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 367 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 367 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_evaluate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 367, __pyx_L1_error)

#line 367 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 367 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_11_evaluate)) {

#line 367 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 367 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 367 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_5 = 0;

#line 367 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 367 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 367 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 367 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 367 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 367 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 367 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 367 "ms_deisotope/_c/scoring.pyx"
            __pyx_t_5 = 1;

#line 367 "ms_deisotope/_c/scoring.pyx"
          }

#line 367 "ms_deisotope/_c/scoring.pyx"
        }

#line 367 "ms_deisotope/_c/scoring.pyx"
        #if CYTHON_FAST_PYCALL

#line 367 "ms_deisotope/_c/scoring.pyx"
        if (PyFunction_Check(__pyx_t_3)) {

#line 367 "ms_deisotope/_c/scoring.pyx"
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_peaklist), __pyx_v_observed, __pyx_v_expected};

#line 367 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 367, __pyx_L1_error)

#line 367 "ms_deisotope/_c/scoring.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 367 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 367 "ms_deisotope/_c/scoring.pyx"
        } else

#line 367 "ms_deisotope/_c/scoring.pyx"
        #endif

#line 367 "ms_deisotope/_c/scoring.pyx"
        #if CYTHON_FAST_PYCCALL

#line 367 "ms_deisotope/_c/scoring.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {

#line 367 "ms_deisotope/_c/scoring.pyx"
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_peaklist), __pyx_v_observed, __pyx_v_expected};

#line 367 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 367, __pyx_L1_error)

#line 367 "ms_deisotope/_c/scoring.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 367 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 367 "ms_deisotope/_c/scoring.pyx"
        } else

#line 367 "ms_deisotope/_c/scoring.pyx"
        #endif

#line 367 "ms_deisotope/_c/scoring.pyx"
        {

#line 367 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_6 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 367, __pyx_L1_error)

#line 367 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_6);

#line 367 "ms_deisotope/_c/scoring.pyx"
          if (__pyx_t_4) {

#line 367 "ms_deisotope/_c/scoring.pyx"
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;

#line 367 "ms_deisotope/_c/scoring.pyx"
          }

#line 367 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(((PyObject *)__pyx_v_peaklist));

#line 367 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(((PyObject *)__pyx_v_peaklist));

#line 367 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_peaklist));

#line 367 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(__pyx_v_observed);

#line 367 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(__pyx_v_observed);

#line 367 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_observed);

#line 367 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(__pyx_v_expected);

#line 367 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(__pyx_v_expected);

#line 367 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, __pyx_v_expected);

#line 367 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 367, __pyx_L1_error)

#line 367 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 367 "ms_deisotope/_c/scoring.pyx"
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 367 "ms_deisotope/_c/scoring.pyx"
        }

#line 367 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 367 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_7 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_7 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 367, __pyx_L1_error)

#line 367 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 367 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = __pyx_t_7;

#line 367 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 367 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 367 "ms_deisotope/_c/scoring.pyx"
      }

#line 367 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 367 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 367 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 367 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 367 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 367 "ms_deisotope/_c/scoring.pyx"
      }

#line 367 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 367 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 367 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 367 "ms_deisotope/_c/scoring.pyx"
    }

#line 367 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 367 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":368
 * 
 *     cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):
 *         raise NotImplementedError()             # <<<<<<<<<<<<<<
 * 
 *     def __call__(self, *args, **kwargs):
 */

#line 368 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_NotImplementedError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 368, __pyx_L1_error)

#line 368 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 368 "ms_deisotope/_c/scoring.pyx"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 368 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 368 "ms_deisotope/_c/scoring.pyx"
  __PYX_ERR(0, 368, __pyx_L1_error)

  /* "ms_deisotope/_c/scoring.pyx":367
 *         return self._evaluate(peaklist, observed, expected)
 * 
 *     cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

#line 367 "ms_deisotope/_c/scoring.pyx"


#line 367 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 367 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 367 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 367 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 367 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 367 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 367 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 367 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.IsotopicFitterBase._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 367 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 367 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 367 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 367 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 367 "ms_deisotope/_c/scoring.pyx"
}

#line 367 "ms_deisotope/_c/scoring.pyx"


#line 367 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 367 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_11_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 367 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_10_evaluate[] = "IsotopicFitterBase._evaluate(self, PeakSet peaklist, list observed, list expected) -> double";

#line 367 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_11_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist = 0
#line 367 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_observed = 0
#line 367 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_expected = 0
#line 367 "ms_deisotope/_c/scoring.pyx"
;

#line 367 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 367 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 367 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 367 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 367 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 367 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate (wrapper)", 0);

#line 367 "ms_deisotope/_c/scoring.pyx"
  {

#line 367 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peaklist,&__pyx_n_s_observed,&__pyx_n_s_expected,0};

#line 367 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[3] = {0,0,0};

#line 367 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 367 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 367 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 367 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  3: 
#line 367 "ms_deisotope/_c/scoring.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 367 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 367 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 367 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 367 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 367 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 367 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 367 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 367 "ms_deisotope/_c/scoring.pyx"
      }

#line 367 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 367 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 367 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 367 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaklist)) != 0)) kw_args--;
        else 
#line 367 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 367 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 367 "ms_deisotope/_c/scoring.pyx"
        case  1:

#line 367 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_observed)) != 0)) kw_args--;

#line 367 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, 1); 
#line 367 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 367, __pyx_L3_error)

#line 367 "ms_deisotope/_c/scoring.pyx"
        }

#line 367 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 367 "ms_deisotope/_c/scoring.pyx"
        case  2:

#line 367 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_expected)) != 0)) kw_args--;

#line 367 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, 2); 
#line 367 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 367, __pyx_L3_error)

#line 367 "ms_deisotope/_c/scoring.pyx"
        }

#line 367 "ms_deisotope/_c/scoring.pyx"
      }

#line 367 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 367 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_evaluate") < 0)) __PYX_ERR(0, 367, __pyx_L3_error)

#line 367 "ms_deisotope/_c/scoring.pyx"
      }

#line 367 "ms_deisotope/_c/scoring.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 367 "ms_deisotope/_c/scoring.pyx"
      goto __pyx_L5_argtuple_error;

#line 367 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 367 "ms_deisotope/_c/scoring.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 367 "ms_deisotope/_c/scoring.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 367 "ms_deisotope/_c/scoring.pyx"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 367 "ms_deisotope/_c/scoring.pyx"
    }

#line 367 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_peaklist = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[0]);

#line 367 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_observed = ((PyObject*)values[1]);

#line 367 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_expected = ((PyObject*)values[2]);

#line 367 "ms_deisotope/_c/scoring.pyx"
  }

#line 367 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 367 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 367 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 367, __pyx_L3_error)

#line 367 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 367 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitterBase._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 367 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 367 "ms_deisotope/_c/scoring.pyx"
  return NULL;

#line 367 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 367 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peaklist), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peaklist", 0))) __PYX_ERR(0, 367, __pyx_L1_error)

#line 367 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_observed), (&PyList_Type), 1, "observed", 1))) __PYX_ERR(0, 367, __pyx_L1_error)

#line 367 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_expected), (&PyList_Type), 1, "expected", 1))) __PYX_ERR(0, 367, __pyx_L1_error)
  __pyx_r = 
#line 367 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_10_evaluate(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self), __pyx_v_peaklist, __pyx_v_observed, __pyx_v_expected);

#line 367 "ms_deisotope/_c/scoring.pyx"


#line 367 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 367 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 367 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 367 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 367 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 367 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 367 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 367 "ms_deisotope/_c/scoring.pyx"
}

#line 367 "ms_deisotope/_c/scoring.pyx"


#line 367 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_10_evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected) {

#line 367 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 367 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 367 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 367 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 367 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 367 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 367 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate", 0);

#line 367 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 367 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase__evaluate(__pyx_v_self, __pyx_v_peaklist, __pyx_v_observed, __pyx_v_expected, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 367, __pyx_L1_error)

#line 367 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 367 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 367 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 367 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 367 "ms_deisotope/_c/scoring.pyx"


#line 367 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 367 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 367 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 367 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitterBase._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 367 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 367 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 367 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 367 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 367 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 367 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":370
 *         raise NotImplementedError()
 * 
 *     def __call__(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         """Invokes :meth:`evaluate`
 * 
 */

#line 370 "ms_deisotope/_c/scoring.pyx"


#line 370 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 370 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_13__call__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 370 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_12__call__[] = "Invokes :meth:`evaluate`\n\n        Parameters\n        ----------\n        *args\n            Forwarded to :meth:`evaluate`\n        **kwargs\n            Forwarded to :meth:`evaluate`\n\n        Returns\n        -------\n        float\n            The score\n        ";

#line 370 "ms_deisotope/_c/scoring.pyx"
#if CYTHON_UPDATE_DESCRIPTOR_DOC

#line 370 "ms_deisotope/_c/scoring.pyx"
struct wrapperbase __pyx_wrapperbase_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_12__call__;

#line 370 "ms_deisotope/_c/scoring.pyx"
#endif

#line 370 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_13__call__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0
#line 370 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_kwargs = 0
#line 370 "ms_deisotope/_c/scoring.pyx"
;

#line 370 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 370 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 370 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__call__ (wrapper)", 0);

#line 370 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__call__", 1))) return NULL;
  __pyx_v_kwargs = (__pyx_kwds) ? PyDict_Copy(__pyx_kwds) : PyDict_New(); 
#line 370 "ms_deisotope/_c/scoring.pyx"
if (unlikely(!__pyx_v_kwargs)) return NULL;

#line 370 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_kwargs);

#line 370 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_args);

#line 370 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_args = __pyx_args;
  __pyx_r = 
#line 370 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_12__call__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

#line 370 "ms_deisotope/_c/scoring.pyx"


#line 370 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 370 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_v_args);

#line 370 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_v_kwargs);

#line 370 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 370 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 370 "ms_deisotope/_c/scoring.pyx"
}

#line 370 "ms_deisotope/_c/scoring.pyx"


#line 370 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_12__call__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {

#line 370 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 370 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 370 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 370 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 370 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 370 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 370 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 370 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 370 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__call__", 0);

  /* "ms_deisotope/_c/scoring.pyx":385
 *             The score
 *         """
 *         return self.evaluate(*args, **kwargs)             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint reject(self, IsotopicFitRecord fit):
 */

#line 385 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 385 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_evaluate_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 385, __pyx_L1_error)

#line 385 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 385 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 385, __pyx_L1_error)

#line 385 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 385 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_v_args, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 385, __pyx_L1_error)

#line 385 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 385 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 385 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 385 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_3;

#line 385 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = 0;

#line 385 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":370
 *         raise NotImplementedError()
 * 
 *     def __call__(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         """Invokes :meth:`evaluate`
 * 
 */

#line 370 "ms_deisotope/_c/scoring.pyx"


#line 370 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 370 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 370 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 370 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 370 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 370 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitterBase.__call__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 370 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 370 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 370 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 370 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 370 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 370 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":387
 *         return self.evaluate(*args, **kwargs)
 * 
 *     cpdef bint reject(self, IsotopicFitRecord fit):             # <<<<<<<<<<<<<<
 *         """Test whether this fit is too poor to be used
 * 
 */

#line 387 "ms_deisotope/_c/scoring.pyx"


#line 387 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_15reject(PyObject *__pyx_v_self, PyObject *__pyx_v_fit); /*proto*/

#line 387 "ms_deisotope/_c/scoring.pyx"
static int __pyx_f_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_reject(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_fit, int __pyx_skip_dispatch) {

#line 387 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 387 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 387 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 387 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 387 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 387 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 387 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_5;

#line 387 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 387 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 387 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 387 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("reject", 0);

#line 387 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 387 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 387 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 387 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 387 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 387 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 387 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 387 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 387 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 387 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_reject); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 387, __pyx_L1_error)

#line 387 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 387 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_15reject)) {

#line 387 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 387 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 387 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 387 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 387 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 387 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 387 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 387 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 387 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 387 "ms_deisotope/_c/scoring.pyx"
          }

#line 387 "ms_deisotope/_c/scoring.pyx"
        }

#line 387 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_fit)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_fit));

#line 387 "ms_deisotope/_c/scoring.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 387 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 387, __pyx_L1_error)

#line 387 "ms_deisotope/_c/scoring.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 387 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 387 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 387, __pyx_L1_error)

#line 387 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 387 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = __pyx_t_5;

#line 387 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 387 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 387 "ms_deisotope/_c/scoring.pyx"
      }

#line 387 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 387 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 387 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 387 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 387 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 387 "ms_deisotope/_c/scoring.pyx"
      }

#line 387 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 387 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 387 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 387 "ms_deisotope/_c/scoring.pyx"
    }

#line 387 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 387 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":399
 *         bool
 *         """
 *         return self.select.reject(fit)             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint reject_score(self, double score):
 */

#line 399 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_v_self->select->__pyx_vtab)->reject(__pyx_v_self->select, __pyx_v_fit, 0);

#line 399 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":387
 *         return self.evaluate(*args, **kwargs)
 * 
 *     cpdef bint reject(self, IsotopicFitRecord fit):             # <<<<<<<<<<<<<<
 *         """Test whether this fit is too poor to be used
 * 
 */

#line 387 "ms_deisotope/_c/scoring.pyx"


#line 387 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 387 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 387 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 387 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 387 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 387 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 387 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.IsotopicFitterBase.reject", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 387 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 387 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 387 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 387 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 387 "ms_deisotope/_c/scoring.pyx"
}

#line 387 "ms_deisotope/_c/scoring.pyx"


#line 387 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 387 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_15reject(PyObject *__pyx_v_self, PyObject *__pyx_v_fit); /*proto*/

#line 387 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_14reject[] = "IsotopicFitterBase.reject(self, IsotopicFitRecord fit) -> bool\nTest whether this fit is too poor to be used\n\n        Parameters\n        ----------\n        fit : :class:`~IsotopicFitRecord`\n            The fit to test\n\n        Returns\n        -------\n        bool\n        ";

#line 387 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_15reject(PyObject *__pyx_v_self, PyObject *__pyx_v_fit) {

#line 387 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 387 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 387 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 387 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 387 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 387 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("reject (wrapper)", 0);

#line 387 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_fit), __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord, 1, "fit", 0))) __PYX_ERR(0, 387, __pyx_L1_error)
  __pyx_r = 
#line 387 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_14reject(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)__pyx_v_fit));

#line 387 "ms_deisotope/_c/scoring.pyx"


#line 387 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 387 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 387 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 387 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 387 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 387 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 387 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 387 "ms_deisotope/_c/scoring.pyx"
}

#line 387 "ms_deisotope/_c/scoring.pyx"


#line 387 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_14reject(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_v_fit) {

#line 387 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 387 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 387 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 387 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 387 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 387 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 387 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("reject", 0);

#line 387 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 387 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_reject(__pyx_v_self, __pyx_v_fit, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 387, __pyx_L1_error)

#line 387 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 387 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 387 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 387 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 387 "ms_deisotope/_c/scoring.pyx"


#line 387 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 387 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 387 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 387 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitterBase.reject", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 387 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 387 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 387 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 387 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 387 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 387 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":401
 *         return self.select.reject(fit)
 * 
 *     cpdef bint reject_score(self, double score):             # <<<<<<<<<<<<<<
 *         return self.select.reject_score(score)
 * 
 */

#line 401 "ms_deisotope/_c/scoring.pyx"


#line 401 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_17reject_score(PyObject *__pyx_v_self, PyObject *__pyx_arg_score); /*proto*/

#line 401 "ms_deisotope/_c/scoring.pyx"
static int __pyx_f_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_reject_score(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, double __pyx_v_score, int __pyx_skip_dispatch) {

#line 401 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 401 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 401 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 401 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 401 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 401 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 401 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 401 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_6;

#line 401 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 401 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 401 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 401 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("reject_score", 0);

#line 401 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 401 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 401 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 401 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 401 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 401 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 401 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 401 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 401 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 401 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_reject_score); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 401, __pyx_L1_error)

#line 401 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 401 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_17reject_score)) {

#line 401 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_score); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 401, __pyx_L1_error)

#line 401 "ms_deisotope/_c/scoring.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 401 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
#line 401 "ms_deisotope/_c/scoring.pyx"
__pyx_t_5 = NULL;

#line 401 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {

#line 401 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 401 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_5)) {

#line 401 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 401 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_5);

#line 401 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 401 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_4, function);

#line 401 "ms_deisotope/_c/scoring.pyx"
          }

#line 401 "ms_deisotope/_c/scoring.pyx"
        }

#line 401 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);

#line 401 "ms_deisotope/_c/scoring.pyx"
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 401 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 401 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 401, __pyx_L1_error)

#line 401 "ms_deisotope/_c/scoring.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 401 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 401 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 401, __pyx_L1_error)

#line 401 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 401 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = __pyx_t_6;

#line 401 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 401 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 401 "ms_deisotope/_c/scoring.pyx"
      }

#line 401 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 401 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 401 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 401 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 401 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 401 "ms_deisotope/_c/scoring.pyx"
      }

#line 401 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 401 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 401 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 401 "ms_deisotope/_c/scoring.pyx"
    }

#line 401 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 401 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":402
 * 
 *     cpdef bint reject_score(self, double score):
 *         return self.select.reject_score(score)             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint is_maximizing(self):
 */

#line 402 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_v_self->select->__pyx_vtab)->reject_score(__pyx_v_self->select, __pyx_v_score, 0);

#line 402 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":401
 *         return self.select.reject(fit)
 * 
 *     cpdef bint reject_score(self, double score):             # <<<<<<<<<<<<<<
 *         return self.select.reject_score(score)
 * 
 */

#line 401 "ms_deisotope/_c/scoring.pyx"


#line 401 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 401 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 401 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 401 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 401 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 401 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 401 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 401 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.IsotopicFitterBase.reject_score", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 401 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 401 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 401 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 401 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 401 "ms_deisotope/_c/scoring.pyx"
}

#line 401 "ms_deisotope/_c/scoring.pyx"


#line 401 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 401 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_17reject_score(PyObject *__pyx_v_self, PyObject *__pyx_arg_score); /*proto*/

#line 401 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_16reject_score[] = "IsotopicFitterBase.reject_score(self, double score) -> bool";

#line 401 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_17reject_score(PyObject *__pyx_v_self, PyObject *__pyx_arg_score) {
  double __pyx_v_score
#line 401 "ms_deisotope/_c/scoring.pyx"
;

#line 401 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 401 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 401 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 401 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 401 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 401 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("reject_score (wrapper)", 0);

#line 401 "ms_deisotope/_c/scoring.pyx"
  assert(__pyx_arg_score); {

#line 401 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_score = __pyx_PyFloat_AsDouble(__pyx_arg_score); if (unlikely((__pyx_v_score == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 401, __pyx_L3_error)

#line 401 "ms_deisotope/_c/scoring.pyx"
  }

#line 401 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 401 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 401 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitterBase.reject_score", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 401 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 401 "ms_deisotope/_c/scoring.pyx"
  return NULL;

#line 401 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 401 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_16reject_score(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self), ((double)__pyx_v_score));

#line 401 "ms_deisotope/_c/scoring.pyx"


#line 401 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 401 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 401 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 401 "ms_deisotope/_c/scoring.pyx"
}

#line 401 "ms_deisotope/_c/scoring.pyx"


#line 401 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_16reject_score(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, double __pyx_v_score) {

#line 401 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 401 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 401 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 401 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 401 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 401 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 401 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("reject_score", 0);

#line 401 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 401 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_reject_score(__pyx_v_self, __pyx_v_score, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 401, __pyx_L1_error)

#line 401 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 401 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 401 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 401 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 401 "ms_deisotope/_c/scoring.pyx"


#line 401 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 401 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 401 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 401 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitterBase.reject_score", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 401 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 401 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 401 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 401 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 401 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 401 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":404
 *         return self.select.reject_score(score)
 * 
 *     cpdef bint is_maximizing(self):             # <<<<<<<<<<<<<<
 *         """Whether or not this fitter's score gets better as it grows
 * 
 */

#line 404 "ms_deisotope/_c/scoring.pyx"


#line 404 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_19is_maximizing(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 404 "ms_deisotope/_c/scoring.pyx"
static int __pyx_f_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_is_maximizing(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, int __pyx_skip_dispatch) {

#line 404 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 404 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 404 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 404 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 404 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 404 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 404 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_5;

#line 404 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 404 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 404 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 404 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("is_maximizing", 0);

#line 404 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 404 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 404 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 404 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 404 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 404 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 404 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 404 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 404 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 404 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_maximizing); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 404, __pyx_L1_error)

#line 404 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 404 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_19is_maximizing)) {

#line 404 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 404 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 404 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 404 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 404 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 404 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 404 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 404 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 404 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 404 "ms_deisotope/_c/scoring.pyx"
          }

#line 404 "ms_deisotope/_c/scoring.pyx"
        }

#line 404 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 404 "ms_deisotope/_c/scoring.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 404 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 404, __pyx_L1_error)

#line 404 "ms_deisotope/_c/scoring.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 404 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 404 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 404, __pyx_L1_error)

#line 404 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 404 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = __pyx_t_5;

#line 404 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 404 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 404 "ms_deisotope/_c/scoring.pyx"
      }

#line 404 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 404 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 404 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 404 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 404 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 404 "ms_deisotope/_c/scoring.pyx"
      }

#line 404 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 404 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 404 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 404 "ms_deisotope/_c/scoring.pyx"
    }

#line 404 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 404 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":412
 *             Whether or not this fitter is a maximizing fitter
 *         """
 *         return self.select.is_maximizing()             # <<<<<<<<<<<<<<
 * 
 *     cpdef IsotopicFitterBase _configure(self, DeconvoluterBase deconvoluter, dict kwargs):
 */

#line 412 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_v_self->select->__pyx_vtab)->is_maximizing(__pyx_v_self->select, 0);

#line 412 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":404
 *         return self.select.reject_score(score)
 * 
 *     cpdef bint is_maximizing(self):             # <<<<<<<<<<<<<<
 *         """Whether or not this fitter's score gets better as it grows
 * 
 */

#line 404 "ms_deisotope/_c/scoring.pyx"


#line 404 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 404 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 404 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 404 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 404 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 404 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 404 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.IsotopicFitterBase.is_maximizing", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 404 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 404 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 404 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 404 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 404 "ms_deisotope/_c/scoring.pyx"
}

#line 404 "ms_deisotope/_c/scoring.pyx"


#line 404 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 404 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_19is_maximizing(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 404 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_18is_maximizing[] = "IsotopicFitterBase.is_maximizing(self) -> bool\nWhether or not this fitter's score gets better as it grows\n\n        Returns\n        -------\n        bool\n            Whether or not this fitter is a maximizing fitter\n        ";

#line 404 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_19is_maximizing(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 404 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 404 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 404 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("is_maximizing (wrapper)", 0);
  __pyx_r = 
#line 404 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_18is_maximizing(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self));

#line 404 "ms_deisotope/_c/scoring.pyx"


#line 404 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 404 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 404 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 404 "ms_deisotope/_c/scoring.pyx"
}

#line 404 "ms_deisotope/_c/scoring.pyx"


#line 404 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_18is_maximizing(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self) {

#line 404 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 404 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 404 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 404 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 404 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 404 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 404 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("is_maximizing", 0);

#line 404 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 404 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_is_maximizing(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 404, __pyx_L1_error)

#line 404 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 404 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 404 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 404 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 404 "ms_deisotope/_c/scoring.pyx"


#line 404 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 404 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 404 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 404 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitterBase.is_maximizing", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 404 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 404 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 404 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 404 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 404 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 404 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":414
 *         return self.select.is_maximizing()
 * 
 *     cpdef IsotopicFitterBase _configure(self, DeconvoluterBase deconvoluter, dict kwargs):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

#line 414 "ms_deisotope/_c/scoring.pyx"


#line 414 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_21_configure(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 414 "ms_deisotope/_c/scoring.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_f_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase__configure(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_deconvoluter, CYTHON_UNUSED PyObject *__pyx_v_kwargs, int __pyx_skip_dispatch) {

#line 414 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_r = NULL;

#line 414 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 414 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 414 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 414 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 414 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 414 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_5;

#line 414 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 414 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 414 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 414 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 414 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_configure", 0);

#line 414 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 414 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 414 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 414 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 414 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 414 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 414 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 414 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 414 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 414 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_configure); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 414, __pyx_L1_error)

#line 414 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 414 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_21_configure)) {

#line 414 "ms_deisotope/_c/scoring.pyx"
        __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 414 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 414 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 414 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_5 = 0;

#line 414 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 414 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 414 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 414 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 414 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 414 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 414 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 414 "ms_deisotope/_c/scoring.pyx"
            __pyx_t_5 = 1;

#line 414 "ms_deisotope/_c/scoring.pyx"
          }

#line 414 "ms_deisotope/_c/scoring.pyx"
        }

#line 414 "ms_deisotope/_c/scoring.pyx"
        #if CYTHON_FAST_PYCALL

#line 414 "ms_deisotope/_c/scoring.pyx"
        if (PyFunction_Check(__pyx_t_3)) {

#line 414 "ms_deisotope/_c/scoring.pyx"
          PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_deconvoluter), __pyx_v_kwargs};

#line 414 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 414, __pyx_L1_error)

#line 414 "ms_deisotope/_c/scoring.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 414 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 414 "ms_deisotope/_c/scoring.pyx"
        } else

#line 414 "ms_deisotope/_c/scoring.pyx"
        #endif

#line 414 "ms_deisotope/_c/scoring.pyx"
        #if CYTHON_FAST_PYCCALL

#line 414 "ms_deisotope/_c/scoring.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {

#line 414 "ms_deisotope/_c/scoring.pyx"
          PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_deconvoluter), __pyx_v_kwargs};

#line 414 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 414, __pyx_L1_error)

#line 414 "ms_deisotope/_c/scoring.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 414 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 414 "ms_deisotope/_c/scoring.pyx"
        } else

#line 414 "ms_deisotope/_c/scoring.pyx"
        #endif

#line 414 "ms_deisotope/_c/scoring.pyx"
        {

#line 414 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 414, __pyx_L1_error)

#line 414 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_6);

#line 414 "ms_deisotope/_c/scoring.pyx"
          if (__pyx_t_4) {

#line 414 "ms_deisotope/_c/scoring.pyx"
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;

#line 414 "ms_deisotope/_c/scoring.pyx"
          }

#line 414 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(((PyObject *)__pyx_v_deconvoluter));

#line 414 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(((PyObject *)__pyx_v_deconvoluter));

#line 414 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_deconvoluter));

#line 414 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(__pyx_v_kwargs);

#line 414 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(__pyx_v_kwargs);

#line 414 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_kwargs);

#line 414 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 414, __pyx_L1_error)

#line 414 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 414 "ms_deisotope/_c/scoring.pyx"
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 414 "ms_deisotope/_c/scoring.pyx"
        }

#line 414 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 414 "ms_deisotope/_c/scoring.pyx"
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase))))) __PYX_ERR(0, 414, __pyx_L1_error)

#line 414 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_t_2);

#line 414 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_2 = 0;

#line 414 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 414 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 414 "ms_deisotope/_c/scoring.pyx"
      }

#line 414 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 414 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 414 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 414 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 414 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 414 "ms_deisotope/_c/scoring.pyx"
      }

#line 414 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 414 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 414 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 414 "ms_deisotope/_c/scoring.pyx"
    }

#line 414 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 414 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":415
 * 
 *     cpdef IsotopicFitterBase _configure(self, DeconvoluterBase deconvoluter, dict kwargs):
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def configure(self, DeconvoluterBase deconvoluter, **kwargs):
 */

#line 415 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 415 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 415 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_v_self;

#line 415 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":414
 *         return self.select.is_maximizing()
 * 
 *     cpdef IsotopicFitterBase _configure(self, DeconvoluterBase deconvoluter, dict kwargs):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

#line 414 "ms_deisotope/_c/scoring.pyx"


#line 414 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 414 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 414 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 414 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 414 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 414 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 414 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 414 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitterBase._configure", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 414 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 414 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 414 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 414 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 414 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 414 "ms_deisotope/_c/scoring.pyx"
}

#line 414 "ms_deisotope/_c/scoring.pyx"


#line 414 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 414 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_21_configure(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 414 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_20_configure[] = "IsotopicFitterBase._configure(self, DeconvoluterBase deconvoluter, dict kwargs) -> IsotopicFitterBase";

#line 414 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_21_configure(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_deconvoluter = 0
#line 414 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_kwargs = 0
#line 414 "ms_deisotope/_c/scoring.pyx"
;

#line 414 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 414 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 414 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 414 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 414 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 414 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_configure (wrapper)", 0);

#line 414 "ms_deisotope/_c/scoring.pyx"
  {

#line 414 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_deconvoluter,&__pyx_n_s_kwargs,0};

#line 414 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[2] = {0,0};

#line 414 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 414 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 414 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 414 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  2: 
#line 414 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 414 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 414 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 414 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 414 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 414 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 414 "ms_deisotope/_c/scoring.pyx"
      }

#line 414 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 414 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 414 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 414 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_deconvoluter)) != 0)) kw_args--;
        else 
#line 414 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 414 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 414 "ms_deisotope/_c/scoring.pyx"
        case  1:

#line 414 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;

#line 414 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_configure", 1, 2, 2, 1); 
#line 414 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 414, __pyx_L3_error)

#line 414 "ms_deisotope/_c/scoring.pyx"
        }

#line 414 "ms_deisotope/_c/scoring.pyx"
      }

#line 414 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 414 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_configure") < 0)) __PYX_ERR(0, 414, __pyx_L3_error)

#line 414 "ms_deisotope/_c/scoring.pyx"
      }

#line 414 "ms_deisotope/_c/scoring.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {

#line 414 "ms_deisotope/_c/scoring.pyx"
      goto __pyx_L5_argtuple_error;

#line 414 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 414 "ms_deisotope/_c/scoring.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 414 "ms_deisotope/_c/scoring.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 414 "ms_deisotope/_c/scoring.pyx"
    }

#line 414 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_deconvoluter = ((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)values[0]);

#line 414 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_kwargs = ((PyObject*)values[1]);

#line 414 "ms_deisotope/_c/scoring.pyx"
  }

#line 414 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 414 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_configure", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 414 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 414, __pyx_L3_error)

#line 414 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 414 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitterBase._configure", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 414 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 414 "ms_deisotope/_c/scoring.pyx"
  return NULL;

#line 414 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 414 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_deconvoluter), __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase, 1, "deconvoluter", 0))) __PYX_ERR(0, 414, __pyx_L1_error)

#line 414 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_kwargs), (&PyDict_Type), 1, "kwargs", 1))) __PYX_ERR(0, 414, __pyx_L1_error)
  __pyx_r = 
#line 414 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_20_configure(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self), __pyx_v_deconvoluter, __pyx_v_kwargs);

#line 414 "ms_deisotope/_c/scoring.pyx"


#line 414 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 414 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 414 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 414 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 414 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 414 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 414 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 414 "ms_deisotope/_c/scoring.pyx"
}

#line 414 "ms_deisotope/_c/scoring.pyx"


#line 414 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_20_configure(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_deconvoluter, PyObject *__pyx_v_kwargs) {

#line 414 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 414 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 414 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 414 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 414 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 414 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 414 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_configure", 0);

#line 414 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 414 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase__configure(__pyx_v_self, __pyx_v_deconvoluter, __pyx_v_kwargs, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 414, __pyx_L1_error)

#line 414 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 414 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 414 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 414 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 414 "ms_deisotope/_c/scoring.pyx"


#line 414 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 414 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 414 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 414 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitterBase._configure", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 414 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 414 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 414 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 414 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 414 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 414 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":417
 *         return self
 * 
 *     def configure(self, DeconvoluterBase deconvoluter, **kwargs):             # <<<<<<<<<<<<<<
 *         return self._configure(deconvoluter, kwargs)
 * 
 */

#line 417 "ms_deisotope/_c/scoring.pyx"


#line 417 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 417 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_23configure(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 417 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_22configure[] = "IsotopicFitterBase.configure(self, DeconvoluterBase deconvoluter, **kwargs)";

#line 417 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_23configure(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_deconvoluter = 0
#line 417 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_kwargs = 0
#line 417 "ms_deisotope/_c/scoring.pyx"
;

#line 417 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 417 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 417 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 417 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 417 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 417 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("configure (wrapper)", 0);

#line 417 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;

#line 417 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_kwargs);

#line 417 "ms_deisotope/_c/scoring.pyx"
  {

#line 417 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_deconvoluter,0};

#line 417 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[1] = {0};

#line 417 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 417 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 417 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 417 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  1: 
#line 417 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 417 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 417 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 417 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 417 "ms_deisotope/_c/scoring.pyx"
      }

#line 417 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 417 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 417 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 417 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_deconvoluter)) != 0)) kw_args--;
        else 
#line 417 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 417 "ms_deisotope/_c/scoring.pyx"
      }

#line 417 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 417 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, pos_args, "configure") < 0)) __PYX_ERR(0, 417, __pyx_L3_error)

#line 417 "ms_deisotope/_c/scoring.pyx"
      }

#line 417 "ms_deisotope/_c/scoring.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {

#line 417 "ms_deisotope/_c/scoring.pyx"
      goto __pyx_L5_argtuple_error;

#line 417 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 417 "ms_deisotope/_c/scoring.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 417 "ms_deisotope/_c/scoring.pyx"
    }

#line 417 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_deconvoluter = ((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)values[0]);

#line 417 "ms_deisotope/_c/scoring.pyx"
  }

#line 417 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 417 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("configure", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); 
#line 417 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 417, __pyx_L3_error)

#line 417 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 417 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;

#line 417 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitterBase.configure", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 417 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 417 "ms_deisotope/_c/scoring.pyx"
  return NULL;

#line 417 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 417 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_deconvoluter), __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase, 1, "deconvoluter", 0))) __PYX_ERR(0, 417, __pyx_L1_error)
  __pyx_r = 
#line 417 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_22configure(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self), __pyx_v_deconvoluter, __pyx_v_kwargs);

#line 417 "ms_deisotope/_c/scoring.pyx"


#line 417 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 417 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 417 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 417 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 417 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 417 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_v_kwargs);

#line 417 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 417 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 417 "ms_deisotope/_c/scoring.pyx"
}

#line 417 "ms_deisotope/_c/scoring.pyx"


#line 417 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_22configure(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_deconvoluter, PyObject *__pyx_v_kwargs) {

#line 417 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 417 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 417 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 417 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 417 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 417 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 417 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("configure", 0);

  /* "ms_deisotope/_c/scoring.pyx":418
 * 
 *     def configure(self, DeconvoluterBase deconvoluter, **kwargs):
 *         return self._configure(deconvoluter, kwargs)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */

#line 418 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 418 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self->__pyx_vtab)->_configure(__pyx_v_self, __pyx_v_deconvoluter, __pyx_v_kwargs, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 418, __pyx_L1_error)

#line 418 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 418 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 418 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 418 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":417
 *         return self
 * 
 *     def configure(self, DeconvoluterBase deconvoluter, **kwargs):             # <<<<<<<<<<<<<<
 *         return self._configure(deconvoluter, kwargs)
 * 
 */

#line 417 "ms_deisotope/_c/scoring.pyx"


#line 417 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 417 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 417 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 417 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitterBase.configure", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 417 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 417 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 417 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 417 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 417 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 417 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":420
 *         return self._configure(deconvoluter, kwargs)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "{self.__class__.__name__}({fields})".format(self=self, fields=self.__getstate__())
 * 
 */

#line 420 "ms_deisotope/_c/scoring.pyx"


#line 420 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 420 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_25__repr__(PyObject *__pyx_v_self); /*proto*/

#line 420 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_25__repr__(PyObject *__pyx_v_self) {

#line 420 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 420 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 420 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = 
#line 420 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_24__repr__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self));

#line 420 "ms_deisotope/_c/scoring.pyx"


#line 420 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 420 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 420 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 420 "ms_deisotope/_c/scoring.pyx"
}

#line 420 "ms_deisotope/_c/scoring.pyx"


#line 420 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_24__repr__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self) {

#line 420 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 420 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 420 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 420 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 420 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 420 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 420 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 420 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 420 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 420 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 420 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "ms_deisotope/_c/scoring.pyx":421
 * 
 *     def __repr__(self):
 *         return "{self.__class__.__name__}({fields})".format(self=self, fields=self.__getstate__())             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 421 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 421 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_self___class_____name___fields, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 421, __pyx_L1_error)

#line 421 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 421 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 421, __pyx_L1_error)

#line 421 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 421 "ms_deisotope/_c/scoring.pyx"
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_self, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 421, __pyx_L1_error)

#line 421 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_getstate); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 421, __pyx_L1_error)

#line 421 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 421 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_5 = NULL;

#line 421 "ms_deisotope/_c/scoring.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {

#line 421 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 421 "ms_deisotope/_c/scoring.pyx"
    if (likely(__pyx_t_5)) {

#line 421 "ms_deisotope/_c/scoring.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 421 "ms_deisotope/_c/scoring.pyx"
      __Pyx_INCREF(__pyx_t_5);

#line 421 "ms_deisotope/_c/scoring.pyx"
      __Pyx_INCREF(function);

#line 421 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF_SET(__pyx_t_4, function);

#line 421 "ms_deisotope/_c/scoring.pyx"
    }

#line 421 "ms_deisotope/_c/scoring.pyx"
  }

#line 421 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);

#line 421 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 421 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 421, __pyx_L1_error)

#line 421 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 421 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 421 "ms_deisotope/_c/scoring.pyx"
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_fields, __pyx_t_3) < 0) __PYX_ERR(0, 421, __pyx_L1_error)

#line 421 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 421 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 421, __pyx_L1_error)

#line 421 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 421 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 421 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 421 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_3;

#line 421 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = 0;

#line 421 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":420
 *         return self._configure(deconvoluter, kwargs)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "{self.__class__.__name__}({fields})".format(self=self, fields=self.__getstate__())
 * 
 */

#line 420 "ms_deisotope/_c/scoring.pyx"


#line 420 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 420 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 420 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 420 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 420 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 420 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 420 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 420 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitterBase.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 420 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 420 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 420 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 420 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 420 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 420 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pxd":52
 * cdef class IsotopicFitterBase(object):
 *     cdef:
 *         public FitSelectorBase select             # <<<<<<<<<<<<<<
 * 
 *     cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected)
 */

#line 52 "ms_deisotope/_c/scoring.pxd"


#line 52 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 52 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_6select_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 52 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_6select_1__get__(PyObject *__pyx_v_self) {

#line 52 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = 0;

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 52 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_6select___get__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self));

#line 52 "ms_deisotope/_c/scoring.pxd"


#line 52 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 52 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 52 "ms_deisotope/_c/scoring.pxd"
}

#line 52 "ms_deisotope/_c/scoring.pxd"


#line 52 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_6select___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self) {

#line 52 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = NULL;

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->select));

#line 52 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = ((PyObject *)__pyx_v_self->select);

#line 52 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 52 "ms_deisotope/_c/scoring.pxd"


#line 52 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 52 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 52 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 52 "ms_deisotope/_c/scoring.pxd"
}

#line 52 "ms_deisotope/_c/scoring.pxd"


#line 52 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 52 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_6select_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 52 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_6select_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 52 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 52 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_6select_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 52 "ms_deisotope/_c/scoring.pxd"


#line 52 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 52 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 52 "ms_deisotope/_c/scoring.pxd"
}

#line 52 "ms_deisotope/_c/scoring.pxd"


#line 52 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_6select_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self, PyObject *__pyx_v_value) {

#line 52 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 52 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 52 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 52 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 52 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 52 "ms_deisotope/_c/scoring.pxd"
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase))))) __PYX_ERR(1, 52, __pyx_L1_error)

#line 52 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = __pyx_v_value;

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(__pyx_t_1);

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GIVEREF(__pyx_t_1);

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_v_self->select);

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->select));

#line 52 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->select = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_t_1);

#line 52 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = 0;

#line 52 "ms_deisotope/_c/scoring.pxd"


#line 52 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 52 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 52 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 52 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.IsotopicFitterBase.select.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 52 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = -1;

#line 52 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 52 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 52 "ms_deisotope/_c/scoring.pxd"
}

#line 52 "ms_deisotope/_c/scoring.pxd"


#line 52 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 52 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_6select_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 52 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_6select_5__del__(PyObject *__pyx_v_self) {

#line 52 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 52 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_6select_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self));

#line 52 "ms_deisotope/_c/scoring.pxd"


#line 52 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 52 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 52 "ms_deisotope/_c/scoring.pxd"
}

#line 52 "ms_deisotope/_c/scoring.pxd"


#line 52 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_6select_4__del__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_v_self) {

#line 52 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(Py_None);

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GIVEREF(Py_None);

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_v_self->select);

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->select));

#line 52 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->select = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)Py_None);

#line 52 "ms_deisotope/_c/scoring.pxd"


#line 52 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 52 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 52 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 52 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 52 "ms_deisotope/_c/scoring.pxd"
}

/* "ms_deisotope/_c/scoring.pyx":424
 * 
 * 
 * cdef double sum_intensity_theoretical(list peaklist):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double summed
 */

#line 424 "ms_deisotope/_c/scoring.pyx"


#line 424 "ms_deisotope/_c/scoring.pyx"
static double __pyx_f_12ms_deisotope_2_c_7scoring_sum_intensity_theoretical(PyObject *__pyx_v_peaklist) {
  double __pyx_v_summed
#line 424 "ms_deisotope/_c/scoring.pyx"
;
  size_t __pyx_v_i
#line 424 "ms_deisotope/_c/scoring.pyx"
;
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_peak = 0
#line 424 "ms_deisotope/_c/scoring.pyx"
;

#line 424 "ms_deisotope/_c/scoring.pyx"
  double __pyx_r;

#line 424 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 424 "ms_deisotope/_c/scoring.pyx"
  Py_ssize_t __pyx_t_1;

#line 424 "ms_deisotope/_c/scoring.pyx"
  Py_ssize_t __pyx_t_2;

#line 424 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_3;

#line 424 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4;

#line 424 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 424 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("sum_intensity_theoretical", 0);

  /* "ms_deisotope/_c/scoring.pyx":429
 *         size_t i
 *         TheoreticalPeak peak
 *     summed = 0             # <<<<<<<<<<<<<<
 *     for i in range(PyList_GET_SIZE(peaklist)):
 *         peak = <TheoreticalPeak>PyList_GET_ITEM(peaklist, i)
 */

#line 429 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_summed = 0.0;

  /* "ms_deisotope/_c/scoring.pyx":430
 *         TheoreticalPeak peak
 *     summed = 0
 *     for i in range(PyList_GET_SIZE(peaklist)):             # <<<<<<<<<<<<<<
 *         peak = <TheoreticalPeak>PyList_GET_ITEM(peaklist, i)
 *         summed += peak.intensity
 */

#line 430 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyList_GET_SIZE(__pyx_v_peaklist);

#line 430 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = __pyx_t_1;

#line 430 "ms_deisotope/_c/scoring.pyx"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 430 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/scoring.pyx":431
 *     summed = 0
 *     for i in range(PyList_GET_SIZE(peaklist)):
 *         peak = <TheoreticalPeak>PyList_GET_ITEM(peaklist, i)             # <<<<<<<<<<<<<<
 *         summed += peak.intensity
 *     return summed
 */

#line 431 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_4 = PyList_GET_ITEM(__pyx_v_peaklist, __pyx_v_i);

#line 431 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = ((PyObject *)__pyx_t_4);

#line 431 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_5);

#line 431 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_5));

#line 431 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = 0;

    /* "ms_deisotope/_c/scoring.pyx":432
 *     for i in range(PyList_GET_SIZE(peaklist)):
 *         peak = <TheoreticalPeak>PyList_GET_ITEM(peaklist, i)
 *         summed += peak.intensity             # <<<<<<<<<<<<<<
 *     return summed
 * 
 */

#line 432 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_summed = (__pyx_v_summed + __pyx_v_peak->intensity);

#line 432 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":433
 *         peak = <TheoreticalPeak>PyList_GET_ITEM(peaklist, i)
 *         summed += peak.intensity
 *     return summed             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 433 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_v_summed;

#line 433 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":424
 * 
 * 
 * cdef double sum_intensity_theoretical(list peaklist):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double summed
 */

#line 424 "ms_deisotope/_c/scoring.pyx"


#line 424 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 424 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 424 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 424 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 424 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 424 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":436
 * 
 * 
 * cdef double sum_intensity_fitted(list peaklist):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double summed
 */

#line 436 "ms_deisotope/_c/scoring.pyx"


#line 436 "ms_deisotope/_c/scoring.pyx"
static double __pyx_f_12ms_deisotope_2_c_7scoring_sum_intensity_fitted(PyObject *__pyx_v_peaklist) {
  double __pyx_v_summed
#line 436 "ms_deisotope/_c/scoring.pyx"
;
  size_t __pyx_v_i
#line 436 "ms_deisotope/_c/scoring.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = 0
#line 436 "ms_deisotope/_c/scoring.pyx"
;

#line 436 "ms_deisotope/_c/scoring.pyx"
  double __pyx_r;

#line 436 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 436 "ms_deisotope/_c/scoring.pyx"
  Py_ssize_t __pyx_t_1;

#line 436 "ms_deisotope/_c/scoring.pyx"
  Py_ssize_t __pyx_t_2;

#line 436 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_3;

#line 436 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4;

#line 436 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 436 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("sum_intensity_fitted", 0);

  /* "ms_deisotope/_c/scoring.pyx":441
 *         size_t i
 *         FittedPeak peak
 *     summed = 0             # <<<<<<<<<<<<<<
 *     for i in range(PyList_GET_SIZE(peaklist)):
 *         peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
 */

#line 441 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_summed = 0.0;

  /* "ms_deisotope/_c/scoring.pyx":442
 *         FittedPeak peak
 *     summed = 0
 *     for i in range(PyList_GET_SIZE(peaklist)):             # <<<<<<<<<<<<<<
 *         peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
 *         summed += peak.intensity
 */

#line 442 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyList_GET_SIZE(__pyx_v_peaklist);

#line 442 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = __pyx_t_1;

#line 442 "ms_deisotope/_c/scoring.pyx"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 442 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/scoring.pyx":443
 *     summed = 0
 *     for i in range(PyList_GET_SIZE(peaklist)):
 *         peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)             # <<<<<<<<<<<<<<
 *         summed += peak.intensity
 *     return summed
 */

#line 443 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_4 = PyList_GET_ITEM(__pyx_v_peaklist, __pyx_v_i);

#line 443 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = ((PyObject *)__pyx_t_4);

#line 443 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_5);

#line 443 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_5));

#line 443 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = 0;

    /* "ms_deisotope/_c/scoring.pyx":444
 *     for i in range(PyList_GET_SIZE(peaklist)):
 *         peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
 *         summed += peak.intensity             # <<<<<<<<<<<<<<
 *     return summed
 * 
 */

#line 444 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_summed = (__pyx_v_summed + __pyx_v_peak->__pyx_base.intensity);

#line 444 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":445
 *         peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
 *         summed += peak.intensity
 *     return summed             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 445 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_v_summed;

#line 445 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":436
 * 
 * 
 * cdef double sum_intensity_fitted(list peaklist):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double summed
 */

#line 436 "ms_deisotope/_c/scoring.pyx"


#line 436 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 436 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 436 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 436 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 436 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 436 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":449
 * 
 * @cython.cdivision
 * cdef double* normalize_intensity_theoretical(list peaklist, double* out, double total):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

#line 449 "ms_deisotope/_c/scoring.pyx"


#line 449 "ms_deisotope/_c/scoring.pyx"
static double *__pyx_f_12ms_deisotope_2_c_7scoring_normalize_intensity_theoretical(PyObject *__pyx_v_peaklist, double *__pyx_v_out, double __pyx_v_total) {
  size_t __pyx_v_i
#line 449 "ms_deisotope/_c/scoring.pyx"
;
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_peak = 0
#line 449 "ms_deisotope/_c/scoring.pyx"
;

#line 449 "ms_deisotope/_c/scoring.pyx"
  double *__pyx_r;

#line 449 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 449 "ms_deisotope/_c/scoring.pyx"
  Py_ssize_t __pyx_t_1;

#line 449 "ms_deisotope/_c/scoring.pyx"
  Py_ssize_t __pyx_t_2;

#line 449 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_3;

#line 449 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4;

#line 449 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 449 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("normalize_intensity_theoretical", 0);

  /* "ms_deisotope/_c/scoring.pyx":454
 *         TheoreticalPeak peak
 * 
 *     for i in range(PyList_GET_SIZE(peaklist)):             # <<<<<<<<<<<<<<
 *         peak = <TheoreticalPeak>PyList_GET_ITEM(peaklist, i)
 *         out[i] = peak.intensity / total
 */

#line 454 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyList_GET_SIZE(__pyx_v_peaklist);

#line 454 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = __pyx_t_1;

#line 454 "ms_deisotope/_c/scoring.pyx"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 454 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/scoring.pyx":455
 * 
 *     for i in range(PyList_GET_SIZE(peaklist)):
 *         peak = <TheoreticalPeak>PyList_GET_ITEM(peaklist, i)             # <<<<<<<<<<<<<<
 *         out[i] = peak.intensity / total
 * 
 */

#line 455 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_4 = PyList_GET_ITEM(__pyx_v_peaklist, __pyx_v_i);

#line 455 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = ((PyObject *)__pyx_t_4);

#line 455 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_5);

#line 455 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_5));

#line 455 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = 0;

    /* "ms_deisotope/_c/scoring.pyx":456
 *     for i in range(PyList_GET_SIZE(peaklist)):
 *         peak = <TheoreticalPeak>PyList_GET_ITEM(peaklist, i)
 *         out[i] = peak.intensity / total             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 456 "ms_deisotope/_c/scoring.pyx"
    (__pyx_v_out[__pyx_v_i]) = (__pyx_v_peak->intensity / __pyx_v_total);

#line 456 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":449
 * 
 * @cython.cdivision
 * cdef double* normalize_intensity_theoretical(list peaklist, double* out, double total):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

#line 449 "ms_deisotope/_c/scoring.pyx"


#line 449 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 449 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 449 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 449 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 449 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 449 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":460
 * 
 * @cython.cdivision
 * cdef double* normalize_intensity_fitted(list peaklist, double* out, double total):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

#line 460 "ms_deisotope/_c/scoring.pyx"


#line 460 "ms_deisotope/_c/scoring.pyx"
static double *__pyx_f_12ms_deisotope_2_c_7scoring_normalize_intensity_fitted(PyObject *__pyx_v_peaklist, double *__pyx_v_out, double __pyx_v_total) {
  size_t __pyx_v_i
#line 460 "ms_deisotope/_c/scoring.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = 0
#line 460 "ms_deisotope/_c/scoring.pyx"
;

#line 460 "ms_deisotope/_c/scoring.pyx"
  double *__pyx_r;

#line 460 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 460 "ms_deisotope/_c/scoring.pyx"
  Py_ssize_t __pyx_t_1;

#line 460 "ms_deisotope/_c/scoring.pyx"
  Py_ssize_t __pyx_t_2;

#line 460 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_3;

#line 460 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4;

#line 460 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 460 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("normalize_intensity_fitted", 0);

  /* "ms_deisotope/_c/scoring.pyx":465
 *         FittedPeak peak
 * 
 *     for i in range(PyList_GET_SIZE(peaklist)):             # <<<<<<<<<<<<<<
 *         peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
 *         out[i] = peak.intensity / total
 */

#line 465 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyList_GET_SIZE(__pyx_v_peaklist);

#line 465 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = __pyx_t_1;

#line 465 "ms_deisotope/_c/scoring.pyx"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 465 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/scoring.pyx":466
 * 
 *     for i in range(PyList_GET_SIZE(peaklist)):
 *         peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)             # <<<<<<<<<<<<<<
 *         out[i] = peak.intensity / total
 * 
 */

#line 466 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_4 = PyList_GET_ITEM(__pyx_v_peaklist, __pyx_v_i);

#line 466 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = ((PyObject *)__pyx_t_4);

#line 466 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_5);

#line 466 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_5));

#line 466 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = 0;

    /* "ms_deisotope/_c/scoring.pyx":467
 *     for i in range(PyList_GET_SIZE(peaklist)):
 *         peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
 *         out[i] = peak.intensity / total             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 467 "ms_deisotope/_c/scoring.pyx"
    (__pyx_v_out[__pyx_v_i]) = (__pyx_v_peak->__pyx_base.intensity / __pyx_v_total);

#line 467 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":460
 * 
 * @cython.cdivision
 * cdef double* normalize_intensity_fitted(list peaklist, double* out, double total):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

#line 460 "ms_deisotope/_c/scoring.pyx"


#line 460 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 460 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 460 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 460 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 460 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 460 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":484
 *     """
 *     @cython.cdivision
 *     cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double g_score, obs, theo, log_ratio
 */

#line 484 "ms_deisotope/_c/scoring.pyx"


#line 484 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_11GTestFitter_1_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 484 "ms_deisotope/_c/scoring.pyx"
static double __pyx_f_12ms_deisotope_2_c_7scoring_11GTestFitter__evaluate(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected, int __pyx_skip_dispatch) {
  double __pyx_v_g_score
#line 484 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_obs
#line 484 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_theo
#line 484 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_log_ratio
#line 484 "ms_deisotope/_c/scoring.pyx"
;
  size_t __pyx_v_n
#line 484 "ms_deisotope/_c/scoring.pyx"
;
  size_t __pyx_v_i
#line 484 "ms_deisotope/_c/scoring.pyx"
;

#line 484 "ms_deisotope/_c/scoring.pyx"
  double __pyx_r;

#line 484 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 484 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 484 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 484 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 484 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 484 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_5;

#line 484 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 484 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_7;

#line 484 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_8;

#line 484 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_9;

#line 484 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_10;

#line 484 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_11;

#line 484 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 484 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 484 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 484 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate", 0);

#line 484 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 484 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 484 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 484 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 484 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 484 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 484 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 484 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 484 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 484 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_evaluate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 484, __pyx_L1_error)

#line 484 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 484 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_11GTestFitter_1_evaluate)) {

#line 484 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 484 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 484 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_5 = 0;

#line 484 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 484 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 484 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 484 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 484 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 484 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 484 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 484 "ms_deisotope/_c/scoring.pyx"
            __pyx_t_5 = 1;

#line 484 "ms_deisotope/_c/scoring.pyx"
          }

#line 484 "ms_deisotope/_c/scoring.pyx"
        }

#line 484 "ms_deisotope/_c/scoring.pyx"
        #if CYTHON_FAST_PYCALL

#line 484 "ms_deisotope/_c/scoring.pyx"
        if (PyFunction_Check(__pyx_t_3)) {

#line 484 "ms_deisotope/_c/scoring.pyx"
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_peaklist), __pyx_v_observed, __pyx_v_expected};

#line 484 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 484, __pyx_L1_error)

#line 484 "ms_deisotope/_c/scoring.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 484 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 484 "ms_deisotope/_c/scoring.pyx"
        } else

#line 484 "ms_deisotope/_c/scoring.pyx"
        #endif

#line 484 "ms_deisotope/_c/scoring.pyx"
        #if CYTHON_FAST_PYCCALL

#line 484 "ms_deisotope/_c/scoring.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {

#line 484 "ms_deisotope/_c/scoring.pyx"
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_peaklist), __pyx_v_observed, __pyx_v_expected};

#line 484 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 484, __pyx_L1_error)

#line 484 "ms_deisotope/_c/scoring.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 484 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 484 "ms_deisotope/_c/scoring.pyx"
        } else

#line 484 "ms_deisotope/_c/scoring.pyx"
        #endif

#line 484 "ms_deisotope/_c/scoring.pyx"
        {

#line 484 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_6 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 484, __pyx_L1_error)

#line 484 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_6);

#line 484 "ms_deisotope/_c/scoring.pyx"
          if (__pyx_t_4) {

#line 484 "ms_deisotope/_c/scoring.pyx"
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;

#line 484 "ms_deisotope/_c/scoring.pyx"
          }

#line 484 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(((PyObject *)__pyx_v_peaklist));

#line 484 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(((PyObject *)__pyx_v_peaklist));

#line 484 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_peaklist));

#line 484 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(__pyx_v_observed);

#line 484 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(__pyx_v_observed);

#line 484 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_observed);

#line 484 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(__pyx_v_expected);

#line 484 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(__pyx_v_expected);

#line 484 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, __pyx_v_expected);

#line 484 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 484, __pyx_L1_error)

#line 484 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 484 "ms_deisotope/_c/scoring.pyx"
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 484 "ms_deisotope/_c/scoring.pyx"
        }

#line 484 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 484 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_7 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_7 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 484, __pyx_L1_error)

#line 484 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 484 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = __pyx_t_7;

#line 484 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 484 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 484 "ms_deisotope/_c/scoring.pyx"
      }

#line 484 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 484 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 484 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 484 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 484 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 484 "ms_deisotope/_c/scoring.pyx"
      }

#line 484 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 484 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 484 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 484 "ms_deisotope/_c/scoring.pyx"
    }

#line 484 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 484 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":489
 *             size_t n
 * 
 *         n = PyList_GET_SIZE(observed)             # <<<<<<<<<<<<<<
 * 
 *         g_score = 0.
 */

#line 489 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_n = PyList_GET_SIZE(__pyx_v_observed);

  /* "ms_deisotope/_c/scoring.pyx":491
 *         n = PyList_GET_SIZE(observed)
 * 
 *         g_score = 0.             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             obs = (<FittedPeak>PyList_GET_ITEM(observed, i)).intensity
 */

#line 491 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_g_score = 0.;

  /* "ms_deisotope/_c/scoring.pyx":492
 * 
 *         g_score = 0.
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             obs = (<FittedPeak>PyList_GET_ITEM(observed, i)).intensity
 *             theo = (<TheoreticalPeak>PyList_GET_ITEM(expected, i)).intensity
 */

#line 492 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_8 = __pyx_v_n;

#line 492 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_9 = __pyx_t_8;

#line 492 "ms_deisotope/_c/scoring.pyx"
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {

#line 492 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_i = __pyx_t_10;

    /* "ms_deisotope/_c/scoring.pyx":493
 *         g_score = 0.
 *         for i in range(n):
 *             obs = (<FittedPeak>PyList_GET_ITEM(observed, i)).intensity             # <<<<<<<<<<<<<<
 *             theo = (<TheoreticalPeak>PyList_GET_ITEM(expected, i)).intensity
 * 
 */

#line 493 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_11 = PyList_GET_ITEM(__pyx_v_observed, __pyx_v_i);

#line 493 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_7 = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_11)->__pyx_base.intensity;

#line 493 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_obs = __pyx_t_7;

    /* "ms_deisotope/_c/scoring.pyx":494
 *         for i in range(n):
 *             obs = (<FittedPeak>PyList_GET_ITEM(observed, i)).intensity
 *             theo = (<TheoreticalPeak>PyList_GET_ITEM(expected, i)).intensity             # <<<<<<<<<<<<<<
 * 
 *             log_ratio = log(obs) - log(theo)
 */

#line 494 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_11 = PyList_GET_ITEM(__pyx_v_expected, __pyx_v_i);

#line 494 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_7 = ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_11)->intensity;

#line 494 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_theo = __pyx_t_7;

    /* "ms_deisotope/_c/scoring.pyx":496
 *             theo = (<TheoreticalPeak>PyList_GET_ITEM(expected, i)).intensity
 * 
 *             log_ratio = log(obs) - log(theo)             # <<<<<<<<<<<<<<
 *             g_score += obs * log_ratio
 * 
 */

#line 496 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_log_ratio = (log(__pyx_v_obs) - log(__pyx_v_theo));

    /* "ms_deisotope/_c/scoring.pyx":497
 * 
 *             log_ratio = log(obs) - log(theo)
 *             g_score += obs * log_ratio             # <<<<<<<<<<<<<<
 * 
 *         return g_score * 2.
 */

#line 497 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_g_score = (__pyx_v_g_score + (__pyx_v_obs * __pyx_v_log_ratio));

#line 497 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":499
 *             g_score += obs * log_ratio
 * 
 *         return g_score * 2.             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 499 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = (__pyx_v_g_score * 2.);

#line 499 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":484
 *     """
 *     @cython.cdivision
 *     cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double g_score, obs, theo, log_ratio
 */

#line 484 "ms_deisotope/_c/scoring.pyx"


#line 484 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 484 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 484 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 484 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 484 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 484 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 484 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 484 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.GTestFitter._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 484 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 484 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 484 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 484 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 484 "ms_deisotope/_c/scoring.pyx"
}

#line 484 "ms_deisotope/_c/scoring.pyx"


#line 484 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 484 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_11GTestFitter_1_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 484 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_11GTestFitter__evaluate[] = "GTestFitter._evaluate(self, PeakSet peaklist, list observed, list expected) -> double";

#line 484 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_11GTestFitter_1_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist = 0
#line 484 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_observed = 0
#line 484 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_expected = 0
#line 484 "ms_deisotope/_c/scoring.pyx"
;

#line 484 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 484 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 484 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 484 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 484 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 484 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate (wrapper)", 0);

#line 484 "ms_deisotope/_c/scoring.pyx"
  {

#line 484 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peaklist,&__pyx_n_s_observed,&__pyx_n_s_expected,0};

#line 484 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[3] = {0,0,0};

#line 484 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 484 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 484 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 484 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  3: 
#line 484 "ms_deisotope/_c/scoring.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 484 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 484 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 484 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 484 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 484 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 484 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 484 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 484 "ms_deisotope/_c/scoring.pyx"
      }

#line 484 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 484 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 484 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 484 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaklist)) != 0)) kw_args--;
        else 
#line 484 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 484 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 484 "ms_deisotope/_c/scoring.pyx"
        case  1:

#line 484 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_observed)) != 0)) kw_args--;

#line 484 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, 1); 
#line 484 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 484, __pyx_L3_error)

#line 484 "ms_deisotope/_c/scoring.pyx"
        }

#line 484 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 484 "ms_deisotope/_c/scoring.pyx"
        case  2:

#line 484 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_expected)) != 0)) kw_args--;

#line 484 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, 2); 
#line 484 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 484, __pyx_L3_error)

#line 484 "ms_deisotope/_c/scoring.pyx"
        }

#line 484 "ms_deisotope/_c/scoring.pyx"
      }

#line 484 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 484 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_evaluate") < 0)) __PYX_ERR(0, 484, __pyx_L3_error)

#line 484 "ms_deisotope/_c/scoring.pyx"
      }

#line 484 "ms_deisotope/_c/scoring.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 484 "ms_deisotope/_c/scoring.pyx"
      goto __pyx_L5_argtuple_error;

#line 484 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 484 "ms_deisotope/_c/scoring.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 484 "ms_deisotope/_c/scoring.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 484 "ms_deisotope/_c/scoring.pyx"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 484 "ms_deisotope/_c/scoring.pyx"
    }

#line 484 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_peaklist = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[0]);

#line 484 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_observed = ((PyObject*)values[1]);

#line 484 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_expected = ((PyObject*)values[2]);

#line 484 "ms_deisotope/_c/scoring.pyx"
  }

#line 484 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 484 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 484 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 484, __pyx_L3_error)

#line 484 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 484 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.GTestFitter._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 484 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 484 "ms_deisotope/_c/scoring.pyx"
  return NULL;

#line 484 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 484 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peaklist), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peaklist", 0))) __PYX_ERR(0, 484, __pyx_L1_error)

#line 484 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_observed), (&PyList_Type), 1, "observed", 1))) __PYX_ERR(0, 484, __pyx_L1_error)

#line 484 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_expected), (&PyList_Type), 1, "expected", 1))) __PYX_ERR(0, 484, __pyx_L1_error)
  __pyx_r = 
#line 484 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_11GTestFitter__evaluate(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter *)__pyx_v_self), __pyx_v_peaklist, __pyx_v_observed, __pyx_v_expected);

#line 484 "ms_deisotope/_c/scoring.pyx"


#line 484 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 484 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 484 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 484 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 484 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 484 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 484 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 484 "ms_deisotope/_c/scoring.pyx"
}

#line 484 "ms_deisotope/_c/scoring.pyx"


#line 484 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_11GTestFitter__evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected) {

#line 484 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 484 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 484 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 484 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 484 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 484 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 484 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate", 0);

#line 484 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 484 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_12ms_deisotope_2_c_7scoring_11GTestFitter__evaluate(__pyx_v_self, __pyx_v_peaklist, __pyx_v_observed, __pyx_v_expected, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 484, __pyx_L1_error)

#line 484 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 484 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 484 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 484 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 484 "ms_deisotope/_c/scoring.pyx"


#line 484 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 484 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 484 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 484 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.GTestFitter._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 484 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 484 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 484 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 484 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 484 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 484 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":514
 * 
 *     @cython.cdivision
 *     cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double total_observed
 */

#line 514 "ms_deisotope/_c/scoring.pyx"


#line 514 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17ScaledGTestFitter_1_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 514 "ms_deisotope/_c/scoring.pyx"
static double __pyx_f_12ms_deisotope_2_c_7scoring_17ScaledGTestFitter__evaluate(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected, int __pyx_skip_dispatch) {
  double __pyx_v_total_observed
#line 514 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_total_expected
#line 514 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_g_score
#line 514 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_obs
#line 514 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_theo
#line 514 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_log_ratio
#line 514 "ms_deisotope/_c/scoring.pyx"
;
  size_t __pyx_v_n
#line 514 "ms_deisotope/_c/scoring.pyx"
;
  size_t __pyx_v_i
#line 514 "ms_deisotope/_c/scoring.pyx"
;

#line 514 "ms_deisotope/_c/scoring.pyx"
  double __pyx_r;

#line 514 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 514 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 514 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 514 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 514 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 514 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_5;

#line 514 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 514 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_7;

#line 514 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_8;

#line 514 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_9;

#line 514 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_10;

#line 514 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_11;

#line 514 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 514 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 514 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 514 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate", 0);

#line 514 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 514 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 514 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 514 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 514 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 514 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 514 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 514 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 514 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 514 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_evaluate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 514, __pyx_L1_error)

#line 514 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 514 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_17ScaledGTestFitter_1_evaluate)) {

#line 514 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 514 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 514 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_5 = 0;

#line 514 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 514 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 514 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 514 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 514 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 514 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 514 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 514 "ms_deisotope/_c/scoring.pyx"
            __pyx_t_5 = 1;

#line 514 "ms_deisotope/_c/scoring.pyx"
          }

#line 514 "ms_deisotope/_c/scoring.pyx"
        }

#line 514 "ms_deisotope/_c/scoring.pyx"
        #if CYTHON_FAST_PYCALL

#line 514 "ms_deisotope/_c/scoring.pyx"
        if (PyFunction_Check(__pyx_t_3)) {

#line 514 "ms_deisotope/_c/scoring.pyx"
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_peaklist), __pyx_v_observed, __pyx_v_expected};

#line 514 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 514, __pyx_L1_error)

#line 514 "ms_deisotope/_c/scoring.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 514 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 514 "ms_deisotope/_c/scoring.pyx"
        } else

#line 514 "ms_deisotope/_c/scoring.pyx"
        #endif

#line 514 "ms_deisotope/_c/scoring.pyx"
        #if CYTHON_FAST_PYCCALL

#line 514 "ms_deisotope/_c/scoring.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {

#line 514 "ms_deisotope/_c/scoring.pyx"
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_peaklist), __pyx_v_observed, __pyx_v_expected};

#line 514 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 514, __pyx_L1_error)

#line 514 "ms_deisotope/_c/scoring.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 514 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 514 "ms_deisotope/_c/scoring.pyx"
        } else

#line 514 "ms_deisotope/_c/scoring.pyx"
        #endif

#line 514 "ms_deisotope/_c/scoring.pyx"
        {

#line 514 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_6 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 514, __pyx_L1_error)

#line 514 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_6);

#line 514 "ms_deisotope/_c/scoring.pyx"
          if (__pyx_t_4) {

#line 514 "ms_deisotope/_c/scoring.pyx"
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;

#line 514 "ms_deisotope/_c/scoring.pyx"
          }

#line 514 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(((PyObject *)__pyx_v_peaklist));

#line 514 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(((PyObject *)__pyx_v_peaklist));

#line 514 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_peaklist));

#line 514 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(__pyx_v_observed);

#line 514 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(__pyx_v_observed);

#line 514 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_observed);

#line 514 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(__pyx_v_expected);

#line 514 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(__pyx_v_expected);

#line 514 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, __pyx_v_expected);

#line 514 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 514, __pyx_L1_error)

#line 514 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 514 "ms_deisotope/_c/scoring.pyx"
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 514 "ms_deisotope/_c/scoring.pyx"
        }

#line 514 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 514 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_7 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_7 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 514, __pyx_L1_error)

#line 514 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 514 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = __pyx_t_7;

#line 514 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 514 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 514 "ms_deisotope/_c/scoring.pyx"
      }

#line 514 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 514 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 514 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 514 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 514 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 514 "ms_deisotope/_c/scoring.pyx"
      }

#line 514 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 514 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 514 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 514 "ms_deisotope/_c/scoring.pyx"
    }

#line 514 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 514 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":520
 *             double g_score, obs, theo, log_ratio
 *             size_t n
 *         total_observed = sum_intensity_fitted(observed)             # <<<<<<<<<<<<<<
 *         total_expected = sum_intensity_theoretical(expected)
 *         n = PyList_GET_SIZE(observed)
 */

#line 520 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_total_observed = __pyx_f_12ms_deisotope_2_c_7scoring_sum_intensity_fitted(__pyx_v_observed);

  /* "ms_deisotope/_c/scoring.pyx":521
 *             size_t n
 *         total_observed = sum_intensity_fitted(observed)
 *         total_expected = sum_intensity_theoretical(expected)             # <<<<<<<<<<<<<<
 *         n = PyList_GET_SIZE(observed)
 * 
 */

#line 521 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_total_expected = __pyx_f_12ms_deisotope_2_c_7scoring_sum_intensity_theoretical(__pyx_v_expected);

  /* "ms_deisotope/_c/scoring.pyx":522
 *         total_observed = sum_intensity_fitted(observed)
 *         total_expected = sum_intensity_theoretical(expected)
 *         n = PyList_GET_SIZE(observed)             # <<<<<<<<<<<<<<
 * 
 *         g_score = 0.
 */

#line 522 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_n = PyList_GET_SIZE(__pyx_v_observed);

  /* "ms_deisotope/_c/scoring.pyx":524
 *         n = PyList_GET_SIZE(observed)
 * 
 *         g_score = 0.             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             obs = (<FittedPeak>PyList_GET_ITEM(observed, i)).intensity / total_observed
 */

#line 524 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_g_score = 0.;

  /* "ms_deisotope/_c/scoring.pyx":525
 * 
 *         g_score = 0.
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             obs = (<FittedPeak>PyList_GET_ITEM(observed, i)).intensity / total_observed
 *             theo = (<TheoreticalPeak>PyList_GET_ITEM(expected, i)).intensity / total_expected
 */

#line 525 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_8 = __pyx_v_n;

#line 525 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_9 = __pyx_t_8;

#line 525 "ms_deisotope/_c/scoring.pyx"
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {

#line 525 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_i = __pyx_t_10;

    /* "ms_deisotope/_c/scoring.pyx":526
 *         g_score = 0.
 *         for i in range(n):
 *             obs = (<FittedPeak>PyList_GET_ITEM(observed, i)).intensity / total_observed             # <<<<<<<<<<<<<<
 *             theo = (<TheoreticalPeak>PyList_GET_ITEM(expected, i)).intensity / total_expected
 * 
 */

#line 526 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_11 = PyList_GET_ITEM(__pyx_v_observed, __pyx_v_i);

#line 526 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_obs = (((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_11)->__pyx_base.intensity / __pyx_v_total_observed);

    /* "ms_deisotope/_c/scoring.pyx":527
 *         for i in range(n):
 *             obs = (<FittedPeak>PyList_GET_ITEM(observed, i)).intensity / total_observed
 *             theo = (<TheoreticalPeak>PyList_GET_ITEM(expected, i)).intensity / total_expected             # <<<<<<<<<<<<<<
 * 
 *             log_ratio = log(obs) - log(theo)
 */

#line 527 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_11 = PyList_GET_ITEM(__pyx_v_expected, __pyx_v_i);

#line 527 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_theo = (((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_11)->intensity / __pyx_v_total_expected);

    /* "ms_deisotope/_c/scoring.pyx":529
 *             theo = (<TheoreticalPeak>PyList_GET_ITEM(expected, i)).intensity / total_expected
 * 
 *             log_ratio = log(obs) - log(theo)             # <<<<<<<<<<<<<<
 *             g_score += obs * log_ratio
 * 
 */

#line 529 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_log_ratio = (log(__pyx_v_obs) - log(__pyx_v_theo));

    /* "ms_deisotope/_c/scoring.pyx":530
 * 
 *             log_ratio = log(obs) - log(theo)
 *             g_score += obs * log_ratio             # <<<<<<<<<<<<<<
 * 
 *         return g_score * 2.
 */

#line 530 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_g_score = (__pyx_v_g_score + (__pyx_v_obs * __pyx_v_log_ratio));

#line 530 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":532
 *             g_score += obs * log_ratio
 * 
 *         return g_score * 2.             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 532 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = (__pyx_v_g_score * 2.);

#line 532 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":514
 * 
 *     @cython.cdivision
 *     cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double total_observed
 */

#line 514 "ms_deisotope/_c/scoring.pyx"


#line 514 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 514 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 514 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 514 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 514 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 514 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 514 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 514 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.ScaledGTestFitter._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 514 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 514 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 514 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 514 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 514 "ms_deisotope/_c/scoring.pyx"
}

#line 514 "ms_deisotope/_c/scoring.pyx"


#line 514 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 514 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17ScaledGTestFitter_1_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 514 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_17ScaledGTestFitter__evaluate[] = "ScaledGTestFitter._evaluate(self, PeakSet peaklist, list observed, list expected) -> double";

#line 514 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_17ScaledGTestFitter_1_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist = 0
#line 514 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_observed = 0
#line 514 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_expected = 0
#line 514 "ms_deisotope/_c/scoring.pyx"
;

#line 514 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 514 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 514 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 514 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 514 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 514 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate (wrapper)", 0);

#line 514 "ms_deisotope/_c/scoring.pyx"
  {

#line 514 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peaklist,&__pyx_n_s_observed,&__pyx_n_s_expected,0};

#line 514 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[3] = {0,0,0};

#line 514 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 514 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 514 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 514 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  3: 
#line 514 "ms_deisotope/_c/scoring.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 514 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 514 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 514 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 514 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 514 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 514 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 514 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 514 "ms_deisotope/_c/scoring.pyx"
      }

#line 514 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 514 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 514 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 514 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaklist)) != 0)) kw_args--;
        else 
#line 514 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 514 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 514 "ms_deisotope/_c/scoring.pyx"
        case  1:

#line 514 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_observed)) != 0)) kw_args--;

#line 514 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, 1); 
#line 514 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 514, __pyx_L3_error)

#line 514 "ms_deisotope/_c/scoring.pyx"
        }

#line 514 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 514 "ms_deisotope/_c/scoring.pyx"
        case  2:

#line 514 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_expected)) != 0)) kw_args--;

#line 514 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, 2); 
#line 514 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 514, __pyx_L3_error)

#line 514 "ms_deisotope/_c/scoring.pyx"
        }

#line 514 "ms_deisotope/_c/scoring.pyx"
      }

#line 514 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 514 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_evaluate") < 0)) __PYX_ERR(0, 514, __pyx_L3_error)

#line 514 "ms_deisotope/_c/scoring.pyx"
      }

#line 514 "ms_deisotope/_c/scoring.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 514 "ms_deisotope/_c/scoring.pyx"
      goto __pyx_L5_argtuple_error;

#line 514 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 514 "ms_deisotope/_c/scoring.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 514 "ms_deisotope/_c/scoring.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 514 "ms_deisotope/_c/scoring.pyx"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 514 "ms_deisotope/_c/scoring.pyx"
    }

#line 514 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_peaklist = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[0]);

#line 514 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_observed = ((PyObject*)values[1]);

#line 514 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_expected = ((PyObject*)values[2]);

#line 514 "ms_deisotope/_c/scoring.pyx"
  }

#line 514 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 514 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 514 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 514, __pyx_L3_error)

#line 514 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 514 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.ScaledGTestFitter._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 514 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 514 "ms_deisotope/_c/scoring.pyx"
  return NULL;

#line 514 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 514 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peaklist), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peaklist", 0))) __PYX_ERR(0, 514, __pyx_L1_error)

#line 514 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_observed), (&PyList_Type), 1, "observed", 1))) __PYX_ERR(0, 514, __pyx_L1_error)

#line 514 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_expected), (&PyList_Type), 1, "expected", 1))) __PYX_ERR(0, 514, __pyx_L1_error)
  __pyx_r = 
#line 514 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_17ScaledGTestFitter__evaluate(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *)__pyx_v_self), __pyx_v_peaklist, __pyx_v_observed, __pyx_v_expected);

#line 514 "ms_deisotope/_c/scoring.pyx"


#line 514 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 514 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 514 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 514 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 514 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 514 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 514 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 514 "ms_deisotope/_c/scoring.pyx"
}

#line 514 "ms_deisotope/_c/scoring.pyx"


#line 514 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_17ScaledGTestFitter__evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected) {

#line 514 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 514 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 514 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 514 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 514 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 514 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 514 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate", 0);

#line 514 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 514 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_12ms_deisotope_2_c_7scoring_17ScaledGTestFitter__evaluate(__pyx_v_self, __pyx_v_peaklist, __pyx_v_observed, __pyx_v_expected, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 514, __pyx_L1_error)

#line 514 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 514 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 514 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 514 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 514 "ms_deisotope/_c/scoring.pyx"


#line 514 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 514 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 514 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 514 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.ScaledGTestFitter._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 514 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 514 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 514 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 514 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 514 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 514 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":540
 * 
 * 
 * cdef double max_intensity_theoretical(list peaklist):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double maximum
 */

#line 540 "ms_deisotope/_c/scoring.pyx"


#line 540 "ms_deisotope/_c/scoring.pyx"
static double __pyx_f_12ms_deisotope_2_c_7scoring_max_intensity_theoretical(PyObject *__pyx_v_peaklist) {
  double __pyx_v_maximum
#line 540 "ms_deisotope/_c/scoring.pyx"
;
  size_t __pyx_v_i
#line 540 "ms_deisotope/_c/scoring.pyx"
;
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_peak = 0
#line 540 "ms_deisotope/_c/scoring.pyx"
;

#line 540 "ms_deisotope/_c/scoring.pyx"
  double __pyx_r;

#line 540 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 540 "ms_deisotope/_c/scoring.pyx"
  Py_ssize_t __pyx_t_1;

#line 540 "ms_deisotope/_c/scoring.pyx"
  Py_ssize_t __pyx_t_2;

#line 540 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_3;

#line 540 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4;

#line 540 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 540 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_6;

#line 540 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_7;

#line 540 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("max_intensity_theoretical", 0);

  /* "ms_deisotope/_c/scoring.pyx":545
 *         size_t i
 *         TheoreticalPeak peak
 *     maximum = 0             # <<<<<<<<<<<<<<
 *     for i in range(PyList_GET_SIZE(peaklist)):
 *         peak = <TheoreticalPeak>PyList_GET_ITEM(peaklist, i)
 */

#line 545 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_maximum = 0.0;

  /* "ms_deisotope/_c/scoring.pyx":546
 *         TheoreticalPeak peak
 *     maximum = 0
 *     for i in range(PyList_GET_SIZE(peaklist)):             # <<<<<<<<<<<<<<
 *         peak = <TheoreticalPeak>PyList_GET_ITEM(peaklist, i)
 *         if peak.intensity > maximum:
 */

#line 546 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyList_GET_SIZE(__pyx_v_peaklist);

#line 546 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = __pyx_t_1;

#line 546 "ms_deisotope/_c/scoring.pyx"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 546 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/scoring.pyx":547
 *     maximum = 0
 *     for i in range(PyList_GET_SIZE(peaklist)):
 *         peak = <TheoreticalPeak>PyList_GET_ITEM(peaklist, i)             # <<<<<<<<<<<<<<
 *         if peak.intensity > maximum:
 *             maximum = peak.intensity
 */

#line 547 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_4 = PyList_GET_ITEM(__pyx_v_peaklist, __pyx_v_i);

#line 547 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = ((PyObject *)__pyx_t_4);

#line 547 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_5);

#line 547 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_5));

#line 547 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = 0;

    /* "ms_deisotope/_c/scoring.pyx":548
 *     for i in range(PyList_GET_SIZE(peaklist)):
 *         peak = <TheoreticalPeak>PyList_GET_ITEM(peaklist, i)
 *         if peak.intensity > maximum:             # <<<<<<<<<<<<<<
 *             maximum = peak.intensity
 *     return maximum
 */

#line 548 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_6 = ((__pyx_v_peak->intensity > __pyx_v_maximum) != 0);

#line 548 "ms_deisotope/_c/scoring.pyx"
    if (__pyx_t_6) {

      /* "ms_deisotope/_c/scoring.pyx":549
 *         peak = <TheoreticalPeak>PyList_GET_ITEM(peaklist, i)
 *         if peak.intensity > maximum:
 *             maximum = peak.intensity             # <<<<<<<<<<<<<<
 *     return maximum
 * 
 */

#line 549 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_7 = __pyx_v_peak->intensity;

#line 549 "ms_deisotope/_c/scoring.pyx"
      __pyx_v_maximum = __pyx_t_7;

      /* "ms_deisotope/_c/scoring.pyx":548
 *     for i in range(PyList_GET_SIZE(peaklist)):
 *         peak = <TheoreticalPeak>PyList_GET_ITEM(peaklist, i)
 *         if peak.intensity > maximum:             # <<<<<<<<<<<<<<
 *             maximum = peak.intensity
 *     return maximum
 */

#line 548 "ms_deisotope/_c/scoring.pyx"
    }

#line 548 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":550
 *         if peak.intensity > maximum:
 *             maximum = peak.intensity
 *     return maximum             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 550 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_v_maximum;

#line 550 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":540
 * 
 * 
 * cdef double max_intensity_theoretical(list peaklist):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double maximum
 */

#line 540 "ms_deisotope/_c/scoring.pyx"


#line 540 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 540 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 540 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 540 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 540 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 540 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":553
 * 
 * 
 * cdef double max_intensity_fitted(list peaklist):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double maximum
 */

#line 553 "ms_deisotope/_c/scoring.pyx"


#line 553 "ms_deisotope/_c/scoring.pyx"
static double __pyx_f_12ms_deisotope_2_c_7scoring_max_intensity_fitted(PyObject *__pyx_v_peaklist) {
  double __pyx_v_maximum
#line 553 "ms_deisotope/_c/scoring.pyx"
;
  size_t __pyx_v_i
#line 553 "ms_deisotope/_c/scoring.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = 0
#line 553 "ms_deisotope/_c/scoring.pyx"
;

#line 553 "ms_deisotope/_c/scoring.pyx"
  double __pyx_r;

#line 553 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 553 "ms_deisotope/_c/scoring.pyx"
  Py_ssize_t __pyx_t_1;

#line 553 "ms_deisotope/_c/scoring.pyx"
  Py_ssize_t __pyx_t_2;

#line 553 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_3;

#line 553 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4;

#line 553 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 553 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_6;

#line 553 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_7;

#line 553 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("max_intensity_fitted", 0);

  /* "ms_deisotope/_c/scoring.pyx":558
 *         size_t i
 *         FittedPeak peak
 *     maximum = 0             # <<<<<<<<<<<<<<
 *     for i in range(PyList_GET_SIZE(peaklist)):
 *         peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
 */

#line 558 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_maximum = 0.0;

  /* "ms_deisotope/_c/scoring.pyx":559
 *         FittedPeak peak
 *     maximum = 0
 *     for i in range(PyList_GET_SIZE(peaklist)):             # <<<<<<<<<<<<<<
 *         peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
 *         if peak.intensity > maximum:
 */

#line 559 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyList_GET_SIZE(__pyx_v_peaklist);

#line 559 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = __pyx_t_1;

#line 559 "ms_deisotope/_c/scoring.pyx"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 559 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/scoring.pyx":560
 *     maximum = 0
 *     for i in range(PyList_GET_SIZE(peaklist)):
 *         peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)             # <<<<<<<<<<<<<<
 *         if peak.intensity > maximum:
 *             maximum = peak.intensity
 */

#line 560 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_4 = PyList_GET_ITEM(__pyx_v_peaklist, __pyx_v_i);

#line 560 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = ((PyObject *)__pyx_t_4);

#line 560 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_5);

#line 560 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_5));

#line 560 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = 0;

    /* "ms_deisotope/_c/scoring.pyx":561
 *     for i in range(PyList_GET_SIZE(peaklist)):
 *         peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
 *         if peak.intensity > maximum:             # <<<<<<<<<<<<<<
 *             maximum = peak.intensity
 *     return maximum
 */

#line 561 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_6 = ((__pyx_v_peak->__pyx_base.intensity > __pyx_v_maximum) != 0);

#line 561 "ms_deisotope/_c/scoring.pyx"
    if (__pyx_t_6) {

      /* "ms_deisotope/_c/scoring.pyx":562
 *         peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
 *         if peak.intensity > maximum:
 *             maximum = peak.intensity             # <<<<<<<<<<<<<<
 *     return maximum
 * 
 */

#line 562 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_7 = __pyx_v_peak->__pyx_base.intensity;

#line 562 "ms_deisotope/_c/scoring.pyx"
      __pyx_v_maximum = __pyx_t_7;

      /* "ms_deisotope/_c/scoring.pyx":561
 *     for i in range(PyList_GET_SIZE(peaklist)):
 *         peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
 *         if peak.intensity > maximum:             # <<<<<<<<<<<<<<
 *             maximum = peak.intensity
 *     return maximum
 */

#line 561 "ms_deisotope/_c/scoring.pyx"
    }

#line 561 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":563
 *         if peak.intensity > maximum:
 *             maximum = peak.intensity
 *     return maximum             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 563 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_v_maximum;

#line 563 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":553
 * 
 * 
 * cdef double max_intensity_fitted(list peaklist):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double maximum
 */

#line 553 "ms_deisotope/_c/scoring.pyx"


#line 553 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 553 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 553 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 553 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 553 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 553 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":583
 *     """
 *     @cython.cdivision
 *     cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double exp_max, theo_max, sum_of_squared_errors, sum_of_squared_theoreticals
 */

#line 583 "ms_deisotope/_c/scoring.pyx"


#line 583 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18LeastSquaresFitter_1_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 583 "ms_deisotope/_c/scoring.pyx"
static double __pyx_f_12ms_deisotope_2_c_7scoring_18LeastSquaresFitter__evaluate(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected, int __pyx_skip_dispatch) {
  double __pyx_v_exp_max
#line 583 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_theo_max
#line 583 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_sum_of_squared_errors
#line 583 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_sum_of_squared_theoreticals
#line 583 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_normed_theo
#line 583 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_normed_expr
#line 583 "ms_deisotope/_c/scoring.pyx"
;
  size_t __pyx_v_i
#line 583 "ms_deisotope/_c/scoring.pyx"
;
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_t = 0
#line 583 "ms_deisotope/_c/scoring.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_e = 0
#line 583 "ms_deisotope/_c/scoring.pyx"
;

#line 583 "ms_deisotope/_c/scoring.pyx"
  double __pyx_r;

#line 583 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 583 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 583 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 583 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 583 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 583 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_5;

#line 583 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 583 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_7;

#line 583 "ms_deisotope/_c/scoring.pyx"
  Py_ssize_t __pyx_t_8;

#line 583 "ms_deisotope/_c/scoring.pyx"
  Py_ssize_t __pyx_t_9;

#line 583 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_10;

#line 583 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_11;

#line 583 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 583 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 583 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 583 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate", 0);

#line 583 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 583 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 583 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 583 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 583 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 583 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 583 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 583 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 583 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 583 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_evaluate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 583, __pyx_L1_error)

#line 583 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 583 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_18LeastSquaresFitter_1_evaluate)) {

#line 583 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 583 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 583 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_5 = 0;

#line 583 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 583 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 583 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 583 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 583 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 583 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 583 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 583 "ms_deisotope/_c/scoring.pyx"
            __pyx_t_5 = 1;

#line 583 "ms_deisotope/_c/scoring.pyx"
          }

#line 583 "ms_deisotope/_c/scoring.pyx"
        }

#line 583 "ms_deisotope/_c/scoring.pyx"
        #if CYTHON_FAST_PYCALL

#line 583 "ms_deisotope/_c/scoring.pyx"
        if (PyFunction_Check(__pyx_t_3)) {

#line 583 "ms_deisotope/_c/scoring.pyx"
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_peaklist), __pyx_v_observed, __pyx_v_expected};

#line 583 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 583, __pyx_L1_error)

#line 583 "ms_deisotope/_c/scoring.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 583 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 583 "ms_deisotope/_c/scoring.pyx"
        } else

#line 583 "ms_deisotope/_c/scoring.pyx"
        #endif

#line 583 "ms_deisotope/_c/scoring.pyx"
        #if CYTHON_FAST_PYCCALL

#line 583 "ms_deisotope/_c/scoring.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {

#line 583 "ms_deisotope/_c/scoring.pyx"
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_peaklist), __pyx_v_observed, __pyx_v_expected};

#line 583 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 583, __pyx_L1_error)

#line 583 "ms_deisotope/_c/scoring.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 583 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 583 "ms_deisotope/_c/scoring.pyx"
        } else

#line 583 "ms_deisotope/_c/scoring.pyx"
        #endif

#line 583 "ms_deisotope/_c/scoring.pyx"
        {

#line 583 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_6 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 583, __pyx_L1_error)

#line 583 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_6);

#line 583 "ms_deisotope/_c/scoring.pyx"
          if (__pyx_t_4) {

#line 583 "ms_deisotope/_c/scoring.pyx"
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;

#line 583 "ms_deisotope/_c/scoring.pyx"
          }

#line 583 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(((PyObject *)__pyx_v_peaklist));

#line 583 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(((PyObject *)__pyx_v_peaklist));

#line 583 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_peaklist));

#line 583 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(__pyx_v_observed);

#line 583 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(__pyx_v_observed);

#line 583 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_observed);

#line 583 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(__pyx_v_expected);

#line 583 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(__pyx_v_expected);

#line 583 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, __pyx_v_expected);

#line 583 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 583, __pyx_L1_error)

#line 583 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 583 "ms_deisotope/_c/scoring.pyx"
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 583 "ms_deisotope/_c/scoring.pyx"
        }

#line 583 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 583 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_7 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_7 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 583, __pyx_L1_error)

#line 583 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 583 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = __pyx_t_7;

#line 583 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 583 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 583 "ms_deisotope/_c/scoring.pyx"
      }

#line 583 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 583 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 583 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 583 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 583 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 583 "ms_deisotope/_c/scoring.pyx"
      }

#line 583 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 583 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 583 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 583 "ms_deisotope/_c/scoring.pyx"
    }

#line 583 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 583 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":591
 *             FittedPeak e
 * 
 *         exp_max = max_intensity_fitted(observed)             # <<<<<<<<<<<<<<
 *         theo_max = max_intensity_theoretical(expected)
 * 
 */

#line 591 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_exp_max = __pyx_f_12ms_deisotope_2_c_7scoring_max_intensity_fitted(__pyx_v_observed);

  /* "ms_deisotope/_c/scoring.pyx":592
 * 
 *         exp_max = max_intensity_fitted(observed)
 *         theo_max = max_intensity_theoretical(expected)             # <<<<<<<<<<<<<<
 * 
 *         sum_of_squared_errors = 0
 */

#line 592 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_theo_max = __pyx_f_12ms_deisotope_2_c_7scoring_max_intensity_theoretical(__pyx_v_expected);

  /* "ms_deisotope/_c/scoring.pyx":594
 *         theo_max = max_intensity_theoretical(expected)
 * 
 *         sum_of_squared_errors = 0             # <<<<<<<<<<<<<<
 *         sum_of_squared_theoreticals = 0
 * 
 */

#line 594 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_sum_of_squared_errors = 0.0;

  /* "ms_deisotope/_c/scoring.pyx":595
 * 
 *         sum_of_squared_errors = 0
 *         sum_of_squared_theoreticals = 0             # <<<<<<<<<<<<<<
 * 
 *         assert len(observed) == len(expected)
 */

#line 595 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_sum_of_squared_theoreticals = 0.0;

  /* "ms_deisotope/_c/scoring.pyx":597
 *         sum_of_squared_theoreticals = 0
 * 
 *         assert len(observed) == len(expected)             # <<<<<<<<<<<<<<
 * 
 *         # for e, t in zip(observed, expected):
 */

#line 597 "ms_deisotope/_c/scoring.pyx"
  #ifndef CYTHON_WITHOUT_ASSERTIONS

#line 597 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!Py_OptimizeFlag)) {

#line 597 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_v_observed == Py_None)) {

#line 597 "ms_deisotope/_c/scoring.pyx"
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 597 "ms_deisotope/_c/scoring.pyx"
      __PYX_ERR(0, 597, __pyx_L1_error)

#line 597 "ms_deisotope/_c/scoring.pyx"
    }

#line 597 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_8 = PyList_GET_SIZE(__pyx_v_observed); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 597, __pyx_L1_error)

#line 597 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_v_expected == Py_None)) {

#line 597 "ms_deisotope/_c/scoring.pyx"
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 597 "ms_deisotope/_c/scoring.pyx"
      __PYX_ERR(0, 597, __pyx_L1_error)

#line 597 "ms_deisotope/_c/scoring.pyx"
    }

#line 597 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_9 = PyList_GET_SIZE(__pyx_v_expected); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 597, __pyx_L1_error)

#line 597 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!((__pyx_t_8 == __pyx_t_9) != 0))) {

#line 597 "ms_deisotope/_c/scoring.pyx"
      PyErr_SetNone(PyExc_AssertionError);

#line 597 "ms_deisotope/_c/scoring.pyx"
      __PYX_ERR(0, 597, __pyx_L1_error)

#line 597 "ms_deisotope/_c/scoring.pyx"
    }

#line 597 "ms_deisotope/_c/scoring.pyx"
  }

#line 597 "ms_deisotope/_c/scoring.pyx"
  #endif

  /* "ms_deisotope/_c/scoring.pyx":600
 * 
 *         # for e, t in zip(observed, expected):
 *         for i in range(PyList_GET_SIZE(observed)):             # <<<<<<<<<<<<<<
 *             e = <FittedPeak>PyList_GetItem(observed, i)
 *             t = <TheoreticalPeak>PyList_GetItem(expected, i)
 */

#line 600 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_9 = PyList_GET_SIZE(__pyx_v_observed);

#line 600 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_8 = __pyx_t_9;

#line 600 "ms_deisotope/_c/scoring.pyx"
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_8; __pyx_t_10+=1) {

#line 600 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_i = __pyx_t_10;

    /* "ms_deisotope/_c/scoring.pyx":601
 *         # for e, t in zip(observed, expected):
 *         for i in range(PyList_GET_SIZE(observed)):
 *             e = <FittedPeak>PyList_GetItem(observed, i)             # <<<<<<<<<<<<<<
 *             t = <TheoreticalPeak>PyList_GetItem(expected, i)
 *             normed_expr = e.intensity / exp_max
 */

#line 601 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_11 = PyList_GetItem(__pyx_v_observed, __pyx_v_i); if (unlikely(__pyx_t_11 == ((PyObject *)NULL))) __PYX_ERR(0, 601, __pyx_L1_error)

#line 601 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_11);

#line 601 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 601 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF_SET(__pyx_v_e, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1));

#line 601 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/scoring.pyx":602
 *         for i in range(PyList_GET_SIZE(observed)):
 *             e = <FittedPeak>PyList_GetItem(observed, i)
 *             t = <TheoreticalPeak>PyList_GetItem(expected, i)             # <<<<<<<<<<<<<<
 *             normed_expr = e.intensity / exp_max
 *             normed_theo = t.intensity / theo_max
 */

#line 602 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_11 = PyList_GetItem(__pyx_v_expected, __pyx_v_i); if (unlikely(__pyx_t_11 == ((PyObject *)NULL))) __PYX_ERR(0, 602, __pyx_L1_error)

#line 602 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_11);

#line 602 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 602 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF_SET(__pyx_v_t, ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_1));

#line 602 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/scoring.pyx":603
 *             e = <FittedPeak>PyList_GetItem(observed, i)
 *             t = <TheoreticalPeak>PyList_GetItem(expected, i)
 *             normed_expr = e.intensity / exp_max             # <<<<<<<<<<<<<<
 *             normed_theo = t.intensity / theo_max
 *             sum_of_squared_errors += (normed_theo - normed_expr) ** 2
 */

#line 603 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_normed_expr = (__pyx_v_e->__pyx_base.intensity / __pyx_v_exp_max);

    /* "ms_deisotope/_c/scoring.pyx":604
 *             t = <TheoreticalPeak>PyList_GetItem(expected, i)
 *             normed_expr = e.intensity / exp_max
 *             normed_theo = t.intensity / theo_max             # <<<<<<<<<<<<<<
 *             sum_of_squared_errors += (normed_theo - normed_expr) ** 2
 *             sum_of_squared_theoreticals += normed_theo ** 2
 */

#line 604 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_normed_theo = (__pyx_v_t->intensity / __pyx_v_theo_max);

    /* "ms_deisotope/_c/scoring.pyx":605
 *             normed_expr = e.intensity / exp_max
 *             normed_theo = t.intensity / theo_max
 *             sum_of_squared_errors += (normed_theo - normed_expr) ** 2             # <<<<<<<<<<<<<<
 *             sum_of_squared_theoreticals += normed_theo ** 2
 *         return sum_of_squared_errors / sum_of_squared_theoreticals
 */

#line 605 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_sum_of_squared_errors = (__pyx_v_sum_of_squared_errors + pow((__pyx_v_normed_theo - __pyx_v_normed_expr), 2.0));

    /* "ms_deisotope/_c/scoring.pyx":606
 *             normed_theo = t.intensity / theo_max
 *             sum_of_squared_errors += (normed_theo - normed_expr) ** 2
 *             sum_of_squared_theoreticals += normed_theo ** 2             # <<<<<<<<<<<<<<
 *         return sum_of_squared_errors / sum_of_squared_theoreticals
 * 
 */

#line 606 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_sum_of_squared_theoreticals = (__pyx_v_sum_of_squared_theoreticals + pow(__pyx_v_normed_theo, 2.0));

#line 606 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":607
 *             sum_of_squared_errors += (normed_theo - normed_expr) ** 2
 *             sum_of_squared_theoreticals += normed_theo ** 2
 *         return sum_of_squared_errors / sum_of_squared_theoreticals             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 607 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = (__pyx_v_sum_of_squared_errors / __pyx_v_sum_of_squared_theoreticals);

#line 607 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":583
 *     """
 *     @cython.cdivision
 *     cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double exp_max, theo_max, sum_of_squared_errors, sum_of_squared_theoreticals
 */

#line 583 "ms_deisotope/_c/scoring.pyx"


#line 583 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 583 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 583 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 583 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 583 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 583 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 583 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 583 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.LeastSquaresFitter._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 583 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 583 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 583 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_t);

#line 583 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_e);

#line 583 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 583 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 583 "ms_deisotope/_c/scoring.pyx"
}

#line 583 "ms_deisotope/_c/scoring.pyx"


#line 583 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 583 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18LeastSquaresFitter_1_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 583 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_18LeastSquaresFitter__evaluate[] = "LeastSquaresFitter._evaluate(self, PeakSet peaklist, list observed, list expected) -> double";

#line 583 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_18LeastSquaresFitter_1_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist = 0
#line 583 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_observed = 0
#line 583 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_expected = 0
#line 583 "ms_deisotope/_c/scoring.pyx"
;

#line 583 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 583 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 583 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 583 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 583 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 583 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate (wrapper)", 0);

#line 583 "ms_deisotope/_c/scoring.pyx"
  {

#line 583 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peaklist,&__pyx_n_s_observed,&__pyx_n_s_expected,0};

#line 583 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[3] = {0,0,0};

#line 583 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 583 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 583 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 583 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  3: 
#line 583 "ms_deisotope/_c/scoring.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 583 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 583 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 583 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 583 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 583 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 583 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 583 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 583 "ms_deisotope/_c/scoring.pyx"
      }

#line 583 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 583 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 583 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 583 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaklist)) != 0)) kw_args--;
        else 
#line 583 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 583 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 583 "ms_deisotope/_c/scoring.pyx"
        case  1:

#line 583 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_observed)) != 0)) kw_args--;

#line 583 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, 1); 
#line 583 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 583, __pyx_L3_error)

#line 583 "ms_deisotope/_c/scoring.pyx"
        }

#line 583 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 583 "ms_deisotope/_c/scoring.pyx"
        case  2:

#line 583 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_expected)) != 0)) kw_args--;

#line 583 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, 2); 
#line 583 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 583, __pyx_L3_error)

#line 583 "ms_deisotope/_c/scoring.pyx"
        }

#line 583 "ms_deisotope/_c/scoring.pyx"
      }

#line 583 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 583 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_evaluate") < 0)) __PYX_ERR(0, 583, __pyx_L3_error)

#line 583 "ms_deisotope/_c/scoring.pyx"
      }

#line 583 "ms_deisotope/_c/scoring.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 583 "ms_deisotope/_c/scoring.pyx"
      goto __pyx_L5_argtuple_error;

#line 583 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 583 "ms_deisotope/_c/scoring.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 583 "ms_deisotope/_c/scoring.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 583 "ms_deisotope/_c/scoring.pyx"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 583 "ms_deisotope/_c/scoring.pyx"
    }

#line 583 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_peaklist = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[0]);

#line 583 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_observed = ((PyObject*)values[1]);

#line 583 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_expected = ((PyObject*)values[2]);

#line 583 "ms_deisotope/_c/scoring.pyx"
  }

#line 583 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 583 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 583 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 583, __pyx_L3_error)

#line 583 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 583 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.LeastSquaresFitter._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 583 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 583 "ms_deisotope/_c/scoring.pyx"
  return NULL;

#line 583 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 583 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peaklist), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peaklist", 0))) __PYX_ERR(0, 583, __pyx_L1_error)

#line 583 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_observed), (&PyList_Type), 1, "observed", 1))) __PYX_ERR(0, 583, __pyx_L1_error)

#line 583 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_expected), (&PyList_Type), 1, "expected", 1))) __PYX_ERR(0, 583, __pyx_L1_error)
  __pyx_r = 
#line 583 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_18LeastSquaresFitter__evaluate(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter *)__pyx_v_self), __pyx_v_peaklist, __pyx_v_observed, __pyx_v_expected);

#line 583 "ms_deisotope/_c/scoring.pyx"


#line 583 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 583 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 583 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 583 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 583 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 583 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 583 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 583 "ms_deisotope/_c/scoring.pyx"
}

#line 583 "ms_deisotope/_c/scoring.pyx"


#line 583 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_18LeastSquaresFitter__evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected) {

#line 583 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 583 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 583 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 583 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 583 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 583 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 583 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate", 0);

#line 583 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 583 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_12ms_deisotope_2_c_7scoring_18LeastSquaresFitter__evaluate(__pyx_v_self, __pyx_v_peaklist, __pyx_v_observed, __pyx_v_expected, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 583, __pyx_L1_error)

#line 583 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 583 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 583 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 583 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 583 "ms_deisotope/_c/scoring.pyx"


#line 583 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 583 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 583 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 583 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.LeastSquaresFitter._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 583 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 583 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 583 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 583 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 583 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 583 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":618
 * 
 *     @cython.cdivision
 *     cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double running_total
 */

#line 618 "ms_deisotope/_c/scoring.pyx"


#line 618 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_15ChiSquareFitter_1_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 618 "ms_deisotope/_c/scoring.pyx"
static double __pyx_f_12ms_deisotope_2_c_7scoring_15ChiSquareFitter__evaluate(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_ChiSquareFitter *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected, int __pyx_skip_dispatch) {
  double __pyx_v_running_total
#line 618 "ms_deisotope/_c/scoring.pyx"
;
  size_t __pyx_v_i
#line 618 "ms_deisotope/_c/scoring.pyx"
;
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_t = 0
#line 618 "ms_deisotope/_c/scoring.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_e = 0
#line 618 "ms_deisotope/_c/scoring.pyx"
;

#line 618 "ms_deisotope/_c/scoring.pyx"
  double __pyx_r;

#line 618 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 618 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 618 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 618 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 618 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 618 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_5;

#line 618 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 618 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_7;

#line 618 "ms_deisotope/_c/scoring.pyx"
  Py_ssize_t __pyx_t_8;

#line 618 "ms_deisotope/_c/scoring.pyx"
  Py_ssize_t __pyx_t_9;

#line 618 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_10;

#line 618 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_11;

#line 618 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 618 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 618 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 618 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate", 0);

#line 618 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 618 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 618 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 618 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 618 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 618 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 618 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 618 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 618 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 618 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_evaluate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 618, __pyx_L1_error)

#line 618 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 618 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_15ChiSquareFitter_1_evaluate)) {

#line 618 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 618 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 618 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_5 = 0;

#line 618 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 618 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 618 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 618 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 618 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 618 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 618 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 618 "ms_deisotope/_c/scoring.pyx"
            __pyx_t_5 = 1;

#line 618 "ms_deisotope/_c/scoring.pyx"
          }

#line 618 "ms_deisotope/_c/scoring.pyx"
        }

#line 618 "ms_deisotope/_c/scoring.pyx"
        #if CYTHON_FAST_PYCALL

#line 618 "ms_deisotope/_c/scoring.pyx"
        if (PyFunction_Check(__pyx_t_3)) {

#line 618 "ms_deisotope/_c/scoring.pyx"
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_peaklist), __pyx_v_observed, __pyx_v_expected};

#line 618 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 618, __pyx_L1_error)

#line 618 "ms_deisotope/_c/scoring.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 618 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 618 "ms_deisotope/_c/scoring.pyx"
        } else

#line 618 "ms_deisotope/_c/scoring.pyx"
        #endif

#line 618 "ms_deisotope/_c/scoring.pyx"
        #if CYTHON_FAST_PYCCALL

#line 618 "ms_deisotope/_c/scoring.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {

#line 618 "ms_deisotope/_c/scoring.pyx"
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_peaklist), __pyx_v_observed, __pyx_v_expected};

#line 618 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 618, __pyx_L1_error)

#line 618 "ms_deisotope/_c/scoring.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 618 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 618 "ms_deisotope/_c/scoring.pyx"
        } else

#line 618 "ms_deisotope/_c/scoring.pyx"
        #endif

#line 618 "ms_deisotope/_c/scoring.pyx"
        {

#line 618 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_6 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 618, __pyx_L1_error)

#line 618 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_6);

#line 618 "ms_deisotope/_c/scoring.pyx"
          if (__pyx_t_4) {

#line 618 "ms_deisotope/_c/scoring.pyx"
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;

#line 618 "ms_deisotope/_c/scoring.pyx"
          }

#line 618 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(((PyObject *)__pyx_v_peaklist));

#line 618 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(((PyObject *)__pyx_v_peaklist));

#line 618 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_peaklist));

#line 618 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(__pyx_v_observed);

#line 618 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(__pyx_v_observed);

#line 618 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_observed);

#line 618 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(__pyx_v_expected);

#line 618 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(__pyx_v_expected);

#line 618 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, __pyx_v_expected);

#line 618 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 618, __pyx_L1_error)

#line 618 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 618 "ms_deisotope/_c/scoring.pyx"
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 618 "ms_deisotope/_c/scoring.pyx"
        }

#line 618 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 618 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_7 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_7 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 618, __pyx_L1_error)

#line 618 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 618 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = __pyx_t_7;

#line 618 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 618 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 618 "ms_deisotope/_c/scoring.pyx"
      }

#line 618 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 618 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 618 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 618 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 618 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 618 "ms_deisotope/_c/scoring.pyx"
      }

#line 618 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 618 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 618 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 618 "ms_deisotope/_c/scoring.pyx"
    }

#line 618 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 618 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":624
 *             TheoreticalPeak t
 *             FittedPeak e
 *         running_total = 0             # <<<<<<<<<<<<<<
 *         for i in range(PyList_GET_SIZE(observed)):
 *             e = <FittedPeak>PyList_GetItem(observed, i)
 */

#line 624 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_running_total = 0.0;

  /* "ms_deisotope/_c/scoring.pyx":625
 *             FittedPeak e
 *         running_total = 0
 *         for i in range(PyList_GET_SIZE(observed)):             # <<<<<<<<<<<<<<
 *             e = <FittedPeak>PyList_GetItem(observed, i)
 *             t = <TheoreticalPeak>PyList_GetItem(expected, i)
 */

#line 625 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_8 = PyList_GET_SIZE(__pyx_v_observed);

#line 625 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_9 = __pyx_t_8;

#line 625 "ms_deisotope/_c/scoring.pyx"
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {

#line 625 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_i = __pyx_t_10;

    /* "ms_deisotope/_c/scoring.pyx":626
 *         running_total = 0
 *         for i in range(PyList_GET_SIZE(observed)):
 *             e = <FittedPeak>PyList_GetItem(observed, i)             # <<<<<<<<<<<<<<
 *             t = <TheoreticalPeak>PyList_GetItem(expected, i)
 *             running_total += ((e - t) ** 2) / t
 */

#line 626 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_11 = PyList_GetItem(__pyx_v_observed, __pyx_v_i); if (unlikely(__pyx_t_11 == ((PyObject *)NULL))) __PYX_ERR(0, 626, __pyx_L1_error)

#line 626 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_11);

#line 626 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 626 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF_SET(__pyx_v_e, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1));

#line 626 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/scoring.pyx":627
 *         for i in range(PyList_GET_SIZE(observed)):
 *             e = <FittedPeak>PyList_GetItem(observed, i)
 *             t = <TheoreticalPeak>PyList_GetItem(expected, i)             # <<<<<<<<<<<<<<
 *             running_total += ((e - t) ** 2) / t
 *         return running_total
 */

#line 627 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_11 = PyList_GetItem(__pyx_v_expected, __pyx_v_i); if (unlikely(__pyx_t_11 == ((PyObject *)NULL))) __PYX_ERR(0, 627, __pyx_L1_error)

#line 627 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_11);

#line 627 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 627 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF_SET(__pyx_v_t, ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_1));

#line 627 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/scoring.pyx":628
 *             e = <FittedPeak>PyList_GetItem(observed, i)
 *             t = <TheoreticalPeak>PyList_GetItem(expected, i)
 *             running_total += ((e - t) ** 2) / t             # <<<<<<<<<<<<<<
 *         return running_total
 * 
 */

#line 628 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = PyFloat_FromDouble(__pyx_v_running_total); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 628, __pyx_L1_error)

#line 628 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 628 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_2 = PyNumber_Subtract(((PyObject *)__pyx_v_e), ((PyObject *)__pyx_v_t)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 628, __pyx_L1_error)

#line 628 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 628 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_3 = PyNumber_Power(__pyx_t_2, __pyx_int_2, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 628, __pyx_L1_error)

#line 628 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 628 "ms_deisotope/_c/scoring.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 628 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_2 = __Pyx_PyNumber_Divide(__pyx_t_3, ((PyObject *)__pyx_v_t)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 628, __pyx_L1_error)

#line 628 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 628 "ms_deisotope/_c/scoring.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 628 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 628, __pyx_L1_error)

#line 628 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 628 "ms_deisotope/_c/scoring.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 628 "ms_deisotope/_c/scoring.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 628 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_7 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_7 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 628, __pyx_L1_error)

#line 628 "ms_deisotope/_c/scoring.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 628 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_running_total = __pyx_t_7;

#line 628 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":629
 *             t = <TheoreticalPeak>PyList_GetItem(expected, i)
 *             running_total += ((e - t) ** 2) / t
 *         return running_total             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 629 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_v_running_total;

#line 629 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":618
 * 
 *     @cython.cdivision
 *     cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double running_total
 */

#line 618 "ms_deisotope/_c/scoring.pyx"


#line 618 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 618 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 618 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 618 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 618 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 618 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 618 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 618 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.ChiSquareFitter._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 618 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 618 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 618 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_t);

#line 618 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_e);

#line 618 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 618 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 618 "ms_deisotope/_c/scoring.pyx"
}

#line 618 "ms_deisotope/_c/scoring.pyx"


#line 618 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 618 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_15ChiSquareFitter_1_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 618 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_15ChiSquareFitter__evaluate[] = "ChiSquareFitter._evaluate(self, PeakSet peaklist, list observed, list expected) -> double";

#line 618 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_15ChiSquareFitter_1_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist = 0
#line 618 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_observed = 0
#line 618 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_expected = 0
#line 618 "ms_deisotope/_c/scoring.pyx"
;

#line 618 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 618 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 618 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 618 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 618 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 618 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate (wrapper)", 0);

#line 618 "ms_deisotope/_c/scoring.pyx"
  {

#line 618 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peaklist,&__pyx_n_s_observed,&__pyx_n_s_expected,0};

#line 618 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[3] = {0,0,0};

#line 618 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 618 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 618 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 618 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  3: 
#line 618 "ms_deisotope/_c/scoring.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 618 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 618 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 618 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 618 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 618 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 618 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 618 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 618 "ms_deisotope/_c/scoring.pyx"
      }

#line 618 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 618 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 618 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 618 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaklist)) != 0)) kw_args--;
        else 
#line 618 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 618 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 618 "ms_deisotope/_c/scoring.pyx"
        case  1:

#line 618 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_observed)) != 0)) kw_args--;

#line 618 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, 1); 
#line 618 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 618, __pyx_L3_error)

#line 618 "ms_deisotope/_c/scoring.pyx"
        }

#line 618 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 618 "ms_deisotope/_c/scoring.pyx"
        case  2:

#line 618 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_expected)) != 0)) kw_args--;

#line 618 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, 2); 
#line 618 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 618, __pyx_L3_error)

#line 618 "ms_deisotope/_c/scoring.pyx"
        }

#line 618 "ms_deisotope/_c/scoring.pyx"
      }

#line 618 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 618 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_evaluate") < 0)) __PYX_ERR(0, 618, __pyx_L3_error)

#line 618 "ms_deisotope/_c/scoring.pyx"
      }

#line 618 "ms_deisotope/_c/scoring.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 618 "ms_deisotope/_c/scoring.pyx"
      goto __pyx_L5_argtuple_error;

#line 618 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 618 "ms_deisotope/_c/scoring.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 618 "ms_deisotope/_c/scoring.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 618 "ms_deisotope/_c/scoring.pyx"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 618 "ms_deisotope/_c/scoring.pyx"
    }

#line 618 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_peaklist = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[0]);

#line 618 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_observed = ((PyObject*)values[1]);

#line 618 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_expected = ((PyObject*)values[2]);

#line 618 "ms_deisotope/_c/scoring.pyx"
  }

#line 618 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 618 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 618 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 618, __pyx_L3_error)

#line 618 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 618 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.ChiSquareFitter._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 618 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 618 "ms_deisotope/_c/scoring.pyx"
  return NULL;

#line 618 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 618 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peaklist), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peaklist", 0))) __PYX_ERR(0, 618, __pyx_L1_error)

#line 618 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_observed), (&PyList_Type), 1, "observed", 1))) __PYX_ERR(0, 618, __pyx_L1_error)

#line 618 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_expected), (&PyList_Type), 1, "expected", 1))) __PYX_ERR(0, 618, __pyx_L1_error)
  __pyx_r = 
#line 618 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_15ChiSquareFitter__evaluate(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_ChiSquareFitter *)__pyx_v_self), __pyx_v_peaklist, __pyx_v_observed, __pyx_v_expected);

#line 618 "ms_deisotope/_c/scoring.pyx"


#line 618 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 618 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 618 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 618 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 618 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 618 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 618 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 618 "ms_deisotope/_c/scoring.pyx"
}

#line 618 "ms_deisotope/_c/scoring.pyx"


#line 618 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_15ChiSquareFitter__evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ChiSquareFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected) {

#line 618 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 618 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 618 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 618 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 618 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 618 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 618 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate", 0);

#line 618 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 618 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_12ms_deisotope_2_c_7scoring_15ChiSquareFitter__evaluate(__pyx_v_self, __pyx_v_peaklist, __pyx_v_observed, __pyx_v_expected, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 618, __pyx_L1_error)

#line 618 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 618 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 618 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 618 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 618 "ms_deisotope/_c/scoring.pyx"


#line 618 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 618 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 618 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 618 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.ChiSquareFitter._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 618 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 618 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 618 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 618 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 618 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 618 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":633
 * 
 * @cython.cdivision
 * cdef double ms_deconv_score_peak(FittedPeak obs, TheoreticalPeak theo, double mass_error_tolerance=0.02, double minimum_signal_to_noise=1) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         double mass_error, abundance_diff
 */

#line 633 "ms_deisotope/_c/scoring.pyx"


#line 633 "ms_deisotope/_c/scoring.pyx"
static double __pyx_f_12ms_deisotope_2_c_7scoring_ms_deconv_score_peak(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_obs, struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_theo, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_ms_deconv_score_peak *__pyx_optional_args) {

#line 633 "ms_deisotope/_c/scoring.pyx"
  double __pyx_v_mass_error_tolerance = ((double)0.02);

#line 633 "ms_deisotope/_c/scoring.pyx"
  double __pyx_v_minimum_signal_to_noise = ((double)1.0);
  double __pyx_v_mass_error
#line 633 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_abundance_diff
#line 633 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_mass_accuracy
#line 633 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_score
#line 633 "ms_deisotope/_c/scoring.pyx"
;

#line 633 "ms_deisotope/_c/scoring.pyx"
  double __pyx_r;

#line 633 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_1;

#line 633 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_2;

#line 633 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_optional_args) {

#line 633 "ms_deisotope/_c/scoring.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 633 "ms_deisotope/_c/scoring.pyx"
      __pyx_v_mass_error_tolerance = __pyx_optional_args->mass_error_tolerance;

#line 633 "ms_deisotope/_c/scoring.pyx"
      if (__pyx_optional_args->__pyx_n > 1) {

#line 633 "ms_deisotope/_c/scoring.pyx"
        __pyx_v_minimum_signal_to_noise = __pyx_optional_args->minimum_signal_to_noise;

#line 633 "ms_deisotope/_c/scoring.pyx"
      }

#line 633 "ms_deisotope/_c/scoring.pyx"
    }

#line 633 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":636
 *     cdef:
 *         double mass_error, abundance_diff
 *     if obs.signal_to_noise < minimum_signal_to_noise:             # <<<<<<<<<<<<<<
 *         return 0.
 * 
 */

#line 636 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = ((__pyx_v_obs->signal_to_noise < __pyx_v_minimum_signal_to_noise) != 0);

#line 636 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/scoring.pyx":637
 *         double mass_error, abundance_diff
 *     if obs.signal_to_noise < minimum_signal_to_noise:
 *         return 0.             # <<<<<<<<<<<<<<
 * 
 *     mass_error = fabs(obs.mz - theo.mz)
 */

#line 637 "ms_deisotope/_c/scoring.pyx"
    __pyx_r = 0.;

#line 637 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/scoring.pyx":636
 *     cdef:
 *         double mass_error, abundance_diff
 *     if obs.signal_to_noise < minimum_signal_to_noise:             # <<<<<<<<<<<<<<
 *         return 0.
 * 
 */

#line 636 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":639
 *         return 0.
 * 
 *     mass_error = fabs(obs.mz - theo.mz)             # <<<<<<<<<<<<<<
 * 
 *     if mass_error <= mass_error_tolerance:
 */

#line 639 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_mass_error = fabs((__pyx_v_obs->__pyx_base.mz - __pyx_v_theo->mz));

  /* "ms_deisotope/_c/scoring.pyx":641
 *     mass_error = fabs(obs.mz - theo.mz)
 * 
 *     if mass_error <= mass_error_tolerance:             # <<<<<<<<<<<<<<
 *         mass_accuracy = 1 - mass_error / mass_error_tolerance
 *     else:
 */

#line 641 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = ((__pyx_v_mass_error <= __pyx_v_mass_error_tolerance) != 0);

#line 641 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/scoring.pyx":642
 * 
 *     if mass_error <= mass_error_tolerance:
 *         mass_accuracy = 1 - mass_error / mass_error_tolerance             # <<<<<<<<<<<<<<
 *     else:
 *         return 0.
 */

#line 642 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_mass_accuracy = (1.0 - (__pyx_v_mass_error / __pyx_v_mass_error_tolerance));

    /* "ms_deisotope/_c/scoring.pyx":641
 *     mass_error = fabs(obs.mz - theo.mz)
 * 
 *     if mass_error <= mass_error_tolerance:             # <<<<<<<<<<<<<<
 *         mass_accuracy = 1 - mass_error / mass_error_tolerance
 *     else:
 */

#line 641 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L4;

#line 641 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":644
 *         mass_accuracy = 1 - mass_error / mass_error_tolerance
 *     else:
 *         return 0.             # <<<<<<<<<<<<<<
 * 
 *     if obs.intensity < theo.intensity and (((theo.intensity - obs.intensity) / obs.intensity) <= 1):
 */

#line 644 "ms_deisotope/_c/scoring.pyx"
  /*else*/ {

#line 644 "ms_deisotope/_c/scoring.pyx"
    __pyx_r = 0.;

#line 644 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L0;

#line 644 "ms_deisotope/_c/scoring.pyx"
  }

#line 644 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4:;

  /* "ms_deisotope/_c/scoring.pyx":646
 *         return 0.
 * 
 *     if obs.intensity < theo.intensity and (((theo.intensity - obs.intensity) / obs.intensity) <= 1):             # <<<<<<<<<<<<<<
 *         abundance_diff = 1 - ((theo.intensity - obs.intensity) / obs.intensity)
 *     elif obs.intensity >= theo.intensity and (((obs.intensity - theo.intensity) / obs.intensity) <= 1):
 */

#line 646 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = ((__pyx_v_obs->__pyx_base.intensity < __pyx_v_theo->intensity) != 0);

#line 646 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_t_2) {

#line 646 "ms_deisotope/_c/scoring.pyx"
  } else {

#line 646 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = __pyx_t_2;

#line 646 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L6_bool_binop_done;

#line 646 "ms_deisotope/_c/scoring.pyx"
  }

#line 646 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = ((((__pyx_v_theo->intensity - __pyx_v_obs->__pyx_base.intensity) / __pyx_v_obs->__pyx_base.intensity) <= 1.0) != 0);

#line 646 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __pyx_t_2;

#line 646 "ms_deisotope/_c/scoring.pyx"
  __pyx_L6_bool_binop_done:;

#line 646 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/scoring.pyx":647
 * 
 *     if obs.intensity < theo.intensity and (((theo.intensity - obs.intensity) / obs.intensity) <= 1):
 *         abundance_diff = 1 - ((theo.intensity - obs.intensity) / obs.intensity)             # <<<<<<<<<<<<<<
 *     elif obs.intensity >= theo.intensity and (((obs.intensity - theo.intensity) / obs.intensity) <= 1):
 *         abundance_diff = sqrt(1 - ((obs.intensity - theo.intensity) / obs.intensity))
 */

#line 647 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_abundance_diff = (1.0 - ((__pyx_v_theo->intensity - __pyx_v_obs->__pyx_base.intensity) / __pyx_v_obs->__pyx_base.intensity));

    /* "ms_deisotope/_c/scoring.pyx":646
 *         return 0.
 * 
 *     if obs.intensity < theo.intensity and (((theo.intensity - obs.intensity) / obs.intensity) <= 1):             # <<<<<<<<<<<<<<
 *         abundance_diff = 1 - ((theo.intensity - obs.intensity) / obs.intensity)
 *     elif obs.intensity >= theo.intensity and (((obs.intensity - theo.intensity) / obs.intensity) <= 1):
 */

#line 646 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L5;

#line 646 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":648
 *     if obs.intensity < theo.intensity and (((theo.intensity - obs.intensity) / obs.intensity) <= 1):
 *         abundance_diff = 1 - ((theo.intensity - obs.intensity) / obs.intensity)
 *     elif obs.intensity >= theo.intensity and (((obs.intensity - theo.intensity) / obs.intensity) <= 1):             # <<<<<<<<<<<<<<
 *         abundance_diff = sqrt(1 - ((obs.intensity - theo.intensity) / obs.intensity))
 *     else:
 */

#line 648 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = ((__pyx_v_obs->__pyx_base.intensity >= __pyx_v_theo->intensity) != 0);

#line 648 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_t_2) {

#line 648 "ms_deisotope/_c/scoring.pyx"
  } else {

#line 648 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = __pyx_t_2;

#line 648 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L8_bool_binop_done;

#line 648 "ms_deisotope/_c/scoring.pyx"
  }

#line 648 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = ((((__pyx_v_obs->__pyx_base.intensity - __pyx_v_theo->intensity) / __pyx_v_obs->__pyx_base.intensity) <= 1.0) != 0);

#line 648 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __pyx_t_2;

#line 648 "ms_deisotope/_c/scoring.pyx"
  __pyx_L8_bool_binop_done:;

#line 648 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/scoring.pyx":649
 *         abundance_diff = 1 - ((theo.intensity - obs.intensity) / obs.intensity)
 *     elif obs.intensity >= theo.intensity and (((obs.intensity - theo.intensity) / obs.intensity) <= 1):
 *         abundance_diff = sqrt(1 - ((obs.intensity - theo.intensity) / obs.intensity))             # <<<<<<<<<<<<<<
 *     else:
 *         return 0.
 */

#line 649 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_abundance_diff = sqrt((1.0 - ((__pyx_v_obs->__pyx_base.intensity - __pyx_v_theo->intensity) / __pyx_v_obs->__pyx_base.intensity)));

    /* "ms_deisotope/_c/scoring.pyx":648
 *     if obs.intensity < theo.intensity and (((theo.intensity - obs.intensity) / obs.intensity) <= 1):
 *         abundance_diff = 1 - ((theo.intensity - obs.intensity) / obs.intensity)
 *     elif obs.intensity >= theo.intensity and (((obs.intensity - theo.intensity) / obs.intensity) <= 1):             # <<<<<<<<<<<<<<
 *         abundance_diff = sqrt(1 - ((obs.intensity - theo.intensity) / obs.intensity))
 *     else:
 */

#line 648 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L5;

#line 648 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":651
 *         abundance_diff = sqrt(1 - ((obs.intensity - theo.intensity) / obs.intensity))
 *     else:
 *         return 0.             # <<<<<<<<<<<<<<
 *     score = sqrt(theo.intensity) * mass_accuracy * abundance_diff
 *     return score
 */

#line 651 "ms_deisotope/_c/scoring.pyx"
  /*else*/ {

#line 651 "ms_deisotope/_c/scoring.pyx"
    __pyx_r = 0.;

#line 651 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L0;

#line 651 "ms_deisotope/_c/scoring.pyx"
  }

#line 651 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5:;

  /* "ms_deisotope/_c/scoring.pyx":652
 *     else:
 *         return 0.
 *     score = sqrt(theo.intensity) * mass_accuracy * abundance_diff             # <<<<<<<<<<<<<<
 *     return score
 * 
 */

#line 652 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_score = ((sqrt(__pyx_v_theo->intensity) * __pyx_v_mass_accuracy) * __pyx_v_abundance_diff);

  /* "ms_deisotope/_c/scoring.pyx":653
 *         return 0.
 *     score = sqrt(theo.intensity) * mass_accuracy * abundance_diff
 *     return score             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 653 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_v_score;

#line 653 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":633
 * 
 * @cython.cdivision
 * cdef double ms_deconv_score_peak(FittedPeak obs, TheoreticalPeak theo, double mass_error_tolerance=0.02, double minimum_signal_to_noise=1) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         double mass_error, abundance_diff
 */

#line 633 "ms_deisotope/_c/scoring.pyx"


#line 633 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 633 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 633 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 633 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":684
 *     Proteomics : MCP, 9(12), 27722782. https://doi.org/10.1074/mcp.M110.002766
 *     """
 *     def __init__(self, minimum_score=10, mass_error_tolerance=0.02):             # <<<<<<<<<<<<<<
 *         self.mass_error_tolerance = mass_error_tolerance
 *         self.select = MaximizeFitSelector()
 */

#line 684 "ms_deisotope/_c/scoring.pyx"


#line 684 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 684 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 684 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_minimum_score = 0
#line 684 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_mass_error_tolerance = 0
#line 684 "ms_deisotope/_c/scoring.pyx"
;

#line 684 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 684 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 684 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 684 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 684 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 684 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 684 "ms_deisotope/_c/scoring.pyx"
  {

#line 684 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_minimum_score,&__pyx_n_s_mass_error_tolerance,0};

#line 684 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[2] = {0,0};

#line 684 "ms_deisotope/_c/scoring.pyx"
    values[0] = ((PyObject *)__pyx_int_10);

#line 684 "ms_deisotope/_c/scoring.pyx"
    values[1] = ((PyObject *)__pyx_float_0_02);

#line 684 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 684 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 684 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 684 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  2: 
#line 684 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 684 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 684 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 684 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 684 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 684 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 684 "ms_deisotope/_c/scoring.pyx"
      }

#line 684 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 684 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 684 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 684 "ms_deisotope/_c/scoring.pyx"
        if (kw_args > 0) {

#line 684 "ms_deisotope/_c/scoring.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_minimum_score);

#line 684 "ms_deisotope/_c/scoring.pyx"
          if (value) { values[0] = value; kw_args--; }

#line 684 "ms_deisotope/_c/scoring.pyx"
        }

#line 684 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 684 "ms_deisotope/_c/scoring.pyx"
        case  1:

#line 684 "ms_deisotope/_c/scoring.pyx"
        if (kw_args > 0) {

#line 684 "ms_deisotope/_c/scoring.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mass_error_tolerance);

#line 684 "ms_deisotope/_c/scoring.pyx"
          if (value) { values[1] = value; kw_args--; }

#line 684 "ms_deisotope/_c/scoring.pyx"
        }

#line 684 "ms_deisotope/_c/scoring.pyx"
      }

#line 684 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 684 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 684, __pyx_L3_error)

#line 684 "ms_deisotope/_c/scoring.pyx"
      }

#line 684 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 684 "ms_deisotope/_c/scoring.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: 
#line 684 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 684 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 684 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 684 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  0: 
#line 684 "ms_deisotope/_c/scoring.pyx"
break;
        default: 
#line 684 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 684 "ms_deisotope/_c/scoring.pyx"
      }

#line 684 "ms_deisotope/_c/scoring.pyx"
    }

#line 684 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_minimum_score = values[0];

#line 684 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_mass_error_tolerance = values[1];

#line 684 "ms_deisotope/_c/scoring.pyx"
  }

#line 684 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 684 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 684 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 684, __pyx_L3_error)

#line 684 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 684 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.MSDeconVFitter.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 684 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 684 "ms_deisotope/_c/scoring.pyx"
  return -1;

#line 684 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 684 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_14MSDeconVFitter___init__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *)__pyx_v_self), __pyx_v_minimum_score, __pyx_v_mass_error_tolerance);

#line 684 "ms_deisotope/_c/scoring.pyx"


#line 684 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 684 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 684 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 684 "ms_deisotope/_c/scoring.pyx"
}

#line 684 "ms_deisotope/_c/scoring.pyx"


#line 684 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_14MSDeconVFitter___init__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *__pyx_v_self, PyObject *__pyx_v_minimum_score, PyObject *__pyx_v_mass_error_tolerance) {

#line 684 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 684 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 684 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_1;

#line 684 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 684 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 684 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 684 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 684 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_deisotope/_c/scoring.pyx":685
 *     """
 *     def __init__(self, minimum_score=10, mass_error_tolerance=0.02):
 *         self.mass_error_tolerance = mass_error_tolerance             # <<<<<<<<<<<<<<
 *         self.select = MaximizeFitSelector()
 *         self.select.minimum_score = minimum_score
 */

#line 685 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_mass_error_tolerance); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 685, __pyx_L1_error)

#line 685 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->mass_error_tolerance = __pyx_t_1;

  /* "ms_deisotope/_c/scoring.pyx":686
 *     def __init__(self, minimum_score=10, mass_error_tolerance=0.02):
 *         self.mass_error_tolerance = mass_error_tolerance
 *         self.select = MaximizeFitSelector()             # <<<<<<<<<<<<<<
 *         self.select.minimum_score = minimum_score
 * 
 */

#line 686 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_7scoring_MaximizeFitSelector)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 686, __pyx_L1_error)

#line 686 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 686 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 686 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.select);

#line 686 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx_base.select));

#line 686 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->__pyx_base.select = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_t_2);

#line 686 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/scoring.pyx":687
 *         self.mass_error_tolerance = mass_error_tolerance
 *         self.select = MaximizeFitSelector()
 *         self.select.minimum_score = minimum_score             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision
 */

#line 687 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_minimum_score); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 687, __pyx_L1_error)

#line 687 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->__pyx_base.select->minimum_score = __pyx_t_1;

  /* "ms_deisotope/_c/scoring.pyx":684
 *     Proteomics : MCP, 9(12), 27722782. https://doi.org/10.1074/mcp.M110.002766
 *     """
 *     def __init__(self, minimum_score=10, mass_error_tolerance=0.02):             # <<<<<<<<<<<<<<
 *         self.mass_error_tolerance = mass_error_tolerance
 *         self.select = MaximizeFitSelector()
 */

#line 684 "ms_deisotope/_c/scoring.pyx"


#line 684 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 684 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 684 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 684 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 684 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 684 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.MSDeconVFitter.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 684 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = -1;

#line 684 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 684 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 684 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 684 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":690
 * 
 *     @cython.cdivision
 *     cdef double score_peak(self, FittedPeak obs, TheoreticalPeak theo, double mass_error_tolerance=0.02, double minimum_signal_to_noise=1) nogil:             # <<<<<<<<<<<<<<
 *         return ms_deconv_score_peak(obs, theo, mass_error_tolerance, minimum_signal_to_noise)
 * 
 */

#line 690 "ms_deisotope/_c/scoring.pyx"


#line 690 "ms_deisotope/_c/scoring.pyx"
static double __pyx_f_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_obs, struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_theo, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak *__pyx_optional_args) {

#line 690 "ms_deisotope/_c/scoring.pyx"
  double __pyx_v_mass_error_tolerance = ((double)0.02);

#line 690 "ms_deisotope/_c/scoring.pyx"
  double __pyx_v_minimum_signal_to_noise = ((double)1.0);

#line 690 "ms_deisotope/_c/scoring.pyx"
  double __pyx_r;

#line 690 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_1;

#line 690 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_ms_deconv_score_peak __pyx_t_2;

#line 690 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_optional_args) {

#line 690 "ms_deisotope/_c/scoring.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 690 "ms_deisotope/_c/scoring.pyx"
      __pyx_v_mass_error_tolerance = __pyx_optional_args->mass_error_tolerance;

#line 690 "ms_deisotope/_c/scoring.pyx"
      if (__pyx_optional_args->__pyx_n > 1) {

#line 690 "ms_deisotope/_c/scoring.pyx"
        __pyx_v_minimum_signal_to_noise = __pyx_optional_args->minimum_signal_to_noise;

#line 690 "ms_deisotope/_c/scoring.pyx"
      }

#line 690 "ms_deisotope/_c/scoring.pyx"
    }

#line 690 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":691
 *     @cython.cdivision
 *     cdef double score_peak(self, FittedPeak obs, TheoreticalPeak theo, double mass_error_tolerance=0.02, double minimum_signal_to_noise=1) nogil:
 *         return ms_deconv_score_peak(obs, theo, mass_error_tolerance, minimum_signal_to_noise)             # <<<<<<<<<<<<<<
 * 
 *     def evaluate(self, PeakSet peaklist, list observed, list expected):
 */

#line 691 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2.__pyx_n = 2;

#line 691 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2.mass_error_tolerance = __pyx_v_mass_error_tolerance;

#line 691 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2.minimum_signal_to_noise = __pyx_v_minimum_signal_to_noise;

#line 691 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_7scoring_ms_deconv_score_peak(__pyx_v_obs, __pyx_v_theo, &__pyx_t_2); 

#line 691 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 691 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":690
 * 
 *     @cython.cdivision
 *     cdef double score_peak(self, FittedPeak obs, TheoreticalPeak theo, double mass_error_tolerance=0.02, double minimum_signal_to_noise=1) nogil:             # <<<<<<<<<<<<<<
 *         return ms_deconv_score_peak(obs, theo, mass_error_tolerance, minimum_signal_to_noise)
 * 
 */

#line 690 "ms_deisotope/_c/scoring.pyx"


#line 690 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 690 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 690 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 690 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":693
 *         return ms_deconv_score_peak(obs, theo, mass_error_tolerance, minimum_signal_to_noise)
 * 
 *     def evaluate(self, PeakSet peaklist, list observed, list expected):             # <<<<<<<<<<<<<<
 *         return self._evaluate(peaklist, observed, expected)
 * 
 */

#line 693 "ms_deisotope/_c/scoring.pyx"


#line 693 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 693 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_3evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 693 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_2evaluate[] = "MSDeconVFitter.evaluate(self, PeakSet peaklist, list observed, list expected)";

#line 693 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_3evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist = 0
#line 693 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_observed = 0
#line 693 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_expected = 0
#line 693 "ms_deisotope/_c/scoring.pyx"
;

#line 693 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 693 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 693 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 693 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 693 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 693 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("evaluate (wrapper)", 0);

#line 693 "ms_deisotope/_c/scoring.pyx"
  {

#line 693 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peaklist,&__pyx_n_s_observed,&__pyx_n_s_expected,0};

#line 693 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[3] = {0,0,0};

#line 693 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 693 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 693 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 693 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  3: 
#line 693 "ms_deisotope/_c/scoring.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 693 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 693 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 693 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 693 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 693 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 693 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 693 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 693 "ms_deisotope/_c/scoring.pyx"
      }

#line 693 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 693 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 693 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 693 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaklist)) != 0)) kw_args--;
        else 
#line 693 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 693 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 693 "ms_deisotope/_c/scoring.pyx"
        case  1:

#line 693 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_observed)) != 0)) kw_args--;

#line 693 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("evaluate", 1, 3, 3, 1); 
#line 693 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 693, __pyx_L3_error)

#line 693 "ms_deisotope/_c/scoring.pyx"
        }

#line 693 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 693 "ms_deisotope/_c/scoring.pyx"
        case  2:

#line 693 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_expected)) != 0)) kw_args--;

#line 693 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("evaluate", 1, 3, 3, 2); 
#line 693 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 693, __pyx_L3_error)

#line 693 "ms_deisotope/_c/scoring.pyx"
        }

#line 693 "ms_deisotope/_c/scoring.pyx"
      }

#line 693 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 693 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "evaluate") < 0)) __PYX_ERR(0, 693, __pyx_L3_error)

#line 693 "ms_deisotope/_c/scoring.pyx"
      }

#line 693 "ms_deisotope/_c/scoring.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 693 "ms_deisotope/_c/scoring.pyx"
      goto __pyx_L5_argtuple_error;

#line 693 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 693 "ms_deisotope/_c/scoring.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 693 "ms_deisotope/_c/scoring.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 693 "ms_deisotope/_c/scoring.pyx"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 693 "ms_deisotope/_c/scoring.pyx"
    }

#line 693 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_peaklist = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[0]);

#line 693 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_observed = ((PyObject*)values[1]);

#line 693 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_expected = ((PyObject*)values[2]);

#line 693 "ms_deisotope/_c/scoring.pyx"
  }

#line 693 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 693 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("evaluate", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 693 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 693, __pyx_L3_error)

#line 693 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 693 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.MSDeconVFitter.evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 693 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 693 "ms_deisotope/_c/scoring.pyx"
  return NULL;

#line 693 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 693 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peaklist), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peaklist", 0))) __PYX_ERR(0, 693, __pyx_L1_error)

#line 693 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_observed), (&PyList_Type), 1, "observed", 1))) __PYX_ERR(0, 693, __pyx_L1_error)

#line 693 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_expected), (&PyList_Type), 1, "expected", 1))) __PYX_ERR(0, 693, __pyx_L1_error)
  __pyx_r = 
#line 693 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_2evaluate(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *)__pyx_v_self), __pyx_v_peaklist, __pyx_v_observed, __pyx_v_expected);

#line 693 "ms_deisotope/_c/scoring.pyx"


#line 693 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 693 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 693 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 693 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 693 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 693 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 693 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 693 "ms_deisotope/_c/scoring.pyx"
}

#line 693 "ms_deisotope/_c/scoring.pyx"


#line 693 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_2evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected) {

#line 693 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 693 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 693 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 693 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 693 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 693 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 693 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("evaluate", 0);

  /* "ms_deisotope/_c/scoring.pyx":694
 * 
 *     def evaluate(self, PeakSet peaklist, list observed, list expected):
 *         return self._evaluate(peaklist, observed, expected)             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */

#line 694 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 694 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyFloat_FromDouble(((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MSDeconVFitter *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._evaluate(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self), __pyx_v_peaklist, __pyx_v_observed, __pyx_v_expected, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 694, __pyx_L1_error)

#line 694 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 694 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 694 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 694 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":693
 *         return ms_deconv_score_peak(obs, theo, mass_error_tolerance, minimum_signal_to_noise)
 * 
 *     def evaluate(self, PeakSet peaklist, list observed, list expected):             # <<<<<<<<<<<<<<
 *         return self._evaluate(peaklist, observed, expected)
 * 
 */

#line 693 "ms_deisotope/_c/scoring.pyx"


#line 693 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 693 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 693 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 693 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.MSDeconVFitter.evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 693 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 693 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 693 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 693 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 693 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 693 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":696
 *         return self._evaluate(peaklist, observed, expected)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (0,), self.__getstate__()
 * 
 */

#line 696 "ms_deisotope/_c/scoring.pyx"


#line 696 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 696 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_5__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 696 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_4__reduce__[] = "MSDeconVFitter.__reduce__(self)";

#line 696 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_5__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 696 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 696 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 696 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = 
#line 696 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_4__reduce__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *)__pyx_v_self));

#line 696 "ms_deisotope/_c/scoring.pyx"


#line 696 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 696 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 696 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 696 "ms_deisotope/_c/scoring.pyx"
}

#line 696 "ms_deisotope/_c/scoring.pyx"


#line 696 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_4__reduce__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *__pyx_v_self) {

#line 696 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 696 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 696 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 696 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 696 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 696 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 696 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 696 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 696 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 696 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "ms_deisotope/_c/scoring.pyx":697
 * 
 *     def __reduce__(self):
 *         return self.__class__, (0,), self.__getstate__()             # <<<<<<<<<<<<<<
 * 
 *     def __getstate__(self):
 */

#line 697 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 697 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 697, __pyx_L1_error)

#line 697 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 697 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_getstate); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 697, __pyx_L1_error)

#line 697 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 697 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_4 = NULL;

#line 697 "ms_deisotope/_c/scoring.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {

#line 697 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 697 "ms_deisotope/_c/scoring.pyx"
    if (likely(__pyx_t_4)) {

#line 697 "ms_deisotope/_c/scoring.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 697 "ms_deisotope/_c/scoring.pyx"
      __Pyx_INCREF(__pyx_t_4);

#line 697 "ms_deisotope/_c/scoring.pyx"
      __Pyx_INCREF(function);

#line 697 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF_SET(__pyx_t_3, function);

#line 697 "ms_deisotope/_c/scoring.pyx"
    }

#line 697 "ms_deisotope/_c/scoring.pyx"
  }

#line 697 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 697 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 697 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 697, __pyx_L1_error)

#line 697 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 697 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 697 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 697, __pyx_L1_error)

#line 697 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 697 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 697 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);

#line 697 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_tuple_);

#line 697 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_tuple_);

#line 697 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_tuple_);

#line 697 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 697 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);

#line 697 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 697 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = 0;

#line 697 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_3;

#line 697 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = 0;

#line 697 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":696
 *         return self._evaluate(peaklist, observed, expected)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (0,), self.__getstate__()
 * 
 */

#line 696 "ms_deisotope/_c/scoring.pyx"


#line 696 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 696 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 696 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 696 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 696 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 696 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 696 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.MSDeconVFitter.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 696 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 696 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 696 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 696 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 696 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 696 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":699
 *         return self.__class__, (0,), self.__getstate__()
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return (self.select, self.mass_error_tolerance)
 * 
 */

#line 699 "ms_deisotope/_c/scoring.pyx"


#line 699 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 699 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_7__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 699 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_6__getstate__[] = "MSDeconVFitter.__getstate__(self)";

#line 699 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_7__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 699 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 699 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 699 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = 
#line 699 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_6__getstate__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *)__pyx_v_self));

#line 699 "ms_deisotope/_c/scoring.pyx"


#line 699 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 699 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 699 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 699 "ms_deisotope/_c/scoring.pyx"
}

#line 699 "ms_deisotope/_c/scoring.pyx"


#line 699 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_6__getstate__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *__pyx_v_self) {

#line 699 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 699 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 699 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 699 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 699 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 699 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 699 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 699 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "ms_deisotope/_c/scoring.pyx":700
 * 
 *     def __getstate__(self):
 *         return (self.select, self.mass_error_tolerance)             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, state):
 */

#line 700 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 700 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->mass_error_tolerance); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 700, __pyx_L1_error)

#line 700 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 700 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 700, __pyx_L1_error)

#line 700 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 700 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx_base.select));

#line 700 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx_base.select));

#line 700 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_self->__pyx_base.select));

#line 700 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 700 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);

#line 700 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 700 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_2;

#line 700 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = 0;

#line 700 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":699
 *         return self.__class__, (0,), self.__getstate__()
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return (self.select, self.mass_error_tolerance)
 * 
 */

#line 699 "ms_deisotope/_c/scoring.pyx"


#line 699 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 699 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 699 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 699 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 699 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.MSDeconVFitter.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 699 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 699 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 699 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 699 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 699 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 699 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":702
 *         return (self.select, self.mass_error_tolerance)
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.select, self.mass_error_tolerance = state
 * 
 */

#line 702 "ms_deisotope/_c/scoring.pyx"


#line 702 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 702 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_9__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state); /*proto*/

#line 702 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_8__setstate__[] = "MSDeconVFitter.__setstate__(self, state)";

#line 702 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_9__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state) {

#line 702 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 702 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 702 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  __pyx_r = 
#line 702 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_8__setstate__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *)__pyx_v_self), ((PyObject *)__pyx_v_state));

#line 702 "ms_deisotope/_c/scoring.pyx"


#line 702 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 702 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 702 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 702 "ms_deisotope/_c/scoring.pyx"
}

#line 702 "ms_deisotope/_c/scoring.pyx"


#line 702 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_8__setstate__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *__pyx_v_self, PyObject *__pyx_v_state) {

#line 702 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 702 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 702 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 702 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 702 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 702 "ms_deisotope/_c/scoring.pyx"
  PyObject *(*__pyx_t_4)(PyObject *);

#line 702 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_5;

#line 702 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 702 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 702 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 702 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* "ms_deisotope/_c/scoring.pyx":703
 * 
 *     def __setstate__(self, state):
 *         self.select, self.mass_error_tolerance = state             # <<<<<<<<<<<<<<
 * 
 *     cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):
 */

#line 703 "ms_deisotope/_c/scoring.pyx"
  if ((likely(PyTuple_CheckExact(__pyx_v_state))) || (PyList_CheckExact(__pyx_v_state))) {

#line 703 "ms_deisotope/_c/scoring.pyx"
    PyObject* sequence = __pyx_v_state;

#line 703 "ms_deisotope/_c/scoring.pyx"
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);

#line 703 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(size != 2)) {

#line 703 "ms_deisotope/_c/scoring.pyx"
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);

#line 703 "ms_deisotope/_c/scoring.pyx"
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);

#line 703 "ms_deisotope/_c/scoring.pyx"
      __PYX_ERR(0, 703, __pyx_L1_error)

#line 703 "ms_deisotope/_c/scoring.pyx"
    }

#line 703 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 703 "ms_deisotope/_c/scoring.pyx"
    if (likely(PyTuple_CheckExact(sequence))) {

#line 703 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 

#line 703 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 

#line 703 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 703 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 

#line 703 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 

#line 703 "ms_deisotope/_c/scoring.pyx"
    }

#line 703 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 703 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_2);

#line 703 "ms_deisotope/_c/scoring.pyx"
    #else

#line 703 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 703, __pyx_L1_error)

#line 703 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 703 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 703, __pyx_L1_error)

#line 703 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 703 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 703 "ms_deisotope/_c/scoring.pyx"
  } else {

#line 703 "ms_deisotope/_c/scoring.pyx"
    Py_ssize_t index = -1;

#line 703 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_3 = PyObject_GetIter(__pyx_v_state); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 703, __pyx_L1_error)

#line 703 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 703 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_4 = Py_TYPE(__pyx_t_3)->tp_iternext;
    index = 0; __pyx_t_1 = __pyx_t_4(__pyx_t_3); if (unlikely(!__pyx_t_1)) 
#line 703 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L3_unpacking_failed;

#line 703 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_1);
    index = 1; __pyx_t_2 = __pyx_t_4(__pyx_t_3); if (unlikely(!__pyx_t_2)) 
#line 703 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L3_unpacking_failed;

#line 703 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 703 "ms_deisotope/_c/scoring.pyx"
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_4(__pyx_t_3), 2) < 0) __PYX_ERR(0, 703, __pyx_L1_error)

#line 703 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_4 = NULL;

#line 703 "ms_deisotope/_c/scoring.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 703 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L4_unpacking_done;

#line 703 "ms_deisotope/_c/scoring.pyx"
    __pyx_L3_unpacking_failed:;

#line 703 "ms_deisotope/_c/scoring.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 703 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_4 = NULL;

#line 703 "ms_deisotope/_c/scoring.pyx"
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);

#line 703 "ms_deisotope/_c/scoring.pyx"
    __PYX_ERR(0, 703, __pyx_L1_error)

#line 703 "ms_deisotope/_c/scoring.pyx"
    __pyx_L4_unpacking_done:;

#line 703 "ms_deisotope/_c/scoring.pyx"
  }

#line 703 "ms_deisotope/_c/scoring.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase))))) __PYX_ERR(0, 703, __pyx_L1_error)

#line 703 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 703, __pyx_L1_error)

#line 703 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 703 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 703 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.select);

#line 703 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx_base.select));

#line 703 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->__pyx_base.select = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_t_1);

#line 703 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 703 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->mass_error_tolerance = __pyx_t_5;

  /* "ms_deisotope/_c/scoring.pyx":702
 *         return (self.select, self.mass_error_tolerance)
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.select, self.mass_error_tolerance = state
 * 
 */

#line 702 "ms_deisotope/_c/scoring.pyx"


#line 702 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 702 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 702 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 702 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 702 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 702 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 702 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 702 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.MSDeconVFitter.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 702 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 702 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 702 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 702 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 702 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 702 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":705
 *         self.select, self.mass_error_tolerance = state
 * 
 *     cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

#line 705 "ms_deisotope/_c/scoring.pyx"


#line 705 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_11_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 705 "ms_deisotope/_c/scoring.pyx"
static double __pyx_f_12ms_deisotope_2_c_7scoring_14MSDeconVFitter__evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected, int __pyx_skip_dispatch) {
  size_t __pyx_v_i
#line 705 "ms_deisotope/_c/scoring.pyx"
;
  size_t __pyx_v_n
#line 705 "ms_deisotope/_c/scoring.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_obs = 0
#line 705 "ms_deisotope/_c/scoring.pyx"
;
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_theo = 0
#line 705 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_score
#line 705 "ms_deisotope/_c/scoring.pyx"
;

#line 705 "ms_deisotope/_c/scoring.pyx"
  double __pyx_r;

#line 705 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 705 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 705 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 705 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 705 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 705 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_5;

#line 705 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 705 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_7;

#line 705 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_8;

#line 705 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_9;

#line 705 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_10;

#line 705 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_11;

#line 705 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak __pyx_t_12;

#line 705 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 705 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 705 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 705 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate", 0);

#line 705 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 705 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 705 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 705 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 705 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 705 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 705 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 705 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 705 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 705 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_evaluate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 705, __pyx_L1_error)

#line 705 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 705 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_11_evaluate)) {

#line 705 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 705 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 705 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_5 = 0;

#line 705 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 705 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 705 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 705 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 705 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 705 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 705 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 705 "ms_deisotope/_c/scoring.pyx"
            __pyx_t_5 = 1;

#line 705 "ms_deisotope/_c/scoring.pyx"
          }

#line 705 "ms_deisotope/_c/scoring.pyx"
        }

#line 705 "ms_deisotope/_c/scoring.pyx"
        #if CYTHON_FAST_PYCALL

#line 705 "ms_deisotope/_c/scoring.pyx"
        if (PyFunction_Check(__pyx_t_3)) {

#line 705 "ms_deisotope/_c/scoring.pyx"
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_peaklist), __pyx_v_observed, __pyx_v_expected};

#line 705 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 705, __pyx_L1_error)

#line 705 "ms_deisotope/_c/scoring.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 705 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 705 "ms_deisotope/_c/scoring.pyx"
        } else

#line 705 "ms_deisotope/_c/scoring.pyx"
        #endif

#line 705 "ms_deisotope/_c/scoring.pyx"
        #if CYTHON_FAST_PYCCALL

#line 705 "ms_deisotope/_c/scoring.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {

#line 705 "ms_deisotope/_c/scoring.pyx"
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_peaklist), __pyx_v_observed, __pyx_v_expected};

#line 705 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 705, __pyx_L1_error)

#line 705 "ms_deisotope/_c/scoring.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 705 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 705 "ms_deisotope/_c/scoring.pyx"
        } else

#line 705 "ms_deisotope/_c/scoring.pyx"
        #endif

#line 705 "ms_deisotope/_c/scoring.pyx"
        {

#line 705 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_6 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 705, __pyx_L1_error)

#line 705 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_6);

#line 705 "ms_deisotope/_c/scoring.pyx"
          if (__pyx_t_4) {

#line 705 "ms_deisotope/_c/scoring.pyx"
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;

#line 705 "ms_deisotope/_c/scoring.pyx"
          }

#line 705 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(((PyObject *)__pyx_v_peaklist));

#line 705 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(((PyObject *)__pyx_v_peaklist));

#line 705 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_peaklist));

#line 705 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(__pyx_v_observed);

#line 705 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(__pyx_v_observed);

#line 705 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_observed);

#line 705 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(__pyx_v_expected);

#line 705 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(__pyx_v_expected);

#line 705 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, __pyx_v_expected);

#line 705 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 705, __pyx_L1_error)

#line 705 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 705 "ms_deisotope/_c/scoring.pyx"
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 705 "ms_deisotope/_c/scoring.pyx"
        }

#line 705 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 705 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_7 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_7 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 705, __pyx_L1_error)

#line 705 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 705 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = __pyx_t_7;

#line 705 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 705 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 705 "ms_deisotope/_c/scoring.pyx"
      }

#line 705 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 705 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 705 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 705 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 705 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 705 "ms_deisotope/_c/scoring.pyx"
      }

#line 705 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 705 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 705 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 705 "ms_deisotope/_c/scoring.pyx"
    }

#line 705 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 705 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":712
 *             double score
 * 
 *         n = PyList_GET_SIZE(observed)             # <<<<<<<<<<<<<<
 *         score = 0
 *         for i in range(n):
 */

#line 712 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_n = PyList_GET_SIZE(__pyx_v_observed);

  /* "ms_deisotope/_c/scoring.pyx":713
 * 
 *         n = PyList_GET_SIZE(observed)
 *         score = 0             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             obs = <FittedPeak>PyList_GET_ITEM(observed, i)
 */

#line 713 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_score = 0.0;

  /* "ms_deisotope/_c/scoring.pyx":714
 *         n = PyList_GET_SIZE(observed)
 *         score = 0
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             obs = <FittedPeak>PyList_GET_ITEM(observed, i)
 *             theo = <TheoreticalPeak>PyList_GET_ITEM(expected, i)
 */

#line 714 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_8 = __pyx_v_n;

#line 714 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_9 = __pyx_t_8;

#line 714 "ms_deisotope/_c/scoring.pyx"
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {

#line 714 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_i = __pyx_t_10;

    /* "ms_deisotope/_c/scoring.pyx":715
 *         score = 0
 *         for i in range(n):
 *             obs = <FittedPeak>PyList_GET_ITEM(observed, i)             # <<<<<<<<<<<<<<
 *             theo = <TheoreticalPeak>PyList_GET_ITEM(expected, i)
 *             score += self.score_peak(obs, theo, self.mass_error_tolerance, 1)
 */

#line 715 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_11 = PyList_GET_ITEM(__pyx_v_observed, __pyx_v_i);

#line 715 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_11);

#line 715 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 715 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF_SET(__pyx_v_obs, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1));

#line 715 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/scoring.pyx":716
 *         for i in range(n):
 *             obs = <FittedPeak>PyList_GET_ITEM(observed, i)
 *             theo = <TheoreticalPeak>PyList_GET_ITEM(expected, i)             # <<<<<<<<<<<<<<
 *             score += self.score_peak(obs, theo, self.mass_error_tolerance, 1)
 * 
 */

#line 716 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_11 = PyList_GET_ITEM(__pyx_v_expected, __pyx_v_i);

#line 716 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_11);

#line 716 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 716 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF_SET(__pyx_v_theo, ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_1));

#line 716 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/scoring.pyx":717
 *             obs = <FittedPeak>PyList_GET_ITEM(observed, i)
 *             theo = <TheoreticalPeak>PyList_GET_ITEM(expected, i)
 *             score += self.score_peak(obs, theo, self.mass_error_tolerance, 1)             # <<<<<<<<<<<<<<
 * 
 *         return score
 */

#line 717 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_12.__pyx_n = 2;

#line 717 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_12.mass_error_tolerance = __pyx_v_self->mass_error_tolerance;

#line 717 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_12.minimum_signal_to_noise = 1.0;

#line 717 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_7 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MSDeconVFitter *)__pyx_v_self->__pyx_base.__pyx_vtab)->score_peak(__pyx_v_self, __pyx_v_obs, __pyx_v_theo, &__pyx_t_12); 

#line 717 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_score = (__pyx_v_score + __pyx_t_7);

#line 717 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":719
 *             score += self.score_peak(obs, theo, self.mass_error_tolerance, 1)
 * 
 *         return score             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 719 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_v_score;

#line 719 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":705
 *         self.select, self.mass_error_tolerance = state
 * 
 *     cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

#line 705 "ms_deisotope/_c/scoring.pyx"


#line 705 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 705 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 705 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 705 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 705 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 705 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 705 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 705 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.MSDeconVFitter._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 705 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 705 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 705 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_obs);

#line 705 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_theo);

#line 705 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 705 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 705 "ms_deisotope/_c/scoring.pyx"
}

#line 705 "ms_deisotope/_c/scoring.pyx"


#line 705 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 705 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_11_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 705 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_10_evaluate[] = "MSDeconVFitter._evaluate(self, PeakSet peaklist, list observed, list expected) -> double";

#line 705 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_11_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist = 0
#line 705 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_observed = 0
#line 705 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_expected = 0
#line 705 "ms_deisotope/_c/scoring.pyx"
;

#line 705 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 705 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 705 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 705 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 705 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 705 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate (wrapper)", 0);

#line 705 "ms_deisotope/_c/scoring.pyx"
  {

#line 705 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peaklist,&__pyx_n_s_observed,&__pyx_n_s_expected,0};

#line 705 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[3] = {0,0,0};

#line 705 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 705 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 705 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 705 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  3: 
#line 705 "ms_deisotope/_c/scoring.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 705 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 705 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 705 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 705 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 705 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 705 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 705 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 705 "ms_deisotope/_c/scoring.pyx"
      }

#line 705 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 705 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 705 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 705 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaklist)) != 0)) kw_args--;
        else 
#line 705 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 705 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 705 "ms_deisotope/_c/scoring.pyx"
        case  1:

#line 705 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_observed)) != 0)) kw_args--;

#line 705 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, 1); 
#line 705 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 705, __pyx_L3_error)

#line 705 "ms_deisotope/_c/scoring.pyx"
        }

#line 705 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 705 "ms_deisotope/_c/scoring.pyx"
        case  2:

#line 705 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_expected)) != 0)) kw_args--;

#line 705 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, 2); 
#line 705 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 705, __pyx_L3_error)

#line 705 "ms_deisotope/_c/scoring.pyx"
        }

#line 705 "ms_deisotope/_c/scoring.pyx"
      }

#line 705 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 705 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_evaluate") < 0)) __PYX_ERR(0, 705, __pyx_L3_error)

#line 705 "ms_deisotope/_c/scoring.pyx"
      }

#line 705 "ms_deisotope/_c/scoring.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 705 "ms_deisotope/_c/scoring.pyx"
      goto __pyx_L5_argtuple_error;

#line 705 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 705 "ms_deisotope/_c/scoring.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 705 "ms_deisotope/_c/scoring.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 705 "ms_deisotope/_c/scoring.pyx"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 705 "ms_deisotope/_c/scoring.pyx"
    }

#line 705 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_peaklist = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[0]);

#line 705 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_observed = ((PyObject*)values[1]);

#line 705 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_expected = ((PyObject*)values[2]);

#line 705 "ms_deisotope/_c/scoring.pyx"
  }

#line 705 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 705 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 705 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 705, __pyx_L3_error)

#line 705 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 705 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.MSDeconVFitter._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 705 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 705 "ms_deisotope/_c/scoring.pyx"
  return NULL;

#line 705 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 705 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peaklist), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peaklist", 0))) __PYX_ERR(0, 705, __pyx_L1_error)

#line 705 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_observed), (&PyList_Type), 1, "observed", 1))) __PYX_ERR(0, 705, __pyx_L1_error)

#line 705 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_expected), (&PyList_Type), 1, "expected", 1))) __PYX_ERR(0, 705, __pyx_L1_error)
  __pyx_r = 
#line 705 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_10_evaluate(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *)__pyx_v_self), __pyx_v_peaklist, __pyx_v_observed, __pyx_v_expected);

#line 705 "ms_deisotope/_c/scoring.pyx"


#line 705 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 705 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 705 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 705 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 705 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 705 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 705 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 705 "ms_deisotope/_c/scoring.pyx"
}

#line 705 "ms_deisotope/_c/scoring.pyx"


#line 705 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_10_evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected) {

#line 705 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 705 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 705 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 705 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 705 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 705 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 705 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate", 0);

#line 705 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 705 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_12ms_deisotope_2_c_7scoring_14MSDeconVFitter__evaluate(__pyx_v_self, __pyx_v_peaklist, __pyx_v_observed, __pyx_v_expected, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 705, __pyx_L1_error)

#line 705 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 705 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 705 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 705 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 705 "ms_deisotope/_c/scoring.pyx"


#line 705 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 705 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 705 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 705 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.MSDeconVFitter._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 705 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 705 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 705 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 705 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 705 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 705 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pxd":73
 * 
 * cdef class MSDeconVFitter(IsotopicFitterBase):
 *     cdef public double mass_error_tolerance             # <<<<<<<<<<<<<<
 * 
 *     cdef double score_peak(self, FittedPeak obs, TheoreticalPeak theo, double mass_error_tolerance=*, double minimum_signal_to_noise=*) nogil
 */

#line 73 "ms_deisotope/_c/scoring.pxd"


#line 73 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 73 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_20mass_error_tolerance_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 73 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_20mass_error_tolerance_1__get__(PyObject *__pyx_v_self) {

#line 73 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = 0;

#line 73 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 73 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 73 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_20mass_error_tolerance___get__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *)__pyx_v_self));

#line 73 "ms_deisotope/_c/scoring.pxd"


#line 73 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 73 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 73 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 73 "ms_deisotope/_c/scoring.pxd"
}

#line 73 "ms_deisotope/_c/scoring.pxd"


#line 73 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_20mass_error_tolerance___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *__pyx_v_self) {

#line 73 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = NULL;

#line 73 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 73 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 73 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 73 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 73 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 73 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 73 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 73 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->mass_error_tolerance); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 73, __pyx_L1_error)

#line 73 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 73 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = __pyx_t_1;

#line 73 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = 0;

#line 73 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 73 "ms_deisotope/_c/scoring.pxd"


#line 73 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 73 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 73 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 73 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.MSDeconVFitter.mass_error_tolerance.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 73 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = NULL;

#line 73 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 73 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 73 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 73 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 73 "ms_deisotope/_c/scoring.pxd"
}

#line 73 "ms_deisotope/_c/scoring.pxd"


#line 73 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 73 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_20mass_error_tolerance_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 73 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_20mass_error_tolerance_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 73 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 73 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 73 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 73 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_20mass_error_tolerance_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 73 "ms_deisotope/_c/scoring.pxd"


#line 73 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 73 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 73 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 73 "ms_deisotope/_c/scoring.pxd"
}

#line 73 "ms_deisotope/_c/scoring.pxd"


#line 73 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_20mass_error_tolerance_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *__pyx_v_self, PyObject *__pyx_v_value) {

#line 73 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 73 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 73 "ms_deisotope/_c/scoring.pxd"
  double __pyx_t_1;

#line 73 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 73 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 73 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 73 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 73 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 73, __pyx_L1_error)

#line 73 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->mass_error_tolerance = __pyx_t_1;

#line 73 "ms_deisotope/_c/scoring.pxd"


#line 73 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 73 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 73 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 73 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 73 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.MSDeconVFitter.mass_error_tolerance.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 73 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = -1;

#line 73 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 73 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 73 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 73 "ms_deisotope/_c/scoring.pxd"
}

/* "ms_deisotope/_c/scoring.pyx":732
 *     peak list
 *     """
 *     def __init__(self, minimum_score=10, penalty_factor=1, mass_error_tolerance=0.02):             # <<<<<<<<<<<<<<
 *         self.select = MaximizeFitSelector(minimum_score)
 *         self.penalty_factor = penalty_factor
 */

#line 732 "ms_deisotope/_c/scoring.pyx"


#line 732 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 732 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 732 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_minimum_score = 0
#line 732 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_penalty_factor = 0
#line 732 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_mass_error_tolerance = 0
#line 732 "ms_deisotope/_c/scoring.pyx"
;

#line 732 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 732 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 732 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 732 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 732 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 732 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 732 "ms_deisotope/_c/scoring.pyx"
  {

#line 732 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_minimum_score,&__pyx_n_s_penalty_factor,&__pyx_n_s_mass_error_tolerance,0};

#line 732 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[3] = {0,0,0};

#line 732 "ms_deisotope/_c/scoring.pyx"
    values[0] = ((PyObject *)__pyx_int_10);

#line 732 "ms_deisotope/_c/scoring.pyx"
    values[1] = ((PyObject *)__pyx_int_1);

#line 732 "ms_deisotope/_c/scoring.pyx"
    values[2] = ((PyObject *)__pyx_float_0_02);

#line 732 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 732 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 732 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 732 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  3: 
#line 732 "ms_deisotope/_c/scoring.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 732 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 732 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 732 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 732 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 732 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 732 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 732 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 732 "ms_deisotope/_c/scoring.pyx"
      }

#line 732 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 732 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 732 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 732 "ms_deisotope/_c/scoring.pyx"
        if (kw_args > 0) {

#line 732 "ms_deisotope/_c/scoring.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_minimum_score);

#line 732 "ms_deisotope/_c/scoring.pyx"
          if (value) { values[0] = value; kw_args--; }

#line 732 "ms_deisotope/_c/scoring.pyx"
        }

#line 732 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 732 "ms_deisotope/_c/scoring.pyx"
        case  1:

#line 732 "ms_deisotope/_c/scoring.pyx"
        if (kw_args > 0) {

#line 732 "ms_deisotope/_c/scoring.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_penalty_factor);

#line 732 "ms_deisotope/_c/scoring.pyx"
          if (value) { values[1] = value; kw_args--; }

#line 732 "ms_deisotope/_c/scoring.pyx"
        }

#line 732 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 732 "ms_deisotope/_c/scoring.pyx"
        case  2:

#line 732 "ms_deisotope/_c/scoring.pyx"
        if (kw_args > 0) {

#line 732 "ms_deisotope/_c/scoring.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mass_error_tolerance);

#line 732 "ms_deisotope/_c/scoring.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 732 "ms_deisotope/_c/scoring.pyx"
        }

#line 732 "ms_deisotope/_c/scoring.pyx"
      }

#line 732 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 732 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 732, __pyx_L3_error)

#line 732 "ms_deisotope/_c/scoring.pyx"
      }

#line 732 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 732 "ms_deisotope/_c/scoring.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 732 "ms_deisotope/_c/scoring.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 732 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 732 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 732 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 732 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 732 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  0: 
#line 732 "ms_deisotope/_c/scoring.pyx"
break;
        default: 
#line 732 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 732 "ms_deisotope/_c/scoring.pyx"
      }

#line 732 "ms_deisotope/_c/scoring.pyx"
    }

#line 732 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_minimum_score = values[0];

#line 732 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_penalty_factor = values[1];

#line 732 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_mass_error_tolerance = values[2];

#line 732 "ms_deisotope/_c/scoring.pyx"
  }

#line 732 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 732 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 732 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 732, __pyx_L3_error)

#line 732 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 732 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.PenalizedMSDeconVFitter.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 732 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 732 "ms_deisotope/_c/scoring.pyx"
  return -1;

#line 732 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 732 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter___init__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *)__pyx_v_self), __pyx_v_minimum_score, __pyx_v_penalty_factor, __pyx_v_mass_error_tolerance);

#line 732 "ms_deisotope/_c/scoring.pyx"


#line 732 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 732 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 732 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 732 "ms_deisotope/_c/scoring.pyx"
}

#line 732 "ms_deisotope/_c/scoring.pyx"


#line 732 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter___init__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_v_self, PyObject *__pyx_v_minimum_score, PyObject *__pyx_v_penalty_factor, PyObject *__pyx_v_mass_error_tolerance) {

#line 732 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 732 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 732 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 732 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_2;

#line 732 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 732 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 732 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 732 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_deisotope/_c/scoring.pyx":733
 *     """
 *     def __init__(self, minimum_score=10, penalty_factor=1, mass_error_tolerance=0.02):
 *         self.select = MaximizeFitSelector(minimum_score)             # <<<<<<<<<<<<<<
 *         self.penalty_factor = penalty_factor
 *         self.mass_error_tolerance = mass_error_tolerance
 */

#line 733 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_7scoring_MaximizeFitSelector), __pyx_v_minimum_score); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 733, __pyx_L1_error)

#line 733 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 733 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 733 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.select);

#line 733 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx_base.select));

#line 733 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->__pyx_base.select = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_t_1);

#line 733 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/scoring.pyx":734
 *     def __init__(self, minimum_score=10, penalty_factor=1, mass_error_tolerance=0.02):
 *         self.select = MaximizeFitSelector(minimum_score)
 *         self.penalty_factor = penalty_factor             # <<<<<<<<<<<<<<
 *         self.mass_error_tolerance = mass_error_tolerance
 * 
 */

#line 734 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_v_penalty_factor); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 734, __pyx_L1_error)

#line 734 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->penalty_factor = __pyx_t_2;

  /* "ms_deisotope/_c/scoring.pyx":735
 *         self.select = MaximizeFitSelector(minimum_score)
 *         self.penalty_factor = penalty_factor
 *         self.mass_error_tolerance = mass_error_tolerance             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */

#line 735 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_v_mass_error_tolerance); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 735, __pyx_L1_error)

#line 735 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->mass_error_tolerance = __pyx_t_2;

  /* "ms_deisotope/_c/scoring.pyx":732
 *     peak list
 *     """
 *     def __init__(self, minimum_score=10, penalty_factor=1, mass_error_tolerance=0.02):             # <<<<<<<<<<<<<<
 *         self.select = MaximizeFitSelector(minimum_score)
 *         self.penalty_factor = penalty_factor
 */

#line 732 "ms_deisotope/_c/scoring.pyx"


#line 732 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 732 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 732 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 732 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 732 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 732 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.PenalizedMSDeconVFitter.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 732 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = -1;

#line 732 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 732 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 732 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 732 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":737
 *         self.mass_error_tolerance = mass_error_tolerance
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (0,), self.__getstate__()
 * 
 */

#line 737 "ms_deisotope/_c/scoring.pyx"


#line 737 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 737 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_3__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 737 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_2__reduce__[] = "PenalizedMSDeconVFitter.__reduce__(self)";

#line 737 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_3__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 737 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 737 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 737 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = 
#line 737 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_2__reduce__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *)__pyx_v_self));

#line 737 "ms_deisotope/_c/scoring.pyx"


#line 737 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 737 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 737 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 737 "ms_deisotope/_c/scoring.pyx"
}

#line 737 "ms_deisotope/_c/scoring.pyx"


#line 737 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_2__reduce__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_v_self) {

#line 737 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 737 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 737 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 737 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 737 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 737 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 737 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 737 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 737 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 737 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "ms_deisotope/_c/scoring.pyx":738
 * 
 *     def __reduce__(self):
 *         return self.__class__, (0,), self.__getstate__()             # <<<<<<<<<<<<<<
 * 
 *     def __getstate__(self):
 */

#line 738 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 738 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 738, __pyx_L1_error)

#line 738 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 738 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_getstate); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 738, __pyx_L1_error)

#line 738 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 738 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_4 = NULL;

#line 738 "ms_deisotope/_c/scoring.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {

#line 738 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 738 "ms_deisotope/_c/scoring.pyx"
    if (likely(__pyx_t_4)) {

#line 738 "ms_deisotope/_c/scoring.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 738 "ms_deisotope/_c/scoring.pyx"
      __Pyx_INCREF(__pyx_t_4);

#line 738 "ms_deisotope/_c/scoring.pyx"
      __Pyx_INCREF(function);

#line 738 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF_SET(__pyx_t_3, function);

#line 738 "ms_deisotope/_c/scoring.pyx"
    }

#line 738 "ms_deisotope/_c/scoring.pyx"
  }

#line 738 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 738 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 738 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 738, __pyx_L1_error)

#line 738 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 738 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 738 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 738, __pyx_L1_error)

#line 738 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 738 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 738 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);

#line 738 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_tuple_);

#line 738 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_tuple_);

#line 738 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_tuple_);

#line 738 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 738 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);

#line 738 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 738 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = 0;

#line 738 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_3;

#line 738 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = 0;

#line 738 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":737
 *         self.mass_error_tolerance = mass_error_tolerance
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (0,), self.__getstate__()
 * 
 */

#line 737 "ms_deisotope/_c/scoring.pyx"


#line 737 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 737 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 737 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 737 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 737 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 737 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 737 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.PenalizedMSDeconVFitter.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 737 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 737 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 737 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 737 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 737 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 737 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":740
 *         return self.__class__, (0,), self.__getstate__()
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return (self.select, self.penalty_factor, self.mass_error_tolerance)
 * 
 */

#line 740 "ms_deisotope/_c/scoring.pyx"


#line 740 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 740 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_5__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 740 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_4__getstate__[] = "PenalizedMSDeconVFitter.__getstate__(self)";

#line 740 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_5__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 740 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 740 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 740 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = 
#line 740 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_4__getstate__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *)__pyx_v_self));

#line 740 "ms_deisotope/_c/scoring.pyx"


#line 740 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 740 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 740 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 740 "ms_deisotope/_c/scoring.pyx"
}

#line 740 "ms_deisotope/_c/scoring.pyx"


#line 740 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_4__getstate__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_v_self) {

#line 740 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 740 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 740 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 740 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 740 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 740 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 740 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 740 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 740 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "ms_deisotope/_c/scoring.pyx":741
 * 
 *     def __getstate__(self):
 *         return (self.select, self.penalty_factor, self.mass_error_tolerance)             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, state):
 */

#line 741 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 741 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->penalty_factor); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 741, __pyx_L1_error)

#line 741 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 741 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->mass_error_tolerance); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 741, __pyx_L1_error)

#line 741 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 741 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 741, __pyx_L1_error)

#line 741 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 741 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx_base.select));

#line 741 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx_base.select));

#line 741 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self->__pyx_base.select));

#line 741 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 741 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);

#line 741 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 741 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);

#line 741 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 741 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = 0;

#line 741 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_3;

#line 741 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = 0;

#line 741 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":740
 *         return self.__class__, (0,), self.__getstate__()
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return (self.select, self.penalty_factor, self.mass_error_tolerance)
 * 
 */

#line 740 "ms_deisotope/_c/scoring.pyx"


#line 740 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 740 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 740 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 740 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 740 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 740 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.PenalizedMSDeconVFitter.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 740 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 740 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 740 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 740 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 740 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 740 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":743
 *         return (self.select, self.penalty_factor, self.mass_error_tolerance)
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.select, self.penalty_factor, self.mass_error_tolerance = state
 * 
 */

#line 743 "ms_deisotope/_c/scoring.pyx"


#line 743 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 743 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_7__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state); /*proto*/

#line 743 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_6__setstate__[] = "PenalizedMSDeconVFitter.__setstate__(self, state)";

#line 743 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_7__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state) {

#line 743 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 743 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 743 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  __pyx_r = 
#line 743 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_6__setstate__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *)__pyx_v_self), ((PyObject *)__pyx_v_state));

#line 743 "ms_deisotope/_c/scoring.pyx"


#line 743 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 743 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 743 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 743 "ms_deisotope/_c/scoring.pyx"
}

#line 743 "ms_deisotope/_c/scoring.pyx"


#line 743 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_6__setstate__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_v_self, PyObject *__pyx_v_state) {

#line 743 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 743 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 743 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 743 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 743 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 743 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 743 "ms_deisotope/_c/scoring.pyx"
  PyObject *(*__pyx_t_5)(PyObject *);

#line 743 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_6;

#line 743 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_7;

#line 743 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 743 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 743 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 743 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* "ms_deisotope/_c/scoring.pyx":744
 * 
 *     def __setstate__(self, state):
 *         self.select, self.penalty_factor, self.mass_error_tolerance = state             # <<<<<<<<<<<<<<
 * 
 *     def evaluate(self, PeakSet peaklist, list observed, list expected):
 */

#line 744 "ms_deisotope/_c/scoring.pyx"
  if ((likely(PyTuple_CheckExact(__pyx_v_state))) || (PyList_CheckExact(__pyx_v_state))) {

#line 744 "ms_deisotope/_c/scoring.pyx"
    PyObject* sequence = __pyx_v_state;

#line 744 "ms_deisotope/_c/scoring.pyx"
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);

#line 744 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(size != 3)) {

#line 744 "ms_deisotope/_c/scoring.pyx"
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);

#line 744 "ms_deisotope/_c/scoring.pyx"
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);

#line 744 "ms_deisotope/_c/scoring.pyx"
      __PYX_ERR(0, 744, __pyx_L1_error)

#line 744 "ms_deisotope/_c/scoring.pyx"
    }

#line 744 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 744 "ms_deisotope/_c/scoring.pyx"
    if (likely(PyTuple_CheckExact(sequence))) {

#line 744 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 

#line 744 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 

#line 744 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 2); 

#line 744 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 744 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 

#line 744 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 

#line 744 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_3 = PyList_GET_ITEM(sequence, 2); 

#line 744 "ms_deisotope/_c/scoring.pyx"
    }

#line 744 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 744 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_2);

#line 744 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_3);

#line 744 "ms_deisotope/_c/scoring.pyx"
    #else

#line 744 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 744, __pyx_L1_error)

#line 744 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 744 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 744, __pyx_L1_error)

#line 744 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 744 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_3 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 744, __pyx_L1_error)

#line 744 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 744 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 744 "ms_deisotope/_c/scoring.pyx"
  } else {

#line 744 "ms_deisotope/_c/scoring.pyx"
    Py_ssize_t index = -1;

#line 744 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_4 = PyObject_GetIter(__pyx_v_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 744, __pyx_L1_error)

#line 744 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 744 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_1 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_1)) 
#line 744 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L3_unpacking_failed;

#line 744 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_1);
    index = 1; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) 
#line 744 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L3_unpacking_failed;

#line 744 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_2);
    index = 2; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) 
#line 744 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L3_unpacking_failed;

#line 744 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 744 "ms_deisotope/_c/scoring.pyx"
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 3) < 0) __PYX_ERR(0, 744, __pyx_L1_error)

#line 744 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = NULL;

#line 744 "ms_deisotope/_c/scoring.pyx"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 744 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L4_unpacking_done;

#line 744 "ms_deisotope/_c/scoring.pyx"
    __pyx_L3_unpacking_failed:;

#line 744 "ms_deisotope/_c/scoring.pyx"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 744 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = NULL;

#line 744 "ms_deisotope/_c/scoring.pyx"
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);

#line 744 "ms_deisotope/_c/scoring.pyx"
    __PYX_ERR(0, 744, __pyx_L1_error)

#line 744 "ms_deisotope/_c/scoring.pyx"
    __pyx_L4_unpacking_done:;

#line 744 "ms_deisotope/_c/scoring.pyx"
  }

#line 744 "ms_deisotope/_c/scoring.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase))))) __PYX_ERR(0, 744, __pyx_L1_error)

#line 744 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_6 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 744, __pyx_L1_error)

#line 744 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 744 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_7 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_7 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 744, __pyx_L1_error)

#line 744 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 744 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 744 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.select);

#line 744 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx_base.select));

#line 744 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->__pyx_base.select = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_t_1);

#line 744 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 744 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->penalty_factor = __pyx_t_6;

#line 744 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->mass_error_tolerance = __pyx_t_7;

  /* "ms_deisotope/_c/scoring.pyx":743
 *         return (self.select, self.penalty_factor, self.mass_error_tolerance)
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.select, self.penalty_factor, self.mass_error_tolerance = state
 * 
 */

#line 743 "ms_deisotope/_c/scoring.pyx"


#line 743 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 743 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 743 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 743 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 743 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 743 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 743 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 743 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 743 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.PenalizedMSDeconVFitter.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 743 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 743 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 743 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 743 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 743 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 743 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":746
 *         self.select, self.penalty_factor, self.mass_error_tolerance = state
 * 
 *     def evaluate(self, PeakSet peaklist, list observed, list expected):             # <<<<<<<<<<<<<<
 *         return self._evaluate(peaklist, observed, expected)
 * 
 */

#line 746 "ms_deisotope/_c/scoring.pyx"


#line 746 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 746 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_9evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 746 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_8evaluate[] = "PenalizedMSDeconVFitter.evaluate(self, PeakSet peaklist, list observed, list expected)";

#line 746 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_9evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist = 0
#line 746 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_observed = 0
#line 746 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_expected = 0
#line 746 "ms_deisotope/_c/scoring.pyx"
;

#line 746 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 746 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 746 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 746 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 746 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 746 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("evaluate (wrapper)", 0);

#line 746 "ms_deisotope/_c/scoring.pyx"
  {

#line 746 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peaklist,&__pyx_n_s_observed,&__pyx_n_s_expected,0};

#line 746 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[3] = {0,0,0};

#line 746 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 746 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 746 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 746 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  3: 
#line 746 "ms_deisotope/_c/scoring.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 746 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 746 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 746 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 746 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 746 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 746 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 746 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 746 "ms_deisotope/_c/scoring.pyx"
      }

#line 746 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 746 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 746 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 746 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaklist)) != 0)) kw_args--;
        else 
#line 746 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 746 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 746 "ms_deisotope/_c/scoring.pyx"
        case  1:

#line 746 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_observed)) != 0)) kw_args--;

#line 746 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("evaluate", 1, 3, 3, 1); 
#line 746 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 746, __pyx_L3_error)

#line 746 "ms_deisotope/_c/scoring.pyx"
        }

#line 746 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 746 "ms_deisotope/_c/scoring.pyx"
        case  2:

#line 746 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_expected)) != 0)) kw_args--;

#line 746 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("evaluate", 1, 3, 3, 2); 
#line 746 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 746, __pyx_L3_error)

#line 746 "ms_deisotope/_c/scoring.pyx"
        }

#line 746 "ms_deisotope/_c/scoring.pyx"
      }

#line 746 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 746 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "evaluate") < 0)) __PYX_ERR(0, 746, __pyx_L3_error)

#line 746 "ms_deisotope/_c/scoring.pyx"
      }

#line 746 "ms_deisotope/_c/scoring.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 746 "ms_deisotope/_c/scoring.pyx"
      goto __pyx_L5_argtuple_error;

#line 746 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 746 "ms_deisotope/_c/scoring.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 746 "ms_deisotope/_c/scoring.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 746 "ms_deisotope/_c/scoring.pyx"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 746 "ms_deisotope/_c/scoring.pyx"
    }

#line 746 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_peaklist = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[0]);

#line 746 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_observed = ((PyObject*)values[1]);

#line 746 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_expected = ((PyObject*)values[2]);

#line 746 "ms_deisotope/_c/scoring.pyx"
  }

#line 746 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 746 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("evaluate", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 746 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 746, __pyx_L3_error)

#line 746 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 746 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.PenalizedMSDeconVFitter.evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 746 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 746 "ms_deisotope/_c/scoring.pyx"
  return NULL;

#line 746 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 746 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peaklist), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peaklist", 0))) __PYX_ERR(0, 746, __pyx_L1_error)

#line 746 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_observed), (&PyList_Type), 1, "observed", 1))) __PYX_ERR(0, 746, __pyx_L1_error)

#line 746 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_expected), (&PyList_Type), 1, "expected", 1))) __PYX_ERR(0, 746, __pyx_L1_error)
  __pyx_r = 
#line 746 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_8evaluate(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *)__pyx_v_self), __pyx_v_peaklist, __pyx_v_observed, __pyx_v_expected);

#line 746 "ms_deisotope/_c/scoring.pyx"


#line 746 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 746 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 746 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 746 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 746 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 746 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 746 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 746 "ms_deisotope/_c/scoring.pyx"
}

#line 746 "ms_deisotope/_c/scoring.pyx"


#line 746 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_8evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected) {

#line 746 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 746 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 746 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 746 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 746 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 746 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 746 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("evaluate", 0);

  /* "ms_deisotope/_c/scoring.pyx":747
 * 
 *     def evaluate(self, PeakSet peaklist, list observed, list expected):
 *         return self._evaluate(peaklist, observed, expected)             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision
 */

#line 747 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 747 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyFloat_FromDouble(((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._evaluate(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self), __pyx_v_peaklist, __pyx_v_observed, __pyx_v_expected, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 747, __pyx_L1_error)

#line 747 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 747 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 747 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 747 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":746
 *         self.select, self.penalty_factor, self.mass_error_tolerance = state
 * 
 *     def evaluate(self, PeakSet peaklist, list observed, list expected):             # <<<<<<<<<<<<<<
 *         return self._evaluate(peaklist, observed, expected)
 * 
 */

#line 746 "ms_deisotope/_c/scoring.pyx"


#line 746 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 746 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 746 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 746 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.PenalizedMSDeconVFitter.evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 746 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 746 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 746 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 746 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 746 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 746 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":750
 * 
 *     @cython.cdivision
 *     cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

#line 750 "ms_deisotope/_c/scoring.pyx"


#line 750 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_11_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 750 "ms_deisotope/_c/scoring.pyx"
static double __pyx_f_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter__evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected, int __pyx_skip_dispatch) {
  size_t __pyx_v_i
#line 750 "ms_deisotope/_c/scoring.pyx"
;
  size_t __pyx_v_n
#line 750 "ms_deisotope/_c/scoring.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_obs = 0
#line 750 "ms_deisotope/_c/scoring.pyx"
;
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_theo = 0
#line 750 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_score
#line 750 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_total_intensity_observed
#line 750 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_total_intensity_expected
#line 750 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_penalty
#line 750 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v__obs
#line 750 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v__theo
#line 750 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_log_ratio
#line 750 "ms_deisotope/_c/scoring.pyx"
;

#line 750 "ms_deisotope/_c/scoring.pyx"
  double __pyx_r;

#line 750 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 750 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 750 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 750 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 750 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 750 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_5;

#line 750 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 750 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_7;

#line 750 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_8;

#line 750 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_9;

#line 750 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_10;

#line 750 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_11;

#line 750 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_ms_deconv_score_peak __pyx_t_12;

#line 750 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 750 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 750 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 750 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate", 0);

#line 750 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 750 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 750 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 750 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 750 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 750 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 750 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 750 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 750 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 750 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_evaluate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 750, __pyx_L1_error)

#line 750 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 750 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_11_evaluate)) {

#line 750 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 750 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 750 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_5 = 0;

#line 750 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 750 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 750 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 750 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 750 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 750 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 750 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 750 "ms_deisotope/_c/scoring.pyx"
            __pyx_t_5 = 1;

#line 750 "ms_deisotope/_c/scoring.pyx"
          }

#line 750 "ms_deisotope/_c/scoring.pyx"
        }

#line 750 "ms_deisotope/_c/scoring.pyx"
        #if CYTHON_FAST_PYCALL

#line 750 "ms_deisotope/_c/scoring.pyx"
        if (PyFunction_Check(__pyx_t_3)) {

#line 750 "ms_deisotope/_c/scoring.pyx"
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_peaklist), __pyx_v_observed, __pyx_v_expected};

#line 750 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 750, __pyx_L1_error)

#line 750 "ms_deisotope/_c/scoring.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 750 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 750 "ms_deisotope/_c/scoring.pyx"
        } else

#line 750 "ms_deisotope/_c/scoring.pyx"
        #endif

#line 750 "ms_deisotope/_c/scoring.pyx"
        #if CYTHON_FAST_PYCCALL

#line 750 "ms_deisotope/_c/scoring.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {

#line 750 "ms_deisotope/_c/scoring.pyx"
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_peaklist), __pyx_v_observed, __pyx_v_expected};

#line 750 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 750, __pyx_L1_error)

#line 750 "ms_deisotope/_c/scoring.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 750 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 750 "ms_deisotope/_c/scoring.pyx"
        } else

#line 750 "ms_deisotope/_c/scoring.pyx"
        #endif

#line 750 "ms_deisotope/_c/scoring.pyx"
        {

#line 750 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_6 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 750, __pyx_L1_error)

#line 750 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_6);

#line 750 "ms_deisotope/_c/scoring.pyx"
          if (__pyx_t_4) {

#line 750 "ms_deisotope/_c/scoring.pyx"
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;

#line 750 "ms_deisotope/_c/scoring.pyx"
          }

#line 750 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(((PyObject *)__pyx_v_peaklist));

#line 750 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(((PyObject *)__pyx_v_peaklist));

#line 750 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_peaklist));

#line 750 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(__pyx_v_observed);

#line 750 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(__pyx_v_observed);

#line 750 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_observed);

#line 750 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(__pyx_v_expected);

#line 750 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(__pyx_v_expected);

#line 750 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, __pyx_v_expected);

#line 750 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 750, __pyx_L1_error)

#line 750 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 750 "ms_deisotope/_c/scoring.pyx"
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 750 "ms_deisotope/_c/scoring.pyx"
        }

#line 750 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 750 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_7 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_7 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 750, __pyx_L1_error)

#line 750 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 750 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = __pyx_t_7;

#line 750 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 750 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 750 "ms_deisotope/_c/scoring.pyx"
      }

#line 750 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 750 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 750 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 750 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 750 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 750 "ms_deisotope/_c/scoring.pyx"
      }

#line 750 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 750 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 750 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 750 "ms_deisotope/_c/scoring.pyx"
    }

#line 750 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 750 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":758
 *             double penalty, _obs, _theo, log_ratio
 * 
 *         total_intensity_observed = 0             # <<<<<<<<<<<<<<
 *         total_intensity_expected = 0
 * 
 */

#line 758 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_total_intensity_observed = 0.0;

  /* "ms_deisotope/_c/scoring.pyx":759
 * 
 *         total_intensity_observed = 0
 *         total_intensity_expected = 0             # <<<<<<<<<<<<<<
 * 
 *         n = PyList_GET_SIZE(observed)
 */

#line 759 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_total_intensity_expected = 0.0;

  /* "ms_deisotope/_c/scoring.pyx":761
 *         total_intensity_expected = 0
 * 
 *         n = PyList_GET_SIZE(observed)             # <<<<<<<<<<<<<<
 *         score = 0
 *         for i in range(n):
 */

#line 761 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_n = PyList_GET_SIZE(__pyx_v_observed);

  /* "ms_deisotope/_c/scoring.pyx":762
 * 
 *         n = PyList_GET_SIZE(observed)
 *         score = 0             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             obs = <FittedPeak>PyList_GET_ITEM(observed, i)
 */

#line 762 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_score = 0.0;

  /* "ms_deisotope/_c/scoring.pyx":763
 *         n = PyList_GET_SIZE(observed)
 *         score = 0
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             obs = <FittedPeak>PyList_GET_ITEM(observed, i)
 *             theo = <TheoreticalPeak>PyList_GET_ITEM(expected, i)
 */

#line 763 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_8 = __pyx_v_n;

#line 763 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_9 = __pyx_t_8;

#line 763 "ms_deisotope/_c/scoring.pyx"
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {

#line 763 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_i = __pyx_t_10;

    /* "ms_deisotope/_c/scoring.pyx":764
 *         score = 0
 *         for i in range(n):
 *             obs = <FittedPeak>PyList_GET_ITEM(observed, i)             # <<<<<<<<<<<<<<
 *             theo = <TheoreticalPeak>PyList_GET_ITEM(expected, i)
 *             score += ms_deconv_score_peak(obs, theo, self.mass_error_tolerance, 1)
 */

#line 764 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_11 = PyList_GET_ITEM(__pyx_v_observed, __pyx_v_i);

#line 764 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_11);

#line 764 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 764 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF_SET(__pyx_v_obs, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1));

#line 764 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/scoring.pyx":765
 *         for i in range(n):
 *             obs = <FittedPeak>PyList_GET_ITEM(observed, i)
 *             theo = <TheoreticalPeak>PyList_GET_ITEM(expected, i)             # <<<<<<<<<<<<<<
 *             score += ms_deconv_score_peak(obs, theo, self.mass_error_tolerance, 1)
 *             total_intensity_observed += obs.intensity
 */

#line 765 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_11 = PyList_GET_ITEM(__pyx_v_expected, __pyx_v_i);

#line 765 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_11);

#line 765 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 765 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF_SET(__pyx_v_theo, ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_1));

#line 765 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/scoring.pyx":766
 *             obs = <FittedPeak>PyList_GET_ITEM(observed, i)
 *             theo = <TheoreticalPeak>PyList_GET_ITEM(expected, i)
 *             score += ms_deconv_score_peak(obs, theo, self.mass_error_tolerance, 1)             # <<<<<<<<<<<<<<
 *             total_intensity_observed += obs.intensity
 *             total_intensity_expected += theo.intensity
 */

#line 766 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_12.__pyx_n = 2;

#line 766 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_12.mass_error_tolerance = __pyx_v_self->mass_error_tolerance;

#line 766 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_12.minimum_signal_to_noise = 1.0;

#line 766 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_7 = __pyx_f_12ms_deisotope_2_c_7scoring_ms_deconv_score_peak(__pyx_v_obs, __pyx_v_theo, &__pyx_t_12); 

#line 766 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_score = (__pyx_v_score + __pyx_t_7);

    /* "ms_deisotope/_c/scoring.pyx":767
 *             theo = <TheoreticalPeak>PyList_GET_ITEM(expected, i)
 *             score += ms_deconv_score_peak(obs, theo, self.mass_error_tolerance, 1)
 *             total_intensity_observed += obs.intensity             # <<<<<<<<<<<<<<
 *             total_intensity_expected += theo.intensity
 *         penalty = 0
 */

#line 767 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_total_intensity_observed = (__pyx_v_total_intensity_observed + __pyx_v_obs->__pyx_base.intensity);

    /* "ms_deisotope/_c/scoring.pyx":768
 *             score += ms_deconv_score_peak(obs, theo, self.mass_error_tolerance, 1)
 *             total_intensity_observed += obs.intensity
 *             total_intensity_expected += theo.intensity             # <<<<<<<<<<<<<<
 *         penalty = 0
 *         for i in range(n):
 */

#line 768 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_total_intensity_expected = (__pyx_v_total_intensity_expected + __pyx_v_theo->intensity);

#line 768 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":769
 *             total_intensity_observed += obs.intensity
 *             total_intensity_expected += theo.intensity
 *         penalty = 0             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             _obs = (<FittedPeak>PyList_GET_ITEM(observed, i)).intensity / total_intensity_observed
 */

#line 769 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_penalty = 0.0;

  /* "ms_deisotope/_c/scoring.pyx":770
 *             total_intensity_expected += theo.intensity
 *         penalty = 0
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             _obs = (<FittedPeak>PyList_GET_ITEM(observed, i)).intensity / total_intensity_observed
 *             _theo = (<TheoreticalPeak>PyList_GET_ITEM(expected, i)).intensity / total_intensity_expected
 */

#line 770 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_8 = __pyx_v_n;

#line 770 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_9 = __pyx_t_8;

#line 770 "ms_deisotope/_c/scoring.pyx"
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {

#line 770 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_i = __pyx_t_10;

    /* "ms_deisotope/_c/scoring.pyx":771
 *         penalty = 0
 *         for i in range(n):
 *             _obs = (<FittedPeak>PyList_GET_ITEM(observed, i)).intensity / total_intensity_observed             # <<<<<<<<<<<<<<
 *             _theo = (<TheoreticalPeak>PyList_GET_ITEM(expected, i)).intensity / total_intensity_expected
 * 
 */

#line 771 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_11 = PyList_GET_ITEM(__pyx_v_observed, __pyx_v_i);

#line 771 "ms_deisotope/_c/scoring.pyx"
    __pyx_v__obs = (((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_11)->__pyx_base.intensity / __pyx_v_total_intensity_observed);

    /* "ms_deisotope/_c/scoring.pyx":772
 *         for i in range(n):
 *             _obs = (<FittedPeak>PyList_GET_ITEM(observed, i)).intensity / total_intensity_observed
 *             _theo = (<TheoreticalPeak>PyList_GET_ITEM(expected, i)).intensity / total_intensity_expected             # <<<<<<<<<<<<<<
 * 
 *             log_ratio = log(_obs) - log(_theo)
 */

#line 772 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_11 = PyList_GET_ITEM(__pyx_v_expected, __pyx_v_i);

#line 772 "ms_deisotope/_c/scoring.pyx"
    __pyx_v__theo = (((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_11)->intensity / __pyx_v_total_intensity_expected);

    /* "ms_deisotope/_c/scoring.pyx":774
 *             _theo = (<TheoreticalPeak>PyList_GET_ITEM(expected, i)).intensity / total_intensity_expected
 * 
 *             log_ratio = log(_obs) - log(_theo)             # <<<<<<<<<<<<<<
 *             penalty += _obs * log_ratio
 *         penalty = abs(2 * penalty)
 */

#line 774 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_log_ratio = (log(__pyx_v__obs) - log(__pyx_v__theo));

    /* "ms_deisotope/_c/scoring.pyx":775
 * 
 *             log_ratio = log(_obs) - log(_theo)
 *             penalty += _obs * log_ratio             # <<<<<<<<<<<<<<
 *         penalty = abs(2 * penalty)
 *         score *= ((1 - penalty * self.penalty_factor))
 */

#line 775 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_penalty = (__pyx_v_penalty + (__pyx_v__obs * __pyx_v_log_ratio));

#line 775 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":776
 *             log_ratio = log(_obs) - log(_theo)
 *             penalty += _obs * log_ratio
 *         penalty = abs(2 * penalty)             # <<<<<<<<<<<<<<
 *         score *= ((1 - penalty * self.penalty_factor))
 *         return score
 */

#line 776 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_penalty = fabs((2.0 * __pyx_v_penalty));

  /* "ms_deisotope/_c/scoring.pyx":777
 *             penalty += _obs * log_ratio
 *         penalty = abs(2 * penalty)
 *         score *= ((1 - penalty * self.penalty_factor))             # <<<<<<<<<<<<<<
 *         return score
 * 
 */

#line 777 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_score = (__pyx_v_score * (1.0 - (__pyx_v_penalty * __pyx_v_self->penalty_factor)));

  /* "ms_deisotope/_c/scoring.pyx":778
 *         penalty = abs(2 * penalty)
 *         score *= ((1 - penalty * self.penalty_factor))
 *         return score             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 778 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_v_score;

#line 778 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":750
 * 
 *     @cython.cdivision
 *     cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

#line 750 "ms_deisotope/_c/scoring.pyx"


#line 750 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 750 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 750 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 750 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 750 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 750 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 750 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 750 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.PenalizedMSDeconVFitter._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 750 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 750 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 750 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_obs);

#line 750 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_theo);

#line 750 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 750 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 750 "ms_deisotope/_c/scoring.pyx"
}

#line 750 "ms_deisotope/_c/scoring.pyx"


#line 750 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 750 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_11_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 750 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_10_evaluate[] = "PenalizedMSDeconVFitter._evaluate(self, PeakSet peaklist, list observed, list expected) -> double";

#line 750 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_11_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist = 0
#line 750 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_observed = 0
#line 750 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_expected = 0
#line 750 "ms_deisotope/_c/scoring.pyx"
;

#line 750 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 750 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 750 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 750 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 750 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 750 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate (wrapper)", 0);

#line 750 "ms_deisotope/_c/scoring.pyx"
  {

#line 750 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peaklist,&__pyx_n_s_observed,&__pyx_n_s_expected,0};

#line 750 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[3] = {0,0,0};

#line 750 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 750 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 750 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 750 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  3: 
#line 750 "ms_deisotope/_c/scoring.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 750 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 750 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 750 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 750 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 750 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 750 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 750 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 750 "ms_deisotope/_c/scoring.pyx"
      }

#line 750 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 750 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 750 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 750 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaklist)) != 0)) kw_args--;
        else 
#line 750 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 750 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 750 "ms_deisotope/_c/scoring.pyx"
        case  1:

#line 750 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_observed)) != 0)) kw_args--;

#line 750 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, 1); 
#line 750 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 750, __pyx_L3_error)

#line 750 "ms_deisotope/_c/scoring.pyx"
        }

#line 750 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 750 "ms_deisotope/_c/scoring.pyx"
        case  2:

#line 750 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_expected)) != 0)) kw_args--;

#line 750 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, 2); 
#line 750 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 750, __pyx_L3_error)

#line 750 "ms_deisotope/_c/scoring.pyx"
        }

#line 750 "ms_deisotope/_c/scoring.pyx"
      }

#line 750 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 750 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_evaluate") < 0)) __PYX_ERR(0, 750, __pyx_L3_error)

#line 750 "ms_deisotope/_c/scoring.pyx"
      }

#line 750 "ms_deisotope/_c/scoring.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 750 "ms_deisotope/_c/scoring.pyx"
      goto __pyx_L5_argtuple_error;

#line 750 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 750 "ms_deisotope/_c/scoring.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 750 "ms_deisotope/_c/scoring.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 750 "ms_deisotope/_c/scoring.pyx"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 750 "ms_deisotope/_c/scoring.pyx"
    }

#line 750 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_peaklist = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[0]);

#line 750 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_observed = ((PyObject*)values[1]);

#line 750 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_expected = ((PyObject*)values[2]);

#line 750 "ms_deisotope/_c/scoring.pyx"
  }

#line 750 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 750 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 750 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 750, __pyx_L3_error)

#line 750 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 750 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.PenalizedMSDeconVFitter._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 750 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 750 "ms_deisotope/_c/scoring.pyx"
  return NULL;

#line 750 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 750 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peaklist), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peaklist", 0))) __PYX_ERR(0, 750, __pyx_L1_error)

#line 750 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_observed), (&PyList_Type), 1, "observed", 1))) __PYX_ERR(0, 750, __pyx_L1_error)

#line 750 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_expected), (&PyList_Type), 1, "expected", 1))) __PYX_ERR(0, 750, __pyx_L1_error)
  __pyx_r = 
#line 750 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_10_evaluate(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *)__pyx_v_self), __pyx_v_peaklist, __pyx_v_observed, __pyx_v_expected);

#line 750 "ms_deisotope/_c/scoring.pyx"


#line 750 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 750 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 750 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 750 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 750 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 750 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 750 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 750 "ms_deisotope/_c/scoring.pyx"
}

#line 750 "ms_deisotope/_c/scoring.pyx"


#line 750 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_10_evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected) {

#line 750 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 750 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 750 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 750 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 750 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 750 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 750 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate", 0);

#line 750 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 750 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter__evaluate(__pyx_v_self, __pyx_v_peaklist, __pyx_v_observed, __pyx_v_expected, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 750, __pyx_L1_error)

#line 750 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 750 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 750 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 750 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 750 "ms_deisotope/_c/scoring.pyx"


#line 750 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 750 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 750 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 750 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.PenalizedMSDeconVFitter._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 750 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 750 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 750 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 750 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 750 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 750 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pxd":80
 * cdef class PenalizedMSDeconVFitter(IsotopicFitterBase):
 *     cdef:
 *         public double mass_error_tolerance             # <<<<<<<<<<<<<<
 *         public double penalty_factor
 * 
 */

#line 80 "ms_deisotope/_c/scoring.pxd"


#line 80 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 80 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_20mass_error_tolerance_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 80 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_20mass_error_tolerance_1__get__(PyObject *__pyx_v_self) {

#line 80 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = 0;

#line 80 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 80 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 80 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_20mass_error_tolerance___get__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *)__pyx_v_self));

#line 80 "ms_deisotope/_c/scoring.pxd"


#line 80 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 80 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 80 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 80 "ms_deisotope/_c/scoring.pxd"
}

#line 80 "ms_deisotope/_c/scoring.pxd"


#line 80 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_20mass_error_tolerance___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_v_self) {

#line 80 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = NULL;

#line 80 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 80 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 80 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 80 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 80 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 80 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 80 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 80 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->mass_error_tolerance); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 80, __pyx_L1_error)

#line 80 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 80 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = __pyx_t_1;

#line 80 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = 0;

#line 80 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 80 "ms_deisotope/_c/scoring.pxd"


#line 80 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 80 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 80 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 80 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.PenalizedMSDeconVFitter.mass_error_tolerance.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 80 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = NULL;

#line 80 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 80 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 80 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 80 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 80 "ms_deisotope/_c/scoring.pxd"
}

#line 80 "ms_deisotope/_c/scoring.pxd"


#line 80 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 80 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_20mass_error_tolerance_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 80 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_20mass_error_tolerance_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 80 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 80 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 80 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 80 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_20mass_error_tolerance_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 80 "ms_deisotope/_c/scoring.pxd"


#line 80 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 80 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 80 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 80 "ms_deisotope/_c/scoring.pxd"
}

#line 80 "ms_deisotope/_c/scoring.pxd"


#line 80 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_20mass_error_tolerance_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_v_self, PyObject *__pyx_v_value) {

#line 80 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 80 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 80 "ms_deisotope/_c/scoring.pxd"
  double __pyx_t_1;

#line 80 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 80 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 80 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 80 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 80 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 80, __pyx_L1_error)

#line 80 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->mass_error_tolerance = __pyx_t_1;

#line 80 "ms_deisotope/_c/scoring.pxd"


#line 80 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 80 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 80 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 80 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 80 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.PenalizedMSDeconVFitter.mass_error_tolerance.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 80 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = -1;

#line 80 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 80 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 80 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 80 "ms_deisotope/_c/scoring.pxd"
}

/* "ms_deisotope/_c/scoring.pxd":81
 *     cdef:
 *         public double mass_error_tolerance
 *         public double penalty_factor             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 81 "ms_deisotope/_c/scoring.pxd"


#line 81 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 81 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_14penalty_factor_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 81 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_14penalty_factor_1__get__(PyObject *__pyx_v_self) {

#line 81 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = 0;

#line 81 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 81 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 81 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_14penalty_factor___get__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *)__pyx_v_self));

#line 81 "ms_deisotope/_c/scoring.pxd"


#line 81 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 81 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 81 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 81 "ms_deisotope/_c/scoring.pxd"
}

#line 81 "ms_deisotope/_c/scoring.pxd"


#line 81 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_14penalty_factor___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_v_self) {

#line 81 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = NULL;

#line 81 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 81 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 81 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 81 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 81 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 81 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 81 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 81 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->penalty_factor); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 81, __pyx_L1_error)

#line 81 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 81 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = __pyx_t_1;

#line 81 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = 0;

#line 81 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 81 "ms_deisotope/_c/scoring.pxd"


#line 81 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 81 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 81 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 81 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.PenalizedMSDeconVFitter.penalty_factor.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 81 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = NULL;

#line 81 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 81 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 81 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 81 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 81 "ms_deisotope/_c/scoring.pxd"
}

#line 81 "ms_deisotope/_c/scoring.pxd"


#line 81 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 81 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_14penalty_factor_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 81 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_14penalty_factor_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 81 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 81 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 81 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 81 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_14penalty_factor_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 81 "ms_deisotope/_c/scoring.pxd"


#line 81 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 81 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 81 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 81 "ms_deisotope/_c/scoring.pxd"
}

#line 81 "ms_deisotope/_c/scoring.pxd"


#line 81 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_14penalty_factor_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_v_self, PyObject *__pyx_v_value) {

#line 81 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 81 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 81 "ms_deisotope/_c/scoring.pxd"
  double __pyx_t_1;

#line 81 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 81 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 81 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 81 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 81 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 81, __pyx_L1_error)

#line 81 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->penalty_factor = __pyx_t_1;

#line 81 "ms_deisotope/_c/scoring.pxd"


#line 81 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 81 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 81 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 81 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 81 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.PenalizedMSDeconVFitter.penalty_factor.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 81 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = -1;

#line 81 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 81 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 81 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 81 "ms_deisotope/_c/scoring.pxd"
}

/* "ms_deisotope/_c/scoring.pyx":795
 *     """
 * 
 *     def __init__(self, function, minimum_score=10, selector_type=MaximizeFitSelector):             # <<<<<<<<<<<<<<
 *         self.function = function
 *         self.select = selector_type(minimum_score)
 */

#line 795 "ms_deisotope/_c/scoring.pyx"


#line 795 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 795 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_14FunctionScorer_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 795 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_14FunctionScorer_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_function = 0
#line 795 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_minimum_score = 0
#line 795 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_selector_type = 0
#line 795 "ms_deisotope/_c/scoring.pyx"
;

#line 795 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 795 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 795 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 795 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 795 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 795 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 795 "ms_deisotope/_c/scoring.pyx"
  {

#line 795 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_function,&__pyx_n_s_minimum_score,&__pyx_n_s_selector_type,0};

#line 795 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[3] = {0,0,0};

#line 795 "ms_deisotope/_c/scoring.pyx"
    values[1] = ((PyObject *)__pyx_int_10);

#line 795 "ms_deisotope/_c/scoring.pyx"
    values[2] = __pyx_k__2;

#line 795 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 795 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 795 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 795 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  3: 
#line 795 "ms_deisotope/_c/scoring.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 795 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 795 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 795 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 795 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 795 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 795 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 795 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 795 "ms_deisotope/_c/scoring.pyx"
      }

#line 795 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 795 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 795 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 795 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_function)) != 0)) kw_args--;
        else 
#line 795 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 795 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 795 "ms_deisotope/_c/scoring.pyx"
        case  1:

#line 795 "ms_deisotope/_c/scoring.pyx"
        if (kw_args > 0) {

#line 795 "ms_deisotope/_c/scoring.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_minimum_score);

#line 795 "ms_deisotope/_c/scoring.pyx"
          if (value) { values[1] = value; kw_args--; }

#line 795 "ms_deisotope/_c/scoring.pyx"
        }

#line 795 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 795 "ms_deisotope/_c/scoring.pyx"
        case  2:

#line 795 "ms_deisotope/_c/scoring.pyx"
        if (kw_args > 0) {

#line 795 "ms_deisotope/_c/scoring.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_selector_type);

#line 795 "ms_deisotope/_c/scoring.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 795 "ms_deisotope/_c/scoring.pyx"
        }

#line 795 "ms_deisotope/_c/scoring.pyx"
      }

#line 795 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 795 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 795, __pyx_L3_error)

#line 795 "ms_deisotope/_c/scoring.pyx"
      }

#line 795 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 795 "ms_deisotope/_c/scoring.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 795 "ms_deisotope/_c/scoring.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 795 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 795 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 795 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 795 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 795 "ms_deisotope/_c/scoring.pyx"
        break;
        default: 
#line 795 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 795 "ms_deisotope/_c/scoring.pyx"
      }

#line 795 "ms_deisotope/_c/scoring.pyx"
    }

#line 795 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_function = values[0];

#line 795 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_minimum_score = values[1];

#line 795 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_selector_type = values[2];

#line 795 "ms_deisotope/_c/scoring.pyx"
  }

#line 795 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 795 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 795 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 795, __pyx_L3_error)

#line 795 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 795 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.FunctionScorer.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 795 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 795 "ms_deisotope/_c/scoring.pyx"
  return -1;

#line 795 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 795 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_14FunctionScorer___init__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer *)__pyx_v_self), __pyx_v_function, __pyx_v_minimum_score, __pyx_v_selector_type);

#line 795 "ms_deisotope/_c/scoring.pyx"


#line 795 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 795 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 795 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 795 "ms_deisotope/_c/scoring.pyx"
}

#line 795 "ms_deisotope/_c/scoring.pyx"


#line 795 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_14FunctionScorer___init__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer *__pyx_v_self, PyObject *__pyx_v_function, PyObject *__pyx_v_minimum_score, PyObject *__pyx_v_selector_type) {

#line 795 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 795 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 795 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 795 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 795 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 795 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 795 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 795 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 795 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_deisotope/_c/scoring.pyx":796
 * 
 *     def __init__(self, function, minimum_score=10, selector_type=MaximizeFitSelector):
 *         self.function = function             # <<<<<<<<<<<<<<
 *         self.select = selector_type(minimum_score)
 * 
 */

#line 796 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_v_function);

#line 796 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_v_function);

#line 796 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->function);

#line 796 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_v_self->function);

#line 796 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->function = __pyx_v_function;

  /* "ms_deisotope/_c/scoring.pyx":797
 *     def __init__(self, function, minimum_score=10, selector_type=MaximizeFitSelector):
 *         self.function = function
 *         self.select = selector_type(minimum_score)             # <<<<<<<<<<<<<<
 * 
 *     def __getstate__(self):
 */

#line 797 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_v_selector_type);
  __pyx_t_2 = __pyx_v_selector_type; 
#line 797 "ms_deisotope/_c/scoring.pyx"
__pyx_t_3 = NULL;

#line 797 "ms_deisotope/_c/scoring.pyx"
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {

#line 797 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);

#line 797 "ms_deisotope/_c/scoring.pyx"
    if (likely(__pyx_t_3)) {

#line 797 "ms_deisotope/_c/scoring.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 797 "ms_deisotope/_c/scoring.pyx"
      __Pyx_INCREF(__pyx_t_3);

#line 797 "ms_deisotope/_c/scoring.pyx"
      __Pyx_INCREF(function);

#line 797 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF_SET(__pyx_t_2, function);

#line 797 "ms_deisotope/_c/scoring.pyx"
    }

#line 797 "ms_deisotope/_c/scoring.pyx"
  }

#line 797 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_minimum_score) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_minimum_score);

#line 797 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 797 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 797, __pyx_L1_error)

#line 797 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 797 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 797 "ms_deisotope/_c/scoring.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase))))) __PYX_ERR(0, 797, __pyx_L1_error)

#line 797 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 797 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.select);

#line 797 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx_base.select));

#line 797 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->__pyx_base.select = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_t_1);

#line 797 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/scoring.pyx":795
 *     """
 * 
 *     def __init__(self, function, minimum_score=10, selector_type=MaximizeFitSelector):             # <<<<<<<<<<<<<<
 *         self.function = function
 *         self.select = selector_type(minimum_score)
 */

#line 795 "ms_deisotope/_c/scoring.pyx"


#line 795 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 795 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 795 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 795 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 795 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 795 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 795 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 795 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.FunctionScorer.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 795 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = -1;

#line 795 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 795 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 795 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 795 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":799
 *         self.select = selector_type(minimum_score)
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self.select, self.function
 * 
 */

#line 799 "ms_deisotope/_c/scoring.pyx"


#line 799 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 799 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_14FunctionScorer_3__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 799 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_14FunctionScorer_2__getstate__[] = "FunctionScorer.__getstate__(self)";

#line 799 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_14FunctionScorer_3__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 799 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 799 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 799 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = 
#line 799 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_14FunctionScorer_2__getstate__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer *)__pyx_v_self));

#line 799 "ms_deisotope/_c/scoring.pyx"


#line 799 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 799 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 799 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 799 "ms_deisotope/_c/scoring.pyx"
}

#line 799 "ms_deisotope/_c/scoring.pyx"


#line 799 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_14FunctionScorer_2__getstate__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer *__pyx_v_self) {

#line 799 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 799 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 799 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 799 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 799 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 799 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 799 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "ms_deisotope/_c/scoring.pyx":800
 * 
 *     def __getstate__(self):
 *         return self.select, self.function             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, state):
 */

#line 800 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 800 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 800, __pyx_L1_error)

#line 800 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 800 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx_base.select));

#line 800 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx_base.select));

#line 800 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->__pyx_base.select));

#line 800 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_v_self->function);

#line 800 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_v_self->function);

#line 800 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->function);

#line 800 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 800 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 800 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":799
 *         self.select = selector_type(minimum_score)
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self.select, self.function
 * 
 */

#line 799 "ms_deisotope/_c/scoring.pyx"


#line 799 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 799 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 799 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 799 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.FunctionScorer.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 799 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 799 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 799 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 799 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 799 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 799 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":802
 *         return self.select, self.function
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.select, self.function = state
 * 
 */

#line 802 "ms_deisotope/_c/scoring.pyx"


#line 802 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 802 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_14FunctionScorer_5__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state); /*proto*/

#line 802 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_14FunctionScorer_4__setstate__[] = "FunctionScorer.__setstate__(self, state)";

#line 802 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_14FunctionScorer_5__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state) {

#line 802 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 802 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 802 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  __pyx_r = 
#line 802 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_14FunctionScorer_4__setstate__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer *)__pyx_v_self), ((PyObject *)__pyx_v_state));

#line 802 "ms_deisotope/_c/scoring.pyx"


#line 802 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 802 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 802 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 802 "ms_deisotope/_c/scoring.pyx"
}

#line 802 "ms_deisotope/_c/scoring.pyx"


#line 802 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_14FunctionScorer_4__setstate__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer *__pyx_v_self, PyObject *__pyx_v_state) {

#line 802 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 802 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 802 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 802 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 802 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 802 "ms_deisotope/_c/scoring.pyx"
  PyObject *(*__pyx_t_4)(PyObject *);

#line 802 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 802 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 802 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 802 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* "ms_deisotope/_c/scoring.pyx":803
 * 
 *     def __setstate__(self, state):
 *         self.select, self.function = state             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */

#line 803 "ms_deisotope/_c/scoring.pyx"
  if ((likely(PyTuple_CheckExact(__pyx_v_state))) || (PyList_CheckExact(__pyx_v_state))) {

#line 803 "ms_deisotope/_c/scoring.pyx"
    PyObject* sequence = __pyx_v_state;

#line 803 "ms_deisotope/_c/scoring.pyx"
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);

#line 803 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(size != 2)) {

#line 803 "ms_deisotope/_c/scoring.pyx"
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);

#line 803 "ms_deisotope/_c/scoring.pyx"
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);

#line 803 "ms_deisotope/_c/scoring.pyx"
      __PYX_ERR(0, 803, __pyx_L1_error)

#line 803 "ms_deisotope/_c/scoring.pyx"
    }

#line 803 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 803 "ms_deisotope/_c/scoring.pyx"
    if (likely(PyTuple_CheckExact(sequence))) {

#line 803 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 

#line 803 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 

#line 803 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 803 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 

#line 803 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 

#line 803 "ms_deisotope/_c/scoring.pyx"
    }

#line 803 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 803 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_2);

#line 803 "ms_deisotope/_c/scoring.pyx"
    #else

#line 803 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 803, __pyx_L1_error)

#line 803 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 803 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 803, __pyx_L1_error)

#line 803 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 803 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 803 "ms_deisotope/_c/scoring.pyx"
  } else {

#line 803 "ms_deisotope/_c/scoring.pyx"
    Py_ssize_t index = -1;

#line 803 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_3 = PyObject_GetIter(__pyx_v_state); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 803, __pyx_L1_error)

#line 803 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 803 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_4 = Py_TYPE(__pyx_t_3)->tp_iternext;
    index = 0; __pyx_t_1 = __pyx_t_4(__pyx_t_3); if (unlikely(!__pyx_t_1)) 
#line 803 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L3_unpacking_failed;

#line 803 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_1);
    index = 1; __pyx_t_2 = __pyx_t_4(__pyx_t_3); if (unlikely(!__pyx_t_2)) 
#line 803 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L3_unpacking_failed;

#line 803 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 803 "ms_deisotope/_c/scoring.pyx"
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_4(__pyx_t_3), 2) < 0) __PYX_ERR(0, 803, __pyx_L1_error)

#line 803 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_4 = NULL;

#line 803 "ms_deisotope/_c/scoring.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 803 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L4_unpacking_done;

#line 803 "ms_deisotope/_c/scoring.pyx"
    __pyx_L3_unpacking_failed:;

#line 803 "ms_deisotope/_c/scoring.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 803 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_4 = NULL;

#line 803 "ms_deisotope/_c/scoring.pyx"
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);

#line 803 "ms_deisotope/_c/scoring.pyx"
    __PYX_ERR(0, 803, __pyx_L1_error)

#line 803 "ms_deisotope/_c/scoring.pyx"
    __pyx_L4_unpacking_done:;

#line 803 "ms_deisotope/_c/scoring.pyx"
  }

#line 803 "ms_deisotope/_c/scoring.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase))))) __PYX_ERR(0, 803, __pyx_L1_error)

#line 803 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 803 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.select);

#line 803 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx_base.select));

#line 803 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->__pyx_base.select = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_t_1);

#line 803 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 803 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 803 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->function);

#line 803 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_v_self->function);

#line 803 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->function = __pyx_t_2;

#line 803 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/scoring.pyx":802
 *         return self.select, self.function
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.select, self.function = state
 * 
 */

#line 802 "ms_deisotope/_c/scoring.pyx"


#line 802 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 802 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 802 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 802 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 802 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 802 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 802 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 802 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.FunctionScorer.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 802 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 802 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 802 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 802 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 802 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 802 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":805
 *         self.select, self.function = state
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (None,), self.__getstate__()
 * 
 */

#line 805 "ms_deisotope/_c/scoring.pyx"


#line 805 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 805 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_14FunctionScorer_7__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 805 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_14FunctionScorer_6__reduce__[] = "FunctionScorer.__reduce__(self)";

#line 805 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_14FunctionScorer_7__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 805 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 805 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 805 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = 
#line 805 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_14FunctionScorer_6__reduce__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer *)__pyx_v_self));

#line 805 "ms_deisotope/_c/scoring.pyx"


#line 805 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 805 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 805 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 805 "ms_deisotope/_c/scoring.pyx"
}

#line 805 "ms_deisotope/_c/scoring.pyx"


#line 805 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_14FunctionScorer_6__reduce__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer *__pyx_v_self) {

#line 805 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 805 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 805 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 805 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 805 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 805 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 805 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 805 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 805 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 805 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "ms_deisotope/_c/scoring.pyx":806
 * 
 *     def __reduce__(self):
 *         return self.__class__, (None,), self.__getstate__()             # <<<<<<<<<<<<<<
 * 
 *     cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):
 */

#line 806 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 806 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 806, __pyx_L1_error)

#line 806 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 806 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_getstate); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 806, __pyx_L1_error)

#line 806 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 806 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_4 = NULL;

#line 806 "ms_deisotope/_c/scoring.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {

#line 806 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 806 "ms_deisotope/_c/scoring.pyx"
    if (likely(__pyx_t_4)) {

#line 806 "ms_deisotope/_c/scoring.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 806 "ms_deisotope/_c/scoring.pyx"
      __Pyx_INCREF(__pyx_t_4);

#line 806 "ms_deisotope/_c/scoring.pyx"
      __Pyx_INCREF(function);

#line 806 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF_SET(__pyx_t_3, function);

#line 806 "ms_deisotope/_c/scoring.pyx"
    }

#line 806 "ms_deisotope/_c/scoring.pyx"
  }

#line 806 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 806 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 806 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 806, __pyx_L1_error)

#line 806 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 806 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 806 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 806, __pyx_L1_error)

#line 806 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 806 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 806 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);

#line 806 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_tuple__3);

#line 806 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_tuple__3);

#line 806 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_tuple__3);

#line 806 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 806 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);

#line 806 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 806 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = 0;

#line 806 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_3;

#line 806 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = 0;

#line 806 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":805
 *         self.select, self.function = state
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (None,), self.__getstate__()
 * 
 */

#line 805 "ms_deisotope/_c/scoring.pyx"


#line 805 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 805 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 805 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 805 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 805 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 805 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 805 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.FunctionScorer.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 805 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 805 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 805 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 805 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 805 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 805 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":808
 *         return self.__class__, (None,), self.__getstate__()
 * 
 *     cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):             # <<<<<<<<<<<<<<
 *         return self.function(observed, expected)
 * 
 */

#line 808 "ms_deisotope/_c/scoring.pyx"


#line 808 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_14FunctionScorer_9_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 808 "ms_deisotope/_c/scoring.pyx"
static double __pyx_f_12ms_deisotope_2_c_7scoring_14FunctionScorer__evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected, int __pyx_skip_dispatch) {

#line 808 "ms_deisotope/_c/scoring.pyx"
  double __pyx_r;

#line 808 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 808 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 808 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 808 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 808 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 808 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_5;

#line 808 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 808 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_7;

#line 808 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 808 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 808 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 808 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate", 0);

#line 808 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 808 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 808 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 808 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 808 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 808 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 808 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 808 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 808 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 808 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_evaluate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 808, __pyx_L1_error)

#line 808 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 808 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_14FunctionScorer_9_evaluate)) {

#line 808 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 808 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 808 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_5 = 0;

#line 808 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 808 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 808 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 808 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 808 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 808 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 808 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 808 "ms_deisotope/_c/scoring.pyx"
            __pyx_t_5 = 1;

#line 808 "ms_deisotope/_c/scoring.pyx"
          }

#line 808 "ms_deisotope/_c/scoring.pyx"
        }

#line 808 "ms_deisotope/_c/scoring.pyx"
        #if CYTHON_FAST_PYCALL

#line 808 "ms_deisotope/_c/scoring.pyx"
        if (PyFunction_Check(__pyx_t_3)) {

#line 808 "ms_deisotope/_c/scoring.pyx"
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_peaklist), __pyx_v_observed, __pyx_v_expected};

#line 808 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 808, __pyx_L1_error)

#line 808 "ms_deisotope/_c/scoring.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 808 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 808 "ms_deisotope/_c/scoring.pyx"
        } else

#line 808 "ms_deisotope/_c/scoring.pyx"
        #endif

#line 808 "ms_deisotope/_c/scoring.pyx"
        #if CYTHON_FAST_PYCCALL

#line 808 "ms_deisotope/_c/scoring.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {

#line 808 "ms_deisotope/_c/scoring.pyx"
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_peaklist), __pyx_v_observed, __pyx_v_expected};

#line 808 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 808, __pyx_L1_error)

#line 808 "ms_deisotope/_c/scoring.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 808 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 808 "ms_deisotope/_c/scoring.pyx"
        } else

#line 808 "ms_deisotope/_c/scoring.pyx"
        #endif

#line 808 "ms_deisotope/_c/scoring.pyx"
        {

#line 808 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_6 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 808, __pyx_L1_error)

#line 808 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_6);

#line 808 "ms_deisotope/_c/scoring.pyx"
          if (__pyx_t_4) {

#line 808 "ms_deisotope/_c/scoring.pyx"
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;

#line 808 "ms_deisotope/_c/scoring.pyx"
          }

#line 808 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(((PyObject *)__pyx_v_peaklist));

#line 808 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(((PyObject *)__pyx_v_peaklist));

#line 808 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_peaklist));

#line 808 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(__pyx_v_observed);

#line 808 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(__pyx_v_observed);

#line 808 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_observed);

#line 808 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(__pyx_v_expected);

#line 808 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(__pyx_v_expected);

#line 808 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, __pyx_v_expected);

#line 808 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 808, __pyx_L1_error)

#line 808 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 808 "ms_deisotope/_c/scoring.pyx"
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 808 "ms_deisotope/_c/scoring.pyx"
        }

#line 808 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 808 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_7 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_7 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 808, __pyx_L1_error)

#line 808 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 808 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = __pyx_t_7;

#line 808 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 808 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 808 "ms_deisotope/_c/scoring.pyx"
      }

#line 808 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 808 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 808 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 808 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 808 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 808 "ms_deisotope/_c/scoring.pyx"
      }

#line 808 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 808 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 808 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 808 "ms_deisotope/_c/scoring.pyx"
    }

#line 808 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 808 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":809
 * 
 *     cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):
 *         return self.function(observed, expected)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 809 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_v_self->function);
  __pyx_t_2 = __pyx_v_self->function; 
#line 809 "ms_deisotope/_c/scoring.pyx"
__pyx_t_3 = NULL;

#line 809 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_5 = 0;

#line 809 "ms_deisotope/_c/scoring.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 809 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);

#line 809 "ms_deisotope/_c/scoring.pyx"
    if (likely(__pyx_t_3)) {

#line 809 "ms_deisotope/_c/scoring.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 809 "ms_deisotope/_c/scoring.pyx"
      __Pyx_INCREF(__pyx_t_3);

#line 809 "ms_deisotope/_c/scoring.pyx"
      __Pyx_INCREF(function);

#line 809 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF_SET(__pyx_t_2, function);

#line 809 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_5 = 1;

#line 809 "ms_deisotope/_c/scoring.pyx"
    }

#line 809 "ms_deisotope/_c/scoring.pyx"
  }

#line 809 "ms_deisotope/_c/scoring.pyx"
  #if CYTHON_FAST_PYCALL

#line 809 "ms_deisotope/_c/scoring.pyx"
  if (PyFunction_Check(__pyx_t_2)) {

#line 809 "ms_deisotope/_c/scoring.pyx"
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_observed, __pyx_v_expected};

#line 809 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 809, __pyx_L1_error)

#line 809 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 809 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 809 "ms_deisotope/_c/scoring.pyx"
  } else

#line 809 "ms_deisotope/_c/scoring.pyx"
  #endif

#line 809 "ms_deisotope/_c/scoring.pyx"
  #if CYTHON_FAST_PYCCALL

#line 809 "ms_deisotope/_c/scoring.pyx"
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {

#line 809 "ms_deisotope/_c/scoring.pyx"
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_observed, __pyx_v_expected};

#line 809 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 809, __pyx_L1_error)

#line 809 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 809 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 809 "ms_deisotope/_c/scoring.pyx"
  } else

#line 809 "ms_deisotope/_c/scoring.pyx"
  #endif

#line 809 "ms_deisotope/_c/scoring.pyx"
  {

#line 809 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 809, __pyx_L1_error)

#line 809 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 809 "ms_deisotope/_c/scoring.pyx"
    if (__pyx_t_3) {

#line 809 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;

#line 809 "ms_deisotope/_c/scoring.pyx"
    }

#line 809 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_v_observed);

#line 809 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GIVEREF(__pyx_v_observed);

#line 809 "ms_deisotope/_c/scoring.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_observed);

#line 809 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_v_expected);

#line 809 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GIVEREF(__pyx_v_expected);

#line 809 "ms_deisotope/_c/scoring.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_expected);

#line 809 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 809, __pyx_L1_error)

#line 809 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 809 "ms_deisotope/_c/scoring.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 809 "ms_deisotope/_c/scoring.pyx"
  }

#line 809 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 809 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_7 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_7 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 809, __pyx_L1_error)

#line 809 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 809 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_7;

#line 809 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":808
 *         return self.__class__, (None,), self.__getstate__()
 * 
 *     cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):             # <<<<<<<<<<<<<<
 *         return self.function(observed, expected)
 * 
 */

#line 808 "ms_deisotope/_c/scoring.pyx"


#line 808 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 808 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 808 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 808 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 808 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 808 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 808 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 808 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.FunctionScorer._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 808 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 808 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 808 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 808 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 808 "ms_deisotope/_c/scoring.pyx"
}

#line 808 "ms_deisotope/_c/scoring.pyx"


#line 808 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 808 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_14FunctionScorer_9_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 808 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_14FunctionScorer_8_evaluate[] = "FunctionScorer._evaluate(self, PeakSet peaklist, list observed, list expected) -> double";

#line 808 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_14FunctionScorer_9_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist = 0
#line 808 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_observed = 0
#line 808 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_expected = 0
#line 808 "ms_deisotope/_c/scoring.pyx"
;

#line 808 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 808 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 808 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 808 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 808 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 808 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate (wrapper)", 0);

#line 808 "ms_deisotope/_c/scoring.pyx"
  {

#line 808 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peaklist,&__pyx_n_s_observed,&__pyx_n_s_expected,0};

#line 808 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[3] = {0,0,0};

#line 808 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 808 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 808 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 808 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  3: 
#line 808 "ms_deisotope/_c/scoring.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 808 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 808 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 808 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 808 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 808 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 808 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 808 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 808 "ms_deisotope/_c/scoring.pyx"
      }

#line 808 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 808 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 808 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 808 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaklist)) != 0)) kw_args--;
        else 
#line 808 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 808 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 808 "ms_deisotope/_c/scoring.pyx"
        case  1:

#line 808 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_observed)) != 0)) kw_args--;

#line 808 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, 1); 
#line 808 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 808, __pyx_L3_error)

#line 808 "ms_deisotope/_c/scoring.pyx"
        }

#line 808 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 808 "ms_deisotope/_c/scoring.pyx"
        case  2:

#line 808 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_expected)) != 0)) kw_args--;

#line 808 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, 2); 
#line 808 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 808, __pyx_L3_error)

#line 808 "ms_deisotope/_c/scoring.pyx"
        }

#line 808 "ms_deisotope/_c/scoring.pyx"
      }

#line 808 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 808 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_evaluate") < 0)) __PYX_ERR(0, 808, __pyx_L3_error)

#line 808 "ms_deisotope/_c/scoring.pyx"
      }

#line 808 "ms_deisotope/_c/scoring.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 808 "ms_deisotope/_c/scoring.pyx"
      goto __pyx_L5_argtuple_error;

#line 808 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 808 "ms_deisotope/_c/scoring.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 808 "ms_deisotope/_c/scoring.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 808 "ms_deisotope/_c/scoring.pyx"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 808 "ms_deisotope/_c/scoring.pyx"
    }

#line 808 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_peaklist = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[0]);

#line 808 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_observed = ((PyObject*)values[1]);

#line 808 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_expected = ((PyObject*)values[2]);

#line 808 "ms_deisotope/_c/scoring.pyx"
  }

#line 808 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 808 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 808 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 808, __pyx_L3_error)

#line 808 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 808 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.FunctionScorer._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 808 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 808 "ms_deisotope/_c/scoring.pyx"
  return NULL;

#line 808 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 808 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peaklist), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peaklist", 0))) __PYX_ERR(0, 808, __pyx_L1_error)

#line 808 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_observed), (&PyList_Type), 1, "observed", 1))) __PYX_ERR(0, 808, __pyx_L1_error)

#line 808 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_expected), (&PyList_Type), 1, "expected", 1))) __PYX_ERR(0, 808, __pyx_L1_error)
  __pyx_r = 
#line 808 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_14FunctionScorer_8_evaluate(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer *)__pyx_v_self), __pyx_v_peaklist, __pyx_v_observed, __pyx_v_expected);

#line 808 "ms_deisotope/_c/scoring.pyx"


#line 808 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 808 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 808 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 808 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 808 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 808 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 808 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 808 "ms_deisotope/_c/scoring.pyx"
}

#line 808 "ms_deisotope/_c/scoring.pyx"


#line 808 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_14FunctionScorer_8_evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected) {

#line 808 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 808 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 808 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 808 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 808 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 808 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 808 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate", 0);

#line 808 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 808 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_12ms_deisotope_2_c_7scoring_14FunctionScorer__evaluate(__pyx_v_self, __pyx_v_peaklist, __pyx_v_observed, __pyx_v_expected, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 808, __pyx_L1_error)

#line 808 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 808 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 808 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 808 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 808 "ms_deisotope/_c/scoring.pyx"


#line 808 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 808 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 808 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 808 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.FunctionScorer._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 808 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 808 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 808 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 808 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 808 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 808 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":813
 * 
 * cdef class InterferenceDetection(object):
 *     def __init__(self, PeakSet peaklist):             # <<<<<<<<<<<<<<
 *         self.peaklist = peaklist
 * 
 */

#line 813 "ms_deisotope/_c/scoring.pyx"


#line 813 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 813 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_21InterferenceDetection_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 813 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_21InterferenceDetection_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist = 0
#line 813 "ms_deisotope/_c/scoring.pyx"
;

#line 813 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 813 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 813 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 813 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 813 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 813 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 813 "ms_deisotope/_c/scoring.pyx"
  {

#line 813 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peaklist,0};

#line 813 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[1] = {0};

#line 813 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 813 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 813 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 813 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  1: 
#line 813 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 813 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 813 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 813 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 813 "ms_deisotope/_c/scoring.pyx"
      }

#line 813 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 813 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 813 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 813 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaklist)) != 0)) kw_args--;
        else 
#line 813 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 813 "ms_deisotope/_c/scoring.pyx"
      }

#line 813 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 813 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 813, __pyx_L3_error)

#line 813 "ms_deisotope/_c/scoring.pyx"
      }

#line 813 "ms_deisotope/_c/scoring.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {

#line 813 "ms_deisotope/_c/scoring.pyx"
      goto __pyx_L5_argtuple_error;

#line 813 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 813 "ms_deisotope/_c/scoring.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 813 "ms_deisotope/_c/scoring.pyx"
    }

#line 813 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_peaklist = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[0]);

#line 813 "ms_deisotope/_c/scoring.pyx"
  }

#line 813 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 813 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); 
#line 813 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 813, __pyx_L3_error)

#line 813 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 813 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.InterferenceDetection.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 813 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 813 "ms_deisotope/_c/scoring.pyx"
  return -1;

#line 813 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 813 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peaklist), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peaklist", 0))) __PYX_ERR(0, 813, __pyx_L1_error)
  __pyx_r = 
#line 813 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_21InterferenceDetection___init__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *)__pyx_v_self), __pyx_v_peaklist);

#line 813 "ms_deisotope/_c/scoring.pyx"


#line 813 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 813 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 813 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 813 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = -1;

#line 813 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 813 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 813 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 813 "ms_deisotope/_c/scoring.pyx"
}

#line 813 "ms_deisotope/_c/scoring.pyx"


#line 813 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_21InterferenceDetection___init__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist) {

#line 813 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 813 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 813 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_deisotope/_c/scoring.pyx":814
 * cdef class InterferenceDetection(object):
 *     def __init__(self, PeakSet peaklist):
 *         self.peaklist = peaklist             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */

#line 814 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_peaklist));

#line 814 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_peaklist));

#line 814 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->peaklist);

#line 814 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->peaklist));

#line 814 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->peaklist = __pyx_v_peaklist;

  /* "ms_deisotope/_c/scoring.pyx":813
 * 
 * cdef class InterferenceDetection(object):
 *     def __init__(self, PeakSet peaklist):             # <<<<<<<<<<<<<<
 *         self.peaklist = peaklist
 * 
 */

#line 813 "ms_deisotope/_c/scoring.pyx"


#line 813 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 813 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 813 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 813 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 813 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":816
 *         self.peaklist = peaklist
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (None,), self.__getstate__()
 * 
 */

#line 816 "ms_deisotope/_c/scoring.pyx"


#line 816 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 816 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_21InterferenceDetection_3__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 816 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_21InterferenceDetection_2__reduce__[] = "InterferenceDetection.__reduce__(self)";

#line 816 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_21InterferenceDetection_3__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 816 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 816 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 816 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = 
#line 816 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_21InterferenceDetection_2__reduce__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *)__pyx_v_self));

#line 816 "ms_deisotope/_c/scoring.pyx"


#line 816 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 816 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 816 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 816 "ms_deisotope/_c/scoring.pyx"
}

#line 816 "ms_deisotope/_c/scoring.pyx"


#line 816 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_21InterferenceDetection_2__reduce__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_v_self) {

#line 816 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 816 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 816 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 816 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 816 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 816 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 816 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 816 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 816 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 816 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "ms_deisotope/_c/scoring.pyx":817
 * 
 *     def __reduce__(self):
 *         return self.__class__, (None,), self.__getstate__()             # <<<<<<<<<<<<<<
 * 
 *     def __getstate__(self):
 */

#line 817 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 817 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 817, __pyx_L1_error)

#line 817 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 817 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_getstate); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 817, __pyx_L1_error)

#line 817 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 817 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_4 = NULL;

#line 817 "ms_deisotope/_c/scoring.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {

#line 817 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 817 "ms_deisotope/_c/scoring.pyx"
    if (likely(__pyx_t_4)) {

#line 817 "ms_deisotope/_c/scoring.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 817 "ms_deisotope/_c/scoring.pyx"
      __Pyx_INCREF(__pyx_t_4);

#line 817 "ms_deisotope/_c/scoring.pyx"
      __Pyx_INCREF(function);

#line 817 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF_SET(__pyx_t_3, function);

#line 817 "ms_deisotope/_c/scoring.pyx"
    }

#line 817 "ms_deisotope/_c/scoring.pyx"
  }

#line 817 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 817 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 817 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 817, __pyx_L1_error)

#line 817 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 817 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 817 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 817, __pyx_L1_error)

#line 817 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 817 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 817 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);

#line 817 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_tuple__3);

#line 817 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_tuple__3);

#line 817 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_tuple__3);

#line 817 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 817 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);

#line 817 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 817 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = 0;

#line 817 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_3;

#line 817 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = 0;

#line 817 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":816
 *         self.peaklist = peaklist
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (None,), self.__getstate__()
 * 
 */

#line 816 "ms_deisotope/_c/scoring.pyx"


#line 816 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 816 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 816 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 816 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 816 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 816 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 816 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.InterferenceDetection.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 816 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 816 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 816 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 816 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 816 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 816 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":819
 *         return self.__class__, (None,), self.__getstate__()
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return (self.peaklist,)
 * 
 */

#line 819 "ms_deisotope/_c/scoring.pyx"


#line 819 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 819 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_21InterferenceDetection_5__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 819 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_21InterferenceDetection_4__getstate__[] = "InterferenceDetection.__getstate__(self)";

#line 819 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_21InterferenceDetection_5__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 819 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 819 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 819 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = 
#line 819 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_21InterferenceDetection_4__getstate__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *)__pyx_v_self));

#line 819 "ms_deisotope/_c/scoring.pyx"


#line 819 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 819 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 819 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 819 "ms_deisotope/_c/scoring.pyx"
}

#line 819 "ms_deisotope/_c/scoring.pyx"


#line 819 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_21InterferenceDetection_4__getstate__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_v_self) {

#line 819 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 819 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 819 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 819 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 819 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 819 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 819 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "ms_deisotope/_c/scoring.pyx":820
 * 
 *     def __getstate__(self):
 *         return (self.peaklist,)             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, state):
 */

#line 820 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 820 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 820, __pyx_L1_error)

#line 820 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 820 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->peaklist));

#line 820 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->peaklist));

#line 820 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->peaklist));

#line 820 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 820 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 820 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":819
 *         return self.__class__, (None,), self.__getstate__()
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return (self.peaklist,)
 * 
 */

#line 819 "ms_deisotope/_c/scoring.pyx"


#line 819 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 819 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 819 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 819 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.InterferenceDetection.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 819 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 819 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 819 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 819 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 819 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 819 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":822
 *         return (self.peaklist,)
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.peaklist, = state
 * 
 */

#line 822 "ms_deisotope/_c/scoring.pyx"


#line 822 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 822 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_21InterferenceDetection_7__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state); /*proto*/

#line 822 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_21InterferenceDetection_6__setstate__[] = "InterferenceDetection.__setstate__(self, state)";

#line 822 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_21InterferenceDetection_7__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state) {

#line 822 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 822 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 822 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  __pyx_r = 
#line 822 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_21InterferenceDetection_6__setstate__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *)__pyx_v_self), ((PyObject *)__pyx_v_state));

#line 822 "ms_deisotope/_c/scoring.pyx"


#line 822 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 822 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 822 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 822 "ms_deisotope/_c/scoring.pyx"
}

#line 822 "ms_deisotope/_c/scoring.pyx"


#line 822 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_21InterferenceDetection_6__setstate__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_v_self, PyObject *__pyx_v_state) {

#line 822 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 822 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 822 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 822 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 822 "ms_deisotope/_c/scoring.pyx"
  PyObject *(*__pyx_t_3)(PyObject *);

#line 822 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 822 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 822 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 822 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* "ms_deisotope/_c/scoring.pyx":823
 * 
 *     def __setstate__(self, state):
 *         self.peaklist, = state             # <<<<<<<<<<<<<<
 * 
 *     cdef double detect_interference(self, list experimental_peaks, double lower=-1, double upper=-1):
 */

#line 823 "ms_deisotope/_c/scoring.pyx"
  if ((likely(PyTuple_CheckExact(__pyx_v_state))) || (PyList_CheckExact(__pyx_v_state))) {

#line 823 "ms_deisotope/_c/scoring.pyx"
    PyObject* sequence = __pyx_v_state;

#line 823 "ms_deisotope/_c/scoring.pyx"
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);

#line 823 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(size != 1)) {

#line 823 "ms_deisotope/_c/scoring.pyx"
      if (size > 1) __Pyx_RaiseTooManyValuesError(1);

#line 823 "ms_deisotope/_c/scoring.pyx"
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);

#line 823 "ms_deisotope/_c/scoring.pyx"
      __PYX_ERR(0, 823, __pyx_L1_error)

#line 823 "ms_deisotope/_c/scoring.pyx"
    }

#line 823 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 823 "ms_deisotope/_c/scoring.pyx"
    if (likely(PyTuple_CheckExact(sequence))) {

#line 823 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 

#line 823 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 823 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 

#line 823 "ms_deisotope/_c/scoring.pyx"
    }

#line 823 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 823 "ms_deisotope/_c/scoring.pyx"
    #else

#line 823 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 823, __pyx_L1_error)

#line 823 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 823 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 823 "ms_deisotope/_c/scoring.pyx"
  } else {

#line 823 "ms_deisotope/_c/scoring.pyx"
    Py_ssize_t index = -1;

#line 823 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_2 = PyObject_GetIter(__pyx_v_state); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 823, __pyx_L1_error)

#line 823 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 823 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_3 = Py_TYPE(__pyx_t_2)->tp_iternext;
    index = 0; __pyx_t_1 = __pyx_t_3(__pyx_t_2); if (unlikely(!__pyx_t_1)) 
#line 823 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L3_unpacking_failed;

#line 823 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 823 "ms_deisotope/_c/scoring.pyx"
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_3(__pyx_t_2), 1) < 0) __PYX_ERR(0, 823, __pyx_L1_error)

#line 823 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_3 = NULL;

#line 823 "ms_deisotope/_c/scoring.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 823 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L4_unpacking_done;

#line 823 "ms_deisotope/_c/scoring.pyx"
    __pyx_L3_unpacking_failed:;

#line 823 "ms_deisotope/_c/scoring.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 823 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_3 = NULL;

#line 823 "ms_deisotope/_c/scoring.pyx"
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);

#line 823 "ms_deisotope/_c/scoring.pyx"
    __PYX_ERR(0, 823, __pyx_L1_error)

#line 823 "ms_deisotope/_c/scoring.pyx"
    __pyx_L4_unpacking_done:;

#line 823 "ms_deisotope/_c/scoring.pyx"
  }

#line 823 "ms_deisotope/_c/scoring.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet))))) __PYX_ERR(0, 823, __pyx_L1_error)

#line 823 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 823 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->peaklist);

#line 823 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->peaklist));

#line 823 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->peaklist = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_1);

#line 823 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/scoring.pyx":822
 *         return (self.peaklist,)
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.peaklist, = state
 * 
 */

#line 822 "ms_deisotope/_c/scoring.pyx"


#line 822 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 822 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 822 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 822 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 822 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 822 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 822 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.InterferenceDetection.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 822 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 822 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 822 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 822 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 822 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 822 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":825
 *         self.peaklist, = state
 * 
 *     cdef double detect_interference(self, list experimental_peaks, double lower=-1, double upper=-1):             # <<<<<<<<<<<<<<
 *         cdef:
 *             FittedPeak min_peak, max_peak
 */

#line 825 "ms_deisotope/_c/scoring.pyx"


#line 825 "ms_deisotope/_c/scoring.pyx"
static double __pyx_f_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_v_self, PyObject *__pyx_v_experimental_peaks, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference *__pyx_optional_args) {

#line 825 "ms_deisotope/_c/scoring.pyx"
  double __pyx_v_lower = ((double)-1.0);

#line 825 "ms_deisotope/_c/scoring.pyx"
  double __pyx_v_upper = ((double)-1.0);
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_min_peak = 0
#line 825 "ms_deisotope/_c/scoring.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_max_peak = 0
#line 825 "ms_deisotope/_c/scoring.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_region = 0
#line 825 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_included_intensity
#line 825 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_region_intensity
#line 825 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_score
#line 825 "ms_deisotope/_c/scoring.pyx"
;
  size_t __pyx_v_n
#line 825 "ms_deisotope/_c/scoring.pyx"
;

#line 825 "ms_deisotope/_c/scoring.pyx"
  double __pyx_r;

#line 825 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 825 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_1;

#line 825 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2;

#line 825 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 825 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 825 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 825 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 825 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("detect_interference", 0);

#line 825 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_optional_args) {

#line 825 "ms_deisotope/_c/scoring.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 825 "ms_deisotope/_c/scoring.pyx"
      __pyx_v_lower = __pyx_optional_args->lower;

#line 825 "ms_deisotope/_c/scoring.pyx"
      if (__pyx_optional_args->__pyx_n > 1) {

#line 825 "ms_deisotope/_c/scoring.pyx"
        __pyx_v_upper = __pyx_optional_args->upper;

#line 825 "ms_deisotope/_c/scoring.pyx"
      }

#line 825 "ms_deisotope/_c/scoring.pyx"
    }

#line 825 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":832
 *             size_t n
 * 
 *         n = PyList_GET_SIZE(experimental_peaks)             # <<<<<<<<<<<<<<
 *         if n == 0:
 *             return 0.0
 */

#line 832 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_n = PyList_GET_SIZE(__pyx_v_experimental_peaks);

  /* "ms_deisotope/_c/scoring.pyx":833
 * 
 *         n = PyList_GET_SIZE(experimental_peaks)
 *         if n == 0:             # <<<<<<<<<<<<<<
 *             return 0.0
 * 
 */

#line 833 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = ((__pyx_v_n == 0) != 0);

#line 833 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/scoring.pyx":834
 *         n = PyList_GET_SIZE(experimental_peaks)
 *         if n == 0:
 *             return 0.0             # <<<<<<<<<<<<<<
 * 
 *         min_peak = <FittedPeak>PyList_GET_ITEM(experimental_peaks, 0)
 */

#line 834 "ms_deisotope/_c/scoring.pyx"
    __pyx_r = 0.0;

#line 834 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/scoring.pyx":833
 * 
 *         n = PyList_GET_SIZE(experimental_peaks)
 *         if n == 0:             # <<<<<<<<<<<<<<
 *             return 0.0
 * 
 */

#line 833 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":836
 *             return 0.0
 * 
 *         min_peak = <FittedPeak>PyList_GET_ITEM(experimental_peaks, 0)             # <<<<<<<<<<<<<<
 *         max_peak = <FittedPeak>PyList_GET_ITEM(experimental_peaks, PyList_GET_SIZE(experimental_peaks) - 1)
 * 
 */

#line 836 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = PyList_GET_ITEM(__pyx_v_experimental_peaks, 0);

#line 836 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = ((PyObject *)__pyx_t_2);

#line 836 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_t_3);

#line 836 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_min_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_3);

#line 836 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/scoring.pyx":837
 * 
 *         min_peak = <FittedPeak>PyList_GET_ITEM(experimental_peaks, 0)
 *         max_peak = <FittedPeak>PyList_GET_ITEM(experimental_peaks, PyList_GET_SIZE(experimental_peaks) - 1)             # <<<<<<<<<<<<<<
 * 
 *         if lower < 0:
 */

#line 837 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = PyList_GET_ITEM(__pyx_v_experimental_peaks, (PyList_GET_SIZE(__pyx_v_experimental_peaks) - 1));

#line 837 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = ((PyObject *)__pyx_t_2);

#line 837 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_t_3);

#line 837 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_max_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_3);

#line 837 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/scoring.pyx":839
 *         max_peak = <FittedPeak>PyList_GET_ITEM(experimental_peaks, PyList_GET_SIZE(experimental_peaks) - 1)
 * 
 *         if lower < 0:             # <<<<<<<<<<<<<<
 *             lower = min_peak.mz - min_peak.full_width_at_half_max
 *         if upper < 0:
 */

#line 839 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = ((__pyx_v_lower < 0.0) != 0);

#line 839 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/scoring.pyx":840
 * 
 *         if lower < 0:
 *             lower = min_peak.mz - min_peak.full_width_at_half_max             # <<<<<<<<<<<<<<
 *         if upper < 0:
 *             upper = max_peak.mz + max_peak.full_width_at_half_max
 */

#line 840 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_lower = (__pyx_v_min_peak->__pyx_base.mz - __pyx_v_min_peak->full_width_at_half_max);

    /* "ms_deisotope/_c/scoring.pyx":839
 *         max_peak = <FittedPeak>PyList_GET_ITEM(experimental_peaks, PyList_GET_SIZE(experimental_peaks) - 1)
 * 
 *         if lower < 0:             # <<<<<<<<<<<<<<
 *             lower = min_peak.mz - min_peak.full_width_at_half_max
 *         if upper < 0:
 */

#line 839 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":841
 *         if lower < 0:
 *             lower = min_peak.mz - min_peak.full_width_at_half_max
 *         if upper < 0:             # <<<<<<<<<<<<<<
 *             upper = max_peak.mz + max_peak.full_width_at_half_max
 * 
 */

#line 841 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = ((__pyx_v_upper < 0.0) != 0);

#line 841 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/scoring.pyx":842
 *             lower = min_peak.mz - min_peak.full_width_at_half_max
 *         if upper < 0:
 *             upper = max_peak.mz + max_peak.full_width_at_half_max             # <<<<<<<<<<<<<<
 * 
 *         region = self.peaklist._between(
 */

#line 842 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_upper = (__pyx_v_max_peak->__pyx_base.mz + __pyx_v_max_peak->full_width_at_half_max);

    /* "ms_deisotope/_c/scoring.pyx":841
 *         if lower < 0:
 *             lower = min_peak.mz - min_peak.full_width_at_half_max
 *         if upper < 0:             # <<<<<<<<<<<<<<
 *             upper = max_peak.mz + max_peak.full_width_at_half_max
 * 
 */

#line 841 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":844
 *             upper = max_peak.mz + max_peak.full_width_at_half_max
 * 
 *         region = self.peaklist._between(             # <<<<<<<<<<<<<<
 *             lower,
 *             upper)
 */

#line 844 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_self->peaklist->__pyx_vtab)->_between(__pyx_v_self->peaklist, __pyx_v_lower, __pyx_v_upper)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 844, __pyx_L1_error)

#line 844 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 844 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_region = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_3);

#line 844 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/scoring.pyx":848
 *             upper)
 * 
 *         included_intensity = sum_intensity_fitted(experimental_peaks)             # <<<<<<<<<<<<<<
 *         region_intensity = sum_intensity_fitted(list(region))
 *         if region_intensity == 0:
 */

#line 848 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_included_intensity = __pyx_f_12ms_deisotope_2_c_7scoring_sum_intensity_fitted(__pyx_v_experimental_peaks);

  /* "ms_deisotope/_c/scoring.pyx":849
 * 
 *         included_intensity = sum_intensity_fitted(experimental_peaks)
 *         region_intensity = sum_intensity_fitted(list(region))             # <<<<<<<<<<<<<<
 *         if region_intensity == 0:
 *             return 0.0
 */

#line 849 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = PySequence_List(((PyObject *)__pyx_v_region)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 849, __pyx_L1_error)

#line 849 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 849 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_region_intensity = __pyx_f_12ms_deisotope_2_c_7scoring_sum_intensity_fitted(((PyObject*)__pyx_t_3));

#line 849 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ms_deisotope/_c/scoring.pyx":850
 *         included_intensity = sum_intensity_fitted(experimental_peaks)
 *         region_intensity = sum_intensity_fitted(list(region))
 *         if region_intensity == 0:             # <<<<<<<<<<<<<<
 *             return 0.0
 * 
 */

#line 850 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = ((__pyx_v_region_intensity == 0.0) != 0);

#line 850 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/scoring.pyx":851
 *         region_intensity = sum_intensity_fitted(list(region))
 *         if region_intensity == 0:
 *             return 0.0             # <<<<<<<<<<<<<<
 * 
 *         score = 1 - (included_intensity / region_intensity)
 */

#line 851 "ms_deisotope/_c/scoring.pyx"
    __pyx_r = 0.0;

#line 851 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/scoring.pyx":850
 *         included_intensity = sum_intensity_fitted(experimental_peaks)
 *         region_intensity = sum_intensity_fitted(list(region))
 *         if region_intensity == 0:             # <<<<<<<<<<<<<<
 *             return 0.0
 * 
 */

#line 850 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":853
 *             return 0.0
 * 
 *         score = 1 - (included_intensity / region_intensity)             # <<<<<<<<<<<<<<
 *         return score
 * 
 */

#line 853 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_v_region_intensity == 0)) {

#line 853 "ms_deisotope/_c/scoring.pyx"
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");

#line 853 "ms_deisotope/_c/scoring.pyx"
    __PYX_ERR(0, 853, __pyx_L1_error)

#line 853 "ms_deisotope/_c/scoring.pyx"
  }

#line 853 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_score = (1.0 - (__pyx_v_included_intensity / __pyx_v_region_intensity));

  /* "ms_deisotope/_c/scoring.pyx":854
 * 
 *         score = 1 - (included_intensity / region_intensity)
 *         return score             # <<<<<<<<<<<<<<
 * 
 *     def __call__(self, list experimental_peaks, lower=None, upper=None):
 */

#line 854 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_v_score;

#line 854 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":825
 *         self.peaklist, = state
 * 
 *     cdef double detect_interference(self, list experimental_peaks, double lower=-1, double upper=-1):             # <<<<<<<<<<<<<<
 *         cdef:
 *             FittedPeak min_peak, max_peak
 */

#line 825 "ms_deisotope/_c/scoring.pyx"


#line 825 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 825 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 825 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 825 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.InterferenceDetection.detect_interference", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 825 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 825 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 825 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_min_peak);

#line 825 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_max_peak);

#line 825 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_region);

#line 825 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 825 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 825 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":856
 *         return score
 * 
 *     def __call__(self, list experimental_peaks, lower=None, upper=None):             # <<<<<<<<<<<<<<
 *         if lower is None:
 *             lower = -1
 */

#line 856 "ms_deisotope/_c/scoring.pyx"


#line 856 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 856 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_21InterferenceDetection_9__call__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 856 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_21InterferenceDetection_9__call__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_experimental_peaks = 0
#line 856 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_lower = 0
#line 856 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_upper = 0
#line 856 "ms_deisotope/_c/scoring.pyx"
;

#line 856 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 856 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 856 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 856 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 856 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 856 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__call__ (wrapper)", 0);

#line 856 "ms_deisotope/_c/scoring.pyx"
  {

#line 856 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_experimental_peaks,&__pyx_n_s_lower,&__pyx_n_s_upper,0};

#line 856 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[3] = {0,0,0};

#line 856 "ms_deisotope/_c/scoring.pyx"
    values[1] = ((PyObject *)Py_None);

#line 856 "ms_deisotope/_c/scoring.pyx"
    values[2] = ((PyObject *)Py_None);

#line 856 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 856 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 856 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 856 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  3: 
#line 856 "ms_deisotope/_c/scoring.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 856 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 856 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 856 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 856 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 856 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 856 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 856 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 856 "ms_deisotope/_c/scoring.pyx"
      }

#line 856 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 856 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 856 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 856 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_experimental_peaks)) != 0)) kw_args--;
        else 
#line 856 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 856 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 856 "ms_deisotope/_c/scoring.pyx"
        case  1:

#line 856 "ms_deisotope/_c/scoring.pyx"
        if (kw_args > 0) {

#line 856 "ms_deisotope/_c/scoring.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lower);

#line 856 "ms_deisotope/_c/scoring.pyx"
          if (value) { values[1] = value; kw_args--; }

#line 856 "ms_deisotope/_c/scoring.pyx"
        }

#line 856 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 856 "ms_deisotope/_c/scoring.pyx"
        case  2:

#line 856 "ms_deisotope/_c/scoring.pyx"
        if (kw_args > 0) {

#line 856 "ms_deisotope/_c/scoring.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_upper);

#line 856 "ms_deisotope/_c/scoring.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 856 "ms_deisotope/_c/scoring.pyx"
        }

#line 856 "ms_deisotope/_c/scoring.pyx"
      }

#line 856 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 856 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__call__") < 0)) __PYX_ERR(0, 856, __pyx_L3_error)

#line 856 "ms_deisotope/_c/scoring.pyx"
      }

#line 856 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 856 "ms_deisotope/_c/scoring.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 856 "ms_deisotope/_c/scoring.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 856 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 856 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 856 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 856 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 856 "ms_deisotope/_c/scoring.pyx"
        break;
        default: 
#line 856 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 856 "ms_deisotope/_c/scoring.pyx"
      }

#line 856 "ms_deisotope/_c/scoring.pyx"
    }

#line 856 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_experimental_peaks = ((PyObject*)values[0]);

#line 856 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_lower = values[1];

#line 856 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_upper = values[2];

#line 856 "ms_deisotope/_c/scoring.pyx"
  }

#line 856 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 856 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__call__", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 856 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 856, __pyx_L3_error)

#line 856 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 856 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.InterferenceDetection.__call__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 856 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 856 "ms_deisotope/_c/scoring.pyx"
  return NULL;

#line 856 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 856 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_experimental_peaks), (&PyList_Type), 1, "experimental_peaks", 1))) __PYX_ERR(0, 856, __pyx_L1_error)
  __pyx_r = 
#line 856 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_21InterferenceDetection_8__call__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *)__pyx_v_self), __pyx_v_experimental_peaks, __pyx_v_lower, __pyx_v_upper);

#line 856 "ms_deisotope/_c/scoring.pyx"


#line 856 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 856 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 856 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 856 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 856 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 856 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 856 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 856 "ms_deisotope/_c/scoring.pyx"
}

#line 856 "ms_deisotope/_c/scoring.pyx"


#line 856 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_21InterferenceDetection_8__call__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_v_self, PyObject *__pyx_v_experimental_peaks, PyObject *__pyx_v_lower, PyObject *__pyx_v_upper) {

#line 856 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 856 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 856 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_1;

#line 856 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_2;

#line 856 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_3;

#line 856 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_4;

#line 856 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_5;

#line 856 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference __pyx_t_6;

#line 856 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_7 = NULL;

#line 856 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 856 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 856 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 856 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__call__", 0);

#line 856 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_v_lower);

#line 856 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_v_upper);

  /* "ms_deisotope/_c/scoring.pyx":857
 * 
 *     def __call__(self, list experimental_peaks, lower=None, upper=None):
 *         if lower is None:             # <<<<<<<<<<<<<<
 *             lower = -1
 *         if upper is None:
 */

#line 857 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = (__pyx_v_lower == Py_None);

#line 857 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 857 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/scoring.pyx":858
 *     def __call__(self, list experimental_peaks, lower=None, upper=None):
 *         if lower is None:
 *             lower = -1             # <<<<<<<<<<<<<<
 *         if upper is None:
 *             upper = -1
 */

#line 858 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_int_neg_1);

#line 858 "ms_deisotope/_c/scoring.pyx"
    __Pyx_DECREF_SET(__pyx_v_lower, __pyx_int_neg_1);

    /* "ms_deisotope/_c/scoring.pyx":857
 * 
 *     def __call__(self, list experimental_peaks, lower=None, upper=None):
 *         if lower is None:             # <<<<<<<<<<<<<<
 *             lower = -1
 *         if upper is None:
 */

#line 857 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":859
 *         if lower is None:
 *             lower = -1
 *         if upper is None:             # <<<<<<<<<<<<<<
 *             upper = -1
 *         return self.detect_interference(experimental_peaks, lower, upper)
 */

#line 859 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = (__pyx_v_upper == Py_None);

#line 859 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = (__pyx_t_2 != 0);

#line 859 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/scoring.pyx":860
 *             lower = -1
 *         if upper is None:
 *             upper = -1             # <<<<<<<<<<<<<<
 *         return self.detect_interference(experimental_peaks, lower, upper)
 * 
 */

#line 860 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_int_neg_1);

#line 860 "ms_deisotope/_c/scoring.pyx"
    __Pyx_DECREF_SET(__pyx_v_upper, __pyx_int_neg_1);

    /* "ms_deisotope/_c/scoring.pyx":859
 *         if lower is None:
 *             lower = -1
 *         if upper is None:             # <<<<<<<<<<<<<<
 *             upper = -1
 *         return self.detect_interference(experimental_peaks, lower, upper)
 */

#line 859 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":861
 *         if upper is None:
 *             upper = -1
 *         return self.detect_interference(experimental_peaks, lower, upper)             # <<<<<<<<<<<<<<
 * 
 * cdef class DistinctPatternFitter(IsotopicFitterBase):
 */

#line 861 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 861 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_v_lower); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 861, __pyx_L1_error)

#line 861 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_v_upper); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 861, __pyx_L1_error)

#line 861 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_6.__pyx_n = 2;

#line 861 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_6.lower = __pyx_t_3;

#line 861 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_6.upper = __pyx_t_4;

#line 861 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_5 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection *)__pyx_v_self->__pyx_vtab)->detect_interference(__pyx_v_self, __pyx_v_experimental_peaks, &__pyx_t_6); 

#line 861 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_7 = PyFloat_FromDouble(__pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 861, __pyx_L1_error)

#line 861 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_7);

#line 861 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_7;

#line 861 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_7 = 0;

#line 861 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":856
 *         return score
 * 
 *     def __call__(self, list experimental_peaks, lower=None, upper=None):             # <<<<<<<<<<<<<<
 *         if lower is None:
 *             lower = -1
 */

#line 856 "ms_deisotope/_c/scoring.pyx"


#line 856 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 856 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 856 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_7);

#line 856 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.InterferenceDetection.__call__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 856 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 856 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 856 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_v_lower);

#line 856 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_v_upper);

#line 856 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 856 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 856 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 856 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pxd":90
 * 
 * cdef class InterferenceDetection(object):
 *     cdef public PeakSet peaklist             # <<<<<<<<<<<<<<
 * 
 *     cdef double detect_interference(self, list experimental_peaks, double lower=*, double upper=*)
 */

#line 90 "ms_deisotope/_c/scoring.pxd"


#line 90 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 90 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_21InterferenceDetection_8peaklist_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 90 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_21InterferenceDetection_8peaklist_1__get__(PyObject *__pyx_v_self) {

#line 90 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = 0;

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 90 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_21InterferenceDetection_8peaklist___get__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *)__pyx_v_self));

#line 90 "ms_deisotope/_c/scoring.pxd"


#line 90 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 90 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 90 "ms_deisotope/_c/scoring.pxd"
}

#line 90 "ms_deisotope/_c/scoring.pxd"


#line 90 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_21InterferenceDetection_8peaklist___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_v_self) {

#line 90 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = NULL;

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->peaklist));

#line 90 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = ((PyObject *)__pyx_v_self->peaklist);

#line 90 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 90 "ms_deisotope/_c/scoring.pxd"


#line 90 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 90 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 90 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 90 "ms_deisotope/_c/scoring.pxd"
}

#line 90 "ms_deisotope/_c/scoring.pxd"


#line 90 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 90 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_21InterferenceDetection_8peaklist_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 90 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_21InterferenceDetection_8peaklist_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 90 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 90 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_21InterferenceDetection_8peaklist_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 90 "ms_deisotope/_c/scoring.pxd"


#line 90 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 90 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 90 "ms_deisotope/_c/scoring.pxd"
}

#line 90 "ms_deisotope/_c/scoring.pxd"


#line 90 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_21InterferenceDetection_8peaklist_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_v_self, PyObject *__pyx_v_value) {

#line 90 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 90 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 90 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 90 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 90 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 90 "ms_deisotope/_c/scoring.pxd"
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet))))) __PYX_ERR(1, 90, __pyx_L1_error)

#line 90 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = __pyx_v_value;

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(__pyx_t_1);

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GIVEREF(__pyx_t_1);

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_v_self->peaklist);

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->peaklist));

#line 90 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->peaklist = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_1);

#line 90 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = 0;

#line 90 "ms_deisotope/_c/scoring.pxd"


#line 90 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 90 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 90 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 90 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.InterferenceDetection.peaklist.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 90 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = -1;

#line 90 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 90 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 90 "ms_deisotope/_c/scoring.pxd"
}

#line 90 "ms_deisotope/_c/scoring.pxd"


#line 90 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 90 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_21InterferenceDetection_8peaklist_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 90 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_21InterferenceDetection_8peaklist_5__del__(PyObject *__pyx_v_self) {

#line 90 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 90 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_21InterferenceDetection_8peaklist_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *)__pyx_v_self));

#line 90 "ms_deisotope/_c/scoring.pxd"


#line 90 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 90 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 90 "ms_deisotope/_c/scoring.pxd"
}

#line 90 "ms_deisotope/_c/scoring.pxd"


#line 90 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_21InterferenceDetection_8peaklist_4__del__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_v_self) {

#line 90 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(Py_None);

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GIVEREF(Py_None);

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_v_self->peaklist);

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->peaklist));

#line 90 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->peaklist = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)Py_None);

#line 90 "ms_deisotope/_c/scoring.pxd"


#line 90 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 90 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 90 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 90 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 90 "ms_deisotope/_c/scoring.pxd"
}

/* "ms_deisotope/_c/scoring.pyx":865
 * cdef class DistinctPatternFitter(IsotopicFitterBase):
 * 
 *     def __init__(self, minimum_score=0.3, peak_count_scale=1.5, domain_scale=100.):             # <<<<<<<<<<<<<<
 *         self.select = MinimizeFitSelector(minimum_score)
 *         self.interference_detector = None
 */

#line 865 "ms_deisotope/_c/scoring.pyx"


#line 865 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 865 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 865 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_minimum_score = 0
#line 865 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_peak_count_scale = 0
#line 865 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_domain_scale = 0
#line 865 "ms_deisotope/_c/scoring.pyx"
;

#line 865 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 865 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 865 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 865 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 865 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 865 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 865 "ms_deisotope/_c/scoring.pyx"
  {

#line 865 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_minimum_score,&__pyx_n_s_peak_count_scale,&__pyx_n_s_domain_scale,0};

#line 865 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[3] = {0,0,0};

#line 865 "ms_deisotope/_c/scoring.pyx"
    values[0] = ((PyObject *)__pyx_float_0_3);

#line 865 "ms_deisotope/_c/scoring.pyx"
    values[1] = ((PyObject *)__pyx_float_1_5);

#line 865 "ms_deisotope/_c/scoring.pyx"
    values[2] = ((PyObject *)__pyx_float_100_);

#line 865 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 865 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 865 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 865 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  3: 
#line 865 "ms_deisotope/_c/scoring.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 865 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 865 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 865 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 865 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 865 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 865 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 865 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 865 "ms_deisotope/_c/scoring.pyx"
      }

#line 865 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 865 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 865 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 865 "ms_deisotope/_c/scoring.pyx"
        if (kw_args > 0) {

#line 865 "ms_deisotope/_c/scoring.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_minimum_score);

#line 865 "ms_deisotope/_c/scoring.pyx"
          if (value) { values[0] = value; kw_args--; }

#line 865 "ms_deisotope/_c/scoring.pyx"
        }

#line 865 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 865 "ms_deisotope/_c/scoring.pyx"
        case  1:

#line 865 "ms_deisotope/_c/scoring.pyx"
        if (kw_args > 0) {

#line 865 "ms_deisotope/_c/scoring.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak_count_scale);

#line 865 "ms_deisotope/_c/scoring.pyx"
          if (value) { values[1] = value; kw_args--; }

#line 865 "ms_deisotope/_c/scoring.pyx"
        }

#line 865 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 865 "ms_deisotope/_c/scoring.pyx"
        case  2:

#line 865 "ms_deisotope/_c/scoring.pyx"
        if (kw_args > 0) {

#line 865 "ms_deisotope/_c/scoring.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_domain_scale);

#line 865 "ms_deisotope/_c/scoring.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 865 "ms_deisotope/_c/scoring.pyx"
        }

#line 865 "ms_deisotope/_c/scoring.pyx"
      }

#line 865 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 865 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 865, __pyx_L3_error)

#line 865 "ms_deisotope/_c/scoring.pyx"
      }

#line 865 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 865 "ms_deisotope/_c/scoring.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 865 "ms_deisotope/_c/scoring.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 865 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 865 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 865 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 865 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 865 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  0: 
#line 865 "ms_deisotope/_c/scoring.pyx"
break;
        default: 
#line 865 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 865 "ms_deisotope/_c/scoring.pyx"
      }

#line 865 "ms_deisotope/_c/scoring.pyx"
    }

#line 865 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_minimum_score = values[0];

#line 865 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_peak_count_scale = values[1];

#line 865 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_domain_scale = values[2];

#line 865 "ms_deisotope/_c/scoring.pyx"
  }

#line 865 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 865 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 865 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 865, __pyx_L3_error)

#line 865 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 865 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.DistinctPatternFitter.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 865 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 865 "ms_deisotope/_c/scoring.pyx"
  return -1;

#line 865 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 865 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter___init__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *)__pyx_v_self), __pyx_v_minimum_score, __pyx_v_peak_count_scale, __pyx_v_domain_scale);

#line 865 "ms_deisotope/_c/scoring.pyx"


#line 865 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 865 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 865 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 865 "ms_deisotope/_c/scoring.pyx"
}

#line 865 "ms_deisotope/_c/scoring.pyx"


#line 865 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter___init__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self, PyObject *__pyx_v_minimum_score, PyObject *__pyx_v_peak_count_scale, PyObject *__pyx_v_domain_scale) {

#line 865 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 865 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 865 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 865 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_2;

#line 865 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 865 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 865 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 865 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_deisotope/_c/scoring.pyx":866
 * 
 *     def __init__(self, minimum_score=0.3, peak_count_scale=1.5, domain_scale=100.):
 *         self.select = MinimizeFitSelector(minimum_score)             # <<<<<<<<<<<<<<
 *         self.interference_detector = None
 *         self.g_test_scaled = ScaledGTestFitter()
 */

#line 866 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_7scoring_MinimizeFitSelector), __pyx_v_minimum_score); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 866, __pyx_L1_error)

#line 866 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 866 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 866 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.select);

#line 866 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx_base.select));

#line 866 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->__pyx_base.select = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_t_1);

#line 866 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/scoring.pyx":867
 *     def __init__(self, minimum_score=0.3, peak_count_scale=1.5, domain_scale=100.):
 *         self.select = MinimizeFitSelector(minimum_score)
 *         self.interference_detector = None             # <<<<<<<<<<<<<<
 *         self.g_test_scaled = ScaledGTestFitter()
 *         self.peak_count_scale = peak_count_scale
 */

#line 867 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(Py_None);

#line 867 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(Py_None);

#line 867 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->interference_detector);

#line 867 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->interference_detector));

#line 867 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->interference_detector = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *)Py_None);

  /* "ms_deisotope/_c/scoring.pyx":868
 *         self.select = MinimizeFitSelector(minimum_score)
 *         self.interference_detector = None
 *         self.g_test_scaled = ScaledGTestFitter()             # <<<<<<<<<<<<<<
 *         self.peak_count_scale = peak_count_scale
 *         self.domain_scale = domain_scale
 */

#line 868 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 868, __pyx_L1_error)

#line 868 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 868 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 868 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->g_test_scaled);

#line 868 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->g_test_scaled));

#line 868 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->g_test_scaled = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *)__pyx_t_1);

#line 868 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/scoring.pyx":869
 *         self.interference_detector = None
 *         self.g_test_scaled = ScaledGTestFitter()
 *         self.peak_count_scale = peak_count_scale             # <<<<<<<<<<<<<<
 *         self.domain_scale = domain_scale
 * 
 */

#line 869 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_v_peak_count_scale); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 869, __pyx_L1_error)

#line 869 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->peak_count_scale = __pyx_t_2;

  /* "ms_deisotope/_c/scoring.pyx":870
 *         self.g_test_scaled = ScaledGTestFitter()
 *         self.peak_count_scale = peak_count_scale
 *         self.domain_scale = domain_scale             # <<<<<<<<<<<<<<
 * 
 *     def __getstate__(self):
 */

#line 870 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_v_domain_scale); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 870, __pyx_L1_error)

#line 870 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->domain_scale = __pyx_t_2;

  /* "ms_deisotope/_c/scoring.pyx":865
 * cdef class DistinctPatternFitter(IsotopicFitterBase):
 * 
 *     def __init__(self, minimum_score=0.3, peak_count_scale=1.5, domain_scale=100.):             # <<<<<<<<<<<<<<
 *         self.select = MinimizeFitSelector(minimum_score)
 *         self.interference_detector = None
 */

#line 865 "ms_deisotope/_c/scoring.pyx"


#line 865 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 865 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 865 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 865 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 865 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 865 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.DistinctPatternFitter.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 865 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = -1;

#line 865 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 865 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 865 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 865 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":872
 *         self.domain_scale = domain_scale
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self.select, self.interference_detector, self.g_test_scaled, self.peak_count_scale, self.domain_scale
 * 
 */

#line 872 "ms_deisotope/_c/scoring.pyx"


#line 872 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 872 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_3__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 872 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_2__getstate__[] = "DistinctPatternFitter.__getstate__(self)";

#line 872 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_3__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 872 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 872 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 872 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = 
#line 872 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_2__getstate__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *)__pyx_v_self));

#line 872 "ms_deisotope/_c/scoring.pyx"


#line 872 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 872 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 872 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 872 "ms_deisotope/_c/scoring.pyx"
}

#line 872 "ms_deisotope/_c/scoring.pyx"


#line 872 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_2__getstate__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self) {

#line 872 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 872 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 872 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 872 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 872 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 872 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 872 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 872 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 872 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "ms_deisotope/_c/scoring.pyx":873
 * 
 *     def __getstate__(self):
 *         return self.select, self.interference_detector, self.g_test_scaled, self.peak_count_scale, self.domain_scale             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, state):
 */

#line 873 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 873 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->peak_count_scale); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 873, __pyx_L1_error)

#line 873 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 873 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->domain_scale); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 873, __pyx_L1_error)

#line 873 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 873 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = PyTuple_New(5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 873, __pyx_L1_error)

#line 873 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 873 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx_base.select));

#line 873 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx_base.select));

#line 873 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self->__pyx_base.select));

#line 873 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->interference_detector));

#line 873 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->interference_detector));

#line 873 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_self->interference_detector));

#line 873 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->g_test_scaled));

#line 873 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->g_test_scaled));

#line 873 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 2, ((PyObject *)__pyx_v_self->g_test_scaled));

#line 873 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 873 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_1);

#line 873 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 873 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_t_2);

#line 873 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 873 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = 0;

#line 873 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_3;

#line 873 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = 0;

#line 873 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":872
 *         self.domain_scale = domain_scale
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self.select, self.interference_detector, self.g_test_scaled, self.peak_count_scale, self.domain_scale
 * 
 */

#line 872 "ms_deisotope/_c/scoring.pyx"


#line 872 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 872 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 872 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 872 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 872 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 872 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.DistinctPatternFitter.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 872 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 872 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 872 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 872 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 872 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 872 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":875
 *         return self.select, self.interference_detector, self.g_test_scaled, self.peak_count_scale, self.domain_scale
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.select, self.interference_detector, self.g_test_scaled, self.peak_count_scale, self.domain_scale = state
 * 
 */

#line 875 "ms_deisotope/_c/scoring.pyx"


#line 875 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 875 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_5__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state); /*proto*/

#line 875 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_4__setstate__[] = "DistinctPatternFitter.__setstate__(self, state)";

#line 875 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_5__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state) {

#line 875 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 875 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 875 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  __pyx_r = 
#line 875 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_4__setstate__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *)__pyx_v_self), ((PyObject *)__pyx_v_state));

#line 875 "ms_deisotope/_c/scoring.pyx"


#line 875 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 875 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 875 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 875 "ms_deisotope/_c/scoring.pyx"
}

#line 875 "ms_deisotope/_c/scoring.pyx"


#line 875 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_4__setstate__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self, PyObject *__pyx_v_state) {

#line 875 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 875 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 875 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 875 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 875 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 875 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 875 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 875 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 875 "ms_deisotope/_c/scoring.pyx"
  PyObject *(*__pyx_t_7)(PyObject *);

#line 875 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_8;

#line 875 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_9;

#line 875 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 875 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 875 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 875 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* "ms_deisotope/_c/scoring.pyx":876
 * 
 *     def __setstate__(self, state):
 *         self.select, self.interference_detector, self.g_test_scaled, self.peak_count_scale, self.domain_scale = state             # <<<<<<<<<<<<<<
 * 
 *     cpdef IsotopicFitterBase _configure(self, DeconvoluterBase deconvoluter, dict kwargs):
 */

#line 876 "ms_deisotope/_c/scoring.pyx"
  if ((likely(PyTuple_CheckExact(__pyx_v_state))) || (PyList_CheckExact(__pyx_v_state))) {

#line 876 "ms_deisotope/_c/scoring.pyx"
    PyObject* sequence = __pyx_v_state;

#line 876 "ms_deisotope/_c/scoring.pyx"
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);

#line 876 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(size != 5)) {

#line 876 "ms_deisotope/_c/scoring.pyx"
      if (size > 5) __Pyx_RaiseTooManyValuesError(5);

#line 876 "ms_deisotope/_c/scoring.pyx"
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);

#line 876 "ms_deisotope/_c/scoring.pyx"
      __PYX_ERR(0, 876, __pyx_L1_error)

#line 876 "ms_deisotope/_c/scoring.pyx"
    }

#line 876 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 876 "ms_deisotope/_c/scoring.pyx"
    if (likely(PyTuple_CheckExact(sequence))) {

#line 876 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 

#line 876 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 

#line 876 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 2); 

#line 876 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 3); 

#line 876 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 4); 

#line 876 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 876 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 

#line 876 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 

#line 876 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_3 = PyList_GET_ITEM(sequence, 2); 

#line 876 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_4 = PyList_GET_ITEM(sequence, 3); 

#line 876 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_5 = PyList_GET_ITEM(sequence, 4); 

#line 876 "ms_deisotope/_c/scoring.pyx"
    }

#line 876 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 876 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_2);

#line 876 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_3);

#line 876 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_4);

#line 876 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_5);

#line 876 "ms_deisotope/_c/scoring.pyx"
    #else

#line 876 "ms_deisotope/_c/scoring.pyx"
    {

#line 876 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t i;

#line 876 "ms_deisotope/_c/scoring.pyx"
      PyObject** temps[5] = {&__pyx_t_1,&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5};

#line 876 "ms_deisotope/_c/scoring.pyx"
      for (i=0; i < 5; i++) {

#line 876 "ms_deisotope/_c/scoring.pyx"
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 876, __pyx_L1_error)

#line 876 "ms_deisotope/_c/scoring.pyx"
        __Pyx_GOTREF(item);

#line 876 "ms_deisotope/_c/scoring.pyx"
        *(temps[i]) = item;

#line 876 "ms_deisotope/_c/scoring.pyx"
      }

#line 876 "ms_deisotope/_c/scoring.pyx"
    }

#line 876 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 876 "ms_deisotope/_c/scoring.pyx"
  } else {

#line 876 "ms_deisotope/_c/scoring.pyx"
    Py_ssize_t index = -1;

#line 876 "ms_deisotope/_c/scoring.pyx"
    PyObject** temps[5] = {&__pyx_t_1,&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5};

#line 876 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_6 = PyObject_GetIter(__pyx_v_state); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 876, __pyx_L1_error)

#line 876 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 876 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_7 = Py_TYPE(__pyx_t_6)->tp_iternext;

#line 876 "ms_deisotope/_c/scoring.pyx"
    for (index=0; index < 5; index++) {
      PyObject* item = __pyx_t_7(__pyx_t_6); if (unlikely(!item)) 
#line 876 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L3_unpacking_failed;

#line 876 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(item);

#line 876 "ms_deisotope/_c/scoring.pyx"
      *(temps[index]) = item;

#line 876 "ms_deisotope/_c/scoring.pyx"
    }

#line 876 "ms_deisotope/_c/scoring.pyx"
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_6), 5) < 0) __PYX_ERR(0, 876, __pyx_L1_error)

#line 876 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_7 = NULL;

#line 876 "ms_deisotope/_c/scoring.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 876 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L4_unpacking_done;

#line 876 "ms_deisotope/_c/scoring.pyx"
    __pyx_L3_unpacking_failed:;

#line 876 "ms_deisotope/_c/scoring.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 876 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_7 = NULL;

#line 876 "ms_deisotope/_c/scoring.pyx"
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);

#line 876 "ms_deisotope/_c/scoring.pyx"
    __PYX_ERR(0, 876, __pyx_L1_error)

#line 876 "ms_deisotope/_c/scoring.pyx"
    __pyx_L4_unpacking_done:;

#line 876 "ms_deisotope/_c/scoring.pyx"
  }

#line 876 "ms_deisotope/_c/scoring.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase))))) __PYX_ERR(0, 876, __pyx_L1_error)

#line 876 "ms_deisotope/_c/scoring.pyx"
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection))))) __PYX_ERR(0, 876, __pyx_L1_error)

#line 876 "ms_deisotope/_c/scoring.pyx"
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter))))) __PYX_ERR(0, 876, __pyx_L1_error)

#line 876 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 876, __pyx_L1_error)

#line 876 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 876 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_9 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_9 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 876, __pyx_L1_error)

#line 876 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 876 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 876 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.select);

#line 876 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx_base.select));

#line 876 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->__pyx_base.select = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_t_1);

#line 876 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 876 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 876 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->interference_detector);

#line 876 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->interference_detector));

#line 876 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->interference_detector = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *)__pyx_t_2);

#line 876 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = 0;

#line 876 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 876 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->g_test_scaled);

#line 876 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->g_test_scaled));

#line 876 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->g_test_scaled = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *)__pyx_t_3);

#line 876 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = 0;

#line 876 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->peak_count_scale = __pyx_t_8;

#line 876 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->domain_scale = __pyx_t_9;

  /* "ms_deisotope/_c/scoring.pyx":875
 *         return self.select, self.interference_detector, self.g_test_scaled, self.peak_count_scale, self.domain_scale
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.select, self.interference_detector, self.g_test_scaled, self.peak_count_scale, self.domain_scale = state
 * 
 */

#line 875 "ms_deisotope/_c/scoring.pyx"


#line 875 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 875 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 875 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 875 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 875 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 875 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 875 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 875 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 875 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 875 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 875 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.DistinctPatternFitter.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 875 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 875 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 875 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 875 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 875 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 875 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":878
 *         self.select, self.interference_detector, self.g_test_scaled, self.peak_count_scale, self.domain_scale = state
 * 
 *     cpdef IsotopicFitterBase _configure(self, DeconvoluterBase deconvoluter, dict kwargs):             # <<<<<<<<<<<<<<
 *         self.interference_detector = InterferenceDetection(deconvoluter.peaklist)
 *         return self
 */

#line 878 "ms_deisotope/_c/scoring.pyx"


#line 878 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_7_configure(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 878 "ms_deisotope/_c/scoring.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_f_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter__configure(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_deconvoluter, CYTHON_UNUSED PyObject *__pyx_v_kwargs, int __pyx_skip_dispatch) {

#line 878 "ms_deisotope/_c/scoring.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_r = NULL;

#line 878 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 878 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 878 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 878 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 878 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 878 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_5;

#line 878 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 878 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 878 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 878 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 878 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_configure", 0);

#line 878 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 878 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 878 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 878 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 878 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 878 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 878 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 878 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 878 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 878 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_configure); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 878, __pyx_L1_error)

#line 878 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 878 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_7_configure)) {

#line 878 "ms_deisotope/_c/scoring.pyx"
        __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 878 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 878 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 878 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_5 = 0;

#line 878 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 878 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 878 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 878 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 878 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 878 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 878 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 878 "ms_deisotope/_c/scoring.pyx"
            __pyx_t_5 = 1;

#line 878 "ms_deisotope/_c/scoring.pyx"
          }

#line 878 "ms_deisotope/_c/scoring.pyx"
        }

#line 878 "ms_deisotope/_c/scoring.pyx"
        #if CYTHON_FAST_PYCALL

#line 878 "ms_deisotope/_c/scoring.pyx"
        if (PyFunction_Check(__pyx_t_3)) {

#line 878 "ms_deisotope/_c/scoring.pyx"
          PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_deconvoluter), __pyx_v_kwargs};

#line 878 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 878, __pyx_L1_error)

#line 878 "ms_deisotope/_c/scoring.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 878 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 878 "ms_deisotope/_c/scoring.pyx"
        } else

#line 878 "ms_deisotope/_c/scoring.pyx"
        #endif

#line 878 "ms_deisotope/_c/scoring.pyx"
        #if CYTHON_FAST_PYCCALL

#line 878 "ms_deisotope/_c/scoring.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {

#line 878 "ms_deisotope/_c/scoring.pyx"
          PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_deconvoluter), __pyx_v_kwargs};

#line 878 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 878, __pyx_L1_error)

#line 878 "ms_deisotope/_c/scoring.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 878 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 878 "ms_deisotope/_c/scoring.pyx"
        } else

#line 878 "ms_deisotope/_c/scoring.pyx"
        #endif

#line 878 "ms_deisotope/_c/scoring.pyx"
        {

#line 878 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 878, __pyx_L1_error)

#line 878 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_6);

#line 878 "ms_deisotope/_c/scoring.pyx"
          if (__pyx_t_4) {

#line 878 "ms_deisotope/_c/scoring.pyx"
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;

#line 878 "ms_deisotope/_c/scoring.pyx"
          }

#line 878 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(((PyObject *)__pyx_v_deconvoluter));

#line 878 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(((PyObject *)__pyx_v_deconvoluter));

#line 878 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_deconvoluter));

#line 878 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(__pyx_v_kwargs);

#line 878 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(__pyx_v_kwargs);

#line 878 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_kwargs);

#line 878 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 878, __pyx_L1_error)

#line 878 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 878 "ms_deisotope/_c/scoring.pyx"
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 878 "ms_deisotope/_c/scoring.pyx"
        }

#line 878 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 878 "ms_deisotope/_c/scoring.pyx"
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase))))) __PYX_ERR(0, 878, __pyx_L1_error)

#line 878 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_t_2);

#line 878 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_2 = 0;

#line 878 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 878 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 878 "ms_deisotope/_c/scoring.pyx"
      }

#line 878 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 878 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 878 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 878 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 878 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 878 "ms_deisotope/_c/scoring.pyx"
      }

#line 878 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 878 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 878 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 878 "ms_deisotope/_c/scoring.pyx"
    }

#line 878 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 878 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":879
 * 
 *     cpdef IsotopicFitterBase _configure(self, DeconvoluterBase deconvoluter, dict kwargs):
 *         self.interference_detector = InterferenceDetection(deconvoluter.peaklist)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

#line 879 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection), ((PyObject *)__pyx_v_deconvoluter->peaklist)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 879, __pyx_L1_error)

#line 879 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 879 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 879 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->interference_detector);

#line 879 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->interference_detector));

#line 879 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->interference_detector = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *)__pyx_t_1);

#line 879 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/scoring.pyx":880
 *     cpdef IsotopicFitterBase _configure(self, DeconvoluterBase deconvoluter, dict kwargs):
 *         self.interference_detector = InterferenceDetection(deconvoluter.peaklist)
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef double _evaluate(self, PeakSet peaklist, list experimental, list theoretical):
 */

#line 880 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 880 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 880 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self);

#line 880 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":878
 *         self.select, self.interference_detector, self.g_test_scaled, self.peak_count_scale, self.domain_scale = state
 * 
 *     cpdef IsotopicFitterBase _configure(self, DeconvoluterBase deconvoluter, dict kwargs):             # <<<<<<<<<<<<<<
 *         self.interference_detector = InterferenceDetection(deconvoluter.peaklist)
 *         return self
 */

#line 878 "ms_deisotope/_c/scoring.pyx"


#line 878 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 878 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 878 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 878 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 878 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 878 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 878 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 878 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.DistinctPatternFitter._configure", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 878 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 878 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 878 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 878 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 878 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 878 "ms_deisotope/_c/scoring.pyx"
}

#line 878 "ms_deisotope/_c/scoring.pyx"


#line 878 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 878 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_7_configure(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 878 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_6_configure[] = "DistinctPatternFitter._configure(self, DeconvoluterBase deconvoluter, dict kwargs) -> IsotopicFitterBase";

#line 878 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_7_configure(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_deconvoluter = 0
#line 878 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_kwargs = 0
#line 878 "ms_deisotope/_c/scoring.pyx"
;

#line 878 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 878 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 878 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 878 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 878 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 878 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_configure (wrapper)", 0);

#line 878 "ms_deisotope/_c/scoring.pyx"
  {

#line 878 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_deconvoluter,&__pyx_n_s_kwargs,0};

#line 878 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[2] = {0,0};

#line 878 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 878 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 878 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 878 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  2: 
#line 878 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 878 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 878 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 878 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 878 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 878 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 878 "ms_deisotope/_c/scoring.pyx"
      }

#line 878 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 878 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 878 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 878 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_deconvoluter)) != 0)) kw_args--;
        else 
#line 878 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 878 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 878 "ms_deisotope/_c/scoring.pyx"
        case  1:

#line 878 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;

#line 878 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_configure", 1, 2, 2, 1); 
#line 878 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 878, __pyx_L3_error)

#line 878 "ms_deisotope/_c/scoring.pyx"
        }

#line 878 "ms_deisotope/_c/scoring.pyx"
      }

#line 878 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 878 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_configure") < 0)) __PYX_ERR(0, 878, __pyx_L3_error)

#line 878 "ms_deisotope/_c/scoring.pyx"
      }

#line 878 "ms_deisotope/_c/scoring.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {

#line 878 "ms_deisotope/_c/scoring.pyx"
      goto __pyx_L5_argtuple_error;

#line 878 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 878 "ms_deisotope/_c/scoring.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 878 "ms_deisotope/_c/scoring.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 878 "ms_deisotope/_c/scoring.pyx"
    }

#line 878 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_deconvoluter = ((struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *)values[0]);

#line 878 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_kwargs = ((PyObject*)values[1]);

#line 878 "ms_deisotope/_c/scoring.pyx"
  }

#line 878 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 878 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_configure", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 878 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 878, __pyx_L3_error)

#line 878 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 878 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.DistinctPatternFitter._configure", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 878 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 878 "ms_deisotope/_c/scoring.pyx"
  return NULL;

#line 878 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 878 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_deconvoluter), __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase, 1, "deconvoluter", 0))) __PYX_ERR(0, 878, __pyx_L1_error)

#line 878 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_kwargs), (&PyDict_Type), 1, "kwargs", 1))) __PYX_ERR(0, 878, __pyx_L1_error)
  __pyx_r = 
#line 878 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_6_configure(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *)__pyx_v_self), __pyx_v_deconvoluter, __pyx_v_kwargs);

#line 878 "ms_deisotope/_c/scoring.pyx"


#line 878 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 878 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 878 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 878 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 878 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 878 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 878 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 878 "ms_deisotope/_c/scoring.pyx"
}

#line 878 "ms_deisotope/_c/scoring.pyx"


#line 878 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_6_configure(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_v_deconvoluter, PyObject *__pyx_v_kwargs) {

#line 878 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 878 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 878 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 878 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 878 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 878 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 878 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_configure", 0);

#line 878 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 878 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter__configure(__pyx_v_self, __pyx_v_deconvoluter, __pyx_v_kwargs, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 878, __pyx_L1_error)

#line 878 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 878 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 878 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 878 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 878 "ms_deisotope/_c/scoring.pyx"


#line 878 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 878 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 878 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 878 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.DistinctPatternFitter._configure", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 878 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 878 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 878 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 878 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 878 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 878 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":882
 *         return self
 * 
 *     cpdef double _evaluate(self, PeakSet peaklist, list experimental, list theoretical):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double score
 */

#line 882 "ms_deisotope/_c/scoring.pyx"


#line 882 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_9_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 882 "ms_deisotope/_c/scoring.pyx"
static double __pyx_f_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter__evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_experimental, PyObject *__pyx_v_theoretical, int __pyx_skip_dispatch) {
  double __pyx_v_score
#line 882 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_npeaks
#line 882 "ms_deisotope/_c/scoring.pyx"
;

#line 882 "ms_deisotope/_c/scoring.pyx"
  double __pyx_r;

#line 882 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 882 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 882 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 882 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 882 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 882 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_5;

#line 882 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 882 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_7;

#line 882 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_8;

#line 882 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_9;

#line 882 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_10;

#line 882 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 882 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 882 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 882 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate", 0);

#line 882 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 882 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 882 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 882 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 882 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 882 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 882 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 882 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 882 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 882 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_evaluate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 882, __pyx_L1_error)

#line 882 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 882 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_9_evaluate)) {

#line 882 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 882 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 882 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_5 = 0;

#line 882 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 882 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 882 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 882 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 882 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 882 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 882 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 882 "ms_deisotope/_c/scoring.pyx"
            __pyx_t_5 = 1;

#line 882 "ms_deisotope/_c/scoring.pyx"
          }

#line 882 "ms_deisotope/_c/scoring.pyx"
        }

#line 882 "ms_deisotope/_c/scoring.pyx"
        #if CYTHON_FAST_PYCALL

#line 882 "ms_deisotope/_c/scoring.pyx"
        if (PyFunction_Check(__pyx_t_3)) {

#line 882 "ms_deisotope/_c/scoring.pyx"
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_peaklist), __pyx_v_experimental, __pyx_v_theoretical};

#line 882 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 882, __pyx_L1_error)

#line 882 "ms_deisotope/_c/scoring.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 882 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 882 "ms_deisotope/_c/scoring.pyx"
        } else

#line 882 "ms_deisotope/_c/scoring.pyx"
        #endif

#line 882 "ms_deisotope/_c/scoring.pyx"
        #if CYTHON_FAST_PYCCALL

#line 882 "ms_deisotope/_c/scoring.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {

#line 882 "ms_deisotope/_c/scoring.pyx"
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_peaklist), __pyx_v_experimental, __pyx_v_theoretical};

#line 882 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 882, __pyx_L1_error)

#line 882 "ms_deisotope/_c/scoring.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 882 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 882 "ms_deisotope/_c/scoring.pyx"
        } else

#line 882 "ms_deisotope/_c/scoring.pyx"
        #endif

#line 882 "ms_deisotope/_c/scoring.pyx"
        {

#line 882 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_6 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 882, __pyx_L1_error)

#line 882 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_6);

#line 882 "ms_deisotope/_c/scoring.pyx"
          if (__pyx_t_4) {

#line 882 "ms_deisotope/_c/scoring.pyx"
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;

#line 882 "ms_deisotope/_c/scoring.pyx"
          }

#line 882 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(((PyObject *)__pyx_v_peaklist));

#line 882 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(((PyObject *)__pyx_v_peaklist));

#line 882 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_peaklist));

#line 882 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(__pyx_v_experimental);

#line 882 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(__pyx_v_experimental);

#line 882 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_experimental);

#line 882 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(__pyx_v_theoretical);

#line 882 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(__pyx_v_theoretical);

#line 882 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, __pyx_v_theoretical);

#line 882 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 882, __pyx_L1_error)

#line 882 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 882 "ms_deisotope/_c/scoring.pyx"
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 882 "ms_deisotope/_c/scoring.pyx"
        }

#line 882 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 882 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_7 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_7 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 882, __pyx_L1_error)

#line 882 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 882 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = __pyx_t_7;

#line 882 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 882 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 882 "ms_deisotope/_c/scoring.pyx"
      }

#line 882 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 882 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 882 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 882 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 882 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 882 "ms_deisotope/_c/scoring.pyx"
      }

#line 882 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 882 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 882 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 882 "ms_deisotope/_c/scoring.pyx"
    }

#line 882 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 882 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":887
 *             double npeaks
 * 
 *         npeaks = PyList_GET_SIZE(experimental)             # <<<<<<<<<<<<<<
 * 
 *         if self.interference_detector is None:
 */

#line 887 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_npeaks = PyList_GET_SIZE(__pyx_v_experimental);

  /* "ms_deisotope/_c/scoring.pyx":889
 *         npeaks = PyList_GET_SIZE(experimental)
 * 
 *         if self.interference_detector is None:             # <<<<<<<<<<<<<<
 *             self.interference_detector = InterferenceDetection(peaklist)
 * 
 */

#line 889 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_8 = (((PyObject *)__pyx_v_self->interference_detector) == Py_None);

#line 889 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_9 = (__pyx_t_8 != 0);

#line 889 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_t_9) {

    /* "ms_deisotope/_c/scoring.pyx":890
 * 
 *         if self.interference_detector is None:
 *             self.interference_detector = InterferenceDetection(peaklist)             # <<<<<<<<<<<<<<
 * 
 *         score = self.g_test_scaled._evaluate(peaklist, experimental, theoretical)
 */

#line 890 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection), ((PyObject *)__pyx_v_peaklist)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 890, __pyx_L1_error)

#line 890 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 890 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GIVEREF(__pyx_t_1);

#line 890 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_v_self->interference_detector);

#line 890 "ms_deisotope/_c/scoring.pyx"
    __Pyx_DECREF(((PyObject *)__pyx_v_self->interference_detector));

#line 890 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_self->interference_detector = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *)__pyx_t_1);

#line 890 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/scoring.pyx":889
 *         npeaks = PyList_GET_SIZE(experimental)
 * 
 *         if self.interference_detector is None:             # <<<<<<<<<<<<<<
 *             self.interference_detector = InterferenceDetection(peaklist)
 * 
 */

#line 889 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":892
 *             self.interference_detector = InterferenceDetection(peaklist)
 * 
 *         score = self.g_test_scaled._evaluate(peaklist, experimental, theoretical)             # <<<<<<<<<<<<<<
 *         score *= abs((self.interference_detector.detect_interference(experimental) + 0.00001) / (
 *             npeaks * self.peak_count_scale)) * self.domain_scale
 */

#line 892 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_score = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *)__pyx_v_self->g_test_scaled->__pyx_base.__pyx_vtab)->__pyx_base._evaluate(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self->g_test_scaled), __pyx_v_peaklist, __pyx_v_experimental, __pyx_v_theoretical, 0);

  /* "ms_deisotope/_c/scoring.pyx":893
 * 
 *         score = self.g_test_scaled._evaluate(peaklist, experimental, theoretical)
 *         score *= abs((self.interference_detector.detect_interference(experimental) + 0.00001) / (             # <<<<<<<<<<<<<<
 *             npeaks * self.peak_count_scale)) * self.domain_scale
 *         return score
 */

#line 893 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_7 = (((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection *)__pyx_v_self->interference_detector->__pyx_vtab)->detect_interference(__pyx_v_self->interference_detector, __pyx_v_experimental, NULL) + 0.00001);

  /* "ms_deisotope/_c/scoring.pyx":894
 *         score = self.g_test_scaled._evaluate(peaklist, experimental, theoretical)
 *         score *= abs((self.interference_detector.detect_interference(experimental) + 0.00001) / (
 *             npeaks * self.peak_count_scale)) * self.domain_scale             # <<<<<<<<<<<<<<
 *         return score
 * 
 */

#line 894 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_10 = (__pyx_v_npeaks * __pyx_v_self->peak_count_scale);

  /* "ms_deisotope/_c/scoring.pyx":893
 * 
 *         score = self.g_test_scaled._evaluate(peaklist, experimental, theoretical)
 *         score *= abs((self.interference_detector.detect_interference(experimental) + 0.00001) / (             # <<<<<<<<<<<<<<
 *             npeaks * self.peak_count_scale)) * self.domain_scale
 *         return score
 */

#line 893 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_t_10 == 0)) {

#line 893 "ms_deisotope/_c/scoring.pyx"
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");

#line 893 "ms_deisotope/_c/scoring.pyx"
    __PYX_ERR(0, 893, __pyx_L1_error)

#line 893 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":894
 *         score = self.g_test_scaled._evaluate(peaklist, experimental, theoretical)
 *         score *= abs((self.interference_detector.detect_interference(experimental) + 0.00001) / (
 *             npeaks * self.peak_count_scale)) * self.domain_scale             # <<<<<<<<<<<<<<
 *         return score
 * 
 */

#line 894 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_score = (__pyx_v_score * (fabs((__pyx_t_7 / __pyx_t_10)) * __pyx_v_self->domain_scale));

  /* "ms_deisotope/_c/scoring.pyx":895
 *         score *= abs((self.interference_detector.detect_interference(experimental) + 0.00001) / (
 *             npeaks * self.peak_count_scale)) * self.domain_scale
 *         return score             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 895 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_v_score;

#line 895 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":882
 *         return self
 * 
 *     cpdef double _evaluate(self, PeakSet peaklist, list experimental, list theoretical):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double score
 */

#line 882 "ms_deisotope/_c/scoring.pyx"


#line 882 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 882 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 882 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 882 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 882 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 882 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 882 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 882 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.DistinctPatternFitter._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 882 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 882 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 882 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 882 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 882 "ms_deisotope/_c/scoring.pyx"
}

#line 882 "ms_deisotope/_c/scoring.pyx"


#line 882 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 882 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_9_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 882 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_8_evaluate[] = "DistinctPatternFitter._evaluate(self, PeakSet peaklist, list experimental, list theoretical) -> double";

#line 882 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_9_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist = 0
#line 882 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_experimental = 0
#line 882 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_theoretical = 0
#line 882 "ms_deisotope/_c/scoring.pyx"
;

#line 882 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 882 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 882 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 882 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 882 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 882 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate (wrapper)", 0);

#line 882 "ms_deisotope/_c/scoring.pyx"
  {

#line 882 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peaklist,&__pyx_n_s_experimental,&__pyx_n_s_theoretical,0};

#line 882 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[3] = {0,0,0};

#line 882 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 882 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 882 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 882 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  3: 
#line 882 "ms_deisotope/_c/scoring.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 882 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 882 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 882 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 882 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 882 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 882 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 882 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 882 "ms_deisotope/_c/scoring.pyx"
      }

#line 882 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 882 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 882 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 882 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaklist)) != 0)) kw_args--;
        else 
#line 882 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 882 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 882 "ms_deisotope/_c/scoring.pyx"
        case  1:

#line 882 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_experimental)) != 0)) kw_args--;

#line 882 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, 1); 
#line 882 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 882, __pyx_L3_error)

#line 882 "ms_deisotope/_c/scoring.pyx"
        }

#line 882 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 882 "ms_deisotope/_c/scoring.pyx"
        case  2:

#line 882 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_theoretical)) != 0)) kw_args--;

#line 882 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, 2); 
#line 882 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 882, __pyx_L3_error)

#line 882 "ms_deisotope/_c/scoring.pyx"
        }

#line 882 "ms_deisotope/_c/scoring.pyx"
      }

#line 882 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 882 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_evaluate") < 0)) __PYX_ERR(0, 882, __pyx_L3_error)

#line 882 "ms_deisotope/_c/scoring.pyx"
      }

#line 882 "ms_deisotope/_c/scoring.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 882 "ms_deisotope/_c/scoring.pyx"
      goto __pyx_L5_argtuple_error;

#line 882 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 882 "ms_deisotope/_c/scoring.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 882 "ms_deisotope/_c/scoring.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 882 "ms_deisotope/_c/scoring.pyx"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 882 "ms_deisotope/_c/scoring.pyx"
    }

#line 882 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_peaklist = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[0]);

#line 882 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_experimental = ((PyObject*)values[1]);

#line 882 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_theoretical = ((PyObject*)values[2]);

#line 882 "ms_deisotope/_c/scoring.pyx"
  }

#line 882 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 882 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 882 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 882, __pyx_L3_error)

#line 882 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 882 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.DistinctPatternFitter._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 882 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 882 "ms_deisotope/_c/scoring.pyx"
  return NULL;

#line 882 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 882 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peaklist), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peaklist", 0))) __PYX_ERR(0, 882, __pyx_L1_error)

#line 882 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_experimental), (&PyList_Type), 1, "experimental", 1))) __PYX_ERR(0, 882, __pyx_L1_error)

#line 882 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_theoretical), (&PyList_Type), 1, "theoretical", 1))) __PYX_ERR(0, 882, __pyx_L1_error)
  __pyx_r = 
#line 882 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_8_evaluate(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *)__pyx_v_self), __pyx_v_peaklist, __pyx_v_experimental, __pyx_v_theoretical);

#line 882 "ms_deisotope/_c/scoring.pyx"


#line 882 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 882 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 882 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 882 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 882 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 882 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 882 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 882 "ms_deisotope/_c/scoring.pyx"
}

#line 882 "ms_deisotope/_c/scoring.pyx"


#line 882 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_8_evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_experimental, PyObject *__pyx_v_theoretical) {

#line 882 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 882 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 882 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 882 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 882 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 882 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 882 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate", 0);

#line 882 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 882 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter__evaluate(__pyx_v_self, __pyx_v_peaklist, __pyx_v_experimental, __pyx_v_theoretical, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 882, __pyx_L1_error)

#line 882 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 882 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 882 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 882 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 882 "ms_deisotope/_c/scoring.pyx"


#line 882 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 882 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 882 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 882 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.DistinctPatternFitter._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 882 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 882 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 882 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 882 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 882 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 882 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pxd":97
 * cdef class DistinctPatternFitter(IsotopicFitterBase):
 *     cdef:
 *         public InterferenceDetection interference_detector             # <<<<<<<<<<<<<<
 *         public ScaledGTestFitter g_test_scaled
 *         public double peak_count_scale
 */

#line 97 "ms_deisotope/_c/scoring.pxd"


#line 97 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 97 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_21interference_detector_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 97 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_21interference_detector_1__get__(PyObject *__pyx_v_self) {

#line 97 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = 0;

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 97 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_21interference_detector___get__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *)__pyx_v_self));

#line 97 "ms_deisotope/_c/scoring.pxd"


#line 97 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 97 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 97 "ms_deisotope/_c/scoring.pxd"
}

#line 97 "ms_deisotope/_c/scoring.pxd"


#line 97 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_21interference_detector___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self) {

#line 97 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = NULL;

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->interference_detector));

#line 97 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = ((PyObject *)__pyx_v_self->interference_detector);

#line 97 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 97 "ms_deisotope/_c/scoring.pxd"


#line 97 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 97 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 97 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 97 "ms_deisotope/_c/scoring.pxd"
}

#line 97 "ms_deisotope/_c/scoring.pxd"


#line 97 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 97 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_21interference_detector_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 97 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_21interference_detector_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 97 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 97 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_21interference_detector_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 97 "ms_deisotope/_c/scoring.pxd"


#line 97 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 97 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 97 "ms_deisotope/_c/scoring.pxd"
}

#line 97 "ms_deisotope/_c/scoring.pxd"


#line 97 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_21interference_detector_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self, PyObject *__pyx_v_value) {

#line 97 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 97 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 97 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 97 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 97 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 97 "ms_deisotope/_c/scoring.pxd"
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection))))) __PYX_ERR(1, 97, __pyx_L1_error)

#line 97 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = __pyx_v_value;

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(__pyx_t_1);

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GIVEREF(__pyx_t_1);

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_v_self->interference_detector);

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->interference_detector));

#line 97 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->interference_detector = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *)__pyx_t_1);

#line 97 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = 0;

#line 97 "ms_deisotope/_c/scoring.pxd"


#line 97 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 97 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 97 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 97 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.DistinctPatternFitter.interference_detector.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 97 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = -1;

#line 97 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 97 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 97 "ms_deisotope/_c/scoring.pxd"
}

#line 97 "ms_deisotope/_c/scoring.pxd"


#line 97 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 97 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_21interference_detector_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 97 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_21interference_detector_5__del__(PyObject *__pyx_v_self) {

#line 97 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 97 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_21interference_detector_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *)__pyx_v_self));

#line 97 "ms_deisotope/_c/scoring.pxd"


#line 97 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 97 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 97 "ms_deisotope/_c/scoring.pxd"
}

#line 97 "ms_deisotope/_c/scoring.pxd"


#line 97 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_21interference_detector_4__del__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self) {

#line 97 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(Py_None);

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GIVEREF(Py_None);

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_v_self->interference_detector);

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->interference_detector));

#line 97 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->interference_detector = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *)Py_None);

#line 97 "ms_deisotope/_c/scoring.pxd"


#line 97 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 97 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 97 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 97 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 97 "ms_deisotope/_c/scoring.pxd"
}

/* "ms_deisotope/_c/scoring.pxd":98
 *     cdef:
 *         public InterferenceDetection interference_detector
 *         public ScaledGTestFitter g_test_scaled             # <<<<<<<<<<<<<<
 *         public double peak_count_scale
 *         public double domain_scale
 */

#line 98 "ms_deisotope/_c/scoring.pxd"


#line 98 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 98 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_13g_test_scaled_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 98 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_13g_test_scaled_1__get__(PyObject *__pyx_v_self) {

#line 98 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = 0;

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 98 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_13g_test_scaled___get__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *)__pyx_v_self));

#line 98 "ms_deisotope/_c/scoring.pxd"


#line 98 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 98 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 98 "ms_deisotope/_c/scoring.pxd"
}

#line 98 "ms_deisotope/_c/scoring.pxd"


#line 98 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_13g_test_scaled___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self) {

#line 98 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = NULL;

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->g_test_scaled));

#line 98 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = ((PyObject *)__pyx_v_self->g_test_scaled);

#line 98 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 98 "ms_deisotope/_c/scoring.pxd"


#line 98 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 98 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 98 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 98 "ms_deisotope/_c/scoring.pxd"
}

#line 98 "ms_deisotope/_c/scoring.pxd"


#line 98 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 98 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_13g_test_scaled_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 98 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_13g_test_scaled_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 98 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 98 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_13g_test_scaled_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 98 "ms_deisotope/_c/scoring.pxd"


#line 98 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 98 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 98 "ms_deisotope/_c/scoring.pxd"
}

#line 98 "ms_deisotope/_c/scoring.pxd"


#line 98 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_13g_test_scaled_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self, PyObject *__pyx_v_value) {

#line 98 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 98 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 98 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 98 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 98 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 98 "ms_deisotope/_c/scoring.pxd"
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter))))) __PYX_ERR(1, 98, __pyx_L1_error)

#line 98 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = __pyx_v_value;

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(__pyx_t_1);

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GIVEREF(__pyx_t_1);

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_v_self->g_test_scaled);

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->g_test_scaled));

#line 98 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->g_test_scaled = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *)__pyx_t_1);

#line 98 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = 0;

#line 98 "ms_deisotope/_c/scoring.pxd"


#line 98 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 98 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 98 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 98 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.DistinctPatternFitter.g_test_scaled.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 98 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = -1;

#line 98 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 98 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 98 "ms_deisotope/_c/scoring.pxd"
}

#line 98 "ms_deisotope/_c/scoring.pxd"


#line 98 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 98 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_13g_test_scaled_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 98 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_13g_test_scaled_5__del__(PyObject *__pyx_v_self) {

#line 98 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 98 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_13g_test_scaled_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *)__pyx_v_self));

#line 98 "ms_deisotope/_c/scoring.pxd"


#line 98 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 98 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 98 "ms_deisotope/_c/scoring.pxd"
}

#line 98 "ms_deisotope/_c/scoring.pxd"


#line 98 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_13g_test_scaled_4__del__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self) {

#line 98 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(Py_None);

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GIVEREF(Py_None);

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_v_self->g_test_scaled);

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->g_test_scaled));

#line 98 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->g_test_scaled = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *)Py_None);

#line 98 "ms_deisotope/_c/scoring.pxd"


#line 98 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 98 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 98 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 98 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 98 "ms_deisotope/_c/scoring.pxd"
}

/* "ms_deisotope/_c/scoring.pxd":99
 *         public InterferenceDetection interference_detector
 *         public ScaledGTestFitter g_test_scaled
 *         public double peak_count_scale             # <<<<<<<<<<<<<<
 *         public double domain_scale
 * 
 */

#line 99 "ms_deisotope/_c/scoring.pxd"


#line 99 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 99 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_16peak_count_scale_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 99 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_16peak_count_scale_1__get__(PyObject *__pyx_v_self) {

#line 99 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = 0;

#line 99 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 99 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 99 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_16peak_count_scale___get__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *)__pyx_v_self));

#line 99 "ms_deisotope/_c/scoring.pxd"


#line 99 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 99 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 99 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 99 "ms_deisotope/_c/scoring.pxd"
}

#line 99 "ms_deisotope/_c/scoring.pxd"


#line 99 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_16peak_count_scale___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self) {

#line 99 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = NULL;

#line 99 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 99 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 99 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 99 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 99 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 99 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 99 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 99 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->peak_count_scale); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 99, __pyx_L1_error)

#line 99 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 99 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = __pyx_t_1;

#line 99 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = 0;

#line 99 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 99 "ms_deisotope/_c/scoring.pxd"


#line 99 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 99 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 99 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 99 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.DistinctPatternFitter.peak_count_scale.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 99 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = NULL;

#line 99 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 99 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 99 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 99 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 99 "ms_deisotope/_c/scoring.pxd"
}

#line 99 "ms_deisotope/_c/scoring.pxd"


#line 99 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 99 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_16peak_count_scale_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 99 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_16peak_count_scale_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 99 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 99 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 99 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 99 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_16peak_count_scale_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 99 "ms_deisotope/_c/scoring.pxd"


#line 99 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 99 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 99 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 99 "ms_deisotope/_c/scoring.pxd"
}

#line 99 "ms_deisotope/_c/scoring.pxd"


#line 99 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_16peak_count_scale_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self, PyObject *__pyx_v_value) {

#line 99 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 99 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 99 "ms_deisotope/_c/scoring.pxd"
  double __pyx_t_1;

#line 99 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 99 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 99 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 99 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 99 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 99, __pyx_L1_error)

#line 99 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->peak_count_scale = __pyx_t_1;

#line 99 "ms_deisotope/_c/scoring.pxd"


#line 99 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 99 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 99 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 99 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 99 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.DistinctPatternFitter.peak_count_scale.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 99 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = -1;

#line 99 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 99 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 99 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 99 "ms_deisotope/_c/scoring.pxd"
}

/* "ms_deisotope/_c/scoring.pxd":100
 *         public ScaledGTestFitter g_test_scaled
 *         public double peak_count_scale
 *         public double domain_scale             # <<<<<<<<<<<<<<
 * 
 * cdef class ScaledPenalizedMSDeconvFitter(IsotopicFitterBase):
 */

#line 100 "ms_deisotope/_c/scoring.pxd"


#line 100 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 100 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_12domain_scale_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 100 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_12domain_scale_1__get__(PyObject *__pyx_v_self) {

#line 100 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = 0;

#line 100 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 100 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 100 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_12domain_scale___get__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *)__pyx_v_self));

#line 100 "ms_deisotope/_c/scoring.pxd"


#line 100 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 100 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 100 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 100 "ms_deisotope/_c/scoring.pxd"
}

#line 100 "ms_deisotope/_c/scoring.pxd"


#line 100 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_12domain_scale___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self) {

#line 100 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = NULL;

#line 100 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 100 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 100 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 100 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 100 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 100 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 100 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 100 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->domain_scale); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 100, __pyx_L1_error)

#line 100 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 100 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = __pyx_t_1;

#line 100 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = 0;

#line 100 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 100 "ms_deisotope/_c/scoring.pxd"


#line 100 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 100 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 100 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 100 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.DistinctPatternFitter.domain_scale.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 100 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = NULL;

#line 100 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 100 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 100 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 100 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 100 "ms_deisotope/_c/scoring.pxd"
}

#line 100 "ms_deisotope/_c/scoring.pxd"


#line 100 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 100 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_12domain_scale_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 100 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_12domain_scale_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 100 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 100 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 100 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 100 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_12domain_scale_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 100 "ms_deisotope/_c/scoring.pxd"


#line 100 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 100 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 100 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 100 "ms_deisotope/_c/scoring.pxd"
}

#line 100 "ms_deisotope/_c/scoring.pxd"


#line 100 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_12domain_scale_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_v_self, PyObject *__pyx_v_value) {

#line 100 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 100 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 100 "ms_deisotope/_c/scoring.pxd"
  double __pyx_t_1;

#line 100 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 100 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 100 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 100 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 100 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 100, __pyx_L1_error)

#line 100 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->domain_scale = __pyx_t_1;

#line 100 "ms_deisotope/_c/scoring.pxd"


#line 100 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 100 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 100 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 100 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 100 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.DistinctPatternFitter.domain_scale.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 100 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = -1;

#line 100 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 100 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 100 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 100 "ms_deisotope/_c/scoring.pxd"
}

/* "ms_deisotope/_c/scoring.pyx":898
 * 
 * 
 * cdef double percentile(double[:] N, double percent):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double k, f, c, d0, d1
 */

#line 898 "ms_deisotope/_c/scoring.pyx"


#line 898 "ms_deisotope/_c/scoring.pyx"
static double __pyx_f_12ms_deisotope_2_c_7scoring_percentile(__Pyx_memviewslice __pyx_v_N, double __pyx_v_percent) {
  double __pyx_v_k
#line 898 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_f
#line 898 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_c
#line 898 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_d0
#line 898 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_d1
#line 898 "ms_deisotope/_c/scoring.pyx"
;

#line 898 "ms_deisotope/_c/scoring.pyx"
  double __pyx_r;

#line 898 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 898 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_1;

#line 898 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_2;

#line 898 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_3;

#line 898 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 898 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 898 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 898 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("percentile", 0);

  /* "ms_deisotope/_c/scoring.pyx":901
 *     cdef:
 *         double k, f, c, d0, d1
 *     k = (N.shape[0] - 1) * percent             # <<<<<<<<<<<<<<
 *     f = floor(k)
 *     c = ceil(k)
 */

#line 901 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_k = (((__pyx_v_N.shape[0]) - 1) * __pyx_v_percent);

  /* "ms_deisotope/_c/scoring.pyx":902
 *         double k, f, c, d0, d1
 *     k = (N.shape[0] - 1) * percent
 *     f = floor(k)             # <<<<<<<<<<<<<<
 *     c = ceil(k)
 *     if f == c:
 */

#line 902 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_f = floor(__pyx_v_k);

  /* "ms_deisotope/_c/scoring.pyx":903
 *     k = (N.shape[0] - 1) * percent
 *     f = floor(k)
 *     c = ceil(k)             # <<<<<<<<<<<<<<
 *     if f == c:
 *         return N[<size_t>(k)]
 */

#line 903 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_c = ceil(__pyx_v_k);

  /* "ms_deisotope/_c/scoring.pyx":904
 *     f = floor(k)
 *     c = ceil(k)
 *     if f == c:             # <<<<<<<<<<<<<<
 *         return N[<size_t>(k)]
 *     d0 = N[<size_t>(f)] * (c - k)
 */

#line 904 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = ((__pyx_v_f == __pyx_v_c) != 0);

#line 904 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/scoring.pyx":905
 *     c = ceil(k)
 *     if f == c:
 *         return N[<size_t>(k)]             # <<<<<<<<<<<<<<
 *     d0 = N[<size_t>(f)] * (c - k)
 *     d1 = N[<size_t>(c)] * (k - f)
 */

#line 905 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_2 = ((size_t)__pyx_v_k);

#line 905 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_3 = -1;

#line 905 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_t_2 >= (size_t)__pyx_v_N.shape[0])) __pyx_t_3 = 0;

#line 905 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_t_3 != -1)) {

#line 905 "ms_deisotope/_c/scoring.pyx"
      __Pyx_RaiseBufferIndexError(__pyx_t_3);

#line 905 "ms_deisotope/_c/scoring.pyx"
      __PYX_ERR(0, 905, __pyx_L1_error)

#line 905 "ms_deisotope/_c/scoring.pyx"
    }

#line 905 "ms_deisotope/_c/scoring.pyx"
    __pyx_r = (*((double *) ( /* dim=0 */ (__pyx_v_N.data + __pyx_t_2 * __pyx_v_N.strides[0]) )));

#line 905 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/scoring.pyx":904
 *     f = floor(k)
 *     c = ceil(k)
 *     if f == c:             # <<<<<<<<<<<<<<
 *         return N[<size_t>(k)]
 *     d0 = N[<size_t>(f)] * (c - k)
 */

#line 904 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":906
 *     if f == c:
 *         return N[<size_t>(k)]
 *     d0 = N[<size_t>(f)] * (c - k)             # <<<<<<<<<<<<<<
 *     d1 = N[<size_t>(c)] * (k - f)
 *     return d0 + d1
 */

#line 906 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = ((size_t)__pyx_v_f);

#line 906 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = -1;

#line 906 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_t_2 >= (size_t)__pyx_v_N.shape[0])) __pyx_t_3 = 0;

#line 906 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_t_3 != -1)) {

#line 906 "ms_deisotope/_c/scoring.pyx"
    __Pyx_RaiseBufferIndexError(__pyx_t_3);

#line 906 "ms_deisotope/_c/scoring.pyx"
    __PYX_ERR(0, 906, __pyx_L1_error)

#line 906 "ms_deisotope/_c/scoring.pyx"
  }

#line 906 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_d0 = ((*((double *) ( /* dim=0 */ (__pyx_v_N.data + __pyx_t_2 * __pyx_v_N.strides[0]) ))) * (__pyx_v_c - __pyx_v_k));

  /* "ms_deisotope/_c/scoring.pyx":907
 *         return N[<size_t>(k)]
 *     d0 = N[<size_t>(f)] * (c - k)
 *     d1 = N[<size_t>(c)] * (k - f)             # <<<<<<<<<<<<<<
 *     return d0 + d1
 * 
 */

#line 907 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = ((size_t)__pyx_v_c);

#line 907 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = -1;

#line 907 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_t_2 >= (size_t)__pyx_v_N.shape[0])) __pyx_t_3 = 0;

#line 907 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_t_3 != -1)) {

#line 907 "ms_deisotope/_c/scoring.pyx"
    __Pyx_RaiseBufferIndexError(__pyx_t_3);

#line 907 "ms_deisotope/_c/scoring.pyx"
    __PYX_ERR(0, 907, __pyx_L1_error)

#line 907 "ms_deisotope/_c/scoring.pyx"
  }

#line 907 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_d1 = ((*((double *) ( /* dim=0 */ (__pyx_v_N.data + __pyx_t_2 * __pyx_v_N.strides[0]) ))) * (__pyx_v_k - __pyx_v_f));

  /* "ms_deisotope/_c/scoring.pyx":908
 *     d0 = N[<size_t>(f)] * (c - k)
 *     d1 = N[<size_t>(c)] * (k - f)
 *     return d0 + d1             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 908 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = (__pyx_v_d0 + __pyx_v_d1);

#line 908 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":898
 * 
 * 
 * cdef double percentile(double[:] N, double percent):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double k, f, c, d0, d1
 */

#line 898 "ms_deisotope/_c/scoring.pyx"


#line 898 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 898 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 898 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.percentile", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 898 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 898 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 898 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 898 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 898 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":913
 * cdef class ScaledPenalizedMSDeconvFitter(IsotopicFitterBase):
 * 
 *     def __init__(self,  minimum_score=0.3, penalty_factor=1., mass_error_tolerance=0.02):             # <<<<<<<<<<<<<<
 *         self.select = MaximizeFitSelector(minimum_score)
 *         self.scale_factor = 0.
 */

#line 913 "ms_deisotope/_c/scoring.pyx"


#line 913 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 913 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 913 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_minimum_score = 0
#line 913 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_penalty_factor = 0
#line 913 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_mass_error_tolerance = 0
#line 913 "ms_deisotope/_c/scoring.pyx"
;

#line 913 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 913 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 913 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 913 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 913 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 913 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 913 "ms_deisotope/_c/scoring.pyx"
  {

#line 913 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_minimum_score,&__pyx_n_s_penalty_factor,&__pyx_n_s_mass_error_tolerance,0};

#line 913 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[3] = {0,0,0};

#line 913 "ms_deisotope/_c/scoring.pyx"
    values[0] = ((PyObject *)__pyx_float_0_3);

#line 913 "ms_deisotope/_c/scoring.pyx"
    values[1] = ((PyObject *)__pyx_float_1_);

#line 913 "ms_deisotope/_c/scoring.pyx"
    values[2] = ((PyObject *)__pyx_float_0_02);

#line 913 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 913 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 913 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 913 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  3: 
#line 913 "ms_deisotope/_c/scoring.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 913 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 913 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 913 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 913 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 913 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 913 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 913 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 913 "ms_deisotope/_c/scoring.pyx"
      }

#line 913 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 913 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 913 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 913 "ms_deisotope/_c/scoring.pyx"
        if (kw_args > 0) {

#line 913 "ms_deisotope/_c/scoring.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_minimum_score);

#line 913 "ms_deisotope/_c/scoring.pyx"
          if (value) { values[0] = value; kw_args--; }

#line 913 "ms_deisotope/_c/scoring.pyx"
        }

#line 913 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 913 "ms_deisotope/_c/scoring.pyx"
        case  1:

#line 913 "ms_deisotope/_c/scoring.pyx"
        if (kw_args > 0) {

#line 913 "ms_deisotope/_c/scoring.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_penalty_factor);

#line 913 "ms_deisotope/_c/scoring.pyx"
          if (value) { values[1] = value; kw_args--; }

#line 913 "ms_deisotope/_c/scoring.pyx"
        }

#line 913 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 913 "ms_deisotope/_c/scoring.pyx"
        case  2:

#line 913 "ms_deisotope/_c/scoring.pyx"
        if (kw_args > 0) {

#line 913 "ms_deisotope/_c/scoring.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mass_error_tolerance);

#line 913 "ms_deisotope/_c/scoring.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 913 "ms_deisotope/_c/scoring.pyx"
        }

#line 913 "ms_deisotope/_c/scoring.pyx"
      }

#line 913 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 913 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 913, __pyx_L3_error)

#line 913 "ms_deisotope/_c/scoring.pyx"
      }

#line 913 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 913 "ms_deisotope/_c/scoring.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 913 "ms_deisotope/_c/scoring.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 913 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 913 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 913 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 913 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 913 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  0: 
#line 913 "ms_deisotope/_c/scoring.pyx"
break;
        default: 
#line 913 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 913 "ms_deisotope/_c/scoring.pyx"
      }

#line 913 "ms_deisotope/_c/scoring.pyx"
    }

#line 913 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_minimum_score = values[0];

#line 913 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_penalty_factor = values[1];

#line 913 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_mass_error_tolerance = values[2];

#line 913 "ms_deisotope/_c/scoring.pyx"
  }

#line 913 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 913 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 913 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 913, __pyx_L3_error)

#line 913 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 913 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.ScaledPenalizedMSDeconvFitter.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 913 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 913 "ms_deisotope/_c/scoring.pyx"
  return -1;

#line 913 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 913 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter___init__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *)__pyx_v_self), __pyx_v_minimum_score, __pyx_v_penalty_factor, __pyx_v_mass_error_tolerance);

#line 913 "ms_deisotope/_c/scoring.pyx"


#line 913 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 913 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 913 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 913 "ms_deisotope/_c/scoring.pyx"
}

#line 913 "ms_deisotope/_c/scoring.pyx"


#line 913 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter___init__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self, PyObject *__pyx_v_minimum_score, PyObject *__pyx_v_penalty_factor, PyObject *__pyx_v_mass_error_tolerance) {

#line 913 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 913 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 913 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 913 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 913 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 913 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 913 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 913 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_deisotope/_c/scoring.pyx":914
 * 
 *     def __init__(self,  minimum_score=0.3, penalty_factor=1., mass_error_tolerance=0.02):
 *         self.select = MaximizeFitSelector(minimum_score)             # <<<<<<<<<<<<<<
 *         self.scale_factor = 0.
 *         self.scorer = PenalizedMSDeconVFitter(
 */

#line 914 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_7scoring_MaximizeFitSelector), __pyx_v_minimum_score); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 914, __pyx_L1_error)

#line 914 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 914 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 914 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.select);

#line 914 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx_base.select));

#line 914 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->__pyx_base.select = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_t_1);

#line 914 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/scoring.pyx":915
 *     def __init__(self,  minimum_score=0.3, penalty_factor=1., mass_error_tolerance=0.02):
 *         self.select = MaximizeFitSelector(minimum_score)
 *         self.scale_factor = 0.             # <<<<<<<<<<<<<<
 *         self.scorer = PenalizedMSDeconVFitter(
 *             penalty_factor=penalty_factor, mass_error_tolerance=mass_error_tolerance)
 */

#line 915 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->scale_factor = 0.;

  /* "ms_deisotope/_c/scoring.pyx":917
 *         self.scale_factor = 0.
 *         self.scorer = PenalizedMSDeconVFitter(
 *             penalty_factor=penalty_factor, mass_error_tolerance=mass_error_tolerance)             # <<<<<<<<<<<<<<
 * 
 *     def __getstate__(self):
 */

#line 917 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 917, __pyx_L1_error)

#line 917 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 917 "ms_deisotope/_c/scoring.pyx"
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_penalty_factor, __pyx_v_penalty_factor) < 0) __PYX_ERR(0, 917, __pyx_L1_error)

#line 917 "ms_deisotope/_c/scoring.pyx"
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_mass_error_tolerance, __pyx_v_mass_error_tolerance) < 0) __PYX_ERR(0, 917, __pyx_L1_error)

  /* "ms_deisotope/_c/scoring.pyx":916
 *         self.select = MaximizeFitSelector(minimum_score)
 *         self.scale_factor = 0.
 *         self.scorer = PenalizedMSDeconVFitter(             # <<<<<<<<<<<<<<
 *             penalty_factor=penalty_factor, mass_error_tolerance=mass_error_tolerance)
 * 
 */

#line 916 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter), __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 916, __pyx_L1_error)

#line 916 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 916 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 916 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 916 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->scorer);

#line 916 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->scorer));

#line 916 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->scorer = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *)__pyx_t_2);

#line 916 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/scoring.pyx":913
 * cdef class ScaledPenalizedMSDeconvFitter(IsotopicFitterBase):
 * 
 *     def __init__(self,  minimum_score=0.3, penalty_factor=1., mass_error_tolerance=0.02):             # <<<<<<<<<<<<<<
 *         self.select = MaximizeFitSelector(minimum_score)
 *         self.scale_factor = 0.
 */

#line 913 "ms_deisotope/_c/scoring.pyx"


#line 913 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 913 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 913 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 913 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 913 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 913 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 913 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.ScaledPenalizedMSDeconvFitter.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 913 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = -1;

#line 913 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 913 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 913 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 913 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":919
 *             penalty_factor=penalty_factor, mass_error_tolerance=mass_error_tolerance)
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self.select, self.scale_factor, self.scorer
 * 
 */

#line 919 "ms_deisotope/_c/scoring.pyx"


#line 919 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 919 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_3__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 919 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_2__getstate__[] = "ScaledPenalizedMSDeconvFitter.__getstate__(self)";

#line 919 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_3__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 919 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 919 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 919 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = 
#line 919 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_2__getstate__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *)__pyx_v_self));

#line 919 "ms_deisotope/_c/scoring.pyx"


#line 919 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 919 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 919 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 919 "ms_deisotope/_c/scoring.pyx"
}

#line 919 "ms_deisotope/_c/scoring.pyx"


#line 919 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_2__getstate__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self) {

#line 919 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 919 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 919 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 919 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 919 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 919 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 919 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 919 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "ms_deisotope/_c/scoring.pyx":920
 * 
 *     def __getstate__(self):
 *         return self.select, self.scale_factor, self.scorer             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, state):
 */

#line 920 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 920 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->scale_factor); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 920, __pyx_L1_error)

#line 920 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 920 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 920, __pyx_L1_error)

#line 920 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 920 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx_base.select));

#line 920 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx_base.select));

#line 920 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_self->__pyx_base.select));

#line 920 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 920 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);

#line 920 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->scorer));

#line 920 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->scorer));

#line 920 "ms_deisotope/_c/scoring.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 2, ((PyObject *)__pyx_v_self->scorer));

#line 920 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 920 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_2;

#line 920 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = 0;

#line 920 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":919
 *             penalty_factor=penalty_factor, mass_error_tolerance=mass_error_tolerance)
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self.select, self.scale_factor, self.scorer
 * 
 */

#line 919 "ms_deisotope/_c/scoring.pyx"


#line 919 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 919 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 919 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 919 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 919 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.ScaledPenalizedMSDeconvFitter.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 919 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 919 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 919 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 919 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 919 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 919 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":922
 *         return self.select, self.scale_factor, self.scorer
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.select, self.scale_factor, self.scorer = state
 * 
 */

#line 922 "ms_deisotope/_c/scoring.pyx"


#line 922 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 922 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_5__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state); /*proto*/

#line 922 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_4__setstate__[] = "ScaledPenalizedMSDeconvFitter.__setstate__(self, state)";

#line 922 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_5__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state) {

#line 922 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 922 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 922 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  __pyx_r = 
#line 922 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_4__setstate__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *)__pyx_v_self), ((PyObject *)__pyx_v_state));

#line 922 "ms_deisotope/_c/scoring.pyx"


#line 922 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 922 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 922 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 922 "ms_deisotope/_c/scoring.pyx"
}

#line 922 "ms_deisotope/_c/scoring.pyx"


#line 922 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_4__setstate__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self, PyObject *__pyx_v_state) {

#line 922 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 922 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 922 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 922 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 922 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 922 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 922 "ms_deisotope/_c/scoring.pyx"
  PyObject *(*__pyx_t_5)(PyObject *);

#line 922 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_6;

#line 922 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 922 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 922 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 922 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* "ms_deisotope/_c/scoring.pyx":923
 * 
 *     def __setstate__(self, state):
 *         self.select, self.scale_factor, self.scorer = state             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 923 "ms_deisotope/_c/scoring.pyx"
  if ((likely(PyTuple_CheckExact(__pyx_v_state))) || (PyList_CheckExact(__pyx_v_state))) {

#line 923 "ms_deisotope/_c/scoring.pyx"
    PyObject* sequence = __pyx_v_state;

#line 923 "ms_deisotope/_c/scoring.pyx"
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);

#line 923 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(size != 3)) {

#line 923 "ms_deisotope/_c/scoring.pyx"
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);

#line 923 "ms_deisotope/_c/scoring.pyx"
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);

#line 923 "ms_deisotope/_c/scoring.pyx"
      __PYX_ERR(0, 923, __pyx_L1_error)

#line 923 "ms_deisotope/_c/scoring.pyx"
    }

#line 923 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 923 "ms_deisotope/_c/scoring.pyx"
    if (likely(PyTuple_CheckExact(sequence))) {

#line 923 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 

#line 923 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 

#line 923 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 2); 

#line 923 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 923 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 

#line 923 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 

#line 923 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_3 = PyList_GET_ITEM(sequence, 2); 

#line 923 "ms_deisotope/_c/scoring.pyx"
    }

#line 923 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 923 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_2);

#line 923 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_3);

#line 923 "ms_deisotope/_c/scoring.pyx"
    #else

#line 923 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 923, __pyx_L1_error)

#line 923 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 923 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 923, __pyx_L1_error)

#line 923 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 923 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_3 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 923, __pyx_L1_error)

#line 923 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 923 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 923 "ms_deisotope/_c/scoring.pyx"
  } else {

#line 923 "ms_deisotope/_c/scoring.pyx"
    Py_ssize_t index = -1;

#line 923 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_4 = PyObject_GetIter(__pyx_v_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 923, __pyx_L1_error)

#line 923 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 923 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_1 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_1)) 
#line 923 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L3_unpacking_failed;

#line 923 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_1);
    index = 1; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) 
#line 923 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L3_unpacking_failed;

#line 923 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_2);
    index = 2; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) 
#line 923 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L3_unpacking_failed;

#line 923 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 923 "ms_deisotope/_c/scoring.pyx"
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 3) < 0) __PYX_ERR(0, 923, __pyx_L1_error)

#line 923 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = NULL;

#line 923 "ms_deisotope/_c/scoring.pyx"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 923 "ms_deisotope/_c/scoring.pyx"
    goto __pyx_L4_unpacking_done;

#line 923 "ms_deisotope/_c/scoring.pyx"
    __pyx_L3_unpacking_failed:;

#line 923 "ms_deisotope/_c/scoring.pyx"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 923 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = NULL;

#line 923 "ms_deisotope/_c/scoring.pyx"
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);

#line 923 "ms_deisotope/_c/scoring.pyx"
    __PYX_ERR(0, 923, __pyx_L1_error)

#line 923 "ms_deisotope/_c/scoring.pyx"
    __pyx_L4_unpacking_done:;

#line 923 "ms_deisotope/_c/scoring.pyx"
  }

#line 923 "ms_deisotope/_c/scoring.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase))))) __PYX_ERR(0, 923, __pyx_L1_error)

#line 923 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_6 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 923, __pyx_L1_error)

#line 923 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 923 "ms_deisotope/_c/scoring.pyx"
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter))))) __PYX_ERR(0, 923, __pyx_L1_error)

#line 923 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 923 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.select);

#line 923 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx_base.select));

#line 923 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->__pyx_base.select = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_t_1);

#line 923 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 923 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->scale_factor = __pyx_t_6;

#line 923 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 923 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->scorer);

#line 923 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->scorer));

#line 923 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->scorer = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *)__pyx_t_3);

#line 923 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/scoring.pyx":922
 *         return self.select, self.scale_factor, self.scorer
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.select, self.scale_factor, self.scorer = state
 * 
 */

#line 922 "ms_deisotope/_c/scoring.pyx"


#line 922 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 922 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 922 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 922 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 922 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 922 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 922 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 922 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 922 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.ScaledPenalizedMSDeconvFitter.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 922 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 922 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 922 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 922 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 922 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 922 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":926
 * 
 *     @property
 *     def penalty_factor(self):             # <<<<<<<<<<<<<<
 *         return self.scorer.penalty_factor
 * 
 */

#line 926 "ms_deisotope/_c/scoring.pyx"


#line 926 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 926 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_14penalty_factor_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 926 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_14penalty_factor_1__get__(PyObject *__pyx_v_self) {

#line 926 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 926 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 926 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 926 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_14penalty_factor___get__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *)__pyx_v_self));

#line 926 "ms_deisotope/_c/scoring.pyx"


#line 926 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 926 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 926 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 926 "ms_deisotope/_c/scoring.pyx"
}

#line 926 "ms_deisotope/_c/scoring.pyx"


#line 926 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_14penalty_factor___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self) {

#line 926 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 926 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 926 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 926 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 926 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 926 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 926 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ms_deisotope/_c/scoring.pyx":927
 *     @property
 *     def penalty_factor(self):
 *         return self.scorer.penalty_factor             # <<<<<<<<<<<<<<
 * 
 *     @penalty_factor.setter
 */

#line 927 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 927 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->scorer->penalty_factor); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 927, __pyx_L1_error)

#line 927 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 927 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 927 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 927 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":926
 * 
 *     @property
 *     def penalty_factor(self):             # <<<<<<<<<<<<<<
 *         return self.scorer.penalty_factor
 * 
 */

#line 926 "ms_deisotope/_c/scoring.pyx"


#line 926 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 926 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 926 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 926 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.ScaledPenalizedMSDeconvFitter.penalty_factor.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 926 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 926 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 926 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 926 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 926 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 926 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":930
 * 
 *     @penalty_factor.setter
 *     def penalty_factor(self, value):             # <<<<<<<<<<<<<<
 *         self.scorer.penalty_factor = value
 * 
 */

#line 930 "ms_deisotope/_c/scoring.pyx"


#line 930 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 930 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_14penalty_factor_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 930 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_14penalty_factor_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 930 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 930 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 930 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 930 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_14penalty_factor_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 930 "ms_deisotope/_c/scoring.pyx"


#line 930 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 930 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 930 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 930 "ms_deisotope/_c/scoring.pyx"
}

#line 930 "ms_deisotope/_c/scoring.pyx"


#line 930 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_14penalty_factor_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self, PyObject *__pyx_v_value) {

#line 930 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 930 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 930 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_1;

#line 930 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 930 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 930 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 930 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "ms_deisotope/_c/scoring.pyx":931
 *     @penalty_factor.setter
 *     def penalty_factor(self, value):
 *         self.scorer.penalty_factor = value             # <<<<<<<<<<<<<<
 * 
 *     cpdef double _calculate_scale_factor(self, PeakSet peaklist):
 */

#line 931 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 931, __pyx_L1_error)

#line 931 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->scorer->penalty_factor = __pyx_t_1;

  /* "ms_deisotope/_c/scoring.pyx":930
 * 
 *     @penalty_factor.setter
 *     def penalty_factor(self, value):             # <<<<<<<<<<<<<<
 *         self.scorer.penalty_factor = value
 * 
 */

#line 930 "ms_deisotope/_c/scoring.pyx"


#line 930 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 930 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 930 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 930 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 930 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.ScaledPenalizedMSDeconvFitter.penalty_factor.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 930 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = -1;

#line 930 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 930 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 930 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 930 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":933
 *         self.scorer.penalty_factor = value
 * 
 *     cpdef double _calculate_scale_factor(self, PeakSet peaklist):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i
 */

#line 933 "ms_deisotope/_c/scoring.pyx"


#line 933 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_7_calculate_scale_factor(PyObject *__pyx_v_self, PyObject *__pyx_v_peaklist); /*proto*/

#line 933 "ms_deisotope/_c/scoring.pyx"
static double __pyx_f_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter__calculate_scale_factor(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, int __pyx_skip_dispatch) {
  size_t __pyx_v_i
#line 933 "ms_deisotope/_c/scoring.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = 0
#line 933 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_maximum
#line 933 "ms_deisotope/_c/scoring.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaks = 0
#line 933 "ms_deisotope/_c/scoring.pyx"
;

#line 933 "ms_deisotope/_c/scoring.pyx"
  double __pyx_r;

#line 933 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 933 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 933 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 933 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 933 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 933 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_5;

#line 933 "ms_deisotope/_c/scoring.pyx"
  Py_ssize_t __pyx_t_6;

#line 933 "ms_deisotope/_c/scoring.pyx"
  Py_ssize_t __pyx_t_7;

#line 933 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_8;

#line 933 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_9;

#line 933 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 933 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 933 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 933 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_calculate_scale_factor", 0);

#line 933 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 933 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 933 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 933 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 933 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 933 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 933 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 933 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 933 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 933 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_calculate_scale_factor); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 933, __pyx_L1_error)

#line 933 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 933 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_7_calculate_scale_factor)) {

#line 933 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 933 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 933 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 933 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 933 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 933 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 933 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 933 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 933 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 933 "ms_deisotope/_c/scoring.pyx"
          }

#line 933 "ms_deisotope/_c/scoring.pyx"
        }

#line 933 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_peaklist)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_peaklist));

#line 933 "ms_deisotope/_c/scoring.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 933 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 933, __pyx_L1_error)

#line 933 "ms_deisotope/_c/scoring.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 933 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 933 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 933, __pyx_L1_error)

#line 933 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 933 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = __pyx_t_5;

#line 933 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 933 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 933 "ms_deisotope/_c/scoring.pyx"
      }

#line 933 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 933 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 933 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 933 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 933 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 933 "ms_deisotope/_c/scoring.pyx"
      }

#line 933 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 933 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 933 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 933 "ms_deisotope/_c/scoring.pyx"
    }

#line 933 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 933 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":941
 *             PeakSet peaks
 * 
 *         maximum = 0.             # <<<<<<<<<<<<<<
 *         peaks = peaklist.peaks
 *         for i in range(len(peaks)):
 */

#line 941 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_maximum = 0.;

  /* "ms_deisotope/_c/scoring.pyx":942
 * 
 *         maximum = 0.
 *         peaks = peaklist.peaks             # <<<<<<<<<<<<<<
 *         for i in range(len(peaks)):
 *             peak = peaks.getitem(i)
 */

#line 942 "ms_deisotope/_c/scoring.pyx"
  if (!(likely(((__pyx_v_peaklist->peaks) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_peaklist->peaks, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet))))) __PYX_ERR(0, 942, __pyx_L1_error)

#line 942 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __pyx_v_peaklist->peaks;

#line 942 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 942 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_peaks = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_t_1);

#line 942 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/scoring.pyx":943
 *         maximum = 0.
 *         peaks = peaklist.peaks
 *         for i in range(len(peaks)):             # <<<<<<<<<<<<<<
 *             peak = peaks.getitem(i)
 *             if peak.intensity > maximum:
 */

#line 943 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_6 = PyObject_Length(((PyObject *)__pyx_v_peaks)); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 943, __pyx_L1_error)

#line 943 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_7 = __pyx_t_6;

#line 943 "ms_deisotope/_c/scoring.pyx"
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {

#line 943 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_i = __pyx_t_8;

    /* "ms_deisotope/_c/scoring.pyx":944
 *         peaks = peaklist.peaks
 *         for i in range(len(peaks)):
 *             peak = peaks.getitem(i)             # <<<<<<<<<<<<<<
 *             if peak.intensity > maximum:
 *                 maximum = peak.intensity
 */

#line 944 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaks->__pyx_vtab)->getitem(__pyx_v_peaks, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 944, __pyx_L1_error)

#line 944 "ms_deisotope/_c/scoring.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 944 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1));

#line 944 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/scoring.pyx":945
 *         for i in range(len(peaks)):
 *             peak = peaks.getitem(i)
 *             if peak.intensity > maximum:             # <<<<<<<<<<<<<<
 *                 maximum = peak.intensity
 *         return maximum
 */

#line 945 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_9 = ((__pyx_v_peak->__pyx_base.intensity > __pyx_v_maximum) != 0);

#line 945 "ms_deisotope/_c/scoring.pyx"
    if (__pyx_t_9) {

      /* "ms_deisotope/_c/scoring.pyx":946
 *             peak = peaks.getitem(i)
 *             if peak.intensity > maximum:
 *                 maximum = peak.intensity             # <<<<<<<<<<<<<<
 *         return maximum
 * 
 */

#line 946 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_5 = __pyx_v_peak->__pyx_base.intensity;

#line 946 "ms_deisotope/_c/scoring.pyx"
      __pyx_v_maximum = __pyx_t_5;

      /* "ms_deisotope/_c/scoring.pyx":945
 *         for i in range(len(peaks)):
 *             peak = peaks.getitem(i)
 *             if peak.intensity > maximum:             # <<<<<<<<<<<<<<
 *                 maximum = peak.intensity
 *         return maximum
 */

#line 945 "ms_deisotope/_c/scoring.pyx"
    }

#line 945 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":947
 *             if peak.intensity > maximum:
 *                 maximum = peak.intensity
 *         return maximum             # <<<<<<<<<<<<<<
 * 
 *     cdef void scale_fitted_peaks(self, list experimental, double factor):
 */

#line 947 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_v_maximum;

#line 947 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":933
 *         self.scorer.penalty_factor = value
 * 
 *     cpdef double _calculate_scale_factor(self, PeakSet peaklist):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i
 */

#line 933 "ms_deisotope/_c/scoring.pyx"


#line 933 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 933 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 933 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 933 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 933 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 933 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 933 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.ScaledPenalizedMSDeconvFitter._calculate_scale_factor", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 933 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 933 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 933 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 933 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peaks);

#line 933 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 933 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 933 "ms_deisotope/_c/scoring.pyx"
}

#line 933 "ms_deisotope/_c/scoring.pyx"


#line 933 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 933 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_7_calculate_scale_factor(PyObject *__pyx_v_self, PyObject *__pyx_v_peaklist); /*proto*/

#line 933 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_6_calculate_scale_factor[] = "ScaledPenalizedMSDeconvFitter._calculate_scale_factor(self, PeakSet peaklist) -> double";

#line 933 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_7_calculate_scale_factor(PyObject *__pyx_v_self, PyObject *__pyx_v_peaklist) {

#line 933 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 933 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 933 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 933 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 933 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 933 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_calculate_scale_factor (wrapper)", 0);

#line 933 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peaklist), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peaklist", 0))) __PYX_ERR(0, 933, __pyx_L1_error)
  __pyx_r = 
#line 933 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_6_calculate_scale_factor(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *)__pyx_v_self), ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)__pyx_v_peaklist));

#line 933 "ms_deisotope/_c/scoring.pyx"


#line 933 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 933 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 933 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 933 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 933 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 933 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 933 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 933 "ms_deisotope/_c/scoring.pyx"
}

#line 933 "ms_deisotope/_c/scoring.pyx"


#line 933 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_6_calculate_scale_factor(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist) {

#line 933 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 933 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 933 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 933 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 933 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 933 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 933 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_calculate_scale_factor", 0);

#line 933 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 933 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter__calculate_scale_factor(__pyx_v_self, __pyx_v_peaklist, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 933, __pyx_L1_error)

#line 933 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 933 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 933 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 933 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 933 "ms_deisotope/_c/scoring.pyx"


#line 933 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 933 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 933 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 933 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.ScaledPenalizedMSDeconvFitter._calculate_scale_factor", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 933 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 933 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 933 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 933 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 933 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 933 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":949
 *         return maximum
 * 
 *     cdef void scale_fitted_peaks(self, list experimental, double factor):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i
 */

#line 949 "ms_deisotope/_c/scoring.pyx"


#line 949 "ms_deisotope/_c/scoring.pyx"
static void __pyx_f_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_scale_fitted_peaks(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self, PyObject *__pyx_v_experimental, double __pyx_v_factor) {
  size_t __pyx_v_i
#line 949 "ms_deisotope/_c/scoring.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_peak = 0
#line 949 "ms_deisotope/_c/scoring.pyx"
;

#line 949 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 949 "ms_deisotope/_c/scoring.pyx"
  Py_ssize_t __pyx_t_1;

#line 949 "ms_deisotope/_c/scoring.pyx"
  Py_ssize_t __pyx_t_2;

#line 949 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_3;

#line 949 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4;

#line 949 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 949 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 949 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 949 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 949 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("scale_fitted_peaks", 0);

  /* "ms_deisotope/_c/scoring.pyx":953
 *             size_t i
 *             FittedPeak peak
 *         for i in range(len(experimental)):             # <<<<<<<<<<<<<<
 *             peak = <FittedPeak>PyList_GET_ITEM(experimental, i)
 *             peak.intensity *= factor
 */

#line 953 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_v_experimental == Py_None)) {

#line 953 "ms_deisotope/_c/scoring.pyx"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 953 "ms_deisotope/_c/scoring.pyx"
    __PYX_ERR(0, 953, __pyx_L1_error)

#line 953 "ms_deisotope/_c/scoring.pyx"
  }

#line 953 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyList_GET_SIZE(__pyx_v_experimental); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 953, __pyx_L1_error)

#line 953 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = __pyx_t_1;

#line 953 "ms_deisotope/_c/scoring.pyx"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 953 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/scoring.pyx":954
 *             FittedPeak peak
 *         for i in range(len(experimental)):
 *             peak = <FittedPeak>PyList_GET_ITEM(experimental, i)             # <<<<<<<<<<<<<<
 *             peak.intensity *= factor
 * 
 */

#line 954 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_4 = PyList_GET_ITEM(__pyx_v_experimental, __pyx_v_i);

#line 954 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = ((PyObject *)__pyx_t_4);

#line 954 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_5);

#line 954 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_5));

#line 954 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = 0;

    /* "ms_deisotope/_c/scoring.pyx":955
 *         for i in range(len(experimental)):
 *             peak = <FittedPeak>PyList_GET_ITEM(experimental, i)
 *             peak.intensity *= factor             # <<<<<<<<<<<<<<
 * 
 *     cdef void scale_theoretical_peaks(self, list theoretical, double factor):
 */

#line 955 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_peak->__pyx_base.intensity = (__pyx_v_peak->__pyx_base.intensity * __pyx_v_factor);

#line 955 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":949
 *         return maximum
 * 
 *     cdef void scale_fitted_peaks(self, list experimental, double factor):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i
 */

#line 949 "ms_deisotope/_c/scoring.pyx"


#line 949 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 949 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 949 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 949 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 949 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.ScaledPenalizedMSDeconvFitter.scale_fitted_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 949 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 949 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 949 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 949 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":957
 *             peak.intensity *= factor
 * 
 *     cdef void scale_theoretical_peaks(self, list theoretical, double factor):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i
 */

#line 957 "ms_deisotope/_c/scoring.pyx"


#line 957 "ms_deisotope/_c/scoring.pyx"
static void __pyx_f_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_scale_theoretical_peaks(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self, PyObject *__pyx_v_theoretical, double __pyx_v_factor) {
  size_t __pyx_v_i
#line 957 "ms_deisotope/_c/scoring.pyx"
;
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_peak = 0
#line 957 "ms_deisotope/_c/scoring.pyx"
;

#line 957 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 957 "ms_deisotope/_c/scoring.pyx"
  Py_ssize_t __pyx_t_1;

#line 957 "ms_deisotope/_c/scoring.pyx"
  Py_ssize_t __pyx_t_2;

#line 957 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_3;

#line 957 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4;

#line 957 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 957 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 957 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 957 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 957 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("scale_theoretical_peaks", 0);

  /* "ms_deisotope/_c/scoring.pyx":961
 *             size_t i
 *             TheoreticalPeak peak
 *         for i in range(len(theoretical)):             # <<<<<<<<<<<<<<
 *             peak = <TheoreticalPeak>PyList_GET_ITEM(theoretical, i)
 *             peak.intensity *= factor
 */

#line 961 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_v_theoretical == Py_None)) {

#line 961 "ms_deisotope/_c/scoring.pyx"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 961 "ms_deisotope/_c/scoring.pyx"
    __PYX_ERR(0, 961, __pyx_L1_error)

#line 961 "ms_deisotope/_c/scoring.pyx"
  }

#line 961 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyList_GET_SIZE(__pyx_v_theoretical); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 961, __pyx_L1_error)

#line 961 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_2 = __pyx_t_1;

#line 961 "ms_deisotope/_c/scoring.pyx"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 961 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/scoring.pyx":962
 *             TheoreticalPeak peak
 *         for i in range(len(theoretical)):
 *             peak = <TheoreticalPeak>PyList_GET_ITEM(theoretical, i)             # <<<<<<<<<<<<<<
 *             peak.intensity *= factor
 * 
 */

#line 962 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_4 = PyList_GET_ITEM(__pyx_v_theoretical, __pyx_v_i);

#line 962 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = ((PyObject *)__pyx_t_4);

#line 962 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_5);

#line 962 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_5));

#line 962 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_5 = 0;

    /* "ms_deisotope/_c/scoring.pyx":963
 *         for i in range(len(theoretical)):
 *             peak = <TheoreticalPeak>PyList_GET_ITEM(theoretical, i)
 *             peak.intensity *= factor             # <<<<<<<<<<<<<<
 * 
 *     def evaluate(self, PeakSet peaklist, list observed, list expected):
 */

#line 963 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_peak->intensity = (__pyx_v_peak->intensity * __pyx_v_factor);

#line 963 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":957
 *             peak.intensity *= factor
 * 
 *     cdef void scale_theoretical_peaks(self, list theoretical, double factor):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i
 */

#line 957 "ms_deisotope/_c/scoring.pyx"


#line 957 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 957 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 957 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 957 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 957 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.ScaledPenalizedMSDeconvFitter.scale_theoretical_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 957 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 957 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 957 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 957 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":965
 *             peak.intensity *= factor
 * 
 *     def evaluate(self, PeakSet peaklist, list observed, list expected):             # <<<<<<<<<<<<<<
 *         return self._evaluate(peaklist, observed, expected)
 * 
 */

#line 965 "ms_deisotope/_c/scoring.pyx"


#line 965 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 965 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_9evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 965 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_8evaluate[] = "ScaledPenalizedMSDeconvFitter.evaluate(self, PeakSet peaklist, list observed, list expected)";

#line 965 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_9evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist = 0
#line 965 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_observed = 0
#line 965 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_expected = 0
#line 965 "ms_deisotope/_c/scoring.pyx"
;

#line 965 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 965 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 965 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 965 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 965 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 965 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("evaluate (wrapper)", 0);

#line 965 "ms_deisotope/_c/scoring.pyx"
  {

#line 965 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peaklist,&__pyx_n_s_observed,&__pyx_n_s_expected,0};

#line 965 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[3] = {0,0,0};

#line 965 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 965 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 965 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 965 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  3: 
#line 965 "ms_deisotope/_c/scoring.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 965 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 965 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 965 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 965 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 965 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 965 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 965 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 965 "ms_deisotope/_c/scoring.pyx"
      }

#line 965 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 965 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 965 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 965 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaklist)) != 0)) kw_args--;
        else 
#line 965 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 965 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 965 "ms_deisotope/_c/scoring.pyx"
        case  1:

#line 965 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_observed)) != 0)) kw_args--;

#line 965 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("evaluate", 1, 3, 3, 1); 
#line 965 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 965, __pyx_L3_error)

#line 965 "ms_deisotope/_c/scoring.pyx"
        }

#line 965 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 965 "ms_deisotope/_c/scoring.pyx"
        case  2:

#line 965 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_expected)) != 0)) kw_args--;

#line 965 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("evaluate", 1, 3, 3, 2); 
#line 965 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 965, __pyx_L3_error)

#line 965 "ms_deisotope/_c/scoring.pyx"
        }

#line 965 "ms_deisotope/_c/scoring.pyx"
      }

#line 965 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 965 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "evaluate") < 0)) __PYX_ERR(0, 965, __pyx_L3_error)

#line 965 "ms_deisotope/_c/scoring.pyx"
      }

#line 965 "ms_deisotope/_c/scoring.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 965 "ms_deisotope/_c/scoring.pyx"
      goto __pyx_L5_argtuple_error;

#line 965 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 965 "ms_deisotope/_c/scoring.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 965 "ms_deisotope/_c/scoring.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 965 "ms_deisotope/_c/scoring.pyx"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 965 "ms_deisotope/_c/scoring.pyx"
    }

#line 965 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_peaklist = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[0]);

#line 965 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_observed = ((PyObject*)values[1]);

#line 965 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_expected = ((PyObject*)values[2]);

#line 965 "ms_deisotope/_c/scoring.pyx"
  }

#line 965 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 965 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("evaluate", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 965 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 965, __pyx_L3_error)

#line 965 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 965 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.ScaledPenalizedMSDeconvFitter.evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 965 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 965 "ms_deisotope/_c/scoring.pyx"
  return NULL;

#line 965 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 965 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peaklist), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peaklist", 0))) __PYX_ERR(0, 965, __pyx_L1_error)

#line 965 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_observed), (&PyList_Type), 1, "observed", 1))) __PYX_ERR(0, 965, __pyx_L1_error)

#line 965 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_expected), (&PyList_Type), 1, "expected", 1))) __PYX_ERR(0, 965, __pyx_L1_error)
  __pyx_r = 
#line 965 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_8evaluate(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *)__pyx_v_self), __pyx_v_peaklist, __pyx_v_observed, __pyx_v_expected);

#line 965 "ms_deisotope/_c/scoring.pyx"


#line 965 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 965 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 965 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 965 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 965 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 965 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 965 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 965 "ms_deisotope/_c/scoring.pyx"
}

#line 965 "ms_deisotope/_c/scoring.pyx"


#line 965 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_8evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_observed, PyObject *__pyx_v_expected) {

#line 965 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 965 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 965 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 965 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 965 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 965 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 965 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("evaluate", 0);

  /* "ms_deisotope/_c/scoring.pyx":966
 * 
 *     def evaluate(self, PeakSet peaklist, list observed, list expected):
 *         return self._evaluate(peaklist, observed, expected)             # <<<<<<<<<<<<<<
 * 
 *     cpdef double _evaluate(self, PeakSet peaklist, list experimental, list theoretical):
 */

#line 966 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 966 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyFloat_FromDouble(((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._evaluate(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self), __pyx_v_peaklist, __pyx_v_observed, __pyx_v_expected, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 966, __pyx_L1_error)

#line 966 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 966 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 966 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 966 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":965
 *             peak.intensity *= factor
 * 
 *     def evaluate(self, PeakSet peaklist, list observed, list expected):             # <<<<<<<<<<<<<<
 *         return self._evaluate(peaklist, observed, expected)
 * 
 */

#line 965 "ms_deisotope/_c/scoring.pyx"


#line 965 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 965 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 965 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 965 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.ScaledPenalizedMSDeconvFitter.evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 965 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 965 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 965 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 965 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 965 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 965 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":968
 *         return self._evaluate(peaklist, observed, expected)
 * 
 *     cpdef double _evaluate(self, PeakSet peaklist, list experimental, list theoretical):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double score
 */

#line 968 "ms_deisotope/_c/scoring.pyx"


#line 968 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_11_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 968 "ms_deisotope/_c/scoring.pyx"
static double __pyx_f_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter__evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_experimental, PyObject *__pyx_v_theoretical, int __pyx_skip_dispatch) {
  double __pyx_v_score
#line 968 "ms_deisotope/_c/scoring.pyx"
;

#line 968 "ms_deisotope/_c/scoring.pyx"
  double __pyx_r;

#line 968 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 968 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 968 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 968 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 968 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 968 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_5;

#line 968 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 968 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_7;

#line 968 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_8;

#line 968 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 968 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 968 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 968 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate", 0);

#line 968 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 968 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 968 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 968 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 968 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 968 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 968 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 968 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 968 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 968 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_evaluate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 968, __pyx_L1_error)

#line 968 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 968 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_11_evaluate)) {

#line 968 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 968 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 968 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_5 = 0;

#line 968 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 968 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 968 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 968 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 968 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 968 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 968 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 968 "ms_deisotope/_c/scoring.pyx"
            __pyx_t_5 = 1;

#line 968 "ms_deisotope/_c/scoring.pyx"
          }

#line 968 "ms_deisotope/_c/scoring.pyx"
        }

#line 968 "ms_deisotope/_c/scoring.pyx"
        #if CYTHON_FAST_PYCALL

#line 968 "ms_deisotope/_c/scoring.pyx"
        if (PyFunction_Check(__pyx_t_3)) {

#line 968 "ms_deisotope/_c/scoring.pyx"
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_peaklist), __pyx_v_experimental, __pyx_v_theoretical};

#line 968 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 968, __pyx_L1_error)

#line 968 "ms_deisotope/_c/scoring.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 968 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 968 "ms_deisotope/_c/scoring.pyx"
        } else

#line 968 "ms_deisotope/_c/scoring.pyx"
        #endif

#line 968 "ms_deisotope/_c/scoring.pyx"
        #if CYTHON_FAST_PYCCALL

#line 968 "ms_deisotope/_c/scoring.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {

#line 968 "ms_deisotope/_c/scoring.pyx"
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_peaklist), __pyx_v_experimental, __pyx_v_theoretical};

#line 968 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 968, __pyx_L1_error)

#line 968 "ms_deisotope/_c/scoring.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 968 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 968 "ms_deisotope/_c/scoring.pyx"
        } else

#line 968 "ms_deisotope/_c/scoring.pyx"
        #endif

#line 968 "ms_deisotope/_c/scoring.pyx"
        {

#line 968 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_6 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 968, __pyx_L1_error)

#line 968 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_6);

#line 968 "ms_deisotope/_c/scoring.pyx"
          if (__pyx_t_4) {

#line 968 "ms_deisotope/_c/scoring.pyx"
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;

#line 968 "ms_deisotope/_c/scoring.pyx"
          }

#line 968 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(((PyObject *)__pyx_v_peaklist));

#line 968 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(((PyObject *)__pyx_v_peaklist));

#line 968 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_peaklist));

#line 968 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(__pyx_v_experimental);

#line 968 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(__pyx_v_experimental);

#line 968 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_experimental);

#line 968 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(__pyx_v_theoretical);

#line 968 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(__pyx_v_theoretical);

#line 968 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, __pyx_v_theoretical);

#line 968 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 968, __pyx_L1_error)

#line 968 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 968 "ms_deisotope/_c/scoring.pyx"
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 968 "ms_deisotope/_c/scoring.pyx"
        }

#line 968 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 968 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_7 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_7 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 968, __pyx_L1_error)

#line 968 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 968 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = __pyx_t_7;

#line 968 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 968 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 968 "ms_deisotope/_c/scoring.pyx"
      }

#line 968 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 968 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 968 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 968 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 968 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 968 "ms_deisotope/_c/scoring.pyx"
      }

#line 968 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 968 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 968 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 968 "ms_deisotope/_c/scoring.pyx"
    }

#line 968 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 968 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":971
 *         cdef:
 *             double score
 *         if self.scale_factor < 1:             # <<<<<<<<<<<<<<
 *             self.scale_factor = self._calculate_scale_factor(peaklist)
 *         self.scale_fitted_peaks(experimental, 1. / self.scale_factor)
 */

#line 971 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_8 = ((__pyx_v_self->scale_factor < 1.0) != 0);

#line 971 "ms_deisotope/_c/scoring.pyx"
  if (__pyx_t_8) {

    /* "ms_deisotope/_c/scoring.pyx":972
 *             double score
 *         if self.scale_factor < 1:
 *             self.scale_factor = self._calculate_scale_factor(peaklist)             # <<<<<<<<<<<<<<
 *         self.scale_fitted_peaks(experimental, 1. / self.scale_factor)
 *         self.scale_theoretical_peaks(theoretical, 1. / self.scale_factor)
 */

#line 972 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_self->scale_factor = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *)__pyx_v_self->__pyx_base.__pyx_vtab)->_calculate_scale_factor(__pyx_v_self, __pyx_v_peaklist, 0);

    /* "ms_deisotope/_c/scoring.pyx":971
 *         cdef:
 *             double score
 *         if self.scale_factor < 1:             # <<<<<<<<<<<<<<
 *             self.scale_factor = self._calculate_scale_factor(peaklist)
 *         self.scale_fitted_peaks(experimental, 1. / self.scale_factor)
 */

#line 971 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":973
 *         if self.scale_factor < 1:
 *             self.scale_factor = self._calculate_scale_factor(peaklist)
 *         self.scale_fitted_peaks(experimental, 1. / self.scale_factor)             # <<<<<<<<<<<<<<
 *         self.scale_theoretical_peaks(theoretical, 1. / self.scale_factor)
 *         score = self.scorer._evaluate(peaklist, experimental, theoretical)
 */

#line 973 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_v_self->scale_factor == 0)) {

#line 973 "ms_deisotope/_c/scoring.pyx"
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");

#line 973 "ms_deisotope/_c/scoring.pyx"
    __PYX_ERR(0, 973, __pyx_L1_error)

#line 973 "ms_deisotope/_c/scoring.pyx"
  }

#line 973 "ms_deisotope/_c/scoring.pyx"
  ((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *)__pyx_v_self->__pyx_base.__pyx_vtab)->scale_fitted_peaks(__pyx_v_self, __pyx_v_experimental, (1. / __pyx_v_self->scale_factor));

  /* "ms_deisotope/_c/scoring.pyx":974
 *             self.scale_factor = self._calculate_scale_factor(peaklist)
 *         self.scale_fitted_peaks(experimental, 1. / self.scale_factor)
 *         self.scale_theoretical_peaks(theoretical, 1. / self.scale_factor)             # <<<<<<<<<<<<<<
 *         score = self.scorer._evaluate(peaklist, experimental, theoretical)
 *         self.scale_fitted_peaks(experimental, self.scale_factor)
 */

#line 974 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_v_self->scale_factor == 0)) {

#line 974 "ms_deisotope/_c/scoring.pyx"
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");

#line 974 "ms_deisotope/_c/scoring.pyx"
    __PYX_ERR(0, 974, __pyx_L1_error)

#line 974 "ms_deisotope/_c/scoring.pyx"
  }

#line 974 "ms_deisotope/_c/scoring.pyx"
  ((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *)__pyx_v_self->__pyx_base.__pyx_vtab)->scale_theoretical_peaks(__pyx_v_self, __pyx_v_theoretical, (1. / __pyx_v_self->scale_factor));

  /* "ms_deisotope/_c/scoring.pyx":975
 *         self.scale_fitted_peaks(experimental, 1. / self.scale_factor)
 *         self.scale_theoretical_peaks(theoretical, 1. / self.scale_factor)
 *         score = self.scorer._evaluate(peaklist, experimental, theoretical)             # <<<<<<<<<<<<<<
 *         self.scale_fitted_peaks(experimental, self.scale_factor)
 *         self.scale_theoretical_peaks(theoretical, self.scale_factor)
 */

#line 975 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_score = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *)__pyx_v_self->scorer->__pyx_base.__pyx_vtab)->__pyx_base._evaluate(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)__pyx_v_self->scorer), __pyx_v_peaklist, __pyx_v_experimental, __pyx_v_theoretical, 0);

  /* "ms_deisotope/_c/scoring.pyx":976
 *         self.scale_theoretical_peaks(theoretical, 1. / self.scale_factor)
 *         score = self.scorer._evaluate(peaklist, experimental, theoretical)
 *         self.scale_fitted_peaks(experimental, self.scale_factor)             # <<<<<<<<<<<<<<
 *         self.scale_theoretical_peaks(theoretical, self.scale_factor)
 *         return score
 */

#line 976 "ms_deisotope/_c/scoring.pyx"
  ((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *)__pyx_v_self->__pyx_base.__pyx_vtab)->scale_fitted_peaks(__pyx_v_self, __pyx_v_experimental, __pyx_v_self->scale_factor);

  /* "ms_deisotope/_c/scoring.pyx":977
 *         score = self.scorer._evaluate(peaklist, experimental, theoretical)
 *         self.scale_fitted_peaks(experimental, self.scale_factor)
 *         self.scale_theoretical_peaks(theoretical, self.scale_factor)             # <<<<<<<<<<<<<<
 *         return score
 * 
 */

#line 977 "ms_deisotope/_c/scoring.pyx"
  ((struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *)__pyx_v_self->__pyx_base.__pyx_vtab)->scale_theoretical_peaks(__pyx_v_self, __pyx_v_theoretical, __pyx_v_self->scale_factor);

  /* "ms_deisotope/_c/scoring.pyx":978
 *         self.scale_fitted_peaks(experimental, self.scale_factor)
 *         self.scale_theoretical_peaks(theoretical, self.scale_factor)
 *         return score             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 978 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_v_score;

#line 978 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":968
 *         return self._evaluate(peaklist, observed, expected)
 * 
 *     cpdef double _evaluate(self, PeakSet peaklist, list experimental, list theoretical):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double score
 */

#line 968 "ms_deisotope/_c/scoring.pyx"


#line 968 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 968 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 968 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 968 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 968 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 968 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 968 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 968 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.ScaledPenalizedMSDeconvFitter._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 968 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 968 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 968 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 968 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 968 "ms_deisotope/_c/scoring.pyx"
}

#line 968 "ms_deisotope/_c/scoring.pyx"


#line 968 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 968 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_11_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 968 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_10_evaluate[] = "ScaledPenalizedMSDeconvFitter._evaluate(self, PeakSet peaklist, list experimental, list theoretical) -> double";

#line 968 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_11_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist = 0
#line 968 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_experimental = 0
#line 968 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_theoretical = 0
#line 968 "ms_deisotope/_c/scoring.pyx"
;

#line 968 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 968 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 968 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 968 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 968 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 968 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate (wrapper)", 0);

#line 968 "ms_deisotope/_c/scoring.pyx"
  {

#line 968 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peaklist,&__pyx_n_s_experimental,&__pyx_n_s_theoretical,0};

#line 968 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[3] = {0,0,0};

#line 968 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 968 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 968 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 968 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  3: 
#line 968 "ms_deisotope/_c/scoring.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 968 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 968 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 968 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 968 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 968 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 968 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 968 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 968 "ms_deisotope/_c/scoring.pyx"
      }

#line 968 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 968 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 968 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 968 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaklist)) != 0)) kw_args--;
        else 
#line 968 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 968 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 968 "ms_deisotope/_c/scoring.pyx"
        case  1:

#line 968 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_experimental)) != 0)) kw_args--;

#line 968 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, 1); 
#line 968 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 968, __pyx_L3_error)

#line 968 "ms_deisotope/_c/scoring.pyx"
        }

#line 968 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 968 "ms_deisotope/_c/scoring.pyx"
        case  2:

#line 968 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_theoretical)) != 0)) kw_args--;

#line 968 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, 2); 
#line 968 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 968, __pyx_L3_error)

#line 968 "ms_deisotope/_c/scoring.pyx"
        }

#line 968 "ms_deisotope/_c/scoring.pyx"
      }

#line 968 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 968 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_evaluate") < 0)) __PYX_ERR(0, 968, __pyx_L3_error)

#line 968 "ms_deisotope/_c/scoring.pyx"
      }

#line 968 "ms_deisotope/_c/scoring.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 968 "ms_deisotope/_c/scoring.pyx"
      goto __pyx_L5_argtuple_error;

#line 968 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 968 "ms_deisotope/_c/scoring.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 968 "ms_deisotope/_c/scoring.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 968 "ms_deisotope/_c/scoring.pyx"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 968 "ms_deisotope/_c/scoring.pyx"
    }

#line 968 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_peaklist = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[0]);

#line 968 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_experimental = ((PyObject*)values[1]);

#line 968 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_theoretical = ((PyObject*)values[2]);

#line 968 "ms_deisotope/_c/scoring.pyx"
  }

#line 968 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 968 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 968 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 968, __pyx_L3_error)

#line 968 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 968 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.ScaledPenalizedMSDeconvFitter._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 968 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 968 "ms_deisotope/_c/scoring.pyx"
  return NULL;

#line 968 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 968 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peaklist), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peaklist", 0))) __PYX_ERR(0, 968, __pyx_L1_error)

#line 968 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_experimental), (&PyList_Type), 1, "experimental", 1))) __PYX_ERR(0, 968, __pyx_L1_error)

#line 968 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_theoretical), (&PyList_Type), 1, "theoretical", 1))) __PYX_ERR(0, 968, __pyx_L1_error)
  __pyx_r = 
#line 968 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_10_evaluate(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *)__pyx_v_self), __pyx_v_peaklist, __pyx_v_experimental, __pyx_v_theoretical);

#line 968 "ms_deisotope/_c/scoring.pyx"


#line 968 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 968 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 968 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 968 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 968 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 968 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 968 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 968 "ms_deisotope/_c/scoring.pyx"
}

#line 968 "ms_deisotope/_c/scoring.pyx"


#line 968 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_10_evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_experimental, PyObject *__pyx_v_theoretical) {

#line 968 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 968 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 968 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 968 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 968 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 968 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 968 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate", 0);

#line 968 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 968 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter__evaluate(__pyx_v_self, __pyx_v_peaklist, __pyx_v_experimental, __pyx_v_theoretical, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 968, __pyx_L1_error)

#line 968 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 968 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 968 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 968 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 968 "ms_deisotope/_c/scoring.pyx"


#line 968 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 968 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 968 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 968 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.ScaledPenalizedMSDeconvFitter._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 968 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 968 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 968 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 968 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 968 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 968 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pxd":104
 * cdef class ScaledPenalizedMSDeconvFitter(IsotopicFitterBase):
 *     cdef:
 *         public double scale_factor             # <<<<<<<<<<<<<<
 *         public PenalizedMSDeconVFitter scorer
 * 
 */

#line 104 "ms_deisotope/_c/scoring.pxd"


#line 104 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 104 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_12scale_factor_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 104 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_12scale_factor_1__get__(PyObject *__pyx_v_self) {

#line 104 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = 0;

#line 104 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 104 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 104 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_12scale_factor___get__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *)__pyx_v_self));

#line 104 "ms_deisotope/_c/scoring.pxd"


#line 104 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 104 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 104 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 104 "ms_deisotope/_c/scoring.pxd"
}

#line 104 "ms_deisotope/_c/scoring.pxd"


#line 104 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_12scale_factor___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self) {

#line 104 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = NULL;

#line 104 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 104 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 104 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 104 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 104 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 104 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 104 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 104 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->scale_factor); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 104, __pyx_L1_error)

#line 104 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 104 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = __pyx_t_1;

#line 104 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = 0;

#line 104 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 104 "ms_deisotope/_c/scoring.pxd"


#line 104 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 104 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 104 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 104 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.ScaledPenalizedMSDeconvFitter.scale_factor.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 104 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = NULL;

#line 104 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 104 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 104 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 104 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 104 "ms_deisotope/_c/scoring.pxd"
}

#line 104 "ms_deisotope/_c/scoring.pxd"


#line 104 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 104 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_12scale_factor_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 104 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_12scale_factor_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 104 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 104 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 104 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 104 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_12scale_factor_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 104 "ms_deisotope/_c/scoring.pxd"


#line 104 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 104 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 104 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 104 "ms_deisotope/_c/scoring.pxd"
}

#line 104 "ms_deisotope/_c/scoring.pxd"


#line 104 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_12scale_factor_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self, PyObject *__pyx_v_value) {

#line 104 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 104 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 104 "ms_deisotope/_c/scoring.pxd"
  double __pyx_t_1;

#line 104 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 104 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 104 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 104 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 104 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 104, __pyx_L1_error)

#line 104 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->scale_factor = __pyx_t_1;

#line 104 "ms_deisotope/_c/scoring.pxd"


#line 104 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 104 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 104 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 104 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 104 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.ScaledPenalizedMSDeconvFitter.scale_factor.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 104 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = -1;

#line 104 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 104 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 104 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 104 "ms_deisotope/_c/scoring.pxd"
}

/* "ms_deisotope/_c/scoring.pxd":105
 *     cdef:
 *         public double scale_factor
 *         public PenalizedMSDeconVFitter scorer             # <<<<<<<<<<<<<<
 * 
 *     cpdef double _calculate_scale_factor(self, PeakSet peaklist)
 */

#line 105 "ms_deisotope/_c/scoring.pxd"


#line 105 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 105 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_6scorer_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 105 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_6scorer_1__get__(PyObject *__pyx_v_self) {

#line 105 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = 0;

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 105 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_6scorer___get__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *)__pyx_v_self));

#line 105 "ms_deisotope/_c/scoring.pxd"


#line 105 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 105 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 105 "ms_deisotope/_c/scoring.pxd"
}

#line 105 "ms_deisotope/_c/scoring.pxd"


#line 105 "ms_deisotope/_c/scoring.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_6scorer___get__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self) {

#line 105 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_r = NULL;

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->scorer));

#line 105 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = ((PyObject *)__pyx_v_self->scorer);

#line 105 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 105 "ms_deisotope/_c/scoring.pxd"


#line 105 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 105 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 105 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 105 "ms_deisotope/_c/scoring.pxd"
}

#line 105 "ms_deisotope/_c/scoring.pxd"


#line 105 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 105 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_6scorer_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 105 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_6scorer_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 105 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 105 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_6scorer_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 105 "ms_deisotope/_c/scoring.pxd"


#line 105 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 105 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 105 "ms_deisotope/_c/scoring.pxd"
}

#line 105 "ms_deisotope/_c/scoring.pxd"


#line 105 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_6scorer_2__set__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self, PyObject *__pyx_v_value) {

#line 105 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 105 "ms_deisotope/_c/scoring.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 105 "ms_deisotope/_c/scoring.pxd"
  int __pyx_lineno = 0;

#line 105 "ms_deisotope/_c/scoring.pxd"
  const char *__pyx_filename = NULL;

#line 105 "ms_deisotope/_c/scoring.pxd"
  int __pyx_clineno = 0;

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 105 "ms_deisotope/_c/scoring.pxd"
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter))))) __PYX_ERR(1, 105, __pyx_L1_error)

#line 105 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = __pyx_v_value;

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(__pyx_t_1);

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GIVEREF(__pyx_t_1);

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_v_self->scorer);

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->scorer));

#line 105 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->scorer = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *)__pyx_t_1);

#line 105 "ms_deisotope/_c/scoring.pxd"
  __pyx_t_1 = 0;

#line 105 "ms_deisotope/_c/scoring.pxd"


#line 105 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 105 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 105 "ms_deisotope/_c/scoring.pxd"
  goto __pyx_L0;

#line 105 "ms_deisotope/_c/scoring.pxd"
  __pyx_L1_error:;

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.ScaledPenalizedMSDeconvFitter.scorer.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 105 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = -1;

#line 105 "ms_deisotope/_c/scoring.pxd"
  __pyx_L0:;

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 105 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 105 "ms_deisotope/_c/scoring.pxd"
}

#line 105 "ms_deisotope/_c/scoring.pxd"


#line 105 "ms_deisotope/_c/scoring.pxd"
/* Python wrapper */

#line 105 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_6scorer_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 105 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_6scorer_5__del__(PyObject *__pyx_v_self) {

#line 105 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 105 "ms_deisotope/_c/scoring.pxd"
__pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_6scorer_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *)__pyx_v_self));

#line 105 "ms_deisotope/_c/scoring.pxd"


#line 105 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 105 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 105 "ms_deisotope/_c/scoring.pxd"
}

#line 105 "ms_deisotope/_c/scoring.pxd"


#line 105 "ms_deisotope/_c/scoring.pxd"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_6scorer_4__del__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_v_self) {

#line 105 "ms_deisotope/_c/scoring.pxd"
  int __pyx_r;

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyDeclarations

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_INCREF(Py_None);

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GIVEREF(Py_None);

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_GOTREF(__pyx_v_self->scorer);

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->scorer));

#line 105 "ms_deisotope/_c/scoring.pxd"
  __pyx_v_self->scorer = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *)Py_None);

#line 105 "ms_deisotope/_c/scoring.pxd"


#line 105 "ms_deisotope/_c/scoring.pxd"
  /* function exit code */

#line 105 "ms_deisotope/_c/scoring.pxd"
  __pyx_r = 0;

#line 105 "ms_deisotope/_c/scoring.pxd"
  __Pyx_RefNannyFinishContext();

#line 105 "ms_deisotope/_c/scoring.pxd"
  return __pyx_r;

#line 105 "ms_deisotope/_c/scoring.pxd"
}

/* "ms_deisotope/_c/scoring.pyx":983
 * cdef class DotProductFitter(IsotopicFitterBase):
 * 
 *     def __init__(self,  minimum_score=100):             # <<<<<<<<<<<<<<
 *         self.select = MaximizeFitSelector(minimum_score)
 * 
 */

#line 983 "ms_deisotope/_c/scoring.pyx"


#line 983 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 983 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_16DotProductFitter_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 983 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pw_12ms_deisotope_2_c_7scoring_16DotProductFitter_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_minimum_score = 0
#line 983 "ms_deisotope/_c/scoring.pyx"
;

#line 983 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 983 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 983 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 983 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 983 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 983 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 983 "ms_deisotope/_c/scoring.pyx"
  {

#line 983 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_minimum_score,0};

#line 983 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[1] = {0};

#line 983 "ms_deisotope/_c/scoring.pyx"
    values[0] = ((PyObject *)__pyx_int_100);

#line 983 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 983 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 983 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 983 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  1: 
#line 983 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 983 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 983 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 983 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 983 "ms_deisotope/_c/scoring.pyx"
      }

#line 983 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 983 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 983 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 983 "ms_deisotope/_c/scoring.pyx"
        if (kw_args > 0) {

#line 983 "ms_deisotope/_c/scoring.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_minimum_score);

#line 983 "ms_deisotope/_c/scoring.pyx"
          if (value) { values[0] = value; kw_args--; }

#line 983 "ms_deisotope/_c/scoring.pyx"
        }

#line 983 "ms_deisotope/_c/scoring.pyx"
      }

#line 983 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 983 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 983, __pyx_L3_error)

#line 983 "ms_deisotope/_c/scoring.pyx"
      }

#line 983 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 983 "ms_deisotope/_c/scoring.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: 
#line 983 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 983 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  0: 
#line 983 "ms_deisotope/_c/scoring.pyx"
break;
        default: 
#line 983 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 983 "ms_deisotope/_c/scoring.pyx"
      }

#line 983 "ms_deisotope/_c/scoring.pyx"
    }

#line 983 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_minimum_score = values[0];

#line 983 "ms_deisotope/_c/scoring.pyx"
  }

#line 983 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 983 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); 
#line 983 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 983, __pyx_L3_error)

#line 983 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 983 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.DotProductFitter.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 983 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 983 "ms_deisotope/_c/scoring.pyx"
  return -1;

#line 983 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 983 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_16DotProductFitter___init__(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter *)__pyx_v_self), __pyx_v_minimum_score);

#line 983 "ms_deisotope/_c/scoring.pyx"


#line 983 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 983 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 983 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 983 "ms_deisotope/_c/scoring.pyx"
}

#line 983 "ms_deisotope/_c/scoring.pyx"


#line 983 "ms_deisotope/_c/scoring.pyx"
static int __pyx_pf_12ms_deisotope_2_c_7scoring_16DotProductFitter___init__(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter *__pyx_v_self, PyObject *__pyx_v_minimum_score) {

#line 983 "ms_deisotope/_c/scoring.pyx"
  int __pyx_r;

#line 983 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 983 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 983 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 983 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 983 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 983 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_deisotope/_c/scoring.pyx":984
 * 
 *     def __init__(self,  minimum_score=100):
 *         self.select = MaximizeFitSelector(minimum_score)             # <<<<<<<<<<<<<<
 * 
 *     cpdef double _evaluate(self, PeakSet peaklist, list experimental, list theoretical):
 */

#line 984 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_7scoring_MaximizeFitSelector), __pyx_v_minimum_score); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 984, __pyx_L1_error)

#line 984 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 984 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 984 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.select);

#line 984 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx_base.select));

#line 984 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_self->__pyx_base.select = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)__pyx_t_1);

#line 984 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/scoring.pyx":983
 * cdef class DotProductFitter(IsotopicFitterBase):
 * 
 *     def __init__(self,  minimum_score=100):             # <<<<<<<<<<<<<<
 *         self.select = MaximizeFitSelector(minimum_score)
 * 
 */

#line 983 "ms_deisotope/_c/scoring.pyx"


#line 983 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 983 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 983 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 983 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 983 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 983 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.DotProductFitter.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 983 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = -1;

#line 983 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 983 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 983 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 983 "ms_deisotope/_c/scoring.pyx"
}

/* "ms_deisotope/_c/scoring.pyx":986
 *         self.select = MaximizeFitSelector(minimum_score)
 * 
 *     cpdef double _evaluate(self, PeakSet peaklist, list experimental, list theoretical):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

#line 986 "ms_deisotope/_c/scoring.pyx"


#line 986 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_16DotProductFitter_3_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 986 "ms_deisotope/_c/scoring.pyx"
static double __pyx_f_12ms_deisotope_2_c_7scoring_16DotProductFitter__evaluate(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_experimental, PyObject *__pyx_v_theoretical, int __pyx_skip_dispatch) {
  size_t __pyx_v_i
#line 986 "ms_deisotope/_c/scoring.pyx"
;
  size_t __pyx_v_n
#line 986 "ms_deisotope/_c/scoring.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_v_obs = 0
#line 986 "ms_deisotope/_c/scoring.pyx"
;
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_theo = 0
#line 986 "ms_deisotope/_c/scoring.pyx"
;
  double __pyx_v_score
#line 986 "ms_deisotope/_c/scoring.pyx"
;

#line 986 "ms_deisotope/_c/scoring.pyx"
  double __pyx_r;

#line 986 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 986 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 986 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 986 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 986 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 986 "ms_deisotope/_c/scoring.pyx"
  int __pyx_t_5;

#line 986 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 986 "ms_deisotope/_c/scoring.pyx"
  double __pyx_t_7;

#line 986 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_8;

#line 986 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_9;

#line 986 "ms_deisotope/_c/scoring.pyx"
  size_t __pyx_t_10;

#line 986 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_11;

#line 986 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 986 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 986 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 986 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate", 0);

#line 986 "ms_deisotope/_c/scoring.pyx"
  /* Check if called by wrapper */

#line 986 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 986 "ms_deisotope/_c/scoring.pyx"
  /* Check if overridden in Python */

#line 986 "ms_deisotope/_c/scoring.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 986 "ms_deisotope/_c/scoring.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 986 "ms_deisotope/_c/scoring.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 986 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 986 "ms_deisotope/_c/scoring.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 986 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 986 "ms_deisotope/_c/scoring.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_evaluate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 986, __pyx_L1_error)

#line 986 "ms_deisotope/_c/scoring.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 986 "ms_deisotope/_c/scoring.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_7scoring_16DotProductFitter_3_evaluate)) {

#line 986 "ms_deisotope/_c/scoring.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 986 "ms_deisotope/_c/scoring.pyx"
__pyx_t_4 = NULL;

#line 986 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_5 = 0;

#line 986 "ms_deisotope/_c/scoring.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 986 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 986 "ms_deisotope/_c/scoring.pyx"
          if (likely(__pyx_t_4)) {

#line 986 "ms_deisotope/_c/scoring.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 986 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 986 "ms_deisotope/_c/scoring.pyx"
            __Pyx_INCREF(function);

#line 986 "ms_deisotope/_c/scoring.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 986 "ms_deisotope/_c/scoring.pyx"
            __pyx_t_5 = 1;

#line 986 "ms_deisotope/_c/scoring.pyx"
          }

#line 986 "ms_deisotope/_c/scoring.pyx"
        }

#line 986 "ms_deisotope/_c/scoring.pyx"
        #if CYTHON_FAST_PYCALL

#line 986 "ms_deisotope/_c/scoring.pyx"
        if (PyFunction_Check(__pyx_t_3)) {

#line 986 "ms_deisotope/_c/scoring.pyx"
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_peaklist), __pyx_v_experimental, __pyx_v_theoretical};

#line 986 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 986, __pyx_L1_error)

#line 986 "ms_deisotope/_c/scoring.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 986 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 986 "ms_deisotope/_c/scoring.pyx"
        } else

#line 986 "ms_deisotope/_c/scoring.pyx"
        #endif

#line 986 "ms_deisotope/_c/scoring.pyx"
        #if CYTHON_FAST_PYCCALL

#line 986 "ms_deisotope/_c/scoring.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {

#line 986 "ms_deisotope/_c/scoring.pyx"
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_peaklist), __pyx_v_experimental, __pyx_v_theoretical};

#line 986 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 986, __pyx_L1_error)

#line 986 "ms_deisotope/_c/scoring.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 986 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 986 "ms_deisotope/_c/scoring.pyx"
        } else

#line 986 "ms_deisotope/_c/scoring.pyx"
        #endif

#line 986 "ms_deisotope/_c/scoring.pyx"
        {

#line 986 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_6 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 986, __pyx_L1_error)

#line 986 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_6);

#line 986 "ms_deisotope/_c/scoring.pyx"
          if (__pyx_t_4) {

#line 986 "ms_deisotope/_c/scoring.pyx"
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;

#line 986 "ms_deisotope/_c/scoring.pyx"
          }

#line 986 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(((PyObject *)__pyx_v_peaklist));

#line 986 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(((PyObject *)__pyx_v_peaklist));

#line 986 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_peaklist));

#line 986 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(__pyx_v_experimental);

#line 986 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(__pyx_v_experimental);

#line 986 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_experimental);

#line 986 "ms_deisotope/_c/scoring.pyx"
          __Pyx_INCREF(__pyx_v_theoretical);

#line 986 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GIVEREF(__pyx_v_theoretical);

#line 986 "ms_deisotope/_c/scoring.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, __pyx_v_theoretical);

#line 986 "ms_deisotope/_c/scoring.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 986, __pyx_L1_error)

#line 986 "ms_deisotope/_c/scoring.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 986 "ms_deisotope/_c/scoring.pyx"
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 986 "ms_deisotope/_c/scoring.pyx"
        }

#line 986 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 986 "ms_deisotope/_c/scoring.pyx"
        __pyx_t_7 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_7 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 986, __pyx_L1_error)

#line 986 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 986 "ms_deisotope/_c/scoring.pyx"
        __pyx_r = __pyx_t_7;

#line 986 "ms_deisotope/_c/scoring.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 986 "ms_deisotope/_c/scoring.pyx"
        goto __pyx_L0;

#line 986 "ms_deisotope/_c/scoring.pyx"
      }

#line 986 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 986 "ms_deisotope/_c/scoring.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 986 "ms_deisotope/_c/scoring.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 986 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 986 "ms_deisotope/_c/scoring.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 986 "ms_deisotope/_c/scoring.pyx"
      }

#line 986 "ms_deisotope/_c/scoring.pyx"
      #endif

#line 986 "ms_deisotope/_c/scoring.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 986 "ms_deisotope/_c/scoring.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 986 "ms_deisotope/_c/scoring.pyx"
    }

#line 986 "ms_deisotope/_c/scoring.pyx"
    #endif

#line 986 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":992
 *             TheoreticalPeak theo
 *             double score
 *         n = PyList_GET_SIZE(experimental)             # <<<<<<<<<<<<<<
 *         score = 0
 *         for i in range(n):
 */

#line 992 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_n = PyList_GET_SIZE(__pyx_v_experimental);

  /* "ms_deisotope/_c/scoring.pyx":993
 *             double score
 *         n = PyList_GET_SIZE(experimental)
 *         score = 0             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             obs = <FittedPeak>PyList_GET_ITEM(experimental, i)
 */

#line 993 "ms_deisotope/_c/scoring.pyx"
  __pyx_v_score = 0.0;

  /* "ms_deisotope/_c/scoring.pyx":994
 *         n = PyList_GET_SIZE(experimental)
 *         score = 0
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             obs = <FittedPeak>PyList_GET_ITEM(experimental, i)
 *             theo = <TheoreticalPeak>PyList_GET_ITEM(theoretical, i)
 */

#line 994 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_8 = __pyx_v_n;

#line 994 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_9 = __pyx_t_8;

#line 994 "ms_deisotope/_c/scoring.pyx"
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {

#line 994 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_i = __pyx_t_10;

    /* "ms_deisotope/_c/scoring.pyx":995
 *         score = 0
 *         for i in range(n):
 *             obs = <FittedPeak>PyList_GET_ITEM(experimental, i)             # <<<<<<<<<<<<<<
 *             theo = <TheoreticalPeak>PyList_GET_ITEM(theoretical, i)
 *             score += obs.intensity * theo.intensity
 */

#line 995 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_11 = PyList_GET_ITEM(__pyx_v_experimental, __pyx_v_i);

#line 995 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_11);

#line 995 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 995 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF_SET(__pyx_v_obs, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_1));

#line 995 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/scoring.pyx":996
 *         for i in range(n):
 *             obs = <FittedPeak>PyList_GET_ITEM(experimental, i)
 *             theo = <TheoreticalPeak>PyList_GET_ITEM(theoretical, i)             # <<<<<<<<<<<<<<
 *             score += obs.intensity * theo.intensity
 *         return score
 */

#line 996 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_11 = PyList_GET_ITEM(__pyx_v_theoretical, __pyx_v_i);

#line 996 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_11);

#line 996 "ms_deisotope/_c/scoring.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 996 "ms_deisotope/_c/scoring.pyx"
    __Pyx_XDECREF_SET(__pyx_v_theo, ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_1));

#line 996 "ms_deisotope/_c/scoring.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/scoring.pyx":997
 *             obs = <FittedPeak>PyList_GET_ITEM(experimental, i)
 *             theo = <TheoreticalPeak>PyList_GET_ITEM(theoretical, i)
 *             score += obs.intensity * theo.intensity             # <<<<<<<<<<<<<<
 *         return score
 */

#line 997 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_score = (__pyx_v_score + (__pyx_v_obs->__pyx_base.intensity * __pyx_v_theo->intensity));

#line 997 "ms_deisotope/_c/scoring.pyx"
  }

  /* "ms_deisotope/_c/scoring.pyx":998
 *             theo = <TheoreticalPeak>PyList_GET_ITEM(theoretical, i)
 *             score += obs.intensity * theo.intensity
 *         return score             # <<<<<<<<<<<<<<
 */

#line 998 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_v_score;

#line 998 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/scoring.pyx":986
 *         self.select = MaximizeFitSelector(minimum_score)
 * 
 *     cpdef double _evaluate(self, PeakSet peaklist, list experimental, list theoretical):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

#line 986 "ms_deisotope/_c/scoring.pyx"


#line 986 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 986 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 986 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 986 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 986 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 986 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 986 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 986 "ms_deisotope/_c/scoring.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.scoring.DotProductFitter._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 986 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = 0;

#line 986 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 986 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_obs);

#line 986 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_theo);

#line 986 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 986 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 986 "ms_deisotope/_c/scoring.pyx"
}

#line 986 "ms_deisotope/_c/scoring.pyx"


#line 986 "ms_deisotope/_c/scoring.pyx"
/* Python wrapper */

#line 986 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_16DotProductFitter_3_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 986 "ms_deisotope/_c/scoring.pyx"
static char __pyx_doc_12ms_deisotope_2_c_7scoring_16DotProductFitter_2_evaluate[] = "DotProductFitter._evaluate(self, PeakSet peaklist, list experimental, list theoretical) -> double";

#line 986 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_7scoring_16DotProductFitter_3_evaluate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist = 0
#line 986 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_experimental = 0
#line 986 "ms_deisotope/_c/scoring.pyx"
;
  PyObject *__pyx_v_theoretical = 0
#line 986 "ms_deisotope/_c/scoring.pyx"
;

#line 986 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 986 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 986 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 986 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = 0;

#line 986 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 986 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate (wrapper)", 0);

#line 986 "ms_deisotope/_c/scoring.pyx"
  {

#line 986 "ms_deisotope/_c/scoring.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peaklist,&__pyx_n_s_experimental,&__pyx_n_s_theoretical,0};

#line 986 "ms_deisotope/_c/scoring.pyx"
    PyObject* values[3] = {0,0,0};

#line 986 "ms_deisotope/_c/scoring.pyx"
    if (unlikely(__pyx_kwds)) {

#line 986 "ms_deisotope/_c/scoring.pyx"
      Py_ssize_t kw_args;

#line 986 "ms_deisotope/_c/scoring.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 986 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {
        case  3: 
#line 986 "ms_deisotope/_c/scoring.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 986 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 986 "ms_deisotope/_c/scoring.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 986 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 986 "ms_deisotope/_c/scoring.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 986 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 986 "ms_deisotope/_c/scoring.pyx"
        case  0: break;
        default: 
#line 986 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 986 "ms_deisotope/_c/scoring.pyx"
      }

#line 986 "ms_deisotope/_c/scoring.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 986 "ms_deisotope/_c/scoring.pyx"
      switch (pos_args) {

#line 986 "ms_deisotope/_c/scoring.pyx"
        case  0:

#line 986 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaklist)) != 0)) kw_args--;
        else 
#line 986 "ms_deisotope/_c/scoring.pyx"
goto __pyx_L5_argtuple_error;

#line 986 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 986 "ms_deisotope/_c/scoring.pyx"
        case  1:

#line 986 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_experimental)) != 0)) kw_args--;

#line 986 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, 1); 
#line 986 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 986, __pyx_L3_error)

#line 986 "ms_deisotope/_c/scoring.pyx"
        }

#line 986 "ms_deisotope/_c/scoring.pyx"
        CYTHON_FALLTHROUGH;

#line 986 "ms_deisotope/_c/scoring.pyx"
        case  2:

#line 986 "ms_deisotope/_c/scoring.pyx"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_theoretical)) != 0)) kw_args--;

#line 986 "ms_deisotope/_c/scoring.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, 2); 
#line 986 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 986, __pyx_L3_error)

#line 986 "ms_deisotope/_c/scoring.pyx"
        }

#line 986 "ms_deisotope/_c/scoring.pyx"
      }

#line 986 "ms_deisotope/_c/scoring.pyx"
      if (unlikely(kw_args > 0)) {

#line 986 "ms_deisotope/_c/scoring.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_evaluate") < 0)) __PYX_ERR(0, 986, __pyx_L3_error)

#line 986 "ms_deisotope/_c/scoring.pyx"
      }

#line 986 "ms_deisotope/_c/scoring.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 986 "ms_deisotope/_c/scoring.pyx"
      goto __pyx_L5_argtuple_error;

#line 986 "ms_deisotope/_c/scoring.pyx"
    } else {

#line 986 "ms_deisotope/_c/scoring.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 986 "ms_deisotope/_c/scoring.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 986 "ms_deisotope/_c/scoring.pyx"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 986 "ms_deisotope/_c/scoring.pyx"
    }

#line 986 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_peaklist = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)values[0]);

#line 986 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_experimental = ((PyObject*)values[1]);

#line 986 "ms_deisotope/_c/scoring.pyx"
    __pyx_v_theoretical = ((PyObject*)values[2]);

#line 986 "ms_deisotope/_c/scoring.pyx"
  }

#line 986 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 986 "ms_deisotope/_c/scoring.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_evaluate", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 986 "ms_deisotope/_c/scoring.pyx"
__PYX_ERR(0, 986, __pyx_L3_error)

#line 986 "ms_deisotope/_c/scoring.pyx"
  __pyx_L3_error:;

#line 986 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.DotProductFitter._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 986 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 986 "ms_deisotope/_c/scoring.pyx"
  return NULL;

#line 986 "ms_deisotope/_c/scoring.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 986 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peaklist), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet, 1, "peaklist", 0))) __PYX_ERR(0, 986, __pyx_L1_error)

#line 986 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_experimental), (&PyList_Type), 1, "experimental", 1))) __PYX_ERR(0, 986, __pyx_L1_error)

#line 986 "ms_deisotope/_c/scoring.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_theoretical), (&PyList_Type), 1, "theoretical", 1))) __PYX_ERR(0, 986, __pyx_L1_error)
  __pyx_r = 
#line 986 "ms_deisotope/_c/scoring.pyx"
__pyx_pf_12ms_deisotope_2_c_7scoring_16DotProductFitter_2_evaluate(((struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter *)__pyx_v_self), __pyx_v_peaklist, __pyx_v_experimental, __pyx_v_theoretical);

#line 986 "ms_deisotope/_c/scoring.pyx"


#line 986 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 986 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 986 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 986 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 986 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 986 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 986 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 986 "ms_deisotope/_c/scoring.pyx"
}

#line 986 "ms_deisotope/_c/scoring.pyx"


#line 986 "ms_deisotope/_c/scoring.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_7scoring_16DotProductFitter_2_evaluate(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_v_peaklist, PyObject *__pyx_v_experimental, PyObject *__pyx_v_theoretical) {

#line 986 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_r = NULL;

#line 986 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyDeclarations

#line 986 "ms_deisotope/_c/scoring.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 986 "ms_deisotope/_c/scoring.pyx"
  int __pyx_lineno = 0;

#line 986 "ms_deisotope/_c/scoring.pyx"
  const char *__pyx_filename = NULL;

#line 986 "ms_deisotope/_c/scoring.pyx"
  int __pyx_clineno = 0;

#line 986 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannySetupContext("_evaluate", 0);

#line 986 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 986 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_12ms_deisotope_2_c_7scoring_16DotProductFitter__evaluate(__pyx_v_self, __pyx_v_peaklist, __pyx_v_experimental, __pyx_v_theoretical, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 986, __pyx_L1_error)

#line 986 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 986 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = __pyx_t_1;

#line 986 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

#line 986 "ms_deisotope/_c/scoring.pyx"
  goto __pyx_L0;

#line 986 "ms_deisotope/_c/scoring.pyx"


#line 986 "ms_deisotope/_c/scoring.pyx"
  /* function exit code */

#line 986 "ms_deisotope/_c/scoring.pyx"
  __pyx_L1_error:;

#line 986 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 986 "ms_deisotope/_c/scoring.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.scoring.DotProductFitter._evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 986 "ms_deisotope/_c/scoring.pyx"
  __pyx_r = NULL;

#line 986 "ms_deisotope/_c/scoring.pyx"
  __pyx_L0:;

#line 986 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 986 "ms_deisotope/_c/scoring.pyx"
  __Pyx_RefNannyFinishContext();

#line 986 "ms_deisotope/_c/scoring.pyx"
  return __pyx_r;

#line 986 "ms_deisotope/_c/scoring.pyx"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":734
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":735
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 735, __pyx_L1_error)

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":734
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":737
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":738
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 738, __pyx_L1_error)

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":737
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":740
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":741
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 741, __pyx_L1_error)

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":740
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":743
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":744
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 744, __pyx_L1_error)

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":743
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":746
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":747
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 747, __pyx_L1_error)

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":746
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":749
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_1;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":750
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */

#line 750 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);

#line 750 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (__pyx_t_1) {

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":751
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */

#line 751 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_r);

#line 751 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));

#line 751 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);

#line 751 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L0;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":750
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */

#line 750 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":753
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /*else*/ {

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_r);

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_INCREF(__pyx_empty_tuple);

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_r = __pyx_empty_tuple;

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L0;

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":749
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":928
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":929
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */

#line 929 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_INCREF(__pyx_v_base);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":930
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */

#line 930 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":928
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":932
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base
#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_1;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":933
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */

#line 933 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":934
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */

#line 934 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);

#line 934 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (__pyx_t_1) {

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":935
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */

#line 935 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_r);

#line 935 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 935 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L0;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":934
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */

#line 934 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":936
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */

#line 936 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 936 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_v_base));

#line 936 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = ((PyObject *)__pyx_v_base);

#line 936 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":932
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":940
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_2 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_3 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_4;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_5 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_6 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_7 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_8 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":941
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_declare

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_assign

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_1);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_2);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_3);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    /*try:*/ {

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":942
 * cdef inline int import_array() except -1:
 *     try:
 *         __pyx_import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */

#line 942 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 942, __pyx_L3_error)

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":941
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L8_try_end;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L3_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":943
 *     try:
 *         __pyx_import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (__pyx_t_4) {

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 943, __pyx_L5_except_error)

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_5);

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_6);

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":944
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 944, __pyx_L5_except_error)

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_8);

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __PYX_ERR(2, 944, __pyx_L5_except_error)

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L5_except_error;

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L5_except_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":941
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_1);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L1_error;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L8_try_end:;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":940
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_5);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_6);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_7);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_8);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = -1;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":946
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_2 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_3 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_4;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_5 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_6 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_7 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_8 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":947
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_declare

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_assign

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_1);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_2);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_3);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    /*try:*/ {

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":948
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */

#line 948 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 948, __pyx_L3_error)

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":947
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L8_try_end;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L3_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":949
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (__pyx_t_4) {

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 949, __pyx_L5_except_error)

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_5);

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_6);

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":950
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 950, __pyx_L5_except_error)

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_8);

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __PYX_ERR(2, 950, __pyx_L5_except_error)

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L5_except_error;

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L5_except_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":947
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_1);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L1_error;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L8_try_end:;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":946
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_5);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_6);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_7);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_8);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = -1;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":952
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_2 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_3 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_4;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_5 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_6 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_7 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_8 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":953
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_declare

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_assign

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_1);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_2);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_3);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    /*try:*/ {

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":954
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */

#line 954 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 954, __pyx_L3_error)

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":953
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L8_try_end;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L3_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":955
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (__pyx_t_4) {

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 955, __pyx_L5_except_error)

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_5);

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_6);

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":956
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef extern from *:
 */

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 956, __pyx_L5_except_error)

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_8);

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __PYX_ERR(2, 956, __pyx_L5_except_error)

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L5_except_error;

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L5_except_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":953
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_1);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L1_error;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L8_try_end:;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":952
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_5);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_6);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_7);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_8);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = -1;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":966
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_is_timedelta64_object(PyObject *__pyx_v_obj) {

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("is_timedelta64_object", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":978
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyTimedeltaArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 978 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyTimedeltaArrType_Type));

#line 978 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":966
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":981
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_is_datetime64_object(PyObject *__pyx_v_obj) {

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("is_datetime64_object", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":993
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyDatetimeArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 993 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyDatetimeArrType_Type));

#line 993 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":981
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":996
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE npy_datetime __pyx_f_5numpy_get_datetime64_value(PyObject *__pyx_v_obj) {

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  npy_datetime __pyx_r;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1003
 *     also needed.  That can be found using `get_datetime64_unit`.
 *     """
 *     return (<PyDatetimeScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1003 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = ((PyDatetimeScalarObject *)__pyx_v_obj)->obval;

#line 1003 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":996
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1006
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE npy_timedelta __pyx_f_5numpy_get_timedelta64_value(PyObject *__pyx_v_obj) {

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  npy_timedelta __pyx_r;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1010
 *     returns the int64 value underlying scalar numpy timedelta64 object
 *     """
 *     return (<PyTimedeltaScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1010 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = ((PyTimedeltaScalarObject *)__pyx_v_obj)->obval;

#line 1010 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1006
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1013
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE NPY_DATETIMEUNIT __pyx_f_5numpy_get_datetime64_unit(PyObject *__pyx_v_obj) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  NPY_DATETIMEUNIT __pyx_r;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1017
 *     returns the unit part of the dtype for a numpy datetime64 object.
 *     """
 *     return <NPY_DATETIMEUNIT>(<PyDatetimeScalarObject*>obj).obmeta.base             # <<<<<<<<<<<<<<
 */

#line 1017 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = ((NPY_DATETIMEUNIT)((PyDatetimeScalarObject *)__pyx_v_obj)->obmeta.base);

#line 1017 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1013
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "View.MemoryView":123
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

#line 123 "View.MemoryView"


#line 123 "View.MemoryView"
/* Python wrapper */

#line 123 "View.MemoryView"
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 123 "View.MemoryView"
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_shape = 0
#line 123 "View.MemoryView"
;
  Py_ssize_t __pyx_v_itemsize
#line 123 "View.MemoryView"
;
  PyObject *__pyx_v_format = 0
#line 123 "View.MemoryView"
;
  PyObject *__pyx_v_mode = 0
#line 123 "View.MemoryView"
;
  int __pyx_v_allocate_buffer
#line 123 "View.MemoryView"
;

#line 123 "View.MemoryView"
  int __pyx_lineno = 0;

#line 123 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 123 "View.MemoryView"
  int __pyx_clineno = 0;

#line 123 "View.MemoryView"
  int __pyx_r;

#line 123 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 123 "View.MemoryView"
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);

#line 123 "View.MemoryView"
  {

#line 123 "View.MemoryView"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_shape,&__pyx_n_s_itemsize,&__pyx_n_s_format,&__pyx_n_s_mode,&__pyx_n_s_allocate_buffer,0};

#line 123 "View.MemoryView"
    PyObject* values[5] = {0,0,0,0,0};

#line 123 "View.MemoryView"
    values[3] = ((PyObject *)__pyx_n_s_c);

#line 123 "View.MemoryView"
    if (unlikely(__pyx_kwds)) {

#line 123 "View.MemoryView"
      Py_ssize_t kw_args;

#line 123 "View.MemoryView"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 123 "View.MemoryView"
      switch (pos_args) {
        case  5: 
#line 123 "View.MemoryView"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 123 "View.MemoryView"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 123 "View.MemoryView"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 123 "View.MemoryView"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 123 "View.MemoryView"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;

#line 123 "View.MemoryView"
        case  0: break;
        default: 
#line 123 "View.MemoryView"
goto __pyx_L5_argtuple_error;

#line 123 "View.MemoryView"
      }

#line 123 "View.MemoryView"
      kw_args = PyDict_Size(__pyx_kwds);

#line 123 "View.MemoryView"
      switch (pos_args) {

#line 123 "View.MemoryView"
        case  0:

#line 123 "View.MemoryView"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape)) != 0)) kw_args--;
        else 
#line 123 "View.MemoryView"
goto __pyx_L5_argtuple_error;

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;

#line 123 "View.MemoryView"
        case  1:

#line 123 "View.MemoryView"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_itemsize)) != 0)) kw_args--;

#line 123 "View.MemoryView"
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 1); 
#line 123 "View.MemoryView"
__PYX_ERR(3, 123, __pyx_L3_error)

#line 123 "View.MemoryView"
        }

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;

#line 123 "View.MemoryView"
        case  2:

#line 123 "View.MemoryView"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_format)) != 0)) kw_args--;

#line 123 "View.MemoryView"
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 2); 
#line 123 "View.MemoryView"
__PYX_ERR(3, 123, __pyx_L3_error)

#line 123 "View.MemoryView"
        }

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;

#line 123 "View.MemoryView"
        case  3:

#line 123 "View.MemoryView"
        if (kw_args > 0) {

#line 123 "View.MemoryView"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mode);

#line 123 "View.MemoryView"
          if (value) { values[3] = value; kw_args--; }

#line 123 "View.MemoryView"
        }

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;

#line 123 "View.MemoryView"
        case  4:

#line 123 "View.MemoryView"
        if (kw_args > 0) {

#line 123 "View.MemoryView"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_allocate_buffer);

#line 123 "View.MemoryView"
          if (value) { values[4] = value; kw_args--; }

#line 123 "View.MemoryView"
        }

#line 123 "View.MemoryView"
      }

#line 123 "View.MemoryView"
      if (unlikely(kw_args > 0)) {

#line 123 "View.MemoryView"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(3, 123, __pyx_L3_error)

#line 123 "View.MemoryView"
      }

#line 123 "View.MemoryView"
    } else {

#line 123 "View.MemoryView"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: 
#line 123 "View.MemoryView"
values[4] = PyTuple_GET_ITEM(__pyx_args, 4);

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;
        case  4: 
#line 123 "View.MemoryView"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 123 "View.MemoryView"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 123 "View.MemoryView"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 123 "View.MemoryView"
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 123 "View.MemoryView"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 123 "View.MemoryView"
        break;
        default: 
#line 123 "View.MemoryView"
goto __pyx_L5_argtuple_error;

#line 123 "View.MemoryView"
      }

#line 123 "View.MemoryView"
    }

#line 123 "View.MemoryView"
    __pyx_v_shape = ((PyObject*)values[0]);

#line 123 "View.MemoryView"
    __pyx_v_itemsize = __Pyx_PyIndex_AsSsize_t(values[1]); if (unlikely((__pyx_v_itemsize == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(3, 123, __pyx_L3_error)

#line 123 "View.MemoryView"
    __pyx_v_format = values[2];

#line 123 "View.MemoryView"
    __pyx_v_mode = values[3];

#line 123 "View.MemoryView"
    if (values[4]) {

#line 123 "View.MemoryView"
      __pyx_v_allocate_buffer = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_allocate_buffer == (int)-1) && PyErr_Occurred())) __PYX_ERR(3, 124, __pyx_L3_error)

#line 123 "View.MemoryView"
    } else {

      /* "View.MemoryView":124
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,
 *                   mode="c", bint allocate_buffer=True):             # <<<<<<<<<<<<<<
 * 
 *         cdef int idx
 */

#line 124 "View.MemoryView"
      __pyx_v_allocate_buffer = ((int)1);

#line 124 "View.MemoryView"
    }

#line 124 "View.MemoryView"
  }

#line 124 "View.MemoryView"
  goto __pyx_L4_argument_unpacking_done;

#line 124 "View.MemoryView"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, PyTuple_GET_SIZE(__pyx_args)); 
#line 124 "View.MemoryView"
__PYX_ERR(3, 123, __pyx_L3_error)

#line 124 "View.MemoryView"
  __pyx_L3_error:;

#line 124 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 124 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 124 "View.MemoryView"
  return -1;

#line 124 "View.MemoryView"
  __pyx_L4_argument_unpacking_done:;

#line 124 "View.MemoryView"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_shape), (&PyTuple_Type), 1, "shape", 1))) __PYX_ERR(3, 123, __pyx_L1_error)

#line 124 "View.MemoryView"
  if (unlikely(((PyObject *)__pyx_v_format) == Py_None)) {

#line 124 "View.MemoryView"
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "format"); __PYX_ERR(3, 123, __pyx_L1_error)

#line 124 "View.MemoryView"
  }
  __pyx_r = 
#line 124 "View.MemoryView"
__pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(((struct __pyx_array_obj *)__pyx_v_self), __pyx_v_shape, __pyx_v_itemsize, __pyx_v_format, __pyx_v_mode, __pyx_v_allocate_buffer);

  /* "View.MemoryView":123
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

#line 123 "View.MemoryView"


#line 123 "View.MemoryView"
  /* function exit code */

#line 123 "View.MemoryView"
  goto __pyx_L0;

#line 123 "View.MemoryView"
  __pyx_L1_error:;

#line 123 "View.MemoryView"
  __pyx_r = -1;

#line 123 "View.MemoryView"
  __pyx_L0:;

#line 123 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 123 "View.MemoryView"
  return __pyx_r;

#line 123 "View.MemoryView"
}

#line 123 "View.MemoryView"


#line 123 "View.MemoryView"
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer) {
  int __pyx_v_idx
#line 123 "View.MemoryView"
;
  Py_ssize_t __pyx_v_i
#line 123 "View.MemoryView"
;
  Py_ssize_t __pyx_v_dim
#line 123 "View.MemoryView"
;
  PyObject **__pyx_v_p
#line 123 "View.MemoryView"
;
  char __pyx_v_order
#line 123 "View.MemoryView"
;

#line 123 "View.MemoryView"
  int __pyx_r;

#line 123 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 123 "View.MemoryView"
  Py_ssize_t __pyx_t_1;

#line 123 "View.MemoryView"
  int __pyx_t_2;

#line 123 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 123 "View.MemoryView"
  int __pyx_t_4;

#line 123 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 123 "View.MemoryView"
  PyObject *__pyx_t_6 = NULL;

#line 123 "View.MemoryView"
  char *__pyx_t_7;

#line 123 "View.MemoryView"
  int __pyx_t_8;

#line 123 "View.MemoryView"
  Py_ssize_t __pyx_t_9;

#line 123 "View.MemoryView"
  PyObject *__pyx_t_10 = NULL;

#line 123 "View.MemoryView"
  Py_ssize_t __pyx_t_11;

#line 123 "View.MemoryView"
  int __pyx_lineno = 0;

#line 123 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 123 "View.MemoryView"
  int __pyx_clineno = 0;

#line 123 "View.MemoryView"
  __Pyx_RefNannySetupContext("__cinit__", 0);

#line 123 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_format);

  /* "View.MemoryView":130
 *         cdef PyObject **p
 * 
 *         self.ndim = <int> len(shape)             # <<<<<<<<<<<<<<
 *         self.itemsize = itemsize
 * 
 */

#line 130 "View.MemoryView"
  if (unlikely(__pyx_v_shape == Py_None)) {

#line 130 "View.MemoryView"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 130 "View.MemoryView"
    __PYX_ERR(3, 130, __pyx_L1_error)

#line 130 "View.MemoryView"
  }

#line 130 "View.MemoryView"
  __pyx_t_1 = PyTuple_GET_SIZE(__pyx_v_shape); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(3, 130, __pyx_L1_error)

#line 130 "View.MemoryView"
  __pyx_v_self->ndim = ((int)__pyx_t_1);

  /* "View.MemoryView":131
 * 
 *         self.ndim = <int> len(shape)
 *         self.itemsize = itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not self.ndim:
 */

#line 131 "View.MemoryView"
  __pyx_v_self->itemsize = __pyx_v_itemsize;

  /* "View.MemoryView":133
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 */

#line 133 "View.MemoryView"
  __pyx_t_2 = ((!(__pyx_v_self->ndim != 0)) != 0);

#line 133 "View.MemoryView"
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":134
 * 
 *         if not self.ndim:
 *             raise ValueError("Empty shape tuple for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if itemsize <= 0:
 */

#line 134 "View.MemoryView"
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 134, __pyx_L1_error)

#line 134 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 134 "View.MemoryView"
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);

#line 134 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 134 "View.MemoryView"
    __PYX_ERR(3, 134, __pyx_L1_error)

    /* "View.MemoryView":133
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 */

#line 133 "View.MemoryView"
  }

  /* "View.MemoryView":136
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 */

#line 136 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_itemsize <= 0) != 0);

#line 136 "View.MemoryView"
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":137
 * 
 *         if itemsize <= 0:
 *             raise ValueError("itemsize <= 0 for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(format, bytes):
 */

#line 137 "View.MemoryView"
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 137, __pyx_L1_error)

#line 137 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 137 "View.MemoryView"
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);

#line 137 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 137 "View.MemoryView"
    __PYX_ERR(3, 137, __pyx_L1_error)

    /* "View.MemoryView":136
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 */

#line 136 "View.MemoryView"
  }

  /* "View.MemoryView":139
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */

#line 139 "View.MemoryView"
  __pyx_t_2 = PyBytes_Check(__pyx_v_format); 

#line 139 "View.MemoryView"
  __pyx_t_4 = ((!(__pyx_t_2 != 0)) != 0);

#line 139 "View.MemoryView"
  if (__pyx_t_4) {

    /* "View.MemoryView":140
 * 
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')             # <<<<<<<<<<<<<<
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format
 */

#line 140 "View.MemoryView"
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_format, __pyx_n_s_encode); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 140, __pyx_L1_error)

#line 140 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_5);

#line 140 "View.MemoryView"
    __pyx_t_6 = NULL;

#line 140 "View.MemoryView"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {

#line 140 "View.MemoryView"
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);

#line 140 "View.MemoryView"
      if (likely(__pyx_t_6)) {

#line 140 "View.MemoryView"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);

#line 140 "View.MemoryView"
        __Pyx_INCREF(__pyx_t_6);

#line 140 "View.MemoryView"
        __Pyx_INCREF(function);

#line 140 "View.MemoryView"
        __Pyx_DECREF_SET(__pyx_t_5, function);

#line 140 "View.MemoryView"
      }

#line 140 "View.MemoryView"
    }

#line 140 "View.MemoryView"
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_n_s_ASCII) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_n_s_ASCII);

#line 140 "View.MemoryView"
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 140 "View.MemoryView"
    if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 140, __pyx_L1_error)

#line 140 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 140 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 140 "View.MemoryView"
    __Pyx_DECREF_SET(__pyx_v_format, __pyx_t_3);

#line 140 "View.MemoryView"
    __pyx_t_3 = 0;

    /* "View.MemoryView":139
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */

#line 139 "View.MemoryView"
  }

  /* "View.MemoryView":141
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string             # <<<<<<<<<<<<<<
 *         self.format = self._format
 * 
 */

#line 141 "View.MemoryView"
  if (!(likely(PyBytes_CheckExact(__pyx_v_format))||((__pyx_v_format) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_format)->tp_name), 0))) __PYX_ERR(3, 141, __pyx_L1_error)

#line 141 "View.MemoryView"
  __pyx_t_3 = __pyx_v_format;

#line 141 "View.MemoryView"
  __Pyx_INCREF(__pyx_t_3);

#line 141 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_3);

#line 141 "View.MemoryView"
  __Pyx_GOTREF(__pyx_v_self->_format);

#line 141 "View.MemoryView"
  __Pyx_DECREF(__pyx_v_self->_format);

#line 141 "View.MemoryView"
  __pyx_v_self->_format = ((PyObject*)__pyx_t_3);

#line 141 "View.MemoryView"
  __pyx_t_3 = 0;

  /* "View.MemoryView":142
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 142 "View.MemoryView"
  if (unlikely(__pyx_v_self->_format == Py_None)) {

#line 142 "View.MemoryView"
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");

#line 142 "View.MemoryView"
    __PYX_ERR(3, 142, __pyx_L1_error)

#line 142 "View.MemoryView"
  }

#line 142 "View.MemoryView"
  __pyx_t_7 = __Pyx_PyBytes_AsWritableString(__pyx_v_self->_format); if (unlikely((!__pyx_t_7) && PyErr_Occurred())) __PYX_ERR(3, 142, __pyx_L1_error)

#line 142 "View.MemoryView"
  __pyx_v_self->format = __pyx_t_7;

  /* "View.MemoryView":145
 * 
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)             # <<<<<<<<<<<<<<
 *         self._strides = self._shape + self.ndim
 * 
 */

#line 145 "View.MemoryView"
  __pyx_v_self->_shape = ((Py_ssize_t *)PyObject_Malloc((((sizeof(Py_ssize_t)) * __pyx_v_self->ndim) * 2)));

  /* "View.MemoryView":146
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)
 *         self._strides = self._shape + self.ndim             # <<<<<<<<<<<<<<
 * 
 *         if not self._shape:
 */

#line 146 "View.MemoryView"
  __pyx_v_self->_strides = (__pyx_v_self->_shape + __pyx_v_self->ndim);

  /* "View.MemoryView":148
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError("unable to allocate shape and strides.")
 * 
 */

#line 148 "View.MemoryView"
  __pyx_t_4 = ((!(__pyx_v_self->_shape != 0)) != 0);

#line 148 "View.MemoryView"
  if (unlikely(__pyx_t_4)) {

    /* "View.MemoryView":149
 * 
 *         if not self._shape:
 *             raise MemoryError("unable to allocate shape and strides.")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 149 "View.MemoryView"
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 149, __pyx_L1_error)

#line 149 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 149 "View.MemoryView"
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);

#line 149 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 149 "View.MemoryView"
    __PYX_ERR(3, 149, __pyx_L1_error)

    /* "View.MemoryView":148
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError("unable to allocate shape and strides.")
 * 
 */

#line 148 "View.MemoryView"
  }

  /* "View.MemoryView":152
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 */

#line 152 "View.MemoryView"
  __pyx_t_8 = 0;

#line 152 "View.MemoryView"
  __pyx_t_3 = __pyx_v_shape; __Pyx_INCREF(__pyx_t_3); __pyx_t_1 = 0;

#line 152 "View.MemoryView"
  for (;;) {

#line 152 "View.MemoryView"
    if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_3)) break;

#line 152 "View.MemoryView"
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 152 "View.MemoryView"
    __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_1); __Pyx_INCREF(__pyx_t_5); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(3, 152, __pyx_L1_error)

#line 152 "View.MemoryView"
    #else

#line 152 "View.MemoryView"
    __pyx_t_5 = PySequence_ITEM(__pyx_t_3, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 152, __pyx_L1_error)

#line 152 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_5);

#line 152 "View.MemoryView"
    #endif

#line 152 "View.MemoryView"
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_5); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(3, 152, __pyx_L1_error)

#line 152 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 152 "View.MemoryView"
    __pyx_v_dim = __pyx_t_9;

#line 152 "View.MemoryView"
    __pyx_v_idx = __pyx_t_8;

#line 152 "View.MemoryView"
    __pyx_t_8 = (__pyx_t_8 + 1);

    /* "View.MemoryView":153
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim
 */

#line 153 "View.MemoryView"
    __pyx_t_4 = ((__pyx_v_dim <= 0) != 0);

#line 153 "View.MemoryView"
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":154
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))             # <<<<<<<<<<<<<<
 *             self._shape[idx] = dim
 * 
 */

#line 154 "View.MemoryView"
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_idx); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 154, __pyx_L1_error)

#line 154 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_5);

#line 154 "View.MemoryView"
      __pyx_t_6 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 154, __pyx_L1_error)

#line 154 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_6);

#line 154 "View.MemoryView"
      __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(3, 154, __pyx_L1_error)

#line 154 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_10);

#line 154 "View.MemoryView"
      __Pyx_GIVEREF(__pyx_t_5);

#line 154 "View.MemoryView"
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_5);

#line 154 "View.MemoryView"
      __Pyx_GIVEREF(__pyx_t_6);

#line 154 "View.MemoryView"
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_6);

#line 154 "View.MemoryView"
      __pyx_t_5 = 0;

#line 154 "View.MemoryView"
      __pyx_t_6 = 0;

#line 154 "View.MemoryView"
      __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_Invalid_shape_in_axis_d_d, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 154, __pyx_L1_error)

#line 154 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_6);

#line 154 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 154 "View.MemoryView"
      __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(3, 154, __pyx_L1_error)

#line 154 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_10);

#line 154 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 154 "View.MemoryView"
      __Pyx_Raise(__pyx_t_10, 0, 0, 0);

#line 154 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 154 "View.MemoryView"
      __PYX_ERR(3, 154, __pyx_L1_error)

      /* "View.MemoryView":153
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim
 */

#line 153 "View.MemoryView"
    }

    /* "View.MemoryView":155
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim             # <<<<<<<<<<<<<<
 * 
 *         cdef char order
 */

#line 155 "View.MemoryView"
    (__pyx_v_self->_shape[__pyx_v_idx]) = __pyx_v_dim;

    /* "View.MemoryView":152
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 */

#line 152 "View.MemoryView"
  }

#line 152 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":158
 * 
 *         cdef char order
 *         if mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */

#line 158 "View.MemoryView"
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_fortran, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(3, 158, __pyx_L1_error)

#line 158 "View.MemoryView"
  if (__pyx_t_4) {

    /* "View.MemoryView":159
 *         cdef char order
 *         if mode == 'fortran':
 *             order = b'F'             # <<<<<<<<<<<<<<
 *             self.mode = u'fortran'
 *         elif mode == 'c':
 */

#line 159 "View.MemoryView"
    __pyx_v_order = 'F';

    /* "View.MemoryView":160
 *         if mode == 'fortran':
 *             order = b'F'
 *             self.mode = u'fortran'             # <<<<<<<<<<<<<<
 *         elif mode == 'c':
 *             order = b'C'
 */

#line 160 "View.MemoryView"
    __Pyx_INCREF(__pyx_n_u_fortran);

#line 160 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_n_u_fortran);

#line 160 "View.MemoryView"
    __Pyx_GOTREF(__pyx_v_self->mode);

#line 160 "View.MemoryView"
    __Pyx_DECREF(__pyx_v_self->mode);

#line 160 "View.MemoryView"
    __pyx_v_self->mode = __pyx_n_u_fortran;

    /* "View.MemoryView":158
 * 
 *         cdef char order
 *         if mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */

#line 158 "View.MemoryView"
    goto __pyx_L10;

#line 158 "View.MemoryView"
  }

  /* "View.MemoryView":161
 *             order = b'F'
 *             self.mode = u'fortran'
 *         elif mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */

#line 161 "View.MemoryView"
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_c, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(3, 161, __pyx_L1_error)

#line 161 "View.MemoryView"
  if (likely(__pyx_t_4)) {

    /* "View.MemoryView":162
 *             self.mode = u'fortran'
 *         elif mode == 'c':
 *             order = b'C'             # <<<<<<<<<<<<<<
 *             self.mode = u'c'
 *         else:
 */

#line 162 "View.MemoryView"
    __pyx_v_order = 'C';

    /* "View.MemoryView":163
 *         elif mode == 'c':
 *             order = b'C'
 *             self.mode = u'c'             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)
 */

#line 163 "View.MemoryView"
    __Pyx_INCREF(__pyx_n_u_c);

#line 163 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_n_u_c);

#line 163 "View.MemoryView"
    __Pyx_GOTREF(__pyx_v_self->mode);

#line 163 "View.MemoryView"
    __Pyx_DECREF(__pyx_v_self->mode);

#line 163 "View.MemoryView"
    __pyx_v_self->mode = __pyx_n_u_c;

    /* "View.MemoryView":161
 *             order = b'F'
 *             self.mode = u'fortran'
 *         elif mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */

#line 161 "View.MemoryView"
    goto __pyx_L10;

#line 161 "View.MemoryView"
  }

  /* "View.MemoryView":165
 *             self.mode = u'c'
 *         else:
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)             # <<<<<<<<<<<<<<
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides,
 */

#line 165 "View.MemoryView"
  /*else*/ {

#line 165 "View.MemoryView"
    __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Invalid_mode_expected_c_or_fortr, __pyx_v_mode); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 165, __pyx_L1_error)

#line 165 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 165 "View.MemoryView"
    __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(3, 165, __pyx_L1_error)

#line 165 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_10);

#line 165 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 165 "View.MemoryView"
    __Pyx_Raise(__pyx_t_10, 0, 0, 0);

#line 165 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 165 "View.MemoryView"
    __PYX_ERR(3, 165, __pyx_L1_error)

#line 165 "View.MemoryView"
  }

#line 165 "View.MemoryView"
  __pyx_L10:;

  /* "View.MemoryView":167
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides,             # <<<<<<<<<<<<<<
 *                                              itemsize, self.ndim, order)
 * 
 */

#line 167 "View.MemoryView"
  __pyx_v_self->len = __pyx_fill_contig_strides_array(__pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_itemsize, __pyx_v_self->ndim, __pyx_v_order);

  /* "View.MemoryView":170
 *                                              itemsize, self.ndim, order)
 * 
 *         self.free_data = allocate_buffer             # <<<<<<<<<<<<<<
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:
 */

#line 170 "View.MemoryView"
  __pyx_v_self->free_data = __pyx_v_allocate_buffer;

  /* "View.MemoryView":171
 * 
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'             # <<<<<<<<<<<<<<
 *         if allocate_buffer:
 * 
 */

#line 171 "View.MemoryView"
  __pyx_t_10 = PyObject_RichCompare(__pyx_v_format, __pyx_n_b_O, Py_EQ); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(3, 171, __pyx_L1_error)

#line 171 "View.MemoryView"
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(3, 171, __pyx_L1_error)

#line 171 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 171 "View.MemoryView"
  __pyx_v_self->dtype_is_object = __pyx_t_4;

  /* "View.MemoryView":172
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 172 "View.MemoryView"
  __pyx_t_4 = (__pyx_v_allocate_buffer != 0);

#line 172 "View.MemoryView"
  if (__pyx_t_4) {

    /* "View.MemoryView":175
 * 
 * 
 *             self.data = <char *>malloc(self.len)             # <<<<<<<<<<<<<<
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")
 */

#line 175 "View.MemoryView"
    __pyx_v_self->data = ((char *)malloc(__pyx_v_self->len));

    /* "View.MemoryView":176
 * 
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:             # <<<<<<<<<<<<<<
 *                 raise MemoryError("unable to allocate array data.")
 * 
 */

#line 176 "View.MemoryView"
    __pyx_t_4 = ((!(__pyx_v_self->data != 0)) != 0);

#line 176 "View.MemoryView"
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":177
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")             # <<<<<<<<<<<<<<
 * 
 *             if self.dtype_is_object:
 */

#line 177 "View.MemoryView"
      __pyx_t_10 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(3, 177, __pyx_L1_error)

#line 177 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_10);

#line 177 "View.MemoryView"
      __Pyx_Raise(__pyx_t_10, 0, 0, 0);

#line 177 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 177 "View.MemoryView"
      __PYX_ERR(3, 177, __pyx_L1_error)

      /* "View.MemoryView":176
 * 
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:             # <<<<<<<<<<<<<<
 *                 raise MemoryError("unable to allocate array data.")
 * 
 */

#line 176 "View.MemoryView"
    }

    /* "View.MemoryView":179
 *                 raise MemoryError("unable to allocate array data.")
 * 
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 */

#line 179 "View.MemoryView"
    __pyx_t_4 = (__pyx_v_self->dtype_is_object != 0);

#line 179 "View.MemoryView"
    if (__pyx_t_4) {

      /* "View.MemoryView":180
 * 
 *             if self.dtype_is_object:
 *                 p = <PyObject **> self.data             # <<<<<<<<<<<<<<
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None
 */

#line 180 "View.MemoryView"
      __pyx_v_p = ((PyObject **)__pyx_v_self->data);

      /* "View.MemoryView":181
 *             if self.dtype_is_object:
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):             # <<<<<<<<<<<<<<
 *                     p[i] = Py_None
 *                     Py_INCREF(Py_None)
 */

#line 181 "View.MemoryView"
      if (unlikely(__pyx_v_itemsize == 0)) {

#line 181 "View.MemoryView"
        PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");

#line 181 "View.MemoryView"
        __PYX_ERR(3, 181, __pyx_L1_error)

#line 181 "View.MemoryView"
      }

#line 181 "View.MemoryView"
      else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_itemsize == (Py_ssize_t)-1)  && unlikely(UNARY_NEG_WOULD_OVERFLOW(__pyx_v_self->len))) {

#line 181 "View.MemoryView"
        PyErr_SetString(PyExc_OverflowError, "value too large to perform division");

#line 181 "View.MemoryView"
        __PYX_ERR(3, 181, __pyx_L1_error)

#line 181 "View.MemoryView"
      }

#line 181 "View.MemoryView"
      __pyx_t_1 = __Pyx_div_Py_ssize_t(__pyx_v_self->len, __pyx_v_itemsize);

#line 181 "View.MemoryView"
      __pyx_t_9 = __pyx_t_1;

#line 181 "View.MemoryView"
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_9; __pyx_t_11+=1) {

#line 181 "View.MemoryView"
        __pyx_v_i = __pyx_t_11;

        /* "View.MemoryView":182
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None             # <<<<<<<<<<<<<<
 *                     Py_INCREF(Py_None)
 * 
 */

#line 182 "View.MemoryView"
        (__pyx_v_p[__pyx_v_i]) = Py_None;

        /* "View.MemoryView":183
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None
 *                     Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */

#line 183 "View.MemoryView"
        Py_INCREF(Py_None);

#line 183 "View.MemoryView"
      }

      /* "View.MemoryView":179
 *                 raise MemoryError("unable to allocate array data.")
 * 
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 */

#line 179 "View.MemoryView"
    }

    /* "View.MemoryView":172
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 172 "View.MemoryView"
  }

  /* "View.MemoryView":123
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

#line 123 "View.MemoryView"


#line 123 "View.MemoryView"
  /* function exit code */

#line 123 "View.MemoryView"
  __pyx_r = 0;

#line 123 "View.MemoryView"
  goto __pyx_L0;

#line 123 "View.MemoryView"
  __pyx_L1_error:;

#line 123 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 123 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 123 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_6);

#line 123 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_10);

#line 123 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 123 "View.MemoryView"
  __pyx_r = -1;

#line 123 "View.MemoryView"
  __pyx_L0:;

#line 123 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_format);

#line 123 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 123 "View.MemoryView"
  return __pyx_r;

#line 123 "View.MemoryView"
}

/* "View.MemoryView":186
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 */

#line 186 "View.MemoryView"


#line 186 "View.MemoryView"
/* Python wrapper */

#line 186 "View.MemoryView"
static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/

#line 186 "View.MemoryView"
static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {

#line 186 "View.MemoryView"
  int __pyx_r;

#line 186 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 186 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = 
#line 186 "View.MemoryView"
__pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(((struct __pyx_array_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

#line 186 "View.MemoryView"


#line 186 "View.MemoryView"
  /* function exit code */

#line 186 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 186 "View.MemoryView"
  return __pyx_r;

#line 186 "View.MemoryView"
}

#line 186 "View.MemoryView"


#line 186 "View.MemoryView"
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_bufmode
#line 186 "View.MemoryView"
;

#line 186 "View.MemoryView"
  int __pyx_r;

#line 186 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 186 "View.MemoryView"
  int __pyx_t_1;

#line 186 "View.MemoryView"
  int __pyx_t_2;

#line 186 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 186 "View.MemoryView"
  char *__pyx_t_4;

#line 186 "View.MemoryView"
  Py_ssize_t __pyx_t_5;

#line 186 "View.MemoryView"
  int __pyx_t_6;

#line 186 "View.MemoryView"
  Py_ssize_t *__pyx_t_7;

#line 186 "View.MemoryView"
  int __pyx_lineno = 0;

#line 186 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 186 "View.MemoryView"
  int __pyx_clineno = 0;

#line 186 "View.MemoryView"
  if (__pyx_v_info == NULL) {

#line 186 "View.MemoryView"
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");

#line 186 "View.MemoryView"
    return -1;

#line 186 "View.MemoryView"
  }

#line 186 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getbuffer__", 0);

#line 186 "View.MemoryView"
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);

#line 186 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "View.MemoryView":187
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1             # <<<<<<<<<<<<<<
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */

#line 187 "View.MemoryView"
  __pyx_v_bufmode = -1;

  /* "View.MemoryView":188
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if self.mode == u"c":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 */

#line 188 "View.MemoryView"
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_c, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(3, 188, __pyx_L1_error)

#line 188 "View.MemoryView"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 188 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":189
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */

#line 189 "View.MemoryView"
    __pyx_v_bufmode = (PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

    /* "View.MemoryView":188
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if self.mode == u"c":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 */

#line 188 "View.MemoryView"
    goto __pyx_L3;

#line 188 "View.MemoryView"
  }

  /* "View.MemoryView":190
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 */

#line 190 "View.MemoryView"
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_fortran, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(3, 190, __pyx_L1_error)

#line 190 "View.MemoryView"
  __pyx_t_1 = (__pyx_t_2 != 0);

#line 190 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":191
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 */

#line 191 "View.MemoryView"
    __pyx_v_bufmode = (PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

    /* "View.MemoryView":190
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 */

#line 190 "View.MemoryView"
  }

#line 190 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":192
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 */

#line 192 "View.MemoryView"
  __pyx_t_1 = ((!((__pyx_v_flags & __pyx_v_bufmode) != 0)) != 0);

#line 192 "View.MemoryView"
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":193
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")             # <<<<<<<<<<<<<<
 *         info.buf = self.data
 *         info.len = self.len
 */

#line 193 "View.MemoryView"
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 193, __pyx_L1_error)

#line 193 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 193 "View.MemoryView"
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);

#line 193 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 193 "View.MemoryView"
    __PYX_ERR(3, 193, __pyx_L1_error)

    /* "View.MemoryView":192
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 */

#line 192 "View.MemoryView"
  }

  /* "View.MemoryView":194
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data             # <<<<<<<<<<<<<<
 *         info.len = self.len
 *         info.ndim = self.ndim
 */

#line 194 "View.MemoryView"
  __pyx_t_4 = __pyx_v_self->data;

#line 194 "View.MemoryView"
  __pyx_v_info->buf = __pyx_t_4;

  /* "View.MemoryView":195
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 *         info.len = self.len             # <<<<<<<<<<<<<<
 *         info.ndim = self.ndim
 *         info.shape = self._shape
 */

#line 195 "View.MemoryView"
  __pyx_t_5 = __pyx_v_self->len;

#line 195 "View.MemoryView"
  __pyx_v_info->len = __pyx_t_5;

  /* "View.MemoryView":196
 *         info.buf = self.data
 *         info.len = self.len
 *         info.ndim = self.ndim             # <<<<<<<<<<<<<<
 *         info.shape = self._shape
 *         info.strides = self._strides
 */

#line 196 "View.MemoryView"
  __pyx_t_6 = __pyx_v_self->ndim;

#line 196 "View.MemoryView"
  __pyx_v_info->ndim = __pyx_t_6;

  /* "View.MemoryView":197
 *         info.len = self.len
 *         info.ndim = self.ndim
 *         info.shape = self._shape             # <<<<<<<<<<<<<<
 *         info.strides = self._strides
 *         info.suboffsets = NULL
 */

#line 197 "View.MemoryView"
  __pyx_t_7 = __pyx_v_self->_shape;

#line 197 "View.MemoryView"
  __pyx_v_info->shape = __pyx_t_7;

  /* "View.MemoryView":198
 *         info.ndim = self.ndim
 *         info.shape = self._shape
 *         info.strides = self._strides             # <<<<<<<<<<<<<<
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize
 */

#line 198 "View.MemoryView"
  __pyx_t_7 = __pyx_v_self->_strides;

#line 198 "View.MemoryView"
  __pyx_v_info->strides = __pyx_t_7;

  /* "View.MemoryView":199
 *         info.shape = self._shape
 *         info.strides = self._strides
 *         info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *         info.itemsize = self.itemsize
 *         info.readonly = 0
 */

#line 199 "View.MemoryView"
  __pyx_v_info->suboffsets = NULL;

  /* "View.MemoryView":200
 *         info.strides = self._strides
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize             # <<<<<<<<<<<<<<
 *         info.readonly = 0
 * 
 */

#line 200 "View.MemoryView"
  __pyx_t_5 = __pyx_v_self->itemsize;

#line 200 "View.MemoryView"
  __pyx_v_info->itemsize = __pyx_t_5;

  /* "View.MemoryView":201
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize
 *         info.readonly = 0             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */

#line 201 "View.MemoryView"
  __pyx_v_info->readonly = 0;

  /* "View.MemoryView":203
 *         info.readonly = 0
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.format
 *         else:
 */

#line 203 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);

#line 203 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":204
 * 
 *         if flags & PyBUF_FORMAT:
 *             info.format = self.format             # <<<<<<<<<<<<<<
 *         else:
 *             info.format = NULL
 */

#line 204 "View.MemoryView"
    __pyx_t_4 = __pyx_v_self->format;

#line 204 "View.MemoryView"
    __pyx_v_info->format = __pyx_t_4;

    /* "View.MemoryView":203
 *         info.readonly = 0
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.format
 *         else:
 */

#line 203 "View.MemoryView"
    goto __pyx_L5;

#line 203 "View.MemoryView"
  }

  /* "View.MemoryView":206
 *             info.format = self.format
 *         else:
 *             info.format = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.obj = self
 */

#line 206 "View.MemoryView"
  /*else*/ {

#line 206 "View.MemoryView"
    __pyx_v_info->format = NULL;

#line 206 "View.MemoryView"
  }

#line 206 "View.MemoryView"
  __pyx_L5:;

  /* "View.MemoryView":208
 *             info.format = NULL
 * 
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 */

#line 208 "View.MemoryView"
  __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 208 "View.MemoryView"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));

#line 208 "View.MemoryView"
  __Pyx_GOTREF(__pyx_v_info->obj);

#line 208 "View.MemoryView"
  __Pyx_DECREF(__pyx_v_info->obj);

#line 208 "View.MemoryView"
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":186
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 */

#line 186 "View.MemoryView"


#line 186 "View.MemoryView"
  /* function exit code */

#line 186 "View.MemoryView"
  __pyx_r = 0;

#line 186 "View.MemoryView"
  goto __pyx_L0;

#line 186 "View.MemoryView"
  __pyx_L1_error:;

#line 186 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 186 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.array.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 186 "View.MemoryView"
  __pyx_r = -1;

#line 186 "View.MemoryView"
  if (__pyx_v_info->obj != NULL) {

#line 186 "View.MemoryView"
    __Pyx_GOTREF(__pyx_v_info->obj);

#line 186 "View.MemoryView"
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;

#line 186 "View.MemoryView"
  }

#line 186 "View.MemoryView"
  goto __pyx_L2;

#line 186 "View.MemoryView"
  __pyx_L0:;

#line 186 "View.MemoryView"
  if (__pyx_v_info->obj == Py_None) {

#line 186 "View.MemoryView"
    __Pyx_GOTREF(__pyx_v_info->obj);

#line 186 "View.MemoryView"
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;

#line 186 "View.MemoryView"
  }

#line 186 "View.MemoryView"
  __pyx_L2:;

#line 186 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 186 "View.MemoryView"
  return __pyx_r;

#line 186 "View.MemoryView"
}

/* "View.MemoryView":212
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

#line 212 "View.MemoryView"


#line 212 "View.MemoryView"
/* Python wrapper */

#line 212 "View.MemoryView"
static void __pyx_array___dealloc__(PyObject *__pyx_v_self); /*proto*/

#line 212 "View.MemoryView"
static void __pyx_array___dealloc__(PyObject *__pyx_v_self) {

#line 212 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 212 "View.MemoryView"
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);

#line 212 "View.MemoryView"
  __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(((struct __pyx_array_obj *)__pyx_v_self));

#line 212 "View.MemoryView"


#line 212 "View.MemoryView"
  /* function exit code */

#line 212 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 212 "View.MemoryView"
}

#line 212 "View.MemoryView"


#line 212 "View.MemoryView"
static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self) {

#line 212 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 212 "View.MemoryView"
  int __pyx_t_1;

#line 212 "View.MemoryView"
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":213
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 */

#line 213 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_self->callback_free_data != NULL) != 0);

#line 213 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":214
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)             # <<<<<<<<<<<<<<
 *         elif self.free_data:
 *             if self.dtype_is_object:
 */

#line 214 "View.MemoryView"
    __pyx_v_self->callback_free_data(__pyx_v_self->data);

    /* "View.MemoryView":213
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 */

#line 213 "View.MemoryView"
    goto __pyx_L3;

#line 213 "View.MemoryView"
  }

  /* "View.MemoryView":215
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,
 */

#line 215 "View.MemoryView"
  __pyx_t_1 = (__pyx_v_self->free_data != 0);

#line 215 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":216
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 */

#line 216 "View.MemoryView"
    __pyx_t_1 = (__pyx_v_self->dtype_is_object != 0);

#line 216 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":217
 *         elif self.free_data:
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,             # <<<<<<<<<<<<<<
 *                                           self._strides, self.ndim, False)
 *             free(self.data)
 */

#line 217 "View.MemoryView"
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_self->data, __pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_self->ndim, 0);

      /* "View.MemoryView":216
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 */

#line 216 "View.MemoryView"
    }

    /* "View.MemoryView":219
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 *             free(self.data)             # <<<<<<<<<<<<<<
 *         PyObject_Free(self._shape)
 * 
 */

#line 219 "View.MemoryView"
    free(__pyx_v_self->data);

    /* "View.MemoryView":215
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,
 */

#line 215 "View.MemoryView"
  }

#line 215 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":220
 *                                           self._strides, self.ndim, False)
 *             free(self.data)
 *         PyObject_Free(self._shape)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 220 "View.MemoryView"
  PyObject_Free(__pyx_v_self->_shape);

  /* "View.MemoryView":212
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

#line 212 "View.MemoryView"


#line 212 "View.MemoryView"
  /* function exit code */

#line 212 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 212 "View.MemoryView"
}

/* "View.MemoryView":223
 * 
 *     @property
 *     def memview(self):             # <<<<<<<<<<<<<<
 *         return self.get_memview()
 * 
 */

#line 223 "View.MemoryView"


#line 223 "View.MemoryView"
/* Python wrapper */

#line 223 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 223 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self) {

#line 223 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 223 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 223 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 223 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(((struct __pyx_array_obj *)__pyx_v_self));

#line 223 "View.MemoryView"


#line 223 "View.MemoryView"
  /* function exit code */

#line 223 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 223 "View.MemoryView"
  return __pyx_r;

#line 223 "View.MemoryView"
}

#line 223 "View.MemoryView"


#line 223 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self) {

#line 223 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 223 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 223 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 223 "View.MemoryView"
  int __pyx_lineno = 0;

#line 223 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 223 "View.MemoryView"
  int __pyx_clineno = 0;

#line 223 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":224
 *     @property
 *     def memview(self):
 *         return self.get_memview()             # <<<<<<<<<<<<<<
 * 
 *     @cname('get_memview')
 */

#line 224 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 224 "View.MemoryView"
  __pyx_t_1 = ((struct __pyx_vtabstruct_array *)__pyx_v_self->__pyx_vtab)->get_memview(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 224, __pyx_L1_error)

#line 224 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 224 "View.MemoryView"
  __pyx_r = __pyx_t_1;

#line 224 "View.MemoryView"
  __pyx_t_1 = 0;

#line 224 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":223
 * 
 *     @property
 *     def memview(self):             # <<<<<<<<<<<<<<
 *         return self.get_memview()
 * 
 */

#line 223 "View.MemoryView"


#line 223 "View.MemoryView"
  /* function exit code */

#line 223 "View.MemoryView"
  __pyx_L1_error:;

#line 223 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 223 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.array.memview.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 223 "View.MemoryView"
  __pyx_r = NULL;

#line 223 "View.MemoryView"
  __pyx_L0:;

#line 223 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 223 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 223 "View.MemoryView"
  return __pyx_r;

#line 223 "View.MemoryView"
}

/* "View.MemoryView":227
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

#line 227 "View.MemoryView"


#line 227 "View.MemoryView"
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self) {
  int __pyx_v_flags
#line 227 "View.MemoryView"
;

#line 227 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 227 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 227 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 227 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 227 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 227 "View.MemoryView"
  int __pyx_lineno = 0;

#line 227 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 227 "View.MemoryView"
  int __pyx_clineno = 0;

#line 227 "View.MemoryView"
  __Pyx_RefNannySetupContext("get_memview", 0);

  /* "View.MemoryView":228
 *     @cname('get_memview')
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE             # <<<<<<<<<<<<<<
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 */

#line 228 "View.MemoryView"
  __pyx_v_flags = ((PyBUF_ANY_CONTIGUOUS | PyBUF_FORMAT) | PyBUF_WRITABLE);

  /* "View.MemoryView":229
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */

#line 229 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 229 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 229, __pyx_L1_error)

#line 229 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 229 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 229, __pyx_L1_error)

#line 229 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 229 "View.MemoryView"
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 229, __pyx_L1_error)

#line 229 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 229 "View.MemoryView"
  __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 229 "View.MemoryView"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));

#line 229 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self));

#line 229 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_1);

#line 229 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);

#line 229 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_2);

#line 229 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);

#line 229 "View.MemoryView"
  __pyx_t_1 = 0;

#line 229 "View.MemoryView"
  __pyx_t_2 = 0;

#line 229 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 229, __pyx_L1_error)

#line 229 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 229 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 229 "View.MemoryView"
  __pyx_r = __pyx_t_2;

#line 229 "View.MemoryView"
  __pyx_t_2 = 0;

#line 229 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":227
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

#line 227 "View.MemoryView"


#line 227 "View.MemoryView"
  /* function exit code */

#line 227 "View.MemoryView"
  __pyx_L1_error:;

#line 227 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 227 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 227 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 227 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.array.get_memview", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 227 "View.MemoryView"
  __pyx_r = 0;

#line 227 "View.MemoryView"
  __pyx_L0:;

#line 227 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 227 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 227 "View.MemoryView"
  return __pyx_r;

#line 227 "View.MemoryView"
}

/* "View.MemoryView":231
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

#line 231 "View.MemoryView"


#line 231 "View.MemoryView"
/* Python wrapper */

#line 231 "View.MemoryView"
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self); /*proto*/

#line 231 "View.MemoryView"
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self) {

#line 231 "View.MemoryView"
  Py_ssize_t __pyx_r;

#line 231 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 231 "View.MemoryView"
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = 
#line 231 "View.MemoryView"
__pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(((struct __pyx_array_obj *)__pyx_v_self));

#line 231 "View.MemoryView"


#line 231 "View.MemoryView"
  /* function exit code */

#line 231 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 231 "View.MemoryView"
  return __pyx_r;

#line 231 "View.MemoryView"
}

#line 231 "View.MemoryView"


#line 231 "View.MemoryView"
static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self) {

#line 231 "View.MemoryView"
  Py_ssize_t __pyx_r;

#line 231 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 231 "View.MemoryView"
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "View.MemoryView":232
 * 
 *     def __len__(self):
 *         return self._shape[0]             # <<<<<<<<<<<<<<
 * 
 *     def __getattr__(self, attr):
 */

#line 232 "View.MemoryView"
  __pyx_r = (__pyx_v_self->_shape[0]);

#line 232 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":231
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

#line 231 "View.MemoryView"


#line 231 "View.MemoryView"
  /* function exit code */

#line 231 "View.MemoryView"
  __pyx_L0:;

#line 231 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 231 "View.MemoryView"
  return __pyx_r;

#line 231 "View.MemoryView"
}

/* "View.MemoryView":234
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

#line 234 "View.MemoryView"


#line 234 "View.MemoryView"
/* Python wrapper */

#line 234 "View.MemoryView"
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr); /*proto*/

#line 234 "View.MemoryView"
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr) {

#line 234 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 234 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 234 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getattr__ (wrapper)", 0);
  __pyx_r = 
#line 234 "View.MemoryView"
__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_attr));

#line 234 "View.MemoryView"


#line 234 "View.MemoryView"
  /* function exit code */

#line 234 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 234 "View.MemoryView"
  return __pyx_r;

#line 234 "View.MemoryView"
}

#line 234 "View.MemoryView"


#line 234 "View.MemoryView"
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr) {

#line 234 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 234 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 234 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 234 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 234 "View.MemoryView"
  int __pyx_lineno = 0;

#line 234 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 234 "View.MemoryView"
  int __pyx_clineno = 0;

#line 234 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getattr__", 0);

  /* "View.MemoryView":235
 * 
 *     def __getattr__(self, attr):
 *         return getattr(self.memview, attr)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, item):
 */

#line 235 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 235 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 235, __pyx_L1_error)

#line 235 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 235 "View.MemoryView"
  __pyx_t_2 = __Pyx_GetAttr(__pyx_t_1, __pyx_v_attr); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 235, __pyx_L1_error)

#line 235 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 235 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 235 "View.MemoryView"
  __pyx_r = __pyx_t_2;

#line 235 "View.MemoryView"
  __pyx_t_2 = 0;

#line 235 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":234
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

#line 234 "View.MemoryView"


#line 234 "View.MemoryView"
  /* function exit code */

#line 234 "View.MemoryView"
  __pyx_L1_error:;

#line 234 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 234 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 234 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.array.__getattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 234 "View.MemoryView"
  __pyx_r = NULL;

#line 234 "View.MemoryView"
  __pyx_L0:;

#line 234 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 234 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 234 "View.MemoryView"
  return __pyx_r;

#line 234 "View.MemoryView"
}

/* "View.MemoryView":237
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

#line 237 "View.MemoryView"


#line 237 "View.MemoryView"
/* Python wrapper */

#line 237 "View.MemoryView"
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item); /*proto*/

#line 237 "View.MemoryView"
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item) {

#line 237 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 237 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 237 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = 
#line 237 "View.MemoryView"
__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item));

#line 237 "View.MemoryView"


#line 237 "View.MemoryView"
  /* function exit code */

#line 237 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 237 "View.MemoryView"
  return __pyx_r;

#line 237 "View.MemoryView"
}

#line 237 "View.MemoryView"


#line 237 "View.MemoryView"
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item) {

#line 237 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 237 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 237 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 237 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 237 "View.MemoryView"
  int __pyx_lineno = 0;

#line 237 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 237 "View.MemoryView"
  int __pyx_clineno = 0;

#line 237 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "View.MemoryView":238
 * 
 *     def __getitem__(self, item):
 *         return self.memview[item]             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, item, value):
 */

#line 238 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 238 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 238, __pyx_L1_error)

#line 238 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 238 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_item); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 238, __pyx_L1_error)

#line 238 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 238 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 238 "View.MemoryView"
  __pyx_r = __pyx_t_2;

#line 238 "View.MemoryView"
  __pyx_t_2 = 0;

#line 238 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":237
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

#line 237 "View.MemoryView"


#line 237 "View.MemoryView"
  /* function exit code */

#line 237 "View.MemoryView"
  __pyx_L1_error:;

#line 237 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 237 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 237 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.array.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 237 "View.MemoryView"
  __pyx_r = NULL;

#line 237 "View.MemoryView"
  __pyx_L0:;

#line 237 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 237 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 237 "View.MemoryView"
  return __pyx_r;

#line 237 "View.MemoryView"
}

/* "View.MemoryView":240
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

#line 240 "View.MemoryView"


#line 240 "View.MemoryView"
/* Python wrapper */

#line 240 "View.MemoryView"
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /*proto*/

#line 240 "View.MemoryView"
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {

#line 240 "View.MemoryView"
  int __pyx_r;

#line 240 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 240 "View.MemoryView"
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = 
#line 240 "View.MemoryView"
__pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item), ((PyObject *)__pyx_v_value));

#line 240 "View.MemoryView"


#line 240 "View.MemoryView"
  /* function exit code */

#line 240 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 240 "View.MemoryView"
  return __pyx_r;

#line 240 "View.MemoryView"
}

#line 240 "View.MemoryView"


#line 240 "View.MemoryView"
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {

#line 240 "View.MemoryView"
  int __pyx_r;

#line 240 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 240 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 240 "View.MemoryView"
  int __pyx_lineno = 0;

#line 240 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 240 "View.MemoryView"
  int __pyx_clineno = 0;

#line 240 "View.MemoryView"
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "View.MemoryView":241
 * 
 *     def __setitem__(self, item, value):
 *         self.memview[item] = value             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 241 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 241, __pyx_L1_error)

#line 241 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 241 "View.MemoryView"
  if (unlikely(PyObject_SetItem(__pyx_t_1, __pyx_v_item, __pyx_v_value) < 0)) __PYX_ERR(3, 241, __pyx_L1_error)

#line 241 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "View.MemoryView":240
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

#line 240 "View.MemoryView"


#line 240 "View.MemoryView"
  /* function exit code */

#line 240 "View.MemoryView"
  __pyx_r = 0;

#line 240 "View.MemoryView"
  goto __pyx_L0;

#line 240 "View.MemoryView"
  __pyx_L1_error:;

#line 240 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 240 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.array.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 240 "View.MemoryView"
  __pyx_r = -1;

#line 240 "View.MemoryView"
  __pyx_L0:;

#line 240 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 240 "View.MemoryView"
  return __pyx_r;

#line 240 "View.MemoryView"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf___pyx_array___reduce_cython__(((struct __pyx_array_obj *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 2 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 2, __pyx_L1_error)

#line 2 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 2 "(tree fragment)"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 2 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 2 "(tree fragment)"
  __PYX_ERR(3, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView.array.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
/* Python wrapper */

#line 3 "(tree fragment)"
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 3 "(tree fragment)"
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 3 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 3 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 3 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 3 "(tree fragment)"
__pyx_pf___pyx_array_2__setstate_cython__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
  /* function exit code */

#line 3 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 3 "(tree fragment)"
  return __pyx_r;

#line 3 "(tree fragment)"
}

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {

#line 3 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 3 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 3 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 3 "(tree fragment)"
  int __pyx_lineno = 0;

#line 3 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 3 "(tree fragment)"
  int __pyx_clineno = 0;

#line 3 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __PYX_ERR(3, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
  /* function exit code */

#line 3 "(tree fragment)"
  __pyx_L1_error:;

#line 3 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 3 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView.array.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 3 "(tree fragment)"
  __pyx_r = NULL;

#line 3 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 3 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 3 "(tree fragment)"
  return __pyx_r;

#line 3 "(tree fragment)"
}

/* "View.MemoryView":245
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format,             # <<<<<<<<<<<<<<
 *                           char *mode, char *buf):
 *     cdef array result
 */

#line 245 "View.MemoryView"


#line 245 "View.MemoryView"
static struct __pyx_array_obj *__pyx_array_new(PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, char *__pyx_v_format, char *__pyx_v_mode, char *__pyx_v_buf) {
  struct __pyx_array_obj *__pyx_v_result = 0
#line 245 "View.MemoryView"
;

#line 245 "View.MemoryView"
  struct __pyx_array_obj *__pyx_r = NULL;

#line 245 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 245 "View.MemoryView"
  int __pyx_t_1;

#line 245 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 245 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 245 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 245 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 245 "View.MemoryView"
  int __pyx_lineno = 0;

#line 245 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 245 "View.MemoryView"
  int __pyx_clineno = 0;

#line 245 "View.MemoryView"
  __Pyx_RefNannySetupContext("array_cwrapper", 0);

  /* "View.MemoryView":249
 *     cdef array result
 * 
 *     if buf == NULL:             # <<<<<<<<<<<<<<
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 */

#line 249 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_buf == NULL) != 0);

#line 249 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":250
 * 
 *     if buf == NULL:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))             # <<<<<<<<<<<<<<
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 */

#line 250 "View.MemoryView"
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 250, __pyx_L1_error)

#line 250 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 250 "View.MemoryView"
    __pyx_t_3 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 250, __pyx_L1_error)

#line 250 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 250 "View.MemoryView"
    __pyx_t_4 = __Pyx_decode_c_string(__pyx_v_mode, 0, strlen(__pyx_v_mode), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 250, __pyx_L1_error)

#line 250 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 250 "View.MemoryView"
    __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 250, __pyx_L1_error)

#line 250 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_5);

#line 250 "View.MemoryView"
    __Pyx_INCREF(__pyx_v_shape);

#line 250 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_v_shape);

#line 250 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_shape);

#line 250 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_t_2);

#line 250 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);

#line 250 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_t_3);

#line 250 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);

#line 250 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_t_4);

#line 250 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);

#line 250 "View.MemoryView"
    __pyx_t_2 = 0;

#line 250 "View.MemoryView"
    __pyx_t_3 = 0;

#line 250 "View.MemoryView"
    __pyx_t_4 = 0;

#line 250 "View.MemoryView"
    __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_array_type), __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 250, __pyx_L1_error)

#line 250 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 250 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 250 "View.MemoryView"
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_4);

#line 250 "View.MemoryView"
    __pyx_t_4 = 0;

    /* "View.MemoryView":249
 *     cdef array result
 * 
 *     if buf == NULL:             # <<<<<<<<<<<<<<
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 */

#line 249 "View.MemoryView"
    goto __pyx_L3;

#line 249 "View.MemoryView"
  }

  /* "View.MemoryView":252
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),             # <<<<<<<<<<<<<<
 *                        allocate_buffer=False)
 *         result.data = buf
 */

#line 252 "View.MemoryView"
  /*else*/ {

#line 252 "View.MemoryView"
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 252, __pyx_L1_error)

#line 252 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 252 "View.MemoryView"
    __pyx_t_5 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 252, __pyx_L1_error)

#line 252 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_5);

#line 252 "View.MemoryView"
    __pyx_t_3 = __Pyx_decode_c_string(__pyx_v_mode, 0, strlen(__pyx_v_mode), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 252, __pyx_L1_error)

#line 252 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 252 "View.MemoryView"
    __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 252, __pyx_L1_error)

#line 252 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 252 "View.MemoryView"
    __Pyx_INCREF(__pyx_v_shape);

#line 252 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_v_shape);

#line 252 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_shape);

#line 252 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_t_4);

#line 252 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);

#line 252 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_t_5);

#line 252 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_5);

#line 252 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_t_3);

#line 252 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_3);

#line 252 "View.MemoryView"
    __pyx_t_4 = 0;

#line 252 "View.MemoryView"
    __pyx_t_5 = 0;

#line 252 "View.MemoryView"
    __pyx_t_3 = 0;

    /* "View.MemoryView":253
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 *                        allocate_buffer=False)             # <<<<<<<<<<<<<<
 *         result.data = buf
 * 
 */

#line 253 "View.MemoryView"
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 253, __pyx_L1_error)

#line 253 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 253 "View.MemoryView"
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_allocate_buffer, Py_False) < 0) __PYX_ERR(3, 253, __pyx_L1_error)

    /* "View.MemoryView":252
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),             # <<<<<<<<<<<<<<
 *                        allocate_buffer=False)
 *         result.data = buf
 */

#line 252 "View.MemoryView"
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_array_type), __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 252, __pyx_L1_error)

#line 252 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_5);

#line 252 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 252 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 252 "View.MemoryView"
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_5);

#line 252 "View.MemoryView"
    __pyx_t_5 = 0;

    /* "View.MemoryView":254
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 *                        allocate_buffer=False)
 *         result.data = buf             # <<<<<<<<<<<<<<
 * 
 *     return result
 */

#line 254 "View.MemoryView"
    __pyx_v_result->data = __pyx_v_buf;

#line 254 "View.MemoryView"
  }

#line 254 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":256
 *         result.data = buf
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 256 "View.MemoryView"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 256 "View.MemoryView"
  __Pyx_INCREF(((PyObject *)__pyx_v_result));

#line 256 "View.MemoryView"
  __pyx_r = __pyx_v_result;

#line 256 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":245
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format,             # <<<<<<<<<<<<<<
 *                           char *mode, char *buf):
 *     cdef array result
 */

#line 245 "View.MemoryView"


#line 245 "View.MemoryView"
  /* function exit code */

#line 245 "View.MemoryView"
  __pyx_L1_error:;

#line 245 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 245 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 245 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_4);

#line 245 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 245 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.array_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 245 "View.MemoryView"
  __pyx_r = 0;

#line 245 "View.MemoryView"
  __pyx_L0:;

#line 245 "View.MemoryView"
  __Pyx_XDECREF((PyObject *)__pyx_v_result);

#line 245 "View.MemoryView"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 245 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 245 "View.MemoryView"
  return __pyx_r;

#line 245 "View.MemoryView"
}

/* "View.MemoryView":282
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

#line 282 "View.MemoryView"


#line 282 "View.MemoryView"
/* Python wrapper */

#line 282 "View.MemoryView"
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 282 "View.MemoryView"
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0
#line 282 "View.MemoryView"
;

#line 282 "View.MemoryView"
  int __pyx_lineno = 0;

#line 282 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 282 "View.MemoryView"
  int __pyx_clineno = 0;

#line 282 "View.MemoryView"
  int __pyx_r;

#line 282 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 282 "View.MemoryView"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 282 "View.MemoryView"
  {

#line 282 "View.MemoryView"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,0};

#line 282 "View.MemoryView"
    PyObject* values[1] = {0};

#line 282 "View.MemoryView"
    if (unlikely(__pyx_kwds)) {

#line 282 "View.MemoryView"
      Py_ssize_t kw_args;

#line 282 "View.MemoryView"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 282 "View.MemoryView"
      switch (pos_args) {
        case  1: 
#line 282 "View.MemoryView"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 282 "View.MemoryView"
        CYTHON_FALLTHROUGH;

#line 282 "View.MemoryView"
        case  0: break;
        default: 
#line 282 "View.MemoryView"
goto __pyx_L5_argtuple_error;

#line 282 "View.MemoryView"
      }

#line 282 "View.MemoryView"
      kw_args = PyDict_Size(__pyx_kwds);

#line 282 "View.MemoryView"
      switch (pos_args) {

#line 282 "View.MemoryView"
        case  0:

#line 282 "View.MemoryView"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else 
#line 282 "View.MemoryView"
goto __pyx_L5_argtuple_error;

#line 282 "View.MemoryView"
      }

#line 282 "View.MemoryView"
      if (unlikely(kw_args > 0)) {

#line 282 "View.MemoryView"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(3, 282, __pyx_L3_error)

#line 282 "View.MemoryView"
      }

#line 282 "View.MemoryView"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {

#line 282 "View.MemoryView"
      goto __pyx_L5_argtuple_error;

#line 282 "View.MemoryView"
    } else {

#line 282 "View.MemoryView"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 282 "View.MemoryView"
    }

#line 282 "View.MemoryView"
    __pyx_v_name = values[0];

#line 282 "View.MemoryView"
  }

#line 282 "View.MemoryView"
  goto __pyx_L4_argument_unpacking_done;

#line 282 "View.MemoryView"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); 
#line 282 "View.MemoryView"
__PYX_ERR(3, 282, __pyx_L3_error)

#line 282 "View.MemoryView"
  __pyx_L3_error:;

#line 282 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.Enum.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 282 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 282 "View.MemoryView"
  return -1;

#line 282 "View.MemoryView"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 282 "View.MemoryView"
__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), __pyx_v_name);

#line 282 "View.MemoryView"


#line 282 "View.MemoryView"
  /* function exit code */

#line 282 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 282 "View.MemoryView"
  return __pyx_r;

#line 282 "View.MemoryView"
}

#line 282 "View.MemoryView"


#line 282 "View.MemoryView"
static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name) {

#line 282 "View.MemoryView"
  int __pyx_r;

#line 282 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 282 "View.MemoryView"
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "View.MemoryView":283
 *     cdef object name
 *     def __init__(self, name):
 *         self.name = name             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return self.name
 */

#line 283 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_name);

#line 283 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_v_name);

#line 283 "View.MemoryView"
  __Pyx_GOTREF(__pyx_v_self->name);

#line 283 "View.MemoryView"
  __Pyx_DECREF(__pyx_v_self->name);

#line 283 "View.MemoryView"
  __pyx_v_self->name = __pyx_v_name;

  /* "View.MemoryView":282
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

#line 282 "View.MemoryView"


#line 282 "View.MemoryView"
  /* function exit code */

#line 282 "View.MemoryView"
  __pyx_r = 0;

#line 282 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 282 "View.MemoryView"
  return __pyx_r;

#line 282 "View.MemoryView"
}

/* "View.MemoryView":284
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

#line 284 "View.MemoryView"


#line 284 "View.MemoryView"
/* Python wrapper */

#line 284 "View.MemoryView"
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self); /*proto*/

#line 284 "View.MemoryView"
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self) {

#line 284 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 284 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 284 "View.MemoryView"
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = 
#line 284 "View.MemoryView"
__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

#line 284 "View.MemoryView"


#line 284 "View.MemoryView"
  /* function exit code */

#line 284 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 284 "View.MemoryView"
  return __pyx_r;

#line 284 "View.MemoryView"
}

#line 284 "View.MemoryView"


#line 284 "View.MemoryView"
static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {

#line 284 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 284 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 284 "View.MemoryView"
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "View.MemoryView":285
 *         self.name = name
 *     def __repr__(self):
 *         return self.name             # <<<<<<<<<<<<<<
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 */

#line 285 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 285 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_self->name);

#line 285 "View.MemoryView"
  __pyx_r = __pyx_v_self->name;

#line 285 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":284
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

#line 284 "View.MemoryView"


#line 284 "View.MemoryView"
  /* function exit code */

#line 284 "View.MemoryView"
  __pyx_L0:;

#line 284 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 284 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 284 "View.MemoryView"
  return __pyx_r;

#line 284 "View.MemoryView"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf___pyx_MemviewEnum___reduce_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {
  PyObject *__pyx_v_state = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v__dict = 0
#line 1 "(tree fragment)"
;
  int __pyx_v_use_setstate
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_2;

#line 1 "(tree fragment)"
  int __pyx_t_3;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.name,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */

#line 5 "(tree fragment)"
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
  __Pyx_INCREF(__pyx_v_self->name);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_v_self->name);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->name);

#line 5 "(tree fragment)"
  __pyx_v_state = ((PyObject*)__pyx_t_1);

#line 5 "(tree fragment)"
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */

#line 6 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 6 "(tree fragment)"
  __pyx_v__dict = __pyx_t_1;

#line 6 "(tree fragment)"
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
  __pyx_t_2 = (__pyx_v__dict != Py_None);

#line 7 "(tree fragment)"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 7 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */

#line 8 "(tree fragment)"
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 8 "(tree fragment)"
    __Pyx_INCREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);

#line 8 "(tree fragment)"
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 8 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 8 "(tree fragment)"
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));

#line 8 "(tree fragment)"
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.name is not None
 */

#line 9 "(tree fragment)"
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
    goto __pyx_L3;

#line 7 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.name is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 */

#line 11 "(tree fragment)"
  /*else*/ {

#line 11 "(tree fragment)"
    __pyx_t_3 = (__pyx_v_self->name != Py_None);

#line 11 "(tree fragment)"
    __pyx_v_use_setstate = __pyx_t_3;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  __pyx_t_3 = (__pyx_v_use_setstate != 0);

#line 12 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = self.name is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 */

#line 13 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 13 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 13 "(tree fragment)"
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 13 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_184977713);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_184977713);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_184977713);

#line 13 "(tree fragment)"
    __Pyx_INCREF(Py_None);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(Py_None);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);

#line 13 "(tree fragment)"
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_4);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_1);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);

#line 13 "(tree fragment)"
    __pyx_t_4 = 0;

#line 13 "(tree fragment)"
    __pyx_t_1 = 0;

#line 13 "(tree fragment)"
    __pyx_r = __pyx_t_5;

#line 13 "(tree fragment)"
    __pyx_t_5 = 0;

#line 13 "(tree fragment)"
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

#line 15 "(tree fragment)"
  /*else*/ {

#line 15 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 15 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 15 "(tree fragment)"
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 15 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_184977713);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_184977713);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_184977713);

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);

#line 15 "(tree fragment)"
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_5);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_1);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);

#line 15 "(tree fragment)"
    __pyx_t_5 = 0;

#line 15 "(tree fragment)"
    __pyx_t_1 = 0;

#line 15 "(tree fragment)"
    __pyx_r = __pyx_t_4;

#line 15 "(tree fragment)"
    __pyx_t_4 = 0;

#line 15 "(tree fragment)"
    goto __pyx_L0;

#line 15 "(tree fragment)"
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView.Enum.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v_state);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v__dict);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
/* Python wrapper */

#line 16 "(tree fragment)"
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 16 "(tree fragment)"
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 16 "(tree fragment)"
__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 16 "(tree fragment)"
  int __pyx_lineno = 0;

#line 16 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 16 "(tree fragment)"
  int __pyx_clineno = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */

#line 17 "(tree fragment)"
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(3, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __pyx_t_1 = __pyx_unpickle_Enum__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 17 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 16 "(tree fragment)"
  goto __pyx_L0;

#line 16 "(tree fragment)"
  __pyx_L1_error:;

#line 16 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 16 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView.Enum.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 16 "(tree fragment)"
  __pyx_r = NULL;

#line 16 "(tree fragment)"
  __pyx_L0:;

#line 16 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

/* "View.MemoryView":299
 * 
 * @cname('__pyx_align_pointer')
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory
 */

#line 299 "View.MemoryView"


#line 299 "View.MemoryView"
static void *__pyx_align_pointer(void *__pyx_v_memory, size_t __pyx_v_alignment) {
  Py_intptr_t __pyx_v_aligned_p
#line 299 "View.MemoryView"
;
  size_t __pyx_v_offset
#line 299 "View.MemoryView"
;

#line 299 "View.MemoryView"
  void *__pyx_r;

#line 299 "View.MemoryView"
  int __pyx_t_1;

  /* "View.MemoryView":301
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory             # <<<<<<<<<<<<<<
 *     cdef size_t offset
 * 
 */

#line 301 "View.MemoryView"
  __pyx_v_aligned_p = ((Py_intptr_t)__pyx_v_memory);

  /* "View.MemoryView":305
 * 
 *     with cython.cdivision(True):
 *         offset = aligned_p % alignment             # <<<<<<<<<<<<<<
 * 
 *     if offset > 0:
 */

#line 305 "View.MemoryView"
  __pyx_v_offset = (__pyx_v_aligned_p % __pyx_v_alignment);

  /* "View.MemoryView":307
 *         offset = aligned_p % alignment
 * 
 *     if offset > 0:             # <<<<<<<<<<<<<<
 *         aligned_p += alignment - offset
 * 
 */

#line 307 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_offset > 0) != 0);

#line 307 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":308
 * 
 *     if offset > 0:
 *         aligned_p += alignment - offset             # <<<<<<<<<<<<<<
 * 
 *     return <void *> aligned_p
 */

#line 308 "View.MemoryView"
    __pyx_v_aligned_p = (__pyx_v_aligned_p + (__pyx_v_alignment - __pyx_v_offset));

    /* "View.MemoryView":307
 *         offset = aligned_p % alignment
 * 
 *     if offset > 0:             # <<<<<<<<<<<<<<
 *         aligned_p += alignment - offset
 * 
 */

#line 307 "View.MemoryView"
  }

  /* "View.MemoryView":310
 *         aligned_p += alignment - offset
 * 
 *     return <void *> aligned_p             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 310 "View.MemoryView"
  __pyx_r = ((void *)__pyx_v_aligned_p);

#line 310 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":299
 * 
 * @cname('__pyx_align_pointer')
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory
 */

#line 299 "View.MemoryView"


#line 299 "View.MemoryView"
  /* function exit code */

#line 299 "View.MemoryView"
  __pyx_L0:;

#line 299 "View.MemoryView"
  return __pyx_r;

#line 299 "View.MemoryView"
}

/* "View.MemoryView":346
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

#line 346 "View.MemoryView"


#line 346 "View.MemoryView"
/* Python wrapper */

#line 346 "View.MemoryView"
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 346 "View.MemoryView"
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_obj = 0
#line 346 "View.MemoryView"
;
  int __pyx_v_flags
#line 346 "View.MemoryView"
;
  int __pyx_v_dtype_is_object
#line 346 "View.MemoryView"
;

#line 346 "View.MemoryView"
  int __pyx_lineno = 0;

#line 346 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 346 "View.MemoryView"
  int __pyx_clineno = 0;

#line 346 "View.MemoryView"
  int __pyx_r;

#line 346 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 346 "View.MemoryView"
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);

#line 346 "View.MemoryView"
  {

#line 346 "View.MemoryView"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_obj,&__pyx_n_s_flags,&__pyx_n_s_dtype_is_object,0};

#line 346 "View.MemoryView"
    PyObject* values[3] = {0,0,0};

#line 346 "View.MemoryView"
    if (unlikely(__pyx_kwds)) {

#line 346 "View.MemoryView"
      Py_ssize_t kw_args;

#line 346 "View.MemoryView"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 346 "View.MemoryView"
      switch (pos_args) {
        case  3: 
#line 346 "View.MemoryView"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 346 "View.MemoryView"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 346 "View.MemoryView"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 346 "View.MemoryView"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 346 "View.MemoryView"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 346 "View.MemoryView"
        CYTHON_FALLTHROUGH;

#line 346 "View.MemoryView"
        case  0: break;
        default: 
#line 346 "View.MemoryView"
goto __pyx_L5_argtuple_error;

#line 346 "View.MemoryView"
      }

#line 346 "View.MemoryView"
      kw_args = PyDict_Size(__pyx_kwds);

#line 346 "View.MemoryView"
      switch (pos_args) {

#line 346 "View.MemoryView"
        case  0:

#line 346 "View.MemoryView"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_obj)) != 0)) kw_args--;
        else 
#line 346 "View.MemoryView"
goto __pyx_L5_argtuple_error;

#line 346 "View.MemoryView"
        CYTHON_FALLTHROUGH;

#line 346 "View.MemoryView"
        case  1:

#line 346 "View.MemoryView"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_flags)) != 0)) kw_args--;

#line 346 "View.MemoryView"
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, 1); 
#line 346 "View.MemoryView"
__PYX_ERR(3, 346, __pyx_L3_error)

#line 346 "View.MemoryView"
        }

#line 346 "View.MemoryView"
        CYTHON_FALLTHROUGH;

#line 346 "View.MemoryView"
        case  2:

#line 346 "View.MemoryView"
        if (kw_args > 0) {

#line 346 "View.MemoryView"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dtype_is_object);

#line 346 "View.MemoryView"
          if (value) { values[2] = value; kw_args--; }

#line 346 "View.MemoryView"
        }

#line 346 "View.MemoryView"
      }

#line 346 "View.MemoryView"
      if (unlikely(kw_args > 0)) {

#line 346 "View.MemoryView"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(3, 346, __pyx_L3_error)

#line 346 "View.MemoryView"
      }

#line 346 "View.MemoryView"
    } else {

#line 346 "View.MemoryView"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 346 "View.MemoryView"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 346 "View.MemoryView"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 346 "View.MemoryView"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 346 "View.MemoryView"
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 346 "View.MemoryView"
        break;
        default: 
#line 346 "View.MemoryView"
goto __pyx_L5_argtuple_error;

#line 346 "View.MemoryView"
      }

#line 346 "View.MemoryView"
    }

#line 346 "View.MemoryView"
    __pyx_v_obj = values[0];

#line 346 "View.MemoryView"
    __pyx_v_flags = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(3, 346, __pyx_L3_error)

#line 346 "View.MemoryView"
    if (values[2]) {

#line 346 "View.MemoryView"
      __pyx_v_dtype_is_object = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_dtype_is_object == (int)-1) && PyErr_Occurred())) __PYX_ERR(3, 346, __pyx_L3_error)

#line 346 "View.MemoryView"
    } else {

#line 346 "View.MemoryView"
      __pyx_v_dtype_is_object = ((int)0);

#line 346 "View.MemoryView"
    }

#line 346 "View.MemoryView"
  }

#line 346 "View.MemoryView"
  goto __pyx_L4_argument_unpacking_done;

#line 346 "View.MemoryView"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 346 "View.MemoryView"
__PYX_ERR(3, 346, __pyx_L3_error)

#line 346 "View.MemoryView"
  __pyx_L3_error:;

#line 346 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 346 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 346 "View.MemoryView"
  return -1;

#line 346 "View.MemoryView"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 346 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_obj, __pyx_v_flags, __pyx_v_dtype_is_object);

#line 346 "View.MemoryView"


#line 346 "View.MemoryView"
  /* function exit code */

#line 346 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 346 "View.MemoryView"
  return __pyx_r;

#line 346 "View.MemoryView"
}

#line 346 "View.MemoryView"


#line 346 "View.MemoryView"
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object) {

#line 346 "View.MemoryView"
  int __pyx_r;

#line 346 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 346 "View.MemoryView"
  int __pyx_t_1;

#line 346 "View.MemoryView"
  int __pyx_t_2;

#line 346 "View.MemoryView"
  int __pyx_t_3;

#line 346 "View.MemoryView"
  int __pyx_t_4;

#line 346 "View.MemoryView"
  int __pyx_lineno = 0;

#line 346 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 346 "View.MemoryView"
  int __pyx_clineno = 0;

#line 346 "View.MemoryView"
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "View.MemoryView":347
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj             # <<<<<<<<<<<<<<
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 */

#line 347 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_obj);

#line 347 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_v_obj);

#line 347 "View.MemoryView"
  __Pyx_GOTREF(__pyx_v_self->obj);

#line 347 "View.MemoryView"
  __Pyx_DECREF(__pyx_v_self->obj);

#line 347 "View.MemoryView"
  __pyx_v_self->obj = __pyx_v_obj;

  /* "View.MemoryView":348
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj
 *         self.flags = flags             # <<<<<<<<<<<<<<
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 */

#line 348 "View.MemoryView"
  __pyx_v_self->flags = __pyx_v_flags;

  /* "View.MemoryView":349
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */

#line 349 "View.MemoryView"
  __pyx_t_2 = (((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))) == ((PyObject *)__pyx_memoryview_type));

#line 349 "View.MemoryView"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 349 "View.MemoryView"
  if (!__pyx_t_3) {

#line 349 "View.MemoryView"
  } else {

#line 349 "View.MemoryView"
    __pyx_t_1 = __pyx_t_3;

#line 349 "View.MemoryView"
    goto __pyx_L4_bool_binop_done;

#line 349 "View.MemoryView"
  }

#line 349 "View.MemoryView"
  __pyx_t_3 = (__pyx_v_obj != Py_None);

#line 349 "View.MemoryView"
  __pyx_t_2 = (__pyx_t_3 != 0);

#line 349 "View.MemoryView"
  __pyx_t_1 = __pyx_t_2;

#line 349 "View.MemoryView"
  __pyx_L4_bool_binop_done:;

#line 349 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":350
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)             # <<<<<<<<<<<<<<
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 */

#line 350 "View.MemoryView"
    __pyx_t_4 = __Pyx_GetBuffer(__pyx_v_obj, (&__pyx_v_self->view), __pyx_v_flags); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(3, 350, __pyx_L1_error)

    /* "View.MemoryView":351
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */

#line 351 "View.MemoryView"
    __pyx_t_1 = ((((PyObject *)__pyx_v_self->view.obj) == NULL) != 0);

#line 351 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":352
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None             # <<<<<<<<<<<<<<
 *                 Py_INCREF(Py_None)
 * 
 */

#line 352 "View.MemoryView"
      ((Py_buffer *)(&__pyx_v_self->view))->obj = Py_None;

      /* "View.MemoryView":353
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():
 */

#line 353 "View.MemoryView"
      Py_INCREF(Py_None);

      /* "View.MemoryView":351
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */

#line 351 "View.MemoryView"
    }

    /* "View.MemoryView":349
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */

#line 349 "View.MemoryView"
  }

  /* "View.MemoryView":355
 *                 Py_INCREF(Py_None)
 * 
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():             # <<<<<<<<<<<<<<
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 */

#line 355 "View.MemoryView"
  __pyx_t_1 = ((!(__PYX_CYTHON_ATOMICS_ENABLED() != 0)) != 0);

#line 355 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":357
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:             # <<<<<<<<<<<<<<
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1
 */

#line 357 "View.MemoryView"
    __pyx_t_1 = ((__pyx_memoryview_thread_locks_used < 8) != 0);

#line 357 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":358
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]             # <<<<<<<<<<<<<<
 *                 __pyx_memoryview_thread_locks_used += 1
 *             if self.lock is NULL:
 */

#line 358 "View.MemoryView"
      __pyx_v_self->lock = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);

      /* "View.MemoryView":359
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1             # <<<<<<<<<<<<<<
 *             if self.lock is NULL:
 *                 self.lock = PyThread_allocate_lock()
 */

#line 359 "View.MemoryView"
      __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used + 1);

      /* "View.MemoryView":357
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:             # <<<<<<<<<<<<<<
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1
 */

#line 357 "View.MemoryView"
    }

    /* "View.MemoryView":360
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:
 */

#line 360 "View.MemoryView"
    __pyx_t_1 = ((__pyx_v_self->lock == NULL) != 0);

#line 360 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":361
 *                 __pyx_memoryview_thread_locks_used += 1
 *             if self.lock is NULL:
 *                 self.lock = PyThread_allocate_lock()             # <<<<<<<<<<<<<<
 *                 if self.lock is NULL:
 *                     raise MemoryError
 */

#line 361 "View.MemoryView"
      __pyx_v_self->lock = PyThread_allocate_lock();

      /* "View.MemoryView":362
 *             if self.lock is NULL:
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                     raise MemoryError
 * 
 */

#line 362 "View.MemoryView"
      __pyx_t_1 = ((__pyx_v_self->lock == NULL) != 0);

#line 362 "View.MemoryView"
      if (unlikely(__pyx_t_1)) {

        /* "View.MemoryView":363
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:
 *                     raise MemoryError             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */

#line 363 "View.MemoryView"
        PyErr_NoMemory(); __PYX_ERR(3, 363, __pyx_L1_error)

        /* "View.MemoryView":362
 *             if self.lock is NULL:
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                     raise MemoryError
 * 
 */

#line 362 "View.MemoryView"
      }

      /* "View.MemoryView":360
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:
 */

#line 360 "View.MemoryView"
    }

    /* "View.MemoryView":355
 *                 Py_INCREF(Py_None)
 * 
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():             # <<<<<<<<<<<<<<
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 */

#line 355 "View.MemoryView"
  }

  /* "View.MemoryView":365
 *                     raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */

#line 365 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);

#line 365 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":366
 * 
 *         if flags & PyBUF_FORMAT:
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')             # <<<<<<<<<<<<<<
 *         else:
 *             self.dtype_is_object = dtype_is_object
 */

#line 366 "View.MemoryView"
    __pyx_t_2 = (((__pyx_v_self->view.format[0]) == 'O') != 0);

#line 366 "View.MemoryView"
    if (__pyx_t_2) {

#line 366 "View.MemoryView"
    } else {

#line 366 "View.MemoryView"
      __pyx_t_1 = __pyx_t_2;

#line 366 "View.MemoryView"
      goto __pyx_L12_bool_binop_done;

#line 366 "View.MemoryView"
    }

#line 366 "View.MemoryView"
    __pyx_t_2 = (((__pyx_v_self->view.format[1]) == '\x00') != 0);

#line 366 "View.MemoryView"
    __pyx_t_1 = __pyx_t_2;

#line 366 "View.MemoryView"
    __pyx_L12_bool_binop_done:;

#line 366 "View.MemoryView"
    __pyx_v_self->dtype_is_object = __pyx_t_1;

    /* "View.MemoryView":365
 *                     raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */

#line 365 "View.MemoryView"
    goto __pyx_L11;

#line 365 "View.MemoryView"
  }

  /* "View.MemoryView":368
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 *             self.dtype_is_object = dtype_is_object             # <<<<<<<<<<<<<<
 * 
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(
 */

#line 368 "View.MemoryView"
  /*else*/ {

#line 368 "View.MemoryView"
    __pyx_v_self->dtype_is_object = __pyx_v_dtype_is_object;

#line 368 "View.MemoryView"
  }

#line 368 "View.MemoryView"
  __pyx_L11:;

  /* "View.MemoryView":370
 *             self.dtype_is_object = dtype_is_object
 * 
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(             # <<<<<<<<<<<<<<
 *                   <void *> &self.acquisition_count[0], sizeof(__pyx_atomic_int))
 *         self.typeinfo = NULL
 */

#line 370 "View.MemoryView"
  __pyx_v_self->acquisition_count_aligned_p = ((__pyx_atomic_int *)__pyx_align_pointer(((void *)(&(__pyx_v_self->acquisition_count[0]))), (sizeof(__pyx_atomic_int))));

  /* "View.MemoryView":372
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(
 *                   <void *> &self.acquisition_count[0], sizeof(__pyx_atomic_int))
 *         self.typeinfo = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(memoryview self):
 */

#line 372 "View.MemoryView"
  __pyx_v_self->typeinfo = NULL;

  /* "View.MemoryView":346
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

#line 346 "View.MemoryView"


#line 346 "View.MemoryView"
  /* function exit code */

#line 346 "View.MemoryView"
  __pyx_r = 0;

#line 346 "View.MemoryView"
  goto __pyx_L0;

#line 346 "View.MemoryView"
  __pyx_L1_error:;

#line 346 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 346 "View.MemoryView"
  __pyx_r = -1;

#line 346 "View.MemoryView"
  __pyx_L0:;

#line 346 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 346 "View.MemoryView"
  return __pyx_r;

#line 346 "View.MemoryView"
}

/* "View.MemoryView":374
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

#line 374 "View.MemoryView"


#line 374 "View.MemoryView"
/* Python wrapper */

#line 374 "View.MemoryView"
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self); /*proto*/

#line 374 "View.MemoryView"
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self) {

#line 374 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 374 "View.MemoryView"
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);

#line 374 "View.MemoryView"
  __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 374 "View.MemoryView"


#line 374 "View.MemoryView"
  /* function exit code */

#line 374 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 374 "View.MemoryView"
}

#line 374 "View.MemoryView"


#line 374 "View.MemoryView"
static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self) {
  int __pyx_v_i
#line 374 "View.MemoryView"
;

#line 374 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 374 "View.MemoryView"
  int __pyx_t_1;

#line 374 "View.MemoryView"
  int __pyx_t_2;

#line 374 "View.MemoryView"
  int __pyx_t_3;

#line 374 "View.MemoryView"
  int __pyx_t_4;

#line 374 "View.MemoryView"
  int __pyx_t_5;

#line 374 "View.MemoryView"
  PyThread_type_lock __pyx_t_6;

#line 374 "View.MemoryView"
  PyThread_type_lock __pyx_t_7;

#line 374 "View.MemoryView"
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":375
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */

#line 375 "View.MemoryView"
  __pyx_t_1 = (__pyx_v_self->obj != Py_None);

#line 375 "View.MemoryView"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 375 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":376
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)             # <<<<<<<<<<<<<<
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 */

#line 376 "View.MemoryView"
    __Pyx_ReleaseBuffer((&__pyx_v_self->view));

    /* "View.MemoryView":375
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */

#line 375 "View.MemoryView"
    goto __pyx_L3;

#line 375 "View.MemoryView"
  }

  /* "View.MemoryView":377
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */

#line 377 "View.MemoryView"
  __pyx_t_2 = ((((Py_buffer *)(&__pyx_v_self->view))->obj == Py_None) != 0);

#line 377 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":379
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL             # <<<<<<<<<<<<<<
 *             Py_DECREF(Py_None)
 * 
 */

#line 379 "View.MemoryView"
    ((Py_buffer *)(&__pyx_v_self->view))->obj = NULL;

    /* "View.MemoryView":380
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 *             Py_DECREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         cdef int i
 */

#line 380 "View.MemoryView"
    Py_DECREF(Py_None);

    /* "View.MemoryView":377
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */

#line 377 "View.MemoryView"
  }

#line 377 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":384
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */

#line 384 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_self->lock != NULL) != 0);

#line 384 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":385
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):             # <<<<<<<<<<<<<<
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 */

#line 385 "View.MemoryView"
    __pyx_t_3 = __pyx_memoryview_thread_locks_used;

#line 385 "View.MemoryView"
    __pyx_t_4 = __pyx_t_3;

#line 385 "View.MemoryView"
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {

#line 385 "View.MemoryView"
      __pyx_v_i = __pyx_t_5;

      /* "View.MemoryView":386
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */

#line 386 "View.MemoryView"
      __pyx_t_2 = (((__pyx_memoryview_thread_locks[__pyx_v_i]) == __pyx_v_self->lock) != 0);

#line 386 "View.MemoryView"
      if (__pyx_t_2) {

        /* "View.MemoryView":387
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1             # <<<<<<<<<<<<<<
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 */

#line 387 "View.MemoryView"
        __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used - 1);

        /* "View.MemoryView":388
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */

#line 388 "View.MemoryView"
        __pyx_t_2 = ((__pyx_v_i != __pyx_memoryview_thread_locks_used) != 0);

#line 388 "View.MemoryView"
        if (__pyx_t_2) {

          /* "View.MemoryView":390
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])             # <<<<<<<<<<<<<<
 *                     break
 *             else:
 */

#line 390 "View.MemoryView"
          __pyx_t_6 = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);

#line 390 "View.MemoryView"
          __pyx_t_7 = (__pyx_memoryview_thread_locks[__pyx_v_i]);

          /* "View.MemoryView":389
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (             # <<<<<<<<<<<<<<
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break
 */

#line 389 "View.MemoryView"
          (__pyx_memoryview_thread_locks[__pyx_v_i]) = __pyx_t_6;

#line 389 "View.MemoryView"
          (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]) = __pyx_t_7;

          /* "View.MemoryView":388
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */

#line 388 "View.MemoryView"
        }

        /* "View.MemoryView":391
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break             # <<<<<<<<<<<<<<
 *             else:
 *                 PyThread_free_lock(self.lock)
 */

#line 391 "View.MemoryView"
        goto __pyx_L6_break;

        /* "View.MemoryView":386
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */

#line 386 "View.MemoryView"
      }

#line 386 "View.MemoryView"
    }

#line 386 "View.MemoryView"
    /*else*/ {

      /* "View.MemoryView":393
 *                     break
 *             else:
 *                 PyThread_free_lock(self.lock)             # <<<<<<<<<<<<<<
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 */

#line 393 "View.MemoryView"
      PyThread_free_lock(__pyx_v_self->lock);

#line 393 "View.MemoryView"
    }

#line 393 "View.MemoryView"
    __pyx_L6_break:;

    /* "View.MemoryView":384
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */

#line 384 "View.MemoryView"
  }

  /* "View.MemoryView":374
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

#line 374 "View.MemoryView"


#line 374 "View.MemoryView"
  /* function exit code */

#line 374 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 374 "View.MemoryView"
}

/* "View.MemoryView":395
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

#line 395 "View.MemoryView"


#line 395 "View.MemoryView"
static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  Py_ssize_t __pyx_v_dim
#line 395 "View.MemoryView"
;
  char *__pyx_v_itemp
#line 395 "View.MemoryView"
;
  PyObject *__pyx_v_idx = NULL
#line 395 "View.MemoryView"
;

#line 395 "View.MemoryView"
  char *__pyx_r;

#line 395 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 395 "View.MemoryView"
  Py_ssize_t __pyx_t_1;

#line 395 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 395 "View.MemoryView"
  Py_ssize_t __pyx_t_3;

#line 395 "View.MemoryView"
  PyObject *(*__pyx_t_4)(PyObject *);

#line 395 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 395 "View.MemoryView"
  Py_ssize_t __pyx_t_6;

#line 395 "View.MemoryView"
  char *__pyx_t_7;

#line 395 "View.MemoryView"
  int __pyx_lineno = 0;

#line 395 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 395 "View.MemoryView"
  int __pyx_clineno = 0;

#line 395 "View.MemoryView"
  __Pyx_RefNannySetupContext("get_item_pointer", 0);

  /* "View.MemoryView":397
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf             # <<<<<<<<<<<<<<
 * 
 *         for dim, idx in enumerate(index):
 */

#line 397 "View.MemoryView"
  __pyx_v_itemp = ((char *)__pyx_v_self->view.buf);

  /* "View.MemoryView":399
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */

#line 399 "View.MemoryView"
  __pyx_t_1 = 0;

#line 399 "View.MemoryView"
  if (likely(PyList_CheckExact(__pyx_v_index)) || PyTuple_CheckExact(__pyx_v_index)) {

#line 399 "View.MemoryView"
    __pyx_t_2 = __pyx_v_index; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;

#line 399 "View.MemoryView"
    __pyx_t_4 = NULL;

#line 399 "View.MemoryView"
  } else {
    __pyx_t_3 = -1; 
#line 399 "View.MemoryView"
__pyx_t_2 = PyObject_GetIter(__pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 399, __pyx_L1_error)

#line 399 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 399 "View.MemoryView"
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 399, __pyx_L1_error)

#line 399 "View.MemoryView"
  }

#line 399 "View.MemoryView"
  for (;;) {

#line 399 "View.MemoryView"
    if (likely(!__pyx_t_4)) {

#line 399 "View.MemoryView"
      if (likely(PyList_CheckExact(__pyx_t_2))) {

#line 399 "View.MemoryView"
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;

#line 399 "View.MemoryView"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 399 "View.MemoryView"
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(3, 399, __pyx_L1_error)

#line 399 "View.MemoryView"
        #else

#line 399 "View.MemoryView"
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 399, __pyx_L1_error)

#line 399 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_5);

#line 399 "View.MemoryView"
        #endif

#line 399 "View.MemoryView"
      } else {

#line 399 "View.MemoryView"
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;

#line 399 "View.MemoryView"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 399 "View.MemoryView"
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(3, 399, __pyx_L1_error)

#line 399 "View.MemoryView"
        #else

#line 399 "View.MemoryView"
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 399, __pyx_L1_error)

#line 399 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_5);

#line 399 "View.MemoryView"
        #endif

#line 399 "View.MemoryView"
      }
    } else 
#line 399 "View.MemoryView"
{

#line 399 "View.MemoryView"
      __pyx_t_5 = __pyx_t_4(__pyx_t_2);

#line 399 "View.MemoryView"
      if (unlikely(!__pyx_t_5)) {

#line 399 "View.MemoryView"
        PyObject* exc_type = PyErr_Occurred();

#line 399 "View.MemoryView"
        if (exc_type) {

#line 399 "View.MemoryView"
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();

#line 399 "View.MemoryView"
          else __PYX_ERR(3, 399, __pyx_L1_error)

#line 399 "View.MemoryView"
        }

#line 399 "View.MemoryView"
        break;

#line 399 "View.MemoryView"
      }

#line 399 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_5);

#line 399 "View.MemoryView"
    }

#line 399 "View.MemoryView"
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_5);

#line 399 "View.MemoryView"
    __pyx_t_5 = 0;

#line 399 "View.MemoryView"
    __pyx_v_dim = __pyx_t_1;

#line 399 "View.MemoryView"
    __pyx_t_1 = (__pyx_t_1 + 1);

    /* "View.MemoryView":400
 * 
 *         for dim, idx in enumerate(index):
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)             # <<<<<<<<<<<<<<
 * 
 *         return itemp
 */

#line 400 "View.MemoryView"
    __pyx_t_6 = __Pyx_PyIndex_AsSsize_t(__pyx_v_idx); if (unlikely((__pyx_t_6 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(3, 400, __pyx_L1_error)

#line 400 "View.MemoryView"
    __pyx_t_7 = __pyx_pybuffer_index((&__pyx_v_self->view), __pyx_v_itemp, __pyx_t_6, __pyx_v_dim); if (unlikely(__pyx_t_7 == ((char *)NULL))) __PYX_ERR(3, 400, __pyx_L1_error)

#line 400 "View.MemoryView"
    __pyx_v_itemp = __pyx_t_7;

    /* "View.MemoryView":399
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */

#line 399 "View.MemoryView"
  }

#line 399 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":402
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 *         return itemp             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 402 "View.MemoryView"
  __pyx_r = __pyx_v_itemp;

#line 402 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":395
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

#line 395 "View.MemoryView"


#line 395 "View.MemoryView"
  /* function exit code */

#line 395 "View.MemoryView"
  __pyx_L1_error:;

#line 395 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 395 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 395 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.get_item_pointer", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 395 "View.MemoryView"
  __pyx_r = NULL;

#line 395 "View.MemoryView"
  __pyx_L0:;

#line 395 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_idx);

#line 395 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 395 "View.MemoryView"
  return __pyx_r;

#line 395 "View.MemoryView"
}

/* "View.MemoryView":405
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

#line 405 "View.MemoryView"


#line 405 "View.MemoryView"
/* Python wrapper */

#line 405 "View.MemoryView"
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index); /*proto*/

#line 405 "View.MemoryView"
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index) {

#line 405 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 405 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 405 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = 
#line 405 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index));

#line 405 "View.MemoryView"


#line 405 "View.MemoryView"
  /* function exit code */

#line 405 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 405 "View.MemoryView"
  return __pyx_r;

#line 405 "View.MemoryView"
}

#line 405 "View.MemoryView"


#line 405 "View.MemoryView"
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  PyObject *__pyx_v_have_slices = NULL
#line 405 "View.MemoryView"
;
  PyObject *__pyx_v_indices = NULL
#line 405 "View.MemoryView"
;
  char *__pyx_v_itemp
#line 405 "View.MemoryView"
;

#line 405 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 405 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 405 "View.MemoryView"
  int __pyx_t_1;

#line 405 "View.MemoryView"
  int __pyx_t_2;

#line 405 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 405 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 405 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 405 "View.MemoryView"
  char *__pyx_t_6;

#line 405 "View.MemoryView"
  int __pyx_lineno = 0;

#line 405 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 405 "View.MemoryView"
  int __pyx_clineno = 0;

#line 405 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "View.MemoryView":406
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */

#line 406 "View.MemoryView"
  __pyx_t_1 = (__pyx_v_index == __pyx_builtin_Ellipsis);

#line 406 "View.MemoryView"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 406 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":407
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:
 *             return self             # <<<<<<<<<<<<<<
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)
 */

#line 407 "View.MemoryView"
    __Pyx_XDECREF(__pyx_r);

#line 407 "View.MemoryView"
    __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 407 "View.MemoryView"
    __pyx_r = ((PyObject *)__pyx_v_self);

#line 407 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":406
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */

#line 406 "View.MemoryView"
  }

  /* "View.MemoryView":409
 *             return self
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         cdef char *itemp
 */

#line 409 "View.MemoryView"
  __pyx_t_3 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 409, __pyx_L1_error)

#line 409 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 409 "View.MemoryView"
  if (likely(__pyx_t_3 != Py_None)) {

#line 409 "View.MemoryView"
    PyObject* sequence = __pyx_t_3;

#line 409 "View.MemoryView"
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);

#line 409 "View.MemoryView"
    if (unlikely(size != 2)) {

#line 409 "View.MemoryView"
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);

#line 409 "View.MemoryView"
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);

#line 409 "View.MemoryView"
      __PYX_ERR(3, 409, __pyx_L1_error)

#line 409 "View.MemoryView"
    }

#line 409 "View.MemoryView"
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 409 "View.MemoryView"
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 

#line 409 "View.MemoryView"
    __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 

#line 409 "View.MemoryView"
    __Pyx_INCREF(__pyx_t_4);

#line 409 "View.MemoryView"
    __Pyx_INCREF(__pyx_t_5);

#line 409 "View.MemoryView"
    #else

#line 409 "View.MemoryView"
    __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 409, __pyx_L1_error)

#line 409 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 409 "View.MemoryView"
    __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 409, __pyx_L1_error)

#line 409 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_5);

#line 409 "View.MemoryView"
    #endif

#line 409 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 409 "View.MemoryView"
  } else {

#line 409 "View.MemoryView"
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(3, 409, __pyx_L1_error)

#line 409 "View.MemoryView"
  }

#line 409 "View.MemoryView"
  __pyx_v_have_slices = __pyx_t_4;

#line 409 "View.MemoryView"
  __pyx_t_4 = 0;

#line 409 "View.MemoryView"
  __pyx_v_indices = __pyx_t_5;

#line 409 "View.MemoryView"
  __pyx_t_5 = 0;

  /* "View.MemoryView":412
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */

#line 412 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(3, 412, __pyx_L1_error)

#line 412 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":413
 *         cdef char *itemp
 *         if have_slices:
 *             return memview_slice(self, indices)             # <<<<<<<<<<<<<<
 *         else:
 *             itemp = self.get_item_pointer(indices)
 */

#line 413 "View.MemoryView"
    __Pyx_XDECREF(__pyx_r);

#line 413 "View.MemoryView"
    __pyx_t_3 = ((PyObject *)__pyx_memview_slice(__pyx_v_self, __pyx_v_indices)); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 413, __pyx_L1_error)

#line 413 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 413 "View.MemoryView"
    __pyx_r = __pyx_t_3;

#line 413 "View.MemoryView"
    __pyx_t_3 = 0;

#line 413 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":412
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */

#line 412 "View.MemoryView"
  }

  /* "View.MemoryView":415
 *             return memview_slice(self, indices)
 *         else:
 *             itemp = self.get_item_pointer(indices)             # <<<<<<<<<<<<<<
 *             return self.convert_item_to_object(itemp)
 * 
 */

#line 415 "View.MemoryView"
  /*else*/ {

#line 415 "View.MemoryView"
    __pyx_t_6 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_indices); if (unlikely(__pyx_t_6 == ((char *)NULL))) __PYX_ERR(3, 415, __pyx_L1_error)

#line 415 "View.MemoryView"
    __pyx_v_itemp = __pyx_t_6;

    /* "View.MemoryView":416
 *         else:
 *             itemp = self.get_item_pointer(indices)
 *             return self.convert_item_to_object(itemp)             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(memoryview self, object index, object value):
 */

#line 416 "View.MemoryView"
    __Pyx_XDECREF(__pyx_r);

#line 416 "View.MemoryView"
    __pyx_t_3 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->convert_item_to_object(__pyx_v_self, __pyx_v_itemp); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 416, __pyx_L1_error)

#line 416 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 416 "View.MemoryView"
    __pyx_r = __pyx_t_3;

#line 416 "View.MemoryView"
    __pyx_t_3 = 0;

#line 416 "View.MemoryView"
    goto __pyx_L0;

#line 416 "View.MemoryView"
  }

  /* "View.MemoryView":405
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

#line 405 "View.MemoryView"


#line 405 "View.MemoryView"
  /* function exit code */

#line 405 "View.MemoryView"
  __pyx_L1_error:;

#line 405 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 405 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_4);

#line 405 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 405 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 405 "View.MemoryView"
  __pyx_r = NULL;

#line 405 "View.MemoryView"
  __pyx_L0:;

#line 405 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_have_slices);

#line 405 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_indices);

#line 405 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 405 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 405 "View.MemoryView"
  return __pyx_r;

#line 405 "View.MemoryView"
}

/* "View.MemoryView":418
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")
 */

#line 418 "View.MemoryView"


#line 418 "View.MemoryView"
/* Python wrapper */

#line 418 "View.MemoryView"
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /*proto*/

#line 418 "View.MemoryView"
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {

#line 418 "View.MemoryView"
  int __pyx_r;

#line 418 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 418 "View.MemoryView"
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = 
#line 418 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index), ((PyObject *)__pyx_v_value));

#line 418 "View.MemoryView"


#line 418 "View.MemoryView"
  /* function exit code */

#line 418 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 418 "View.MemoryView"
  return __pyx_r;

#line 418 "View.MemoryView"
}

#line 418 "View.MemoryView"


#line 418 "View.MemoryView"
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_have_slices = NULL
#line 418 "View.MemoryView"
;
  PyObject *__pyx_v_obj = NULL
#line 418 "View.MemoryView"
;

#line 418 "View.MemoryView"
  int __pyx_r;

#line 418 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 418 "View.MemoryView"
  int __pyx_t_1;

#line 418 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 418 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 418 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 418 "View.MemoryView"
  int __pyx_lineno = 0;

#line 418 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 418 "View.MemoryView"
  int __pyx_clineno = 0;

#line 418 "View.MemoryView"
  __Pyx_RefNannySetupContext("__setitem__", 0);

#line 418 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_index);

  /* "View.MemoryView":419
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 */

#line 419 "View.MemoryView"
  __pyx_t_1 = (__pyx_v_self->view.readonly != 0);

#line 419 "View.MemoryView"
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":420
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 */

#line 420 "View.MemoryView"
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 420, __pyx_L1_error)

#line 420 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 420 "View.MemoryView"
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);

#line 420 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 420 "View.MemoryView"
    __PYX_ERR(3, 420, __pyx_L1_error)

    /* "View.MemoryView":419
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 */

#line 419 "View.MemoryView"
  }

  /* "View.MemoryView":422
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         if have_slices:
 */

#line 422 "View.MemoryView"
  __pyx_t_2 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 422, __pyx_L1_error)

#line 422 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 422 "View.MemoryView"
  if (likely(__pyx_t_2 != Py_None)) {

#line 422 "View.MemoryView"
    PyObject* sequence = __pyx_t_2;

#line 422 "View.MemoryView"
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);

#line 422 "View.MemoryView"
    if (unlikely(size != 2)) {

#line 422 "View.MemoryView"
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);

#line 422 "View.MemoryView"
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);

#line 422 "View.MemoryView"
      __PYX_ERR(3, 422, __pyx_L1_error)

#line 422 "View.MemoryView"
    }

#line 422 "View.MemoryView"
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 422 "View.MemoryView"
    __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 

#line 422 "View.MemoryView"
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 

#line 422 "View.MemoryView"
    __Pyx_INCREF(__pyx_t_3);

#line 422 "View.MemoryView"
    __Pyx_INCREF(__pyx_t_4);

#line 422 "View.MemoryView"
    #else

#line 422 "View.MemoryView"
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 422, __pyx_L1_error)

#line 422 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 422 "View.MemoryView"
    __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 422, __pyx_L1_error)

#line 422 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 422 "View.MemoryView"
    #endif

#line 422 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 422 "View.MemoryView"
  } else {

#line 422 "View.MemoryView"
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(3, 422, __pyx_L1_error)

#line 422 "View.MemoryView"
  }

#line 422 "View.MemoryView"
  __pyx_v_have_slices = __pyx_t_3;

#line 422 "View.MemoryView"
  __pyx_t_3 = 0;

#line 422 "View.MemoryView"
  __Pyx_DECREF_SET(__pyx_v_index, __pyx_t_4);

#line 422 "View.MemoryView"
  __pyx_t_4 = 0;

  /* "View.MemoryView":424
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */

#line 424 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(3, 424, __pyx_L1_error)

#line 424 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":425
 * 
 *         if have_slices:
 *             obj = self.is_slice(value)             # <<<<<<<<<<<<<<
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)
 */

#line 425 "View.MemoryView"
    __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->is_slice(__pyx_v_self, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 425, __pyx_L1_error)

#line 425 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 425 "View.MemoryView"
    __pyx_v_obj = __pyx_t_2;

#line 425 "View.MemoryView"
    __pyx_t_2 = 0;

    /* "View.MemoryView":426
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */

#line 426 "View.MemoryView"
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_obj); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(3, 426, __pyx_L1_error)

#line 426 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":427
 *             obj = self.is_slice(value)
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)             # <<<<<<<<<<<<<<
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)
 */

#line 427 "View.MemoryView"
      __pyx_t_2 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 427, __pyx_L1_error)

#line 427 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_2);

#line 427 "View.MemoryView"
      __pyx_t_4 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assignment(__pyx_v_self, __pyx_t_2, __pyx_v_obj); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 427, __pyx_L1_error)

#line 427 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_4);

#line 427 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 427 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "View.MemoryView":426
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */

#line 426 "View.MemoryView"
      goto __pyx_L5;

#line 426 "View.MemoryView"
    }

    /* "View.MemoryView":429
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)             # <<<<<<<<<<<<<<
 *         else:
 *             self.setitem_indexed(index, value)
 */

#line 429 "View.MemoryView"
    /*else*/ {

#line 429 "View.MemoryView"
      __pyx_t_4 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 429, __pyx_L1_error)

#line 429 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_4);

#line 429 "View.MemoryView"
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_memoryview_type))))) __PYX_ERR(3, 429, __pyx_L1_error)

#line 429 "View.MemoryView"
      __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assign_scalar(__pyx_v_self, ((struct __pyx_memoryview_obj *)__pyx_t_4), __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 429, __pyx_L1_error)

#line 429 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_2);

#line 429 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 429 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 429 "View.MemoryView"
    }

#line 429 "View.MemoryView"
    __pyx_L5:;

    /* "View.MemoryView":424
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */

#line 424 "View.MemoryView"
    goto __pyx_L4;

#line 424 "View.MemoryView"
  }

  /* "View.MemoryView":431
 *                 self.setitem_slice_assign_scalar(self[index], value)
 *         else:
 *             self.setitem_indexed(index, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef is_slice(self, obj):
 */

#line 431 "View.MemoryView"
  /*else*/ {

#line 431 "View.MemoryView"
    __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_indexed(__pyx_v_self, __pyx_v_index, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 431, __pyx_L1_error)

#line 431 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 431 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 431 "View.MemoryView"
  }

#line 431 "View.MemoryView"
  __pyx_L4:;

  /* "View.MemoryView":418
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")
 */

#line 418 "View.MemoryView"


#line 418 "View.MemoryView"
  /* function exit code */

#line 418 "View.MemoryView"
  __pyx_r = 0;

#line 418 "View.MemoryView"
  goto __pyx_L0;

#line 418 "View.MemoryView"
  __pyx_L1_error:;

#line 418 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 418 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 418 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_4);

#line 418 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 418 "View.MemoryView"
  __pyx_r = -1;

#line 418 "View.MemoryView"
  __pyx_L0:;

#line 418 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_have_slices);

#line 418 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_obj);

#line 418 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_index);

#line 418 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 418 "View.MemoryView"
  return __pyx_r;

#line 418 "View.MemoryView"
}

/* "View.MemoryView":433
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

#line 433 "View.MemoryView"


#line 433 "View.MemoryView"
static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj) {

#line 433 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 433 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 433 "View.MemoryView"
  int __pyx_t_1;

#line 433 "View.MemoryView"
  int __pyx_t_2;

#line 433 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 433 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 433 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 433 "View.MemoryView"
  PyObject *__pyx_t_6 = NULL;

#line 433 "View.MemoryView"
  PyObject *__pyx_t_7 = NULL;

#line 433 "View.MemoryView"
  PyObject *__pyx_t_8 = NULL;

#line 433 "View.MemoryView"
  int __pyx_t_9;

#line 433 "View.MemoryView"
  int __pyx_lineno = 0;

#line 433 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 433 "View.MemoryView"
  int __pyx_clineno = 0;

#line 433 "View.MemoryView"
  __Pyx_RefNannySetupContext("is_slice", 0);

#line 433 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_obj);

  /* "View.MemoryView":434
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */

#line 434 "View.MemoryView"
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_memoryview_type); 

#line 434 "View.MemoryView"
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);

#line 434 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":435
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */

#line 435 "View.MemoryView"
    {

#line 435 "View.MemoryView"
      __Pyx_PyThreadState_declare

#line 435 "View.MemoryView"
      __Pyx_PyThreadState_assign

#line 435 "View.MemoryView"
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);

#line 435 "View.MemoryView"
      __Pyx_XGOTREF(__pyx_t_3);

#line 435 "View.MemoryView"
      __Pyx_XGOTREF(__pyx_t_4);

#line 435 "View.MemoryView"
      __Pyx_XGOTREF(__pyx_t_5);

#line 435 "View.MemoryView"
      /*try:*/ {

        /* "View.MemoryView":436
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */

#line 436 "View.MemoryView"
        __pyx_t_6 = __Pyx_PyInt_From_int(((__pyx_v_self->flags & (~PyBUF_WRITABLE)) | PyBUF_ANY_CONTIGUOUS)); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 436, __pyx_L4_error)

#line 436 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_6);

        /* "View.MemoryView":437
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)             # <<<<<<<<<<<<<<
 *             except TypeError:
 *                 return None
 */

#line 437 "View.MemoryView"
        __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 437, __pyx_L4_error)

#line 437 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_7);

        /* "View.MemoryView":436
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */

#line 436 "View.MemoryView"
        __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 436, __pyx_L4_error)

#line 436 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_8);

#line 436 "View.MemoryView"
        __Pyx_INCREF(__pyx_v_obj);

#line 436 "View.MemoryView"
        __Pyx_GIVEREF(__pyx_v_obj);

#line 436 "View.MemoryView"
        PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_obj);

#line 436 "View.MemoryView"
        __Pyx_GIVEREF(__pyx_t_6);

#line 436 "View.MemoryView"
        PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_6);

#line 436 "View.MemoryView"
        __Pyx_GIVEREF(__pyx_t_7);

#line 436 "View.MemoryView"
        PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_7);

#line 436 "View.MemoryView"
        __pyx_t_6 = 0;

#line 436 "View.MemoryView"
        __pyx_t_7 = 0;

#line 436 "View.MemoryView"
        __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_8, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 436, __pyx_L4_error)

#line 436 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_7);

#line 436 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 436 "View.MemoryView"
        __Pyx_DECREF_SET(__pyx_v_obj, __pyx_t_7);

#line 436 "View.MemoryView"
        __pyx_t_7 = 0;

        /* "View.MemoryView":435
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */

#line 435 "View.MemoryView"
      }

#line 435 "View.MemoryView"
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 435 "View.MemoryView"
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 435 "View.MemoryView"
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 435 "View.MemoryView"
      goto __pyx_L9_try_end;

#line 435 "View.MemoryView"
      __pyx_L4_error:;

#line 435 "View.MemoryView"
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 435 "View.MemoryView"
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 435 "View.MemoryView"
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "View.MemoryView":438
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 *             except TypeError:             # <<<<<<<<<<<<<<
 *                 return None
 * 
 */

#line 438 "View.MemoryView"
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);

#line 438 "View.MemoryView"
      if (__pyx_t_9) {

#line 438 "View.MemoryView"
        __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 438 "View.MemoryView"
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_6) < 0) __PYX_ERR(3, 438, __pyx_L6_except_error)

#line 438 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_7);

#line 438 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_8);

#line 438 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_6);

        /* "View.MemoryView":439
 *                                  self.dtype_is_object)
 *             except TypeError:
 *                 return None             # <<<<<<<<<<<<<<
 * 
 *         return obj
 */

#line 439 "View.MemoryView"
        __Pyx_XDECREF(__pyx_r);

#line 439 "View.MemoryView"
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 439 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 439 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 439 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 439 "View.MemoryView"
        goto __pyx_L7_except_return;

#line 439 "View.MemoryView"
      }

#line 439 "View.MemoryView"
      goto __pyx_L6_except_error;

#line 439 "View.MemoryView"
      __pyx_L6_except_error:;

      /* "View.MemoryView":435
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */

#line 435 "View.MemoryView"
      __Pyx_XGIVEREF(__pyx_t_3);

#line 435 "View.MemoryView"
      __Pyx_XGIVEREF(__pyx_t_4);

#line 435 "View.MemoryView"
      __Pyx_XGIVEREF(__pyx_t_5);

#line 435 "View.MemoryView"
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);

#line 435 "View.MemoryView"
      goto __pyx_L1_error;

#line 435 "View.MemoryView"
      __pyx_L7_except_return:;

#line 435 "View.MemoryView"
      __Pyx_XGIVEREF(__pyx_t_3);

#line 435 "View.MemoryView"
      __Pyx_XGIVEREF(__pyx_t_4);

#line 435 "View.MemoryView"
      __Pyx_XGIVEREF(__pyx_t_5);

#line 435 "View.MemoryView"
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);

#line 435 "View.MemoryView"
      goto __pyx_L0;

#line 435 "View.MemoryView"
      __pyx_L9_try_end:;

#line 435 "View.MemoryView"
    }

    /* "View.MemoryView":434
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */

#line 434 "View.MemoryView"
  }

  /* "View.MemoryView":441
 *                 return None
 * 
 *         return obj             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assignment(self, dst, src):
 */

#line 441 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 441 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_obj);

#line 441 "View.MemoryView"
  __pyx_r = __pyx_v_obj;

#line 441 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":433
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

#line 433 "View.MemoryView"


#line 433 "View.MemoryView"
  /* function exit code */

#line 433 "View.MemoryView"
  __pyx_L1_error:;

#line 433 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_6);

#line 433 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_7);

#line 433 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_8);

#line 433 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 433 "View.MemoryView"
  __pyx_r = 0;

#line 433 "View.MemoryView"
  __pyx_L0:;

#line 433 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_obj);

#line 433 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 433 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 433 "View.MemoryView"
  return __pyx_r;

#line 433 "View.MemoryView"
}

/* "View.MemoryView":443
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

#line 443 "View.MemoryView"


#line 443 "View.MemoryView"
static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src) {
  __Pyx_memviewslice __pyx_v_dst_slice
#line 443 "View.MemoryView"
;
  __Pyx_memviewslice __pyx_v_src_slice
#line 443 "View.MemoryView"
;

#line 443 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 443 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 443 "View.MemoryView"
  __Pyx_memviewslice *__pyx_t_1;

#line 443 "View.MemoryView"
  __Pyx_memviewslice *__pyx_t_2;

#line 443 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 443 "View.MemoryView"
  int __pyx_t_4;

#line 443 "View.MemoryView"
  int __pyx_t_5;

#line 443 "View.MemoryView"
  int __pyx_t_6;

#line 443 "View.MemoryView"
  int __pyx_lineno = 0;

#line 443 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 443 "View.MemoryView"
  int __pyx_clineno = 0;

#line 443 "View.MemoryView"
  __Pyx_RefNannySetupContext("setitem_slice_assignment", 0);

  /* "View.MemoryView":447
 *         cdef __Pyx_memviewslice src_slice
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],             # <<<<<<<<<<<<<<
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 */

#line 447 "View.MemoryView"
  if (!(likely(((__pyx_v_src) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_src, __pyx_memoryview_type))))) __PYX_ERR(3, 447, __pyx_L1_error)

#line 447 "View.MemoryView"
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_src), (&__pyx_v_src_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(3, 447, __pyx_L1_error)

  /* "View.MemoryView":448
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],
 *                                  get_slice_from_memview(dst, &dst_slice)[0],             # <<<<<<<<<<<<<<
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 */

#line 448 "View.MemoryView"
  if (!(likely(((__pyx_v_dst) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_dst, __pyx_memoryview_type))))) __PYX_ERR(3, 448, __pyx_L1_error)

#line 448 "View.MemoryView"
  __pyx_t_2 = __pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_dst), (&__pyx_v_dst_slice)); if (unlikely(__pyx_t_2 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(3, 448, __pyx_L1_error)

  /* "View.MemoryView":449
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 */

#line 449 "View.MemoryView"
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 449, __pyx_L1_error)

#line 449 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 449 "View.MemoryView"
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(3, 449, __pyx_L1_error)

#line 449 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 449 "View.MemoryView"
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_dst, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 449, __pyx_L1_error)

#line 449 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 449 "View.MemoryView"
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(3, 449, __pyx_L1_error)

#line 449 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":447
 *         cdef __Pyx_memviewslice src_slice
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],             # <<<<<<<<<<<<<<
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 */

#line 447 "View.MemoryView"
  __pyx_t_6 = __pyx_memoryview_copy_contents((__pyx_t_1[0]), (__pyx_t_2[0]), __pyx_t_4, __pyx_t_5, __pyx_v_self->dtype_is_object); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(3, 447, __pyx_L1_error)

  /* "View.MemoryView":443
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

#line 443 "View.MemoryView"


#line 443 "View.MemoryView"
  /* function exit code */

#line 443 "View.MemoryView"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 443 "View.MemoryView"
  goto __pyx_L0;

#line 443 "View.MemoryView"
  __pyx_L1_error:;

#line 443 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 443 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assignment", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 443 "View.MemoryView"
  __pyx_r = 0;

#line 443 "View.MemoryView"
  __pyx_L0:;

#line 443 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 443 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 443 "View.MemoryView"
  return __pyx_r;

#line 443 "View.MemoryView"
}

/* "View.MemoryView":451
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

#line 451 "View.MemoryView"


#line 451 "View.MemoryView"
static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value) {
  int __pyx_v_array[0x80]
#line 451 "View.MemoryView"
;
  void *__pyx_v_tmp
#line 451 "View.MemoryView"
;
  void *__pyx_v_item
#line 451 "View.MemoryView"
;
  __Pyx_memviewslice *__pyx_v_dst_slice
#line 451 "View.MemoryView"
;
  __Pyx_memviewslice __pyx_v_tmp_slice
#line 451 "View.MemoryView"
;

#line 451 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 451 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 451 "View.MemoryView"
  __Pyx_memviewslice *__pyx_t_1;

#line 451 "View.MemoryView"
  int __pyx_t_2;

#line 451 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 451 "View.MemoryView"
  int __pyx_t_4;

#line 451 "View.MemoryView"
  int __pyx_t_5;

#line 451 "View.MemoryView"
  char const *__pyx_t_6;

#line 451 "View.MemoryView"
  PyObject *__pyx_t_7 = NULL;

#line 451 "View.MemoryView"
  PyObject *__pyx_t_8 = NULL;

#line 451 "View.MemoryView"
  PyObject *__pyx_t_9 = NULL;

#line 451 "View.MemoryView"
  PyObject *__pyx_t_10 = NULL;

#line 451 "View.MemoryView"
  PyObject *__pyx_t_11 = NULL;

#line 451 "View.MemoryView"
  PyObject *__pyx_t_12 = NULL;

#line 451 "View.MemoryView"
  int __pyx_lineno = 0;

#line 451 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 451 "View.MemoryView"
  int __pyx_clineno = 0;

#line 451 "View.MemoryView"
  __Pyx_RefNannySetupContext("setitem_slice_assign_scalar", 0);

  /* "View.MemoryView":453
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 *         cdef int array[128]
 *         cdef void *tmp = NULL             # <<<<<<<<<<<<<<
 *         cdef void *item
 * 
 */

#line 453 "View.MemoryView"
  __pyx_v_tmp = NULL;

  /* "View.MemoryView":458
 *         cdef __Pyx_memviewslice *dst_slice
 *         cdef __Pyx_memviewslice tmp_slice
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)             # <<<<<<<<<<<<<<
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 */

#line 458 "View.MemoryView"
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_dst, (&__pyx_v_tmp_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(3, 458, __pyx_L1_error)

#line 458 "View.MemoryView"
  __pyx_v_dst_slice = __pyx_t_1;

  /* "View.MemoryView":460
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */

#line 460 "View.MemoryView"
  __pyx_t_2 = ((((size_t)__pyx_v_self->view.itemsize) > (sizeof(__pyx_v_array))) != 0);

#line 460 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":461
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)             # <<<<<<<<<<<<<<
 *             if tmp == NULL:
 *                 raise MemoryError
 */

#line 461 "View.MemoryView"
    __pyx_v_tmp = PyMem_Malloc(__pyx_v_self->view.itemsize);

    /* "View.MemoryView":462
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */

#line 462 "View.MemoryView"
    __pyx_t_2 = ((__pyx_v_tmp == NULL) != 0);

#line 462 "View.MemoryView"
    if (unlikely(__pyx_t_2)) {

      /* "View.MemoryView":463
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 *                 raise MemoryError             # <<<<<<<<<<<<<<
 *             item = tmp
 *         else:
 */

#line 463 "View.MemoryView"
      PyErr_NoMemory(); __PYX_ERR(3, 463, __pyx_L1_error)

      /* "View.MemoryView":462
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */

#line 462 "View.MemoryView"
    }

    /* "View.MemoryView":464
 *             if tmp == NULL:
 *                 raise MemoryError
 *             item = tmp             # <<<<<<<<<<<<<<
 *         else:
 *             item = <void *> array
 */

#line 464 "View.MemoryView"
    __pyx_v_item = __pyx_v_tmp;

    /* "View.MemoryView":460
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */

#line 460 "View.MemoryView"
    goto __pyx_L3;

#line 460 "View.MemoryView"
  }

  /* "View.MemoryView":466
 *             item = tmp
 *         else:
 *             item = <void *> array             # <<<<<<<<<<<<<<
 * 
 *         try:
 */

#line 466 "View.MemoryView"
  /*else*/ {

#line 466 "View.MemoryView"
    __pyx_v_item = ((void *)__pyx_v_array);

#line 466 "View.MemoryView"
  }

#line 466 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":468
 *             item = <void *> array
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value
 */

#line 468 "View.MemoryView"
  /*try:*/ {

    /* "View.MemoryView":469
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */

#line 469 "View.MemoryView"
    __pyx_t_2 = (__pyx_v_self->dtype_is_object != 0);

#line 469 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":470
 *         try:
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value             # <<<<<<<<<<<<<<
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)
 */

#line 470 "View.MemoryView"
      (((PyObject **)__pyx_v_item)[0]) = ((PyObject *)__pyx_v_value);

      /* "View.MemoryView":469
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */

#line 469 "View.MemoryView"
      goto __pyx_L8;

#line 469 "View.MemoryView"
    }

    /* "View.MemoryView":472
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 472 "View.MemoryView"
    /*else*/ {

#line 472 "View.MemoryView"
      __pyx_t_3 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, ((char *)__pyx_v_item), __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 472, __pyx_L6_error)

#line 472 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_3);

#line 472 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 472 "View.MemoryView"
    }

#line 472 "View.MemoryView"
    __pyx_L8:;

    /* "View.MemoryView":476
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */

#line 476 "View.MemoryView"
    __pyx_t_2 = ((__pyx_v_self->view.suboffsets != NULL) != 0);

#line 476 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":477
 * 
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)             # <<<<<<<<<<<<<<
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 *                                 item, self.dtype_is_object)
 */

#line 477 "View.MemoryView"
      __pyx_t_3 = assert_direct_dimensions(__pyx_v_self->view.suboffsets, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 477, __pyx_L6_error)

#line 477 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_3);

#line 477 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "View.MemoryView":476
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */

#line 476 "View.MemoryView"
    }

    /* "View.MemoryView":478
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,             # <<<<<<<<<<<<<<
 *                                 item, self.dtype_is_object)
 *         finally:
 */

#line 478 "View.MemoryView"
    __pyx_memoryview_slice_assign_scalar(__pyx_v_dst_slice, __pyx_v_dst->view.ndim, __pyx_v_self->view.itemsize, __pyx_v_item, __pyx_v_self->dtype_is_object);

#line 478 "View.MemoryView"
  }

  /* "View.MemoryView":481
 *                                 item, self.dtype_is_object)
 *         finally:
 *             PyMem_Free(tmp)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_indexed(self, index, value):
 */

#line 481 "View.MemoryView"
  /*finally:*/ {

#line 481 "View.MemoryView"
    /*normal exit:*/{

#line 481 "View.MemoryView"
      PyMem_Free(__pyx_v_tmp);

#line 481 "View.MemoryView"
      goto __pyx_L7;

#line 481 "View.MemoryView"
    }

#line 481 "View.MemoryView"
    __pyx_L6_error:;

#line 481 "View.MemoryView"
    /*exception exit:*/{

#line 481 "View.MemoryView"
      __Pyx_PyThreadState_declare

#line 481 "View.MemoryView"
      __Pyx_PyThreadState_assign

#line 481 "View.MemoryView"
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;

#line 481 "View.MemoryView"
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 481 "View.MemoryView"
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);

#line 481 "View.MemoryView"
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9) < 0)) __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);

#line 481 "View.MemoryView"
      __Pyx_XGOTREF(__pyx_t_7);

#line 481 "View.MemoryView"
      __Pyx_XGOTREF(__pyx_t_8);

#line 481 "View.MemoryView"
      __Pyx_XGOTREF(__pyx_t_9);

#line 481 "View.MemoryView"
      __Pyx_XGOTREF(__pyx_t_10);

#line 481 "View.MemoryView"
      __Pyx_XGOTREF(__pyx_t_11);

#line 481 "View.MemoryView"
      __Pyx_XGOTREF(__pyx_t_12);

#line 481 "View.MemoryView"
      __pyx_t_4 = __pyx_lineno; __pyx_t_5 = __pyx_clineno; __pyx_t_6 = __pyx_filename;

#line 481 "View.MemoryView"
      {

#line 481 "View.MemoryView"
        PyMem_Free(__pyx_v_tmp);

#line 481 "View.MemoryView"
      }

#line 481 "View.MemoryView"
      if (PY_MAJOR_VERSION >= 3) {

#line 481 "View.MemoryView"
        __Pyx_XGIVEREF(__pyx_t_10);

#line 481 "View.MemoryView"
        __Pyx_XGIVEREF(__pyx_t_11);

#line 481 "View.MemoryView"
        __Pyx_XGIVEREF(__pyx_t_12);

#line 481 "View.MemoryView"
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);

#line 481 "View.MemoryView"
      }

#line 481 "View.MemoryView"
      __Pyx_XGIVEREF(__pyx_t_7);

#line 481 "View.MemoryView"
      __Pyx_XGIVEREF(__pyx_t_8);

#line 481 "View.MemoryView"
      __Pyx_XGIVEREF(__pyx_t_9);

#line 481 "View.MemoryView"
      __Pyx_ErrRestore(__pyx_t_7, __pyx_t_8, __pyx_t_9);

#line 481 "View.MemoryView"
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;

#line 481 "View.MemoryView"
      __pyx_lineno = __pyx_t_4; __pyx_clineno = __pyx_t_5; __pyx_filename = __pyx_t_6;

#line 481 "View.MemoryView"
      goto __pyx_L1_error;

#line 481 "View.MemoryView"
    }

#line 481 "View.MemoryView"
    __pyx_L7:;

#line 481 "View.MemoryView"
  }

  /* "View.MemoryView":451
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

#line 451 "View.MemoryView"


#line 451 "View.MemoryView"
  /* function exit code */

#line 451 "View.MemoryView"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 451 "View.MemoryView"
  goto __pyx_L0;

#line 451 "View.MemoryView"
  __pyx_L1_error:;

#line 451 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 451 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assign_scalar", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 451 "View.MemoryView"
  __pyx_r = 0;

#line 451 "View.MemoryView"
  __pyx_L0:;

#line 451 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 451 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 451 "View.MemoryView"
  return __pyx_r;

#line 451 "View.MemoryView"
}

/* "View.MemoryView":483
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

#line 483 "View.MemoryView"


#line 483 "View.MemoryView"
static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  char *__pyx_v_itemp
#line 483 "View.MemoryView"
;

#line 483 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 483 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 483 "View.MemoryView"
  char *__pyx_t_1;

#line 483 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 483 "View.MemoryView"
  int __pyx_lineno = 0;

#line 483 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 483 "View.MemoryView"
  int __pyx_clineno = 0;

#line 483 "View.MemoryView"
  __Pyx_RefNannySetupContext("setitem_indexed", 0);

  /* "View.MemoryView":484
 * 
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)             # <<<<<<<<<<<<<<
 *         self.assign_item_from_object(itemp, value)
 * 
 */

#line 484 "View.MemoryView"
  __pyx_t_1 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_index); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(3, 484, __pyx_L1_error)

#line 484 "View.MemoryView"
  __pyx_v_itemp = __pyx_t_1;

  /* "View.MemoryView":485
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */

#line 485 "View.MemoryView"
  __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 485, __pyx_L1_error)

#line 485 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 485 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":483
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

#line 483 "View.MemoryView"


#line 483 "View.MemoryView"
  /* function exit code */

#line 483 "View.MemoryView"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 483 "View.MemoryView"
  goto __pyx_L0;

#line 483 "View.MemoryView"
  __pyx_L1_error:;

#line 483 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 483 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_indexed", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 483 "View.MemoryView"
  __pyx_r = 0;

#line 483 "View.MemoryView"
  __pyx_L0:;

#line 483 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 483 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 483 "View.MemoryView"
  return __pyx_r;

#line 483 "View.MemoryView"
}

/* "View.MemoryView":487
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

#line 487 "View.MemoryView"


#line 487 "View.MemoryView"
static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp) {
  PyObject *__pyx_v_struct = NULL
#line 487 "View.MemoryView"
;
  PyObject *__pyx_v_bytesitem = 0
#line 487 "View.MemoryView"
;
  PyObject *__pyx_v_result = NULL
#line 487 "View.MemoryView"
;

#line 487 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 487 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 487 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 487 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 487 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 487 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 487 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 487 "View.MemoryView"
  PyObject *__pyx_t_6 = NULL;

#line 487 "View.MemoryView"
  PyObject *__pyx_t_7 = NULL;

#line 487 "View.MemoryView"
  int __pyx_t_8;

#line 487 "View.MemoryView"
  PyObject *__pyx_t_9 = NULL;

#line 487 "View.MemoryView"
  size_t __pyx_t_10;

#line 487 "View.MemoryView"
  int __pyx_t_11;

#line 487 "View.MemoryView"
  int __pyx_lineno = 0;

#line 487 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 487 "View.MemoryView"
  int __pyx_clineno = 0;

#line 487 "View.MemoryView"
  __Pyx_RefNannySetupContext("convert_item_to_object", 0);

  /* "View.MemoryView":490
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef bytes bytesitem
 * 
 */

#line 490 "View.MemoryView"
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 490, __pyx_L1_error)

#line 490 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 490 "View.MemoryView"
  __pyx_v_struct = __pyx_t_1;

#line 490 "View.MemoryView"
  __pyx_t_1 = 0;

  /* "View.MemoryView":493
 *         cdef bytes bytesitem
 * 
 *         bytesitem = itemp[:self.view.itemsize]             # <<<<<<<<<<<<<<
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 */

#line 493 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_itemp + 0, __pyx_v_self->view.itemsize - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 493, __pyx_L1_error)

#line 493 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 493 "View.MemoryView"
  __pyx_v_bytesitem = ((PyObject*)__pyx_t_1);

#line 493 "View.MemoryView"
  __pyx_t_1 = 0;

  /* "View.MemoryView":494
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */

#line 494 "View.MemoryView"
  {

#line 494 "View.MemoryView"
    __Pyx_PyThreadState_declare

#line 494 "View.MemoryView"
    __Pyx_PyThreadState_assign

#line 494 "View.MemoryView"
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);

#line 494 "View.MemoryView"
    __Pyx_XGOTREF(__pyx_t_2);

#line 494 "View.MemoryView"
    __Pyx_XGOTREF(__pyx_t_3);

#line 494 "View.MemoryView"
    __Pyx_XGOTREF(__pyx_t_4);

#line 494 "View.MemoryView"
    /*try:*/ {

      /* "View.MemoryView":495
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)             # <<<<<<<<<<<<<<
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")
 */

#line 495 "View.MemoryView"
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_unpack); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 495, __pyx_L3_error)

#line 495 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_5);

#line 495 "View.MemoryView"
      __pyx_t_6 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 495, __pyx_L3_error)

#line 495 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_6);

#line 495 "View.MemoryView"
      __pyx_t_7 = NULL;

#line 495 "View.MemoryView"
      __pyx_t_8 = 0;

#line 495 "View.MemoryView"
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {

#line 495 "View.MemoryView"
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);

#line 495 "View.MemoryView"
        if (likely(__pyx_t_7)) {

#line 495 "View.MemoryView"
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);

#line 495 "View.MemoryView"
          __Pyx_INCREF(__pyx_t_7);

#line 495 "View.MemoryView"
          __Pyx_INCREF(function);

#line 495 "View.MemoryView"
          __Pyx_DECREF_SET(__pyx_t_5, function);

#line 495 "View.MemoryView"
          __pyx_t_8 = 1;

#line 495 "View.MemoryView"
        }

#line 495 "View.MemoryView"
      }

#line 495 "View.MemoryView"
      #if CYTHON_FAST_PYCALL

#line 495 "View.MemoryView"
      if (PyFunction_Check(__pyx_t_5)) {

#line 495 "View.MemoryView"
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_bytesitem};

#line 495 "View.MemoryView"
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 495, __pyx_L3_error)

#line 495 "View.MemoryView"
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 495 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_1);

#line 495 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 495 "View.MemoryView"
      } else

#line 495 "View.MemoryView"
      #endif

#line 495 "View.MemoryView"
      #if CYTHON_FAST_PYCCALL

#line 495 "View.MemoryView"
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {

#line 495 "View.MemoryView"
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_bytesitem};

#line 495 "View.MemoryView"
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 495, __pyx_L3_error)

#line 495 "View.MemoryView"
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 495 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_1);

#line 495 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 495 "View.MemoryView"
      } else

#line 495 "View.MemoryView"
      #endif

#line 495 "View.MemoryView"
      {

#line 495 "View.MemoryView"
        __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(3, 495, __pyx_L3_error)

#line 495 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_9);

#line 495 "View.MemoryView"
        if (__pyx_t_7) {

#line 495 "View.MemoryView"
          __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;

#line 495 "View.MemoryView"
        }

#line 495 "View.MemoryView"
        __Pyx_GIVEREF(__pyx_t_6);

#line 495 "View.MemoryView"
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_t_6);

#line 495 "View.MemoryView"
        __Pyx_INCREF(__pyx_v_bytesitem);

#line 495 "View.MemoryView"
        __Pyx_GIVEREF(__pyx_v_bytesitem);

#line 495 "View.MemoryView"
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_bytesitem);

#line 495 "View.MemoryView"
        __pyx_t_6 = 0;

#line 495 "View.MemoryView"
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 495, __pyx_L3_error)

#line 495 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_1);

#line 495 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 495 "View.MemoryView"
      }

#line 495 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 495 "View.MemoryView"
      __pyx_v_result = __pyx_t_1;

#line 495 "View.MemoryView"
      __pyx_t_1 = 0;

      /* "View.MemoryView":494
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */

#line 494 "View.MemoryView"
    }

    /* "View.MemoryView":499
 *             raise ValueError("Unable to convert item to object")
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */

#line 499 "View.MemoryView"
    /*else:*/ {

#line 499 "View.MemoryView"
      __pyx_t_10 = strlen(__pyx_v_self->view.format); 

#line 499 "View.MemoryView"
      __pyx_t_11 = ((__pyx_t_10 == 1) != 0);

#line 499 "View.MemoryView"
      if (__pyx_t_11) {

        /* "View.MemoryView":500
 *         else:
 *             if len(self.view.format) == 1:
 *                 return result[0]             # <<<<<<<<<<<<<<
 *             return result
 * 
 */

#line 500 "View.MemoryView"
        __Pyx_XDECREF(__pyx_r);

#line 500 "View.MemoryView"
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_result, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 500, __pyx_L5_except_error)

#line 500 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_1);

#line 500 "View.MemoryView"
        __pyx_r = __pyx_t_1;

#line 500 "View.MemoryView"
        __pyx_t_1 = 0;

#line 500 "View.MemoryView"
        goto __pyx_L6_except_return;

        /* "View.MemoryView":499
 *             raise ValueError("Unable to convert item to object")
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */

#line 499 "View.MemoryView"
      }

      /* "View.MemoryView":501
 *             if len(self.view.format) == 1:
 *                 return result[0]
 *             return result             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */

#line 501 "View.MemoryView"
      __Pyx_XDECREF(__pyx_r);

#line 501 "View.MemoryView"
      __Pyx_INCREF(__pyx_v_result);

#line 501 "View.MemoryView"
      __pyx_r = __pyx_v_result;

#line 501 "View.MemoryView"
      goto __pyx_L6_except_return;

#line 501 "View.MemoryView"
    }

#line 501 "View.MemoryView"
    __pyx_L3_error:;

#line 501 "View.MemoryView"
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 501 "View.MemoryView"
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 501 "View.MemoryView"
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 501 "View.MemoryView"
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 501 "View.MemoryView"
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "View.MemoryView":496
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:             # <<<<<<<<<<<<<<
 *             raise ValueError("Unable to convert item to object")
 *         else:
 */

#line 496 "View.MemoryView"
    __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_5, &__pyx_t_9);

#line 496 "View.MemoryView"
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_error); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 496, __pyx_L5_except_error)

#line 496 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_6);

#line 496 "View.MemoryView"
    __pyx_t_8 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_6);

#line 496 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 496 "View.MemoryView"
    __Pyx_ErrRestore(__pyx_t_1, __pyx_t_5, __pyx_t_9);

#line 496 "View.MemoryView"
    __pyx_t_1 = 0; __pyx_t_5 = 0; __pyx_t_9 = 0;

#line 496 "View.MemoryView"
    if (__pyx_t_8) {

#line 496 "View.MemoryView"
      __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 496 "View.MemoryView"
      if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_5, &__pyx_t_1) < 0) __PYX_ERR(3, 496, __pyx_L5_except_error)

#line 496 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_9);

#line 496 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_5);

#line 496 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_1);

      /* "View.MemoryView":497
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")             # <<<<<<<<<<<<<<
 *         else:
 *             if len(self.view.format) == 1:
 */

#line 497 "View.MemoryView"
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 497, __pyx_L5_except_error)

#line 497 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_6);

#line 497 "View.MemoryView"
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);

#line 497 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 497 "View.MemoryView"
      __PYX_ERR(3, 497, __pyx_L5_except_error)

#line 497 "View.MemoryView"
    }

#line 497 "View.MemoryView"
    goto __pyx_L5_except_error;

#line 497 "View.MemoryView"
    __pyx_L5_except_error:;

    /* "View.MemoryView":494
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */

#line 494 "View.MemoryView"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 494 "View.MemoryView"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 494 "View.MemoryView"
    __Pyx_XGIVEREF(__pyx_t_4);

#line 494 "View.MemoryView"
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);

#line 494 "View.MemoryView"
    goto __pyx_L1_error;

#line 494 "View.MemoryView"
    __pyx_L6_except_return:;

#line 494 "View.MemoryView"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 494 "View.MemoryView"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 494 "View.MemoryView"
    __Pyx_XGIVEREF(__pyx_t_4);

#line 494 "View.MemoryView"
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);

#line 494 "View.MemoryView"
    goto __pyx_L0;

#line 494 "View.MemoryView"
  }

  /* "View.MemoryView":487
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

#line 487 "View.MemoryView"


#line 487 "View.MemoryView"
  /* function exit code */

#line 487 "View.MemoryView"
  __pyx_L1_error:;

#line 487 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 487 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 487 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_6);

#line 487 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_7);

#line 487 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_9);

#line 487 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 487 "View.MemoryView"
  __pyx_r = 0;

#line 487 "View.MemoryView"
  __pyx_L0:;

#line 487 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_struct);

#line 487 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_bytesitem);

#line 487 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_result);

#line 487 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 487 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 487 "View.MemoryView"
  return __pyx_r;

#line 487 "View.MemoryView"
}

/* "View.MemoryView":503
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

#line 503 "View.MemoryView"


#line 503 "View.MemoryView"
static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_struct = NULL
#line 503 "View.MemoryView"
;
  char __pyx_v_c
#line 503 "View.MemoryView"
;
  PyObject *__pyx_v_bytesvalue = 0
#line 503 "View.MemoryView"
;
  Py_ssize_t __pyx_v_i
#line 503 "View.MemoryView"
;

#line 503 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 503 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 503 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 503 "View.MemoryView"
  int __pyx_t_2;

#line 503 "View.MemoryView"
  int __pyx_t_3;

#line 503 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 503 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 503 "View.MemoryView"
  PyObject *__pyx_t_6 = NULL;

#line 503 "View.MemoryView"
  int __pyx_t_7;

#line 503 "View.MemoryView"
  PyObject *__pyx_t_8 = NULL;

#line 503 "View.MemoryView"
  Py_ssize_t __pyx_t_9;

#line 503 "View.MemoryView"
  PyObject *__pyx_t_10 = NULL;

#line 503 "View.MemoryView"
  char *__pyx_t_11;

#line 503 "View.MemoryView"
  char *__pyx_t_12;

#line 503 "View.MemoryView"
  char *__pyx_t_13;

#line 503 "View.MemoryView"
  char *__pyx_t_14;

#line 503 "View.MemoryView"
  int __pyx_lineno = 0;

#line 503 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 503 "View.MemoryView"
  int __pyx_clineno = 0;

#line 503 "View.MemoryView"
  __Pyx_RefNannySetupContext("assign_item_from_object", 0);

  /* "View.MemoryView":506
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef char c
 *         cdef bytes bytesvalue
 */

#line 506 "View.MemoryView"
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 506, __pyx_L1_error)

#line 506 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 506 "View.MemoryView"
  __pyx_v_struct = __pyx_t_1;

#line 506 "View.MemoryView"
  __pyx_t_1 = 0;

  /* "View.MemoryView":511
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */

#line 511 "View.MemoryView"
  __pyx_t_2 = PyTuple_Check(__pyx_v_value); 

#line 511 "View.MemoryView"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 511 "View.MemoryView"
  if (__pyx_t_3) {

    /* "View.MemoryView":512
 * 
 *         if isinstance(value, tuple):
 *             bytesvalue = struct.pack(self.view.format, *value)             # <<<<<<<<<<<<<<
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)
 */

#line 512 "View.MemoryView"
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 512, __pyx_L1_error)

#line 512 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_1);

#line 512 "View.MemoryView"
    __pyx_t_4 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 512, __pyx_L1_error)

#line 512 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 512 "View.MemoryView"
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 512, __pyx_L1_error)

#line 512 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_5);

#line 512 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_t_4);

#line 512 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);

#line 512 "View.MemoryView"
    __pyx_t_4 = 0;

#line 512 "View.MemoryView"
    __pyx_t_4 = __Pyx_PySequence_Tuple(__pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 512, __pyx_L1_error)

#line 512 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 512 "View.MemoryView"
    __pyx_t_6 = PyNumber_Add(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 512, __pyx_L1_error)

#line 512 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_6);

#line 512 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 512 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 512 "View.MemoryView"
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 512, __pyx_L1_error)

#line 512 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 512 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 512 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 512 "View.MemoryView"
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(3, 512, __pyx_L1_error)

#line 512 "View.MemoryView"
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_4);

#line 512 "View.MemoryView"
    __pyx_t_4 = 0;

    /* "View.MemoryView":511
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */

#line 511 "View.MemoryView"
    goto __pyx_L3;

#line 511 "View.MemoryView"
  }

  /* "View.MemoryView":514
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)             # <<<<<<<<<<<<<<
 * 
 *         for i, c in enumerate(bytesvalue):
 */

#line 514 "View.MemoryView"
  /*else*/ {

#line 514 "View.MemoryView"
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 514, __pyx_L1_error)

#line 514 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_6);

#line 514 "View.MemoryView"
    __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 514, __pyx_L1_error)

#line 514 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_1);

#line 514 "View.MemoryView"
    __pyx_t_5 = NULL;

#line 514 "View.MemoryView"
    __pyx_t_7 = 0;

#line 514 "View.MemoryView"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {

#line 514 "View.MemoryView"
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);

#line 514 "View.MemoryView"
      if (likely(__pyx_t_5)) {

#line 514 "View.MemoryView"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);

#line 514 "View.MemoryView"
        __Pyx_INCREF(__pyx_t_5);

#line 514 "View.MemoryView"
        __Pyx_INCREF(function);

#line 514 "View.MemoryView"
        __Pyx_DECREF_SET(__pyx_t_6, function);

#line 514 "View.MemoryView"
        __pyx_t_7 = 1;

#line 514 "View.MemoryView"
      }

#line 514 "View.MemoryView"
    }

#line 514 "View.MemoryView"
    #if CYTHON_FAST_PYCALL

#line 514 "View.MemoryView"
    if (PyFunction_Check(__pyx_t_6)) {

#line 514 "View.MemoryView"
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_1, __pyx_v_value};

#line 514 "View.MemoryView"
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 514, __pyx_L1_error)

#line 514 "View.MemoryView"
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 514 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_4);

#line 514 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 514 "View.MemoryView"
    } else

#line 514 "View.MemoryView"
    #endif

#line 514 "View.MemoryView"
    #if CYTHON_FAST_PYCCALL

#line 514 "View.MemoryView"
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {

#line 514 "View.MemoryView"
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_1, __pyx_v_value};

#line 514 "View.MemoryView"
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 514, __pyx_L1_error)

#line 514 "View.MemoryView"
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 514 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_4);

#line 514 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 514 "View.MemoryView"
    } else

#line 514 "View.MemoryView"
    #endif

#line 514 "View.MemoryView"
    {

#line 514 "View.MemoryView"
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 514, __pyx_L1_error)

#line 514 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_8);

#line 514 "View.MemoryView"
      if (__pyx_t_5) {

#line 514 "View.MemoryView"
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5); __pyx_t_5 = NULL;

#line 514 "View.MemoryView"
      }

#line 514 "View.MemoryView"
      __Pyx_GIVEREF(__pyx_t_1);

#line 514 "View.MemoryView"
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_1);

#line 514 "View.MemoryView"
      __Pyx_INCREF(__pyx_v_value);

#line 514 "View.MemoryView"
      __Pyx_GIVEREF(__pyx_v_value);

#line 514 "View.MemoryView"
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_value);

#line 514 "View.MemoryView"
      __pyx_t_1 = 0;

#line 514 "View.MemoryView"
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 514, __pyx_L1_error)

#line 514 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_4);

#line 514 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 514 "View.MemoryView"
    }

#line 514 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 514 "View.MemoryView"
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(3, 514, __pyx_L1_error)

#line 514 "View.MemoryView"
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_4);

#line 514 "View.MemoryView"
    __pyx_t_4 = 0;

#line 514 "View.MemoryView"
  }

#line 514 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":516
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */

#line 516 "View.MemoryView"
  __pyx_t_9 = 0;

#line 516 "View.MemoryView"
  if (unlikely(__pyx_v_bytesvalue == Py_None)) {

#line 516 "View.MemoryView"
    PyErr_SetString(PyExc_TypeError, "'NoneType' is not iterable");

#line 516 "View.MemoryView"
    __PYX_ERR(3, 516, __pyx_L1_error)

#line 516 "View.MemoryView"
  }

#line 516 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_bytesvalue);

#line 516 "View.MemoryView"
  __pyx_t_10 = __pyx_v_bytesvalue;

#line 516 "View.MemoryView"
  __pyx_t_12 = PyBytes_AS_STRING(__pyx_t_10);

#line 516 "View.MemoryView"
  __pyx_t_13 = (__pyx_t_12 + PyBytes_GET_SIZE(__pyx_t_10));

#line 516 "View.MemoryView"
  for (__pyx_t_14 = __pyx_t_12; __pyx_t_14 < __pyx_t_13; __pyx_t_14++) {

#line 516 "View.MemoryView"
    __pyx_t_11 = __pyx_t_14;

#line 516 "View.MemoryView"
    __pyx_v_c = (__pyx_t_11[0]);

    /* "View.MemoryView":517
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */

#line 517 "View.MemoryView"
    __pyx_v_i = __pyx_t_9;

    /* "View.MemoryView":516
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */

#line 516 "View.MemoryView"
    __pyx_t_9 = (__pyx_t_9 + 1);

    /* "View.MemoryView":517
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */

#line 517 "View.MemoryView"
    (__pyx_v_itemp[__pyx_v_i]) = __pyx_v_c;

#line 517 "View.MemoryView"
  }

#line 517 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "View.MemoryView":503
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

#line 503 "View.MemoryView"


#line 503 "View.MemoryView"
  /* function exit code */

#line 503 "View.MemoryView"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 503 "View.MemoryView"
  goto __pyx_L0;

#line 503 "View.MemoryView"
  __pyx_L1_error:;

#line 503 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 503 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_4);

#line 503 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 503 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_6);

#line 503 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_8);

#line 503 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_10);

#line 503 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 503 "View.MemoryView"
  __pyx_r = 0;

#line 503 "View.MemoryView"
  __pyx_L0:;

#line 503 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_struct);

#line 503 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_bytesvalue);

#line 503 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 503 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 503 "View.MemoryView"
  return __pyx_r;

#line 503 "View.MemoryView"
}

/* "View.MemoryView":520
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 */

#line 520 "View.MemoryView"


#line 520 "View.MemoryView"
/* Python wrapper */

#line 520 "View.MemoryView"
static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/

#line 520 "View.MemoryView"
static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {

#line 520 "View.MemoryView"
  int __pyx_r;

#line 520 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 520 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = 
#line 520 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

#line 520 "View.MemoryView"


#line 520 "View.MemoryView"
  /* function exit code */

#line 520 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 520 "View.MemoryView"
  return __pyx_r;

#line 520 "View.MemoryView"
}

#line 520 "View.MemoryView"


#line 520 "View.MemoryView"
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {

#line 520 "View.MemoryView"
  int __pyx_r;

#line 520 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 520 "View.MemoryView"
  int __pyx_t_1;

#line 520 "View.MemoryView"
  int __pyx_t_2;

#line 520 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 520 "View.MemoryView"
  Py_ssize_t *__pyx_t_4;

#line 520 "View.MemoryView"
  char *__pyx_t_5;

#line 520 "View.MemoryView"
  void *__pyx_t_6;

#line 520 "View.MemoryView"
  int __pyx_t_7;

#line 520 "View.MemoryView"
  Py_ssize_t __pyx_t_8;

#line 520 "View.MemoryView"
  int __pyx_lineno = 0;

#line 520 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 520 "View.MemoryView"
  int __pyx_clineno = 0;

#line 520 "View.MemoryView"
  if (__pyx_v_info == NULL) {

#line 520 "View.MemoryView"
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");

#line 520 "View.MemoryView"
    return -1;

#line 520 "View.MemoryView"
  }

#line 520 "View.MemoryView"
  __Pyx_RefNannySetupContext("__getbuffer__", 0);

#line 520 "View.MemoryView"
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);

#line 520 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "View.MemoryView":521
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 */

#line 521 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_flags & PyBUF_WRITABLE) != 0);

#line 521 "View.MemoryView"
  if (__pyx_t_2) {

#line 521 "View.MemoryView"
  } else {

#line 521 "View.MemoryView"
    __pyx_t_1 = __pyx_t_2;

#line 521 "View.MemoryView"
    goto __pyx_L4_bool_binop_done;

#line 521 "View.MemoryView"
  }

#line 521 "View.MemoryView"
  __pyx_t_2 = (__pyx_v_self->view.readonly != 0);

#line 521 "View.MemoryView"
  __pyx_t_1 = __pyx_t_2;

#line 521 "View.MemoryView"
  __pyx_L4_bool_binop_done:;

#line 521 "View.MemoryView"
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":522
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_ND:
 */

#line 522 "View.MemoryView"
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 522, __pyx_L1_error)

#line 522 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 522 "View.MemoryView"
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);

#line 522 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 522 "View.MemoryView"
    __PYX_ERR(3, 522, __pyx_L1_error)

    /* "View.MemoryView":521
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 */

#line 521 "View.MemoryView"
  }

  /* "View.MemoryView":524
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */

#line 524 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_ND) != 0);

#line 524 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":525
 * 
 *         if flags & PyBUF_ND:
 *             info.shape = self.view.shape             # <<<<<<<<<<<<<<
 *         else:
 *             info.shape = NULL
 */

#line 525 "View.MemoryView"
    __pyx_t_4 = __pyx_v_self->view.shape;

#line 525 "View.MemoryView"
    __pyx_v_info->shape = __pyx_t_4;

    /* "View.MemoryView":524
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */

#line 524 "View.MemoryView"
    goto __pyx_L6;

#line 524 "View.MemoryView"
  }

  /* "View.MemoryView":527
 *             info.shape = self.view.shape
 *         else:
 *             info.shape = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_STRIDES:
 */

#line 527 "View.MemoryView"
  /*else*/ {

#line 527 "View.MemoryView"
    __pyx_v_info->shape = NULL;

#line 527 "View.MemoryView"
  }

#line 527 "View.MemoryView"
  __pyx_L6:;

  /* "View.MemoryView":529
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */

#line 529 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_STRIDES) != 0);

#line 529 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":530
 * 
 *         if flags & PyBUF_STRIDES:
 *             info.strides = self.view.strides             # <<<<<<<<<<<<<<
 *         else:
 *             info.strides = NULL
 */

#line 530 "View.MemoryView"
    __pyx_t_4 = __pyx_v_self->view.strides;

#line 530 "View.MemoryView"
    __pyx_v_info->strides = __pyx_t_4;

    /* "View.MemoryView":529
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */

#line 529 "View.MemoryView"
    goto __pyx_L7;

#line 529 "View.MemoryView"
  }

  /* "View.MemoryView":532
 *             info.strides = self.view.strides
 *         else:
 *             info.strides = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_INDIRECT:
 */

#line 532 "View.MemoryView"
  /*else*/ {

#line 532 "View.MemoryView"
    __pyx_v_info->strides = NULL;

#line 532 "View.MemoryView"
  }

#line 532 "View.MemoryView"
  __pyx_L7:;

  /* "View.MemoryView":534
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */

#line 534 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_INDIRECT) != 0);

#line 534 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":535
 * 
 *         if flags & PyBUF_INDIRECT:
 *             info.suboffsets = self.view.suboffsets             # <<<<<<<<<<<<<<
 *         else:
 *             info.suboffsets = NULL
 */

#line 535 "View.MemoryView"
    __pyx_t_4 = __pyx_v_self->view.suboffsets;

#line 535 "View.MemoryView"
    __pyx_v_info->suboffsets = __pyx_t_4;

    /* "View.MemoryView":534
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */

#line 534 "View.MemoryView"
    goto __pyx_L8;

#line 534 "View.MemoryView"
  }

  /* "View.MemoryView":537
 *             info.suboffsets = self.view.suboffsets
 *         else:
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */

#line 537 "View.MemoryView"
  /*else*/ {

#line 537 "View.MemoryView"
    __pyx_v_info->suboffsets = NULL;

#line 537 "View.MemoryView"
  }

#line 537 "View.MemoryView"
  __pyx_L8:;

  /* "View.MemoryView":539
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */

#line 539 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);

#line 539 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":540
 * 
 *         if flags & PyBUF_FORMAT:
 *             info.format = self.view.format             # <<<<<<<<<<<<<<
 *         else:
 *             info.format = NULL
 */

#line 540 "View.MemoryView"
    __pyx_t_5 = __pyx_v_self->view.format;

#line 540 "View.MemoryView"
    __pyx_v_info->format = __pyx_t_5;

    /* "View.MemoryView":539
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */

#line 539 "View.MemoryView"
    goto __pyx_L9;

#line 539 "View.MemoryView"
  }

  /* "View.MemoryView":542
 *             info.format = self.view.format
 *         else:
 *             info.format = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.buf = self.view.buf
 */

#line 542 "View.MemoryView"
  /*else*/ {

#line 542 "View.MemoryView"
    __pyx_v_info->format = NULL;

#line 542 "View.MemoryView"
  }

#line 542 "View.MemoryView"
  __pyx_L9:;

  /* "View.MemoryView":544
 *             info.format = NULL
 * 
 *         info.buf = self.view.buf             # <<<<<<<<<<<<<<
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 */

#line 544 "View.MemoryView"
  __pyx_t_6 = __pyx_v_self->view.buf;

#line 544 "View.MemoryView"
  __pyx_v_info->buf = __pyx_t_6;

  /* "View.MemoryView":545
 * 
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim             # <<<<<<<<<<<<<<
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 */

#line 545 "View.MemoryView"
  __pyx_t_7 = __pyx_v_self->view.ndim;

#line 545 "View.MemoryView"
  __pyx_v_info->ndim = __pyx_t_7;

  /* "View.MemoryView":546
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize             # <<<<<<<<<<<<<<
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 */

#line 546 "View.MemoryView"
  __pyx_t_8 = __pyx_v_self->view.itemsize;

#line 546 "View.MemoryView"
  __pyx_v_info->itemsize = __pyx_t_8;

  /* "View.MemoryView":547
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len             # <<<<<<<<<<<<<<
 *         info.readonly = self.view.readonly
 *         info.obj = self
 */

#line 547 "View.MemoryView"
  __pyx_t_8 = __pyx_v_self->view.len;

#line 547 "View.MemoryView"
  __pyx_v_info->len = __pyx_t_8;

  /* "View.MemoryView":548
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly             # <<<<<<<<<<<<<<
 *         info.obj = self
 * 
 */

#line 548 "View.MemoryView"
  __pyx_t_1 = __pyx_v_self->view.readonly;

#line 548 "View.MemoryView"
  __pyx_v_info->readonly = __pyx_t_1;

  /* "View.MemoryView":549
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")
 */

#line 549 "View.MemoryView"
  __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 549 "View.MemoryView"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));

#line 549 "View.MemoryView"
  __Pyx_GOTREF(__pyx_v_info->obj);

#line 549 "View.MemoryView"
  __Pyx_DECREF(__pyx_v_info->obj);

#line 549 "View.MemoryView"
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":520
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 */

#line 520 "View.MemoryView"


#line 520 "View.MemoryView"
  /* function exit code */

#line 520 "View.MemoryView"
  __pyx_r = 0;

#line 520 "View.MemoryView"
  goto __pyx_L0;

#line 520 "View.MemoryView"
  __pyx_L1_error:;

#line 520 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 520 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 520 "View.MemoryView"
  __pyx_r = -1;

#line 520 "View.MemoryView"
  if (__pyx_v_info->obj != NULL) {

#line 520 "View.MemoryView"
    __Pyx_GOTREF(__pyx_v_info->obj);

#line 520 "View.MemoryView"
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;

#line 520 "View.MemoryView"
  }

#line 520 "View.MemoryView"
  goto __pyx_L2;

#line 520 "View.MemoryView"
  __pyx_L0:;

#line 520 "View.MemoryView"
  if (__pyx_v_info->obj == Py_None) {

#line 520 "View.MemoryView"
    __Pyx_GOTREF(__pyx_v_info->obj);

#line 520 "View.MemoryView"
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;

#line 520 "View.MemoryView"
  }

#line 520 "View.MemoryView"
  __pyx_L2:;

#line 520 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 520 "View.MemoryView"
  return __pyx_r;

#line 520 "View.MemoryView"
}

/* "View.MemoryView":555
 * 
 *     @property
 *     def T(self):             # <<<<<<<<<<<<<<
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 */

#line 555 "View.MemoryView"


#line 555 "View.MemoryView"
/* Python wrapper */

#line 555 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 555 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self) {

#line 555 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 555 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 555 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 555 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 555 "View.MemoryView"


#line 555 "View.MemoryView"
  /* function exit code */

#line 555 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 555 "View.MemoryView"
  return __pyx_r;

#line 555 "View.MemoryView"
}

#line 555 "View.MemoryView"


#line 555 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0
#line 555 "View.MemoryView"
;

#line 555 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 555 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 555 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 555 "View.MemoryView"
  int __pyx_t_2;

#line 555 "View.MemoryView"
  int __pyx_lineno = 0;

#line 555 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 555 "View.MemoryView"
  int __pyx_clineno = 0;

#line 555 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":556
 *     @property
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&result.from_slice)
 *         return result
 */

#line 556 "View.MemoryView"
  __pyx_t_1 = __pyx_memoryview_copy_object(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 556, __pyx_L1_error)

#line 556 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 556 "View.MemoryView"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_memoryviewslice_type))))) __PYX_ERR(3, 556, __pyx_L1_error)

#line 556 "View.MemoryView"
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_1);

#line 556 "View.MemoryView"
  __pyx_t_1 = 0;

  /* "View.MemoryView":557
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */

#line 557 "View.MemoryView"
  __pyx_t_2 = __pyx_memslice_transpose((&__pyx_v_result->from_slice)); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(3, 557, __pyx_L1_error)

  /* "View.MemoryView":558
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 558 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 558 "View.MemoryView"
  __Pyx_INCREF(((PyObject *)__pyx_v_result));

#line 558 "View.MemoryView"
  __pyx_r = ((PyObject *)__pyx_v_result);

#line 558 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":555
 * 
 *     @property
 *     def T(self):             # <<<<<<<<<<<<<<
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 */

#line 555 "View.MemoryView"


#line 555 "View.MemoryView"
  /* function exit code */

#line 555 "View.MemoryView"
  __pyx_L1_error:;

#line 555 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 555 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.T.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 555 "View.MemoryView"
  __pyx_r = NULL;

#line 555 "View.MemoryView"
  __pyx_L0:;

#line 555 "View.MemoryView"
  __Pyx_XDECREF((PyObject *)__pyx_v_result);

#line 555 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 555 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 555 "View.MemoryView"
  return __pyx_r;

#line 555 "View.MemoryView"
}

/* "View.MemoryView":561
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

#line 561 "View.MemoryView"


#line 561 "View.MemoryView"
/* Python wrapper */

#line 561 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 561 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self) {

#line 561 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 561 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 561 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 561 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 561 "View.MemoryView"


#line 561 "View.MemoryView"
  /* function exit code */

#line 561 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 561 "View.MemoryView"
  return __pyx_r;

#line 561 "View.MemoryView"
}

#line 561 "View.MemoryView"


#line 561 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self) {

#line 561 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 561 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 561 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":562
 *     @property
 *     def base(self):
 *         return self.obj             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 562 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 562 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_self->obj);

#line 562 "View.MemoryView"
  __pyx_r = __pyx_v_self->obj;

#line 562 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":561
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

#line 561 "View.MemoryView"


#line 561 "View.MemoryView"
  /* function exit code */

#line 561 "View.MemoryView"
  __pyx_L0:;

#line 561 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 561 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 561 "View.MemoryView"
  return __pyx_r;

#line 561 "View.MemoryView"
}

/* "View.MemoryView":565
 * 
 *     @property
 *     def shape(self):             # <<<<<<<<<<<<<<
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 */

#line 565 "View.MemoryView"


#line 565 "View.MemoryView"
/* Python wrapper */

#line 565 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 565 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self) {

#line 565 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 565 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 565 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 565 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 565 "View.MemoryView"


#line 565 "View.MemoryView"
  /* function exit code */

#line 565 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 565 "View.MemoryView"
  return __pyx_r;

#line 565 "View.MemoryView"
}

#line 565 "View.MemoryView"


#line 565 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_length
#line 565 "View.MemoryView"
;

#line 565 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 565 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 565 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 565 "View.MemoryView"
  Py_ssize_t *__pyx_t_2;

#line 565 "View.MemoryView"
  Py_ssize_t *__pyx_t_3;

#line 565 "View.MemoryView"
  Py_ssize_t *__pyx_t_4;

#line 565 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 565 "View.MemoryView"
  int __pyx_lineno = 0;

#line 565 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 565 "View.MemoryView"
  int __pyx_clineno = 0;

#line 565 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":566
 *     @property
 *     def shape(self):
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 566 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 566 "View.MemoryView"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 566, __pyx_L1_error)

#line 566 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 566 "View.MemoryView"
  __pyx_t_3 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);

#line 566 "View.MemoryView"
  for (__pyx_t_4 = __pyx_v_self->view.shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {

#line 566 "View.MemoryView"
    __pyx_t_2 = __pyx_t_4;

#line 566 "View.MemoryView"
    __pyx_v_length = (__pyx_t_2[0]);

#line 566 "View.MemoryView"
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_length); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 566, __pyx_L1_error)

#line 566 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_5);

#line 566 "View.MemoryView"
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(3, 566, __pyx_L1_error)

#line 566 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 566 "View.MemoryView"
  }

#line 566 "View.MemoryView"
  __pyx_t_5 = PyList_AsTuple(((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 566, __pyx_L1_error)

#line 566 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_5);

#line 566 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 566 "View.MemoryView"
  __pyx_r = __pyx_t_5;

#line 566 "View.MemoryView"
  __pyx_t_5 = 0;

#line 566 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":565
 * 
 *     @property
 *     def shape(self):             # <<<<<<<<<<<<<<
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 */

#line 565 "View.MemoryView"


#line 565 "View.MemoryView"
  /* function exit code */

#line 565 "View.MemoryView"
  __pyx_L1_error:;

#line 565 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 565 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 565 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.shape.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 565 "View.MemoryView"
  __pyx_r = NULL;

#line 565 "View.MemoryView"
  __pyx_L0:;

#line 565 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 565 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 565 "View.MemoryView"
  return __pyx_r;

#line 565 "View.MemoryView"
}

/* "View.MemoryView":569
 * 
 *     @property
 *     def strides(self):             # <<<<<<<<<<<<<<
 *         if self.view.strides == NULL:
 * 
 */

#line 569 "View.MemoryView"


#line 569 "View.MemoryView"
/* Python wrapper */

#line 569 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 569 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self) {

#line 569 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 569 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 569 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 569 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 569 "View.MemoryView"


#line 569 "View.MemoryView"
  /* function exit code */

#line 569 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 569 "View.MemoryView"
  return __pyx_r;

#line 569 "View.MemoryView"
}

#line 569 "View.MemoryView"


#line 569 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_stride
#line 569 "View.MemoryView"
;

#line 569 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 569 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 569 "View.MemoryView"
  int __pyx_t_1;

#line 569 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 569 "View.MemoryView"
  Py_ssize_t *__pyx_t_3;

#line 569 "View.MemoryView"
  Py_ssize_t *__pyx_t_4;

#line 569 "View.MemoryView"
  Py_ssize_t *__pyx_t_5;

#line 569 "View.MemoryView"
  PyObject *__pyx_t_6 = NULL;

#line 569 "View.MemoryView"
  int __pyx_lineno = 0;

#line 569 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 569 "View.MemoryView"
  int __pyx_clineno = 0;

#line 569 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":570
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError("Buffer view does not expose strides")
 */

#line 570 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_self->view.strides == NULL) != 0);

#line 570 "View.MemoryView"
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":572
 *         if self.view.strides == NULL:
 * 
 *             raise ValueError("Buffer view does not expose strides")             # <<<<<<<<<<<<<<
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 */

#line 572 "View.MemoryView"
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 572, __pyx_L1_error)

#line 572 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 572 "View.MemoryView"
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);

#line 572 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 572 "View.MemoryView"
    __PYX_ERR(3, 572, __pyx_L1_error)

    /* "View.MemoryView":570
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError("Buffer view does not expose strides")
 */

#line 570 "View.MemoryView"
  }

  /* "View.MemoryView":574
 *             raise ValueError("Buffer view does not expose strides")
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 574 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 574 "View.MemoryView"
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 574, __pyx_L1_error)

#line 574 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 574 "View.MemoryView"
  __pyx_t_4 = (__pyx_v_self->view.strides + __pyx_v_self->view.ndim);

#line 574 "View.MemoryView"
  for (__pyx_t_5 = __pyx_v_self->view.strides; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {

#line 574 "View.MemoryView"
    __pyx_t_3 = __pyx_t_5;

#line 574 "View.MemoryView"
    __pyx_v_stride = (__pyx_t_3[0]);

#line 574 "View.MemoryView"
    __pyx_t_6 = PyInt_FromSsize_t(__pyx_v_stride); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 574, __pyx_L1_error)

#line 574 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_6);

#line 574 "View.MemoryView"
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(3, 574, __pyx_L1_error)

#line 574 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 574 "View.MemoryView"
  }

#line 574 "View.MemoryView"
  __pyx_t_6 = PyList_AsTuple(((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 574, __pyx_L1_error)

#line 574 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_6);

#line 574 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 574 "View.MemoryView"
  __pyx_r = __pyx_t_6;

#line 574 "View.MemoryView"
  __pyx_t_6 = 0;

#line 574 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":569
 * 
 *     @property
 *     def strides(self):             # <<<<<<<<<<<<<<
 *         if self.view.strides == NULL:
 * 
 */

#line 569 "View.MemoryView"


#line 569 "View.MemoryView"
  /* function exit code */

#line 569 "View.MemoryView"
  __pyx_L1_error:;

#line 569 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 569 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_6);

#line 569 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.strides.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 569 "View.MemoryView"
  __pyx_r = NULL;

#line 569 "View.MemoryView"
  __pyx_L0:;

#line 569 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 569 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 569 "View.MemoryView"
  return __pyx_r;

#line 569 "View.MemoryView"
}

/* "View.MemoryView":577
 * 
 *     @property
 *     def suboffsets(self):             # <<<<<<<<<<<<<<
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim
 */

#line 577 "View.MemoryView"


#line 577 "View.MemoryView"
/* Python wrapper */

#line 577 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 577 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self) {

#line 577 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 577 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 577 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 577 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 577 "View.MemoryView"


#line 577 "View.MemoryView"
  /* function exit code */

#line 577 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 577 "View.MemoryView"
  return __pyx_r;

#line 577 "View.MemoryView"
}

#line 577 "View.MemoryView"


#line 577 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_suboffset
#line 577 "View.MemoryView"
;

#line 577 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 577 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 577 "View.MemoryView"
  int __pyx_t_1;

#line 577 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 577 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 577 "View.MemoryView"
  Py_ssize_t *__pyx_t_4;

#line 577 "View.MemoryView"
  Py_ssize_t *__pyx_t_5;

#line 577 "View.MemoryView"
  Py_ssize_t *__pyx_t_6;

#line 577 "View.MemoryView"
  int __pyx_lineno = 0;

#line 577 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 577 "View.MemoryView"
  int __pyx_clineno = 0;

#line 577 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":578
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */

#line 578 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_self->view.suboffsets == NULL) != 0);

#line 578 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":579
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */

#line 579 "View.MemoryView"
    __Pyx_XDECREF(__pyx_r);

#line 579 "View.MemoryView"
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 579, __pyx_L1_error)

#line 579 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 579 "View.MemoryView"
    __pyx_t_3 = PyNumber_Multiply(__pyx_tuple__17, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 579, __pyx_L1_error)

#line 579 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 579 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 579 "View.MemoryView"
    __pyx_r = __pyx_t_3;

#line 579 "View.MemoryView"
    __pyx_t_3 = 0;

#line 579 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":578
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */

#line 578 "View.MemoryView"
  }

  /* "View.MemoryView":581
 *             return (-1,) * self.view.ndim
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 581 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 581 "View.MemoryView"
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 581, __pyx_L1_error)

#line 581 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 581 "View.MemoryView"
  __pyx_t_5 = (__pyx_v_self->view.suboffsets + __pyx_v_self->view.ndim);

#line 581 "View.MemoryView"
  for (__pyx_t_6 = __pyx_v_self->view.suboffsets; __pyx_t_6 < __pyx_t_5; __pyx_t_6++) {

#line 581 "View.MemoryView"
    __pyx_t_4 = __pyx_t_6;

#line 581 "View.MemoryView"
    __pyx_v_suboffset = (__pyx_t_4[0]);

#line 581 "View.MemoryView"
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_suboffset); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 581, __pyx_L1_error)

#line 581 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 581 "View.MemoryView"
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(3, 581, __pyx_L1_error)

#line 581 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 581 "View.MemoryView"
  }

#line 581 "View.MemoryView"
  __pyx_t_2 = PyList_AsTuple(((PyObject*)__pyx_t_3)); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 581, __pyx_L1_error)

#line 581 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 581 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 581 "View.MemoryView"
  __pyx_r = __pyx_t_2;

#line 581 "View.MemoryView"
  __pyx_t_2 = 0;

#line 581 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":577
 * 
 *     @property
 *     def suboffsets(self):             # <<<<<<<<<<<<<<
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim
 */

#line 577 "View.MemoryView"


#line 577 "View.MemoryView"
  /* function exit code */

#line 577 "View.MemoryView"
  __pyx_L1_error:;

#line 577 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 577 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 577 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.suboffsets.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 577 "View.MemoryView"
  __pyx_r = NULL;

#line 577 "View.MemoryView"
  __pyx_L0:;

#line 577 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 577 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 577 "View.MemoryView"
  return __pyx_r;

#line 577 "View.MemoryView"
}

/* "View.MemoryView":584
 * 
 *     @property
 *     def ndim(self):             # <<<<<<<<<<<<<<
 *         return self.view.ndim
 * 
 */

#line 584 "View.MemoryView"


#line 584 "View.MemoryView"
/* Python wrapper */

#line 584 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 584 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self) {

#line 584 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 584 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 584 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 584 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 584 "View.MemoryView"


#line 584 "View.MemoryView"
  /* function exit code */

#line 584 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 584 "View.MemoryView"
  return __pyx_r;

#line 584 "View.MemoryView"
}

#line 584 "View.MemoryView"


#line 584 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self) {

#line 584 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 584 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 584 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 584 "View.MemoryView"
  int __pyx_lineno = 0;

#line 584 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 584 "View.MemoryView"
  int __pyx_clineno = 0;

#line 584 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":585
 *     @property
 *     def ndim(self):
 *         return self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 585 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 585 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->view.ndim); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 585, __pyx_L1_error)

#line 585 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 585 "View.MemoryView"
  __pyx_r = __pyx_t_1;

#line 585 "View.MemoryView"
  __pyx_t_1 = 0;

#line 585 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":584
 * 
 *     @property
 *     def ndim(self):             # <<<<<<<<<<<<<<
 *         return self.view.ndim
 * 
 */

#line 584 "View.MemoryView"


#line 584 "View.MemoryView"
  /* function exit code */

#line 584 "View.MemoryView"
  __pyx_L1_error:;

#line 584 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 584 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.ndim.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 584 "View.MemoryView"
  __pyx_r = NULL;

#line 584 "View.MemoryView"
  __pyx_L0:;

#line 584 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 584 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 584 "View.MemoryView"
  return __pyx_r;

#line 584 "View.MemoryView"
}

/* "View.MemoryView":588
 * 
 *     @property
 *     def itemsize(self):             # <<<<<<<<<<<<<<
 *         return self.view.itemsize
 * 
 */

#line 588 "View.MemoryView"


#line 588 "View.MemoryView"
/* Python wrapper */

#line 588 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 588 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self) {

#line 588 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 588 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 588 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 588 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 588 "View.MemoryView"


#line 588 "View.MemoryView"
  /* function exit code */

#line 588 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 588 "View.MemoryView"
  return __pyx_r;

#line 588 "View.MemoryView"
}

#line 588 "View.MemoryView"


#line 588 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self) {

#line 588 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 588 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 588 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 588 "View.MemoryView"
  int __pyx_lineno = 0;

#line 588 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 588 "View.MemoryView"
  int __pyx_clineno = 0;

#line 588 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":589
 *     @property
 *     def itemsize(self):
 *         return self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 589 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 589 "View.MemoryView"
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 589, __pyx_L1_error)

#line 589 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 589 "View.MemoryView"
  __pyx_r = __pyx_t_1;

#line 589 "View.MemoryView"
  __pyx_t_1 = 0;

#line 589 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":588
 * 
 *     @property
 *     def itemsize(self):             # <<<<<<<<<<<<<<
 *         return self.view.itemsize
 * 
 */

#line 588 "View.MemoryView"


#line 588 "View.MemoryView"
  /* function exit code */

#line 588 "View.MemoryView"
  __pyx_L1_error:;

#line 588 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 588 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.itemsize.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 588 "View.MemoryView"
  __pyx_r = NULL;

#line 588 "View.MemoryView"
  __pyx_L0:;

#line 588 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 588 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 588 "View.MemoryView"
  return __pyx_r;

#line 588 "View.MemoryView"
}

/* "View.MemoryView":592
 * 
 *     @property
 *     def nbytes(self):             # <<<<<<<<<<<<<<
 *         return self.size * self.view.itemsize
 * 
 */

#line 592 "View.MemoryView"


#line 592 "View.MemoryView"
/* Python wrapper */

#line 592 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 592 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self) {

#line 592 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 592 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 592 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 592 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 592 "View.MemoryView"


#line 592 "View.MemoryView"
  /* function exit code */

#line 592 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 592 "View.MemoryView"
  return __pyx_r;

#line 592 "View.MemoryView"
}

#line 592 "View.MemoryView"


#line 592 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self) {

#line 592 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 592 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 592 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 592 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 592 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 592 "View.MemoryView"
  int __pyx_lineno = 0;

#line 592 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 592 "View.MemoryView"
  int __pyx_clineno = 0;

#line 592 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":593
 *     @property
 *     def nbytes(self):
 *         return self.size * self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 593 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 593 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 593, __pyx_L1_error)

#line 593 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 593 "View.MemoryView"
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 593, __pyx_L1_error)

#line 593 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 593 "View.MemoryView"
  __pyx_t_3 = PyNumber_Multiply(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 593, __pyx_L1_error)

#line 593 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 593 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 593 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 593 "View.MemoryView"
  __pyx_r = __pyx_t_3;

#line 593 "View.MemoryView"
  __pyx_t_3 = 0;

#line 593 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":592
 * 
 *     @property
 *     def nbytes(self):             # <<<<<<<<<<<<<<
 *         return self.size * self.view.itemsize
 * 
 */

#line 592 "View.MemoryView"


#line 592 "View.MemoryView"
  /* function exit code */

#line 592 "View.MemoryView"
  __pyx_L1_error:;

#line 592 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 592 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 592 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 592 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.nbytes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 592 "View.MemoryView"
  __pyx_r = NULL;

#line 592 "View.MemoryView"
  __pyx_L0:;

#line 592 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 592 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 592 "View.MemoryView"
  return __pyx_r;

#line 592 "View.MemoryView"
}

/* "View.MemoryView":596
 * 
 *     @property
 *     def size(self):             # <<<<<<<<<<<<<<
 *         if self._size is None:
 *             result = 1
 */

#line 596 "View.MemoryView"


#line 596 "View.MemoryView"
/* Python wrapper */

#line 596 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 596 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self) {

#line 596 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 596 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 596 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 596 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 596 "View.MemoryView"


#line 596 "View.MemoryView"
  /* function exit code */

#line 596 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 596 "View.MemoryView"
  return __pyx_r;

#line 596 "View.MemoryView"
}

#line 596 "View.MemoryView"


#line 596 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_v_result = NULL
#line 596 "View.MemoryView"
;
  PyObject *__pyx_v_length = NULL
#line 596 "View.MemoryView"
;

#line 596 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 596 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 596 "View.MemoryView"
  int __pyx_t_1;

#line 596 "View.MemoryView"
  int __pyx_t_2;

#line 596 "View.MemoryView"
  Py_ssize_t *__pyx_t_3;

#line 596 "View.MemoryView"
  Py_ssize_t *__pyx_t_4;

#line 596 "View.MemoryView"
  Py_ssize_t *__pyx_t_5;

#line 596 "View.MemoryView"
  PyObject *__pyx_t_6 = NULL;

#line 596 "View.MemoryView"
  int __pyx_lineno = 0;

#line 596 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 596 "View.MemoryView"
  int __pyx_clineno = 0;

#line 596 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":597
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */

#line 597 "View.MemoryView"
  __pyx_t_1 = (__pyx_v_self->_size == Py_None);

#line 597 "View.MemoryView"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 597 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":598
 *     def size(self):
 *         if self._size is None:
 *             result = 1             # <<<<<<<<<<<<<<
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 */

#line 598 "View.MemoryView"
    __Pyx_INCREF(__pyx_int_1);

#line 598 "View.MemoryView"
    __pyx_v_result = __pyx_int_1;

    /* "View.MemoryView":600
 *             result = 1
 * 
 *             for length in self.view.shape[:self.view.ndim]:             # <<<<<<<<<<<<<<
 *                 result *= length
 * 
 */

#line 600 "View.MemoryView"
    __pyx_t_4 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);

#line 600 "View.MemoryView"
    for (__pyx_t_5 = __pyx_v_self->view.shape; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {

#line 600 "View.MemoryView"
      __pyx_t_3 = __pyx_t_5;

#line 600 "View.MemoryView"
      __pyx_t_6 = PyInt_FromSsize_t((__pyx_t_3[0])); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 600, __pyx_L1_error)

#line 600 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_6);

#line 600 "View.MemoryView"
      __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_6);

#line 600 "View.MemoryView"
      __pyx_t_6 = 0;

      /* "View.MemoryView":601
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 *                 result *= length             # <<<<<<<<<<<<<<
 * 
 *             self._size = result
 */

#line 601 "View.MemoryView"
      __pyx_t_6 = PyNumber_InPlaceMultiply(__pyx_v_result, __pyx_v_length); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 601, __pyx_L1_error)

#line 601 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_6);

#line 601 "View.MemoryView"
      __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_6);

#line 601 "View.MemoryView"
      __pyx_t_6 = 0;

#line 601 "View.MemoryView"
    }

    /* "View.MemoryView":603
 *                 result *= length
 * 
 *             self._size = result             # <<<<<<<<<<<<<<
 * 
 *         return self._size
 */

#line 603 "View.MemoryView"
    __Pyx_INCREF(__pyx_v_result);

#line 603 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_v_result);

#line 603 "View.MemoryView"
    __Pyx_GOTREF(__pyx_v_self->_size);

#line 603 "View.MemoryView"
    __Pyx_DECREF(__pyx_v_self->_size);

#line 603 "View.MemoryView"
    __pyx_v_self->_size = __pyx_v_result;

    /* "View.MemoryView":597
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */

#line 597 "View.MemoryView"
  }

  /* "View.MemoryView":605
 *             self._size = result
 * 
 *         return self._size             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */

#line 605 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 605 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_self->_size);

#line 605 "View.MemoryView"
  __pyx_r = __pyx_v_self->_size;

#line 605 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":596
 * 
 *     @property
 *     def size(self):             # <<<<<<<<<<<<<<
 *         if self._size is None:
 *             result = 1
 */

#line 596 "View.MemoryView"


#line 596 "View.MemoryView"
  /* function exit code */

#line 596 "View.MemoryView"
  __pyx_L1_error:;

#line 596 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_6);

#line 596 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 596 "View.MemoryView"
  __pyx_r = NULL;

#line 596 "View.MemoryView"
  __pyx_L0:;

#line 596 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_result);

#line 596 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_length);

#line 596 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 596 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 596 "View.MemoryView"
  return __pyx_r;

#line 596 "View.MemoryView"
}

/* "View.MemoryView":607
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

#line 607 "View.MemoryView"


#line 607 "View.MemoryView"
/* Python wrapper */

#line 607 "View.MemoryView"
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self); /*proto*/

#line 607 "View.MemoryView"
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self) {

#line 607 "View.MemoryView"
  Py_ssize_t __pyx_r;

#line 607 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 607 "View.MemoryView"
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = 
#line 607 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 607 "View.MemoryView"


#line 607 "View.MemoryView"
  /* function exit code */

#line 607 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 607 "View.MemoryView"
  return __pyx_r;

#line 607 "View.MemoryView"
}

#line 607 "View.MemoryView"


#line 607 "View.MemoryView"
static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self) {

#line 607 "View.MemoryView"
  Py_ssize_t __pyx_r;

#line 607 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 607 "View.MemoryView"
  int __pyx_t_1;

#line 607 "View.MemoryView"
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "View.MemoryView":608
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */

#line 608 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_self->view.ndim >= 1) != 0);

#line 608 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":609
 *     def __len__(self):
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]             # <<<<<<<<<<<<<<
 * 
 *         return 0
 */

#line 609 "View.MemoryView"
    __pyx_r = (__pyx_v_self->view.shape[0]);

#line 609 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":608
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */

#line 608 "View.MemoryView"
  }

  /* "View.MemoryView":611
 *             return self.view.shape[0]
 * 
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */

#line 611 "View.MemoryView"
  __pyx_r = 0;

#line 611 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":607
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

#line 607 "View.MemoryView"


#line 607 "View.MemoryView"
  /* function exit code */

#line 607 "View.MemoryView"
  __pyx_L0:;

#line 607 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 607 "View.MemoryView"
  return __pyx_r;

#line 607 "View.MemoryView"
}

/* "View.MemoryView":613
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

#line 613 "View.MemoryView"


#line 613 "View.MemoryView"
/* Python wrapper */

#line 613 "View.MemoryView"
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self); /*proto*/

#line 613 "View.MemoryView"
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self) {

#line 613 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 613 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 613 "View.MemoryView"
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = 
#line 613 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 613 "View.MemoryView"


#line 613 "View.MemoryView"
  /* function exit code */

#line 613 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 613 "View.MemoryView"
  return __pyx_r;

#line 613 "View.MemoryView"
}

#line 613 "View.MemoryView"


#line 613 "View.MemoryView"
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self) {

#line 613 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 613 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 613 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 613 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 613 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 613 "View.MemoryView"
  int __pyx_lineno = 0;

#line 613 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 613 "View.MemoryView"
  int __pyx_clineno = 0;

#line 613 "View.MemoryView"
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "View.MemoryView":614
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */

#line 614 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 614 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 614, __pyx_L1_error)

#line 614 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 614 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 614, __pyx_L1_error)

#line 614 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 614 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 614 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 614, __pyx_L1_error)

#line 614 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 614 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":615
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */

#line 615 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 615, __pyx_L1_error)

#line 615 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

  /* "View.MemoryView":614
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */

#line 614 "View.MemoryView"
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 614, __pyx_L1_error)

#line 614 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 614 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_1);

#line 614 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);

#line 614 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_2);

#line 614 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);

#line 614 "View.MemoryView"
  __pyx_t_1 = 0;

#line 614 "View.MemoryView"
  __pyx_t_2 = 0;

#line 614 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 614, __pyx_L1_error)

#line 614 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 614 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 614 "View.MemoryView"
  __pyx_r = __pyx_t_2;

#line 614 "View.MemoryView"
  __pyx_t_2 = 0;

#line 614 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":613
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

#line 613 "View.MemoryView"


#line 613 "View.MemoryView"
  /* function exit code */

#line 613 "View.MemoryView"
  __pyx_L1_error:;

#line 613 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 613 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 613 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 613 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 613 "View.MemoryView"
  __pyx_r = NULL;

#line 613 "View.MemoryView"
  __pyx_L0:;

#line 613 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 613 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 613 "View.MemoryView"
  return __pyx_r;

#line 613 "View.MemoryView"
}

/* "View.MemoryView":617
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

#line 617 "View.MemoryView"


#line 617 "View.MemoryView"
/* Python wrapper */

#line 617 "View.MemoryView"
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self); /*proto*/

#line 617 "View.MemoryView"
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self) {

#line 617 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 617 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 617 "View.MemoryView"
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = 
#line 617 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 617 "View.MemoryView"


#line 617 "View.MemoryView"
  /* function exit code */

#line 617 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 617 "View.MemoryView"
  return __pyx_r;

#line 617 "View.MemoryView"
}

#line 617 "View.MemoryView"


#line 617 "View.MemoryView"
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self) {

#line 617 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 617 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 617 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 617 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 617 "View.MemoryView"
  int __pyx_lineno = 0;

#line 617 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 617 "View.MemoryView"
  int __pyx_clineno = 0;

#line 617 "View.MemoryView"
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "View.MemoryView":618
 * 
 *     def __str__(self):
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 618 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 618 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 618, __pyx_L1_error)

#line 618 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 618 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 618, __pyx_L1_error)

#line 618 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 618 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 618 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 618, __pyx_L1_error)

#line 618 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 618 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 618 "View.MemoryView"
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 618, __pyx_L1_error)

#line 618 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 618 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_1);

#line 618 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);

#line 618 "View.MemoryView"
  __pyx_t_1 = 0;

#line 618 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_object, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 618, __pyx_L1_error)

#line 618 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 618 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 618 "View.MemoryView"
  __pyx_r = __pyx_t_1;

#line 618 "View.MemoryView"
  __pyx_t_1 = 0;

#line 618 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":617
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

#line 617 "View.MemoryView"


#line 617 "View.MemoryView"
  /* function exit code */

#line 617 "View.MemoryView"
  __pyx_L1_error:;

#line 617 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 617 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 617 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 617 "View.MemoryView"
  __pyx_r = NULL;

#line 617 "View.MemoryView"
  __pyx_L0:;

#line 617 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 617 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 617 "View.MemoryView"
  return __pyx_r;

#line 617 "View.MemoryView"
}

/* "View.MemoryView":621
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

#line 621 "View.MemoryView"


#line 621 "View.MemoryView"
/* Python wrapper */

#line 621 "View.MemoryView"
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 621 "View.MemoryView"
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 621 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 621 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 621 "View.MemoryView"
  __Pyx_RefNannySetupContext("is_c_contig (wrapper)", 0);
  __pyx_r = 
#line 621 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 621 "View.MemoryView"


#line 621 "View.MemoryView"
  /* function exit code */

#line 621 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 621 "View.MemoryView"
  return __pyx_r;

#line 621 "View.MemoryView"
}

#line 621 "View.MemoryView"


#line 621 "View.MemoryView"
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice
#line 621 "View.MemoryView"
;
  __Pyx_memviewslice __pyx_v_tmp
#line 621 "View.MemoryView"
;

#line 621 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 621 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 621 "View.MemoryView"
  __Pyx_memviewslice *__pyx_t_1;

#line 621 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 621 "View.MemoryView"
  int __pyx_lineno = 0;

#line 621 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 621 "View.MemoryView"
  int __pyx_clineno = 0;

#line 621 "View.MemoryView"
  __Pyx_RefNannySetupContext("is_c_contig", 0);

  /* "View.MemoryView":624
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 */

#line 624 "View.MemoryView"
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(3, 624, __pyx_L1_error)

#line 624 "View.MemoryView"
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":625
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def is_f_contig(self):
 */

#line 625 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 625 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'C', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 625, __pyx_L1_error)

#line 625 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 625 "View.MemoryView"
  __pyx_r = __pyx_t_2;

#line 625 "View.MemoryView"
  __pyx_t_2 = 0;

#line 625 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":621
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

#line 621 "View.MemoryView"


#line 621 "View.MemoryView"
  /* function exit code */

#line 621 "View.MemoryView"
  __pyx_L1_error:;

#line 621 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 621 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_c_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 621 "View.MemoryView"
  __pyx_r = NULL;

#line 621 "View.MemoryView"
  __pyx_L0:;

#line 621 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 621 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 621 "View.MemoryView"
  return __pyx_r;

#line 621 "View.MemoryView"
}

/* "View.MemoryView":627
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

#line 627 "View.MemoryView"


#line 627 "View.MemoryView"
/* Python wrapper */

#line 627 "View.MemoryView"
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 627 "View.MemoryView"
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 627 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 627 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 627 "View.MemoryView"
  __Pyx_RefNannySetupContext("is_f_contig (wrapper)", 0);
  __pyx_r = 
#line 627 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 627 "View.MemoryView"


#line 627 "View.MemoryView"
  /* function exit code */

#line 627 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 627 "View.MemoryView"
  return __pyx_r;

#line 627 "View.MemoryView"
}

#line 627 "View.MemoryView"


#line 627 "View.MemoryView"
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice
#line 627 "View.MemoryView"
;
  __Pyx_memviewslice __pyx_v_tmp
#line 627 "View.MemoryView"
;

#line 627 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 627 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 627 "View.MemoryView"
  __Pyx_memviewslice *__pyx_t_1;

#line 627 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 627 "View.MemoryView"
  int __pyx_lineno = 0;

#line 627 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 627 "View.MemoryView"
  int __pyx_clineno = 0;

#line 627 "View.MemoryView"
  __Pyx_RefNannySetupContext("is_f_contig", 0);

  /* "View.MemoryView":630
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 */

#line 630 "View.MemoryView"
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(3, 630, __pyx_L1_error)

#line 630 "View.MemoryView"
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":631
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */

#line 631 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 631 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'F', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 631, __pyx_L1_error)

#line 631 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 631 "View.MemoryView"
  __pyx_r = __pyx_t_2;

#line 631 "View.MemoryView"
  __pyx_t_2 = 0;

#line 631 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":627
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

#line 627 "View.MemoryView"


#line 627 "View.MemoryView"
  /* function exit code */

#line 627 "View.MemoryView"
  __pyx_L1_error:;

#line 627 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 627 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_f_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 627 "View.MemoryView"
  __pyx_r = NULL;

#line 627 "View.MemoryView"
  __pyx_L0:;

#line 627 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 627 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 627 "View.MemoryView"
  return __pyx_r;

#line 627 "View.MemoryView"
}

/* "View.MemoryView":633
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

#line 633 "View.MemoryView"


#line 633 "View.MemoryView"
/* Python wrapper */

#line 633 "View.MemoryView"
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 633 "View.MemoryView"
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 633 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 633 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 633 "View.MemoryView"
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = 
#line 633 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 633 "View.MemoryView"


#line 633 "View.MemoryView"
  /* function exit code */

#line 633 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 633 "View.MemoryView"
  return __pyx_r;

#line 633 "View.MemoryView"
}

#line 633 "View.MemoryView"


#line 633 "View.MemoryView"
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_mslice
#line 633 "View.MemoryView"
;
  int __pyx_v_flags
#line 633 "View.MemoryView"
;

#line 633 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 633 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 633 "View.MemoryView"
  __Pyx_memviewslice __pyx_t_1;

#line 633 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 633 "View.MemoryView"
  int __pyx_lineno = 0;

#line 633 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 633 "View.MemoryView"
  int __pyx_clineno = 0;

#line 633 "View.MemoryView"
  __Pyx_RefNannySetupContext("copy", 0);

  /* "View.MemoryView":635
 *     def copy(self):
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &mslice)
 */

#line 635 "View.MemoryView"
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_F_CONTIGUOUS));

  /* "View.MemoryView":637
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 * 
 *         slice_copy(self, &mslice)             # <<<<<<<<<<<<<<
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,
 *                                    self.view.itemsize,
 */

#line 637 "View.MemoryView"
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_mslice));

  /* "View.MemoryView":638
 * 
 *         slice_copy(self, &mslice)
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                    self.view.itemsize,
 *                                    flags|PyBUF_C_CONTIGUOUS,
 */

#line 638 "View.MemoryView"
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_mslice), ((char *)"c"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_C_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 638, __pyx_L1_error)

#line 638 "View.MemoryView"
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":643
 *                                    self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &mslice)             # <<<<<<<<<<<<<<
 * 
 *     def copy_fortran(self):
 */

#line 643 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 643 "View.MemoryView"
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_mslice)); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 643, __pyx_L1_error)

#line 643 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 643 "View.MemoryView"
  __pyx_r = __pyx_t_2;

#line 643 "View.MemoryView"
  __pyx_t_2 = 0;

#line 643 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":633
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

#line 633 "View.MemoryView"


#line 633 "View.MemoryView"
  /* function exit code */

#line 633 "View.MemoryView"
  __pyx_L1_error:;

#line 633 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 633 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 633 "View.MemoryView"
  __pyx_r = NULL;

#line 633 "View.MemoryView"
  __pyx_L0:;

#line 633 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 633 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 633 "View.MemoryView"
  return __pyx_r;

#line 633 "View.MemoryView"
}

/* "View.MemoryView":645
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

#line 645 "View.MemoryView"


#line 645 "View.MemoryView"
/* Python wrapper */

#line 645 "View.MemoryView"
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 645 "View.MemoryView"
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 645 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 645 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 645 "View.MemoryView"
  __Pyx_RefNannySetupContext("copy_fortran (wrapper)", 0);
  __pyx_r = 
#line 645 "View.MemoryView"
__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 645 "View.MemoryView"


#line 645 "View.MemoryView"
  /* function exit code */

#line 645 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 645 "View.MemoryView"
  return __pyx_r;

#line 645 "View.MemoryView"
}

#line 645 "View.MemoryView"


#line 645 "View.MemoryView"
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_src
#line 645 "View.MemoryView"
;
  __Pyx_memviewslice __pyx_v_dst
#line 645 "View.MemoryView"
;
  int __pyx_v_flags
#line 645 "View.MemoryView"
;

#line 645 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 645 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 645 "View.MemoryView"
  __Pyx_memviewslice __pyx_t_1;

#line 645 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 645 "View.MemoryView"
  int __pyx_lineno = 0;

#line 645 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 645 "View.MemoryView"
  int __pyx_clineno = 0;

#line 645 "View.MemoryView"
  __Pyx_RefNannySetupContext("copy_fortran", 0);

  /* "View.MemoryView":647
 *     def copy_fortran(self):
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &src)
 */

#line 647 "View.MemoryView"
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_C_CONTIGUOUS));

  /* "View.MemoryView":649
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 * 
 *         slice_copy(self, &src)             # <<<<<<<<<<<<<<
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,
 *                                 self.view.itemsize,
 */

#line 649 "View.MemoryView"
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_src));

  /* "View.MemoryView":650
 * 
 *         slice_copy(self, &src)
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                 self.view.itemsize,
 *                                 flags|PyBUF_F_CONTIGUOUS,
 */

#line 650 "View.MemoryView"
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_src), ((char *)"fortran"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_F_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 650, __pyx_L1_error)

#line 650 "View.MemoryView"
  __pyx_v_dst = __pyx_t_1;

  /* "View.MemoryView":655
 *                                 self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &dst)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 655 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 655 "View.MemoryView"
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_dst)); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 655, __pyx_L1_error)

#line 655 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 655 "View.MemoryView"
  __pyx_r = __pyx_t_2;

#line 655 "View.MemoryView"
  __pyx_t_2 = 0;

#line 655 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":645
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

#line 645 "View.MemoryView"


#line 645 "View.MemoryView"
  /* function exit code */

#line 645 "View.MemoryView"
  __pyx_L1_error:;

#line 645 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 645 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy_fortran", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 645 "View.MemoryView"
  __pyx_r = NULL;

#line 645 "View.MemoryView"
  __pyx_L0:;

#line 645 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 645 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 645 "View.MemoryView"
  return __pyx_r;

#line 645 "View.MemoryView"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf___pyx_memoryview___reduce_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 2 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 2, __pyx_L1_error)

#line 2 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 2 "(tree fragment)"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 2 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 2 "(tree fragment)"
  __PYX_ERR(3, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView.memoryview.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
/* Python wrapper */

#line 3 "(tree fragment)"
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 3 "(tree fragment)"
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 3 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 3 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 3 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 3 "(tree fragment)"
__pyx_pf___pyx_memoryview_2__setstate_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
  /* function exit code */

#line 3 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 3 "(tree fragment)"
  return __pyx_r;

#line 3 "(tree fragment)"
}

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {

#line 3 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 3 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 3 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 3 "(tree fragment)"
  int __pyx_lineno = 0;

#line 3 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 3 "(tree fragment)"
  int __pyx_clineno = 0;

#line 3 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __PYX_ERR(3, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
  /* function exit code */

#line 3 "(tree fragment)"
  __pyx_L1_error:;

#line 3 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 3 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 3 "(tree fragment)"
  __pyx_r = NULL;

#line 3 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 3 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 3 "(tree fragment)"
  return __pyx_r;

#line 3 "(tree fragment)"
}

/* "View.MemoryView":659
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

#line 659 "View.MemoryView"


#line 659 "View.MemoryView"
static PyObject *__pyx_memoryview_new(PyObject *__pyx_v_o, int __pyx_v_flags, int __pyx_v_dtype_is_object, __Pyx_TypeInfo *__pyx_v_typeinfo) {
  struct __pyx_memoryview_obj *__pyx_v_result = 0
#line 659 "View.MemoryView"
;

#line 659 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 659 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 659 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 659 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 659 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 659 "View.MemoryView"
  int __pyx_lineno = 0;

#line 659 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 659 "View.MemoryView"
  int __pyx_clineno = 0;

#line 659 "View.MemoryView"
  __Pyx_RefNannySetupContext("memoryview_cwrapper", 0);

  /* "View.MemoryView":660
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)             # <<<<<<<<<<<<<<
 *     result.typeinfo = typeinfo
 *     return result
 */

#line 660 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 660, __pyx_L1_error)

#line 660 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 660 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 660, __pyx_L1_error)

#line 660 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 660 "View.MemoryView"
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 660, __pyx_L1_error)

#line 660 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 660 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_o);

#line 660 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_v_o);

#line 660 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_o);

#line 660 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_1);

#line 660 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);

#line 660 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_2);

#line 660 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);

#line 660 "View.MemoryView"
  __pyx_t_1 = 0;

#line 660 "View.MemoryView"
  __pyx_t_2 = 0;

#line 660 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 660, __pyx_L1_error)

#line 660 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 660 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 660 "View.MemoryView"
  __pyx_v_result = ((struct __pyx_memoryview_obj *)__pyx_t_2);

#line 660 "View.MemoryView"
  __pyx_t_2 = 0;

  /* "View.MemoryView":661
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo             # <<<<<<<<<<<<<<
 *     return result
 * 
 */

#line 661 "View.MemoryView"
  __pyx_v_result->typeinfo = __pyx_v_typeinfo;

  /* "View.MemoryView":662
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_check')
 */

#line 662 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 662 "View.MemoryView"
  __Pyx_INCREF(((PyObject *)__pyx_v_result));

#line 662 "View.MemoryView"
  __pyx_r = ((PyObject *)__pyx_v_result);

#line 662 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":659
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

#line 659 "View.MemoryView"


#line 659 "View.MemoryView"
  /* function exit code */

#line 659 "View.MemoryView"
  __pyx_L1_error:;

#line 659 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 659 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 659 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 659 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 659 "View.MemoryView"
  __pyx_r = 0;

#line 659 "View.MemoryView"
  __pyx_L0:;

#line 659 "View.MemoryView"
  __Pyx_XDECREF((PyObject *)__pyx_v_result);

#line 659 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 659 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 659 "View.MemoryView"
  return __pyx_r;

#line 659 "View.MemoryView"
}

/* "View.MemoryView":665
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

#line 665 "View.MemoryView"


#line 665 "View.MemoryView"
static CYTHON_INLINE int __pyx_memoryview_check(PyObject *__pyx_v_o) {

#line 665 "View.MemoryView"
  int __pyx_r;

#line 665 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 665 "View.MemoryView"
  int __pyx_t_1;

#line 665 "View.MemoryView"
  __Pyx_RefNannySetupContext("memoryview_check", 0);

  /* "View.MemoryView":666
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):
 *     return isinstance(o, memoryview)             # <<<<<<<<<<<<<<
 * 
 * cdef tuple _unellipsify(object index, int ndim):
 */

#line 666 "View.MemoryView"
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_o, __pyx_memoryview_type); 

#line 666 "View.MemoryView"
  __pyx_r = __pyx_t_1;

#line 666 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":665
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

#line 665 "View.MemoryView"


#line 665 "View.MemoryView"
  /* function exit code */

#line 665 "View.MemoryView"
  __pyx_L0:;

#line 665 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 665 "View.MemoryView"
  return __pyx_r;

#line 665 "View.MemoryView"
}

/* "View.MemoryView":668
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

#line 668 "View.MemoryView"


#line 668 "View.MemoryView"
static PyObject *_unellipsify(PyObject *__pyx_v_index, int __pyx_v_ndim) {
  PyObject *__pyx_v_tup = NULL
#line 668 "View.MemoryView"
;
  PyObject *__pyx_v_result = NULL
#line 668 "View.MemoryView"
;
  int __pyx_v_have_slices
#line 668 "View.MemoryView"
;
  int __pyx_v_seen_ellipsis
#line 668 "View.MemoryView"
;
  CYTHON_UNUSED PyObject *__pyx_v_idx = NULL
#line 668 "View.MemoryView"
;
  PyObject *__pyx_v_item = NULL
#line 668 "View.MemoryView"
;
  Py_ssize_t __pyx_v_nslices
#line 668 "View.MemoryView"
;

#line 668 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 668 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 668 "View.MemoryView"
  int __pyx_t_1;

#line 668 "View.MemoryView"
  int __pyx_t_2;

#line 668 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 668 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 668 "View.MemoryView"
  Py_ssize_t __pyx_t_5;

#line 668 "View.MemoryView"
  PyObject *(*__pyx_t_6)(PyObject *);

#line 668 "View.MemoryView"
  PyObject *__pyx_t_7 = NULL;

#line 668 "View.MemoryView"
  Py_ssize_t __pyx_t_8;

#line 668 "View.MemoryView"
  int __pyx_t_9;

#line 668 "View.MemoryView"
  int __pyx_t_10;

#line 668 "View.MemoryView"
  PyObject *__pyx_t_11 = NULL;

#line 668 "View.MemoryView"
  int __pyx_lineno = 0;

#line 668 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 668 "View.MemoryView"
  int __pyx_clineno = 0;

#line 668 "View.MemoryView"
  __Pyx_RefNannySetupContext("_unellipsify", 0);

  /* "View.MemoryView":673
 *     full slices.
 *     """
 *     if not isinstance(index, tuple):             # <<<<<<<<<<<<<<
 *         tup = (index,)
 *     else:
 */

#line 673 "View.MemoryView"
  __pyx_t_1 = PyTuple_Check(__pyx_v_index); 

#line 673 "View.MemoryView"
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);

#line 673 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":674
 *     """
 *     if not isinstance(index, tuple):
 *         tup = (index,)             # <<<<<<<<<<<<<<
 *     else:
 *         tup = index
 */

#line 674 "View.MemoryView"
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 674, __pyx_L1_error)

#line 674 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 674 "View.MemoryView"
    __Pyx_INCREF(__pyx_v_index);

#line 674 "View.MemoryView"
    __Pyx_GIVEREF(__pyx_v_index);

#line 674 "View.MemoryView"
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_index);

#line 674 "View.MemoryView"
    __pyx_v_tup = __pyx_t_3;

#line 674 "View.MemoryView"
    __pyx_t_3 = 0;

    /* "View.MemoryView":673
 *     full slices.
 *     """
 *     if not isinstance(index, tuple):             # <<<<<<<<<<<<<<
 *         tup = (index,)
 *     else:
 */

#line 673 "View.MemoryView"
    goto __pyx_L3;

#line 673 "View.MemoryView"
  }

  /* "View.MemoryView":676
 *         tup = (index,)
 *     else:
 *         tup = index             # <<<<<<<<<<<<<<
 * 
 *     result = []
 */

#line 676 "View.MemoryView"
  /*else*/ {

#line 676 "View.MemoryView"
    __Pyx_INCREF(__pyx_v_index);

#line 676 "View.MemoryView"
    __pyx_v_tup = __pyx_v_index;

#line 676 "View.MemoryView"
  }

#line 676 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":678
 *         tup = index
 * 
 *     result = []             # <<<<<<<<<<<<<<
 *     have_slices = False
 *     seen_ellipsis = False
 */

#line 678 "View.MemoryView"
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 678, __pyx_L1_error)

#line 678 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 678 "View.MemoryView"
  __pyx_v_result = ((PyObject*)__pyx_t_3);

#line 678 "View.MemoryView"
  __pyx_t_3 = 0;

  /* "View.MemoryView":679
 * 
 *     result = []
 *     have_slices = False             # <<<<<<<<<<<<<<
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 */

#line 679 "View.MemoryView"
  __pyx_v_have_slices = 0;

  /* "View.MemoryView":680
 *     result = []
 *     have_slices = False
 *     seen_ellipsis = False             # <<<<<<<<<<<<<<
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 */

#line 680 "View.MemoryView"
  __pyx_v_seen_ellipsis = 0;

  /* "View.MemoryView":681
 *     have_slices = False
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */

#line 681 "View.MemoryView"
  __Pyx_INCREF(__pyx_int_0);

#line 681 "View.MemoryView"
  __pyx_t_3 = __pyx_int_0;

#line 681 "View.MemoryView"
  if (likely(PyList_CheckExact(__pyx_v_tup)) || PyTuple_CheckExact(__pyx_v_tup)) {

#line 681 "View.MemoryView"
    __pyx_t_4 = __pyx_v_tup; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;

#line 681 "View.MemoryView"
    __pyx_t_6 = NULL;

#line 681 "View.MemoryView"
  } else {
    __pyx_t_5 = -1; 
#line 681 "View.MemoryView"
__pyx_t_4 = PyObject_GetIter(__pyx_v_tup); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 681, __pyx_L1_error)

#line 681 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 681 "View.MemoryView"
    __pyx_t_6 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 681, __pyx_L1_error)

#line 681 "View.MemoryView"
  }

#line 681 "View.MemoryView"
  for (;;) {

#line 681 "View.MemoryView"
    if (likely(!__pyx_t_6)) {

#line 681 "View.MemoryView"
      if (likely(PyList_CheckExact(__pyx_t_4))) {

#line 681 "View.MemoryView"
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;

#line 681 "View.MemoryView"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 681 "View.MemoryView"
        __pyx_t_7 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(3, 681, __pyx_L1_error)

#line 681 "View.MemoryView"
        #else

#line 681 "View.MemoryView"
        __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 681, __pyx_L1_error)

#line 681 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_7);

#line 681 "View.MemoryView"
        #endif

#line 681 "View.MemoryView"
      } else {

#line 681 "View.MemoryView"
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;

#line 681 "View.MemoryView"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 681 "View.MemoryView"
        __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(3, 681, __pyx_L1_error)

#line 681 "View.MemoryView"
        #else

#line 681 "View.MemoryView"
        __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 681, __pyx_L1_error)

#line 681 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_7);

#line 681 "View.MemoryView"
        #endif

#line 681 "View.MemoryView"
      }
    } else 
#line 681 "View.MemoryView"
{

#line 681 "View.MemoryView"
      __pyx_t_7 = __pyx_t_6(__pyx_t_4);

#line 681 "View.MemoryView"
      if (unlikely(!__pyx_t_7)) {

#line 681 "View.MemoryView"
        PyObject* exc_type = PyErr_Occurred();

#line 681 "View.MemoryView"
        if (exc_type) {

#line 681 "View.MemoryView"
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();

#line 681 "View.MemoryView"
          else __PYX_ERR(3, 681, __pyx_L1_error)

#line 681 "View.MemoryView"
        }

#line 681 "View.MemoryView"
        break;

#line 681 "View.MemoryView"
      }

#line 681 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_7);

#line 681 "View.MemoryView"
    }

#line 681 "View.MemoryView"
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_7);

#line 681 "View.MemoryView"
    __pyx_t_7 = 0;

#line 681 "View.MemoryView"
    __Pyx_INCREF(__pyx_t_3);

#line 681 "View.MemoryView"
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_3);

#line 681 "View.MemoryView"
    __pyx_t_7 = __Pyx_PyInt_AddObjC(__pyx_t_3, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 681, __pyx_L1_error)

#line 681 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_7);

#line 681 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3);

#line 681 "View.MemoryView"
    __pyx_t_3 = __pyx_t_7;

#line 681 "View.MemoryView"
    __pyx_t_7 = 0;

    /* "View.MemoryView":682
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 */

#line 682 "View.MemoryView"
    __pyx_t_2 = (__pyx_v_item == __pyx_builtin_Ellipsis);

#line 682 "View.MemoryView"
    __pyx_t_1 = (__pyx_t_2 != 0);

#line 682 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":683
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True
 */

#line 683 "View.MemoryView"
      __pyx_t_1 = ((!(__pyx_v_seen_ellipsis != 0)) != 0);

#line 683 "View.MemoryView"
      if (__pyx_t_1) {

        /* "View.MemoryView":684
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))             # <<<<<<<<<<<<<<
 *                 seen_ellipsis = True
 *             else:
 */

#line 684 "View.MemoryView"
        __pyx_t_8 = PyObject_Length(__pyx_v_tup); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(3, 684, __pyx_L1_error)

#line 684 "View.MemoryView"
        __pyx_t_7 = PyList_New(1 * ((((__pyx_v_ndim - __pyx_t_8) + 1)<0) ? 0:((__pyx_v_ndim - __pyx_t_8) + 1))); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 684, __pyx_L1_error)

#line 684 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_7);

#line 684 "View.MemoryView"
        { Py_ssize_t __pyx_temp;

#line 684 "View.MemoryView"
          for (__pyx_temp=0; __pyx_temp < ((__pyx_v_ndim - __pyx_t_8) + 1); __pyx_temp++) {

#line 684 "View.MemoryView"
            __Pyx_INCREF(__pyx_slice__20);

#line 684 "View.MemoryView"
            __Pyx_GIVEREF(__pyx_slice__20);

#line 684 "View.MemoryView"
            PyList_SET_ITEM(__pyx_t_7, __pyx_temp, __pyx_slice__20);

#line 684 "View.MemoryView"
          }

#line 684 "View.MemoryView"
        }

#line 684 "View.MemoryView"
        __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_7); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(3, 684, __pyx_L1_error)

#line 684 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "View.MemoryView":685
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True             # <<<<<<<<<<<<<<
 *             else:
 *                 result.append(slice(None))
 */

#line 685 "View.MemoryView"
        __pyx_v_seen_ellipsis = 1;

        /* "View.MemoryView":683
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True
 */

#line 683 "View.MemoryView"
        goto __pyx_L7;

#line 683 "View.MemoryView"
      }

      /* "View.MemoryView":687
 *                 seen_ellipsis = True
 *             else:
 *                 result.append(slice(None))             # <<<<<<<<<<<<<<
 *             have_slices = True
 *         else:
 */

#line 687 "View.MemoryView"
      /*else*/ {

#line 687 "View.MemoryView"
        __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_slice__20); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(3, 687, __pyx_L1_error)

#line 687 "View.MemoryView"
      }

#line 687 "View.MemoryView"
      __pyx_L7:;

      /* "View.MemoryView":688
 *             else:
 *                 result.append(slice(None))
 *             have_slices = True             # <<<<<<<<<<<<<<
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):
 */

#line 688 "View.MemoryView"
      __pyx_v_have_slices = 1;

      /* "View.MemoryView":682
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 */

#line 682 "View.MemoryView"
      goto __pyx_L6;

#line 682 "View.MemoryView"
    }

    /* "View.MemoryView":690
 *             have_slices = True
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 */

#line 690 "View.MemoryView"
    /*else*/ {

#line 690 "View.MemoryView"
      __pyx_t_2 = PySlice_Check(__pyx_v_item); 

#line 690 "View.MemoryView"
      __pyx_t_10 = ((!(__pyx_t_2 != 0)) != 0);

#line 690 "View.MemoryView"
      if (__pyx_t_10) {

#line 690 "View.MemoryView"
      } else {

#line 690 "View.MemoryView"
        __pyx_t_1 = __pyx_t_10;

#line 690 "View.MemoryView"
        goto __pyx_L9_bool_binop_done;

#line 690 "View.MemoryView"
      }

#line 690 "View.MemoryView"
      __pyx_t_10 = ((!(PyIndex_Check(__pyx_v_item) != 0)) != 0);

#line 690 "View.MemoryView"
      __pyx_t_1 = __pyx_t_10;

#line 690 "View.MemoryView"
      __pyx_L9_bool_binop_done:;

#line 690 "View.MemoryView"
      if (unlikely(__pyx_t_1)) {

        /* "View.MemoryView":691
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):
 *                 raise TypeError("Cannot index with type '%s'" % type(item))             # <<<<<<<<<<<<<<
 * 
 *             have_slices = have_slices or isinstance(item, slice)
 */

#line 691 "View.MemoryView"
        __pyx_t_7 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Cannot_index_with_type_s, ((PyObject *)Py_TYPE(__pyx_v_item))); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 691, __pyx_L1_error)

#line 691 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_7);

#line 691 "View.MemoryView"
        __pyx_t_11 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_7); if (unlikely(!__pyx_t_11)) __PYX_ERR(3, 691, __pyx_L1_error)

#line 691 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_11);

#line 691 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 691 "View.MemoryView"
        __Pyx_Raise(__pyx_t_11, 0, 0, 0);

#line 691 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

#line 691 "View.MemoryView"
        __PYX_ERR(3, 691, __pyx_L1_error)

        /* "View.MemoryView":690
 *             have_slices = True
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 */

#line 690 "View.MemoryView"
      }

      /* "View.MemoryView":693
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 *             have_slices = have_slices or isinstance(item, slice)             # <<<<<<<<<<<<<<
 *             result.append(item)
 * 
 */

#line 693 "View.MemoryView"
      __pyx_t_10 = (__pyx_v_have_slices != 0);

#line 693 "View.MemoryView"
      if (!__pyx_t_10) {

#line 693 "View.MemoryView"
      } else {

#line 693 "View.MemoryView"
        __pyx_t_1 = __pyx_t_10;

#line 693 "View.MemoryView"
        goto __pyx_L11_bool_binop_done;

#line 693 "View.MemoryView"
      }

#line 693 "View.MemoryView"
      __pyx_t_10 = PySlice_Check(__pyx_v_item); 

#line 693 "View.MemoryView"
      __pyx_t_2 = (__pyx_t_10 != 0);

#line 693 "View.MemoryView"
      __pyx_t_1 = __pyx_t_2;

#line 693 "View.MemoryView"
      __pyx_L11_bool_binop_done:;

#line 693 "View.MemoryView"
      __pyx_v_have_slices = __pyx_t_1;

      /* "View.MemoryView":694
 * 
 *             have_slices = have_slices or isinstance(item, slice)
 *             result.append(item)             # <<<<<<<<<<<<<<
 * 
 *     nslices = ndim - len(result)
 */

#line 694 "View.MemoryView"
      __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_v_item); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(3, 694, __pyx_L1_error)

#line 694 "View.MemoryView"
    }

#line 694 "View.MemoryView"
    __pyx_L6:;

    /* "View.MemoryView":681
 *     have_slices = False
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */

#line 681 "View.MemoryView"
  }

#line 681 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 681 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":696
 *             result.append(item)
 * 
 *     nslices = ndim - len(result)             # <<<<<<<<<<<<<<
 *     if nslices:
 *         result.extend([slice(None)] * nslices)
 */

#line 696 "View.MemoryView"
  __pyx_t_5 = PyList_GET_SIZE(__pyx_v_result); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(3, 696, __pyx_L1_error)

#line 696 "View.MemoryView"
  __pyx_v_nslices = (__pyx_v_ndim - __pyx_t_5);

  /* "View.MemoryView":697
 * 
 *     nslices = ndim - len(result)
 *     if nslices:             # <<<<<<<<<<<<<<
 *         result.extend([slice(None)] * nslices)
 * 
 */

#line 697 "View.MemoryView"
  __pyx_t_1 = (__pyx_v_nslices != 0);

#line 697 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":698
 *     nslices = ndim - len(result)
 *     if nslices:
 *         result.extend([slice(None)] * nslices)             # <<<<<<<<<<<<<<
 * 
 *     return have_slices or nslices, tuple(result)
 */

#line 698 "View.MemoryView"
    __pyx_t_3 = PyList_New(1 * ((__pyx_v_nslices<0) ? 0:__pyx_v_nslices)); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 698, __pyx_L1_error)

#line 698 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 698 "View.MemoryView"
    { Py_ssize_t __pyx_temp;

#line 698 "View.MemoryView"
      for (__pyx_temp=0; __pyx_temp < __pyx_v_nslices; __pyx_temp++) {

#line 698 "View.MemoryView"
        __Pyx_INCREF(__pyx_slice__20);

#line 698 "View.MemoryView"
        __Pyx_GIVEREF(__pyx_slice__20);

#line 698 "View.MemoryView"
        PyList_SET_ITEM(__pyx_t_3, __pyx_temp, __pyx_slice__20);

#line 698 "View.MemoryView"
      }

#line 698 "View.MemoryView"
    }

#line 698 "View.MemoryView"
    __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_3); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(3, 698, __pyx_L1_error)

#line 698 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "View.MemoryView":697
 * 
 *     nslices = ndim - len(result)
 *     if nslices:             # <<<<<<<<<<<<<<
 *         result.extend([slice(None)] * nslices)
 * 
 */

#line 697 "View.MemoryView"
  }

  /* "View.MemoryView":700
 *         result.extend([slice(None)] * nslices)
 * 
 *     return have_slices or nslices, tuple(result)             # <<<<<<<<<<<<<<
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 */

#line 700 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 700 "View.MemoryView"
  if (!__pyx_v_have_slices) {

#line 700 "View.MemoryView"
  } else {

#line 700 "View.MemoryView"
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_have_slices); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 700, __pyx_L1_error)

#line 700 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 700 "View.MemoryView"
    __pyx_t_3 = __pyx_t_4;

#line 700 "View.MemoryView"
    __pyx_t_4 = 0;

#line 700 "View.MemoryView"
    goto __pyx_L14_bool_binop_done;

#line 700 "View.MemoryView"
  }

#line 700 "View.MemoryView"
  __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_nslices); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 700, __pyx_L1_error)

#line 700 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_4);

#line 700 "View.MemoryView"
  __pyx_t_3 = __pyx_t_4;

#line 700 "View.MemoryView"
  __pyx_t_4 = 0;

#line 700 "View.MemoryView"
  __pyx_L14_bool_binop_done:;

#line 700 "View.MemoryView"
  __pyx_t_4 = PyList_AsTuple(__pyx_v_result); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 700, __pyx_L1_error)

#line 700 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_4);

#line 700 "View.MemoryView"
  __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(3, 700, __pyx_L1_error)

#line 700 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_11);

#line 700 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_3);

#line 700 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_3);

#line 700 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_4);

#line 700 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_4);

#line 700 "View.MemoryView"
  __pyx_t_3 = 0;

#line 700 "View.MemoryView"
  __pyx_t_4 = 0;

#line 700 "View.MemoryView"
  __pyx_r = ((PyObject*)__pyx_t_11);

#line 700 "View.MemoryView"
  __pyx_t_11 = 0;

#line 700 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":668
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

#line 668 "View.MemoryView"


#line 668 "View.MemoryView"
  /* function exit code */

#line 668 "View.MemoryView"
  __pyx_L1_error:;

#line 668 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 668 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_4);

#line 668 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_7);

#line 668 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_11);

#line 668 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView._unellipsify", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 668 "View.MemoryView"
  __pyx_r = 0;

#line 668 "View.MemoryView"
  __pyx_L0:;

#line 668 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_tup);

#line 668 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_result);

#line 668 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_idx);

#line 668 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_item);

#line 668 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 668 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 668 "View.MemoryView"
  return __pyx_r;

#line 668 "View.MemoryView"
}

/* "View.MemoryView":702
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

#line 702 "View.MemoryView"


#line 702 "View.MemoryView"
static PyObject *assert_direct_dimensions(Py_ssize_t *__pyx_v_suboffsets, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_suboffset
#line 702 "View.MemoryView"
;

#line 702 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 702 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 702 "View.MemoryView"
  Py_ssize_t *__pyx_t_1;

#line 702 "View.MemoryView"
  Py_ssize_t *__pyx_t_2;

#line 702 "View.MemoryView"
  Py_ssize_t *__pyx_t_3;

#line 702 "View.MemoryView"
  int __pyx_t_4;

#line 702 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 702 "View.MemoryView"
  int __pyx_lineno = 0;

#line 702 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 702 "View.MemoryView"
  int __pyx_clineno = 0;

#line 702 "View.MemoryView"
  __Pyx_RefNannySetupContext("assert_direct_dimensions", 0);

  /* "View.MemoryView":703
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")
 */

#line 703 "View.MemoryView"
  __pyx_t_2 = (__pyx_v_suboffsets + __pyx_v_ndim);

#line 703 "View.MemoryView"
  for (__pyx_t_3 = __pyx_v_suboffsets; __pyx_t_3 < __pyx_t_2; __pyx_t_3++) {

#line 703 "View.MemoryView"
    __pyx_t_1 = __pyx_t_3;

#line 703 "View.MemoryView"
    __pyx_v_suboffset = (__pyx_t_1[0]);

    /* "View.MemoryView":704
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("Indirect dimensions not supported")
 * 
 */

#line 704 "View.MemoryView"
    __pyx_t_4 = ((__pyx_v_suboffset >= 0) != 0);

#line 704 "View.MemoryView"
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":705
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 705 "View.MemoryView"
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 705, __pyx_L1_error)

#line 705 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_5);

#line 705 "View.MemoryView"
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);

#line 705 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 705 "View.MemoryView"
      __PYX_ERR(3, 705, __pyx_L1_error)

      /* "View.MemoryView":704
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("Indirect dimensions not supported")
 * 
 */

#line 704 "View.MemoryView"
    }

#line 704 "View.MemoryView"
  }

  /* "View.MemoryView":702
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

#line 702 "View.MemoryView"


#line 702 "View.MemoryView"
  /* function exit code */

#line 702 "View.MemoryView"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 702 "View.MemoryView"
  goto __pyx_L0;

#line 702 "View.MemoryView"
  __pyx_L1_error:;

#line 702 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 702 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.assert_direct_dimensions", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 702 "View.MemoryView"
  __pyx_r = 0;

#line 702 "View.MemoryView"
  __pyx_L0:;

#line 702 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 702 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 702 "View.MemoryView"
  return __pyx_r;

#line 702 "View.MemoryView"
}

/* "View.MemoryView":712
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

#line 712 "View.MemoryView"


#line 712 "View.MemoryView"
static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *__pyx_v_memview, PyObject *__pyx_v_indices) {
  int __pyx_v_new_ndim
#line 712 "View.MemoryView"
;
  int __pyx_v_suboffset_dim
#line 712 "View.MemoryView"
;
  int __pyx_v_dim
#line 712 "View.MemoryView"
;
  __Pyx_memviewslice __pyx_v_src
#line 712 "View.MemoryView"
;
  __Pyx_memviewslice __pyx_v_dst
#line 712 "View.MemoryView"
;
  __Pyx_memviewslice *__pyx_v_p_src
#line 712 "View.MemoryView"
;
  struct __pyx_memoryviewslice_obj *__pyx_v_memviewsliceobj = 0
#line 712 "View.MemoryView"
;
  __Pyx_memviewslice *__pyx_v_p_dst
#line 712 "View.MemoryView"
;
  int *__pyx_v_p_suboffset_dim
#line 712 "View.MemoryView"
;
  Py_ssize_t __pyx_v_start
#line 712 "View.MemoryView"
;
  Py_ssize_t __pyx_v_stop
#line 712 "View.MemoryView"
;
  Py_ssize_t __pyx_v_step
#line 712 "View.MemoryView"
;
  int __pyx_v_have_start
#line 712 "View.MemoryView"
;
  int __pyx_v_have_stop
#line 712 "View.MemoryView"
;
  int __pyx_v_have_step
#line 712 "View.MemoryView"
;
  PyObject *__pyx_v_index = NULL
#line 712 "View.MemoryView"
;

#line 712 "View.MemoryView"
  struct __pyx_memoryview_obj *__pyx_r = NULL;

#line 712 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 712 "View.MemoryView"
  int __pyx_t_1;

#line 712 "View.MemoryView"
  int __pyx_t_2;

#line 712 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 712 "View.MemoryView"
  struct __pyx_memoryview_obj *__pyx_t_4;

#line 712 "View.MemoryView"
  char *__pyx_t_5;

#line 712 "View.MemoryView"
  int __pyx_t_6;

#line 712 "View.MemoryView"
  Py_ssize_t __pyx_t_7;

#line 712 "View.MemoryView"
  PyObject *(*__pyx_t_8)(PyObject *);

#line 712 "View.MemoryView"
  PyObject *__pyx_t_9 = NULL;

#line 712 "View.MemoryView"
  Py_ssize_t __pyx_t_10;

#line 712 "View.MemoryView"
  int __pyx_t_11;

#line 712 "View.MemoryView"
  Py_ssize_t __pyx_t_12;

#line 712 "View.MemoryView"
  int __pyx_lineno = 0;

#line 712 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 712 "View.MemoryView"
  int __pyx_clineno = 0;

#line 712 "View.MemoryView"
  __Pyx_RefNannySetupContext("memview_slice", 0);

  /* "View.MemoryView":713
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim             # <<<<<<<<<<<<<<
 *     cdef bint negative_step
 *     cdef __Pyx_memviewslice src, dst
 */

#line 713 "View.MemoryView"
  __pyx_v_new_ndim = 0;

#line 713 "View.MemoryView"
  __pyx_v_suboffset_dim = -1;

  /* "View.MemoryView":720
 * 
 * 
 *     memset(&dst, 0, sizeof(dst))             # <<<<<<<<<<<<<<
 * 
 *     cdef _memoryviewslice memviewsliceobj
 */

#line 720 "View.MemoryView"
  (void)(memset((&__pyx_v_dst), 0, (sizeof(__pyx_v_dst))));

  /* "View.MemoryView":724
 *     cdef _memoryviewslice memviewsliceobj
 * 
 *     assert memview.view.ndim > 0             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */

#line 724 "View.MemoryView"
  #ifndef CYTHON_WITHOUT_ASSERTIONS

#line 724 "View.MemoryView"
  if (unlikely(!Py_OptimizeFlag)) {

#line 724 "View.MemoryView"
    if (unlikely(!((__pyx_v_memview->view.ndim > 0) != 0))) {

#line 724 "View.MemoryView"
      PyErr_SetNone(PyExc_AssertionError);

#line 724 "View.MemoryView"
      __PYX_ERR(3, 724, __pyx_L1_error)

#line 724 "View.MemoryView"
    }

#line 724 "View.MemoryView"
  }

#line 724 "View.MemoryView"
  #endif

  /* "View.MemoryView":726
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */

#line 726 "View.MemoryView"
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 

#line 726 "View.MemoryView"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 726 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":727
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview             # <<<<<<<<<<<<<<
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 */

#line 727 "View.MemoryView"
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(3, 727, __pyx_L1_error)

#line 727 "View.MemoryView"
    __pyx_t_3 = ((PyObject *)__pyx_v_memview);

#line 727 "View.MemoryView"
    __Pyx_INCREF(__pyx_t_3);

#line 727 "View.MemoryView"
    __pyx_v_memviewsliceobj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_3);

#line 727 "View.MemoryView"
    __pyx_t_3 = 0;

    /* "View.MemoryView":728
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, &src)
 */

#line 728 "View.MemoryView"
    __pyx_v_p_src = (&__pyx_v_memviewsliceobj->from_slice);

    /* "View.MemoryView":726
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */

#line 726 "View.MemoryView"
    goto __pyx_L3;

#line 726 "View.MemoryView"
  }

  /* "View.MemoryView":730
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 *         slice_copy(memview, &src)             # <<<<<<<<<<<<<<
 *         p_src = &src
 * 
 */

#line 730 "View.MemoryView"
  /*else*/ {

#line 730 "View.MemoryView"
    __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_src));

    /* "View.MemoryView":731
 *     else:
 *         slice_copy(memview, &src)
 *         p_src = &src             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 731 "View.MemoryView"
    __pyx_v_p_src = (&__pyx_v_src);

#line 731 "View.MemoryView"
  }

#line 731 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":737
 * 
 * 
 *     dst.memview = p_src.memview             # <<<<<<<<<<<<<<
 *     dst.data = p_src.data
 * 
 */

#line 737 "View.MemoryView"
  __pyx_t_4 = __pyx_v_p_src->memview;

#line 737 "View.MemoryView"
  __pyx_v_dst.memview = __pyx_t_4;

  /* "View.MemoryView":738
 * 
 *     dst.memview = p_src.memview
 *     dst.data = p_src.data             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 738 "View.MemoryView"
  __pyx_t_5 = __pyx_v_p_src->data;

#line 738 "View.MemoryView"
  __pyx_v_dst.data = __pyx_t_5;

  /* "View.MemoryView":743
 * 
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst             # <<<<<<<<<<<<<<
 *     cdef int *p_suboffset_dim = &suboffset_dim
 *     cdef Py_ssize_t start, stop, step
 */

#line 743 "View.MemoryView"
  __pyx_v_p_dst = (&__pyx_v_dst);

  /* "View.MemoryView":744
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst
 *     cdef int *p_suboffset_dim = &suboffset_dim             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t start, stop, step
 *     cdef bint have_start, have_stop, have_step
 */

#line 744 "View.MemoryView"
  __pyx_v_p_suboffset_dim = (&__pyx_v_suboffset_dim);

  /* "View.MemoryView":748
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             slice_memviewslice(
 */

#line 748 "View.MemoryView"
  __pyx_t_6 = 0;

#line 748 "View.MemoryView"
  if (likely(PyList_CheckExact(__pyx_v_indices)) || PyTuple_CheckExact(__pyx_v_indices)) {

#line 748 "View.MemoryView"
    __pyx_t_3 = __pyx_v_indices; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;

#line 748 "View.MemoryView"
    __pyx_t_8 = NULL;

#line 748 "View.MemoryView"
  } else {
    __pyx_t_7 = -1; 
#line 748 "View.MemoryView"
__pyx_t_3 = PyObject_GetIter(__pyx_v_indices); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 748, __pyx_L1_error)

#line 748 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 748 "View.MemoryView"
    __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 748, __pyx_L1_error)

#line 748 "View.MemoryView"
  }

#line 748 "View.MemoryView"
  for (;;) {

#line 748 "View.MemoryView"
    if (likely(!__pyx_t_8)) {

#line 748 "View.MemoryView"
      if (likely(PyList_CheckExact(__pyx_t_3))) {

#line 748 "View.MemoryView"
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;

#line 748 "View.MemoryView"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 748 "View.MemoryView"
        __pyx_t_9 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(3, 748, __pyx_L1_error)

#line 748 "View.MemoryView"
        #else

#line 748 "View.MemoryView"
        __pyx_t_9 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(3, 748, __pyx_L1_error)

#line 748 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_9);

#line 748 "View.MemoryView"
        #endif

#line 748 "View.MemoryView"
      } else {

#line 748 "View.MemoryView"
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_3)) break;

#line 748 "View.MemoryView"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 748 "View.MemoryView"
        __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(3, 748, __pyx_L1_error)

#line 748 "View.MemoryView"
        #else

#line 748 "View.MemoryView"
        __pyx_t_9 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(3, 748, __pyx_L1_error)

#line 748 "View.MemoryView"
        __Pyx_GOTREF(__pyx_t_9);

#line 748 "View.MemoryView"
        #endif

#line 748 "View.MemoryView"
      }
    } else 
#line 748 "View.MemoryView"
{

#line 748 "View.MemoryView"
      __pyx_t_9 = __pyx_t_8(__pyx_t_3);

#line 748 "View.MemoryView"
      if (unlikely(!__pyx_t_9)) {

#line 748 "View.MemoryView"
        PyObject* exc_type = PyErr_Occurred();

#line 748 "View.MemoryView"
        if (exc_type) {

#line 748 "View.MemoryView"
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();

#line 748 "View.MemoryView"
          else __PYX_ERR(3, 748, __pyx_L1_error)

#line 748 "View.MemoryView"
        }

#line 748 "View.MemoryView"
        break;

#line 748 "View.MemoryView"
      }

#line 748 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_9);

#line 748 "View.MemoryView"
    }

#line 748 "View.MemoryView"
    __Pyx_XDECREF_SET(__pyx_v_index, __pyx_t_9);

#line 748 "View.MemoryView"
    __pyx_t_9 = 0;

#line 748 "View.MemoryView"
    __pyx_v_dim = __pyx_t_6;

#line 748 "View.MemoryView"
    __pyx_t_6 = (__pyx_t_6 + 1);

    /* "View.MemoryView":749
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             slice_memviewslice(
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 */

#line 749 "View.MemoryView"
    __pyx_t_2 = (PyIndex_Check(__pyx_v_index) != 0);

#line 749 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":753
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 *                 index, 0, 0, # start, stop, step             # <<<<<<<<<<<<<<
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 */

#line 753 "View.MemoryView"
      __pyx_t_10 = __Pyx_PyIndex_AsSsize_t(__pyx_v_index); if (unlikely((__pyx_t_10 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(3, 753, __pyx_L1_error)

      /* "View.MemoryView":750
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */

#line 750 "View.MemoryView"
      __pyx_t_11 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_t_10, 0, 0, 0, 0, 0, 0); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(3, 750, __pyx_L1_error)

      /* "View.MemoryView":749
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             slice_memviewslice(
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 */

#line 749 "View.MemoryView"
      goto __pyx_L6;

#line 749 "View.MemoryView"
    }

    /* "View.MemoryView":756
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */

#line 756 "View.MemoryView"
    __pyx_t_2 = (__pyx_v_index == Py_None);

#line 756 "View.MemoryView"
    __pyx_t_1 = (__pyx_t_2 != 0);

#line 756 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":757
 *                 False)
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1             # <<<<<<<<<<<<<<
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 */

#line 757 "View.MemoryView"
      (__pyx_v_p_dst->shape[__pyx_v_new_ndim]) = 1;

      /* "View.MemoryView":758
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0             # <<<<<<<<<<<<<<
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1
 */

#line 758 "View.MemoryView"
      (__pyx_v_p_dst->strides[__pyx_v_new_ndim]) = 0;

      /* "View.MemoryView":759
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1             # <<<<<<<<<<<<<<
 *             new_ndim += 1
 *         else:
 */

#line 759 "View.MemoryView"
      (__pyx_v_p_dst->suboffsets[__pyx_v_new_ndim]) = -1L;

      /* "View.MemoryView":760
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 *         else:
 *             start = index.start or 0
 */

#line 760 "View.MemoryView"
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);

      /* "View.MemoryView":756
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */

#line 756 "View.MemoryView"
      goto __pyx_L6;

#line 756 "View.MemoryView"
    }

    /* "View.MemoryView":762
 *             new_ndim += 1
 *         else:
 *             start = index.start or 0             # <<<<<<<<<<<<<<
 *             stop = index.stop or 0
 *             step = index.step or 0
 */

#line 762 "View.MemoryView"
    /*else*/ {

#line 762 "View.MemoryView"
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_9)) __PYX_ERR(3, 762, __pyx_L1_error)

#line 762 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_9);

#line 762 "View.MemoryView"
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(3, 762, __pyx_L1_error)

#line 762 "View.MemoryView"
      if (!__pyx_t_1) {

#line 762 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 762 "View.MemoryView"
      } else {

#line 762 "View.MemoryView"
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(3, 762, __pyx_L1_error)

#line 762 "View.MemoryView"
        __pyx_t_10 = __pyx_t_12;

#line 762 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 762 "View.MemoryView"
        goto __pyx_L7_bool_binop_done;

#line 762 "View.MemoryView"
      }

#line 762 "View.MemoryView"
      __pyx_t_10 = 0;

#line 762 "View.MemoryView"
      __pyx_L7_bool_binop_done:;

#line 762 "View.MemoryView"
      __pyx_v_start = __pyx_t_10;

      /* "View.MemoryView":763
 *         else:
 *             start = index.start or 0
 *             stop = index.stop or 0             # <<<<<<<<<<<<<<
 *             step = index.step or 0
 * 
 */

#line 763 "View.MemoryView"
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_9)) __PYX_ERR(3, 763, __pyx_L1_error)

#line 763 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_9);

#line 763 "View.MemoryView"
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(3, 763, __pyx_L1_error)

#line 763 "View.MemoryView"
      if (!__pyx_t_1) {

#line 763 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 763 "View.MemoryView"
      } else {

#line 763 "View.MemoryView"
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(3, 763, __pyx_L1_error)

#line 763 "View.MemoryView"
        __pyx_t_10 = __pyx_t_12;

#line 763 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 763 "View.MemoryView"
        goto __pyx_L9_bool_binop_done;

#line 763 "View.MemoryView"
      }

#line 763 "View.MemoryView"
      __pyx_t_10 = 0;

#line 763 "View.MemoryView"
      __pyx_L9_bool_binop_done:;

#line 763 "View.MemoryView"
      __pyx_v_stop = __pyx_t_10;

      /* "View.MemoryView":764
 *             start = index.start or 0
 *             stop = index.stop or 0
 *             step = index.step or 0             # <<<<<<<<<<<<<<
 * 
 *             have_start = index.start is not None
 */

#line 764 "View.MemoryView"
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_9)) __PYX_ERR(3, 764, __pyx_L1_error)

#line 764 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_9);

#line 764 "View.MemoryView"
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(3, 764, __pyx_L1_error)

#line 764 "View.MemoryView"
      if (!__pyx_t_1) {

#line 764 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 764 "View.MemoryView"
      } else {

#line 764 "View.MemoryView"
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(3, 764, __pyx_L1_error)

#line 764 "View.MemoryView"
        __pyx_t_10 = __pyx_t_12;

#line 764 "View.MemoryView"
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 764 "View.MemoryView"
        goto __pyx_L11_bool_binop_done;

#line 764 "View.MemoryView"
      }

#line 764 "View.MemoryView"
      __pyx_t_10 = 0;

#line 764 "View.MemoryView"
      __pyx_L11_bool_binop_done:;

#line 764 "View.MemoryView"
      __pyx_v_step = __pyx_t_10;

      /* "View.MemoryView":766
 *             step = index.step or 0
 * 
 *             have_start = index.start is not None             # <<<<<<<<<<<<<<
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None
 */

#line 766 "View.MemoryView"
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_9)) __PYX_ERR(3, 766, __pyx_L1_error)

#line 766 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_9);

#line 766 "View.MemoryView"
      __pyx_t_1 = (__pyx_t_9 != Py_None);

#line 766 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 766 "View.MemoryView"
      __pyx_v_have_start = __pyx_t_1;

      /* "View.MemoryView":767
 * 
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None             # <<<<<<<<<<<<<<
 *             have_step = index.step is not None
 * 
 */

#line 767 "View.MemoryView"
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_9)) __PYX_ERR(3, 767, __pyx_L1_error)

#line 767 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_9);

#line 767 "View.MemoryView"
      __pyx_t_1 = (__pyx_t_9 != Py_None);

#line 767 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 767 "View.MemoryView"
      __pyx_v_have_stop = __pyx_t_1;

      /* "View.MemoryView":768
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None             # <<<<<<<<<<<<<<
 * 
 *             slice_memviewslice(
 */

#line 768 "View.MemoryView"
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_9)) __PYX_ERR(3, 768, __pyx_L1_error)

#line 768 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_9);

#line 768 "View.MemoryView"
      __pyx_t_1 = (__pyx_t_9 != Py_None);

#line 768 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 768 "View.MemoryView"
      __pyx_v_have_step = __pyx_t_1;

      /* "View.MemoryView":770
 *             have_step = index.step is not None
 * 
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */

#line 770 "View.MemoryView"
      __pyx_t_11 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_v_start, __pyx_v_stop, __pyx_v_step, __pyx_v_have_start, __pyx_v_have_stop, __pyx_v_have_step, 1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(3, 770, __pyx_L1_error)

      /* "View.MemoryView":776
 *                 have_start, have_stop, have_step,
 *                 True)
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */

#line 776 "View.MemoryView"
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);

#line 776 "View.MemoryView"
    }

#line 776 "View.MemoryView"
    __pyx_L6:;

    /* "View.MemoryView":748
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             slice_memviewslice(
 */

#line 748 "View.MemoryView"
  }

#line 748 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":778
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */

#line 778 "View.MemoryView"
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 

#line 778 "View.MemoryView"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 778 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":779
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */

#line 779 "View.MemoryView"
    __Pyx_XDECREF(((PyObject *)__pyx_r));

    /* "View.MemoryView":780
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_dtype_func,
 *                                     memview.dtype_is_object)
 */

#line 780 "View.MemoryView"
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(3, 780, __pyx_L1_error) }

    /* "View.MemoryView":781
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 *     else:
 */

#line 781 "View.MemoryView"
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(3, 781, __pyx_L1_error) }

    /* "View.MemoryView":779
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */

#line 779 "View.MemoryView"
    __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, __pyx_v_memviewsliceobj->to_object_func, __pyx_v_memviewsliceobj->to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 779, __pyx_L1_error)

#line 779 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 779 "View.MemoryView"
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_memoryview_type))))) __PYX_ERR(3, 779, __pyx_L1_error)

#line 779 "View.MemoryView"
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_3);

#line 779 "View.MemoryView"
    __pyx_t_3 = 0;

#line 779 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":778
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */

#line 778 "View.MemoryView"
  }

  /* "View.MemoryView":784
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */

#line 784 "View.MemoryView"
  /*else*/ {

#line 784 "View.MemoryView"
    __Pyx_XDECREF(((PyObject *)__pyx_r));

    /* "View.MemoryView":785
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,
 *                                     memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 785 "View.MemoryView"
    __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, NULL, NULL, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 784, __pyx_L1_error)

#line 785 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

    /* "View.MemoryView":784
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */

#line 784 "View.MemoryView"
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_memoryview_type))))) __PYX_ERR(3, 784, __pyx_L1_error)

#line 784 "View.MemoryView"
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_3);

#line 784 "View.MemoryView"
    __pyx_t_3 = 0;

#line 784 "View.MemoryView"
    goto __pyx_L0;

#line 784 "View.MemoryView"
  }

  /* "View.MemoryView":712
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

#line 712 "View.MemoryView"


#line 712 "View.MemoryView"
  /* function exit code */

#line 712 "View.MemoryView"
  __pyx_L1_error:;

#line 712 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 712 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_9);

#line 712 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memview_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 712 "View.MemoryView"
  __pyx_r = 0;

#line 712 "View.MemoryView"
  __pyx_L0:;

#line 712 "View.MemoryView"
  __Pyx_XDECREF((PyObject *)__pyx_v_memviewsliceobj);

#line 712 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_index);

#line 712 "View.MemoryView"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 712 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 712 "View.MemoryView"
  return __pyx_r;

#line 712 "View.MemoryView"
}

/* "View.MemoryView":809
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

#line 809 "View.MemoryView"


#line 809 "View.MemoryView"
static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *__pyx_v_dst, Py_ssize_t __pyx_v_shape, Py_ssize_t __pyx_v_stride, Py_ssize_t __pyx_v_suboffset, int __pyx_v_dim, int __pyx_v_new_ndim, int *__pyx_v_suboffset_dim, Py_ssize_t __pyx_v_start, Py_ssize_t __pyx_v_stop, Py_ssize_t __pyx_v_step, int __pyx_v_have_start, int __pyx_v_have_stop, int __pyx_v_have_step, int __pyx_v_is_slice) {
  Py_ssize_t __pyx_v_new_shape
#line 809 "View.MemoryView"
;
  int __pyx_v_negative_step
#line 809 "View.MemoryView"
;

#line 809 "View.MemoryView"
  int __pyx_r;

#line 809 "View.MemoryView"
  int __pyx_t_1;

#line 809 "View.MemoryView"
  int __pyx_t_2;

#line 809 "View.MemoryView"
  int __pyx_t_3;

#line 809 "View.MemoryView"
  int __pyx_lineno = 0;

#line 809 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 809 "View.MemoryView"
  int __pyx_clineno = 0;

  /* "View.MemoryView":829
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */

#line 829 "View.MemoryView"
  __pyx_t_1 = ((!(__pyx_v_is_slice != 0)) != 0);

#line 829 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":831
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */

#line 831 "View.MemoryView"
    __pyx_t_1 = ((__pyx_v_start < 0) != 0);

#line 831 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":832
 * 
 *         if start < 0:
 *             start += shape             # <<<<<<<<<<<<<<
 *         if not 0 <= start < shape:
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 */

#line 832 "View.MemoryView"
      __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

      /* "View.MemoryView":831
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */

#line 831 "View.MemoryView"
    }

    /* "View.MemoryView":833
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */

#line 833 "View.MemoryView"
    __pyx_t_1 = (0 <= __pyx_v_start);

#line 833 "View.MemoryView"
    if (__pyx_t_1) {

#line 833 "View.MemoryView"
      __pyx_t_1 = (__pyx_v_start < __pyx_v_shape);

#line 833 "View.MemoryView"
    }

#line 833 "View.MemoryView"
    __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);

#line 833 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":834
 *             start += shape
 *         if not 0 <= start < shape:
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)             # <<<<<<<<<<<<<<
 *     else:
 * 
 */

#line 834 "View.MemoryView"
      __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_IndexError, ((char *)"Index out of bounds (axis %d)"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(3, 834, __pyx_L1_error)

      /* "View.MemoryView":833
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */

#line 833 "View.MemoryView"
    }

    /* "View.MemoryView":829
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */

#line 829 "View.MemoryView"
    goto __pyx_L3;

#line 829 "View.MemoryView"
  }

  /* "View.MemoryView":837
 *     else:
 * 
 *         negative_step = have_step != 0 and step < 0             # <<<<<<<<<<<<<<
 * 
 *         if have_step and step == 0:
 */

#line 837 "View.MemoryView"
  /*else*/ {

#line 837 "View.MemoryView"
    __pyx_t_1 = ((__pyx_v_have_step != 0) != 0);

#line 837 "View.MemoryView"
    if (__pyx_t_1) {

#line 837 "View.MemoryView"
    } else {

#line 837 "View.MemoryView"
      __pyx_t_2 = __pyx_t_1;

#line 837 "View.MemoryView"
      goto __pyx_L6_bool_binop_done;

#line 837 "View.MemoryView"
    }

#line 837 "View.MemoryView"
    __pyx_t_1 = ((__pyx_v_step < 0) != 0);

#line 837 "View.MemoryView"
    __pyx_t_2 = __pyx_t_1;

#line 837 "View.MemoryView"
    __pyx_L6_bool_binop_done:;

#line 837 "View.MemoryView"
    __pyx_v_negative_step = __pyx_t_2;

    /* "View.MemoryView":839
 *         negative_step = have_step != 0 and step < 0
 * 
 *         if have_step and step == 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)
 * 
 */

#line 839 "View.MemoryView"
    __pyx_t_1 = (__pyx_v_have_step != 0);

#line 839 "View.MemoryView"
    if (__pyx_t_1) {

#line 839 "View.MemoryView"
    } else {

#line 839 "View.MemoryView"
      __pyx_t_2 = __pyx_t_1;

#line 839 "View.MemoryView"
      goto __pyx_L9_bool_binop_done;

#line 839 "View.MemoryView"
    }

#line 839 "View.MemoryView"
    __pyx_t_1 = ((__pyx_v_step == 0) != 0);

#line 839 "View.MemoryView"
    __pyx_t_2 = __pyx_t_1;

#line 839 "View.MemoryView"
    __pyx_L9_bool_binop_done:;

#line 839 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":840
 * 
 *         if have_step and step == 0:
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 840 "View.MemoryView"
      __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_ValueError, ((char *)"Step may not be zero (axis %d)"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(3, 840, __pyx_L1_error)

      /* "View.MemoryView":839
 *         negative_step = have_step != 0 and step < 0
 * 
 *         if have_step and step == 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)
 * 
 */

#line 839 "View.MemoryView"
    }

    /* "View.MemoryView":843
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */

#line 843 "View.MemoryView"
    __pyx_t_2 = (__pyx_v_have_start != 0);

#line 843 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":844
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */

#line 844 "View.MemoryView"
      __pyx_t_2 = ((__pyx_v_start < 0) != 0);

#line 844 "View.MemoryView"
      if (__pyx_t_2) {

        /* "View.MemoryView":845
 *         if have_start:
 *             if start < 0:
 *                 start += shape             # <<<<<<<<<<<<<<
 *                 if start < 0:
 *                     start = 0
 */

#line 845 "View.MemoryView"
        __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

        /* "View.MemoryView":846
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */

#line 846 "View.MemoryView"
        __pyx_t_2 = ((__pyx_v_start < 0) != 0);

#line 846 "View.MemoryView"
        if (__pyx_t_2) {

          /* "View.MemoryView":847
 *                 start += shape
 *                 if start < 0:
 *                     start = 0             # <<<<<<<<<<<<<<
 *             elif start >= shape:
 *                 if negative_step:
 */

#line 847 "View.MemoryView"
          __pyx_v_start = 0;

          /* "View.MemoryView":846
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */

#line 846 "View.MemoryView"
        }

        /* "View.MemoryView":844
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */

#line 844 "View.MemoryView"
        goto __pyx_L12;

#line 844 "View.MemoryView"
      }

      /* "View.MemoryView":848
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */

#line 848 "View.MemoryView"
      __pyx_t_2 = ((__pyx_v_start >= __pyx_v_shape) != 0);

#line 848 "View.MemoryView"
      if (__pyx_t_2) {

        /* "View.MemoryView":849
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */

#line 849 "View.MemoryView"
        __pyx_t_2 = (__pyx_v_negative_step != 0);

#line 849 "View.MemoryView"
        if (__pyx_t_2) {

          /* "View.MemoryView":850
 *             elif start >= shape:
 *                 if negative_step:
 *                     start = shape - 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     start = shape
 */

#line 850 "View.MemoryView"
          __pyx_v_start = (__pyx_v_shape - 1);

          /* "View.MemoryView":849
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */

#line 849 "View.MemoryView"
          goto __pyx_L14;

#line 849 "View.MemoryView"
        }

        /* "View.MemoryView":852
 *                     start = shape - 1
 *                 else:
 *                     start = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */

#line 852 "View.MemoryView"
        /*else*/ {

#line 852 "View.MemoryView"
          __pyx_v_start = __pyx_v_shape;

#line 852 "View.MemoryView"
        }

#line 852 "View.MemoryView"
        __pyx_L14:;

        /* "View.MemoryView":848
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */

#line 848 "View.MemoryView"
      }

#line 848 "View.MemoryView"
      __pyx_L12:;

      /* "View.MemoryView":843
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */

#line 843 "View.MemoryView"
      goto __pyx_L11;

#line 843 "View.MemoryView"
    }

    /* "View.MemoryView":854
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */

#line 854 "View.MemoryView"
    /*else*/ {

#line 854 "View.MemoryView"
      __pyx_t_2 = (__pyx_v_negative_step != 0);

#line 854 "View.MemoryView"
      if (__pyx_t_2) {

        /* "View.MemoryView":855
 *         else:
 *             if negative_step:
 *                 start = shape - 1             # <<<<<<<<<<<<<<
 *             else:
 *                 start = 0
 */

#line 855 "View.MemoryView"
        __pyx_v_start = (__pyx_v_shape - 1);

        /* "View.MemoryView":854
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */

#line 854 "View.MemoryView"
        goto __pyx_L15;

#line 854 "View.MemoryView"
      }

      /* "View.MemoryView":857
 *                 start = shape - 1
 *             else:
 *                 start = 0             # <<<<<<<<<<<<<<
 * 
 *         if have_stop:
 */

#line 857 "View.MemoryView"
      /*else*/ {

#line 857 "View.MemoryView"
        __pyx_v_start = 0;

#line 857 "View.MemoryView"
      }

#line 857 "View.MemoryView"
      __pyx_L15:;

#line 857 "View.MemoryView"
    }

#line 857 "View.MemoryView"
    __pyx_L11:;

    /* "View.MemoryView":859
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */

#line 859 "View.MemoryView"
    __pyx_t_2 = (__pyx_v_have_stop != 0);

#line 859 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":860
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */

#line 860 "View.MemoryView"
      __pyx_t_2 = ((__pyx_v_stop < 0) != 0);

#line 860 "View.MemoryView"
      if (__pyx_t_2) {

        /* "View.MemoryView":861
 *         if have_stop:
 *             if stop < 0:
 *                 stop += shape             # <<<<<<<<<<<<<<
 *                 if stop < 0:
 *                     stop = 0
 */

#line 861 "View.MemoryView"
        __pyx_v_stop = (__pyx_v_stop + __pyx_v_shape);

        /* "View.MemoryView":862
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */

#line 862 "View.MemoryView"
        __pyx_t_2 = ((__pyx_v_stop < 0) != 0);

#line 862 "View.MemoryView"
        if (__pyx_t_2) {

          /* "View.MemoryView":863
 *                 stop += shape
 *                 if stop < 0:
 *                     stop = 0             # <<<<<<<<<<<<<<
 *             elif stop > shape:
 *                 stop = shape
 */

#line 863 "View.MemoryView"
          __pyx_v_stop = 0;

          /* "View.MemoryView":862
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */

#line 862 "View.MemoryView"
        }

        /* "View.MemoryView":860
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */

#line 860 "View.MemoryView"
        goto __pyx_L17;

#line 860 "View.MemoryView"
      }

      /* "View.MemoryView":864
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */

#line 864 "View.MemoryView"
      __pyx_t_2 = ((__pyx_v_stop > __pyx_v_shape) != 0);

#line 864 "View.MemoryView"
      if (__pyx_t_2) {

        /* "View.MemoryView":865
 *                     stop = 0
 *             elif stop > shape:
 *                 stop = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */

#line 865 "View.MemoryView"
        __pyx_v_stop = __pyx_v_shape;

        /* "View.MemoryView":864
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */

#line 864 "View.MemoryView"
      }

#line 864 "View.MemoryView"
      __pyx_L17:;

      /* "View.MemoryView":859
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */

#line 859 "View.MemoryView"
      goto __pyx_L16;

#line 859 "View.MemoryView"
    }

    /* "View.MemoryView":867
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */

#line 867 "View.MemoryView"
    /*else*/ {

#line 867 "View.MemoryView"
      __pyx_t_2 = (__pyx_v_negative_step != 0);

#line 867 "View.MemoryView"
      if (__pyx_t_2) {

        /* "View.MemoryView":868
 *         else:
 *             if negative_step:
 *                 stop = -1             # <<<<<<<<<<<<<<
 *             else:
 *                 stop = shape
 */

#line 868 "View.MemoryView"
        __pyx_v_stop = -1L;

        /* "View.MemoryView":867
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */

#line 867 "View.MemoryView"
        goto __pyx_L19;

#line 867 "View.MemoryView"
      }

      /* "View.MemoryView":870
 *                 stop = -1
 *             else:
 *                 stop = shape             # <<<<<<<<<<<<<<
 * 
 *         if not have_step:
 */

#line 870 "View.MemoryView"
      /*else*/ {

#line 870 "View.MemoryView"
        __pyx_v_stop = __pyx_v_shape;

#line 870 "View.MemoryView"
      }

#line 870 "View.MemoryView"
      __pyx_L19:;

#line 870 "View.MemoryView"
    }

#line 870 "View.MemoryView"
    __pyx_L16:;

    /* "View.MemoryView":872
 *                 stop = shape
 * 
 *         if not have_step:             # <<<<<<<<<<<<<<
 *             step = 1
 * 
 */

#line 872 "View.MemoryView"
    __pyx_t_2 = ((!(__pyx_v_have_step != 0)) != 0);

#line 872 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":873
 * 
 *         if not have_step:
 *             step = 1             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 873 "View.MemoryView"
      __pyx_v_step = 1;

      /* "View.MemoryView":872
 *                 stop = shape
 * 
 *         if not have_step:             # <<<<<<<<<<<<<<
 *             step = 1
 * 
 */

#line 872 "View.MemoryView"
    }

    /* "View.MemoryView":877
 * 
 *         with cython.cdivision(True):
 *             new_shape = (stop - start) // step             # <<<<<<<<<<<<<<
 * 
 *             if (stop - start) - step * new_shape:
 */

#line 877 "View.MemoryView"
    __pyx_v_new_shape = ((__pyx_v_stop - __pyx_v_start) / __pyx_v_step);

    /* "View.MemoryView":879
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */

#line 879 "View.MemoryView"
    __pyx_t_2 = (((__pyx_v_stop - __pyx_v_start) - (__pyx_v_step * __pyx_v_new_shape)) != 0);

#line 879 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":880
 * 
 *             if (stop - start) - step * new_shape:
 *                 new_shape += 1             # <<<<<<<<<<<<<<
 * 
 *         if new_shape < 0:
 */

#line 880 "View.MemoryView"
      __pyx_v_new_shape = (__pyx_v_new_shape + 1);

      /* "View.MemoryView":879
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */

#line 879 "View.MemoryView"
    }

    /* "View.MemoryView":882
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */

#line 882 "View.MemoryView"
    __pyx_t_2 = ((__pyx_v_new_shape < 0) != 0);

#line 882 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":883
 * 
 *         if new_shape < 0:
 *             new_shape = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 883 "View.MemoryView"
      __pyx_v_new_shape = 0;

      /* "View.MemoryView":882
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */

#line 882 "View.MemoryView"
    }

    /* "View.MemoryView":886
 * 
 * 
 *         dst.strides[new_ndim] = stride * step             # <<<<<<<<<<<<<<
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset
 */

#line 886 "View.MemoryView"
    (__pyx_v_dst->strides[__pyx_v_new_ndim]) = (__pyx_v_stride * __pyx_v_step);

    /* "View.MemoryView":887
 * 
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape             # <<<<<<<<<<<<<<
 *         dst.suboffsets[new_ndim] = suboffset
 * 
 */

#line 887 "View.MemoryView"
    (__pyx_v_dst->shape[__pyx_v_new_ndim]) = __pyx_v_new_shape;

    /* "View.MemoryView":888
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 888 "View.MemoryView"
    (__pyx_v_dst->suboffsets[__pyx_v_new_ndim]) = __pyx_v_suboffset;

#line 888 "View.MemoryView"
  }

#line 888 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":891
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */

#line 891 "View.MemoryView"
  __pyx_t_2 = (((__pyx_v_suboffset_dim[0]) < 0) != 0);

#line 891 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":892
 * 
 *     if suboffset_dim[0] < 0:
 *         dst.data += start * stride             # <<<<<<<<<<<<<<
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 */

#line 892 "View.MemoryView"
    __pyx_v_dst->data = (__pyx_v_dst->data + (__pyx_v_start * __pyx_v_stride));

    /* "View.MemoryView":891
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */

#line 891 "View.MemoryView"
    goto __pyx_L23;

#line 891 "View.MemoryView"
  }

  /* "View.MemoryView":894
 *         dst.data += start * stride
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride             # <<<<<<<<<<<<<<
 * 
 *     if suboffset >= 0:
 */

#line 894 "View.MemoryView"
  /*else*/ {

#line 894 "View.MemoryView"
    __pyx_t_3 = (__pyx_v_suboffset_dim[0]);

#line 894 "View.MemoryView"
    (__pyx_v_dst->suboffsets[__pyx_t_3]) = ((__pyx_v_dst->suboffsets[__pyx_t_3]) + (__pyx_v_start * __pyx_v_stride));

#line 894 "View.MemoryView"
  }

#line 894 "View.MemoryView"
  __pyx_L23:;

  /* "View.MemoryView":896
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */

#line 896 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_suboffset >= 0) != 0);

#line 896 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":897
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */

#line 897 "View.MemoryView"
    __pyx_t_2 = ((!(__pyx_v_is_slice != 0)) != 0);

#line 897 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":898
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */

#line 898 "View.MemoryView"
      __pyx_t_2 = ((__pyx_v_new_ndim == 0) != 0);

#line 898 "View.MemoryView"
      if (__pyx_t_2) {

        /* "View.MemoryView":899
 *         if not is_slice:
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "
 */

#line 899 "View.MemoryView"
        __pyx_v_dst->data = ((((char **)__pyx_v_dst->data)[0]) + __pyx_v_suboffset);

        /* "View.MemoryView":898
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */

#line 898 "View.MemoryView"
        goto __pyx_L26;

#line 898 "View.MemoryView"
      }

      /* "View.MemoryView":901
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "             # <<<<<<<<<<<<<<
 *                                      "must be indexed and not sliced", dim)
 *         else:
 */

#line 901 "View.MemoryView"
      /*else*/ {

        /* "View.MemoryView":902
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "
 *                                      "must be indexed and not sliced", dim)             # <<<<<<<<<<<<<<
 *         else:
 *             suboffset_dim[0] = new_ndim
 */

#line 902 "View.MemoryView"
        __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_IndexError, ((char *)"All dimensions preceding dimension %d must be indexed and not sliced"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(3, 901, __pyx_L1_error)

#line 902 "View.MemoryView"
      }

#line 902 "View.MemoryView"
      __pyx_L26:;

      /* "View.MemoryView":897
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */

#line 897 "View.MemoryView"
      goto __pyx_L25;

#line 897 "View.MemoryView"
    }

    /* "View.MemoryView":904
 *                                      "must be indexed and not sliced", dim)
 *         else:
 *             suboffset_dim[0] = new_ndim             # <<<<<<<<<<<<<<
 * 
 *     return 0
 */

#line 904 "View.MemoryView"
    /*else*/ {

#line 904 "View.MemoryView"
      (__pyx_v_suboffset_dim[0]) = __pyx_v_new_ndim;

#line 904 "View.MemoryView"
    }

#line 904 "View.MemoryView"
    __pyx_L25:;

    /* "View.MemoryView":896
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */

#line 896 "View.MemoryView"
  }

  /* "View.MemoryView":906
 *             suboffset_dim[0] = new_ndim
 * 
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 906 "View.MemoryView"
  __pyx_r = 0;

#line 906 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":809
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

#line 809 "View.MemoryView"


#line 809 "View.MemoryView"
  /* function exit code */

#line 809 "View.MemoryView"
  __pyx_L1_error:;

#line 809 "View.MemoryView"
  {

#line 809 "View.MemoryView"
    #ifdef WITH_THREAD
    PyGILState_STATE 
#line 809 "View.MemoryView"
__pyx_gilstate_save = __Pyx_PyGILState_Ensure();

#line 809 "View.MemoryView"
    #endif

#line 809 "View.MemoryView"
    __Pyx_AddTraceback("View.MemoryView.slice_memviewslice", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 809 "View.MemoryView"
    #ifdef WITH_THREAD

#line 809 "View.MemoryView"
    __Pyx_PyGILState_Release(__pyx_gilstate_save);

#line 809 "View.MemoryView"
    #endif

#line 809 "View.MemoryView"
  }

#line 809 "View.MemoryView"
  __pyx_r = -1;

#line 809 "View.MemoryView"
  __pyx_L0:;

#line 809 "View.MemoryView"
  return __pyx_r;

#line 809 "View.MemoryView"
}

/* "View.MemoryView":912
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

#line 912 "View.MemoryView"


#line 912 "View.MemoryView"
static char *__pyx_pybuffer_index(Py_buffer *__pyx_v_view, char *__pyx_v_bufp, Py_ssize_t __pyx_v_index, Py_ssize_t __pyx_v_dim) {
  Py_ssize_t __pyx_v_shape
#line 912 "View.MemoryView"
;
  Py_ssize_t __pyx_v_stride
#line 912 "View.MemoryView"
;
  Py_ssize_t __pyx_v_suboffset
#line 912 "View.MemoryView"
;
  Py_ssize_t __pyx_v_itemsize
#line 912 "View.MemoryView"
;
  char *__pyx_v_resultp
#line 912 "View.MemoryView"
;

#line 912 "View.MemoryView"
  char *__pyx_r;

#line 912 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 912 "View.MemoryView"
  Py_ssize_t __pyx_t_1;

#line 912 "View.MemoryView"
  int __pyx_t_2;

#line 912 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 912 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 912 "View.MemoryView"
  int __pyx_lineno = 0;

#line 912 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 912 "View.MemoryView"
  int __pyx_clineno = 0;

#line 912 "View.MemoryView"
  __Pyx_RefNannySetupContext("pybuffer_index", 0);

  /* "View.MemoryView":914
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t itemsize = view.itemsize
 *     cdef char *resultp
 */

#line 914 "View.MemoryView"
  __pyx_v_suboffset = -1L;

  /* "View.MemoryView":915
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 *     cdef Py_ssize_t itemsize = view.itemsize             # <<<<<<<<<<<<<<
 *     cdef char *resultp
 * 
 */

#line 915 "View.MemoryView"
  __pyx_t_1 = __pyx_v_view->itemsize;

#line 915 "View.MemoryView"
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":918
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len / itemsize
 *         stride = itemsize
 */

#line 918 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_view->ndim == 0) != 0);

#line 918 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":919
 * 
 *     if view.ndim == 0:
 *         shape = view.len / itemsize             # <<<<<<<<<<<<<<
 *         stride = itemsize
 *     else:
 */

#line 919 "View.MemoryView"
    if (unlikely(__pyx_v_itemsize == 0)) {

#line 919 "View.MemoryView"
      PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");

#line 919 "View.MemoryView"
      __PYX_ERR(3, 919, __pyx_L1_error)

#line 919 "View.MemoryView"
    }

#line 919 "View.MemoryView"
    else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_itemsize == (Py_ssize_t)-1)  && unlikely(UNARY_NEG_WOULD_OVERFLOW(__pyx_v_view->len))) {

#line 919 "View.MemoryView"
      PyErr_SetString(PyExc_OverflowError, "value too large to perform division");

#line 919 "View.MemoryView"
      __PYX_ERR(3, 919, __pyx_L1_error)

#line 919 "View.MemoryView"
    }

#line 919 "View.MemoryView"
    __pyx_v_shape = __Pyx_div_Py_ssize_t(__pyx_v_view->len, __pyx_v_itemsize);

    /* "View.MemoryView":920
 *     if view.ndim == 0:
 *         shape = view.len / itemsize
 *         stride = itemsize             # <<<<<<<<<<<<<<
 *     else:
 *         shape = view.shape[dim]
 */

#line 920 "View.MemoryView"
    __pyx_v_stride = __pyx_v_itemsize;

    /* "View.MemoryView":918
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len / itemsize
 *         stride = itemsize
 */

#line 918 "View.MemoryView"
    goto __pyx_L3;

#line 918 "View.MemoryView"
  }

  /* "View.MemoryView":922
 *         stride = itemsize
 *     else:
 *         shape = view.shape[dim]             # <<<<<<<<<<<<<<
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 */

#line 922 "View.MemoryView"
  /*else*/ {

#line 922 "View.MemoryView"
    __pyx_v_shape = (__pyx_v_view->shape[__pyx_v_dim]);

    /* "View.MemoryView":923
 *     else:
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]             # <<<<<<<<<<<<<<
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]
 */

#line 923 "View.MemoryView"
    __pyx_v_stride = (__pyx_v_view->strides[__pyx_v_dim]);

    /* "View.MemoryView":924
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */

#line 924 "View.MemoryView"
    __pyx_t_2 = ((__pyx_v_view->suboffsets != NULL) != 0);

#line 924 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":925
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]             # <<<<<<<<<<<<<<
 * 
 *     if index < 0:
 */

#line 925 "View.MemoryView"
      __pyx_v_suboffset = (__pyx_v_view->suboffsets[__pyx_v_dim]);

      /* "View.MemoryView":924
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */

#line 924 "View.MemoryView"
    }

#line 924 "View.MemoryView"
  }

#line 924 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":927
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */

#line 927 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_index < 0) != 0);

#line 927 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":928
 * 
 *     if index < 0:
 *         index += view.shape[dim]             # <<<<<<<<<<<<<<
 *         if index < 0:
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 */

#line 928 "View.MemoryView"
    __pyx_v_index = (__pyx_v_index + (__pyx_v_view->shape[__pyx_v_dim]));

    /* "View.MemoryView":929
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */

#line 929 "View.MemoryView"
    __pyx_t_2 = ((__pyx_v_index < 0) != 0);

#line 929 "View.MemoryView"
    if (unlikely(__pyx_t_2)) {

      /* "View.MemoryView":930
 *         index += view.shape[dim]
 *         if index < 0:
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)             # <<<<<<<<<<<<<<
 * 
 *     if index >= shape:
 */

#line 930 "View.MemoryView"
      __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 930, __pyx_L1_error)

#line 930 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_3);

#line 930 "View.MemoryView"
      __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 930, __pyx_L1_error)

#line 930 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_4);

#line 930 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 930 "View.MemoryView"
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 930, __pyx_L1_error)

#line 930 "View.MemoryView"
      __Pyx_GOTREF(__pyx_t_3);

#line 930 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 930 "View.MemoryView"
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);

#line 930 "View.MemoryView"
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 930 "View.MemoryView"
      __PYX_ERR(3, 930, __pyx_L1_error)

      /* "View.MemoryView":929
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */

#line 929 "View.MemoryView"
    }

    /* "View.MemoryView":927
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */

#line 927 "View.MemoryView"
  }

  /* "View.MemoryView":932
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */

#line 932 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_index >= __pyx_v_shape) != 0);

#line 932 "View.MemoryView"
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":933
 * 
 *     if index >= shape:
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)             # <<<<<<<<<<<<<<
 * 
 *     resultp = bufp + index * stride
 */

#line 933 "View.MemoryView"
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 933, __pyx_L1_error)

#line 933 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 933 "View.MemoryView"
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 933, __pyx_L1_error)

#line 933 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_4);

#line 933 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 933 "View.MemoryView"
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 933, __pyx_L1_error)

#line 933 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 933 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 933 "View.MemoryView"
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);

#line 933 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 933 "View.MemoryView"
    __PYX_ERR(3, 933, __pyx_L1_error)

    /* "View.MemoryView":932
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */

#line 932 "View.MemoryView"
  }

  /* "View.MemoryView":935
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     resultp = bufp + index * stride             # <<<<<<<<<<<<<<
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset
 */

#line 935 "View.MemoryView"
  __pyx_v_resultp = (__pyx_v_bufp + (__pyx_v_index * __pyx_v_stride));

  /* "View.MemoryView":936
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */

#line 936 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_suboffset >= 0) != 0);

#line 936 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":937
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset             # <<<<<<<<<<<<<<
 * 
 *     return resultp
 */

#line 937 "View.MemoryView"
    __pyx_v_resultp = ((((char **)__pyx_v_resultp)[0]) + __pyx_v_suboffset);

    /* "View.MemoryView":936
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */

#line 936 "View.MemoryView"
  }

  /* "View.MemoryView":939
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 *     return resultp             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 939 "View.MemoryView"
  __pyx_r = __pyx_v_resultp;

#line 939 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":912
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

#line 912 "View.MemoryView"


#line 912 "View.MemoryView"
  /* function exit code */

#line 912 "View.MemoryView"
  __pyx_L1_error:;

#line 912 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 912 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_4);

#line 912 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.pybuffer_index", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 912 "View.MemoryView"
  __pyx_r = NULL;

#line 912 "View.MemoryView"
  __pyx_L0:;

#line 912 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 912 "View.MemoryView"
  return __pyx_r;

#line 912 "View.MemoryView"
}

/* "View.MemoryView":945
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

#line 945 "View.MemoryView"


#line 945 "View.MemoryView"
static int __pyx_memslice_transpose(__Pyx_memviewslice *__pyx_v_memslice) {
  int __pyx_v_ndim
#line 945 "View.MemoryView"
;
  Py_ssize_t *__pyx_v_shape
#line 945 "View.MemoryView"
;
  Py_ssize_t *__pyx_v_strides
#line 945 "View.MemoryView"
;
  int __pyx_v_i
#line 945 "View.MemoryView"
;
  int __pyx_v_j
#line 945 "View.MemoryView"
;

#line 945 "View.MemoryView"
  int __pyx_r;

#line 945 "View.MemoryView"
  int __pyx_t_1;

#line 945 "View.MemoryView"
  Py_ssize_t *__pyx_t_2;

#line 945 "View.MemoryView"
  long __pyx_t_3;

#line 945 "View.MemoryView"
  long __pyx_t_4;

#line 945 "View.MemoryView"
  Py_ssize_t __pyx_t_5;

#line 945 "View.MemoryView"
  Py_ssize_t __pyx_t_6;

#line 945 "View.MemoryView"
  int __pyx_t_7;

#line 945 "View.MemoryView"
  int __pyx_t_8;

#line 945 "View.MemoryView"
  int __pyx_t_9;

#line 945 "View.MemoryView"
  int __pyx_lineno = 0;

#line 945 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 945 "View.MemoryView"
  int __pyx_clineno = 0;

  /* "View.MemoryView":946
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:
 *     cdef int ndim = memslice.memview.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 */

#line 946 "View.MemoryView"
  __pyx_t_1 = __pyx_v_memslice->memview->view.ndim;

#line 946 "View.MemoryView"
  __pyx_v_ndim = __pyx_t_1;

  /* "View.MemoryView":948
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 *     cdef Py_ssize_t *shape = memslice.shape             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t *strides = memslice.strides
 * 
 */

#line 948 "View.MemoryView"
  __pyx_t_2 = __pyx_v_memslice->shape;

#line 948 "View.MemoryView"
  __pyx_v_shape = __pyx_t_2;

  /* "View.MemoryView":949
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 *     cdef Py_ssize_t *strides = memslice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 949 "View.MemoryView"
  __pyx_t_2 = __pyx_v_memslice->strides;

#line 949 "View.MemoryView"
  __pyx_v_strides = __pyx_t_2;

  /* "View.MemoryView":953
 * 
 *     cdef int i, j
 *     for i in range(ndim / 2):             # <<<<<<<<<<<<<<
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 */

#line 953 "View.MemoryView"
  __pyx_t_3 = __Pyx_div_long(__pyx_v_ndim, 2);

#line 953 "View.MemoryView"
  __pyx_t_4 = __pyx_t_3;

#line 953 "View.MemoryView"
  for (__pyx_t_1 = 0; __pyx_t_1 < __pyx_t_4; __pyx_t_1+=1) {

#line 953 "View.MemoryView"
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":954
 *     cdef int i, j
 *     for i in range(ndim / 2):
 *         j = ndim - 1 - i             # <<<<<<<<<<<<<<
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]
 */

#line 954 "View.MemoryView"
    __pyx_v_j = ((__pyx_v_ndim - 1) - __pyx_v_i);

    /* "View.MemoryView":955
 *     for i in range(ndim / 2):
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]             # <<<<<<<<<<<<<<
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 */

#line 955 "View.MemoryView"
    __pyx_t_5 = (__pyx_v_strides[__pyx_v_j]);

#line 955 "View.MemoryView"
    __pyx_t_6 = (__pyx_v_strides[__pyx_v_i]);

#line 955 "View.MemoryView"
    (__pyx_v_strides[__pyx_v_i]) = __pyx_t_5;

#line 955 "View.MemoryView"
    (__pyx_v_strides[__pyx_v_j]) = __pyx_t_6;

    /* "View.MemoryView":956
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]             # <<<<<<<<<<<<<<
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 */

#line 956 "View.MemoryView"
    __pyx_t_6 = (__pyx_v_shape[__pyx_v_j]);

#line 956 "View.MemoryView"
    __pyx_t_5 = (__pyx_v_shape[__pyx_v_i]);

#line 956 "View.MemoryView"
    (__pyx_v_shape[__pyx_v_i]) = __pyx_t_6;

#line 956 "View.MemoryView"
    (__pyx_v_shape[__pyx_v_j]) = __pyx_t_5;

    /* "View.MemoryView":958
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */

#line 958 "View.MemoryView"
    __pyx_t_8 = (((__pyx_v_memslice->suboffsets[__pyx_v_i]) >= 0) != 0);

#line 958 "View.MemoryView"
    if (!__pyx_t_8) {

#line 958 "View.MemoryView"
    } else {

#line 958 "View.MemoryView"
      __pyx_t_7 = __pyx_t_8;

#line 958 "View.MemoryView"
      goto __pyx_L6_bool_binop_done;

#line 958 "View.MemoryView"
    }

#line 958 "View.MemoryView"
    __pyx_t_8 = (((__pyx_v_memslice->suboffsets[__pyx_v_j]) >= 0) != 0);

#line 958 "View.MemoryView"
    __pyx_t_7 = __pyx_t_8;

#line 958 "View.MemoryView"
    __pyx_L6_bool_binop_done:;

#line 958 "View.MemoryView"
    if (__pyx_t_7) {

      /* "View.MemoryView":959
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")             # <<<<<<<<<<<<<<
 * 
 *     return 1
 */

#line 959 "View.MemoryView"
      __pyx_t_9 = __pyx_memoryview_err(__pyx_builtin_ValueError, ((char *)"Cannot transpose memoryview with indirect dimensions")); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(3, 959, __pyx_L1_error)

      /* "View.MemoryView":958
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */

#line 958 "View.MemoryView"
    }

#line 958 "View.MemoryView"
  }

  /* "View.MemoryView":961
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 961 "View.MemoryView"
  __pyx_r = 1;

#line 961 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":945
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

#line 945 "View.MemoryView"


#line 945 "View.MemoryView"
  /* function exit code */

#line 945 "View.MemoryView"
  __pyx_L1_error:;

#line 945 "View.MemoryView"
  {

#line 945 "View.MemoryView"
    #ifdef WITH_THREAD
    PyGILState_STATE 
#line 945 "View.MemoryView"
__pyx_gilstate_save = __Pyx_PyGILState_Ensure();

#line 945 "View.MemoryView"
    #endif

#line 945 "View.MemoryView"
    __Pyx_AddTraceback("View.MemoryView.transpose_memslice", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 945 "View.MemoryView"
    #ifdef WITH_THREAD

#line 945 "View.MemoryView"
    __Pyx_PyGILState_Release(__pyx_gilstate_save);

#line 945 "View.MemoryView"
    #endif

#line 945 "View.MemoryView"
  }

#line 945 "View.MemoryView"
  __pyx_r = 0;

#line 945 "View.MemoryView"
  __pyx_L0:;

#line 945 "View.MemoryView"
  return __pyx_r;

#line 945 "View.MemoryView"
}

/* "View.MemoryView":978
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 */

#line 978 "View.MemoryView"


#line 978 "View.MemoryView"
/* Python wrapper */

#line 978 "View.MemoryView"
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self); /*proto*/

#line 978 "View.MemoryView"
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self) {

#line 978 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 978 "View.MemoryView"
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);

#line 978 "View.MemoryView"
  __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

#line 978 "View.MemoryView"


#line 978 "View.MemoryView"
  /* function exit code */

#line 978 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 978 "View.MemoryView"
}

#line 978 "View.MemoryView"


#line 978 "View.MemoryView"
static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self) {

#line 978 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 978 "View.MemoryView"
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":979
 * 
 *     def __dealloc__(self):
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */

#line 979 "View.MemoryView"
  __PYX_XDEC_MEMVIEW((&__pyx_v_self->from_slice), 1);

  /* "View.MemoryView":978
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 */

#line 978 "View.MemoryView"


#line 978 "View.MemoryView"
  /* function exit code */

#line 978 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 978 "View.MemoryView"
}

/* "View.MemoryView":981
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

#line 981 "View.MemoryView"


#line 981 "View.MemoryView"
static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp) {

#line 981 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 981 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 981 "View.MemoryView"
  int __pyx_t_1;

#line 981 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 981 "View.MemoryView"
  int __pyx_lineno = 0;

#line 981 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 981 "View.MemoryView"
  int __pyx_clineno = 0;

#line 981 "View.MemoryView"
  __Pyx_RefNannySetupContext("convert_item_to_object", 0);

  /* "View.MemoryView":982
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */

#line 982 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_self->to_object_func != NULL) != 0);

#line 982 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":983
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)             # <<<<<<<<<<<<<<
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)
 */

#line 983 "View.MemoryView"
    __Pyx_XDECREF(__pyx_r);

#line 983 "View.MemoryView"
    __pyx_t_2 = __pyx_v_self->to_object_func(__pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 983, __pyx_L1_error)

#line 983 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 983 "View.MemoryView"
    __pyx_r = __pyx_t_2;

#line 983 "View.MemoryView"
    __pyx_t_2 = 0;

#line 983 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":982
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */

#line 982 "View.MemoryView"
  }

  /* "View.MemoryView":985
 *             return self.to_object_func(itemp)
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */

#line 985 "View.MemoryView"
  /*else*/ {

#line 985 "View.MemoryView"
    __Pyx_XDECREF(__pyx_r);

#line 985 "View.MemoryView"
    __pyx_t_2 = __pyx_memoryview_convert_item_to_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 985, __pyx_L1_error)

#line 985 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 985 "View.MemoryView"
    __pyx_r = __pyx_t_2;

#line 985 "View.MemoryView"
    __pyx_t_2 = 0;

#line 985 "View.MemoryView"
    goto __pyx_L0;

#line 985 "View.MemoryView"
  }

  /* "View.MemoryView":981
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

#line 981 "View.MemoryView"


#line 981 "View.MemoryView"
  /* function exit code */

#line 981 "View.MemoryView"
  __pyx_L1_error:;

#line 981 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 981 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 981 "View.MemoryView"
  __pyx_r = 0;

#line 981 "View.MemoryView"
  __pyx_L0:;

#line 981 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 981 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 981 "View.MemoryView"
  return __pyx_r;

#line 981 "View.MemoryView"
}

/* "View.MemoryView":987
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

#line 987 "View.MemoryView"


#line 987 "View.MemoryView"
static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {

#line 987 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 987 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 987 "View.MemoryView"
  int __pyx_t_1;

#line 987 "View.MemoryView"
  int __pyx_t_2;

#line 987 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 987 "View.MemoryView"
  int __pyx_lineno = 0;

#line 987 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 987 "View.MemoryView"
  int __pyx_clineno = 0;

#line 987 "View.MemoryView"
  __Pyx_RefNannySetupContext("assign_item_from_object", 0);

  /* "View.MemoryView":988
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */

#line 988 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_self->to_dtype_func != NULL) != 0);

#line 988 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":989
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)             # <<<<<<<<<<<<<<
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)
 */

#line 989 "View.MemoryView"
    __pyx_t_2 = __pyx_v_self->to_dtype_func(__pyx_v_itemp, __pyx_v_value); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(3, 989, __pyx_L1_error)

    /* "View.MemoryView":988
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */

#line 988 "View.MemoryView"
    goto __pyx_L3;

#line 988 "View.MemoryView"
  }

  /* "View.MemoryView":991
 *             self.to_dtype_func(itemp, value)
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 991 "View.MemoryView"
  /*else*/ {

#line 991 "View.MemoryView"
    __pyx_t_3 = __pyx_memoryview_assign_item_from_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 991, __pyx_L1_error)

#line 991 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 991 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 991 "View.MemoryView"
  }

#line 991 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":987
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

#line 987 "View.MemoryView"


#line 987 "View.MemoryView"
  /* function exit code */

#line 987 "View.MemoryView"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 987 "View.MemoryView"
  goto __pyx_L0;

#line 987 "View.MemoryView"
  __pyx_L1_error:;

#line 987 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 987 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 987 "View.MemoryView"
  __pyx_r = 0;

#line 987 "View.MemoryView"
  __pyx_L0:;

#line 987 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 987 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 987 "View.MemoryView"
  return __pyx_r;

#line 987 "View.MemoryView"
}

/* "View.MemoryView":994
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

#line 994 "View.MemoryView"


#line 994 "View.MemoryView"
/* Python wrapper */

#line 994 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 994 "View.MemoryView"
static PyObject *__pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(PyObject *__pyx_v_self) {

#line 994 "View.MemoryView"
  PyObject *__pyx_r = 0;

#line 994 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 994 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 994 "View.MemoryView"
__pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

#line 994 "View.MemoryView"


#line 994 "View.MemoryView"
  /* function exit code */

#line 994 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 994 "View.MemoryView"
  return __pyx_r;

#line 994 "View.MemoryView"
}

#line 994 "View.MemoryView"


#line 994 "View.MemoryView"
static PyObject *__pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(struct __pyx_memoryviewslice_obj *__pyx_v_self) {

#line 994 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 994 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 994 "View.MemoryView"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":995
 *     @property
 *     def base(self):
 *         return self.from_object             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")
 */

#line 995 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 995 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_self->from_object);

#line 995 "View.MemoryView"
  __pyx_r = __pyx_v_self->from_object;

#line 995 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":994
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

#line 994 "View.MemoryView"


#line 994 "View.MemoryView"
  /* function exit code */

#line 994 "View.MemoryView"
  __pyx_L0:;

#line 994 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 994 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 994 "View.MemoryView"
  return __pyx_r;

#line 994 "View.MemoryView"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf___pyx_memoryviewslice___reduce_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 2 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__22, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 2, __pyx_L1_error)

#line 2 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 2 "(tree fragment)"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 2 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 2 "(tree fragment)"
  __PYX_ERR(3, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
/* Python wrapper */

#line 3 "(tree fragment)"
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 3 "(tree fragment)"
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 3 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 3 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 3 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 3 "(tree fragment)"
__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
  /* function exit code */

#line 3 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 3 "(tree fragment)"
  return __pyx_r;

#line 3 "(tree fragment)"
}

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {

#line 3 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 3 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 3 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 3 "(tree fragment)"
  int __pyx_lineno = 0;

#line 3 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 3 "(tree fragment)"
  int __pyx_clineno = 0;

#line 3 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__23, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __PYX_ERR(3, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
  /* function exit code */

#line 3 "(tree fragment)"
  __pyx_L1_error:;

#line 3 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 3 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 3 "(tree fragment)"
  __pyx_r = NULL;

#line 3 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 3 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 3 "(tree fragment)"
  return __pyx_r;

#line 3 "(tree fragment)"
}

/* "View.MemoryView":1001
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

#line 1001 "View.MemoryView"


#line 1001 "View.MemoryView"
static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice __pyx_v_memviewslice, int __pyx_v_ndim, PyObject *(*__pyx_v_to_object_func)(char *), int (*__pyx_v_to_dtype_func)(char *, PyObject *), int __pyx_v_dtype_is_object) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0
#line 1001 "View.MemoryView"
;
  Py_ssize_t __pyx_v_suboffset
#line 1001 "View.MemoryView"
;
  PyObject *__pyx_v_length = NULL
#line 1001 "View.MemoryView"
;

#line 1001 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 1001 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1001 "View.MemoryView"
  int __pyx_t_1;

#line 1001 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 1001 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 1001 "View.MemoryView"
  __Pyx_TypeInfo *__pyx_t_4;

#line 1001 "View.MemoryView"
  Py_buffer __pyx_t_5;

#line 1001 "View.MemoryView"
  Py_ssize_t *__pyx_t_6;

#line 1001 "View.MemoryView"
  Py_ssize_t *__pyx_t_7;

#line 1001 "View.MemoryView"
  Py_ssize_t *__pyx_t_8;

#line 1001 "View.MemoryView"
  Py_ssize_t __pyx_t_9;

#line 1001 "View.MemoryView"
  int __pyx_lineno = 0;

#line 1001 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 1001 "View.MemoryView"
  int __pyx_clineno = 0;

#line 1001 "View.MemoryView"
  __Pyx_RefNannySetupContext("memoryview_fromslice", 0);

  /* "View.MemoryView":1009
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */

#line 1009 "View.MemoryView"
  __pyx_t_1 = ((((PyObject *)__pyx_v_memviewslice.memview) == Py_None) != 0);

#line 1009 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":1010
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:
 *         return None             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1010 "View.MemoryView"
    __Pyx_XDECREF(__pyx_r);

#line 1010 "View.MemoryView"
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 1010 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":1009
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */

#line 1009 "View.MemoryView"
  }

  /* "View.MemoryView":1015
 * 
 * 
 *     result = _memoryviewslice(None, 0, dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     result.from_slice = memviewslice
 */

#line 1015 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 1015, __pyx_L1_error)

#line 1015 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 1015 "View.MemoryView"
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 1015, __pyx_L1_error)

#line 1015 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 1015 "View.MemoryView"
  __Pyx_INCREF(Py_None);

#line 1015 "View.MemoryView"
  __Pyx_GIVEREF(Py_None);

#line 1015 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 0, Py_None);

#line 1015 "View.MemoryView"
  __Pyx_INCREF(__pyx_int_0);

#line 1015 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_int_0);

#line 1015 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_0);

#line 1015 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_2);

#line 1015 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);

#line 1015 "View.MemoryView"
  __pyx_t_2 = 0;

#line 1015 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryviewslice_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 1015, __pyx_L1_error)

#line 1015 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 1015 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1015 "View.MemoryView"
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_2);

#line 1015 "View.MemoryView"
  __pyx_t_2 = 0;

  /* "View.MemoryView":1017
 *     result = _memoryviewslice(None, 0, dtype_is_object)
 * 
 *     result.from_slice = memviewslice             # <<<<<<<<<<<<<<
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 */

#line 1017 "View.MemoryView"
  __pyx_v_result->from_slice = __pyx_v_memviewslice;

  /* "View.MemoryView":1018
 * 
 *     result.from_slice = memviewslice
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)             # <<<<<<<<<<<<<<
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base
 */

#line 1018 "View.MemoryView"
  __PYX_INC_MEMVIEW((&__pyx_v_memviewslice), 1);

  /* "View.MemoryView":1020
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base             # <<<<<<<<<<<<<<
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 */

#line 1020 "View.MemoryView"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_memviewslice.memview), __pyx_n_s_base); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 1020, __pyx_L1_error)

#line 1020 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 1020 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_2);

#line 1020 "View.MemoryView"
  __Pyx_GOTREF(__pyx_v_result->from_object);

#line 1020 "View.MemoryView"
  __Pyx_DECREF(__pyx_v_result->from_object);

#line 1020 "View.MemoryView"
  __pyx_v_result->from_object = __pyx_t_2;

#line 1020 "View.MemoryView"
  __pyx_t_2 = 0;

  /* "View.MemoryView":1021
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base
 *     result.typeinfo = memviewslice.memview.typeinfo             # <<<<<<<<<<<<<<
 * 
 *     result.view = memviewslice.memview.view
 */

#line 1021 "View.MemoryView"
  __pyx_t_4 = __pyx_v_memviewslice.memview->typeinfo;

#line 1021 "View.MemoryView"
  __pyx_v_result->__pyx_base.typeinfo = __pyx_t_4;

  /* "View.MemoryView":1023
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 *     result.view = memviewslice.memview.view             # <<<<<<<<<<<<<<
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 */

#line 1023 "View.MemoryView"
  __pyx_t_5 = __pyx_v_memviewslice.memview->view;

#line 1023 "View.MemoryView"
  __pyx_v_result->__pyx_base.view = __pyx_t_5;

  /* "View.MemoryView":1024
 * 
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data             # <<<<<<<<<<<<<<
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 */

#line 1024 "View.MemoryView"
  __pyx_v_result->__pyx_base.view.buf = ((void *)__pyx_v_memviewslice.data);

  /* "View.MemoryView":1025
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim             # <<<<<<<<<<<<<<
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)
 */

#line 1025 "View.MemoryView"
  __pyx_v_result->__pyx_base.view.ndim = __pyx_v_ndim;

  /* "View.MemoryView":1026
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None             # <<<<<<<<<<<<<<
 *     Py_INCREF(Py_None)
 * 
 */

#line 1026 "View.MemoryView"
  ((Py_buffer *)(&__pyx_v_result->__pyx_base.view))->obj = Py_None;

  /* "View.MemoryView":1027
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 */

#line 1027 "View.MemoryView"
  Py_INCREF(Py_None);

  /* "View.MemoryView":1029
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */

#line 1029 "View.MemoryView"
  __pyx_t_1 = ((((struct __pyx_memoryview_obj *)__pyx_v_memviewslice.memview)->flags & PyBUF_WRITABLE) != 0);

#line 1029 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":1030
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 *         result.flags = PyBUF_RECORDS             # <<<<<<<<<<<<<<
 *     else:
 *         result.flags = PyBUF_RECORDS_RO
 */

#line 1030 "View.MemoryView"
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS;

    /* "View.MemoryView":1029
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */

#line 1029 "View.MemoryView"
    goto __pyx_L4;

#line 1029 "View.MemoryView"
  }

  /* "View.MemoryView":1032
 *         result.flags = PyBUF_RECORDS
 *     else:
 *         result.flags = PyBUF_RECORDS_RO             # <<<<<<<<<<<<<<
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 */

#line 1032 "View.MemoryView"
  /*else*/ {

#line 1032 "View.MemoryView"
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS_RO;

#line 1032 "View.MemoryView"
  }

#line 1032 "View.MemoryView"
  __pyx_L4:;

  /* "View.MemoryView":1034
 *         result.flags = PyBUF_RECORDS_RO
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape             # <<<<<<<<<<<<<<
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides
 * 
 */

#line 1034 "View.MemoryView"
  __pyx_v_result->__pyx_base.view.shape = ((Py_ssize_t *)__pyx_v_result->from_slice.shape);

  /* "View.MemoryView":1035
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1035 "View.MemoryView"
  __pyx_v_result->__pyx_base.view.strides = ((Py_ssize_t *)__pyx_v_result->from_slice.strides);

  /* "View.MemoryView":1038
 * 
 * 
 *     result.view.suboffsets = NULL             # <<<<<<<<<<<<<<
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

#line 1038 "View.MemoryView"
  __pyx_v_result->__pyx_base.view.suboffsets = NULL;

  /* "View.MemoryView":1039
 * 
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 */

#line 1039 "View.MemoryView"
  __pyx_t_7 = (__pyx_v_result->from_slice.suboffsets + __pyx_v_ndim);

#line 1039 "View.MemoryView"
  for (__pyx_t_8 = __pyx_v_result->from_slice.suboffsets; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {

#line 1039 "View.MemoryView"
    __pyx_t_6 = __pyx_t_8;

#line 1039 "View.MemoryView"
    __pyx_v_suboffset = (__pyx_t_6[0]);

    /* "View.MemoryView":1040
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */

#line 1040 "View.MemoryView"
    __pyx_t_1 = ((__pyx_v_suboffset >= 0) != 0);

#line 1040 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":1041
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets             # <<<<<<<<<<<<<<
 *             break
 * 
 */

#line 1041 "View.MemoryView"
      __pyx_v_result->__pyx_base.view.suboffsets = ((Py_ssize_t *)__pyx_v_result->from_slice.suboffsets);

      /* "View.MemoryView":1042
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break             # <<<<<<<<<<<<<<
 * 
 *     result.view.len = result.view.itemsize
 */

#line 1042 "View.MemoryView"
      goto __pyx_L6_break;

      /* "View.MemoryView":1040
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */

#line 1040 "View.MemoryView"
    }

#line 1040 "View.MemoryView"
  }

#line 1040 "View.MemoryView"
  __pyx_L6_break:;

  /* "View.MemoryView":1044
 *             break
 * 
 *     result.view.len = result.view.itemsize             # <<<<<<<<<<<<<<
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length
 */

#line 1044 "View.MemoryView"
  __pyx_t_9 = __pyx_v_result->__pyx_base.view.itemsize;

#line 1044 "View.MemoryView"
  __pyx_v_result->__pyx_base.view.len = __pyx_t_9;

  /* "View.MemoryView":1045
 * 
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:             # <<<<<<<<<<<<<<
 *         result.view.len *= length
 * 
 */

#line 1045 "View.MemoryView"
  __pyx_t_7 = (__pyx_v_result->__pyx_base.view.shape + __pyx_v_ndim);

#line 1045 "View.MemoryView"
  for (__pyx_t_8 = __pyx_v_result->__pyx_base.view.shape; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {

#line 1045 "View.MemoryView"
    __pyx_t_6 = __pyx_t_8;

#line 1045 "View.MemoryView"
    __pyx_t_2 = PyInt_FromSsize_t((__pyx_t_6[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 1045, __pyx_L1_error)

#line 1045 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 1045 "View.MemoryView"
    __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_2);

#line 1045 "View.MemoryView"
    __pyx_t_2 = 0;

    /* "View.MemoryView":1046
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length             # <<<<<<<<<<<<<<
 * 
 *     result.to_object_func = to_object_func
 */

#line 1046 "View.MemoryView"
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_result->__pyx_base.view.len); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 1046, __pyx_L1_error)

#line 1046 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 1046 "View.MemoryView"
    __pyx_t_3 = PyNumber_InPlaceMultiply(__pyx_t_2, __pyx_v_length); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 1046, __pyx_L1_error)

#line 1046 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 1046 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 1046 "View.MemoryView"
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_3); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(3, 1046, __pyx_L1_error)

#line 1046 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1046 "View.MemoryView"
    __pyx_v_result->__pyx_base.view.len = __pyx_t_9;

#line 1046 "View.MemoryView"
  }

  /* "View.MemoryView":1048
 *         result.view.len *= length
 * 
 *     result.to_object_func = to_object_func             # <<<<<<<<<<<<<<
 *     result.to_dtype_func = to_dtype_func
 * 
 */

#line 1048 "View.MemoryView"
  __pyx_v_result->to_object_func = __pyx_v_to_object_func;

  /* "View.MemoryView":1049
 * 
 *     result.to_object_func = to_object_func
 *     result.to_dtype_func = to_dtype_func             # <<<<<<<<<<<<<<
 * 
 *     return result
 */

#line 1049 "View.MemoryView"
  __pyx_v_result->to_dtype_func = __pyx_v_to_dtype_func;

  /* "View.MemoryView":1051
 *     result.to_dtype_func = to_dtype_func
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 */

#line 1051 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 1051 "View.MemoryView"
  __Pyx_INCREF(((PyObject *)__pyx_v_result));

#line 1051 "View.MemoryView"
  __pyx_r = ((PyObject *)__pyx_v_result);

#line 1051 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":1001
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

#line 1001 "View.MemoryView"


#line 1001 "View.MemoryView"
  /* function exit code */

#line 1001 "View.MemoryView"
  __pyx_L1_error:;

#line 1001 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 1001 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 1001 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview_fromslice", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1001 "View.MemoryView"
  __pyx_r = 0;

#line 1001 "View.MemoryView"
  __pyx_L0:;

#line 1001 "View.MemoryView"
  __Pyx_XDECREF((PyObject *)__pyx_v_result);

#line 1001 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_length);

#line 1001 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 1001 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1001 "View.MemoryView"
  return __pyx_r;

#line 1001 "View.MemoryView"
}

/* "View.MemoryView":1054
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 */

#line 1054 "View.MemoryView"


#line 1054 "View.MemoryView"
static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_mslice) {
  struct __pyx_memoryviewslice_obj *__pyx_v_obj = 0
#line 1054 "View.MemoryView"
;

#line 1054 "View.MemoryView"
  __Pyx_memviewslice *__pyx_r;

#line 1054 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1054 "View.MemoryView"
  int __pyx_t_1;

#line 1054 "View.MemoryView"
  int __pyx_t_2;

#line 1054 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 1054 "View.MemoryView"
  int __pyx_lineno = 0;

#line 1054 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 1054 "View.MemoryView"
  int __pyx_clineno = 0;

#line 1054 "View.MemoryView"
  __Pyx_RefNannySetupContext("get_slice_from_memview", 0);

  /* "View.MemoryView":1057
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */

#line 1057 "View.MemoryView"
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 

#line 1057 "View.MemoryView"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 1057 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":1058
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview             # <<<<<<<<<<<<<<
 *         return &obj.from_slice
 *     else:
 */

#line 1058 "View.MemoryView"
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(3, 1058, __pyx_L1_error)

#line 1058 "View.MemoryView"
    __pyx_t_3 = ((PyObject *)__pyx_v_memview);

#line 1058 "View.MemoryView"
    __Pyx_INCREF(__pyx_t_3);

#line 1058 "View.MemoryView"
    __pyx_v_obj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_3);

#line 1058 "View.MemoryView"
    __pyx_t_3 = 0;

    /* "View.MemoryView":1059
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview
 *         return &obj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, mslice)
 */

#line 1059 "View.MemoryView"
    __pyx_r = (&__pyx_v_obj->from_slice);

#line 1059 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":1057
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */

#line 1057 "View.MemoryView"
  }

  /* "View.MemoryView":1061
 *         return &obj.from_slice
 *     else:
 *         slice_copy(memview, mslice)             # <<<<<<<<<<<<<<
 *         return mslice
 * 
 */

#line 1061 "View.MemoryView"
  /*else*/ {

#line 1061 "View.MemoryView"
    __pyx_memoryview_slice_copy(__pyx_v_memview, __pyx_v_mslice);

    /* "View.MemoryView":1062
 *     else:
 *         slice_copy(memview, mslice)
 *         return mslice             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_slice_copy')
 */

#line 1062 "View.MemoryView"
    __pyx_r = __pyx_v_mslice;

#line 1062 "View.MemoryView"
    goto __pyx_L0;

#line 1062 "View.MemoryView"
  }

  /* "View.MemoryView":1054
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 */

#line 1054 "View.MemoryView"


#line 1054 "View.MemoryView"
  /* function exit code */

#line 1054 "View.MemoryView"
  __pyx_L1_error:;

#line 1054 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 1054 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.get_slice_from_memview", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1054 "View.MemoryView"
  __pyx_r = NULL;

#line 1054 "View.MemoryView"
  __pyx_L0:;

#line 1054 "View.MemoryView"
  __Pyx_XDECREF((PyObject *)__pyx_v_obj);

#line 1054 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1054 "View.MemoryView"
  return __pyx_r;

#line 1054 "View.MemoryView"
}

/* "View.MemoryView":1065
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst):             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

#line 1065 "View.MemoryView"


#line 1065 "View.MemoryView"
static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_dst) {
  int __pyx_v_dim
#line 1065 "View.MemoryView"
;
  Py_ssize_t *__pyx_v_shape
#line 1065 "View.MemoryView"
;
  Py_ssize_t *__pyx_v_strides
#line 1065 "View.MemoryView"
;
  Py_ssize_t *__pyx_v_suboffsets
#line 1065 "View.MemoryView"
;

#line 1065 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1065 "View.MemoryView"
  Py_ssize_t *__pyx_t_1;

#line 1065 "View.MemoryView"
  int __pyx_t_2;

#line 1065 "View.MemoryView"
  int __pyx_t_3;

#line 1065 "View.MemoryView"
  int __pyx_t_4;

#line 1065 "View.MemoryView"
  Py_ssize_t __pyx_t_5;

#line 1065 "View.MemoryView"
  __Pyx_RefNannySetupContext("slice_copy", 0);

  /* "View.MemoryView":1069
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 * 
 *     shape = memview.view.shape             # <<<<<<<<<<<<<<
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets
 */

#line 1069 "View.MemoryView"
  __pyx_t_1 = __pyx_v_memview->view.shape;

#line 1069 "View.MemoryView"
  __pyx_v_shape = __pyx_t_1;

  /* "View.MemoryView":1070
 * 
 *     shape = memview.view.shape
 *     strides = memview.view.strides             # <<<<<<<<<<<<<<
 *     suboffsets = memview.view.suboffsets
 * 
 */

#line 1070 "View.MemoryView"
  __pyx_t_1 = __pyx_v_memview->view.strides;

#line 1070 "View.MemoryView"
  __pyx_v_strides = __pyx_t_1;

  /* "View.MemoryView":1071
 *     shape = memview.view.shape
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets             # <<<<<<<<<<<<<<
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 */

#line 1071 "View.MemoryView"
  __pyx_t_1 = __pyx_v_memview->view.suboffsets;

#line 1071 "View.MemoryView"
  __pyx_v_suboffsets = __pyx_t_1;

  /* "View.MemoryView":1073
 *     suboffsets = memview.view.suboffsets
 * 
 *     dst.memview = <__pyx_memoryview *> memview             # <<<<<<<<<<<<<<
 *     dst.data = <char *> memview.view.buf
 * 
 */

#line 1073 "View.MemoryView"
  __pyx_v_dst->memview = ((struct __pyx_memoryview_obj *)__pyx_v_memview);

  /* "View.MemoryView":1074
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 *     dst.data = <char *> memview.view.buf             # <<<<<<<<<<<<<<
 * 
 *     for dim in range(memview.view.ndim):
 */

#line 1074 "View.MemoryView"
  __pyx_v_dst->data = ((char *)__pyx_v_memview->view.buf);

  /* "View.MemoryView":1076
 *     dst.data = <char *> memview.view.buf
 * 
 *     for dim in range(memview.view.ndim):             # <<<<<<<<<<<<<<
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 */

#line 1076 "View.MemoryView"
  __pyx_t_2 = __pyx_v_memview->view.ndim;

#line 1076 "View.MemoryView"
  __pyx_t_3 = __pyx_t_2;

#line 1076 "View.MemoryView"
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 1076 "View.MemoryView"
    __pyx_v_dim = __pyx_t_4;

    /* "View.MemoryView":1077
 * 
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]             # <<<<<<<<<<<<<<
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 */

#line 1077 "View.MemoryView"
    (__pyx_v_dst->shape[__pyx_v_dim]) = (__pyx_v_shape[__pyx_v_dim]);

    /* "View.MemoryView":1078
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]             # <<<<<<<<<<<<<<
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 * 
 */

#line 1078 "View.MemoryView"
    (__pyx_v_dst->strides[__pyx_v_dim]) = (__pyx_v_strides[__pyx_v_dim]);

    /* "View.MemoryView":1079
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object')
 */

#line 1079 "View.MemoryView"
    if ((__pyx_v_suboffsets != 0)) {

#line 1079 "View.MemoryView"
      __pyx_t_5 = (__pyx_v_suboffsets[__pyx_v_dim]);

#line 1079 "View.MemoryView"
    } else {

#line 1079 "View.MemoryView"
      __pyx_t_5 = -1L;

#line 1079 "View.MemoryView"
    }

#line 1079 "View.MemoryView"
    (__pyx_v_dst->suboffsets[__pyx_v_dim]) = __pyx_t_5;

#line 1079 "View.MemoryView"
  }

  /* "View.MemoryView":1065
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst):             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

#line 1065 "View.MemoryView"


#line 1065 "View.MemoryView"
  /* function exit code */

#line 1065 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1065 "View.MemoryView"
}

/* "View.MemoryView":1082
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

#line 1082 "View.MemoryView"


#line 1082 "View.MemoryView"
static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *__pyx_v_memview) {
  __Pyx_memviewslice __pyx_v_memviewslice
#line 1082 "View.MemoryView"
;

#line 1082 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 1082 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1082 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 1082 "View.MemoryView"
  int __pyx_lineno = 0;

#line 1082 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 1082 "View.MemoryView"
  int __pyx_clineno = 0;

#line 1082 "View.MemoryView"
  __Pyx_RefNannySetupContext("memoryview_copy", 0);

  /* "View.MemoryView":1085
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)             # <<<<<<<<<<<<<<
 *     return memoryview_copy_from_slice(memview, &memviewslice)
 * 
 */

#line 1085 "View.MemoryView"
  __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_memviewslice));

  /* "View.MemoryView":1086
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)
 *     return memoryview_copy_from_slice(memview, &memviewslice)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 */

#line 1086 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

#line 1086 "View.MemoryView"
  __pyx_t_1 = __pyx_memoryview_copy_object_from_slice(__pyx_v_memview, (&__pyx_v_memviewslice)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 1086, __pyx_L1_error)

#line 1086 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 1086 "View.MemoryView"
  __pyx_r = __pyx_t_1;

#line 1086 "View.MemoryView"
  __pyx_t_1 = 0;

#line 1086 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":1082
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

#line 1082 "View.MemoryView"


#line 1082 "View.MemoryView"
  /* function exit code */

#line 1082 "View.MemoryView"
  __pyx_L1_error:;

#line 1082 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 1082 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1082 "View.MemoryView"
  __pyx_r = 0;

#line 1082 "View.MemoryView"
  __pyx_L0:;

#line 1082 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 1082 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1082 "View.MemoryView"
  return __pyx_r;

#line 1082 "View.MemoryView"
}

/* "View.MemoryView":1089
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

#line 1089 "View.MemoryView"


#line 1089 "View.MemoryView"
static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_memviewslice) {
  PyObject *(*__pyx_v_to_object_func)(char *)
#line 1089 "View.MemoryView"
;
  int (*__pyx_v_to_dtype_func)(char *, PyObject *)
#line 1089 "View.MemoryView"
;

#line 1089 "View.MemoryView"
  PyObject *__pyx_r = NULL;

#line 1089 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1089 "View.MemoryView"
  int __pyx_t_1;

#line 1089 "View.MemoryView"
  int __pyx_t_2;

#line 1089 "View.MemoryView"
  PyObject *(*__pyx_t_3)(char *);

#line 1089 "View.MemoryView"
  int (*__pyx_t_4)(char *, PyObject *);

#line 1089 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 1089 "View.MemoryView"
  int __pyx_lineno = 0;

#line 1089 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 1089 "View.MemoryView"
  int __pyx_clineno = 0;

#line 1089 "View.MemoryView"
  __Pyx_RefNannySetupContext("memoryview_copy_from_slice", 0);

  /* "View.MemoryView":1096
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */

#line 1096 "View.MemoryView"
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 

#line 1096 "View.MemoryView"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 1096 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":1097
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func             # <<<<<<<<<<<<<<
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 */

#line 1097 "View.MemoryView"
    __pyx_t_3 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_object_func;

#line 1097 "View.MemoryView"
    __pyx_v_to_object_func = __pyx_t_3;

    /* "View.MemoryView":1098
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func             # <<<<<<<<<<<<<<
 *     else:
 *         to_object_func = NULL
 */

#line 1098 "View.MemoryView"
    __pyx_t_4 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_dtype_func;

#line 1098 "View.MemoryView"
    __pyx_v_to_dtype_func = __pyx_t_4;

    /* "View.MemoryView":1096
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */

#line 1096 "View.MemoryView"
    goto __pyx_L3;

#line 1096 "View.MemoryView"
  }

  /* "View.MemoryView":1100
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 *         to_object_func = NULL             # <<<<<<<<<<<<<<
 *         to_dtype_func = NULL
 * 
 */

#line 1100 "View.MemoryView"
  /*else*/ {

#line 1100 "View.MemoryView"
    __pyx_v_to_object_func = NULL;

    /* "View.MemoryView":1101
 *     else:
 *         to_object_func = NULL
 *         to_dtype_func = NULL             # <<<<<<<<<<<<<<
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 */

#line 1101 "View.MemoryView"
    __pyx_v_to_dtype_func = NULL;

#line 1101 "View.MemoryView"
  }

#line 1101 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":1103
 *         to_dtype_func = NULL
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,             # <<<<<<<<<<<<<<
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)
 */

#line 1103 "View.MemoryView"
  __Pyx_XDECREF(__pyx_r);

  /* "View.MemoryView":1105
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1105 "View.MemoryView"
  __pyx_t_5 = __pyx_memoryview_fromslice((__pyx_v_memviewslice[0]), __pyx_v_memview->view.ndim, __pyx_v_to_object_func, __pyx_v_to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 1103, __pyx_L1_error)

#line 1105 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_5);

#line 1105 "View.MemoryView"
  __pyx_r = __pyx_t_5;

#line 1105 "View.MemoryView"
  __pyx_t_5 = 0;

#line 1105 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":1089
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

#line 1089 "View.MemoryView"


#line 1089 "View.MemoryView"
  /* function exit code */

#line 1089 "View.MemoryView"
  __pyx_L1_error:;

#line 1089 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 1089 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy_from_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1089 "View.MemoryView"
  __pyx_r = 0;

#line 1089 "View.MemoryView"
  __pyx_L0:;

#line 1089 "View.MemoryView"
  __Pyx_XGIVEREF(__pyx_r);

#line 1089 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1089 "View.MemoryView"
  return __pyx_r;

#line 1089 "View.MemoryView"
}

/* "View.MemoryView":1111
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:             # <<<<<<<<<<<<<<
 *     if arg < 0:
 *         return -arg
 */

#line 1111 "View.MemoryView"


#line 1111 "View.MemoryView"
static Py_ssize_t abs_py_ssize_t(Py_ssize_t __pyx_v_arg) {

#line 1111 "View.MemoryView"
  Py_ssize_t __pyx_r;

#line 1111 "View.MemoryView"
  int __pyx_t_1;

  /* "View.MemoryView":1112
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:             # <<<<<<<<<<<<<<
 *         return -arg
 *     else:
 */

#line 1112 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_arg < 0) != 0);

#line 1112 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":1113
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:
 *         return -arg             # <<<<<<<<<<<<<<
 *     else:
 *         return arg
 */

#line 1113 "View.MemoryView"
    __pyx_r = (-__pyx_v_arg);

#line 1113 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":1112
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:             # <<<<<<<<<<<<<<
 *         return -arg
 *     else:
 */

#line 1112 "View.MemoryView"
  }

  /* "View.MemoryView":1115
 *         return -arg
 *     else:
 *         return arg             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_get_best_slice_order')
 */

#line 1115 "View.MemoryView"
  /*else*/ {

#line 1115 "View.MemoryView"
    __pyx_r = __pyx_v_arg;

#line 1115 "View.MemoryView"
    goto __pyx_L0;

#line 1115 "View.MemoryView"
  }

  /* "View.MemoryView":1111
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:             # <<<<<<<<<<<<<<
 *     if arg < 0:
 *         return -arg
 */

#line 1111 "View.MemoryView"


#line 1111 "View.MemoryView"
  /* function exit code */

#line 1111 "View.MemoryView"
  __pyx_L0:;

#line 1111 "View.MemoryView"
  return __pyx_r;

#line 1111 "View.MemoryView"
}

/* "View.MemoryView":1118
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

#line 1118 "View.MemoryView"


#line 1118 "View.MemoryView"
static char __pyx_get_best_slice_order(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim) {
  int __pyx_v_i
#line 1118 "View.MemoryView"
;
  Py_ssize_t __pyx_v_c_stride
#line 1118 "View.MemoryView"
;
  Py_ssize_t __pyx_v_f_stride
#line 1118 "View.MemoryView"
;

#line 1118 "View.MemoryView"
  char __pyx_r;

#line 1118 "View.MemoryView"
  int __pyx_t_1;

#line 1118 "View.MemoryView"
  int __pyx_t_2;

#line 1118 "View.MemoryView"
  int __pyx_t_3;

#line 1118 "View.MemoryView"
  int __pyx_t_4;

  /* "View.MemoryView":1123
 *     """
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t f_stride = 0
 * 
 */

#line 1123 "View.MemoryView"
  __pyx_v_c_stride = 0;

  /* "View.MemoryView":1124
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0
 *     cdef Py_ssize_t f_stride = 0             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */

#line 1124 "View.MemoryView"
  __pyx_v_f_stride = 0;

  /* "View.MemoryView":1126
 *     cdef Py_ssize_t f_stride = 0
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 */

#line 1126 "View.MemoryView"
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {

#line 1126 "View.MemoryView"
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1127
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */

#line 1127 "View.MemoryView"
    __pyx_t_2 = (((__pyx_v_mslice->shape[__pyx_v_i]) > 1) != 0);

#line 1127 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":1128
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */

#line 1128 "View.MemoryView"
      __pyx_v_c_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1129
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */

#line 1129 "View.MemoryView"
      goto __pyx_L4_break;

      /* "View.MemoryView":1127
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */

#line 1127 "View.MemoryView"
    }

#line 1127 "View.MemoryView"
  }

#line 1127 "View.MemoryView"
  __pyx_L4_break:;

  /* "View.MemoryView":1131
 *             break
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 */

#line 1131 "View.MemoryView"
  __pyx_t_1 = __pyx_v_ndim;

#line 1131 "View.MemoryView"
  __pyx_t_3 = __pyx_t_1;

#line 1131 "View.MemoryView"
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 1131 "View.MemoryView"
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1132
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */

#line 1132 "View.MemoryView"
    __pyx_t_2 = (((__pyx_v_mslice->shape[__pyx_v_i]) > 1) != 0);

#line 1132 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":1133
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */

#line 1133 "View.MemoryView"
      __pyx_v_f_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1134
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 */

#line 1134 "View.MemoryView"
      goto __pyx_L7_break;

      /* "View.MemoryView":1132
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */

#line 1132 "View.MemoryView"
    }

#line 1132 "View.MemoryView"
  }

#line 1132 "View.MemoryView"
  __pyx_L7_break:;

  /* "View.MemoryView":1136
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */

#line 1136 "View.MemoryView"
  __pyx_t_2 = ((abs_py_ssize_t(__pyx_v_c_stride) <= abs_py_ssize_t(__pyx_v_f_stride)) != 0);

#line 1136 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":1137
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 *         return 'C'             # <<<<<<<<<<<<<<
 *     else:
 *         return 'F'
 */

#line 1137 "View.MemoryView"
    __pyx_r = 'C';

#line 1137 "View.MemoryView"
    goto __pyx_L0;

    /* "View.MemoryView":1136
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */

#line 1136 "View.MemoryView"
  }

  /* "View.MemoryView":1139
 *         return 'C'
 *     else:
 *         return 'F'             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */

#line 1139 "View.MemoryView"
  /*else*/ {

#line 1139 "View.MemoryView"
    __pyx_r = 'F';

#line 1139 "View.MemoryView"
    goto __pyx_L0;

#line 1139 "View.MemoryView"
  }

  /* "View.MemoryView":1118
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

#line 1118 "View.MemoryView"


#line 1118 "View.MemoryView"
  /* function exit code */

#line 1118 "View.MemoryView"
  __pyx_L0:;

#line 1118 "View.MemoryView"
  return __pyx_r;

#line 1118 "View.MemoryView"
}

/* "View.MemoryView":1142
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

#line 1142 "View.MemoryView"


#line 1142 "View.MemoryView"
static void _copy_strided_to_strided(char *__pyx_v_src_data, Py_ssize_t *__pyx_v_src_strides, char *__pyx_v_dst_data, Py_ssize_t *__pyx_v_dst_strides, Py_ssize_t *__pyx_v_src_shape, Py_ssize_t *__pyx_v_dst_shape, int __pyx_v_ndim, size_t __pyx_v_itemsize) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i
#line 1142 "View.MemoryView"
;
  CYTHON_UNUSED Py_ssize_t __pyx_v_src_extent
#line 1142 "View.MemoryView"
;
  Py_ssize_t __pyx_v_dst_extent
#line 1142 "View.MemoryView"
;
  Py_ssize_t __pyx_v_src_stride
#line 1142 "View.MemoryView"
;
  Py_ssize_t __pyx_v_dst_stride
#line 1142 "View.MemoryView"
;

#line 1142 "View.MemoryView"
  int __pyx_t_1;

#line 1142 "View.MemoryView"
  int __pyx_t_2;

#line 1142 "View.MemoryView"
  int __pyx_t_3;

#line 1142 "View.MemoryView"
  Py_ssize_t __pyx_t_4;

#line 1142 "View.MemoryView"
  Py_ssize_t __pyx_t_5;

#line 1142 "View.MemoryView"
  Py_ssize_t __pyx_t_6;

  /* "View.MemoryView":1149
 * 
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 */

#line 1149 "View.MemoryView"
  __pyx_v_src_extent = (__pyx_v_src_shape[0]);

  /* "View.MemoryView":1150
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 */

#line 1150 "View.MemoryView"
  __pyx_v_dst_extent = (__pyx_v_dst_shape[0]);

  /* "View.MemoryView":1151
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 */

#line 1151 "View.MemoryView"
  __pyx_v_src_stride = (__pyx_v_src_strides[0]);

  /* "View.MemoryView":1152
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */

#line 1152 "View.MemoryView"
  __pyx_v_dst_stride = (__pyx_v_dst_strides[0]);

  /* "View.MemoryView":1154
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 */

#line 1154 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_ndim == 1) != 0);

#line 1154 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":1155
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */

#line 1155 "View.MemoryView"
    __pyx_t_2 = ((__pyx_v_src_stride > 0) != 0);

#line 1155 "View.MemoryView"
    if (__pyx_t_2) {

#line 1155 "View.MemoryView"
    } else {

#line 1155 "View.MemoryView"
      __pyx_t_1 = __pyx_t_2;

#line 1155 "View.MemoryView"
      goto __pyx_L5_bool_binop_done;

#line 1155 "View.MemoryView"
    }

#line 1155 "View.MemoryView"
    __pyx_t_2 = ((__pyx_v_dst_stride > 0) != 0);

#line 1155 "View.MemoryView"
    if (__pyx_t_2) {

#line 1155 "View.MemoryView"
    } else {

#line 1155 "View.MemoryView"
      __pyx_t_1 = __pyx_t_2;

#line 1155 "View.MemoryView"
      goto __pyx_L5_bool_binop_done;

#line 1155 "View.MemoryView"
    }

    /* "View.MemoryView":1156
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):             # <<<<<<<<<<<<<<
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 *        else:
 */

#line 1156 "View.MemoryView"
    __pyx_t_2 = (((size_t)__pyx_v_src_stride) == __pyx_v_itemsize);

#line 1156 "View.MemoryView"
    if (__pyx_t_2) {

#line 1156 "View.MemoryView"
      __pyx_t_2 = (__pyx_v_itemsize == ((size_t)__pyx_v_dst_stride));

#line 1156 "View.MemoryView"
    }

#line 1156 "View.MemoryView"
    __pyx_t_3 = (__pyx_t_2 != 0);

#line 1156 "View.MemoryView"
    __pyx_t_1 = __pyx_t_3;

#line 1156 "View.MemoryView"
    __pyx_L5_bool_binop_done:;

    /* "View.MemoryView":1155
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */

#line 1155 "View.MemoryView"
    if (__pyx_t_1) {

      /* "View.MemoryView":1157
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)             # <<<<<<<<<<<<<<
 *        else:
 *            for i in range(dst_extent):
 */

#line 1157 "View.MemoryView"
      (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, (__pyx_v_itemsize * __pyx_v_dst_extent)));

      /* "View.MemoryView":1155
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */

#line 1155 "View.MemoryView"
      goto __pyx_L4;

#line 1155 "View.MemoryView"
    }

    /* "View.MemoryView":1159
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 *        else:
 *            for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride
 */

#line 1159 "View.MemoryView"
    /*else*/ {

#line 1159 "View.MemoryView"
      __pyx_t_4 = __pyx_v_dst_extent;

#line 1159 "View.MemoryView"
      __pyx_t_5 = __pyx_t_4;

#line 1159 "View.MemoryView"
      for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {

#line 1159 "View.MemoryView"
        __pyx_v_i = __pyx_t_6;

        /* "View.MemoryView":1160
 *        else:
 *            for i in range(dst_extent):
 *                memcpy(dst_data, src_data, itemsize)             # <<<<<<<<<<<<<<
 *                src_data += src_stride
 *                dst_data += dst_stride
 */

#line 1160 "View.MemoryView"
        (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, __pyx_v_itemsize));

        /* "View.MemoryView":1161
 *            for i in range(dst_extent):
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride             # <<<<<<<<<<<<<<
 *                dst_data += dst_stride
 *     else:
 */

#line 1161 "View.MemoryView"
        __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

        /* "View.MemoryView":1162
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride
 *                dst_data += dst_stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(dst_extent):
 */

#line 1162 "View.MemoryView"
        __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);

#line 1162 "View.MemoryView"
      }

#line 1162 "View.MemoryView"
    }

#line 1162 "View.MemoryView"
    __pyx_L4:;

    /* "View.MemoryView":1154
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 */

#line 1154 "View.MemoryView"
    goto __pyx_L3;

#line 1154 "View.MemoryView"
  }

  /* "View.MemoryView":1164
 *                dst_data += dst_stride
 *     else:
 *         for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *             _copy_strided_to_strided(src_data, src_strides + 1,
 *                                      dst_data, dst_strides + 1,
 */

#line 1164 "View.MemoryView"
  /*else*/ {

#line 1164 "View.MemoryView"
    __pyx_t_4 = __pyx_v_dst_extent;

#line 1164 "View.MemoryView"
    __pyx_t_5 = __pyx_t_4;

#line 1164 "View.MemoryView"
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {

#line 1164 "View.MemoryView"
      __pyx_v_i = __pyx_t_6;

      /* "View.MemoryView":1165
 *     else:
 *         for i in range(dst_extent):
 *             _copy_strided_to_strided(src_data, src_strides + 1,             # <<<<<<<<<<<<<<
 *                                      dst_data, dst_strides + 1,
 *                                      src_shape + 1, dst_shape + 1,
 */

#line 1165 "View.MemoryView"
      _copy_strided_to_strided(__pyx_v_src_data, (__pyx_v_src_strides + 1), __pyx_v_dst_data, (__pyx_v_dst_strides + 1), (__pyx_v_src_shape + 1), (__pyx_v_dst_shape + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize);

      /* "View.MemoryView":1169
 *                                      src_shape + 1, dst_shape + 1,
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride             # <<<<<<<<<<<<<<
 *             dst_data += dst_stride
 * 
 */

#line 1169 "View.MemoryView"
      __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

      /* "View.MemoryView":1170
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride
 *             dst_data += dst_stride             # <<<<<<<<<<<<<<
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,
 */

#line 1170 "View.MemoryView"
      __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);

#line 1170 "View.MemoryView"
    }

#line 1170 "View.MemoryView"
  }

#line 1170 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":1142
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

#line 1142 "View.MemoryView"


#line 1142 "View.MemoryView"
  /* function exit code */

#line 1142 "View.MemoryView"
}

/* "View.MemoryView":1172
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 */

#line 1172 "View.MemoryView"


#line 1172 "View.MemoryView"
static void copy_strided_to_strided(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize) {

  /* "View.MemoryView":1175
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 *     _copy_strided_to_strided(src.data, src.strides, dst.data, dst.strides,             # <<<<<<<<<<<<<<
 *                              src.shape, dst.shape, ndim, itemsize)
 * 
 */

#line 1175 "View.MemoryView"
  _copy_strided_to_strided(__pyx_v_src->data, __pyx_v_src->strides, __pyx_v_dst->data, __pyx_v_dst->strides, __pyx_v_src->shape, __pyx_v_dst->shape, __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1172
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 */

#line 1172 "View.MemoryView"


#line 1172 "View.MemoryView"
  /* function exit code */

#line 1172 "View.MemoryView"
}

/* "View.MemoryView":1179
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 */

#line 1179 "View.MemoryView"


#line 1179 "View.MemoryView"
static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *__pyx_v_src, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_shape
#line 1179 "View.MemoryView"
;
  Py_ssize_t __pyx_v_size
#line 1179 "View.MemoryView"
;

#line 1179 "View.MemoryView"
  Py_ssize_t __pyx_r;

#line 1179 "View.MemoryView"
  Py_ssize_t __pyx_t_1;

#line 1179 "View.MemoryView"
  Py_ssize_t *__pyx_t_2;

#line 1179 "View.MemoryView"
  Py_ssize_t *__pyx_t_3;

#line 1179 "View.MemoryView"
  Py_ssize_t *__pyx_t_4;

  /* "View.MemoryView":1181
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     for shape in src.shape[:ndim]:
 */

#line 1181 "View.MemoryView"
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;

#line 1181 "View.MemoryView"
  __pyx_v_size = __pyx_t_1;

  /* "View.MemoryView":1183
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 * 
 *     for shape in src.shape[:ndim]:             # <<<<<<<<<<<<<<
 *         size *= shape
 * 
 */

#line 1183 "View.MemoryView"
  __pyx_t_3 = (__pyx_v_src->shape + __pyx_v_ndim);

#line 1183 "View.MemoryView"
  for (__pyx_t_4 = __pyx_v_src->shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {

#line 1183 "View.MemoryView"
    __pyx_t_2 = __pyx_t_4;

#line 1183 "View.MemoryView"
    __pyx_v_shape = (__pyx_t_2[0]);

    /* "View.MemoryView":1184
 * 
 *     for shape in src.shape[:ndim]:
 *         size *= shape             # <<<<<<<<<<<<<<
 * 
 *     return size
 */

#line 1184 "View.MemoryView"
    __pyx_v_size = (__pyx_v_size * __pyx_v_shape);

#line 1184 "View.MemoryView"
  }

  /* "View.MemoryView":1186
 *         size *= shape
 * 
 *     return size             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_fill_contig_strides_array')
 */

#line 1186 "View.MemoryView"
  __pyx_r = __pyx_v_size;

#line 1186 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":1179
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 */

#line 1179 "View.MemoryView"


#line 1179 "View.MemoryView"
  /* function exit code */

#line 1179 "View.MemoryView"
  __pyx_L0:;

#line 1179 "View.MemoryView"
  return __pyx_r;

#line 1179 "View.MemoryView"
}

/* "View.MemoryView":1189
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) nogil:
 */

#line 1189 "View.MemoryView"


#line 1189 "View.MemoryView"
static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, Py_ssize_t __pyx_v_stride, int __pyx_v_ndim, char __pyx_v_order) {
  int __pyx_v_idx
#line 1189 "View.MemoryView"
;

#line 1189 "View.MemoryView"
  Py_ssize_t __pyx_r;

#line 1189 "View.MemoryView"
  int __pyx_t_1;

#line 1189 "View.MemoryView"
  int __pyx_t_2;

#line 1189 "View.MemoryView"
  int __pyx_t_3;

#line 1189 "View.MemoryView"
  int __pyx_t_4;

  /* "View.MemoryView":1198
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */

#line 1198 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_order == 'F') != 0);

#line 1198 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":1199
 * 
 *     if order == 'F':
 *         for idx in range(ndim):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride *= shape[idx]
 */

#line 1199 "View.MemoryView"
    __pyx_t_2 = __pyx_v_ndim;

#line 1199 "View.MemoryView"
    __pyx_t_3 = __pyx_t_2;

#line 1199 "View.MemoryView"
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 1199 "View.MemoryView"
      __pyx_v_idx = __pyx_t_4;

      /* "View.MemoryView":1200
 *     if order == 'F':
 *         for idx in range(ndim):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride *= shape[idx]
 *     else:
 */

#line 1200 "View.MemoryView"
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1201
 *         for idx in range(ndim):
 *             strides[idx] = stride
 *             stride *= shape[idx]             # <<<<<<<<<<<<<<
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 */

#line 1201 "View.MemoryView"
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));

#line 1201 "View.MemoryView"
    }

    /* "View.MemoryView":1198
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */

#line 1198 "View.MemoryView"
    goto __pyx_L3;

#line 1198 "View.MemoryView"
  }

  /* "View.MemoryView":1203
 *             stride *= shape[idx]
 *     else:
 *         for idx in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride *= shape[idx]
 */

#line 1203 "View.MemoryView"
  /*else*/ {

#line 1203 "View.MemoryView"
    for (__pyx_t_2 = (__pyx_v_ndim - 1); __pyx_t_2 > -1; __pyx_t_2-=1) {

#line 1203 "View.MemoryView"
      __pyx_v_idx = __pyx_t_2;

      /* "View.MemoryView":1204
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride *= shape[idx]
 * 
 */

#line 1204 "View.MemoryView"
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1205
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride
 *             stride *= shape[idx]             # <<<<<<<<<<<<<<
 * 
 *     return stride
 */

#line 1205 "View.MemoryView"
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));

#line 1205 "View.MemoryView"
    }

#line 1205 "View.MemoryView"
  }

#line 1205 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":1207
 *             stride *= shape[idx]
 * 
 *     return stride             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 */

#line 1207 "View.MemoryView"
  __pyx_r = __pyx_v_stride;

#line 1207 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":1189
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) nogil:
 */

#line 1189 "View.MemoryView"


#line 1189 "View.MemoryView"
  /* function exit code */

#line 1189 "View.MemoryView"
  __pyx_L0:;

#line 1189 "View.MemoryView"
  return __pyx_r;

#line 1189 "View.MemoryView"
}

/* "View.MemoryView":1210
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

#line 1210 "View.MemoryView"


#line 1210 "View.MemoryView"
static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_tmpslice, char __pyx_v_order, int __pyx_v_ndim) {
  int __pyx_v_i
#line 1210 "View.MemoryView"
;
  void *__pyx_v_result
#line 1210 "View.MemoryView"
;
  size_t __pyx_v_itemsize
#line 1210 "View.MemoryView"
;
  size_t __pyx_v_size
#line 1210 "View.MemoryView"
;

#line 1210 "View.MemoryView"
  void *__pyx_r;

#line 1210 "View.MemoryView"
  Py_ssize_t __pyx_t_1;

#line 1210 "View.MemoryView"
  int __pyx_t_2;

#line 1210 "View.MemoryView"
  int __pyx_t_3;

#line 1210 "View.MemoryView"
  struct __pyx_memoryview_obj *__pyx_t_4;

#line 1210 "View.MemoryView"
  int __pyx_t_5;

#line 1210 "View.MemoryView"
  int __pyx_t_6;

#line 1210 "View.MemoryView"
  int __pyx_lineno = 0;

#line 1210 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 1210 "View.MemoryView"
  int __pyx_clineno = 0;

  /* "View.MemoryView":1221
 *     cdef void *result
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 */

#line 1221 "View.MemoryView"
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;

#line 1221 "View.MemoryView"
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1222
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef size_t size = slice_get_size(src, ndim)             # <<<<<<<<<<<<<<
 * 
 *     result = malloc(size)
 */

#line 1222 "View.MemoryView"
  __pyx_v_size = __pyx_memoryview_slice_get_size(__pyx_v_src, __pyx_v_ndim);

  /* "View.MemoryView":1224
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 *     result = malloc(size)             # <<<<<<<<<<<<<<
 *     if not result:
 *         _err(MemoryError, NULL)
 */

#line 1224 "View.MemoryView"
  __pyx_v_result = malloc(__pyx_v_size);

  /* "View.MemoryView":1225
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err(MemoryError, NULL)
 * 
 */

#line 1225 "View.MemoryView"
  __pyx_t_2 = ((!(__pyx_v_result != 0)) != 0);

#line 1225 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":1226
 *     result = malloc(size)
 *     if not result:
 *         _err(MemoryError, NULL)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1226 "View.MemoryView"
    __pyx_t_3 = __pyx_memoryview_err(__pyx_builtin_MemoryError, NULL); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(3, 1226, __pyx_L1_error)

    /* "View.MemoryView":1225
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err(MemoryError, NULL)
 * 
 */

#line 1225 "View.MemoryView"
  }

  /* "View.MemoryView":1229
 * 
 * 
 *     tmpslice.data = <char *> result             # <<<<<<<<<<<<<<
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 */

#line 1229 "View.MemoryView"
  __pyx_v_tmpslice->data = ((char *)__pyx_v_result);

  /* "View.MemoryView":1230
 * 
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview             # <<<<<<<<<<<<<<
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 */

#line 1230 "View.MemoryView"
  __pyx_t_4 = __pyx_v_src->memview;

#line 1230 "View.MemoryView"
  __pyx_v_tmpslice->memview = __pyx_t_4;

  /* "View.MemoryView":1231
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1
 */

#line 1231 "View.MemoryView"
  __pyx_t_3 = __pyx_v_ndim;

#line 1231 "View.MemoryView"
  __pyx_t_5 = __pyx_t_3;

#line 1231 "View.MemoryView"
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {

#line 1231 "View.MemoryView"
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1232
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]             # <<<<<<<<<<<<<<
 *         tmpslice.suboffsets[i] = -1
 * 
 */

#line 1232 "View.MemoryView"
    (__pyx_v_tmpslice->shape[__pyx_v_i]) = (__pyx_v_src->shape[__pyx_v_i]);

    /* "View.MemoryView":1233
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize,
 */

#line 1233 "View.MemoryView"
    (__pyx_v_tmpslice->suboffsets[__pyx_v_i]) = -1L;

#line 1233 "View.MemoryView"
  }

  /* "View.MemoryView":1235
 *         tmpslice.suboffsets[i] = -1
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize,             # <<<<<<<<<<<<<<
 *                               ndim, order)
 * 
 */

#line 1235 "View.MemoryView"
  (void)(__pyx_fill_contig_strides_array((&(__pyx_v_tmpslice->shape[0])), (&(__pyx_v_tmpslice->strides[0])), __pyx_v_itemsize, __pyx_v_ndim, __pyx_v_order));

  /* "View.MemoryView":1239
 * 
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0
 */

#line 1239 "View.MemoryView"
  __pyx_t_3 = __pyx_v_ndim;

#line 1239 "View.MemoryView"
  __pyx_t_5 = __pyx_t_3;

#line 1239 "View.MemoryView"
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {

#line 1239 "View.MemoryView"
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1240
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */

#line 1240 "View.MemoryView"
    __pyx_t_2 = (((__pyx_v_tmpslice->shape[__pyx_v_i]) == 1) != 0);

#line 1240 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":1241
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0             # <<<<<<<<<<<<<<
 * 
 *     if slice_is_contig(src[0], order, ndim):
 */

#line 1241 "View.MemoryView"
      (__pyx_v_tmpslice->strides[__pyx_v_i]) = 0;

      /* "View.MemoryView":1240
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */

#line 1240 "View.MemoryView"
    }

#line 1240 "View.MemoryView"
  }

  /* "View.MemoryView":1243
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */

#line 1243 "View.MemoryView"
  __pyx_t_2 = (__pyx_memviewslice_is_contig((__pyx_v_src[0]), __pyx_v_order, __pyx_v_ndim) != 0);

#line 1243 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":1244
 * 
 *     if slice_is_contig(src[0], order, ndim):
 *         memcpy(result, src.data, size)             # <<<<<<<<<<<<<<
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 */

#line 1244 "View.MemoryView"
    (void)(memcpy(__pyx_v_result, __pyx_v_src->data, __pyx_v_size));

    /* "View.MemoryView":1243
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */

#line 1243 "View.MemoryView"
    goto __pyx_L9;

#line 1243 "View.MemoryView"
  }

  /* "View.MemoryView":1246
 *         memcpy(result, src.data, size)
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)             # <<<<<<<<<<<<<<
 * 
 *     return result
 */

#line 1246 "View.MemoryView"
  /*else*/ {

#line 1246 "View.MemoryView"
    copy_strided_to_strided(__pyx_v_src, __pyx_v_tmpslice, __pyx_v_ndim, __pyx_v_itemsize);

#line 1246 "View.MemoryView"
  }

#line 1246 "View.MemoryView"
  __pyx_L9:;

  /* "View.MemoryView":1248
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1248 "View.MemoryView"
  __pyx_r = __pyx_v_result;

#line 1248 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":1210
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

#line 1210 "View.MemoryView"


#line 1210 "View.MemoryView"
  /* function exit code */

#line 1210 "View.MemoryView"
  __pyx_L1_error:;

#line 1210 "View.MemoryView"
  {

#line 1210 "View.MemoryView"
    #ifdef WITH_THREAD
    PyGILState_STATE 
#line 1210 "View.MemoryView"
__pyx_gilstate_save = __Pyx_PyGILState_Ensure();

#line 1210 "View.MemoryView"
    #endif

#line 1210 "View.MemoryView"
    __Pyx_AddTraceback("View.MemoryView.copy_data_to_temp", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1210 "View.MemoryView"
    #ifdef WITH_THREAD

#line 1210 "View.MemoryView"
    __Pyx_PyGILState_Release(__pyx_gilstate_save);

#line 1210 "View.MemoryView"
    #endif

#line 1210 "View.MemoryView"
  }

#line 1210 "View.MemoryView"
  __pyx_r = NULL;

#line 1210 "View.MemoryView"
  __pyx_L0:;

#line 1210 "View.MemoryView"
  return __pyx_r;

#line 1210 "View.MemoryView"
}

/* "View.MemoryView":1253
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 */

#line 1253 "View.MemoryView"


#line 1253 "View.MemoryView"
static int __pyx_memoryview_err_extents(int __pyx_v_i, Py_ssize_t __pyx_v_extent1, Py_ssize_t __pyx_v_extent2) {

#line 1253 "View.MemoryView"
  int __pyx_r;

#line 1253 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1253 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 1253 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 1253 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 1253 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 1253 "View.MemoryView"
  int __pyx_lineno = 0;

#line 1253 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 1253 "View.MemoryView"
  int __pyx_clineno = 0;

#line 1253 "View.MemoryView"
  #ifdef WITH_THREAD
  PyGILState_STATE 
#line 1253 "View.MemoryView"
__pyx_gilstate_save = __Pyx_PyGILState_Ensure();

#line 1253 "View.MemoryView"
  #endif

#line 1253 "View.MemoryView"
  __Pyx_RefNannySetupContext("_err_extents", 0);

  /* "View.MemoryView":1256
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 *                                                         (i, extent1, extent2))             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err_dim')
 */

#line 1256 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 1256, __pyx_L1_error)

#line 1256 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 1256 "View.MemoryView"
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_extent1); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 1256, __pyx_L1_error)

#line 1256 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 1256 "View.MemoryView"
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_extent2); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 1256, __pyx_L1_error)

#line 1256 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 1256 "View.MemoryView"
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 1256, __pyx_L1_error)

#line 1256 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_4);

#line 1256 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_1);

#line 1256 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);

#line 1256 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_2);

#line 1256 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);

#line 1256 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_3);

#line 1256 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);

#line 1256 "View.MemoryView"
  __pyx_t_1 = 0;

#line 1256 "View.MemoryView"
  __pyx_t_2 = 0;

#line 1256 "View.MemoryView"
  __pyx_t_3 = 0;

  /* "View.MemoryView":1255
 * cdef int _err_extents(int i, Py_ssize_t extent1,
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %             # <<<<<<<<<<<<<<
 *                                                         (i, extent1, extent2))
 * 
 */

#line 1255 "View.MemoryView"
  __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_got_differing_extents_in_dimensi, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 1255, __pyx_L1_error)

#line 1255 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 1255 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 1255 "View.MemoryView"
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 1255, __pyx_L1_error)

#line 1255 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_4);

#line 1255 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1255 "View.MemoryView"
  __Pyx_Raise(__pyx_t_4, 0, 0, 0);

#line 1255 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 1255 "View.MemoryView"
  __PYX_ERR(3, 1255, __pyx_L1_error)

  /* "View.MemoryView":1253
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 */

#line 1253 "View.MemoryView"


#line 1253 "View.MemoryView"
  /* function exit code */

#line 1253 "View.MemoryView"
  __pyx_L1_error:;

#line 1253 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 1253 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 1253 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 1253 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_4);

#line 1253 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView._err_extents", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1253 "View.MemoryView"
  __pyx_r = -1;

#line 1253 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1253 "View.MemoryView"
  #ifdef WITH_THREAD

#line 1253 "View.MemoryView"
  __Pyx_PyGILState_Release(__pyx_gilstate_save);

#line 1253 "View.MemoryView"
  #endif

#line 1253 "View.MemoryView"
  return __pyx_r;

#line 1253 "View.MemoryView"
}

/* "View.MemoryView":1259
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise error(msg.decode('ascii') % dim)
 * 
 */

#line 1259 "View.MemoryView"


#line 1259 "View.MemoryView"
static int __pyx_memoryview_err_dim(PyObject *__pyx_v_error, char *__pyx_v_msg, int __pyx_v_dim) {

#line 1259 "View.MemoryView"
  int __pyx_r;

#line 1259 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1259 "View.MemoryView"
  PyObject *__pyx_t_1 = NULL;

#line 1259 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 1259 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 1259 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 1259 "View.MemoryView"
  int __pyx_lineno = 0;

#line 1259 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 1259 "View.MemoryView"
  int __pyx_clineno = 0;

#line 1259 "View.MemoryView"
  #ifdef WITH_THREAD
  PyGILState_STATE 
#line 1259 "View.MemoryView"
__pyx_gilstate_save = __Pyx_PyGILState_Ensure();

#line 1259 "View.MemoryView"
  #endif

#line 1259 "View.MemoryView"
  __Pyx_RefNannySetupContext("_err_dim", 0);

#line 1259 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_error);

  /* "View.MemoryView":1260
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:
 *     raise error(msg.decode('ascii') % dim)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err')
 */

#line 1260 "View.MemoryView"
  __pyx_t_2 = __Pyx_decode_c_string(__pyx_v_msg, 0, strlen(__pyx_v_msg), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 1260, __pyx_L1_error)

#line 1260 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_2);

#line 1260 "View.MemoryView"
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 1260, __pyx_L1_error)

#line 1260 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_3);

#line 1260 "View.MemoryView"
  __pyx_t_4 = PyUnicode_Format(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 1260, __pyx_L1_error)

#line 1260 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_4);

#line 1260 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 1260 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1260 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_error);
  __pyx_t_3 = __pyx_v_error; 
#line 1260 "View.MemoryView"
__pyx_t_2 = NULL;

#line 1260 "View.MemoryView"
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 1260 "View.MemoryView"
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);

#line 1260 "View.MemoryView"
    if (likely(__pyx_t_2)) {

#line 1260 "View.MemoryView"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 1260 "View.MemoryView"
      __Pyx_INCREF(__pyx_t_2);

#line 1260 "View.MemoryView"
      __Pyx_INCREF(function);

#line 1260 "View.MemoryView"
      __Pyx_DECREF_SET(__pyx_t_3, function);

#line 1260 "View.MemoryView"
    }

#line 1260 "View.MemoryView"
  }

#line 1260 "View.MemoryView"
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);

#line 1260 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 1260 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 1260 "View.MemoryView"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 1260, __pyx_L1_error)

#line 1260 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 1260 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1260 "View.MemoryView"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 1260 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1260 "View.MemoryView"
  __PYX_ERR(3, 1260, __pyx_L1_error)

  /* "View.MemoryView":1259
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise error(msg.decode('ascii') % dim)
 * 
 */

#line 1259 "View.MemoryView"


#line 1259 "View.MemoryView"
  /* function exit code */

#line 1259 "View.MemoryView"
  __pyx_L1_error:;

#line 1259 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_1);

#line 1259 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 1259 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 1259 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_4);

#line 1259 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView._err_dim", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1259 "View.MemoryView"
  __pyx_r = -1;

#line 1259 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_error);

#line 1259 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1259 "View.MemoryView"
  #ifdef WITH_THREAD

#line 1259 "View.MemoryView"
  __Pyx_PyGILState_Release(__pyx_gilstate_save);

#line 1259 "View.MemoryView"
  #endif

#line 1259 "View.MemoryView"
  return __pyx_r;

#line 1259 "View.MemoryView"
}

/* "View.MemoryView":1263
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))
 */

#line 1263 "View.MemoryView"


#line 1263 "View.MemoryView"
static int __pyx_memoryview_err(PyObject *__pyx_v_error, char *__pyx_v_msg) {

#line 1263 "View.MemoryView"
  int __pyx_r;

#line 1263 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1263 "View.MemoryView"
  int __pyx_t_1;

#line 1263 "View.MemoryView"
  PyObject *__pyx_t_2 = NULL;

#line 1263 "View.MemoryView"
  PyObject *__pyx_t_3 = NULL;

#line 1263 "View.MemoryView"
  PyObject *__pyx_t_4 = NULL;

#line 1263 "View.MemoryView"
  PyObject *__pyx_t_5 = NULL;

#line 1263 "View.MemoryView"
  int __pyx_lineno = 0;

#line 1263 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 1263 "View.MemoryView"
  int __pyx_clineno = 0;

#line 1263 "View.MemoryView"
  #ifdef WITH_THREAD
  PyGILState_STATE 
#line 1263 "View.MemoryView"
__pyx_gilstate_save = __Pyx_PyGILState_Ensure();

#line 1263 "View.MemoryView"
  #endif

#line 1263 "View.MemoryView"
  __Pyx_RefNannySetupContext("_err", 0);

#line 1263 "View.MemoryView"
  __Pyx_INCREF(__pyx_v_error);

  /* "View.MemoryView":1264
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:             # <<<<<<<<<<<<<<
 *         raise error(msg.decode('ascii'))
 *     else:
 */

#line 1264 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_msg != NULL) != 0);

#line 1264 "View.MemoryView"
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":1265
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))             # <<<<<<<<<<<<<<
 *     else:
 *         raise error
 */

#line 1265 "View.MemoryView"
    __pyx_t_3 = __Pyx_decode_c_string(__pyx_v_msg, 0, strlen(__pyx_v_msg), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 1265, __pyx_L1_error)

#line 1265 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_3);

#line 1265 "View.MemoryView"
    __Pyx_INCREF(__pyx_v_error);
    __pyx_t_4 = __pyx_v_error; 
#line 1265 "View.MemoryView"
__pyx_t_5 = NULL;

#line 1265 "View.MemoryView"
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {

#line 1265 "View.MemoryView"
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 1265 "View.MemoryView"
      if (likely(__pyx_t_5)) {

#line 1265 "View.MemoryView"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 1265 "View.MemoryView"
        __Pyx_INCREF(__pyx_t_5);

#line 1265 "View.MemoryView"
        __Pyx_INCREF(function);

#line 1265 "View.MemoryView"
        __Pyx_DECREF_SET(__pyx_t_4, function);

#line 1265 "View.MemoryView"
      }

#line 1265 "View.MemoryView"
    }

#line 1265 "View.MemoryView"
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);

#line 1265 "View.MemoryView"
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 1265 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1265 "View.MemoryView"
    if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 1265, __pyx_L1_error)

#line 1265 "View.MemoryView"
    __Pyx_GOTREF(__pyx_t_2);

#line 1265 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 1265 "View.MemoryView"
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);

#line 1265 "View.MemoryView"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 1265 "View.MemoryView"
    __PYX_ERR(3, 1265, __pyx_L1_error)

    /* "View.MemoryView":1264
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:             # <<<<<<<<<<<<<<
 *         raise error(msg.decode('ascii'))
 *     else:
 */

#line 1264 "View.MemoryView"
  }

  /* "View.MemoryView":1267
 *         raise error(msg.decode('ascii'))
 *     else:
 *         raise error             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_contents')
 */

#line 1267 "View.MemoryView"
  /*else*/ {

#line 1267 "View.MemoryView"
    __Pyx_Raise(__pyx_v_error, 0, 0, 0);

#line 1267 "View.MemoryView"
    __PYX_ERR(3, 1267, __pyx_L1_error)

#line 1267 "View.MemoryView"
  }

  /* "View.MemoryView":1263
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))
 */

#line 1263 "View.MemoryView"


#line 1263 "View.MemoryView"
  /* function exit code */

#line 1263 "View.MemoryView"
  __pyx_L1_error:;

#line 1263 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_2);

#line 1263 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_3);

#line 1263 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_4);

#line 1263 "View.MemoryView"
  __Pyx_XDECREF(__pyx_t_5);

#line 1263 "View.MemoryView"
  __Pyx_AddTraceback("View.MemoryView._err", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1263 "View.MemoryView"
  __pyx_r = -1;

#line 1263 "View.MemoryView"
  __Pyx_XDECREF(__pyx_v_error);

#line 1263 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1263 "View.MemoryView"
  #ifdef WITH_THREAD

#line 1263 "View.MemoryView"
  __Pyx_PyGILState_Release(__pyx_gilstate_save);

#line 1263 "View.MemoryView"
  #endif

#line 1263 "View.MemoryView"
  return __pyx_r;

#line 1263 "View.MemoryView"
}

/* "View.MemoryView":1270
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

#line 1270 "View.MemoryView"


#line 1270 "View.MemoryView"
static int __pyx_memoryview_copy_contents(__Pyx_memviewslice __pyx_v_src, __Pyx_memviewslice __pyx_v_dst, int __pyx_v_src_ndim, int __pyx_v_dst_ndim, int __pyx_v_dtype_is_object) {
  void *__pyx_v_tmpdata
#line 1270 "View.MemoryView"
;
  size_t __pyx_v_itemsize
#line 1270 "View.MemoryView"
;
  int __pyx_v_i
#line 1270 "View.MemoryView"
;
  char __pyx_v_order
#line 1270 "View.MemoryView"
;
  int __pyx_v_broadcasting
#line 1270 "View.MemoryView"
;
  int __pyx_v_direct_copy
#line 1270 "View.MemoryView"
;
  __Pyx_memviewslice __pyx_v_tmp
#line 1270 "View.MemoryView"
;
  int __pyx_v_ndim
#line 1270 "View.MemoryView"
;

#line 1270 "View.MemoryView"
  int __pyx_r;

#line 1270 "View.MemoryView"
  Py_ssize_t __pyx_t_1;

#line 1270 "View.MemoryView"
  int __pyx_t_2;

#line 1270 "View.MemoryView"
  int __pyx_t_3;

#line 1270 "View.MemoryView"
  int __pyx_t_4;

#line 1270 "View.MemoryView"
  int __pyx_t_5;

#line 1270 "View.MemoryView"
  int __pyx_t_6;

#line 1270 "View.MemoryView"
  void *__pyx_t_7;

#line 1270 "View.MemoryView"
  int __pyx_t_8;

#line 1270 "View.MemoryView"
  int __pyx_lineno = 0;

#line 1270 "View.MemoryView"
  const char *__pyx_filename = NULL;

#line 1270 "View.MemoryView"
  int __pyx_clineno = 0;

  /* "View.MemoryView":1278
 *     Check for overlapping memory and verify the shapes.
 *     """
 *     cdef void *tmpdata = NULL             # <<<<<<<<<<<<<<
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 */

#line 1278 "View.MemoryView"
  __pyx_v_tmpdata = NULL;

  /* "View.MemoryView":1279
 *     """
 *     cdef void *tmpdata = NULL
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 */

#line 1279 "View.MemoryView"
  __pyx_t_1 = __pyx_v_src.memview->view.itemsize;

#line 1279 "View.MemoryView"
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1281
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)             # <<<<<<<<<<<<<<
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False
 */

#line 1281 "View.MemoryView"
  __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_src), __pyx_v_src_ndim);

  /* "View.MemoryView":1282
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False             # <<<<<<<<<<<<<<
 *     cdef bint direct_copy = False
 *     cdef __Pyx_memviewslice tmp
 */

#line 1282 "View.MemoryView"
  __pyx_v_broadcasting = 0;

  /* "View.MemoryView":1283
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False             # <<<<<<<<<<<<<<
 *     cdef __Pyx_memviewslice tmp
 * 
 */

#line 1283 "View.MemoryView"
  __pyx_v_direct_copy = 0;

  /* "View.MemoryView":1286
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */

#line 1286 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_src_ndim < __pyx_v_dst_ndim) != 0);

#line 1286 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":1287
 * 
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 */

#line 1287 "View.MemoryView"
    __pyx_memoryview_broadcast_leading((&__pyx_v_src), __pyx_v_src_ndim, __pyx_v_dst_ndim);

    /* "View.MemoryView":1286
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */

#line 1286 "View.MemoryView"
    goto __pyx_L3;

#line 1286 "View.MemoryView"
  }

  /* "View.MemoryView":1288
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */

#line 1288 "View.MemoryView"
  __pyx_t_2 = ((__pyx_v_dst_ndim < __pyx_v_src_ndim) != 0);

#line 1288 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":1289
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)             # <<<<<<<<<<<<<<
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)
 */

#line 1289 "View.MemoryView"
    __pyx_memoryview_broadcast_leading((&__pyx_v_dst), __pyx_v_dst_ndim, __pyx_v_src_ndim);

    /* "View.MemoryView":1288
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */

#line 1288 "View.MemoryView"
  }

#line 1288 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":1291
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */

#line 1291 "View.MemoryView"
  __pyx_t_3 = __pyx_v_dst_ndim;

#line 1291 "View.MemoryView"
  __pyx_t_4 = __pyx_v_src_ndim;

#line 1291 "View.MemoryView"
  if (((__pyx_t_3 > __pyx_t_4) != 0)) {

#line 1291 "View.MemoryView"
    __pyx_t_5 = __pyx_t_3;

#line 1291 "View.MemoryView"
  } else {

#line 1291 "View.MemoryView"
    __pyx_t_5 = __pyx_t_4;

#line 1291 "View.MemoryView"
  }

#line 1291 "View.MemoryView"
  __pyx_v_ndim = __pyx_t_5;

  /* "View.MemoryView":1293
 *     cdef int ndim = max(src_ndim, dst_ndim)
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 */

#line 1293 "View.MemoryView"
  __pyx_t_5 = __pyx_v_ndim;

#line 1293 "View.MemoryView"
  __pyx_t_3 = __pyx_t_5;

#line 1293 "View.MemoryView"
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 1293 "View.MemoryView"
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1294
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */

#line 1294 "View.MemoryView"
    __pyx_t_2 = (((__pyx_v_src.shape[__pyx_v_i]) != (__pyx_v_dst.shape[__pyx_v_i])) != 0);

#line 1294 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":1295
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */

#line 1295 "View.MemoryView"
      __pyx_t_2 = (((__pyx_v_src.shape[__pyx_v_i]) == 1) != 0);

#line 1295 "View.MemoryView"
      if (__pyx_t_2) {

        /* "View.MemoryView":1296
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 *                 broadcasting = True             # <<<<<<<<<<<<<<
 *                 src.strides[i] = 0
 *             else:
 */

#line 1296 "View.MemoryView"
        __pyx_v_broadcasting = 1;

        /* "View.MemoryView":1297
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 *                 src.strides[i] = 0             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 */

#line 1297 "View.MemoryView"
        (__pyx_v_src.strides[__pyx_v_i]) = 0;

        /* "View.MemoryView":1295
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */

#line 1295 "View.MemoryView"
        goto __pyx_L7;

#line 1295 "View.MemoryView"
      }

      /* "View.MemoryView":1299
 *                 src.strides[i] = 0
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])             # <<<<<<<<<<<<<<
 * 
 *         if src.suboffsets[i] >= 0:
 */

#line 1299 "View.MemoryView"
      /*else*/ {

#line 1299 "View.MemoryView"
        __pyx_t_6 = __pyx_memoryview_err_extents(__pyx_v_i, (__pyx_v_dst.shape[__pyx_v_i]), (__pyx_v_src.shape[__pyx_v_i])); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(3, 1299, __pyx_L1_error)

#line 1299 "View.MemoryView"
      }

#line 1299 "View.MemoryView"
      __pyx_L7:;

      /* "View.MemoryView":1294
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */

#line 1294 "View.MemoryView"
    }

    /* "View.MemoryView":1301
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 */

#line 1301 "View.MemoryView"
    __pyx_t_2 = (((__pyx_v_src.suboffsets[__pyx_v_i]) >= 0) != 0);

#line 1301 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":1302
 * 
 *         if src.suboffsets[i] >= 0:
 *             _err_dim(ValueError, "Dimension %d is not direct", i)             # <<<<<<<<<<<<<<
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 */

#line 1302 "View.MemoryView"
      __pyx_t_6 = __pyx_memoryview_err_dim(__pyx_builtin_ValueError, ((char *)"Dimension %d is not direct"), __pyx_v_i); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(3, 1302, __pyx_L1_error)

      /* "View.MemoryView":1301
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 */

#line 1301 "View.MemoryView"
    }

#line 1301 "View.MemoryView"
  }

  /* "View.MemoryView":1304
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */

#line 1304 "View.MemoryView"
  __pyx_t_2 = (__pyx_slices_overlap((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize) != 0);

#line 1304 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":1306
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */

#line 1306 "View.MemoryView"
    __pyx_t_2 = ((!(__pyx_memviewslice_is_contig(__pyx_v_src, __pyx_v_order, __pyx_v_ndim) != 0)) != 0);

#line 1306 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":1307
 * 
 *         if not slice_is_contig(src, order, ndim):
 *             order = get_best_order(&dst, ndim)             # <<<<<<<<<<<<<<
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 */

#line 1307 "View.MemoryView"
      __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim);

      /* "View.MemoryView":1306
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */

#line 1306 "View.MemoryView"
    }

    /* "View.MemoryView":1309
 *             order = get_best_order(&dst, ndim)
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)             # <<<<<<<<<<<<<<
 *         src = tmp
 * 
 */

#line 1309 "View.MemoryView"
    __pyx_t_7 = __pyx_memoryview_copy_data_to_temp((&__pyx_v_src), (&__pyx_v_tmp), __pyx_v_order, __pyx_v_ndim); if (unlikely(__pyx_t_7 == ((void *)NULL))) __PYX_ERR(3, 1309, __pyx_L1_error)

#line 1309 "View.MemoryView"
    __pyx_v_tmpdata = __pyx_t_7;

    /* "View.MemoryView":1310
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 *         src = tmp             # <<<<<<<<<<<<<<
 * 
 *     if not broadcasting:
 */

#line 1310 "View.MemoryView"
    __pyx_v_src = __pyx_v_tmp;

    /* "View.MemoryView":1304
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */

#line 1304 "View.MemoryView"
  }

  /* "View.MemoryView":1312
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1312 "View.MemoryView"
  __pyx_t_2 = ((!(__pyx_v_broadcasting != 0)) != 0);

#line 1312 "View.MemoryView"
  if (__pyx_t_2) {

    /* "View.MemoryView":1315
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */

#line 1315 "View.MemoryView"
    __pyx_t_2 = (__pyx_memviewslice_is_contig(__pyx_v_src, 'C', __pyx_v_ndim) != 0);

#line 1315 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":1316
 * 
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)             # <<<<<<<<<<<<<<
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 */

#line 1316 "View.MemoryView"
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'C', __pyx_v_ndim);

      /* "View.MemoryView":1315
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */

#line 1315 "View.MemoryView"
      goto __pyx_L12;

#line 1315 "View.MemoryView"
    }

    /* "View.MemoryView":1317
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */

#line 1317 "View.MemoryView"
    __pyx_t_2 = (__pyx_memviewslice_is_contig(__pyx_v_src, 'F', __pyx_v_ndim) != 0);

#line 1317 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":1318
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)             # <<<<<<<<<<<<<<
 * 
 *         if direct_copy:
 */

#line 1318 "View.MemoryView"
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'F', __pyx_v_ndim);

      /* "View.MemoryView":1317
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */

#line 1317 "View.MemoryView"
    }

#line 1317 "View.MemoryView"
    __pyx_L12:;

    /* "View.MemoryView":1320
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 */

#line 1320 "View.MemoryView"
    __pyx_t_2 = (__pyx_v_direct_copy != 0);

#line 1320 "View.MemoryView"
    if (__pyx_t_2) {

      /* "View.MemoryView":1322
 *         if direct_copy:
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 */

#line 1322 "View.MemoryView"
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

      /* "View.MemoryView":1323
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))             # <<<<<<<<<<<<<<
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)
 */

#line 1323 "View.MemoryView"
      (void)(memcpy(__pyx_v_dst.data, __pyx_v_src.data, __pyx_memoryview_slice_get_size((&__pyx_v_src), __pyx_v_ndim)));

      /* "View.MemoryView":1324
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 *             free(tmpdata)
 *             return 0
 */

#line 1324 "View.MemoryView"
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

      /* "View.MemoryView":1325
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */

#line 1325 "View.MemoryView"
      free(__pyx_v_tmpdata);

      /* "View.MemoryView":1326
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):
 */

#line 1326 "View.MemoryView"
      __pyx_r = 0;

#line 1326 "View.MemoryView"
      goto __pyx_L0;

      /* "View.MemoryView":1320
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 */

#line 1320 "View.MemoryView"
    }

    /* "View.MemoryView":1312
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1312 "View.MemoryView"
  }

  /* "View.MemoryView":1328
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1328 "View.MemoryView"
  __pyx_t_2 = (__pyx_v_order == 'F');

#line 1328 "View.MemoryView"
  if (__pyx_t_2) {

#line 1328 "View.MemoryView"
    __pyx_t_2 = ('F' == __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim));

#line 1328 "View.MemoryView"
  }

#line 1328 "View.MemoryView"
  __pyx_t_8 = (__pyx_t_2 != 0);

#line 1328 "View.MemoryView"
  if (__pyx_t_8) {

    /* "View.MemoryView":1331
 * 
 * 
 *         transpose_memslice(&src)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&dst)
 * 
 */

#line 1331 "View.MemoryView"
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_src)); if (unlikely(__pyx_t_5 == ((int)0))) __PYX_ERR(3, 1331, __pyx_L1_error)

    /* "View.MemoryView":1332
 * 
 *         transpose_memslice(&src)
 *         transpose_memslice(&dst)             # <<<<<<<<<<<<<<
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 */

#line 1332 "View.MemoryView"
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_dst)); if (unlikely(__pyx_t_5 == ((int)0))) __PYX_ERR(3, 1332, __pyx_L1_error)

    /* "View.MemoryView":1328
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1328 "View.MemoryView"
  }

  /* "View.MemoryView":1334
 *         transpose_memslice(&dst)
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 */

#line 1334 "View.MemoryView"
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1335
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)             # <<<<<<<<<<<<<<
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 * 
 */

#line 1335 "View.MemoryView"
  copy_strided_to_strided((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1336
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 * 
 *     free(tmpdata)
 */

#line 1336 "View.MemoryView"
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1338
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 * 
 *     free(tmpdata)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */

#line 1338 "View.MemoryView"
  free(__pyx_v_tmpdata);

  /* "View.MemoryView":1339
 * 
 *     free(tmpdata)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 */

#line 1339 "View.MemoryView"
  __pyx_r = 0;

#line 1339 "View.MemoryView"
  goto __pyx_L0;

  /* "View.MemoryView":1270
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

#line 1270 "View.MemoryView"


#line 1270 "View.MemoryView"
  /* function exit code */

#line 1270 "View.MemoryView"
  __pyx_L1_error:;

#line 1270 "View.MemoryView"
  {

#line 1270 "View.MemoryView"
    #ifdef WITH_THREAD
    PyGILState_STATE 
#line 1270 "View.MemoryView"
__pyx_gilstate_save = __Pyx_PyGILState_Ensure();

#line 1270 "View.MemoryView"
    #endif

#line 1270 "View.MemoryView"
    __Pyx_AddTraceback("View.MemoryView.memoryview_copy_contents", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1270 "View.MemoryView"
    #ifdef WITH_THREAD

#line 1270 "View.MemoryView"
    __Pyx_PyGILState_Release(__pyx_gilstate_save);

#line 1270 "View.MemoryView"
    #endif

#line 1270 "View.MemoryView"
  }

#line 1270 "View.MemoryView"
  __pyx_r = -1;

#line 1270 "View.MemoryView"
  __pyx_L0:;

#line 1270 "View.MemoryView"
  return __pyx_r;

#line 1270 "View.MemoryView"
}

/* "View.MemoryView":1342
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) nogil:
 */

#line 1342 "View.MemoryView"


#line 1342 "View.MemoryView"
static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim, int __pyx_v_ndim_other) {
  int __pyx_v_i
#line 1342 "View.MemoryView"
;
  int __pyx_v_offset
#line 1342 "View.MemoryView"
;

#line 1342 "View.MemoryView"
  int __pyx_t_1;

#line 1342 "View.MemoryView"
  int __pyx_t_2;

#line 1342 "View.MemoryView"
  int __pyx_t_3;

  /* "View.MemoryView":1346
 *                             int ndim_other) nogil:
 *     cdef int i
 *     cdef int offset = ndim_other - ndim             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */

#line 1346 "View.MemoryView"
  __pyx_v_offset = (__pyx_v_ndim_other - __pyx_v_ndim);

  /* "View.MemoryView":1348
 *     cdef int offset = ndim_other - ndim
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 */

#line 1348 "View.MemoryView"
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {

#line 1348 "View.MemoryView"
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1349
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]             # <<<<<<<<<<<<<<
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 */

#line 1349 "View.MemoryView"
    (__pyx_v_mslice->shape[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->shape[__pyx_v_i]);

    /* "View.MemoryView":1350
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 */

#line 1350 "View.MemoryView"
    (__pyx_v_mslice->strides[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->strides[__pyx_v_i]);

    /* "View.MemoryView":1351
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]             # <<<<<<<<<<<<<<
 * 
 *     for i in range(offset):
 */

#line 1351 "View.MemoryView"
    (__pyx_v_mslice->suboffsets[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->suboffsets[__pyx_v_i]);

#line 1351 "View.MemoryView"
  }

  /* "View.MemoryView":1353
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 *     for i in range(offset):             # <<<<<<<<<<<<<<
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 */

#line 1353 "View.MemoryView"
  __pyx_t_1 = __pyx_v_offset;

#line 1353 "View.MemoryView"
  __pyx_t_2 = __pyx_t_1;

#line 1353 "View.MemoryView"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 1353 "View.MemoryView"
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1354
 * 
 *     for i in range(offset):
 *         mslice.shape[i] = 1             # <<<<<<<<<<<<<<
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1
 */

#line 1354 "View.MemoryView"
    (__pyx_v_mslice->shape[__pyx_v_i]) = 1;

    /* "View.MemoryView":1355
 *     for i in range(offset):
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i] = -1
 * 
 */

#line 1355 "View.MemoryView"
    (__pyx_v_mslice->strides[__pyx_v_i]) = (__pyx_v_mslice->strides[0]);

    /* "View.MemoryView":1356
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1356 "View.MemoryView"
    (__pyx_v_mslice->suboffsets[__pyx_v_i]) = -1L;

#line 1356 "View.MemoryView"
  }

  /* "View.MemoryView":1342
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) nogil:
 */

#line 1342 "View.MemoryView"


#line 1342 "View.MemoryView"
  /* function exit code */

#line 1342 "View.MemoryView"
}

/* "View.MemoryView":1364
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object,             # <<<<<<<<<<<<<<
 *                            int ndim, bint inc) nogil:
 * 
 */

#line 1364 "View.MemoryView"


#line 1364 "View.MemoryView"
static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_dtype_is_object, int __pyx_v_ndim, int __pyx_v_inc) {

#line 1364 "View.MemoryView"
  int __pyx_t_1;

  /* "View.MemoryView":1368
 * 
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,
 *                                            dst.strides, ndim, inc)
 */

#line 1368 "View.MemoryView"
  __pyx_t_1 = (__pyx_v_dtype_is_object != 0);

#line 1368 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":1369
 * 
 *     if dtype_is_object:
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,             # <<<<<<<<<<<<<<
 *                                            dst.strides, ndim, inc)
 * 
 */

#line 1369 "View.MemoryView"
    __pyx_memoryview_refcount_objects_in_slice_with_gil(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_inc);

    /* "View.MemoryView":1368
 * 
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,
 *                                            dst.strides, ndim, inc)
 */

#line 1368 "View.MemoryView"
  }

  /* "View.MemoryView":1364
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object,             # <<<<<<<<<<<<<<
 *                            int ndim, bint inc) nogil:
 * 
 */

#line 1364 "View.MemoryView"


#line 1364 "View.MemoryView"
  /* function exit code */

#line 1364 "View.MemoryView"
}

/* "View.MemoryView":1373
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 */

#line 1373 "View.MemoryView"


#line 1373 "View.MemoryView"
static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {

#line 1373 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1373 "View.MemoryView"
  #ifdef WITH_THREAD
  PyGILState_STATE 
#line 1373 "View.MemoryView"
__pyx_gilstate_save = __Pyx_PyGILState_Ensure();

#line 1373 "View.MemoryView"
  #endif

#line 1373 "View.MemoryView"
  __Pyx_RefNannySetupContext("refcount_objects_in_slice_with_gil", 0);

  /* "View.MemoryView":1376
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 *     refcount_objects_in_slice(data, shape, strides, ndim, inc)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 */

#line 1376 "View.MemoryView"
  __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, __pyx_v_shape, __pyx_v_strides, __pyx_v_ndim, __pyx_v_inc);

  /* "View.MemoryView":1373
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 */

#line 1373 "View.MemoryView"


#line 1373 "View.MemoryView"
  /* function exit code */

#line 1373 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1373 "View.MemoryView"
  #ifdef WITH_THREAD

#line 1373 "View.MemoryView"
  __Pyx_PyGILState_Release(__pyx_gilstate_save);

#line 1373 "View.MemoryView"
  #endif

#line 1373 "View.MemoryView"
}

/* "View.MemoryView":1379
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc):
 *     cdef Py_ssize_t i
 */

#line 1379 "View.MemoryView"


#line 1379 "View.MemoryView"
static void __pyx_memoryview_refcount_objects_in_slice(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i
#line 1379 "View.MemoryView"
;

#line 1379 "View.MemoryView"
  __Pyx_RefNannyDeclarations

#line 1379 "View.MemoryView"
  Py_ssize_t __pyx_t_1;

#line 1379 "View.MemoryView"
  Py_ssize_t __pyx_t_2;

#line 1379 "View.MemoryView"
  Py_ssize_t __pyx_t_3;

#line 1379 "View.MemoryView"
  int __pyx_t_4;

#line 1379 "View.MemoryView"
  __Pyx_RefNannySetupContext("refcount_objects_in_slice", 0);

  /* "View.MemoryView":1383
 *     cdef Py_ssize_t i
 * 
 *     for i in range(shape[0]):             # <<<<<<<<<<<<<<
 *         if ndim == 1:
 *             if inc:
 */

#line 1383 "View.MemoryView"
  __pyx_t_1 = (__pyx_v_shape[0]);

#line 1383 "View.MemoryView"
  __pyx_t_2 = __pyx_t_1;

#line 1383 "View.MemoryView"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 1383 "View.MemoryView"
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1384
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */

#line 1384 "View.MemoryView"
    __pyx_t_4 = ((__pyx_v_ndim == 1) != 0);

#line 1384 "View.MemoryView"
    if (__pyx_t_4) {

      /* "View.MemoryView":1385
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */

#line 1385 "View.MemoryView"
      __pyx_t_4 = (__pyx_v_inc != 0);

#line 1385 "View.MemoryView"
      if (__pyx_t_4) {

        /* "View.MemoryView":1386
 *         if ndim == 1:
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])
 */

#line 1386 "View.MemoryView"
        Py_INCREF((((PyObject **)__pyx_v_data)[0]));

        /* "View.MemoryView":1385
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */

#line 1385 "View.MemoryView"
        goto __pyx_L6;

#line 1385 "View.MemoryView"
      }

      /* "View.MemoryView":1388
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,
 */

#line 1388 "View.MemoryView"
      /*else*/ {

#line 1388 "View.MemoryView"
        Py_DECREF((((PyObject **)__pyx_v_data)[0]));

#line 1388 "View.MemoryView"
      }

#line 1388 "View.MemoryView"
      __pyx_L6:;

      /* "View.MemoryView":1384
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */

#line 1384 "View.MemoryView"
      goto __pyx_L5;

#line 1384 "View.MemoryView"
    }

    /* "View.MemoryView":1390
 *                 Py_DECREF((<PyObject **> data)[0])
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,             # <<<<<<<<<<<<<<
 *                                       ndim - 1, inc)
 * 
 */

#line 1390 "View.MemoryView"
    /*else*/ {

      /* "View.MemoryView":1391
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,
 *                                       ndim - 1, inc)             # <<<<<<<<<<<<<<
 * 
 *         data += strides[0]
 */

#line 1391 "View.MemoryView"
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_inc);

#line 1391 "View.MemoryView"
    }

#line 1391 "View.MemoryView"
    __pyx_L5:;

    /* "View.MemoryView":1393
 *                                       ndim - 1, inc)
 * 
 *         data += strides[0]             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1393 "View.MemoryView"
    __pyx_v_data = (__pyx_v_data + (__pyx_v_strides[0]));

#line 1393 "View.MemoryView"
  }

  /* "View.MemoryView":1379
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc):
 *     cdef Py_ssize_t i
 */

#line 1379 "View.MemoryView"


#line 1379 "View.MemoryView"
  /* function exit code */

#line 1379 "View.MemoryView"
  __Pyx_RefNannyFinishContext();

#line 1379 "View.MemoryView"
}

/* "View.MemoryView":1399
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 */

#line 1399 "View.MemoryView"


#line 1399 "View.MemoryView"
static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item, int __pyx_v_dtype_is_object) {

  /* "View.MemoryView":1402
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,
 *                          itemsize, item)
 */

#line 1402 "View.MemoryView"
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1403
 *                               bint dtype_is_object) nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, False)
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,             # <<<<<<<<<<<<<<
 *                          itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, True)
 */

#line 1403 "View.MemoryView"
  __pyx_memoryview__slice_assign_scalar(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_itemsize, __pyx_v_item);

  /* "View.MemoryView":1405
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,
 *                          itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1405 "View.MemoryView"
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1399
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 */

#line 1399 "View.MemoryView"


#line 1399 "View.MemoryView"
  /* function exit code */

#line 1399 "View.MemoryView"
}

/* "View.MemoryView":1409
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) nogil:
 */

#line 1409 "View.MemoryView"


#line 1409 "View.MemoryView"
static void __pyx_memoryview__slice_assign_scalar(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i
#line 1409 "View.MemoryView"
;
  Py_ssize_t __pyx_v_stride
#line 1409 "View.MemoryView"
;
  Py_ssize_t __pyx_v_extent
#line 1409 "View.MemoryView"
;

#line 1409 "View.MemoryView"
  int __pyx_t_1;

#line 1409 "View.MemoryView"
  Py_ssize_t __pyx_t_2;

#line 1409 "View.MemoryView"
  Py_ssize_t __pyx_t_3;

#line 1409 "View.MemoryView"
  Py_ssize_t __pyx_t_4;

  /* "View.MemoryView":1413
 *                               size_t itemsize, void *item) nogil:
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t extent = shape[0]
 * 
 */

#line 1413 "View.MemoryView"
  __pyx_v_stride = (__pyx_v_strides[0]);

  /* "View.MemoryView":1414
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]
 *     cdef Py_ssize_t extent = shape[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */

#line 1414 "View.MemoryView"
  __pyx_v_extent = (__pyx_v_shape[0]);

  /* "View.MemoryView":1416
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */

#line 1416 "View.MemoryView"
  __pyx_t_1 = ((__pyx_v_ndim == 1) != 0);

#line 1416 "View.MemoryView"
  if (__pyx_t_1) {

    /* "View.MemoryView":1417
 * 
 *     if ndim == 1:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             memcpy(data, item, itemsize)
 *             data += stride
 */

#line 1417 "View.MemoryView"
    __pyx_t_2 = __pyx_v_extent;

#line 1417 "View.MemoryView"
    __pyx_t_3 = __pyx_t_2;

#line 1417 "View.MemoryView"
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 1417 "View.MemoryView"
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1418
 *     if ndim == 1:
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)             # <<<<<<<<<<<<<<
 *             data += stride
 *     else:
 */

#line 1418 "View.MemoryView"
      (void)(memcpy(__pyx_v_data, __pyx_v_item, __pyx_v_itemsize));

      /* "View.MemoryView":1419
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 *             data += stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(extent):
 */

#line 1419 "View.MemoryView"
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);

#line 1419 "View.MemoryView"
    }

    /* "View.MemoryView":1416
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */

#line 1416 "View.MemoryView"
    goto __pyx_L3;

#line 1416 "View.MemoryView"
  }

  /* "View.MemoryView":1421
 *             data += stride
 *     else:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             _slice_assign_scalar(data, shape + 1, strides + 1,
 *                                 ndim - 1, itemsize, item)
 */

#line 1421 "View.MemoryView"
  /*else*/ {

#line 1421 "View.MemoryView"
    __pyx_t_2 = __pyx_v_extent;

#line 1421 "View.MemoryView"
    __pyx_t_3 = __pyx_t_2;

#line 1421 "View.MemoryView"
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 1421 "View.MemoryView"
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1422
 *     else:
 *         for i in range(extent):
 *             _slice_assign_scalar(data, shape + 1, strides + 1,             # <<<<<<<<<<<<<<
 *                                 ndim - 1, itemsize, item)
 *             data += stride
 */

#line 1422 "View.MemoryView"
      __pyx_memoryview__slice_assign_scalar(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize, __pyx_v_item);

      /* "View.MemoryView":1424
 *             _slice_assign_scalar(data, shape + 1, strides + 1,
 *                                 ndim - 1, itemsize, item)
 *             data += stride             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1424 "View.MemoryView"
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);

#line 1424 "View.MemoryView"
    }

#line 1424 "View.MemoryView"
  }

#line 1424 "View.MemoryView"
  __pyx_L3:;

  /* "View.MemoryView":1409
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) nogil:
 */

#line 1409 "View.MemoryView"


#line 1409 "View.MemoryView"
  /* function exit code */

#line 1409 "View.MemoryView"
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum = 
#line 1 "(tree fragment)"
{"__pyx_unpickle_Enum", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum, METH_VARARGS|METH_KEYWORDS, 0};

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0
#line 1 "(tree fragment)"
;
  long __pyx_v___pyx_checksum
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_state = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum (wrapper)", 0);

#line 1 "(tree fragment)"
  {

#line 1 "(tree fragment)"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};

#line 1 "(tree fragment)"
    PyObject* values[3] = {0,0,0};

#line 1 "(tree fragment)"
    if (unlikely(__pyx_kwds)) {

#line 1 "(tree fragment)"
      Py_ssize_t kw_args;

#line 1 "(tree fragment)"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 1 "(tree fragment)"
      switch (pos_args) {
        case  3: 
#line 1 "(tree fragment)"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 1 "(tree fragment)"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 1 "(tree fragment)"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  0: break;
        default: 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      kw_args = PyDict_Size(__pyx_kwds);

#line 1 "(tree fragment)"
      switch (pos_args) {

#line 1 "(tree fragment)"
        case  0:

#line 1 "(tree fragment)"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  1:

#line 1 "(tree fragment)"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 1); 
#line 1 "(tree fragment)"
__PYX_ERR(3, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  2:

#line 1 "(tree fragment)"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 2); 
#line 1 "(tree fragment)"
__PYX_ERR(3, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      if (unlikely(kw_args > 0)) {

#line 1 "(tree fragment)"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Enum") < 0)) __PYX_ERR(3, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 1 "(tree fragment)"
      goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
    } else {

#line 1 "(tree fragment)"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
    }

#line 1 "(tree fragment)"
    __pyx_v___pyx_type = values[0];

#line 1 "(tree fragment)"
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(3, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
    __pyx_v___pyx_state = values[2];

#line 1 "(tree fragment)"
  }

#line 1 "(tree fragment)"
  goto __pyx_L4_argument_unpacking_done;

#line 1 "(tree fragment)"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 1 "(tree fragment)"
__PYX_ERR(3, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
  __pyx_L3_error:;

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return NULL;

#line 1 "(tree fragment)"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_result = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_2;

#line 1 "(tree fragment)"
  int __pyx_t_3;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xb068931, 0x82a3537, 0x6ae9995):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__24, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(3, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 4 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xb068931, 0x82a3537, 0x6ae9995):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 */

#line 5 "(tree fragment)"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __pyx_v___pyx_PickleError = __pyx_t_1;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xb068931, 0x82a3537, 0x6ae9995):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */

#line 6 "(tree fragment)"
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 6 "(tree fragment)"
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; 
#line 6 "(tree fragment)"
__pyx_t_6 = NULL;

#line 6 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {

#line 6 "(tree fragment)"
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);

#line 6 "(tree fragment)"
      if (likely(__pyx_t_6)) {

#line 6 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 6 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_6);

#line 6 "(tree fragment)"
        __Pyx_INCREF(function);

#line 6 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_1, function);

#line 6 "(tree fragment)"
      }

#line 6 "(tree fragment)"
    }

#line 6 "(tree fragment)"
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 6 "(tree fragment)"
    if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 6 "(tree fragment)"
    __PYX_ERR(3, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xb068931, 0x82a3537, 0x6ae9995):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 */

#line 7 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_MemviewEnum_type), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 7 "(tree fragment)"
  __pyx_t_5 = NULL;

#line 7 "(tree fragment)"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {

#line 7 "(tree fragment)"
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);

#line 7 "(tree fragment)"
    if (likely(__pyx_t_5)) {

#line 7 "(tree fragment)"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 7 "(tree fragment)"
      __Pyx_INCREF(__pyx_t_5);

#line 7 "(tree fragment)"
      __Pyx_INCREF(function);

#line 7 "(tree fragment)"
      __Pyx_DECREF_SET(__pyx_t_1, function);

#line 7 "(tree fragment)"
    }

#line 7 "(tree fragment)"
  }

#line 7 "(tree fragment)"
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);

#line 7 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 7 "(tree fragment)"
  if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 7 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 7 "(tree fragment)"
  __pyx_v___pyx_result = __pyx_t_4;

#line 7 "(tree fragment)"
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);

#line 8 "(tree fragment)"
  __pyx_t_2 = (__pyx_t_3 != 0);

#line 8 "(tree fragment)"
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 */

#line 9 "(tree fragment)"
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(3, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __pyx_t_4 = __pyx_unpickle_Enum__set_state(((struct __pyx_MemviewEnum_obj *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 9 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 */

#line 10 "(tree fragment)"
  __Pyx_XDECREF(__pyx_r);

#line 10 "(tree fragment)"
  __Pyx_INCREF(__pyx_v___pyx_result);

#line 10 "(tree fragment)"
  __pyx_r = __pyx_v___pyx_result;

#line 10 "(tree fragment)"
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_result);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {

#line 11 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 11 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 11 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 11 "(tree fragment)"
  int __pyx_t_2;

#line 11 "(tree fragment)"
  Py_ssize_t __pyx_t_3;

#line 11 "(tree fragment)"
  int __pyx_t_4;

#line 11 "(tree fragment)"
  int __pyx_t_5;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_7 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_8 = NULL;

#line 11 "(tree fragment)"
  int __pyx_lineno = 0;

#line 11 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 11 "(tree fragment)"
  int __pyx_clineno = 0;

#line 11 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(3, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->name);

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_v___pyx_result->name);

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->name = __pyx_t_1;

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */

#line 13 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 13 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 13 "(tree fragment)"
    __PYX_ERR(3, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(3, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_4 = ((__pyx_t_3 > 1) != 0);

#line 13 "(tree fragment)"
  if (__pyx_t_4) {

#line 13 "(tree fragment)"
  } else {

#line 13 "(tree fragment)"
    __pyx_t_2 = __pyx_t_4;

#line 13 "(tree fragment)"
    goto __pyx_L4_bool_binop_done;

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(3, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_5 = (__pyx_t_4 != 0);

#line 13 "(tree fragment)"
  __pyx_t_2 = __pyx_t_5;

#line 13 "(tree fragment)"
  __pyx_L4_bool_binop_done:;

#line 13 "(tree fragment)"
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */

#line 14 "(tree fragment)"
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_6);

#line 14 "(tree fragment)"
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_7);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 14 "(tree fragment)"
    if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 14 "(tree fragment)"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 14 "(tree fragment)"
      __PYX_ERR(3, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_6);

#line 14 "(tree fragment)"
    __pyx_t_8 = NULL;

#line 14 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {

#line 14 "(tree fragment)"
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);

#line 14 "(tree fragment)"
      if (likely(__pyx_t_8)) {

#line 14 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);

#line 14 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_8);

#line 14 "(tree fragment)"
        __Pyx_INCREF(function);

#line 14 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_7, function);

#line 14 "(tree fragment)"
      }

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);

#line 14 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 14 "(tree fragment)"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */

#line 13 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
  /* function exit code */

#line 11 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 11 "(tree fragment)"
  goto __pyx_L0;

#line 11 "(tree fragment)"
  __pyx_L1_error:;

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_7);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_8);

#line 11 "(tree fragment)"
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 11 "(tree fragment)"
  __pyx_r = 0;

#line 11 "(tree fragment)"
  __pyx_L0:;

#line 11 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 11 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 11 "(tree fragment)"
  return __pyx_r;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord __pyx_vtable_12ms_deisotope_2_c_7scoring_IsotopicFitRecord;

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_freelist_12ms_deisotope_2_c_7scoring_IsotopicFitRecord[50000];

#line 11 "(tree fragment)"
static int __pyx_freecount_12ms_deisotope_2_c_7scoring_IsotopicFitRecord = 0;

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_IsotopicFitRecord(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *p;

#line 11 "(tree fragment)"
  PyObject *o;

#line 11 "(tree fragment)"
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_12ms_deisotope_2_c_7scoring_IsotopicFitRecord > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord)) & ((t->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)) == 0))) {

#line 11 "(tree fragment)"
    o = (PyObject*)__pyx_freelist_12ms_deisotope_2_c_7scoring_IsotopicFitRecord[--__pyx_freecount_12ms_deisotope_2_c_7scoring_IsotopicFitRecord];

#line 11 "(tree fragment)"
    memset(o, 0, sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord));

#line 11 "(tree fragment)"
    (void) PyObject_INIT(o, t);

#line 11 "(tree fragment)"
    PyObject_GC_Track(o);

#line 11 "(tree fragment)"
  } else {

#line 11 "(tree fragment)"
    if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {

#line 11 "(tree fragment)"
      o = (*t->tp_alloc)(t, 0);

#line 11 "(tree fragment)"
    } else {

#line 11 "(tree fragment)"
      o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);

#line 11 "(tree fragment)"
    }

#line 11 "(tree fragment)"
    if (unlikely(!o)) return 0;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)o);

#line 11 "(tree fragment)"
  p->__pyx_vtab = __pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitRecord;

#line 11 "(tree fragment)"
  p->seed_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)Py_None); Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  p->experimental = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  p->theoretical = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)Py_None); Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  p->monoisotopic_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)Py_None); Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  p->data = Py_None; Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  return o;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_7scoring_IsotopicFitRecord(PyObject *o) {

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *p = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)o;

#line 11 "(tree fragment)"
  #if CYTHON_USE_TP_FINALIZE

#line 11 "(tree fragment)"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {

#line 11 "(tree fragment)"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  PyObject_GC_UnTrack(o);

#line 11 "(tree fragment)"
  Py_CLEAR(p->seed_peak);

#line 11 "(tree fragment)"
  Py_CLEAR(p->experimental);

#line 11 "(tree fragment)"
  Py_CLEAR(p->theoretical);

#line 11 "(tree fragment)"
  Py_CLEAR(p->monoisotopic_peak);

#line 11 "(tree fragment)"
  Py_CLEAR(p->data);

#line 11 "(tree fragment)"
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_12ms_deisotope_2_c_7scoring_IsotopicFitRecord < 50000) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord)) & ((Py_TYPE(o)->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)) == 0))) {

#line 11 "(tree fragment)"
    __pyx_freelist_12ms_deisotope_2_c_7scoring_IsotopicFitRecord[__pyx_freecount_12ms_deisotope_2_c_7scoring_IsotopicFitRecord++] = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)o);

#line 11 "(tree fragment)"
  } else {

#line 11 "(tree fragment)"
    (*Py_TYPE(o)->tp_free)(o);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_traverse_12ms_deisotope_2_c_7scoring_IsotopicFitRecord(PyObject *o, visitproc v, void *a) {

#line 11 "(tree fragment)"
  int e;

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *p = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)o;

#line 11 "(tree fragment)"
  if (p->seed_peak) {

#line 11 "(tree fragment)"
    e = (*v)(((PyObject *)p->seed_peak), a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  if (p->experimental) {

#line 11 "(tree fragment)"
    e = (*v)(p->experimental, a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  if (p->theoretical) {

#line 11 "(tree fragment)"
    e = (*v)(((PyObject *)p->theoretical), a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  if (p->monoisotopic_peak) {

#line 11 "(tree fragment)"
    e = (*v)(((PyObject *)p->monoisotopic_peak), a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  if (p->data) {

#line 11 "(tree fragment)"
    e = (*v)(p->data, a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_clear_12ms_deisotope_2_c_7scoring_IsotopicFitRecord(PyObject *o) {

#line 11 "(tree fragment)"
  PyObject* tmp;

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *p = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *)o;

#line 11 "(tree fragment)"
  tmp = ((PyObject*)p->seed_peak);

#line 11 "(tree fragment)"
  p->seed_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)Py_None); Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  Py_XDECREF(tmp);

#line 11 "(tree fragment)"
  tmp = ((PyObject*)p->experimental);

#line 11 "(tree fragment)"
  p->experimental = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  Py_XDECREF(tmp);

#line 11 "(tree fragment)"
  tmp = ((PyObject*)p->theoretical);

#line 11 "(tree fragment)"
  p->theoretical = ((struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *)Py_None); Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  Py_XDECREF(tmp);

#line 11 "(tree fragment)"
  tmp = ((PyObject*)p->monoisotopic_peak);

#line 11 "(tree fragment)"
  p->monoisotopic_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)Py_None); Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  Py_XDECREF(tmp);

#line 11 "(tree fragment)"
  tmp = ((PyObject*)p->data);

#line 11 "(tree fragment)"
  p->data = Py_None; Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  Py_XDECREF(tmp);

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_npeaks(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_6npeaks_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_seed_peak(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_9seed_peak_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_seed_peak(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_9seed_peak_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_9seed_peak_5__del__(o);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_score(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_5score_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_score(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_5score_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 11 "(tree fragment)"
    return -1;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_charge(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_6charge_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_charge(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_6charge_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 11 "(tree fragment)"
    return -1;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_experimental(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12experimental_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_experimental(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12experimental_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12experimental_5__del__(o);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_theoretical(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_11theoretical_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_theoretical(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_11theoretical_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_11theoretical_5__del__(o);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_monoisotopic_peak(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_17monoisotopic_peak_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_monoisotopic_peak(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_17monoisotopic_peak_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_17monoisotopic_peak_5__del__(o);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_missed_peaks(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12missed_peaks_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_missed_peaks(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12missed_peaks_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 11 "(tree fragment)"
    return -1;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_data(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_4data_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_data(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_4data_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_4data_5__del__(o);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__hash(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_5_hash_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__hash(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_5_hash_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 11 "(tree fragment)"
    return -1;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_7scoring_IsotopicFitRecord[] = {

#line 11 "(tree fragment)"
  {"clone", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_3clone, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_2clone},

#line 11 "(tree fragment)"
  {"__reduce__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_5__reduce__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_4__reduce__},

#line 11 "(tree fragment)"
  {"_eq", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_7_eq, METH_O, __pyx_doc_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_6_eq},

#line 11 "(tree fragment)"
  {"_ne", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_9_ne, METH_O, __pyx_doc_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_8_ne},

#line 11 "(tree fragment)"
  {"_lt", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_11_lt, METH_O, __pyx_doc_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_10_lt},

#line 11 "(tree fragment)"
  {"_gt", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_13_gt, METH_O, __pyx_doc_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_12_gt},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_7scoring_IsotopicFitRecord[] = {

#line 11 "(tree fragment)"
  {(char *)"npeaks", __pyx_getprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_npeaks, 0, (char *)0, 0},

#line 11 "(tree fragment)"
  {(char *)"seed_peak", __pyx_getprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_seed_peak, __pyx_setprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_seed_peak, (char *)"seed_peak: ms_peak_picker._c.peak_set.FittedPeak", 0},

#line 11 "(tree fragment)"
  {(char *)"score", __pyx_getprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_score, __pyx_setprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_score, (char *)"score: 'double'", 0},

#line 11 "(tree fragment)"
  {(char *)"charge", __pyx_getprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_charge, __pyx_setprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_charge, (char *)"charge: 'int'", 0},

#line 11 "(tree fragment)"
  {(char *)"experimental", __pyx_getprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_experimental, __pyx_setprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_experimental, (char *)"experimental: list", 0},

#line 11 "(tree fragment)"
  {(char *)"theoretical", __pyx_getprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_theoretical, __pyx_setprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_theoretical, (char *)"theoretical: ms_deisotope._c.averagine.TheoreticalIsotopicPattern", 0},

#line 11 "(tree fragment)"
  {(char *)"monoisotopic_peak", __pyx_getprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_monoisotopic_peak, __pyx_setprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_monoisotopic_peak, (char *)"monoisotopic_peak: ms_peak_picker._c.peak_set.FittedPeak", 0},

#line 11 "(tree fragment)"
  {(char *)"missed_peaks", __pyx_getprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_missed_peaks, __pyx_setprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_missed_peaks, (char *)"missed_peaks: 'int'", 0},

#line 11 "(tree fragment)"
  {(char *)"data", __pyx_getprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_data, __pyx_setprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_data, (char *)"data: object", 0},

#line 11 "(tree fragment)"
  {(char *)"_hash", __pyx_getprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__hash, __pyx_setprop_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__hash, (char *)"_hash: 'Py_hash_t'", 0},

#line 11 "(tree fragment)"
  {0, 0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_7scoring_IsotopicFitRecord = {

#line 11 "(tree fragment)"
  PyVarObject_HEAD_INIT(0, 0)

#line 11 "(tree fragment)"
  "ms_deisotope._c.scoring.IsotopicFitRecord", /*tp_name*/

#line 11 "(tree fragment)"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord), /*tp_basicsize*/

#line 11 "(tree fragment)"
  0, /*tp_itemsize*/

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_12ms_deisotope_2_c_7scoring_IsotopicFitRecord, /*tp_dealloc*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX < 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_vectorcall_offset*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattr*/

#line 11 "(tree fragment)"
  0, /*tp_setattr*/

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*tp_compare*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  0, /*tp_as_async*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_22__repr__, /*tp_repr*/

#line 11 "(tree fragment)"
  0, /*tp_as_number*/

#line 11 "(tree fragment)"
  0, /*tp_as_sequence*/

#line 11 "(tree fragment)"
  0, /*tp_as_mapping*/

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_17__hash__, /*tp_hash*/

#line 11 "(tree fragment)"
  0, /*tp_call*/

#line 11 "(tree fragment)"
  0, /*tp_str*/

#line 11 "(tree fragment)"
  0, /*tp_getattro*/

#line 11 "(tree fragment)"
  0, /*tp_setattro*/

#line 11 "(tree fragment)"
  0, /*tp_as_buffer*/

#line 11 "(tree fragment)"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 11 "(tree fragment)"
  "IsotopicFitRecord(FittedPeak seed_peak, double score, int charge, TheoreticalIsotopicPattern theoretical, list experimental, data=None, int missed_peaks=0)\nDescribes a single isotopic pattern fit, comparing how well an\n    experimentally observed sequence of peaks matches a theoretical isotopic\n    pattern.\n\n    IsotopicFitRecord instances are hashable and orderable (by score).\n\n    Attributes\n    ----------\n    charge : int\n        The charge state used to generate the theoretical pattern\n    data : object\n        An arbitrary Python object containing extra information\n    experimental : list of FittedPeak\n        The observed experimental peaks to be fitted\n    missed_peaks : int\n        The number of peaks in the theoretical pattern that do not\n        have a matching experimental peak\n    monoisotopic_peak : FittedPeak\n        The fitted peak which corresponds to the monoisotopic peak\n    score : float\n        The score assigned to the fit by an IsotopicFitter object\n    seed_peak : FittedPeak\n        The peak that was used to initiate the fit. This may be unused\n        if not using an Averagine method\n    theoretical : TheoreticalIsotopicPattern\n        The theoretical isotopic pattern being fitted on the experimental data\n    ", /*tp_doc*/

#line 11 "(tree fragment)"
  __pyx_tp_traverse_12ms_deisotope_2_c_7scoring_IsotopicFitRecord, /*tp_traverse*/

#line 11 "(tree fragment)"
  __pyx_tp_clear_12ms_deisotope_2_c_7scoring_IsotopicFitRecord, /*tp_clear*/

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_15__richcmp__, /*tp_richcompare*/

#line 11 "(tree fragment)"
  0, /*tp_weaklistoffset*/

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_19__iter__, /*tp_iter*/

#line 11 "(tree fragment)"
  0, /*tp_iternext*/

#line 11 "(tree fragment)"
  __pyx_methods_12ms_deisotope_2_c_7scoring_IsotopicFitRecord, /*tp_methods*/

#line 11 "(tree fragment)"
  0, /*tp_members*/

#line 11 "(tree fragment)"
  __pyx_getsets_12ms_deisotope_2_c_7scoring_IsotopicFitRecord, /*tp_getset*/

#line 11 "(tree fragment)"
  0, /*tp_base*/

#line 11 "(tree fragment)"
  0, /*tp_dict*/

#line 11 "(tree fragment)"
  0, /*tp_descr_get*/

#line 11 "(tree fragment)"
  0, /*tp_descr_set*/

#line 11 "(tree fragment)"
  0, /*tp_dictoffset*/

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord_1__init__, /*tp_init*/

#line 11 "(tree fragment)"
  0, /*tp_alloc*/

#line 11 "(tree fragment)"
  __pyx_tp_new_12ms_deisotope_2_c_7scoring_IsotopicFitRecord, /*tp_new*/

#line 11 "(tree fragment)"
  0, /*tp_free*/

#line 11 "(tree fragment)"
  0, /*tp_is_gc*/

#line 11 "(tree fragment)"
  0, /*tp_bases*/

#line 11 "(tree fragment)"
  0, /*tp_mro*/

#line 11 "(tree fragment)"
  0, /*tp_cache*/

#line 11 "(tree fragment)"
  0, /*tp_subclasses*/

#line 11 "(tree fragment)"
  0, /*tp_weaklist*/

#line 11 "(tree fragment)"
  0, /*tp_del*/

#line 11 "(tree fragment)"
  0, /*tp_version_tag*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030400a1

#line 11 "(tree fragment)"
  0, /*tp_finalize*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 11 "(tree fragment)"
  0, /*tp_vectorcall*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_pypy_flags*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase __pyx_vtable_12ms_deisotope_2_c_7scoring_FitSelectorBase;

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_FitSelectorBase(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *p;

#line 11 "(tree fragment)"
  PyObject *o;

#line 11 "(tree fragment)"
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {

#line 11 "(tree fragment)"
    o = (*t->tp_alloc)(t, 0);

#line 11 "(tree fragment)"
  } else {

#line 11 "(tree fragment)"
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  if (unlikely(!o)) return 0;

#line 11 "(tree fragment)"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)o);

#line 11 "(tree fragment)"
  p->__pyx_vtab = __pyx_vtabptr_12ms_deisotope_2_c_7scoring_FitSelectorBase;

#line 11 "(tree fragment)"
  return o;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_7scoring_FitSelectorBase(PyObject *o) {

#line 11 "(tree fragment)"
  #if CYTHON_USE_TP_FINALIZE

#line 11 "(tree fragment)"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {

#line 11 "(tree fragment)"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  (*Py_TYPE(o)->tp_free)(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_7scoring_15FitSelectorBase_minimum_score(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_13minimum_score_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_7scoring_15FitSelectorBase_minimum_score(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_13minimum_score_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 11 "(tree fragment)"
    return -1;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_7scoring_FitSelectorBase[] = {

#line 11 "(tree fragment)"
  {"__reduce__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_3__reduce__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_7scoring_15FitSelectorBase_2__reduce__},

#line 11 "(tree fragment)"
  {"best", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_5best, METH_O, __pyx_doc_12ms_deisotope_2_c_7scoring_15FitSelectorBase_4best},

#line 11 "(tree fragment)"
  {"reject", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_9reject, METH_O, __pyx_doc_12ms_deisotope_2_c_7scoring_15FitSelectorBase_8reject},

#line 11 "(tree fragment)"
  {"reject_score", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_11reject_score, METH_O, __pyx_doc_12ms_deisotope_2_c_7scoring_15FitSelectorBase_10reject_score},

#line 11 "(tree fragment)"
  {"is_maximizing", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_13is_maximizing, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_7scoring_15FitSelectorBase_12is_maximizing},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_7scoring_FitSelectorBase[] = {

#line 11 "(tree fragment)"
  {(char *)"minimum_score", __pyx_getprop_12ms_deisotope_2_c_7scoring_15FitSelectorBase_minimum_score, __pyx_setprop_12ms_deisotope_2_c_7scoring_15FitSelectorBase_minimum_score, (char *)"minimum_score: 'double'", 0},

#line 11 "(tree fragment)"
  {0, 0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_7scoring_FitSelectorBase = {

#line 11 "(tree fragment)"
  PyVarObject_HEAD_INIT(0, 0)

#line 11 "(tree fragment)"
  "ms_deisotope._c.scoring.FitSelectorBase", /*tp_name*/

#line 11 "(tree fragment)"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase), /*tp_basicsize*/

#line 11 "(tree fragment)"
  0, /*tp_itemsize*/

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_12ms_deisotope_2_c_7scoring_FitSelectorBase, /*tp_dealloc*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX < 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_vectorcall_offset*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattr*/

#line 11 "(tree fragment)"
  0, /*tp_setattr*/

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*tp_compare*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  0, /*tp_as_async*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_15__repr__, /*tp_repr*/

#line 11 "(tree fragment)"
  0, /*tp_as_number*/

#line 11 "(tree fragment)"
  0, /*tp_as_sequence*/

#line 11 "(tree fragment)"
  0, /*tp_as_mapping*/

#line 11 "(tree fragment)"
  0, /*tp_hash*/

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_7__call__, /*tp_call*/

#line 11 "(tree fragment)"
  0, /*tp_str*/

#line 11 "(tree fragment)"
  0, /*tp_getattro*/

#line 11 "(tree fragment)"
  0, /*tp_setattro*/

#line 11 "(tree fragment)"
  0, /*tp_as_buffer*/

#line 11 "(tree fragment)"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/

#line 11 "(tree fragment)"
  "FitSelectorBase(minimum_score=0)\nAn object that controls the filtering and\n    selection of IsotopicFitRecord\n\n    Attributes\n    ----------\n    minimum_score : int\n        The minimum score needed to be a candidate for selection. If the\n        FitSelector is `minimizing` it is the maximal score to be a candidate.\n    ", /*tp_doc*/

#line 11 "(tree fragment)"
  0, /*tp_traverse*/

#line 11 "(tree fragment)"
  0, /*tp_clear*/

#line 11 "(tree fragment)"
  0, /*tp_richcompare*/

#line 11 "(tree fragment)"
  0, /*tp_weaklistoffset*/

#line 11 "(tree fragment)"
  0, /*tp_iter*/

#line 11 "(tree fragment)"
  0, /*tp_iternext*/

#line 11 "(tree fragment)"
  __pyx_methods_12ms_deisotope_2_c_7scoring_FitSelectorBase, /*tp_methods*/

#line 11 "(tree fragment)"
  0, /*tp_members*/

#line 11 "(tree fragment)"
  __pyx_getsets_12ms_deisotope_2_c_7scoring_FitSelectorBase, /*tp_getset*/

#line 11 "(tree fragment)"
  0, /*tp_base*/

#line 11 "(tree fragment)"
  0, /*tp_dict*/

#line 11 "(tree fragment)"
  0, /*tp_descr_get*/

#line 11 "(tree fragment)"
  0, /*tp_descr_set*/

#line 11 "(tree fragment)"
  0, /*tp_dictoffset*/

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_1__init__, /*tp_init*/

#line 11 "(tree fragment)"
  0, /*tp_alloc*/

#line 11 "(tree fragment)"
  __pyx_tp_new_12ms_deisotope_2_c_7scoring_FitSelectorBase, /*tp_new*/

#line 11 "(tree fragment)"
  0, /*tp_free*/

#line 11 "(tree fragment)"
  0, /*tp_is_gc*/

#line 11 "(tree fragment)"
  0, /*tp_bases*/

#line 11 "(tree fragment)"
  0, /*tp_mro*/

#line 11 "(tree fragment)"
  0, /*tp_cache*/

#line 11 "(tree fragment)"
  0, /*tp_subclasses*/

#line 11 "(tree fragment)"
  0, /*tp_weaklist*/

#line 11 "(tree fragment)"
  0, /*tp_del*/

#line 11 "(tree fragment)"
  0, /*tp_version_tag*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030400a1

#line 11 "(tree fragment)"
  0, /*tp_finalize*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 11 "(tree fragment)"
  0, /*tp_vectorcall*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_pypy_flags*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_vtable_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_IsotopicFitterBase(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *p;

#line 11 "(tree fragment)"
  PyObject *o;

#line 11 "(tree fragment)"
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {

#line 11 "(tree fragment)"
    o = (*t->tp_alloc)(t, 0);

#line 11 "(tree fragment)"
  } else {

#line 11 "(tree fragment)"
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  if (unlikely(!o)) return 0;

#line 11 "(tree fragment)"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)o);

#line 11 "(tree fragment)"
  p->__pyx_vtab = __pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;

#line 11 "(tree fragment)"
  p->select = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)Py_None); Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  return o;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_7scoring_IsotopicFitterBase(PyObject *o) {

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *p = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)o;

#line 11 "(tree fragment)"
  #if CYTHON_USE_TP_FINALIZE

#line 11 "(tree fragment)"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {

#line 11 "(tree fragment)"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  PyObject_GC_UnTrack(o);

#line 11 "(tree fragment)"
  Py_CLEAR(p->select);

#line 11 "(tree fragment)"
  (*Py_TYPE(o)->tp_free)(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_traverse_12ms_deisotope_2_c_7scoring_IsotopicFitterBase(PyObject *o, visitproc v, void *a) {

#line 11 "(tree fragment)"
  int e;

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *p = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)o;

#line 11 "(tree fragment)"
  if (p->select) {

#line 11 "(tree fragment)"
    e = (*v)(((PyObject *)p->select), a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_clear_12ms_deisotope_2_c_7scoring_IsotopicFitterBase(PyObject *o) {

#line 11 "(tree fragment)"
  PyObject* tmp;

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *p = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *)o;

#line 11 "(tree fragment)"
  tmp = ((PyObject*)p->select);

#line 11 "(tree fragment)"
  p->select = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *)Py_None); Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  Py_XDECREF(tmp);

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_select(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_6select_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_select(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_6select_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_6select_5__del__(o);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_7scoring_IsotopicFitterBase[] = {

#line 11 "(tree fragment)"
  {"__reduce__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_3__reduce__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_2__reduce__},

#line 11 "(tree fragment)"
  {"__getstate__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_5__getstate__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_4__getstate__},

#line 11 "(tree fragment)"
  {"__setstate__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_7__setstate__, METH_O, __pyx_doc_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_6__setstate__},

#line 11 "(tree fragment)"
  {"evaluate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_9evaluate, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_8evaluate},

#line 11 "(tree fragment)"
  {"_evaluate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_11_evaluate, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_10_evaluate},

#line 11 "(tree fragment)"
  {"reject", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_15reject, METH_O, __pyx_doc_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_14reject},

#line 11 "(tree fragment)"
  {"reject_score", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_17reject_score, METH_O, __pyx_doc_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_16reject_score},

#line 11 "(tree fragment)"
  {"is_maximizing", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_19is_maximizing, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_18is_maximizing},

#line 11 "(tree fragment)"
  {"_configure", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_21_configure, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_20_configure},

#line 11 "(tree fragment)"
  {"configure", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_23configure, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_22configure},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_7scoring_IsotopicFitterBase[] = {

#line 11 "(tree fragment)"
  {(char *)"select", __pyx_getprop_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_select, __pyx_setprop_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_select, (char *)"select: ms_deisotope._c.scoring.FitSelectorBase", 0},

#line 11 "(tree fragment)"
  {0, 0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_7scoring_IsotopicFitterBase = {

#line 11 "(tree fragment)"
  PyVarObject_HEAD_INIT(0, 0)

#line 11 "(tree fragment)"
  "ms_deisotope._c.scoring.IsotopicFitterBase", /*tp_name*/

#line 11 "(tree fragment)"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase), /*tp_basicsize*/

#line 11 "(tree fragment)"
  0, /*tp_itemsize*/

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_12ms_deisotope_2_c_7scoring_IsotopicFitterBase, /*tp_dealloc*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX < 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_vectorcall_offset*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattr*/

#line 11 "(tree fragment)"
  0, /*tp_setattr*/

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*tp_compare*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  0, /*tp_as_async*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_25__repr__, /*tp_repr*/

#line 11 "(tree fragment)"
  0, /*tp_as_number*/

#line 11 "(tree fragment)"
  0, /*tp_as_sequence*/

#line 11 "(tree fragment)"
  0, /*tp_as_mapping*/

#line 11 "(tree fragment)"
  0, /*tp_hash*/

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_13__call__, /*tp_call*/

#line 11 "(tree fragment)"
  0, /*tp_str*/

#line 11 "(tree fragment)"
  0, /*tp_getattro*/

#line 11 "(tree fragment)"
  0, /*tp_setattro*/

#line 11 "(tree fragment)"
  0, /*tp_as_buffer*/

#line 11 "(tree fragment)"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 11 "(tree fragment)"
  "IsotopicFitterBase(score_threshold=0.5)\nA base class for Isotopic Pattern Fitters, objects\n    which given a set of experimental peaks and a set of matching\n    theoretical peaks, returns a fit score describing how good the\n    match is.\n\n    An IsotopicFitter may be optimal when the score is small (minimizing)\n    or when the score is large (maximizing), and the appropriate\n    :class:`FitSelectorBase` type will be used for :attr:`select`. This\n    will also be reflected by :meth:`is_maximizing`.\n    ", /*tp_doc*/

#line 11 "(tree fragment)"
  __pyx_tp_traverse_12ms_deisotope_2_c_7scoring_IsotopicFitterBase, /*tp_traverse*/

#line 11 "(tree fragment)"
  __pyx_tp_clear_12ms_deisotope_2_c_7scoring_IsotopicFitterBase, /*tp_clear*/

#line 11 "(tree fragment)"
  0, /*tp_richcompare*/

#line 11 "(tree fragment)"
  0, /*tp_weaklistoffset*/

#line 11 "(tree fragment)"
  0, /*tp_iter*/

#line 11 "(tree fragment)"
  0, /*tp_iternext*/

#line 11 "(tree fragment)"
  __pyx_methods_12ms_deisotope_2_c_7scoring_IsotopicFitterBase, /*tp_methods*/

#line 11 "(tree fragment)"
  0, /*tp_members*/

#line 11 "(tree fragment)"
  __pyx_getsets_12ms_deisotope_2_c_7scoring_IsotopicFitterBase, /*tp_getset*/

#line 11 "(tree fragment)"
  0, /*tp_base*/

#line 11 "(tree fragment)"
  0, /*tp_dict*/

#line 11 "(tree fragment)"
  0, /*tp_descr_get*/

#line 11 "(tree fragment)"
  0, /*tp_descr_set*/

#line 11 "(tree fragment)"
  0, /*tp_dictoffset*/

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_1__init__, /*tp_init*/

#line 11 "(tree fragment)"
  0, /*tp_alloc*/

#line 11 "(tree fragment)"
  __pyx_tp_new_12ms_deisotope_2_c_7scoring_IsotopicFitterBase, /*tp_new*/

#line 11 "(tree fragment)"
  0, /*tp_free*/

#line 11 "(tree fragment)"
  0, /*tp_is_gc*/

#line 11 "(tree fragment)"
  0, /*tp_bases*/

#line 11 "(tree fragment)"
  0, /*tp_mro*/

#line 11 "(tree fragment)"
  0, /*tp_cache*/

#line 11 "(tree fragment)"
  0, /*tp_subclasses*/

#line 11 "(tree fragment)"
  0, /*tp_weaklist*/

#line 11 "(tree fragment)"
  0, /*tp_del*/

#line 11 "(tree fragment)"
  0, /*tp_version_tag*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030400a1

#line 11 "(tree fragment)"
  0, /*tp_finalize*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 11 "(tree fragment)"
  0, /*tp_vectorcall*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_pypy_flags*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_LeastSquaresFitter __pyx_vtable_12ms_deisotope_2_c_7scoring_LeastSquaresFitter;

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_LeastSquaresFitter(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter *p;

#line 11 "(tree fragment)"
  PyObject *o = __pyx_tp_new_12ms_deisotope_2_c_7scoring_IsotopicFitterBase(t, a, k);

#line 11 "(tree fragment)"
  if (unlikely(!o)) return 0;

#line 11 "(tree fragment)"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter *)o);

#line 11 "(tree fragment)"
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase*)__pyx_vtabptr_12ms_deisotope_2_c_7scoring_LeastSquaresFitter;

#line 11 "(tree fragment)"
  return o;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_7scoring_LeastSquaresFitter[] = {

#line 11 "(tree fragment)"
  {"_evaluate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_7scoring_18LeastSquaresFitter_1_evaluate, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_7scoring_18LeastSquaresFitter__evaluate},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_7scoring_LeastSquaresFitter = {

#line 11 "(tree fragment)"
  PyVarObject_HEAD_INIT(0, 0)

#line 11 "(tree fragment)"
  "ms_deisotope._c.scoring.LeastSquaresFitter", /*tp_name*/

#line 11 "(tree fragment)"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter), /*tp_basicsize*/

#line 11 "(tree fragment)"
  0, /*tp_itemsize*/

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_12ms_deisotope_2_c_7scoring_IsotopicFitterBase, /*tp_dealloc*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX < 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_vectorcall_offset*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattr*/

#line 11 "(tree fragment)"
  0, /*tp_setattr*/

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*tp_compare*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  0, /*tp_as_async*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_25__repr__, /*tp_repr*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_repr*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_as_number*/

#line 11 "(tree fragment)"
  0, /*tp_as_sequence*/

#line 11 "(tree fragment)"
  0, /*tp_as_mapping*/

#line 11 "(tree fragment)"
  0, /*tp_hash*/

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_13__call__, /*tp_call*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_call*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_str*/

#line 11 "(tree fragment)"
  0, /*tp_getattro*/

#line 11 "(tree fragment)"
  0, /*tp_setattro*/

#line 11 "(tree fragment)"
  0, /*tp_as_buffer*/

#line 11 "(tree fragment)"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 11 "(tree fragment)"
  "Evaluate an isotopic fit using least squares coefficient of\n    determination :math:`R^2`.\n\n    .. math::\n        {\\hat e_i} &= e_i / max(e)\n\n        {\\hat t_i} &= t_i / max(t)\n\n        {\\hat t} &= \\sum_i^n {\\hat t_i}^2\n\n        R^2 &= \\frac{1}{{\\hat t}}\\sum_i^n ({\\hat e_i} - {\\hat t_i})^2\n\n    where :math:`e_i` is the ith experimental peak intensity and :math:`t_i`\n    is the ith theoretical peak intensity\n    ", /*tp_doc*/

#line 11 "(tree fragment)"
  __pyx_tp_traverse_12ms_deisotope_2_c_7scoring_IsotopicFitterBase, /*tp_traverse*/

#line 11 "(tree fragment)"
  __pyx_tp_clear_12ms_deisotope_2_c_7scoring_IsotopicFitterBase, /*tp_clear*/

#line 11 "(tree fragment)"
  0, /*tp_richcompare*/

#line 11 "(tree fragment)"
  0, /*tp_weaklistoffset*/

#line 11 "(tree fragment)"
  0, /*tp_iter*/

#line 11 "(tree fragment)"
  0, /*tp_iternext*/

#line 11 "(tree fragment)"
  __pyx_methods_12ms_deisotope_2_c_7scoring_LeastSquaresFitter, /*tp_methods*/

#line 11 "(tree fragment)"
  0, /*tp_members*/

#line 11 "(tree fragment)"
  0, /*tp_getset*/

#line 11 "(tree fragment)"
  0, /*tp_base*/

#line 11 "(tree fragment)"
  0, /*tp_dict*/

#line 11 "(tree fragment)"
  0, /*tp_descr_get*/

#line 11 "(tree fragment)"
  0, /*tp_descr_set*/

#line 11 "(tree fragment)"
  0, /*tp_dictoffset*/

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_1__init__, /*tp_init*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_init*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_alloc*/

#line 11 "(tree fragment)"
  __pyx_tp_new_12ms_deisotope_2_c_7scoring_LeastSquaresFitter, /*tp_new*/

#line 11 "(tree fragment)"
  0, /*tp_free*/

#line 11 "(tree fragment)"
  0, /*tp_is_gc*/

#line 11 "(tree fragment)"
  0, /*tp_bases*/

#line 11 "(tree fragment)"
  0, /*tp_mro*/

#line 11 "(tree fragment)"
  0, /*tp_cache*/

#line 11 "(tree fragment)"
  0, /*tp_subclasses*/

#line 11 "(tree fragment)"
  0, /*tp_weaklist*/

#line 11 "(tree fragment)"
  0, /*tp_del*/

#line 11 "(tree fragment)"
  0, /*tp_version_tag*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030400a1

#line 11 "(tree fragment)"
  0, /*tp_finalize*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 11 "(tree fragment)"
  0, /*tp_vectorcall*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_pypy_flags*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_GTestFitter __pyx_vtable_12ms_deisotope_2_c_7scoring_GTestFitter;

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_GTestFitter(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter *p;

#line 11 "(tree fragment)"
  PyObject *o = __pyx_tp_new_12ms_deisotope_2_c_7scoring_IsotopicFitterBase(t, a, k);

#line 11 "(tree fragment)"
  if (unlikely(!o)) return 0;

#line 11 "(tree fragment)"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter *)o);

#line 11 "(tree fragment)"
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase*)__pyx_vtabptr_12ms_deisotope_2_c_7scoring_GTestFitter;

#line 11 "(tree fragment)"
  return o;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_7scoring_GTestFitter[] = {

#line 11 "(tree fragment)"
  {"_evaluate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_7scoring_11GTestFitter_1_evaluate, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_7scoring_11GTestFitter__evaluate},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_7scoring_GTestFitter = {

#line 11 "(tree fragment)"
  PyVarObject_HEAD_INIT(0, 0)

#line 11 "(tree fragment)"
  "ms_deisotope._c.scoring.GTestFitter", /*tp_name*/

#line 11 "(tree fragment)"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter), /*tp_basicsize*/

#line 11 "(tree fragment)"
  0, /*tp_itemsize*/

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_12ms_deisotope_2_c_7scoring_IsotopicFitterBase, /*tp_dealloc*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX < 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_vectorcall_offset*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattr*/

#line 11 "(tree fragment)"
  0, /*tp_setattr*/

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*tp_compare*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  0, /*tp_as_async*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_25__repr__, /*tp_repr*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_repr*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_as_number*/

#line 11 "(tree fragment)"
  0, /*tp_as_sequence*/

#line 11 "(tree fragment)"
  0, /*tp_as_mapping*/

#line 11 "(tree fragment)"
  0, /*tp_hash*/

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_13__call__, /*tp_call*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_call*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_str*/

#line 11 "(tree fragment)"
  0, /*tp_getattro*/

#line 11 "(tree fragment)"
  0, /*tp_setattro*/

#line 11 "(tree fragment)"
  0, /*tp_as_buffer*/

#line 11 "(tree fragment)"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 11 "(tree fragment)"
  "Evaluate an isotopic fit using a G-test\n\n    .. math::\n        G = 2\\sum_i^n{o_i * ({log}o_i - {log}e_i)}\n\n    where :math:`o_i` is the intensity of the ith experimental peak\n    and :math:`e_i` is the intensity of the ith theoretical peak.\n\n    .. note::\n\n        The G statistic is on the scale of the signal used.\n    ", /*tp_doc*/

#line 11 "(tree fragment)"
  __pyx_tp_traverse_12ms_deisotope_2_c_7scoring_IsotopicFitterBase, /*tp_traverse*/

#line 11 "(tree fragment)"
  __pyx_tp_clear_12ms_deisotope_2_c_7scoring_IsotopicFitterBase, /*tp_clear*/

#line 11 "(tree fragment)"
  0, /*tp_richcompare*/

#line 11 "(tree fragment)"
  0, /*tp_weaklistoffset*/

#line 11 "(tree fragment)"
  0, /*tp_iter*/

#line 11 "(tree fragment)"
  0, /*tp_iternext*/

#line 11 "(tree fragment)"
  __pyx_methods_12ms_deisotope_2_c_7scoring_GTestFitter, /*tp_methods*/

#line 11 "(tree fragment)"
  0, /*tp_members*/

#line 11 "(tree fragment)"
  0, /*tp_getset*/

#line 11 "(tree fragment)"
  0, /*tp_base*/

#line 11 "(tree fragment)"
  0, /*tp_dict*/

#line 11 "(tree fragment)"
  0, /*tp_descr_get*/

#line 11 "(tree fragment)"
  0, /*tp_descr_set*/

#line 11 "(tree fragment)"
  0, /*tp_dictoffset*/

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_1__init__, /*tp_init*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_init*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_alloc*/

#line 11 "(tree fragment)"
  __pyx_tp_new_12ms_deisotope_2_c_7scoring_GTestFitter, /*tp_new*/

#line 11 "(tree fragment)"
  0, /*tp_free*/

#line 11 "(tree fragment)"
  0, /*tp_is_gc*/

#line 11 "(tree fragment)"
  0, /*tp_bases*/

#line 11 "(tree fragment)"
  0, /*tp_mro*/

#line 11 "(tree fragment)"
  0, /*tp_cache*/

#line 11 "(tree fragment)"
  0, /*tp_subclasses*/

#line 11 "(tree fragment)"
  0, /*tp_weaklist*/

#line 11 "(tree fragment)"
  0, /*tp_del*/

#line 11 "(tree fragment)"
  0, /*tp_version_tag*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030400a1

#line 11 "(tree fragment)"
  0, /*tp_finalize*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 11 "(tree fragment)"
  0, /*tp_vectorcall*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_pypy_flags*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledGTestFitter __pyx_vtable_12ms_deisotope_2_c_7scoring_ScaledGTestFitter;

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_ScaledGTestFitter(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *p;

#line 11 "(tree fragment)"
  PyObject *o = __pyx_tp_new_12ms_deisotope_2_c_7scoring_IsotopicFitterBase(t, a, k);

#line 11 "(tree fragment)"
  if (unlikely(!o)) return 0;

#line 11 "(tree fragment)"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *)o);

#line 11 "(tree fragment)"
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase*)__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledGTestFitter;

#line 11 "(tree fragment)"
  return o;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_7scoring_ScaledGTestFitter[] = {

#line 11 "(tree fragment)"
  {"_evaluate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_7scoring_17ScaledGTestFitter_1_evaluate, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_7scoring_17ScaledGTestFitter__evaluate},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_7scoring_ScaledGTestFitter = {

#line 11 "(tree fragment)"
  PyVarObject_HEAD_INIT(0, 0)

#line 11 "(tree fragment)"
  "ms_deisotope._c.scoring.ScaledGTestFitter", /*tp_name*/

#line 11 "(tree fragment)"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter), /*tp_basicsize*/

#line 11 "(tree fragment)"
  0, /*tp_itemsize*/

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_12ms_deisotope_2_c_7scoring_IsotopicFitterBase, /*tp_dealloc*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX < 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_vectorcall_offset*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattr*/

#line 11 "(tree fragment)"
  0, /*tp_setattr*/

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*tp_compare*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  0, /*tp_as_async*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_25__repr__, /*tp_repr*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_repr*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_as_number*/

#line 11 "(tree fragment)"
  0, /*tp_as_sequence*/

#line 11 "(tree fragment)"
  0, /*tp_as_mapping*/

#line 11 "(tree fragment)"
  0, /*tp_hash*/

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_13__call__, /*tp_call*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_call*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_str*/

#line 11 "(tree fragment)"
  0, /*tp_getattro*/

#line 11 "(tree fragment)"
  0, /*tp_setattro*/

#line 11 "(tree fragment)"
  0, /*tp_as_buffer*/

#line 11 "(tree fragment)"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 11 "(tree fragment)"
  "Evaluate an isotopic fit using a G-test after normalizing the\n    list of experimental and theoretical peaks to both sum to 1.\n\n    .. math::\n        G = 2\\sum_i^n{o_i * ({log}o_i - {log}e_i)}\n\n    where :math:`o_i` is the intensity of the ith experimental peak\n    and :math:`e_i` is the intensity of the ith theoretical peak.\n    ", /*tp_doc*/

#line 11 "(tree fragment)"
  __pyx_tp_traverse_12ms_deisotope_2_c_7scoring_IsotopicFitterBase, /*tp_traverse*/

#line 11 "(tree fragment)"
  __pyx_tp_clear_12ms_deisotope_2_c_7scoring_IsotopicFitterBase, /*tp_clear*/

#line 11 "(tree fragment)"
  0, /*tp_richcompare*/

#line 11 "(tree fragment)"
  0, /*tp_weaklistoffset*/

#line 11 "(tree fragment)"
  0, /*tp_iter*/

#line 11 "(tree fragment)"
  0, /*tp_iternext*/

#line 11 "(tree fragment)"
  __pyx_methods_12ms_deisotope_2_c_7scoring_ScaledGTestFitter, /*tp_methods*/

#line 11 "(tree fragment)"
  0, /*tp_members*/

#line 11 "(tree fragment)"
  0, /*tp_getset*/

#line 11 "(tree fragment)"
  0, /*tp_base*/

#line 11 "(tree fragment)"
  0, /*tp_dict*/

#line 11 "(tree fragment)"
  0, /*tp_descr_get*/

#line 11 "(tree fragment)"
  0, /*tp_descr_set*/

#line 11 "(tree fragment)"
  0, /*tp_dictoffset*/

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_1__init__, /*tp_init*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_init*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_alloc*/

#line 11 "(tree fragment)"
  __pyx_tp_new_12ms_deisotope_2_c_7scoring_ScaledGTestFitter, /*tp_new*/

#line 11 "(tree fragment)"
  0, /*tp_free*/

#line 11 "(tree fragment)"
  0, /*tp_is_gc*/

#line 11 "(tree fragment)"
  0, /*tp_bases*/

#line 11 "(tree fragment)"
  0, /*tp_mro*/

#line 11 "(tree fragment)"
  0, /*tp_cache*/

#line 11 "(tree fragment)"
  0, /*tp_subclasses*/

#line 11 "(tree fragment)"
  0, /*tp_weaklist*/

#line 11 "(tree fragment)"
  0, /*tp_del*/

#line 11 "(tree fragment)"
  0, /*tp_version_tag*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030400a1

#line 11 "(tree fragment)"
  0, /*tp_finalize*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 11 "(tree fragment)"
  0, /*tp_vectorcall*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_pypy_flags*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MSDeconVFitter __pyx_vtable_12ms_deisotope_2_c_7scoring_MSDeconVFitter;

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_MSDeconVFitter(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *p;

#line 11 "(tree fragment)"
  PyObject *o = __pyx_tp_new_12ms_deisotope_2_c_7scoring_IsotopicFitterBase(t, a, k);

#line 11 "(tree fragment)"
  if (unlikely(!o)) return 0;

#line 11 "(tree fragment)"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *)o);

#line 11 "(tree fragment)"
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase*)__pyx_vtabptr_12ms_deisotope_2_c_7scoring_MSDeconVFitter;

#line 11 "(tree fragment)"
  return o;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_mass_error_tolerance(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_20mass_error_tolerance_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_mass_error_tolerance(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_20mass_error_tolerance_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 11 "(tree fragment)"
    return -1;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_7scoring_MSDeconVFitter[] = {

#line 11 "(tree fragment)"
  {"evaluate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_3evaluate, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_2evaluate},

#line 11 "(tree fragment)"
  {"__reduce__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_5__reduce__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_4__reduce__},

#line 11 "(tree fragment)"
  {"__getstate__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_7__getstate__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_6__getstate__},

#line 11 "(tree fragment)"
  {"__setstate__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_9__setstate__, METH_O, __pyx_doc_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_8__setstate__},

#line 11 "(tree fragment)"
  {"_evaluate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_11_evaluate, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_10_evaluate},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_7scoring_MSDeconVFitter[] = {

#line 11 "(tree fragment)"
  {(char *)"mass_error_tolerance", __pyx_getprop_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_mass_error_tolerance, __pyx_setprop_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_mass_error_tolerance, (char *)"mass_error_tolerance: 'double'", 0},

#line 11 "(tree fragment)"
  {0, 0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_7scoring_MSDeconVFitter = {

#line 11 "(tree fragment)"
  PyVarObject_HEAD_INIT(0, 0)

#line 11 "(tree fragment)"
  "ms_deisotope._c.scoring.MSDeconVFitter", /*tp_name*/

#line 11 "(tree fragment)"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter), /*tp_basicsize*/

#line 11 "(tree fragment)"
  0, /*tp_itemsize*/

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_12ms_deisotope_2_c_7scoring_IsotopicFitterBase, /*tp_dealloc*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX < 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_vectorcall_offset*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattr*/

#line 11 "(tree fragment)"
  0, /*tp_setattr*/

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*tp_compare*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  0, /*tp_as_async*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_25__repr__, /*tp_repr*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_repr*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_as_number*/

#line 11 "(tree fragment)"
  0, /*tp_as_sequence*/

#line 11 "(tree fragment)"
  0, /*tp_as_mapping*/

#line 11 "(tree fragment)"
  0, /*tp_hash*/

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_13__call__, /*tp_call*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_call*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_str*/

#line 11 "(tree fragment)"
  0, /*tp_getattro*/

#line 11 "(tree fragment)"
  0, /*tp_setattro*/

#line 11 "(tree fragment)"
  0, /*tp_as_buffer*/

#line 11 "(tree fragment)"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 11 "(tree fragment)"
  "MSDeconVFitter(minimum_score=10, mass_error_tolerance=0.02)\nAn implementation of the scoring function used in :title-reference:`MSDeconV`\n\n    .. math::\n        :nowrap:\n\n        \\begin{split}\n            s_{mz}(e, t) &= \\begin{cases}\n                1 - \\frac{\\|mz(e) - mz(t)\\|}{d} & \\text{if } \\|mz(e) - mz(t)\\| < d,\\\\\n                0 & \\text{otherwise}\n            \\end{cases}\\\\\n\n            s_{int}(e, t) &= \\begin{cases}\n                1 - \\frac{int(t) - int(e)}{int(e)} & \\text{if } int(e) < int(t) \\text{ and } \\frac{int(t) - int(e)}{int(e)} \\le 1, \\\\\n                \\sqrt{1 - \\frac{int(e) - int(t)}{int(t)}} & \\text{if } int(e) \\ge int(t) \\text{ and } \\frac{int(e) - int(t)}{int(t)} \\le 1,\\\\\n                0 & \\text{otherwise}\n            \\end{cases}\\\\\n\n            \\text{S}(e, t) &= \\sqrt{int(t)}\\times s_{mz}(e,t) \\times s_{int}(e, t)\n        \\end{split}\n\n    References\n    ----------\n    Liu, X., Inbar, Y., Dorrestein, P. C., Wynne, C., Edwards, N., Souda, P., \342\200\246\n    Pevzner, P. A. (2010). Deconvolution and database search of complex tandem\n    mass spectra of intact proteins: a combinatorial approach. Molecular & Cellular\n    Proteomics : MCP, 9(12), 2772\342\200\2232782. https://doi.org/10.1074/mcp.M110.002766\n    ", /*tp_doc*/

#line 11 "(tree fragment)"
  __pyx_tp_traverse_12ms_deisotope_2_c_7scoring_IsotopicFitterBase, /*tp_traverse*/

#line 11 "(tree fragment)"
  __pyx_tp_clear_12ms_deisotope_2_c_7scoring_IsotopicFitterBase, /*tp_clear*/

#line 11 "(tree fragment)"
  0, /*tp_richcompare*/

#line 11 "(tree fragment)"
  0, /*tp_weaklistoffset*/

#line 11 "(tree fragment)"
  0, /*tp_iter*/

#line 11 "(tree fragment)"
  0, /*tp_iternext*/

#line 11 "(tree fragment)"
  __pyx_methods_12ms_deisotope_2_c_7scoring_MSDeconVFitter, /*tp_methods*/

#line 11 "(tree fragment)"
  0, /*tp_members*/

#line 11 "(tree fragment)"
  __pyx_getsets_12ms_deisotope_2_c_7scoring_MSDeconVFitter, /*tp_getset*/

#line 11 "(tree fragment)"
  0, /*tp_base*/

#line 11 "(tree fragment)"
  0, /*tp_dict*/

#line 11 "(tree fragment)"
  0, /*tp_descr_get*/

#line 11 "(tree fragment)"
  0, /*tp_descr_set*/

#line 11 "(tree fragment)"
  0, /*tp_dictoffset*/

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_1__init__, /*tp_init*/

#line 11 "(tree fragment)"
  0, /*tp_alloc*/

#line 11 "(tree fragment)"
  __pyx_tp_new_12ms_deisotope_2_c_7scoring_MSDeconVFitter, /*tp_new*/

#line 11 "(tree fragment)"
  0, /*tp_free*/

#line 11 "(tree fragment)"
  0, /*tp_is_gc*/

#line 11 "(tree fragment)"
  0, /*tp_bases*/

#line 11 "(tree fragment)"
  0, /*tp_mro*/

#line 11 "(tree fragment)"
  0, /*tp_cache*/

#line 11 "(tree fragment)"
  0, /*tp_subclasses*/

#line 11 "(tree fragment)"
  0, /*tp_weaklist*/

#line 11 "(tree fragment)"
  0, /*tp_del*/

#line 11 "(tree fragment)"
  0, /*tp_version_tag*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030400a1

#line 11 "(tree fragment)"
  0, /*tp_finalize*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 11 "(tree fragment)"
  0, /*tp_vectorcall*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_pypy_flags*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter __pyx_vtable_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter;

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *p;

#line 11 "(tree fragment)"
  PyObject *o = __pyx_tp_new_12ms_deisotope_2_c_7scoring_IsotopicFitterBase(t, a, k);

#line 11 "(tree fragment)"
  if (unlikely(!o)) return 0;

#line 11 "(tree fragment)"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *)o);

#line 11 "(tree fragment)"
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase*)__pyx_vtabptr_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter;

#line 11 "(tree fragment)"
  return o;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_mass_error_tolerance(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_20mass_error_tolerance_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_mass_error_tolerance(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_20mass_error_tolerance_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 11 "(tree fragment)"
    return -1;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_penalty_factor(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_14penalty_factor_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_penalty_factor(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_14penalty_factor_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 11 "(tree fragment)"
    return -1;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter[] = {

#line 11 "(tree fragment)"
  {"__reduce__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_3__reduce__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_2__reduce__},

#line 11 "(tree fragment)"
  {"__getstate__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_5__getstate__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_4__getstate__},

#line 11 "(tree fragment)"
  {"__setstate__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_7__setstate__, METH_O, __pyx_doc_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_6__setstate__},

#line 11 "(tree fragment)"
  {"evaluate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_9evaluate, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_8evaluate},

#line 11 "(tree fragment)"
  {"_evaluate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_11_evaluate, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_10_evaluate},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter[] = {

#line 11 "(tree fragment)"
  {(char *)"mass_error_tolerance", __pyx_getprop_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_mass_error_tolerance, __pyx_setprop_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_mass_error_tolerance, (char *)"mass_error_tolerance: 'double'", 0},

#line 11 "(tree fragment)"
  {(char *)"penalty_factor", __pyx_getprop_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_penalty_factor, __pyx_setprop_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_penalty_factor, (char *)"penalty_factor: 'double'", 0},

#line 11 "(tree fragment)"
  {0, 0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter = {

#line 11 "(tree fragment)"
  PyVarObject_HEAD_INIT(0, 0)

#line 11 "(tree fragment)"
  "ms_deisotope._c.scoring.PenalizedMSDeconVFitter", /*tp_name*/

#line 11 "(tree fragment)"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter), /*tp_basicsize*/

#line 11 "(tree fragment)"
  0, /*tp_itemsize*/

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_12ms_deisotope_2_c_7scoring_IsotopicFitterBase, /*tp_dealloc*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX < 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_vectorcall_offset*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattr*/

#line 11 "(tree fragment)"
  0, /*tp_setattr*/

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*tp_compare*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  0, /*tp_as_async*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_25__repr__, /*tp_repr*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_repr*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_as_number*/

#line 11 "(tree fragment)"
  0, /*tp_as_sequence*/

#line 11 "(tree fragment)"
  0, /*tp_as_mapping*/

#line 11 "(tree fragment)"
  0, /*tp_hash*/

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_13__call__, /*tp_call*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_call*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_str*/

#line 11 "(tree fragment)"
  0, /*tp_getattro*/

#line 11 "(tree fragment)"
  0, /*tp_setattro*/

#line 11 "(tree fragment)"
  0, /*tp_as_buffer*/

#line 11 "(tree fragment)"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 11 "(tree fragment)"
  "PenalizedMSDeconVFitter(minimum_score=10, penalty_factor=1, mass_error_tolerance=0.02)\nAn Isotopic Fitter which uses the :class:`MSDeconVFitter` score\n    weighted by 1 - :attr:`penalty_factor` * :class:`ScaledGTestFitter` score\n\n    .. math::\n        S(e, t) = M(e, t) * (1 - G(e, t))\n\n    where :math:`e` is the experimental peak list and :math:`t` is the theoretical\n    peak list\n    ", /*tp_doc*/

#line 11 "(tree fragment)"
  __pyx_tp_traverse_12ms_deisotope_2_c_7scoring_IsotopicFitterBase, /*tp_traverse*/

#line 11 "(tree fragment)"
  __pyx_tp_clear_12ms_deisotope_2_c_7scoring_IsotopicFitterBase, /*tp_clear*/

#line 11 "(tree fragment)"
  0, /*tp_richcompare*/

#line 11 "(tree fragment)"
  0, /*tp_weaklistoffset*/

#line 11 "(tree fragment)"
  0, /*tp_iter*/

#line 11 "(tree fragment)"
  0, /*tp_iternext*/

#line 11 "(tree fragment)"
  __pyx_methods_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter, /*tp_methods*/

#line 11 "(tree fragment)"
  0, /*tp_members*/

#line 11 "(tree fragment)"
  __pyx_getsets_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter, /*tp_getset*/

#line 11 "(tree fragment)"
  0, /*tp_base*/

#line 11 "(tree fragment)"
  0, /*tp_dict*/

#line 11 "(tree fragment)"
  0, /*tp_descr_get*/

#line 11 "(tree fragment)"
  0, /*tp_descr_set*/

#line 11 "(tree fragment)"
  0, /*tp_dictoffset*/

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter_1__init__, /*tp_init*/

#line 11 "(tree fragment)"
  0, /*tp_alloc*/

#line 11 "(tree fragment)"
  __pyx_tp_new_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter, /*tp_new*/

#line 11 "(tree fragment)"
  0, /*tp_free*/

#line 11 "(tree fragment)"
  0, /*tp_is_gc*/

#line 11 "(tree fragment)"
  0, /*tp_bases*/

#line 11 "(tree fragment)"
  0, /*tp_mro*/

#line 11 "(tree fragment)"
  0, /*tp_cache*/

#line 11 "(tree fragment)"
  0, /*tp_subclasses*/

#line 11 "(tree fragment)"
  0, /*tp_weaklist*/

#line 11 "(tree fragment)"
  0, /*tp_del*/

#line 11 "(tree fragment)"
  0, /*tp_version_tag*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030400a1

#line 11 "(tree fragment)"
  0, /*tp_finalize*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 11 "(tree fragment)"
  0, /*tp_vectorcall*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_pypy_flags*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FunctionScorer __pyx_vtable_12ms_deisotope_2_c_7scoring_FunctionScorer;

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_FunctionScorer(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer *p;

#line 11 "(tree fragment)"
  PyObject *o = __pyx_tp_new_12ms_deisotope_2_c_7scoring_IsotopicFitterBase(t, a, k);

#line 11 "(tree fragment)"
  if (unlikely(!o)) return 0;

#line 11 "(tree fragment)"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer *)o);

#line 11 "(tree fragment)"
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase*)__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FunctionScorer;

#line 11 "(tree fragment)"
  p->function = Py_None; Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  return o;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_7scoring_FunctionScorer(PyObject *o) {

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer *p = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer *)o;

#line 11 "(tree fragment)"
  #if CYTHON_USE_TP_FINALIZE

#line 11 "(tree fragment)"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {

#line 11 "(tree fragment)"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  PyObject_GC_UnTrack(o);

#line 11 "(tree fragment)"
  Py_CLEAR(p->function);

#line 11 "(tree fragment)"
  PyObject_GC_Track(o);

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_12ms_deisotope_2_c_7scoring_IsotopicFitterBase(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_traverse_12ms_deisotope_2_c_7scoring_FunctionScorer(PyObject *o, visitproc v, void *a) {

#line 11 "(tree fragment)"
  int e;

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer *p = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer *)o;

#line 11 "(tree fragment)"
  e = __pyx_tp_traverse_12ms_deisotope_2_c_7scoring_IsotopicFitterBase(o, v, a); if (e) return e;

#line 11 "(tree fragment)"
  if (p->function) {

#line 11 "(tree fragment)"
    e = (*v)(p->function, a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_clear_12ms_deisotope_2_c_7scoring_FunctionScorer(PyObject *o) {

#line 11 "(tree fragment)"
  PyObject* tmp;

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer *p = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer *)o;

#line 11 "(tree fragment)"
  __pyx_tp_clear_12ms_deisotope_2_c_7scoring_IsotopicFitterBase(o);

#line 11 "(tree fragment)"
  tmp = ((PyObject*)p->function);

#line 11 "(tree fragment)"
  p->function = Py_None; Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  Py_XDECREF(tmp);

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_7scoring_FunctionScorer[] = {

#line 11 "(tree fragment)"
  {"__getstate__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_14FunctionScorer_3__getstate__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_7scoring_14FunctionScorer_2__getstate__},

#line 11 "(tree fragment)"
  {"__setstate__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_14FunctionScorer_5__setstate__, METH_O, __pyx_doc_12ms_deisotope_2_c_7scoring_14FunctionScorer_4__setstate__},

#line 11 "(tree fragment)"
  {"__reduce__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_14FunctionScorer_7__reduce__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_7scoring_14FunctionScorer_6__reduce__},

#line 11 "(tree fragment)"
  {"_evaluate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_7scoring_14FunctionScorer_9_evaluate, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_7scoring_14FunctionScorer_8_evaluate},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_7scoring_FunctionScorer = {

#line 11 "(tree fragment)"
  PyVarObject_HEAD_INIT(0, 0)

#line 11 "(tree fragment)"
  "ms_deisotope._c.scoring.FunctionScorer", /*tp_name*/

#line 11 "(tree fragment)"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer), /*tp_basicsize*/

#line 11 "(tree fragment)"
  0, /*tp_itemsize*/

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_12ms_deisotope_2_c_7scoring_FunctionScorer, /*tp_dealloc*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX < 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_vectorcall_offset*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattr*/

#line 11 "(tree fragment)"
  0, /*tp_setattr*/

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*tp_compare*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  0, /*tp_as_async*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_25__repr__, /*tp_repr*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_repr*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_as_number*/

#line 11 "(tree fragment)"
  0, /*tp_as_sequence*/

#line 11 "(tree fragment)"
  0, /*tp_as_mapping*/

#line 11 "(tree fragment)"
  0, /*tp_hash*/

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_13__call__, /*tp_call*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_call*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_str*/

#line 11 "(tree fragment)"
  0, /*tp_getattro*/

#line 11 "(tree fragment)"
  0, /*tp_setattro*/

#line 11 "(tree fragment)"
  0, /*tp_as_buffer*/

#line 11 "(tree fragment)"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 11 "(tree fragment)"
  "FunctionScorer(function, minimum_score=10, selector_type=MaximizeFitSelector)\nUse a user-provided Python function or callable object to evaluate an isotopic\n    envelope fit.\n\n    The provided function should take two arguments, a list of experimental fitted peaks\n    and a list of theoretical peaks, and return a single number as an output.\n\n    Make sure to pass the right thresholds and selector types when creating an instance\n    of this class that match the semantics of the provided function.\n\n    While this type is available for convenience, there is considerable overhead in using\n    it compared to one of the C accelerated fitters.\n    ", /*tp_doc*/

#line 11 "(tree fragment)"
  __pyx_tp_traverse_12ms_deisotope_2_c_7scoring_FunctionScorer, /*tp_traverse*/

#line 11 "(tree fragment)"
  __pyx_tp_clear_12ms_deisotope_2_c_7scoring_FunctionScorer, /*tp_clear*/

#line 11 "(tree fragment)"
  0, /*tp_richcompare*/

#line 11 "(tree fragment)"
  0, /*tp_weaklistoffset*/

#line 11 "(tree fragment)"
  0, /*tp_iter*/

#line 11 "(tree fragment)"
  0, /*tp_iternext*/

#line 11 "(tree fragment)"
  __pyx_methods_12ms_deisotope_2_c_7scoring_FunctionScorer, /*tp_methods*/

#line 11 "(tree fragment)"
  0, /*tp_members*/

#line 11 "(tree fragment)"
  0, /*tp_getset*/

#line 11 "(tree fragment)"
  0, /*tp_base*/

#line 11 "(tree fragment)"
  0, /*tp_dict*/

#line 11 "(tree fragment)"
  0, /*tp_descr_get*/

#line 11 "(tree fragment)"
  0, /*tp_descr_set*/

#line 11 "(tree fragment)"
  0, /*tp_dictoffset*/

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_14FunctionScorer_1__init__, /*tp_init*/

#line 11 "(tree fragment)"
  0, /*tp_alloc*/

#line 11 "(tree fragment)"
  __pyx_tp_new_12ms_deisotope_2_c_7scoring_FunctionScorer, /*tp_new*/

#line 11 "(tree fragment)"
  0, /*tp_free*/

#line 11 "(tree fragment)"
  0, /*tp_is_gc*/

#line 11 "(tree fragment)"
  0, /*tp_bases*/

#line 11 "(tree fragment)"
  0, /*tp_mro*/

#line 11 "(tree fragment)"
  0, /*tp_cache*/

#line 11 "(tree fragment)"
  0, /*tp_subclasses*/

#line 11 "(tree fragment)"
  0, /*tp_weaklist*/

#line 11 "(tree fragment)"
  0, /*tp_del*/

#line 11 "(tree fragment)"
  0, /*tp_version_tag*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030400a1

#line 11 "(tree fragment)"
  0, /*tp_finalize*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 11 "(tree fragment)"
  0, /*tp_vectorcall*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_pypy_flags*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection __pyx_vtable_12ms_deisotope_2_c_7scoring_InterferenceDetection;

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_InterferenceDetection(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *p;

#line 11 "(tree fragment)"
  PyObject *o;

#line 11 "(tree fragment)"
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {

#line 11 "(tree fragment)"
    o = (*t->tp_alloc)(t, 0);

#line 11 "(tree fragment)"
  } else {

#line 11 "(tree fragment)"
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  if (unlikely(!o)) return 0;

#line 11 "(tree fragment)"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *)o);

#line 11 "(tree fragment)"
  p->__pyx_vtab = __pyx_vtabptr_12ms_deisotope_2_c_7scoring_InterferenceDetection;

#line 11 "(tree fragment)"
  p->peaklist = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)Py_None); Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  return o;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_7scoring_InterferenceDetection(PyObject *o) {

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *p = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *)o;

#line 11 "(tree fragment)"
  #if CYTHON_USE_TP_FINALIZE

#line 11 "(tree fragment)"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {

#line 11 "(tree fragment)"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  PyObject_GC_UnTrack(o);

#line 11 "(tree fragment)"
  Py_CLEAR(p->peaklist);

#line 11 "(tree fragment)"
  (*Py_TYPE(o)->tp_free)(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_traverse_12ms_deisotope_2_c_7scoring_InterferenceDetection(PyObject *o, visitproc v, void *a) {

#line 11 "(tree fragment)"
  int e;

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *p = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *)o;

#line 11 "(tree fragment)"
  if (p->peaklist) {

#line 11 "(tree fragment)"
    e = (*v)(((PyObject *)p->peaklist), a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_clear_12ms_deisotope_2_c_7scoring_InterferenceDetection(PyObject *o) {

#line 11 "(tree fragment)"
  PyObject* tmp;

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *p = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *)o;

#line 11 "(tree fragment)"
  tmp = ((PyObject*)p->peaklist);

#line 11 "(tree fragment)"
  p->peaklist = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *)Py_None); Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  Py_XDECREF(tmp);

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_7scoring_21InterferenceDetection_peaklist(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_7scoring_21InterferenceDetection_8peaklist_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_7scoring_21InterferenceDetection_peaklist(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_21InterferenceDetection_8peaklist_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_21InterferenceDetection_8peaklist_5__del__(o);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_7scoring_InterferenceDetection[] = {

#line 11 "(tree fragment)"
  {"__reduce__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_21InterferenceDetection_3__reduce__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_7scoring_21InterferenceDetection_2__reduce__},

#line 11 "(tree fragment)"
  {"__getstate__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_21InterferenceDetection_5__getstate__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_7scoring_21InterferenceDetection_4__getstate__},

#line 11 "(tree fragment)"
  {"__setstate__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_21InterferenceDetection_7__setstate__, METH_O, __pyx_doc_12ms_deisotope_2_c_7scoring_21InterferenceDetection_6__setstate__},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_7scoring_InterferenceDetection[] = {

#line 11 "(tree fragment)"
  {(char *)"peaklist", __pyx_getprop_12ms_deisotope_2_c_7scoring_21InterferenceDetection_peaklist, __pyx_setprop_12ms_deisotope_2_c_7scoring_21InterferenceDetection_peaklist, (char *)"peaklist: ms_peak_picker._c.peak_set.PeakSet", 0},

#line 11 "(tree fragment)"
  {0, 0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_7scoring_InterferenceDetection = {

#line 11 "(tree fragment)"
  PyVarObject_HEAD_INIT(0, 0)

#line 11 "(tree fragment)"
  "ms_deisotope._c.scoring.InterferenceDetection", /*tp_name*/

#line 11 "(tree fragment)"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection), /*tp_basicsize*/

#line 11 "(tree fragment)"
  0, /*tp_itemsize*/

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_12ms_deisotope_2_c_7scoring_InterferenceDetection, /*tp_dealloc*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX < 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_vectorcall_offset*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattr*/

#line 11 "(tree fragment)"
  0, /*tp_setattr*/

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*tp_compare*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  0, /*tp_as_async*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_repr*/

#line 11 "(tree fragment)"
  0, /*tp_as_number*/

#line 11 "(tree fragment)"
  0, /*tp_as_sequence*/

#line 11 "(tree fragment)"
  0, /*tp_as_mapping*/

#line 11 "(tree fragment)"
  0, /*tp_hash*/

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_21InterferenceDetection_9__call__, /*tp_call*/

#line 11 "(tree fragment)"
  0, /*tp_str*/

#line 11 "(tree fragment)"
  0, /*tp_getattro*/

#line 11 "(tree fragment)"
  0, /*tp_setattro*/

#line 11 "(tree fragment)"
  0, /*tp_as_buffer*/

#line 11 "(tree fragment)"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 11 "(tree fragment)"
  "InterferenceDetection(PeakSet peaklist)", /*tp_doc*/

#line 11 "(tree fragment)"
  __pyx_tp_traverse_12ms_deisotope_2_c_7scoring_InterferenceDetection, /*tp_traverse*/

#line 11 "(tree fragment)"
  __pyx_tp_clear_12ms_deisotope_2_c_7scoring_InterferenceDetection, /*tp_clear*/

#line 11 "(tree fragment)"
  0, /*tp_richcompare*/

#line 11 "(tree fragment)"
  0, /*tp_weaklistoffset*/

#line 11 "(tree fragment)"
  0, /*tp_iter*/

#line 11 "(tree fragment)"
  0, /*tp_iternext*/

#line 11 "(tree fragment)"
  __pyx_methods_12ms_deisotope_2_c_7scoring_InterferenceDetection, /*tp_methods*/

#line 11 "(tree fragment)"
  0, /*tp_members*/

#line 11 "(tree fragment)"
  __pyx_getsets_12ms_deisotope_2_c_7scoring_InterferenceDetection, /*tp_getset*/

#line 11 "(tree fragment)"
  0, /*tp_base*/

#line 11 "(tree fragment)"
  0, /*tp_dict*/

#line 11 "(tree fragment)"
  0, /*tp_descr_get*/

#line 11 "(tree fragment)"
  0, /*tp_descr_set*/

#line 11 "(tree fragment)"
  0, /*tp_dictoffset*/

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_21InterferenceDetection_1__init__, /*tp_init*/

#line 11 "(tree fragment)"
  0, /*tp_alloc*/

#line 11 "(tree fragment)"
  __pyx_tp_new_12ms_deisotope_2_c_7scoring_InterferenceDetection, /*tp_new*/

#line 11 "(tree fragment)"
  0, /*tp_free*/

#line 11 "(tree fragment)"
  0, /*tp_is_gc*/

#line 11 "(tree fragment)"
  0, /*tp_bases*/

#line 11 "(tree fragment)"
  0, /*tp_mro*/

#line 11 "(tree fragment)"
  0, /*tp_cache*/

#line 11 "(tree fragment)"
  0, /*tp_subclasses*/

#line 11 "(tree fragment)"
  0, /*tp_weaklist*/

#line 11 "(tree fragment)"
  0, /*tp_del*/

#line 11 "(tree fragment)"
  0, /*tp_version_tag*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030400a1

#line 11 "(tree fragment)"
  0, /*tp_finalize*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 11 "(tree fragment)"
  0, /*tp_vectorcall*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_pypy_flags*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DistinctPatternFitter __pyx_vtable_12ms_deisotope_2_c_7scoring_DistinctPatternFitter;

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_DistinctPatternFitter(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *p;

#line 11 "(tree fragment)"
  PyObject *o = __pyx_tp_new_12ms_deisotope_2_c_7scoring_IsotopicFitterBase(t, a, k);

#line 11 "(tree fragment)"
  if (unlikely(!o)) return 0;

#line 11 "(tree fragment)"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *)o);

#line 11 "(tree fragment)"
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase*)__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DistinctPatternFitter;

#line 11 "(tree fragment)"
  p->interference_detector = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *)Py_None); Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  p->g_test_scaled = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *)Py_None); Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  return o;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_7scoring_DistinctPatternFitter(PyObject *o) {

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *p = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *)o;

#line 11 "(tree fragment)"
  #if CYTHON_USE_TP_FINALIZE

#line 11 "(tree fragment)"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {

#line 11 "(tree fragment)"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  PyObject_GC_UnTrack(o);

#line 11 "(tree fragment)"
  Py_CLEAR(p->interference_detector);

#line 11 "(tree fragment)"
  Py_CLEAR(p->g_test_scaled);

#line 11 "(tree fragment)"
  PyObject_GC_Track(o);

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_12ms_deisotope_2_c_7scoring_IsotopicFitterBase(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_traverse_12ms_deisotope_2_c_7scoring_DistinctPatternFitter(PyObject *o, visitproc v, void *a) {

#line 11 "(tree fragment)"
  int e;

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *p = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *)o;

#line 11 "(tree fragment)"
  e = __pyx_tp_traverse_12ms_deisotope_2_c_7scoring_IsotopicFitterBase(o, v, a); if (e) return e;

#line 11 "(tree fragment)"
  if (p->interference_detector) {

#line 11 "(tree fragment)"
    e = (*v)(((PyObject *)p->interference_detector), a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  if (p->g_test_scaled) {

#line 11 "(tree fragment)"
    e = (*v)(((PyObject *)p->g_test_scaled), a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_clear_12ms_deisotope_2_c_7scoring_DistinctPatternFitter(PyObject *o) {

#line 11 "(tree fragment)"
  PyObject* tmp;

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *p = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *)o;

#line 11 "(tree fragment)"
  __pyx_tp_clear_12ms_deisotope_2_c_7scoring_IsotopicFitterBase(o);

#line 11 "(tree fragment)"
  tmp = ((PyObject*)p->interference_detector);

#line 11 "(tree fragment)"
  p->interference_detector = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *)Py_None); Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  Py_XDECREF(tmp);

#line 11 "(tree fragment)"
  tmp = ((PyObject*)p->g_test_scaled);

#line 11 "(tree fragment)"
  p->g_test_scaled = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *)Py_None); Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  Py_XDECREF(tmp);

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_interference_detector(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_21interference_detector_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_interference_detector(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_21interference_detector_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_21interference_detector_5__del__(o);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_g_test_scaled(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_13g_test_scaled_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_g_test_scaled(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_13g_test_scaled_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_13g_test_scaled_5__del__(o);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_peak_count_scale(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_16peak_count_scale_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_peak_count_scale(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_16peak_count_scale_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 11 "(tree fragment)"
    return -1;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_domain_scale(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_12domain_scale_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_domain_scale(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_12domain_scale_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 11 "(tree fragment)"
    return -1;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_7scoring_DistinctPatternFitter[] = {

#line 11 "(tree fragment)"
  {"__getstate__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_3__getstate__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_2__getstate__},

#line 11 "(tree fragment)"
  {"__setstate__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_5__setstate__, METH_O, __pyx_doc_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_4__setstate__},

#line 11 "(tree fragment)"
  {"_configure", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_7_configure, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_6_configure},

#line 11 "(tree fragment)"
  {"_evaluate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_9_evaluate, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_8_evaluate},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_7scoring_DistinctPatternFitter[] = {

#line 11 "(tree fragment)"
  {(char *)"interference_detector", __pyx_getprop_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_interference_detector, __pyx_setprop_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_interference_detector, (char *)"interference_detector: ms_deisotope._c.scoring.InterferenceDetection", 0},

#line 11 "(tree fragment)"
  {(char *)"g_test_scaled", __pyx_getprop_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_g_test_scaled, __pyx_setprop_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_g_test_scaled, (char *)"g_test_scaled: ms_deisotope._c.scoring.ScaledGTestFitter", 0},

#line 11 "(tree fragment)"
  {(char *)"peak_count_scale", __pyx_getprop_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_peak_count_scale, __pyx_setprop_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_peak_count_scale, (char *)"peak_count_scale: 'double'", 0},

#line 11 "(tree fragment)"
  {(char *)"domain_scale", __pyx_getprop_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_domain_scale, __pyx_setprop_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_domain_scale, (char *)"domain_scale: 'double'", 0},

#line 11 "(tree fragment)"
  {0, 0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_7scoring_DistinctPatternFitter = {

#line 11 "(tree fragment)"
  PyVarObject_HEAD_INIT(0, 0)

#line 11 "(tree fragment)"
  "ms_deisotope._c.scoring.DistinctPatternFitter", /*tp_name*/

#line 11 "(tree fragment)"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter), /*tp_basicsize*/

#line 11 "(tree fragment)"
  0, /*tp_itemsize*/

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_12ms_deisotope_2_c_7scoring_DistinctPatternFitter, /*tp_dealloc*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX < 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_vectorcall_offset*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattr*/

#line 11 "(tree fragment)"
  0, /*tp_setattr*/

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*tp_compare*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  0, /*tp_as_async*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_25__repr__, /*tp_repr*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_repr*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_as_number*/

#line 11 "(tree fragment)"
  0, /*tp_as_sequence*/

#line 11 "(tree fragment)"
  0, /*tp_as_mapping*/

#line 11 "(tree fragment)"
  0, /*tp_hash*/

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_13__call__, /*tp_call*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_call*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_str*/

#line 11 "(tree fragment)"
  0, /*tp_getattro*/

#line 11 "(tree fragment)"
  0, /*tp_setattro*/

#line 11 "(tree fragment)"
  0, /*tp_as_buffer*/

#line 11 "(tree fragment)"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 11 "(tree fragment)"
  "DistinctPatternFitter(minimum_score=0.3, peak_count_scale=1.5, domain_scale=100.)", /*tp_doc*/

#line 11 "(tree fragment)"
  __pyx_tp_traverse_12ms_deisotope_2_c_7scoring_DistinctPatternFitter, /*tp_traverse*/

#line 11 "(tree fragment)"
  __pyx_tp_clear_12ms_deisotope_2_c_7scoring_DistinctPatternFitter, /*tp_clear*/

#line 11 "(tree fragment)"
  0, /*tp_richcompare*/

#line 11 "(tree fragment)"
  0, /*tp_weaklistoffset*/

#line 11 "(tree fragment)"
  0, /*tp_iter*/

#line 11 "(tree fragment)"
  0, /*tp_iternext*/

#line 11 "(tree fragment)"
  __pyx_methods_12ms_deisotope_2_c_7scoring_DistinctPatternFitter, /*tp_methods*/

#line 11 "(tree fragment)"
  0, /*tp_members*/

#line 11 "(tree fragment)"
  __pyx_getsets_12ms_deisotope_2_c_7scoring_DistinctPatternFitter, /*tp_getset*/

#line 11 "(tree fragment)"
  0, /*tp_base*/

#line 11 "(tree fragment)"
  0, /*tp_dict*/

#line 11 "(tree fragment)"
  0, /*tp_descr_get*/

#line 11 "(tree fragment)"
  0, /*tp_descr_set*/

#line 11 "(tree fragment)"
  0, /*tp_dictoffset*/

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter_1__init__, /*tp_init*/

#line 11 "(tree fragment)"
  0, /*tp_alloc*/

#line 11 "(tree fragment)"
  __pyx_tp_new_12ms_deisotope_2_c_7scoring_DistinctPatternFitter, /*tp_new*/

#line 11 "(tree fragment)"
  0, /*tp_free*/

#line 11 "(tree fragment)"
  0, /*tp_is_gc*/

#line 11 "(tree fragment)"
  0, /*tp_bases*/

#line 11 "(tree fragment)"
  0, /*tp_mro*/

#line 11 "(tree fragment)"
  0, /*tp_cache*/

#line 11 "(tree fragment)"
  0, /*tp_subclasses*/

#line 11 "(tree fragment)"
  0, /*tp_weaklist*/

#line 11 "(tree fragment)"
  0, /*tp_del*/

#line 11 "(tree fragment)"
  0, /*tp_version_tag*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030400a1

#line 11 "(tree fragment)"
  0, /*tp_finalize*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 11 "(tree fragment)"
  0, /*tp_vectorcall*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_pypy_flags*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter __pyx_vtable_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter;

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *p;

#line 11 "(tree fragment)"
  PyObject *o = __pyx_tp_new_12ms_deisotope_2_c_7scoring_IsotopicFitterBase(t, a, k);

#line 11 "(tree fragment)"
  if (unlikely(!o)) return 0;

#line 11 "(tree fragment)"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *)o);

#line 11 "(tree fragment)"
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase*)__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter;

#line 11 "(tree fragment)"
  p->scorer = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *)Py_None); Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  return o;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter(PyObject *o) {

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *p = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *)o;

#line 11 "(tree fragment)"
  #if CYTHON_USE_TP_FINALIZE

#line 11 "(tree fragment)"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {

#line 11 "(tree fragment)"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  PyObject_GC_UnTrack(o);

#line 11 "(tree fragment)"
  Py_CLEAR(p->scorer);

#line 11 "(tree fragment)"
  PyObject_GC_Track(o);

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_12ms_deisotope_2_c_7scoring_IsotopicFitterBase(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_traverse_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter(PyObject *o, visitproc v, void *a) {

#line 11 "(tree fragment)"
  int e;

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *p = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *)o;

#line 11 "(tree fragment)"
  e = __pyx_tp_traverse_12ms_deisotope_2_c_7scoring_IsotopicFitterBase(o, v, a); if (e) return e;

#line 11 "(tree fragment)"
  if (p->scorer) {

#line 11 "(tree fragment)"
    e = (*v)(((PyObject *)p->scorer), a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_clear_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter(PyObject *o) {

#line 11 "(tree fragment)"
  PyObject* tmp;

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *p = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *)o;

#line 11 "(tree fragment)"
  __pyx_tp_clear_12ms_deisotope_2_c_7scoring_IsotopicFitterBase(o);

#line 11 "(tree fragment)"
  tmp = ((PyObject*)p->scorer);

#line 11 "(tree fragment)"
  p->scorer = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *)Py_None); Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  Py_XDECREF(tmp);

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_penalty_factor(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_14penalty_factor_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_penalty_factor(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_14penalty_factor_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 11 "(tree fragment)"
    return -1;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_scale_factor(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_12scale_factor_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_scale_factor(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_12scale_factor_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 11 "(tree fragment)"
    return -1;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_scorer(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_6scorer_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_setprop_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_scorer(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_6scorer_3__set__(o, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    return __pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_6scorer_5__del__(o);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter[] = {

#line 11 "(tree fragment)"
  {"__getstate__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_3__getstate__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_2__getstate__},

#line 11 "(tree fragment)"
  {"__setstate__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_5__setstate__, METH_O, __pyx_doc_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_4__setstate__},

#line 11 "(tree fragment)"
  {"_calculate_scale_factor", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_7_calculate_scale_factor, METH_O, __pyx_doc_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_6_calculate_scale_factor},

#line 11 "(tree fragment)"
  {"evaluate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_9evaluate, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_8evaluate},

#line 11 "(tree fragment)"
  {"_evaluate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_11_evaluate, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_10_evaluate},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter[] = {

#line 11 "(tree fragment)"
  {(char *)"penalty_factor", __pyx_getprop_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_penalty_factor, __pyx_setprop_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_penalty_factor, (char *)0, 0},

#line 11 "(tree fragment)"
  {(char *)"scale_factor", __pyx_getprop_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_scale_factor, __pyx_setprop_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_scale_factor, (char *)"scale_factor: 'double'", 0},

#line 11 "(tree fragment)"
  {(char *)"scorer", __pyx_getprop_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_scorer, __pyx_setprop_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_scorer, (char *)"scorer: ms_deisotope._c.scoring.PenalizedMSDeconVFitter", 0},

#line 11 "(tree fragment)"
  {0, 0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter = {

#line 11 "(tree fragment)"
  PyVarObject_HEAD_INIT(0, 0)

#line 11 "(tree fragment)"
  "ms_deisotope._c.scoring.ScaledPenalizedMSDeconvFitter", /*tp_name*/

#line 11 "(tree fragment)"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter), /*tp_basicsize*/

#line 11 "(tree fragment)"
  0, /*tp_itemsize*/

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter, /*tp_dealloc*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX < 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_vectorcall_offset*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattr*/

#line 11 "(tree fragment)"
  0, /*tp_setattr*/

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*tp_compare*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  0, /*tp_as_async*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_25__repr__, /*tp_repr*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_repr*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_as_number*/

#line 11 "(tree fragment)"
  0, /*tp_as_sequence*/

#line 11 "(tree fragment)"
  0, /*tp_as_mapping*/

#line 11 "(tree fragment)"
  0, /*tp_hash*/

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_13__call__, /*tp_call*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_call*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_str*/

#line 11 "(tree fragment)"
  0, /*tp_getattro*/

#line 11 "(tree fragment)"
  0, /*tp_setattro*/

#line 11 "(tree fragment)"
  0, /*tp_as_buffer*/

#line 11 "(tree fragment)"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 11 "(tree fragment)"
  "ScaledPenalizedMSDeconvFitter(minimum_score=0.3, penalty_factor=1., mass_error_tolerance=0.02)", /*tp_doc*/

#line 11 "(tree fragment)"
  __pyx_tp_traverse_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter, /*tp_traverse*/

#line 11 "(tree fragment)"
  __pyx_tp_clear_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter, /*tp_clear*/

#line 11 "(tree fragment)"
  0, /*tp_richcompare*/

#line 11 "(tree fragment)"
  0, /*tp_weaklistoffset*/

#line 11 "(tree fragment)"
  0, /*tp_iter*/

#line 11 "(tree fragment)"
  0, /*tp_iternext*/

#line 11 "(tree fragment)"
  __pyx_methods_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter, /*tp_methods*/

#line 11 "(tree fragment)"
  0, /*tp_members*/

#line 11 "(tree fragment)"
  __pyx_getsets_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter, /*tp_getset*/

#line 11 "(tree fragment)"
  0, /*tp_base*/

#line 11 "(tree fragment)"
  0, /*tp_dict*/

#line 11 "(tree fragment)"
  0, /*tp_descr_get*/

#line 11 "(tree fragment)"
  0, /*tp_descr_set*/

#line 11 "(tree fragment)"
  0, /*tp_dictoffset*/

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_1__init__, /*tp_init*/

#line 11 "(tree fragment)"
  0, /*tp_alloc*/

#line 11 "(tree fragment)"
  __pyx_tp_new_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter, /*tp_new*/

#line 11 "(tree fragment)"
  0, /*tp_free*/

#line 11 "(tree fragment)"
  0, /*tp_is_gc*/

#line 11 "(tree fragment)"
  0, /*tp_bases*/

#line 11 "(tree fragment)"
  0, /*tp_mro*/

#line 11 "(tree fragment)"
  0, /*tp_cache*/

#line 11 "(tree fragment)"
  0, /*tp_subclasses*/

#line 11 "(tree fragment)"
  0, /*tp_weaklist*/

#line 11 "(tree fragment)"
  0, /*tp_del*/

#line 11 "(tree fragment)"
  0, /*tp_version_tag*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030400a1

#line 11 "(tree fragment)"
  0, /*tp_finalize*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 11 "(tree fragment)"
  0, /*tp_vectorcall*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_pypy_flags*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DotProductFitter __pyx_vtable_12ms_deisotope_2_c_7scoring_DotProductFitter;

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_DotProductFitter(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter *p;

#line 11 "(tree fragment)"
  PyObject *o = __pyx_tp_new_12ms_deisotope_2_c_7scoring_IsotopicFitterBase(t, a, k);

#line 11 "(tree fragment)"
  if (unlikely(!o)) return 0;

#line 11 "(tree fragment)"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter *)o);

#line 11 "(tree fragment)"
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase*)__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DotProductFitter;

#line 11 "(tree fragment)"
  return o;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_7scoring_DotProductFitter[] = {

#line 11 "(tree fragment)"
  {"_evaluate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_7scoring_16DotProductFitter_3_evaluate, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_7scoring_16DotProductFitter_2_evaluate},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_7scoring_DotProductFitter = {

#line 11 "(tree fragment)"
  PyVarObject_HEAD_INIT(0, 0)

#line 11 "(tree fragment)"
  "ms_deisotope._c.scoring.DotProductFitter", /*tp_name*/

#line 11 "(tree fragment)"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter), /*tp_basicsize*/

#line 11 "(tree fragment)"
  0, /*tp_itemsize*/

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_12ms_deisotope_2_c_7scoring_IsotopicFitterBase, /*tp_dealloc*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX < 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_vectorcall_offset*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattr*/

#line 11 "(tree fragment)"
  0, /*tp_setattr*/

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*tp_compare*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  0, /*tp_as_async*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_25__repr__, /*tp_repr*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_repr*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_as_number*/

#line 11 "(tree fragment)"
  0, /*tp_as_sequence*/

#line 11 "(tree fragment)"
  0, /*tp_as_mapping*/

#line 11 "(tree fragment)"
  0, /*tp_hash*/

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_13__call__, /*tp_call*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_call*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_str*/

#line 11 "(tree fragment)"
  0, /*tp_getattro*/

#line 11 "(tree fragment)"
  0, /*tp_setattro*/

#line 11 "(tree fragment)"
  0, /*tp_as_buffer*/

#line 11 "(tree fragment)"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 11 "(tree fragment)"
  "DotProductFitter(minimum_score=100)", /*tp_doc*/

#line 11 "(tree fragment)"
  __pyx_tp_traverse_12ms_deisotope_2_c_7scoring_IsotopicFitterBase, /*tp_traverse*/

#line 11 "(tree fragment)"
  __pyx_tp_clear_12ms_deisotope_2_c_7scoring_IsotopicFitterBase, /*tp_clear*/

#line 11 "(tree fragment)"
  0, /*tp_richcompare*/

#line 11 "(tree fragment)"
  0, /*tp_weaklistoffset*/

#line 11 "(tree fragment)"
  0, /*tp_iter*/

#line 11 "(tree fragment)"
  0, /*tp_iternext*/

#line 11 "(tree fragment)"
  __pyx_methods_12ms_deisotope_2_c_7scoring_DotProductFitter, /*tp_methods*/

#line 11 "(tree fragment)"
  0, /*tp_members*/

#line 11 "(tree fragment)"
  0, /*tp_getset*/

#line 11 "(tree fragment)"
  0, /*tp_base*/

#line 11 "(tree fragment)"
  0, /*tp_dict*/

#line 11 "(tree fragment)"
  0, /*tp_descr_get*/

#line 11 "(tree fragment)"
  0, /*tp_descr_set*/

#line 11 "(tree fragment)"
  0, /*tp_dictoffset*/

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_16DotProductFitter_1__init__, /*tp_init*/

#line 11 "(tree fragment)"
  0, /*tp_alloc*/

#line 11 "(tree fragment)"
  __pyx_tp_new_12ms_deisotope_2_c_7scoring_DotProductFitter, /*tp_new*/

#line 11 "(tree fragment)"
  0, /*tp_free*/

#line 11 "(tree fragment)"
  0, /*tp_is_gc*/

#line 11 "(tree fragment)"
  0, /*tp_bases*/

#line 11 "(tree fragment)"
  0, /*tp_mro*/

#line 11 "(tree fragment)"
  0, /*tp_cache*/

#line 11 "(tree fragment)"
  0, /*tp_subclasses*/

#line 11 "(tree fragment)"
  0, /*tp_weaklist*/

#line 11 "(tree fragment)"
  0, /*tp_del*/

#line 11 "(tree fragment)"
  0, /*tp_version_tag*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030400a1

#line 11 "(tree fragment)"
  0, /*tp_finalize*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 11 "(tree fragment)"
  0, /*tp_vectorcall*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_pypy_flags*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MinimizeFitSelector __pyx_vtable_12ms_deisotope_2_c_7scoring_MinimizeFitSelector;

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_MinimizeFitSelector(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_MinimizeFitSelector *p;

#line 11 "(tree fragment)"
  PyObject *o = __pyx_tp_new_12ms_deisotope_2_c_7scoring_FitSelectorBase(t, a, k);

#line 11 "(tree fragment)"
  if (unlikely(!o)) return 0;

#line 11 "(tree fragment)"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_MinimizeFitSelector *)o);

#line 11 "(tree fragment)"
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase*)__pyx_vtabptr_12ms_deisotope_2_c_7scoring_MinimizeFitSelector;

#line 11 "(tree fragment)"
  return o;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_7scoring_MinimizeFitSelector[] = {

#line 11 "(tree fragment)"
  {"best", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_1best, METH_O, __pyx_doc_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_best},

#line 11 "(tree fragment)"
  {"reject", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_3reject, METH_O, __pyx_doc_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_2reject},

#line 11 "(tree fragment)"
  {"reject_score", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_5reject_score, METH_O, __pyx_doc_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_4reject_score},

#line 11 "(tree fragment)"
  {"is_maximizing", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_7is_maximizing, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_6is_maximizing},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_7scoring_MinimizeFitSelector = {

#line 11 "(tree fragment)"
  PyVarObject_HEAD_INIT(0, 0)

#line 11 "(tree fragment)"
  "ms_deisotope._c.scoring.MinimizeFitSelector", /*tp_name*/

#line 11 "(tree fragment)"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MinimizeFitSelector), /*tp_basicsize*/

#line 11 "(tree fragment)"
  0, /*tp_itemsize*/

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_12ms_deisotope_2_c_7scoring_FitSelectorBase, /*tp_dealloc*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX < 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_vectorcall_offset*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattr*/

#line 11 "(tree fragment)"
  0, /*tp_setattr*/

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*tp_compare*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  0, /*tp_as_async*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_15__repr__, /*tp_repr*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_repr*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_as_number*/

#line 11 "(tree fragment)"
  0, /*tp_as_sequence*/

#line 11 "(tree fragment)"
  0, /*tp_as_mapping*/

#line 11 "(tree fragment)"
  0, /*tp_hash*/

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_7__call__, /*tp_call*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_call*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_str*/

#line 11 "(tree fragment)"
  0, /*tp_getattro*/

#line 11 "(tree fragment)"
  0, /*tp_setattro*/

#line 11 "(tree fragment)"
  0, /*tp_as_buffer*/

#line 11 "(tree fragment)"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/

#line 11 "(tree fragment)"
  "A FitSelector which tries to minimize the score of the best fit.\n    ", /*tp_doc*/

#line 11 "(tree fragment)"
  0, /*tp_traverse*/

#line 11 "(tree fragment)"
  0, /*tp_clear*/

#line 11 "(tree fragment)"
  0, /*tp_richcompare*/

#line 11 "(tree fragment)"
  0, /*tp_weaklistoffset*/

#line 11 "(tree fragment)"
  0, /*tp_iter*/

#line 11 "(tree fragment)"
  0, /*tp_iternext*/

#line 11 "(tree fragment)"
  __pyx_methods_12ms_deisotope_2_c_7scoring_MinimizeFitSelector, /*tp_methods*/

#line 11 "(tree fragment)"
  0, /*tp_members*/

#line 11 "(tree fragment)"
  0, /*tp_getset*/

#line 11 "(tree fragment)"
  0, /*tp_base*/

#line 11 "(tree fragment)"
  0, /*tp_dict*/

#line 11 "(tree fragment)"
  0, /*tp_descr_get*/

#line 11 "(tree fragment)"
  0, /*tp_descr_set*/

#line 11 "(tree fragment)"
  0, /*tp_dictoffset*/

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_1__init__, /*tp_init*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_init*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_alloc*/

#line 11 "(tree fragment)"
  __pyx_tp_new_12ms_deisotope_2_c_7scoring_MinimizeFitSelector, /*tp_new*/

#line 11 "(tree fragment)"
  0, /*tp_free*/

#line 11 "(tree fragment)"
  0, /*tp_is_gc*/

#line 11 "(tree fragment)"
  0, /*tp_bases*/

#line 11 "(tree fragment)"
  0, /*tp_mro*/

#line 11 "(tree fragment)"
  0, /*tp_cache*/

#line 11 "(tree fragment)"
  0, /*tp_subclasses*/

#line 11 "(tree fragment)"
  0, /*tp_weaklist*/

#line 11 "(tree fragment)"
  0, /*tp_del*/

#line 11 "(tree fragment)"
  0, /*tp_version_tag*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030400a1

#line 11 "(tree fragment)"
  0, /*tp_finalize*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 11 "(tree fragment)"
  0, /*tp_vectorcall*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_pypy_flags*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MaximizeFitSelector __pyx_vtable_12ms_deisotope_2_c_7scoring_MaximizeFitSelector;

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_MaximizeFitSelector(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_MaximizeFitSelector *p;

#line 11 "(tree fragment)"
  PyObject *o = __pyx_tp_new_12ms_deisotope_2_c_7scoring_FitSelectorBase(t, a, k);

#line 11 "(tree fragment)"
  if (unlikely(!o)) return 0;

#line 11 "(tree fragment)"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_MaximizeFitSelector *)o);

#line 11 "(tree fragment)"
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase*)__pyx_vtabptr_12ms_deisotope_2_c_7scoring_MaximizeFitSelector;

#line 11 "(tree fragment)"
  return o;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_7scoring_MaximizeFitSelector[] = {

#line 11 "(tree fragment)"
  {"best", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_1best, METH_O, __pyx_doc_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_best},

#line 11 "(tree fragment)"
  {"reject", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_3reject, METH_O, __pyx_doc_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_2reject},

#line 11 "(tree fragment)"
  {"reject_score", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_5reject_score, METH_O, __pyx_doc_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_4reject_score},

#line 11 "(tree fragment)"
  {"is_maximizing", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_7is_maximizing, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_6is_maximizing},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_7scoring_MaximizeFitSelector = {

#line 11 "(tree fragment)"
  PyVarObject_HEAD_INIT(0, 0)

#line 11 "(tree fragment)"
  "ms_deisotope._c.scoring.MaximizeFitSelector", /*tp_name*/

#line 11 "(tree fragment)"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MaximizeFitSelector), /*tp_basicsize*/

#line 11 "(tree fragment)"
  0, /*tp_itemsize*/

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_12ms_deisotope_2_c_7scoring_FitSelectorBase, /*tp_dealloc*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX < 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_vectorcall_offset*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattr*/

#line 11 "(tree fragment)"
  0, /*tp_setattr*/

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*tp_compare*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  0, /*tp_as_async*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_15__repr__, /*tp_repr*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_repr*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_as_number*/

#line 11 "(tree fragment)"
  0, /*tp_as_sequence*/

#line 11 "(tree fragment)"
  0, /*tp_as_mapping*/

#line 11 "(tree fragment)"
  0, /*tp_hash*/

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_7__call__, /*tp_call*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_call*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_str*/

#line 11 "(tree fragment)"
  0, /*tp_getattro*/

#line 11 "(tree fragment)"
  0, /*tp_setattro*/

#line 11 "(tree fragment)"
  0, /*tp_as_buffer*/

#line 11 "(tree fragment)"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/

#line 11 "(tree fragment)"
  "A FitSelector which tries to maximize the score of the best fit.\n    ", /*tp_doc*/

#line 11 "(tree fragment)"
  0, /*tp_traverse*/

#line 11 "(tree fragment)"
  0, /*tp_clear*/

#line 11 "(tree fragment)"
  0, /*tp_richcompare*/

#line 11 "(tree fragment)"
  0, /*tp_weaklistoffset*/

#line 11 "(tree fragment)"
  0, /*tp_iter*/

#line 11 "(tree fragment)"
  0, /*tp_iternext*/

#line 11 "(tree fragment)"
  __pyx_methods_12ms_deisotope_2_c_7scoring_MaximizeFitSelector, /*tp_methods*/

#line 11 "(tree fragment)"
  0, /*tp_members*/

#line 11 "(tree fragment)"
  0, /*tp_getset*/

#line 11 "(tree fragment)"
  0, /*tp_base*/

#line 11 "(tree fragment)"
  0, /*tp_dict*/

#line 11 "(tree fragment)"
  0, /*tp_descr_get*/

#line 11 "(tree fragment)"
  0, /*tp_descr_set*/

#line 11 "(tree fragment)"
  0, /*tp_dictoffset*/

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_15FitSelectorBase_1__init__, /*tp_init*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_init*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_alloc*/

#line 11 "(tree fragment)"
  __pyx_tp_new_12ms_deisotope_2_c_7scoring_MaximizeFitSelector, /*tp_new*/

#line 11 "(tree fragment)"
  0, /*tp_free*/

#line 11 "(tree fragment)"
  0, /*tp_is_gc*/

#line 11 "(tree fragment)"
  0, /*tp_bases*/

#line 11 "(tree fragment)"
  0, /*tp_mro*/

#line 11 "(tree fragment)"
  0, /*tp_cache*/

#line 11 "(tree fragment)"
  0, /*tp_subclasses*/

#line 11 "(tree fragment)"
  0, /*tp_weaklist*/

#line 11 "(tree fragment)"
  0, /*tp_del*/

#line 11 "(tree fragment)"
  0, /*tp_version_tag*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030400a1

#line 11 "(tree fragment)"
  0, /*tp_finalize*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 11 "(tree fragment)"
  0, /*tp_vectorcall*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_pypy_flags*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ChiSquareFitter __pyx_vtable_12ms_deisotope_2_c_7scoring_ChiSquareFitter;

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring_ChiSquareFitter(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_ChiSquareFitter *p;

#line 11 "(tree fragment)"
  PyObject *o = __pyx_tp_new_12ms_deisotope_2_c_7scoring_IsotopicFitterBase(t, a, k);

#line 11 "(tree fragment)"
  if (unlikely(!o)) return 0;

#line 11 "(tree fragment)"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_ChiSquareFitter *)o);

#line 11 "(tree fragment)"
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase*)__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ChiSquareFitter;

#line 11 "(tree fragment)"
  return o;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_7scoring_ChiSquareFitter[] = {

#line 11 "(tree fragment)"
  {"_evaluate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_7scoring_15ChiSquareFitter_1_evaluate, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_7scoring_15ChiSquareFitter__evaluate},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_7scoring_ChiSquareFitter = {

#line 11 "(tree fragment)"
  PyVarObject_HEAD_INIT(0, 0)

#line 11 "(tree fragment)"
  "ms_deisotope._c.scoring.ChiSquareFitter", /*tp_name*/

#line 11 "(tree fragment)"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ChiSquareFitter), /*tp_basicsize*/

#line 11 "(tree fragment)"
  0, /*tp_itemsize*/

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_12ms_deisotope_2_c_7scoring_IsotopicFitterBase, /*tp_dealloc*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX < 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_vectorcall_offset*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattr*/

#line 11 "(tree fragment)"
  0, /*tp_setattr*/

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*tp_compare*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  0, /*tp_as_async*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_25__repr__, /*tp_repr*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_repr*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_as_number*/

#line 11 "(tree fragment)"
  0, /*tp_as_sequence*/

#line 11 "(tree fragment)"
  0, /*tp_as_mapping*/

#line 11 "(tree fragment)"
  0, /*tp_hash*/

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_13__call__, /*tp_call*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_call*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_str*/

#line 11 "(tree fragment)"
  0, /*tp_getattro*/

#line 11 "(tree fragment)"
  0, /*tp_setattro*/

#line 11 "(tree fragment)"
  0, /*tp_as_buffer*/

#line 11 "(tree fragment)"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 11 "(tree fragment)"
  0, /*tp_doc*/

#line 11 "(tree fragment)"
  __pyx_tp_traverse_12ms_deisotope_2_c_7scoring_IsotopicFitterBase, /*tp_traverse*/

#line 11 "(tree fragment)"
  __pyx_tp_clear_12ms_deisotope_2_c_7scoring_IsotopicFitterBase, /*tp_clear*/

#line 11 "(tree fragment)"
  0, /*tp_richcompare*/

#line 11 "(tree fragment)"
  0, /*tp_weaklistoffset*/

#line 11 "(tree fragment)"
  0, /*tp_iter*/

#line 11 "(tree fragment)"
  0, /*tp_iternext*/

#line 11 "(tree fragment)"
  __pyx_methods_12ms_deisotope_2_c_7scoring_ChiSquareFitter, /*tp_methods*/

#line 11 "(tree fragment)"
  0, /*tp_members*/

#line 11 "(tree fragment)"
  0, /*tp_getset*/

#line 11 "(tree fragment)"
  0, /*tp_base*/

#line 11 "(tree fragment)"
  0, /*tp_dict*/

#line 11 "(tree fragment)"
  0, /*tp_descr_get*/

#line 11 "(tree fragment)"
  0, /*tp_descr_set*/

#line 11 "(tree fragment)"
  0, /*tp_dictoffset*/

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_pw_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_1__init__, /*tp_init*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_init*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_alloc*/

#line 11 "(tree fragment)"
  __pyx_tp_new_12ms_deisotope_2_c_7scoring_ChiSquareFitter, /*tp_new*/

#line 11 "(tree fragment)"
  0, /*tp_free*/

#line 11 "(tree fragment)"
  0, /*tp_is_gc*/

#line 11 "(tree fragment)"
  0, /*tp_bases*/

#line 11 "(tree fragment)"
  0, /*tp_mro*/

#line 11 "(tree fragment)"
  0, /*tp_cache*/

#line 11 "(tree fragment)"
  0, /*tp_subclasses*/

#line 11 "(tree fragment)"
  0, /*tp_weaklist*/

#line 11 "(tree fragment)"
  0, /*tp_del*/

#line 11 "(tree fragment)"
  0, /*tp_version_tag*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030400a1

#line 11 "(tree fragment)"
  0, /*tp_finalize*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 11 "(tree fragment)"
  0, /*tp_vectorcall*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_pypy_flags*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static struct __pyx_obj_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__ *__pyx_freelist_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__[8];

#line 11 "(tree fragment)"
static int __pyx_freecount_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__ = 0;

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {

#line 11 "(tree fragment)"
  PyObject *o;

#line 11 "(tree fragment)"
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__)))) {

#line 11 "(tree fragment)"
    o = (PyObject*)__pyx_freelist_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__[--__pyx_freecount_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__];

#line 11 "(tree fragment)"
    memset(o, 0, sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__));

#line 11 "(tree fragment)"
    (void) PyObject_INIT(o, t);

#line 11 "(tree fragment)"
    PyObject_GC_Track(o);

#line 11 "(tree fragment)"
  } else {

#line 11 "(tree fragment)"
    o = (*t->tp_alloc)(t, 0);

#line 11 "(tree fragment)"
    if (unlikely(!o)) return 0;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  return o;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__(PyObject *o) {

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__ *p = (struct __pyx_obj_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__ *)o;

#line 11 "(tree fragment)"
  PyObject_GC_UnTrack(o);

#line 11 "(tree fragment)"
  Py_CLEAR(p->__pyx_v_self);

#line 11 "(tree fragment)"
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__)))) {

#line 11 "(tree fragment)"
    __pyx_freelist_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__[__pyx_freecount_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__++] = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__ *)o);

#line 11 "(tree fragment)"
  } else {

#line 11 "(tree fragment)"
    (*Py_TYPE(o)->tp_free)(o);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_traverse_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__(PyObject *o, visitproc v, void *a) {

#line 11 "(tree fragment)"
  int e;

#line 11 "(tree fragment)"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__ *p = (struct __pyx_obj_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__ *)o;

#line 11 "(tree fragment)"
  if (p->__pyx_v_self) {

#line 11 "(tree fragment)"
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__ = {

#line 11 "(tree fragment)"
  PyVarObject_HEAD_INIT(0, 0)

#line 11 "(tree fragment)"
  "ms_deisotope._c.scoring.__pyx_scope_struct____iter__", /*tp_name*/

#line 11 "(tree fragment)"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__), /*tp_basicsize*/

#line 11 "(tree fragment)"
  0, /*tp_itemsize*/

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__, /*tp_dealloc*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX < 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_vectorcall_offset*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattr*/

#line 11 "(tree fragment)"
  0, /*tp_setattr*/

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*tp_compare*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  0, /*tp_as_async*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_repr*/

#line 11 "(tree fragment)"
  0, /*tp_as_number*/

#line 11 "(tree fragment)"
  0, /*tp_as_sequence*/

#line 11 "(tree fragment)"
  0, /*tp_as_mapping*/

#line 11 "(tree fragment)"
  0, /*tp_hash*/

#line 11 "(tree fragment)"
  0, /*tp_call*/

#line 11 "(tree fragment)"
  0, /*tp_str*/

#line 11 "(tree fragment)"
  0, /*tp_getattro*/

#line 11 "(tree fragment)"
  0, /*tp_setattro*/

#line 11 "(tree fragment)"
  0, /*tp_as_buffer*/

#line 11 "(tree fragment)"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 11 "(tree fragment)"
  0, /*tp_doc*/

#line 11 "(tree fragment)"
  __pyx_tp_traverse_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__, /*tp_traverse*/

#line 11 "(tree fragment)"
  0, /*tp_clear*/

#line 11 "(tree fragment)"
  0, /*tp_richcompare*/

#line 11 "(tree fragment)"
  0, /*tp_weaklistoffset*/

#line 11 "(tree fragment)"
  0, /*tp_iter*/

#line 11 "(tree fragment)"
  0, /*tp_iternext*/

#line 11 "(tree fragment)"
  0, /*tp_methods*/

#line 11 "(tree fragment)"
  0, /*tp_members*/

#line 11 "(tree fragment)"
  0, /*tp_getset*/

#line 11 "(tree fragment)"
  0, /*tp_base*/

#line 11 "(tree fragment)"
  0, /*tp_dict*/

#line 11 "(tree fragment)"
  0, /*tp_descr_get*/

#line 11 "(tree fragment)"
  0, /*tp_descr_set*/

#line 11 "(tree fragment)"
  0, /*tp_dictoffset*/

#line 11 "(tree fragment)"
  0, /*tp_init*/

#line 11 "(tree fragment)"
  0, /*tp_alloc*/

#line 11 "(tree fragment)"
  __pyx_tp_new_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__, /*tp_new*/

#line 11 "(tree fragment)"
  0, /*tp_free*/

#line 11 "(tree fragment)"
  0, /*tp_is_gc*/

#line 11 "(tree fragment)"
  0, /*tp_bases*/

#line 11 "(tree fragment)"
  0, /*tp_mro*/

#line 11 "(tree fragment)"
  0, /*tp_cache*/

#line 11 "(tree fragment)"
  0, /*tp_subclasses*/

#line 11 "(tree fragment)"
  0, /*tp_weaklist*/

#line 11 "(tree fragment)"
  0, /*tp_del*/

#line 11 "(tree fragment)"
  0, /*tp_version_tag*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030400a1

#line 11 "(tree fragment)"
  0, /*tp_finalize*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 11 "(tree fragment)"
  0, /*tp_vectorcall*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_pypy_flags*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"
static struct __pyx_vtabstruct_array __pyx_vtable_array;

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 11 "(tree fragment)"
  struct __pyx_array_obj *p;

#line 11 "(tree fragment)"
  PyObject *o;

#line 11 "(tree fragment)"
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {

#line 11 "(tree fragment)"
    o = (*t->tp_alloc)(t, 0);

#line 11 "(tree fragment)"
  } else {

#line 11 "(tree fragment)"
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  if (unlikely(!o)) return 0;

#line 11 "(tree fragment)"
  p = ((struct __pyx_array_obj *)o);

#line 11 "(tree fragment)"
  p->__pyx_vtab = __pyx_vtabptr_array;

#line 11 "(tree fragment)"
  p->mode = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  p->_format = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  if (unlikely(__pyx_array___cinit__(o, a, k) < 0)) goto bad;

#line 11 "(tree fragment)"
  return o;

#line 11 "(tree fragment)"
  bad:

#line 11 "(tree fragment)"
  Py_DECREF(o); o = 0;

#line 11 "(tree fragment)"
  return NULL;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static void __pyx_tp_dealloc_array(PyObject *o) {

#line 11 "(tree fragment)"
  struct __pyx_array_obj *p = (struct __pyx_array_obj *)o;

#line 11 "(tree fragment)"
  #if CYTHON_USE_TP_FINALIZE

#line 11 "(tree fragment)"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {

#line 11 "(tree fragment)"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  {

#line 11 "(tree fragment)"
    PyObject *etype, *eval, *etb;

#line 11 "(tree fragment)"
    PyErr_Fetch(&etype, &eval, &etb);

#line 11 "(tree fragment)"
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);

#line 11 "(tree fragment)"
    __pyx_array___dealloc__(o);

#line 11 "(tree fragment)"
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);

#line 11 "(tree fragment)"
    PyErr_Restore(etype, eval, etb);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  Py_CLEAR(p->mode);

#line 11 "(tree fragment)"
  Py_CLEAR(p->_format);

#line 11 "(tree fragment)"
  (*Py_TYPE(o)->tp_free)(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"
static PyObject *__pyx_sq_item_array(PyObject *o, Py_ssize_t i) {

#line 11 "(tree fragment)"
  PyObject *r;

#line 11 "(tree fragment)"
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;

#line 11 "(tree fragment)"
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);

#line 11 "(tree fragment)"
  Py_DECREF(x);

#line 11 "(tree fragment)"
  return r;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_mp_ass_subscript_array(PyObject *o, PyObject *i, PyObject *v) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_array___setitem__(o, i, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    PyErr_Format(PyExc_NotImplementedError,

#line 11 "(tree fragment)"
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);

#line 11 "(tree fragment)"
    return -1;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_getattro_array(PyObject *o, PyObject *n) {

#line 11 "(tree fragment)"
  PyObject *v = __Pyx_PyObject_GenericGetAttr(o, n);

#line 11 "(tree fragment)"
  if (!v && PyErr_ExceptionMatches(PyExc_AttributeError)) {

#line 11 "(tree fragment)"
    PyErr_Clear();

#line 11 "(tree fragment)"
    v = __pyx_array___getattr__(o, n);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  return v;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop___pyx_array_memview(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods_array[] = {

#line 11 "(tree fragment)"
  {"__getattr__", (PyCFunction)__pyx_array___getattr__, METH_O|METH_COEXIST, 0},

#line 11 "(tree fragment)"
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_array_1__reduce_cython__, METH_NOARGS, 0},

#line 11 "(tree fragment)"
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_array_3__setstate_cython__, METH_O, 0},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static struct PyGetSetDef __pyx_getsets_array[] = {

#line 11 "(tree fragment)"
  {(char *)"memview", __pyx_getprop___pyx_array_memview, 0, (char *)0, 0},

#line 11 "(tree fragment)"
  {0, 0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PySequenceMethods __pyx_tp_as_sequence_array = {

#line 11 "(tree fragment)"
  __pyx_array___len__, /*sq_length*/

#line 11 "(tree fragment)"
  0, /*sq_concat*/

#line 11 "(tree fragment)"
  0, /*sq_repeat*/

#line 11 "(tree fragment)"
  __pyx_sq_item_array, /*sq_item*/

#line 11 "(tree fragment)"
  0, /*sq_slice*/

#line 11 "(tree fragment)"
  0, /*sq_ass_item*/

#line 11 "(tree fragment)"
  0, /*sq_ass_slice*/

#line 11 "(tree fragment)"
  0, /*sq_contains*/

#line 11 "(tree fragment)"
  0, /*sq_inplace_concat*/

#line 11 "(tree fragment)"
  0, /*sq_inplace_repeat*/

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMappingMethods __pyx_tp_as_mapping_array = {

#line 11 "(tree fragment)"
  __pyx_array___len__, /*mp_length*/

#line 11 "(tree fragment)"
  __pyx_array___getitem__, /*mp_subscript*/

#line 11 "(tree fragment)"
  __pyx_mp_ass_subscript_array, /*mp_ass_subscript*/

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyBufferProcs __pyx_tp_as_buffer_array = {

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*bf_getreadbuffer*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*bf_getwritebuffer*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*bf_getsegcount*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*bf_getcharbuffer*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  __pyx_array_getbuffer, /*bf_getbuffer*/

#line 11 "(tree fragment)"
  0, /*bf_releasebuffer*/

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyTypeObject __pyx_type___pyx_array = {

#line 11 "(tree fragment)"
  PyVarObject_HEAD_INIT(0, 0)

#line 11 "(tree fragment)"
  "ms_deisotope._c.scoring.array", /*tp_name*/

#line 11 "(tree fragment)"
  sizeof(struct __pyx_array_obj), /*tp_basicsize*/

#line 11 "(tree fragment)"
  0, /*tp_itemsize*/

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_array, /*tp_dealloc*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX < 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_vectorcall_offset*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattr*/

#line 11 "(tree fragment)"
  0, /*tp_setattr*/

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*tp_compare*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  0, /*tp_as_async*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_repr*/

#line 11 "(tree fragment)"
  0, /*tp_as_number*/

#line 11 "(tree fragment)"
  &__pyx_tp_as_sequence_array, /*tp_as_sequence*/

#line 11 "(tree fragment)"
  &__pyx_tp_as_mapping_array, /*tp_as_mapping*/

#line 11 "(tree fragment)"
  0, /*tp_hash*/

#line 11 "(tree fragment)"
  0, /*tp_call*/

#line 11 "(tree fragment)"
  0, /*tp_str*/

#line 11 "(tree fragment)"
  __pyx_tp_getattro_array, /*tp_getattro*/

#line 11 "(tree fragment)"
  0, /*tp_setattro*/

#line 11 "(tree fragment)"
  &__pyx_tp_as_buffer_array, /*tp_as_buffer*/

#line 11 "(tree fragment)"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/

#line 11 "(tree fragment)"
  0, /*tp_doc*/

#line 11 "(tree fragment)"
  0, /*tp_traverse*/

#line 11 "(tree fragment)"
  0, /*tp_clear*/

#line 11 "(tree fragment)"
  0, /*tp_richcompare*/

#line 11 "(tree fragment)"
  0, /*tp_weaklistoffset*/

#line 11 "(tree fragment)"
  0, /*tp_iter*/

#line 11 "(tree fragment)"
  0, /*tp_iternext*/

#line 11 "(tree fragment)"
  __pyx_methods_array, /*tp_methods*/

#line 11 "(tree fragment)"
  0, /*tp_members*/

#line 11 "(tree fragment)"
  __pyx_getsets_array, /*tp_getset*/

#line 11 "(tree fragment)"
  0, /*tp_base*/

#line 11 "(tree fragment)"
  0, /*tp_dict*/

#line 11 "(tree fragment)"
  0, /*tp_descr_get*/

#line 11 "(tree fragment)"
  0, /*tp_descr_set*/

#line 11 "(tree fragment)"
  0, /*tp_dictoffset*/

#line 11 "(tree fragment)"
  0, /*tp_init*/

#line 11 "(tree fragment)"
  0, /*tp_alloc*/

#line 11 "(tree fragment)"
  __pyx_tp_new_array, /*tp_new*/

#line 11 "(tree fragment)"
  0, /*tp_free*/

#line 11 "(tree fragment)"
  0, /*tp_is_gc*/

#line 11 "(tree fragment)"
  0, /*tp_bases*/

#line 11 "(tree fragment)"
  0, /*tp_mro*/

#line 11 "(tree fragment)"
  0, /*tp_cache*/

#line 11 "(tree fragment)"
  0, /*tp_subclasses*/

#line 11 "(tree fragment)"
  0, /*tp_weaklist*/

#line 11 "(tree fragment)"
  0, /*tp_del*/

#line 11 "(tree fragment)"
  0, /*tp_version_tag*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030400a1

#line 11 "(tree fragment)"
  0, /*tp_finalize*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 11 "(tree fragment)"
  0, /*tp_vectorcall*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_pypy_flags*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {

#line 11 "(tree fragment)"
  struct __pyx_MemviewEnum_obj *p;

#line 11 "(tree fragment)"
  PyObject *o;

#line 11 "(tree fragment)"
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {

#line 11 "(tree fragment)"
    o = (*t->tp_alloc)(t, 0);

#line 11 "(tree fragment)"
  } else {

#line 11 "(tree fragment)"
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  if (unlikely(!o)) return 0;

#line 11 "(tree fragment)"
  p = ((struct __pyx_MemviewEnum_obj *)o);

#line 11 "(tree fragment)"
  p->name = Py_None; Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  return o;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static void __pyx_tp_dealloc_Enum(PyObject *o) {

#line 11 "(tree fragment)"
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;

#line 11 "(tree fragment)"
  #if CYTHON_USE_TP_FINALIZE

#line 11 "(tree fragment)"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {

#line 11 "(tree fragment)"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  PyObject_GC_UnTrack(o);

#line 11 "(tree fragment)"
  Py_CLEAR(p->name);

#line 11 "(tree fragment)"
  (*Py_TYPE(o)->tp_free)(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_traverse_Enum(PyObject *o, visitproc v, void *a) {

#line 11 "(tree fragment)"
  int e;

#line 11 "(tree fragment)"
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;

#line 11 "(tree fragment)"
  if (p->name) {

#line 11 "(tree fragment)"
    e = (*v)(p->name, a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_clear_Enum(PyObject *o) {

#line 11 "(tree fragment)"
  PyObject* tmp;

#line 11 "(tree fragment)"
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;

#line 11 "(tree fragment)"
  tmp = ((PyObject*)p->name);

#line 11 "(tree fragment)"
  p->name = Py_None; Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  Py_XDECREF(tmp);

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods_Enum[] = {

#line 11 "(tree fragment)"
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_MemviewEnum_1__reduce_cython__, METH_NOARGS, 0},

#line 11 "(tree fragment)"
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_MemviewEnum_3__setstate_cython__, METH_O, 0},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyTypeObject __pyx_type___pyx_MemviewEnum = {

#line 11 "(tree fragment)"
  PyVarObject_HEAD_INIT(0, 0)

#line 11 "(tree fragment)"
  "ms_deisotope._c.scoring.Enum", /*tp_name*/

#line 11 "(tree fragment)"
  sizeof(struct __pyx_MemviewEnum_obj), /*tp_basicsize*/

#line 11 "(tree fragment)"
  0, /*tp_itemsize*/

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_Enum, /*tp_dealloc*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX < 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_vectorcall_offset*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattr*/

#line 11 "(tree fragment)"
  0, /*tp_setattr*/

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*tp_compare*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  0, /*tp_as_async*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  __pyx_MemviewEnum___repr__, /*tp_repr*/

#line 11 "(tree fragment)"
  0, /*tp_as_number*/

#line 11 "(tree fragment)"
  0, /*tp_as_sequence*/

#line 11 "(tree fragment)"
  0, /*tp_as_mapping*/

#line 11 "(tree fragment)"
  0, /*tp_hash*/

#line 11 "(tree fragment)"
  0, /*tp_call*/

#line 11 "(tree fragment)"
  0, /*tp_str*/

#line 11 "(tree fragment)"
  0, /*tp_getattro*/

#line 11 "(tree fragment)"
  0, /*tp_setattro*/

#line 11 "(tree fragment)"
  0, /*tp_as_buffer*/

#line 11 "(tree fragment)"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 11 "(tree fragment)"
  0, /*tp_doc*/

#line 11 "(tree fragment)"
  __pyx_tp_traverse_Enum, /*tp_traverse*/

#line 11 "(tree fragment)"
  __pyx_tp_clear_Enum, /*tp_clear*/

#line 11 "(tree fragment)"
  0, /*tp_richcompare*/

#line 11 "(tree fragment)"
  0, /*tp_weaklistoffset*/

#line 11 "(tree fragment)"
  0, /*tp_iter*/

#line 11 "(tree fragment)"
  0, /*tp_iternext*/

#line 11 "(tree fragment)"
  __pyx_methods_Enum, /*tp_methods*/

#line 11 "(tree fragment)"
  0, /*tp_members*/

#line 11 "(tree fragment)"
  0, /*tp_getset*/

#line 11 "(tree fragment)"
  0, /*tp_base*/

#line 11 "(tree fragment)"
  0, /*tp_dict*/

#line 11 "(tree fragment)"
  0, /*tp_descr_get*/

#line 11 "(tree fragment)"
  0, /*tp_descr_set*/

#line 11 "(tree fragment)"
  0, /*tp_dictoffset*/

#line 11 "(tree fragment)"
  __pyx_MemviewEnum___init__, /*tp_init*/

#line 11 "(tree fragment)"
  0, /*tp_alloc*/

#line 11 "(tree fragment)"
  __pyx_tp_new_Enum, /*tp_new*/

#line 11 "(tree fragment)"
  0, /*tp_free*/

#line 11 "(tree fragment)"
  0, /*tp_is_gc*/

#line 11 "(tree fragment)"
  0, /*tp_bases*/

#line 11 "(tree fragment)"
  0, /*tp_mro*/

#line 11 "(tree fragment)"
  0, /*tp_cache*/

#line 11 "(tree fragment)"
  0, /*tp_subclasses*/

#line 11 "(tree fragment)"
  0, /*tp_weaklist*/

#line 11 "(tree fragment)"
  0, /*tp_del*/

#line 11 "(tree fragment)"
  0, /*tp_version_tag*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030400a1

#line 11 "(tree fragment)"
  0, /*tp_finalize*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 11 "(tree fragment)"
  0, /*tp_vectorcall*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_pypy_flags*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"
static struct __pyx_vtabstruct_memoryview __pyx_vtable_memoryview;

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 11 "(tree fragment)"
  struct __pyx_memoryview_obj *p;

#line 11 "(tree fragment)"
  PyObject *o;

#line 11 "(tree fragment)"
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {

#line 11 "(tree fragment)"
    o = (*t->tp_alloc)(t, 0);

#line 11 "(tree fragment)"
  } else {

#line 11 "(tree fragment)"
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  if (unlikely(!o)) return 0;

#line 11 "(tree fragment)"
  p = ((struct __pyx_memoryview_obj *)o);

#line 11 "(tree fragment)"
  p->__pyx_vtab = __pyx_vtabptr_memoryview;

#line 11 "(tree fragment)"
  p->obj = Py_None; Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  p->_size = Py_None; Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  p->_array_interface = Py_None; Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  p->view.obj = NULL;

#line 11 "(tree fragment)"
  if (unlikely(__pyx_memoryview___cinit__(o, a, k) < 0)) goto bad;

#line 11 "(tree fragment)"
  return o;

#line 11 "(tree fragment)"
  bad:

#line 11 "(tree fragment)"
  Py_DECREF(o); o = 0;

#line 11 "(tree fragment)"
  return NULL;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static void __pyx_tp_dealloc_memoryview(PyObject *o) {

#line 11 "(tree fragment)"
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;

#line 11 "(tree fragment)"
  #if CYTHON_USE_TP_FINALIZE

#line 11 "(tree fragment)"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {

#line 11 "(tree fragment)"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  PyObject_GC_UnTrack(o);

#line 11 "(tree fragment)"
  {

#line 11 "(tree fragment)"
    PyObject *etype, *eval, *etb;

#line 11 "(tree fragment)"
    PyErr_Fetch(&etype, &eval, &etb);

#line 11 "(tree fragment)"
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);

#line 11 "(tree fragment)"
    __pyx_memoryview___dealloc__(o);

#line 11 "(tree fragment)"
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);

#line 11 "(tree fragment)"
    PyErr_Restore(etype, eval, etb);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  Py_CLEAR(p->obj);

#line 11 "(tree fragment)"
  Py_CLEAR(p->_size);

#line 11 "(tree fragment)"
  Py_CLEAR(p->_array_interface);

#line 11 "(tree fragment)"
  (*Py_TYPE(o)->tp_free)(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_traverse_memoryview(PyObject *o, visitproc v, void *a) {

#line 11 "(tree fragment)"
  int e;

#line 11 "(tree fragment)"
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;

#line 11 "(tree fragment)"
  if (p->obj) {

#line 11 "(tree fragment)"
    e = (*v)(p->obj, a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  if (p->_size) {

#line 11 "(tree fragment)"
    e = (*v)(p->_size, a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  if (p->_array_interface) {

#line 11 "(tree fragment)"
    e = (*v)(p->_array_interface, a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  if (p->view.obj) {

#line 11 "(tree fragment)"
    e = (*v)(p->view.obj, a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_clear_memoryview(PyObject *o) {

#line 11 "(tree fragment)"
  PyObject* tmp;

#line 11 "(tree fragment)"
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;

#line 11 "(tree fragment)"
  tmp = ((PyObject*)p->obj);

#line 11 "(tree fragment)"
  p->obj = Py_None; Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  Py_XDECREF(tmp);

#line 11 "(tree fragment)"
  tmp = ((PyObject*)p->_size);

#line 11 "(tree fragment)"
  p->_size = Py_None; Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  Py_XDECREF(tmp);

#line 11 "(tree fragment)"
  tmp = ((PyObject*)p->_array_interface);

#line 11 "(tree fragment)"
  p->_array_interface = Py_None; Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  Py_XDECREF(tmp);

#line 11 "(tree fragment)"
  Py_CLEAR(p->view.obj);

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"
static PyObject *__pyx_sq_item_memoryview(PyObject *o, Py_ssize_t i) {

#line 11 "(tree fragment)"
  PyObject *r;

#line 11 "(tree fragment)"
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;

#line 11 "(tree fragment)"
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);

#line 11 "(tree fragment)"
  Py_DECREF(x);

#line 11 "(tree fragment)"
  return r;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_mp_ass_subscript_memoryview(PyObject *o, PyObject *i, PyObject *v) {

#line 11 "(tree fragment)"
  if (v) {

#line 11 "(tree fragment)"
    return __pyx_memoryview___setitem__(o, i, v);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  else {

#line 11 "(tree fragment)"
    PyErr_Format(PyExc_NotImplementedError,

#line 11 "(tree fragment)"
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);

#line 11 "(tree fragment)"
    return -1;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop___pyx_memoryview_T(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop___pyx_memoryview_base(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop___pyx_memoryview_shape(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop___pyx_memoryview_strides(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop___pyx_memoryview_suboffsets(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop___pyx_memoryview_ndim(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop___pyx_memoryview_itemsize(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop___pyx_memoryview_nbytes(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop___pyx_memoryview_size(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods_memoryview[] = {

#line 11 "(tree fragment)"
  {"is_c_contig", (PyCFunction)__pyx_memoryview_is_c_contig, METH_NOARGS, 0},

#line 11 "(tree fragment)"
  {"is_f_contig", (PyCFunction)__pyx_memoryview_is_f_contig, METH_NOARGS, 0},

#line 11 "(tree fragment)"
  {"copy", (PyCFunction)__pyx_memoryview_copy, METH_NOARGS, 0},

#line 11 "(tree fragment)"
  {"copy_fortran", (PyCFunction)__pyx_memoryview_copy_fortran, METH_NOARGS, 0},

#line 11 "(tree fragment)"
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_memoryview_1__reduce_cython__, METH_NOARGS, 0},

#line 11 "(tree fragment)"
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_memoryview_3__setstate_cython__, METH_O, 0},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static struct PyGetSetDef __pyx_getsets_memoryview[] = {

#line 11 "(tree fragment)"
  {(char *)"T", __pyx_getprop___pyx_memoryview_T, 0, (char *)0, 0},

#line 11 "(tree fragment)"
  {(char *)"base", __pyx_getprop___pyx_memoryview_base, 0, (char *)0, 0},

#line 11 "(tree fragment)"
  {(char *)"shape", __pyx_getprop___pyx_memoryview_shape, 0, (char *)0, 0},

#line 11 "(tree fragment)"
  {(char *)"strides", __pyx_getprop___pyx_memoryview_strides, 0, (char *)0, 0},

#line 11 "(tree fragment)"
  {(char *)"suboffsets", __pyx_getprop___pyx_memoryview_suboffsets, 0, (char *)0, 0},

#line 11 "(tree fragment)"
  {(char *)"ndim", __pyx_getprop___pyx_memoryview_ndim, 0, (char *)0, 0},

#line 11 "(tree fragment)"
  {(char *)"itemsize", __pyx_getprop___pyx_memoryview_itemsize, 0, (char *)0, 0},

#line 11 "(tree fragment)"
  {(char *)"nbytes", __pyx_getprop___pyx_memoryview_nbytes, 0, (char *)0, 0},

#line 11 "(tree fragment)"
  {(char *)"size", __pyx_getprop___pyx_memoryview_size, 0, (char *)0, 0},

#line 11 "(tree fragment)"
  {0, 0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PySequenceMethods __pyx_tp_as_sequence_memoryview = {

#line 11 "(tree fragment)"
  __pyx_memoryview___len__, /*sq_length*/

#line 11 "(tree fragment)"
  0, /*sq_concat*/

#line 11 "(tree fragment)"
  0, /*sq_repeat*/

#line 11 "(tree fragment)"
  __pyx_sq_item_memoryview, /*sq_item*/

#line 11 "(tree fragment)"
  0, /*sq_slice*/

#line 11 "(tree fragment)"
  0, /*sq_ass_item*/

#line 11 "(tree fragment)"
  0, /*sq_ass_slice*/

#line 11 "(tree fragment)"
  0, /*sq_contains*/

#line 11 "(tree fragment)"
  0, /*sq_inplace_concat*/

#line 11 "(tree fragment)"
  0, /*sq_inplace_repeat*/

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMappingMethods __pyx_tp_as_mapping_memoryview = {

#line 11 "(tree fragment)"
  __pyx_memoryview___len__, /*mp_length*/

#line 11 "(tree fragment)"
  __pyx_memoryview___getitem__, /*mp_subscript*/

#line 11 "(tree fragment)"
  __pyx_mp_ass_subscript_memoryview, /*mp_ass_subscript*/

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyBufferProcs __pyx_tp_as_buffer_memoryview = {

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*bf_getreadbuffer*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*bf_getwritebuffer*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*bf_getsegcount*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*bf_getcharbuffer*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  __pyx_memoryview_getbuffer, /*bf_getbuffer*/

#line 11 "(tree fragment)"
  0, /*bf_releasebuffer*/

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyTypeObject __pyx_type___pyx_memoryview = {

#line 11 "(tree fragment)"
  PyVarObject_HEAD_INIT(0, 0)

#line 11 "(tree fragment)"
  "ms_deisotope._c.scoring.memoryview", /*tp_name*/

#line 11 "(tree fragment)"
  sizeof(struct __pyx_memoryview_obj), /*tp_basicsize*/

#line 11 "(tree fragment)"
  0, /*tp_itemsize*/

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_memoryview, /*tp_dealloc*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX < 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_vectorcall_offset*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattr*/

#line 11 "(tree fragment)"
  0, /*tp_setattr*/

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*tp_compare*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  0, /*tp_as_async*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  __pyx_memoryview___repr__, /*tp_repr*/

#line 11 "(tree fragment)"
  0, /*tp_as_number*/

#line 11 "(tree fragment)"
  &__pyx_tp_as_sequence_memoryview, /*tp_as_sequence*/

#line 11 "(tree fragment)"
  &__pyx_tp_as_mapping_memoryview, /*tp_as_mapping*/

#line 11 "(tree fragment)"
  0, /*tp_hash*/

#line 11 "(tree fragment)"
  0, /*tp_call*/

#line 11 "(tree fragment)"
  __pyx_memoryview___str__, /*tp_str*/

#line 11 "(tree fragment)"
  0, /*tp_getattro*/

#line 11 "(tree fragment)"
  0, /*tp_setattro*/

#line 11 "(tree fragment)"
  &__pyx_tp_as_buffer_memoryview, /*tp_as_buffer*/

#line 11 "(tree fragment)"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 11 "(tree fragment)"
  0, /*tp_doc*/

#line 11 "(tree fragment)"
  __pyx_tp_traverse_memoryview, /*tp_traverse*/

#line 11 "(tree fragment)"
  __pyx_tp_clear_memoryview, /*tp_clear*/

#line 11 "(tree fragment)"
  0, /*tp_richcompare*/

#line 11 "(tree fragment)"
  0, /*tp_weaklistoffset*/

#line 11 "(tree fragment)"
  0, /*tp_iter*/

#line 11 "(tree fragment)"
  0, /*tp_iternext*/

#line 11 "(tree fragment)"
  __pyx_methods_memoryview, /*tp_methods*/

#line 11 "(tree fragment)"
  0, /*tp_members*/

#line 11 "(tree fragment)"
  __pyx_getsets_memoryview, /*tp_getset*/

#line 11 "(tree fragment)"
  0, /*tp_base*/

#line 11 "(tree fragment)"
  0, /*tp_dict*/

#line 11 "(tree fragment)"
  0, /*tp_descr_get*/

#line 11 "(tree fragment)"
  0, /*tp_descr_set*/

#line 11 "(tree fragment)"
  0, /*tp_dictoffset*/

#line 11 "(tree fragment)"
  0, /*tp_init*/

#line 11 "(tree fragment)"
  0, /*tp_alloc*/

#line 11 "(tree fragment)"
  __pyx_tp_new_memoryview, /*tp_new*/

#line 11 "(tree fragment)"
  0, /*tp_free*/

#line 11 "(tree fragment)"
  0, /*tp_is_gc*/

#line 11 "(tree fragment)"
  0, /*tp_bases*/

#line 11 "(tree fragment)"
  0, /*tp_mro*/

#line 11 "(tree fragment)"
  0, /*tp_cache*/

#line 11 "(tree fragment)"
  0, /*tp_subclasses*/

#line 11 "(tree fragment)"
  0, /*tp_weaklist*/

#line 11 "(tree fragment)"
  0, /*tp_del*/

#line 11 "(tree fragment)"
  0, /*tp_version_tag*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030400a1

#line 11 "(tree fragment)"
  0, /*tp_finalize*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 11 "(tree fragment)"
  0, /*tp_vectorcall*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_pypy_flags*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"
static struct __pyx_vtabstruct__memoryviewslice __pyx_vtable__memoryviewslice;

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 11 "(tree fragment)"
  struct __pyx_memoryviewslice_obj *p;

#line 11 "(tree fragment)"
  PyObject *o = __pyx_tp_new_memoryview(t, a, k);

#line 11 "(tree fragment)"
  if (unlikely(!o)) return 0;

#line 11 "(tree fragment)"
  p = ((struct __pyx_memoryviewslice_obj *)o);

#line 11 "(tree fragment)"
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_memoryview*)__pyx_vtabptr__memoryviewslice;

#line 11 "(tree fragment)"
  p->from_object = Py_None; Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  p->from_slice.memview = NULL;

#line 11 "(tree fragment)"
  return o;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static void __pyx_tp_dealloc__memoryviewslice(PyObject *o) {

#line 11 "(tree fragment)"
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;

#line 11 "(tree fragment)"
  #if CYTHON_USE_TP_FINALIZE

#line 11 "(tree fragment)"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {

#line 11 "(tree fragment)"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  PyObject_GC_UnTrack(o);

#line 11 "(tree fragment)"
  {

#line 11 "(tree fragment)"
    PyObject *etype, *eval, *etb;

#line 11 "(tree fragment)"
    PyErr_Fetch(&etype, &eval, &etb);

#line 11 "(tree fragment)"
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);

#line 11 "(tree fragment)"
    __pyx_memoryviewslice___dealloc__(o);

#line 11 "(tree fragment)"
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);

#line 11 "(tree fragment)"
    PyErr_Restore(etype, eval, etb);

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  Py_CLEAR(p->from_object);

#line 11 "(tree fragment)"
  PyObject_GC_Track(o);

#line 11 "(tree fragment)"
  __pyx_tp_dealloc_memoryview(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_traverse__memoryviewslice(PyObject *o, visitproc v, void *a) {

#line 11 "(tree fragment)"
  int e;

#line 11 "(tree fragment)"
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;

#line 11 "(tree fragment)"
  e = __pyx_tp_traverse_memoryview(o, v, a); if (e) return e;

#line 11 "(tree fragment)"
  if (p->from_object) {

#line 11 "(tree fragment)"
    e = (*v)(p->from_object, a); if (e) return e;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static int __pyx_tp_clear__memoryviewslice(PyObject *o) {

#line 11 "(tree fragment)"
  PyObject* tmp;

#line 11 "(tree fragment)"
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;

#line 11 "(tree fragment)"
  __pyx_tp_clear_memoryview(o);

#line 11 "(tree fragment)"
  tmp = ((PyObject*)p->from_object);

#line 11 "(tree fragment)"
  p->from_object = Py_None; Py_INCREF(Py_None);

#line 11 "(tree fragment)"
  Py_XDECREF(tmp);

#line 11 "(tree fragment)"
  __PYX_XDEC_MEMVIEW(&p->from_slice, 1);

#line 11 "(tree fragment)"
  return 0;

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_getprop___pyx_memoryviewslice_base(PyObject *o, CYTHON_UNUSED void *x) {

#line 11 "(tree fragment)"
  return __pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(o);

#line 11 "(tree fragment)"
}

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods__memoryviewslice[] = {

#line 11 "(tree fragment)"
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_memoryviewslice_1__reduce_cython__, METH_NOARGS, 0},

#line 11 "(tree fragment)"
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_memoryviewslice_3__setstate_cython__, METH_O, 0},

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static struct PyGetSetDef __pyx_getsets__memoryviewslice[] = {

#line 11 "(tree fragment)"
  {(char *)"base", __pyx_getprop___pyx_memoryviewslice_base, 0, (char *)0, 0},

#line 11 "(tree fragment)"
  {0, 0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyTypeObject __pyx_type___pyx_memoryviewslice = {

#line 11 "(tree fragment)"
  PyVarObject_HEAD_INIT(0, 0)

#line 11 "(tree fragment)"
  "ms_deisotope._c.scoring._memoryviewslice", /*tp_name*/

#line 11 "(tree fragment)"
  sizeof(struct __pyx_memoryviewslice_obj), /*tp_basicsize*/

#line 11 "(tree fragment)"
  0, /*tp_itemsize*/

#line 11 "(tree fragment)"
  __pyx_tp_dealloc__memoryviewslice, /*tp_dealloc*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX < 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4

#line 11 "(tree fragment)"
  0, /*tp_vectorcall_offset*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattr*/

#line 11 "(tree fragment)"
  0, /*tp_setattr*/

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION < 3

#line 11 "(tree fragment)"
  0, /*tp_compare*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  0, /*tp_as_async*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_memoryview___repr__, /*tp_repr*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_repr*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_as_number*/

#line 11 "(tree fragment)"
  0, /*tp_as_sequence*/

#line 11 "(tree fragment)"
  0, /*tp_as_mapping*/

#line 11 "(tree fragment)"
  0, /*tp_hash*/

#line 11 "(tree fragment)"
  0, /*tp_call*/

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY

#line 11 "(tree fragment)"
  __pyx_memoryview___str__, /*tp_str*/

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  0, /*tp_str*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  0, /*tp_getattro*/

#line 11 "(tree fragment)"
  0, /*tp_setattro*/

#line 11 "(tree fragment)"
  0, /*tp_as_buffer*/

#line 11 "(tree fragment)"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 11 "(tree fragment)"
  "Internal class for passing memoryview slices to Python", /*tp_doc*/

#line 11 "(tree fragment)"
  __pyx_tp_traverse__memoryviewslice, /*tp_traverse*/

#line 11 "(tree fragment)"
  __pyx_tp_clear__memoryviewslice, /*tp_clear*/

#line 11 "(tree fragment)"
  0, /*tp_richcompare*/

#line 11 "(tree fragment)"
  0, /*tp_weaklistoffset*/

#line 11 "(tree fragment)"
  0, /*tp_iter*/

#line 11 "(tree fragment)"
  0, /*tp_iternext*/

#line 11 "(tree fragment)"
  __pyx_methods__memoryviewslice, /*tp_methods*/

#line 11 "(tree fragment)"
  0, /*tp_members*/

#line 11 "(tree fragment)"
  __pyx_getsets__memoryviewslice, /*tp_getset*/

#line 11 "(tree fragment)"
  0, /*tp_base*/

#line 11 "(tree fragment)"
  0, /*tp_dict*/

#line 11 "(tree fragment)"
  0, /*tp_descr_get*/

#line 11 "(tree fragment)"
  0, /*tp_descr_set*/

#line 11 "(tree fragment)"
  0, /*tp_dictoffset*/

#line 11 "(tree fragment)"
  0, /*tp_init*/

#line 11 "(tree fragment)"
  0, /*tp_alloc*/

#line 11 "(tree fragment)"
  __pyx_tp_new__memoryviewslice, /*tp_new*/

#line 11 "(tree fragment)"
  0, /*tp_free*/

#line 11 "(tree fragment)"
  0, /*tp_is_gc*/

#line 11 "(tree fragment)"
  0, /*tp_bases*/

#line 11 "(tree fragment)"
  0, /*tp_mro*/

#line 11 "(tree fragment)"
  0, /*tp_cache*/

#line 11 "(tree fragment)"
  0, /*tp_subclasses*/

#line 11 "(tree fragment)"
  0, /*tp_weaklist*/

#line 11 "(tree fragment)"
  0, /*tp_del*/

#line 11 "(tree fragment)"
  0, /*tp_version_tag*/

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030400a1

#line 11 "(tree fragment)"
  0, /*tp_finalize*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 11 "(tree fragment)"
  0, /*tp_vectorcall*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_print*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 11 "(tree fragment)"
  0, /*tp_pypy_flags*/

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyMethodDef __pyx_methods[] = {

#line 11 "(tree fragment)"
  {0, 0, 0, 0}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
#if PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
#if CYTHON_PEP489_MULTI_PHASE_INIT

#line 11 "(tree fragment)"
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/

#line 11 "(tree fragment)"
static int __pyx_pymod_exec_scoring(PyObject* module); /*proto*/

#line 11 "(tree fragment)"
static PyModuleDef_Slot __pyx_moduledef_slots[] = {

#line 11 "(tree fragment)"
  {Py_mod_create, (void*)__pyx_pymod_create},

#line 11 "(tree fragment)"
  {Py_mod_exec, (void*)__pyx_pymod_exec_scoring},

#line 11 "(tree fragment)"
  {0, NULL}

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"
#endif

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static struct PyModuleDef __pyx_moduledef = {

#line 11 "(tree fragment)"
    PyModuleDef_HEAD_INIT,

#line 11 "(tree fragment)"
    "scoring",

#line 11 "(tree fragment)"
    0, /* m_doc */

#line 11 "(tree fragment)"
  #if CYTHON_PEP489_MULTI_PHASE_INIT

#line 11 "(tree fragment)"
    0, /* m_size */

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
    -1, /* m_size */

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
    __pyx_methods /* m_methods */,

#line 11 "(tree fragment)"
  #if CYTHON_PEP489_MULTI_PHASE_INIT

#line 11 "(tree fragment)"
    __pyx_moduledef_slots, /* m_slots */

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
    NULL, /* m_reload */

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
    NULL, /* m_traverse */

#line 11 "(tree fragment)"
    NULL, /* m_clear */

#line 11 "(tree fragment)"
    NULL /* m_free */

#line 11 "(tree fragment)"
};

#line 11 "(tree fragment)"
#endif

#line 11 "(tree fragment)"
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_ASCII, __pyx_k_ASCII, sizeof(__pyx_k_ASCII), 0, 0, 1, 1},
  {&__pyx_kp_s_Buffer_view_does_not_expose_stri, __pyx_k_Buffer_view_does_not_expose_stri, sizeof(__pyx_k_Buffer_view_does_not_expose_stri), 0, 0, 1, 0},
  {&__pyx_kp_s_Can_only_create_a_buffer_that_is, __pyx_k_Can_only_create_a_buffer_that_is, sizeof(__pyx_k_Can_only_create_a_buffer_that_is), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_assign_to_read_only_memor, __pyx_k_Cannot_assign_to_read_only_memor, sizeof(__pyx_k_Cannot_assign_to_read_only_memor), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_create_writable_memory_vi, __pyx_k_Cannot_create_writable_memory_vi, sizeof(__pyx_k_Cannot_create_writable_memory_vi), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_index_with_type_s, __pyx_k_Cannot_index_with_type_s, sizeof(__pyx_k_Cannot_index_with_type_s), 0, 0, 1, 0},
  {&__pyx_n_s_ChiSquareFitter, __pyx_k_ChiSquareFitter, sizeof(__pyx_k_ChiSquareFitter), 0, 0, 1, 1},
  {&__pyx_n_s_DistinctPatternFitter, __pyx_k_DistinctPatternFitter, sizeof(__pyx_k_DistinctPatternFitter), 0, 0, 1, 1},
  {&__pyx_n_s_DotProductFitter, __pyx_k_DotProductFitter, sizeof(__pyx_k_DotProductFitter), 0, 0, 1, 1},
  {&__pyx_n_s_Ellipsis, __pyx_k_Ellipsis, sizeof(__pyx_k_Ellipsis), 0, 0, 1, 1},
  {&__pyx_kp_s_Empty_shape_tuple_for_cython_arr, __pyx_k_Empty_shape_tuple_for_cython_arr, sizeof(__pyx_k_Empty_shape_tuple_for_cython_arr), 0, 0, 1, 0},
  {&__pyx_n_s_FitSelectorBase, __pyx_k_FitSelectorBase, sizeof(__pyx_k_FitSelectorBase), 0, 0, 1, 1},
  {&__pyx_n_s_FunctionScorer, __pyx_k_FunctionScorer, sizeof(__pyx_k_FunctionScorer), 0, 0, 1, 1},
  {&__pyx_n_s_GTestFitter, __pyx_k_GTestFitter, sizeof(__pyx_k_GTestFitter), 0, 0, 1, 1},
  {&__pyx_n_s_INFINITY, __pyx_k_INFINITY, sizeof(__pyx_k_INFINITY), 0, 0, 1, 1},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_k_Incompatible_checksums_0x_x_vs_0, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0), 0, 0, 1, 0},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_kp_s_Indirect_dimensions_not_supporte, __pyx_k_Indirect_dimensions_not_supporte, sizeof(__pyx_k_Indirect_dimensions_not_supporte), 0, 0, 1, 0},
  {&__pyx_n_s_InterferenceDetection, __pyx_k_InterferenceDetection, sizeof(__pyx_k_InterferenceDetection), 0, 0, 1, 1},
  {&__pyx_kp_s_Invalid_mode_expected_c_or_fortr, __pyx_k_Invalid_mode_expected_c_or_fortr, sizeof(__pyx_k_Invalid_mode_expected_c_or_fortr), 0, 0, 1, 0},
  {&__pyx_kp_s_Invalid_shape_in_axis_d_d, __pyx_k_Invalid_shape_in_axis_d_d, sizeof(__pyx_k_Invalid_shape_in_axis_d_d), 0, 0, 1, 0},
  {&__pyx_n_s_IsotopicFitRecord, __pyx_k_IsotopicFitRecord, sizeof(__pyx_k_IsotopicFitRecord), 0, 0, 1, 1},
  {&__pyx_n_s_IsotopicFitRecord___iter, __pyx_k_IsotopicFitRecord___iter, sizeof(__pyx_k_IsotopicFitRecord___iter), 0, 0, 1, 1},
  {&__pyx_kp_s_IsotopicFitRecord_score_0_5f_cha, __pyx_k_IsotopicFitRecord_score_0_5f_cha, sizeof(__pyx_k_IsotopicFitRecord_score_0_5f_cha), 0, 0, 1, 0},
  {&__pyx_n_s_IsotopicFitterBase, __pyx_k_IsotopicFitterBase, sizeof(__pyx_k_IsotopicFitterBase), 0, 0, 1, 1},
  {&__pyx_n_s_LeastSquaresFitter, __pyx_k_LeastSquaresFitter, sizeof(__pyx_k_LeastSquaresFitter), 0, 0, 1, 1},
  {&__pyx_n_s_MSDeconVFitter, __pyx_k_MSDeconVFitter, sizeof(__pyx_k_MSDeconVFitter), 0, 0, 1, 1},
  {&__pyx_n_s_MaximizeFitSelector, __pyx_k_MaximizeFitSelector, sizeof(__pyx_k_MaximizeFitSelector), 0, 0, 1, 1},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_k_MemoryView_of_r_at_0x_x, sizeof(__pyx_k_MemoryView_of_r_at_0x_x), 0, 0, 1, 0},
  {&__pyx_kp_s_MemoryView_of_r_object, __pyx_k_MemoryView_of_r_object, sizeof(__pyx_k_MemoryView_of_r_object), 0, 0, 1, 0},
  {&__pyx_n_s_MinimizeFitSelector, __pyx_k_MinimizeFitSelector, sizeof(__pyx_k_MinimizeFitSelector), 0, 0, 1, 1},
  {&__pyx_n_s_NotImplementedError, __pyx_k_NotImplementedError, sizeof(__pyx_k_NotImplementedError), 0, 0, 1, 1},
  {&__pyx_n_b_O, __pyx_k_O, sizeof(__pyx_k_O), 0, 0, 0, 1},
  {&__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_k_Out_of_bounds_on_buffer_access_a, sizeof(__pyx_k_Out_of_bounds_on_buffer_access_a), 0, 0, 1, 0},
  {&__pyx_n_s_PenalizedMSDeconVFitter, __pyx_k_PenalizedMSDeconVFitter, sizeof(__pyx_k_PenalizedMSDeconVFitter), 0, 0, 1, 1},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_ScaledGTestFitter, __pyx_k_ScaledGTestFitter, sizeof(__pyx_k_ScaledGTestFitter), 0, 0, 1, 1},
  {&__pyx_n_s_ScaledPenalizedMSDeconvFitter, __pyx_k_ScaledPenalizedMSDeconvFitter, sizeof(__pyx_k_ScaledPenalizedMSDeconvFitter), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_kp_s_Unable_to_convert_item_to_object, __pyx_k_Unable_to_convert_item_to_object, sizeof(__pyx_k_Unable_to_convert_item_to_object), 0, 0, 1, 0},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_View_MemoryView, __pyx_k_View_MemoryView, sizeof(__pyx_k_View_MemoryView), 0, 0, 1, 1},
  {&__pyx_n_s_allocate_buffer, __pyx_k_allocate_buffer, sizeof(__pyx_k_allocate_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_attrgetter, __pyx_k_attrgetter, sizeof(__pyx_k_attrgetter), 0, 0, 1, 1},
  {&__pyx_n_s_base, __pyx_k_base, sizeof(__pyx_k_base), 0, 0, 1, 1},
  {&__pyx_n_s_best, __pyx_k_best, sizeof(__pyx_k_best), 0, 0, 1, 1},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_n_u_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 1, 0, 1},
  {&__pyx_n_s_calculate_scale_factor, __pyx_k_calculate_scale_factor, sizeof(__pyx_k_calculate_scale_factor), 0, 0, 1, 1},
  {&__pyx_n_s_charge, __pyx_k_charge, sizeof(__pyx_k_charge), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_configure, __pyx_k_configure, sizeof(__pyx_k_configure), 0, 0, 1, 1},
  {&__pyx_kp_s_contiguous_and_direct, __pyx_k_contiguous_and_direct, sizeof(__pyx_k_contiguous_and_direct), 0, 0, 1, 0},
  {&__pyx_kp_s_contiguous_and_indirect, __pyx_k_contiguous_and_indirect, sizeof(__pyx_k_contiguous_and_indirect), 0, 0, 1, 0},
  {&__pyx_n_s_data, __pyx_k_data, sizeof(__pyx_k_data), 0, 0, 1, 1},
  {&__pyx_n_s_deconvoluter, __pyx_k_deconvoluter, sizeof(__pyx_k_deconvoluter), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_domain_scale, __pyx_k_domain_scale, sizeof(__pyx_k_domain_scale), 0, 0, 1, 1},
  {&__pyx_n_s_dtype_is_object, __pyx_k_dtype_is_object, sizeof(__pyx_k_dtype_is_object), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_eq, __pyx_k_eq, sizeof(__pyx_k_eq), 0, 0, 1, 1},
  {&__pyx_n_s_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
  {&__pyx_n_s_evaluate, __pyx_k_evaluate, sizeof(__pyx_k_evaluate), 0, 0, 1, 1},
  {&__pyx_n_s_evaluate_2, __pyx_k_evaluate_2, sizeof(__pyx_k_evaluate_2), 0, 0, 1, 1},
  {&__pyx_n_s_expected, __pyx_k_expected, sizeof(__pyx_k_expected), 0, 0, 1, 1},
  {&__pyx_n_s_experimental, __pyx_k_experimental, sizeof(__pyx_k_experimental), 0, 0, 1, 1},
  {&__pyx_n_s_experimental_peaks, __pyx_k_experimental_peaks, sizeof(__pyx_k_experimental_peaks), 0, 0, 1, 1},
  {&__pyx_n_s_fields, __pyx_k_fields, sizeof(__pyx_k_fields), 0, 0, 1, 1},
  {&__pyx_n_s_flags, __pyx_k_flags, sizeof(__pyx_k_flags), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 0, 1, 1},
  {&__pyx_n_u_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 1, 0, 1},
  {&__pyx_n_s_function, __pyx_k_function, sizeof(__pyx_k_function), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_kp_s_got_differing_extents_in_dimensi, __pyx_k_got_differing_extents_in_dimensi, sizeof(__pyx_k_got_differing_extents_in_dimensi), 0, 0, 1, 0},
  {&__pyx_n_s_gt, __pyx_k_gt, sizeof(__pyx_k_gt), 0, 0, 1, 1},
  {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_is_maximizing, __pyx_k_is_maximizing, sizeof(__pyx_k_is_maximizing), 0, 0, 1, 1},
  {&__pyx_n_s_itemsize, __pyx_k_itemsize, sizeof(__pyx_k_itemsize), 0, 0, 1, 1},
  {&__pyx_kp_s_itemsize_0_for_cython_array, __pyx_k_itemsize_0_for_cython_array, sizeof(__pyx_k_itemsize_0_for_cython_array), 0, 0, 1, 0},
  {&__pyx_n_s_iter, __pyx_k_iter, sizeof(__pyx_k_iter), 0, 0, 1, 1},
  {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
  {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
  {&__pyx_n_s_lower, __pyx_k_lower, sizeof(__pyx_k_lower), 0, 0, 1, 1},
  {&__pyx_n_s_lt, __pyx_k_lt, sizeof(__pyx_k_lt), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_mass_error_tolerance, __pyx_k_mass_error_tolerance, sizeof(__pyx_k_mass_error_tolerance), 0, 0, 1, 1},
  {&__pyx_n_s_max, __pyx_k_max, sizeof(__pyx_k_max), 0, 0, 1, 1},
  {&__pyx_n_s_memview, __pyx_k_memview, sizeof(__pyx_k_memview), 0, 0, 1, 1},
  {&__pyx_n_s_min, __pyx_k_min, sizeof(__pyx_k_min), 0, 0, 1, 1},
  {&__pyx_n_s_minimum_score, __pyx_k_minimum_score, sizeof(__pyx_k_minimum_score), 0, 0, 1, 1},
  {&__pyx_n_s_missed_peaks, __pyx_k_missed_peaks, sizeof(__pyx_k_missed_peaks), 0, 0, 1, 1},
  {&__pyx_n_s_mode, __pyx_k_mode, sizeof(__pyx_k_mode), 0, 0, 1, 1},
  {&__pyx_n_s_ms_deisotope__c_scoring, __pyx_k_ms_deisotope__c_scoring, sizeof(__pyx_k_ms_deisotope__c_scoring), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_ndim, __pyx_k_ndim, sizeof(__pyx_k_ndim), 0, 0, 1, 1},
  {&__pyx_n_s_ne, __pyx_k_ne, sizeof(__pyx_k_ne), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_npeaks, __pyx_k_npeaks, sizeof(__pyx_k_npeaks), 0, 0, 1, 1},
  {&__pyx_kp_s_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 0, 1, 0},
  {&__pyx_kp_s_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 0, 1, 0},
  {&__pyx_n_s_obj, __pyx_k_obj, sizeof(__pyx_k_obj), 0, 0, 1, 1},
  {&__pyx_n_s_observed, __pyx_k_observed, sizeof(__pyx_k_observed), 0, 0, 1, 1},
  {&__pyx_n_s_operator, __pyx_k_operator, sizeof(__pyx_k_operator), 0, 0, 1, 1},
  {&__pyx_n_s_pack, __pyx_k_pack, sizeof(__pyx_k_pack), 0, 0, 1, 1},
  {&__pyx_n_s_peak_count_scale, __pyx_k_peak_count_scale, sizeof(__pyx_k_peak_count_scale), 0, 0, 1, 1},
  {&__pyx_n_s_peaklist, __pyx_k_peaklist, sizeof(__pyx_k_peaklist), 0, 0, 1, 1},
  {&__pyx_n_s_penalty_factor, __pyx_k_penalty_factor, sizeof(__pyx_k_penalty_factor), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_capi, __pyx_k_pyx_capi, sizeof(__pyx_k_pyx_capi), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_getbuffer, __pyx_k_pyx_getbuffer, sizeof(__pyx_k_pyx_getbuffer), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Enum, __pyx_k_pyx_unpickle_Enum, sizeof(__pyx_k_pyx_unpickle_Enum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_reject, __pyx_k_reject, sizeof(__pyx_k_reject), 0, 0, 1, 1},
  {&__pyx_n_s_reject_score, __pyx_k_reject_score, sizeof(__pyx_k_reject_score), 0, 0, 1, 1},
  {&__pyx_n_s_score, __pyx_k_score, sizeof(__pyx_k_score), 0, 0, 1, 1},
  {&__pyx_n_s_score_threshold, __pyx_k_score_threshold, sizeof(__pyx_k_score_threshold), 0, 0, 1, 1},
  {&__pyx_n_s_seed_peak, __pyx_k_seed_peak, sizeof(__pyx_k_seed_peak), 0, 0, 1, 1},
  {&__pyx_n_s_selector_type, __pyx_k_selector_type, sizeof(__pyx_k_selector_type), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_kp_s_self___class_____name___fields, __pyx_k_self___class_____name___fields, sizeof(__pyx_k_self___class_____name___fields), 0, 0, 1, 0},
  {&__pyx_kp_s_self___class_____name___minimum, __pyx_k_self___class_____name___minimum, sizeof(__pyx_k_self___class_____name___minimum), 0, 0, 1, 0},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_step, __pyx_k_step, sizeof(__pyx_k_step), 0, 0, 1, 1},
  {&__pyx_n_s_stop, __pyx_k_stop, sizeof(__pyx_k_stop), 0, 0, 1, 1},
  {&__pyx_kp_s_strided_and_direct, __pyx_k_strided_and_direct, sizeof(__pyx_k_strided_and_direct), 0, 0, 1, 0},
  {&__pyx_kp_s_strided_and_direct_or_indirect, __pyx_k_strided_and_direct_or_indirect, sizeof(__pyx_k_strided_and_direct_or_indirect), 0, 0, 1, 0},
  {&__pyx_kp_s_strided_and_indirect, __pyx_k_strided_and_indirect, sizeof(__pyx_k_strided_and_indirect), 0, 0, 1, 0},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_struct, __pyx_k_struct, sizeof(__pyx_k_struct), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_theoretical, __pyx_k_theoretical, sizeof(__pyx_k_theoretical), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_kp_s_unable_to_allocate_array_data, __pyx_k_unable_to_allocate_array_data, sizeof(__pyx_k_unable_to_allocate_array_data), 0, 0, 1, 0},
  {&__pyx_kp_s_unable_to_allocate_shape_and_str, __pyx_k_unable_to_allocate_shape_and_str, sizeof(__pyx_k_unable_to_allocate_shape_and_str), 0, 0, 1, 0},
  {&__pyx_n_s_unpack, __pyx_k_unpack, sizeof(__pyx_k_unpack), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_upper, __pyx_k_upper, sizeof(__pyx_k_upper), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 102, __pyx_L1_error)
  __pyx_builtin_NotImplementedError = __Pyx_GetBuiltinName(__pyx_n_s_NotImplementedError); if (!__pyx_builtin_NotImplementedError) __PYX_ERR(0, 178, __pyx_L1_error)
  __pyx_builtin_min = __Pyx_GetBuiltinName(__pyx_n_s_min); if (!__pyx_builtin_min) __PYX_ERR(0, 217, __pyx_L1_error)
  __pyx_builtin_max = __Pyx_GetBuiltinName(__pyx_n_s_max); if (!__pyx_builtin_max) __PYX_ERR(0, 280, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(2, 944, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(3, 134, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(3, 149, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(3, 152, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(3, 2, __pyx_L1_error)
  __pyx_builtin_Ellipsis = __Pyx_GetBuiltinName(__pyx_n_s_Ellipsis); if (!__pyx_builtin_Ellipsis) __PYX_ERR(3, 406, __pyx_L1_error)
  __pyx_builtin_id = __Pyx_GetBuiltinName(__pyx_n_s_id); if (!__pyx_builtin_id) __PYX_ERR(3, 615, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(3, 834, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "ms_deisotope/_c/scoring.pyx":339
 * 
 *     def __reduce__(self):
 *         return self.__class__, (0,), self.__getstate__()             # <<<<<<<<<<<<<<
 * 
 *     def __getstate__(self):
 */

#line 339 "ms_deisotope/_c/scoring.pyx"
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_int_0); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 339, __pyx_L1_error)

#line 339 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_tuple_);

#line 339 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "ms_deisotope/_c/scoring.pyx":806
 * 
 *     def __reduce__(self):
 *         return self.__class__, (None,), self.__getstate__()             # <<<<<<<<<<<<<<
 * 
 *     cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):
 */

#line 806 "ms_deisotope/_c/scoring.pyx"
  __pyx_tuple__3 = PyTuple_Pack(1, Py_None); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 806, __pyx_L1_error)

#line 806 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_tuple__3);

#line 806 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":944
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(2, 944, __pyx_L1_error)

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_tuple__4);

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":950
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(2, 950, __pyx_L1_error)

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_tuple__5);

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "View.MemoryView":134
 * 
 *         if not self.ndim:
 *             raise ValueError("Empty shape tuple for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if itemsize <= 0:
 */

#line 134 "View.MemoryView"
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_Empty_shape_tuple_for_cython_arr); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(3, 134, __pyx_L1_error)

#line 134 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__6);

#line 134 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "View.MemoryView":137
 * 
 *         if itemsize <= 0:
 *             raise ValueError("itemsize <= 0 for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(format, bytes):
 */

#line 137 "View.MemoryView"
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_itemsize_0_for_cython_array); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(3, 137, __pyx_L1_error)

#line 137 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__7);

#line 137 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "View.MemoryView":149
 * 
 *         if not self._shape:
 *             raise MemoryError("unable to allocate shape and strides.")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 149 "View.MemoryView"
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_s_unable_to_allocate_shape_and_str); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(3, 149, __pyx_L1_error)

#line 149 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__8);

#line 149 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "View.MemoryView":177
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")             # <<<<<<<<<<<<<<
 * 
 *             if self.dtype_is_object:
 */

#line 177 "View.MemoryView"
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_s_unable_to_allocate_array_data); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(3, 177, __pyx_L1_error)

#line 177 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__9);

#line 177 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "View.MemoryView":193
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")             # <<<<<<<<<<<<<<
 *         info.buf = self.data
 *         info.len = self.len
 */

#line 193 "View.MemoryView"
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_s_Can_only_create_a_buffer_that_is); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(3, 193, __pyx_L1_error)

#line 193 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__10);

#line 193 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 2 "(tree fragment)"
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(3, 2, __pyx_L1_error)

#line 2 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__11);

#line 2 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */

#line 4 "(tree fragment)"
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(3, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__12);

#line 4 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "View.MemoryView":420
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 */

#line 420 "View.MemoryView"
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_s_Cannot_assign_to_read_only_memor); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(3, 420, __pyx_L1_error)

#line 420 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__13);

#line 420 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "View.MemoryView":497
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")             # <<<<<<<<<<<<<<
 *         else:
 *             if len(self.view.format) == 1:
 */

#line 497 "View.MemoryView"
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_s_Unable_to_convert_item_to_object); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(3, 497, __pyx_L1_error)

#line 497 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__14);

#line 497 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "View.MemoryView":522
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_ND:
 */

#line 522 "View.MemoryView"
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_s_Cannot_create_writable_memory_vi); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(3, 522, __pyx_L1_error)

#line 522 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__15);

#line 522 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "View.MemoryView":572
 *         if self.view.strides == NULL:
 * 
 *             raise ValueError("Buffer view does not expose strides")             # <<<<<<<<<<<<<<
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 */

#line 572 "View.MemoryView"
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_Buffer_view_does_not_expose_stri); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(3, 572, __pyx_L1_error)

#line 572 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__16);

#line 572 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "View.MemoryView":579
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */

#line 579 "View.MemoryView"
  __pyx_tuple__17 = PyTuple_New(1); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(3, 579, __pyx_L1_error)

#line 579 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__17);

#line 579 "View.MemoryView"
  __Pyx_INCREF(__pyx_int_neg_1);

#line 579 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_int_neg_1);

#line 579 "View.MemoryView"
  PyTuple_SET_ITEM(__pyx_tuple__17, 0, __pyx_int_neg_1);

#line 579 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 2 "(tree fragment)"
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(3, 2, __pyx_L1_error)

#line 2 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__18);

#line 2 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */

#line 4 "(tree fragment)"
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(3, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__19);

#line 4 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "View.MemoryView":684
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))             # <<<<<<<<<<<<<<
 *                 seen_ellipsis = True
 *             else:
 */

#line 684 "View.MemoryView"
  __pyx_slice__20 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__20)) __PYX_ERR(3, 684, __pyx_L1_error)

#line 684 "View.MemoryView"
  __Pyx_GOTREF(__pyx_slice__20);

#line 684 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_slice__20);

  /* "View.MemoryView":705
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 705 "View.MemoryView"
  __pyx_tuple__21 = PyTuple_Pack(1, __pyx_kp_s_Indirect_dimensions_not_supporte); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(3, 705, __pyx_L1_error)

#line 705 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__21);

#line 705 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

#line 2 "(tree fragment)"
  __pyx_tuple__22 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(3, 2, __pyx_L1_error)

#line 2 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__22);

#line 2 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__22);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */

#line 4 "(tree fragment)"
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(3, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__23);

#line 4 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__23);

#line 4 "(tree fragment)"
  __pyx_tuple__24 = PyTuple_Pack(3, __pyx_int_184977713, __pyx_int_136983863, __pyx_int_112105877); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(3, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__24);

#line 4 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__24);

  /* "View.MemoryView":287
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */

#line 287 "View.MemoryView"
  __pyx_tuple__25 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct_or_indirect); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(3, 287, __pyx_L1_error)

#line 287 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__25);

#line 287 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__25);

  /* "View.MemoryView":288
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */

#line 288 "View.MemoryView"
  __pyx_tuple__26 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(3, 288, __pyx_L1_error)

#line 288 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__26);

#line 288 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "View.MemoryView":289
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 289 "View.MemoryView"
  __pyx_tuple__27 = PyTuple_Pack(1, __pyx_kp_s_strided_and_indirect); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(3, 289, __pyx_L1_error)

#line 289 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__27);

#line 289 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__27);

  /* "View.MemoryView":292
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */

#line 292 "View.MemoryView"
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_direct); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(3, 292, __pyx_L1_error)

#line 292 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__28);

#line 292 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "View.MemoryView":293
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 293 "View.MemoryView"
  __pyx_tuple__29 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_indirect); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(3, 293, __pyx_L1_error)

#line 293 "View.MemoryView"
  __Pyx_GOTREF(__pyx_tuple__29);

#line 293 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_tuple__29);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"
  __pyx_tuple__30 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(3, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__30);

#line 1 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__30);

#line 1 "(tree fragment)"
  __pyx_codeobj__31 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__30, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Enum, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__31)) __PYX_ERR(3, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return 0;

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return -1;

#line 1 "(tree fragment)"
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_1_ = PyFloat_FromDouble(1.); if (unlikely(!__pyx_float_1_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_0_3 = PyFloat_FromDouble(0.3); if (unlikely(!__pyx_float_0_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_0_5 = PyFloat_FromDouble(0.5); if (unlikely(!__pyx_float_0_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_1_5 = PyFloat_FromDouble(1.5); if (unlikely(!__pyx_float_1_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_0_02 = PyFloat_FromDouble(0.02); if (unlikely(!__pyx_float_0_02)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_100_ = PyFloat_FromDouble(100.); if (unlikely(!__pyx_float_100_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_10 = PyInt_FromLong(10); if (unlikely(!__pyx_int_10)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_100 = PyInt_FromLong(100); if (unlikely(!__pyx_int_100)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_112105877 = PyInt_FromLong(112105877L); if (unlikely(!__pyx_int_112105877)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_136983863 = PyInt_FromLong(136983863L); if (unlikely(!__pyx_int_136983863)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_184977713 = PyInt_FromLong(184977713L); if (unlikely(!__pyx_int_184977713)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_12ms_deisotope_2_c_7scoring_g_test_scaled = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *)Py_None); Py_INCREF(Py_None);
  __pyx_v_12ms_deisotope_2_c_7scoring_least_squares = ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter *)Py_None); Py_INCREF(Py_None);
  generic = Py_None; Py_INCREF(Py_None);
  strided = Py_None; Py_INCREF(Py_None);
  indirect = Py_None; Py_INCREF(Py_None);
  contiguous = Py_None; Py_INCREF(Py_None);
  indirect_contiguous = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  if (__Pyx_ExportVoidPtr(__pyx_n_s_INFINITY, (void *)&__pyx_v_12ms_deisotope_2_c_7scoring_INFINITY, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitRecord = &__pyx_vtable_12ms_deisotope_2_c_7scoring_IsotopicFitRecord;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_IsotopicFitRecord._create = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, double, int, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_IsotopicFitRecord._eq = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__eq;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_IsotopicFitRecord._ne = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__ne;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_IsotopicFitRecord._lt = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__lt;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_IsotopicFitRecord._gt = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__gt;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_7scoring_IsotopicFitRecord) < 0) __PYX_ERR(0, 22, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_7scoring_IsotopicFitRecord.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_7scoring_IsotopicFitRecord.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_7scoring_IsotopicFitRecord.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_7scoring_IsotopicFitRecord.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_7scoring_IsotopicFitRecord.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitRecord) < 0) __PYX_ERR(0, 22, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_IsotopicFitRecord, (PyObject *)&__pyx_type_12ms_deisotope_2_c_7scoring_IsotopicFitRecord) < 0) __PYX_ERR(0, 22, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord = &__pyx_type_12ms_deisotope_2_c_7scoring_IsotopicFitRecord;
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_FitSelectorBase = &__pyx_vtable_12ms_deisotope_2_c_7scoring_FitSelectorBase;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_FitSelectorBase.best = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_15FitSelectorBase_best;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_FitSelectorBase._best_from_set = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, PyObject *))__pyx_f_12ms_deisotope_2_c_7scoring_15FitSelectorBase__best_from_set;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_FitSelectorBase.reject = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_15FitSelectorBase_reject;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_FitSelectorBase.reject_score = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, double, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_15FitSelectorBase_reject_score;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_FitSelectorBase.is_maximizing = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_15FitSelectorBase_is_maximizing;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_7scoring_FitSelectorBase) < 0) __PYX_ERR(0, 161, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_7scoring_FitSelectorBase.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_7scoring_FitSelectorBase.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_7scoring_FitSelectorBase.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_7scoring_FitSelectorBase.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_7scoring_FitSelectorBase.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_7scoring_FitSelectorBase) < 0) __PYX_ERR(0, 161, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_FitSelectorBase, (PyObject *)&__pyx_type_12ms_deisotope_2_c_7scoring_FitSelectorBase) < 0) __PYX_ERR(0, 161, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase = &__pyx_type_12ms_deisotope_2_c_7scoring_FitSelectorBase;
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase = &__pyx_vtable_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_IsotopicFitterBase._evaluate = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, PyObject *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase__evaluate;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_IsotopicFitterBase.reject = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_reject;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_IsotopicFitterBase.reject_score = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, double, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_reject_score;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_IsotopicFitterBase.is_maximizing = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_is_maximizing;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_IsotopicFitterBase._configure = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *(*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase__configure;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_7scoring_IsotopicFitterBase) < 0) __PYX_ERR(0, 323, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_7scoring_IsotopicFitterBase.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_7scoring_IsotopicFitterBase.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_7scoring_IsotopicFitterBase.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_7scoring_IsotopicFitterBase.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)&__pyx_type_12ms_deisotope_2_c_7scoring_IsotopicFitterBase, "__call__"); if (unlikely(!wrapper)) __PYX_ERR(0, 323, __pyx_L1_error)
    if (Py_TYPE(wrapper) == &PyWrapperDescr_Type) {
      __pyx_wrapperbase_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_12__call__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_12__call__.doc = __pyx_doc_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_12__call__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_12ms_deisotope_2_c_7scoring_18IsotopicFitterBase_12__call__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_7scoring_IsotopicFitterBase.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase) < 0) __PYX_ERR(0, 323, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_IsotopicFitterBase, (PyObject *)&__pyx_type_12ms_deisotope_2_c_7scoring_IsotopicFitterBase) < 0) __PYX_ERR(0, 323, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase = &__pyx_type_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_LeastSquaresFitter = &__pyx_vtable_12ms_deisotope_2_c_7scoring_LeastSquaresFitter;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_LeastSquaresFitter.__pyx_base = *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_LeastSquaresFitter.__pyx_base._evaluate = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, PyObject *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_18LeastSquaresFitter__evaluate;
  __pyx_type_12ms_deisotope_2_c_7scoring_LeastSquaresFitter.tp_base = __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_7scoring_LeastSquaresFitter) < 0) __PYX_ERR(0, 566, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_7scoring_LeastSquaresFitter.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_7scoring_LeastSquaresFitter.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_7scoring_LeastSquaresFitter.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_7scoring_LeastSquaresFitter.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_7scoring_LeastSquaresFitter.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_7scoring_LeastSquaresFitter) < 0) __PYX_ERR(0, 566, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_LeastSquaresFitter, (PyObject *)&__pyx_type_12ms_deisotope_2_c_7scoring_LeastSquaresFitter) < 0) __PYX_ERR(0, 566, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter = &__pyx_type_12ms_deisotope_2_c_7scoring_LeastSquaresFitter;
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_GTestFitter = &__pyx_vtable_12ms_deisotope_2_c_7scoring_GTestFitter;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_GTestFitter.__pyx_base = *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_GTestFitter.__pyx_base._evaluate = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, PyObject *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_11GTestFitter__evaluate;
  __pyx_type_12ms_deisotope_2_c_7scoring_GTestFitter.tp_base = __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_7scoring_GTestFitter) < 0) __PYX_ERR(0, 470, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_7scoring_GTestFitter.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_7scoring_GTestFitter.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_7scoring_GTestFitter.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_7scoring_GTestFitter.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_7scoring_GTestFitter.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_7scoring_GTestFitter) < 0) __PYX_ERR(0, 470, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_GTestFitter, (PyObject *)&__pyx_type_12ms_deisotope_2_c_7scoring_GTestFitter) < 0) __PYX_ERR(0, 470, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter = &__pyx_type_12ms_deisotope_2_c_7scoring_GTestFitter;
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledGTestFitter = &__pyx_vtable_12ms_deisotope_2_c_7scoring_ScaledGTestFitter;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_ScaledGTestFitter.__pyx_base = *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_ScaledGTestFitter.__pyx_base._evaluate = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, PyObject *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_17ScaledGTestFitter__evaluate;
  __pyx_type_12ms_deisotope_2_c_7scoring_ScaledGTestFitter.tp_base = __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_7scoring_ScaledGTestFitter) < 0) __PYX_ERR(0, 502, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_7scoring_ScaledGTestFitter.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_7scoring_ScaledGTestFitter.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_7scoring_ScaledGTestFitter.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_7scoring_ScaledGTestFitter.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_7scoring_ScaledGTestFitter.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledGTestFitter) < 0) __PYX_ERR(0, 502, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ScaledGTestFitter, (PyObject *)&__pyx_type_12ms_deisotope_2_c_7scoring_ScaledGTestFitter) < 0) __PYX_ERR(0, 502, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter = &__pyx_type_12ms_deisotope_2_c_7scoring_ScaledGTestFitter;
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_MSDeconVFitter = &__pyx_vtable_12ms_deisotope_2_c_7scoring_MSDeconVFitter;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_MSDeconVFitter.__pyx_base = *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_MSDeconVFitter.__pyx_base._evaluate = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, PyObject *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_14MSDeconVFitter__evaluate;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_MSDeconVFitter.score_peak = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak;
  __pyx_type_12ms_deisotope_2_c_7scoring_MSDeconVFitter.tp_base = __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_7scoring_MSDeconVFitter) < 0) __PYX_ERR(0, 656, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_7scoring_MSDeconVFitter.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_7scoring_MSDeconVFitter.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_7scoring_MSDeconVFitter.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_7scoring_MSDeconVFitter.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_7scoring_MSDeconVFitter.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_7scoring_MSDeconVFitter) < 0) __PYX_ERR(0, 656, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_MSDeconVFitter, (PyObject *)&__pyx_type_12ms_deisotope_2_c_7scoring_MSDeconVFitter) < 0) __PYX_ERR(0, 656, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter = &__pyx_type_12ms_deisotope_2_c_7scoring_MSDeconVFitter;
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter = &__pyx_vtable_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter.__pyx_base = *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter.__pyx_base._evaluate = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, PyObject *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_23PenalizedMSDeconVFitter__evaluate;
  __pyx_type_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter.tp_base = __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter) < 0) __PYX_ERR(0, 722, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter) < 0) __PYX_ERR(0, 722, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PenalizedMSDeconVFitter, (PyObject *)&__pyx_type_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter) < 0) __PYX_ERR(0, 722, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter = &__pyx_type_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter;
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_FunctionScorer = &__pyx_vtable_12ms_deisotope_2_c_7scoring_FunctionScorer;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_FunctionScorer.__pyx_base = *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_FunctionScorer.__pyx_base._evaluate = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, PyObject *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_14FunctionScorer__evaluate;
  __pyx_type_12ms_deisotope_2_c_7scoring_FunctionScorer.tp_base = __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_7scoring_FunctionScorer) < 0) __PYX_ERR(0, 781, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_7scoring_FunctionScorer.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_7scoring_FunctionScorer.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_7scoring_FunctionScorer.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_7scoring_FunctionScorer.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_7scoring_FunctionScorer.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_7scoring_FunctionScorer) < 0) __PYX_ERR(0, 781, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_FunctionScorer, (PyObject *)&__pyx_type_12ms_deisotope_2_c_7scoring_FunctionScorer) < 0) __PYX_ERR(0, 781, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer = &__pyx_type_12ms_deisotope_2_c_7scoring_FunctionScorer;
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_InterferenceDetection = &__pyx_vtable_12ms_deisotope_2_c_7scoring_InterferenceDetection;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_InterferenceDetection.detect_interference = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_7scoring_InterferenceDetection) < 0) __PYX_ERR(0, 812, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_7scoring_InterferenceDetection.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_7scoring_InterferenceDetection.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_7scoring_InterferenceDetection.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_7scoring_InterferenceDetection.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_7scoring_InterferenceDetection.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_7scoring_InterferenceDetection) < 0) __PYX_ERR(0, 812, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_InterferenceDetection, (PyObject *)&__pyx_type_12ms_deisotope_2_c_7scoring_InterferenceDetection) < 0) __PYX_ERR(0, 812, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection = &__pyx_type_12ms_deisotope_2_c_7scoring_InterferenceDetection;
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_DistinctPatternFitter = &__pyx_vtable_12ms_deisotope_2_c_7scoring_DistinctPatternFitter;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_DistinctPatternFitter.__pyx_base = *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_DistinctPatternFitter.__pyx_base._evaluate = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, PyObject *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter__evaluate;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_DistinctPatternFitter.__pyx_base._configure = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *(*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_21DistinctPatternFitter__configure;
  __pyx_type_12ms_deisotope_2_c_7scoring_DistinctPatternFitter.tp_base = __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_7scoring_DistinctPatternFitter) < 0) __PYX_ERR(0, 863, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_7scoring_DistinctPatternFitter.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_7scoring_DistinctPatternFitter.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_7scoring_DistinctPatternFitter.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_7scoring_DistinctPatternFitter.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_7scoring_DistinctPatternFitter.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_7scoring_DistinctPatternFitter) < 0) __PYX_ERR(0, 863, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_DistinctPatternFitter, (PyObject *)&__pyx_type_12ms_deisotope_2_c_7scoring_DistinctPatternFitter) < 0) __PYX_ERR(0, 863, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter = &__pyx_type_12ms_deisotope_2_c_7scoring_DistinctPatternFitter;
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter = &__pyx_vtable_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter.__pyx_base = *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter.__pyx_base._evaluate = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, PyObject *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter__evaluate;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter._calculate_scale_factor = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter__calculate_scale_factor;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter.scale_fitted_peaks = (void (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *, PyObject *, double))__pyx_f_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_scale_fitted_peaks;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter.scale_theoretical_peaks = (void (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *, PyObject *, double))__pyx_f_12ms_deisotope_2_c_7scoring_29ScaledPenalizedMSDeconvFitter_scale_theoretical_peaks;
  __pyx_type_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter.tp_base = __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter) < 0) __PYX_ERR(0, 911, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter) < 0) __PYX_ERR(0, 911, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ScaledPenalizedMSDeconvFitter, (PyObject *)&__pyx_type_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter) < 0) __PYX_ERR(0, 911, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter = &__pyx_type_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter;
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_DotProductFitter = &__pyx_vtable_12ms_deisotope_2_c_7scoring_DotProductFitter;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_DotProductFitter.__pyx_base = *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_DotProductFitter.__pyx_base._evaluate = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, PyObject *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_16DotProductFitter__evaluate;
  __pyx_type_12ms_deisotope_2_c_7scoring_DotProductFitter.tp_base = __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_7scoring_DotProductFitter) < 0) __PYX_ERR(0, 981, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_7scoring_DotProductFitter.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_7scoring_DotProductFitter.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_7scoring_DotProductFitter.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_7scoring_DotProductFitter.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_7scoring_DotProductFitter.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_7scoring_DotProductFitter) < 0) __PYX_ERR(0, 981, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_DotProductFitter, (PyObject *)&__pyx_type_12ms_deisotope_2_c_7scoring_DotProductFitter) < 0) __PYX_ERR(0, 981, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter = &__pyx_type_12ms_deisotope_2_c_7scoring_DotProductFitter;
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_MinimizeFitSelector = &__pyx_vtable_12ms_deisotope_2_c_7scoring_MinimizeFitSelector;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_MinimizeFitSelector.__pyx_base = *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FitSelectorBase;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_MinimizeFitSelector.__pyx_base.best = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_best;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_MinimizeFitSelector.__pyx_base._best_from_set = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, PyObject *))__pyx_f_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector__best_from_set;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_MinimizeFitSelector.__pyx_base.reject = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_reject;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_MinimizeFitSelector.__pyx_base.reject_score = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, double, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_reject_score;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_MinimizeFitSelector.__pyx_base.is_maximizing = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_19MinimizeFitSelector_is_maximizing;
  __pyx_type_12ms_deisotope_2_c_7scoring_MinimizeFitSelector.tp_base = __pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_7scoring_MinimizeFitSelector) < 0) __PYX_ERR(0, 199, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_7scoring_MinimizeFitSelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_7scoring_MinimizeFitSelector.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_7scoring_MinimizeFitSelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_7scoring_MinimizeFitSelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_7scoring_MinimizeFitSelector.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_7scoring_MinimizeFitSelector) < 0) __PYX_ERR(0, 199, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_MinimizeFitSelector, (PyObject *)&__pyx_type_12ms_deisotope_2_c_7scoring_MinimizeFitSelector) < 0) __PYX_ERR(0, 199, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_MinimizeFitSelector = &__pyx_type_12ms_deisotope_2_c_7scoring_MinimizeFitSelector;
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_MaximizeFitSelector = &__pyx_vtable_12ms_deisotope_2_c_7scoring_MaximizeFitSelector;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_MaximizeFitSelector.__pyx_base = *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FitSelectorBase;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_MaximizeFitSelector.__pyx_base.best = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_best;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_MaximizeFitSelector.__pyx_base._best_from_set = (struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, PyObject *))__pyx_f_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector__best_from_set;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_MaximizeFitSelector.__pyx_base.reject = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_reject;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_MaximizeFitSelector.__pyx_base.reject_score = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, double, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_reject_score;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_MaximizeFitSelector.__pyx_base.is_maximizing = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_19MaximizeFitSelector_is_maximizing;
  __pyx_type_12ms_deisotope_2_c_7scoring_MaximizeFitSelector.tp_base = __pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_7scoring_MaximizeFitSelector) < 0) __PYX_ERR(0, 262, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_7scoring_MaximizeFitSelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_7scoring_MaximizeFitSelector.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_7scoring_MaximizeFitSelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_7scoring_MaximizeFitSelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_7scoring_MaximizeFitSelector.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_7scoring_MaximizeFitSelector) < 0) __PYX_ERR(0, 262, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_MaximizeFitSelector, (PyObject *)&__pyx_type_12ms_deisotope_2_c_7scoring_MaximizeFitSelector) < 0) __PYX_ERR(0, 262, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_MaximizeFitSelector = &__pyx_type_12ms_deisotope_2_c_7scoring_MaximizeFitSelector;
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_ChiSquareFitter = &__pyx_vtable_12ms_deisotope_2_c_7scoring_ChiSquareFitter;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_ChiSquareFitter.__pyx_base = *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
  __pyx_vtable_12ms_deisotope_2_c_7scoring_ChiSquareFitter.__pyx_base._evaluate = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, PyObject *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_7scoring_15ChiSquareFitter__evaluate;
  __pyx_type_12ms_deisotope_2_c_7scoring_ChiSquareFitter.tp_base = __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_7scoring_ChiSquareFitter) < 0) __PYX_ERR(0, 615, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_7scoring_ChiSquareFitter.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_7scoring_ChiSquareFitter.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_7scoring_ChiSquareFitter.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_7scoring_ChiSquareFitter.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_7scoring_ChiSquareFitter.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_7scoring_ChiSquareFitter) < 0) __PYX_ERR(0, 615, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ChiSquareFitter, (PyObject *)&__pyx_type_12ms_deisotope_2_c_7scoring_ChiSquareFitter) < 0) __PYX_ERR(0, 615, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_ChiSquareFitter = &__pyx_type_12ms_deisotope_2_c_7scoring_ChiSquareFitter;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__) < 0) __PYX_ERR(0, 146, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__ = &__pyx_type_12ms_deisotope_2_c_7scoring___pyx_scope_struct____iter__;
  __pyx_vtabptr_array = &__pyx_vtable_array;
  __pyx_vtable_array.get_memview = (PyObject *(*)(struct __pyx_array_obj *))__pyx_array_get_memview;
  if (PyType_Ready(&__pyx_type___pyx_array) < 0) __PYX_ERR(3, 106, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_array.tp_print = 0;
  #endif
  if (__Pyx_SetVtable(__pyx_type___pyx_array.tp_dict, __pyx_vtabptr_array) < 0) __PYX_ERR(3, 106, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_array) < 0) __PYX_ERR(3, 106, __pyx_L1_error)
  __pyx_array_type = &__pyx_type___pyx_array;
  if (PyType_Ready(&__pyx_type___pyx_MemviewEnum) < 0) __PYX_ERR(3, 280, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_MemviewEnum.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_MemviewEnum.tp_dictoffset && __pyx_type___pyx_MemviewEnum.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_MemviewEnum.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_MemviewEnum) < 0) __PYX_ERR(3, 280, __pyx_L1_error)
  __pyx_MemviewEnum_type = &__pyx_type___pyx_MemviewEnum;
  __pyx_vtabptr_memoryview = &__pyx_vtable_memoryview;
  __pyx_vtable_memoryview.get_item_pointer = (char *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_get_item_pointer;
  __pyx_vtable_memoryview.is_slice = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_is_slice;
  __pyx_vtable_memoryview.setitem_slice_assignment = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_slice_assignment;
  __pyx_vtable_memoryview.setitem_slice_assign_scalar = (PyObject *(*)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_setitem_slice_assign_scalar;
  __pyx_vtable_memoryview.setitem_indexed = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_indexed;
  __pyx_vtable_memoryview.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryview_convert_item_to_object;
  __pyx_vtable_memoryview.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryview_assign_item_from_object;
  if (PyType_Ready(&__pyx_type___pyx_memoryview) < 0) __PYX_ERR(3, 331, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_memoryview.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_memoryview.tp_dictoffset && __pyx_type___pyx_memoryview.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_memoryview.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type___pyx_memoryview.tp_dict, __pyx_vtabptr_memoryview) < 0) __PYX_ERR(3, 331, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_memoryview) < 0) __PYX_ERR(3, 331, __pyx_L1_error)
  __pyx_memoryview_type = &__pyx_type___pyx_memoryview;
  __pyx_vtabptr__memoryviewslice = &__pyx_vtable__memoryviewslice;
  __pyx_vtable__memoryviewslice.__pyx_base = *__pyx_vtabptr_memoryview;
  __pyx_vtable__memoryviewslice.__pyx_base.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryviewslice_convert_item_to_object;
  __pyx_vtable__memoryviewslice.__pyx_base.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryviewslice_assign_item_from_object;
  __pyx_type___pyx_memoryviewslice.tp_base = __pyx_memoryview_type;
  if (PyType_Ready(&__pyx_type___pyx_memoryviewslice) < 0) __PYX_ERR(3, 967, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_memoryviewslice.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_memoryviewslice.tp_dictoffset && __pyx_type___pyx_memoryviewslice.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_memoryviewslice.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type___pyx_memoryviewslice.tp_dict, __pyx_vtabptr__memoryviewslice) < 0) __PYX_ERR(3, 967, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_memoryviewslice) < 0) __PYX_ERR(3, 967, __pyx_L1_error)
  __pyx_memoryviewslice_type = &__pyx_type___pyx_memoryviewslice;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.composition"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7brainpy_2_c_11composition_PyComposition = __Pyx_ImportType(__pyx_t_1, "brainpy._c.composition", "PyComposition", sizeof(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7brainpy_2_c_11composition_PyComposition) __PYX_ERR(4, 123, __pyx_L1_error)
  __pyx_vtabptr_7brainpy_2_c_11composition_PyComposition = (struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition*)__Pyx_GetVtable(__pyx_ptype_7brainpy_2_c_11composition_PyComposition->tp_dict); if (unlikely(!__pyx_vtabptr_7brainpy_2_c_11composition_PyComposition)) __PYX_ERR(4, 123, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.isotopic_distribution"); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak = __Pyx_ImportType(__pyx_t_1, "brainpy._c.isotopic_distribution", "TheoreticalPeak", sizeof(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak) __PYX_ERR(5, 63, __pyx_L1_error)
  __pyx_vtabptr_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak = (struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak*)__Pyx_GetVtable(__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak->tp_dict); if (unlikely(!__pyx_vtabptr_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak)) __PYX_ERR(5, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_peak_picker._c.peak_set"); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "PeakBase", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase) __PYX_ERR(6, 2, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase)) __PYX_ERR(6, 2, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "FittedPeak", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak) __PYX_ERR(6, 12, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak)) __PYX_ERR(6, 12, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "PeakSet", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet) __PYX_ERR(6, 29, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet)) __PYX_ERR(6, 29, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "PeakSetIndexed", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed) __PYX_ERR(6, 83, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed)) __PYX_ERR(6, 83, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(7, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(2, 199, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(2, 222, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(2, 226, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(2, 238, __pyx_L1_error)
  __pyx_ptype_5numpy_generic = __Pyx_ImportType(__pyx_t_1, "numpy", "generic", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_generic) __PYX_ERR(2, 770, __pyx_L1_error)
  __pyx_ptype_5numpy_number = __Pyx_ImportType(__pyx_t_1, "numpy", "number", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_number) __PYX_ERR(2, 772, __pyx_L1_error)
  __pyx_ptype_5numpy_integer = __Pyx_ImportType(__pyx_t_1, "numpy", "integer", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_integer) __PYX_ERR(2, 774, __pyx_L1_error)
  __pyx_ptype_5numpy_signedinteger = __Pyx_ImportType(__pyx_t_1, "numpy", "signedinteger", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_signedinteger) __PYX_ERR(2, 776, __pyx_L1_error)
  __pyx_ptype_5numpy_unsignedinteger = __Pyx_ImportType(__pyx_t_1, "numpy", "unsignedinteger", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_unsignedinteger) __PYX_ERR(2, 778, __pyx_L1_error)
  __pyx_ptype_5numpy_inexact = __Pyx_ImportType(__pyx_t_1, "numpy", "inexact", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_inexact) __PYX_ERR(2, 780, __pyx_L1_error)
  __pyx_ptype_5numpy_floating = __Pyx_ImportType(__pyx_t_1, "numpy", "floating", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_floating) __PYX_ERR(2, 782, __pyx_L1_error)
  __pyx_ptype_5numpy_complexfloating = __Pyx_ImportType(__pyx_t_1, "numpy", "complexfloating", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_complexfloating) __PYX_ERR(2, 784, __pyx_L1_error)
  __pyx_ptype_5numpy_flexible = __Pyx_ImportType(__pyx_t_1, "numpy", "flexible", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_flexible) __PYX_ERR(2, 786, __pyx_L1_error)
  __pyx_ptype_5numpy_character = __Pyx_ImportType(__pyx_t_1, "numpy", "character", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_character) __PYX_ERR(2, 788, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(2, 826, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_peak_picker._c.peak_index"); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_index", "PeakIndex", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex) __PYX_ERR(8, 5, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_10peak_index_PeakIndex = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_10peak_index_PeakIndex)) __PYX_ERR(8, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.averagine"); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.averagine", "Averagine", sizeof(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine) __PYX_ERR(9, 12, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_9averagine_Averagine = (struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_9averagine_Averagine)) __PYX_ERR(9, 12, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.averagine", "TheoreticalIsotopicPattern", sizeof(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern) __PYX_ERR(9, 23, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern = (struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern)) __PYX_ERR(9, 23, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.averagine", "AveragineCache", sizeof(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache) __PYX_ERR(9, 79, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_9averagine_AveragineCache = (struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_9averagine_AveragineCache)) __PYX_ERR(9, 79, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.deconvoluter_base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.deconvoluter_base", "DeconvoluterBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase) __PYX_ERR(10, 25, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase)) __PYX_ERR(10, 25, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.deconvoluter_base", "AveragineDeconvoluterBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase) __PYX_ERR(10, 57, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase)) __PYX_ERR(10, 57, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.deconvoluter_base", "MultiAveragineDeconvoluterBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase) __PYX_ERR(10, 69, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase)) __PYX_ERR(10, 69, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.deconvoluter_base", "ChargeIterator", sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator) __PYX_ERR(10, 91, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator)) __PYX_ERR(10, 91, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.composition"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "nist_mass", (void **)&__pyx_vp_7brainpy_2_c_11composition_nist_mass, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "PROTON", (void **)&__pyx_vp_7brainpy_2_c_11composition_PROTON, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "_ElementTable", (void **)&__pyx_vp_7brainpy_2_c_11composition__ElementTable, "struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.isotopic_constants"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "DEFAULT_ISOTOPIC_CONSTANTS_SIZE", (void **)&__pyx_vp_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE, "size_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_peak_picker._c.peak_set"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "INTERVAL_INDEX_SIZE", (void **)&__pyx_vp_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE, "size_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.averagine"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "PROTON", (void **)&__pyx_vp_12ms_deisotope_2_c_9averagine_PROTON, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initscoring(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initscoring(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_scoring(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_scoring(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_scoring(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  static PyThread_type_lock __pyx_t_2[8];
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'scoring' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_scoring(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("scoring", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_ms_deisotope___c__scoring) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "ms_deisotope._c.scoring")) {
      if (unlikely(PyDict_SetItemString(modules, "ms_deisotope._c.scoring", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  if (unlikely(__Pyx_modinit_variable_export_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_variable_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "ms_deisotope/_c/scoring.pyx":6
 * from libc.math cimport fabs, sqrt, log, ceil, floor
 * from libc.stdlib cimport malloc, free
 * import operator             # <<<<<<<<<<<<<<
 * 
 * from cpython.list cimport PyList_New, PyList_GetItem, PyList_Size, PyList_GET_ITEM, PyList_SET_ITEM, PyList_GET_SIZE
 */

#line 6 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_operator, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)

#line 6 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 6 "ms_deisotope/_c/scoring.pyx"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_operator, __pyx_t_1) < 0) __PYX_ERR(0, 6, __pyx_L1_error)

#line 6 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/scoring.pyx":537
 * cdef ScaledGTestFitter g_test_scaled
 * 
 * g_test_scaled = ScaledGTestFitter()             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 537 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 537, __pyx_L1_error)

#line 537 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 537 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGOTREF(((PyObject *)__pyx_v_12ms_deisotope_2_c_7scoring_g_test_scaled));

#line 537 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF_SET(__pyx_v_12ms_deisotope_2_c_7scoring_g_test_scaled, ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *)__pyx_t_1));

#line 537 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 537 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/scoring.pyx":612
 * cdef LeastSquaresFitter least_squares
 * 
 * least_squares = LeastSquaresFitter()             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 612 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 612, __pyx_L1_error)

#line 612 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 612 "ms_deisotope/_c/scoring.pyx"
  __Pyx_XGOTREF(((PyObject *)__pyx_v_12ms_deisotope_2_c_7scoring_least_squares));

#line 612 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF_SET(__pyx_v_12ms_deisotope_2_c_7scoring_least_squares, ((struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter *)__pyx_t_1));

#line 612 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 612 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/scoring.pyx":795
 *     """
 * 
 *     def __init__(self, function, minimum_score=10, selector_type=MaximizeFitSelector):             # <<<<<<<<<<<<<<
 *         self.function = function
 *         self.select = selector_type(minimum_score)
 */

#line 795 "ms_deisotope/_c/scoring.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_7scoring_MaximizeFitSelector));

#line 795 "ms_deisotope/_c/scoring.pyx"
  __pyx_k__2 = ((PyObject *)__pyx_ptype_12ms_deisotope_2_c_7scoring_MaximizeFitSelector);

#line 795 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GIVEREF(__pyx_ptype_12ms_deisotope_2_c_7scoring_MaximizeFitSelector);

  /* "ms_deisotope/_c/scoring.pyx":1
 * # cython: embedsignature=True             # <<<<<<<<<<<<<<
 * 
 * cimport cython
 */

#line 1 "ms_deisotope/_c/scoring.pyx"
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)

#line 1 "ms_deisotope/_c/scoring.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 1 "ms_deisotope/_c/scoring.pyx"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)

#line 1 "ms_deisotope/_c/scoring.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "View.MemoryView":210
 *         info.obj = self
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(array self):
 */

#line 210 "View.MemoryView"
  __pyx_t_1 = __pyx_capsule_create(((void *)(&__pyx_array_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 210, __pyx_L1_error)

#line 210 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 210 "View.MemoryView"
  if (PyDict_SetItem((PyObject *)__pyx_array_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_1) < 0) __PYX_ERR(3, 210, __pyx_L1_error)

#line 210 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 210 "View.MemoryView"
  PyType_Modified(__pyx_array_type);

  /* "View.MemoryView":287
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */

#line 287 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__25, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 287, __pyx_L1_error)

#line 287 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 287 "View.MemoryView"
  __Pyx_XGOTREF(generic);

#line 287 "View.MemoryView"
  __Pyx_DECREF_SET(generic, __pyx_t_1);

#line 287 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_1);

#line 287 "View.MemoryView"
  __pyx_t_1 = 0;

  /* "View.MemoryView":288
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */

#line 288 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__26, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 288, __pyx_L1_error)

#line 288 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 288 "View.MemoryView"
  __Pyx_XGOTREF(strided);

#line 288 "View.MemoryView"
  __Pyx_DECREF_SET(strided, __pyx_t_1);

#line 288 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_1);

#line 288 "View.MemoryView"
  __pyx_t_1 = 0;

  /* "View.MemoryView":289
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 289 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__27, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 289, __pyx_L1_error)

#line 289 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 289 "View.MemoryView"
  __Pyx_XGOTREF(indirect);

#line 289 "View.MemoryView"
  __Pyx_DECREF_SET(indirect, __pyx_t_1);

#line 289 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_1);

#line 289 "View.MemoryView"
  __pyx_t_1 = 0;

  /* "View.MemoryView":292
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */

#line 292 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 292, __pyx_L1_error)

#line 292 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 292 "View.MemoryView"
  __Pyx_XGOTREF(contiguous);

#line 292 "View.MemoryView"
  __Pyx_DECREF_SET(contiguous, __pyx_t_1);

#line 292 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_1);

#line 292 "View.MemoryView"
  __pyx_t_1 = 0;

  /* "View.MemoryView":293
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 293 "View.MemoryView"
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__29, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 293, __pyx_L1_error)

#line 293 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 293 "View.MemoryView"
  __Pyx_XGOTREF(indirect_contiguous);

#line 293 "View.MemoryView"
  __Pyx_DECREF_SET(indirect_contiguous, __pyx_t_1);

#line 293 "View.MemoryView"
  __Pyx_GIVEREF(__pyx_t_1);

#line 293 "View.MemoryView"
  __pyx_t_1 = 0;

  /* "View.MemoryView":317
 * 
 * DEF THREAD_LOCKS_PREALLOCATED = 8
 * cdef int __pyx_memoryview_thread_locks_used = 0             # <<<<<<<<<<<<<<
 * cdef PyThread_type_lock[THREAD_LOCKS_PREALLOCATED] __pyx_memoryview_thread_locks = [
 *     PyThread_allocate_lock(),
 */

#line 317 "View.MemoryView"
  __pyx_memoryview_thread_locks_used = 0;

  /* "View.MemoryView":318
 * DEF THREAD_LOCKS_PREALLOCATED = 8
 * cdef int __pyx_memoryview_thread_locks_used = 0
 * cdef PyThread_type_lock[THREAD_LOCKS_PREALLOCATED] __pyx_memoryview_thread_locks = [             # <<<<<<<<<<<<<<
 *     PyThread_allocate_lock(),
 *     PyThread_allocate_lock(),
 */

#line 318 "View.MemoryView"
  __pyx_t_2[0] = PyThread_allocate_lock();

#line 318 "View.MemoryView"
  __pyx_t_2[1] = PyThread_allocate_lock();

#line 318 "View.MemoryView"
  __pyx_t_2[2] = PyThread_allocate_lock();

#line 318 "View.MemoryView"
  __pyx_t_2[3] = PyThread_allocate_lock();

#line 318 "View.MemoryView"
  __pyx_t_2[4] = PyThread_allocate_lock();

#line 318 "View.MemoryView"
  __pyx_t_2[5] = PyThread_allocate_lock();

#line 318 "View.MemoryView"
  __pyx_t_2[6] = PyThread_allocate_lock();

#line 318 "View.MemoryView"
  __pyx_t_2[7] = PyThread_allocate_lock();

#line 318 "View.MemoryView"
  memcpy(&(__pyx_memoryview_thread_locks[0]), __pyx_t_2, sizeof(__pyx_memoryview_thread_locks[0]) * (8));

  /* "View.MemoryView":551
 *         info.obj = self
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 551 "View.MemoryView"
  __pyx_t_1 = __pyx_capsule_create(((void *)(&__pyx_memoryview_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 551, __pyx_L1_error)

#line 551 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 551 "View.MemoryView"
  if (PyDict_SetItem((PyObject *)__pyx_memoryview_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_1) < 0) __PYX_ERR(3, 551, __pyx_L1_error)

#line 551 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 551 "View.MemoryView"
  PyType_Modified(__pyx_memoryview_type);

  /* "View.MemoryView":997
 *         return self.from_object
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 997 "View.MemoryView"
  __pyx_t_1 = __pyx_capsule_create(((void *)(&__pyx_memoryview_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 997, __pyx_L1_error)

#line 997 "View.MemoryView"
  __Pyx_GOTREF(__pyx_t_1);

#line 997 "View.MemoryView"
  if (PyDict_SetItem((PyObject *)__pyx_memoryviewslice_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_1) < 0) __PYX_ERR(3, 997, __pyx_L1_error)

#line 997 "View.MemoryView"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 997 "View.MemoryView"
  PyType_Modified(__pyx_memoryviewslice_type);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum, NULL, __pyx_n_s_View_MemoryView); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 1 "(tree fragment)"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Enum, __pyx_t_1) < 0) __PYX_ERR(3, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
  /*--- Wrapped vars code ---*/

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
  goto __pyx_L0;

#line 11 "(tree fragment)"
  __pyx_L1_error:;

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 11 "(tree fragment)"
  if (__pyx_m) {

#line 11 "(tree fragment)"
    if (__pyx_d) {

#line 11 "(tree fragment)"
      __Pyx_AddTraceback("init ms_deisotope._c.scoring", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 11 "(tree fragment)"
    }

#line 11 "(tree fragment)"
    Py_CLEAR(__pyx_m);

#line 11 "(tree fragment)"
  } else if (!PyErr_Occurred()) {

#line 11 "(tree fragment)"
    PyErr_SetString(PyExc_ImportError, "init ms_deisotope._c.scoring");

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  __pyx_L0:;

#line 11 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 11 "(tree fragment)"
  #if CYTHON_PEP489_MULTI_PHASE_INIT

#line 11 "(tree fragment)"
  return (__pyx_m != NULL) ? 0 : -1;

#line 11 "(tree fragment)"
  #elif PY_MAJOR_VERSION >= 3

#line 11 "(tree fragment)"
  return __pyx_m;

#line 11 "(tree fragment)"
  #else

#line 11 "(tree fragment)"
  return;

#line 11 "(tree fragment)"
  #endif

#line 11 "(tree fragment)"
}

#line 1 "cython_utility"

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#if defined(__Pyx_CyFunction_USED) && defined(NDEBUG)
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* set_iter */
static CYTHON_INLINE PyObject* __Pyx_set_iterator(PyObject* iterable, int is_set,
                                                  Py_ssize_t* p_orig_length, int* p_source_is_set) {
#if CYTHON_COMPILING_IN_CPYTHON
    is_set = is_set || likely(PySet_CheckExact(iterable) || PyFrozenSet_CheckExact(iterable));
    *p_source_is_set = is_set;
    if (likely(is_set)) {
        *p_orig_length = PySet_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
    }
#else
    (void)is_set;
    *p_source_is_set = 0;
#endif
    *p_orig_length = 0;
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_set_iter_next(
        PyObject* iter_obj, Py_ssize_t orig_length,
        Py_ssize_t* ppos, PyObject **value,
        int source_is_set) {
    if (!CYTHON_COMPILING_IN_CPYTHON || unlikely(!source_is_set)) {
        *value = PyIter_Next(iter_obj);
        if (unlikely(!*value)) {
            return __Pyx_IterFinish();
        }
        (void)orig_length;
        (void)ppos;
        return 1;
    }
#if CYTHON_COMPILING_IN_CPYTHON
    if (unlikely(PySet_GET_SIZE(iter_obj) != orig_length)) {
        PyErr_SetString(
            PyExc_RuntimeError,
            "set changed size during iteration");
        return -1;
    }
    {
        Py_hash_t hash;
        int ret = _PySet_NextEntry(iter_obj, ppos, value, &hash);
        assert (ret != -1);
        if (likely(ret)) {
            Py_INCREF(*value);
            return 1;
        }
    }
#endif
    return 0;
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* BufferIndexError */
static void __Pyx_RaiseBufferIndexError(int axis) {
  PyErr_Format(PyExc_IndexError,
     "Out of bounds on buffer access (axis %d)", axis);
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* DivInt[Py_ssize_t] */
static CYTHON_INLINE Py_ssize_t __Pyx_div_Py_ssize_t(Py_ssize_t a, Py_ssize_t b) {
    Py_ssize_t q = a / b;
    Py_ssize_t r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* decode_c_string */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    Py_ssize_t length;
    if (unlikely((start < 0) | (stop < 0))) {
        size_t slen = strlen(cstring);
        if (unlikely(slen > (size_t) PY_SSIZE_T_MAX)) {
            PyErr_SetString(PyExc_OverflowError,
                            "c-string too long to convert to Python");
            return NULL;
        }
        length = (Py_ssize_t) slen;
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (unlikely(stop <= start))
        return __Pyx_NewRef(__pyx_empty_unicode);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* None */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* DivInt[long] */
static CYTHON_INLINE long __Pyx_div_long(long a, long b) {
    long q = a / b;
    long r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_array_type)) return __pyx_array_getbuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_memoryview_type)) return __pyx_memoryview_getbuffer(obj, view, flags);
    PyErr_Format(PyExc_TypeError, "'%.200s' does not have the buffer interface", Py_TYPE(obj)->tp_name);
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
    if ((0)) {}
    view->obj = NULL;
    Py_DECREF(obj);
}
#endif


/* MemviewSliceIsContig */
static int
__pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim)
{
    int i, index, step, start;
    Py_ssize_t itemsize = mvs.memview->view.itemsize;
    if (order == 'F') {
        step = 1;
        start = 0;
    } else {
        step = -1;
        start = ndim - 1;
    }
    for (i = 0; i < ndim; i++) {
        index = start + step * i;
        if (mvs.suboffsets[index] >= 0 || mvs.strides[index] != itemsize)
            return 0;
        itemsize *= mvs.shape[index];
    }
    return 1;
}

/* OverlappingSlices */
static void
__pyx_get_array_memory_extents(__Pyx_memviewslice *slice,
                               void **out_start, void **out_end,
                               int ndim, size_t itemsize)
{
    char *start, *end;
    int i;
    start = end = slice->data;
    for (i = 0; i < ndim; i++) {
        Py_ssize_t stride = slice->strides[i];
        Py_ssize_t extent = slice->shape[i];
        if (extent == 0) {
            *out_start = *out_end = start;
            return;
        } else {
            if (stride > 0)
                end += stride * (extent - 1);
            else
                start += stride * (extent - 1);
        }
    }
    *out_start = start;
    *out_end = end + itemsize;
}
static int
__pyx_slices_overlap(__Pyx_memviewslice *slice1,
                     __Pyx_memviewslice *slice2,
                     int ndim, size_t itemsize)
{
    void *start1, *end1, *start2, *end2;
    __pyx_get_array_memory_extents(slice1, &start1, &end1, ndim, itemsize);
    __pyx_get_array_memory_extents(slice2, &start2, &end2, ndim, itemsize);
    return (start1 < end2) && (start2 < end1);
}

/* Capsule */
static CYTHON_INLINE PyObject *
__pyx_capsule_create(void *p, CYTHON_UNUSED const char *sig)
{
    PyObject *cobj;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(p, sig, NULL);
#else
    cobj = PyCObject_FromVoidPtr(p, NULL);
#endif
    return cobj;
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* Declarations */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* MemviewSliceCopyTemplate */
static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object)
{
    __Pyx_RefNannyDeclarations
    int i;
    __Pyx_memviewslice new_mvs = { 0, 0, { 0 }, { 0 }, { 0 } };
    struct __pyx_memoryview_obj *from_memview = from_mvs->memview;
    Py_buffer *buf = &from_memview->view;
    PyObject *shape_tuple = NULL;
    PyObject *temp_int = NULL;
    struct __pyx_array_obj *array_obj = NULL;
    struct __pyx_memoryview_obj *memview_obj = NULL;
    __Pyx_RefNannySetupContext("__pyx_memoryview_copy_new_contig", 0);
    for (i = 0; i < ndim; i++) {
        if (unlikely(from_mvs->suboffsets[i] >= 0)) {
            PyErr_Format(PyExc_ValueError, "Cannot copy memoryview slice with "
                                           "indirect dimensions (axis %d)", i);
            goto fail;
        }
    }
    shape_tuple = PyTuple_New(ndim);
    if (unlikely(!shape_tuple)) {
        goto fail;
    }
    __Pyx_GOTREF(shape_tuple);
    for(i = 0; i < ndim; i++) {
        temp_int = PyInt_FromSsize_t(from_mvs->shape[i]);
        if(unlikely(!temp_int)) {
            goto fail;
        } else {
            PyTuple_SET_ITEM(shape_tuple, i, temp_int);
            temp_int = NULL;
        }
    }
    array_obj = __pyx_array_new(shape_tuple, sizeof_dtype, buf->format, (char *) mode, NULL);
    if (unlikely(!array_obj)) {
        goto fail;
    }
    __Pyx_GOTREF(array_obj);
    memview_obj = (struct __pyx_memoryview_obj *) __pyx_memoryview_new(
                                    (PyObject *) array_obj, contig_flag,
                                    dtype_is_object,
                                    from_mvs->memview->typeinfo);
    if (unlikely(!memview_obj))
        goto fail;
    if (unlikely(__Pyx_init_memviewslice(memview_obj, ndim, &new_mvs, 1) < 0))
        goto fail;
    if (unlikely(__pyx_memoryview_copy_contents(*from_mvs, new_mvs, ndim, ndim,
                                                dtype_is_object) < 0))
        goto fail;
    goto no_fail;
fail:
    __Pyx_XDECREF(new_mvs.memview);
    new_mvs.memview = NULL;
    new_mvs.data = NULL;
no_fail:
    __Pyx_XDECREF(shape_tuple);
    __Pyx_XDECREF(temp_int);
    __Pyx_XDECREF(array_obj);
    __Pyx_RefNannyFinishContext();
    return new_mvs;
}

/* MemviewSliceInit */
static int
__Pyx_init_memviewslice(struct __pyx_memoryview_obj *memview,
                        int ndim,
                        __Pyx_memviewslice *memviewslice,
                        int memview_is_new_reference)
{
    __Pyx_RefNannyDeclarations
    int i, retval=-1;
    Py_buffer *buf = &memview->view;
    __Pyx_RefNannySetupContext("init_memviewslice", 0);
    if (unlikely(memviewslice->memview || memviewslice->data)) {
        PyErr_SetString(PyExc_ValueError,
            "memviewslice is already initialized!");
        goto fail;
    }
    if (buf->strides) {
        for (i = 0; i < ndim; i++) {
            memviewslice->strides[i] = buf->strides[i];
        }
    } else {
        Py_ssize_t stride = buf->itemsize;
        for (i = ndim - 1; i >= 0; i--) {
            memviewslice->strides[i] = stride;
            stride *= buf->shape[i];
        }
    }
    for (i = 0; i < ndim; i++) {
        memviewslice->shape[i]   = buf->shape[i];
        if (buf->suboffsets) {
            memviewslice->suboffsets[i] = buf->suboffsets[i];
        } else {
            memviewslice->suboffsets[i] = -1;
        }
    }
    memviewslice->memview = memview;
    memviewslice->data = (char *)buf->buf;
    if (__pyx_add_acquisition_count(memview) == 0 && !memview_is_new_reference) {
        Py_INCREF(memview);
    }
    retval = 0;
    goto no_fail;
fail:
    memviewslice->memview = 0;
    memviewslice->data = 0;
    retval = -1;
no_fail:
    __Pyx_RefNannyFinishContext();
    return retval;
}
#ifndef Py_NO_RETURN
#define Py_NO_RETURN
#endif
static void __pyx_fatalerror(const char *fmt, ...) Py_NO_RETURN {
    va_list vargs;
    char msg[200];
#if PY_VERSION_HEX >= 0x030A0000 || defined(HAVE_STDARG_PROTOTYPES)
    va_start(vargs, fmt);
#else
    va_start(vargs);
#endif
    vsnprintf(msg, 200, fmt, vargs);
    va_end(vargs);
    Py_FatalError(msg);
}
static CYTHON_INLINE int
__pyx_add_acquisition_count_locked(__pyx_atomic_int *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)++;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE int
__pyx_sub_acquisition_count_locked(__pyx_atomic_int *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)--;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE void
__Pyx_INC_MEMVIEW(__Pyx_memviewslice *memslice, int have_gil, int lineno)
{
    int first_time;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (unlikely(!memview || (PyObject *) memview == Py_None))
        return;
    if (unlikely(__pyx_get_slice_count(memview) < 0))
        __pyx_fatalerror("Acquisition count is %d (line %d)",
                         __pyx_get_slice_count(memview), lineno);
    first_time = __pyx_add_acquisition_count(memview) == 0;
    if (unlikely(first_time)) {
        if (have_gil) {
            Py_INCREF((PyObject *) memview);
        } else {
            PyGILState_STATE _gilstate = PyGILState_Ensure();
            Py_INCREF((PyObject *) memview);
            PyGILState_Release(_gilstate);
        }
    }
}
static CYTHON_INLINE void __Pyx_XDEC_MEMVIEW(__Pyx_memviewslice *memslice,
                                             int have_gil, int lineno) {
    int last_time;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (unlikely(!memview || (PyObject *) memview == Py_None)) {
        memslice->memview = NULL;
        return;
    }
    if (unlikely(__pyx_get_slice_count(memview) <= 0))
        __pyx_fatalerror("Acquisition count is %d (line %d)",
                         __pyx_get_slice_count(memview), lineno);
    last_time = __pyx_sub_acquisition_count(memview) == 1;
    memslice->data = NULL;
    if (unlikely(last_time)) {
        if (have_gil) {
            Py_CLEAR(memslice->memview);
        } else {
            PyGILState_STATE _gilstate = PyGILState_Ensure();
            Py_CLEAR(memslice->memview);
            PyGILState_Release(_gilstate);
        }
    } else {
        memslice->memview = NULL;
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const char neg_one = (char) -1, const_zero = (char) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(char) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(char, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (char) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case  1: __PYX_VERIFY_RETURN_INT(char, digit, digits[0])
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 2 * PyLong_SHIFT) {
                            return (char) (((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 3 * PyLong_SHIFT) {
                            return (char) (((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 4 * PyLong_SHIFT) {
                            return (char) (((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (char) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(char) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case -1: __PYX_VERIFY_RETURN_INT(char, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(char,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(char) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) ((((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) ((((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) ((((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(char) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            char val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (char) -1;
        }
    } else {
        char val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (char) -1;
        val = __Pyx_PyInt_As_char(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to char");
    return (char) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to char");
    return (char) -1;
}

/* FetchCommonType */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* CoroutineBase */
#include <structmember.h>
#include <frameobject.h>
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            assert(f->f_back == NULL);
            #if PY_VERSION_HEX >= 0x030B00A1
            f->f_back = PyThreadState_GetFrame(tstate);
            #else
            Py_XINCREF(tstate->frame);
            f->f_back = tstate->frame;
            #endif
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
            _PyGen_SetStopIterationValue(result);
        }
        Py_CLEAR(result);
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (Py_REFCNT(self) > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(Py_REFCNT(self) > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = Py_REFCNT(self);
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(Py_TYPE(self)) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *frame = self->gi_frame;
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {(char *) "gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* VoidPtrExport */
static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig) {
    PyObject *d;
    PyObject *cobj = 0;
    d = PyDict_GetItem(__pyx_d, __pyx_n_s_pyx_capi);
    Py_XINCREF(d);
    if (!d) {
        d = PyDict_New();
        if (!d)
            goto bad;
        if (__Pyx_PyObject_SetAttrStr(__pyx_m, __pyx_n_s_pyx_capi, d) < 0)
            goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItem(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* VoidPtrImport */
#ifndef __PYX_HAVE_RT_ImportVoidPtr
#define __PYX_HAVE_RT_ImportVoidPtr
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, name);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C variable %.200s",
                PyModule_GetName(module), name);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, desc);
        goto bad;
    }
    *p = PyCObject_AsVoidPtr(cobj);}
#endif
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
