/* Generated by Cython 0.29.33 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/arrayobject.h",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/arrayscalars.h",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/ndarrayobject.h",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/ndarraytypes.h",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include/numpy/ufuncobject.h"
        ],
        "include_dirs": [
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/core/include",
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/ms_peak_picker/_c"
        ],
        "name": "ms_deisotope._c.feature_map.lcms_feature",
        "sources": [
            "src/ms_deisotope/_c/feature_map/lcms_feature.pyx"
        ]
    },
    "module_name": "ms_deisotope._c.feature_map.lcms_feature"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_33"
#define CYTHON_HEX_VERSION 0x001D21F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 0
  #elif !defined(CYTHON_FAST_THREAD_STATE)
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030A0000)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__ms_deisotope___c__feature_map__lcms_feature
#define __PYX_HAVE_API__ms_deisotope___c__feature_map__lcms_feature
/* Early includes */
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "pythread.h"
#include "numpy/arrayobject.h"
#include "numpy/ndarrayobject.h"
#include "numpy/ndarraytypes.h"
#include "numpy/arrayscalars.h"
#include "numpy/ufuncobject.h"

    /* NumPy API declarations from "numpy/__init__.pxd" */
    
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "src/ms_deisotope/_c/feature_map/lcms_feature.pyx",
  "stringsource",
  "src/ms_deisotope/_c/feature_map/lcms_feature.pxd",
  "__init__.pxd",
  "peak_set.pxd",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
  "composition.pxd",
  "isotopic_distribution.pxd",
  "peak_index.pxd",
  "src/ms_deisotope/_c/averagine.pxd",
  "src/ms_deisotope/_c/deconvoluter_base.pxd",
  "src/ms_deisotope/_c/scoring.pxd",
  "src/ms_deisotope/_c/peak_set.pxd",
};
/* BufferFormatStructs.proto */
#define IS_UNSIGNED(type) (((type) -1) > 0)
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;


/* "brainpy/_c/composition.pxd":3
 * from brainpy.mass_dict import nist_mass as __nist_mass
 * 
 * ctypedef long count_type             # <<<<<<<<<<<<<<
 * 
 * cdef dict nist_mass
 */

#line 3 "brainpy/_c/composition.pxd"
typedef long __pyx_t_7brainpy_2_c_11composition_count_type;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":689
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */

#line 689 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":690
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */

#line 690 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":691
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */

#line 691 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":692
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */

#line 692 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":696
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */

#line 696 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":697
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */

#line 697 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":698
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */

#line 698 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":699
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */

#line 699 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":703
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */

#line 703 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":704
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */

#line 704 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":713
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */

#line 713 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":714
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */

#line 714 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":715
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */

#line 715 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":717
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */

#line 717 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":718
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */

#line 718 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":719
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */

#line 719 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":721
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */

#line 721 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":722
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */

#line 722 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":724
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */

#line 724 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":725
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */

#line 725 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":726
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */

#line 726 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet;
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed;
struct __pyx_obj_7brainpy_2_c_11composition_PyComposition;
struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak;
struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex;
struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine;
struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern;
struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase;
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter;
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed;
struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__;
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr;
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak;
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for;
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell;
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list;
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted;

/* "ms_peak_picker/_c/peak_set.pxd":42
 *     cdef FittedPeak _get_nearest_peak(self, double mz, double* errout)
 * 
 *     cdef FittedPeak _has_peak(self, double mz, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef PeakSet clone(self)
 */

#line 42 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak {

#line 42 "ms_peak_picker/_c/peak_set.pxd"
  int __pyx_n;

#line 42 "ms_peak_picker/_c/peak_set.pxd"
  double tolerance;

#line 42 "ms_peak_picker/_c/peak_set.pxd"
};

/* "ms_peak_picker/_c/peak_set.pxd":50
 *     cdef int _between_bounds(self, double m1, double m2, size_t* startp, size_t* endp)
 * 
 *     cpdef tuple all_peaks_for(self, double mz, double error_tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef FittedPeak getitem(self, size_t i)
 */

#line 50 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for {

#line 50 "ms_peak_picker/_c/peak_set.pxd"
  int __pyx_n;

#line 50 "ms_peak_picker/_c/peak_set.pxd"
  double error_tolerance;

#line 50 "ms_peak_picker/_c/peak_set.pxd"
};

/* "ms_peak_picker/_c/peak_set.pxd":64
 * 
 * 
 * cdef struct index_cell:             # <<<<<<<<<<<<<<
 *     double center_value
 *     size_t start
 */

#line 64 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell {

#line 64 "ms_peak_picker/_c/peak_set.pxd"
  double center_value;

#line 64 "ms_peak_picker/_c/peak_set.pxd"
  size_t start;

#line 64 "ms_peak_picker/_c/peak_set.pxd"
  size_t end;

#line 64 "ms_peak_picker/_c/peak_set.pxd"
};

/* "ms_peak_picker/_c/peak_set.pxd":70
 * 
 * 
 * cdef struct index_list:             # <<<<<<<<<<<<<<
 *     index_cell* index
 *     size_t size
 */

#line 70 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list {

#line 70 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_cell *index;

#line 70 "ms_peak_picker/_c/peak_set.pxd"
  size_t size;

#line 70 "ms_peak_picker/_c/peak_set.pxd"
  double low;

#line 70 "ms_peak_picker/_c/peak_set.pxd"
  double high;

#line 70 "ms_peak_picker/_c/peak_set.pxd"
};

/* "ms_peak_picker/_c/peak_set.pxd":90
 *     cpdef _allocate_index(self, size_t interval_index_size)
 * 
 *     cpdef FittedPeak has_peak_hinted(self, double mz, double tolerance=*, size_t hint_start=*, size_t hint_end=*)             # <<<<<<<<<<<<<<
 * 
 * cdef size_t double_binary_search_ppm(double* array, double value, double tolerance, size_t n)
 */

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted {

#line 90 "ms_peak_picker/_c/peak_set.pxd"
  int __pyx_n;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
  double tolerance;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
  size_t hint_start;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
  size_t hint_end;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
};

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_neutral_mass;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_mass_charge_ratio;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_t_7brainpy_2_c_11composition_Isotope;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_t_7brainpy_2_c_11composition_Element;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_t_7brainpy_2_c_11composition_Composition;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl;

#line 90 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition_pop;

/* "brainpy/_c/composition.pxd":10
 * cdef double PROTON = nist_mass["H+"][0][0]
 * 
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*) nogil             # <<<<<<<<<<<<<<
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*) nogil
 * cdef char* _parse_isotope_string(char* label, int* isotope_num, char* element_name) nogil
 */

#line 10 "brainpy/_c/composition.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_neutral_mass {

#line 10 "brainpy/_c/composition.pxd"
  int __pyx_n;

#line 10 "brainpy/_c/composition.pxd"
  double charge_carrier;

#line 10 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":11
 * 
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*) nogil
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*) nogil             # <<<<<<<<<<<<<<
 * cdef char* _parse_isotope_string(char* label, int* isotope_num, char* element_name) nogil
 * 
 */

#line 11 "brainpy/_c/composition.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_mass_charge_ratio {

#line 11 "brainpy/_c/composition.pxd"
  int __pyx_n;

#line 11 "brainpy/_c/composition.pxd"
  double charge_carrier;

#line 11 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":17
 * # Isotope and IsotopeMap Declarations
 * 
 * cdef struct Isotope:             # <<<<<<<<<<<<<<
 *     double mass
 *     double abundance
 */

#line 17 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_Isotope {

#line 17 "brainpy/_c/composition.pxd"
  double mass;

#line 17 "brainpy/_c/composition.pxd"
  double abundance;

#line 17 "brainpy/_c/composition.pxd"
  int neutrons;

#line 17 "brainpy/_c/composition.pxd"
  int neutron_shift;

#line 17 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":23
 *     int neutron_shift
 * 
 * cdef struct IsotopeMap:             # <<<<<<<<<<<<<<
 *     Isotope* bins
 *     size_t size
 */

#line 23 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap {

#line 23 "brainpy/_c/composition.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Isotope *bins;

#line 23 "brainpy/_c/composition.pxd"
  size_t size;

#line 23 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":36
 * # Element Declarations
 * 
 * cdef struct Element:             # <<<<<<<<<<<<<<
 *     char* symbol
 *     IsotopeMap* isotopes
 */

#line 36 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_Element {

#line 36 "brainpy/_c/composition.pxd"
  char *symbol;

#line 36 "brainpy/_c/composition.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *isotopes;

#line 36 "brainpy/_c/composition.pxd"
  int monoisotopic_isotope_index;

#line 36 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":56
 * # ElementHashTable and ElementHashBucket Declarations
 * 
 * cdef struct ElementHashBucket:             # <<<<<<<<<<<<<<
 *     Element** elements
 *     size_t used
 */

#line 56 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket {

#line 56 "brainpy/_c/composition.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Element **elements;

#line 56 "brainpy/_c/composition.pxd"
  size_t used;

#line 56 "brainpy/_c/composition.pxd"
  size_t size;

#line 56 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":65
 * 
 * 
 * cdef struct ElementHashTable:             # <<<<<<<<<<<<<<
 *     ElementHashBucket* buckets
 *     size_t size
 */

#line 65 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable {

#line 65 "brainpy/_c/composition.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket *buckets;

#line 65 "brainpy/_c/composition.pxd"
  size_t size;

#line 65 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":94
 * # Composition Declarations
 * 
 * cdef struct Composition:             # <<<<<<<<<<<<<<
 *     char** elements
 *     count_type* counts
 */

#line 94 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_11composition_Composition {

#line 94 "brainpy/_c/composition.pxd"
  char **elements;

#line 94 "brainpy/_c/composition.pxd"
  __pyx_t_7brainpy_2_c_11composition_count_type *counts;

#line 94 "brainpy/_c/composition.pxd"
  size_t size;

#line 94 "brainpy/_c/composition.pxd"
  size_t used;

#line 94 "brainpy/_c/composition.pxd"
  int max_variants;

#line 94 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":130
 *     @staticmethod
 *     cdef PyComposition _create(Composition* base)
 *     cdef void _set_impl(self, Composition* composition, bint free_existing=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _initialize_from_formula(self, str formula)
 */

#line 130 "brainpy/_c/composition.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl {

#line 130 "brainpy/_c/composition.pxd"
  int __pyx_n;

#line 130 "brainpy/_c/composition.pxd"
  int free_existing;

#line 130 "brainpy/_c/composition.pxd"
};

/* "brainpy/_c/composition.pxd":144
 *     cpdef list items(self)
 * 
 *     cpdef pop(self, str key, object default=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef count_type getitem(self, str key)
 */

#line 144 "brainpy/_c/composition.pxd"
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition_pop {

#line 144 "brainpy/_c/composition.pxd"
  int __pyx_n;

#line 144 "brainpy/_c/composition.pxd"
  PyObject *__pyx_default;

#line 144 "brainpy/_c/composition.pxd"
};

#line 144 "brainpy/_c/composition.pxd"
struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector;

/* "brainpy/_c/double_vector.pxd":1
 * cdef struct DoubleVector:             # <<<<<<<<<<<<<<
 *     double* v
 *     size_t used
 */

#line 1 "brainpy/_c/double_vector.pxd"
struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector {

#line 1 "brainpy/_c/double_vector.pxd"
  double *v;

#line 1 "brainpy/_c/double_vector.pxd"
  size_t used;

#line 1 "brainpy/_c/double_vector.pxd"
  size_t size;

#line 1 "brainpy/_c/double_vector.pxd"
};

#line 1 "brainpy/_c/double_vector.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters;

#line 1 "brainpy/_c/double_vector.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants;

#line 1 "brainpy/_c/double_vector.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants;

/* "brainpy/_c/isotopic_constants.pxd":6
 * from brainpy._c.double_vector cimport DoubleVector
 * 
 * ctypedef DoubleVector dvec             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 6 "brainpy/_c/isotopic_constants.pxd"
typedef struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector __pyx_t_7brainpy_2_c_18isotopic_constants_dvec;

/* "brainpy/_c/isotopic_constants.pxd":9
 * 
 * 
 * cdef struct PolynomialParameters:             # <<<<<<<<<<<<<<
 *     dvec* elementary_symmetric_polynomial
 *     dvec* power_sum
 */

#line 9 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters {

#line 9 "brainpy/_c/isotopic_constants.pxd"
  __pyx_t_7brainpy_2_c_18isotopic_constants_dvec *elementary_symmetric_polynomial;

#line 9 "brainpy/_c/isotopic_constants.pxd"
  __pyx_t_7brainpy_2_c_18isotopic_constants_dvec *power_sum;

#line 9 "brainpy/_c/isotopic_constants.pxd"
};

/* "brainpy/_c/isotopic_constants.pxd":14
 * 
 * 
 * cdef struct PhiConstants:             # <<<<<<<<<<<<<<
 *     int order
 *     Element* element
 */

#line 14 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants {

#line 14 "brainpy/_c/isotopic_constants.pxd"
  int order;

#line 14 "brainpy/_c/isotopic_constants.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Element *element;

#line 14 "brainpy/_c/isotopic_constants.pxd"
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters *element_coefficients;

#line 14 "brainpy/_c/isotopic_constants.pxd"
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters *mass_coefficients;

#line 14 "brainpy/_c/isotopic_constants.pxd"
};

/* "brainpy/_c/isotopic_constants.pxd":21
 * 
 * 
 * cdef struct IsotopicConstants:             # <<<<<<<<<<<<<<
 *     int order
 *     PhiConstants** constants
 */

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants {

#line 21 "brainpy/_c/isotopic_constants.pxd"
  int order;

#line 21 "brainpy/_c/isotopic_constants.pxd"
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants **constants;

#line 21 "brainpy/_c/isotopic_constants.pxd"
  size_t size;

#line 21 "brainpy/_c/isotopic_constants.pxd"
  size_t used;

#line 21 "brainpy/_c/isotopic_constants.pxd"
};

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_guess_npeaks;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_isotopic_variants;

#line 21 "brainpy/_c/isotopic_constants.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_max_variants_approx;

/* "brainpy/_c/isotopic_distribution.pxd":7
 * 
 * 
 * cdef struct Peak:             # <<<<<<<<<<<<<<
 *     double mz
 *     double intensity
 */

#line 7 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak {

#line 7 "brainpy/_c/isotopic_distribution.pxd"
  double mz;

#line 7 "brainpy/_c/isotopic_distribution.pxd"
  double intensity;

#line 7 "brainpy/_c/isotopic_distribution.pxd"
  int charge;

#line 7 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":13
 * 
 * 
 * cdef struct PeakList:             # <<<<<<<<<<<<<<
 *     Peak* peaks
 *     size_t used
 */

#line 13 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList {

#line 13 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *peaks;

#line 13 "brainpy/_c/isotopic_distribution.pxd"
  size_t used;

#line 13 "brainpy/_c/isotopic_distribution.pxd"
  size_t size;

#line 13 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":35
 * 
 * 
 * cdef struct ElementCache:             # <<<<<<<<<<<<<<
 *     Element** elements
 *     size_t used
 */

#line 35 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache {

#line 35 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Element **elements;

#line 35 "brainpy/_c/isotopic_distribution.pxd"
  size_t used;

#line 35 "brainpy/_c/isotopic_distribution.pxd"
  size_t size;

#line 35 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *source;

#line 35 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":49
 * 
 * 
 * cdef struct IsotopicDistribution:             # <<<<<<<<<<<<<<
 *     Composition* composition
 *     IsotopicConstants* _isotopic_constants
 */

#line 49 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution {

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Composition *composition;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *_isotopic_constants;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  int order;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  double average_mass;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *monoisotopic_peak;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *cache;

#line 49 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":58
 * 
 * 
 * cdef IsotopicDistribution* make_isotopic_distribution(Composition* composition, int order, ElementCache* cache=*) nogil             # <<<<<<<<<<<<<<
 * cdef void free_isotopic_distribution(IsotopicDistribution* distribution)  nogil
 * 
 */

#line 58 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution {

#line 58 "brainpy/_c/isotopic_distribution.pxd"
  int __pyx_n;

#line 58 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *cache;

#line 58 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":77
 * 
 * cdef int max_variants(Composition* composition, ElementCache* cache) nogil
 * cdef int guess_npeaks(Composition* composition_struct, size_t max_npeaks, ElementCache* cache=*) nogil             # <<<<<<<<<<<<<<
 * 
 * cpdef list _isotopic_variants(object composition, object npeaks=*, int charge=*, double charge_carrier=*)
 */

#line 77 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_guess_npeaks {

#line 77 "brainpy/_c/isotopic_distribution.pxd"
  int __pyx_n;

#line 77 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *cache;

#line 77 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":79
 * cdef int guess_npeaks(Composition* composition_struct, size_t max_npeaks, ElementCache* cache=*) nogil
 * 
 * cpdef list _isotopic_variants(object composition, object npeaks=*, int charge=*, double charge_carrier=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 79 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants {

#line 79 "brainpy/_c/isotopic_distribution.pxd"
  int __pyx_n;

#line 79 "brainpy/_c/isotopic_distribution.pxd"
  PyObject *npeaks;

#line 79 "brainpy/_c/isotopic_distribution.pxd"
  int charge;

#line 79 "brainpy/_c/isotopic_distribution.pxd"
  double charge_carrier;

#line 79 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":82
 * 
 * 
 * cdef PeakList* isotopic_variants(Composition* composition, int npeaks, int charge=*, double charge_carrier=*) nogil             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 82 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_isotopic_variants {

#line 82 "brainpy/_c/isotopic_distribution.pxd"
  int __pyx_n;

#line 82 "brainpy/_c/isotopic_distribution.pxd"
  int charge;

#line 82 "brainpy/_c/isotopic_distribution.pxd"
  double charge_carrier;

#line 82 "brainpy/_c/isotopic_distribution.pxd"
};

/* "brainpy/_c/isotopic_distribution.pxd":85
 * 
 * 
 * cdef size_t max_variants_approx(double mass, double lambda_factor=*, size_t maxiter=*, double threshold=*) nogil             # <<<<<<<<<<<<<<
 */

#line 85 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_max_variants_approx {

#line 85 "brainpy/_c/isotopic_distribution.pxd"
  int __pyx_n;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
  double lambda_factor;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
  size_t maxiter;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
  double threshold;

#line 85 "brainpy/_c/isotopic_distribution.pxd"
};

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":728
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */

#line 728 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":729
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */

#line 729 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":730
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */

#line 730 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":732
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */

#line 732 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
typedef npy_cdouble __pyx_t_5numpy_complex_t;

#line 732 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex__has_peak;

#line 732 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex_all_peaks_for;

/* "ms_peak_picker/_c/peak_index.pxd":11
 *         public PeakSet peaks
 * 
 *     cdef FittedPeak _has_peak(self, double mz, double tolerance=*)             # <<<<<<<<<<<<<<
 *     cdef PeakSet _between(self, double start, double stop)
 * 
 */

#line 11 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex__has_peak {

#line 11 "ms_peak_picker/_c/peak_index.pxd"
  int __pyx_n;

#line 11 "ms_peak_picker/_c/peak_index.pxd"
  double tolerance;

#line 11 "ms_peak_picker/_c/peak_index.pxd"
};

/* "ms_peak_picker/_c/peak_index.pxd":14
 *     cdef PeakSet _between(self, double start, double stop)
 * 
 *     cpdef tuple all_peaks_for(self, double mz, double error_tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef size_t get_size(self)
 */

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex_all_peaks_for {

#line 14 "ms_peak_picker/_c/peak_index.pxd"
  int __pyx_n;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
  double error_tolerance;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
};

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_mass_charge_ratio;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_scale;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_isotopic_cluster;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine__isotopic_cluster;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_ignore_below;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_truncate_after;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_scale;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern__scale;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_has_mz_charge_pair;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_isotopic_cluster;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_isotopic_shift;

#line 14 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_poisson_approximate;

/* "ms_deisotope/_c/averagine.pxd":7
 * cdef double PROTON
 * 
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*)             # <<<<<<<<<<<<<<
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*)
 * cdef void slide(double mz, list peaklist)
 */

#line 7 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_mass_charge_ratio {

#line 7 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 7 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 7 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":8
 * 
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*)
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*)             # <<<<<<<<<<<<<<
 * cdef void slide(double mz, list peaklist)
 * cdef dict scale_dict(dict data, double factor)
 */

#line 8 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_neutral_mass {

#line 8 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 8 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 8 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":17
 *         public dict base_composition
 * 
 *     cpdef dict scale(self, double mz, int charge=*, double charge_carrier=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 *     cdef TheoreticalIsotopicPattern _isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 */

#line 17 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_scale {

#line 17 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 17 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 17 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 17 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":18
 * 
 *     cpdef dict scale(self, double mz, int charge=*, double charge_carrier=*)
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 *     cdef TheoreticalIsotopicPattern _isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 * 
 */

#line 18 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_isotopic_cluster {

#line 18 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 18 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 18 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 18 "ms_deisotope/_c/averagine.pxd"
  double truncate_after;

#line 18 "ms_deisotope/_c/averagine.pxd"
  double ignore_below;

#line 18 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":19
 *     cpdef dict scale(self, double mz, int charge=*, double charge_carrier=*)
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 *     cdef TheoreticalIsotopicPattern _isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 19 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine__isotopic_cluster {

#line 19 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 19 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 19 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 19 "ms_deisotope/_c/averagine.pxd"
  double truncate_after;

#line 19 "ms_deisotope/_c/averagine.pxd"
  double ignore_below;

#line 19 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":39
 *     cdef bint _eq_inst(self, TheoreticalIsotopicPattern other)
 * 
 *     cpdef TheoreticalIsotopicPattern ignore_below(self, double ignore_below=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern truncate_after(self, double truncate_after=*)
 *     cpdef TheoreticalIsotopicPattern shift(self, double mz)
 */

#line 39 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_ignore_below {

#line 39 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 39 "ms_deisotope/_c/averagine.pxd"
  double ignore_below;

#line 39 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":40
 * 
 *     cpdef TheoreticalIsotopicPattern ignore_below(self, double ignore_below=*)
 *     cpdef TheoreticalIsotopicPattern truncate_after(self, double truncate_after=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern shift(self, double mz)
 *     cpdef TheoreticalIsotopicPattern scale(self, list experimental_distribution, str method=*)
 */

#line 40 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_truncate_after {

#line 40 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 40 "ms_deisotope/_c/averagine.pxd"
  double truncate_after;

#line 40 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":42
 *     cpdef TheoreticalIsotopicPattern truncate_after(self, double truncate_after=*)
 *     cpdef TheoreticalIsotopicPattern shift(self, double mz)
 *     cpdef TheoreticalIsotopicPattern scale(self, list experimental_distribution, str method=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern scale_raw(self, double scale_factor)
 *     cpdef double drop_last_peak(self)
 */

#line 42 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_scale {

#line 42 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 42 "ms_deisotope/_c/averagine.pxd"
  PyObject *method;

#line 42 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":53
 * 
 *     @cython.final
 *     cdef inline TheoreticalIsotopicPattern _scale(self, list experimental_distribution, str method=*)             # <<<<<<<<<<<<<<
 * 
 *     @cython.final
 */

#line 53 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern__scale {

#line 53 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 53 "ms_deisotope/_c/averagine.pxd"
  PyObject *method;

#line 53 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":88
 *     cdef double _make_cache_key(self, double mz)
 * 
 *     cdef TheoreticalIsotopicPattern has_mz_charge_pair(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 * 
 */

#line 88 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_has_mz_charge_pair {

#line 88 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 88 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 88 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 88 "ms_deisotope/_c/averagine.pxd"
  double truncate_after;

#line 88 "ms_deisotope/_c/averagine.pxd"
  double ignore_below;

#line 88 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":89
 * 
 *     cdef TheoreticalIsotopicPattern has_mz_charge_pair(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)
 *     cpdef TheoreticalIsotopicPattern isotopic_cluster(self, double mz, int charge=*, double charge_carrier=*, double truncate_after=*, double ignore_below=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 89 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_isotopic_cluster {

#line 89 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 89 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 89 "ms_deisotope/_c/averagine.pxd"
  double charge_carrier;

#line 89 "ms_deisotope/_c/averagine.pxd"
  double truncate_after;

#line 89 "ms_deisotope/_c/averagine.pxd"
  double ignore_below;

#line 89 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":92
 * 
 * 
 * cpdef double isotopic_shift(int charge=*)             # <<<<<<<<<<<<<<
 * 
 * cpdef TheoreticalIsotopicPattern poisson_approximate(double mass, size_t n_peaks, double lambda_factor=*, int charge=*)
 */

#line 92 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_isotopic_shift {

#line 92 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 92 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 92 "ms_deisotope/_c/averagine.pxd"
};

/* "ms_deisotope/_c/averagine.pxd":94
 * cpdef double isotopic_shift(int charge=*)
 * 
 * cpdef TheoreticalIsotopicPattern poisson_approximate(double mass, size_t n_peaks, double lambda_factor=*, int charge=*)             # <<<<<<<<<<<<<<
 */

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_poisson_approximate {

#line 94 "ms_deisotope/_c/averagine.pxd"
  int __pyx_n;

#line 94 "ms_deisotope/_c/averagine.pxd"
  double lambda_factor;

#line 94 "ms_deisotope/_c/averagine.pxd"
  int charge;

#line 94 "ms_deisotope/_c/averagine.pxd"
};

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges;

#line 94 "ms_deisotope/_c/averagine.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base__get_all_peak_charge_pairs;

/* "ms_deisotope/_c/deconvoluter_base.pxd":42
 *     cdef FittedPeak _has_peak(self, double mz, double error_tolerance)
 * 
 *     cpdef list match_theoretical_isotopic_distribution(self, list theoretical_distribution, double error_tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef scale_theoretical_distribution(self, TheoreticalIsotopicPattern theoretical_distribution, list experimental_distribution)
 *     cpdef IsotopicFitRecord _evaluate_theoretical_distribution(self, list experimental, TheoreticalIsotopicPattern theoretical, FittedPeak peak, int charge)
 */

#line 42 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution {

#line 42 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 42 "ms_deisotope/_c/deconvoluter_base.pxd"
  double error_tolerance;

#line 42 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":45
 *     cpdef scale_theoretical_distribution(self, TheoreticalIsotopicPattern theoretical_distribution, list experimental_distribution)
 *     cpdef IsotopicFitRecord _evaluate_theoretical_distribution(self, list experimental, TheoreticalIsotopicPattern theoretical, FittedPeak peak, int charge)
 *     cpdef subtraction(self, TheoreticalIsotopicPattern isotopic_cluster, double error_tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef list fit_incremental_truncation(self, IsotopicFitRecord seed_fit, double lower_bound)
 * 
 */

#line 45 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction {

#line 45 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 45 "ms_deisotope/_c/deconvoluter_base.pxd"
  double error_tolerance;

#line 45 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":50
 *     cpdef bint _check_fit(self, IsotopicFitRecord fit)
 * 
 *     cpdef list _find_next_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)
 * 
 */

#line 50 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak {

#line 50 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 50 "ms_deisotope/_c/deconvoluter_base.pxd"
  int step;

#line 50 "ms_deisotope/_c/deconvoluter_base.pxd"
  double tolerance;

#line 50 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":51
 * 
 *     cpdef list _find_next_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)
 *     cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)
 */

#line 51 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak {

#line 51 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 51 "ms_deisotope/_c/deconvoluter_base.pxd"
  int step;

#line 51 "ms_deisotope/_c/deconvoluter_base.pxd"
  double tolerance;

#line 51 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":53
 *     cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=*, double tolerance=*)
 * 
 *     cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 *     cdef int _find_previous_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)
 * 
 */

#line 53 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace {

#line 53 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 53 "ms_deisotope/_c/deconvoluter_base.pxd"
  int step;

#line 53 "ms_deisotope/_c/deconvoluter_base.pxd"
  double tolerance;

#line 53 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":54
 * 
 *     cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)
 *     cdef int _find_previous_putative_peak_inplace(self, double mz, int charge, set result, int step=*, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 54 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace {

#line 54 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 54 "ms_deisotope/_c/deconvoluter_base.pxd"
  int step;

#line 54 "ms_deisotope/_c/deconvoluter_base.pxd"
  double tolerance;

#line 54 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":61
 *         public AveragineCache averagine
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,             # <<<<<<<<<<<<<<
 *                                                          double charge_carrier=*, double truncate_after=*,
 *                                                          double ignore_below=*)
 */

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution {

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
  double charge_carrier;

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
  double truncate_after;

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
  double ignore_below;

#line 61 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":65
 *                                                          double ignore_below=*)
 * 
 *     cpdef set _fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=*, double truncate_after=*,             # <<<<<<<<<<<<<<
 *                                     double ignore_below=*)
 * 
 */

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges {

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
  double charge_carrier;

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
  double truncate_after;

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
  double ignore_below;

#line 65 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":73
 *         public list averagines
 * 
 *     cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,             # <<<<<<<<<<<<<<
 *                                                          AveragineCache  averagine, double charge_carrier=*, double truncate_after=*,
 *                                                          double ignore_below=*)
 */

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution {

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
  double charge_carrier;

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
  double truncate_after;

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
  double ignore_below;

#line 73 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":76
 *                                                          AveragineCache  averagine, double charge_carrier=*, double truncate_after=*,
 *                                                          double ignore_below=*)
 *     cpdef set _fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=*, double truncate_after=*,             # <<<<<<<<<<<<<<
 *                                     double ignore_below=*)
 * 
 */

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges {

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
  double charge_carrier;

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
  double truncate_after;

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
  double ignore_below;

#line 76 "ms_deisotope/_c/deconvoluter_base.pxd"
};

/* "ms_deisotope/_c/deconvoluter_base.pxd":83
 * 
 * 
 * cpdef set _get_all_peak_charge_pairs(DeconvoluterBase self, FittedPeak peak, double error_tolerance=*,             # <<<<<<<<<<<<<<
 *                                      object charge_range=*,
 *                                      int left_search_limit=*, int right_search_limit=*,
 */

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base__get_all_peak_charge_pairs {

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  int __pyx_n;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  double error_tolerance;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *charge_range;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  int left_search_limit;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  int right_search_limit;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  int recalculate_starting_peak;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
  int use_quick_charge;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak;

#line 83 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference;

/* "ms_deisotope/_c/scoring.pxd":29
 * 
 *     @staticmethod
 *     cdef IsotopicFitRecord _create(FittedPeak seed_peak, double score, int charge, TheoreticalIsotopicPattern theoretical,             # <<<<<<<<<<<<<<
 *                                    list experimental, object data=*, int missed_peaks=*)
 * 
 */

#line 29 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create {

#line 29 "ms_deisotope/_c/scoring.pxd"
  int __pyx_n;

#line 29 "ms_deisotope/_c/scoring.pxd"
  PyObject *data;

#line 29 "ms_deisotope/_c/scoring.pxd"
  int missed_peaks;

#line 29 "ms_deisotope/_c/scoring.pxd"
};

/* "ms_deisotope/_c/scoring.pxd":75
 *     cdef public double mass_error_tolerance
 * 
 *     cdef double score_peak(self, FittedPeak obs, TheoreticalPeak theo, double mass_error_tolerance=*, double minimum_signal_to_noise=*) nogil             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 75 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak {

#line 75 "ms_deisotope/_c/scoring.pxd"
  int __pyx_n;

#line 75 "ms_deisotope/_c/scoring.pxd"
  double mass_error_tolerance;

#line 75 "ms_deisotope/_c/scoring.pxd"
  double minimum_signal_to_noise;

#line 75 "ms_deisotope/_c/scoring.pxd"
};

/* "ms_deisotope/_c/scoring.pxd":92
 *     cdef public PeakSet peaklist
 * 
 *     cdef double detect_interference(self, list experimental_peaks, double lower=*, double upper=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference {

#line 92 "ms_deisotope/_c/scoring.pxd"
  int __pyx_n;

#line 92 "ms_deisotope/_c/scoring.pxd"
  double lower;

#line 92 "ms_deisotope/_c/scoring.pxd"
  double upper;

#line 92 "ms_deisotope/_c/scoring.pxd"
};

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet__has_peak;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_has_peak;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_all_peaks_for;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_cell;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_list;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_t;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_t;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_all_peaks_for;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_has_peak;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__has_peak;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__all_peaks_for;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_has_peak;

#line 92 "ms_deisotope/_c/scoring.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_all_peaks_for;

/* "ms_deisotope/_c/peak_set.pxd":138
 * 
 * 
 * cdef enum CPeaksFlags:             # <<<<<<<<<<<<<<
 *     borrowing = 0
 *     owns_peaks = 1
 */

#line 138 "ms_deisotope/_c/peak_set.pxd"
enum __pyx_t_12ms_deisotope_2_c_8peak_set_CPeaksFlags {

#line 138 "ms_deisotope/_c/peak_set.pxd"
  __pyx_e_12ms_deisotope_2_c_8peak_set_borrowing = 0,

#line 138 "ms_deisotope/_c/peak_set.pxd"
  __pyx_e_12ms_deisotope_2_c_8peak_set_owns_peaks = 1,

#line 138 "ms_deisotope/_c/peak_set.pxd"
  __pyx_e_12ms_deisotope_2_c_8peak_set_owns_index = 2,

#line 138 "ms_deisotope/_c/peak_set.pxd"
  __pyx_e_12ms_deisotope_2_c_8peak_set_owns_peaks_and_index = 3

#line 138 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":87
 *     cpdef DeconvolutedPeakSet copy(self)
 * 
 *     cdef DeconvolutedPeak _has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef DeconvolutedPeak has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)
 */

#line 87 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet__has_peak {

#line 87 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 87 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 87 "ms_deisotope/_c/peak_set.pxd"
  int use_mz;

#line 87 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":89
 *     cdef DeconvolutedPeak _has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)
 * 
 *     cpdef DeconvolutedPeak has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef tuple all_peaks_for(self, double neutral_mass, double tolerance=*)
 */

#line 89 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_has_peak {

#line 89 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 89 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 89 "ms_deisotope/_c/peak_set.pxd"
  int use_mz;

#line 89 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":91
 *     cpdef DeconvolutedPeak has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)
 * 
 *     cpdef tuple all_peaks_for(self, double neutral_mass, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef DeconvolutedPeak _get_nearest_peak(self, double neutral_mass, double* errout)
 */

#line 91 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_all_peaks_for {

#line 91 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 91 "ms_deisotope/_c/peak_set.pxd"
  double tolerance;

#line 91 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":118
 * cdef size_t INTERVAL_INDEX_SIZE
 * 
 * cdef struct index_cell:             # <<<<<<<<<<<<<<
 *     double center_value
 *     size_t start
 */

#line 118 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_cell {

#line 118 "ms_deisotope/_c/peak_set.pxd"
  double center_value;

#line 118 "ms_deisotope/_c/peak_set.pxd"
  size_t start;

#line 118 "ms_deisotope/_c/peak_set.pxd"
  size_t end;

#line 118 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":124
 * 
 * 
 * cdef struct index_list:             # <<<<<<<<<<<<<<
 *     index_cell* index
 *     size_t size
 */

#line 124 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_list {

#line 124 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_cell *index;

#line 124 "ms_deisotope/_c/peak_set.pxd"
  size_t size;

#line 124 "ms_deisotope/_c/peak_set.pxd"
  double low;

#line 124 "ms_deisotope/_c/peak_set.pxd"
  double high;

#line 124 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":131
 * 
 * 
 * cdef struct deconvoluted_peak_t:             # <<<<<<<<<<<<<<
 *     double neutral_mass
 *     double intensity
 */

#line 131 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_t {

#line 131 "ms_deisotope/_c/peak_set.pxd"
  double neutral_mass;

#line 131 "ms_deisotope/_c/peak_set.pxd"
  double intensity;

#line 131 "ms_deisotope/_c/peak_set.pxd"
  int charge;

#line 131 "ms_deisotope/_c/peak_set.pxd"
  unsigned int index;

#line 131 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":144
 *     owns_peaks_and_index = 3
 * 
 * cdef struct deconvoluted_peak_set_t:             # <<<<<<<<<<<<<<
 *     deconvoluted_peak_t* peaks
 *     double* mass_index
 */

#line 144 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_t {

#line 144 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_t *peaks;

#line 144 "ms_deisotope/_c/peak_set.pxd"
  double *mass_index;

#line 144 "ms_deisotope/_c/peak_set.pxd"
  size_t size;

#line 144 "ms_deisotope/_c/peak_set.pxd"
  enum __pyx_t_12ms_deisotope_2_c_8peak_set_CPeaksFlags flags;

#line 144 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":154
 * cdef int free_deconvoluted_peak_set_t(deconvoluted_peak_set_t* destination) nogil
 * 
 * cdef deconvoluted_peak_set_t deconvoluted_peak_set_all_peaks_for(deconvoluted_peak_set_t* self, double neutral_mass, double error_tolerance=*) nogil             # <<<<<<<<<<<<<<
 * cdef deconvoluted_peak_t* deconvoluted_peak_set_has_peak(deconvoluted_peak_set_t* self, double neutral_mass, double error_tolerance=*) nogil
 * 
 */

#line 154 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_all_peaks_for {

#line 154 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 154 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 154 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":155
 * 
 * cdef deconvoluted_peak_set_t deconvoluted_peak_set_all_peaks_for(deconvoluted_peak_set_t* self, double neutral_mass, double error_tolerance=*) nogil
 * cdef deconvoluted_peak_t* deconvoluted_peak_set_has_peak(deconvoluted_peak_set_t* self, double neutral_mass, double error_tolerance=*) nogil             # <<<<<<<<<<<<<<
 * 
 * cdef int deconvoluted_peak_eq(deconvoluted_peak_t* self, deconvoluted_peak_t* other) nogil
 */

#line 155 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_has_peak {

#line 155 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 155 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 155 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":167
 *     cdef deconvoluted_peak_t* getitem(self, size_t i) nogil
 * 
 *     cdef deconvoluted_peak_t* _has_peak(self, double neutral_mass, double error_tolerance=*) nogil             # <<<<<<<<<<<<<<
 *     cdef deconvoluted_peak_set_t _all_peaks_for(self, double neutral_mass, double error_tolerance=*) nogil
 * 
 */

#line 167 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__has_peak {

#line 167 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 167 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 167 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":168
 * 
 *     cdef deconvoluted_peak_t* _has_peak(self, double neutral_mass, double error_tolerance=*) nogil
 *     cdef deconvoluted_peak_set_t _all_peaks_for(self, double neutral_mass, double error_tolerance=*) nogil             # <<<<<<<<<<<<<<
 * 
 *     cpdef object has_peak(self, double neutral_mass, double error_tolerance=*)
 */

#line 168 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__all_peaks_for {

#line 168 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 168 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 168 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":170
 *     cdef deconvoluted_peak_set_t _all_peaks_for(self, double neutral_mass, double error_tolerance=*) nogil
 * 
 *     cpdef object has_peak(self, double neutral_mass, double error_tolerance=*)             # <<<<<<<<<<<<<<
 *     cpdef tuple all_peaks_for(self, double neutral_mass, double error_tolerance=*)
 */

#line 170 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_has_peak {

#line 170 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 170 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 170 "ms_deisotope/_c/peak_set.pxd"
};

/* "ms_deisotope/_c/peak_set.pxd":171
 * 
 *     cpdef object has_peak(self, double neutral_mass, double error_tolerance=*)
 *     cpdef tuple all_peaks_for(self, double neutral_mass, double error_tolerance=*)             # <<<<<<<<<<<<<<
 */

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_all_peaks_for {

#line 171 "ms_deisotope/_c/peak_set.pxd"
  int __pyx_n;

#line 171 "ms_deisotope/_c/peak_set.pxd"
  double error_tolerance;

#line 171 "ms_deisotope/_c/peak_set.pxd"
};

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__invalidate;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_split_sparse;

#line 171 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_clone;

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":98
 *     cpdef insert_node(self, LCMSFeatureTreeNode node)
 *     cpdef insert(self, PeakBase peak, double time)
 *     cpdef _invalidate(self, bint reaverage=*)             # <<<<<<<<<<<<<<
 *     cpdef list split_sparse(self, double delta_rt=*)
 *     cpdef LCMSFeature clone(self, deep=*, cls=*)
 */

#line 98 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__invalidate {

#line 98 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_n;

#line 98 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int reaverage;

#line 98 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":99
 *     cpdef insert(self, PeakBase peak, double time)
 *     cpdef _invalidate(self, bint reaverage=*)
 *     cpdef list split_sparse(self, double delta_rt=*)             # <<<<<<<<<<<<<<
 *     cpdef LCMSFeature clone(self, deep=*, cls=*)
 *     cpdef double max_intensity(self)
 */

#line 99 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_split_sparse {

#line 99 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_n;

#line 99 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double delta_rt;

#line 99 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":100
 *     cpdef _invalidate(self, bint reaverage=*)
 *     cpdef list split_sparse(self, double delta_rt=*)
 *     cpdef LCMSFeature clone(self, deep=*, cls=*)             # <<<<<<<<<<<<<<
 *     cpdef double max_intensity(self)
 * 
 */

#line 100 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_clone {

#line 100 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_n;

#line 100 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *deep;

#line 100 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *cls;

#line 100 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

/* "ms_peak_picker/_c/peak_set.pxd":2
 * 
 * cdef class PeakBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 2 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase {

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  PyObject_HEAD

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_vtab;

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  double mz;

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  double intensity;

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  double area;

#line 2 "ms_peak_picker/_c/peak_set.pxd"
};

#line 2 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":12
 * 
 * 
 * cdef class FittedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double signal_to_noise
 */

#line 12 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak {

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  double signal_to_noise;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  double full_width_at_half_max;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  double left_width;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  double right_width;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  long peak_count;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  long index;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
};

#line 12 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":29
 *                             double area)
 * 
 * cdef class PeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */

#line 29 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet {

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  PyObject_HEAD

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_vtab;

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  PyObject *peaks;

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  int indexed;

#line 29 "ms_peak_picker/_c/peak_set.pxd"
};

#line 29 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":83
 * 
 * 
 * cdef class PeakSetIndexed(PeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* mz_index
 */

#line 83 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed {

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet __pyx_base;

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  double *mz_index;

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_t_14ms_peak_picker_2_c_8peak_set_index_list *interval_index;

#line 83 "ms_peak_picker/_c/peak_set.pxd"
};

#line 83 "ms_peak_picker/_c/peak_set.pxd"


/* "brainpy/_c/composition.pxd":123
 * 
 * 
 * cdef class PyComposition(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* impl
 */

#line 123 "brainpy/_c/composition.pxd"
struct __pyx_obj_7brainpy_2_c_11composition_PyComposition {

#line 123 "brainpy/_c/composition.pxd"
  PyObject_HEAD

#line 123 "brainpy/_c/composition.pxd"
  struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition *__pyx_vtab;

#line 123 "brainpy/_c/composition.pxd"
  struct __pyx_t_7brainpy_2_c_11composition_Composition *impl;

#line 123 "brainpy/_c/composition.pxd"
  double cached_mass;

#line 123 "brainpy/_c/composition.pxd"
  int _clean;

#line 123 "brainpy/_c/composition.pxd"
};

#line 123 "brainpy/_c/composition.pxd"


/* "brainpy/_c/isotopic_distribution.pxd":63
 * 
 * @freelist(100000)
 * cdef class TheoreticalPeak(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 63 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak {

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  PyObject_HEAD

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_vtab;

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  double mz;

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  double intensity;

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  int charge;

#line 63 "brainpy/_c/isotopic_distribution.pxd"
};

#line 63 "brainpy/_c/isotopic_distribution.pxd"


/* "ms_peak_picker/_c/peak_index.pxd":5
 * 
 * 
 * cdef class PeakIndex(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public np.ndarray mz_array
 */

#line 5 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex {

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  PyObject_HEAD

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_vtab;

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  PyArrayObject *mz_array;

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  PyArrayObject *intensity_array;

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *peaks;

#line 5 "ms_peak_picker/_c/peak_index.pxd"
};

#line 5 "ms_peak_picker/_c/peak_index.pxd"


/* "ms_deisotope/_c/averagine.pxd":12
 * cdef dict scale_dict(dict data, double factor)
 * 
 * cdef class Averagine(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double base_mass
 */

#line 12 "ms_deisotope/_c/averagine.pxd"
struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine {

#line 12 "ms_deisotope/_c/averagine.pxd"
  PyObject_HEAD

#line 12 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine *__pyx_vtab;

#line 12 "ms_deisotope/_c/averagine.pxd"
  double base_mass;

#line 12 "ms_deisotope/_c/averagine.pxd"
  PyObject *base_composition;

#line 12 "ms_deisotope/_c/averagine.pxd"
};

#line 12 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/averagine.pxd":23
 * 
 * @cython.final
 * cdef class TheoreticalIsotopicPattern(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list peaklist
 */

#line 23 "ms_deisotope/_c/averagine.pxd"
struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern {

#line 23 "ms_deisotope/_c/averagine.pxd"
  PyObject_HEAD

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_vtab;

#line 23 "ms_deisotope/_c/averagine.pxd"
  PyObject *peaklist;

#line 23 "ms_deisotope/_c/averagine.pxd"
  double origin;

#line 23 "ms_deisotope/_c/averagine.pxd"
  double offset;

#line 23 "ms_deisotope/_c/averagine.pxd"
};

#line 23 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/averagine.pxd":79
 * 
 * @cython.final
 * cdef class AveragineCache(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public dict backend
 */

#line 79 "ms_deisotope/_c/averagine.pxd"
struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache {

#line 79 "ms_deisotope/_c/averagine.pxd"
  PyObject_HEAD

#line 79 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache *__pyx_vtab;

#line 79 "ms_deisotope/_c/averagine.pxd"
  PyObject *backend;

#line 79 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *averagine;

#line 79 "ms_deisotope/_c/averagine.pxd"
  double cache_truncation;

#line 79 "ms_deisotope/_c/averagine.pxd"
  int enabled;

#line 79 "ms_deisotope/_c/averagine.pxd"
};

#line 79 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":25
 * 
 * 
 * cdef class DeconvoluterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public bint use_subtraction
 */

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase {

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject_HEAD

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_vtab;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int use_subtraction;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *scale_method;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int merge_isobaric_peaks;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  double minimum_intensity;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *peaklist;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *scorer;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int verbose;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *_slice_cache;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":57
 * 
 * 
 * cdef class AveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public AveragineCache averagine
 */

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase {

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *averagine;

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":69
 * 
 * 
 * cdef class MultiAveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list averagines
 */

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase {

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *averagines;

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":91
 * 
 * 
 * cdef class ChargeIterator(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public int lower
 */

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator {

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject_HEAD

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_vtab;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int lower;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int upper;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int sign;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int *values;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  size_t size;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  size_t index;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/scoring.pxd":16
 * 
 * 
 * cdef class IsotopicFitRecord(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FittedPeak seed_peak
 */

#line 16 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord {

#line 16 "ms_deisotope/_c/scoring.pxd"
  PyObject_HEAD

#line 16 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_vtab;

#line 16 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *seed_peak;

#line 16 "ms_deisotope/_c/scoring.pxd"
  double score;

#line 16 "ms_deisotope/_c/scoring.pxd"
  int charge;

#line 16 "ms_deisotope/_c/scoring.pxd"
  PyObject *experimental;

#line 16 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *theoretical;

#line 16 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *monoisotopic_peak;

#line 16 "ms_deisotope/_c/scoring.pxd"
  int missed_peaks;

#line 16 "ms_deisotope/_c/scoring.pxd"
  PyObject *data;

#line 16 "ms_deisotope/_c/scoring.pxd"
  Py_hash_t _hash;

#line 16 "ms_deisotope/_c/scoring.pxd"
};

#line 16 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":38
 * 
 * 
 * cdef class FitSelectorBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double minimum_score
 */

#line 38 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase {

#line 38 "ms_deisotope/_c/scoring.pxd"
  PyObject_HEAD

#line 38 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_vtab;

#line 38 "ms_deisotope/_c/scoring.pxd"
  double minimum_score;

#line 38 "ms_deisotope/_c/scoring.pxd"
};

#line 38 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":50
 * 
 * 
 * cdef class IsotopicFitterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FitSelectorBase select
 */

#line 50 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase {

#line 50 "ms_deisotope/_c/scoring.pxd"
  PyObject_HEAD

#line 50 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_vtab;

#line 50 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *select;

#line 50 "ms_deisotope/_c/scoring.pxd"
};

#line 50 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":62
 * 
 * 
 * cdef class LeastSquaresFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

#line 62 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter {

#line 62 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 62 "ms_deisotope/_c/scoring.pxd"
};

#line 62 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":66
 * 
 * 
 * cdef class GTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

#line 66 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter {

#line 66 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 66 "ms_deisotope/_c/scoring.pxd"
};

#line 66 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":69
 *     pass
 * 
 * cdef class ScaledGTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

#line 69 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter {

#line 69 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 69 "ms_deisotope/_c/scoring.pxd"
};

#line 69 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":72
 *     pass
 * 
 * cdef class MSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef public double mass_error_tolerance
 * 
 */

#line 72 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter {

#line 72 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 72 "ms_deisotope/_c/scoring.pxd"
  double mass_error_tolerance;

#line 72 "ms_deisotope/_c/scoring.pxd"
};

#line 72 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":78
 * 
 * 
 * cdef class PenalizedMSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mass_error_tolerance
 */

#line 78 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter {

#line 78 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 78 "ms_deisotope/_c/scoring.pxd"
  double mass_error_tolerance;

#line 78 "ms_deisotope/_c/scoring.pxd"
  double penalty_factor;

#line 78 "ms_deisotope/_c/scoring.pxd"
};

#line 78 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":84
 * 
 * 
 * cdef class FunctionScorer(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object function
 */

#line 84 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer {

#line 84 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 84 "ms_deisotope/_c/scoring.pxd"
  PyObject *function;

#line 84 "ms_deisotope/_c/scoring.pxd"
};

#line 84 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":89
 * 
 * 
 * cdef class InterferenceDetection(object):             # <<<<<<<<<<<<<<
 *     cdef public PeakSet peaklist
 * 
 */

#line 89 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection {

#line 89 "ms_deisotope/_c/scoring.pxd"
  PyObject_HEAD

#line 89 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_vtab;

#line 89 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *peaklist;

#line 89 "ms_deisotope/_c/scoring.pxd"
};

#line 89 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":95
 * 
 * 
 * cdef class DistinctPatternFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public InterferenceDetection interference_detector
 */

#line 95 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter {

#line 95 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 95 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *interference_detector;

#line 95 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *g_test_scaled;

#line 95 "ms_deisotope/_c/scoring.pxd"
  double peak_count_scale;

#line 95 "ms_deisotope/_c/scoring.pxd"
  double domain_scale;

#line 95 "ms_deisotope/_c/scoring.pxd"
};

#line 95 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":102
 *         public double domain_scale
 * 
 * cdef class ScaledPenalizedMSDeconvFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double scale_factor
 */

#line 102 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter {

#line 102 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 102 "ms_deisotope/_c/scoring.pxd"
  double scale_factor;

#line 102 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *scorer;

#line 102 "ms_deisotope/_c/scoring.pxd"
};

#line 102 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":112
 * 
 * 
 * cdef class DotProductFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 */

#line 112 "ms_deisotope/_c/scoring.pxd"
struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter {

#line 112 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 112 "ms_deisotope/_c/scoring.pxd"
};

#line 112 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/peak_set.pxd":5
 * 
 * 
 * cdef class _Index:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public size_t neutral_mass
 */

#line 5 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index {

#line 5 "ms_deisotope/_c/peak_set.pxd"
  PyObject_HEAD

#line 5 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__Index *__pyx_vtab;

#line 5 "ms_deisotope/_c/peak_set.pxd"
  size_t neutral_mass;

#line 5 "ms_deisotope/_c/peak_set.pxd"
  size_t mz;

#line 5 "ms_deisotope/_c/peak_set.pxd"
};

#line 5 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":14
 * 
 * 
 * cdef class EnvelopePair:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 14 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair {

#line 14 "ms_deisotope/_c/peak_set.pxd"
  PyObject_HEAD

#line 14 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_EnvelopePair *__pyx_vtab;

#line 14 "ms_deisotope/_c/peak_set.pxd"
  double mz;

#line 14 "ms_deisotope/_c/peak_set.pxd"
  double intensity;

#line 14 "ms_deisotope/_c/peak_set.pxd"
};

#line 14 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":25
 * 
 * 
 * cdef class Envelope:             # <<<<<<<<<<<<<<
 *     cdef public tuple pairs
 * 
 */

#line 25 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope {

#line 25 "ms_deisotope/_c/peak_set.pxd"
  PyObject_HEAD

#line 25 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope *__pyx_vtab;

#line 25 "ms_deisotope/_c/peak_set.pxd"
  PyObject *pairs;

#line 25 "ms_deisotope/_c/peak_set.pxd"
};

#line 25 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":40
 * 
 * 
 * cdef class DeconvolutedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double neutral_mass
 */

#line 40 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak {

#line 40 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  double neutral_mass;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  double signal_to_noise;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index *_index;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  double full_width_at_half_max;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  int charge;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  double a_to_a2_ratio;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  double most_abundant_mass;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  double average_mass;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  double score;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *envelope;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *fit;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  int chosen_for_msms;

#line 40 "ms_deisotope/_c/peak_set.pxd"
};

#line 40 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":66
 * 
 * 
 * cdef class IonMobilityDeconvolutedPeak(DeconvolutedPeak):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double drift_time
 */

#line 66 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak {

#line 66 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak __pyx_base;

#line 66 "ms_deisotope/_c/peak_set.pxd"
  double drift_time;

#line 66 "ms_deisotope/_c/peak_set.pxd"
};

#line 66 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":71
 * 
 * 
 * cdef class DeconvolutedPeakSolution(DeconvolutedPeak):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object solution
 */

#line 71 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution {

#line 71 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak __pyx_base;

#line 71 "ms_deisotope/_c/peak_set.pxd"
  PyObject *solution;

#line 71 "ms_deisotope/_c/peak_set.pxd"
};

#line 71 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":75
 *         public object solution
 * 
 * cdef class DeconvolutedPeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */

#line 75 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet {

#line 75 "ms_deisotope/_c/peak_set.pxd"
  PyObject_HEAD

#line 75 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_vtab;

#line 75 "ms_deisotope/_c/peak_set.pxd"
  PyObject *peaks;

#line 75 "ms_deisotope/_c/peak_set.pxd"
  PyObject *_mz_ordered;

#line 75 "ms_deisotope/_c/peak_set.pxd"
  int indexed;

#line 75 "ms_deisotope/_c/peak_set.pxd"
};

#line 75 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":104
 * 
 * 
 * cdef class DeconvolutedPeakSetIndexed(DeconvolutedPeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* neutral_mass_array
 */

#line 104 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed {

#line 104 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet __pyx_base;

#line 104 "ms_deisotope/_c/peak_set.pxd"
  double *neutral_mass_array;

#line 104 "ms_deisotope/_c/peak_set.pxd"
  double *mz_array;

#line 104 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_index_list *interval_index;

#line 104 "ms_deisotope/_c/peak_set.pxd"
  size_t _size;

#line 104 "ms_deisotope/_c/peak_set.pxd"
};

#line 104 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":161
 * 
 * 
 * cdef class _CPeakSet:             # <<<<<<<<<<<<<<
 *     cdef:
 *         deconvoluted_peak_set_t* ptr
 */

#line 161 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet {

#line 161 "ms_deisotope/_c/peak_set.pxd"
  PyObject_HEAD

#line 161 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__CPeakSet *__pyx_vtab;

#line 161 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_t *ptr;

#line 161 "ms_deisotope/_c/peak_set.pxd"
};

#line 161 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":4
 * 
 * 
 * cdef class LCMSFeatureTreeList(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list roots
 */

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList {

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject_HEAD

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_vtab;

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *roots;

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *_node_id_hash;

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 4 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":23
 * 
 * 
 * cdef class LCMSFeatureTreeNodeBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double time
 */

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase {

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject_HEAD

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_vtab;

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double time;

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *members;

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *_most_abundant_member;

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *node_id;

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":44
 * 
 * 
 * cdef class LCMSFeatureTreeNode(LCMSFeatureTreeNodeBase):             # <<<<<<<<<<<<<<
 *     cdef:
 * 
 */

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode {

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase __pyx_base;

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double _mz;

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":56
 * 
 * 
 * cdef class FeatureBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public LCMSFeatureTreeList nodes
 */

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase {

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject_HEAD

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_vtab;

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *nodes;

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double _mz;

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double _start_time;

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double _end_time;

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 56 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":73
 * 
 * 
 * cdef class LCMSFeature(FeatureBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double _total_intensity
 */

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature {

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase __pyx_base;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double _total_intensity;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double _last_mz;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *_times;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *_peaks;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *adducts;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *used_as_adduct;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *created_at;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *feature_id;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *_peak_averager;

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":104
 * 
 * 
 * cdef class EmptyFeature(FeatureBase):             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef EmptyFeature _create(double mz)
 */

#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature {

#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase __pyx_base;

#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":109
 * 
 * 
 * cdef class FeatureSetIterator(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list features
 */

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator {

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject_HEAD

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_vtab;

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *features;

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *real_features;

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double start_time;

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double end_time;

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double last_time_seen;

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  size_t *index_list;

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":137
 * 
 * 
 * cdef class RunningWeightedAverage(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double current_mean
 */

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage {

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject_HEAD

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_vtab;

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double current_mean;

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  size_t current_count;

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double total_weight;

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 137 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":154
 * 
 * 
 * cdef class RunningWeightedAverageNeutralMass(RunningWeightedAverage):             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass {

#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage __pyx_base;

#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
};

#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":144
 *         return self.__class__, (self.roots,)
 * 
 *     def clone(self, deep=False):             # <<<<<<<<<<<<<<
 *         return LCMSFeatureTreeList(node.clone(deep=deep) for node in self)
 * 
 */

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone {

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject_HEAD

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_v_deep;

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self;

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
};

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":145
 * 
 *     def clone(self, deep=False):
 *         return LCMSFeatureTreeList(node.clone(deep=deep) for node in self)             # <<<<<<<<<<<<<<
 * 
 *     def unspool(self):
 */

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr {

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject_HEAD

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone *__pyx_outer_scope;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_v_node;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_0;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_ssize_t __pyx_t_1;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*__pyx_t_2)(PyObject *);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
};

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":196
 *         return node
 * 
 *     def _unspool_strip_children(self):             # <<<<<<<<<<<<<<
 *         node = self.__class__(
 *             self.time, list(self.members))
 */

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children {

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject_HEAD

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_v_node;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
};

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":476
 * 
 *     @property
 *     def peaks(self):             # <<<<<<<<<<<<<<
 *         if self._peaks is None:
 *             self._peaks = tuple(node.peaks for node in self.nodes)
 */

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__ {

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject_HEAD

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self;

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
};

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":478
 *     def peaks(self):
 *         if self._peaks is None:
 *             self._peaks = tuple(node.peaks for node in self.nodes)             # <<<<<<<<<<<<<<
 *         return self._peaks
 * 
 */

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr {

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject_HEAD

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__ *__pyx_outer_scope;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_v_node;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_0;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_ssize_t __pyx_t_1;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*__pyx_t_2)(PyObject *);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
};

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


/* "ms_peak_picker/_c/peak_set.pxd":2
 * 
 * cdef class PeakBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 2 "ms_peak_picker/_c/peak_set.pxd"


#line 2 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase {

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*clone)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch);

#line 2 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch);

#line 2 "ms_peak_picker/_c/peak_set.pxd"
};

#line 2 "ms_peak_picker/_c/peak_set.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase;

#line 2 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":12
 * 
 * 
 * cdef class FittedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double signal_to_noise
 */

#line 12 "ms_peak_picker/_c/peak_set.pxd"


#line 12 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak {

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  int (*_eq)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int __pyx_skip_dispatch);

#line 12 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_create)(double, double, double, double, double, double, long, long, double);

#line 12 "ms_peak_picker/_c/peak_set.pxd"
};

#line 12 "ms_peak_picker/_c/peak_set.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak;

#line 12 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":29
 *                             double area)
 * 
 * cdef class PeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */

#line 29 "ms_peak_picker/_c/peak_set.pxd"


#line 29 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet {

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*_create)(PyObject *);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  PyObject *(*reindex)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  size_t (*_index)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_get_nearest_peak)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double *);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_has_peak)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet__has_peak *__pyx_optional_args);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*clone)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*_between)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  int (*_between_bounds)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, double, size_t *, size_t *);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  PyObject *(*all_peaks_for)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_7PeakSet_all_peaks_for *__pyx_optional_args);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*getitem)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, size_t);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  size_t (*get_size)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
  int (*_eq)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 29 "ms_peak_picker/_c/peak_set.pxd"
};

#line 29 "ms_peak_picker/_c/peak_set.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet;

#line 29 "ms_peak_picker/_c/peak_set.pxd"


/* "ms_peak_picker/_c/peak_set.pxd":83
 * 
 * 
 * cdef class PeakSetIndexed(PeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* mz_index
 */

#line 83 "ms_peak_picker/_c/peak_set.pxd"


#line 83 "ms_peak_picker/_c/peak_set.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed {

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet __pyx_base;

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  PyObject *(*_allocate_index)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *, size_t, int __pyx_skip_dispatch);

#line 83 "ms_peak_picker/_c/peak_set.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*has_peak_hinted)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_8peak_set_14PeakSetIndexed_has_peak_hinted *__pyx_optional_args);

#line 83 "ms_peak_picker/_c/peak_set.pxd"
};

#line 83 "ms_peak_picker/_c/peak_set.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed *__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed;

#line 83 "ms_peak_picker/_c/peak_set.pxd"


/* "brainpy/_c/composition.pxd":123
 * 
 * 
 * cdef class PyComposition(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* impl
 */

#line 123 "brainpy/_c/composition.pxd"


#line 123 "brainpy/_c/composition.pxd"
struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition {

#line 123 "brainpy/_c/composition.pxd"
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *(*_create)(struct __pyx_t_7brainpy_2_c_11composition_Composition *);

#line 123 "brainpy/_c/composition.pxd"
  void (*_set_impl)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_t_7brainpy_2_c_11composition_Composition *, struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl *__pyx_optional_args);

#line 123 "brainpy/_c/composition.pxd"
  void (*_initialize_from_formula)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *);

#line 123 "brainpy/_c/composition.pxd"
  double (*mass)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  int (*__pyx___equality_pycomposition)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  int (*__pyx___equality_dict)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *(*copy)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  PyObject *(*update)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  PyObject *(*keys)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  PyObject *(*values)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  PyObject *(*items)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);

#line 123 "brainpy/_c/composition.pxd"
  PyObject *(*pop)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition_pop *__pyx_optional_args);

#line 123 "brainpy/_c/composition.pxd"
  __pyx_t_7brainpy_2_c_11composition_count_type (*getitem)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *);

#line 123 "brainpy/_c/composition.pxd"
  void (*setitem)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, __pyx_t_7brainpy_2_c_11composition_count_type);

#line 123 "brainpy/_c/composition.pxd"
  void (*increment)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, __pyx_t_7brainpy_2_c_11composition_count_type);

#line 123 "brainpy/_c/composition.pxd"
  void (*add_from)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *);

#line 123 "brainpy/_c/composition.pxd"
  void (*subtract_from)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *);

#line 123 "brainpy/_c/composition.pxd"
  void (*scale_by)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, long);

#line 123 "brainpy/_c/composition.pxd"
};

#line 123 "brainpy/_c/composition.pxd"
static struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition *__pyx_vtabptr_7brainpy_2_c_11composition_PyComposition;

#line 123 "brainpy/_c/composition.pxd"


/* "brainpy/_c/isotopic_distribution.pxd":63
 * 
 * @freelist(100000)
 * cdef class TheoreticalPeak(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 63 "brainpy/_c/isotopic_distribution.pxd"


#line 63 "brainpy/_c/isotopic_distribution.pxd"
struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak {

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *(*_create)(double, double, int);

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  int (*_eq)(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, int __pyx_skip_dispatch);

#line 63 "brainpy/_c/isotopic_distribution.pxd"
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *(*clone)(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, int __pyx_skip_dispatch);

#line 63 "brainpy/_c/isotopic_distribution.pxd"
};

#line 63 "brainpy/_c/isotopic_distribution.pxd"
static struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_vtabptr_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak;

#line 63 "brainpy/_c/isotopic_distribution.pxd"


/* "ms_peak_picker/_c/peak_index.pxd":5
 * 
 * 
 * cdef class PeakIndex(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public np.ndarray mz_array
 */

#line 5 "ms_peak_picker/_c/peak_index.pxd"


#line 5 "ms_peak_picker/_c/peak_index.pxd"
struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex {

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_has_peak)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, double, struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex__has_peak *__pyx_optional_args);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*_between)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, double, double);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  PyObject *(*all_peaks_for)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_14ms_peak_picker_2_c_10peak_index_9PeakIndex_all_peaks_for *__pyx_optional_args);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  size_t (*get_size)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *(*clone)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, int __pyx_skip_dispatch);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *(*copy)(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex *, int __pyx_skip_dispatch);

#line 5 "ms_peak_picker/_c/peak_index.pxd"
};

#line 5 "ms_peak_picker/_c/peak_index.pxd"
static struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex *__pyx_vtabptr_14ms_peak_picker_2_c_10peak_index_PeakIndex;

#line 5 "ms_peak_picker/_c/peak_index.pxd"


/* "ms_deisotope/_c/averagine.pxd":12
 * cdef dict scale_dict(dict data, double factor)
 * 
 * cdef class Averagine(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double base_mass
 */

#line 12 "ms_deisotope/_c/averagine.pxd"


#line 12 "ms_deisotope/_c/averagine.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine {

#line 12 "ms_deisotope/_c/averagine.pxd"
  PyObject *(*scale)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_scale *__pyx_optional_args);

#line 12 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*isotopic_cluster)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine_isotopic_cluster *__pyx_optional_args);

#line 12 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*_isotopic_cluster)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine *, double, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_9Averagine__isotopic_cluster *__pyx_optional_args);

#line 12 "ms_deisotope/_c/averagine.pxd"
};

#line 12 "ms_deisotope/_c/averagine.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine *__pyx_vtabptr_12ms_deisotope_2_c_9averagine_Averagine;

#line 12 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/averagine.pxd":23
 * 
 * @cython.final
 * cdef class TheoreticalIsotopicPattern(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list peaklist
 */

#line 23 "ms_deisotope/_c/averagine.pxd"


#line 23 "ms_deisotope/_c/averagine.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern {

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*_create)(PyObject *, double, double);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*clone)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  int (*_eq_inst)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*ignore_below)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_ignore_below *__pyx_optional_args);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*truncate_after)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_truncate_after *__pyx_optional_args);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*shift)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*scale)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern_scale *__pyx_optional_args);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*scale_raw)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  double (*drop_last_peak)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  double (*total)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*normalize)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*_scale)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_26TheoreticalIsotopicPattern__scale *__pyx_optional_args);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*clone_shift)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *(*get)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, Py_ssize_t);

#line 23 "ms_deisotope/_c/averagine.pxd"
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);

#line 23 "ms_deisotope/_c/averagine.pxd"
  double (*get_monoisotopic_mz)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);

#line 23 "ms_deisotope/_c/averagine.pxd"
  PyObject *(*get_processed_peaks)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);

#line 23 "ms_deisotope/_c/averagine.pxd"
  size_t (*basepeak_index)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  PyObject *(*incremental_truncation)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, double, int __pyx_skip_dispatch);

#line 23 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*clone_drop_last)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *);

#line 23 "ms_deisotope/_c/averagine.pxd"
};

#line 23 "ms_deisotope/_c/averagine.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *__pyx_vtabptr_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern;

#line 23 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/averagine.pxd":79
 * 
 * @cython.final
 * cdef class AveragineCache(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public dict backend
 */

#line 79 "ms_deisotope/_c/averagine.pxd"


#line 79 "ms_deisotope/_c/averagine.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache {

#line 79 "ms_deisotope/_c/averagine.pxd"
  double (*_make_cache_key)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, double);

#line 79 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*has_mz_charge_pair)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, double, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_has_mz_charge_pair *__pyx_optional_args);

#line 79 "ms_deisotope/_c/averagine.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *(*isotopic_cluster)(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_9averagine_14AveragineCache_isotopic_cluster *__pyx_optional_args);

#line 79 "ms_deisotope/_c/averagine.pxd"
};

#line 79 "ms_deisotope/_c/averagine.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache *__pyx_vtabptr_12ms_deisotope_2_c_9averagine_AveragineCache;

#line 79 "ms_deisotope/_c/averagine.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":25
 * 
 * 
 * cdef class DeconvoluterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public bint use_subtraction
 */

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"


#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase {

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *(*between)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, double, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*has_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, double, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*_has_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, double);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*match_theoretical_isotopic_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_match_theoretical_isotopic_distribution *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*scale_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*_evaluate_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*subtraction)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase_subtraction *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*fit_incremental_truncation)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, double, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int (*_check_fit)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*_find_next_putative_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*_find_previous_putative_peak)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int (*_find_next_putative_peak_inplace)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_next_putative_peak_inplace *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
  int (*_find_previous_putative_peak_inplace)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, double, int, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_16DeconvoluterBase__find_previous_putative_peak_inplace *__pyx_optional_args);

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase;

#line 25 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":57
 * 
 * 
 * cdef class AveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public AveragineCache averagine
 */

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"


#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase {

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*fit_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, double, int, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase_fit_theoretical_distribution *__pyx_optional_args);

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*_fit_peaks_at_charges)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *, PyObject *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_25AveragineDeconvoluterBase__fit_peaks_at_charges *__pyx_optional_args);

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase;

#line 57 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":69
 * 
 * 
 * cdef class MultiAveragineDeconvoluterBase(DeconvoluterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list averagines
 */

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"


#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase {

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase __pyx_base;

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*fit_theoretical_distribution)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, double, int, struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase_fit_theoretical_distribution *__pyx_optional_args);

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*_fit_peaks_at_charges)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *, PyObject *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_17deconvoluter_base_30MultiAveragineDeconvoluterBase__fit_peaks_at_charges *__pyx_optional_args);

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase;

#line 69 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/deconvoluter_base.pxd":91
 * 
 * 
 * cdef class ChargeIterator(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public int lower
 */

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"


#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator {

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *(*_create)(int, int);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *(*from_tuple)(PyObject *);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *(*from_quickcharge)(PyObject *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*sequence_from_quickcharge)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, int __pyx_skip_dispatch);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*make_sequence)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  void (*release_sequence)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  void (*set_bounds)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int, int);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int (*has_more)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  PyObject *(*reset)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  int (*get_next_value)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *, int __pyx_skip_dispatch);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *);

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
};

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator *__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator;

#line 91 "ms_deisotope/_c/deconvoluter_base.pxd"


/* "ms_deisotope/_c/scoring.pxd":16
 * 
 * 
 * cdef class IsotopicFitRecord(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FittedPeak seed_peak
 */

#line 16 "ms_deisotope/_c/scoring.pxd"


#line 16 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord {

#line 16 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*_create)(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, double, int, struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern *, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_17IsotopicFitRecord__create *__pyx_optional_args);

#line 16 "ms_deisotope/_c/scoring.pxd"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 16 "ms_deisotope/_c/scoring.pxd"
  int (*_ne)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 16 "ms_deisotope/_c/scoring.pxd"
  int (*_lt)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 16 "ms_deisotope/_c/scoring.pxd"
  int (*_gt)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 16 "ms_deisotope/_c/scoring.pxd"
};

#line 16 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitRecord;

#line 16 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":38
 * 
 * 
 * cdef class FitSelectorBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double minimum_score
 */

#line 38 "ms_deisotope/_c/scoring.pxd"


#line 38 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase {

#line 38 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*best)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, PyObject *, int __pyx_skip_dispatch);

#line 38 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *(*_best_from_set)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, PyObject *);

#line 38 "ms_deisotope/_c/scoring.pxd"
  int (*reject)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 38 "ms_deisotope/_c/scoring.pxd"
  int (*reject_score)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, double, int __pyx_skip_dispatch);

#line 38 "ms_deisotope/_c/scoring.pxd"
  int (*is_maximizing)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase *, int __pyx_skip_dispatch);

#line 38 "ms_deisotope/_c/scoring.pxd"
};

#line 38 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FitSelectorBase;

#line 38 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":50
 * 
 * 
 * cdef class IsotopicFitterBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public FitSelectorBase select
 */

#line 50 "ms_deisotope/_c/scoring.pxd"


#line 50 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase {

#line 50 "ms_deisotope/_c/scoring.pxd"
  double (*_evaluate)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, PyObject *, PyObject *, int __pyx_skip_dispatch);

#line 50 "ms_deisotope/_c/scoring.pxd"
  int (*reject)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord *, int __pyx_skip_dispatch);

#line 50 "ms_deisotope/_c/scoring.pxd"
  int (*reject_score)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, double, int __pyx_skip_dispatch);

#line 50 "ms_deisotope/_c/scoring.pxd"
  int (*is_maximizing)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, int __pyx_skip_dispatch);

#line 50 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *(*_configure)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *, struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase *, PyObject *, int __pyx_skip_dispatch);

#line 50 "ms_deisotope/_c/scoring.pxd"
};

#line 50 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase;

#line 50 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":62
 * 
 * 
 * cdef class LeastSquaresFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

#line 62 "ms_deisotope/_c/scoring.pxd"


#line 62 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_LeastSquaresFitter {

#line 62 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 62 "ms_deisotope/_c/scoring.pxd"
};

#line 62 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_LeastSquaresFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_LeastSquaresFitter;

#line 62 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":66
 * 
 * 
 * cdef class GTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

#line 66 "ms_deisotope/_c/scoring.pxd"


#line 66 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_GTestFitter {

#line 66 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 66 "ms_deisotope/_c/scoring.pxd"
};

#line 66 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_GTestFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_GTestFitter;

#line 66 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":69
 *     pass
 * 
 * cdef class ScaledGTestFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

#line 69 "ms_deisotope/_c/scoring.pxd"


#line 69 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledGTestFitter {

#line 69 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 69 "ms_deisotope/_c/scoring.pxd"
};

#line 69 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledGTestFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledGTestFitter;

#line 69 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":72
 *     pass
 * 
 * cdef class MSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef public double mass_error_tolerance
 * 
 */

#line 72 "ms_deisotope/_c/scoring.pxd"


#line 72 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MSDeconVFitter {

#line 72 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 72 "ms_deisotope/_c/scoring.pxd"
  double (*score_peak)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *, struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_14MSDeconVFitter_score_peak *__pyx_optional_args);

#line 72 "ms_deisotope/_c/scoring.pxd"
};

#line 72 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MSDeconVFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_MSDeconVFitter;

#line 72 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":78
 * 
 * 
 * cdef class PenalizedMSDeconVFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mass_error_tolerance
 */

#line 78 "ms_deisotope/_c/scoring.pxd"


#line 78 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter {

#line 78 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 78 "ms_deisotope/_c/scoring.pxd"
};

#line 78 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter;

#line 78 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":84
 * 
 * 
 * cdef class FunctionScorer(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object function
 */

#line 84 "ms_deisotope/_c/scoring.pxd"


#line 84 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FunctionScorer {

#line 84 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 84 "ms_deisotope/_c/scoring.pxd"
};

#line 84 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FunctionScorer *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FunctionScorer;

#line 84 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":89
 * 
 * 
 * cdef class InterferenceDetection(object):             # <<<<<<<<<<<<<<
 *     cdef public PeakSet peaklist
 * 
 */

#line 89 "ms_deisotope/_c/scoring.pxd"


#line 89 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection {

#line 89 "ms_deisotope/_c/scoring.pxd"
  double (*detect_interference)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection *, PyObject *, struct __pyx_opt_args_12ms_deisotope_2_c_7scoring_21InterferenceDetection_detect_interference *__pyx_optional_args);

#line 89 "ms_deisotope/_c/scoring.pxd"
};

#line 89 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_InterferenceDetection;

#line 89 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":95
 * 
 * 
 * cdef class DistinctPatternFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public InterferenceDetection interference_detector
 */

#line 95 "ms_deisotope/_c/scoring.pxd"


#line 95 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DistinctPatternFitter {

#line 95 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 95 "ms_deisotope/_c/scoring.pxd"
};

#line 95 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DistinctPatternFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DistinctPatternFitter;

#line 95 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":102
 *         public double domain_scale
 * 
 * cdef class ScaledPenalizedMSDeconvFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double scale_factor
 */

#line 102 "ms_deisotope/_c/scoring.pxd"


#line 102 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter {

#line 102 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 102 "ms_deisotope/_c/scoring.pxd"
  double (*_calculate_scale_factor)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet *, int __pyx_skip_dispatch);

#line 102 "ms_deisotope/_c/scoring.pxd"
  void (*scale_fitted_peaks)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *, PyObject *, double);

#line 102 "ms_deisotope/_c/scoring.pxd"
  void (*scale_theoretical_peaks)(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *, PyObject *, double);

#line 102 "ms_deisotope/_c/scoring.pxd"
};

#line 102 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter;

#line 102 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/scoring.pxd":112
 * 
 * 
 * cdef class DotProductFitter(IsotopicFitterBase):             # <<<<<<<<<<<<<<
 *     pass
 */

#line 112 "ms_deisotope/_c/scoring.pxd"


#line 112 "ms_deisotope/_c/scoring.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DotProductFitter {

#line 112 "ms_deisotope/_c/scoring.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase __pyx_base;

#line 112 "ms_deisotope/_c/scoring.pxd"
};

#line 112 "ms_deisotope/_c/scoring.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DotProductFitter *__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DotProductFitter;

#line 112 "ms_deisotope/_c/scoring.pxd"


/* "ms_deisotope/_c/peak_set.pxd":5
 * 
 * 
 * cdef class _Index:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public size_t neutral_mass
 */

#line 5 "ms_deisotope/_c/peak_set.pxd"


#line 5 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__Index {

#line 5 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index *(*_create)(size_t, size_t);

#line 5 "ms_deisotope/_c/peak_set.pxd"
};

#line 5 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__Index *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set__Index;

#line 5 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":14
 * 
 * 
 * cdef class EnvelopePair:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */

#line 14 "ms_deisotope/_c/peak_set.pxd"


#line 14 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_EnvelopePair {

#line 14 "ms_deisotope/_c/peak_set.pxd"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *, int __pyx_skip_dispatch);

#line 14 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *(*_create)(double, double);

#line 14 "ms_deisotope/_c/peak_set.pxd"
};

#line 14 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_EnvelopePair *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_EnvelopePair;

#line 14 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":25
 * 
 * 
 * cdef class Envelope:             # <<<<<<<<<<<<<<
 *     cdef public tuple pairs
 * 
 */

#line 25 "ms_deisotope/_c/peak_set.pxd"


#line 25 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope {

#line 25 "ms_deisotope/_c/peak_set.pxd"
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *);

#line 25 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *, size_t);

#line 25 "ms_deisotope/_c/peak_set.pxd"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *(*clone)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *, int __pyx_skip_dispatch);

#line 25 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *(*_create)(PyObject *);

#line 25 "ms_deisotope/_c/peak_set.pxd"
};

#line 25 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_Envelope;

#line 25 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":40
 * 
 * 
 * cdef class DeconvolutedPeak(PeakBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double neutral_mass
 */

#line 40 "ms_deisotope/_c/peak_set.pxd"


#line 40 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak {

#line 40 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase __pyx_base;

#line 40 "ms_deisotope/_c/peak_set.pxd"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *, int __pyx_skip_dispatch);

#line 40 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*_create_simple)(double, double, int, double, double, struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope *);

#line 40 "ms_deisotope/_c/peak_set.pxd"
};

#line 40 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak;

#line 40 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":66
 * 
 * 
 * cdef class IonMobilityDeconvolutedPeak(DeconvolutedPeak):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double drift_time
 */

#line 66 "ms_deisotope/_c/peak_set.pxd"


#line 66 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak {

#line 66 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak __pyx_base;

#line 66 "ms_deisotope/_c/peak_set.pxd"
};

#line 66 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak;

#line 66 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":71
 * 
 * 
 * cdef class DeconvolutedPeakSolution(DeconvolutedPeak):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object solution
 */

#line 71 "ms_deisotope/_c/peak_set.pxd"


#line 71 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution {

#line 71 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak __pyx_base;

#line 71 "ms_deisotope/_c/peak_set.pxd"
};

#line 71 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution;

#line 71 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":75
 *         public object solution
 * 
 * cdef class DeconvolutedPeakSet(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public tuple peaks
 */

#line 75 "ms_deisotope/_c/peak_set.pxd"


#line 75 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet {

#line 75 "ms_deisotope/_c/peak_set.pxd"
  PyObject *(*reindex)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *(*clone)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *(*copy)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, int __pyx_skip_dispatch);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*_has_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, double, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet__has_peak *__pyx_optional_args);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*has_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_has_peak *__pyx_optional_args);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  PyObject *(*all_peaks_for)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_19DeconvolutedPeakSet_all_peaks_for *__pyx_optional_args);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*_get_nearest_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, double, double *);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, size_t);

#line 75 "ms_deisotope/_c/peak_set.pxd"
  PyObject *(*getslice)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *, size_t, size_t);

#line 75 "ms_deisotope/_c/peak_set.pxd"
};

#line 75 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet;

#line 75 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":104
 * 
 * 
 * cdef class DeconvolutedPeakSetIndexed(DeconvolutedPeakSet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double* neutral_mass_array
 */

#line 104 "ms_deisotope/_c/peak_set.pxd"


#line 104 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed {

#line 104 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet __pyx_base;

#line 104 "ms_deisotope/_c/peak_set.pxd"
  void (*_build_index_arrays)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *);

#line 104 "ms_deisotope/_c/peak_set.pxd"
  int (*_interval_for)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *, double, double, size_t *, size_t *);

#line 104 "ms_deisotope/_c/peak_set.pxd"
};

#line 104 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed;

#line 104 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/peak_set.pxd":161
 * 
 * 
 * cdef class _CPeakSet:             # <<<<<<<<<<<<<<
 *     cdef:
 *         deconvoluted_peak_set_t* ptr
 */

#line 161 "ms_deisotope/_c/peak_set.pxd"


#line 161 "ms_deisotope/_c/peak_set.pxd"
struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__CPeakSet {

#line 161 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_t *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet *, size_t);

#line 161 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_t *(*_has_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet *, double, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__has_peak *__pyx_optional_args);

#line 161 "ms_deisotope/_c/peak_set.pxd"
  struct __pyx_t_12ms_deisotope_2_c_8peak_set_deconvoluted_peak_set_t (*_all_peaks_for)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet *, double, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet__all_peaks_for *__pyx_optional_args);

#line 161 "ms_deisotope/_c/peak_set.pxd"
  PyObject *(*has_peak)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_has_peak *__pyx_optional_args);

#line 161 "ms_deisotope/_c/peak_set.pxd"
  PyObject *(*all_peaks_for)(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet *, double, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_8peak_set_9_CPeakSet_all_peaks_for *__pyx_optional_args);

#line 161 "ms_deisotope/_c/peak_set.pxd"
};

#line 161 "ms_deisotope/_c/peak_set.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__CPeakSet *__pyx_vtabptr_12ms_deisotope_2_c_8peak_set__CPeakSet;

#line 161 "ms_deisotope/_c/peak_set.pxd"


/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":31
 * 
 * 
 * cdef class LCMSFeatureTreeList(object):             # <<<<<<<<<<<<<<
 *     def __init__(self, roots=None):
 *         if roots is None:
 */

#line 31 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 31 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList {

#line 31 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *(*_create)(PyObject *);

#line 31 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  void (*_invalidate)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *);

#line 31 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *(*_make_node)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *, double, PyObject *, int __pyx_skip_dispatch);

#line 31 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*find_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *, double, int __pyx_skip_dispatch);

#line 31 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *(*_find_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *, double, size_t *);

#line 31 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *, size_t);

#line 31 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *);

#line 31 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
};

#line 31 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList;

#line 31 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static CYTHON_INLINE struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList__find_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *, double, size_t *);

#line 31 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static CYTHON_INLINE struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_getitem(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *, size_t);

#line 31 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static CYTHON_INLINE size_t __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_get_size(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *);

#line 31 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":175
 * 
 * 
 * cdef class LCMSFeatureTreeNodeBase(object):             # <<<<<<<<<<<<<<
 * 
 *     def clone(self, deep=False):
 */

#line 175 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 175 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase {

#line 175 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double (*_total_intensity_members)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, int __pyx_skip_dispatch);

#line 175 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double (*max_intensity)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, int __pyx_skip_dispatch);

#line 175 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double (*total_intensity)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, int __pyx_skip_dispatch);

#line 175 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*_find_most_abundant_member)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *);

#line 175 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*_calculate_most_abundant_member)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, int __pyx_skip_dispatch);

#line 175 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*_recalculate)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, int __pyx_skip_dispatch);

#line 175 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, size_t);

#line 175 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t (*get_members_size)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *);

#line 175 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
};

#line 175 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase;

#line 175 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static CYTHON_INLINE size_t __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_get_members_size(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *);

#line 175 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":282
 * 
 * 
 * cdef class LCMSFeatureTreeNode(LCMSFeatureTreeNodeBase):             # <<<<<<<<<<<<<<
 *     def __init__(self, time, members=None):
 *         if members is None:
 */

#line 282 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 282 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode {

#line 282 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase __pyx_base;

#line 282 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, int __pyx_skip_dispatch);

#line 282 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int (*_ne)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, int __pyx_skip_dispatch);

#line 282 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*getpeak)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, size_t);

#line 282 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double (*get_mz)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *);

#line 282 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
};

#line 282 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode;

#line 282 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static CYTHON_INLINE struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_getpeak(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, size_t);

#line 282 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":337
 * 
 * @cython.freelist(10000000)
 * cdef class FeatureBase(object):             # <<<<<<<<<<<<<<
 *     def __init__(self, nodes):
 *         self.nodes = LCMSFeatureTreeList(nodes)
 */

#line 337 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 337 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase {

#line 337 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double (*get_mz)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *);

#line 337 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double (*get_neutral_mass)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *);

#line 337 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double (*get_start_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *);

#line 337 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double (*get_end_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *);

#line 337 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *);

#line 337 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*find_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *, double, int __pyx_skip_dispatch);

#line 337 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *(*_find_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *, double, size_t *);

#line 337 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
};

#line 337 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase;

#line 337 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static CYTHON_INLINE size_t __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_get_size(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *);

#line 337 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static CYTHON_INLINE struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase__find_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *, double, size_t *);

#line 337 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":373
 * 
 * 
 * cdef class LCMSFeature(FeatureBase):             # <<<<<<<<<<<<<<
 *     def __init__(self, nodes=None, adducts=None, used_as_adduct=None, feature_id=None):
 *         if nodes is None:
 */

#line 373 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 373 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature {

#line 373 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase __pyx_base;

#line 373 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int (*_eq)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, PyObject *, int __pyx_skip_dispatch);

#line 373 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int (*_ne)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, PyObject *, int __pyx_skip_dispatch);

#line 373 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int (*overlaps_in_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *, int __pyx_skip_dispatch);

#line 373 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int (*spans_in_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, double, int __pyx_skip_dispatch);

#line 373 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, size_t);

#line 373 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*_initialize_averager)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch);

#line 373 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*_update_from_averager)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch);

#line 373 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*_reaverage_and_update)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch);

#line 373 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*insert_node)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, int __pyx_skip_dispatch);

#line 373 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*insert)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, double, int __pyx_skip_dispatch);

#line 373 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*_invalidate)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__invalidate *__pyx_optional_args);

#line 373 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*split_sparse)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_split_sparse *__pyx_optional_args);

#line 373 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *(*clone)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_clone *__pyx_optional_args);

#line 373 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double (*max_intensity)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch);

#line 373 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
};

#line 373 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature;

#line 373 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":790
 * 
 * 
 * cdef class EmptyFeature(FeatureBase):             # <<<<<<<<<<<<<<
 *     def __init__(self, mz):
 *         FeatureBase.__init__(self, [])
 */

#line 790 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 790 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature {

#line 790 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase __pyx_base;

#line 790 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *(*_create)(double);

#line 790 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
};

#line 790 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature;

#line 790 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":828
 * 
 * 
 * cdef class FeatureSetIterator(object):             # <<<<<<<<<<<<<<
 *     def __init__(self, features, start_time=None, end_time=None):
 *         cdef:
 */

#line 828 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 828 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator {

#line 828 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *(*_create)(PyObject *);

#line 828 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *(*_create_with_threshold)(PyObject *, PyObject *, double);

#line 828 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int (*_initialize)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, PyObject *);

#line 828 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*init_indices)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, int __pyx_skip_dispatch);

#line 828 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double (*get_next_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, int __pyx_skip_dispatch);

#line 828 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double (*get_current_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, int __pyx_skip_dispatch);

#line 828 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int (*has_more)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, int __pyx_skip_dispatch);

#line 828 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*get_peaks_for_time)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, double, int __pyx_skip_dispatch);

#line 828 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*get_next_value)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, int __pyx_skip_dispatch);

#line 828 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t (*get_size)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *);

#line 828 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *(*getitem)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, size_t);

#line 828 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
};

#line 828 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator;

#line 828 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static CYTHON_INLINE size_t __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_get_size(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *);

#line 828 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static CYTHON_INLINE struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_getitem(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, size_t);

#line 828 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1075
 * 
 * 
 * cdef class RunningWeightedAverage(object):             # <<<<<<<<<<<<<<
 * 
 *     cpdef _initialize(self):
 */

#line 1075 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1075 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage {

#line 1075 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*reset)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, int __pyx_skip_dispatch);

#line 1075 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*_initialize)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, int __pyx_skip_dispatch);

#line 1075 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*add)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch);

#line 1075 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *(*update)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, PyObject *, int __pyx_skip_dispatch);

#line 1075 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int (*feed_from_feature)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch);

#line 1075 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *(*_create)(PyObject *);

#line 1075 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  void (*_update)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, PyObject *);

#line 1075 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
};

#line 1075 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage;

#line 1075 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1157
 * 
 * 
 * cdef class RunningWeightedAverageNeutralMass(RunningWeightedAverage):             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

#line 1157 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1157 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass {

#line 1157 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage __pyx_base;

#line 1157 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *(*_create)(PyObject *);

#line 1157 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
};

#line 1157 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if CYTHON_FAST_PYCALL
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif // CYTHON_FAST_PYCALL
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* DivInt[long].proto */
static CYTHON_INLINE long __Pyx_div_long(long, long);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* pop.proto */
static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L);
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L);
#define __Pyx_PyObject_Pop(L) (likely(PyList_CheckExact(L)) ?\
    __Pyx_PyList_Pop(L) : __Pyx__PyObject_Pop(L))
#else
#define __Pyx_PyList_Pop(L)  __Pyx__PyObject_Pop(L)
#define __Pyx_PyObject_Pop(L)  __Pyx__PyObject_Pop(L)
#endif

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* IsLittleEndian.proto */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void);

/* BufferFormatCheck.proto */
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts);
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type);

/* BufferGetAndValidate.proto */
#define __Pyx_GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack)\
    ((obj == Py_None || obj == NULL) ?\
    (__Pyx_ZeroBuffer(buf), 0) :\
    __Pyx__GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack))
static int  __Pyx__GetBufferAndValidate(Py_buffer* buf, PyObject* obj,
    __Pyx_TypeInfo* dtype, int flags, int nd, int cast, __Pyx_BufFmt_StackElem* stack);
static void __Pyx_ZeroBuffer(Py_buffer* buf);
static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info);
static Py_ssize_t __Pyx_minusones[] = { -1, -1, -1, -1, -1, -1, -1, -1 };
static Py_ssize_t __Pyx_zeros[] = { 0, 0, 0, 0, 0, 0, 0, 0 };

/* BufferFallbackError.proto */
static void __Pyx_RaiseBufferFallbackError(void);

/* BufferIndexError.proto */
static void __Pyx_RaiseBufferIndexError(int axis);

#define __Pyx_BufPtrCContig1d(type, buf, i0, s0) ((type)buf + i0)
/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* py_abs.proto */
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject *__Pyx_PyLong_AbsNeg(PyObject *num);
#define __Pyx_PyNumber_Absolute(x)\
    ((likely(PyLong_CheckExact(x))) ?\
         (likely(Py_SIZE(x) >= 0) ? (Py_INCREF(x), (x)) : __Pyx_PyLong_AbsNeg(x)) :\
         PyNumber_Absolute(x))
#else
#define __Pyx_PyNumber_Absolute(x)  PyNumber_Absolute(x)
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* IncludeStringH.proto */
#include <string.h>

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* pyobject_as_double.proto */
static double __Pyx__PyObject_AsDouble(PyObject* obj);
#if CYTHON_COMPILING_IN_PYPY
#define __Pyx_PyObject_AsDouble(obj)\
(likely(PyFloat_CheckExact(obj)) ? PyFloat_AS_DOUBLE(obj) :\
 likely(PyInt_CheckExact(obj)) ?\
 PyFloat_AsDouble(obj) : __Pyx__PyObject_AsDouble(obj))
#else
#define __Pyx_PyObject_AsDouble(obj)\
((likely(PyFloat_CheckExact(obj))) ?\
 PyFloat_AS_DOUBLE(obj) : __Pyx__PyObject_AsDouble(obj))
#endif

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* BufferStructDeclare.proto */
typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


/* Print.proto */
static int __Pyx_Print(PyObject*, PyObject *, int);
#if CYTHON_COMPILING_IN_PYPY || PY_MAJOR_VERSION >= 3
static PyObject* __pyx_print = 0;
static PyObject* __pyx_print_kwargs = 0;
#endif

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* PrintOne.proto */
static int __Pyx_PrintOne(PyObject* stream, PyObject *o);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* VoidPtrImport.proto */
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList__create(PyObject *__pyx_v_roots); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList__make_node(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, double __pyx_v_time, PyObject *__pyx_v_peaks, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList__invalidate(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_find_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, double __pyx_v_time, int __pyx_skip_dispatch); /* proto*/
static CYTHON_INLINE struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList__find_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, double __pyx_v_time, size_t *__pyx_v_indexout); /* proto*/
static CYTHON_INLINE struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_getitem(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, size_t __pyx_v_i); /* proto*/
static CYTHON_INLINE size_t __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_get_size(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase__total_intensity_members(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_max_intensity(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_total_intensity(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase__recalculate(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static CYTHON_INLINE size_t __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_get_members_size(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self); /* proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase__find_most_abundant_member(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase__calculate_most_abundant_member(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_getitem(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self, size_t __pyx_v_i); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode__recalculate(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_get_mz(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_self); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode__eq(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode__ne(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static CYTHON_INLINE struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_getpeak(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_self, size_t __pyx_v_i); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_get_mz(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_get_neutral_mass(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_get_start_time(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_get_end_time(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_find_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self, double __pyx_v_time, int __pyx_skip_dispatch); /* proto*/
static CYTHON_INLINE struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase__find_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self, double __pyx_v_time, size_t *__pyx_v_indexout); /* proto*/
static CYTHON_INLINE size_t __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_get_size(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__initialize_averager(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__update_from_averager(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__reaverage_and_update(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__invalidate(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__invalidate *__pyx_optional_args); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_get_mz(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_get_start_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_get_end_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_overlaps_in_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_interval, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_spans_in_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, double __pyx_v_time, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_split_sparse(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_split_sparse *__pyx_optional_args); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_clone(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_clone *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_insert_node(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_node, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_insert(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak, double __pyx_v_time, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__eq(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__ne(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_getitem(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, size_t __pyx_v_i); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_max_intensity(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature__create(double __pyx_v_mz); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_get_mz(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *__pyx_v_self); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator__initialize(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, PyObject *__pyx_v_features); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator__create(PyObject *__pyx_v_features); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator__create_with_threshold(PyObject *__pyx_v_features, PyObject *__pyx_v_theoretical_distribution, double __pyx_v_detection_threshold); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_init_indices(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_get_next_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_has_more(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static CYTHON_INLINE size_t __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_get_size(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self); /* proto*/
static CYTHON_INLINE struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_getitem(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, size_t __pyx_v_i); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_get_peaks_for_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, double __pyx_v_time, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_get_current_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_get_next_value(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage__initialize(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_reset(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage__create(PyObject *__pyx_v_peaks); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_add(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_feed_from_feature(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_feature, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_update(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self, PyObject *__pyx_v_iterable, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage__update(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self, PyObject *__pyx_v_peaks); /* proto*/
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass__create(PyObject *__pyx_v_peaks); /* proto*/
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass__create__pyx_wrap_1(PyObject *__pyx_v_peaks); /* proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_add(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak, int __pyx_skip_dispatch); /* proto*/

/* Module declarations from 'ms_peak_picker._c.peak_set' */
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase = 0;
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak = 0;
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet = 0;
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = 0;
static size_t *__pyx_vp_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE = 0;
#define __pyx_v_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE (*__pyx_vp_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE)

/* Module declarations from 'cython' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'cpython.version' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'libc.stdio' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'cpython.module' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cpython.tuple' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cpython.sequence' */

/* Module declarations from 'cpython.mapping' */

/* Module declarations from 'cpython.iterator' */

/* Module declarations from 'cpython.number' */

/* Module declarations from 'cpython.int' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.bool' */
static PyTypeObject *__pyx_ptype_7cpython_4bool_bool = 0;

/* Module declarations from 'cpython.long' */

/* Module declarations from 'cpython.float' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.complex' */
static PyTypeObject *__pyx_ptype_7cpython_7complex_complex = 0;

/* Module declarations from 'cpython.string' */

/* Module declarations from 'cpython.unicode' */

/* Module declarations from 'cpython.dict' */

/* Module declarations from 'cpython.instance' */

/* Module declarations from 'cpython.function' */

/* Module declarations from 'cpython.method' */

/* Module declarations from 'cpython.weakref' */

/* Module declarations from 'cpython.getargs' */

/* Module declarations from 'cpython.pythread' */

/* Module declarations from 'cpython.pystate' */

/* Module declarations from 'cpython.cobject' */

/* Module declarations from 'cpython.oldbuffer' */

/* Module declarations from 'cpython.set' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'cpython.pycapsule' */

/* Module declarations from 'cpython' */

/* Module declarations from 'brainpy._c.composition' */
static PyTypeObject *__pyx_ptype_7brainpy_2_c_11composition_PyComposition = 0;
static PyObject **__pyx_vp_7brainpy_2_c_11composition_nist_mass = 0;
#define __pyx_v_7brainpy_2_c_11composition_nist_mass (*__pyx_vp_7brainpy_2_c_11composition_nist_mass)
static double *__pyx_vp_7brainpy_2_c_11composition_PROTON = 0;
#define __pyx_v_7brainpy_2_c_11composition_PROTON (*__pyx_vp_7brainpy_2_c_11composition_PROTON)
static struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable **__pyx_vp_7brainpy_2_c_11composition__ElementTable = 0;
#define __pyx_v_7brainpy_2_c_11composition__ElementTable (*__pyx_vp_7brainpy_2_c_11composition__ElementTable)

/* Module declarations from 'brainpy._c.double_vector' */

/* Module declarations from 'brainpy._c.isotopic_constants' */
static size_t *__pyx_vp_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE = 0;
#define __pyx_v_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE (*__pyx_vp_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE)

/* Module declarations from 'brainpy._c.isotopic_distribution' */
static PyTypeObject *__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak = 0;

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_generic = 0;
static PyTypeObject *__pyx_ptype_5numpy_number = 0;
static PyTypeObject *__pyx_ptype_5numpy_integer = 0;
static PyTypeObject *__pyx_ptype_5numpy_signedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_unsignedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_inexact = 0;
static PyTypeObject *__pyx_ptype_5numpy_floating = 0;
static PyTypeObject *__pyx_ptype_5numpy_complexfloating = 0;
static PyTypeObject *__pyx_ptype_5numpy_flexible = 0;
static PyTypeObject *__pyx_ptype_5numpy_character = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE int __pyx_f_5numpy_import_array(void); /*proto*/

/* Module declarations from 'ms_peak_picker._c.peak_index' */
static PyTypeObject *__pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex = 0;

/* Module declarations from 'ms_deisotope._c.averagine' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache = 0;
static double *__pyx_vp_12ms_deisotope_2_c_9averagine_PROTON = 0;
#define __pyx_v_12ms_deisotope_2_c_9averagine_PROTON (*__pyx_vp_12ms_deisotope_2_c_9averagine_PROTON)

/* Module declarations from 'ms_deisotope._c.deconvoluter_base' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator = 0;

/* Module declarations from 'ms_deisotope._c.scoring' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter = 0;
static double *__pyx_vp_12ms_deisotope_2_c_7scoring_INFINITY = 0;
#define __pyx_v_12ms_deisotope_2_c_7scoring_INFINITY (*__pyx_vp_12ms_deisotope_2_c_7scoring_INFINITY)

/* Module declarations from 'ms_deisotope._c.peak_set' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set__Index = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_8peak_set__CPeakSet = 0;
static size_t *__pyx_vp_12ms_deisotope_2_c_8peak_set_INTERVAL_INDEX_SIZE = 0;
#define __pyx_v_12ms_deisotope_2_c_8peak_set_INTERVAL_INDEX_SIZE (*__pyx_vp_12ms_deisotope_2_c_8peak_set_INTERVAL_INDEX_SIZE)

/* Module declarations from 'ms_deisotope._c.feature_map.lcms_feature' */
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__ = 0;
static PyTypeObject *__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr = 0;
static PyObject *__pyx_v_12ms_deisotope_2_c_11feature_map_12lcms_feature__zeros = 0;
static double __pyx_v_12ms_deisotope_2_c_11feature_map_12lcms_feature_INF;
static PyObject *__pyx_v_12ms_deisotope_2_c_11feature_map_12lcms_feature_uid = 0;
static PyObject *__pyx_v_12ms_deisotope_2_c_11feature_map_12lcms_feature_sample = 0;
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_unpickle_FeatureBase__set_state(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *, PyObject *); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_unpickle_EmptyFeature__set_state(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *, PyObject *); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_unpickle_RunningWeightedAverage__set_state(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, PyObject *); /*proto*/
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_unpickle_RunningWeightedAverageNeutralMass__set_state(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *, PyObject *); /*proto*/
static __Pyx_TypeInfo __Pyx_TypeInfo_double = { "double", NULL, sizeof(double), { 0 }, 0, 'R', 0, 0 };
#define __Pyx_MODULE_NAME "ms_deisotope._c.feature_map.lcms_feature"
extern int __pyx_module_is_main_ms_deisotope___c__feature_map__lcms_feature;
int __pyx_module_is_main_ms_deisotope___c__feature_map__lcms_feature = 0;

/* Implementation of 'ms_deisotope._c.feature_map.lcms_feature' */
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_NotImplemented;
static PyObject *__pyx_builtin_zip;
static PyObject *__pyx_builtin_IndexError;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_StopIteration;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_ImportError;
static const char __pyx_k_[] = "";
static const char __pyx_k_eq[] = "_eq";
static const char __pyx_k_mz[] = "mz";
static const char __pyx_k_ne[] = "_ne";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_add[] = "add";
static const char __pyx_k_cls[] = "cls";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_inf[] = "inf";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_pop[] = "pop";
static const char __pyx_k_uid[] = "uid";
static const char __pyx_k_zip[] = "zip";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_deep[] = "deep";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_file[] = "file";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_node[] = "node";
static const char __pyx_k_peak[] = "peak";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_time[] = "time";
static const char __pyx_k_array[] = "array";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_clone[] = "clone";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_deque[] = "deque";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_merge[] = "merge";
static const char __pyx_k_nodes[] = "nodes";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_peaks[] = "peaks";
static const char __pyx_k_print[] = "print";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_reset[] = "reset";
static const char __pyx_k_roots[] = "roots";
static const char __pyx_k_s_r_d[] = "%s(%r, %d)";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_uid_2[] = "_uid";
static const char __pyx_k_zeros[] = "zeros";
static const char __pyx_k_argmax[] = "argmax";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_insert[] = "insert";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_random[] = "random";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_sample[] = "sample";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_adducts[] = "adducts";
static const char __pyx_k_float64[] = "float64";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_members[] = "members";
static const char __pyx_k_node_id[] = "node_id";
static const char __pyx_k_unspool[] = "unspool";
static const char __pyx_k_children[] = "children";
static const char __pyx_k_delta_rt[] = "delta_rt";
static const char __pyx_k_end_time[] = "end_time";
static const char __pyx_k_features[] = "features";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_has_more[] = "has_more";
static const char __pyx_k_iterable[] = "iterable";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_sample_2[] = "_sample";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_as_arrays[] = "as_arrays";
static const char __pyx_k_best_mz_0[] = "best_mz < 0";
static const char __pyx_k_find_time[] = "find_time";
static const char __pyx_k_make_node[] = "_make_node";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reaverage[] = "reaverage";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_copy_chunk[] = "_copy_chunk";
static const char __pyx_k_created_at[] = "created_at";
static const char __pyx_k_feature_id[] = "feature_id";
static const char __pyx_k_initialize[] = "_initialize";
static const char __pyx_k_invalidate[] = "_invalidate";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_start_time[] = "start_time";
static const char __pyx_k_FeatureBase[] = "FeatureBase";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_LCMSFeature[] = "LCMSFeature";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_insert_node[] = "insert_node";
static const char __pyx_k_recalculate[] = "recalculate";
static const char __pyx_k_EmptyFeature[] = "EmptyFeature";
static const char __pyx_k_init_indices[] = "init_indices";
static const char __pyx_k_node_id_hash[] = "node_id_hash";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_s_f_0_4f_s_d[] = "%s(%f, %0.4f %s|%d)";
static const char __pyx_k_split_sparse[] = "split_sparse";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_total_signal[] = "total_signal";
static const char __pyx_k_StopIteration[] = "StopIteration";
static const char __pyx_k_get_next_time[] = "get_next_time";
static const char __pyx_k_max_intensity[] = "max_intensity";
static const char __pyx_k_recalculate_2[] = "_recalculate";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_spans_in_time[] = "spans_in_time";
static const char __pyx_k_NotImplemented[] = "NotImplemented";
static const char __pyx_k_get_next_value[] = "get_next_value";
static const char __pyx_k_used_as_adduct[] = "used_as_adduct";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_total_intensity[] = "total_intensity";
static const char __pyx_k_Duplicate_Node_s[] = "Duplicate Node %s";
static const char __pyx_k_get_current_time[] = "get_current_time";
static const char __pyx_k_overlaps_in_time[] = "overlaps_in_time";
static const char __pyx_k_s_0_4f_0_2f_0_2f[] = "%s(%0.4f, %0.2f, %0.2f)";
static const char __pyx_k_feed_from_feature[] = "feed_from_feature";
static const char __pyx_k_FeatureSetIterator[] = "FeatureSetIterator";
static const char __pyx_k_build_node_id_hash[] = "_build_node_id_hash";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_get_peaks_for_time[] = "get_peaks_for_time";
static const char __pyx_k_ms_deisotope_utils[] = "ms_deisotope.utils";
static const char __pyx_k_LCMSFeatureTreeList[] = "LCMSFeatureTreeList";
static const char __pyx_k_LCMSFeatureTreeNode[] = "LCMSFeatureTreeNode";
static const char __pyx_k_NaN_produced_in_add[] = "NaN produced in add()";
static const char __pyx_k_initialize_averager[] = "_initialize_averager";
static const char __pyx_k_clone_locals_genexpr[] = "clone.<locals>.genexpr";
static const char __pyx_k_reaverage_and_update[] = "_reaverage_and_update";
static const char __pyx_k_update_from_averager[] = "_update_from_averager";
static const char __pyx_k_RunningWeightedAverage[] = "RunningWeightedAverage";
static const char __pyx_k_unspool_strip_children[] = "_unspool_strip_children";
static const char __pyx_k_LCMSFeatureTreeNodeBase[] = "LCMSFeatureTreeNodeBase";
static const char __pyx_k_total_intensity_members[] = "_total_intensity_members";
static const char __pyx_k_pyx_unpickle_FeatureBase[] = "__pyx_unpickle_FeatureBase";
static const char __pyx_k_pyx_unpickle_EmptyFeature[] = "__pyx_unpickle_EmptyFeature";
static const char __pyx_k_calculate_most_abundant_member[] = "_calculate_most_abundant_member";
static const char __pyx_k_pyx_unpickle_RunningWeightedAv[] = "__pyx_unpickle_RunningWeightedAverage";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_Empty_Feature_is_not_of_type_Emp[] = "Empty Feature is not of type EmptyFeature";
static const char __pyx_k_Failed_to_allocate_index_list_fo[] = "Failed to allocate index list for LCMS Feature Iterator";
static const char __pyx_k_FeatureSetIterator_d_d_0_3f_0_3f[] = "FeatureSetIterator(%d/%d, %0.3f-%0.3f)";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0[] = "Incompatible checksums (0x%x vs (0x9e63e75, 0x870ba3c, 0x5312e73) = (_end_time, _mz, _start_time, nodes))";
static const char __pyx_k_LCMSFeatureTreeList_d_nodes_0_2f[] = "LCMSFeatureTreeList(%d nodes, %0.2f-%0.2f)";
static const char __pyx_k_LCMSFeatureTreeNodeBase__unspool[] = "LCMSFeatureTreeNodeBase._unspool_strip_children";
static const char __pyx_k_LCMSFeature___get___locals_genex[] = "LCMSFeature.__get__.<locals>.genexpr";
static const char __pyx_k_RunningWeightedAverageNeutralMas[] = "RunningWeightedAverageNeutralMass";
static const char __pyx_k_ms_deisotope__c_feature_map_lcms[] = "ms_deisotope._c.feature_map.lcms_feature";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_pyx_unpickle_RunningWeightedAv_2[] = "__pyx_unpickle_RunningWeightedAverageNeutralMass";
static const char __pyx_k_self_index_list_cannot_be_conver[] = "self.index_list cannot be converted to a Python object for pickling";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_2[] = "Incompatible checksums (0x%x vs (0x69a8b2a, 0xbe93219, 0x16980be) = (current_count, current_mean, total_weight))";
static PyObject *__pyx_kp_s_;
static PyObject *__pyx_kp_s_Duplicate_Node_s;
static PyObject *__pyx_n_s_EmptyFeature;
static PyObject *__pyx_kp_s_Empty_Feature_is_not_of_type_Emp;
static PyObject *__pyx_kp_s_Failed_to_allocate_index_list_fo;
static PyObject *__pyx_n_s_FeatureBase;
static PyObject *__pyx_n_s_FeatureSetIterator;
static PyObject *__pyx_kp_s_FeatureSetIterator_d_d_0_3f_0_3f;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2;
static PyObject *__pyx_n_s_IndexError;
static PyObject *__pyx_n_s_LCMSFeature;
static PyObject *__pyx_n_s_LCMSFeatureTreeList;
static PyObject *__pyx_kp_s_LCMSFeatureTreeList_d_nodes_0_2f;
static PyObject *__pyx_n_s_LCMSFeatureTreeNode;
static PyObject *__pyx_n_s_LCMSFeatureTreeNodeBase;
static PyObject *__pyx_n_s_LCMSFeatureTreeNodeBase__unspool;
static PyObject *__pyx_n_s_LCMSFeature___get___locals_genex;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_kp_s_NaN_produced_in_add;
static PyObject *__pyx_n_s_NotImplemented;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_RunningWeightedAverage;
static PyObject *__pyx_n_s_RunningWeightedAverageNeutralMas;
static PyObject *__pyx_n_s_StopIteration;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_add;
static PyObject *__pyx_n_s_adducts;
static PyObject *__pyx_n_s_argmax;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_array;
static PyObject *__pyx_n_s_as_arrays;
static PyObject *__pyx_kp_s_best_mz_0;
static PyObject *__pyx_n_s_build_node_id_hash;
static PyObject *__pyx_n_s_calculate_most_abundant_member;
static PyObject *__pyx_n_s_children;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_clone;
static PyObject *__pyx_n_s_clone_locals_genexpr;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_cls;
static PyObject *__pyx_n_s_collections;
static PyObject *__pyx_n_s_copy_chunk;
static PyObject *__pyx_n_s_created_at;
static PyObject *__pyx_n_s_deep;
static PyObject *__pyx_n_s_delta_rt;
static PyObject *__pyx_n_s_deque;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_dtype;
static PyObject *__pyx_n_s_end;
static PyObject *__pyx_n_s_end_time;
static PyObject *__pyx_n_s_eq;
static PyObject *__pyx_n_s_feature_id;
static PyObject *__pyx_n_s_features;
static PyObject *__pyx_n_s_feed_from_feature;
static PyObject *__pyx_n_s_file;
static PyObject *__pyx_n_s_find_time;
static PyObject *__pyx_n_s_float64;
static PyObject *__pyx_n_s_genexpr;
static PyObject *__pyx_n_s_get_current_time;
static PyObject *__pyx_n_s_get_next_time;
static PyObject *__pyx_n_s_get_next_value;
static PyObject *__pyx_n_s_get_peaks_for_time;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_has_more;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_inf;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_n_s_init_indices;
static PyObject *__pyx_n_s_initialize;
static PyObject *__pyx_n_s_initialize_averager;
static PyObject *__pyx_n_s_insert;
static PyObject *__pyx_n_s_insert_node;
static PyObject *__pyx_n_s_invalidate;
static PyObject *__pyx_n_s_iterable;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_make_node;
static PyObject *__pyx_n_s_max_intensity;
static PyObject *__pyx_n_s_members;
static PyObject *__pyx_n_s_merge;
static PyObject *__pyx_n_s_ms_deisotope__c_feature_map_lcms;
static PyObject *__pyx_n_s_ms_deisotope_utils;
static PyObject *__pyx_n_s_mz;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_ne;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_n_s_node;
static PyObject *__pyx_n_s_node_id;
static PyObject *__pyx_n_s_node_id_hash;
static PyObject *__pyx_n_s_nodes;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_s_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_s_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_overlaps_in_time;
static PyObject *__pyx_n_s_peak;
static PyObject *__pyx_n_s_peaks;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_pop;
static PyObject *__pyx_n_s_print;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_EmptyFeature;
static PyObject *__pyx_n_s_pyx_unpickle_FeatureBase;
static PyObject *__pyx_n_s_pyx_unpickle_RunningWeightedAv;
static PyObject *__pyx_n_s_pyx_unpickle_RunningWeightedAv_2;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_random;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reaverage;
static PyObject *__pyx_n_s_reaverage_and_update;
static PyObject *__pyx_n_s_recalculate;
static PyObject *__pyx_n_s_recalculate_2;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_reset;
static PyObject *__pyx_n_s_roots;
static PyObject *__pyx_kp_s_s_0_4f_0_2f_0_2f;
static PyObject *__pyx_kp_s_s_f_0_4f_s_d;
static PyObject *__pyx_kp_s_s_r_d;
static PyObject *__pyx_n_s_sample;
static PyObject *__pyx_n_s_sample_2;
static PyObject *__pyx_kp_s_self_index_list_cannot_be_conver;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_spans_in_time;
static PyObject *__pyx_n_s_split_sparse;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_s_start_time;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_time;
static PyObject *__pyx_n_s_total_intensity;
static PyObject *__pyx_n_s_total_intensity_members;
static PyObject *__pyx_n_s_total_signal;
static PyObject *__pyx_n_s_uid;
static PyObject *__pyx_n_s_uid_2;
static PyObject *__pyx_n_s_unspool;
static PyObject *__pyx_n_s_unspool_strip_children;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_update_from_averager;
static PyObject *__pyx_n_s_used_as_adduct;
static PyObject *__pyx_n_s_zeros;
static PyObject *__pyx_n_s_zip;
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList___init__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, PyObject *__pyx_v_roots); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_2_make_node(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, double __pyx_v_time, PyObject *__pyx_v_peaks); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_4find_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, double __pyx_v_time); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_6_build_node_id_hash(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_12node_id_hash___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_8insert_node(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_node); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_10insert(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, PyObject *__pyx_v_time, PyObject *__pyx_v_peaks); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_12extend(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, PyObject *__pyx_v_iterable); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_14__getitem__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static Py_ssize_t __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_16__len__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_18__iter__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_20__reduce__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5clone_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_22clone(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, PyObject *__pyx_v_deep); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_24unspool(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_26common_nodes(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_28__repr__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5roots___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5roots_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5roots_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_13_node_id_hash___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_13_node_id_hash_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_13_node_id_hash_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_clone(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self, PyObject *__pyx_v_deep); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_2_unspool_strip_children(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_5__getstate__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7__setstate__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_9__reduce__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_11add(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self, PyObject *__pyx_v_node, PyObject *__pyx_v_recalculate); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_13_total_intensity_members(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_15max_intensity(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_17total_intensity(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_19_recalculate(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_21_calculate_most_abundant_member(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_5peaks___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self); /* proto */
static Py_hash_t __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_23__hash__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_4time___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_4time_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7members___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7members_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7members_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_21_most_abundant_member___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_21_most_abundant_member_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_21_most_abundant_member_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7node_id___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7node_id_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7node_id_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode___init__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_self, PyObject *__pyx_v_time, PyObject *__pyx_v_members); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_2_recalculate(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_2mz___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_4_eq(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_6_ne(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_8__richcmp__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_10__repr__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_3_mz___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_3_mz_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase___init__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self, PyObject *__pyx_v_nodes); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_2find_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self, double __pyx_v_time); /* proto */
static Py_ssize_t __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_4__len__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_6__iter__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_5nodes___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_5nodes_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_5nodes_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_3_mz___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_3_mz_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_11_start_time___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_11_start_time_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_9_end_time___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_9_end_time_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_8__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_10__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature___init__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_nodes, PyObject *__pyx_v_adducts, PyObject *__pyx_v_used_as_adduct, PyObject *__pyx_v_feature_id); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_2_initialize_averager(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_4_update_from_averager(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_reaverage_and_update(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_8invalidate(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_reaverage); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10_invalidate(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, int __pyx_v_reaverage); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_12total_signal___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_9intensity___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_12neutral_mass___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_2mz___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_5times___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_5peaks_7__get___genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_5peaks___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10start_time___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_8end_time___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_12overlaps_in_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_interval); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14spans_in_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, double __pyx_v_time); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_16as_arrays(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_dtype); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_18__repr__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_20_copy_chunk(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_nodes, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_22split_at(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_time); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_24between_times(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_start, PyObject *__pyx_v_end); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_26split_sparse(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, double __pyx_v_delta_rt); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_28truncate_before(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_time); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_30truncate_after(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_time); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_32clone(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_deep, PyObject *__pyx_v_cls); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_34insert_node(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_node); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_36insert(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak, double __pyx_v_time); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_38merge(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_9apex_time___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_40_eq(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_42_ne(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_44__richcmp__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_46__reduce__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static Py_hash_t __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_48__hash__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_50__getitem__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_52max_intensity(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_16_total_intensity___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_16_total_intensity_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_8_last_mz___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_8_last_mz_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_times___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_times_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_times_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_peaks___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_peaks_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_peaks_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_7adducts___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_7adducts_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_7adducts_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14used_as_adduct___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14used_as_adduct_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14used_as_adduct_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10created_at___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10created_at_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10created_at_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10feature_id___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10feature_id_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10feature_id_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14_peak_averager___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14_peak_averager_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14_peak_averager_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature___init__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *__pyx_v_self, PyObject *__pyx_v_mz); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_2mz___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_2__len__(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_4__getitem__(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_10start_time___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_8end_time___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_6__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_8__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator___init__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, PyObject *__pyx_v_features, PyObject *__pyx_v_start_time, PyObject *__pyx_v_end_time); /* proto */
static void __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_2__dealloc__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_4init_indices(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_6get_next_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8has_more(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_10get_index_list(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_12get_peaks_for_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, double __pyx_v_time); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_12current_time___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_14get_current_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_16get_next_value(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_18__next__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_20__iter__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_22__repr__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8features___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8features_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8features_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_13real_features___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_13real_features_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_13real_features_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_10start_time___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_10start_time_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8end_time___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8end_time_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_14last_time_seen___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_14last_time_seen_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_24__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_26__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage__initialize(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_2__init__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self, PyObject *__pyx_v_iterable); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_4reset(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_6add(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_8feed_from_feature(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_feature); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_10update(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self, PyObject *__pyx_v_iterable); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_12__repr__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_12current_mean___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_12current_mean_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_13current_count___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_13current_count_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_12total_weight___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self); /* proto */
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_12total_weight_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_14__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_16__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_add(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_2__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_4__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_unpickle_FeatureBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_2__pyx_unpickle_EmptyFeature(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_4__pyx_unpickle_RunningWeightedAverage(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_6__pyx_unpickle_RunningWeightedAverageNeutralMass(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyList_Type_pop = {0, &__pyx_n_s_pop, 0, 0, 0};
static PyObject *__pyx_float_1eneg_4;
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_23691454;
static PyObject *__pyx_int_87109235;
static PyObject *__pyx_int_110791466;
static PyObject *__pyx_int_141605436;
static PyObject *__pyx_int_166084213;
static PyObject *__pyx_int_199832089;
static PyObject *__pyx_k__3;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_codeobj__13;
static PyObject *__pyx_codeobj__15;
static PyObject *__pyx_codeobj__17;
static PyObject *__pyx_codeobj__19;
/* Late includes */

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":32
 * 
 * cdef class LCMSFeatureTreeList(object):
 *     def __init__(self, roots=None):             # <<<<<<<<<<<<<<
 *         if roots is None:
 *             roots = []
 */

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_roots = 0
#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_r;

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_roots,0};

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject* values[1] = {0};

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    values[0] = ((PyObject *)Py_None);

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(__pyx_kwds)) {

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      Py_ssize_t kw_args;

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {
        case  1: 
#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0: break;
        default: 
#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0:

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (kw_args > 0) {

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_roots);

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (value) { values[0] = value; kw_args--; }

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(kw_args > 0)) {

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 32, __pyx_L3_error)

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: 
#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  0: 
#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
break;
        default: 
#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_roots = values[0];

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); 
#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__PYX_ERR(0, 32, __pyx_L3_error)

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_error:;

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return -1;

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList___init__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_v_self), __pyx_v_roots);

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList___init__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, PyObject *__pyx_v_roots) {

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_r;

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_1;

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_2;

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__init__", 0);

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_roots);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":33
 * cdef class LCMSFeatureTreeList(object):
 *     def __init__(self, roots=None):
 *         if roots is None:             # <<<<<<<<<<<<<<
 *             roots = []
 *         self.roots = list(roots)
 */

#line 33 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = (__pyx_v_roots == Py_None);

#line 33 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 33 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":34
 *     def __init__(self, roots=None):
 *         if roots is None:
 *             roots = []             # <<<<<<<<<<<<<<
 *         self.roots = list(roots)
 *         self._node_id_hash = None
 */

#line 34 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 34, __pyx_L1_error)

#line 34 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 34 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF_SET(__pyx_v_roots, __pyx_t_3);

#line 34 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":33
 * cdef class LCMSFeatureTreeList(object):
 *     def __init__(self, roots=None):
 *         if roots is None:             # <<<<<<<<<<<<<<
 *             roots = []
 *         self.roots = list(roots)
 */

#line 33 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":35
 *         if roots is None:
 *             roots = []
 *         self.roots = list(roots)             # <<<<<<<<<<<<<<
 *         self._node_id_hash = None
 * 
 */

#line 35 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = PySequence_List(__pyx_v_roots); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 35, __pyx_L1_error)

#line 35 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 35 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 35 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_self->roots);

#line 35 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_v_self->roots);

#line 35 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->roots = ((PyObject*)__pyx_t_3);

#line 35 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":36
 *             roots = []
 *         self.roots = list(roots)
 *         self._node_id_hash = None             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

#line 36 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(Py_None);

#line 36 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(Py_None);

#line 36 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_self->_node_id_hash);

#line 36 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_v_self->_node_id_hash);

#line 36 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->_node_id_hash = ((PyObject*)Py_None);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":32
 * 
 * cdef class LCMSFeatureTreeList(object):
 *     def __init__(self, roots=None):             # <<<<<<<<<<<<<<
 *         if roots is None:
 *             roots = []
 */

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = -1;

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_roots);

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 32 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":39
 * 
 *     @staticmethod
 *     cdef LCMSFeatureTreeList _create(list roots):             # <<<<<<<<<<<<<<
 *         cdef:
 *             LCMSFeatureTreeList inst
 */

#line 39 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 39 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList__create(PyObject *__pyx_v_roots) {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_inst = 0
#line 39 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 39 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_r = NULL;

#line 39 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 39 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 39 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 39 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 39 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 39 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_create", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":42
 *         cdef:
 *             LCMSFeatureTreeList inst
 *         inst = LCMSFeatureTreeList.__new__(LCMSFeatureTreeList)             # <<<<<<<<<<<<<<
 *         inst.roots = list(roots)
 *         inst._node_id_hash = None
 */

#line 42 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList(((PyTypeObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 42, __pyx_L1_error)

#line 42 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));

#line 42 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_inst = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_t_1);

#line 42 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":43
 *             LCMSFeatureTreeList inst
 *         inst = LCMSFeatureTreeList.__new__(LCMSFeatureTreeList)
 *         inst.roots = list(roots)             # <<<<<<<<<<<<<<
 *         inst._node_id_hash = None
 *         return inst
 */

#line 43 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PySequence_List(__pyx_v_roots); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 43, __pyx_L1_error)

#line 43 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 43 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 43 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_inst->roots);

#line 43 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_v_inst->roots);

#line 43 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_inst->roots = ((PyObject*)__pyx_t_1);

#line 43 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":44
 *         inst = LCMSFeatureTreeList.__new__(LCMSFeatureTreeList)
 *         inst.roots = list(roots)
 *         inst._node_id_hash = None             # <<<<<<<<<<<<<<
 *         return inst
 * 
 */

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(Py_None);

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(Py_None);

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_inst->_node_id_hash);

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_v_inst->_node_id_hash);

#line 44 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_inst->_node_id_hash = ((PyObject*)Py_None);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":45
 *         inst.roots = list(roots)
 *         inst._node_id_hash = None
 *         return inst             # <<<<<<<<<<<<<<
 * 
 *     cpdef LCMSFeatureTreeNodeBase _make_node(self, double time, list peaks):
 */

#line 45 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 45 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_inst));

#line 45 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_v_inst;

#line 45 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":39
 * 
 *     @staticmethod
 *     cdef LCMSFeatureTreeList _create(list roots):             # <<<<<<<<<<<<<<
 *         cdef:
 *             LCMSFeatureTreeList inst
 */

#line 39 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 39 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 39 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 39 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 39 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList._create", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 39 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 39 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 39 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);

#line 39 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 39 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 39 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 39 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":47
 *         return inst
 * 
 *     cpdef LCMSFeatureTreeNodeBase _make_node(self, double time, list peaks):             # <<<<<<<<<<<<<<
 *         return LCMSFeatureTreeNode(time, peaks)
 * 
 */

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_3_make_node(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList__make_node(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, double __pyx_v_time, PyObject *__pyx_v_peaks, int __pyx_skip_dispatch) {

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_r = NULL;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_6;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_7 = NULL;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_make_node", 0);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_make_node); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L1_error)

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_3_make_node)) {

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 47, __pyx_L1_error)

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_5 = NULL;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_6 = 0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_5)) {

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_5);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_4, function);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __pyx_t_6 = 1;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #if CYTHON_FAST_PYCALL

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (PyFunction_Check(__pyx_t_4)) {

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_3, __pyx_v_peaks};

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 47, __pyx_L1_error)

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        } else

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #endif

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #if CYTHON_FAST_PYCCALL

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_3, __pyx_v_peaks};

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 47, __pyx_L1_error)

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        } else

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #endif

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        {

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 47, __pyx_L1_error)

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_GOTREF(__pyx_t_7);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (__pyx_t_5) {

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_GIVEREF(__pyx_t_3);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_t_3);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_INCREF(__pyx_v_peaks);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_GIVEREF(__pyx_v_peaks);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_peaks);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_3 = 0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 47, __pyx_L1_error)

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase))))) __PYX_ERR(0, 47, __pyx_L1_error)

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_t_2);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = 0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":48
 * 
 *     cpdef LCMSFeatureTreeNodeBase _make_node(self, double time, list peaks):
 *         return LCMSFeatureTreeNode(time, peaks)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _invalidate(self):
 */

#line 48 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 48 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 48, __pyx_L1_error)

#line 48 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 48 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 48, __pyx_L1_error)

#line 48 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 48 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 48 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);

#line 48 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_peaks);

#line 48 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_v_peaks);

#line 48 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_peaks);

#line 48 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 48 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 48, __pyx_L1_error)

#line 48 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 48 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 48 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_t_1);

#line 48 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 48 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":47
 *         return inst
 * 
 *     cpdef LCMSFeatureTreeNodeBase _make_node(self, double time, list peaks):             # <<<<<<<<<<<<<<
 *         return LCMSFeatureTreeNode(time, peaks)
 * 
 */

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_7);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList._make_node", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_3_make_node(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_2_make_node[] = "LCMSFeatureTreeList._make_node(self, double time, list peaks) -> LCMSFeatureTreeNodeBase";

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_3_make_node(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_time
#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyObject *__pyx_v_peaks = 0
#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_make_node (wrapper)", 0);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_time,&__pyx_n_s_peaks,0};

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject* values[2] = {0,0};

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(__pyx_kwds)) {

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      Py_ssize_t kw_args;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {
        case  2: 
#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0: break;
        default: 
#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0:

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_time)) != 0)) kw_args--;
        else 
#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  1:

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaks)) != 0)) kw_args--;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("_make_node", 1, 2, 2, 1); 
#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__PYX_ERR(0, 47, __pyx_L3_error)

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(kw_args > 0)) {

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_make_node") < 0)) __PYX_ERR(0, 47, __pyx_L3_error)

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L5_argtuple_error;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_time = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_time == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 47, __pyx_L3_error)

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_peaks = ((PyObject*)values[1]);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_make_node", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__PYX_ERR(0, 47, __pyx_L3_error)

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_error:;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList._make_node", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return NULL;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peaks), (&PyList_Type), 1, "peaks", 1))) __PYX_ERR(0, 47, __pyx_L1_error)
  __pyx_r = 
#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_2_make_node(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_v_self), __pyx_v_time, __pyx_v_peaks);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_2_make_node(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, double __pyx_v_time, PyObject *__pyx_v_peaks) {

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_make_node", 0);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList__make_node(__pyx_v_self, __pyx_v_time, __pyx_v_peaks, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L1_error)

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList._make_node", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":50
 *         return LCMSFeatureTreeNode(time, peaks)
 * 
 *     cdef void _invalidate(self):             # <<<<<<<<<<<<<<
 *         self._node_id_hash = None
 * 
 */

#line 50 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 50 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static void __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList__invalidate(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self) {

#line 50 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 50 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_invalidate", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":51
 * 
 *     cdef void _invalidate(self):
 *         self._node_id_hash = None             # <<<<<<<<<<<<<<
 * 
 *     cpdef tuple find_time(self, double time):
 */

#line 51 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(Py_None);

#line 51 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(Py_None);

#line 51 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_self->_node_id_hash);

#line 51 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_v_self->_node_id_hash);

#line 51 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->_node_id_hash = ((PyObject*)Py_None);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":50
 *         return LCMSFeatureTreeNode(time, peaks)
 * 
 *     cdef void _invalidate(self):             # <<<<<<<<<<<<<<
 *         self._node_id_hash = None
 * 
 */

#line 50 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 50 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 50 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 50 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":53
 *         self._node_id_hash = None
 * 
 *     cpdef tuple find_time(self, double time):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t indexout
 */

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5find_time(PyObject *__pyx_v_self, PyObject *__pyx_arg_time); /*proto*/

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_find_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, double __pyx_v_time, int __pyx_skip_dispatch) {
  size_t __pyx_v_indexout
#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_node = 0
#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("find_time", 0);

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_find_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 53, __pyx_L1_error)

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5find_time)) {

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 53, __pyx_L1_error)

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_5 = NULL;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_5)) {

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_5);

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_4, function);

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 53, __pyx_L1_error)

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (!(likely(PyTuple_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 53, __pyx_L1_error)

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = ((PyObject*)__pyx_t_2);

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = 0;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":57
 *             size_t indexout
 *             LCMSFeatureTreeNodeBase node
 *         node = self._find_time(time, &indexout)             # <<<<<<<<<<<<<<
 *         return node, indexout
 * 
 */

#line 57 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList__find_time(__pyx_v_self, __pyx_v_time, (&__pyx_v_indexout))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 57, __pyx_L1_error)

#line 57 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 57 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_node = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_t_1);

#line 57 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":58
 *             LCMSFeatureTreeNodeBase node
 *         node = self._find_time(time, &indexout)
 *         return node, indexout             # <<<<<<<<<<<<<<
 * 
 *     cdef LCMSFeatureTreeNodeBase _find_time(self, double time, size_t* indexout):
 */

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_indexout); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 58, __pyx_L1_error)

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 58, __pyx_L1_error)

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_node));

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_node));

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_node));

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = ((PyObject*)__pyx_t_2);

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = 0;

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":53
 *         self._node_id_hash = None
 * 
 *     cpdef tuple find_time(self, double time):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t indexout
 */

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList.find_time", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_node);

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5find_time(PyObject *__pyx_v_self, PyObject *__pyx_arg_time); /*proto*/

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_4find_time[] = "LCMSFeatureTreeList.find_time(self, double time) -> tuple";

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5find_time(PyObject *__pyx_v_self, PyObject *__pyx_arg_time) {
  double __pyx_v_time
#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("find_time (wrapper)", 0);

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  assert(__pyx_arg_time); {

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_time = __pyx_PyFloat_AsDouble(__pyx_arg_time); if (unlikely((__pyx_v_time == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 53, __pyx_L3_error)

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_error:;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList.find_time", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return NULL;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_4find_time(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_v_self), ((double)__pyx_v_time));

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_4find_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, double __pyx_v_time) {

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("find_time", 0);

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_find_time(__pyx_v_self, __pyx_v_time, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 53, __pyx_L1_error)

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList.find_time", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 53 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":60
 *         return node, indexout
 * 
 *     cdef LCMSFeatureTreeNodeBase _find_time(self, double time, size_t* indexout):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int lo, hi
 */

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static CYTHON_INLINE struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList__find_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, double __pyx_v_time, size_t *__pyx_v_indexout) {
  int __pyx_v_lo
#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  int __pyx_v_hi
#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  double __pyx_v_rt
#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_node = 0
#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  long __pyx_v_i
#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_r = NULL;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_2;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_4;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_find_time", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":65
 *             double rt
 *             LCMSFeatureTreeNodeBase node
 *         hi = PyList_GET_SIZE(self.roots)             # <<<<<<<<<<<<<<
 *         if hi == 0:
 *             raise ValueError()
 */

#line 65 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_v_self->roots;

#line 65 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 65 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_hi = PyList_GET_SIZE(__pyx_t_1);

#line 65 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":66
 *             LCMSFeatureTreeNodeBase node
 *         hi = PyList_GET_SIZE(self.roots)
 *         if hi == 0:             # <<<<<<<<<<<<<<
 *             raise ValueError()
 *         lo = 0
 */

#line 66 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = ((__pyx_v_hi == 0) != 0);

#line 66 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_t_2)) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":67
 *         hi = PyList_GET_SIZE(self.roots)
 *         if hi == 0:
 *             raise ValueError()             # <<<<<<<<<<<<<<
 *         lo = 0
 *         while lo != hi:
 */

#line 67 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_ValueError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 67, __pyx_L1_error)

#line 67 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 67 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 67 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 67 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __PYX_ERR(0, 67, __pyx_L1_error)

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":66
 *             LCMSFeatureTreeNodeBase node
 *         hi = PyList_GET_SIZE(self.roots)
 *         if hi == 0:             # <<<<<<<<<<<<<<
 *             raise ValueError()
 *         lo = 0
 */

#line 66 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":68
 *         if hi == 0:
 *             raise ValueError()
 *         lo = 0             # <<<<<<<<<<<<<<
 *         while lo != hi:
 *             i = (lo + hi) // 2
 */

#line 68 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_lo = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":69
 *             raise ValueError()
 *         lo = 0
 *         while lo != hi:             # <<<<<<<<<<<<<<
 *             i = (lo + hi) // 2
 *             node = <LCMSFeatureTreeNodeBase>PyList_GET_ITEM(self.roots, i)
 */

#line 69 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  while (1) {

#line 69 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = ((__pyx_v_lo != __pyx_v_hi) != 0);

#line 69 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (!__pyx_t_2) break;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":70
 *         lo = 0
 *         while lo != hi:
 *             i = (lo + hi) // 2             # <<<<<<<<<<<<<<
 *             node = <LCMSFeatureTreeNodeBase>PyList_GET_ITEM(self.roots, i)
 *             rt = node.time
 */

#line 70 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_i = __Pyx_div_long((__pyx_v_lo + __pyx_v_hi), 2);

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":71
 *         while lo != hi:
 *             i = (lo + hi) // 2
 *             node = <LCMSFeatureTreeNodeBase>PyList_GET_ITEM(self.roots, i)             # <<<<<<<<<<<<<<
 *             rt = node.time
 *             if rt == time:
 */

#line 71 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __pyx_v_self->roots;

#line 71 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 71 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_v_i);

#line 71 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 71 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_3);

#line 71 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 71 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_t_1));

#line 71 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":72
 *             i = (lo + hi) // 2
 *             node = <LCMSFeatureTreeNodeBase>PyList_GET_ITEM(self.roots, i)
 *             rt = node.time             # <<<<<<<<<<<<<<
 *             if rt == time:
 *                 indexout[0] = i
 */

#line 72 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = __pyx_v_node->time;

#line 72 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_rt = __pyx_t_4;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":73
 *             node = <LCMSFeatureTreeNodeBase>PyList_GET_ITEM(self.roots, i)
 *             rt = node.time
 *             if rt == time:             # <<<<<<<<<<<<<<
 *                 indexout[0] = i
 *                 return node
 */

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = ((__pyx_v_rt == __pyx_v_time) != 0);

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":74
 *             rt = node.time
 *             if rt == time:
 *                 indexout[0] = i             # <<<<<<<<<<<<<<
 *                 return node
 *             elif (hi - lo) == 1:
 */

#line 74 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      (__pyx_v_indexout[0]) = __pyx_v_i;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":75
 *             if rt == time:
 *                 indexout[0] = i
 *                 return node             # <<<<<<<<<<<<<<
 *             elif (hi - lo) == 1:
 *                 indexout[0] = i
 */

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(((PyObject *)__pyx_v_node));

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_r = __pyx_v_node;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L0;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":73
 *             node = <LCMSFeatureTreeNodeBase>PyList_GET_ITEM(self.roots, i)
 *             rt = node.time
 *             if rt == time:             # <<<<<<<<<<<<<<
 *                 indexout[0] = i
 *                 return node
 */

#line 73 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":76
 *                 indexout[0] = i
 *                 return node
 *             elif (hi - lo) == 1:             # <<<<<<<<<<<<<<
 *                 indexout[0] = i
 *                 return None
 */

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = (((__pyx_v_hi - __pyx_v_lo) == 1) != 0);

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":77
 *                 return node
 *             elif (hi - lo) == 1:
 *                 indexout[0] = i             # <<<<<<<<<<<<<<
 *                 return None
 *             elif rt < time:
 */

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      (__pyx_v_indexout[0]) = __pyx_v_i;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":78
 *             elif (hi - lo) == 1:
 *                 indexout[0] = i
 *                 return None             # <<<<<<<<<<<<<<
 *             elif rt < time:
 *                 lo = i
 */

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)Py_None); __Pyx_INCREF(Py_None);

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L0;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":76
 *                 indexout[0] = i
 *                 return node
 *             elif (hi - lo) == 1:             # <<<<<<<<<<<<<<
 *                 indexout[0] = i
 *                 return None
 */

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":79
 *                 indexout[0] = i
 *                 return None
 *             elif rt < time:             # <<<<<<<<<<<<<<
 *                 lo = i
 *             elif rt > time:
 */

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = ((__pyx_v_rt < __pyx_v_time) != 0);

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":80
 *                 return None
 *             elif rt < time:
 *                 lo = i             # <<<<<<<<<<<<<<
 *             elif rt > time:
 *                 hi = i
 */

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_lo = __pyx_v_i;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":79
 *                 indexout[0] = i
 *                 return None
 *             elif rt < time:             # <<<<<<<<<<<<<<
 *                 lo = i
 *             elif rt > time:
 */

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L6;

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":81
 *             elif rt < time:
 *                 lo = i
 *             elif rt > time:             # <<<<<<<<<<<<<<
 *                 hi = i
 * 
 */

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = ((__pyx_v_rt > __pyx_v_time) != 0);

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":82
 *                 lo = i
 *             elif rt > time:
 *                 hi = i             # <<<<<<<<<<<<<<
 * 
 *     def _build_node_id_hash(self):
 */

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_hi = __pyx_v_i;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":81
 *             elif rt < time:
 *                 lo = i
 *             elif rt > time:             # <<<<<<<<<<<<<<
 *                 hi = i
 * 
 */

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L6:;

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":60
 *         return node, indexout
 * 
 *     cdef LCMSFeatureTreeNodeBase _find_time(self, double time, size_t* indexout):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int lo, hi
 */

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)Py_None); __Pyx_INCREF(Py_None);

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList._find_time", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_node);

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":84
 *                 hi = i
 * 
 *     def _build_node_id_hash(self):             # <<<<<<<<<<<<<<
 *         self._node_id_hash = set()
 *         for node in self.unspool():
 */

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_7_build_node_id_hash(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_6_build_node_id_hash[] = "LCMSFeatureTreeList._build_node_id_hash(self)";

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_7_build_node_id_hash(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_build_node_id_hash (wrapper)", 0);
  __pyx_r = 
#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_6_build_node_id_hash(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_v_self));

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_6_build_node_id_hash(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self) {
  PyObject *__pyx_v_node = NULL
#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_ssize_t __pyx_t_4;

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*__pyx_t_5)(PyObject *);

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_6;

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_build_node_id_hash", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":85
 * 
 *     def _build_node_id_hash(self):
 *         self._node_id_hash = set()             # <<<<<<<<<<<<<<
 *         for node in self.unspool():
 *             self._node_id_hash.add(node.node_id)
 */

#line 85 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 85, __pyx_L1_error)

#line 85 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 85 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 85 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_self->_node_id_hash);

#line 85 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_v_self->_node_id_hash);

#line 85 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->_node_id_hash = ((PyObject*)__pyx_t_1);

#line 85 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":86
 *     def _build_node_id_hash(self):
 *         self._node_id_hash = set()
 *         for node in self.unspool():             # <<<<<<<<<<<<<<
 *             self._node_id_hash.add(node.node_id)
 * 
 */

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_unspool); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 86, __pyx_L1_error)

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = NULL;

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (likely(__pyx_t_3)) {

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(__pyx_t_3);

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(function);

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF_SET(__pyx_t_2, function);

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 86, __pyx_L1_error)

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = NULL;

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {
    __pyx_t_4 = -1; 
#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 86, __pyx_L1_error)

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 86, __pyx_L1_error)

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  for (;;) {

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (likely(!__pyx_t_5)) {

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (likely(PyList_CheckExact(__pyx_t_2))) {

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 86, __pyx_L1_error)

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #else

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 86, __pyx_L1_error)

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_1);

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #endif

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      } else {

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 86, __pyx_L1_error)

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #else

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 86, __pyx_L1_error)

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_1);

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #endif

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }
    } else 
#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
{

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __pyx_t_5(__pyx_t_2);

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(!__pyx_t_1)) {

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        PyObject* exc_type = PyErr_Occurred();

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (exc_type) {

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          else __PYX_ERR(0, 86, __pyx_L1_error)

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        break;

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF_SET(__pyx_v_node, __pyx_t_1);

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":87
 *         self._node_id_hash = set()
 *         for node in self.unspool():
 *             self._node_id_hash.add(node.node_id)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 87 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(__pyx_v_self->_node_id_hash == Py_None)) {

#line 87 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "add");

#line 87 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __PYX_ERR(0, 87, __pyx_L1_error)

#line 87 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 87 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_node, __pyx_n_s_node_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 87, __pyx_L1_error)

#line 87 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 87 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_6 = PySet_Add(__pyx_v_self->_node_id_hash, __pyx_t_1); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 87, __pyx_L1_error)

#line 87 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":86
 *     def _build_node_id_hash(self):
 *         self._node_id_hash = set()
 *         for node in self.unspool():             # <<<<<<<<<<<<<<
 *             self._node_id_hash.add(node.node_id)
 * 
 */

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 86 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":84
 *                 hi = i
 * 
 *     def _build_node_id_hash(self):             # <<<<<<<<<<<<<<
 *         self._node_id_hash = set()
 *         for node in self.unspool():
 */

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList._build_node_id_hash", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_node);

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 84 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":90
 * 
 *     @property
 *     def node_id_hash(self):             # <<<<<<<<<<<<<<
 *         if self._node_id_hash is None:
 *             self._build_node_id_hash()
 */

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_12node_id_hash_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_12node_id_hash_1__get__(PyObject *__pyx_v_self) {

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_12node_id_hash___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_v_self));

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_12node_id_hash___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self) {

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_1;

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_2;

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":91
 *     @property
 *     def node_id_hash(self):
 *         if self._node_id_hash is None:             # <<<<<<<<<<<<<<
 *             self._build_node_id_hash()
 *         return self._node_id_hash
 */

#line 91 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = (__pyx_v_self->_node_id_hash == ((PyObject*)Py_None));

#line 91 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 91 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":92
 *     def node_id_hash(self):
 *         if self._node_id_hash is None:
 *             self._build_node_id_hash()             # <<<<<<<<<<<<<<
 *         return self._node_id_hash
 * 
 */

#line 92 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_build_node_id_hash); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 92, __pyx_L1_error)

#line 92 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 92 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = NULL;

#line 92 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {

#line 92 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 92 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (likely(__pyx_t_5)) {

#line 92 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 92 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_5);

#line 92 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(function);

#line 92 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF_SET(__pyx_t_4, function);

#line 92 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 92 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 92 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);

#line 92 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 92 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 92, __pyx_L1_error)

#line 92 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 92 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 92 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":91
 *     @property
 *     def node_id_hash(self):
 *         if self._node_id_hash is None:             # <<<<<<<<<<<<<<
 *             self._build_node_id_hash()
 *         return self._node_id_hash
 */

#line 91 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":93
 *         if self._node_id_hash is None:
 *             self._build_node_id_hash()
 *         return self._node_id_hash             # <<<<<<<<<<<<<<
 * 
 *     def insert_node(self, LCMSFeatureTreeNodeBase node):
 */

#line 93 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 93 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_self->_node_id_hash);

#line 93 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_v_self->_node_id_hash;

#line 93 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":90
 * 
 *     @property
 *     def node_id_hash(self):             # <<<<<<<<<<<<<<
 *         if self._node_id_hash is None:
 *             self._build_node_id_hash()
 */

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList.node_id_hash.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 90 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":95
 *         return self._node_id_hash
 * 
 *     def insert_node(self, LCMSFeatureTreeNodeBase node):             # <<<<<<<<<<<<<<
 *         self._invalidate()
 *         cdef:
 */

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_9insert_node(PyObject *__pyx_v_self, PyObject *__pyx_v_node); /*proto*/

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_8insert_node[] = "LCMSFeatureTreeList.insert_node(self, LCMSFeatureTreeNodeBase node)";

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_9insert_node(PyObject *__pyx_v_self, PyObject *__pyx_v_node) {

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("insert_node (wrapper)", 0);

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_node), __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase, 1, "node", 0))) __PYX_ERR(0, 95, __pyx_L1_error)
  __pyx_r = 
#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_8insert_node(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_node));

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_8insert_node(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_node) {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_root = 0
#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_i
#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_slot = NULL
#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_5;

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_6;

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_7;

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_8 = NULL;

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_9 = NULL;

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_10;

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("insert_node", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":96
 * 
 *     def insert_node(self, LCMSFeatureTreeNodeBase node):
 *         self._invalidate()             # <<<<<<<<<<<<<<
 *         cdef:
 *             LCMSFeatureTreeNodeBase root
 */

#line 96 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_v_self->__pyx_vtab)->_invalidate(__pyx_v_self);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":100
 *             LCMSFeatureTreeNodeBase root
 *             size_t i
 *         i = 0             # <<<<<<<<<<<<<<
 *         try:
 *             root = self._find_time(node.time, &i)
 */

#line 100 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_i = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":101
 *             size_t i
 *         i = 0
 *         try:             # <<<<<<<<<<<<<<
 *             root = self._find_time(node.time, &i)
 *             if root is None:
 */

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_PyThreadState_declare

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_PyThreadState_assign

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XGOTREF(__pyx_t_1);

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XGOTREF(__pyx_t_2);

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XGOTREF(__pyx_t_3);

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    /*try:*/ {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":102
 *         i = 0
 *         try:
 *             root = self._find_time(node.time, &i)             # <<<<<<<<<<<<<<
 *             if root is None:
 *                 if i != 0:
 */

#line 102 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_4 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList__find_time(__pyx_v_self, __pyx_v_node->time, (&__pyx_v_i))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 102, __pyx_L3_error)

#line 102 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_4);

#line 102 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_root = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_t_4);

#line 102 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_4 = 0;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":103
 *         try:
 *             root = self._find_time(node.time, &i)
 *             if root is None:             # <<<<<<<<<<<<<<
 *                 if i != 0:
 *                     self.roots.insert(i + 1, node)
 */

#line 103 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_5 = (((PyObject *)__pyx_v_root) == Py_None);

#line 103 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_6 = (__pyx_t_5 != 0);

#line 103 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (__pyx_t_6) {

        /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":104
 *             root = self._find_time(node.time, &i)
 *             if root is None:
 *                 if i != 0:             # <<<<<<<<<<<<<<
 *                     self.roots.insert(i + 1, node)
 *                 else:
 */

#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_6 = ((__pyx_v_i != 0) != 0);

#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (__pyx_t_6) {

          /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":105
 *             if root is None:
 *                 if i != 0:
 *                     self.roots.insert(i + 1, node)             # <<<<<<<<<<<<<<
 *                 else:
 *                     slot = self.getitem(i)
 */

#line 105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (unlikely(__pyx_v_self->roots == Py_None)) {

#line 105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "insert");

#line 105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __PYX_ERR(0, 105, __pyx_L3_error)

#line 105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_7 = PyList_Insert(__pyx_v_self->roots, (__pyx_v_i + 1), ((PyObject *)__pyx_v_node)); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 105, __pyx_L3_error)

          /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":104
 *             root = self._find_time(node.time, &i)
 *             if root is None:
 *                 if i != 0:             # <<<<<<<<<<<<<<
 *                     self.roots.insert(i + 1, node)
 *                 else:
 */

#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          goto __pyx_L10;

#line 104 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

        /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":107
 *                     self.roots.insert(i + 1, node)
 *                 else:
 *                     slot = self.getitem(i)             # <<<<<<<<<<<<<<
 *                     if slot.time < node.time:
 *                         i += 1
 */

#line 107 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        /*else*/ {

#line 107 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_getitem(__pyx_v_self, __pyx_v_i)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 107, __pyx_L3_error)

#line 107 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_GOTREF(__pyx_t_4);

#line 107 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_v_slot = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_t_4);

#line 107 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = 0;

          /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":108
 *                 else:
 *                     slot = self.getitem(i)
 *                     if slot.time < node.time:             # <<<<<<<<<<<<<<
 *                         i += 1
 *                     self.roots.insert(i, node)
 */

#line 108 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_6 = ((__pyx_v_slot->time < __pyx_v_node->time) != 0);

#line 108 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (__pyx_t_6) {

            /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":109
 *                     slot = self.getitem(i)
 *                     if slot.time < node.time:
 *                         i += 1             # <<<<<<<<<<<<<<
 *                     self.roots.insert(i, node)
 *             else:
 */

#line 109 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __pyx_v_i = (__pyx_v_i + 1);

            /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":108
 *                 else:
 *                     slot = self.getitem(i)
 *                     if slot.time < node.time:             # <<<<<<<<<<<<<<
 *                         i += 1
 *                     self.roots.insert(i, node)
 */

#line 108 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

          /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":110
 *                     if slot.time < node.time:
 *                         i += 1
 *                     self.roots.insert(i, node)             # <<<<<<<<<<<<<<
 *             else:
 *                 root.add(node)
 */

#line 110 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (unlikely(__pyx_v_self->roots == Py_None)) {

#line 110 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "insert");

#line 110 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __PYX_ERR(0, 110, __pyx_L3_error)

#line 110 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 110 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_7 = PyList_Insert(__pyx_v_self->roots, __pyx_v_i, ((PyObject *)__pyx_v_node)); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 110, __pyx_L3_error)

#line 110 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 110 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_L10:;

        /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":103
 *         try:
 *             root = self._find_time(node.time, &i)
 *             if root is None:             # <<<<<<<<<<<<<<
 *                 if i != 0:
 *                     self.roots.insert(i + 1, node)
 */

#line 103 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L9;

#line 103 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":112
 *                     self.roots.insert(i, node)
 *             else:
 *                 root.add(node)             # <<<<<<<<<<<<<<
 *             return i
 *         except ValueError:
 */

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      /*else*/ {

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_root), __pyx_n_s_add); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 112, __pyx_L3_error)

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_8);

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_9 = NULL;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_9)) {

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_9);

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_8, function);

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_4 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, ((PyObject *)__pyx_v_node)) : __Pyx_PyObject_CallOneArg(__pyx_t_8, ((PyObject *)__pyx_v_node));

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 112, __pyx_L3_error)

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_L9:;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":113
 *             else:
 *                 root.add(node)
 *             return i             # <<<<<<<<<<<<<<
 *         except ValueError:
 *             self.roots.append(node)
 */

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_XDECREF(__pyx_r);

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 113, __pyx_L3_error)

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_4);

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_r = __pyx_t_4;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_4 = 0;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L7_try_return;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":101
 *             size_t i
 *         i = 0
 *         try:             # <<<<<<<<<<<<<<
 *             root = self._find_time(node.time, &i)
 *             if root is None:
 */

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L3_error:;

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":114
 *                 root.add(node)
 *             return i
 *         except ValueError:             # <<<<<<<<<<<<<<
 *             self.roots.append(node)
 *             return 0
 */

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError);

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_10) {

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList.insert_node", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_8, &__pyx_t_9) < 0) __PYX_ERR(0, 114, __pyx_L5_except_error)

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_4);

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_8);

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_9);

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":115
 *             return i
 *         except ValueError:
 *             self.roots.append(node)             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_v_self->roots == Py_None)) {

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __PYX_ERR(0, 115, __pyx_L5_except_error)

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_self->roots, ((PyObject *)__pyx_v_node)); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 115, __pyx_L5_except_error)

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":116
 *         except ValueError:
 *             self.roots.append(node)
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     def insert(self, time, peaks):
 */

#line 116 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_XDECREF(__pyx_r);

#line 116 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(__pyx_int_0);

#line 116 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_r = __pyx_int_0;

#line 116 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 116 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 116 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 116 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L6_except_return;

#line 116 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 116 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L5_except_error;

#line 116 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L5_except_error:;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":101
 *             size_t i
 *         i = 0
 *         try:             # <<<<<<<<<<<<<<
 *             root = self._find_time(node.time, &i)
 *             if root is None:
 */

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XGIVEREF(__pyx_t_1);

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L1_error;

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L7_try_return:;

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XGIVEREF(__pyx_t_1);

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L0;

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L6_except_return:;

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XGIVEREF(__pyx_t_1);

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L0;

#line 101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":95
 *         return self._node_id_hash
 * 
 *     def insert_node(self, LCMSFeatureTreeNodeBase node):             # <<<<<<<<<<<<<<
 *         self._invalidate()
 *         cdef:
 */

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_8);

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_9);

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList.insert_node", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_root);

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_slot);

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 95 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":118
 *             return 0
 * 
 *     def insert(self, time, peaks):             # <<<<<<<<<<<<<<
 *         node = self._make_node(time, peaks)
 *         return self.insert_node(node)
 */

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_11insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_10insert[] = "LCMSFeatureTreeList.insert(self, time, peaks)";

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_11insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_time = 0
#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyObject *__pyx_v_peaks = 0
#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("insert (wrapper)", 0);

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_time,&__pyx_n_s_peaks,0};

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject* values[2] = {0,0};

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(__pyx_kwds)) {

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      Py_ssize_t kw_args;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {
        case  2: 
#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0: break;
        default: 
#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0:

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_time)) != 0)) kw_args--;
        else 
#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  1:

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peaks)) != 0)) kw_args--;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("insert", 1, 2, 2, 1); 
#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__PYX_ERR(0, 118, __pyx_L3_error)

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(kw_args > 0)) {

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "insert") < 0)) __PYX_ERR(0, 118, __pyx_L3_error)

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L5_argtuple_error;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_time = values[0];

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_peaks = values[1];

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("insert", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__PYX_ERR(0, 118, __pyx_L3_error)

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_error:;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return NULL;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_10insert(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_v_self), __pyx_v_time, __pyx_v_peaks);

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_10insert(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, PyObject *__pyx_v_time, PyObject *__pyx_v_peaks) {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_node = NULL
#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_1;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("insert", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":119
 * 
 *     def insert(self, time, peaks):
 *         node = self._make_node(time, peaks)             # <<<<<<<<<<<<<<
 *         return self.insert_node(node)
 * 
 */

#line 119 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_time); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 119, __pyx_L1_error)

#line 119 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (!(likely(PyList_CheckExact(__pyx_v_peaks))||((__pyx_v_peaks) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_peaks)->tp_name), 0))) __PYX_ERR(0, 119, __pyx_L1_error)

#line 119 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_v_self->__pyx_vtab)->_make_node(__pyx_v_self, __pyx_t_1, ((PyObject*)__pyx_v_peaks), 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 119, __pyx_L1_error)

#line 119 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 119 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_node = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_t_2);

#line 119 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":120
 *     def insert(self, time, peaks):
 *         node = self._make_node(time, peaks)
 *         return self.insert_node(node)             # <<<<<<<<<<<<<<
 * 
 *     def extend(self, iterable):
 */

#line 120 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 120 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_insert_node); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 120, __pyx_L1_error)

#line 120 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 120 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = NULL;

#line 120 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {

#line 120 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 120 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (likely(__pyx_t_4)) {

#line 120 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 120 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(__pyx_t_4);

#line 120 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(function);

#line 120 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF_SET(__pyx_t_3, function);

#line 120 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 120 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 120 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_node)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_node));

#line 120 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 120 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 120, __pyx_L1_error)

#line 120 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 120 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 120 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_2;

#line 120 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = 0;

#line 120 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":118
 *             return 0
 * 
 *     def insert(self, time, peaks):             # <<<<<<<<<<<<<<
 *         node = self._make_node(time, peaks)
 *         return self.insert_node(node)
 */

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_node);

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":122
 *         return self.insert_node(node)
 * 
 *     def extend(self, iterable):             # <<<<<<<<<<<<<<
 *         for peaks, time in iterable:
 *             self.insert(time, peaks)
 */

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_13extend(PyObject *__pyx_v_self, PyObject *__pyx_v_iterable); /*proto*/

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_12extend[] = "LCMSFeatureTreeList.extend(self, iterable)";

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_13extend(PyObject *__pyx_v_self, PyObject *__pyx_v_iterable) {

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("extend (wrapper)", 0);
  __pyx_r = 
#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_12extend(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_v_self), ((PyObject *)__pyx_v_iterable));

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_12extend(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, PyObject *__pyx_v_iterable) {
  PyObject *__pyx_v_peaks = NULL
#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyObject *__pyx_v_time = NULL
#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_ssize_t __pyx_t_2;

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*__pyx_t_3)(PyObject *);

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_7 = NULL;

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*__pyx_t_8)(PyObject *);

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_9;

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("extend", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":123
 * 
 *     def extend(self, iterable):
 *         for peaks, time in iterable:             # <<<<<<<<<<<<<<
 *             self.insert(time, peaks)
 * 
 */

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (likely(PyList_CheckExact(__pyx_v_iterable)) || PyTuple_CheckExact(__pyx_v_iterable)) {

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __pyx_v_iterable; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = NULL;

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {
    __pyx_t_2 = -1; 
#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_1 = PyObject_GetIter(__pyx_v_iterable); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 123, __pyx_L1_error)

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 123, __pyx_L1_error)

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  for (;;) {

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (likely(!__pyx_t_3)) {

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (likely(PyList_CheckExact(__pyx_t_1))) {

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 123, __pyx_L1_error)

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #else

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 123, __pyx_L1_error)

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #endif

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      } else {

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 123, __pyx_L1_error)

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #else

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 123, __pyx_L1_error)

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #endif

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }
    } else 
#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
{

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(!__pyx_t_4)) {

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        PyObject* exc_type = PyErr_Occurred();

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (exc_type) {

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          else __PYX_ERR(0, 123, __pyx_L1_error)

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        break;

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_4);

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PyObject* sequence = __pyx_t_4;

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(size != 2)) {

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __PYX_ERR(0, 123, __pyx_L1_error)

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (likely(PyTuple_CheckExact(sequence))) {

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      } else {

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(__pyx_t_5);

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(__pyx_t_6);

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #else

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 123, __pyx_L1_error)

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_5);

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 123, __pyx_L1_error)

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      Py_ssize_t index = -1;

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_7 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 123, __pyx_L1_error)

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_7);

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_5)) 
#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_unpacking_failed;

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) 
#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_unpacking_failed;

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(0, 123, __pyx_L1_error)

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_8 = NULL;

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L6_unpacking_done;

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_L5_unpacking_failed:;

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_8 = NULL;

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __PYX_ERR(0, 123, __pyx_L1_error)

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_L6_unpacking_done:;

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peaks, __pyx_t_5);

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = 0;

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF_SET(__pyx_v_time, __pyx_t_6);

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_6 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":124
 *     def extend(self, iterable):
 *         for peaks, time in iterable:
 *             self.insert(time, peaks)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, i):
 */

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_insert); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 124, __pyx_L1_error)

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = NULL;

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_9 = 0;

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (likely(__pyx_t_5)) {

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_5);

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(function);

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF_SET(__pyx_t_6, function);

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_9 = 1;

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_FAST_PYCALL

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (PyFunction_Check(__pyx_t_6)) {

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_time, __pyx_v_peaks};

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 124, __pyx_L1_error)

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_4);

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_FAST_PYCCALL

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_time, __pyx_v_peaks};

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 124, __pyx_L1_error)

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_4);

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    {

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_7 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 124, __pyx_L1_error)

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_7);

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (__pyx_t_5) {

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(__pyx_v_time);

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GIVEREF(__pyx_v_time);

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_9, __pyx_v_time);

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(__pyx_v_peaks);

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GIVEREF(__pyx_v_peaks);

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_9, __pyx_v_peaks);

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_7, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 124, __pyx_L1_error)

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_4);

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":123
 * 
 *     def extend(self, iterable):
 *         for peaks, time in iterable:             # <<<<<<<<<<<<<<
 *             self.insert(time, peaks)
 * 
 */

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 123 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":122
 *         return self.insert_node(node)
 * 
 *     def extend(self, iterable):             # <<<<<<<<<<<<<<
 *         for peaks, time in iterable:
 *             self.insert(time, peaks)
 */

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_7);

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_peaks);

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_time);

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":126
 *             self.insert(time, peaks)
 * 
 *     def __getitem__(self, i):             # <<<<<<<<<<<<<<
 *         return self.roots[i]
 * 
 */

#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_15__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_i); /*proto*/

#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_15__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_i) {

#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = 
#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_14__getitem__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_v_self), ((PyObject *)__pyx_v_i));

#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_14__getitem__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, PyObject *__pyx_v_i) {

#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":127
 * 
 *     def __getitem__(self, i):
 *         return self.roots[i]             # <<<<<<<<<<<<<<
 * 
 *     cdef LCMSFeatureTreeNodeBase getitem(self, size_t i):
 */

#line 127 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 127 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_v_self->roots == Py_None)) {

#line 127 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 127 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 127 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_self->roots, __pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)

#line 127 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 127 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 127 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 127 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":126
 *             self.insert(time, peaks)
 * 
 *     def __getitem__(self, i):             # <<<<<<<<<<<<<<
 *         return self.roots[i]
 * 
 */

#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 126 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":129
 *         return self.roots[i]
 * 
 *     cdef LCMSFeatureTreeNodeBase getitem(self, size_t i):             # <<<<<<<<<<<<<<
 *         return <LCMSFeatureTreeNodeBase>PyList_GetItem(self.roots, i)
 * 
 */

#line 129 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 129 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static CYTHON_INLINE struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_getitem(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, size_t __pyx_v_i) {

#line 129 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_r = NULL;

#line 129 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 129 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 129 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2;

#line 129 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 129 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 129 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 129 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("getitem", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":130
 * 
 *     cdef LCMSFeatureTreeNodeBase getitem(self, size_t i):
 *         return <LCMSFeatureTreeNodeBase>PyList_GetItem(self.roots, i)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */

#line 130 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 130 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_v_self->roots;

#line 130 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 130 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = PyList_GetItem(__pyx_t_1, __pyx_v_i); if (unlikely(__pyx_t_2 == ((PyObject *)NULL))) __PYX_ERR(0, 130, __pyx_L1_error)

#line 130 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 130 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(((PyObject *)((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_t_2)));

#line 130 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_t_2);

#line 130 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":129
 *         return self.roots[i]
 * 
 *     cdef LCMSFeatureTreeNodeBase getitem(self, size_t i):             # <<<<<<<<<<<<<<
 *         return <LCMSFeatureTreeNodeBase>PyList_GetItem(self.roots, i)
 * 
 */

#line 129 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 129 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 129 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 129 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 129 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList.getitem", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 129 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 129 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 129 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 129 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 129 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 129 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":132
 *         return <LCMSFeatureTreeNodeBase>PyList_GetItem(self.roots, i)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return PyList_GET_SIZE(self.roots)
 * 
 */

#line 132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static Py_ssize_t __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_17__len__(PyObject *__pyx_v_self); /*proto*/

#line 132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static Py_ssize_t __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_17__len__(PyObject *__pyx_v_self) {

#line 132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_ssize_t __pyx_r;

#line 132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = 
#line 132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_16__len__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_v_self));

#line 132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static Py_ssize_t __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_16__len__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self) {

#line 132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_ssize_t __pyx_r;

#line 132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":133
 * 
 *     def __len__(self):
 *         return PyList_GET_SIZE(self.roots)             # <<<<<<<<<<<<<<
 * 
 *     cdef size_t get_size(self):
 */

#line 133 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_v_self->roots;

#line 133 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 133 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = PyList_GET_SIZE(__pyx_t_1);

#line 133 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 133 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":132
 *         return <LCMSFeatureTreeNodeBase>PyList_GetItem(self.roots, i)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return PyList_GET_SIZE(self.roots)
 * 
 */

#line 132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":135
 *         return PyList_GET_SIZE(self.roots)
 * 
 *     cdef size_t get_size(self):             # <<<<<<<<<<<<<<
 *         return PyList_GET_SIZE(self.roots)
 * 
 */

#line 135 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 135 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static CYTHON_INLINE size_t __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_get_size(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self) {

#line 135 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_r;

#line 135 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 135 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 135 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("get_size", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":136
 * 
 *     cdef size_t get_size(self):
 *         return PyList_GET_SIZE(self.roots)             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */

#line 136 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_v_self->roots;

#line 136 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 136 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = PyList_GET_SIZE(__pyx_t_1);

#line 136 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 136 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":135
 *         return PyList_GET_SIZE(self.roots)
 * 
 *     cdef size_t get_size(self):             # <<<<<<<<<<<<<<
 *         return PyList_GET_SIZE(self.roots)
 * 
 */

#line 135 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 135 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 135 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 135 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 135 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 135 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":138
 *         return PyList_GET_SIZE(self.roots)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.roots)
 * 
 */

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_19__iter__(PyObject *__pyx_v_self); /*proto*/

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_19__iter__(PyObject *__pyx_v_self) {

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = 
#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_18__iter__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_v_self));

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_18__iter__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self) {

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":139
 * 
 *     def __iter__(self):
 *         return iter(self.roots)             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_v_self->roots;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 139, __pyx_L1_error)

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_2;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = 0;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":138
 *         return PyList_GET_SIZE(self.roots)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.roots)
 * 
 */

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 138 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":141
 *         return iter(self.roots)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.roots,)
 * 
 */

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_21__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_20__reduce__[] = "LCMSFeatureTreeList.__reduce__(self)";

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_21__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = 
#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_20__reduce__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_v_self));

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_20__reduce__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self) {

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":142
 * 
 *     def __reduce__(self):
 *         return self.__class__, (self.roots,)             # <<<<<<<<<<<<<<
 * 
 *     def clone(self, deep=False):
 */

#line 142 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 142 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 142, __pyx_L1_error)

#line 142 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 142 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 142, __pyx_L1_error)

#line 142 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 142 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_self->roots);

#line 142 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_v_self->roots);

#line 142 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self->roots);

#line 142 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 142, __pyx_L1_error)

#line 142 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 142 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 142 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);

#line 142 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 142 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);

#line 142 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 142 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = 0;

#line 142 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_3;

#line 142 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = 0;

#line 142 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":141
 *         return iter(self.roots)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.roots,)
 * 
 */

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":144
 *         return self.__class__, (self.roots,)
 * 
 *     def clone(self, deep=False):             # <<<<<<<<<<<<<<
 *         return LCMSFeatureTreeList(node.clone(deep=deep) for node in self)
 * 
 */

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_23clone(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_22clone[] = "LCMSFeatureTreeList.clone(self, deep=False)";

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_23clone(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_deep = 0
#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("clone (wrapper)", 0);

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_deep,0};

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject* values[1] = {0};

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    values[0] = ((PyObject *)Py_False);

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(__pyx_kwds)) {

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      Py_ssize_t kw_args;

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {
        case  1: 
#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0: break;
        default: 
#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0:

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (kw_args > 0) {

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_deep);

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (value) { values[0] = value; kw_args--; }

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(kw_args > 0)) {

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "clone") < 0)) __PYX_ERR(0, 144, __pyx_L3_error)

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: 
#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  0: 
#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
break;
        default: 
#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_deep = values[0];

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("clone", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); 
#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__PYX_ERR(0, 144, __pyx_L3_error)

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_error:;

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return NULL;

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_22clone(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_v_self), __pyx_v_deep);

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_gb_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5clone_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":145
 * 
 *     def clone(self, deep=False):
 *         return LCMSFeatureTreeList(node.clone(deep=deep) for node in self)             # <<<<<<<<<<<<<<
 * 
 *     def unspool(self):
 */

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5clone_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr *__pyx_cur_scope
#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("genexpr", 0);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_cur_scope = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr *)__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr, __pyx_empty_tuple, NULL);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__pyx_cur_scope)) {

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_cur_scope = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr *)Py_None);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(Py_None);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __PYX_ERR(0, 145, __pyx_L1_error)

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_cur_scope);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone *) __pyx_self;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5clone_2generator1, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_clone_locals_genexpr, __pyx_n_s_ms_deisotope__c_feature_map_lcms); if (unlikely(!gen)) __PYX_ERR(0, 145, __pyx_L1_error)

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_cur_scope);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_RefNannyFinishContext();

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    return (PyObject *) gen;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList.clone.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_gb_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5clone_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr *__pyx_cur_scope = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr *)__pyx_generator->closure);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_ssize_t __pyx_t_2;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*__pyx_t_3)(PyObject *);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("genexpr", 0);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  switch (__pyx_generator->resume_label) {

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    case 0: goto __pyx_L3_first_run;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    case 1: goto __pyx_L6_resume_from_yield;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    default: /* CPython raises the right error here */

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_RefNannyFinishContext();

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    return NULL;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_first_run:;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 145, __pyx_L1_error)

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 145, __pyx_L1_error) }

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (likely(PyList_CheckExact(((PyObject *)__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self))) || PyTuple_CheckExact(((PyObject *)__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self))) {

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self); __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = NULL;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {
    __pyx_t_2 = -1; 
#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_1 = PyObject_GetIter(((PyObject *)__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 145, __pyx_L1_error)

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 145, __pyx_L1_error)

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  for (;;) {

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (likely(!__pyx_t_3)) {

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (likely(PyList_CheckExact(__pyx_t_1))) {

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 145, __pyx_L1_error)

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #else

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 145, __pyx_L1_error)

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #endif

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      } else {

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 145, __pyx_L1_error)

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #else

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 145, __pyx_L1_error)

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #endif

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }
    } else 
#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
{

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(!__pyx_t_4)) {

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        PyObject* exc_type = PyErr_Occurred();

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (exc_type) {

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          else __PYX_ERR(0, 145, __pyx_L1_error)

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        break;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_4);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_node);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_node, __pyx_t_4);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GIVEREF(__pyx_t_4);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = 0;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_node, __pyx_n_s_clone); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 145, __pyx_L1_error)

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 145, __pyx_L1_error)

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_5);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_deep)) { __Pyx_RaiseClosureNameError("deep"); __PYX_ERR(0, 145, __pyx_L1_error) }

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_deep, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_deep) < 0) __PYX_ERR(0, 145, __pyx_L1_error)

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_empty_tuple, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 145, __pyx_L1_error)

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_r = __pyx_t_6;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_6 = 0;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XGIVEREF(__pyx_t_1);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XGIVEREF(__pyx_r);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_RefNannyFinishContext();

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    /* return from generator, yielding value */

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_generator->resume_label = 1;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    return __pyx_r;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L6_resume_from_yield:;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_cur_scope->__pyx_t_0 = 0;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XGOTREF(__pyx_t_1);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 145, __pyx_L1_error)

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyErr_SetNone(PyExc_StopIteration);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #if !CYTHON_USE_EXC_INFO_STACK

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #endif

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_generator->resume_label = -1;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":144
 *         return self.__class__, (self.roots,)
 * 
 *     def clone(self, deep=False):             # <<<<<<<<<<<<<<
 *         return LCMSFeatureTreeList(node.clone(deep=deep) for node in self)
 * 
 */

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_22clone(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, PyObject *__pyx_v_deep) {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone *__pyx_cur_scope
#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyObject *__pyx_gb_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5clone_2generator1 = 0
#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("clone", 0);

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_cur_scope = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone *)__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone, __pyx_empty_tuple, NULL);

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__pyx_cur_scope)) {

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_cur_scope = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone *)Py_None);

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(Py_None);

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __PYX_ERR(0, 144, __pyx_L1_error)

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_cur_scope);

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_cur_scope->__pyx_v_deep = __pyx_v_deep;

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_deep);

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_deep);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":145
 * 
 *     def clone(self, deep=False):
 *         return LCMSFeatureTreeList(node.clone(deep=deep) for node in self)             # <<<<<<<<<<<<<<
 * 
 *     def unspool(self):
 */

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5clone_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 145, __pyx_L1_error)

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 145, __pyx_L1_error)

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_2;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = 0;

#line 145 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":144
 *         return self.__class__, (self.roots,)
 * 
 *     def clone(self, deep=False):             # <<<<<<<<<<<<<<
 *         return LCMSFeatureTreeList(node.clone(deep=deep) for node in self)
 * 
 */

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_gb_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5clone_2generator1);

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":147
 *         return LCMSFeatureTreeList(node.clone(deep=deep) for node in self)
 * 
 *     def unspool(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list out_queue, stack
 */

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_25unspool(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_24unspool[] = "LCMSFeatureTreeList.unspool(self)";

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_25unspool(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("unspool (wrapper)", 0);
  __pyx_r = 
#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_24unspool(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_v_self));

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_24unspool(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self) {
  PyObject *__pyx_v_out_queue = 0
#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyObject *__pyx_v_stack = 0
#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_node = 0
#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_root = 0
#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_i
#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_n
#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_2;

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_3;

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_4;

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_ssize_t __pyx_t_5;

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_6;

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_7;

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("unspool", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":152
 *             LCMSFeatureTreeNodeBase node, root
 *             size_t i, n
 *         out_queue = []             # <<<<<<<<<<<<<<
 *         n = self.get_size()
 *         for i in range(n):
 */

#line 152 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 152, __pyx_L1_error)

#line 152 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 152 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_out_queue = ((PyObject*)__pyx_t_1);

#line 152 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":153
 *             size_t i, n
 *         out_queue = []
 *         n = self.get_size()             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             root = self.getitem(i)
 */

#line 153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_n = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_get_size(__pyx_v_self);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":154
 *         out_queue = []
 *         n = self.get_size()
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             root = self.getitem(i)
 *             stack = [root]
 */

#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __pyx_v_n;

#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = __pyx_t_2;

#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_i = __pyx_t_4;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":155
 *         n = self.get_size()
 *         for i in range(n):
 *             root = self.getitem(i)             # <<<<<<<<<<<<<<
 *             stack = [root]
 *             while len(stack) != 0:
 */

#line 155 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_getitem(__pyx_v_self, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 155, __pyx_L1_error)

#line 155 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 155 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF_SET(__pyx_v_root, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_t_1));

#line 155 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":156
 *         for i in range(n):
 *             root = self.getitem(i)
 *             stack = [root]             # <<<<<<<<<<<<<<
 *             while len(stack) != 0:
 *                 node = stack.pop()
 */

#line 156 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 156, __pyx_L1_error)

#line 156 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 156 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(((PyObject *)__pyx_v_root));

#line 156 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GIVEREF(((PyObject *)__pyx_v_root));

#line 156 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyList_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_root));

#line 156 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF_SET(__pyx_v_stack, ((PyObject*)__pyx_t_1));

#line 156 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":157
 *             root = self.getitem(i)
 *             stack = [root]
 *             while len(stack) != 0:             # <<<<<<<<<<<<<<
 *                 node = stack.pop()
 *                 out_queue.append(node)
 */

#line 157 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    while (1) {

#line 157 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_5 = PyList_GET_SIZE(__pyx_v_stack); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 157, __pyx_L1_error)

#line 157 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_6 = ((__pyx_t_5 != 0) != 0);

#line 157 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!__pyx_t_6) break;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":158
 *             stack = [root]
 *             while len(stack) != 0:
 *                 node = stack.pop()             # <<<<<<<<<<<<<<
 *                 out_queue.append(node)
 *                 stack.extend(node.children)
 */

#line 158 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyList_Pop(__pyx_v_stack); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 158, __pyx_L1_error)

#line 158 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 158 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase))))) __PYX_ERR(0, 158, __pyx_L1_error)

#line 158 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_t_1));

#line 158 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = 0;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":159
 *             while len(stack) != 0:
 *                 node = stack.pop()
 *                 out_queue.append(node)             # <<<<<<<<<<<<<<
 *                 stack.extend(node.children)
 *         return out_queue
 */

#line 159 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_out_queue, ((PyObject *)__pyx_v_node)); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 159, __pyx_L1_error)

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":160
 *                 node = stack.pop()
 *                 out_queue.append(node)
 *                 stack.extend(node.children)             # <<<<<<<<<<<<<<
 *         return out_queue
 * 
 */

#line 160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_node), __pyx_n_s_children); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 160, __pyx_L1_error)

#line 160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_7 = __Pyx_PyList_Extend(__pyx_v_stack, __pyx_t_1); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 160, __pyx_L1_error)

#line 160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":161
 *                 out_queue.append(node)
 *                 stack.extend(node.children)
 *         return out_queue             # <<<<<<<<<<<<<<
 * 
 *     def common_nodes(self, other):
 */

#line 161 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 161 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_out_queue);

#line 161 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_v_out_queue;

#line 161 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":147
 *         return LCMSFeatureTreeList(node.clone(deep=deep) for node in self)
 * 
 *     def unspool(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list out_queue, stack
 */

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList.unspool", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_out_queue);

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_stack);

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_node);

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_root);

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 147 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":163
 *         return out_queue
 * 
 *     def common_nodes(self, other):             # <<<<<<<<<<<<<<
 *         return len(self.node_id_hash & other.node_id_hash)
 * 
 */

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_27common_nodes(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_26common_nodes[] = "LCMSFeatureTreeList.common_nodes(self, other)";

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_27common_nodes(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("common_nodes (wrapper)", 0);
  __pyx_r = 
#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_26common_nodes(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_v_self), ((PyObject *)__pyx_v_other));

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_26common_nodes(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, PyObject *__pyx_v_other) {

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_ssize_t __pyx_t_4;

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("common_nodes", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":164
 * 
 *     def common_nodes(self, other):
 *         return len(self.node_id_hash & other.node_id_hash)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */

#line 164 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 164 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_node_id_hash); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 164, __pyx_L1_error)

#line 164 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 164 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_node_id_hash); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 164, __pyx_L1_error)

#line 164 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 164 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = PyNumber_And(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 164, __pyx_L1_error)

#line 164 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 164 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 164 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 164 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 164, __pyx_L1_error)

#line 164 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 164 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 164, __pyx_L1_error)

#line 164 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 164 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_3;

#line 164 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = 0;

#line 164 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":163
 *         return out_queue
 * 
 *     def common_nodes(self, other):             # <<<<<<<<<<<<<<
 *         return len(self.node_id_hash & other.node_id_hash)
 * 
 */

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList.common_nodes", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 163 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":166
 *         return len(self.node_id_hash & other.node_id_hash)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "LCMSFeatureTreeList(%d nodes, %0.2f-%0.2f)" % (
 *             len(self), self[0].time, self[-1].time)
 */

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_29__repr__(PyObject *__pyx_v_self); /*proto*/

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_29__repr__(PyObject *__pyx_v_self) {

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = 
#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_28__repr__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_v_self));

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_28__repr__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self) {

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_ssize_t __pyx_t_1;

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":167
 * 
 *     def __repr__(self):
 *         return "LCMSFeatureTreeList(%d nodes, %0.2f-%0.2f)" % (             # <<<<<<<<<<<<<<
 *             len(self), self[0].time, self[-1].time)
 * 
 */

#line 167 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":168
 *     def __repr__(self):
 *         return "LCMSFeatureTreeList(%d nodes, %0.2f-%0.2f)" % (
 *             len(self), self[0].time, self[-1].time)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 168 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyObject_Length(((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 168, __pyx_L1_error)

#line 168 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 168, __pyx_L1_error)

#line 168 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 168 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = __Pyx_GetItemInt(((PyObject *)__pyx_v_self), 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 168, __pyx_L1_error)

#line 168 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 168 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_time); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 168, __pyx_L1_error)

#line 168 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 168 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 168 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = __Pyx_GetItemInt(((PyObject *)__pyx_v_self), -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 168, __pyx_L1_error)

#line 168 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 168 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_time); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 168, __pyx_L1_error)

#line 168 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 168 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 168 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 168, __pyx_L1_error)

#line 168 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 168 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 168 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);

#line 168 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_4);

#line 168 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);

#line 168 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_5);

#line 168 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_5);

#line 168 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = 0;

#line 168 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = 0;

#line 168 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":167
 * 
 *     def __repr__(self):
 *         return "LCMSFeatureTreeList(%d nodes, %0.2f-%0.2f)" % (             # <<<<<<<<<<<<<<
 *             len(self), self[0].time, self[-1].time)
 * 
 */

#line 167 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_LCMSFeatureTreeList_d_nodes_0_2f, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 167, __pyx_L1_error)

#line 167 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 167 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 167 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_5;

#line 167 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = 0;

#line 167 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":166
 *         return len(self.node_id_hash & other.node_id_hash)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "LCMSFeatureTreeList(%d nodes, %0.2f-%0.2f)" % (
 *             len(self), self[0].time, self[-1].time)
 */

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":6
 * cdef class LCMSFeatureTreeList(object):
 *     cdef:
 *         public list roots             # <<<<<<<<<<<<<<
 *         public set _node_id_hash
 * 
 */

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5roots_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5roots_1__get__(PyObject *__pyx_v_self) {

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = 0;

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5roots___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_v_self));

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5roots___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self) {

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = NULL;

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(__pyx_v_self->roots);

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = __pyx_v_self->roots;

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5roots_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5roots_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5roots_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5roots_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, PyObject *__pyx_v_value) {

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  if (!(likely(PyList_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(2, 6, __pyx_L1_error)

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = __pyx_v_value;

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(__pyx_t_1);

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(__pyx_t_1);

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->roots);

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(__pyx_v_self->roots);

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->roots = ((PyObject*)__pyx_t_1);

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = 0;

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList.roots.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = -1;

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5roots_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5roots_5__del__(PyObject *__pyx_v_self) {

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5roots_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_v_self));

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5roots_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self) {

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(Py_None);

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(Py_None);

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->roots);

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(__pyx_v_self->roots);

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->roots = ((PyObject*)Py_None);

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 6 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":7
 *     cdef:
 *         public list roots
 *         public set _node_id_hash             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_13_node_id_hash_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_13_node_id_hash_1__get__(PyObject *__pyx_v_self) {

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = 0;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_13_node_id_hash___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_v_self));

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_13_node_id_hash___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self) {

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = NULL;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(__pyx_v_self->_node_id_hash);

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = __pyx_v_self->_node_id_hash;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_13_node_id_hash_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_13_node_id_hash_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_13_node_id_hash_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_13_node_id_hash_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self, PyObject *__pyx_v_value) {

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  if (!(likely(PySet_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "set", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(2, 7, __pyx_L1_error)

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = __pyx_v_value;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(__pyx_t_1);

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(__pyx_t_1);

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->_node_id_hash);

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(__pyx_v_self->_node_id_hash);

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->_node_id_hash = ((PyObject*)__pyx_t_1);

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = 0;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList._node_id_hash.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = -1;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_13_node_id_hash_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_13_node_id_hash_5__del__(PyObject *__pyx_v_self) {

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_13_node_id_hash_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_v_self));

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_13_node_id_hash_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *__pyx_v_self) {

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(Py_None);

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(Py_None);

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->_node_id_hash);

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(__pyx_v_self->_node_id_hash);

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->_node_id_hash = ((PyObject*)Py_None);

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":177
 * cdef class LCMSFeatureTreeNodeBase(object):
 * 
 *     def clone(self, deep=False):             # <<<<<<<<<<<<<<
 *         cdef LCMSFeatureTreeNodeBase node
 *         cdef list peaks
 */

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_1clone(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_clone[] = "LCMSFeatureTreeNodeBase.clone(self, deep=False)";

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_1clone(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_deep = 0
#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("clone (wrapper)", 0);

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_deep,0};

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject* values[1] = {0};

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    values[0] = ((PyObject *)Py_False);

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(__pyx_kwds)) {

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      Py_ssize_t kw_args;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {
        case  1: 
#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0: break;
        default: 
#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0:

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (kw_args > 0) {

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_deep);

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (value) { values[0] = value; kw_args--; }

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(kw_args > 0)) {

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "clone") < 0)) __PYX_ERR(0, 177, __pyx_L3_error)

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: 
#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  0: 
#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
break;
        default: 
#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_deep = values[0];

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("clone", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); 
#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__PYX_ERR(0, 177, __pyx_L3_error)

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_error:;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNodeBase.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return NULL;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_clone(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self), __pyx_v_deep);

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_clone(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self, PyObject *__pyx_v_deep) {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_node = 0
#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyObject *__pyx_v_peaks = 0
#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_i
#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_n
#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak = 0
#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_1;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_3;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_4;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_5;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_6;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_7;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_8 = NULL;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_9 = NULL;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_10 = NULL;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_11;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_12 = NULL;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("clone", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":182
 *         cdef size_t i, n
 *         cdef PeakBase peak
 *         if deep:             # <<<<<<<<<<<<<<
 *             peaks = []
 *             n = self.get_members_size()
 */

#line 182 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_deep); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 182, __pyx_L1_error)

#line 182 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":183
 *         cdef PeakBase peak
 *         if deep:
 *             peaks = []             # <<<<<<<<<<<<<<
 *             n = self.get_members_size()
 *             for i in range(n):
 */

#line 183 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 183, __pyx_L1_error)

#line 183 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 183 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_peaks = ((PyObject*)__pyx_t_2);

#line 183 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":184
 *         if deep:
 *             peaks = []
 *             n = self.get_members_size()             # <<<<<<<<<<<<<<
 *             for i in range(n):
 *                 peak = self.getitem(i)
 */

#line 184 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_n = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_get_members_size(__pyx_v_self);

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":185
 *             peaks = []
 *             n = self.get_members_size()
 *             for i in range(n):             # <<<<<<<<<<<<<<
 *                 peak = self.getitem(i)
 *                 if peak is not None:
 */

#line 185 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = __pyx_v_n;

#line 185 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = __pyx_t_3;

#line 185 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {

#line 185 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_i = __pyx_t_5;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":186
 *             n = self.get_members_size()
 *             for i in range(n):
 *                 peak = self.getitem(i)             # <<<<<<<<<<<<<<
 *                 if peak is not None:
 *                     peak = peak.clone()
 */

#line 186 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self->__pyx_vtab)->getitem(__pyx_v_self, __pyx_v_i)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 186, __pyx_L1_error)

#line 186 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_2);

#line 186 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_2));

#line 186 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_2 = 0;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":187
 *             for i in range(n):
 *                 peak = self.getitem(i)
 *                 if peak is not None:             # <<<<<<<<<<<<<<
 *                     peak = peak.clone()
 *                 peaks.append(peak)
 */

#line 187 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = (((PyObject *)__pyx_v_peak) != Py_None);

#line 187 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_6 = (__pyx_t_1 != 0);

#line 187 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (__pyx_t_6) {

        /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":188
 *                 peak = self.getitem(i)
 *                 if peak is not None:
 *                     peak = peak.clone()             # <<<<<<<<<<<<<<
 *                 peaks.append(peak)
 *         else:
 */

#line 188 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_peak->__pyx_vtab)->clone(__pyx_v_peak, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 188, __pyx_L1_error)

#line 188 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 188 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_2));

#line 188 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = 0;

        /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":187
 *             for i in range(n):
 *                 peak = self.getitem(i)
 *                 if peak is not None:             # <<<<<<<<<<<<<<
 *                     peak = peak.clone()
 *                 peaks.append(peak)
 */

#line 187 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":189
 *                 if peak is not None:
 *                     peak = peak.clone()
 *                 peaks.append(peak)             # <<<<<<<<<<<<<<
 *         else:
 *             peaks = list(self.members)
 */

#line 189 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_peaks, ((PyObject *)__pyx_v_peak)); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 189, __pyx_L1_error)

#line 189 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":182
 *         cdef size_t i, n
 *         cdef PeakBase peak
 *         if deep:             # <<<<<<<<<<<<<<
 *             peaks = []
 *             n = self.get_members_size()
 */

#line 182 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L3;

#line 182 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":191
 *                 peaks.append(peak)
 *         else:
 *             peaks = list(self.members)             # <<<<<<<<<<<<<<
 *         node = self.__class__(self.time, peaks)
 *         node.node_id = self.node_id
 */

#line 191 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /*else*/ {

#line 191 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = PySequence_List(__pyx_v_self->members); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 191, __pyx_L1_error)

#line 191 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 191 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_peaks = ((PyObject*)__pyx_t_2);

#line 191 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = 0;

#line 191 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 191 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3:;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":192
 *         else:
 *             peaks = list(self.members)
 *         node = self.__class__(self.time, peaks)             # <<<<<<<<<<<<<<
 *         node.node_id = self.node_id
 *         return node
 */

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 192, __pyx_L1_error)

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_8);

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_9 = PyFloat_FromDouble(__pyx_v_self->time); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 192, __pyx_L1_error)

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_9);

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_10 = NULL;

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_11 = 0;

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_8);

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (likely(__pyx_t_10)) {

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(__pyx_t_10);

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(function);

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF_SET(__pyx_t_8, function);

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_11 = 1;

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #if CYTHON_FAST_PYCALL

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (PyFunction_Check(__pyx_t_8)) {

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_9, __pyx_v_peaks};

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 192, __pyx_L1_error)

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #endif

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #if CYTHON_FAST_PYCCALL

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_9, __pyx_v_peaks};

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 192, __pyx_L1_error)

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #endif

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_12 = PyTuple_New(2+__pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 192, __pyx_L1_error)

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_12);

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_10) {

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_10); __pyx_t_10 = NULL;

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GIVEREF(__pyx_t_9);

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_11, __pyx_t_9);

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(__pyx_v_peaks);

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GIVEREF(__pyx_v_peaks);

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_11, __pyx_v_peaks);

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_9 = 0;

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_12, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 192, __pyx_L1_error)

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase))))) __PYX_ERR(0, 192, __pyx_L1_error)

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_node = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_t_2);

#line 192 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":193
 *             peaks = list(self.members)
 *         node = self.__class__(self.time, peaks)
 *         node.node_id = self.node_id             # <<<<<<<<<<<<<<
 *         return node
 * 
 */

#line 193 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __pyx_v_self->node_id;

#line 193 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_t_2);

#line 193 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 193 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_node->node_id);

#line 193 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_v_node->node_id);

#line 193 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_node->node_id = __pyx_t_2;

#line 193 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":194
 *         node = self.__class__(self.time, peaks)
 *         node.node_id = self.node_id
 *         return node             # <<<<<<<<<<<<<<
 * 
 *     def _unspool_strip_children(self):
 */

#line 194 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 194 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_node));

#line 194 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = ((PyObject *)__pyx_v_node);

#line 194 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":177
 * cdef class LCMSFeatureTreeNodeBase(object):
 * 
 *     def clone(self, deep=False):             # <<<<<<<<<<<<<<
 *         cdef LCMSFeatureTreeNodeBase node
 *         cdef list peaks
 */

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_8);

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_9);

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_10);

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_12);

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNodeBase.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_node);

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_peaks);

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_gb_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_4generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":196
 *         return node
 * 
 *     def _unspool_strip_children(self):             # <<<<<<<<<<<<<<
 *         node = self.__class__(
 *             self.time, list(self.members))
 */

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_3_unspool_strip_children(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_2_unspool_strip_children[] = "LCMSFeatureTreeNodeBase._unspool_strip_children(self)";

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_3_unspool_strip_children(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_unspool_strip_children (wrapper)", 0);
  __pyx_r = 
#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_2_unspool_strip_children(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self));

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_2_unspool_strip_children(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self) {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children *__pyx_cur_scope
#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_unspool_strip_children", 0);

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_cur_scope = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children *)__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children, __pyx_empty_tuple, NULL);

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__pyx_cur_scope)) {

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_cur_scope = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children *)Py_None);

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(Py_None);

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __PYX_ERR(0, 196, __pyx_L1_error)

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_cur_scope);

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_4generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_unspool_strip_children, __pyx_n_s_LCMSFeatureTreeNodeBase__unspool, __pyx_n_s_ms_deisotope__c_feature_map_lcms); if (unlikely(!gen)) __PYX_ERR(0, 196, __pyx_L1_error)

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_cur_scope);

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_RefNannyFinishContext();

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    return (PyObject *) gen;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNodeBase._unspool_strip_children", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_gb_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_4generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children *__pyx_cur_scope = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children *)__pyx_generator->closure);

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_6;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_7 = NULL;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_unspool_strip_children", 0);

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  switch (__pyx_generator->resume_label) {

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    case 0: goto __pyx_L3_first_run;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    case 1: goto __pyx_L4_resume_from_yield;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    default: /* CPython raises the right error here */

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_RefNannyFinishContext();

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    return NULL;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_first_run:;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 196, __pyx_L1_error)

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":197
 * 
 *     def _unspool_strip_children(self):
 *         node = self.__class__(             # <<<<<<<<<<<<<<
 *             self.time, list(self.members))
 *         yield node
 */

#line 197 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 197, __pyx_L1_error)

#line 197 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":198
 *     def _unspool_strip_children(self):
 *         node = self.__class__(
 *             self.time, list(self.members))             # <<<<<<<<<<<<<<
 *         yield node
 * 
 */

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = PyFloat_FromDouble(__pyx_cur_scope->__pyx_v_self->time); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 198, __pyx_L1_error)

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = PySequence_List(__pyx_cur_scope->__pyx_v_self->members); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 198, __pyx_L1_error)

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = NULL;

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = 0;

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (likely(__pyx_t_5)) {

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(__pyx_t_5);

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(function);

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF_SET(__pyx_t_2, function);

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_6 = 1;

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #if CYTHON_FAST_PYCALL

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (PyFunction_Check(__pyx_t_2)) {

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_3, __pyx_t_4};

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 197, __pyx_L1_error)

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #endif

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #if CYTHON_FAST_PYCCALL

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_3, __pyx_t_4};

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 197, __pyx_L1_error)

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #endif

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 197, __pyx_L1_error)

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_7);

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_5) {

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GIVEREF(__pyx_t_3);

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_t_3);

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GIVEREF(__pyx_t_4);

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_4);

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = 0;

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = 0;

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 197, __pyx_L1_error)

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_cur_scope->__pyx_v_node = __pyx_t_1;

#line 198 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":199
 *         node = self.__class__(
 *             self.time, list(self.members))
 *         yield node             # <<<<<<<<<<<<<<
 * 
 *     def __getstate__(self):
 */

#line 199 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_node);

#line 199 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_cur_scope->__pyx_v_node;

#line 199 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 199 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 199 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);

#line 199 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* return from generator, yielding value */

#line 199 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_generator->resume_label = 1;

#line 199 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 199 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L4_resume_from_yield:;

#line 199 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 199, __pyx_L1_error)

#line 199 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":196
 *         return node
 * 
 *     def _unspool_strip_children(self):             # <<<<<<<<<<<<<<
 *         node = self.__class__(
 *             self.time, list(self.members))
 */

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyErr_SetNone(PyExc_StopIteration);

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_7);

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("_unspool_strip_children", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #if !CYTHON_USE_EXC_INFO_STACK

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #endif

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_generator->resume_label = -1;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 196 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":201
 *         yield node
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return (self.time, self.members, self.node_id)
 * 
 */

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_6__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_5__getstate__[] = "LCMSFeatureTreeNodeBase.__getstate__(self)";

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_6__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = 
#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_5__getstate__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self));

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_5__getstate__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self) {

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":202
 * 
 *     def __getstate__(self):
 *         return (self.time, self.members, self.node_id)             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, state):
 */

#line 202 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 202 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 202, __pyx_L1_error)

#line 202 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 202 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 202, __pyx_L1_error)

#line 202 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 202 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 202 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);

#line 202 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_self->members);

#line 202 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_v_self->members);

#line 202 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_self->members);

#line 202 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_self->node_id);

#line 202 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_v_self->node_id);

#line 202 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_self->node_id);

#line 202 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 202 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_2;

#line 202 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = 0;

#line 202 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":201
 *         yield node
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return (self.time, self.members, self.node_id)
 * 
 */

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNodeBase.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 201 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":204
 *         return (self.time, self.members, self.node_id)
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.time, self.members, self.node_id = state
 *         self._recalculate()
 */

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_8__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state); /*proto*/

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7__setstate__[] = "LCMSFeatureTreeNodeBase.__setstate__(self, state)";

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_8__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state) {

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  __pyx_r = 
#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7__setstate__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self), ((PyObject *)__pyx_v_state));

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7__setstate__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self, PyObject *__pyx_v_state) {

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*__pyx_t_5)(PyObject *);

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_6;

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":205
 * 
 *     def __setstate__(self, state):
 *         self.time, self.members, self.node_id = state             # <<<<<<<<<<<<<<
 *         self._recalculate()
 * 
 */

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if ((likely(PyTuple_CheckExact(__pyx_v_state))) || (PyList_CheckExact(__pyx_v_state))) {

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject* sequence = __pyx_v_state;

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(size != 3)) {

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __PYX_ERR(0, 205, __pyx_L1_error)

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (likely(PyTuple_CheckExact(sequence))) {

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 2); 

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_3 = PyList_GET_ITEM(sequence, 2); 

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(__pyx_t_2);

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(__pyx_t_3);

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #else

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 205, __pyx_L1_error)

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 205, __pyx_L1_error)

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 205, __pyx_L1_error)

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    Py_ssize_t index = -1;

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = PyObject_GetIter(__pyx_v_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 205, __pyx_L1_error)

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_1 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_1)) 
#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L3_unpacking_failed;

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);
    index = 1; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) 
#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L3_unpacking_failed;

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_2);
    index = 2; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) 
#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L3_unpacking_failed;

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 3) < 0) __PYX_ERR(0, 205, __pyx_L1_error)

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = NULL;

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L4_unpacking_done;

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L3_unpacking_failed:;

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = NULL;

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __PYX_ERR(0, 205, __pyx_L1_error)

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L4_unpacking_done:;

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_6 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 205, __pyx_L1_error)

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 205, __pyx_L1_error)

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->time = __pyx_t_6;

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_self->members);

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_v_self->members);

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->members = ((PyObject*)__pyx_t_2);

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = 0;

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_self->node_id);

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_v_self->node_id);

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->node_id = __pyx_t_3;

#line 205 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":206
 *     def __setstate__(self, state):
 *         self.time, self.members, self.node_id = state
 *         self._recalculate()             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */

#line 206 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self->__pyx_vtab)->_recalculate(__pyx_v_self, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 206, __pyx_L1_error)

#line 206 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 206 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":204
 *         return (self.time, self.members, self.node_id)
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.time, self.members, self.node_id = state
 *         self._recalculate()
 */

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNodeBase.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 204 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":208
 *         self._recalculate()
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.time, []), self.__getstate__()
 * 
 */

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_10__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_9__reduce__[] = "LCMSFeatureTreeNodeBase.__reduce__(self)";

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_10__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = 
#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_9__reduce__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self));

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_9__reduce__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self) {

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":209
 * 
 *     def __reduce__(self):
 *         return self.__class__, (self.time, []), self.__getstate__()             # <<<<<<<<<<<<<<
 * 
 *     def add(self, node, recalculate=True):
 */

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 209, __pyx_L1_error)

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->time); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 209, __pyx_L1_error)

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 209, __pyx_L1_error)

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 209, __pyx_L1_error)

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = 0;

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = 0;

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_getstate); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 209, __pyx_L1_error)

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = NULL;

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (likely(__pyx_t_5)) {

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(__pyx_t_5);

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(function);

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF_SET(__pyx_t_2, function);

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 209, __pyx_L1_error)

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 209, __pyx_L1_error)

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_4);

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_3);

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = 0;

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = 0;

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_2;

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = 0;

#line 209 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":208
 *         self._recalculate()
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.time, []), self.__getstate__()
 * 
 */

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNodeBase.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 208 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":211
 *         return self.__class__, (self.time, []), self.__getstate__()
 * 
 *     def add(self, node, recalculate=True):             # <<<<<<<<<<<<<<
 *         if self.node_id == node.node_id:
 *             raise ValueError("Duplicate Node %s" % node)
 */

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_12add(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_11add[] = "LCMSFeatureTreeNodeBase.add(self, node, recalculate=True)";

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_12add(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_node = 0
#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyObject *__pyx_v_recalculate = 0
#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("add (wrapper)", 0);

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_node,&__pyx_n_s_recalculate,0};

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject* values[2] = {0,0};

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    values[1] = ((PyObject *)Py_True);

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(__pyx_kwds)) {

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      Py_ssize_t kw_args;

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {
        case  2: 
#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0: break;
        default: 
#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0:

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_node)) != 0)) kw_args--;
        else 
#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  1:

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (kw_args > 0) {

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_recalculate);

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (value) { values[1] = value; kw_args--; }

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(kw_args > 0)) {

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "add") < 0)) __PYX_ERR(0, 211, __pyx_L3_error)

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: 
#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        break;
        default: 
#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_node = values[0];

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_recalculate = values[1];

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__PYX_ERR(0, 211, __pyx_L3_error)

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_error:;

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNodeBase.add", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return NULL;

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_11add(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self), __pyx_v_node, __pyx_v_recalculate);

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_11add(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self, PyObject *__pyx_v_node, PyObject *__pyx_v_recalculate) {

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_3;

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_4;

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("add", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":212
 * 
 *     def add(self, node, recalculate=True):
 *         if self.node_id == node.node_id:             # <<<<<<<<<<<<<<
 *             raise ValueError("Duplicate Node %s" % node)
 *         self.members.extend(node.members)
 */

#line 212 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_node, __pyx_n_s_node_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 212, __pyx_L1_error)

#line 212 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 212 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_self->node_id, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 212, __pyx_L1_error)

#line 212 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 212 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 212, __pyx_L1_error)

#line 212 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 212 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_t_3)) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":213
 *     def add(self, node, recalculate=True):
 *         if self.node_id == node.node_id:
 *             raise ValueError("Duplicate Node %s" % node)             # <<<<<<<<<<<<<<
 *         self.members.extend(node.members)
 *         if recalculate:
 */

#line 213 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Duplicate_Node_s, __pyx_v_node); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 213, __pyx_L1_error)

#line 213 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 213 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 213, __pyx_L1_error)

#line 213 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 213 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 213 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 213 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 213 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __PYX_ERR(0, 213, __pyx_L1_error)

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":212
 * 
 *     def add(self, node, recalculate=True):
 *         if self.node_id == node.node_id:             # <<<<<<<<<<<<<<
 *             raise ValueError("Duplicate Node %s" % node)
 *         self.members.extend(node.members)
 */

#line 212 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":214
 *         if self.node_id == node.node_id:
 *             raise ValueError("Duplicate Node %s" % node)
 *         self.members.extend(node.members)             # <<<<<<<<<<<<<<
 *         if recalculate:
 *             self._recalculate()
 */

#line 214 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_v_self->members == Py_None)) {

#line 214 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "extend");

#line 214 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __PYX_ERR(0, 214, __pyx_L1_error)

#line 214 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 214 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_node, __pyx_n_s_members); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 214, __pyx_L1_error)

#line 214 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 214 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = __Pyx_PyList_Extend(__pyx_v_self->members, __pyx_t_1); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 214, __pyx_L1_error)

#line 214 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":215
 *             raise ValueError("Duplicate Node %s" % node)
 *         self.members.extend(node.members)
 *         if recalculate:             # <<<<<<<<<<<<<<
 *             self._recalculate()
 * 
 */

#line 215 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_recalculate); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 215, __pyx_L1_error)

#line 215 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_3) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":216
 *         self.members.extend(node.members)
 *         if recalculate:
 *             self._recalculate()             # <<<<<<<<<<<<<<
 * 
 *     cpdef double _total_intensity_members(self):
 */

#line 216 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self->__pyx_vtab)->_recalculate(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)

#line 216 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 216 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":215
 *             raise ValueError("Duplicate Node %s" % node)
 *         self.members.extend(node.members)
 *         if recalculate:             # <<<<<<<<<<<<<<
 *             self._recalculate()
 * 
 */

#line 215 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":211
 *         return self.__class__, (self.time, []), self.__getstate__()
 * 
 *     def add(self, node, recalculate=True):             # <<<<<<<<<<<<<<
 *         if self.node_id == node.node_id:
 *             raise ValueError("Duplicate Node %s" % node)
 */

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNodeBase.add", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 211 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":218
 *             self._recalculate()
 * 
 *     cpdef double _total_intensity_members(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double total
 */

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_14_total_intensity_members(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase__total_intensity_members(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self, int __pyx_skip_dispatch) {
  double __pyx_v_total
#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_n
#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_i
#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak = 0
#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_r;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_5;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_6;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_7;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_8;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_total_intensity_members", 0);

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_total_intensity_members); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 218, __pyx_L1_error)

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_14_total_intensity_members)) {

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_4 = NULL;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_4)) {

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 218, __pyx_L1_error)

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 218, __pyx_L1_error)

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = __pyx_t_5;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":223
 *             size_t n, i
 *             PeakBase peak
 *         total = 0.             # <<<<<<<<<<<<<<
 *         n = self.get_members_size()
 *         for i in range(n):
 */

#line 223 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_total = 0.;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":224
 *             PeakBase peak
 *         total = 0.
 *         n = self.get_members_size()             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             peak = self.getitem(i)
 */

#line 224 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_n = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_get_members_size(__pyx_v_self);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":225
 *         total = 0.
 *         n = self.get_members_size()
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             peak = self.getitem(i)
 *             total += peak.intensity
 */

#line 225 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = __pyx_v_n;

#line 225 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_7 = __pyx_t_6;

#line 225 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {

#line 225 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_i = __pyx_t_8;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":226
 *         n = self.get_members_size()
 *         for i in range(n):
 *             peak = self.getitem(i)             # <<<<<<<<<<<<<<
 *             total += peak.intensity
 *         return total
 */

#line 226 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self->__pyx_vtab)->getitem(__pyx_v_self, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 226, __pyx_L1_error)

#line 226 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 226 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1));

#line 226 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":227
 *         for i in range(n):
 *             peak = self.getitem(i)
 *             total += peak.intensity             # <<<<<<<<<<<<<<
 *         return total
 * 
 */

#line 227 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_total = (__pyx_v_total + __pyx_v_peak->intensity);

#line 227 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":228
 *             peak = self.getitem(i)
 *             total += peak.intensity
 *         return total             # <<<<<<<<<<<<<<
 * 
 *     cpdef double max_intensity(self):
 */

#line 228 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_v_total;

#line 228 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":218
 *             self._recalculate()
 * 
 *     cpdef double _total_intensity_members(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double total
 */

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNodeBase._total_intensity_members", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_14_total_intensity_members(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_13_total_intensity_members[] = "LCMSFeatureTreeNodeBase._total_intensity_members(self) -> double";

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_14_total_intensity_members(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_total_intensity_members (wrapper)", 0);
  __pyx_r = 
#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_13_total_intensity_members(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self));

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_13_total_intensity_members(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self) {

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_total_intensity_members", 0);

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase__total_intensity_members(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 218, __pyx_L1_error)

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNodeBase._total_intensity_members", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 218 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":230
 *         return total
 * 
 *     cpdef double max_intensity(self):             # <<<<<<<<<<<<<<
 *         return self._most_abundant_member.intensity
 * 
 */

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_16max_intensity(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_max_intensity(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self, int __pyx_skip_dispatch) {

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_r;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_5;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("max_intensity", 0);

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_max_intensity); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 230, __pyx_L1_error)

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_16max_intensity)) {

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_4 = NULL;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_4)) {

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 230, __pyx_L1_error)

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 230, __pyx_L1_error)

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = __pyx_t_5;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":231
 * 
 *     cpdef double max_intensity(self):
 *         return self._most_abundant_member.intensity             # <<<<<<<<<<<<<<
 * 
 *     cpdef double total_intensity(self):
 */

#line 231 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_v_self->_most_abundant_member->intensity;

#line 231 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":230
 *         return total
 * 
 *     cpdef double max_intensity(self):             # <<<<<<<<<<<<<<
 *         return self._most_abundant_member.intensity
 * 
 */

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNodeBase.max_intensity", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_16max_intensity(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_15max_intensity[] = "LCMSFeatureTreeNodeBase.max_intensity(self) -> double";

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_16max_intensity(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("max_intensity (wrapper)", 0);
  __pyx_r = 
#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_15max_intensity(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self));

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_15max_intensity(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self) {

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("max_intensity", 0);

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_max_intensity(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 230, __pyx_L1_error)

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNodeBase.max_intensity", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 230 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":233
 *         return self._most_abundant_member.intensity
 * 
 *     cpdef double total_intensity(self):             # <<<<<<<<<<<<<<
 *         return self._total_intensity_members()
 * 
 */

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_18total_intensity(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_total_intensity(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self, int __pyx_skip_dispatch) {

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_r;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_5;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("total_intensity", 0);

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_total_intensity); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 233, __pyx_L1_error)

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_18total_intensity)) {

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_4 = NULL;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_4)) {

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 233, __pyx_L1_error)

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 233, __pyx_L1_error)

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = __pyx_t_5;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":234
 * 
 *     cpdef double total_intensity(self):
 *         return self._total_intensity_members()             # <<<<<<<<<<<<<<
 * 
 *     cpdef _recalculate(self):
 */

#line 234 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self->__pyx_vtab)->_total_intensity_members(__pyx_v_self, 0);

#line 234 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":233
 *         return self._most_abundant_member.intensity
 * 
 *     cpdef double total_intensity(self):             # <<<<<<<<<<<<<<
 *         return self._total_intensity_members()
 * 
 */

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNodeBase.total_intensity", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_18total_intensity(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_17total_intensity[] = "LCMSFeatureTreeNodeBase.total_intensity(self) -> double";

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_18total_intensity(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("total_intensity (wrapper)", 0);
  __pyx_r = 
#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_17total_intensity(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self));

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_17total_intensity(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self) {

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("total_intensity", 0);

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_total_intensity(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 233, __pyx_L1_error)

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNodeBase.total_intensity", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 233 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":236
 *         return self._total_intensity_members()
 * 
 *     cpdef _recalculate(self):             # <<<<<<<<<<<<<<
 *         self._calculate_most_abundant_member()
 * 
 */

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_20_recalculate(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase__recalculate(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self, int __pyx_skip_dispatch) {

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_recalculate", 0);

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_recalculate_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 236, __pyx_L1_error)

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_20_recalculate)) {

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_4 = NULL;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_4)) {

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 236, __pyx_L1_error)

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = __pyx_t_2;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = 0;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":237
 * 
 *     cpdef _recalculate(self):
 *         self._calculate_most_abundant_member()             # <<<<<<<<<<<<<<
 * 
 *     cdef inline size_t get_members_size(self):
 */

#line 237 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self->__pyx_vtab)->_calculate_most_abundant_member(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 237, __pyx_L1_error)

#line 237 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 237 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":236
 *         return self._total_intensity_members()
 * 
 *     cpdef _recalculate(self):             # <<<<<<<<<<<<<<
 *         self._calculate_most_abundant_member()
 * 
 */

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNodeBase._recalculate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_20_recalculate(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_19_recalculate[] = "LCMSFeatureTreeNodeBase._recalculate(self)";

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_20_recalculate(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_recalculate (wrapper)", 0);
  __pyx_r = 
#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_19_recalculate(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self));

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_19_recalculate(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self) {

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_recalculate", 0);

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase__recalculate(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 236, __pyx_L1_error)

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNodeBase._recalculate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 236 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":239
 *         self._calculate_most_abundant_member()
 * 
 *     cdef inline size_t get_members_size(self):             # <<<<<<<<<<<<<<
 *         return PyList_GET_SIZE(self.members)
 * 
 */

#line 239 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 239 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static CYTHON_INLINE size_t __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_get_members_size(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self) {

#line 239 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_r;

#line 239 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 239 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 239 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("get_members_size", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":240
 * 
 *     cdef inline size_t get_members_size(self):
 *         return PyList_GET_SIZE(self.members)             # <<<<<<<<<<<<<<
 * 
 *     cdef PeakBase _find_most_abundant_member(self):
 */

#line 240 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_v_self->members;

#line 240 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 240 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = PyList_GET_SIZE(__pyx_t_1);

#line 240 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 240 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":239
 *         self._calculate_most_abundant_member()
 * 
 *     cdef inline size_t get_members_size(self):             # <<<<<<<<<<<<<<
 *         return PyList_GET_SIZE(self.members)
 * 
 */

#line 239 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 239 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 239 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 239 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 239 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 239 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":242
 *         return PyList_GET_SIZE(self.members)
 * 
 *     cdef PeakBase _find_most_abundant_member(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase__find_most_abundant_member(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self) {
  size_t __pyx_v_i
#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_n
#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak = 0
#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_best = 0
#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  double __pyx_v_best_intensity
#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_r = NULL;

#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_2;

#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_3;

#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_4;

#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_5;

#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_6;

#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_find_most_abundant_member", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":247
 *             PeakBase peak, best
 *             double best_intensity
 *         n = self.get_members_size()             # <<<<<<<<<<<<<<
 *         best = self.getitem(0)
 *         best_intensity = best.intensity
 */

#line 247 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_n = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_get_members_size(__pyx_v_self);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":248
 *             double best_intensity
 *         n = self.get_members_size()
 *         best = self.getitem(0)             # <<<<<<<<<<<<<<
 *         best_intensity = best.intensity
 *         for i in range(n):
 */

#line 248 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self->__pyx_vtab)->getitem(__pyx_v_self, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 248, __pyx_L1_error)

#line 248 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 248 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_best = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1);

#line 248 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":249
 *         n = self.get_members_size()
 *         best = self.getitem(0)
 *         best_intensity = best.intensity             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             peak = self.getitem(i)
 */

#line 249 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __pyx_v_best->intensity;

#line 249 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_best_intensity = __pyx_t_2;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":250
 *         best = self.getitem(0)
 *         best_intensity = best.intensity
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             peak = self.getitem(i)
 *             if peak.intensity > best_intensity:
 */

#line 250 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = __pyx_v_n;

#line 250 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = __pyx_t_3;

#line 250 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {

#line 250 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_i = __pyx_t_5;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":251
 *         best_intensity = best.intensity
 *         for i in range(n):
 *             peak = self.getitem(i)             # <<<<<<<<<<<<<<
 *             if peak.intensity > best_intensity:
 *                 best = peak
 */

#line 251 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self->__pyx_vtab)->getitem(__pyx_v_self, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 251, __pyx_L1_error)

#line 251 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 251 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1));

#line 251 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":252
 *         for i in range(n):
 *             peak = self.getitem(i)
 *             if peak.intensity > best_intensity:             # <<<<<<<<<<<<<<
 *                 best = peak
 *                 best_intensity = best.intensity
 */

#line 252 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_6 = ((__pyx_v_peak->intensity > __pyx_v_best_intensity) != 0);

#line 252 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_6) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":253
 *             peak = self.getitem(i)
 *             if peak.intensity > best_intensity:
 *                 best = peak             # <<<<<<<<<<<<<<
 *                 best_intensity = best.intensity
 *         return best
 */

#line 253 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(((PyObject *)__pyx_v_peak));

#line 253 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF_SET(__pyx_v_best, __pyx_v_peak);

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":254
 *             if peak.intensity > best_intensity:
 *                 best = peak
 *                 best_intensity = best.intensity             # <<<<<<<<<<<<<<
 *         return best
 * 
 */

#line 254 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_2 = __pyx_v_best->intensity;

#line 254 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_best_intensity = __pyx_t_2;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":252
 *         for i in range(n):
 *             peak = self.getitem(i)
 *             if peak.intensity > best_intensity:             # <<<<<<<<<<<<<<
 *                 best = peak
 *                 best_intensity = best.intensity
 */

#line 252 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 252 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":255
 *                 best = peak
 *                 best_intensity = best.intensity
 *         return best             # <<<<<<<<<<<<<<
 * 
 *     cpdef _calculate_most_abundant_member(self):
 */

#line 255 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 255 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_best));

#line 255 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_v_best;

#line 255 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":242
 *         return PyList_GET_SIZE(self.members)
 * 
 *     cdef PeakBase _find_most_abundant_member(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNodeBase._find_most_abundant_member", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_best);

#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 242 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":257
 *         return best
 * 
 *     cpdef _calculate_most_abundant_member(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t n
 */

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_22_calculate_most_abundant_member(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase__calculate_most_abundant_member(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self, int __pyx_skip_dispatch) {
  size_t __pyx_v_n
#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_5;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_calculate_most_abundant_member", 0);

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_calculate_most_abundant_member); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 257, __pyx_L1_error)

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_22_calculate_most_abundant_member)) {

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_4 = NULL;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_4)) {

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 257, __pyx_L1_error)

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = __pyx_t_2;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = 0;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":260
 *         cdef:
 *             size_t n
 *         n = self.get_members_size()             # <<<<<<<<<<<<<<
 *         if n == 1:
 *             self._most_abundant_member = self.getitem(0)
 */

#line 260 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_n = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_get_members_size(__pyx_v_self);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":261
 *             size_t n
 *         n = self.get_members_size()
 *         if n == 1:             # <<<<<<<<<<<<<<
 *             self._most_abundant_member = self.getitem(0)
 *         else:
 */

#line 261 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = ((__pyx_v_n == 1) != 0);

#line 261 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":262
 *         n = self.get_members_size()
 *         if n == 1:
 *             self._most_abundant_member = self.getitem(0)             # <<<<<<<<<<<<<<
 *         else:
 *             if n == 0:
 */

#line 262 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self->__pyx_vtab)->getitem(__pyx_v_self, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 262, __pyx_L1_error)

#line 262 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 262 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GIVEREF(__pyx_t_1);

#line 262 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_v_self->_most_abundant_member);

#line 262 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(((PyObject *)__pyx_v_self->_most_abundant_member));

#line 262 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_self->_most_abundant_member = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1);

#line 262 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":261
 *             size_t n
 *         n = self.get_members_size()
 *         if n == 1:             # <<<<<<<<<<<<<<
 *             self._most_abundant_member = self.getitem(0)
 *         else:
 */

#line 261 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L3;

#line 261 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":264
 *             self._most_abundant_member = self.getitem(0)
 *         else:
 *             if n == 0:             # <<<<<<<<<<<<<<
 *                 self._most_abundant_member = None
 *             else:
 */

#line 264 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /*else*/ {

#line 264 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = ((__pyx_v_n == 0) != 0);

#line 264 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":265
 *         else:
 *             if n == 0:
 *                 self._most_abundant_member = None             # <<<<<<<<<<<<<<
 *             else:
 *                 self._most_abundant_member = self._find_most_abundant_member()
 */

#line 265 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(Py_None);

#line 265 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GIVEREF(Py_None);

#line 265 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_v_self->_most_abundant_member);

#line 265 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(((PyObject *)__pyx_v_self->_most_abundant_member));

#line 265 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_self->_most_abundant_member = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)Py_None);

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":264
 *             self._most_abundant_member = self.getitem(0)
 *         else:
 *             if n == 0:             # <<<<<<<<<<<<<<
 *                 self._most_abundant_member = None
 *             else:
 */

#line 264 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L4;

#line 264 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":267
 *                 self._most_abundant_member = None
 *             else:
 *                 self._most_abundant_member = self._find_most_abundant_member()             # <<<<<<<<<<<<<<
 *         return
 * 
 */

#line 267 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    /*else*/ {

#line 267 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self->__pyx_vtab)->_find_most_abundant_member(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 267, __pyx_L1_error)

#line 267 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 267 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GIVEREF(__pyx_t_1);

#line 267 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_v_self->_most_abundant_member);

#line 267 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(((PyObject *)__pyx_v_self->_most_abundant_member));

#line 267 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_self->_most_abundant_member = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1);

#line 267 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = 0;

#line 267 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 267 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L4:;

#line 267 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 267 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3:;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":268
 *             else:
 *                 self._most_abundant_member = self._find_most_abundant_member()
 *         return             # <<<<<<<<<<<<<<
 * 
 *     cdef PeakBase getitem(self, size_t i):
 */

#line 268 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 268 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 268 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":257
 *         return best
 * 
 *     cpdef _calculate_most_abundant_member(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t n
 */

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNodeBase._calculate_most_abundant_member", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_22_calculate_most_abundant_member(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_21_calculate_most_abundant_member[] = "LCMSFeatureTreeNodeBase._calculate_most_abundant_member(self)";

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_22_calculate_most_abundant_member(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_calculate_most_abundant_member (wrapper)", 0);
  __pyx_r = 
#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_21_calculate_most_abundant_member(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self));

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_21_calculate_most_abundant_member(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self) {

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_calculate_most_abundant_member", 0);

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase__calculate_most_abundant_member(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 257, __pyx_L1_error)

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNodeBase._calculate_most_abundant_member", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 257 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":270
 *         return
 * 
 *     cdef PeakBase getitem(self, size_t i):             # <<<<<<<<<<<<<<
 *         return <PeakBase>PyList_GET_ITEM(self.members, i)
 * 
 */

#line 270 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 270 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_getitem(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self, size_t __pyx_v_i) {

#line 270 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_r = NULL;

#line 270 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 270 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 270 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2;

#line 270 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("getitem", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":271
 * 
 *     cdef PeakBase getitem(self, size_t i):
 *         return <PeakBase>PyList_GET_ITEM(self.members, i)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 271 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 271 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_v_self->members;

#line 271 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 271 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_v_i);

#line 271 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 271 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(((PyObject *)((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_2)));

#line 271 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_2);

#line 271 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":270
 *         return
 * 
 *     cdef PeakBase getitem(self, size_t i):             # <<<<<<<<<<<<<<
 *         return <PeakBase>PyList_GET_ITEM(self.members, i)
 * 
 */

#line 270 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 270 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 270 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 270 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 270 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 270 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 270 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":274
 * 
 *     @property
 *     def peaks(self):             # <<<<<<<<<<<<<<
 *         peaks = list(self.members)
 *         return peaks
 */

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_5peaks_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_5peaks_1__get__(PyObject *__pyx_v_self) {

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_5peaks___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self));

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_5peaks___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self) {
  PyObject *__pyx_v_peaks = NULL
#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":275
 *     @property
 *     def peaks(self):
 *         peaks = list(self.members)             # <<<<<<<<<<<<<<
 *         return peaks
 * 
 */

#line 275 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PySequence_List(__pyx_v_self->members); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 275, __pyx_L1_error)

#line 275 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 275 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_peaks = ((PyObject*)__pyx_t_1);

#line 275 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":276
 *     def peaks(self):
 *         peaks = list(self.members)
 *         return peaks             # <<<<<<<<<<<<<<
 * 
 *     def __hash__(self):
 */

#line 276 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 276 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_peaks);

#line 276 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_v_peaks;

#line 276 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":274
 * 
 *     @property
 *     def peaks(self):             # <<<<<<<<<<<<<<
 *         peaks = list(self.members)
 *         return peaks
 */

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNodeBase.peaks.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_peaks);

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 274 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":278
 *         return peaks
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash(self.uid)
 * 
 */

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static Py_hash_t __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_24__hash__(PyObject *__pyx_v_self); /*proto*/

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static Py_hash_t __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_24__hash__(PyObject *__pyx_v_self) {

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_hash_t __pyx_r;

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_r = 
#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_23__hash__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self));

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static Py_hash_t __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_23__hash__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self) {

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_hash_t __pyx_r;

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_hash_t __pyx_t_2;

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__hash__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":279
 * 
 *     def __hash__(self):
 *         return hash(self.uid)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 279 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_uid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 279, __pyx_L1_error)

#line 279 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 279 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = PyObject_Hash(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_hash_t)-1))) __PYX_ERR(0, 279, __pyx_L1_error)

#line 279 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 279 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_2;

#line 279 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":278
 *         return peaks
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash(self.uid)
 * 
 */

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNodeBase.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = -1;

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 278 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":25
 * cdef class LCMSFeatureTreeNodeBase(object):
 *     cdef:
 *         public double time             # <<<<<<<<<<<<<<
 *         public list members
 *         public PeakBase _most_abundant_member
 */

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_4time_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_4time_1__get__(PyObject *__pyx_v_self) {

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = 0;

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_4time___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self));

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_4time___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self) {

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = NULL;

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->time); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 25, __pyx_L1_error)

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = __pyx_t_1;

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = 0;

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNodeBase.time.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = NULL;

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_4time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_4time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_4time_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_4time_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self, PyObject *__pyx_v_value) {

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double __pyx_t_1;

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 25, __pyx_L1_error)

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->time = __pyx_t_1;

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNodeBase.time.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = -1;

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 25 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":26
 *     cdef:
 *         public double time
 *         public list members             # <<<<<<<<<<<<<<
 *         public PeakBase _most_abundant_member
 *         public object node_id
 */

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7members_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7members_1__get__(PyObject *__pyx_v_self) {

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = 0;

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7members___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self));

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7members___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self) {

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = NULL;

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(__pyx_v_self->members);

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = __pyx_v_self->members;

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7members_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7members_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7members_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7members_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self, PyObject *__pyx_v_value) {

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  if (!(likely(PyList_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(2, 26, __pyx_L1_error)

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = __pyx_v_value;

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(__pyx_t_1);

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(__pyx_t_1);

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->members);

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(__pyx_v_self->members);

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->members = ((PyObject*)__pyx_t_1);

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = 0;

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNodeBase.members.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = -1;

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7members_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7members_5__del__(PyObject *__pyx_v_self) {

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7members_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self));

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7members_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self) {

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(Py_None);

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(Py_None);

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->members);

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(__pyx_v_self->members);

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->members = ((PyObject*)Py_None);

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":27
 *         public double time
 *         public list members
 *         public PeakBase _most_abundant_member             # <<<<<<<<<<<<<<
 *         public object node_id
 * 
 */

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_21_most_abundant_member_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_21_most_abundant_member_1__get__(PyObject *__pyx_v_self) {

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = 0;

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_21_most_abundant_member___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self));

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_21_most_abundant_member___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self) {

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = NULL;

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->_most_abundant_member));

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = ((PyObject *)__pyx_v_self->_most_abundant_member);

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_21_most_abundant_member_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_21_most_abundant_member_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_21_most_abundant_member_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_21_most_abundant_member_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self, PyObject *__pyx_v_value) {

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase))))) __PYX_ERR(2, 27, __pyx_L1_error)

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = __pyx_v_value;

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(__pyx_t_1);

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(__pyx_t_1);

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->_most_abundant_member);

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_most_abundant_member));

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->_most_abundant_member = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1);

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = 0;

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNodeBase._most_abundant_member.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = -1;

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_21_most_abundant_member_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_21_most_abundant_member_5__del__(PyObject *__pyx_v_self) {

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_21_most_abundant_member_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self));

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_21_most_abundant_member_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self) {

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(Py_None);

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(Py_None);

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->_most_abundant_member);

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_most_abundant_member));

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->_most_abundant_member = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)Py_None);

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 27 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":28
 *         public list members
 *         public PeakBase _most_abundant_member
 *         public object node_id             # <<<<<<<<<<<<<<
 * 
 *     cpdef double _total_intensity_members(self)
 */

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7node_id_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7node_id_1__get__(PyObject *__pyx_v_self) {

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = 0;

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7node_id___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self));

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7node_id___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self) {

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = NULL;

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(__pyx_v_self->node_id);

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = __pyx_v_self->node_id;

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7node_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7node_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7node_id_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7node_id_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self, PyObject *__pyx_v_value) {

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(__pyx_v_value);

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(__pyx_v_value);

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->node_id);

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(__pyx_v_self->node_id);

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->node_id = __pyx_v_value;

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7node_id_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7node_id_5__del__(PyObject *__pyx_v_self) {

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7node_id_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self));

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7node_id_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *__pyx_v_self) {

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(Py_None);

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(Py_None);

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->node_id);

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(__pyx_v_self->node_id);

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->node_id = Py_None;

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":283
 * 
 * cdef class LCMSFeatureTreeNode(LCMSFeatureTreeNodeBase):
 *     def __init__(self, time, members=None):             # <<<<<<<<<<<<<<
 *         if members is None:
 *             members = []
 */

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_time = 0
#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyObject *__pyx_v_members = 0
#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_r;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_time,&__pyx_n_s_members,0};

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject* values[2] = {0,0};

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    values[1] = ((PyObject *)Py_None);

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(__pyx_kwds)) {

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      Py_ssize_t kw_args;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {
        case  2: 
#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0: break;
        default: 
#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0:

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_time)) != 0)) kw_args--;
        else 
#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  1:

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (kw_args > 0) {

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_members);

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (value) { values[1] = value; kw_args--; }

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(kw_args > 0)) {

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 283, __pyx_L3_error)

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: 
#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        break;
        default: 
#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_time = values[0];

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_members = values[1];

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__PYX_ERR(0, 283, __pyx_L3_error)

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_error:;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return -1;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode___init__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_v_self), __pyx_v_time, __pyx_v_members);

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode___init__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_self, PyObject *__pyx_v_time, PyObject *__pyx_v_members) {

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_r;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_1;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_2;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_4;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__init__", 0);

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_members);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":284
 * cdef class LCMSFeatureTreeNode(LCMSFeatureTreeNodeBase):
 *     def __init__(self, time, members=None):
 *         if members is None:             # <<<<<<<<<<<<<<
 *             members = []
 *         self.time = time
 */

#line 284 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = (__pyx_v_members == Py_None);

#line 284 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 284 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":285
 *     def __init__(self, time, members=None):
 *         if members is None:
 *             members = []             # <<<<<<<<<<<<<<
 *         self.time = time
 *         self.members = members
 */

#line 285 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 285, __pyx_L1_error)

#line 285 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 285 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF_SET(__pyx_v_members, __pyx_t_3);

#line 285 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":284
 * cdef class LCMSFeatureTreeNode(LCMSFeatureTreeNodeBase):
 *     def __init__(self, time, members=None):
 *         if members is None:             # <<<<<<<<<<<<<<
 *             members = []
 *         self.time = time
 */

#line 284 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":286
 *         if members is None:
 *             members = []
 *         self.time = time             # <<<<<<<<<<<<<<
 *         self.members = members
 *         self._most_abundant_member = None
 */

#line 286 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_v_time); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 286, __pyx_L1_error)

#line 286 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->__pyx_base.time = __pyx_t_4;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":287
 *             members = []
 *         self.time = time
 *         self.members = members             # <<<<<<<<<<<<<<
 *         self._most_abundant_member = None
 *         self._mz = 0
 */

#line 287 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (!(likely(PyList_CheckExact(__pyx_v_members))||((__pyx_v_members) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_members)->tp_name), 0))) __PYX_ERR(0, 287, __pyx_L1_error)

#line 287 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = __pyx_v_members;

#line 287 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_t_3);

#line 287 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 287 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.members);

#line 287 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_v_self->__pyx_base.members);

#line 287 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->__pyx_base.members = ((PyObject*)__pyx_t_3);

#line 287 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":288
 *         self.time = time
 *         self.members = members
 *         self._most_abundant_member = None             # <<<<<<<<<<<<<<
 *         self._mz = 0
 *         self._recalculate()
 */

#line 288 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(Py_None);

#line 288 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(Py_None);

#line 288 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_self->__pyx_base._most_abundant_member);

#line 288 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx_base._most_abundant_member));

#line 288 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->__pyx_base._most_abundant_member = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)Py_None);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":289
 *         self.members = members
 *         self._most_abundant_member = None
 *         self._mz = 0             # <<<<<<<<<<<<<<
 *         self._recalculate()
 *         self.node_id = uid()
 */

#line 289 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->_mz = 0.0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":290
 *         self._most_abundant_member = None
 *         self._mz = 0
 *         self._recalculate()             # <<<<<<<<<<<<<<
 *         self.node_id = uid()
 * 
 */

#line 290 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._recalculate(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self), 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 290, __pyx_L1_error)

#line 290 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 290 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":291
 *         self._mz = 0
 *         self._recalculate()
 *         self.node_id = uid()             # <<<<<<<<<<<<<<
 * 
 *     cpdef _recalculate(self):
 */

#line 291 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_12ms_deisotope_2_c_11feature_map_12lcms_feature_uid);
  __pyx_t_5 = __pyx_v_12ms_deisotope_2_c_11feature_map_12lcms_feature_uid; 
#line 291 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_6 = NULL;

#line 291 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {

#line 291 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);

#line 291 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (likely(__pyx_t_6)) {

#line 291 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);

#line 291 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(__pyx_t_6);

#line 291 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(function);

#line 291 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF_SET(__pyx_t_5, function);

#line 291 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 291 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 291 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_5);

#line 291 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 291 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 291, __pyx_L1_error)

#line 291 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 291 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 291 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 291 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.node_id);

#line 291 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_v_self->__pyx_base.node_id);

#line 291 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->__pyx_base.node_id = __pyx_t_3;

#line 291 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":283
 * 
 * cdef class LCMSFeatureTreeNode(LCMSFeatureTreeNodeBase):
 *     def __init__(self, time, members=None):             # <<<<<<<<<<<<<<
 *         if members is None:
 *             members = []
 */

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = -1;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_members);

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 283 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":293
 *         self.node_id = uid()
 * 
 *     cpdef _recalculate(self):             # <<<<<<<<<<<<<<
 *         self._calculate_most_abundant_member()
 *         if self._most_abundant_member is not None:
 */

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_3_recalculate(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode__recalculate(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_self, int __pyx_skip_dispatch) {

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_5;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_6;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_7;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_recalculate", 0);

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_recalculate_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 293, __pyx_L1_error)

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_3_recalculate)) {

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_4 = NULL;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_4)) {

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 293, __pyx_L1_error)

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = __pyx_t_2;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = 0;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":294
 * 
 *     cpdef _recalculate(self):
 *         self._calculate_most_abundant_member()             # <<<<<<<<<<<<<<
 *         if self._most_abundant_member is not None:
 *             self._mz = self._most_abundant_member.mz
 */

#line 294 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._calculate_most_abundant_member(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_self), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 294, __pyx_L1_error)

#line 294 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 294 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":295
 *     cpdef _recalculate(self):
 *         self._calculate_most_abundant_member()
 *         if self._most_abundant_member is not None:             # <<<<<<<<<<<<<<
 *             self._mz = self._most_abundant_member.mz
 * 
 */

#line 295 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = (((PyObject *)__pyx_v_self->__pyx_base._most_abundant_member) != Py_None);

#line 295 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = (__pyx_t_5 != 0);

#line 295 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_6) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":296
 *         self._calculate_most_abundant_member()
 *         if self._most_abundant_member is not None:
 *             self._mz = self._most_abundant_member.mz             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 296 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_7 = __pyx_v_self->__pyx_base._most_abundant_member->mz;

#line 296 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_self->_mz = __pyx_t_7;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":295
 *     cpdef _recalculate(self):
 *         self._calculate_most_abundant_member()
 *         if self._most_abundant_member is not None:             # <<<<<<<<<<<<<<
 *             self._mz = self._most_abundant_member.mz
 * 
 */

#line 295 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":293
 *         self.node_id = uid()
 * 
 *     cpdef _recalculate(self):             # <<<<<<<<<<<<<<
 *         self._calculate_most_abundant_member()
 *         if self._most_abundant_member is not None:
 */

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNode._recalculate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_3_recalculate(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_2_recalculate[] = "LCMSFeatureTreeNode._recalculate(self)";

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_3_recalculate(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_recalculate (wrapper)", 0);
  __pyx_r = 
#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_2_recalculate(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_v_self));

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_2_recalculate(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_self) {

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_recalculate", 0);

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode__recalculate(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 293, __pyx_L1_error)

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNode._recalculate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 293 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":299
 * 
 *     @property
 *     def mz(self):             # <<<<<<<<<<<<<<
 *         if self._mz == 0:
 *             if self._most_abundant_member is not None:
 */

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_2mz_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_2mz_1__get__(PyObject *__pyx_v_self) {

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_2mz___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_v_self));

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_2mz___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_self) {

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_1;

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_2;

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_3;

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":300
 *     @property
 *     def mz(self):
 *         if self._mz == 0:             # <<<<<<<<<<<<<<
 *             if self._most_abundant_member is not None:
 *                 self._mz = self._most_abundant_member.mz
 */

#line 300 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((__pyx_v_self->_mz == 0.0) != 0);

#line 300 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":301
 *     def mz(self):
 *         if self._mz == 0:
 *             if self._most_abundant_member is not None:             # <<<<<<<<<<<<<<
 *                 self._mz = self._most_abundant_member.mz
 *         return self._mz
 */

#line 301 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = (((PyObject *)__pyx_v_self->__pyx_base._most_abundant_member) != Py_None);

#line 301 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = (__pyx_t_1 != 0);

#line 301 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":302
 *         if self._mz == 0:
 *             if self._most_abundant_member is not None:
 *                 self._mz = self._most_abundant_member.mz             # <<<<<<<<<<<<<<
 *         return self._mz
 * 
 */

#line 302 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_3 = __pyx_v_self->__pyx_base._most_abundant_member->mz;

#line 302 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_self->_mz = __pyx_t_3;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":301
 *     def mz(self):
 *         if self._mz == 0:
 *             if self._most_abundant_member is not None:             # <<<<<<<<<<<<<<
 *                 self._mz = self._most_abundant_member.mz
 *         return self._mz
 */

#line 301 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":300
 *     @property
 *     def mz(self):
 *         if self._mz == 0:             # <<<<<<<<<<<<<<
 *             if self._most_abundant_member is not None:
 *                 self._mz = self._most_abundant_member.mz
 */

#line 300 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":303
 *             if self._most_abundant_member is not None:
 *                 self._mz = self._most_abundant_member.mz
 *         return self._mz             # <<<<<<<<<<<<<<
 * 
 *     cdef double get_mz(self):
 */

#line 303 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 303 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->_mz); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 303, __pyx_L1_error)

#line 303 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 303 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_4;

#line 303 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = 0;

#line 303 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":299
 * 
 *     @property
 *     def mz(self):             # <<<<<<<<<<<<<<
 *         if self._mz == 0:
 *             if self._most_abundant_member is not None:
 */

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNode.mz.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 299 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":305
 *         return self._mz
 * 
 *     cdef double get_mz(self):             # <<<<<<<<<<<<<<
 *         if self._mz == 0:
 *             if self._most_abundant_member is not None:
 */

#line 305 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 305 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_get_mz(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_self) {

#line 305 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_r;

#line 305 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 305 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_1;

#line 305 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_2;

#line 305 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_3;

#line 305 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("get_mz", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":306
 * 
 *     cdef double get_mz(self):
 *         if self._mz == 0:             # <<<<<<<<<<<<<<
 *             if self._most_abundant_member is not None:
 *                 self._mz = self._most_abundant_member.mz
 */

#line 306 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((__pyx_v_self->_mz == 0.0) != 0);

#line 306 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":307
 *     cdef double get_mz(self):
 *         if self._mz == 0:
 *             if self._most_abundant_member is not None:             # <<<<<<<<<<<<<<
 *                 self._mz = self._most_abundant_member.mz
 *         return self._mz
 */

#line 307 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = (((PyObject *)__pyx_v_self->__pyx_base._most_abundant_member) != Py_None);

#line 307 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = (__pyx_t_1 != 0);

#line 307 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":308
 *         if self._mz == 0:
 *             if self._most_abundant_member is not None:
 *                 self._mz = self._most_abundant_member.mz             # <<<<<<<<<<<<<<
 *         return self._mz
 * 
 */

#line 308 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_3 = __pyx_v_self->__pyx_base._most_abundant_member->mz;

#line 308 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_self->_mz = __pyx_t_3;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":307
 *     cdef double get_mz(self):
 *         if self._mz == 0:
 *             if self._most_abundant_member is not None:             # <<<<<<<<<<<<<<
 *                 self._mz = self._most_abundant_member.mz
 *         return self._mz
 */

#line 307 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":306
 * 
 *     cdef double get_mz(self):
 *         if self._mz == 0:             # <<<<<<<<<<<<<<
 *             if self._most_abundant_member is not None:
 *                 self._mz = self._most_abundant_member.mz
 */

#line 306 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":309
 *             if self._most_abundant_member is not None:
 *                 self._mz = self._most_abundant_member.mz
 *         return self._mz             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint _eq(self, LCMSFeatureTreeNode other):
 */

#line 309 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_v_self->_mz;

#line 309 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":305
 *         return self._mz
 * 
 *     cdef double get_mz(self):             # <<<<<<<<<<<<<<
 *         if self._mz == 0:
 *             if self._most_abundant_member is not None:
 */

#line 305 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 305 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 305 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 305 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 305 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 305 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":311
 *         return self._mz
 * 
 *     cpdef bint _eq(self, LCMSFeatureTreeNode other):             # <<<<<<<<<<<<<<
 *         return self.members == other.members and abs(
 *             self.time - other.time) < 1e-4
 */

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_5_eq(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode__eq(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_other, int __pyx_skip_dispatch) {

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_r;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_5;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_6;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_eq", 0);

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_eq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 311, __pyx_L1_error)

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_5_eq)) {

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_4 = NULL;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_4)) {

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_other)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other));

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 311, __pyx_L1_error)

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 311, __pyx_L1_error)

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = __pyx_t_5;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":312
 * 
 *     cpdef bint _eq(self, LCMSFeatureTreeNode other):
 *         return self.members == other.members and abs(             # <<<<<<<<<<<<<<
 *             self.time - other.time) < 1e-4
 * 
 */

#line 312 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_self->__pyx_base.members, __pyx_v_other->__pyx_base.members, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 312, __pyx_L1_error)

#line 312 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 312, __pyx_L1_error)

#line 312 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 312 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_6) {

#line 312 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {

#line 312 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = __pyx_t_6;

#line 312 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L3_bool_binop_done;

#line 312 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":313
 *     cpdef bint _eq(self, LCMSFeatureTreeNode other):
 *         return self.members == other.members and abs(
 *             self.time - other.time) < 1e-4             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint _ne(self, LCMSFeatureTreeNode other):
 */

#line 313 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = ((fabs((__pyx_v_self->__pyx_base.time - __pyx_v_other->__pyx_base.time)) < 1e-4) != 0);

#line 313 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = __pyx_t_6;

#line 313 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_bool_binop_done:;

#line 313 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_5;

#line 313 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":311
 *         return self._mz
 * 
 *     cpdef bint _eq(self, LCMSFeatureTreeNode other):             # <<<<<<<<<<<<<<
 *         return self.members == other.members and abs(
 *             self.time - other.time) < 1e-4
 */

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNode._eq", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_5_eq(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_4_eq[] = "LCMSFeatureTreeNode._eq(self, LCMSFeatureTreeNode other) -> bool";

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_5_eq(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_eq (wrapper)", 0);

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode, 1, "other", 0))) __PYX_ERR(0, 311, __pyx_L1_error)
  __pyx_r = 
#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_4_eq(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_v_other));

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_4_eq(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_other) {

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_eq", 0);

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode__eq(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 311, __pyx_L1_error)

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNode._eq", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 311 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":315
 *             self.time - other.time) < 1e-4
 * 
 *     cpdef bint _ne(self, LCMSFeatureTreeNode other):             # <<<<<<<<<<<<<<
 *         return not (self._eq(other))
 * 
 */

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_7_ne(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode__ne(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_other, int __pyx_skip_dispatch) {

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_r;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_5;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_ne", 0);

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_ne); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 315, __pyx_L1_error)

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_7_ne)) {

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_4 = NULL;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_4)) {

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_other)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other));

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 315, __pyx_L1_error)

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 315, __pyx_L1_error)

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = __pyx_t_5;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":316
 * 
 *     cpdef bint _ne(self, LCMSFeatureTreeNode other):
 *         return not (self._eq(other))             # <<<<<<<<<<<<<<
 * 
 *     def __richcmp__(self, other, int code):
 */

#line 316 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = (!(((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_v_self->__pyx_base.__pyx_vtab)->_eq(__pyx_v_self, __pyx_v_other, 0) != 0));

#line 316 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":315
 *             self.time - other.time) < 1e-4
 * 
 *     cpdef bint _ne(self, LCMSFeatureTreeNode other):             # <<<<<<<<<<<<<<
 *         return not (self._eq(other))
 * 
 */

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNode._ne", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_7_ne(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_6_ne[] = "LCMSFeatureTreeNode._ne(self, LCMSFeatureTreeNode other) -> bool";

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_7_ne(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_ne (wrapper)", 0);

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode, 1, "other", 0))) __PYX_ERR(0, 315, __pyx_L1_error)
  __pyx_r = 
#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_6_ne(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_v_other));

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_6_ne(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_other) {

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_ne", 0);

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode__ne(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 315, __pyx_L1_error)

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNode._ne", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 315 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":318
 *         return not (self._eq(other))
 * 
 *     def __richcmp__(self, other, int code):             # <<<<<<<<<<<<<<
 *         if code == 2:
 *             return self._eq(other)
 */

#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_9__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code); /*proto*/

#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_9__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code) {

#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_r = 
#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_8__richcmp__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_v_self), ((PyObject *)__pyx_v_other), ((int)__pyx_v_code));

#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_8__richcmp__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code) {

#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":319
 * 
 *     def __richcmp__(self, other, int code):
 *         if code == 2:             # <<<<<<<<<<<<<<
 *             return self._eq(other)
 *         elif code == 3:
 */

#line 319 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  switch (__pyx_v_code) {

#line 319 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    case 2:

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":320
 *     def __richcmp__(self, other, int code):
 *         if code == 2:
 *             return self._eq(other)             # <<<<<<<<<<<<<<
 *         elif code == 3:
 *             return self._ne(other)
 */

#line 320 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 320 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (!(likely(((__pyx_v_other) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_other, __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode))))) __PYX_ERR(0, 320, __pyx_L1_error)

#line 320 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __Pyx_PyBool_FromLong(((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_v_self->__pyx_base.__pyx_vtab)->_eq(__pyx_v_self, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_v_other), 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 320, __pyx_L1_error)

#line 320 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 320 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_r = __pyx_t_1;

#line 320 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = 0;

#line 320 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":319
 * 
 *     def __richcmp__(self, other, int code):
 *         if code == 2:             # <<<<<<<<<<<<<<
 *             return self._eq(other)
 *         elif code == 3:
 */

#line 319 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    break;

#line 319 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    case 3:

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":322
 *             return self._eq(other)
 *         elif code == 3:
 *             return self._ne(other)             # <<<<<<<<<<<<<<
 *         else:
 *             return NotImplemented
 */

#line 322 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 322 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (!(likely(((__pyx_v_other) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_other, __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode))))) __PYX_ERR(0, 322, __pyx_L1_error)

#line 322 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __Pyx_PyBool_FromLong(((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_v_self->__pyx_base.__pyx_vtab)->_ne(__pyx_v_self, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_v_other), 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 322, __pyx_L1_error)

#line 322 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 322 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_r = __pyx_t_1;

#line 322 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = 0;

#line 322 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":321
 *         if code == 2:
 *             return self._eq(other)
 *         elif code == 3:             # <<<<<<<<<<<<<<
 *             return self._ne(other)
 *         else:
 */

#line 321 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    break;

#line 321 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    default:

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":324
 *             return self._ne(other)
 *         else:
 *             return NotImplemented             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */

#line 324 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 324 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(__pyx_builtin_NotImplemented);

#line 324 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_r = __pyx_builtin_NotImplemented;

#line 324 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L0;

#line 324 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    break;

#line 324 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":318
 *         return not (self._eq(other))
 * 
 *     def __richcmp__(self, other, int code):             # <<<<<<<<<<<<<<
 *         if code == 2:
 *             return self._eq(other)
 */

#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNode.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 318 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":326
 *             return NotImplemented
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "%s(%f, %0.4f %s|%d)" % (
 *             self.__class__.__name__,
 */

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_11__repr__(PyObject *__pyx_v_self); /*proto*/

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_11__repr__(PyObject *__pyx_v_self) {

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = 
#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_10__repr__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_v_self));

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_10__repr__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_self) {

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_ssize_t __pyx_t_5;

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":327
 * 
 *     def __repr__(self):
 *         return "%s(%f, %0.4f %s|%d)" % (             # <<<<<<<<<<<<<<
 *             self.__class__.__name__,
 *             self.time, self.peaks[0].mz, "",
 */

#line 327 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":328
 *     def __repr__(self):
 *         return "%s(%f, %0.4f %s|%d)" % (
 *             self.__class__.__name__,             # <<<<<<<<<<<<<<
 *             self.time, self.peaks[0].mz, "",
 *             len(self.members))
 */

#line 328 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 328, __pyx_L1_error)

#line 328 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 328 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 328, __pyx_L1_error)

#line 328 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 328 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":329
 *         return "%s(%f, %0.4f %s|%d)" % (
 *             self.__class__.__name__,
 *             self.time, self.peaks[0].mz, "",             # <<<<<<<<<<<<<<
 *             len(self.members))
 * 
 */

#line 329 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 329, __pyx_L1_error)

#line 329 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 329 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_peaks); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 329, __pyx_L1_error)

#line 329 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 329 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 329, __pyx_L1_error)

#line 329 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 329 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 329 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_mz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 329, __pyx_L1_error)

#line 329 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 329 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":330
 *             self.__class__.__name__,
 *             self.time, self.peaks[0].mz, "",
 *             len(self.members))             # <<<<<<<<<<<<<<
 * 
 *     cdef inline FittedPeak getpeak(self, size_t i):
 */

#line 330 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = __pyx_v_self->__pyx_base.members;

#line 330 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_t_4);

#line 330 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_t_4 == Py_None)) {

#line 330 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 330 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __PYX_ERR(0, 330, __pyx_L1_error)

#line 330 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 330 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = PyList_GET_SIZE(__pyx_t_4); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 330, __pyx_L1_error)

#line 330 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 330 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 330, __pyx_L1_error)

#line 330 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_4);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":328
 *     def __repr__(self):
 *         return "%s(%f, %0.4f %s|%d)" % (
 *             self.__class__.__name__,             # <<<<<<<<<<<<<<
 *             self.time, self.peaks[0].mz, "",
 *             len(self.members))
 */

#line 328 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = PyTuple_New(5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 328, __pyx_L1_error)

#line 328 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_6);

#line 328 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 328 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);

#line 328 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 328 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);

#line 328 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 328 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_3);

#line 328 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_kp_s_);

#line 328 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_kp_s_);

#line 328 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_kp_s_);

#line 328 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_4);

#line 328 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_6, 4, __pyx_t_4);

#line 328 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = 0;

#line 328 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 328 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = 0;

#line 328 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":327
 * 
 *     def __repr__(self):
 *         return "%s(%f, %0.4f %s|%d)" % (             # <<<<<<<<<<<<<<
 *             self.__class__.__name__,
 *             self.time, self.peaks[0].mz, "",
 */

#line 327 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_s_f_0_4f_s_d, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 327, __pyx_L1_error)

#line 327 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 327 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 327 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_4;

#line 327 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = 0;

#line 327 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":326
 *             return NotImplemented
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "%s(%f, %0.4f %s|%d)" % (
 *             self.__class__.__name__,
 */

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNode.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 326 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":332
 *             len(self.members))
 * 
 *     cdef inline FittedPeak getpeak(self, size_t i):             # <<<<<<<<<<<<<<
 *         return <FittedPeak>PyList_GET_ITEM(self.members, i)
 * 
 */

#line 332 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 332 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static CYTHON_INLINE struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_getpeak(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_self, size_t __pyx_v_i) {

#line 332 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *__pyx_r = NULL;

#line 332 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 332 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 332 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2;

#line 332 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("getpeak", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":333
 * 
 *     cdef inline FittedPeak getpeak(self, size_t i):
 *         return <FittedPeak>PyList_GET_ITEM(self.members, i)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 333 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 333 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_v_self->__pyx_base.members;

#line 333 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 333 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_v_i);

#line 333 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 333 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(((PyObject *)((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_2)));

#line 333 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *)__pyx_t_2);

#line 333 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":332
 *             len(self.members))
 * 
 *     cdef inline FittedPeak getpeak(self, size_t i):             # <<<<<<<<<<<<<<
 *         return <FittedPeak>PyList_GET_ITEM(self.members, i)
 * 
 */

#line 332 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 332 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 332 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 332 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 332 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 332 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 332 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":47
 *     cdef:
 * 
 *         public double _mz             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint _eq(self, LCMSFeatureTreeNode other)
 */

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_3_mz_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_3_mz_1__get__(PyObject *__pyx_v_self) {

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = 0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_3_mz___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_v_self));

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_3_mz___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_self) {

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = NULL;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_mz); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 47, __pyx_L1_error)

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = __pyx_t_1;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = 0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNode._mz.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = NULL;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_3_mz_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_3_mz_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_3_mz_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_3_mz_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_self, PyObject *__pyx_v_value) {

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double __pyx_t_1;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 47, __pyx_L1_error)

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->_mz = __pyx_t_1;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNode._mz.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = -1;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 47 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":338
 * @cython.freelist(10000000)
 * cdef class FeatureBase(object):
 *     def __init__(self, nodes):             # <<<<<<<<<<<<<<
 *         self.nodes = LCMSFeatureTreeList(nodes)
 * 
 */

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_nodes = 0
#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_r;

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_nodes,0};

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject* values[1] = {0};

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(__pyx_kwds)) {

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      Py_ssize_t kw_args;

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {
        case  1: 
#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0: break;
        default: 
#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0:

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nodes)) != 0)) kw_args--;
        else 
#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(kw_args > 0)) {

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 338, __pyx_L3_error)

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L5_argtuple_error;

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_nodes = values[0];

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); 
#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__PYX_ERR(0, 338, __pyx_L3_error)

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_error:;

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureBase.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return -1;

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase___init__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_self), __pyx_v_nodes);

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase___init__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self, PyObject *__pyx_v_nodes) {

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_r;

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":339
 * cdef class FeatureBase(object):
 *     def __init__(self, nodes):
 *         self.nodes = LCMSFeatureTreeList(nodes)             # <<<<<<<<<<<<<<
 * 
 *     cdef double get_mz(self):
 */

#line 339 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList), __pyx_v_nodes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 339, __pyx_L1_error)

#line 339 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 339 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 339 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_self->nodes);

#line 339 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->nodes));

#line 339 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->nodes = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_t_1);

#line 339 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":338
 * @cython.freelist(10000000)
 * cdef class FeatureBase(object):
 *     def __init__(self, nodes):             # <<<<<<<<<<<<<<
 *         self.nodes = LCMSFeatureTreeList(nodes)
 * 
 */

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureBase.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = -1;

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 338 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":341
 *         self.nodes = LCMSFeatureTreeList(nodes)
 * 
 *     cdef double get_mz(self):             # <<<<<<<<<<<<<<
 *         return 0.0
 * 
 */

#line 341 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 341 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_get_mz(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self) {

#line 341 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_r;

#line 341 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 341 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("get_mz", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":342
 * 
 *     cdef double get_mz(self):
 *         return 0.0             # <<<<<<<<<<<<<<
 * 
 *     cdef double get_neutral_mass(self):
 */

#line 342 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0.0;

#line 342 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":341
 *         self.nodes = LCMSFeatureTreeList(nodes)
 * 
 *     cdef double get_mz(self):             # <<<<<<<<<<<<<<
 *         return 0.0
 * 
 */

#line 341 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 341 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 341 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 341 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 341 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 341 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":344
 *         return 0.0
 * 
 *     cdef double get_neutral_mass(self):             # <<<<<<<<<<<<<<
 *         return self.get_mz()
 * 
 */

#line 344 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 344 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_get_neutral_mass(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self) {

#line 344 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_r;

#line 344 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 344 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("get_neutral_mass", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":345
 * 
 *     cdef double get_neutral_mass(self):
 *         return self.get_mz()             # <<<<<<<<<<<<<<
 * 
 *     cdef double get_start_time(self):
 */

#line 345 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_self->__pyx_vtab)->get_mz(__pyx_v_self);

#line 345 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":344
 *         return 0.0
 * 
 *     cdef double get_neutral_mass(self):             # <<<<<<<<<<<<<<
 *         return self.get_mz()
 * 
 */

#line 344 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 344 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 344 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 344 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 344 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 344 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":347
 *         return self.get_mz()
 * 
 *     cdef double get_start_time(self):             # <<<<<<<<<<<<<<
 *         return 0.0
 * 
 */

#line 347 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 347 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_get_start_time(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self) {

#line 347 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_r;

#line 347 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 347 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("get_start_time", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":348
 * 
 *     cdef double get_start_time(self):
 *         return 0.0             # <<<<<<<<<<<<<<
 * 
 *     cdef double get_end_time(self):
 */

#line 348 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0.0;

#line 348 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":347
 *         return self.get_mz()
 * 
 *     cdef double get_start_time(self):             # <<<<<<<<<<<<<<
 *         return 0.0
 * 
 */

#line 347 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 347 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 347 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 347 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 347 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 347 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":350
 *         return 0.0
 * 
 *     cdef double get_end_time(self):             # <<<<<<<<<<<<<<
 *         return INF
 * 
 */

#line 350 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 350 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_get_end_time(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self) {

#line 350 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_r;

#line 350 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 350 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("get_end_time", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":351
 * 
 *     cdef double get_end_time(self):
 *         return INF             # <<<<<<<<<<<<<<
 * 
 *     cpdef tuple find_time(self, double time):
 */

#line 351 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_v_12ms_deisotope_2_c_11feature_map_12lcms_feature_INF;

#line 351 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":350
 *         return 0.0
 * 
 *     cdef double get_end_time(self):             # <<<<<<<<<<<<<<
 *         return INF
 * 
 */

#line 350 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 350 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 350 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 350 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 350 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 350 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":353
 *         return INF
 * 
 *     cpdef tuple find_time(self, double time):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t indexout
 */

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_3find_time(PyObject *__pyx_v_self, PyObject *__pyx_arg_time); /*proto*/

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_find_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self, double __pyx_v_time, int __pyx_skip_dispatch) {
  size_t __pyx_v_indexout
#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_node = 0
#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("find_time", 0);

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_find_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 353, __pyx_L1_error)

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_3find_time)) {

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 353, __pyx_L1_error)

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_5 = NULL;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_5)) {

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_5);

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_4, function);

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 353, __pyx_L1_error)

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (!(likely(PyTuple_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 353, __pyx_L1_error)

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = ((PyObject*)__pyx_t_2);

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = 0;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":357
 *             size_t indexout
 *             LCMSFeatureTreeNode node
 *         node = self._find_time(time, &indexout)             # <<<<<<<<<<<<<<
 *         return node, indexout
 * 
 */

#line 357 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase__find_time(__pyx_v_self, __pyx_v_time, (&__pyx_v_indexout))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 357, __pyx_L1_error)

#line 357 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 357 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_node = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_t_1);

#line 357 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":358
 *             LCMSFeatureTreeNode node
 *         node = self._find_time(time, &indexout)
 *         return node, indexout             # <<<<<<<<<<<<<<
 * 
 *     cdef inline LCMSFeatureTreeNode _find_time(self, double time, size_t* indexout):
 */

#line 358 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 358 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_indexout); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 358, __pyx_L1_error)

#line 358 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 358 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 358, __pyx_L1_error)

#line 358 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 358 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_node));

#line 358 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_node));

#line 358 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_node));

#line 358 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 358 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);

#line 358 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 358 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = ((PyObject*)__pyx_t_2);

#line 358 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = 0;

#line 358 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":353
 *         return INF
 * 
 *     cpdef tuple find_time(self, double time):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t indexout
 */

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureBase.find_time", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_node);

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_3find_time(PyObject *__pyx_v_self, PyObject *__pyx_arg_time); /*proto*/

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_2find_time[] = "FeatureBase.find_time(self, double time) -> tuple";

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_3find_time(PyObject *__pyx_v_self, PyObject *__pyx_arg_time) {
  double __pyx_v_time
#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("find_time (wrapper)", 0);

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  assert(__pyx_arg_time); {

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_time = __pyx_PyFloat_AsDouble(__pyx_arg_time); if (unlikely((__pyx_v_time == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 353, __pyx_L3_error)

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_error:;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureBase.find_time", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return NULL;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_2find_time(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_self), ((double)__pyx_v_time));

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_2find_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self, double __pyx_v_time) {

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("find_time", 0);

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_find_time(__pyx_v_self, __pyx_v_time, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 353, __pyx_L1_error)

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureBase.find_time", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 353 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":360
 *         return node, indexout
 * 
 *     cdef inline LCMSFeatureTreeNode _find_time(self, double time, size_t* indexout):             # <<<<<<<<<<<<<<
 *         return self.nodes._find_time(time, indexout)
 * 
 */

#line 360 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 360 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static CYTHON_INLINE struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase__find_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self, double __pyx_v_time, size_t *__pyx_v_indexout) {

#line 360 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_r = NULL;

#line 360 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 360 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 360 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 360 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 360 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 360 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_find_time", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":361
 * 
 *     cdef inline LCMSFeatureTreeNode _find_time(self, double time, size_t* indexout):
 *         return self.nodes._find_time(time, indexout)             # <<<<<<<<<<<<<<
 * 
 *     cdef inline size_t get_size(self):
 */

#line 361 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 361 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList__find_time(__pyx_v_self->nodes, __pyx_v_time, __pyx_v_indexout)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 361, __pyx_L1_error)

#line 361 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 361 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode))))) __PYX_ERR(0, 361, __pyx_L1_error)

#line 361 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_t_1);

#line 361 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 361 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":360
 *         return node, indexout
 * 
 *     cdef inline LCMSFeatureTreeNode _find_time(self, double time, size_t* indexout):             # <<<<<<<<<<<<<<
 *         return self.nodes._find_time(time, indexout)
 * 
 */

#line 360 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 360 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 360 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 360 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 360 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureBase._find_time", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 360 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 360 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 360 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 360 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 360 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 360 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":363
 *         return self.nodes._find_time(time, indexout)
 * 
 *     cdef inline size_t get_size(self):             # <<<<<<<<<<<<<<
 *         return self.nodes.get_size()
 * 
 */

#line 363 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 363 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static CYTHON_INLINE size_t __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_get_size(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self) {

#line 363 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_r;

#line 363 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 363 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("get_size", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":364
 * 
 *     cdef inline size_t get_size(self):
 *         return self.nodes.get_size()             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */

#line 364 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_get_size(__pyx_v_self->nodes);

#line 364 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":363
 *         return self.nodes._find_time(time, indexout)
 * 
 *     cdef inline size_t get_size(self):             # <<<<<<<<<<<<<<
 *         return self.nodes.get_size()
 * 
 */

#line 363 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 363 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 363 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 363 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 363 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 363 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":366
 *         return self.nodes.get_size()
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.get_size()
 * 
 */

#line 366 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 366 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 366 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static Py_ssize_t __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_5__len__(PyObject *__pyx_v_self); /*proto*/

#line 366 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static Py_ssize_t __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_5__len__(PyObject *__pyx_v_self) {

#line 366 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_ssize_t __pyx_r;

#line 366 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 366 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = 
#line 366 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_4__len__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_self));

#line 366 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 366 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 366 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 366 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 366 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 366 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 366 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static Py_ssize_t __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_4__len__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self) {

#line 366 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_ssize_t __pyx_r;

#line 366 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 366 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":367
 * 
 *     def __len__(self):
 *         return self.get_size()             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */

#line 367 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_get_size(__pyx_v_self);

#line 367 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":366
 *         return self.nodes.get_size()
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.get_size()
 * 
 */

#line 366 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 366 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 366 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 366 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 366 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 366 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":369
 *         return self.get_size()
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.nodes)
 * 
 */

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_7__iter__(PyObject *__pyx_v_self); /*proto*/

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_7__iter__(PyObject *__pyx_v_self) {

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = 
#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_6__iter__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_self));

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_6__iter__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self) {

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":370
 * 
 *     def __iter__(self):
 *         return iter(self.nodes)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 370 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 370 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_v_self->nodes);

#line 370 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 370 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 370, __pyx_L1_error)

#line 370 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 370 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 370 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_2;

#line 370 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = 0;

#line 370 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":369
 *         return self.get_size()
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.nodes)
 * 
 */

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureBase.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 369 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":58
 * cdef class FeatureBase(object):
 *     cdef:
 *         public LCMSFeatureTreeList nodes             # <<<<<<<<<<<<<<
 *         public double _mz
 *         public double _start_time
 */

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_5nodes_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_5nodes_1__get__(PyObject *__pyx_v_self) {

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = 0;

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_5nodes___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_self));

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_5nodes___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self) {

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = NULL;

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->nodes));

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = ((PyObject *)__pyx_v_self->nodes);

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_5nodes_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_5nodes_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_5nodes_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_5nodes_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self, PyObject *__pyx_v_value) {

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList))))) __PYX_ERR(2, 58, __pyx_L1_error)

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = __pyx_v_value;

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(__pyx_t_1);

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(__pyx_t_1);

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->nodes);

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->nodes));

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->nodes = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_t_1);

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = 0;

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureBase.nodes.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = -1;

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_5nodes_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_5nodes_5__del__(PyObject *__pyx_v_self) {

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_5nodes_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_self));

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_5nodes_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self) {

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(Py_None);

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(Py_None);

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->nodes);

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->nodes));

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->nodes = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)Py_None);

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 58 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":59
 *     cdef:
 *         public LCMSFeatureTreeList nodes
 *         public double _mz             # <<<<<<<<<<<<<<
 *         public double _start_time
 *         public double _end_time
 */

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_3_mz_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_3_mz_1__get__(PyObject *__pyx_v_self) {

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = 0;

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_3_mz___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_self));

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_3_mz___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self) {

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = NULL;

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_mz); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 59, __pyx_L1_error)

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = __pyx_t_1;

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = 0;

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureBase._mz.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = NULL;

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_3_mz_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_3_mz_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_3_mz_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_3_mz_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self, PyObject *__pyx_v_value) {

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double __pyx_t_1;

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 59, __pyx_L1_error)

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->_mz = __pyx_t_1;

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureBase._mz.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = -1;

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 59 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":60
 *         public LCMSFeatureTreeList nodes
 *         public double _mz
 *         public double _start_time             # <<<<<<<<<<<<<<
 *         public double _end_time
 * 
 */

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_11_start_time_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_11_start_time_1__get__(PyObject *__pyx_v_self) {

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = 0;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_11_start_time___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_self));

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_11_start_time___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self) {

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = NULL;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_start_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 60, __pyx_L1_error)

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = __pyx_t_1;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = 0;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureBase._start_time.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = NULL;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_11_start_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_11_start_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_11_start_time_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_11_start_time_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self, PyObject *__pyx_v_value) {

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double __pyx_t_1;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 60, __pyx_L1_error)

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->_start_time = __pyx_t_1;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureBase._start_time.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = -1;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 60 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":61
 *         public double _mz
 *         public double _start_time
 *         public double _end_time             # <<<<<<<<<<<<<<
 * 
 *     cdef double get_mz(self)
 */

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_9_end_time_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_9_end_time_1__get__(PyObject *__pyx_v_self) {

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = 0;

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_9_end_time___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_self));

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_9_end_time___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self) {

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = NULL;

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_end_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 61, __pyx_L1_error)

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = __pyx_t_1;

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = 0;

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureBase._end_time.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = NULL;

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_9_end_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_9_end_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_9_end_time_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_9_end_time_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self, PyObject *__pyx_v_value) {

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double __pyx_t_1;

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 61, __pyx_L1_error)

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->_end_time = __pyx_t_1;

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureBase._end_time.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = -1;

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 61 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_8__reduce_cython__[] = "FeatureBase.__reduce_cython__(self)";

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_8__reduce_cython__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_8__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self) {
  PyObject *__pyx_v_state = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v__dict = 0
#line 1 "(tree fragment)"
;
  int __pyx_v_use_setstate
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_2 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_3 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_5;

#line 1 "(tree fragment)"
  int __pyx_t_6;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._end_time, self._mz, self._start_time, self.nodes)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */

#line 5 "(tree fragment)"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_end_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->_mz); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_2);

#line 5 "(tree fragment)"
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->_start_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_3);

#line 5 "(tree fragment)"
  __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_2);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_3);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);

#line 5 "(tree fragment)"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->nodes));

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->nodes));

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_4, 3, ((PyObject *)__pyx_v_self->nodes));

#line 5 "(tree fragment)"
  __pyx_t_1 = 0;

#line 5 "(tree fragment)"
  __pyx_t_2 = 0;

#line 5 "(tree fragment)"
  __pyx_t_3 = 0;

#line 5 "(tree fragment)"
  __pyx_v_state = ((PyObject*)__pyx_t_4);

#line 5 "(tree fragment)"
  __pyx_t_4 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._end_time, self._mz, self._start_time, self.nodes)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */

#line 6 "(tree fragment)"
  __pyx_t_4 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 6 "(tree fragment)"
  __pyx_v__dict = __pyx_t_4;

#line 6 "(tree fragment)"
  __pyx_t_4 = 0;

  /* "(tree fragment)":7
 *     state = (self._end_time, self._mz, self._start_time, self.nodes)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
  __pyx_t_5 = (__pyx_v__dict != Py_None);

#line 7 "(tree fragment)"
  __pyx_t_6 = (__pyx_t_5 != 0);

#line 7 "(tree fragment)"
  if (__pyx_t_6) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */

#line 8 "(tree fragment)"
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 8 "(tree fragment)"
    __Pyx_INCREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v__dict);

#line 8 "(tree fragment)"
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 8 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 8 "(tree fragment)"
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));

#line 8 "(tree fragment)"
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.nodes is not None
 */

#line 9 "(tree fragment)"
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._end_time, self._mz, self._start_time, self.nodes)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
    goto __pyx_L3;

#line 7 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.nodes is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_FeatureBase, (type(self), 0x9e63e75, None), state
 */

#line 11 "(tree fragment)"
  /*else*/ {

#line 11 "(tree fragment)"
    __pyx_t_6 = (((PyObject *)__pyx_v_self->nodes) != Py_None);

#line 11 "(tree fragment)"
    __pyx_v_use_setstate = __pyx_t_6;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.nodes is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_FeatureBase, (type(self), 0x9e63e75, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  __pyx_t_6 = (__pyx_v_use_setstate != 0);

#line 12 "(tree fragment)"
  if (__pyx_t_6) {

    /* "(tree fragment)":13
 *         use_setstate = self.nodes is not None
 *     if use_setstate:
 *         return __pyx_unpickle_FeatureBase, (type(self), 0x9e63e75, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_FeatureBase, (type(self), 0x9e63e75, state)
 */

#line 13 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 13 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_FeatureBase); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 13 "(tree fragment)"
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 13 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_166084213);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_166084213);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_166084213);

#line 13 "(tree fragment)"
    __Pyx_INCREF(Py_None);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(Py_None);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 2, Py_None);

#line 13 "(tree fragment)"
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_3);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_4);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state);

#line 13 "(tree fragment)"
    __pyx_t_3 = 0;

#line 13 "(tree fragment)"
    __pyx_t_4 = 0;

#line 13 "(tree fragment)"
    __pyx_r = __pyx_t_2;

#line 13 "(tree fragment)"
    __pyx_t_2 = 0;

#line 13 "(tree fragment)"
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.nodes is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_FeatureBase, (type(self), 0x9e63e75, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_FeatureBase, (type(self), 0x9e63e75, None), state
 *     else:
 *         return __pyx_unpickle_FeatureBase, (type(self), 0x9e63e75, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_FeatureBase__set_state(self, __pyx_state)
 */

#line 15 "(tree fragment)"
  /*else*/ {

#line 15 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 15 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_FeatureBase); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 15 "(tree fragment)"
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 15 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_166084213);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_166084213);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_166084213);

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state);

#line 15 "(tree fragment)"
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_2);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_4);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);

#line 15 "(tree fragment)"
    __pyx_t_2 = 0;

#line 15 "(tree fragment)"
    __pyx_t_4 = 0;

#line 15 "(tree fragment)"
    __pyx_r = __pyx_t_3;

#line 15 "(tree fragment)"
    __pyx_t_3 = 0;

#line 15 "(tree fragment)"
    goto __pyx_L0;

#line 15 "(tree fragment)"
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_2);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_3);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureBase.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v_state);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v__dict);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FeatureBase, (type(self), 0x9e63e75, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FeatureBase__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
/* Python wrapper */

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 16 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_10__setstate_cython__[] = "FeatureBase.__setstate_cython__(self, __pyx_state)";

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 16 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_10__setstate_cython__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_10__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 16 "(tree fragment)"
  int __pyx_lineno = 0;

#line 16 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 16 "(tree fragment)"
  int __pyx_clineno = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_FeatureBase, (type(self), 0x9e63e75, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_FeatureBase__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */

#line 17 "(tree fragment)"
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_unpickle_FeatureBase__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 17 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FeatureBase, (type(self), 0x9e63e75, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FeatureBase__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 16 "(tree fragment)"
  goto __pyx_L0;

#line 16 "(tree fragment)"
  __pyx_L1_error:;

#line 16 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 16 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureBase.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 16 "(tree fragment)"
  __pyx_r = NULL;

#line 16 "(tree fragment)"
  __pyx_L0:;

#line 16 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":374
 * 
 * cdef class LCMSFeature(FeatureBase):
 *     def __init__(self, nodes=None, adducts=None, used_as_adduct=None, feature_id=None):             # <<<<<<<<<<<<<<
 *         if nodes is None:
 *             nodes = []
 */

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_nodes = 0
#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyObject *__pyx_v_adducts = 0
#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyObject *__pyx_v_used_as_adduct = 0
#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyObject *__pyx_v_feature_id = 0
#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_r;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_nodes,&__pyx_n_s_adducts,&__pyx_n_s_used_as_adduct,&__pyx_n_s_feature_id,0};

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject* values[4] = {0,0,0,0};

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    values[0] = ((PyObject *)Py_None);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    values[1] = ((PyObject *)Py_None);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    values[2] = ((PyObject *)Py_None);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    values[3] = ((PyObject *)Py_None);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(__pyx_kwds)) {

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      Py_ssize_t kw_args;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {
        case  4: 
#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0: break;
        default: 
#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0:

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (kw_args > 0) {

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nodes);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (value) { values[0] = value; kw_args--; }

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  1:

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (kw_args > 0) {

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_adducts);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (value) { values[1] = value; kw_args--; }

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  2:

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (kw_args > 0) {

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_used_as_adduct);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  3:

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (kw_args > 0) {

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_feature_id);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (value) { values[3] = value; kw_args--; }

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(kw_args > 0)) {

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 374, __pyx_L3_error)

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: 
#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[3] = PyTuple_GET_ITEM(__pyx_args, 3);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  3: 
#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  0: 
#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
break;
        default: 
#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_nodes = values[0];

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_adducts = values[1];

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_used_as_adduct = values[2];

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_feature_id = values[3];

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 4, PyTuple_GET_SIZE(__pyx_args)); 
#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__PYX_ERR(0, 374, __pyx_L3_error)

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_error:;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return -1;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature___init__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), __pyx_v_nodes, __pyx_v_adducts, __pyx_v_used_as_adduct, __pyx_v_feature_id);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature___init__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_nodes, PyObject *__pyx_v_adducts, PyObject *__pyx_v_used_as_adduct, PyObject *__pyx_v_feature_id) {

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_r;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_1;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_2;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_6;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_7 = NULL;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__init__", 0);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_nodes);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_adducts);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_used_as_adduct);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_feature_id);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":375
 * cdef class LCMSFeature(FeatureBase):
 *     def __init__(self, nodes=None, adducts=None, used_as_adduct=None, feature_id=None):
 *         if nodes is None:             # <<<<<<<<<<<<<<
 *             nodes = []
 *         if adducts is None:
 */

#line 375 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = (__pyx_v_nodes == Py_None);

#line 375 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 375 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":376
 *     def __init__(self, nodes=None, adducts=None, used_as_adduct=None, feature_id=None):
 *         if nodes is None:
 *             nodes = []             # <<<<<<<<<<<<<<
 *         if adducts is None:
 *             adducts = []
 */

#line 376 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 376, __pyx_L1_error)

#line 376 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 376 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF_SET(__pyx_v_nodes, __pyx_t_3);

#line 376 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":375
 * cdef class LCMSFeature(FeatureBase):
 *     def __init__(self, nodes=None, adducts=None, used_as_adduct=None, feature_id=None):
 *         if nodes is None:             # <<<<<<<<<<<<<<
 *             nodes = []
 *         if adducts is None:
 */

#line 375 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":377
 *         if nodes is None:
 *             nodes = []
 *         if adducts is None:             # <<<<<<<<<<<<<<
 *             adducts = []
 *         if used_as_adduct is None:
 */

#line 377 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = (__pyx_v_adducts == Py_None);

#line 377 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = (__pyx_t_2 != 0);

#line 377 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":378
 *             nodes = []
 *         if adducts is None:
 *             adducts = []             # <<<<<<<<<<<<<<
 *         if used_as_adduct is None:
 *             used_as_adduct = []
 */

#line 378 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 378, __pyx_L1_error)

#line 378 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 378 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF_SET(__pyx_v_adducts, __pyx_t_3);

#line 378 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":377
 *         if nodes is None:
 *             nodes = []
 *         if adducts is None:             # <<<<<<<<<<<<<<
 *             adducts = []
 *         if used_as_adduct is None:
 */

#line 377 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":379
 *         if adducts is None:
 *             adducts = []
 *         if used_as_adduct is None:             # <<<<<<<<<<<<<<
 *             used_as_adduct = []
 *         if feature_id is None:
 */

#line 379 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = (__pyx_v_used_as_adduct == Py_None);

#line 379 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 379 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":380
 *             adducts = []
 *         if used_as_adduct is None:
 *             used_as_adduct = []             # <<<<<<<<<<<<<<
 *         if feature_id is None:
 *             feature_id = uid()
 */

#line 380 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 380, __pyx_L1_error)

#line 380 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 380 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF_SET(__pyx_v_used_as_adduct, __pyx_t_3);

#line 380 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":379
 *         if adducts is None:
 *             adducts = []
 *         if used_as_adduct is None:             # <<<<<<<<<<<<<<
 *             used_as_adduct = []
 *         if feature_id is None:
 */

#line 379 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":381
 *         if used_as_adduct is None:
 *             used_as_adduct = []
 *         if feature_id is None:             # <<<<<<<<<<<<<<
 *             feature_id = uid()
 * 
 */

#line 381 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = (__pyx_v_feature_id == Py_None);

#line 381 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = (__pyx_t_2 != 0);

#line 381 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":382
 *             used_as_adduct = []
 *         if feature_id is None:
 *             feature_id = uid()             # <<<<<<<<<<<<<<
 * 
 *         FeatureBase.__init__(self, nodes)
 */

#line 382 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(__pyx_v_12ms_deisotope_2_c_11feature_map_12lcms_feature_uid);
    __pyx_t_4 = __pyx_v_12ms_deisotope_2_c_11feature_map_12lcms_feature_uid; 
#line 382 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_5 = NULL;

#line 382 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {

#line 382 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 382 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (likely(__pyx_t_5)) {

#line 382 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 382 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_5);

#line 382 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(function);

#line 382 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF_SET(__pyx_t_4, function);

#line 382 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 382 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 382 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);

#line 382 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 382 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 382, __pyx_L1_error)

#line 382 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 382 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 382 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF_SET(__pyx_v_feature_id, __pyx_t_3);

#line 382 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":381
 *         if used_as_adduct is None:
 *             used_as_adduct = []
 *         if feature_id is None:             # <<<<<<<<<<<<<<
 *             feature_id = uid()
 * 
 */

#line 381 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":384
 *             feature_id = uid()
 * 
 *         FeatureBase.__init__(self, nodes)             # <<<<<<<<<<<<<<
 *         self._total_intensity = -1
 *         self._mz = -1
 */

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase), __pyx_n_s_init); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 384, __pyx_L1_error)

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = NULL;

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = 0;

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (likely(__pyx_t_5)) {

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(__pyx_t_5);

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(function);

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF_SET(__pyx_t_4, function);

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_6 = 1;

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #if CYTHON_FAST_PYCALL

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (PyFunction_Check(__pyx_t_4)) {

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_self), __pyx_v_nodes};

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 384, __pyx_L1_error)

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #endif

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #if CYTHON_FAST_PYCCALL

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_self), __pyx_v_nodes};

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 384, __pyx_L1_error)

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #endif

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 384, __pyx_L1_error)

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_7);

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_5) {

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self));

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_self));

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(__pyx_v_nodes);

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GIVEREF(__pyx_v_nodes);

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_nodes);

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 384, __pyx_L1_error)

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 384 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":385
 * 
 *         FeatureBase.__init__(self, nodes)
 *         self._total_intensity = -1             # <<<<<<<<<<<<<<
 *         self._mz = -1
 *         self._last_mz = 0.0
 */

#line 385 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->_total_intensity = -1.0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":386
 *         FeatureBase.__init__(self, nodes)
 *         self._total_intensity = -1
 *         self._mz = -1             # <<<<<<<<<<<<<<
 *         self._last_mz = 0.0
 *         self._times = None
 */

#line 386 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->__pyx_base._mz = -1.0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":387
 *         self._total_intensity = -1
 *         self._mz = -1
 *         self._last_mz = 0.0             # <<<<<<<<<<<<<<
 *         self._times = None
 *         self._peaks = None
 */

#line 387 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->_last_mz = 0.0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":388
 *         self._mz = -1
 *         self._last_mz = 0.0
 *         self._times = None             # <<<<<<<<<<<<<<
 *         self._peaks = None
 *         self._start_time = -1
 */

#line 388 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(Py_None);

#line 388 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(Py_None);

#line 388 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_self->_times);

#line 388 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_v_self->_times);

#line 388 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->_times = Py_None;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":389
 *         self._last_mz = 0.0
 *         self._times = None
 *         self._peaks = None             # <<<<<<<<<<<<<<
 *         self._start_time = -1
 *         self._end_time = -1
 */

#line 389 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(Py_None);

#line 389 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(Py_None);

#line 389 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_self->_peaks);

#line 389 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_v_self->_peaks);

#line 389 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->_peaks = Py_None;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":390
 *         self._times = None
 *         self._peaks = None
 *         self._start_time = -1             # <<<<<<<<<<<<<<
 *         self._end_time = -1
 *         self.adducts = adducts
 */

#line 390 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->__pyx_base._start_time = -1.0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":391
 *         self._peaks = None
 *         self._start_time = -1
 *         self._end_time = -1             # <<<<<<<<<<<<<<
 *         self.adducts = adducts
 *         self.used_as_adduct = used_as_adduct
 */

#line 391 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->__pyx_base._end_time = -1.0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":392
 *         self._start_time = -1
 *         self._end_time = -1
 *         self.adducts = adducts             # <<<<<<<<<<<<<<
 *         self.used_as_adduct = used_as_adduct
 *         self.feature_id = feature_id
 */

#line 392 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_adducts);

#line 392 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_v_adducts);

#line 392 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_self->adducts);

#line 392 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_v_self->adducts);

#line 392 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->adducts = __pyx_v_adducts;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":393
 *         self._end_time = -1
 *         self.adducts = adducts
 *         self.used_as_adduct = used_as_adduct             # <<<<<<<<<<<<<<
 *         self.feature_id = feature_id
 *         self._initialize_averager()
 */

#line 393 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_used_as_adduct);

#line 393 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_v_used_as_adduct);

#line 393 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_self->used_as_adduct);

#line 393 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_v_self->used_as_adduct);

#line 393 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->used_as_adduct = __pyx_v_used_as_adduct;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":394
 *         self.adducts = adducts
 *         self.used_as_adduct = used_as_adduct
 *         self.feature_id = feature_id             # <<<<<<<<<<<<<<
 *         self._initialize_averager()
 * 
 */

#line 394 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_feature_id);

#line 394 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_v_feature_id);

#line 394 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_self->feature_id);

#line 394 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_v_self->feature_id);

#line 394 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->feature_id = __pyx_v_feature_id;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":395
 *         self.used_as_adduct = used_as_adduct
 *         self.feature_id = feature_id
 *         self._initialize_averager()             # <<<<<<<<<<<<<<
 * 
 *     cpdef _initialize_averager(self):
 */

#line 395 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->_initialize_averager(__pyx_v_self, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 395, __pyx_L1_error)

#line 395 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 395 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":374
 * 
 * cdef class LCMSFeature(FeatureBase):
 *     def __init__(self, nodes=None, adducts=None, used_as_adduct=None, feature_id=None):             # <<<<<<<<<<<<<<
 *         if nodes is None:
 *             nodes = []
 */

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_7);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = -1;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_nodes);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_adducts);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_used_as_adduct);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_feature_id);

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 374 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":397
 *         self._initialize_averager()
 * 
 *     cpdef _initialize_averager(self):             # <<<<<<<<<<<<<<
 *         self._peak_averager = RunningWeightedAverage._create(None)
 * 
 */

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_3_initialize_averager(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__initialize_averager(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, int __pyx_skip_dispatch) {

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_initialize_averager", 0);

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_initialize_averager); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 397, __pyx_L1_error)

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_3_initialize_averager)) {

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_4 = NULL;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_4)) {

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 397, __pyx_L1_error)

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = __pyx_t_2;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = 0;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":398
 * 
 *     cpdef _initialize_averager(self):
 *         self._peak_averager = RunningWeightedAverage._create(None)             # <<<<<<<<<<<<<<
 * 
 *     cpdef _update_from_averager(self):
 */

#line 398 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage__create(((PyObject*)Py_None))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 398, __pyx_L1_error)

#line 398 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 398 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 398 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_self->_peak_averager);

#line 398 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_peak_averager));

#line 398 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->_peak_averager = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_t_1);

#line 398 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":397
 *         self._initialize_averager()
 * 
 *     cpdef _initialize_averager(self):             # <<<<<<<<<<<<<<
 *         self._peak_averager = RunningWeightedAverage._create(None)
 * 
 */

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature._initialize_averager", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_3_initialize_averager(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_2_initialize_averager[] = "LCMSFeature._initialize_averager(self)";

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_3_initialize_averager(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_initialize_averager (wrapper)", 0);
  __pyx_r = 
#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_2_initialize_averager(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_2_initialize_averager(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_initialize_averager", 0);

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__initialize_averager(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 397, __pyx_L1_error)

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature._initialize_averager", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 397 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":400
 *         self._peak_averager = RunningWeightedAverage._create(None)
 * 
 *     cpdef _update_from_averager(self):             # <<<<<<<<<<<<<<
 *         best_mz = self._peak_averager.current_mean
 *         if best_mz < 0:
 */

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_5_update_from_averager(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__update_from_averager(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, int __pyx_skip_dispatch) {
  double __pyx_v_best_mz
#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_5;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_6;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_update_from_averager", 0);

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_update_from_averager); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 400, __pyx_L1_error)

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_5_update_from_averager)) {

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_4 = NULL;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_4)) {

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 400, __pyx_L1_error)

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = __pyx_t_2;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = 0;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":401
 * 
 *     cpdef _update_from_averager(self):
 *         best_mz = self._peak_averager.current_mean             # <<<<<<<<<<<<<<
 *         if best_mz < 0:
 *             raise ValueError("best_mz < 0")
 */

#line 401 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = __pyx_v_self->_peak_averager->current_mean;

#line 401 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_best_mz = __pyx_t_5;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":402
 *     cpdef _update_from_averager(self):
 *         best_mz = self._peak_averager.current_mean
 *         if best_mz < 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("best_mz < 0")
 *         self._last_mz = self._mz = best_mz
 */

#line 402 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = ((__pyx_v_best_mz < 0.0) != 0);

#line 402 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_t_6)) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":403
 *         best_mz = self._peak_averager.current_mean
 *         if best_mz < 0:
 *             raise ValueError("best_mz < 0")             # <<<<<<<<<<<<<<
 *         self._last_mz = self._mz = best_mz
 * 
 */

#line 403 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 403, __pyx_L1_error)

#line 403 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 403 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 403 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 403 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __PYX_ERR(0, 403, __pyx_L1_error)

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":402
 *     cpdef _update_from_averager(self):
 *         best_mz = self._peak_averager.current_mean
 *         if best_mz < 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("best_mz < 0")
 *         self._last_mz = self._mz = best_mz
 */

#line 402 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":404
 *         if best_mz < 0:
 *             raise ValueError("best_mz < 0")
 *         self._last_mz = self._mz = best_mz             # <<<<<<<<<<<<<<
 * 
 *     cpdef _reaverage_and_update(self):
 */

#line 404 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->_last_mz = __pyx_v_best_mz;

#line 404 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->__pyx_base._mz = __pyx_v_best_mz;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":400
 *         self._peak_averager = RunningWeightedAverage._create(None)
 * 
 *     cpdef _update_from_averager(self):             # <<<<<<<<<<<<<<
 *         best_mz = self._peak_averager.current_mean
 *         if best_mz < 0:
 */

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature._update_from_averager", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_5_update_from_averager(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_4_update_from_averager[] = "LCMSFeature._update_from_averager(self)";

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_5_update_from_averager(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_update_from_averager (wrapper)", 0);
  __pyx_r = 
#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_4_update_from_averager(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_4_update_from_averager(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_update_from_averager", 0);

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__update_from_averager(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 400, __pyx_L1_error)

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature._update_from_averager", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 400 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":406
 *         self._last_mz = self._mz = best_mz
 * 
 *     cpdef _reaverage_and_update(self):             # <<<<<<<<<<<<<<
 *         self._peak_averager.reset()
 *         self._peak_averager.feed_from_feature(self)
 */

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_7_reaverage_and_update(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__reaverage_and_update(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, int __pyx_skip_dispatch) {

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_reaverage_and_update", 0);

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_reaverage_and_update); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 406, __pyx_L1_error)

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_7_reaverage_and_update)) {

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_4 = NULL;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_4)) {

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 406, __pyx_L1_error)

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = __pyx_t_2;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = 0;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":407
 * 
 *     cpdef _reaverage_and_update(self):
 *         self._peak_averager.reset()             # <<<<<<<<<<<<<<
 *         self._peak_averager.feed_from_feature(self)
 *         self._update_from_averager()
 */

#line 407 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_self->_peak_averager->__pyx_vtab)->reset(__pyx_v_self->_peak_averager, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 407, __pyx_L1_error)

#line 407 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 407 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":408
 *     cpdef _reaverage_and_update(self):
 *         self._peak_averager.reset()
 *         self._peak_averager.feed_from_feature(self)             # <<<<<<<<<<<<<<
 *         self._update_from_averager()
 * 
 */

#line 408 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  (void)(((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_self->_peak_averager->__pyx_vtab)->feed_from_feature(__pyx_v_self->_peak_averager, __pyx_v_self, 0));

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":409
 *         self._peak_averager.reset()
 *         self._peak_averager.feed_from_feature(self)
 *         self._update_from_averager()             # <<<<<<<<<<<<<<
 * 
 *     def invalidate(self, reaverage=False):
 */

#line 409 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->_update_from_averager(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 409, __pyx_L1_error)

#line 409 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 409 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":406
 *         self._last_mz = self._mz = best_mz
 * 
 *     cpdef _reaverage_and_update(self):             # <<<<<<<<<<<<<<
 *         self._peak_averager.reset()
 *         self._peak_averager.feed_from_feature(self)
 */

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature._reaverage_and_update", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_7_reaverage_and_update(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_reaverage_and_update[] = "LCMSFeature._reaverage_and_update(self)";

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_7_reaverage_and_update(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_reaverage_and_update (wrapper)", 0);
  __pyx_r = 
#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_reaverage_and_update(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_reaverage_and_update(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_reaverage_and_update", 0);

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__reaverage_and_update(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 406, __pyx_L1_error)

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature._reaverage_and_update", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 406 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":411
 *         self._update_from_averager()
 * 
 *     def invalidate(self, reaverage=False):             # <<<<<<<<<<<<<<
 *         self._invalidate(reaverage)
 * 
 */

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_9invalidate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_8invalidate[] = "LCMSFeature.invalidate(self, reaverage=False)";

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_9invalidate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_reaverage = 0
#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("invalidate (wrapper)", 0);

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_reaverage,0};

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject* values[1] = {0};

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    values[0] = ((PyObject *)Py_False);

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(__pyx_kwds)) {

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      Py_ssize_t kw_args;

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {
        case  1: 
#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0: break;
        default: 
#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0:

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (kw_args > 0) {

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_reaverage);

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (value) { values[0] = value; kw_args--; }

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(kw_args > 0)) {

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "invalidate") < 0)) __PYX_ERR(0, 411, __pyx_L3_error)

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: 
#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  0: 
#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
break;
        default: 
#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_reaverage = values[0];

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("invalidate", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); 
#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__PYX_ERR(0, 411, __pyx_L3_error)

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_error:;

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.invalidate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return NULL;

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_8invalidate(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), __pyx_v_reaverage);

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_8invalidate(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_reaverage) {

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_1;

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__invalidate __pyx_t_3;

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("invalidate", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":412
 * 
 *     def invalidate(self, reaverage=False):
 *         self._invalidate(reaverage)             # <<<<<<<<<<<<<<
 * 
 *     cpdef _invalidate(self, bint reaverage=False):
 */

#line 412 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_reaverage); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 412, __pyx_L1_error)

#line 412 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3.__pyx_n = 1;

#line 412 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3.reaverage = __pyx_t_1;

#line 412 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->_invalidate(__pyx_v_self, 0, &__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 412, __pyx_L1_error)

#line 412 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 412 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":411
 *         self._update_from_averager()
 * 
 *     def invalidate(self, reaverage=False):             # <<<<<<<<<<<<<<
 *         self._invalidate(reaverage)
 * 
 */

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.invalidate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 411 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":414
 *         self._invalidate(reaverage)
 * 
 *     cpdef _invalidate(self, bint reaverage=False):             # <<<<<<<<<<<<<<
 *         self._total_intensity = -1
 *         self._last_mz = self._mz if self._mz != -1 else 0.
 */

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_11_invalidate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__invalidate(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__invalidate *__pyx_optional_args) {

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_v_reaverage = ((int)0);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_6;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_7;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_invalidate", 0);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_optional_args) {

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_reaverage = __pyx_optional_args->reaverage;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_invalidate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 414, __pyx_L1_error)

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_11_invalidate)) {

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_reaverage); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 414, __pyx_L1_error)

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_5 = NULL;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_5)) {

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_5);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_4, function);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 414, __pyx_L1_error)

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = __pyx_t_2;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = 0;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":415
 * 
 *     cpdef _invalidate(self, bint reaverage=False):
 *         self._total_intensity = -1             # <<<<<<<<<<<<<<
 *         self._last_mz = self._mz if self._mz != -1 else 0.
 *         self._mz = -1
 */

#line 415 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->_total_intensity = -1.0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":416
 *     cpdef _invalidate(self, bint reaverage=False):
 *         self._total_intensity = -1
 *         self._last_mz = self._mz if self._mz != -1 else 0.             # <<<<<<<<<<<<<<
 *         self._mz = -1
 *         self._times = None
 */

#line 416 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (((__pyx_v_self->__pyx_base._mz != -1.0) != 0)) {

#line 416 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_6 = __pyx_v_self->__pyx_base._mz;

#line 416 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {

#line 416 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_6 = 0.;

#line 416 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 416 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->_last_mz = __pyx_t_6;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":417
 *         self._total_intensity = -1
 *         self._last_mz = self._mz if self._mz != -1 else 0.
 *         self._mz = -1             # <<<<<<<<<<<<<<
 *         self._times = None
 *         self._peaks = None
 */

#line 417 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->__pyx_base._mz = -1.0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":418
 *         self._last_mz = self._mz if self._mz != -1 else 0.
 *         self._mz = -1
 *         self._times = None             # <<<<<<<<<<<<<<
 *         self._peaks = None
 *         self._start_time = -1
 */

#line 418 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(Py_None);

#line 418 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(Py_None);

#line 418 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_self->_times);

#line 418 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_v_self->_times);

#line 418 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->_times = Py_None;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":419
 *         self._mz = -1
 *         self._times = None
 *         self._peaks = None             # <<<<<<<<<<<<<<
 *         self._start_time = -1
 *         self._end_time = -1
 */

#line 419 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(Py_None);

#line 419 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(Py_None);

#line 419 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_self->_peaks);

#line 419 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_v_self->_peaks);

#line 419 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->_peaks = Py_None;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":420
 *         self._times = None
 *         self._peaks = None
 *         self._start_time = -1             # <<<<<<<<<<<<<<
 *         self._end_time = -1
 *         if reaverage:
 */

#line 420 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->__pyx_base._start_time = -1.0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":421
 *         self._peaks = None
 *         self._start_time = -1
 *         self._end_time = -1             # <<<<<<<<<<<<<<
 *         if reaverage:
 *             self._reaverage_and_update()
 */

#line 421 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->__pyx_base._end_time = -1.0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":422
 *         self._start_time = -1
 *         self._end_time = -1
 *         if reaverage:             # <<<<<<<<<<<<<<
 *             self._reaverage_and_update()
 * 
 */

#line 422 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_7 = (__pyx_v_reaverage != 0);

#line 422 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_7) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":423
 *         self._end_time = -1
 *         if reaverage:
 *             self._reaverage_and_update()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 423 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->_reaverage_and_update(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 423, __pyx_L1_error)

#line 423 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 423 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":422
 *         self._start_time = -1
 *         self._end_time = -1
 *         if reaverage:             # <<<<<<<<<<<<<<
 *             self._reaverage_and_update()
 * 
 */

#line 422 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":414
 *         self._invalidate(reaverage)
 * 
 *     cpdef _invalidate(self, bint reaverage=False):             # <<<<<<<<<<<<<<
 *         self._total_intensity = -1
 *         self._last_mz = self._mz if self._mz != -1 else 0.
 */

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature._invalidate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_11_invalidate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10_invalidate[] = "LCMSFeature._invalidate(self, bool reaverage=False)";

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_11_invalidate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_reaverage
#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_invalidate (wrapper)", 0);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_reaverage,0};

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject* values[1] = {0};

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(__pyx_kwds)) {

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      Py_ssize_t kw_args;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {
        case  1: 
#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0: break;
        default: 
#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0:

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (kw_args > 0) {

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_reaverage);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (value) { values[0] = value; kw_args--; }

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(kw_args > 0)) {

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_invalidate") < 0)) __PYX_ERR(0, 414, __pyx_L3_error)

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: 
#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  0: 
#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
break;
        default: 
#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (values[0]) {

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_reaverage = __Pyx_PyObject_IsTrue(values[0]); if (unlikely((__pyx_v_reaverage == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 414, __pyx_L3_error)

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_reaverage = ((int)0);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_invalidate", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); 
#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__PYX_ERR(0, 414, __pyx_L3_error)

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_error:;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature._invalidate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return NULL;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10_invalidate(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), __pyx_v_reaverage);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10_invalidate(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, int __pyx_v_reaverage) {

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__invalidate __pyx_t_2;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_invalidate", 0);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2.reaverage = __pyx_v_reaverage;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature->_invalidate(__pyx_v_self, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 414, __pyx_L1_error)

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature._invalidate", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 414 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":426
 * 
 *     @property
 *     def total_signal(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double total
 */

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_12total_signal_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_12total_signal_1__get__(PyObject *__pyx_v_self) {

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_12total_signal___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_12total_signal___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {
  double __pyx_v_total
#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_i
#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_n
#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_node = 0
#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_1;

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_2;

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_3;

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_4;

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":431
 *             size_t i, n
 *             LCMSFeatureTreeNode node
 *         if self._total_intensity == -1:             # <<<<<<<<<<<<<<
 *             total = 0.
 *             n = self.get_size()
 */

#line 431 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((__pyx_v_self->_total_intensity == -1.0) != 0);

#line 431 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":432
 *             LCMSFeatureTreeNode node
 *         if self._total_intensity == -1:
 *             total = 0.             # <<<<<<<<<<<<<<
 *             n = self.get_size()
 *             for i in range(n):
 */

#line 432 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_total = 0.;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":433
 *         if self._total_intensity == -1:
 *             total = 0.
 *             n = self.get_size()             # <<<<<<<<<<<<<<
 *             for i in range(n):
 *                 node = self.getitem(i)
 */

#line 433 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_n = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.get_size(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_self));

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":434
 *             total = 0.
 *             n = self.get_size()
 *             for i in range(n):             # <<<<<<<<<<<<<<
 *                 node = self.getitem(i)
 *                 total += node.total_intensity()
 */

#line 434 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = __pyx_v_n;

#line 434 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = __pyx_t_2;

#line 434 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 434 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_i = __pyx_t_4;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":435
 *             n = self.get_size()
 *             for i in range(n):
 *                 node = self.getitem(i)             # <<<<<<<<<<<<<<
 *                 total += node.total_intensity()
 *             self._total_intensity = total
 */

#line 435 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->getitem(__pyx_v_self, __pyx_v_i)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 435, __pyx_L1_error)

#line 435 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_5);

#line 435 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_t_5));

#line 435 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_5 = 0;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":436
 *             for i in range(n):
 *                 node = self.getitem(i)
 *                 total += node.total_intensity()             # <<<<<<<<<<<<<<
 *             self._total_intensity = total
 *         return self._total_intensity
 */

#line 436 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_total = (__pyx_v_total + ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_v_node->__pyx_base.__pyx_vtab)->__pyx_base.total_intensity(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_node), 0));

#line 436 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":437
 *                 node = self.getitem(i)
 *                 total += node.total_intensity()
 *             self._total_intensity = total             # <<<<<<<<<<<<<<
 *         return self._total_intensity
 * 
 */

#line 437 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_self->_total_intensity = __pyx_v_total;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":431
 *             size_t i, n
 *             LCMSFeatureTreeNode node
 *         if self._total_intensity == -1:             # <<<<<<<<<<<<<<
 *             total = 0.
 *             n = self.get_size()
 */

#line 431 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":438
 *                 total += node.total_intensity()
 *             self._total_intensity = total
 *         return self._total_intensity             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 438 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 438 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_self->_total_intensity); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 438, __pyx_L1_error)

#line 438 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 438 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_5;

#line 438 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = 0;

#line 438 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":426
 * 
 *     @property
 *     def total_signal(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double total
 */

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.total_signal.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_node);

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 426 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":441
 * 
 *     @property
 *     def intensity(self):             # <<<<<<<<<<<<<<
 *         return self.total_signal
 * 
 */

#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_9intensity_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_9intensity_1__get__(PyObject *__pyx_v_self) {

#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_9intensity___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_9intensity___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":442
 *     @property
 *     def intensity(self):
 *         return self.total_signal             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 442 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 442 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_total_signal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 442, __pyx_L1_error)

#line 442 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 442 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 442 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 442 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":441
 * 
 *     @property
 *     def intensity(self):             # <<<<<<<<<<<<<<
 *         return self.total_signal
 * 
 */

#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.intensity.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 441 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":445
 * 
 *     @property
 *     def neutral_mass(self):             # <<<<<<<<<<<<<<
 *         return self.mz
 * 
 */

#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_12neutral_mass_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_12neutral_mass_1__get__(PyObject *__pyx_v_self) {

#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_12neutral_mass___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_12neutral_mass___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":446
 *     @property
 *     def neutral_mass(self):
 *         return self.mz             # <<<<<<<<<<<<<<
 * 
 *     cdef double get_mz(self):
 */

#line 446 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 446 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_mz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 446, __pyx_L1_error)

#line 446 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 446 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 446 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 446 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":445
 * 
 *     @property
 *     def neutral_mass(self):             # <<<<<<<<<<<<<<
 *         return self.mz
 * 
 */

#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.neutral_mass.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 445 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":448
 *         return self.mz
 * 
 *     cdef double get_mz(self):             # <<<<<<<<<<<<<<
 *         if self._mz == -1:
 *             self._reaverage_and_update()
 */

#line 448 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 448 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_get_mz(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 448 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_r;

#line 448 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 448 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_1;

#line 448 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 448 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 448 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 448 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 448 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("get_mz", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":449
 * 
 *     cdef double get_mz(self):
 *         if self._mz == -1:             # <<<<<<<<<<<<<<
 *             self._reaverage_and_update()
 *         return self._mz
 */

#line 449 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((__pyx_v_self->__pyx_base._mz == -1.0) != 0);

#line 449 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":450
 *     cdef double get_mz(self):
 *         if self._mz == -1:
 *             self._reaverage_and_update()             # <<<<<<<<<<<<<<
 *         return self._mz
 * 
 */

#line 450 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->_reaverage_and_update(__pyx_v_self, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 450, __pyx_L1_error)

#line 450 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 450 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":449
 * 
 *     cdef double get_mz(self):
 *         if self._mz == -1:             # <<<<<<<<<<<<<<
 *             self._reaverage_and_update()
 *         return self._mz
 */

#line 449 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":451
 *         if self._mz == -1:
 *             self._reaverage_and_update()
 *         return self._mz             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 451 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_v_self->__pyx_base._mz;

#line 451 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":448
 *         return self.mz
 * 
 *     cdef double get_mz(self):             # <<<<<<<<<<<<<<
 *         if self._mz == -1:
 *             self._reaverage_and_update()
 */

#line 448 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 448 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 448 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 448 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 448 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.get_mz", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 448 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 448 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 448 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 448 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 448 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":454
 * 
 *     @property
 *     def mz(self):             # <<<<<<<<<<<<<<
 *         return self.get_mz()
 * 
 */

#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_2mz_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_2mz_1__get__(PyObject *__pyx_v_self) {

#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_2mz___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_2mz___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":455
 *     @property
 *     def mz(self):
 *         return self.get_mz()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 455 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 455 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyFloat_FromDouble(((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.get_mz(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_self))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 455, __pyx_L1_error)

#line 455 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 455 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 455 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 455 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":454
 * 
 *     @property
 *     def mz(self):             # <<<<<<<<<<<<<<
 *         return self.get_mz()
 * 
 */

#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.mz.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 454 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":458
 * 
 *     @property
 *     def times(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_5times_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_5times_1__get__(PyObject *__pyx_v_self) {

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_5times___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_5times___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {
  size_t __pyx_v_i
#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_n
#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_node = 0
#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  npy_intp __pyx_v_knd
#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyArrayObject *__pyx_v_acc = 0
#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_acc
#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  __Pyx_Buffer __pyx_pybuffer_acc
#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_1;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_2;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyArrayObject *__pyx_t_4 = NULL;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_5;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_7 = NULL;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_8 = NULL;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_9;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_10;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_11;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_12;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_13;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_pybuffer_acc.pybuffer.buf = NULL;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_pybuffer_acc.refcount = 0;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_pybuffernd_acc.data = NULL;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_pybuffernd_acc.rcbuffer = &__pyx_pybuffer_acc;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":464
 *             np.npy_intp knd
 *             np.ndarray[double, ndim=1, mode='c'] acc
 *         if self._times is None:             # <<<<<<<<<<<<<<
 *             n = self.get_size()
 *             knd = n
 */

#line 464 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = (__pyx_v_self->_times == Py_None);

#line 464 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 464 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":465
 *             np.ndarray[double, ndim=1, mode='c'] acc
 *         if self._times is None:
 *             n = self.get_size()             # <<<<<<<<<<<<<<
 *             knd = n
 *             acc = np.PyArray_ZEROS(1, &knd, np.NPY_FLOAT64, 0)
 */

#line 465 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_n = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.get_size(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_self));

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":466
 *         if self._times is None:
 *             n = self.get_size()
 *             knd = n             # <<<<<<<<<<<<<<
 *             acc = np.PyArray_ZEROS(1, &knd, np.NPY_FLOAT64, 0)
 *             # acc = _zeros(n)
 */

#line 466 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_knd = __pyx_v_n;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":467
 *             n = self.get_size()
 *             knd = n
 *             acc = np.PyArray_ZEROS(1, &knd, np.NPY_FLOAT64, 0)             # <<<<<<<<<<<<<<
 *             # acc = _zeros(n)
 *             for i in range(n):
 */

#line 467 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = PyArray_ZEROS(1, (&__pyx_v_knd), NPY_FLOAT64, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 467, __pyx_L1_error)

#line 467 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 467 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 467, __pyx_L1_error)

#line 467 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = ((PyArrayObject *)__pyx_t_3);

#line 467 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    {

#line 467 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_BufFmt_StackElem __pyx_stack[1];

#line 467 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_acc.rcbuffer->pybuffer);

#line 467 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_5 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_acc.rcbuffer->pybuffer, (PyObject*)__pyx_t_4, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_C_CONTIGUOUS| PyBUF_WRITABLE, 1, 0, __pyx_stack);

#line 467 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_t_5 < 0)) {

#line 467 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        PyErr_Fetch(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);

#line 467 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_acc.rcbuffer->pybuffer, (PyObject*)__pyx_v_acc, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_C_CONTIGUOUS| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {

#line 467 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_7); Py_XDECREF(__pyx_t_8);

#line 467 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_RaiseBufferFallbackError();

#line 467 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        } else {

#line 467 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyErr_Restore(__pyx_t_6, __pyx_t_7, __pyx_t_8);

#line 467 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 467 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_6 = __pyx_t_7 = __pyx_t_8 = 0;

#line 467 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 467 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_pybuffernd_acc.diminfo[0].strides = __pyx_pybuffernd_acc.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_acc.diminfo[0].shape = __pyx_pybuffernd_acc.rcbuffer->pybuffer.shape[0];

#line 467 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 467, __pyx_L1_error)

#line 467 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 467 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = 0;

#line 467 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_acc = ((PyArrayObject *)__pyx_t_3);

#line 467 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":469
 *             acc = np.PyArray_ZEROS(1, &knd, np.NPY_FLOAT64, 0)
 *             # acc = _zeros(n)
 *             for i in range(n):             # <<<<<<<<<<<<<<
 *                 node = self.getitem(i)
 *                 acc[i] = node.time
 */

#line 469 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_9 = __pyx_v_n;

#line 469 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_10 = __pyx_t_9;

#line 469 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {

#line 469 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_i = __pyx_t_11;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":470
 *             # acc = _zeros(n)
 *             for i in range(n):
 *                 node = self.getitem(i)             # <<<<<<<<<<<<<<
 *                 acc[i] = node.time
 *             self._times = acc
 */

#line 470 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->getitem(__pyx_v_self, __pyx_v_i)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 470, __pyx_L1_error)

#line 470 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_3);

#line 470 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_t_3));

#line 470 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_3 = 0;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":471
 *             for i in range(n):
 *                 node = self.getitem(i)
 *                 acc[i] = node.time             # <<<<<<<<<<<<<<
 *             self._times = acc
 *         return self._times
 */

#line 471 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_12 = __pyx_v_node->__pyx_base.time;

#line 471 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_13 = __pyx_v_i;

#line 471 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_5 = -1;

#line 471 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_t_13 >= (size_t)__pyx_pybuffernd_acc.diminfo[0].shape)) __pyx_t_5 = 0;

#line 471 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_t_5 != -1)) {

#line 471 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_RaiseBufferIndexError(__pyx_t_5);

#line 471 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __PYX_ERR(0, 471, __pyx_L1_error)

#line 471 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 471 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      *__Pyx_BufPtrCContig1d(double *, __pyx_pybuffernd_acc.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_acc.diminfo[0].strides) = __pyx_t_12;

#line 471 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":472
 *                 node = self.getitem(i)
 *                 acc[i] = node.time
 *             self._times = acc             # <<<<<<<<<<<<<<
 *         return self._times
 * 
 */

#line 472 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(((PyObject *)__pyx_v_acc));

#line 472 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GIVEREF(((PyObject *)__pyx_v_acc));

#line 472 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_v_self->_times);

#line 472 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_v_self->_times);

#line 472 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_self->_times = ((PyObject *)__pyx_v_acc);

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":464
 *             np.npy_intp knd
 *             np.ndarray[double, ndim=1, mode='c'] acc
 *         if self._times is None:             # <<<<<<<<<<<<<<
 *             n = self.get_size()
 *             knd = n
 */

#line 464 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":473
 *                 acc[i] = node.time
 *             self._times = acc
 *         return self._times             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 473 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 473 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_self->_times);

#line 473 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_v_self->_times;

#line 473 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":458
 * 
 *     @property
 *     def times(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_PyThreadState_declare

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_PyThreadState_assign

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_acc.rcbuffer->pybuffer);

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.times.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L2;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_acc.rcbuffer->pybuffer);

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L2:;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_node);

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_acc);

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 458 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":476
 * 
 *     @property
 *     def peaks(self):             # <<<<<<<<<<<<<<
 *         if self._peaks is None:
 *             self._peaks = tuple(node.peaks for node in self.nodes)
 */

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_5peaks_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_5peaks_1__get__(PyObject *__pyx_v_self) {

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_5peaks___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_gb_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_5peaks_7__get___2generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":478
 *     def peaks(self):
 *         if self._peaks is None:
 *             self._peaks = tuple(node.peaks for node in self.nodes)             # <<<<<<<<<<<<<<
 *         return self._peaks
 * 
 */

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_5peaks_7__get___genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr *__pyx_cur_scope
#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("genexpr", 0);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_cur_scope = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr *)__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr, __pyx_empty_tuple, NULL);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__pyx_cur_scope)) {

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_cur_scope = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr *)Py_None);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(Py_None);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __PYX_ERR(0, 478, __pyx_L1_error)

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_cur_scope);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__ *) __pyx_self;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_5peaks_7__get___2generator2, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_LCMSFeature___get___locals_genex, __pyx_n_s_ms_deisotope__c_feature_map_lcms); if (unlikely(!gen)) __PYX_ERR(0, 478, __pyx_L1_error)

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_cur_scope);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_RefNannyFinishContext();

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    return (PyObject *) gen;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.peaks.__get__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_gb_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_5peaks_7__get___2generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr *__pyx_cur_scope = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr *)__pyx_generator->closure);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_ssize_t __pyx_t_2;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*__pyx_t_3)(PyObject *);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("genexpr", 0);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  switch (__pyx_generator->resume_label) {

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    case 0: goto __pyx_L3_first_run;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    case 1: goto __pyx_L6_resume_from_yield;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    default: /* CPython raises the right error here */

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_RefNannyFinishContext();

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    return NULL;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_first_run:;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 478, __pyx_L1_error)

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 478, __pyx_L1_error) }

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (likely(PyList_CheckExact(((PyObject *)__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->__pyx_base.nodes))) || PyTuple_CheckExact(((PyObject *)__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->__pyx_base.nodes))) {

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->__pyx_base.nodes); __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = NULL;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {
    __pyx_t_2 = -1; 
#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_1 = PyObject_GetIter(((PyObject *)__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->__pyx_base.nodes)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 478, __pyx_L1_error)

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 478, __pyx_L1_error)

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  for (;;) {

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (likely(!__pyx_t_3)) {

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (likely(PyList_CheckExact(__pyx_t_1))) {

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 478, __pyx_L1_error)

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #else

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 478, __pyx_L1_error)

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #endif

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      } else {

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 478, __pyx_L1_error)

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #else

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 478, __pyx_L1_error)

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_4);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #endif

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }
    } else 
#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
{

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(!__pyx_t_4)) {

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        PyObject* exc_type = PyErr_Occurred();

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (exc_type) {

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          else __PYX_ERR(0, 478, __pyx_L1_error)

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        break;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_4);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_node);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_node, __pyx_t_4);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GIVEREF(__pyx_t_4);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = 0;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_node, __pyx_n_s_peaks); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 478, __pyx_L1_error)

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_r = __pyx_t_4;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = 0;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XGIVEREF(__pyx_t_1);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XGIVEREF(__pyx_r);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_RefNannyFinishContext();

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    /* return from generator, yielding value */

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_generator->resume_label = 1;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    return __pyx_r;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L6_resume_from_yield:;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_cur_scope->__pyx_t_0 = 0;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XGOTREF(__pyx_t_1);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 478, __pyx_L1_error)

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyErr_SetNone(PyExc_StopIteration);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #if !CYTHON_USE_EXC_INFO_STACK

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #endif

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_generator->resume_label = -1;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":476
 * 
 *     @property
 *     def peaks(self):             # <<<<<<<<<<<<<<
 *         if self._peaks is None:
 *             self._peaks = tuple(node.peaks for node in self.nodes)
 */

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_5peaks___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__ *__pyx_cur_scope
#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyObject *__pyx_gb_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_5peaks_7__get___2generator2 = 0
#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_1;

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_2;

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_cur_scope = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__ *)__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__(__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__, __pyx_empty_tuple, NULL);

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__pyx_cur_scope)) {

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_cur_scope = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__ *)Py_None);

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(Py_None);

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __PYX_ERR(0, 476, __pyx_L1_error)

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_cur_scope);

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":477
 *     @property
 *     def peaks(self):
 *         if self._peaks is None:             # <<<<<<<<<<<<<<
 *             self._peaks = tuple(node.peaks for node in self.nodes)
 *         return self._peaks
 */

#line 477 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_self->_peaks == Py_None);

#line 477 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = (__pyx_t_1 != 0);

#line 477 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":478
 *     def peaks(self):
 *         if self._peaks is None:
 *             self._peaks = tuple(node.peaks for node in self.nodes)             # <<<<<<<<<<<<<<
 *         return self._peaks
 * 
 */

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_5peaks_7__get___genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 478, __pyx_L1_error)

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = __Pyx_PySequence_Tuple(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 478, __pyx_L1_error)

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GIVEREF(__pyx_t_4);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_self->_peaks);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_cur_scope->__pyx_v_self->_peaks);

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_cur_scope->__pyx_v_self->_peaks = __pyx_t_4;

#line 478 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":477
 *     @property
 *     def peaks(self):
 *         if self._peaks is None:             # <<<<<<<<<<<<<<
 *             self._peaks = tuple(node.peaks for node in self.nodes)
 *         return self._peaks
 */

#line 477 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":479
 *         if self._peaks is None:
 *             self._peaks = tuple(node.peaks for node in self.nodes)
 *         return self._peaks             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 479 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 479 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self->_peaks);

#line 479 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_cur_scope->__pyx_v_self->_peaks;

#line 479 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":476
 * 
 *     @property
 *     def peaks(self):             # <<<<<<<<<<<<<<
 *         if self._peaks is None:
 *             self._peaks = tuple(node.peaks for node in self.nodes)
 */

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.peaks.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_gb_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_5peaks_7__get___2generator2);

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 476 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":482
 * 
 *     @property
 *     def start_time(self):             # <<<<<<<<<<<<<<
 *         if self._start_time == -1:
 *             self._start_time = self.nodes[0].time
 */

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10start_time_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10start_time_1__get__(PyObject *__pyx_v_self) {

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10start_time___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10start_time___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_1;

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_4;

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":483
 *     @property
 *     def start_time(self):
 *         if self._start_time == -1:             # <<<<<<<<<<<<<<
 *             self._start_time = self.nodes[0].time
 *         return self._start_time
 */

#line 483 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((__pyx_v_self->__pyx_base._start_time == -1.0) != 0);

#line 483 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":484
 *     def start_time(self):
 *         if self._start_time == -1:
 *             self._start_time = self.nodes[0].time             # <<<<<<<<<<<<<<
 *         return self._start_time
 * 
 */

#line 484 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = __Pyx_GetItemInt(((PyObject *)__pyx_v_self->__pyx_base.nodes), 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 484, __pyx_L1_error)

#line 484 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 484 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 484, __pyx_L1_error)

#line 484 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 484 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 484 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 484, __pyx_L1_error)

#line 484 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 484 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_self->__pyx_base._start_time = __pyx_t_4;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":483
 *     @property
 *     def start_time(self):
 *         if self._start_time == -1:             # <<<<<<<<<<<<<<
 *             self._start_time = self.nodes[0].time
 *         return self._start_time
 */

#line 483 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":485
 *         if self._start_time == -1:
 *             self._start_time = self.nodes[0].time
 *         return self._start_time             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 485 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 485 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->__pyx_base._start_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 485, __pyx_L1_error)

#line 485 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 485 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_3;

#line 485 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = 0;

#line 485 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":482
 * 
 *     @property
 *     def start_time(self):             # <<<<<<<<<<<<<<
 *         if self._start_time == -1:
 *             self._start_time = self.nodes[0].time
 */

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.start_time.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 482 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":488
 * 
 *     @property
 *     def end_time(self):             # <<<<<<<<<<<<<<
 *         if self._end_time == -1:
 *             self._end_time = self.nodes[-1].time
 */

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_8end_time_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_8end_time_1__get__(PyObject *__pyx_v_self) {

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_8end_time___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_8end_time___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_1;

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_4;

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":489
 *     @property
 *     def end_time(self):
 *         if self._end_time == -1:             # <<<<<<<<<<<<<<
 *             self._end_time = self.nodes[-1].time
 *         return self._end_time
 */

#line 489 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((__pyx_v_self->__pyx_base._end_time == -1.0) != 0);

#line 489 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":490
 *     def end_time(self):
 *         if self._end_time == -1:
 *             self._end_time = self.nodes[-1].time             # <<<<<<<<<<<<<<
 *         return self._end_time
 * 
 */

#line 490 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = __Pyx_GetItemInt(((PyObject *)__pyx_v_self->__pyx_base.nodes), -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 490, __pyx_L1_error)

#line 490 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 490 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 490, __pyx_L1_error)

#line 490 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 490 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 490 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 490, __pyx_L1_error)

#line 490 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 490 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_self->__pyx_base._end_time = __pyx_t_4;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":489
 *     @property
 *     def end_time(self):
 *         if self._end_time == -1:             # <<<<<<<<<<<<<<
 *             self._end_time = self.nodes[-1].time
 *         return self._end_time
 */

#line 489 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":491
 *         if self._end_time == -1:
 *             self._end_time = self.nodes[-1].time
 *         return self._end_time             # <<<<<<<<<<<<<<
 * 
 *     cdef double get_start_time(self):
 */

#line 491 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 491 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->__pyx_base._end_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 491, __pyx_L1_error)

#line 491 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 491 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_3;

#line 491 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = 0;

#line 491 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":488
 * 
 *     @property
 *     def end_time(self):             # <<<<<<<<<<<<<<
 *         if self._end_time == -1:
 *             self._end_time = self.nodes[-1].time
 */

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.end_time.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 488 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":493
 *         return self._end_time
 * 
 *     cdef double get_start_time(self):             # <<<<<<<<<<<<<<
 *         if self._start_time == -1:
 *             self._start_time = self.getitem(0).time
 */

#line 493 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 493 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_get_start_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 493 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_r;

#line 493 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 493 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_1;

#line 493 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 493 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_3;

#line 493 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 493 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 493 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 493 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("get_start_time", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":494
 * 
 *     cdef double get_start_time(self):
 *         if self._start_time == -1:             # <<<<<<<<<<<<<<
 *             self._start_time = self.getitem(0).time
 *         return self._start_time
 */

#line 494 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((__pyx_v_self->__pyx_base._start_time == -1.0) != 0);

#line 494 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":495
 *     cdef double get_start_time(self):
 *         if self._start_time == -1:
 *             self._start_time = self.getitem(0).time             # <<<<<<<<<<<<<<
 *         return self._start_time
 * 
 */

#line 495 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->getitem(__pyx_v_self, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 495, __pyx_L1_error)

#line 495 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 495 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_t_2)->__pyx_base.time;

#line 495 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 495 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_self->__pyx_base._start_time = __pyx_t_3;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":494
 * 
 *     cdef double get_start_time(self):
 *         if self._start_time == -1:             # <<<<<<<<<<<<<<
 *             self._start_time = self.getitem(0).time
 *         return self._start_time
 */

#line 494 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":496
 *         if self._start_time == -1:
 *             self._start_time = self.getitem(0).time
 *         return self._start_time             # <<<<<<<<<<<<<<
 * 
 *     cdef double get_end_time(self):
 */

#line 496 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_v_self->__pyx_base._start_time;

#line 496 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":493
 *         return self._end_time
 * 
 *     cdef double get_start_time(self):             # <<<<<<<<<<<<<<
 *         if self._start_time == -1:
 *             self._start_time = self.getitem(0).time
 */

#line 493 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 493 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 493 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 493 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 493 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.get_start_time", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 493 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 493 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 493 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 493 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 493 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":498
 *         return self._start_time
 * 
 *     cdef double get_end_time(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t n
 */

#line 498 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 498 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_get_end_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {
  size_t __pyx_v_n
#line 498 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 498 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_r;

#line 498 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 498 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_1;

#line 498 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 498 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_3;

#line 498 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 498 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 498 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 498 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("get_end_time", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":501
 *         cdef:
 *             size_t n
 *         if self._end_time == -1:             # <<<<<<<<<<<<<<
 *             n = self.get_size()
 *             self._end_time = self.getitem(n - 1).time
 */

#line 501 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((__pyx_v_self->__pyx_base._end_time == -1.0) != 0);

#line 501 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_1) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":502
 *             size_t n
 *         if self._end_time == -1:
 *             n = self.get_size()             # <<<<<<<<<<<<<<
 *             self._end_time = self.getitem(n - 1).time
 *         return self._end_time
 */

#line 502 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_n = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.get_size(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_self));

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":503
 *         if self._end_time == -1:
 *             n = self.get_size()
 *             self._end_time = self.getitem(n - 1).time             # <<<<<<<<<<<<<<
 *         return self._end_time
 * 
 */

#line 503 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->getitem(__pyx_v_self, (__pyx_v_n - 1))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 503, __pyx_L1_error)

#line 503 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 503 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_t_2)->__pyx_base.time;

#line 503 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 503 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_self->__pyx_base._end_time = __pyx_t_3;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":501
 *         cdef:
 *             size_t n
 *         if self._end_time == -1:             # <<<<<<<<<<<<<<
 *             n = self.get_size()
 *             self._end_time = self.getitem(n - 1).time
 */

#line 501 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":504
 *             n = self.get_size()
 *             self._end_time = self.getitem(n - 1).time
 *         return self._end_time             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint overlaps_in_time(self, FeatureBase interval):
 */

#line 504 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_v_self->__pyx_base._end_time;

#line 504 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":498
 *         return self._start_time
 * 
 *     cdef double get_end_time(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t n
 */

#line 498 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 498 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 498 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 498 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 498 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.get_end_time", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 498 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 498 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 498 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 498 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 498 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":506
 *         return self._end_time
 * 
 *     cpdef bint overlaps_in_time(self, FeatureBase interval):             # <<<<<<<<<<<<<<
 *         """Test whether or not this feature overlaps with `interval`.
 * 
 */

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_13overlaps_in_time(PyObject *__pyx_v_self, PyObject *__pyx_v_interval); /*proto*/

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_overlaps_in_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_interval, int __pyx_skip_dispatch) {
  double __pyx_v_self_start_time
#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  double __pyx_v_self_end_time
#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  double __pyx_v_other_start_time
#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  double __pyx_v_other_end_time
#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  int __pyx_v_cond
#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_r;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_5;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_6;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("overlaps_in_time", 0);

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_overlaps_in_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 506, __pyx_L1_error)

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_13overlaps_in_time)) {

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_4 = NULL;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_4)) {

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_interval)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_interval));

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 506, __pyx_L1_error)

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 506, __pyx_L1_error)

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = __pyx_t_5;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":525
 *             bint cond
 * 
 *         self_start_time = self.get_start_time()             # <<<<<<<<<<<<<<
 *         self_end_time = self.get_end_time()
 *         other_start_time = interval.get_start_time()
 */

#line 525 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self_start_time = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.get_start_time(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_self));

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":526
 * 
 *         self_start_time = self.get_start_time()
 *         self_end_time = self.get_end_time()             # <<<<<<<<<<<<<<
 *         other_start_time = interval.get_start_time()
 *         other_end_time = interval.get_end_time()
 */

#line 526 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self_end_time = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.get_end_time(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_self));

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":527
 *         self_start_time = self.get_start_time()
 *         self_end_time = self.get_end_time()
 *         other_start_time = interval.get_start_time()             # <<<<<<<<<<<<<<
 *         other_end_time = interval.get_end_time()
 * 
 */

#line 527 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_other_start_time = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_interval->__pyx_vtab)->get_start_time(__pyx_v_interval);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":528
 *         self_end_time = self.get_end_time()
 *         other_start_time = interval.get_start_time()
 *         other_end_time = interval.get_end_time()             # <<<<<<<<<<<<<<
 * 
 *         cond = ((self_start_time <= other_start_time and self_end_time >= other_end_time) or (
 */

#line 528 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_other_end_time = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_interval->__pyx_vtab)->get_end_time(__pyx_v_interval);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":530
 *         other_end_time = interval.get_end_time()
 * 
 *         cond = ((self_start_time <= other_start_time and self_end_time >= other_end_time) or (             # <<<<<<<<<<<<<<
 *             self_start_time >= other_start_time and self_end_time <= other_end_time) or (
 *             self_start_time >= other_start_time and self_end_time >= other_end_time and
 */

#line 530 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = ((__pyx_v_self_start_time <= __pyx_v_other_start_time) != 0);

#line 530 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (!__pyx_t_6) {

#line 530 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L4_next_or;

#line 530 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {

#line 530 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 530 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = ((__pyx_v_self_end_time >= __pyx_v_other_end_time) != 0);

#line 530 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (!__pyx_t_6) {

#line 530 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {

#line 530 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = __pyx_t_6;

#line 530 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L3_bool_binop_done;

#line 530 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 530 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L4_next_or:;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":531
 * 
 *         cond = ((self_start_time <= other_start_time and self_end_time >= other_end_time) or (
 *             self_start_time >= other_start_time and self_end_time <= other_end_time) or (             # <<<<<<<<<<<<<<
 *             self_start_time >= other_start_time and self_end_time >= other_end_time and
 *             self_start_time <= other_end_time) or (
 */

#line 531 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = ((__pyx_v_self_start_time >= __pyx_v_other_start_time) != 0);

#line 531 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (!__pyx_t_6) {

#line 531 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L6_next_or;

#line 531 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {

#line 531 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 531 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = ((__pyx_v_self_end_time <= __pyx_v_other_end_time) != 0);

#line 531 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (!__pyx_t_6) {

#line 531 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {

#line 531 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = __pyx_t_6;

#line 531 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L3_bool_binop_done;

#line 531 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 531 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L6_next_or:;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":532
 *         cond = ((self_start_time <= other_start_time and self_end_time >= other_end_time) or (
 *             self_start_time >= other_start_time and self_end_time <= other_end_time) or (
 *             self_start_time >= other_start_time and self_end_time >= other_end_time and             # <<<<<<<<<<<<<<
 *             self_start_time <= other_end_time) or (
 *             self_start_time <= other_start_time and self_end_time >= other_start_time) or (
 */

#line 532 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = ((__pyx_v_self_start_time >= __pyx_v_other_start_time) != 0);

#line 532 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (!__pyx_t_6) {

#line 532 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L8_next_or;

#line 532 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {

#line 532 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 532 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = ((__pyx_v_self_end_time >= __pyx_v_other_end_time) != 0);

#line 532 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (!__pyx_t_6) {

#line 532 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L8_next_or;

#line 532 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {

#line 532 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":533
 *             self_start_time >= other_start_time and self_end_time <= other_end_time) or (
 *             self_start_time >= other_start_time and self_end_time >= other_end_time and
 *             self_start_time <= other_end_time) or (             # <<<<<<<<<<<<<<
 *             self_start_time <= other_start_time and self_end_time >= other_start_time) or (
 *             self_start_time <= other_end_time and self_end_time >= other_end_time))
 */

#line 533 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = ((__pyx_v_self_start_time <= __pyx_v_other_end_time) != 0);

#line 533 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (!__pyx_t_6) {

#line 533 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {

#line 533 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = __pyx_t_6;

#line 533 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L3_bool_binop_done;

#line 533 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 533 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L8_next_or:;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":534
 *             self_start_time >= other_start_time and self_end_time >= other_end_time and
 *             self_start_time <= other_end_time) or (
 *             self_start_time <= other_start_time and self_end_time >= other_start_time) or (             # <<<<<<<<<<<<<<
 *             self_start_time <= other_end_time and self_end_time >= other_end_time))
 *         return cond
 */

#line 534 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = ((__pyx_v_self_start_time <= __pyx_v_other_start_time) != 0);

#line 534 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (!__pyx_t_6) {

#line 534 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L11_next_or;

#line 534 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {

#line 534 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 534 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = ((__pyx_v_self_end_time >= __pyx_v_other_start_time) != 0);

#line 534 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (!__pyx_t_6) {

#line 534 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {

#line 534 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = __pyx_t_6;

#line 534 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L3_bool_binop_done;

#line 534 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 534 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L11_next_or:;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":535
 *             self_start_time <= other_end_time) or (
 *             self_start_time <= other_start_time and self_end_time >= other_start_time) or (
 *             self_start_time <= other_end_time and self_end_time >= other_end_time))             # <<<<<<<<<<<<<<
 *         return cond
 * 
 */

#line 535 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = ((__pyx_v_self_start_time <= __pyx_v_other_end_time) != 0);

#line 535 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_6) {

#line 535 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {

#line 535 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = __pyx_t_6;

#line 535 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L3_bool_binop_done;

#line 535 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 535 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = ((__pyx_v_self_end_time >= __pyx_v_other_end_time) != 0);

#line 535 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = __pyx_t_6;

#line 535 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_bool_binop_done:;

#line 535 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_cond = __pyx_t_5;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":536
 *             self_start_time <= other_start_time and self_end_time >= other_start_time) or (
 *             self_start_time <= other_end_time and self_end_time >= other_end_time))
 *         return cond             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint spans_in_time(self, double time):
 */

#line 536 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_v_cond;

#line 536 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":506
 *         return self._end_time
 * 
 *     cpdef bint overlaps_in_time(self, FeatureBase interval):             # <<<<<<<<<<<<<<
 *         """Test whether or not this feature overlaps with `interval`.
 * 
 */

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.overlaps_in_time", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_13overlaps_in_time(PyObject *__pyx_v_self, PyObject *__pyx_v_interval); /*proto*/

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_12overlaps_in_time[] = "LCMSFeature.overlaps_in_time(self, FeatureBase interval) -> bool\nTest whether or not this feature overlaps with `interval`.\n\n        Parameters\n        ----------\n        interval : FeatureBase\n            The feature to test spanning\n\n        Returns\n        -------\n        bool:\n            Whether this feature spans `interval` or the `interval` spans this feature,\n            partially or one is fully contained in the other.\n        ";

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_13overlaps_in_time(PyObject *__pyx_v_self, PyObject *__pyx_v_interval) {

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("overlaps_in_time (wrapper)", 0);

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_interval), __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase, 1, "interval", 0))) __PYX_ERR(0, 506, __pyx_L1_error)
  __pyx_r = 
#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_12overlaps_in_time(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_interval));

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_12overlaps_in_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_interval) {

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("overlaps_in_time", 0);

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_overlaps_in_time(__pyx_v_self, __pyx_v_interval, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 506, __pyx_L1_error)

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.overlaps_in_time", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 506 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":538
 *         return cond
 * 
 *     cpdef bint spans_in_time(self, double time):             # <<<<<<<<<<<<<<
 *         """Test whether or not `time` is between :attr:`start_time`
 *         and :attr:`end_time`.
 */

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_15spans_in_time(PyObject *__pyx_v_self, PyObject *__pyx_arg_time); /*proto*/

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_spans_in_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, double __pyx_v_time, int __pyx_skip_dispatch) {

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_r;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_6;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("spans_in_time", 0);

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_spans_in_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 538, __pyx_L1_error)

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_15spans_in_time)) {

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 538, __pyx_L1_error)

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_5 = NULL;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_5)) {

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_5);

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_4, function);

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 538, __pyx_L1_error)

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 538, __pyx_L1_error)

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = __pyx_t_6;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":552
 *             Whether or not the time is contained
 *         """
 *         return self.get_start_time() <= time <= self.get_end_time()             # <<<<<<<<<<<<<<
 * 
 *     def as_arrays(self, dtype=np.float64):
 */

#line 552 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = (((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.get_start_time(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_self)) <= __pyx_v_time);

#line 552 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_6) {

#line 552 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_6 = (__pyx_v_time <= ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.get_end_time(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_self)));

#line 552 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 552 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_6;

#line 552 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":538
 *         return cond
 * 
 *     cpdef bint spans_in_time(self, double time):             # <<<<<<<<<<<<<<
 *         """Test whether or not `time` is between :attr:`start_time`
 *         and :attr:`end_time`.
 */

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.spans_in_time", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_15spans_in_time(PyObject *__pyx_v_self, PyObject *__pyx_arg_time); /*proto*/

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14spans_in_time[] = "LCMSFeature.spans_in_time(self, double time) -> bool\nTest whether or not `time` is between :attr:`start_time`\n        and :attr:`end_time`.\n\n        Parameters\n        ----------\n        time : float\n            The time to test for\n\n        Returns\n        -------\n        bool:\n            Whether or not the time is contained\n        ";

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_15spans_in_time(PyObject *__pyx_v_self, PyObject *__pyx_arg_time) {
  double __pyx_v_time
#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("spans_in_time (wrapper)", 0);

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  assert(__pyx_arg_time); {

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_time = __pyx_PyFloat_AsDouble(__pyx_arg_time); if (unlikely((__pyx_v_time == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 538, __pyx_L3_error)

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_error:;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.spans_in_time", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return NULL;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14spans_in_time(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), ((double)__pyx_v_time));

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14spans_in_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, double __pyx_v_time) {

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("spans_in_time", 0);

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_spans_in_time(__pyx_v_self, __pyx_v_time, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 538, __pyx_L1_error)

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.spans_in_time", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 538 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":554
 *         return self.get_start_time() <= time <= self.get_end_time()
 * 
 *     def as_arrays(self, dtype=np.float64):             # <<<<<<<<<<<<<<
 *         """Convert this object into a pair of time and intensity arrays, summing
 *         all peaks at the same time point.
 */

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_17as_arrays(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_16as_arrays[] = "LCMSFeature.as_arrays(self, dtype=np.float64)\nConvert this object into a pair of time and intensity arrays, summing\n        all peaks at the same time point.\n\n        Returns\n        -------\n        times : np.ndarray\n            The time array for this feature.\n        intensities : np.ndarray\n            The intensity array for this feature.\n        ";

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_17as_arrays(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_dtype = 0
#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("as_arrays (wrapper)", 0);

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dtype,0};

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject* values[1] = {0};

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    values[0] = __pyx_k__3;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(__pyx_kwds)) {

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      Py_ssize_t kw_args;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {
        case  1: 
#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0: break;
        default: 
#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0:

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (kw_args > 0) {

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dtype);

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (value) { values[0] = value; kw_args--; }

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(kw_args > 0)) {

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "as_arrays") < 0)) __PYX_ERR(0, 554, __pyx_L3_error)

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: 
#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  0: 
#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
break;
        default: 
#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_dtype = values[0];

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("as_arrays", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); 
#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__PYX_ERR(0, 554, __pyx_L3_error)

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_error:;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.as_arrays", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return NULL;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_16as_arrays(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), __pyx_v_dtype);

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_16as_arrays(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_dtype) {
  PyObject *__pyx_v_rts = NULL
#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyObject *__pyx_v_signal = NULL
#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyObject *__pyx_v_node = NULL
#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_ssize_t __pyx_t_4;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*__pyx_t_5)(PyObject *);

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_7 = NULL;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_8 = NULL;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("as_arrays", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":565
 *             The intensity array for this feature.
 *         """
 *         rts = np.array(             # <<<<<<<<<<<<<<
 *             [node.time for node in self.nodes], dtype=dtype)
 *         signal = np.array([node.total_intensity()
 */

#line 565 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 565, __pyx_L1_error)

#line 565 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 565 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 565, __pyx_L1_error)

#line 565 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 565 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":566
 *         """
 *         rts = np.array(
 *             [node.time for node in self.nodes], dtype=dtype)             # <<<<<<<<<<<<<<
 *         signal = np.array([node.total_intensity()
 *                            for node in self.nodes], dtype=dtype)
 */

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 566, __pyx_L1_error)

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (likely(PyList_CheckExact(((PyObject *)__pyx_v_self->__pyx_base.nodes))) || PyTuple_CheckExact(((PyObject *)__pyx_v_self->__pyx_base.nodes))) {

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = ((PyObject *)__pyx_v_self->__pyx_base.nodes); __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = NULL;

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {
    __pyx_t_4 = -1; 
#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_3 = PyObject_GetIter(((PyObject *)__pyx_v_self->__pyx_base.nodes)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 566, __pyx_L1_error)

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 566, __pyx_L1_error)

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  for (;;) {

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (likely(!__pyx_t_5)) {

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (likely(PyList_CheckExact(__pyx_t_3))) {

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_6 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 566, __pyx_L1_error)

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #else

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 566, __pyx_L1_error)

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_6);

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #endif

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      } else {

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 566, __pyx_L1_error)

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #else

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 566, __pyx_L1_error)

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_6);

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #endif

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }
    } else 
#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
{

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_6 = __pyx_t_5(__pyx_t_3);

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(!__pyx_t_6)) {

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        PyObject* exc_type = PyErr_Occurred();

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (exc_type) {

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          else __PYX_ERR(0, 566, __pyx_L1_error)

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        break;

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_6);

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF_SET(__pyx_v_node, __pyx_t_6);

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_6 = 0;

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_node, __pyx_n_s_time); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 566, __pyx_L1_error)

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 566, __pyx_L1_error)

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":565
 *             The intensity array for this feature.
 *         """
 *         rts = np.array(             # <<<<<<<<<<<<<<
 *             [node.time for node in self.nodes], dtype=dtype)
 *         signal = np.array([node.total_intensity()
 */

#line 565 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 565, __pyx_L1_error)

#line 565 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 565 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 565 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);

#line 565 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":566
 *         """
 *         rts = np.array(
 *             [node.time for node in self.nodes], dtype=dtype)             # <<<<<<<<<<<<<<
 *         signal = np.array([node.total_intensity()
 *                            for node in self.nodes], dtype=dtype)
 */

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 566, __pyx_L1_error)

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 566 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_v_dtype) < 0) __PYX_ERR(0, 566, __pyx_L1_error)

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":565
 *             The intensity array for this feature.
 *         """
 *         rts = np.array(             # <<<<<<<<<<<<<<
 *             [node.time for node in self.nodes], dtype=dtype)
 *         signal = np.array([node.total_intensity()
 */

#line 565 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 565, __pyx_L1_error)

#line 565 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_6);

#line 565 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 565 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 565 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 565 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_rts = __pyx_t_6;

#line 565 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":567
 *         rts = np.array(
 *             [node.time for node in self.nodes], dtype=dtype)
 *         signal = np.array([node.total_intensity()             # <<<<<<<<<<<<<<
 *                            for node in self.nodes], dtype=dtype)
 *         return rts, signal
 */

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 567, __pyx_L1_error)

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_6);

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 567, __pyx_L1_error)

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 567, __pyx_L1_error)

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_6);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":568
 *             [node.time for node in self.nodes], dtype=dtype)
 *         signal = np.array([node.total_intensity()
 *                            for node in self.nodes], dtype=dtype)             # <<<<<<<<<<<<<<
 *         return rts, signal
 * 
 */

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (likely(PyList_CheckExact(((PyObject *)__pyx_v_self->__pyx_base.nodes))) || PyTuple_CheckExact(((PyObject *)__pyx_v_self->__pyx_base.nodes))) {

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = ((PyObject *)__pyx_v_self->__pyx_base.nodes); __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = NULL;

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {
    __pyx_t_4 = -1; 
#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_3 = PyObject_GetIter(((PyObject *)__pyx_v_self->__pyx_base.nodes)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 568, __pyx_L1_error)

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 568, __pyx_L1_error)

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  for (;;) {

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (likely(!__pyx_t_5)) {

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (likely(PyList_CheckExact(__pyx_t_3))) {

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 568, __pyx_L1_error)

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #else

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 568, __pyx_L1_error)

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #endif

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      } else {

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 568, __pyx_L1_error)

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #else

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 568, __pyx_L1_error)

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #endif

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }
    } else 
#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
{

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_2 = __pyx_t_5(__pyx_t_3);

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(!__pyx_t_2)) {

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        PyObject* exc_type = PyErr_Occurred();

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (exc_type) {

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          else __PYX_ERR(0, 568, __pyx_L1_error)

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        break;

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_2);

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF_SET(__pyx_v_node, __pyx_t_2);

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":567
 *         rts = np.array(
 *             [node.time for node in self.nodes], dtype=dtype)
 *         signal = np.array([node.total_intensity()             # <<<<<<<<<<<<<<
 *                            for node in self.nodes], dtype=dtype)
 *         return rts, signal
 */

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_node, __pyx_n_s_total_intensity); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 567, __pyx_L1_error)

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_7);

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_8 = NULL;

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (likely(__pyx_t_8)) {

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_8);

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(function);

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF_SET(__pyx_t_7, function);

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_7);

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 567, __pyx_L1_error)

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_6, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 567, __pyx_L1_error)

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":568
 *             [node.time for node in self.nodes], dtype=dtype)
 *         signal = np.array([node.total_intensity()
 *                            for node in self.nodes], dtype=dtype)             # <<<<<<<<<<<<<<
 *         return rts, signal
 * 
 */

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":567
 *         rts = np.array(
 *             [node.time for node in self.nodes], dtype=dtype)
 *         signal = np.array([node.total_intensity()             # <<<<<<<<<<<<<<
 *                            for node in self.nodes], dtype=dtype)
 *         return rts, signal
 */

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 567, __pyx_L1_error)

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_6);

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_6);

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":568
 *             [node.time for node in self.nodes], dtype=dtype)
 *         signal = np.array([node.total_intensity()
 *                            for node in self.nodes], dtype=dtype)             # <<<<<<<<<<<<<<
 *         return rts, signal
 * 
 */

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 568, __pyx_L1_error)

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_6);

#line 568 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_dtype, __pyx_v_dtype) < 0) __PYX_ERR(0, 568, __pyx_L1_error)

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":567
 *         rts = np.array(
 *             [node.time for node in self.nodes], dtype=dtype)
 *         signal = np.array([node.total_intensity()             # <<<<<<<<<<<<<<
 *                            for node in self.nodes], dtype=dtype)
 *         return rts, signal
 */

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 567, __pyx_L1_error)

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_signal = __pyx_t_2;

#line 567 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":569
 *         signal = np.array([node.total_intensity()
 *                            for node in self.nodes], dtype=dtype)
 *         return rts, signal             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */

#line 569 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 569 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 569, __pyx_L1_error)

#line 569 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 569 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_rts);

#line 569 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_v_rts);

#line 569 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_rts);

#line 569 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_signal);

#line 569 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_v_signal);

#line 569 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_signal);

#line 569 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_2;

#line 569 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = 0;

#line 569 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":554
 *         return self.get_start_time() <= time <= self.get_end_time()
 * 
 *     def as_arrays(self, dtype=np.float64):             # <<<<<<<<<<<<<<
 *         """Convert this object into a pair of time and intensity arrays, summing
 *         all peaks at the same time point.
 */

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_7);

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_8);

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.as_arrays", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_rts);

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_signal);

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_node);

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":571
 *         return rts, signal
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "%s(%0.4f, %0.2f, %0.2f)" % (
 *             self.__class__.__name__, self.mz, self.start_time, self.end_time)
 */

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_19__repr__(PyObject *__pyx_v_self); /*proto*/

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_19__repr__(PyObject *__pyx_v_self) {

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = 
#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_18__repr__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_18__repr__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":572
 * 
 *     def __repr__(self):
 *         return "%s(%0.4f, %0.2f, %0.2f)" % (             # <<<<<<<<<<<<<<
 *             self.__class__.__name__, self.mz, self.start_time, self.end_time)
 * 
 */

#line 572 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":573
 *     def __repr__(self):
 *         return "%s(%0.4f, %0.2f, %0.2f)" % (
 *             self.__class__.__name__, self.mz, self.start_time, self.end_time)             # <<<<<<<<<<<<<<
 * 
 *     def _copy_chunk(self, nodes, *args, **kwargs):
 */

#line 573 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 573, __pyx_L1_error)

#line 573 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 573 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 573, __pyx_L1_error)

#line 573 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 573 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 573 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_mz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 573, __pyx_L1_error)

#line 573 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 573 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_start_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 573, __pyx_L1_error)

#line 573 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 573 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_end_time); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 573, __pyx_L1_error)

#line 573 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 573 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 573, __pyx_L1_error)

#line 573 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 573 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 573 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);

#line 573 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 573 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);

#line 573 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 573 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);

#line 573 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_4);

#line 573 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);

#line 573 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = 0;

#line 573 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 573 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = 0;

#line 573 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":572
 * 
 *     def __repr__(self):
 *         return "%s(%0.4f, %0.2f, %0.2f)" % (             # <<<<<<<<<<<<<<
 *             self.__class__.__name__, self.mz, self.start_time, self.end_time)
 * 
 */

#line 572 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_s_0_4f_0_2f_0_2f, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 572, __pyx_L1_error)

#line 572 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 572 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 572 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_4;

#line 572 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = 0;

#line 572 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":571
 *         return rts, signal
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "%s(%0.4f, %0.2f, %0.2f)" % (
 *             self.__class__.__name__, self.mz, self.start_time, self.end_time)
 */

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 571 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":575
 *             self.__class__.__name__, self.mz, self.start_time, self.end_time)
 * 
 *     def _copy_chunk(self, nodes, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         x = self.__class__(LCMSFeatureTreeList(nodes))
 *         x.used_as_adduct = list(self.used_as_adduct)
 */

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_21_copy_chunk(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_20_copy_chunk[] = "LCMSFeature._copy_chunk(self, nodes, *args, **kwargs)";

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_21_copy_chunk(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_nodes = 0
#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  CYTHON_UNUSED PyObject *__pyx_v_args = 0
#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0
#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_copy_chunk (wrapper)", 0);

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_kwargs);

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (PyTuple_GET_SIZE(__pyx_args) > 1) {

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_args = PyTuple_GetSlice(__pyx_args, 1, PyTuple_GET_SIZE(__pyx_args));

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__pyx_v_args)) {

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_RefNannyFinishContext();

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      return NULL;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_v_args);

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {
    __pyx_v_args = __pyx_empty_tuple; 
#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__Pyx_INCREF(__pyx_empty_tuple);

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_nodes,0};

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject* values[1] = {0};

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(__pyx_kwds)) {

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      Py_ssize_t kw_args;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        default:
        case  1: 
#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0: break;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0:

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nodes)) != 0)) kw_args--;
        else 
#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(kw_args > 0)) {

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        const Py_ssize_t used_pos_args = (pos_args < 1) ? pos_args : 1;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, used_pos_args, "_copy_chunk") < 0)) __PYX_ERR(0, 575, __pyx_L3_error)

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) < 1) {

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L5_argtuple_error;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_nodes = values[0];

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_copy_chunk", 0, 1, 1, PyTuple_GET_SIZE(__pyx_args)); 
#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__PYX_ERR(0, 575, __pyx_L3_error)

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_error:;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature._copy_chunk", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return NULL;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_20_copy_chunk(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), __pyx_v_nodes, __pyx_v_args, __pyx_v_kwargs);

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_args);

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_kwargs);

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_20_copy_chunk(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_nodes, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v_x = NULL
#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_copy_chunk", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":576
 * 
 *     def _copy_chunk(self, nodes, *args, **kwargs):
 *         x = self.__class__(LCMSFeatureTreeList(nodes))             # <<<<<<<<<<<<<<
 *         x.used_as_adduct = list(self.used_as_adduct)
 *         return x
 */

#line 576 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 576, __pyx_L1_error)

#line 576 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 576 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList), __pyx_v_nodes); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 576, __pyx_L1_error)

#line 576 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 576 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = NULL;

#line 576 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 576 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);

#line 576 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (likely(__pyx_t_4)) {

#line 576 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 576 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(__pyx_t_4);

#line 576 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(function);

#line 576 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF_SET(__pyx_t_2, function);

#line 576 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 576 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 576 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);

#line 576 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 576 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 576 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 576, __pyx_L1_error)

#line 576 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 576 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 576 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_x = __pyx_t_1;

#line 576 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":577
 *     def _copy_chunk(self, nodes, *args, **kwargs):
 *         x = self.__class__(LCMSFeatureTreeList(nodes))
 *         x.used_as_adduct = list(self.used_as_adduct)             # <<<<<<<<<<<<<<
 *         return x
 * 
 */

#line 577 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PySequence_List(__pyx_v_self->used_as_adduct); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 577, __pyx_L1_error)

#line 577 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 577 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_x, __pyx_n_s_used_as_adduct, __pyx_t_1) < 0) __PYX_ERR(0, 577, __pyx_L1_error)

#line 577 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":578
 *         x = self.__class__(LCMSFeatureTreeList(nodes))
 *         x.used_as_adduct = list(self.used_as_adduct)
 *         return x             # <<<<<<<<<<<<<<
 * 
 *     def split_at(self, time):
 */

#line 578 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 578 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_x);

#line 578 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_v_x;

#line 578 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":575
 *             self.__class__.__name__, self.mz, self.start_time, self.end_time)
 * 
 *     def _copy_chunk(self, nodes, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         x = self.__class__(LCMSFeatureTreeList(nodes))
 *         x.used_as_adduct = list(self.used_as_adduct)
 */

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature._copy_chunk", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_x);

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 575 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":580
 *         return x
 * 
 *     def split_at(self, time):             # <<<<<<<<<<<<<<
 *         """Split this feature at a specific time point into two pieces, before
 *         and after that point in time.
 */

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_23split_at(PyObject *__pyx_v_self, PyObject *__pyx_v_time); /*proto*/

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_22split_at[] = "LCMSFeature.split_at(self, time)\nSplit this feature at a specific time point into two pieces, before\n        and after that point in time.\n\n        Parameters\n        ----------\n        time : float\n            The time point to split at\n\n        Returns\n        -------\n        before : LCMSFeature\n            The component up to the provided time\n        after : LCMSFeature\n            The component after the provided time\n        ";

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_23split_at(PyObject *__pyx_v_self, PyObject *__pyx_v_time) {

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("split_at (wrapper)", 0);
  __pyx_r = 
#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_22split_at(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), ((PyObject *)__pyx_v_time));

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_22split_at(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_time) {
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL
#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyObject *__pyx_v_i = NULL
#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_1;

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_5;

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("split_at", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":596
 *             The component after the provided time
 *         """
 *         _, i = self.nodes.find_time(time)             # <<<<<<<<<<<<<<
 *         if self.nodes[i].time < time:
 *             i += 1
 */

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_time); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 596, __pyx_L1_error)

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_v_self->__pyx_base.nodes->__pyx_vtab)->find_time(__pyx_v_self->__pyx_base.nodes, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 596, __pyx_L1_error)

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (likely(__pyx_t_2 != Py_None)) {

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject* sequence = __pyx_t_2;

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(size != 2)) {

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __PYX_ERR(0, 596, __pyx_L1_error)

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(__pyx_t_3);

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(__pyx_t_4);

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #else

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 596, __pyx_L1_error)

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 596, __pyx_L1_error)

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(0, 596, __pyx_L1_error)

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v__ = __pyx_t_3;

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = 0;

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_i = __pyx_t_4;

#line 596 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":597
 *         """
 *         _, i = self.nodes.find_time(time)
 *         if self.nodes[i].time < time:             # <<<<<<<<<<<<<<
 *             i += 1
 *         return LCMSFeature(self.nodes[:i]), LCMSFeature(self.nodes[i:])
 */

#line 597 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self->__pyx_base.nodes), __pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 597, __pyx_L1_error)

#line 597 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 597 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_time); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 597, __pyx_L1_error)

#line 597 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 597 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 597 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_4, __pyx_v_time, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 597, __pyx_L1_error)

#line 597 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 597 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 597, __pyx_L1_error)

#line 597 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 597 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":598
 *         _, i = self.nodes.find_time(time)
 *         if self.nodes[i].time < time:
 *             i += 1             # <<<<<<<<<<<<<<
 *         return LCMSFeature(self.nodes[:i]), LCMSFeature(self.nodes[i:])
 * 
 */

#line 598 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_v_i, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 598, __pyx_L1_error)

#line 598 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 598 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF_SET(__pyx_v_i, __pyx_t_2);

#line 598 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":597
 *         """
 *         _, i = self.nodes.find_time(time)
 *         if self.nodes[i].time < time:             # <<<<<<<<<<<<<<
 *             i += 1
 *         return LCMSFeature(self.nodes[:i]), LCMSFeature(self.nodes[i:])
 */

#line 597 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":599
 *         if self.nodes[i].time < time:
 *             i += 1
 *         return LCMSFeature(self.nodes[:i]), LCMSFeature(self.nodes[i:])             # <<<<<<<<<<<<<<
 * 
 *     def between_times(self, start, end):
 */

#line 599 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 599 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetSlice(((PyObject *)__pyx_v_self->__pyx_base.nodes), 0, 0, NULL, &__pyx_v_i, NULL, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 599, __pyx_L1_error)

#line 599 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 599 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature), __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 599, __pyx_L1_error)

#line 599 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 599 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 599 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetSlice(((PyObject *)__pyx_v_self->__pyx_base.nodes), 0, 0, &__pyx_v_i, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 599, __pyx_L1_error)

#line 599 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 599 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 599, __pyx_L1_error)

#line 599 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 599 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 599 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 599, __pyx_L1_error)

#line 599 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 599 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_4);

#line 599 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4);

#line 599 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 599 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);

#line 599 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = 0;

#line 599 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = 0;

#line 599 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_2;

#line 599 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = 0;

#line 599 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":580
 *         return x
 * 
 *     def split_at(self, time):             # <<<<<<<<<<<<<<
 *         """Split this feature at a specific time point into two pieces, before
 *         and after that point in time.
 */

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.split_at", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v__);

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_i);

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 580 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":601
 *         return LCMSFeature(self.nodes[:i]), LCMSFeature(self.nodes[i:])
 * 
 *     def between_times(self, start, end):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, j
 */

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_25between_times(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_24between_times[] = "LCMSFeature.between_times(self, start, end)";

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_25between_times(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_start = 0
#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyObject *__pyx_v_end = 0
#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("between_times (wrapper)", 0);

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_start,&__pyx_n_s_end,0};

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject* values[2] = {0,0};

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(__pyx_kwds)) {

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      Py_ssize_t kw_args;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {
        case  2: 
#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0: break;
        default: 
#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0:

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_start)) != 0)) kw_args--;
        else 
#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  1:

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_end)) != 0)) kw_args--;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("between_times", 1, 2, 2, 1); 
#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__PYX_ERR(0, 601, __pyx_L3_error)

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(kw_args > 0)) {

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "between_times") < 0)) __PYX_ERR(0, 601, __pyx_L3_error)

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L5_argtuple_error;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_start = values[0];

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_end = values[1];

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("between_times", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__PYX_ERR(0, 601, __pyx_L3_error)

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_error:;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.between_times", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return NULL;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_24between_times(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), __pyx_v_start, __pyx_v_end);

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_24between_times(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_start, PyObject *__pyx_v_end) {
  size_t __pyx_v_i
#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_j
#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_node = 0
#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_1;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_3;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_4;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_5;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_7 = NULL;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_8 = NULL;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("between_times", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":605
 *             size_t i, j
 *             LCMSFeatureTreeNode node
 *         i = 0             # <<<<<<<<<<<<<<
 *         j = 0
 *         node = self._find_time(start, &i)
 */

#line 605 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_i = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":606
 *             LCMSFeatureTreeNode node
 *         i = 0
 *         j = 0             # <<<<<<<<<<<<<<
 *         node = self._find_time(start, &i)
 *         if node is not None and node.time < start:
 */

#line 606 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_j = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":607
 *         i = 0
 *         j = 0
 *         node = self._find_time(start, &i)             # <<<<<<<<<<<<<<
 *         if node is not None and node.time < start:
 *             i += 1
 */

#line 607 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_start); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 607, __pyx_L1_error)

#line 607 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._find_time(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_self), __pyx_t_1, (&__pyx_v_i))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 607, __pyx_L1_error)

#line 607 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 607 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_node = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_t_2);

#line 607 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":608
 *         j = 0
 *         node = self._find_time(start, &i)
 *         if node is not None and node.time < start:             # <<<<<<<<<<<<<<
 *             i += 1
 *         node = self._find_time(end, &j)
 */

#line 608 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = (((PyObject *)__pyx_v_node) != Py_None);

#line 608 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = (__pyx_t_4 != 0);

#line 608 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_5) {

#line 608 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {

#line 608 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = __pyx_t_5;

#line 608 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L4_bool_binop_done;

#line 608 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 608 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_node->__pyx_base.time); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 608, __pyx_L1_error)

#line 608 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 608 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = PyObject_RichCompare(__pyx_t_2, __pyx_v_start, Py_LT); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 608, __pyx_L1_error)

#line 608 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 608 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 608, __pyx_L1_error)

#line 608 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 608 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = __pyx_t_5;

#line 608 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L4_bool_binop_done:;

#line 608 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_3) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":609
 *         node = self._find_time(start, &i)
 *         if node is not None and node.time < start:
 *             i += 1             # <<<<<<<<<<<<<<
 *         node = self._find_time(end, &j)
 *         return self._copy_chunk(self.nodes[i:j + 1])
 */

#line 609 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_i = (__pyx_v_i + 1);

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":608
 *         j = 0
 *         node = self._find_time(start, &i)
 *         if node is not None and node.time < start:             # <<<<<<<<<<<<<<
 *             i += 1
 *         node = self._find_time(end, &j)
 */

#line 608 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":610
 *         if node is not None and node.time < start:
 *             i += 1
 *         node = self._find_time(end, &j)             # <<<<<<<<<<<<<<
 *         return self._copy_chunk(self.nodes[i:j + 1])
 * 
 */

#line 610 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_end); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 610, __pyx_L1_error)

#line 610 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._find_time(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_self), __pyx_t_1, (&__pyx_v_j))); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 610, __pyx_L1_error)

#line 610 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_6);

#line 610 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF_SET(__pyx_v_node, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_t_6));

#line 610 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":611
 *             i += 1
 *         node = self._find_time(end, &j)
 *         return self._copy_chunk(self.nodes[i:j + 1])             # <<<<<<<<<<<<<<
 * 
 *     cpdef list split_sparse(self, double delta_rt=1.):
 */

#line 611 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 611 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_copy_chunk); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 611, __pyx_L1_error)

#line 611 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 611 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_7 = __Pyx_PyObject_GetSlice(((PyObject *)__pyx_v_self->__pyx_base.nodes), __pyx_v_i, (__pyx_v_j + 1), NULL, NULL, NULL, 1, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 611, __pyx_L1_error)

#line 611 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_7);

#line 611 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_8 = NULL;

#line 611 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 611 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);

#line 611 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (likely(__pyx_t_8)) {

#line 611 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 611 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(__pyx_t_8);

#line 611 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(function);

#line 611 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF_SET(__pyx_t_2, function);

#line 611 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 611 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 611 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_8, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7);

#line 611 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 611 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 611 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 611, __pyx_L1_error)

#line 611 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_6);

#line 611 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 611 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_6;

#line 611 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = 0;

#line 611 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":601
 *         return LCMSFeature(self.nodes[:i]), LCMSFeature(self.nodes[i:])
 * 
 *     def between_times(self, start, end):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, j
 */

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_7);

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_8);

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.between_times", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_node);

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 601 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":613
 *         return self._copy_chunk(self.nodes[i:j + 1])
 * 
 *     cpdef list split_sparse(self, double delta_rt=1.):             # <<<<<<<<<<<<<<
 *         """Split this feature at any point where the distance from the
 *         time between one observed peak and the next exceeds `delta_rt`.
 */

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_27split_sparse(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_split_sparse(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_split_sparse *__pyx_optional_args) {

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_v_delta_rt = ((double)1.);
  PyObject *__pyx_v_chunks = 0
#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyObject *__pyx_v_current_chunk = 0
#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  double __pyx_v_last_rt
#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_node = 0
#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_i
#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_n
#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyObject *__pyx_v_x = NULL
#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyObject *__pyx_v_chunk = NULL
#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_6;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_7;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_8;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_9;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_10;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_11;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_ssize_t __pyx_t_12;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("split_sparse", 0);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_optional_args) {

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_delta_rt = __pyx_optional_args->delta_rt;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_split_sparse); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 613, __pyx_L1_error)

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_27split_sparse)) {

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_delta_rt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 613, __pyx_L1_error)

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_5 = NULL;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_5)) {

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_5);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_4, function);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 613, __pyx_L1_error)

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 613, __pyx_L1_error)

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = ((PyObject*)__pyx_t_2);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = 0;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":634
 *             size_t i, n
 * 
 *         chunks = []             # <<<<<<<<<<<<<<
 *         current_chunk = []
 *         n = self.get_size()
 */

#line 634 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 634, __pyx_L1_error)

#line 634 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 634 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_chunks = ((PyObject*)__pyx_t_1);

#line 634 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":635
 * 
 *         chunks = []
 *         current_chunk = []             # <<<<<<<<<<<<<<
 *         n = self.get_size()
 *         if n == 0:
 */

#line 635 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)

#line 635 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 635 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_current_chunk = ((PyObject*)__pyx_t_1);

#line 635 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":636
 *         chunks = []
 *         current_chunk = []
 *         n = self.get_size()             # <<<<<<<<<<<<<<
 *         if n == 0:
 *             return chunks
 */

#line 636 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_n = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.get_size(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_self));

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":637
 *         current_chunk = []
 *         n = self.get_size()
 *         if n == 0:             # <<<<<<<<<<<<<<
 *             return chunks
 * 
 */

#line 637 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = ((__pyx_v_n == 0) != 0);

#line 637 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_6) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":638
 *         n = self.get_size()
 *         if n == 0:
 *             return chunks             # <<<<<<<<<<<<<<
 * 
 *         last_rt = self.getitem(0).time
 */

#line 638 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 638 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(__pyx_v_chunks);

#line 638 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_r = __pyx_v_chunks;

#line 638 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":637
 *         current_chunk = []
 *         n = self.get_size()
 *         if n == 0:             # <<<<<<<<<<<<<<
 *             return chunks
 * 
 */

#line 637 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":640
 *             return chunks
 * 
 *         last_rt = self.getitem(0).time             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             node = self.getitem(i)
 */

#line 640 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->getitem(__pyx_v_self, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 640, __pyx_L1_error)

#line 640 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 640 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_7 = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_t_1)->__pyx_base.time;

#line 640 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 640 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_last_rt = __pyx_t_7;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":641
 * 
 *         last_rt = self.getitem(0).time
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             node = self.getitem(i)
 *             if (node.time - last_rt) > delta_rt:
 */

#line 641 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_8 = __pyx_v_n;

#line 641 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_9 = __pyx_t_8;

#line 641 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {

#line 641 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_i = __pyx_t_10;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":642
 *         last_rt = self.getitem(0).time
 *         for i in range(n):
 *             node = self.getitem(i)             # <<<<<<<<<<<<<<
 *             if (node.time - last_rt) > delta_rt:
 *                 x = self._copy_chunk(current_chunk)
 */

#line 642 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->getitem(__pyx_v_self, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 642, __pyx_L1_error)

#line 642 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 642 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_t_1));

#line 642 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":643
 *         for i in range(n):
 *             node = self.getitem(i)
 *             if (node.time - last_rt) > delta_rt:             # <<<<<<<<<<<<<<
 *                 x = self._copy_chunk(current_chunk)
 * 
 */

#line 643 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_6 = (((__pyx_v_node->__pyx_base.time - __pyx_v_last_rt) > __pyx_v_delta_rt) != 0);

#line 643 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_6) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":644
 *             node = self.getitem(i)
 *             if (node.time - last_rt) > delta_rt:
 *                 x = self._copy_chunk(current_chunk)             # <<<<<<<<<<<<<<
 * 
 *                 chunks.append(x)
 */

#line 644 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_copy_chunk); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 644, __pyx_L1_error)

#line 644 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_2);

#line 644 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_4 = NULL;

#line 644 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 644 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);

#line 644 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (likely(__pyx_t_4)) {

#line 644 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 644 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_INCREF(__pyx_t_4);

#line 644 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_INCREF(function);

#line 644 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_DECREF_SET(__pyx_t_2, function);

#line 644 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 644 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 644 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v_current_chunk) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_current_chunk);

#line 644 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 644 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 644, __pyx_L1_error)

#line 644 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 644 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 644 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_XDECREF_SET(__pyx_v_x, __pyx_t_1);

#line 644 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = 0;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":646
 *                 x = self._copy_chunk(current_chunk)
 * 
 *                 chunks.append(x)             # <<<<<<<<<<<<<<
 *                 current_chunk = []
 * 
 */

#line 646 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_chunks, __pyx_v_x); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 646, __pyx_L1_error)

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":647
 * 
 *                 chunks.append(x)
 *                 current_chunk = []             # <<<<<<<<<<<<<<
 * 
 *             last_rt = node.time
 */

#line 647 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 647, __pyx_L1_error)

#line 647 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 647 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF_SET(__pyx_v_current_chunk, ((PyObject*)__pyx_t_1));

#line 647 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = 0;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":643
 *         for i in range(n):
 *             node = self.getitem(i)
 *             if (node.time - last_rt) > delta_rt:             # <<<<<<<<<<<<<<
 *                 x = self._copy_chunk(current_chunk)
 * 
 */

#line 643 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":649
 *                 current_chunk = []
 * 
 *             last_rt = node.time             # <<<<<<<<<<<<<<
 *             current_chunk.append(node)
 * 
 */

#line 649 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_7 = __pyx_v_node->__pyx_base.time;

#line 649 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_last_rt = __pyx_t_7;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":650
 * 
 *             last_rt = node.time
 *             current_chunk.append(node)             # <<<<<<<<<<<<<<
 * 
 *         x = self._copy_chunk(current_chunk)
 */

#line 650 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_current_chunk, ((PyObject *)__pyx_v_node)); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 650, __pyx_L1_error)

#line 650 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":652
 *             current_chunk.append(node)
 * 
 *         x = self._copy_chunk(current_chunk)             # <<<<<<<<<<<<<<
 * 
 *         chunks.append(x)
 */

#line 652 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_copy_chunk); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 652, __pyx_L1_error)

#line 652 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 652 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = NULL;

#line 652 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 652 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);

#line 652 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (likely(__pyx_t_4)) {

#line 652 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 652 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(__pyx_t_4);

#line 652 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(function);

#line 652 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF_SET(__pyx_t_2, function);

#line 652 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 652 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 652 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v_current_chunk) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_current_chunk);

#line 652 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 652 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 652, __pyx_L1_error)

#line 652 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 652 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 652 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF_SET(__pyx_v_x, __pyx_t_1);

#line 652 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":654
 *         x = self._copy_chunk(current_chunk)
 * 
 *         chunks.append(x)             # <<<<<<<<<<<<<<
 *         for chunk in chunks:
 *             chunk.created_at = self.created_at
 */

#line 654 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_chunks, __pyx_v_x); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 654, __pyx_L1_error)

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":655
 * 
 *         chunks.append(x)
 *         for chunk in chunks:             # <<<<<<<<<<<<<<
 *             chunk.created_at = self.created_at
 * 
 */

#line 655 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_v_chunks; __Pyx_INCREF(__pyx_t_1); __pyx_t_12 = 0;

#line 655 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  for (;;) {

#line 655 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_12 >= PyList_GET_SIZE(__pyx_t_1)) break;

#line 655 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 655 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_12); __Pyx_INCREF(__pyx_t_2); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 655, __pyx_L1_error)

#line 655 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #else

#line 655 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 655, __pyx_L1_error)

#line 655 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 655 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 655 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF_SET(__pyx_v_chunk, __pyx_t_2);

#line 655 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":656
 *         chunks.append(x)
 *         for chunk in chunks:
 *             chunk.created_at = self.created_at             # <<<<<<<<<<<<<<
 * 
 *         return chunks
 */

#line 656 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = __pyx_v_self->created_at;

#line 656 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(__pyx_t_2);

#line 656 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_chunk, __pyx_n_s_created_at, __pyx_t_2) < 0) __PYX_ERR(0, 656, __pyx_L1_error)

#line 656 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":655
 * 
 *         chunks.append(x)
 *         for chunk in chunks:             # <<<<<<<<<<<<<<
 *             chunk.created_at = self.created_at
 * 
 */

#line 655 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 655 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":658
 *             chunk.created_at = self.created_at
 * 
 *         return chunks             # <<<<<<<<<<<<<<
 * 
 *     def truncate_before(self, time):
 */

#line 658 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 658 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_chunks);

#line 658 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_v_chunks;

#line 658 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":613
 *         return self._copy_chunk(self.nodes[i:j + 1])
 * 
 *     cpdef list split_sparse(self, double delta_rt=1.):             # <<<<<<<<<<<<<<
 *         """Split this feature at any point where the distance from the
 *         time between one observed peak and the next exceeds `delta_rt`.
 */

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.split_sparse", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_chunks);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_current_chunk);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_node);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_x);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_chunk);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_27split_sparse(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_26split_sparse[] = "LCMSFeature.split_sparse(self, double delta_rt=1.) -> list\nSplit this feature at any point where the distance from the\n        time between one observed peak and the next exceeds `delta_rt`.\n\n        Parameters\n        ----------\n        delta_rt : float\n            The maximum time between observed peaks to tolerate before\n            breaking the feature.\n\n        Returns\n        -------\n        list of :class:`LCMSFeature`\n            The feature components broken at the gaps.\n        ";

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_27split_sparse(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_delta_rt
#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("split_sparse (wrapper)", 0);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_delta_rt,0};

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject* values[1] = {0};

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(__pyx_kwds)) {

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      Py_ssize_t kw_args;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {
        case  1: 
#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0: break;
        default: 
#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0:

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (kw_args > 0) {

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_delta_rt);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (value) { values[0] = value; kw_args--; }

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(kw_args > 0)) {

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "split_sparse") < 0)) __PYX_ERR(0, 613, __pyx_L3_error)

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: 
#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  0: 
#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
break;
        default: 
#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (values[0]) {

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_delta_rt = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_delta_rt == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 613, __pyx_L3_error)

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_delta_rt = ((double)1.);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("split_sparse", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); 
#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__PYX_ERR(0, 613, __pyx_L3_error)

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_error:;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.split_sparse", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return NULL;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_26split_sparse(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), __pyx_v_delta_rt);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_26split_sparse(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, double __pyx_v_delta_rt) {

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_split_sparse __pyx_t_2;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("split_sparse", 0);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2.__pyx_n = 1;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2.delta_rt = __pyx_v_delta_rt;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature->split_sparse(__pyx_v_self, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 613, __pyx_L1_error)

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.split_sparse", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 613 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":660
 *         return chunks
 * 
 *     def truncate_before(self, time):             # <<<<<<<<<<<<<<
 *         _, i = self.nodes.find_time(time)
 *         if self.nodes[i].time < time:
 */

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_29truncate_before(PyObject *__pyx_v_self, PyObject *__pyx_v_time); /*proto*/

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_28truncate_before[] = "LCMSFeature.truncate_before(self, time)";

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_29truncate_before(PyObject *__pyx_v_self, PyObject *__pyx_v_time) {

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("truncate_before (wrapper)", 0);
  __pyx_r = 
#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_28truncate_before(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), ((PyObject *)__pyx_v_time));

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_28truncate_before(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_time) {
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL
#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyObject *__pyx_v_i = NULL
#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_1;

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_5;

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("truncate_before", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":661
 * 
 *     def truncate_before(self, time):
 *         _, i = self.nodes.find_time(time)             # <<<<<<<<<<<<<<
 *         if self.nodes[i].time < time:
 *             i += 1
 */

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_time); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 661, __pyx_L1_error)

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_v_self->__pyx_base.nodes->__pyx_vtab)->find_time(__pyx_v_self->__pyx_base.nodes, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 661, __pyx_L1_error)

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (likely(__pyx_t_2 != Py_None)) {

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject* sequence = __pyx_t_2;

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(size != 2)) {

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __PYX_ERR(0, 661, __pyx_L1_error)

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(__pyx_t_3);

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(__pyx_t_4);

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #else

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 661, __pyx_L1_error)

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 661, __pyx_L1_error)

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(0, 661, __pyx_L1_error)

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v__ = __pyx_t_3;

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = 0;

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_i = __pyx_t_4;

#line 661 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":662
 *     def truncate_before(self, time):
 *         _, i = self.nodes.find_time(time)
 *         if self.nodes[i].time < time:             # <<<<<<<<<<<<<<
 *             i += 1
 *         self.nodes = LCMSFeatureTreeList(self.nodes[i:])
 */

#line 662 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self->__pyx_base.nodes), __pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 662, __pyx_L1_error)

#line 662 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 662 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_time); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 662, __pyx_L1_error)

#line 662 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 662 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 662 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_4, __pyx_v_time, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 662, __pyx_L1_error)

#line 662 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 662 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 662, __pyx_L1_error)

#line 662 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 662 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":663
 *         _, i = self.nodes.find_time(time)
 *         if self.nodes[i].time < time:
 *             i += 1             # <<<<<<<<<<<<<<
 *         self.nodes = LCMSFeatureTreeList(self.nodes[i:])
 *         self._invalidate()
 */

#line 663 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_v_i, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 663, __pyx_L1_error)

#line 663 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 663 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF_SET(__pyx_v_i, __pyx_t_2);

#line 663 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":662
 *     def truncate_before(self, time):
 *         _, i = self.nodes.find_time(time)
 *         if self.nodes[i].time < time:             # <<<<<<<<<<<<<<
 *             i += 1
 *         self.nodes = LCMSFeatureTreeList(self.nodes[i:])
 */

#line 662 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":664
 *         if self.nodes[i].time < time:
 *             i += 1
 *         self.nodes = LCMSFeatureTreeList(self.nodes[i:])             # <<<<<<<<<<<<<<
 *         self._invalidate()
 * 
 */

#line 664 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetSlice(((PyObject *)__pyx_v_self->__pyx_base.nodes), 0, 0, &__pyx_v_i, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 664, __pyx_L1_error)

#line 664 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 664 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList), __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 664, __pyx_L1_error)

#line 664 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 664 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 664 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_4);

#line 664 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.nodes);

#line 664 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx_base.nodes));

#line 664 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->__pyx_base.nodes = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_t_4);

#line 664 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":665
 *             i += 1
 *         self.nodes = LCMSFeatureTreeList(self.nodes[i:])
 *         self._invalidate()             # <<<<<<<<<<<<<<
 * 
 *     def truncate_after(self, time):
 */

#line 665 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->_invalidate(__pyx_v_self, 0, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 665, __pyx_L1_error)

#line 665 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 665 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":660
 *         return chunks
 * 
 *     def truncate_before(self, time):             # <<<<<<<<<<<<<<
 *         _, i = self.nodes.find_time(time)
 *         if self.nodes[i].time < time:
 */

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.truncate_before", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v__);

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_i);

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 660 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":667
 *         self._invalidate()
 * 
 *     def truncate_after(self, time):             # <<<<<<<<<<<<<<
 *         _, i = self.nodes.find_time(time)
 *         if self.nodes[i].time < time:
 */

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_31truncate_after(PyObject *__pyx_v_self, PyObject *__pyx_v_time); /*proto*/

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_30truncate_after[] = "LCMSFeature.truncate_after(self, time)";

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_31truncate_after(PyObject *__pyx_v_self, PyObject *__pyx_v_time) {

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("truncate_after (wrapper)", 0);
  __pyx_r = 
#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_30truncate_after(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), ((PyObject *)__pyx_v_time));

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_30truncate_after(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_time) {
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL
#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyObject *__pyx_v_i = NULL
#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_1;

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_5;

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("truncate_after", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":668
 * 
 *     def truncate_after(self, time):
 *         _, i = self.nodes.find_time(time)             # <<<<<<<<<<<<<<
 *         if self.nodes[i].time < time:
 *             i += 1
 */

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_time); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 668, __pyx_L1_error)

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_v_self->__pyx_base.nodes->__pyx_vtab)->find_time(__pyx_v_self->__pyx_base.nodes, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 668, __pyx_L1_error)

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (likely(__pyx_t_2 != Py_None)) {

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject* sequence = __pyx_t_2;

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(size != 2)) {

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __PYX_ERR(0, 668, __pyx_L1_error)

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(__pyx_t_3);

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(__pyx_t_4);

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #else

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 668, __pyx_L1_error)

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 668, __pyx_L1_error)

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(0, 668, __pyx_L1_error)

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v__ = __pyx_t_3;

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = 0;

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_i = __pyx_t_4;

#line 668 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":669
 *     def truncate_after(self, time):
 *         _, i = self.nodes.find_time(time)
 *         if self.nodes[i].time < time:             # <<<<<<<<<<<<<<
 *             i += 1
 *         self.nodes = LCMSFeatureTreeList(self.nodes[:i])
 */

#line 669 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self->__pyx_base.nodes), __pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 669, __pyx_L1_error)

#line 669 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 669 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_time); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 669, __pyx_L1_error)

#line 669 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 669 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 669 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_4, __pyx_v_time, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 669, __pyx_L1_error)

#line 669 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 669 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 669, __pyx_L1_error)

#line 669 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 669 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":670
 *         _, i = self.nodes.find_time(time)
 *         if self.nodes[i].time < time:
 *             i += 1             # <<<<<<<<<<<<<<
 *         self.nodes = LCMSFeatureTreeList(self.nodes[:i])
 *         self._invalidate()
 */

#line 670 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_v_i, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 670, __pyx_L1_error)

#line 670 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 670 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF_SET(__pyx_v_i, __pyx_t_2);

#line 670 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":669
 *     def truncate_after(self, time):
 *         _, i = self.nodes.find_time(time)
 *         if self.nodes[i].time < time:             # <<<<<<<<<<<<<<
 *             i += 1
 *         self.nodes = LCMSFeatureTreeList(self.nodes[:i])
 */

#line 669 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":671
 *         if self.nodes[i].time < time:
 *             i += 1
 *         self.nodes = LCMSFeatureTreeList(self.nodes[:i])             # <<<<<<<<<<<<<<
 *         self._invalidate()
 * 
 */

#line 671 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetSlice(((PyObject *)__pyx_v_self->__pyx_base.nodes), 0, 0, NULL, &__pyx_v_i, NULL, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 671, __pyx_L1_error)

#line 671 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 671 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList), __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 671, __pyx_L1_error)

#line 671 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 671 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 671 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_4);

#line 671 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.nodes);

#line 671 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx_base.nodes));

#line 671 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->__pyx_base.nodes = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_t_4);

#line 671 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":672
 *             i += 1
 *         self.nodes = LCMSFeatureTreeList(self.nodes[:i])
 *         self._invalidate()             # <<<<<<<<<<<<<<
 * 
 *     cpdef LCMSFeature clone(self, deep=False, cls=None):
 */

#line 672 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->_invalidate(__pyx_v_self, 0, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 672, __pyx_L1_error)

#line 672 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 672 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":667
 *         self._invalidate()
 * 
 *     def truncate_after(self, time):             # <<<<<<<<<<<<<<
 *         _, i = self.nodes.find_time(time)
 *         if self.nodes[i].time < time:
 */

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.truncate_after", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v__);

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_i);

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 667 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":674
 *         self._invalidate()
 * 
 *     cpdef LCMSFeature clone(self, deep=False, cls=None):             # <<<<<<<<<<<<<<
 *         if cls is None:
 *             cls = self.__class__
 */

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_33clone(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_clone(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_clone *__pyx_optional_args) {

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_v_deep = ((PyObject *)Py_False);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_v_cls = ((PyObject *)Py_None);
  PyObject *__pyx_v_c = NULL
#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_r = NULL;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_5;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_7;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_8;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_9 = NULL;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_10 = NULL;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("clone", 0);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_optional_args) {

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_optional_args->__pyx_n > 0) {

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_deep = __pyx_optional_args->deep;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (__pyx_optional_args->__pyx_n > 1) {

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_v_cls = __pyx_optional_args->cls;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_cls);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_clone); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 674, __pyx_L1_error)

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_33clone)) {

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_4 = NULL;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_5 = 0;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_4)) {

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __pyx_t_5 = 1;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #if CYTHON_FAST_PYCALL

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (PyFunction_Check(__pyx_t_3)) {

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_deep, __pyx_v_cls};

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 674, __pyx_L1_error)

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        } else

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #endif

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #if CYTHON_FAST_PYCCALL

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_deep, __pyx_v_cls};

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 674, __pyx_L1_error)

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        } else

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #endif

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        {

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 674, __pyx_L1_error)

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_GOTREF(__pyx_t_6);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (__pyx_t_4) {

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_INCREF(__pyx_v_deep);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_GIVEREF(__pyx_v_deep);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_deep);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_INCREF(__pyx_v_cls);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_GIVEREF(__pyx_v_cls);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_cls);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 674, __pyx_L1_error)

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature))))) __PYX_ERR(0, 674, __pyx_L1_error)

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_2);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = 0;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":675
 * 
 *     cpdef LCMSFeature clone(self, deep=False, cls=None):
 *         if cls is None:             # <<<<<<<<<<<<<<
 *             cls = self.__class__
 *         c = cls(self.nodes.clone(deep=deep), list(self.adducts), list(self.used_as_adduct))
 */

#line 675 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_7 = (__pyx_v_cls == Py_None);

#line 675 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_8 = (__pyx_t_7 != 0);

#line 675 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_8) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":676
 *     cpdef LCMSFeature clone(self, deep=False, cls=None):
 *         if cls is None:
 *             cls = self.__class__             # <<<<<<<<<<<<<<
 *         c = cls(self.nodes.clone(deep=deep), list(self.adducts), list(self.used_as_adduct))
 *         c.feature_id = self.feature_id
 */

#line 676 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 676, __pyx_L1_error)

#line 676 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 676 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF_SET(__pyx_v_cls, __pyx_t_1);

#line 676 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":675
 * 
 *     cpdef LCMSFeature clone(self, deep=False, cls=None):
 *         if cls is None:             # <<<<<<<<<<<<<<
 *             cls = self.__class__
 *         c = cls(self.nodes.clone(deep=deep), list(self.adducts), list(self.used_as_adduct))
 */

#line 675 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":677
 *         if cls is None:
 *             cls = self.__class__
 *         c = cls(self.nodes.clone(deep=deep), list(self.adducts), list(self.used_as_adduct))             # <<<<<<<<<<<<<<
 *         c.feature_id = self.feature_id
 *         c.created_at = self.created_at
 */

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base.nodes), __pyx_n_s_clone); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 677, __pyx_L1_error)

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 677, __pyx_L1_error)

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_deep, __pyx_v_deep) < 0) __PYX_ERR(0, 677, __pyx_L1_error)

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 677, __pyx_L1_error)

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_6);

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = PySequence_List(__pyx_v_self->adducts); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 677, __pyx_L1_error)

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = PySequence_List(__pyx_v_self->used_as_adduct); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 677, __pyx_L1_error)

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_cls);
  __pyx_t_4 = __pyx_v_cls; 
#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_9 = NULL;

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = 0;

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (likely(__pyx_t_9)) {

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(__pyx_t_9);

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(function);

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF_SET(__pyx_t_4, function);

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_5 = 1;

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #if CYTHON_FAST_PYCALL

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (PyFunction_Check(__pyx_t_4)) {

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject *__pyx_temp[4] = {__pyx_t_9, __pyx_t_6, __pyx_t_3, __pyx_t_2};

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 677, __pyx_L1_error)

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #endif

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #if CYTHON_FAST_PYCCALL

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject *__pyx_temp[4] = {__pyx_t_9, __pyx_t_6, __pyx_t_3, __pyx_t_2};

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 677, __pyx_L1_error)

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #endif

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_10 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 677, __pyx_L1_error)

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_10);

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_9) {

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9); __pyx_t_9 = NULL;

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GIVEREF(__pyx_t_6);

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_5, __pyx_t_6);

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GIVEREF(__pyx_t_3);

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_5, __pyx_t_3);

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GIVEREF(__pyx_t_2);

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_5, __pyx_t_2);

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_6 = 0;

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = 0;

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = 0;

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 677, __pyx_L1_error)

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_c = __pyx_t_1;

#line 677 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":678
 *             cls = self.__class__
 *         c = cls(self.nodes.clone(deep=deep), list(self.adducts), list(self.used_as_adduct))
 *         c.feature_id = self.feature_id             # <<<<<<<<<<<<<<
 *         c.created_at = self.created_at
 *         return c
 */

#line 678 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_v_self->feature_id;

#line 678 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 678 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_c, __pyx_n_s_feature_id, __pyx_t_1) < 0) __PYX_ERR(0, 678, __pyx_L1_error)

#line 678 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":679
 *         c = cls(self.nodes.clone(deep=deep), list(self.adducts), list(self.used_as_adduct))
 *         c.feature_id = self.feature_id
 *         c.created_at = self.created_at             # <<<<<<<<<<<<<<
 *         return c
 * 
 */

#line 679 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_v_self->created_at;

#line 679 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 679 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_c, __pyx_n_s_created_at, __pyx_t_1) < 0) __PYX_ERR(0, 679, __pyx_L1_error)

#line 679 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":680
 *         c.feature_id = self.feature_id
 *         c.created_at = self.created_at
 *         return c             # <<<<<<<<<<<<<<
 * 
 *     cpdef insert_node(self, LCMSFeatureTreeNode node):
 */

#line 680 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 680 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (!(likely(((__pyx_v_c) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_c, __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature))))) __PYX_ERR(0, 680, __pyx_L1_error)

#line 680 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_c);

#line 680 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_c);

#line 680 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":674
 *         self._invalidate()
 * 
 *     cpdef LCMSFeature clone(self, deep=False, cls=None):             # <<<<<<<<<<<<<<
 *         if cls is None:
 *             cls = self.__class__
 */

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_9);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_10);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_c);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_cls);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_33clone(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_32clone[] = "LCMSFeature.clone(self, deep=False, cls=None) -> LCMSFeature";

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_33clone(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_deep = 0
#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyObject *__pyx_v_cls = 0
#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("clone (wrapper)", 0);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_deep,&__pyx_n_s_cls,0};

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject* values[2] = {0,0};

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    values[0] = ((PyObject *)Py_False);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    values[1] = ((PyObject *)Py_None);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(__pyx_kwds)) {

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      Py_ssize_t kw_args;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {
        case  2: 
#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0: break;
        default: 
#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0:

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (kw_args > 0) {

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_deep);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (value) { values[0] = value; kw_args--; }

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  1:

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (kw_args > 0) {

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cls);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (value) { values[1] = value; kw_args--; }

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(kw_args > 0)) {

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "clone") < 0)) __PYX_ERR(0, 674, __pyx_L3_error)

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: 
#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  0: 
#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
break;
        default: 
#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_deep = values[0];

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_cls = values[1];

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("clone", 0, 0, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__PYX_ERR(0, 674, __pyx_L3_error)

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_error:;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return NULL;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_32clone(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), __pyx_v_deep, __pyx_v_cls);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_32clone(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_deep, PyObject *__pyx_v_cls) {

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_clone __pyx_t_2;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("clone", 0);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2.__pyx_n = 2;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2.deep = __pyx_v_deep;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2.cls = __pyx_v_cls;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature->clone(__pyx_v_self, 1, &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 674, __pyx_L1_error)

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 674 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":682
 *         return c
 * 
 *     cpdef insert_node(self, LCMSFeatureTreeNode node):             # <<<<<<<<<<<<<<
 *         """Insert a fully formed :class:`LCMSFeatureTreeNode` into this feature
 *         at the appropriate locaton in the time sequence.
 */

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_35insert_node(PyObject *__pyx_v_self, PyObject *__pyx_v_node); /*proto*/

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_insert_node(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_node, int __pyx_skip_dispatch) {

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("insert_node", 0);

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_insert_node); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 682, __pyx_L1_error)

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_35insert_node)) {

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_4 = NULL;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_4)) {

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_node)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_node));

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 682, __pyx_L1_error)

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = __pyx_t_2;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = 0;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":694
 *             The node to add.
 *         """
 *         self._peak_averager._update(node.members)             # <<<<<<<<<<<<<<
 *         self.nodes.insert_node(node)
 *         self._invalidate()
 */

#line 694 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_v_node->__pyx_base.members;

#line 694 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 694 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_self->_peak_averager->__pyx_vtab)->_update(__pyx_v_self->_peak_averager, ((PyObject*)__pyx_t_1));

#line 694 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":695
 *         """
 *         self._peak_averager._update(node.members)
 *         self.nodes.insert_node(node)             # <<<<<<<<<<<<<<
 *         self._invalidate()
 * 
 */

#line 695 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base.nodes), __pyx_n_s_insert_node); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 695, __pyx_L1_error)

#line 695 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 695 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = NULL;

#line 695 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 695 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);

#line 695 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (likely(__pyx_t_3)) {

#line 695 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 695 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(__pyx_t_3);

#line 695 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(function);

#line 695 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF_SET(__pyx_t_2, function);

#line 695 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 695 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 695 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, ((PyObject *)__pyx_v_node)) : __Pyx_PyObject_CallOneArg(__pyx_t_2, ((PyObject *)__pyx_v_node));

#line 695 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 695 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 695, __pyx_L1_error)

#line 695 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 695 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 695 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":696
 *         self._peak_averager._update(node.members)
 *         self.nodes.insert_node(node)
 *         self._invalidate()             # <<<<<<<<<<<<<<
 * 
 *     cpdef insert(self, PeakBase peak, double time):
 */

#line 696 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->_invalidate(__pyx_v_self, 0, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 696, __pyx_L1_error)

#line 696 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 696 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":682
 *         return c
 * 
 *     cpdef insert_node(self, LCMSFeatureTreeNode node):             # <<<<<<<<<<<<<<
 *         """Insert a fully formed :class:`LCMSFeatureTreeNode` into this feature
 *         at the appropriate locaton in the time sequence.
 */

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.insert_node", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_35insert_node(PyObject *__pyx_v_self, PyObject *__pyx_v_node); /*proto*/

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_34insert_node[] = "LCMSFeature.insert_node(self, LCMSFeatureTreeNode node)\nInsert a fully formed :class:`LCMSFeatureTreeNode` into this feature\n        at the appropriate locaton in the time sequence.\n\n        The node's :attr:`members` attribute will update the peak averager and\n        triggers state invalidation.\n\n        Parameters\n        ----------\n        node : LCMSFeatureTreeNode\n            The node to add.\n        ";

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_35insert_node(PyObject *__pyx_v_self, PyObject *__pyx_v_node) {

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("insert_node (wrapper)", 0);

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_node), __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode, 1, "node", 0))) __PYX_ERR(0, 682, __pyx_L1_error)
  __pyx_r = 
#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_34insert_node(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_v_node));

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_34insert_node(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_node) {

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("insert_node", 0);

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_insert_node(__pyx_v_self, __pyx_v_node, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 682, __pyx_L1_error)

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.insert_node", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 682 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":698
 *         self._invalidate()
 * 
 *     cpdef insert(self, PeakBase peak, double time):             # <<<<<<<<<<<<<<
 *         """Insert a peak at a known time into this feature.
 * 
 */

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_37insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_insert(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak, double __pyx_v_time, int __pyx_skip_dispatch) {

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_6;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_7 = NULL;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("insert", 0);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_insert); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 698, __pyx_L1_error)

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_37insert)) {

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 698, __pyx_L1_error)

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_5 = NULL;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_6 = 0;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_5)) {

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_5);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_4, function);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __pyx_t_6 = 1;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #if CYTHON_FAST_PYCALL

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (PyFunction_Check(__pyx_t_4)) {

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_peak), __pyx_t_3};

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 698, __pyx_L1_error)

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        } else

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #endif

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #if CYTHON_FAST_PYCCALL

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_peak), __pyx_t_3};

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 698, __pyx_L1_error)

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        } else

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #endif

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        {

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 698, __pyx_L1_error)

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_GOTREF(__pyx_t_7);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (__pyx_t_5) {

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_INCREF(((PyObject *)__pyx_v_peak));

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_GIVEREF(((PyObject *)__pyx_v_peak));

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_peak));

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_GIVEREF(__pyx_t_3);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_3 = 0;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 698, __pyx_L1_error)

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = __pyx_t_2;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = 0;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":710
 *             The time point at which this peak was observed
 *         """
 *         self._peak_averager.add(peak)             # <<<<<<<<<<<<<<
 *         self.nodes.insert(time, [peak])
 *         self._invalidate()
 */

#line 710 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_self->_peak_averager->__pyx_vtab)->add(__pyx_v_self->_peak_averager, __pyx_v_peak, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 710, __pyx_L1_error)

#line 710 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 710 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":711
 *         """
 *         self._peak_averager.add(peak)
 *         self.nodes.insert(time, [peak])             # <<<<<<<<<<<<<<
 *         self._invalidate()
 * 
 */

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base.nodes), __pyx_n_s_insert); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 711, __pyx_L1_error)

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_time); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 711, __pyx_L1_error)

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_7 = PyList_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 711, __pyx_L1_error)

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_7);

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_peak));

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_peak));

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyList_SET_ITEM(__pyx_t_7, 0, ((PyObject *)__pyx_v_peak));

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = NULL;

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = 0;

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (likely(__pyx_t_3)) {

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(__pyx_t_3);

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(function);

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF_SET(__pyx_t_2, function);

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_6 = 1;

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #if CYTHON_FAST_PYCALL

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (PyFunction_Check(__pyx_t_2)) {

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_t_7};

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 711, __pyx_L1_error)

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #endif

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #if CYTHON_FAST_PYCCALL

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_4, __pyx_t_7};

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 711, __pyx_L1_error)

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #endif

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 711, __pyx_L1_error)

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_5);

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_3) {

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GIVEREF(__pyx_t_4);

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_t_4);

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GIVEREF(__pyx_t_7);

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_7);

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = 0;

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_7 = 0;

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 711, __pyx_L1_error)

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 711 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":712
 *         self._peak_averager.add(peak)
 *         self.nodes.insert(time, [peak])
 *         self._invalidate()             # <<<<<<<<<<<<<<
 * 
 *     def merge(self, other):
 */

#line 712 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->_invalidate(__pyx_v_self, 0, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 712, __pyx_L1_error)

#line 712 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 712 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":698
 *         self._invalidate()
 * 
 *     cpdef insert(self, PeakBase peak, double time):             # <<<<<<<<<<<<<<
 *         """Insert a peak at a known time into this feature.
 * 
 */

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_7);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_37insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_36insert[] = "LCMSFeature.insert(self, PeakBase peak, double time)\nInsert a peak at a known time into this feature.\n\n        The peak updates the peak averager and triggers state invalidation.\n\n        Parameters\n        ----------\n        peak : PeakBase\n            The a peak with known m/z and intensity\n        time : float\n            The time point at which this peak was observed\n        ";

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_37insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak = 0
#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  double __pyx_v_time
#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("insert (wrapper)", 0);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_peak,&__pyx_n_s_time,0};

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject* values[2] = {0,0};

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(__pyx_kwds)) {

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      Py_ssize_t kw_args;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {
        case  2: 
#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0: break;
        default: 
#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0:

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_peak)) != 0)) kw_args--;
        else 
#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  1:

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_time)) != 0)) kw_args--;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        else {
          __Pyx_RaiseArgtupleInvalid("insert", 1, 2, 2, 1); 
#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__PYX_ERR(0, 698, __pyx_L3_error)

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(kw_args > 0)) {

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "insert") < 0)) __PYX_ERR(0, 698, __pyx_L3_error)

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L5_argtuple_error;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_peak = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)values[0]);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_time = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_time == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 698, __pyx_L3_error)

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("insert", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); 
#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__PYX_ERR(0, 698, __pyx_L3_error)

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_error:;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return NULL;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L4_argument_unpacking_done:;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase, 1, "peak", 0))) __PYX_ERR(0, 698, __pyx_L1_error)
  __pyx_r = 
#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_36insert(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), __pyx_v_peak, __pyx_v_time);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_36insert(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak, double __pyx_v_time) {

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("insert", 0);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_insert(__pyx_v_self, __pyx_v_peak, __pyx_v_time, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 698, __pyx_L1_error)

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 698 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":714
 *         self._invalidate()
 * 
 *     def merge(self, other):             # <<<<<<<<<<<<<<
 *         """Create a copy of this feature and copies all nodes from `other` into it,
 *         combining information from both features in a new instance.
 */

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_39merge(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_38merge[] = "LCMSFeature.merge(self, other)\nCreate a copy of this feature and copies all nodes from `other` into it,\n        combining information from both features in a new instance.\n\n        Parameters\n        ----------\n        other : FeatureBase\n            Another feature whose nodes (which are distinct from this feature's)\n            will be copied into a new copy of this feature.\n\n        Returns\n        -------\n        LCMSFeature\n        ";

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_39merge(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("merge (wrapper)", 0);
  __pyx_r = 
#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_38merge(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), ((PyObject *)__pyx_v_other));

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_38merge(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_other) {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_new = NULL
#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyObject *__pyx_v_node = NULL
#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_ssize_t __pyx_t_3;

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*__pyx_t_4)(PyObject *);

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("merge", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":728
 *         LCMSFeature
 *         """
 *         new = self.clone()             # <<<<<<<<<<<<<<
 *         for node in other.nodes:
 *             node = node.clone()
 */

#line 728 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->clone(__pyx_v_self, 0, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 728, __pyx_L1_error)

#line 728 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 728 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_new = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_t_1);

#line 728 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":729
 *         """
 *         new = self.clone()
 *         for node in other.nodes:             # <<<<<<<<<<<<<<
 *             node = node.clone()
 *             new.insert_node(node)
 */

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_nodes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 729, __pyx_L1_error)

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = NULL;

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {
    __pyx_t_3 = -1; 
#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 729, __pyx_L1_error)

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 729, __pyx_L1_error)

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  for (;;) {

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (likely(!__pyx_t_4)) {

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (likely(PyList_CheckExact(__pyx_t_2))) {

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 729, __pyx_L1_error)

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #else

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 729, __pyx_L1_error)

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_1);

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #endif

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      } else {

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 729, __pyx_L1_error)

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #else

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 729, __pyx_L1_error)

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_1);

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #endif

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }
    } else 
#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
{

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(!__pyx_t_1)) {

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        PyObject* exc_type = PyErr_Occurred();

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (exc_type) {

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          else __PYX_ERR(0, 729, __pyx_L1_error)

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        break;

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF_SET(__pyx_v_node, __pyx_t_1);

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":730
 *         new = self.clone()
 *         for node in other.nodes:
 *             node = node.clone()             # <<<<<<<<<<<<<<
 *             new.insert_node(node)
 *         new.created_at = "merge"
 */

#line 730 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_node, __pyx_n_s_clone); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 730, __pyx_L1_error)

#line 730 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_5);

#line 730 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_6 = NULL;

#line 730 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {

#line 730 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);

#line 730 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (likely(__pyx_t_6)) {

#line 730 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);

#line 730 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_6);

#line 730 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(function);

#line 730 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF_SET(__pyx_t_5, function);

#line 730 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 730 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 730 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_5);

#line 730 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 730 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)

#line 730 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 730 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 730 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF_SET(__pyx_v_node, __pyx_t_1);

#line 730 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":731
 *         for node in other.nodes:
 *             node = node.clone()
 *             new.insert_node(node)             # <<<<<<<<<<<<<<
 *         new.created_at = "merge"
 *         return new
 */

#line 731 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (!(likely(((__pyx_v_node) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_node, __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode))))) __PYX_ERR(0, 731, __pyx_L1_error)

#line 731 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_new->__pyx_base.__pyx_vtab)->insert_node(__pyx_v_new, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_v_node), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 731, __pyx_L1_error)

#line 731 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 731 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":729
 *         """
 *         new = self.clone()
 *         for node in other.nodes:             # <<<<<<<<<<<<<<
 *             node = node.clone()
 *             new.insert_node(node)
 */

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 729 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":732
 *             node = node.clone()
 *             new.insert_node(node)
 *         new.created_at = "merge"             # <<<<<<<<<<<<<<
 *         return new
 * 
 */

#line 732 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_n_s_merge);

#line 732 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_n_s_merge);

#line 732 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_new->created_at);

#line 732 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_v_new->created_at);

#line 732 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_new->created_at = __pyx_n_s_merge;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":733
 *             new.insert_node(node)
 *         new.created_at = "merge"
 *         return new             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 733 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 733 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_new));

#line 733 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = ((PyObject *)__pyx_v_new);

#line 733 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":714
 *         self._invalidate()
 * 
 *     def merge(self, other):             # <<<<<<<<<<<<<<
 *         """Create a copy of this feature and copies all nodes from `other` into it,
 *         combining information from both features in a new instance.
 */

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.merge", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_new);

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_node);

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 714 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":736
 * 
 *     @property
 *     def apex_time(self):             # <<<<<<<<<<<<<<
 *         rt, intensity = self.as_arrays()
 *         return rt[np.argmax(intensity)]
 */

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_9apex_time_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_9apex_time_1__get__(PyObject *__pyx_v_self) {

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_9apex_time___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_9apex_time___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {
  PyObject *__pyx_v_rt = NULL
#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyObject *__pyx_v_intensity = NULL
#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*__pyx_t_5)(PyObject *);

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":737
 *     @property
 *     def apex_time(self):
 *         rt, intensity = self.as_arrays()             # <<<<<<<<<<<<<<
 *         return rt[np.argmax(intensity)]
 * 
 */

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_as_arrays); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 737, __pyx_L1_error)

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = NULL;

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (likely(__pyx_t_3)) {

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(__pyx_t_3);

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(function);

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF_SET(__pyx_t_2, function);

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 737, __pyx_L1_error)

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject* sequence = __pyx_t_1;

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(size != 2)) {

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __PYX_ERR(0, 737, __pyx_L1_error)

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (likely(PyTuple_CheckExact(sequence))) {

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(__pyx_t_2);

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(__pyx_t_3);

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #else

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 737, __pyx_L1_error)

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 737, __pyx_L1_error)

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    Py_ssize_t index = -1;

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 737, __pyx_L1_error)

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_4);

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) 
#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L3_unpacking_failed;

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) 
#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L3_unpacking_failed;

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(0, 737, __pyx_L1_error)

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = NULL;

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L4_unpacking_done;

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L3_unpacking_failed:;

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = NULL;

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __PYX_ERR(0, 737, __pyx_L1_error)

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L4_unpacking_done:;

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_rt = __pyx_t_2;

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = 0;

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_intensity = __pyx_t_3;

#line 737 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":738
 *     def apex_time(self):
 *         rt, intensity = self.as_arrays()
 *         return rt[np.argmax(intensity)]             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint _eq(self, other):
 */

#line 738 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 738 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 738, __pyx_L1_error)

#line 738 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 738 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_argmax); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 738, __pyx_L1_error)

#line 738 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 738 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 738 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = NULL;

#line 738 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {

#line 738 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);

#line 738 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (likely(__pyx_t_3)) {

#line 738 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 738 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(__pyx_t_3);

#line 738 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(function);

#line 738 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF_SET(__pyx_t_2, function);

#line 738 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 738 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 738 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_intensity) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_intensity);

#line 738 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 738 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 738, __pyx_L1_error)

#line 738 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 738 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 738 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_rt, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 738, __pyx_L1_error)

#line 738 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 738 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 738 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_2;

#line 738 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = 0;

#line 738 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":736
 * 
 *     @property
 *     def apex_time(self):             # <<<<<<<<<<<<<<
 *         rt, intensity = self.as_arrays()
 *         return rt[np.argmax(intensity)]
 */

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.apex_time.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_rt);

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_intensity);

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 736 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":740
 *         return rt[np.argmax(intensity)]
 * 
 *     cpdef bint _eq(self, other):             # <<<<<<<<<<<<<<
 *         if other is None:
 *             return False
 */

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_41_eq(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__eq(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_a = NULL
#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyObject *__pyx_v_b = NULL
#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_r;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_5;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_6;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_ssize_t __pyx_t_7;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_ssize_t __pyx_t_8;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*__pyx_t_9)(PyObject *);

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_10 = NULL;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*__pyx_t_11)(PyObject *);

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_eq", 0);

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_eq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 740, __pyx_L1_error)

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_41_eq)) {

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_4 = NULL;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_4)) {

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_other);

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 740, __pyx_L1_error)

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 740, __pyx_L1_error)

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = __pyx_t_5;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":741
 * 
 *     cpdef bint _eq(self, other):
 *         if other is None:             # <<<<<<<<<<<<<<
 *             return False
 *         if len(self) != len(other):
 */

#line 741 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = (__pyx_v_other == Py_None);

#line 741 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = (__pyx_t_5 != 0);

#line 741 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_6) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":742
 *     cpdef bint _eq(self, other):
 *         if other is None:
 *             return False             # <<<<<<<<<<<<<<
 *         if len(self) != len(other):
 *             return False
 */

#line 742 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_r = 0;

#line 742 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":741
 * 
 *     cpdef bint _eq(self, other):
 *         if other is None:             # <<<<<<<<<<<<<<
 *             return False
 *         if len(self) != len(other):
 */

#line 741 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":743
 *         if other is None:
 *             return False
 *         if len(self) != len(other):             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */

#line 743 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_7 = PyObject_Length(((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 743, __pyx_L1_error)

#line 743 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_8 = PyObject_Length(__pyx_v_other); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 743, __pyx_L1_error)

#line 743 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = ((__pyx_t_7 != __pyx_t_8) != 0);

#line 743 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_6) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":744
 *             return False
 *         if len(self) != len(other):
 *             return False             # <<<<<<<<<<<<<<
 *         else:
 *             if abs(self.start_time - other.start_time) > 1e-4:
 */

#line 744 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_r = 0;

#line 744 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":743
 *         if other is None:
 *             return False
 *         if len(self) != len(other):             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */

#line 743 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":746
 *             return False
 *         else:
 *             if abs(self.start_time - other.start_time) > 1e-4:             # <<<<<<<<<<<<<<
 *                 return False
 *             elif abs(self.end_time - other.end_time) > 1e-4:
 */

#line 746 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /*else*/ {

#line 746 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_start_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 746, __pyx_L1_error)

#line 746 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 746 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_start_time); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 746, __pyx_L1_error)

#line 746 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 746 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 746, __pyx_L1_error)

#line 746 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 746 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 746 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 746 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = __Pyx_PyNumber_Absolute(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 746, __pyx_L1_error)

#line 746 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 746 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 746 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_float_1eneg_4, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 746, __pyx_L1_error)

#line 746 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 746 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 746, __pyx_L1_error)

#line 746 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 746 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_6) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":747
 *         else:
 *             if abs(self.start_time - other.start_time) > 1e-4:
 *                 return False             # <<<<<<<<<<<<<<
 *             elif abs(self.end_time - other.end_time) > 1e-4:
 *                 return False
 */

#line 747 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_r = 0;

#line 747 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L0;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":746
 *             return False
 *         else:
 *             if abs(self.start_time - other.start_time) > 1e-4:             # <<<<<<<<<<<<<<
 *                 return False
 *             elif abs(self.end_time - other.end_time) > 1e-4:
 */

#line 746 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":748
 *             if abs(self.start_time - other.start_time) > 1e-4:
 *                 return False
 *             elif abs(self.end_time - other.end_time) > 1e-4:             # <<<<<<<<<<<<<<
 *                 return False
 *             else:
 */

#line 748 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_end_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 748, __pyx_L1_error)

#line 748 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_3);

#line 748 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_end_time); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 748, __pyx_L1_error)

#line 748 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 748 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = PyNumber_Subtract(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 748, __pyx_L1_error)

#line 748 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 748 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 748 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 748 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = __Pyx_PyNumber_Absolute(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 748, __pyx_L1_error)

#line 748 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 748 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 748 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = PyObject_RichCompare(__pyx_t_2, __pyx_float_1eneg_4, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 748, __pyx_L1_error)

#line 748 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 748 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 748, __pyx_L1_error)

#line 748 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 748 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_6) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":749
 *                 return False
 *             elif abs(self.end_time - other.end_time) > 1e-4:
 *                 return False             # <<<<<<<<<<<<<<
 *             else:
 *                 for a, b in zip(self, other):
 */

#line 749 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_r = 0;

#line 749 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L0;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":748
 *             if abs(self.start_time - other.start_time) > 1e-4:
 *                 return False
 *             elif abs(self.end_time - other.end_time) > 1e-4:             # <<<<<<<<<<<<<<
 *                 return False
 *             else:
 */

#line 748 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":751
 *                 return False
 *             else:
 *                 for a, b in zip(self, other):             # <<<<<<<<<<<<<<
 *                     if a != b:
 *                         return False
 */

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    /*else*/ {

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 751, __pyx_L1_error)

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GIVEREF(((PyObject *)__pyx_v_self));

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self));

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(__pyx_v_other);

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GIVEREF(__pyx_v_other);

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_other);

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 751, __pyx_L1_error)

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_2);

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_8 = 0;

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_9 = NULL;

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      } else {
        __pyx_t_8 = -1; 
#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 751, __pyx_L1_error)

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_1);

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_9 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 751, __pyx_L1_error)

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      for (;;) {

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (likely(!__pyx_t_9)) {

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(PyList_CheckExact(__pyx_t_1))) {

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_1)) break;

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 751, __pyx_L1_error)

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            #else

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 751, __pyx_L1_error)

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_GOTREF(__pyx_t_2);

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            #endif

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          } else {

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_1)) break;

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 751, __pyx_L1_error)

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            #else

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 751, __pyx_L1_error)

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_GOTREF(__pyx_t_2);

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            #endif

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }
        } else 
#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
{

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_2 = __pyx_t_9(__pyx_t_1);

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (unlikely(!__pyx_t_2)) {

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* exc_type = PyErr_Occurred();

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            if (exc_type) {

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
              else __PYX_ERR(0, 751, __pyx_L1_error)

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            }

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            break;

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_GOTREF(__pyx_t_2);

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyObject* sequence = __pyx_t_2;

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (unlikely(size != 2)) {

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __PYX_ERR(0, 751, __pyx_L1_error)

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(PyTuple_CheckExact(sequence))) {

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          } else {

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_INCREF(__pyx_t_3);

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_INCREF(__pyx_t_4);

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          #else

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 751, __pyx_L1_error)

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_GOTREF(__pyx_t_3);

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 751, __pyx_L1_error)

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_GOTREF(__pyx_t_4);

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          #endif

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        } else {

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          Py_ssize_t index = -1;

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_10 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 751, __pyx_L1_error)

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_GOTREF(__pyx_t_10);

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_11 = Py_TYPE(__pyx_t_10)->tp_iternext;
          index = 0; __pyx_t_3 = __pyx_t_11(__pyx_t_10); if (unlikely(!__pyx_t_3)) 
#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L8_unpacking_failed;

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_GOTREF(__pyx_t_3);
          index = 1; __pyx_t_4 = __pyx_t_11(__pyx_t_10); if (unlikely(!__pyx_t_4)) 
#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L8_unpacking_failed;

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_GOTREF(__pyx_t_4);

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_10), 2) < 0) __PYX_ERR(0, 751, __pyx_L1_error)

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_11 = NULL;

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          goto __pyx_L9_unpacking_done;

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_L8_unpacking_failed:;

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_11 = NULL;

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __PYX_ERR(0, 751, __pyx_L1_error)

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_L9_unpacking_done:;

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF_SET(__pyx_v_a, __pyx_t_3);

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_3 = 0;

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF_SET(__pyx_v_b, __pyx_t_4);

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_4 = 0;

        /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":752
 *             else:
 *                 for a, b in zip(self, other):
 *                     if a != b:             # <<<<<<<<<<<<<<
 *                         return False
 *                 return True
 */

#line 752 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = PyObject_RichCompare(__pyx_v_a, __pyx_v_b, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 752, __pyx_L1_error)

#line 752 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 752, __pyx_L1_error)

#line 752 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 752 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (__pyx_t_6) {

          /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":753
 *                 for a, b in zip(self, other):
 *                     if a != b:
 *                         return False             # <<<<<<<<<<<<<<
 *                 return True
 * 
 */

#line 753 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_r = 0;

#line 753 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 753 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          goto __pyx_L0;

          /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":752
 *             else:
 *                 for a, b in zip(self, other):
 *                     if a != b:             # <<<<<<<<<<<<<<
 *                         return False
 *                 return True
 */

#line 752 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

        /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":751
 *                 return False
 *             else:
 *                 for a, b in zip(self, other):             # <<<<<<<<<<<<<<
 *                     if a != b:
 *                         return False
 */

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 751 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":754
 *                     if a != b:
 *                         return False
 *                 return True             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint _ne(self, other):
 */

#line 754 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_r = 1;

#line 754 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L0;

#line 754 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 754 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":740
 *         return rt[np.argmax(intensity)]
 * 
 *     cpdef bint _eq(self, other):             # <<<<<<<<<<<<<<
 *         if other is None:
 *             return False
 */

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_10);

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature._eq", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_a);

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_b);

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_41_eq(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_40_eq[] = "LCMSFeature._eq(self, other) -> bool";

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_41_eq(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_eq (wrapper)", 0);
  __pyx_r = 
#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_40_eq(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), ((PyObject *)__pyx_v_other));

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_40_eq(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_other) {

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_eq", 0);

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__eq(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 740, __pyx_L1_error)

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature._eq", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 740 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":756
 *                 return True
 * 
 *     cpdef bint _ne(self, other):             # <<<<<<<<<<<<<<
 *         return not self._eq(other)
 * 
 */

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_43_ne(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__ne(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_skip_dispatch) {

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_r;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_5;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_ne", 0);

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_ne); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 756, __pyx_L1_error)

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_43_ne)) {

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_4 = NULL;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_4)) {

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_other);

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 756, __pyx_L1_error)

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 756, __pyx_L1_error)

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = __pyx_t_5;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":757
 * 
 *     cpdef bint _ne(self, other):
 *         return not self._eq(other)             # <<<<<<<<<<<<<<
 * 
 *     def __richcmp__(self, other, int code):
 */

#line 757 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = (!(((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->_eq(__pyx_v_self, __pyx_v_other, 0) != 0));

#line 757 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":756
 *                 return True
 * 
 *     cpdef bint _ne(self, other):             # <<<<<<<<<<<<<<
 *         return not self._eq(other)
 * 
 */

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature._ne", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_43_ne(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_42_ne[] = "LCMSFeature._ne(self, other) -> bool";

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_43_ne(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_ne (wrapper)", 0);
  __pyx_r = 
#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_42_ne(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), ((PyObject *)__pyx_v_other));

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_42_ne(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_other) {

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_ne", 0);

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__ne(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 756, __pyx_L1_error)

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature._ne", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 756 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":759
 *         return not self._eq(other)
 * 
 *     def __richcmp__(self, other, int code):             # <<<<<<<<<<<<<<
 *         if code == 2:
 *             return self._eq(other)
 */

#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_45__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code); /*proto*/

#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_45__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code) {

#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_r = 
#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_44__richcmp__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), ((PyObject *)__pyx_v_other), ((int)__pyx_v_code));

#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_44__richcmp__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code) {

#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":760
 * 
 *     def __richcmp__(self, other, int code):
 *         if code == 2:             # <<<<<<<<<<<<<<
 *             return self._eq(other)
 *         elif code == 3:
 */

#line 760 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  switch (__pyx_v_code) {

#line 760 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    case 2:

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":761
 *     def __richcmp__(self, other, int code):
 *         if code == 2:
 *             return self._eq(other)             # <<<<<<<<<<<<<<
 *         elif code == 3:
 *             return self._ne(other)
 */

#line 761 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 761 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __Pyx_PyBool_FromLong(((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->_eq(__pyx_v_self, __pyx_v_other, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 761, __pyx_L1_error)

#line 761 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 761 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_r = __pyx_t_1;

#line 761 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = 0;

#line 761 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":760
 * 
 *     def __richcmp__(self, other, int code):
 *         if code == 2:             # <<<<<<<<<<<<<<
 *             return self._eq(other)
 *         elif code == 3:
 */

#line 760 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    break;

#line 760 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    case 3:

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":763
 *             return self._eq(other)
 *         elif code == 3:
 *             return self._ne(other)             # <<<<<<<<<<<<<<
 *         else:
 *             return NotImplemented
 */

#line 763 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 763 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __Pyx_PyBool_FromLong(((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->_ne(__pyx_v_self, __pyx_v_other, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 763, __pyx_L1_error)

#line 763 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 763 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_r = __pyx_t_1;

#line 763 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = 0;

#line 763 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":762
 *         if code == 2:
 *             return self._eq(other)
 *         elif code == 3:             # <<<<<<<<<<<<<<
 *             return self._ne(other)
 *         else:
 */

#line 762 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    break;

#line 762 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    default:

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":765
 *             return self._ne(other)
 *         else:
 *             return NotImplemented             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */

#line 765 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 765 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(__pyx_builtin_NotImplemented);

#line 765 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_r = __pyx_builtin_NotImplemented;

#line 765 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L0;

#line 765 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    break;

#line 765 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":759
 *         return not self._eq(other)
 * 
 *     def __richcmp__(self, other, int code):             # <<<<<<<<<<<<<<
 *         if code == 2:
 *             return self._eq(other)
 */

#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 759 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":767
 *             return NotImplemented
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.nodes, self.adducts, self.used_as_adduct, self.feature_id)
 * 
 */

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_47__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_46__reduce__[] = "LCMSFeature.__reduce__(self)";

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_47__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = 
#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_46__reduce__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_46__reduce__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":768
 * 
 *     def __reduce__(self):
 *         return self.__class__, (self.nodes, self.adducts, self.used_as_adduct, self.feature_id)             # <<<<<<<<<<<<<<
 * 
 *     def __hash__(self):
 */

#line 768 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 768 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 768, __pyx_L1_error)

#line 768 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 768 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 768, __pyx_L1_error)

#line 768 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 768 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx_base.nodes));

#line 768 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx_base.nodes));

#line 768 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_self->__pyx_base.nodes));

#line 768 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_self->adducts);

#line 768 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_v_self->adducts);

#line 768 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_self->adducts);

#line 768 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_self->used_as_adduct);

#line 768 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_v_self->used_as_adduct);

#line 768 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_self->used_as_adduct);

#line 768 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_self->feature_id);

#line 768 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_v_self->feature_id);

#line 768 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_v_self->feature_id);

#line 768 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 768, __pyx_L1_error)

#line 768 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 768 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 768 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);

#line 768 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 768 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);

#line 768 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 768 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = 0;

#line 768 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_3;

#line 768 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = 0;

#line 768 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":767
 *             return NotImplemented
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.nodes, self.adducts, self.used_as_adduct, self.feature_id)
 * 
 */

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 767 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":770
 *         return self.__class__, (self.nodes, self.adducts, self.used_as_adduct, self.feature_id)
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash((self.mz, self.start_time, self.end_time))
 * 
 */

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static Py_hash_t __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_49__hash__(PyObject *__pyx_v_self); /*proto*/

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static Py_hash_t __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_49__hash__(PyObject *__pyx_v_self) {

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_hash_t __pyx_r;

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_r = 
#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_48__hash__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static Py_hash_t __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_48__hash__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_hash_t __pyx_r;

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_hash_t __pyx_t_5;

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__hash__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":771
 * 
 *     def __hash__(self):
 *         return hash((self.mz, self.start_time, self.end_time))             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, i):
 */

#line 771 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_mz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 771, __pyx_L1_error)

#line 771 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 771 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_start_time); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 771, __pyx_L1_error)

#line 771 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 771 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_end_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 771, __pyx_L1_error)

#line 771 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 771 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 771, __pyx_L1_error)

#line 771 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 771 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 771 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);

#line 771 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 771 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);

#line 771 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 771 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);

#line 771 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 771 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = 0;

#line 771 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = 0;

#line 771 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = PyObject_Hash(__pyx_t_4); if (unlikely(__pyx_t_5 == ((Py_hash_t)-1))) __PYX_ERR(0, 771, __pyx_L1_error)

#line 771 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 771 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_5;

#line 771 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":770
 *         return self.__class__, (self.nodes, self.adducts, self.used_as_adduct, self.feature_id)
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash((self.mz, self.start_time, self.end_time))
 * 
 */

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = -1;

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 770 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":773
 *         return hash((self.mz, self.start_time, self.end_time))
 * 
 *     def __getitem__(self, i):             # <<<<<<<<<<<<<<
 *         return self.nodes[i]
 * 
 */

#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_51__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_i); /*proto*/

#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_51__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_i) {

#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = 
#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_50__getitem__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), ((PyObject *)__pyx_v_i));

#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_50__getitem__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_i) {

#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":774
 * 
 *     def __getitem__(self, i):
 *         return self.nodes[i]             # <<<<<<<<<<<<<<
 * 
 *     cdef LCMSFeatureTreeNode getitem(self, size_t i):
 */

#line 774 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 774 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self->__pyx_base.nodes), __pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 774, __pyx_L1_error)

#line 774 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 774 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 774 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 774 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":773
 *         return hash((self.mz, self.start_time, self.end_time))
 * 
 *     def __getitem__(self, i):             # <<<<<<<<<<<<<<
 *         return self.nodes[i]
 * 
 */

#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 773 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":776
 *         return self.nodes[i]
 * 
 *     cdef LCMSFeatureTreeNode getitem(self, size_t i):             # <<<<<<<<<<<<<<
 *         return self.nodes.getitem(i)
 * 
 */

#line 776 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 776 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_getitem(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, size_t __pyx_v_i) {

#line 776 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_r = NULL;

#line 776 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 776 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 776 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 776 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 776 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 776 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("getitem", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":777
 * 
 *     cdef LCMSFeatureTreeNode getitem(self, size_t i):
 *         return self.nodes.getitem(i)             # <<<<<<<<<<<<<<
 * 
 *     cpdef double max_intensity(self):
 */

#line 777 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 777 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_getitem(__pyx_v_self->__pyx_base.nodes, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 777, __pyx_L1_error)

#line 777 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 777 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode))))) __PYX_ERR(0, 777, __pyx_L1_error)

#line 777 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_t_1);

#line 777 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 777 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":776
 *         return self.nodes[i]
 * 
 *     cdef LCMSFeatureTreeNode getitem(self, size_t i):             # <<<<<<<<<<<<<<
 *         return self.nodes.getitem(i)
 * 
 */

#line 776 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 776 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 776 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 776 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 776 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.getitem", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 776 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 776 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 776 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 776 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 776 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 776 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":779
 *         return self.nodes.getitem(i)
 * 
 *     cpdef double max_intensity(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_53max_intensity(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_max_intensity(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, int __pyx_skip_dispatch) {
  size_t __pyx_v_i
#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_n
#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  double __pyx_v_max_intensity
#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_r;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_5;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_6;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_7;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_8;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_9;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_10;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("max_intensity", 0);

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_max_intensity); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 779, __pyx_L1_error)

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_53max_intensity)) {

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_4 = NULL;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_4)) {

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 779, __pyx_L1_error)

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 779, __pyx_L1_error)

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = __pyx_t_5;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":783
 *             size_t i, n
 *             double max_intensity
 *         max_intensity = 0.0             # <<<<<<<<<<<<<<
 *         n = self.get_size()
 *         for i in range(n):
 */

#line 783 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_max_intensity = 0.0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":784
 *             double max_intensity
 *         max_intensity = 0.0
 *         n = self.get_size()             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             max_intensity = max(self.getitem(i).max_intensity(), max_intensity)
 */

#line 784 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_n = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.get_size(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_self));

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":785
 *         max_intensity = 0.0
 *         n = self.get_size()
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             max_intensity = max(self.getitem(i).max_intensity(), max_intensity)
 *         return max_intensity
 */

#line 785 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = __pyx_v_n;

#line 785 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_7 = __pyx_t_6;

#line 785 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {

#line 785 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_i = __pyx_t_8;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":786
 *         n = self.get_size()
 *         for i in range(n):
 *             max_intensity = max(self.getitem(i).max_intensity(), max_intensity)             # <<<<<<<<<<<<<<
 *         return max_intensity
 * 
 */

#line 786 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = __pyx_v_max_intensity;

#line 786 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self->__pyx_base.__pyx_vtab)->getitem(__pyx_v_self, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 786, __pyx_L1_error)

#line 786 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 786 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_9 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_t_1)->__pyx_base.__pyx_vtab)->__pyx_base.max_intensity(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_t_1), 0);

#line 786 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 786 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (((__pyx_t_5 > __pyx_t_9) != 0)) {

#line 786 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_10 = __pyx_t_5;

#line 786 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 786 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_10 = __pyx_t_9;

#line 786 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 786 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_max_intensity = __pyx_t_10;

#line 786 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":787
 *         for i in range(n):
 *             max_intensity = max(self.getitem(i).max_intensity(), max_intensity)
 *         return max_intensity             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 787 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_v_max_intensity;

#line 787 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":779
 *         return self.nodes.getitem(i)
 * 
 *     cpdef double max_intensity(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.max_intensity", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_53max_intensity(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_52max_intensity[] = "LCMSFeature.max_intensity(self) -> double";

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_53max_intensity(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("max_intensity (wrapper)", 0);
  __pyx_r = 
#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_52max_intensity(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_52max_intensity(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("max_intensity", 0);

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_max_intensity(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 779, __pyx_L1_error)

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature.max_intensity", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 779 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":75
 * cdef class LCMSFeature(FeatureBase):
 *     cdef:
 *         public double _total_intensity             # <<<<<<<<<<<<<<
 *         public double _last_mz
 *         public object _times
 */

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_16_total_intensity_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_16_total_intensity_1__get__(PyObject *__pyx_v_self) {

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = 0;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_16_total_intensity___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_16_total_intensity___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = NULL;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_total_intensity); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 75, __pyx_L1_error)

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = __pyx_t_1;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = 0;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature._total_intensity.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = NULL;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_16_total_intensity_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_16_total_intensity_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_16_total_intensity_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_16_total_intensity_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_value) {

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double __pyx_t_1;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 75, __pyx_L1_error)

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->_total_intensity = __pyx_t_1;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature._total_intensity.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = -1;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 75 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":76
 *     cdef:
 *         public double _total_intensity
 *         public double _last_mz             # <<<<<<<<<<<<<<
 *         public object _times
 *         public object _peaks
 */

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_8_last_mz_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_8_last_mz_1__get__(PyObject *__pyx_v_self) {

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = 0;

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_8_last_mz___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_8_last_mz___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = NULL;

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_last_mz); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 76, __pyx_L1_error)

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = __pyx_t_1;

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = 0;

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature._last_mz.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = NULL;

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_8_last_mz_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_8_last_mz_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_8_last_mz_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_8_last_mz_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_value) {

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double __pyx_t_1;

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 76, __pyx_L1_error)

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->_last_mz = __pyx_t_1;

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature._last_mz.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = -1;

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 76 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":77
 *         public double _total_intensity
 *         public double _last_mz
 *         public object _times             # <<<<<<<<<<<<<<
 *         public object _peaks
 *         public object adducts
 */

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_times_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_times_1__get__(PyObject *__pyx_v_self) {

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = 0;

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_times___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_times___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = NULL;

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(__pyx_v_self->_times);

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = __pyx_v_self->_times;

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_times_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_times_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_times_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_times_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_value) {

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(__pyx_v_value);

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(__pyx_v_value);

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->_times);

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(__pyx_v_self->_times);

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->_times = __pyx_v_value;

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_times_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_times_5__del__(PyObject *__pyx_v_self) {

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_times_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_times_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(Py_None);

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(Py_None);

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->_times);

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(__pyx_v_self->_times);

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->_times = Py_None;

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 77 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":78
 *         public double _last_mz
 *         public object _times
 *         public object _peaks             # <<<<<<<<<<<<<<
 *         public object adducts
 *         public object used_as_adduct
 */

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_peaks_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_peaks_1__get__(PyObject *__pyx_v_self) {

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = 0;

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_peaks___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_peaks___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = NULL;

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(__pyx_v_self->_peaks);

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = __pyx_v_self->_peaks;

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_peaks_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_peaks_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_peaks_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_peaks_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_value) {

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(__pyx_v_value);

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(__pyx_v_value);

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->_peaks);

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(__pyx_v_self->_peaks);

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->_peaks = __pyx_v_value;

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_peaks_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_peaks_5__del__(PyObject *__pyx_v_self) {

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_peaks_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_peaks_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(Py_None);

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(Py_None);

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->_peaks);

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(__pyx_v_self->_peaks);

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->_peaks = Py_None;

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 78 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":79
 *         public object _times
 *         public object _peaks
 *         public object adducts             # <<<<<<<<<<<<<<
 *         public object used_as_adduct
 *         public object created_at
 */

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_7adducts_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_7adducts_1__get__(PyObject *__pyx_v_self) {

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = 0;

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_7adducts___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_7adducts___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = NULL;

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(__pyx_v_self->adducts);

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = __pyx_v_self->adducts;

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_7adducts_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_7adducts_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_7adducts_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_7adducts_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_value) {

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(__pyx_v_value);

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(__pyx_v_value);

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->adducts);

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(__pyx_v_self->adducts);

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->adducts = __pyx_v_value;

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_7adducts_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_7adducts_5__del__(PyObject *__pyx_v_self) {

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_7adducts_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_7adducts_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(Py_None);

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(Py_None);

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->adducts);

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(__pyx_v_self->adducts);

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->adducts = Py_None;

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 79 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":80
 *         public object _peaks
 *         public object adducts
 *         public object used_as_adduct             # <<<<<<<<<<<<<<
 *         public object created_at
 *         public object feature_id
 */

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14used_as_adduct_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14used_as_adduct_1__get__(PyObject *__pyx_v_self) {

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = 0;

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14used_as_adduct___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14used_as_adduct___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = NULL;

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(__pyx_v_self->used_as_adduct);

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = __pyx_v_self->used_as_adduct;

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14used_as_adduct_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14used_as_adduct_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14used_as_adduct_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14used_as_adduct_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_value) {

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(__pyx_v_value);

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(__pyx_v_value);

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->used_as_adduct);

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(__pyx_v_self->used_as_adduct);

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->used_as_adduct = __pyx_v_value;

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14used_as_adduct_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14used_as_adduct_5__del__(PyObject *__pyx_v_self) {

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14used_as_adduct_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14used_as_adduct_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(Py_None);

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(Py_None);

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->used_as_adduct);

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(__pyx_v_self->used_as_adduct);

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->used_as_adduct = Py_None;

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 80 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":81
 *         public object adducts
 *         public object used_as_adduct
 *         public object created_at             # <<<<<<<<<<<<<<
 *         public object feature_id
 *         public RunningWeightedAverage _peak_averager
 */

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10created_at_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10created_at_1__get__(PyObject *__pyx_v_self) {

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = 0;

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10created_at___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10created_at___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = NULL;

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(__pyx_v_self->created_at);

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = __pyx_v_self->created_at;

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10created_at_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10created_at_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10created_at_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10created_at_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_value) {

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(__pyx_v_value);

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(__pyx_v_value);

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->created_at);

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(__pyx_v_self->created_at);

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->created_at = __pyx_v_value;

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10created_at_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10created_at_5__del__(PyObject *__pyx_v_self) {

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10created_at_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10created_at_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(Py_None);

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(Py_None);

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->created_at);

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(__pyx_v_self->created_at);

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->created_at = Py_None;

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 81 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":82
 *         public object used_as_adduct
 *         public object created_at
 *         public object feature_id             # <<<<<<<<<<<<<<
 *         public RunningWeightedAverage _peak_averager
 * 
 */

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10feature_id_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10feature_id_1__get__(PyObject *__pyx_v_self) {

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = 0;

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10feature_id___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10feature_id___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = NULL;

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(__pyx_v_self->feature_id);

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = __pyx_v_self->feature_id;

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10feature_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10feature_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10feature_id_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10feature_id_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_value) {

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(__pyx_v_value);

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(__pyx_v_value);

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->feature_id);

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(__pyx_v_self->feature_id);

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->feature_id = __pyx_v_value;

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10feature_id_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10feature_id_5__del__(PyObject *__pyx_v_self) {

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10feature_id_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10feature_id_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(Py_None);

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(Py_None);

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->feature_id);

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(__pyx_v_self->feature_id);

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->feature_id = Py_None;

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 82 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":83
 *         public object created_at
 *         public object feature_id
 *         public RunningWeightedAverage _peak_averager             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint _eq(self, other)
 */

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14_peak_averager_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14_peak_averager_1__get__(PyObject *__pyx_v_self) {

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = 0;

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14_peak_averager___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14_peak_averager___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = NULL;

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->_peak_averager));

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = ((PyObject *)__pyx_v_self->_peak_averager);

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14_peak_averager_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14_peak_averager_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14_peak_averager_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14_peak_averager_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self, PyObject *__pyx_v_value) {

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage))))) __PYX_ERR(2, 83, __pyx_L1_error)

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = __pyx_v_value;

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(__pyx_t_1);

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(__pyx_t_1);

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->_peak_averager);

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_peak_averager));

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->_peak_averager = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_t_1);

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = 0;

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.LCMSFeature._peak_averager.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = -1;

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14_peak_averager_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14_peak_averager_5__del__(PyObject *__pyx_v_self) {

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14_peak_averager_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_self));

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14_peak_averager_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_self) {

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(Py_None);

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(Py_None);

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->_peak_averager);

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_peak_averager));

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->_peak_averager = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)Py_None);

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 83 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":791
 * 
 * cdef class EmptyFeature(FeatureBase):
 *     def __init__(self, mz):             # <<<<<<<<<<<<<<
 *         FeatureBase.__init__(self, [])
 *         self._mz = mz
 */

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_mz = 0
#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_r;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mz,0};

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject* values[1] = {0};

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(__pyx_kwds)) {

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      Py_ssize_t kw_args;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {
        case  1: 
#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0: break;
        default: 
#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0:

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mz)) != 0)) kw_args--;
        else 
#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(kw_args > 0)) {

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 791, __pyx_L3_error)

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L5_argtuple_error;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_mz = values[0];

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); 
#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__PYX_ERR(0, 791, __pyx_L3_error)

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_error:;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.EmptyFeature.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return -1;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature___init__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *)__pyx_v_self), __pyx_v_mz);

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature___init__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *__pyx_v_self, PyObject *__pyx_v_mz) {

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_r;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_5;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_7;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":792
 * cdef class EmptyFeature(FeatureBase):
 *     def __init__(self, mz):
 *         FeatureBase.__init__(self, [])             # <<<<<<<<<<<<<<
 *         self._mz = mz
 *         self._start_time = 0
 */

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase), __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 792, __pyx_L1_error)

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 792, __pyx_L1_error)

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = NULL;

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = 0;

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (likely(__pyx_t_4)) {

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(__pyx_t_4);

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_INCREF(function);

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF_SET(__pyx_t_2, function);

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_5 = 1;

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #if CYTHON_FAST_PYCALL

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (PyFunction_Check(__pyx_t_2)) {

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_self), __pyx_t_3};

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 792, __pyx_L1_error)

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #endif

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #if CYTHON_FAST_PYCCALL

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_self), __pyx_t_3};

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 792, __pyx_L1_error)

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  #endif

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 792, __pyx_L1_error)

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_6);

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_4) {

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self));

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_self));

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GIVEREF(__pyx_t_3);

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_t_3);

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = 0;

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 792, __pyx_L1_error)

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 792 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":793
 *     def __init__(self, mz):
 *         FeatureBase.__init__(self, [])
 *         self._mz = mz             # <<<<<<<<<<<<<<
 *         self._start_time = 0
 *         self._end_time = INF
 */

#line 793 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_7 = __pyx_PyFloat_AsDouble(__pyx_v_mz); if (unlikely((__pyx_t_7 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 793, __pyx_L1_error)

#line 793 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->__pyx_base._mz = __pyx_t_7;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":794
 *         FeatureBase.__init__(self, [])
 *         self._mz = mz
 *         self._start_time = 0             # <<<<<<<<<<<<<<
 *         self._end_time = INF
 * 
 */

#line 794 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->__pyx_base._start_time = 0.0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":795
 *         self._mz = mz
 *         self._start_time = 0
 *         self._end_time = INF             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

#line 795 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->__pyx_base._end_time = __pyx_v_12ms_deisotope_2_c_11feature_map_12lcms_feature_INF;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":791
 * 
 * cdef class EmptyFeature(FeatureBase):
 *     def __init__(self, mz):             # <<<<<<<<<<<<<<
 *         FeatureBase.__init__(self, [])
 *         self._mz = mz
 */

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.EmptyFeature.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = -1;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 791 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":798
 * 
 *     @staticmethod
 *     cdef EmptyFeature _create(double mz):             # <<<<<<<<<<<<<<
 *         cdef EmptyFeature inst = EmptyFeature.__new__(EmptyFeature)
 *         inst._mz = mz
 */

#line 798 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 798 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature__create(double __pyx_v_mz) {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *__pyx_v_inst = 0
#line 798 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 798 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *__pyx_r = NULL;

#line 798 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 798 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 798 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 798 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 798 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 798 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 798 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_create", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":799
 *     @staticmethod
 *     cdef EmptyFeature _create(double mz):
 *         cdef EmptyFeature inst = EmptyFeature.__new__(EmptyFeature)             # <<<<<<<<<<<<<<
 *         inst._mz = mz
 *         inst.nodes = LCMSFeatureTreeList._create([])
 */

#line 799 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature(((PyTypeObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 799, __pyx_L1_error)

#line 799 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));

#line 799 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_inst = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *)__pyx_t_1);

#line 799 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":800
 *     cdef EmptyFeature _create(double mz):
 *         cdef EmptyFeature inst = EmptyFeature.__new__(EmptyFeature)
 *         inst._mz = mz             # <<<<<<<<<<<<<<
 *         inst.nodes = LCMSFeatureTreeList._create([])
 *         inst._start_time = 0
 */

#line 800 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_inst->__pyx_base._mz = __pyx_v_mz;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":801
 *         cdef EmptyFeature inst = EmptyFeature.__new__(EmptyFeature)
 *         inst._mz = mz
 *         inst.nodes = LCMSFeatureTreeList._create([])             # <<<<<<<<<<<<<<
 *         inst._start_time = 0
 *         inst._end_time = INF
 */

#line 801 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 801, __pyx_L1_error)

#line 801 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 801 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList__create(((PyObject*)__pyx_t_1))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 801, __pyx_L1_error)

#line 801 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 801 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 801 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 801 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_inst->__pyx_base.nodes);

#line 801 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(((PyObject *)__pyx_v_inst->__pyx_base.nodes));

#line 801 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_inst->__pyx_base.nodes = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_t_2);

#line 801 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":802
 *         inst._mz = mz
 *         inst.nodes = LCMSFeatureTreeList._create([])
 *         inst._start_time = 0             # <<<<<<<<<<<<<<
 *         inst._end_time = INF
 *         return inst
 */

#line 802 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_inst->__pyx_base._start_time = 0.0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":803
 *         inst.nodes = LCMSFeatureTreeList._create([])
 *         inst._start_time = 0
 *         inst._end_time = INF             # <<<<<<<<<<<<<<
 *         return inst
 * 
 */

#line 803 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_inst->__pyx_base._end_time = __pyx_v_12ms_deisotope_2_c_11feature_map_12lcms_feature_INF;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":804
 *         inst._start_time = 0
 *         inst._end_time = INF
 *         return inst             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 804 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 804 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_inst));

#line 804 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_v_inst;

#line 804 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":798
 * 
 *     @staticmethod
 *     cdef EmptyFeature _create(double mz):             # <<<<<<<<<<<<<<
 *         cdef EmptyFeature inst = EmptyFeature.__new__(EmptyFeature)
 *         inst._mz = mz
 */

#line 798 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 798 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 798 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 798 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 798 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 798 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.EmptyFeature._create", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 798 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 798 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 798 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);

#line 798 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 798 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 798 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 798 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":807
 * 
 *     @property
 *     def mz(self):             # <<<<<<<<<<<<<<
 *         return self._mz
 * 
 */

#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_2mz_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_2mz_1__get__(PyObject *__pyx_v_self) {

#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_2mz___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *)__pyx_v_self));

#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_2mz___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *__pyx_v_self) {

#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":808
 *     @property
 *     def mz(self):
 *         return self._mz             # <<<<<<<<<<<<<<
 * 
 *     cdef double get_mz(self):
 */

#line 808 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 808 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->__pyx_base._mz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 808, __pyx_L1_error)

#line 808 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 808 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 808 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 808 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":807
 * 
 *     @property
 *     def mz(self):             # <<<<<<<<<<<<<<
 *         return self._mz
 * 
 */

#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.EmptyFeature.mz.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 807 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":810
 *         return self._mz
 * 
 *     cdef double get_mz(self):             # <<<<<<<<<<<<<<
 *         return self._mz
 * 
 */

#line 810 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 810 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_get_mz(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *__pyx_v_self) {

#line 810 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_r;

#line 810 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 810 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("get_mz", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":811
 * 
 *     cdef double get_mz(self):
 *         return self._mz             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */

#line 811 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_v_self->__pyx_base._mz;

#line 811 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":810
 *         return self._mz
 * 
 *     cdef double get_mz(self):             # <<<<<<<<<<<<<<
 *         return self._mz
 * 
 */

#line 810 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 810 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 810 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 810 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 810 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 810 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":813
 *         return self._mz
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */

#line 813 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 813 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 813 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static Py_ssize_t __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_3__len__(PyObject *__pyx_v_self); /*proto*/

#line 813 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static Py_ssize_t __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_3__len__(PyObject *__pyx_v_self) {

#line 813 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_ssize_t __pyx_r;

#line 813 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 813 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = 
#line 813 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_2__len__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *)__pyx_v_self));

#line 813 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 813 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 813 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 813 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 813 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 813 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 813 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static Py_ssize_t __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_2__len__(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *__pyx_v_self) {

#line 813 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_ssize_t __pyx_r;

#line 813 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 813 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":814
 * 
 *     def __len__(self):
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, i):
 */

#line 814 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 814 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":813
 *         return self._mz
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */

#line 813 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 813 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 813 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 813 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 813 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 813 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":816
 *         return 0
 * 
 *     def __getitem__(self, i):             # <<<<<<<<<<<<<<
 *         raise IndexError()
 * 
 */

#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_5__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_i); /*proto*/

#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_5__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_i) {

#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = 
#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_4__getitem__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *)__pyx_v_self), ((PyObject *)__pyx_v_i));

#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_4__getitem__(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_i) {

#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":817
 * 
 *     def __getitem__(self, i):
 *         raise IndexError()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 817 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_IndexError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 817, __pyx_L1_error)

#line 817 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 817 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 817 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 817 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __PYX_ERR(0, 817, __pyx_L1_error)

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":816
 *         return 0
 * 
 *     def __getitem__(self, i):             # <<<<<<<<<<<<<<
 *         raise IndexError()
 * 
 */

#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.EmptyFeature.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 816 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":820
 * 
 *     @property
 *     def start_time(self):             # <<<<<<<<<<<<<<
 *         return self._start_time
 * 
 */

#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_10start_time_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_10start_time_1__get__(PyObject *__pyx_v_self) {

#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_10start_time___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *)__pyx_v_self));

#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_10start_time___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *__pyx_v_self) {

#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":821
 *     @property
 *     def start_time(self):
 *         return self._start_time             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 821 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 821 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->__pyx_base._start_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 821, __pyx_L1_error)

#line 821 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 821 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 821 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 821 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":820
 * 
 *     @property
 *     def start_time(self):             # <<<<<<<<<<<<<<
 *         return self._start_time
 * 
 */

#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.EmptyFeature.start_time.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 820 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":824
 * 
 *     @property
 *     def end_time(self):             # <<<<<<<<<<<<<<
 *         return self._end_time
 * 
 */

#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_8end_time_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_8end_time_1__get__(PyObject *__pyx_v_self) {

#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_8end_time___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *)__pyx_v_self));

#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_8end_time___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *__pyx_v_self) {

#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":825
 *     @property
 *     def end_time(self):
 *         return self._end_time             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 825 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 825 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->__pyx_base._end_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 825, __pyx_L1_error)

#line 825 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 825 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 825 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 825 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":824
 * 
 *     @property
 *     def end_time(self):             # <<<<<<<<<<<<<<
 *         return self._end_time
 * 
 */

#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.EmptyFeature.end_time.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 824 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_6__reduce_cython__[] = "EmptyFeature.__reduce_cython__(self)";

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_6__reduce_cython__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_6__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *__pyx_v_self) {
  PyObject *__pyx_v_state = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v__dict = 0
#line 1 "(tree fragment)"
;
  int __pyx_v_use_setstate
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_2 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_3 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_5;

#line 1 "(tree fragment)"
  int __pyx_t_6;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._end_time, self._mz, self._start_time, self.nodes)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */

#line 5 "(tree fragment)"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->__pyx_base._end_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->__pyx_base._mz); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_2);

#line 5 "(tree fragment)"
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->__pyx_base._start_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_3);

#line 5 "(tree fragment)"
  __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_2);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_3);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);

#line 5 "(tree fragment)"
  __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx_base.nodes));

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx_base.nodes));

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_4, 3, ((PyObject *)__pyx_v_self->__pyx_base.nodes));

#line 5 "(tree fragment)"
  __pyx_t_1 = 0;

#line 5 "(tree fragment)"
  __pyx_t_2 = 0;

#line 5 "(tree fragment)"
  __pyx_t_3 = 0;

#line 5 "(tree fragment)"
  __pyx_v_state = ((PyObject*)__pyx_t_4);

#line 5 "(tree fragment)"
  __pyx_t_4 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._end_time, self._mz, self._start_time, self.nodes)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */

#line 6 "(tree fragment)"
  __pyx_t_4 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 6 "(tree fragment)"
  __pyx_v__dict = __pyx_t_4;

#line 6 "(tree fragment)"
  __pyx_t_4 = 0;

  /* "(tree fragment)":7
 *     state = (self._end_time, self._mz, self._start_time, self.nodes)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
  __pyx_t_5 = (__pyx_v__dict != Py_None);

#line 7 "(tree fragment)"
  __pyx_t_6 = (__pyx_t_5 != 0);

#line 7 "(tree fragment)"
  if (__pyx_t_6) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */

#line 8 "(tree fragment)"
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 8 "(tree fragment)"
    __Pyx_INCREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v__dict);

#line 8 "(tree fragment)"
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 8 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 8 "(tree fragment)"
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));

#line 8 "(tree fragment)"
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.nodes is not None
 */

#line 9 "(tree fragment)"
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._end_time, self._mz, self._start_time, self.nodes)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
    goto __pyx_L3;

#line 7 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.nodes is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_EmptyFeature, (type(self), 0x9e63e75, None), state
 */

#line 11 "(tree fragment)"
  /*else*/ {

#line 11 "(tree fragment)"
    __pyx_t_6 = (((PyObject *)__pyx_v_self->__pyx_base.nodes) != Py_None);

#line 11 "(tree fragment)"
    __pyx_v_use_setstate = __pyx_t_6;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.nodes is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_EmptyFeature, (type(self), 0x9e63e75, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  __pyx_t_6 = (__pyx_v_use_setstate != 0);

#line 12 "(tree fragment)"
  if (__pyx_t_6) {

    /* "(tree fragment)":13
 *         use_setstate = self.nodes is not None
 *     if use_setstate:
 *         return __pyx_unpickle_EmptyFeature, (type(self), 0x9e63e75, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_EmptyFeature, (type(self), 0x9e63e75, state)
 */

#line 13 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 13 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_EmptyFeature); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 13 "(tree fragment)"
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 13 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_166084213);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_166084213);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_166084213);

#line 13 "(tree fragment)"
    __Pyx_INCREF(Py_None);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(Py_None);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 2, Py_None);

#line 13 "(tree fragment)"
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_3);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_4);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state);

#line 13 "(tree fragment)"
    __pyx_t_3 = 0;

#line 13 "(tree fragment)"
    __pyx_t_4 = 0;

#line 13 "(tree fragment)"
    __pyx_r = __pyx_t_2;

#line 13 "(tree fragment)"
    __pyx_t_2 = 0;

#line 13 "(tree fragment)"
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.nodes is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_EmptyFeature, (type(self), 0x9e63e75, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_EmptyFeature, (type(self), 0x9e63e75, None), state
 *     else:
 *         return __pyx_unpickle_EmptyFeature, (type(self), 0x9e63e75, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_EmptyFeature__set_state(self, __pyx_state)
 */

#line 15 "(tree fragment)"
  /*else*/ {

#line 15 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 15 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_EmptyFeature); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 15 "(tree fragment)"
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 15 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_166084213);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_166084213);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_166084213);

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state);

#line 15 "(tree fragment)"
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_2);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_4);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);

#line 15 "(tree fragment)"
    __pyx_t_2 = 0;

#line 15 "(tree fragment)"
    __pyx_t_4 = 0;

#line 15 "(tree fragment)"
    __pyx_r = __pyx_t_3;

#line 15 "(tree fragment)"
    __pyx_t_3 = 0;

#line 15 "(tree fragment)"
    goto __pyx_L0;

#line 15 "(tree fragment)"
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_2);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_3);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.EmptyFeature.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v_state);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v__dict);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_EmptyFeature, (type(self), 0x9e63e75, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_EmptyFeature__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
/* Python wrapper */

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 16 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_8__setstate_cython__[] = "EmptyFeature.__setstate_cython__(self, __pyx_state)";

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 16 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_8__setstate_cython__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_8__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 16 "(tree fragment)"
  int __pyx_lineno = 0;

#line 16 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 16 "(tree fragment)"
  int __pyx_clineno = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_EmptyFeature, (type(self), 0x9e63e75, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_EmptyFeature__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */

#line 17 "(tree fragment)"
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_unpickle_EmptyFeature__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 17 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_EmptyFeature, (type(self), 0x9e63e75, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_EmptyFeature__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 16 "(tree fragment)"
  goto __pyx_L0;

#line 16 "(tree fragment)"
  __pyx_L1_error:;

#line 16 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 16 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.EmptyFeature.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 16 "(tree fragment)"
  __pyx_r = NULL;

#line 16 "(tree fragment)"
  __pyx_L0:;

#line 16 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":829
 * 
 * cdef class FeatureSetIterator(object):
 *     def __init__(self, features, start_time=None, end_time=None):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t n, i
 */

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_features = 0
#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyObject *__pyx_v_start_time = 0
#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyObject *__pyx_v_end_time = 0
#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_r;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_features,&__pyx_n_s_start_time,&__pyx_n_s_end_time,0};

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject* values[3] = {0,0,0};

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    values[1] = ((PyObject *)Py_None);

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    values[2] = ((PyObject *)Py_None);

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(__pyx_kwds)) {

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      Py_ssize_t kw_args;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {
        case  3: 
#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0: break;
        default: 
#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0:

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_features)) != 0)) kw_args--;
        else 
#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  1:

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (kw_args > 0) {

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_start_time);

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (value) { values[1] = value; kw_args--; }

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  2:

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (kw_args > 0) {

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_end_time);

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (value) { values[2] = value; kw_args--; }

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(kw_args > 0)) {

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 829, __pyx_L3_error)

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: 
#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        break;
        default: 
#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_features = values[0];

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_start_time = values[1];

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_end_time = values[2];

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__PYX_ERR(0, 829, __pyx_L3_error)

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_error:;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return -1;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator___init__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self), __pyx_v_features, __pyx_v_start_time, __pyx_v_end_time);

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator___init__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, PyObject *__pyx_v_features, PyObject *__pyx_v_start_time, PyObject *__pyx_v_end_time) {
  size_t __pyx_v_n
#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_i
#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_feature = NULL
#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  double __pyx_v_time
#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_r;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_2;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_3;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_4;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_5;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_6;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_7;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_8;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_9;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_10;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":833
 *             size_t n, i
 * 
 *         self.features = list(features)             # <<<<<<<<<<<<<<
 *         n = PyList_GET_SIZE(self.features)
 *         self.real_features = []
 */

#line 833 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PySequence_List(__pyx_v_features); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 833, __pyx_L1_error)

#line 833 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 833 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 833 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_self->features);

#line 833 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_v_self->features);

#line 833 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->features = ((PyObject*)__pyx_t_1);

#line 833 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":834
 * 
 *         self.features = list(features)
 *         n = PyList_GET_SIZE(self.features)             # <<<<<<<<<<<<<<
 *         self.real_features = []
 *         self.index_list = <size_t*>malloc(sizeof(size_t) * n)
 */

#line 834 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_v_self->features;

#line 834 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 834 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_n = PyList_GET_SIZE(__pyx_t_1);

#line 834 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":835
 *         self.features = list(features)
 *         n = PyList_GET_SIZE(self.features)
 *         self.real_features = []             # <<<<<<<<<<<<<<
 *         self.index_list = <size_t*>malloc(sizeof(size_t) * n)
 *         if self.index_list == NULL:
 */

#line 835 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 835, __pyx_L1_error)

#line 835 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 835 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 835 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_self->real_features);

#line 835 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_v_self->real_features);

#line 835 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->real_features = ((PyObject*)__pyx_t_1);

#line 835 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":836
 *         n = PyList_GET_SIZE(self.features)
 *         self.real_features = []
 *         self.index_list = <size_t*>malloc(sizeof(size_t) * n)             # <<<<<<<<<<<<<<
 *         if self.index_list == NULL:
 *             raise MemoryError("Failed to allocate index list for LCMS Feature Iterator")
 */

#line 836 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->index_list = ((size_t *)malloc(((sizeof(size_t)) * __pyx_v_n)));

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":837
 *         self.real_features = []
 *         self.index_list = <size_t*>malloc(sizeof(size_t) * n)
 *         if self.index_list == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError("Failed to allocate index list for LCMS Feature Iterator")
 *         self.start_time = 0
 */

#line 837 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = ((__pyx_v_self->index_list == NULL) != 0);

#line 837 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_t_2)) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":838
 *         self.index_list = <size_t*>malloc(sizeof(size_t) * n)
 *         if self.index_list == NULL:
 *             raise MemoryError("Failed to allocate index list for LCMS Feature Iterator")             # <<<<<<<<<<<<<<
 *         self.start_time = 0
 *         self.end_time = INF
 */

#line 838 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 838, __pyx_L1_error)

#line 838 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 838 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 838 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 838 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __PYX_ERR(0, 838, __pyx_L1_error)

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":837
 *         self.real_features = []
 *         self.index_list = <size_t*>malloc(sizeof(size_t) * n)
 *         if self.index_list == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError("Failed to allocate index list for LCMS Feature Iterator")
 *         self.start_time = 0
 */

#line 837 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":839
 *         if self.index_list == NULL:
 *             raise MemoryError("Failed to allocate index list for LCMS Feature Iterator")
 *         self.start_time = 0             # <<<<<<<<<<<<<<
 *         self.end_time = INF
 * 
 */

#line 839 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->start_time = 0.0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":840
 *             raise MemoryError("Failed to allocate index list for LCMS Feature Iterator")
 *         self.start_time = 0
 *         self.end_time = INF             # <<<<<<<<<<<<<<
 * 
 *         if start_time is not None:
 */

#line 840 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->end_time = __pyx_v_12ms_deisotope_2_c_11feature_map_12lcms_feature_INF;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":842
 *         self.end_time = INF
 * 
 *         if start_time is not None:             # <<<<<<<<<<<<<<
 *             self.start_time = start_time
 *         if end_time is not None:
 */

#line 842 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = (__pyx_v_start_time != Py_None);

#line 842 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 842 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_3) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":843
 * 
 *         if start_time is not None:
 *             self.start_time = start_time             # <<<<<<<<<<<<<<
 *         if end_time is not None:
 *             self.end_time = end_time
 */

#line 843 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_v_start_time); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 843, __pyx_L1_error)

#line 843 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_self->start_time = __pyx_t_4;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":842
 *         self.end_time = INF
 * 
 *         if start_time is not None:             # <<<<<<<<<<<<<<
 *             self.start_time = start_time
 *         if end_time is not None:
 */

#line 842 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":844
 *         if start_time is not None:
 *             self.start_time = start_time
 *         if end_time is not None:             # <<<<<<<<<<<<<<
 *             self.end_time = end_time
 * 
 */

#line 844 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = (__pyx_v_end_time != Py_None);

#line 844 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = (__pyx_t_3 != 0);

#line 844 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":845
 *             self.start_time = start_time
 *         if end_time is not None:
 *             self.end_time = end_time             # <<<<<<<<<<<<<<
 * 
 *         for i in range(n):
 */

#line 845 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_v_end_time); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 845, __pyx_L1_error)

#line 845 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_self->end_time = __pyx_t_4;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":844
 *         if start_time is not None:
 *             self.start_time = start_time
 *         if end_time is not None:             # <<<<<<<<<<<<<<
 *             self.end_time = end_time
 * 
 */

#line 844 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":847
 *             self.end_time = end_time
 * 
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             feature = <FeatureBase>PyList_GET_ITEM(self.features, i)
 *             if feature is None or isinstance(feature, EmptyFeature):
 */

#line 847 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = __pyx_v_n;

#line 847 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = __pyx_t_5;

#line 847 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {

#line 847 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_i = __pyx_t_7;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":848
 * 
 *         for i in range(n):
 *             feature = <FeatureBase>PyList_GET_ITEM(self.features, i)             # <<<<<<<<<<<<<<
 *             if feature is None or isinstance(feature, EmptyFeature):
 *                 continue
 */

#line 848 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __pyx_v_self->features;

#line 848 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 848 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_8 = PyList_GET_ITEM(__pyx_t_1, __pyx_v_i);

#line 848 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 848 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_8);

#line 848 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 848 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF_SET(__pyx_v_feature, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_t_1));

#line 848 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":849
 *         for i in range(n):
 *             feature = <FeatureBase>PyList_GET_ITEM(self.features, i)
 *             if feature is None or isinstance(feature, EmptyFeature):             # <<<<<<<<<<<<<<
 *                 continue
 *             self.real_features.append(feature)
 */

#line 849 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = (((PyObject *)__pyx_v_feature) == Py_None);

#line 849 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_9 = (__pyx_t_3 != 0);

#line 849 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (!__pyx_t_9) {

#line 849 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 849 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_2 = __pyx_t_9;

#line 849 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L9_bool_binop_done;

#line 849 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 849 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_9 = __Pyx_TypeCheck(((PyObject *)__pyx_v_feature), __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature); 

#line 849 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = (__pyx_t_9 != 0);

#line 849 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = __pyx_t_3;

#line 849 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L9_bool_binop_done:;

#line 849 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":850
 *             feature = <FeatureBase>PyList_GET_ITEM(self.features, i)
 *             if feature is None or isinstance(feature, EmptyFeature):
 *                 continue             # <<<<<<<<<<<<<<
 *             self.real_features.append(feature)
 *             time = feature.get_start_time()
 */

#line 850 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L6_continue;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":849
 *         for i in range(n):
 *             feature = <FeatureBase>PyList_GET_ITEM(self.features, i)
 *             if feature is None or isinstance(feature, EmptyFeature):             # <<<<<<<<<<<<<<
 *                 continue
 *             self.real_features.append(feature)
 */

#line 849 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":851
 *             if feature is None or isinstance(feature, EmptyFeature):
 *                 continue
 *             self.real_features.append(feature)             # <<<<<<<<<<<<<<
 *             time = feature.get_start_time()
 *             if (self.start_time < time) and (start_time is None):
 */

#line 851 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(__pyx_v_self->real_features == Py_None)) {

#line 851 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");

#line 851 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __PYX_ERR(0, 851, __pyx_L1_error)

#line 851 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 851 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_self->real_features, ((PyObject *)__pyx_v_feature)); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 851, __pyx_L1_error)

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":852
 *                 continue
 *             self.real_features.append(feature)
 *             time = feature.get_start_time()             # <<<<<<<<<<<<<<
 *             if (self.start_time < time) and (start_time is None):
 *                 self.start_time = time
 */

#line 852 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_time = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_feature->__pyx_vtab)->get_start_time(__pyx_v_feature);

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":853
 *             self.real_features.append(feature)
 *             time = feature.get_start_time()
 *             if (self.start_time < time) and (start_time is None):             # <<<<<<<<<<<<<<
 *                 self.start_time = time
 *             time = feature.get_end_time()
 */

#line 853 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = ((__pyx_v_self->start_time < __pyx_v_time) != 0);

#line 853 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_3) {

#line 853 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 853 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_2 = __pyx_t_3;

#line 853 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L12_bool_binop_done;

#line 853 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 853 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = (__pyx_v_start_time == Py_None);

#line 853 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_9 = (__pyx_t_3 != 0);

#line 853 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = __pyx_t_9;

#line 853 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L12_bool_binop_done:;

#line 853 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":854
 *             time = feature.get_start_time()
 *             if (self.start_time < time) and (start_time is None):
 *                 self.start_time = time             # <<<<<<<<<<<<<<
 *             time = feature.get_end_time()
 *             if (self.end_time > time) and (end_time is None):
 */

#line 854 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_self->start_time = __pyx_v_time;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":853
 *             self.real_features.append(feature)
 *             time = feature.get_start_time()
 *             if (self.start_time < time) and (start_time is None):             # <<<<<<<<<<<<<<
 *                 self.start_time = time
 *             time = feature.get_end_time()
 */

#line 853 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":855
 *             if (self.start_time < time) and (start_time is None):
 *                 self.start_time = time
 *             time = feature.get_end_time()             # <<<<<<<<<<<<<<
 *             if (self.end_time > time) and (end_time is None):
 *                 self.end_time = time
 */

#line 855 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_time = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_feature->__pyx_vtab)->get_end_time(__pyx_v_feature);

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":856
 *                 self.start_time = time
 *             time = feature.get_end_time()
 *             if (self.end_time > time) and (end_time is None):             # <<<<<<<<<<<<<<
 *                 self.end_time = time
 *             self.index_list[i] = 0
 */

#line 856 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_9 = ((__pyx_v_self->end_time > __pyx_v_time) != 0);

#line 856 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_9) {

#line 856 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 856 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_2 = __pyx_t_9;

#line 856 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L15_bool_binop_done;

#line 856 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 856 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_9 = (__pyx_v_end_time == Py_None);

#line 856 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_3 = (__pyx_t_9 != 0);

#line 856 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = __pyx_t_3;

#line 856 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L15_bool_binop_done:;

#line 856 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":857
 *             time = feature.get_end_time()
 *             if (self.end_time > time) and (end_time is None):
 *                 self.end_time = time             # <<<<<<<<<<<<<<
 *             self.index_list[i] = 0
 * 
 */

#line 857 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_self->end_time = __pyx_v_time;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":856
 *                 self.start_time = time
 *             time = feature.get_end_time()
 *             if (self.end_time > time) and (end_time is None):             # <<<<<<<<<<<<<<
 *                 self.end_time = time
 *             self.index_list[i] = 0
 */

#line 856 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":858
 *             if (self.end_time > time) and (end_time is None):
 *                 self.end_time = time
 *             self.index_list[i] = 0             # <<<<<<<<<<<<<<
 * 
 *         self.init_indices()
 */

#line 858 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    (__pyx_v_self->index_list[__pyx_v_i]) = 0;

#line 858 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L6_continue:;

#line 858 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":860
 *             self.index_list[i] = 0
 * 
 *         self.init_indices()             # <<<<<<<<<<<<<<
 *         self.last_time_seen = -1
 * 
 */

#line 860 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self->__pyx_vtab)->init_indices(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 860, __pyx_L1_error)

#line 860 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 860 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":861
 * 
 *         self.init_indices()
 *         self.last_time_seen = -1             # <<<<<<<<<<<<<<
 * 
 *     cdef int _initialize(self, list features) except 1:
 */

#line 861 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->last_time_seen = -1.0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":829
 * 
 * cdef class FeatureSetIterator(object):
 *     def __init__(self, features, start_time=None, end_time=None):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t n, i
 */

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = -1;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_feature);

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 829 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":863
 *         self.last_time_seen = -1
 * 
 *     cdef int _initialize(self, list features) except 1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator__initialize(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, PyObject *__pyx_v_features) {
  size_t __pyx_v_i
#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_n
#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_feature = NULL
#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  double __pyx_v_time
#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_r;

#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_2;

#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_3;

#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_4;

#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_5;

#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_6;

#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_7;

#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_8;

#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_9;

#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_initialize", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":866
 *         cdef:
 *             size_t i, n
 *         self.features = features             # <<<<<<<<<<<<<<
 *         n = PyList_GET_SIZE(self.features)
 *         self.real_features = []
 */

#line 866 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_features);

#line 866 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_v_features);

#line 866 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_self->features);

#line 866 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_v_self->features);

#line 866 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->features = __pyx_v_features;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":867
 *             size_t i, n
 *         self.features = features
 *         n = PyList_GET_SIZE(self.features)             # <<<<<<<<<<<<<<
 *         self.real_features = []
 *         self.index_list = <size_t*>malloc(sizeof(size_t) * n)
 */

#line 867 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_v_self->features;

#line 867 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 867 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_n = PyList_GET_SIZE(__pyx_t_1);

#line 867 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":868
 *         self.features = features
 *         n = PyList_GET_SIZE(self.features)
 *         self.real_features = []             # <<<<<<<<<<<<<<
 *         self.index_list = <size_t*>malloc(sizeof(size_t) * n)
 *         if self.index_list == NULL:
 */

#line 868 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 868, __pyx_L1_error)

#line 868 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 868 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 868 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_v_self->real_features);

#line 868 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_v_self->real_features);

#line 868 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->real_features = ((PyObject*)__pyx_t_1);

#line 868 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":869
 *         n = PyList_GET_SIZE(self.features)
 *         self.real_features = []
 *         self.index_list = <size_t*>malloc(sizeof(size_t) * n)             # <<<<<<<<<<<<<<
 *         if self.index_list == NULL:
 *             PyErr_SetString(MemoryError, "Failed to allocate index list for LCMS Feature Iterator")
 */

#line 869 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->index_list = ((size_t *)malloc(((sizeof(size_t)) * __pyx_v_n)));

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":870
 *         self.real_features = []
 *         self.index_list = <size_t*>malloc(sizeof(size_t) * n)
 *         if self.index_list == NULL:             # <<<<<<<<<<<<<<
 *             PyErr_SetString(MemoryError, "Failed to allocate index list for LCMS Feature Iterator")
 *             return 1
 */

#line 870 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = ((__pyx_v_self->index_list == NULL) != 0);

#line 870 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_2) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":871
 *         self.index_list = <size_t*>malloc(sizeof(size_t) * n)
 *         if self.index_list == NULL:
 *             PyErr_SetString(MemoryError, "Failed to allocate index list for LCMS Feature Iterator")             # <<<<<<<<<<<<<<
 *             return 1
 *         self.start_time = 0
 */

#line 871 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyErr_SetString(__pyx_builtin_MemoryError, ((char *)"Failed to allocate index list for LCMS Feature Iterator"));

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":872
 *         if self.index_list == NULL:
 *             PyErr_SetString(MemoryError, "Failed to allocate index list for LCMS Feature Iterator")
 *             return 1             # <<<<<<<<<<<<<<
 *         self.start_time = 0
 *         self.end_time = INF
 */

#line 872 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_r = 1;

#line 872 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":870
 *         self.real_features = []
 *         self.index_list = <size_t*>malloc(sizeof(size_t) * n)
 *         if self.index_list == NULL:             # <<<<<<<<<<<<<<
 *             PyErr_SetString(MemoryError, "Failed to allocate index list for LCMS Feature Iterator")
 *             return 1
 */

#line 870 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":873
 *             PyErr_SetString(MemoryError, "Failed to allocate index list for LCMS Feature Iterator")
 *             return 1
 *         self.start_time = 0             # <<<<<<<<<<<<<<
 *         self.end_time = INF
 * 
 */

#line 873 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->start_time = 0.0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":874
 *             return 1
 *         self.start_time = 0
 *         self.end_time = INF             # <<<<<<<<<<<<<<
 * 
 *         for i in range(n):
 */

#line 874 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->end_time = __pyx_v_12ms_deisotope_2_c_11feature_map_12lcms_feature_INF;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":876
 *         self.end_time = INF
 * 
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             feature = <FeatureBase>PyList_GET_ITEM(self.features, i)
 *             if feature is None or isinstance(feature, EmptyFeature):
 */

#line 876 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = __pyx_v_n;

#line 876 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = __pyx_t_3;

#line 876 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {

#line 876 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_i = __pyx_t_5;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":877
 * 
 *         for i in range(n):
 *             feature = <FeatureBase>PyList_GET_ITEM(self.features, i)             # <<<<<<<<<<<<<<
 *             if feature is None or isinstance(feature, EmptyFeature):
 *                 continue
 */

#line 877 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __pyx_v_self->features;

#line 877 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 877 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_6 = PyList_GET_ITEM(__pyx_t_1, __pyx_v_i);

#line 877 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 877 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_6);

#line 877 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 877 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF_SET(__pyx_v_feature, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_t_1));

#line 877 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":878
 *         for i in range(n):
 *             feature = <FeatureBase>PyList_GET_ITEM(self.features, i)
 *             if feature is None or isinstance(feature, EmptyFeature):             # <<<<<<<<<<<<<<
 *                 continue
 *             self.real_features.append(feature)
 */

#line 878 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_7 = (((PyObject *)__pyx_v_feature) == Py_None);

#line 878 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_8 = (__pyx_t_7 != 0);

#line 878 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (!__pyx_t_8) {

#line 878 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 878 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_2 = __pyx_t_8;

#line 878 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L7_bool_binop_done;

#line 878 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 878 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_8 = __Pyx_TypeCheck(((PyObject *)__pyx_v_feature), __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature); 

#line 878 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_7 = (__pyx_t_8 != 0);

#line 878 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = __pyx_t_7;

#line 878 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L7_bool_binop_done:;

#line 878 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":879
 *             feature = <FeatureBase>PyList_GET_ITEM(self.features, i)
 *             if feature is None or isinstance(feature, EmptyFeature):
 *                 continue             # <<<<<<<<<<<<<<
 *             self.real_features.append(feature)
 *             time = feature.get_start_time()
 */

#line 879 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L4_continue;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":878
 *         for i in range(n):
 *             feature = <FeatureBase>PyList_GET_ITEM(self.features, i)
 *             if feature is None or isinstance(feature, EmptyFeature):             # <<<<<<<<<<<<<<
 *                 continue
 *             self.real_features.append(feature)
 */

#line 878 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":880
 *             if feature is None or isinstance(feature, EmptyFeature):
 *                 continue
 *             self.real_features.append(feature)             # <<<<<<<<<<<<<<
 *             time = feature.get_start_time()
 *             if (self.start_time < time):
 */

#line 880 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(__pyx_v_self->real_features == Py_None)) {

#line 880 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");

#line 880 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __PYX_ERR(0, 880, __pyx_L1_error)

#line 880 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 880 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_self->real_features, ((PyObject *)__pyx_v_feature)); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 880, __pyx_L1_error)

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":881
 *                 continue
 *             self.real_features.append(feature)
 *             time = feature.get_start_time()             # <<<<<<<<<<<<<<
 *             if (self.start_time < time):
 *                 self.start_time = time
 */

#line 881 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_time = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_feature->__pyx_vtab)->get_start_time(__pyx_v_feature);

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":882
 *             self.real_features.append(feature)
 *             time = feature.get_start_time()
 *             if (self.start_time < time):             # <<<<<<<<<<<<<<
 *                 self.start_time = time
 *             time = feature.get_end_time()
 */

#line 882 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = ((__pyx_v_self->start_time < __pyx_v_time) != 0);

#line 882 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":883
 *             time = feature.get_start_time()
 *             if (self.start_time < time):
 *                 self.start_time = time             # <<<<<<<<<<<<<<
 *             time = feature.get_end_time()
 *             if (self.end_time > time):
 */

#line 883 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_self->start_time = __pyx_v_time;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":882
 *             self.real_features.append(feature)
 *             time = feature.get_start_time()
 *             if (self.start_time < time):             # <<<<<<<<<<<<<<
 *                 self.start_time = time
 *             time = feature.get_end_time()
 */

#line 882 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":884
 *             if (self.start_time < time):
 *                 self.start_time = time
 *             time = feature.get_end_time()             # <<<<<<<<<<<<<<
 *             if (self.end_time > time):
 *                 self.end_time = time
 */

#line 884 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_time = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_feature->__pyx_vtab)->get_end_time(__pyx_v_feature);

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":885
 *                 self.start_time = time
 *             time = feature.get_end_time()
 *             if (self.end_time > time):             # <<<<<<<<<<<<<<
 *                 self.end_time = time
 *             self.index_list[i] = 0
 */

#line 885 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = ((__pyx_v_self->end_time > __pyx_v_time) != 0);

#line 885 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_2) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":886
 *             time = feature.get_end_time()
 *             if (self.end_time > time):
 *                 self.end_time = time             # <<<<<<<<<<<<<<
 *             self.index_list[i] = 0
 * 
 */

#line 886 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_self->end_time = __pyx_v_time;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":885
 *                 self.start_time = time
 *             time = feature.get_end_time()
 *             if (self.end_time > time):             # <<<<<<<<<<<<<<
 *                 self.end_time = time
 *             self.index_list[i] = 0
 */

#line 885 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":887
 *             if (self.end_time > time):
 *                 self.end_time = time
 *             self.index_list[i] = 0             # <<<<<<<<<<<<<<
 * 
 *         self.init_indices()
 */

#line 887 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    (__pyx_v_self->index_list[__pyx_v_i]) = 0;

#line 887 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L4_continue:;

#line 887 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":889
 *             self.index_list[i] = 0
 * 
 *         self.init_indices()             # <<<<<<<<<<<<<<
 *         self.last_time_seen = -1
 *         return 0
 */

#line 889 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self->__pyx_vtab)->init_indices(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 889, __pyx_L1_error)

#line 889 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 889 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":890
 * 
 *         self.init_indices()
 *         self.last_time_seen = -1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */

#line 890 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->last_time_seen = -1.0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":891
 *         self.init_indices()
 *         self.last_time_seen = -1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

#line 891 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 891 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":863
 *         self.last_time_seen = -1
 * 
 *     cdef int _initialize(self, list features) except 1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator._initialize", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 1;

#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_feature);

#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 863 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":894
 * 
 *     @staticmethod
 *     cdef FeatureSetIterator _create(list features):             # <<<<<<<<<<<<<<
 *         cdef FeatureSetIterator self
 *         self = FeatureSetIterator.__new__(FeatureSetIterator)
 */

#line 894 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 894 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator__create(PyObject *__pyx_v_features) {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self = 0
#line 894 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 894 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_r = NULL;

#line 894 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 894 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 894 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_2;

#line 894 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 894 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 894 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 894 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_create", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":896
 *     cdef FeatureSetIterator _create(list features):
 *         cdef FeatureSetIterator self
 *         self = FeatureSetIterator.__new__(FeatureSetIterator)             # <<<<<<<<<<<<<<
 *         self._initialize(features)
 *         return self
 */

#line 896 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator(((PyTypeObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 896, __pyx_L1_error)

#line 896 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));

#line 896 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_t_1);

#line 896 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":897
 *         cdef FeatureSetIterator self
 *         self = FeatureSetIterator.__new__(FeatureSetIterator)
 *         self._initialize(features)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

#line 897 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self->__pyx_vtab)->_initialize(__pyx_v_self, __pyx_v_features); if (unlikely(__pyx_t_2 == ((int)1))) __PYX_ERR(0, 897, __pyx_L1_error)

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":898
 *         self = FeatureSetIterator.__new__(FeatureSetIterator)
 *         self._initialize(features)
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

#line 898 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 898 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 898 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_v_self;

#line 898 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":894
 * 
 *     @staticmethod
 *     cdef FeatureSetIterator _create(list features):             # <<<<<<<<<<<<<<
 *         cdef FeatureSetIterator self
 *         self = FeatureSetIterator.__new__(FeatureSetIterator)
 */

#line 894 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 894 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 894 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 894 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 894 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator._create", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 894 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 894 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 894 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_self);

#line 894 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 894 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 894 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 894 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":901
 * 
 *     @staticmethod
 *     cdef FeatureSetIterator _create_with_threshold(list features, list theoretical_distribution, double detection_threshold):             # <<<<<<<<<<<<<<
 *         cdef:
 *             FeatureSetIterator self
 */

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator__create_with_threshold(PyObject *__pyx_v_features, PyObject *__pyx_v_theoretical_distribution, double __pyx_v_detection_threshold) {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self = 0
#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_p = 0
#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_i
#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  int __pyx_v_passed_threshold
#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  double __pyx_v_start_time
#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  double __pyx_v_end_time
#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_feature = NULL
#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  double __pyx_v_time
#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_r = NULL;

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_2;

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_3;

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_4;

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_5;

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_6;

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_7;

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_8;

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_9;

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_create_with_threshold", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":908
 *             bint passed_threshold
 *             double start_time, end_time
 *         self = FeatureSetIterator.__new__(FeatureSetIterator)             # <<<<<<<<<<<<<<
 *         self._initialize(features)
 * 
 */

#line 908 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator(((PyTypeObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 908, __pyx_L1_error)

#line 908 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));

#line 908 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_t_1);

#line 908 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":909
 *             double start_time, end_time
 *         self = FeatureSetIterator.__new__(FeatureSetIterator)
 *         self._initialize(features)             # <<<<<<<<<<<<<<
 * 
 *         start_time = -INF
 */

#line 909 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self->__pyx_vtab)->_initialize(__pyx_v_self, __pyx_v_features); if (unlikely(__pyx_t_2 == ((int)1))) __PYX_ERR(0, 909, __pyx_L1_error)

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":911
 *         self._initialize(features)
 * 
 *         start_time = -INF             # <<<<<<<<<<<<<<
 *         end_time = INF
 *         for i in range(self.get_size()):
 */

#line 911 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_start_time = (-__pyx_v_12ms_deisotope_2_c_11feature_map_12lcms_feature_INF);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":912
 * 
 *         start_time = -INF
 *         end_time = INF             # <<<<<<<<<<<<<<
 *         for i in range(self.get_size()):
 *             feature = self.getitem(i)
 */

#line 912 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_end_time = __pyx_v_12ms_deisotope_2_c_11feature_map_12lcms_feature_INF;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":913
 *         start_time = -INF
 *         end_time = INF
 *         for i in range(self.get_size()):             # <<<<<<<<<<<<<<
 *             feature = self.getitem(i)
 *             if feature is None or isinstance(feature, EmptyFeature):
 */

#line 913 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_get_size(__pyx_v_self);

#line 913 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = __pyx_t_3;

#line 913 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {

#line 913 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_i = __pyx_t_5;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":914
 *         end_time = INF
 *         for i in range(self.get_size()):
 *             feature = self.getitem(i)             # <<<<<<<<<<<<<<
 *             if feature is None or isinstance(feature, EmptyFeature):
 *                 continue
 */

#line 914 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_getitem(__pyx_v_self, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 914, __pyx_L1_error)

#line 914 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 914 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF_SET(__pyx_v_feature, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_t_1));

#line 914 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":915
 *         for i in range(self.get_size()):
 *             feature = self.getitem(i)
 *             if feature is None or isinstance(feature, EmptyFeature):             # <<<<<<<<<<<<<<
 *                 continue
 *             p = <TheoreticalPeak>PyList_GET_ITEM(theoretical_distribution, i)
 */

#line 915 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_7 = (((PyObject *)__pyx_v_feature) == Py_None);

#line 915 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_8 = (__pyx_t_7 != 0);

#line 915 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (!__pyx_t_8) {

#line 915 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 915 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_6 = __pyx_t_8;

#line 915 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L6_bool_binop_done;

#line 915 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 915 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_8 = __Pyx_TypeCheck(((PyObject *)__pyx_v_feature), __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature); 

#line 915 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_7 = (__pyx_t_8 != 0);

#line 915 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_6 = __pyx_t_7;

#line 915 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L6_bool_binop_done:;

#line 915 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_6) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":916
 *             feature = self.getitem(i)
 *             if feature is None or isinstance(feature, EmptyFeature):
 *                 continue             # <<<<<<<<<<<<<<
 *             p = <TheoreticalPeak>PyList_GET_ITEM(theoretical_distribution, i)
 * 
 */

#line 916 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L3_continue;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":915
 *         for i in range(self.get_size()):
 *             feature = self.getitem(i)
 *             if feature is None or isinstance(feature, EmptyFeature):             # <<<<<<<<<<<<<<
 *                 continue
 *             p = <TheoreticalPeak>PyList_GET_ITEM(theoretical_distribution, i)
 */

#line 915 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":917
 *             if feature is None or isinstance(feature, EmptyFeature):
 *                 continue
 *             p = <TheoreticalPeak>PyList_GET_ITEM(theoretical_distribution, i)             # <<<<<<<<<<<<<<
 * 
 *             passed_threshold = p.intensity > detection_threshold
 */

#line 917 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_9 = PyList_GET_ITEM(__pyx_v_theoretical_distribution, __pyx_v_i);

#line 917 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_t_9);

#line 917 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(__pyx_t_1);

#line 917 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF_SET(__pyx_v_p, ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_1));

#line 917 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":919
 *             p = <TheoreticalPeak>PyList_GET_ITEM(theoretical_distribution, i)
 * 
 *             passed_threshold = p.intensity > detection_threshold             # <<<<<<<<<<<<<<
 * 
 *             time = feature.get_start_time()
 */

#line 919 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_passed_threshold = (__pyx_v_p->intensity > __pyx_v_detection_threshold);

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":921
 *             passed_threshold = p.intensity > detection_threshold
 * 
 *             time = feature.get_start_time()             # <<<<<<<<<<<<<<
 *             if (start_time < time) and passed_threshold:
 *                 start_time = time
 */

#line 921 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_time = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_feature->__pyx_vtab)->get_start_time(__pyx_v_feature);

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":922
 * 
 *             time = feature.get_start_time()
 *             if (start_time < time) and passed_threshold:             # <<<<<<<<<<<<<<
 *                 start_time = time
 *             time = feature.get_end_time()
 */

#line 922 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_7 = ((__pyx_v_start_time < __pyx_v_time) != 0);

#line 922 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_7) {

#line 922 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 922 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_6 = __pyx_t_7;

#line 922 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L9_bool_binop_done;

#line 922 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 922 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_7 = (__pyx_v_passed_threshold != 0);

#line 922 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_6 = __pyx_t_7;

#line 922 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L9_bool_binop_done:;

#line 922 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_6) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":923
 *             time = feature.get_start_time()
 *             if (start_time < time) and passed_threshold:
 *                 start_time = time             # <<<<<<<<<<<<<<
 *             time = feature.get_end_time()
 *             if (end_time > time) and passed_threshold:
 */

#line 923 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_start_time = __pyx_v_time;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":922
 * 
 *             time = feature.get_start_time()
 *             if (start_time < time) and passed_threshold:             # <<<<<<<<<<<<<<
 *                 start_time = time
 *             time = feature.get_end_time()
 */

#line 922 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":924
 *             if (start_time < time) and passed_threshold:
 *                 start_time = time
 *             time = feature.get_end_time()             # <<<<<<<<<<<<<<
 *             if (end_time > time) and passed_threshold:
 *                 end_time = time
 */

#line 924 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_time = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_feature->__pyx_vtab)->get_end_time(__pyx_v_feature);

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":925
 *                 start_time = time
 *             time = feature.get_end_time()
 *             if (end_time > time) and passed_threshold:             # <<<<<<<<<<<<<<
 *                 end_time = time
 *         if end_time != INF:
 */

#line 925 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_7 = ((__pyx_v_end_time > __pyx_v_time) != 0);

#line 925 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_7) {

#line 925 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 925 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_6 = __pyx_t_7;

#line 925 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L12_bool_binop_done;

#line 925 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 925 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_7 = (__pyx_v_passed_threshold != 0);

#line 925 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_6 = __pyx_t_7;

#line 925 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L12_bool_binop_done:;

#line 925 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_6) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":926
 *             time = feature.get_end_time()
 *             if (end_time > time) and passed_threshold:
 *                 end_time = time             # <<<<<<<<<<<<<<
 *         if end_time != INF:
 *             self.start_time = start_time
 */

#line 926 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_end_time = __pyx_v_time;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":925
 *                 start_time = time
 *             time = feature.get_end_time()
 *             if (end_time > time) and passed_threshold:             # <<<<<<<<<<<<<<
 *                 end_time = time
 *         if end_time != INF:
 */

#line 925 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 925 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L3_continue:;

#line 925 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":927
 *             if (end_time > time) and passed_threshold:
 *                 end_time = time
 *         if end_time != INF:             # <<<<<<<<<<<<<<
 *             self.start_time = start_time
 *             self.end_time = end_time
 */

#line 927 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = ((__pyx_v_end_time != __pyx_v_12ms_deisotope_2_c_11feature_map_12lcms_feature_INF) != 0);

#line 927 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_6) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":928
 *                 end_time = time
 *         if end_time != INF:
 *             self.start_time = start_time             # <<<<<<<<<<<<<<
 *             self.end_time = end_time
 * 
 */

#line 928 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_self->start_time = __pyx_v_start_time;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":929
 *         if end_time != INF:
 *             self.start_time = start_time
 *             self.end_time = end_time             # <<<<<<<<<<<<<<
 * 
 *         self.init_indices()
 */

#line 929 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_self->end_time = __pyx_v_end_time;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":927
 *             if (end_time > time) and passed_threshold:
 *                 end_time = time
 *         if end_time != INF:             # <<<<<<<<<<<<<<
 *             self.start_time = start_time
 *             self.end_time = end_time
 */

#line 927 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":931
 *             self.end_time = end_time
 * 
 *         self.init_indices()             # <<<<<<<<<<<<<<
 *         self.last_time_seen = -1
 *         return self
 */

#line 931 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self->__pyx_vtab)->init_indices(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 931, __pyx_L1_error)

#line 931 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 931 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":932
 * 
 *         self.init_indices()
 *         self.last_time_seen = -1             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

#line 932 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->last_time_seen = -1.0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":933
 *         self.init_indices()
 *         self.last_time_seen = -1
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */

#line 933 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 933 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 933 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_v_self;

#line 933 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":901
 * 
 *     @staticmethod
 *     cdef FeatureSetIterator _create_with_threshold(list features, list theoretical_distribution, double detection_threshold):             # <<<<<<<<<<<<<<
 *         cdef:
 *             FeatureSetIterator self
 */

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator._create_with_threshold", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_self);

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_p);

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_feature);

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 901 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":935
 *         return self
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         free(self.index_list)
 * 
 */

#line 935 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 935 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 935 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static void __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_3__dealloc__(PyObject *__pyx_v_self); /*proto*/

#line 935 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static void __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_3__dealloc__(PyObject *__pyx_v_self) {

#line 935 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 935 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);

#line 935 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_2__dealloc__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self));

#line 935 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 935 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 935 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 935 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 935 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 935 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static void __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_2__dealloc__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self) {

#line 935 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 935 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":936
 * 
 *     def __dealloc__(self):
 *         free(self.index_list)             # <<<<<<<<<<<<<<
 * 
 *     cpdef init_indices(self):
 */

#line 936 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  free(__pyx_v_self->index_list);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":935
 *         return self
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         free(self.index_list)
 * 
 */

#line 935 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 935 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 935 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 935 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":938
 *         free(self.index_list)
 * 
 *     cpdef init_indices(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, ix, n
 */

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_5init_indices(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_init_indices(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, int __pyx_skip_dispatch) {
  size_t __pyx_v_i
#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_ix
#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_n
#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_f = 0
#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_node = 0
#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_5;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_6;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_7;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_8;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_9;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("init_indices", 0);

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_init_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 938, __pyx_L1_error)

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_5init_indices)) {

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_4 = NULL;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_4)) {

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 938, __pyx_L1_error)

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = __pyx_t_2;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = 0;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":943
 *             FeatureBase f
 *             LCMSFeatureTreeNode node
 *         n = self.get_size()             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             f = self.getitem(i)
 */

#line 943 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_n = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_get_size(__pyx_v_self);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":944
 *             LCMSFeatureTreeNode node
 *         n = self.get_size()
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             f = self.getitem(i)
 *             if f is None:
 */

#line 944 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = __pyx_v_n;

#line 944 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = __pyx_t_5;

#line 944 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {

#line 944 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_i = __pyx_t_7;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":945
 *         n = self.get_size()
 *         for i in range(n):
 *             f = self.getitem(i)             # <<<<<<<<<<<<<<
 *             if f is None:
 *                 self.index_list[i] = 0
 */

#line 945 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_getitem(__pyx_v_self, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 945, __pyx_L1_error)

#line 945 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 945 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF_SET(__pyx_v_f, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_t_1));

#line 945 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":946
 *         for i in range(n):
 *             f = self.getitem(i)
 *             if f is None:             # <<<<<<<<<<<<<<
 *                 self.index_list[i] = 0
 *                 continue
 */

#line 946 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_8 = (((PyObject *)__pyx_v_f) == Py_None);

#line 946 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_9 = (__pyx_t_8 != 0);

#line 946 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_9) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":947
 *             f = self.getitem(i)
 *             if f is None:
 *                 self.index_list[i] = 0             # <<<<<<<<<<<<<<
 *                 continue
 *             ix = 0
 */

#line 947 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      (__pyx_v_self->index_list[__pyx_v_i]) = 0;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":948
 *             if f is None:
 *                 self.index_list[i] = 0
 *                 continue             # <<<<<<<<<<<<<<
 *             ix = 0
 *             if f.get_size() > 0:
 */

#line 948 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L3_continue;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":946
 *         for i in range(n):
 *             f = self.getitem(i)
 *             if f is None:             # <<<<<<<<<<<<<<
 *                 self.index_list[i] = 0
 *                 continue
 */

#line 946 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":949
 *                 self.index_list[i] = 0
 *                 continue
 *             ix = 0             # <<<<<<<<<<<<<<
 *             if f.get_size() > 0:
 *                 node = f._find_time(self.start_time, &ix)
 */

#line 949 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_ix = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":950
 *                 continue
 *             ix = 0
 *             if f.get_size() > 0:             # <<<<<<<<<<<<<<
 *                 node = f._find_time(self.start_time, &ix)
 *             self.index_list[i] = ix
 */

#line 950 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_9 = ((__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_get_size(__pyx_v_f) > 0) != 0);

#line 950 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_9) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":951
 *             ix = 0
 *             if f.get_size() > 0:
 *                 node = f._find_time(self.start_time, &ix)             # <<<<<<<<<<<<<<
 *             self.index_list[i] = ix
 * 
 */

#line 951 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase__find_time(__pyx_v_f, __pyx_v_self->start_time, (&__pyx_v_ix))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 951, __pyx_L1_error)

#line 951 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 951 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_t_1));

#line 951 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = 0;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":950
 *                 continue
 *             ix = 0
 *             if f.get_size() > 0:             # <<<<<<<<<<<<<<
 *                 node = f._find_time(self.start_time, &ix)
 *             self.index_list[i] = ix
 */

#line 950 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":952
 *             if f.get_size() > 0:
 *                 node = f._find_time(self.start_time, &ix)
 *             self.index_list[i] = ix             # <<<<<<<<<<<<<<
 * 
 *     cpdef double get_next_time(self):
 */

#line 952 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    (__pyx_v_self->index_list[__pyx_v_i]) = __pyx_v_ix;

#line 952 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L3_continue:;

#line 952 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":938
 *         free(self.index_list)
 * 
 *     cpdef init_indices(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, ix, n
 */

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.init_indices", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_f);

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_node);

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_5init_indices(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_4init_indices[] = "FeatureSetIterator.init_indices(self)";

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_5init_indices(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("init_indices (wrapper)", 0);
  __pyx_r = 
#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_4init_indices(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self));

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_4init_indices(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self) {

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("init_indices", 0);

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_init_indices(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 938, __pyx_L1_error)

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.init_indices", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 938 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":954
 *             self.index_list[i] = ix
 * 
 *     cpdef double get_next_time(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double time, ix_time
 */

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_7get_next_time(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_get_next_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, int __pyx_skip_dispatch) {
  double __pyx_v_time
#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  double __pyx_v_ix_time
#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_i
#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_n
#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_ix
#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_f = 0
#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_r;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_5;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_6;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_7;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_8;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_9;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_10;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("get_next_time", 0);

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_next_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 954, __pyx_L1_error)

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_7get_next_time)) {

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_4 = NULL;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_4)) {

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 954, __pyx_L1_error)

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 954, __pyx_L1_error)

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = __pyx_t_5;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":959
 *             size_t i, n, ix
 *             FeatureBase f
 *         time = INF             # <<<<<<<<<<<<<<
 *         n = self.get_size()
 *         for i in range(n):
 */

#line 959 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_time = __pyx_v_12ms_deisotope_2_c_11feature_map_12lcms_feature_INF;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":960
 *             FeatureBase f
 *         time = INF
 *         n = self.get_size()             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             f = self.getitem(i)
 */

#line 960 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_n = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_get_size(__pyx_v_self);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":961
 *         time = INF
 *         n = self.get_size()
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             f = self.getitem(i)
 *             if f is None:
 */

#line 961 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = __pyx_v_n;

#line 961 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_7 = __pyx_t_6;

#line 961 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {

#line 961 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_i = __pyx_t_8;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":962
 *         n = self.get_size()
 *         for i in range(n):
 *             f = self.getitem(i)             # <<<<<<<<<<<<<<
 *             if f is None:
 *                 continue
 */

#line 962 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_getitem(__pyx_v_self, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 962, __pyx_L1_error)

#line 962 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 962 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF_SET(__pyx_v_f, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_t_1));

#line 962 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":963
 *         for i in range(n):
 *             f = self.getitem(i)
 *             if f is None:             # <<<<<<<<<<<<<<
 *                 continue
 *             ix = self.index_list[i]
 */

#line 963 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_9 = (((PyObject *)__pyx_v_f) == Py_None);

#line 963 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_10 = (__pyx_t_9 != 0);

#line 963 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_10) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":964
 *             f = self.getitem(i)
 *             if f is None:
 *                 continue             # <<<<<<<<<<<<<<
 *             ix = self.index_list[i]
 *             if ix < f.get_size():
 */

#line 964 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L3_continue;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":963
 *         for i in range(n):
 *             f = self.getitem(i)
 *             if f is None:             # <<<<<<<<<<<<<<
 *                 continue
 *             ix = self.index_list[i]
 */

#line 963 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":965
 *             if f is None:
 *                 continue
 *             ix = self.index_list[i]             # <<<<<<<<<<<<<<
 *             if ix < f.get_size():
 *                 ix_time = f.nodes.getitem(ix).time
 */

#line 965 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_ix = (__pyx_v_self->index_list[__pyx_v_i]);

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":966
 *                 continue
 *             ix = self.index_list[i]
 *             if ix < f.get_size():             # <<<<<<<<<<<<<<
 *                 ix_time = f.nodes.getitem(ix).time
 *                 if ix_time < time and ix_time < self.end_time:
 */

#line 966 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_10 = ((__pyx_v_ix < __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_get_size(__pyx_v_f)) != 0);

#line 966 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_10) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":967
 *             ix = self.index_list[i]
 *             if ix < f.get_size():
 *                 ix_time = f.nodes.getitem(ix).time             # <<<<<<<<<<<<<<
 *                 if ix_time < time and ix_time < self.end_time:
 *                     time = ix_time
 */

#line 967 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_getitem(__pyx_v_f->nodes, __pyx_v_ix)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 967, __pyx_L1_error)

#line 967 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 967 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_5 = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_t_1)->time;

#line 967 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 967 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_ix_time = __pyx_t_5;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":968
 *             if ix < f.get_size():
 *                 ix_time = f.nodes.getitem(ix).time
 *                 if ix_time < time and ix_time < self.end_time:             # <<<<<<<<<<<<<<
 *                     time = ix_time
 *         if time == INF:
 */

#line 968 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_9 = ((__pyx_v_ix_time < __pyx_v_time) != 0);

#line 968 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (__pyx_t_9) {

#line 968 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      } else {

#line 968 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_10 = __pyx_t_9;

#line 968 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L8_bool_binop_done;

#line 968 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 968 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_9 = ((__pyx_v_ix_time < __pyx_v_self->end_time) != 0);

#line 968 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_10 = __pyx_t_9;

#line 968 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_L8_bool_binop_done:;

#line 968 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (__pyx_t_10) {

        /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":969
 *                 ix_time = f.nodes.getitem(ix).time
 *                 if ix_time < time and ix_time < self.end_time:
 *                     time = ix_time             # <<<<<<<<<<<<<<
 *         if time == INF:
 *             return -1
 */

#line 969 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_v_time = __pyx_v_ix_time;

        /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":968
 *             if ix < f.get_size():
 *                 ix_time = f.nodes.getitem(ix).time
 *                 if ix_time < time and ix_time < self.end_time:             # <<<<<<<<<<<<<<
 *                     time = ix_time
 *         if time == INF:
 */

#line 968 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":966
 *                 continue
 *             ix = self.index_list[i]
 *             if ix < f.get_size():             # <<<<<<<<<<<<<<
 *                 ix_time = f.nodes.getitem(ix).time
 *                 if ix_time < time and ix_time < self.end_time:
 */

#line 966 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 966 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L3_continue:;

#line 966 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":970
 *                 if ix_time < time and ix_time < self.end_time:
 *                     time = ix_time
 *         if time == INF:             # <<<<<<<<<<<<<<
 *             return -1
 *         else:
 */

#line 970 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_10 = ((__pyx_v_time == __pyx_v_12ms_deisotope_2_c_11feature_map_12lcms_feature_INF) != 0);

#line 970 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_10) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":971
 *                     time = ix_time
 *         if time == INF:
 *             return -1             # <<<<<<<<<<<<<<
 *         else:
 *             return time
 */

#line 971 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_r = -1.0;

#line 971 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":970
 *                 if ix_time < time and ix_time < self.end_time:
 *                     time = ix_time
 *         if time == INF:             # <<<<<<<<<<<<<<
 *             return -1
 *         else:
 */

#line 970 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":973
 *             return -1
 *         else:
 *             return time             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint has_more(self):
 */

#line 973 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /*else*/ {

#line 973 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_r = __pyx_v_time;

#line 973 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L0;

#line 973 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":954
 *             self.index_list[i] = ix
 * 
 *     cpdef double get_next_time(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double time, ix_time
 */

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.get_next_time", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_f);

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_7get_next_time(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_6get_next_time[] = "FeatureSetIterator.get_next_time(self) -> double";

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_7get_next_time(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("get_next_time (wrapper)", 0);
  __pyx_r = 
#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_6get_next_time(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self));

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_6get_next_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self) {

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("get_next_time", 0);

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_get_next_time(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 954, __pyx_L1_error)

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.get_next_time", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 954 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":975
 *             return time
 * 
 *     cpdef bint has_more(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t j, i, n, ix
 */

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_9has_more(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_has_more(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, int __pyx_skip_dispatch) {
  size_t __pyx_v_j
#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_i
#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_n
#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_ix
#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_f = 0
#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  int __pyx_v_done
#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  int __pyx_v_at_end_time
#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_r;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_5;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_6;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_7;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_8;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_9;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("has_more", 0);

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_has_more); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 975, __pyx_L1_error)

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_9has_more)) {

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_4 = NULL;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_4)) {

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 975, __pyx_L1_error)

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 975, __pyx_L1_error)

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = __pyx_t_5;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":980
 *             FeatureBase f
 *             bint done, at_end_time
 *         j = 0             # <<<<<<<<<<<<<<
 *         n = self.get_size()
 *         for i in range(n):
 */

#line 980 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_j = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":981
 *             bint done, at_end_time
 *         j = 0
 *         n = self.get_size()             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             f = self.getitem(i)
 */

#line 981 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_n = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_get_size(__pyx_v_self);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":982
 *         j = 0
 *         n = self.get_size()
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             f = self.getitem(i)
 *             if f is None:
 */

#line 982 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = __pyx_v_n;

#line 982 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_7 = __pyx_t_6;

#line 982 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {

#line 982 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_i = __pyx_t_8;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":983
 *         n = self.get_size()
 *         for i in range(n):
 *             f = self.getitem(i)             # <<<<<<<<<<<<<<
 *             if f is None:
 *                 j += 1
 */

#line 983 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_getitem(__pyx_v_self, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 983, __pyx_L1_error)

#line 983 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 983 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF_SET(__pyx_v_f, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_t_1));

#line 983 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":984
 *         for i in range(n):
 *             f = self.getitem(i)
 *             if f is None:             # <<<<<<<<<<<<<<
 *                 j += 1
 *                 continue
 */

#line 984 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = (((PyObject *)__pyx_v_f) == Py_None);

#line 984 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_9 = (__pyx_t_5 != 0);

#line 984 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_9) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":985
 *             f = self.getitem(i)
 *             if f is None:
 *                 j += 1             # <<<<<<<<<<<<<<
 *                 continue
 *             ix = self.index_list[i]
 */

#line 985 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_j = (__pyx_v_j + 1);

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":986
 *             if f is None:
 *                 j += 1
 *                 continue             # <<<<<<<<<<<<<<
 *             ix = self.index_list[i]
 *             done = ix >= f.get_size()
 */

#line 986 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L3_continue;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":984
 *         for i in range(n):
 *             f = self.getitem(i)
 *             if f is None:             # <<<<<<<<<<<<<<
 *                 j += 1
 *                 continue
 */

#line 984 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":987
 *                 j += 1
 *                 continue
 *             ix = self.index_list[i]             # <<<<<<<<<<<<<<
 *             done = ix >= f.get_size()
 *             if not done:
 */

#line 987 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_ix = (__pyx_v_self->index_list[__pyx_v_i]);

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":988
 *                 continue
 *             ix = self.index_list[i]
 *             done = ix >= f.get_size()             # <<<<<<<<<<<<<<
 *             if not done:
 *                 at_end_time = f.nodes.getitem(ix).time >= self.end_time
 */

#line 988 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_done = (__pyx_v_ix >= __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_get_size(__pyx_v_f));

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":989
 *             ix = self.index_list[i]
 *             done = ix >= f.get_size()
 *             if not done:             # <<<<<<<<<<<<<<
 *                 at_end_time = f.nodes.getitem(ix).time >= self.end_time
 *             else:
 */

#line 989 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_9 = ((!(__pyx_v_done != 0)) != 0);

#line 989 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_9) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":990
 *             done = ix >= f.get_size()
 *             if not done:
 *                 at_end_time = f.nodes.getitem(ix).time >= self.end_time             # <<<<<<<<<<<<<<
 *             else:
 *                 at_end_time = done
 */

#line 990 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_getitem(__pyx_v_f->nodes, __pyx_v_ix)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 990, __pyx_L1_error)

#line 990 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 990 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_at_end_time = (((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_t_1)->time >= __pyx_v_self->end_time);

#line 990 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":989
 *             ix = self.index_list[i]
 *             done = ix >= f.get_size()
 *             if not done:             # <<<<<<<<<<<<<<
 *                 at_end_time = f.nodes.getitem(ix).time >= self.end_time
 *             else:
 */

#line 989 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L6;

#line 989 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":992
 *                 at_end_time = f.nodes.getitem(ix).time >= self.end_time
 *             else:
 *                 at_end_time = done             # <<<<<<<<<<<<<<
 *             if done or at_end_time:
 *                 j += 1
 */

#line 992 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    /*else*/ {

#line 992 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_at_end_time = __pyx_v_done;

#line 992 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 992 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L6:;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":993
 *             else:
 *                 at_end_time = done
 *             if done or at_end_time:             # <<<<<<<<<<<<<<
 *                 j += 1
 *         return j != n
 */

#line 993 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = (__pyx_v_done != 0);

#line 993 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (!__pyx_t_5) {

#line 993 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 993 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_9 = __pyx_t_5;

#line 993 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L8_bool_binop_done;

#line 993 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 993 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = (__pyx_v_at_end_time != 0);

#line 993 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_9 = __pyx_t_5;

#line 993 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L8_bool_binop_done:;

#line 993 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_9) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":994
 *                 at_end_time = done
 *             if done or at_end_time:
 *                 j += 1             # <<<<<<<<<<<<<<
 *         return j != n
 * 
 */

#line 994 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_j = (__pyx_v_j + 1);

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":993
 *             else:
 *                 at_end_time = done
 *             if done or at_end_time:             # <<<<<<<<<<<<<<
 *                 j += 1
 *         return j != n
 */

#line 993 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 993 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L3_continue:;

#line 993 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":995
 *             if done or at_end_time:
 *                 j += 1
 *         return j != n             # <<<<<<<<<<<<<<
 * 
 *     cdef inline size_t get_size(self):
 */

#line 995 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = (__pyx_v_j != __pyx_v_n);

#line 995 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":975
 *             return time
 * 
 *     cpdef bint has_more(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t j, i, n, ix
 */

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.has_more", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_f);

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_9has_more(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8has_more[] = "FeatureSetIterator.has_more(self) -> bool";

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_9has_more(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("has_more (wrapper)", 0);
  __pyx_r = 
#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8has_more(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self));

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8has_more(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self) {

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("has_more", 0);

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_has_more(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 975, __pyx_L1_error)

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.has_more", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 975 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":997
 *         return j != n
 * 
 *     cdef inline size_t get_size(self):             # <<<<<<<<<<<<<<
 *         return PyList_GET_SIZE(self.features)
 * 
 */

#line 997 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 997 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static CYTHON_INLINE size_t __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_get_size(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self) {

#line 997 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_r;

#line 997 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 997 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 997 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("get_size", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":998
 * 
 *     cdef inline size_t get_size(self):
 *         return PyList_GET_SIZE(self.features)             # <<<<<<<<<<<<<<
 * 
 *     cdef inline FeatureBase getitem(self, size_t i):
 */

#line 998 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_v_self->features;

#line 998 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 998 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = PyList_GET_SIZE(__pyx_t_1);

#line 998 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 998 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":997
 *         return j != n
 * 
 *     cdef inline size_t get_size(self):             # <<<<<<<<<<<<<<
 *         return PyList_GET_SIZE(self.features)
 * 
 */

#line 997 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 997 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 997 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 997 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 997 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 997 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1000
 *         return PyList_GET_SIZE(self.features)
 * 
 *     cdef inline FeatureBase getitem(self, size_t i):             # <<<<<<<<<<<<<<
 *         return <FeatureBase>PyList_GET_ITEM(self.features, i)
 * 
 */

#line 1000 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1000 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static CYTHON_INLINE struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_getitem(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, size_t __pyx_v_i) {

#line 1000 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_r = NULL;

#line 1000 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1000 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 1000 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2;

#line 1000 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("getitem", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1001
 * 
 *     cdef inline FeatureBase getitem(self, size_t i):
 *         return <FeatureBase>PyList_GET_ITEM(self.features, i)             # <<<<<<<<<<<<<<
 * 
 *     def get_index_list(self):
 */

#line 1001 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 1001 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_v_self->features;

#line 1001 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 1001 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_v_i);

#line 1001 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1001 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(((PyObject *)((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_t_2)));

#line 1001 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_t_2);

#line 1001 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1000
 *         return PyList_GET_SIZE(self.features)
 * 
 *     cdef inline FeatureBase getitem(self, size_t i):             # <<<<<<<<<<<<<<
 *         return <FeatureBase>PyList_GET_ITEM(self.features, i)
 * 
 */

#line 1000 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1000 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1000 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1000 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 1000 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1000 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1000 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1003
 *         return <FeatureBase>PyList_GET_ITEM(self.features, i)
 * 
 *     def get_index_list(self):             # <<<<<<<<<<<<<<
 *         out = []
 *         for i in range(self.get_size()):
 */

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_11get_index_list(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_10get_index_list[] = "FeatureSetIterator.get_index_list(self)";

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_11get_index_list(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("get_index_list (wrapper)", 0);
  __pyx_r = 
#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_10get_index_list(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self));

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_10get_index_list(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self) {
  PyObject *__pyx_v_out = NULL
#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_i
#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_2;

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_3;

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_4;

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_5;

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("get_index_list", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1004
 * 
 *     def get_index_list(self):
 *         out = []             # <<<<<<<<<<<<<<
 *         for i in range(self.get_size()):
 *             out.append(self.index_list[i])
 */

#line 1004 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1004, __pyx_L1_error)

#line 1004 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 1004 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_out = ((PyObject*)__pyx_t_1);

#line 1004 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1005
 *     def get_index_list(self):
 *         out = []
 *         for i in range(self.get_size()):             # <<<<<<<<<<<<<<
 *             out.append(self.index_list[i])
 *         return out
 */

#line 1005 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_get_size(__pyx_v_self);

#line 1005 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = __pyx_t_2;

#line 1005 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {

#line 1005 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_i = __pyx_t_4;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1006
 *         out = []
 *         for i in range(self.get_size()):
 *             out.append(self.index_list[i])             # <<<<<<<<<<<<<<
 *         return out
 * 
 */

#line 1006 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __Pyx_PyInt_FromSize_t((__pyx_v_self->index_list[__pyx_v_i])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1006, __pyx_L1_error)

#line 1006 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 1006 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_out, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1006, __pyx_L1_error)

#line 1006 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1006 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1007
 *         for i in range(self.get_size()):
 *             out.append(self.index_list[i])
 *         return out             # <<<<<<<<<<<<<<
 * 
 *     cpdef list get_peaks_for_time(self, double time):
 */

#line 1007 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 1007 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_out);

#line 1007 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_v_out;

#line 1007 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1003
 *         return <FeatureBase>PyList_GET_ITEM(self.features, i)
 * 
 *     def get_index_list(self):             # <<<<<<<<<<<<<<
 *         out = []
 *         for i in range(self.get_size()):
 */

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.get_index_list", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_out);

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1003 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1009
 *         return out
 * 
 *     cpdef list get_peaks_for_time(self, double time):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list peaks
 */

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_13get_peaks_for_time(PyObject *__pyx_v_self, PyObject *__pyx_arg_time); /*proto*/

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_get_peaks_for_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, double __pyx_v_time, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_peaks = 0
#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_node = 0
#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v_feature = 0
#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_i
#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_n
#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_ix
#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_6;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_7;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_8;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_9;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_10;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_11;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_12;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_13;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("get_peaks_for_time", 0);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_peaks_for_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1009, __pyx_L1_error)

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_13get_peaks_for_time)) {

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1009, __pyx_L1_error)

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_3);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_5 = NULL;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_5)) {

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_5);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_4, function);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1009, __pyx_L1_error)

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 1009, __pyx_L1_error)

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = ((PyObject*)__pyx_t_2);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = 0;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1015
 *             FeatureBase feature
 *             size_t i, n, ix
 *         peaks = []             # <<<<<<<<<<<<<<
 *         i = 0
 *         n = self.get_size()
 */

#line 1015 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1015, __pyx_L1_error)

#line 1015 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 1015 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_peaks = ((PyObject*)__pyx_t_1);

#line 1015 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1016
 *             size_t i, n, ix
 *         peaks = []
 *         i = 0             # <<<<<<<<<<<<<<
 *         n = self.get_size()
 *         for i in range(n):
 */

#line 1016 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_i = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1017
 *         peaks = []
 *         i = 0
 *         n = self.get_size()             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             feature = self.getitem(i)
 */

#line 1017 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_n = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_get_size(__pyx_v_self);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1018
 *         i = 0
 *         n = self.get_size()
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             feature = self.getitem(i)
 *             if feature is not None:
 */

#line 1018 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = __pyx_v_n;

#line 1018 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_7 = __pyx_t_6;

#line 1018 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {

#line 1018 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_i = __pyx_t_8;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1019
 *         n = self.get_size()
 *         for i in range(n):
 *             feature = self.getitem(i)             # <<<<<<<<<<<<<<
 *             if feature is not None:
 *                 if feature.get_size() > 0:
 */

#line 1019 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_getitem(__pyx_v_self, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1019, __pyx_L1_error)

#line 1019 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 1019 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF_SET(__pyx_v_feature, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_t_1));

#line 1019 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1020
 *         for i in range(n):
 *             feature = self.getitem(i)
 *             if feature is not None:             # <<<<<<<<<<<<<<
 *                 if feature.get_size() > 0:
 *                     ix = 0
 */

#line 1020 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_9 = (((PyObject *)__pyx_v_feature) != Py_None);

#line 1020 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_10 = (__pyx_t_9 != 0);

#line 1020 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_10) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1021
 *             feature = self.getitem(i)
 *             if feature is not None:
 *                 if feature.get_size() > 0:             # <<<<<<<<<<<<<<
 *                     ix = 0
 *                     node = feature._find_time(time, &ix)
 */

#line 1021 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_10 = ((__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_get_size(__pyx_v_feature) > 0) != 0);

#line 1021 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (__pyx_t_10) {

        /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1022
 *             if feature is not None:
 *                 if feature.get_size() > 0:
 *                     ix = 0             # <<<<<<<<<<<<<<
 *                     node = feature._find_time(time, &ix)
 *                     if ix == self.index_list[i] and self.index_list[i] == 0 and node is None:
 */

#line 1022 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_v_ix = 0;

        /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1023
 *                 if feature.get_size() > 0:
 *                     ix = 0
 *                     node = feature._find_time(time, &ix)             # <<<<<<<<<<<<<<
 *                     if ix == self.index_list[i] and self.index_list[i] == 0 and node is None:
 *                         pass
 */

#line 1023 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase__find_time(__pyx_v_feature, __pyx_v_time, (&__pyx_v_ix))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1023, __pyx_L1_error)

#line 1023 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_1);

#line 1023 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_t_1));

#line 1023 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_1 = 0;

        /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1024
 *                     ix = 0
 *                     node = feature._find_time(time, &ix)
 *                     if ix == self.index_list[i] and self.index_list[i] == 0 and node is None:             # <<<<<<<<<<<<<<
 *                         pass
 *                     elif ix >= self.index_list[i]:
 */

#line 1024 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_9 = ((__pyx_v_ix == (__pyx_v_self->index_list[__pyx_v_i])) != 0);

#line 1024 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (__pyx_t_9) {

#line 1024 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        } else {

#line 1024 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_10 = __pyx_t_9;

#line 1024 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          goto __pyx_L8_bool_binop_done;

#line 1024 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 1024 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_9 = (((__pyx_v_self->index_list[__pyx_v_i]) == 0) != 0);

#line 1024 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (__pyx_t_9) {

#line 1024 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        } else {

#line 1024 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_10 = __pyx_t_9;

#line 1024 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          goto __pyx_L8_bool_binop_done;

#line 1024 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 1024 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_9 = (((PyObject *)__pyx_v_node) == Py_None);

#line 1024 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_11 = (__pyx_t_9 != 0);

#line 1024 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_10 = __pyx_t_11;

#line 1024 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_L8_bool_binop_done:;

#line 1024 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (__pyx_t_10) {

#line 1024 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          goto __pyx_L7;

#line 1024 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

        /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1026
 *                     if ix == self.index_list[i] and self.index_list[i] == 0 and node is None:
 *                         pass
 *                     elif ix >= self.index_list[i]:             # <<<<<<<<<<<<<<
 *                         self.index_list[i] += 1
 *                 else:
 */

#line 1026 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_10 = ((__pyx_v_ix >= (__pyx_v_self->index_list[__pyx_v_i])) != 0);

#line 1026 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (__pyx_t_10) {

          /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1027
 *                         pass
 *                     elif ix >= self.index_list[i]:
 *                         self.index_list[i] += 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     if not isinstance(feature, EmptyFeature):
 */

#line 1027 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_12 = __pyx_v_i;

#line 1027 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          (__pyx_v_self->index_list[__pyx_t_12]) = ((__pyx_v_self->index_list[__pyx_t_12]) + 1);

          /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1026
 *                     if ix == self.index_list[i] and self.index_list[i] == 0 and node is None:
 *                         pass
 *                     elif ix >= self.index_list[i]:             # <<<<<<<<<<<<<<
 *                         self.index_list[i] += 1
 *                 else:
 */

#line 1026 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 1026 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_L7:;

        /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1021
 *             feature = self.getitem(i)
 *             if feature is not None:
 *                 if feature.get_size() > 0:             # <<<<<<<<<<<<<<
 *                     ix = 0
 *                     node = feature._find_time(time, &ix)
 */

#line 1021 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L6;

#line 1021 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1029
 *                         self.index_list[i] += 1
 *                 else:
 *                     if not isinstance(feature, EmptyFeature):             # <<<<<<<<<<<<<<
 *                         raise ValueError("Empty Feature is not of type EmptyFeature")
 *                     node = None
 */

#line 1029 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      /*else*/ {

#line 1029 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_10 = __Pyx_TypeCheck(((PyObject *)__pyx_v_feature), __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature); 

#line 1029 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_11 = ((!(__pyx_t_10 != 0)) != 0);

#line 1029 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(__pyx_t_11)) {

          /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1030
 *                 else:
 *                     if not isinstance(feature, EmptyFeature):
 *                         raise ValueError("Empty Feature is not of type EmptyFeature")             # <<<<<<<<<<<<<<
 *                     node = None
 *                 if node is not None:
 */

#line 1030 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1030, __pyx_L1_error)

#line 1030 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_GOTREF(__pyx_t_1);

#line 1030 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 1030 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1030 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __PYX_ERR(0, 1030, __pyx_L1_error)

          /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1029
 *                         self.index_list[i] += 1
 *                 else:
 *                     if not isinstance(feature, EmptyFeature):             # <<<<<<<<<<<<<<
 *                         raise ValueError("Empty Feature is not of type EmptyFeature")
 *                     node = None
 */

#line 1029 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

        /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1031
 *                     if not isinstance(feature, EmptyFeature):
 *                         raise ValueError("Empty Feature is not of type EmptyFeature")
 *                     node = None             # <<<<<<<<<<<<<<
 *                 if node is not None:
 *                     peaks.append(
 */

#line 1031 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(Py_None);

#line 1031 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)Py_None));

#line 1031 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 1031 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_L6:;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1032
 *                         raise ValueError("Empty Feature is not of type EmptyFeature")
 *                     node = None
 *                 if node is not None:             # <<<<<<<<<<<<<<
 *                     peaks.append(
 *                         node.getpeak(0))
 */

#line 1032 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_11 = (((PyObject *)__pyx_v_node) != Py_None);

#line 1032 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_10 = (__pyx_t_11 != 0);

#line 1032 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (__pyx_t_10) {

        /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1034
 *                 if node is not None:
 *                     peaks.append(
 *                         node.getpeak(0))             # <<<<<<<<<<<<<<
 *                 else:
 *                     peaks.append(None)
 */

#line 1034 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_getpeak(__pyx_v_node, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1034, __pyx_L1_error)

#line 1034 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_1);

        /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1033
 *                     node = None
 *                 if node is not None:
 *                     peaks.append(             # <<<<<<<<<<<<<<
 *                         node.getpeak(0))
 *                 else:
 */

#line 1033 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_peaks, __pyx_t_1); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 1033, __pyx_L1_error)

#line 1033 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1032
 *                         raise ValueError("Empty Feature is not of type EmptyFeature")
 *                     node = None
 *                 if node is not None:             # <<<<<<<<<<<<<<
 *                     peaks.append(
 *                         node.getpeak(0))
 */

#line 1032 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L12;

#line 1032 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1036
 *                         node.getpeak(0))
 *                 else:
 *                     peaks.append(None)             # <<<<<<<<<<<<<<
 *             else:
 *                 peaks.append(None)
 */

#line 1036 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      /*else*/ {

#line 1036 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_peaks, Py_None); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 1036, __pyx_L1_error)

#line 1036 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 1036 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_L12:;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1020
 *         for i in range(n):
 *             feature = self.getitem(i)
 *             if feature is not None:             # <<<<<<<<<<<<<<
 *                 if feature.get_size() > 0:
 *                     ix = 0
 */

#line 1020 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L5;

#line 1020 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1038
 *                     peaks.append(None)
 *             else:
 *                 peaks.append(None)             # <<<<<<<<<<<<<<
 *         return peaks
 * 
 */

#line 1038 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    /*else*/ {

#line 1038 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_peaks, Py_None); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 1038, __pyx_L1_error)

#line 1038 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 1038 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L5:;

#line 1038 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1039
 *             else:
 *                 peaks.append(None)
 *         return peaks             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

#line 1039 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 1039 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_peaks);

#line 1039 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_v_peaks;

#line 1039 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1009
 *         return out
 * 
 *     cpdef list get_peaks_for_time(self, double time):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list peaks
 */

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.get_peaks_for_time", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_peaks);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_node);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_feature);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_13get_peaks_for_time(PyObject *__pyx_v_self, PyObject *__pyx_arg_time); /*proto*/

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_12get_peaks_for_time[] = "FeatureSetIterator.get_peaks_for_time(self, double time) -> list";

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_13get_peaks_for_time(PyObject *__pyx_v_self, PyObject *__pyx_arg_time) {
  double __pyx_v_time
#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("get_peaks_for_time (wrapper)", 0);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  assert(__pyx_arg_time); {

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_time = __pyx_PyFloat_AsDouble(__pyx_arg_time); if (unlikely((__pyx_v_time == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1009, __pyx_L3_error)

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_error:;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.get_peaks_for_time", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return NULL;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_12get_peaks_for_time(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self), ((double)__pyx_v_time));

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_12get_peaks_for_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, double __pyx_v_time) {

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("get_peaks_for_time", 0);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_get_peaks_for_time(__pyx_v_self, __pyx_v_time, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1009, __pyx_L1_error)

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.get_peaks_for_time", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1009 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1042
 * 
 *     @property
 *     def current_time(self):             # <<<<<<<<<<<<<<
 *         return self.last_time_seen
 * 
 */

#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_12current_time_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_12current_time_1__get__(PyObject *__pyx_v_self) {

#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_12current_time___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self));

#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_12current_time___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self) {

#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1043
 *     @property
 *     def current_time(self):
 *         return self.last_time_seen             # <<<<<<<<<<<<<<
 * 
 *     cpdef double get_current_time(self):
 */

#line 1043 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 1043 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->last_time_seen); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1043, __pyx_L1_error)

#line 1043 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 1043 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 1043 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 1043 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1042
 * 
 *     @property
 *     def current_time(self):             # <<<<<<<<<<<<<<
 *         return self.last_time_seen
 * 
 */

#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.current_time.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1042 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1045
 *         return self.last_time_seen
 * 
 *     cpdef double get_current_time(self):             # <<<<<<<<<<<<<<
 *         return self.last_time_seen
 * 
 */

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_15get_current_time(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static double __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_get_current_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, int __pyx_skip_dispatch) {

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_r;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_5;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("get_current_time", 0);

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_current_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1045, __pyx_L1_error)

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_15get_current_time)) {

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_4 = NULL;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_4)) {

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1045, __pyx_L1_error)

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1045, __pyx_L1_error)

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = __pyx_t_5;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1046
 * 
 *     cpdef double get_current_time(self):
 *         return self.last_time_seen             # <<<<<<<<<<<<<<
 * 
 *     cpdef list get_next_value(self):
 */

#line 1046 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_v_self->last_time_seen;

#line 1046 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1045
 *         return self.last_time_seen
 * 
 *     cpdef double get_current_time(self):             # <<<<<<<<<<<<<<
 *         return self.last_time_seen
 * 
 */

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.get_current_time", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_15get_current_time(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_14get_current_time[] = "FeatureSetIterator.get_current_time(self) -> double";

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_15get_current_time(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("get_current_time (wrapper)", 0);
  __pyx_r = 
#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_14get_current_time(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self));

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_14get_current_time(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self) {

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("get_current_time", 0);

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_get_current_time(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1045, __pyx_L1_error)

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.get_current_time", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1045 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1048
 *         return self.last_time_seen
 * 
 *     cpdef list get_next_value(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double time
 */

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_17get_next_value(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_get_next_value(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, int __pyx_skip_dispatch) {
  double __pyx_v_time
#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  PyObject *__pyx_v_peaks = 0
#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_5;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("get_next_value", 0);

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_next_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1048, __pyx_L1_error)

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_17get_next_value)) {

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_4 = NULL;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_4)) {

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1048, __pyx_L1_error)

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 1048, __pyx_L1_error)

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = ((PyObject*)__pyx_t_2);

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = 0;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1052
 *             double time
 *             list peaks
 *         time = self.get_next_time()             # <<<<<<<<<<<<<<
 *         if time < 0:
 *             return None
 */

#line 1052 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_time = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self->__pyx_vtab)->get_next_time(__pyx_v_self, 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1053
 *             list peaks
 *         time = self.get_next_time()
 *         if time < 0:             # <<<<<<<<<<<<<<
 *             return None
 *         self.last_time_seen = time
 */

#line 1053 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = ((__pyx_v_time < 0.0) != 0);

#line 1053 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1054
 *         time = self.get_next_time()
 *         if time < 0:
 *             return None             # <<<<<<<<<<<<<<
 *         self.last_time_seen = time
 *         peaks = self.get_peaks_for_time(time)
 */

#line 1054 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(__pyx_r);

#line 1054 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_r = ((PyObject*)Py_None); __Pyx_INCREF(Py_None);

#line 1054 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1053
 *             list peaks
 *         time = self.get_next_time()
 *         if time < 0:             # <<<<<<<<<<<<<<
 *             return None
 *         self.last_time_seen = time
 */

#line 1053 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1055
 *         if time < 0:
 *             return None
 *         self.last_time_seen = time             # <<<<<<<<<<<<<<
 *         peaks = self.get_peaks_for_time(time)
 *         return peaks
 */

#line 1055 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->last_time_seen = __pyx_v_time;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1056
 *             return None
 *         self.last_time_seen = time
 *         peaks = self.get_peaks_for_time(time)             # <<<<<<<<<<<<<<
 *         return peaks
 * 
 */

#line 1056 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self->__pyx_vtab)->get_peaks_for_time(__pyx_v_self, __pyx_v_time, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1056, __pyx_L1_error)

#line 1056 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 1056 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_peaks = ((PyObject*)__pyx_t_1);

#line 1056 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1057
 *         self.last_time_seen = time
 *         peaks = self.get_peaks_for_time(time)
 *         return peaks             # <<<<<<<<<<<<<<
 * 
 *     def __next__(self):
 */

#line 1057 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 1057 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_peaks);

#line 1057 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_v_peaks;

#line 1057 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1048
 *         return self.last_time_seen
 * 
 *     cpdef list get_next_value(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double time
 */

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.get_next_value", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_peaks);

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_17get_next_value(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_16get_next_value[] = "FeatureSetIterator.get_next_value(self) -> list";

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_17get_next_value(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("get_next_value (wrapper)", 0);
  __pyx_r = 
#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_16get_next_value(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self));

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_16get_next_value(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self) {

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("get_next_value", 0);

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_get_next_value(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1048, __pyx_L1_error)

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.get_next_value", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1048 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1059
 *         return peaks
 * 
 *     def __next__(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list peaks
 */

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_19__next__(PyObject *__pyx_v_self); /*proto*/

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_19__next__(PyObject *__pyx_v_self) {

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__next__ (wrapper)", 0);
  __pyx_r = 
#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_18__next__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self));

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_18__next__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self) {
  PyObject *__pyx_v_peaks = 0
#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_2;

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_3;

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__next__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1062
 *         cdef:
 *             list peaks
 *         peaks = self.get_next_value()             # <<<<<<<<<<<<<<
 *         if peaks is None:
 *             raise StopIteration()
 */

#line 1062 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self->__pyx_vtab)->get_next_value(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1062, __pyx_L1_error)

#line 1062 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 1062 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_peaks = ((PyObject*)__pyx_t_1);

#line 1062 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1063
 *             list peaks
 *         peaks = self.get_next_value()
 *         if peaks is None:             # <<<<<<<<<<<<<<
 *             raise StopIteration()
 *         return peaks
 */

#line 1063 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = (__pyx_v_peaks == ((PyObject*)Py_None));

#line 1063 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 1063 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_t_3)) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1064
 *         peaks = self.get_next_value()
 *         if peaks is None:
 *             raise StopIteration()             # <<<<<<<<<<<<<<
 *         return peaks
 * 
 */

#line 1064 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_StopIteration); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1064, __pyx_L1_error)

#line 1064 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 1064 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 1064 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1064 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __PYX_ERR(0, 1064, __pyx_L1_error)

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1063
 *             list peaks
 *         peaks = self.get_next_value()
 *         if peaks is None:             # <<<<<<<<<<<<<<
 *             raise StopIteration()
 *         return peaks
 */

#line 1063 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1065
 *         if peaks is None:
 *             raise StopIteration()
 *         return peaks             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */

#line 1065 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 1065 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_v_peaks);

#line 1065 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_v_peaks;

#line 1065 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1059
 *         return peaks
 * 
 *     def __next__(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list peaks
 */

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.__next__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_peaks);

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1059 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1067
 *         return peaks
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

#line 1067 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1067 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 1067 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_21__iter__(PyObject *__pyx_v_self); /*proto*/

#line 1067 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_21__iter__(PyObject *__pyx_v_self) {

#line 1067 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 1067 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1067 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = 
#line 1067 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_20__iter__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self));

#line 1067 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1067 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1067 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1067 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1067 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 1067 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1067 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_20__iter__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self) {

#line 1067 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 1067 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1067 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1068
 * 
 *     def __iter__(self):
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */

#line 1068 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 1068 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 1068 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = ((PyObject *)__pyx_v_self);

#line 1068 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1067
 *         return peaks
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

#line 1067 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1067 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1067 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1067 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 1067 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1067 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1067 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1070
 *         return self
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "FeatureSetIterator(%d/%d, %0.3f-%0.3f)" % (
 *             len(self.real_features), len(self.features), self.start_time, self.end_time)
 */

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_23__repr__(PyObject *__pyx_v_self); /*proto*/

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_23__repr__(PyObject *__pyx_v_self) {

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = 
#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_22__repr__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self));

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_22__repr__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self) {

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_ssize_t __pyx_t_2;

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_6 = NULL;

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1071
 * 
 *     def __repr__(self):
 *         return "FeatureSetIterator(%d/%d, %0.3f-%0.3f)" % (             # <<<<<<<<<<<<<<
 *             len(self.real_features), len(self.features), self.start_time, self.end_time)
 * 
 */

#line 1071 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1072
 *     def __repr__(self):
 *         return "FeatureSetIterator(%d/%d, %0.3f-%0.3f)" % (
 *             len(self.real_features), len(self.features), self.start_time, self.end_time)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_v_self->real_features;

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_t_1);

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_t_1 == Py_None)) {

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __PYX_ERR(0, 1072, __pyx_L1_error)

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = PyList_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1072, __pyx_L1_error)

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1072, __pyx_L1_error)

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = __pyx_v_self->features;

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_t_3);

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_t_3 == Py_None)) {

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __PYX_ERR(0, 1072, __pyx_L1_error)

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = PyList_GET_SIZE(__pyx_t_3); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1072, __pyx_L1_error)

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1072, __pyx_L1_error)

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->start_time); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1072, __pyx_L1_error)

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_self->end_time); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1072, __pyx_L1_error)

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = PyTuple_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1072, __pyx_L1_error)

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_6);

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_3);

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_4);

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_4);

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_5);

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_5);

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = 0;

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = 0;

#line 1072 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1071
 * 
 *     def __repr__(self):
 *         return "FeatureSetIterator(%d/%d, %0.3f-%0.3f)" % (             # <<<<<<<<<<<<<<
 *             len(self.real_features), len(self.features), self.start_time, self.end_time)
 * 
 */

#line 1071 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_FeatureSetIterator_d_d_0_3f_0_3f, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1071, __pyx_L1_error)

#line 1071 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_5);

#line 1071 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 1071 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_5;

#line 1071 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = 0;

#line 1071 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1070
 *         return self
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "FeatureSetIterator(%d/%d, %0.3f-%0.3f)" % (
 *             len(self.real_features), len(self.features), self.start_time, self.end_time)
 */

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_6);

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1070 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":111
 * cdef class FeatureSetIterator(object):
 *     cdef:
 *         public list features             # <<<<<<<<<<<<<<
 *         public list real_features
 *         public double start_time
 */

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8features_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8features_1__get__(PyObject *__pyx_v_self) {

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = 0;

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8features___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self));

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8features___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self) {

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = NULL;

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(__pyx_v_self->features);

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = __pyx_v_self->features;

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8features_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8features_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8features_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8features_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, PyObject *__pyx_v_value) {

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  if (!(likely(PyList_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(2, 111, __pyx_L1_error)

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = __pyx_v_value;

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(__pyx_t_1);

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(__pyx_t_1);

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->features);

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(__pyx_v_self->features);

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->features = ((PyObject*)__pyx_t_1);

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = 0;

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.features.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = -1;

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8features_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8features_5__del__(PyObject *__pyx_v_self) {

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8features_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self));

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8features_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self) {

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(Py_None);

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(Py_None);

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->features);

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(__pyx_v_self->features);

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->features = ((PyObject*)Py_None);

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 111 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":112
 *     cdef:
 *         public list features
 *         public list real_features             # <<<<<<<<<<<<<<
 *         public double start_time
 *         public double end_time
 */

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_13real_features_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_13real_features_1__get__(PyObject *__pyx_v_self) {

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = 0;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_13real_features___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self));

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_13real_features___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self) {

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = NULL;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(__pyx_v_self->real_features);

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = __pyx_v_self->real_features;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_13real_features_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_13real_features_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_13real_features_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_13real_features_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, PyObject *__pyx_v_value) {

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  if (!(likely(PyList_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(2, 112, __pyx_L1_error)

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = __pyx_v_value;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(__pyx_t_1);

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(__pyx_t_1);

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->real_features);

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(__pyx_v_self->real_features);

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->real_features = ((PyObject*)__pyx_t_1);

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = 0;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.real_features.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = -1;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_13real_features_5__del__(PyObject *__pyx_v_self); /*proto*/

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_13real_features_5__del__(PyObject *__pyx_v_self) {

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = 
#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_13real_features_4__del__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self));

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_13real_features_4__del__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self) {

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__del__", 0);

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_INCREF(Py_None);

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GIVEREF(Py_None);

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_v_self->real_features);

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_DECREF(__pyx_v_self->real_features);

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->real_features = ((PyObject*)Py_None);

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 112 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":113
 *         public list features
 *         public list real_features
 *         public double start_time             # <<<<<<<<<<<<<<
 *         public double end_time
 *         public double last_time_seen
 */

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_10start_time_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_10start_time_1__get__(PyObject *__pyx_v_self) {

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = 0;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_10start_time___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self));

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_10start_time___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self) {

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = NULL;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->start_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 113, __pyx_L1_error)

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = __pyx_t_1;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = 0;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.start_time.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = NULL;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_10start_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_10start_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_10start_time_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_10start_time_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, PyObject *__pyx_v_value) {

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double __pyx_t_1;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 113, __pyx_L1_error)

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->start_time = __pyx_t_1;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.start_time.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = -1;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 113 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":114
 *         public list real_features
 *         public double start_time
 *         public double end_time             # <<<<<<<<<<<<<<
 *         public double last_time_seen
 *         size_t* index_list
 */

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8end_time_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8end_time_1__get__(PyObject *__pyx_v_self) {

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = 0;

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8end_time___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self));

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8end_time___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self) {

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = NULL;

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->end_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 114, __pyx_L1_error)

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = __pyx_t_1;

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = 0;

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.end_time.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = NULL;

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8end_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8end_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8end_time_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8end_time_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, PyObject *__pyx_v_value) {

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double __pyx_t_1;

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 114, __pyx_L1_error)

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->end_time = __pyx_t_1;

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.end_time.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = -1;

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 114 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":115
 *         public double start_time
 *         public double end_time
 *         public double last_time_seen             # <<<<<<<<<<<<<<
 *         size_t* index_list
 * 
 */

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_14last_time_seen_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_14last_time_seen_1__get__(PyObject *__pyx_v_self) {

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = 0;

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_14last_time_seen___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self));

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_14last_time_seen___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self) {

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = NULL;

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->last_time_seen); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 115, __pyx_L1_error)

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = __pyx_t_1;

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = 0;

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.last_time_seen.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = NULL;

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_14last_time_seen_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_14last_time_seen_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_14last_time_seen_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_14last_time_seen_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, PyObject *__pyx_v_value) {

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double __pyx_t_1;

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 115, __pyx_L1_error)

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->last_time_seen = __pyx_t_1;

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.last_time_seen.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = -1;

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 115 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.index_list cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_25__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_24__reduce_cython__[] = "FeatureSetIterator.__reduce_cython__(self)";

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_25__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_24__reduce_cython__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_24__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.index_list cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.index_list cannot be converted to a Python object for pickling")
 */

#line 2 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)

#line 2 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 2 "(tree fragment)"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 2 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 2 "(tree fragment)"
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.index_list cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.index_list cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.index_list cannot be converted to a Python object for pickling")
 */

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
/* Python wrapper */

#line 3 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_27__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 3 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_26__setstate_cython__[] = "FeatureSetIterator.__setstate_cython__(self, __pyx_state)";

#line 3 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_27__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 3 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 3 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 3 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 3 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_26__setstate_cython__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
  /* function exit code */

#line 3 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 3 "(tree fragment)"
  return __pyx_r;

#line 3 "(tree fragment)"
}

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_26__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {

#line 3 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 3 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 3 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 3 "(tree fragment)"
  int __pyx_lineno = 0;

#line 3 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 3 "(tree fragment)"
  int __pyx_clineno = 0;

#line 3 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self.index_list cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.index_list cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.index_list cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.index_list cannot be converted to a Python object for pickling")
 */

#line 3 "(tree fragment)"


#line 3 "(tree fragment)"
  /* function exit code */

#line 3 "(tree fragment)"
  __pyx_L1_error:;

#line 3 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 3 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 3 "(tree fragment)"
  __pyx_r = NULL;

#line 3 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 3 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 3 "(tree fragment)"
  return __pyx_r;

#line 3 "(tree fragment)"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1077
 * cdef class RunningWeightedAverage(object):
 * 
 *     cpdef _initialize(self):             # <<<<<<<<<<<<<<
 *         self.current_mean = 0
 *         self.total_weight = 0
 */

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_1_initialize(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage__initialize(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self, int __pyx_skip_dispatch) {

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_initialize", 0);

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_initialize); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1077, __pyx_L1_error)

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_1_initialize)) {

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_4 = NULL;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_4)) {

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1077, __pyx_L1_error)

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = __pyx_t_2;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = 0;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1078
 * 
 *     cpdef _initialize(self):
 *         self.current_mean = 0             # <<<<<<<<<<<<<<
 *         self.total_weight = 0
 *         self.current_count = 0
 */

#line 1078 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->current_mean = 0.0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1079
 *     cpdef _initialize(self):
 *         self.current_mean = 0
 *         self.total_weight = 0             # <<<<<<<<<<<<<<
 *         self.current_count = 0
 * 
 */

#line 1079 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->total_weight = 0.0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1080
 *         self.current_mean = 0
 *         self.total_weight = 0
 *         self.current_count = 0             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, iterable=None):
 */

#line 1080 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->current_count = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1077
 * cdef class RunningWeightedAverage(object):
 * 
 *     cpdef _initialize(self):             # <<<<<<<<<<<<<<
 *         self.current_mean = 0
 *         self.total_weight = 0
 */

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverage._initialize", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_1_initialize(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage__initialize[] = "RunningWeightedAverage._initialize(self)";

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_1_initialize(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_initialize (wrapper)", 0);
  __pyx_r = 
#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage__initialize(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_self));

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage__initialize(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self) {

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_initialize", 0);

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage__initialize(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1077, __pyx_L1_error)

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverage._initialize", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1077 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1082
 *         self.current_count = 0
 * 
 *     def __init__(self, iterable=None):             # <<<<<<<<<<<<<<
 *         self._initialize()
 *         if iterable is not None:
 */

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_3__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_3__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_iterable = 0
#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_r;

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  {

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_iterable,0};

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    PyObject* values[1] = {0};

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    values[0] = ((PyObject *)Py_None);

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(__pyx_kwds)) {

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      Py_ssize_t kw_args;

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {
        case  1: 
#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0: break;
        default: 
#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      kw_args = PyDict_Size(__pyx_kwds);

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (pos_args) {

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        case  0:

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (kw_args > 0) {

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_iterable);

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (value) { values[0] = value; kw_args--; }

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(kw_args > 0)) {

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1082, __pyx_L3_error)

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: 
#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        CYTHON_FALLTHROUGH;
        case  0: 
#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
break;
        default: 
#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
goto __pyx_L5_argtuple_error;

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_iterable = values[0];

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L4_argument_unpacking_done;

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); 
#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__PYX_ERR(0, 1082, __pyx_L3_error)

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L3_error:;

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverage.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return -1;

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_2__init__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_self), __pyx_v_iterable);

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_2__init__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self, PyObject *__pyx_v_iterable) {

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_r;

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_2;

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_3;

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1083
 * 
 *     def __init__(self, iterable=None):
 *         self._initialize()             # <<<<<<<<<<<<<<
 *         if iterable is not None:
 *             self.update(iterable)
 */

#line 1083 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_self->__pyx_vtab)->_initialize(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1083, __pyx_L1_error)

#line 1083 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 1083 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1084
 *     def __init__(self, iterable=None):
 *         self._initialize()
 *         if iterable is not None:             # <<<<<<<<<<<<<<
 *             self.update(iterable)
 * 
 */

#line 1084 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = (__pyx_v_iterable != Py_None);

#line 1084 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 1084 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_3) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1085
 *         self._initialize()
 *         if iterable is not None:
 *             self.update(iterable)             # <<<<<<<<<<<<<<
 * 
 *     cpdef reset(self):
 */

#line 1085 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_self->__pyx_vtab)->update(__pyx_v_self, __pyx_v_iterable, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1085, __pyx_L1_error)

#line 1085 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 1085 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1084
 *     def __init__(self, iterable=None):
 *         self._initialize()
 *         if iterable is not None:             # <<<<<<<<<<<<<<
 *             self.update(iterable)
 * 
 */

#line 1084 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1082
 *         self.current_count = 0
 * 
 *     def __init__(self, iterable=None):             # <<<<<<<<<<<<<<
 *         self._initialize()
 *         if iterable is not None:
 */

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverage.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = -1;

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1082 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1087
 *             self.update(iterable)
 * 
 *     cpdef reset(self):             # <<<<<<<<<<<<<<
 *         self._initialize()
 * 
 */

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_5reset(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_reset(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self, int __pyx_skip_dispatch) {

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("reset", 0);

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_reset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1087, __pyx_L1_error)

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_5reset)) {

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_4 = NULL;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_4)) {

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1087, __pyx_L1_error)

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = __pyx_t_2;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = 0;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1088
 * 
 *     cpdef reset(self):
 *         self._initialize()             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

#line 1088 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_self->__pyx_vtab)->_initialize(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1088, __pyx_L1_error)

#line 1088 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 1088 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1087
 *             self.update(iterable)
 * 
 *     cpdef reset(self):             # <<<<<<<<<<<<<<
 *         self._initialize()
 * 
 */

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverage.reset", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_5reset(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_4reset[] = "RunningWeightedAverage.reset(self)";

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_5reset(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("reset (wrapper)", 0);
  __pyx_r = 
#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_4reset(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_self));

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_4reset(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self) {

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("reset", 0);

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_reset(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1087, __pyx_L1_error)

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverage.reset", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1087 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1091
 * 
 *     @staticmethod
 *     cdef RunningWeightedAverage _create(list peaks):             # <<<<<<<<<<<<<<
 *         cdef:
 *             RunningWeightedAverage inst
 */

#line 1091 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1091 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage__create(PyObject *__pyx_v_peaks) {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_inst = 0
#line 1091 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 1091 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_r = NULL;

#line 1091 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1091 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 1091 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_2;

#line 1091 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_3;

#line 1091 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1091 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1091 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1091 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_create", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1095
 *             RunningWeightedAverage inst
 * 
 *         inst = RunningWeightedAverage.__new__(RunningWeightedAverage)             # <<<<<<<<<<<<<<
 *         inst._initialize()
 *         if peaks is None:
 */

#line 1095 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage(((PyTypeObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1095, __pyx_L1_error)

#line 1095 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));

#line 1095 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_inst = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_t_1);

#line 1095 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1096
 * 
 *         inst = RunningWeightedAverage.__new__(RunningWeightedAverage)
 *         inst._initialize()             # <<<<<<<<<<<<<<
 *         if peaks is None:
 *             return inst
 */

#line 1096 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_inst->__pyx_vtab)->_initialize(__pyx_v_inst, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1096, __pyx_L1_error)

#line 1096 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 1096 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1097
 *         inst = RunningWeightedAverage.__new__(RunningWeightedAverage)
 *         inst._initialize()
 *         if peaks is None:             # <<<<<<<<<<<<<<
 *             return inst
 *         else:
 */

#line 1097 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = (__pyx_v_peaks == ((PyObject*)Py_None));

#line 1097 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 1097 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_3) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1098
 *         inst._initialize()
 *         if peaks is None:
 *             return inst             # <<<<<<<<<<<<<<
 *         else:
 *             inst._update(peaks)
 */

#line 1098 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 1098 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(((PyObject *)__pyx_v_inst));

#line 1098 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_r = __pyx_v_inst;

#line 1098 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1097
 *         inst = RunningWeightedAverage.__new__(RunningWeightedAverage)
 *         inst._initialize()
 *         if peaks is None:             # <<<<<<<<<<<<<<
 *             return inst
 *         else:
 */

#line 1097 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1100
 *             return inst
 *         else:
 *             inst._update(peaks)             # <<<<<<<<<<<<<<
 *             return inst
 * 
 */

#line 1100 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /*else*/ {

#line 1100 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_inst->__pyx_vtab)->_update(__pyx_v_inst, __pyx_v_peaks);

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1101
 *         else:
 *             inst._update(peaks)
 *             return inst             # <<<<<<<<<<<<<<
 * 
 *     @cython.nonecheck(False)
 */

#line 1101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 1101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(((PyObject *)__pyx_v_inst));

#line 1101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_r = __pyx_v_inst;

#line 1101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L0;

#line 1101 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1091
 * 
 *     @staticmethod
 *     cdef RunningWeightedAverage _create(list peaks):             # <<<<<<<<<<<<<<
 *         cdef:
 *             RunningWeightedAverage inst
 */

#line 1091 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1091 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1091 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1091 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 1091 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverage._create", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1091 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 1091 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1091 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);

#line 1091 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 1091 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1091 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1091 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1105
 *     @cython.nonecheck(False)
 *     @cython.cdivision(True)
 *     cpdef add(self, PeakBase peak):             # <<<<<<<<<<<<<<
 *         if peak.intensity == 0:
 *             if self.current_mean == 0 and self.total_weight == 0:
 */

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_7add(PyObject *__pyx_v_self, PyObject *__pyx_v_peak); /*proto*/

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_add(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak, int __pyx_skip_dispatch) {
  double __pyx_v_agg
#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_5;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_6;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_7;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("add", 0);

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_add); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1105, __pyx_L1_error)

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_7add)) {

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_4 = NULL;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_4)) {

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_peak)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_peak));

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1105, __pyx_L1_error)

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = __pyx_t_2;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = 0;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1106
 *     @cython.cdivision(True)
 *     cpdef add(self, PeakBase peak):
 *         if peak.intensity == 0:             # <<<<<<<<<<<<<<
 *             if self.current_mean == 0 and self.total_weight == 0:
 *                 self.current_mean = peak.mz
 */

#line 1106 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = ((__pyx_v_peak->intensity == 0.0) != 0);

#line 1106 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1107
 *     cpdef add(self, PeakBase peak):
 *         if peak.intensity == 0:
 *             if self.current_mean == 0 and self.total_weight == 0:             # <<<<<<<<<<<<<<
 *                 self.current_mean = peak.mz
 *                 self.total_weight = 1
 */

#line 1107 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_6 = ((__pyx_v_self->current_mean == 0.0) != 0);

#line 1107 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_6) {

#line 1107 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 1107 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_5 = __pyx_t_6;

#line 1107 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L5_bool_binop_done;

#line 1107 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 1107 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_6 = ((__pyx_v_self->total_weight == 0.0) != 0);

#line 1107 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = __pyx_t_6;

#line 1107 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L5_bool_binop_done:;

#line 1107 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1108
 *         if peak.intensity == 0:
 *             if self.current_mean == 0 and self.total_weight == 0:
 *                 self.current_mean = peak.mz             # <<<<<<<<<<<<<<
 *                 self.total_weight = 1
 *             else:
 */

#line 1108 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_7 = __pyx_v_peak->mz;

#line 1108 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_self->current_mean = __pyx_t_7;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1109
 *             if self.current_mean == 0 and self.total_weight == 0:
 *                 self.current_mean = peak.mz
 *                 self.total_weight = 1             # <<<<<<<<<<<<<<
 *             else:
 *                 return
 */

#line 1109 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_self->total_weight = 1.0;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1107
 *     cpdef add(self, PeakBase peak):
 *         if peak.intensity == 0:
 *             if self.current_mean == 0 and self.total_weight == 0:             # <<<<<<<<<<<<<<
 *                 self.current_mean = peak.mz
 *                 self.total_weight = 1
 */

#line 1107 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L4;

#line 1107 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1111
 *                 self.total_weight = 1
 *             else:
 *                 return             # <<<<<<<<<<<<<<
 * 
 *         agg = (self.total_weight * self.current_mean) + \
 */

#line 1111 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    /*else*/ {

#line 1111 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_XDECREF(__pyx_r);

#line 1111 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 1111 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L0;

#line 1111 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 1111 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L4:;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1106
 *     @cython.cdivision(True)
 *     cpdef add(self, PeakBase peak):
 *         if peak.intensity == 0:             # <<<<<<<<<<<<<<
 *             if self.current_mean == 0 and self.total_weight == 0:
 *                 self.current_mean = peak.mz
 */

#line 1106 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1113
 *                 return
 * 
 *         agg = (self.total_weight * self.current_mean) + \             # <<<<<<<<<<<<<<
 *             (peak.mz * peak.intensity)
 *         self.total_weight += peak.intensity
 */

#line 1113 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_agg = ((__pyx_v_self->total_weight * __pyx_v_self->current_mean) + (__pyx_v_peak->mz * __pyx_v_peak->intensity));

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1115
 *         agg = (self.total_weight * self.current_mean) + \
 *             (peak.mz * peak.intensity)
 *         self.total_weight += peak.intensity             # <<<<<<<<<<<<<<
 *         self.current_count += 1
 * 
 */

#line 1115 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->total_weight = (__pyx_v_self->total_weight + __pyx_v_peak->intensity);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1116
 *             (peak.mz * peak.intensity)
 *         self.total_weight += peak.intensity
 *         self.current_count += 1             # <<<<<<<<<<<<<<
 * 
 *         if self.total_weight != 0:
 */

#line 1116 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->current_count = (__pyx_v_self->current_count + 1);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1118
 *         self.current_count += 1
 * 
 *         if self.total_weight != 0:             # <<<<<<<<<<<<<<
 *             self.current_mean = agg / (self.total_weight)
 *         else:
 */

#line 1118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = ((__pyx_v_self->total_weight != 0.0) != 0);

#line 1118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1119
 * 
 *         if self.total_weight != 0:
 *             self.current_mean = agg / (self.total_weight)             # <<<<<<<<<<<<<<
 *         else:
 *             print("NaN produced in add()")
 */

#line 1119 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_self->current_mean = (__pyx_v_agg / __pyx_v_self->total_weight);

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1118
 *         self.current_count += 1
 * 
 *         if self.total_weight != 0:             # <<<<<<<<<<<<<<
 *             self.current_mean = agg / (self.total_weight)
 *         else:
 */

#line 1118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L7;

#line 1118 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1121
 *             self.current_mean = agg / (self.total_weight)
 *         else:
 *             print("NaN produced in add()")             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

#line 1121 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /*else*/ {

#line 1121 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__Pyx_PrintOne(0, __pyx_kp_s_NaN_produced_in_add) < 0) __PYX_ERR(0, 1121, __pyx_L1_error)

#line 1121 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 1121 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L7:;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1122
 *         else:
 *             print("NaN produced in add()")
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cpdef int feed_from_feature(self, LCMSFeature feature):
 */

#line 1122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 1122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 1122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = ((PyObject *)__pyx_v_self);

#line 1122 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1105
 *     @cython.nonecheck(False)
 *     @cython.cdivision(True)
 *     cpdef add(self, PeakBase peak):             # <<<<<<<<<<<<<<
 *         if peak.intensity == 0:
 *             if self.current_mean == 0 and self.total_weight == 0:
 */

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverage.add", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_7add(PyObject *__pyx_v_self, PyObject *__pyx_v_peak); /*proto*/

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_6add[] = "RunningWeightedAverage.add(self, PeakBase peak)";

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_7add(PyObject *__pyx_v_self, PyObject *__pyx_v_peak) {

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("add (wrapper)", 0);

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase, 1, "peak", 0))) __PYX_ERR(0, 1105, __pyx_L1_error)
  __pyx_r = 
#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_6add(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_self), ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_peak));

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_6add(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak) {

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("add", 0);

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_add(__pyx_v_self, __pyx_v_peak, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1105, __pyx_L1_error)

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverage.add", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1105 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1124
 *         return self
 * 
 *     cpdef int feed_from_feature(self, LCMSFeature feature):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, j, n, m
 */

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_9feed_from_feature(PyObject *__pyx_v_self, PyObject *__pyx_v_feature); /*proto*/

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static int __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_feed_from_feature(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_feature, int __pyx_skip_dispatch) {
  size_t __pyx_v_i
#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_j
#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_n
#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_m
#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *__pyx_v_node = 0
#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak = 0
#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_r;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_5;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_6;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_7;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_8;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_9;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_10;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_11;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("feed_from_feature", 0);

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_feed_from_feature); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1124, __pyx_L1_error)

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_9feed_from_feature)) {

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_4 = NULL;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_4)) {

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_feature)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_feature));

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1124, __pyx_L1_error)

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1124, __pyx_L1_error)

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = __pyx_t_5;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1130
 *             PeakBase peak
 * 
 *         n = feature.get_size()             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             node = feature.getitem(i)
 */

#line 1130 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_n = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_feature->__pyx_base.__pyx_vtab)->__pyx_base.get_size(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v_feature));

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1131
 * 
 *         n = feature.get_size()
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             node = feature.getitem(i)
 *             m = node.get_members_size()
 */

#line 1131 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_6 = __pyx_v_n;

#line 1131 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_7 = __pyx_t_6;

#line 1131 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {

#line 1131 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_i = __pyx_t_8;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1132
 *         n = feature.get_size()
 *         for i in range(n):
 *             node = feature.getitem(i)             # <<<<<<<<<<<<<<
 *             m = node.get_members_size()
 *             for j in range(m):
 */

#line 1132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_feature->__pyx_base.__pyx_vtab)->getitem(__pyx_v_feature, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1132, __pyx_L1_error)

#line 1132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 1132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF_SET(__pyx_v_node, ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_t_1));

#line 1132 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1133
 *         for i in range(n):
 *             node = feature.getitem(i)
 *             m = node.get_members_size()             # <<<<<<<<<<<<<<
 *             for j in range(m):
 *                 peak = node.getitem(j)
 */

#line 1133 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_m = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_v_node->__pyx_base.__pyx_vtab)->__pyx_base.get_members_size(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_node));

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1134
 *             node = feature.getitem(i)
 *             m = node.get_members_size()
 *             for j in range(m):             # <<<<<<<<<<<<<<
 *                 peak = node.getitem(j)
 *                 self.add(peak)
 */

#line 1134 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_9 = __pyx_v_m;

#line 1134 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_10 = __pyx_t_9;

#line 1134 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {

#line 1134 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_j = __pyx_t_11;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1135
 *             m = node.get_members_size()
 *             for j in range(m):
 *                 peak = node.getitem(j)             # <<<<<<<<<<<<<<
 *                 self.add(peak)
 *         return n
 */

#line 1135 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)__pyx_v_node->__pyx_base.__pyx_vtab)->__pyx_base.getitem(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)__pyx_v_node), __pyx_v_j)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1135, __pyx_L1_error)

#line 1135 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 1135 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_XDECREF_SET(__pyx_v_peak, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_1));

#line 1135 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = 0;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1136
 *             for j in range(m):
 *                 peak = node.getitem(j)
 *                 self.add(peak)             # <<<<<<<<<<<<<<
 *         return n
 * 
 */

#line 1136 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_self->__pyx_vtab)->add(__pyx_v_self, __pyx_v_peak, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1136, __pyx_L1_error)

#line 1136 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 1136 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1136 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 1136 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1137
 *                 peak = node.getitem(j)
 *                 self.add(peak)
 *         return n             # <<<<<<<<<<<<<<
 * 
 *     cpdef RunningWeightedAverage update(self, iterable):
 */

#line 1137 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_v_n;

#line 1137 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1124
 *         return self
 * 
 *     cpdef int feed_from_feature(self, LCMSFeature feature):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, j, n, m
 */

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverage.feed_from_feature", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_node);

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_peak);

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_9feed_from_feature(PyObject *__pyx_v_self, PyObject *__pyx_v_feature); /*proto*/

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_8feed_from_feature[] = "RunningWeightedAverage.feed_from_feature(self, LCMSFeature feature) -> int";

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_9feed_from_feature(PyObject *__pyx_v_self, PyObject *__pyx_v_feature) {

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("feed_from_feature (wrapper)", 0);

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_feature), __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature, 1, "feature", 0))) __PYX_ERR(0, 1124, __pyx_L1_error)
  __pyx_r = 
#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_8feed_from_feature(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_self), ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)__pyx_v_feature));

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_8feed_from_feature(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *__pyx_v_feature) {

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("feed_from_feature", 0);

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_feed_from_feature(__pyx_v_self, __pyx_v_feature, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1124, __pyx_L1_error)

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverage.feed_from_feature", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1124 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1139
 *         return n
 * 
 *     cpdef RunningWeightedAverage update(self, iterable):             # <<<<<<<<<<<<<<
 *         for x in iterable:
 *             self.add(x)
 */

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_11update(PyObject *__pyx_v_self, PyObject *__pyx_v_iterable); /*proto*/

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_update(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self, PyObject *__pyx_v_iterable, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_x = NULL
#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_r = NULL;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  Py_ssize_t __pyx_t_5;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *(*__pyx_t_6)(PyObject *);

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("update", 0);

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_update); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1139, __pyx_L1_error)

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_11update)) {

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_4 = NULL;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_4)) {

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_iterable) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_iterable);

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1139, __pyx_L1_error)

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage))))) __PYX_ERR(0, 1139, __pyx_L1_error)

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_t_2);

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = 0;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1140
 * 
 *     cpdef RunningWeightedAverage update(self, iterable):
 *         for x in iterable:             # <<<<<<<<<<<<<<
 *             self.add(x)
 *         return self
 */

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (likely(PyList_CheckExact(__pyx_v_iterable)) || PyTuple_CheckExact(__pyx_v_iterable)) {

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_1 = __pyx_v_iterable; __Pyx_INCREF(__pyx_t_1); __pyx_t_5 = 0;

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_6 = NULL;

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  } else {
    __pyx_t_5 = -1; 
#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_1 = PyObject_GetIter(__pyx_v_iterable); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1140, __pyx_L1_error)

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_1);

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_6 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1140, __pyx_L1_error)

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  for (;;) {

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (likely(!__pyx_t_6)) {

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (likely(PyList_CheckExact(__pyx_t_1))) {

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_1)) break;

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_2); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 1140, __pyx_L1_error)

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #else

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1140, __pyx_L1_error)

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #endif

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      } else {

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_1)) break;

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_2); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 1140, __pyx_L1_error)

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #else

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1140, __pyx_L1_error)

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        #endif

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }
    } else 
#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
{

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_2 = __pyx_t_6(__pyx_t_1);

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(!__pyx_t_2)) {

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        PyObject* exc_type = PyErr_Occurred();

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (exc_type) {

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          else __PYX_ERR(0, 1140, __pyx_L1_error)

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        break;

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_2);

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF_SET(__pyx_v_x, __pyx_t_2);

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1141
 *     cpdef RunningWeightedAverage update(self, iterable):
 *         for x in iterable:
 *             self.add(x)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

#line 1141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (!(likely(((__pyx_v_x) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_x, __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase))))) __PYX_ERR(0, 1141, __pyx_L1_error)

#line 1141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_2 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_self->__pyx_vtab)->add(__pyx_v_self, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_x), 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1141, __pyx_L1_error)

#line 1141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_2);

#line 1141 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1140
 * 
 *     cpdef RunningWeightedAverage update(self, iterable):
 *         for x in iterable:             # <<<<<<<<<<<<<<
 *             self.add(x)
 *         return self
 */

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 1140 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1142
 *         for x in iterable:
 *             self.add(x)
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cdef void _update(self, list peaks):
 */

#line 1142 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 1142 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 1142 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_v_self;

#line 1142 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1139
 *         return n
 * 
 *     cpdef RunningWeightedAverage update(self, iterable):             # <<<<<<<<<<<<<<
 *         for x in iterable:
 *             self.add(x)
 */

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverage.update", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_v_x);

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_11update(PyObject *__pyx_v_self, PyObject *__pyx_v_iterable); /*proto*/

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_10update[] = "RunningWeightedAverage.update(self, iterable) -> RunningWeightedAverage";

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_11update(PyObject *__pyx_v_self, PyObject *__pyx_v_iterable) {

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("update (wrapper)", 0);
  __pyx_r = 
#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_10update(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_self), ((PyObject *)__pyx_v_iterable));

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_10update(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self, PyObject *__pyx_v_iterable) {

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("update", 0);

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_update(__pyx_v_self, __pyx_v_iterable, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1139, __pyx_L1_error)

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverage.update", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1139 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1144
 *         return self
 * 
 *     cdef void _update(self, list peaks):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

#line 1144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static void __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage__update(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self, PyObject *__pyx_v_peaks) {
  size_t __pyx_v_i
#line 1144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  size_t __pyx_v_n
#line 1144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;
  struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_p = 0
#line 1144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 1144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_1;

#line 1144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_2;

#line 1144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  size_t __pyx_t_3;

#line 1144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4;

#line 1144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_5 = NULL;

#line 1144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_update", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1148
 *             size_t i, n
 *             PeakBase p
 *         n = PyList_GET_SIZE(peaks)             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             p = <PeakBase>PyList_GET_ITEM(peaks, i)
 */

#line 1148 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_n = PyList_GET_SIZE(__pyx_v_peaks);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1149
 *             PeakBase p
 *         n = PyList_GET_SIZE(peaks)
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             p = <PeakBase>PyList_GET_ITEM(peaks, i)
 *             self.add(p)
 */

#line 1149 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_v_n;

#line 1149 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __pyx_t_1;

#line 1149 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {

#line 1149 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_i = __pyx_t_3;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1150
 *         n = PyList_GET_SIZE(peaks)
 *         for i in range(n):
 *             p = <PeakBase>PyList_GET_ITEM(peaks, i)             # <<<<<<<<<<<<<<
 *             self.add(p)
 * 
 */

#line 1150 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_4 = PyList_GET_ITEM(__pyx_v_peaks, __pyx_v_i);

#line 1150 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = ((PyObject *)__pyx_t_4);

#line 1150 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(__pyx_t_5);

#line 1150 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF_SET(__pyx_v_p, ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_t_5));

#line 1150 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = 0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1151
 *         for i in range(n):
 *             p = <PeakBase>PyList_GET_ITEM(peaks, i)
 *             self.add(p)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */

#line 1151 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_self->__pyx_vtab)->add(__pyx_v_self, __pyx_v_p, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1151, __pyx_L1_error)

#line 1151 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_GOTREF(__pyx_t_5);

#line 1151 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 1151 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1144
 *         return self
 * 
 *     cdef void _update(self, list peaks):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i, n
 */

#line 1144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 1144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_5);

#line 1144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_WriteUnraisable("ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverage._update", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);

#line 1144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_p);

#line 1144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1144 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1153
 *             self.add(p)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "%s(%r, %d)" % (self.__class__.__name__, self.current_mean, self.current_count)
 * 
 */

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_13__repr__(PyObject *__pyx_v_self); /*proto*/

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_13__repr__(PyObject *__pyx_v_self) {

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = 
#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_12__repr__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_self));

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_12__repr__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self) {

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1154
 * 
 *     def __repr__(self):
 *         return "%s(%r, %d)" % (self.__class__.__name__, self.current_mean, self.current_count)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 1154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1154, __pyx_L1_error)

#line 1154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 1154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1154, __pyx_L1_error)

#line 1154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 1154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->current_mean); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1154, __pyx_L1_error)

#line 1154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 1154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_self->current_count); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1154, __pyx_L1_error)

#line 1154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 1154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1154, __pyx_L1_error)

#line 1154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_4);

#line 1154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 1154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);

#line 1154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 1154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);

#line 1154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_3);

#line 1154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);

#line 1154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = 0;

#line 1154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 1154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = 0;

#line 1154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_s_r_d, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1154, __pyx_L1_error)

#line 1154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_3);

#line 1154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 1154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_3;

#line 1154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = 0;

#line 1154 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1153
 *             self.add(p)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "%s(%r, %d)" % (self.__class__.__name__, self.current_mean, self.current_count)
 * 
 */

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverage.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1153 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":139
 * cdef class RunningWeightedAverage(object):
 *     cdef:
 *         public double current_mean             # <<<<<<<<<<<<<<
 *         public size_t current_count
 *         public double total_weight
 */

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_12current_mean_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_12current_mean_1__get__(PyObject *__pyx_v_self) {

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = 0;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_12current_mean___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_self));

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_12current_mean___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self) {

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = NULL;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->current_mean); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 139, __pyx_L1_error)

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = __pyx_t_1;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = 0;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverage.current_mean.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = NULL;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_12current_mean_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_12current_mean_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_12current_mean_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_12current_mean_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self, PyObject *__pyx_v_value) {

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double __pyx_t_1;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 139, __pyx_L1_error)

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->current_mean = __pyx_t_1;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverage.current_mean.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = -1;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 139 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":140
 *     cdef:
 *         public double current_mean
 *         public size_t current_count             # <<<<<<<<<<<<<<
 *         public double total_weight
 * 
 */

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_13current_count_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_13current_count_1__get__(PyObject *__pyx_v_self) {

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = 0;

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_13current_count___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_self));

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_13current_count___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self) {

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = NULL;

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->current_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 140, __pyx_L1_error)

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = __pyx_t_1;

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = 0;

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverage.current_count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = NULL;

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_13current_count_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_13current_count_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_13current_count_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_13current_count_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self, PyObject *__pyx_v_value) {

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  size_t __pyx_t_1;

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = __Pyx_PyInt_As_size_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 140, __pyx_L1_error)

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->current_count = __pyx_t_1;

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverage.current_count.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = -1;

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 140 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pxd":141
 *         public double current_mean
 *         public size_t current_count
 *         public double total_weight             # <<<<<<<<<<<<<<
 * 
 *     cpdef reset(self)
 */

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_12total_weight_1__get__(PyObject *__pyx_v_self); /*proto*/

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_12total_weight_1__get__(PyObject *__pyx_v_self) {

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = 0;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = 
#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_12total_weight___get__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_self));

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_12total_weight___get__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self) {

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_r = NULL;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__get__", 0);

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->total_weight); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 141, __pyx_L1_error)

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = __pyx_t_1;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = 0;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverage.total_weight.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = NULL;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
/* Python wrapper */

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_12total_weight_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_12total_weight_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = 
#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_12total_weight_2__set__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_self), ((PyObject *)__pyx_v_value));

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
static int __pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_12total_weight_2__set__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self, PyObject *__pyx_v_value) {

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_r;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyDeclarations

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  double __pyx_t_1;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_lineno = 0;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  const char *__pyx_filename = NULL;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  int __pyx_clineno = 0;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannySetupContext("__set__", 0);

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 141, __pyx_L1_error)

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_v_self->total_weight = __pyx_t_1;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"


#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  /* function exit code */

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = 0;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  goto __pyx_L0;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L1_error:;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverage.total_weight.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_r = -1;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __pyx_L0:;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  __Pyx_RefNannyFinishContext();

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
  return __pyx_r;

#line 141 "ms_deisotope/_c/feature_map/lcms_feature.pxd"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_15__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_14__reduce_cython__[] = "RunningWeightedAverage.__reduce_cython__(self)";

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_15__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_14__reduce_cython__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_14__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self) {
  PyObject *__pyx_v_state = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v__dict = 0
#line 1 "(tree fragment)"
;
  int __pyx_v_use_setstate
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_2 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_3 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_5;

#line 1 "(tree fragment)"
  int __pyx_t_6;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.current_count, self.current_mean, self.total_weight)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */

#line 5 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->current_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->current_mean); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_2);

#line 5 "(tree fragment)"
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->total_weight); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_3);

#line 5 "(tree fragment)"
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_2);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_3);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);

#line 5 "(tree fragment)"
  __pyx_t_1 = 0;

#line 5 "(tree fragment)"
  __pyx_t_2 = 0;

#line 5 "(tree fragment)"
  __pyx_t_3 = 0;

#line 5 "(tree fragment)"
  __pyx_v_state = ((PyObject*)__pyx_t_4);

#line 5 "(tree fragment)"
  __pyx_t_4 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.current_count, self.current_mean, self.total_weight)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */

#line 6 "(tree fragment)"
  __pyx_t_4 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 6 "(tree fragment)"
  __pyx_v__dict = __pyx_t_4;

#line 6 "(tree fragment)"
  __pyx_t_4 = 0;

  /* "(tree fragment)":7
 *     state = (self.current_count, self.current_mean, self.total_weight)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
  __pyx_t_5 = (__pyx_v__dict != Py_None);

#line 7 "(tree fragment)"
  __pyx_t_6 = (__pyx_t_5 != 0);

#line 7 "(tree fragment)"
  if (__pyx_t_6) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */

#line 8 "(tree fragment)"
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 8 "(tree fragment)"
    __Pyx_INCREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v__dict);

#line 8 "(tree fragment)"
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 8 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 8 "(tree fragment)"
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));

#line 8 "(tree fragment)"
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */

#line 9 "(tree fragment)"
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.current_count, self.current_mean, self.total_weight)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
    goto __pyx_L3;

#line 7 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_RunningWeightedAverage, (type(self), 0x69a8b2a, None), state
 */

#line 11 "(tree fragment)"
  /*else*/ {

#line 11 "(tree fragment)"
    __pyx_v_use_setstate = 0;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_RunningWeightedAverage, (type(self), 0x69a8b2a, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  __pyx_t_6 = (__pyx_v_use_setstate != 0);

#line 12 "(tree fragment)"
  if (__pyx_t_6) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_RunningWeightedAverage, (type(self), 0x69a8b2a, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_RunningWeightedAverage, (type(self), 0x69a8b2a, state)
 */

#line 13 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 13 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_RunningWeightedAv); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 13 "(tree fragment)"
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 13 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_110791466);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_110791466);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_110791466);

#line 13 "(tree fragment)"
    __Pyx_INCREF(Py_None);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(Py_None);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 2, Py_None);

#line 13 "(tree fragment)"
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_3);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_4);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state);

#line 13 "(tree fragment)"
    __pyx_t_3 = 0;

#line 13 "(tree fragment)"
    __pyx_t_4 = 0;

#line 13 "(tree fragment)"
    __pyx_r = __pyx_t_2;

#line 13 "(tree fragment)"
    __pyx_t_2 = 0;

#line 13 "(tree fragment)"
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_RunningWeightedAverage, (type(self), 0x69a8b2a, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_RunningWeightedAverage, (type(self), 0x69a8b2a, None), state
 *     else:
 *         return __pyx_unpickle_RunningWeightedAverage, (type(self), 0x69a8b2a, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_RunningWeightedAverage__set_state(self, __pyx_state)
 */

#line 15 "(tree fragment)"
  /*else*/ {

#line 15 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 15 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_RunningWeightedAv); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 15 "(tree fragment)"
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 15 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_110791466);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_110791466);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_110791466);

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state);

#line 15 "(tree fragment)"
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_2);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_4);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);

#line 15 "(tree fragment)"
    __pyx_t_2 = 0;

#line 15 "(tree fragment)"
    __pyx_t_4 = 0;

#line 15 "(tree fragment)"
    __pyx_r = __pyx_t_3;

#line 15 "(tree fragment)"
    __pyx_t_3 = 0;

#line 15 "(tree fragment)"
    goto __pyx_L0;

#line 15 "(tree fragment)"
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_2);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_3);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverage.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v_state);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v__dict);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_RunningWeightedAverage, (type(self), 0x69a8b2a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_RunningWeightedAverage__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
/* Python wrapper */

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_17__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 16 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_16__setstate_cython__[] = "RunningWeightedAverage.__setstate_cython__(self, __pyx_state)";

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_17__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 16 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_16__setstate_cython__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_16__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 16 "(tree fragment)"
  int __pyx_lineno = 0;

#line 16 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 16 "(tree fragment)"
  int __pyx_clineno = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_RunningWeightedAverage, (type(self), 0x69a8b2a, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_RunningWeightedAverage__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */

#line 17 "(tree fragment)"
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_unpickle_RunningWeightedAverage__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 17 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_RunningWeightedAverage, (type(self), 0x69a8b2a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_RunningWeightedAverage__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 16 "(tree fragment)"
  goto __pyx_L0;

#line 16 "(tree fragment)"
  __pyx_L1_error:;

#line 16 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 16 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverage.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 16 "(tree fragment)"
  __pyx_r = NULL;

#line 16 "(tree fragment)"
  __pyx_L0:;

#line 16 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1160
 * 
 *     @staticmethod
 *     cdef RunningWeightedAverageNeutralMass _create(list peaks):             # <<<<<<<<<<<<<<
 *         cdef:
 *             RunningWeightedAverageNeutralMass inst
 */

#line 1160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass__create(PyObject *__pyx_v_peaks) {
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *__pyx_v_inst = 0
#line 1160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 1160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *__pyx_r = NULL;

#line 1160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 1160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_2;

#line 1160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_3;

#line 1160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("_create", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1164
 *             RunningWeightedAverageNeutralMass inst
 * 
 *         inst = RunningWeightedAverageNeutralMass.__new__(RunningWeightedAverageNeutralMass)             # <<<<<<<<<<<<<<
 *         inst._initialize()
 *         if peaks is None:
 */

#line 1164 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass(((PyTypeObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1164, __pyx_L1_error)

#line 1164 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));

#line 1164 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_inst = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *)__pyx_t_1);

#line 1164 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1165
 * 
 *         inst = RunningWeightedAverageNeutralMass.__new__(RunningWeightedAverageNeutralMass)
 *         inst._initialize()             # <<<<<<<<<<<<<<
 *         if peaks is None:
 *             return inst
 */

#line 1165 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *)__pyx_v_inst->__pyx_base.__pyx_vtab)->__pyx_base._initialize(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_inst), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1165, __pyx_L1_error)

#line 1165 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 1165 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1166
 *         inst = RunningWeightedAverageNeutralMass.__new__(RunningWeightedAverageNeutralMass)
 *         inst._initialize()
 *         if peaks is None:             # <<<<<<<<<<<<<<
 *             return inst
 *         else:
 */

#line 1166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = (__pyx_v_peaks == ((PyObject*)Py_None));

#line 1166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 1166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_3) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1167
 *         inst._initialize()
 *         if peaks is None:
 *             return inst             # <<<<<<<<<<<<<<
 *         else:
 *             inst._update(peaks)
 */

#line 1167 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 1167 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(((PyObject *)__pyx_v_inst));

#line 1167 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_r = __pyx_v_inst;

#line 1167 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L0;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1166
 *         inst = RunningWeightedAverageNeutralMass.__new__(RunningWeightedAverageNeutralMass)
 *         inst._initialize()
 *         if peaks is None:             # <<<<<<<<<<<<<<
 *             return inst
 *         else:
 */

#line 1166 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1169
 *             return inst
 *         else:
 *             inst._update(peaks)             # <<<<<<<<<<<<<<
 *             return inst
 * 
 */

#line 1169 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /*else*/ {

#line 1169 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    ((struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *)__pyx_v_inst->__pyx_base.__pyx_vtab)->__pyx_base._update(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v_inst), __pyx_v_peaks);

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1170
 *         else:
 *             inst._update(peaks)
 *             return inst             # <<<<<<<<<<<<<<
 * 
 *     @cython.nonecheck(False)
 */

#line 1170 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_XDECREF(((PyObject *)__pyx_r));

#line 1170 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __Pyx_INCREF(((PyObject *)__pyx_v_inst));

#line 1170 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_r = __pyx_v_inst;

#line 1170 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L0;

#line 1170 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1160
 * 
 *     @staticmethod
 *     cdef RunningWeightedAverageNeutralMass _create(list peaks):             # <<<<<<<<<<<<<<
 *         cdef:
 *             RunningWeightedAverageNeutralMass inst
 */

#line 1160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 1160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverageNeutralMass._create", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 1160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);

#line 1160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF((PyObject *)__pyx_r);

#line 1160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 1160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass__create__pyx_wrap_1(PyObject *__pyx_v_peaks) {
  return 
#line 1160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass__create(__pyx_v_peaks);

#line 1160 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1174
 *     @cython.nonecheck(False)
 *     @cython.cdivision(True)
 *     cpdef add(self, PeakBase peak):             # <<<<<<<<<<<<<<
 *         if peak.intensity == 0:
 *             if self.current_mean == 0 and self.total_weight == 0:
 */

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_1add(PyObject *__pyx_v_self, PyObject *__pyx_v_peak); /*proto*/

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_add(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak, int __pyx_skip_dispatch) {
  double __pyx_v_agg
#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_2 = NULL;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_3 = NULL;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_4 = NULL;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_5;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_t_6;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  double __pyx_t_7;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("add", 0);

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if called by wrapper */

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(__pyx_skip_dispatch)) ;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* Check if overridden in Python */

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_add); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1174, __pyx_L1_error)

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_GOTREF(__pyx_t_1);

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_1add)) {

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_r);

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; 
#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_t_4 = NULL;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          if (likely(__pyx_t_4)) {

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(__pyx_t_4);

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_INCREF(function);

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
            __Pyx_DECREF_SET(__pyx_t_3, function);

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
          }

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        }

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_peak)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_peak));

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1174, __pyx_L1_error)

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_GOTREF(__pyx_t_2);

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_r = __pyx_t_2;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_t_2 = 0;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        goto __pyx_L0;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      }

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #endif

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    #endif

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1175
 *     @cython.cdivision(True)
 *     cpdef add(self, PeakBase peak):
 *         if peak.intensity == 0:             # <<<<<<<<<<<<<<
 *             if self.current_mean == 0 and self.total_weight == 0:
 *                 self.current_mean = (<DeconvolutedPeak?>peak).neutral_mass
 */

#line 1175 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = ((__pyx_v_peak->intensity == 0.0) != 0);

#line 1175 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1176
 *     cpdef add(self, PeakBase peak):
 *         if peak.intensity == 0:
 *             if self.current_mean == 0 and self.total_weight == 0:             # <<<<<<<<<<<<<<
 *                 self.current_mean = (<DeconvolutedPeak?>peak).neutral_mass
 *                 self.total_weight = 1
 */

#line 1176 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_6 = ((__pyx_v_self->__pyx_base.current_mean == 0.0) != 0);

#line 1176 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_6) {

#line 1176 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    } else {

#line 1176 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_5 = __pyx_t_6;

#line 1176 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L5_bool_binop_done;

#line 1176 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 1176 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_6 = ((__pyx_v_self->__pyx_base.total_weight == 0.0) != 0);

#line 1176 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_t_5 = __pyx_t_6;

#line 1176 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L5_bool_binop_done:;

#line 1176 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__pyx_t_5) {

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1177
 *         if peak.intensity == 0:
 *             if self.current_mean == 0 and self.total_weight == 0:
 *                 self.current_mean = (<DeconvolutedPeak?>peak).neutral_mass             # <<<<<<<<<<<<<<
 *                 self.total_weight = 1
 *             else:
 */

#line 1177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      if (!(likely(__Pyx_TypeTest(((PyObject *)__pyx_v_peak), __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak)))) __PYX_ERR(0, 1177, __pyx_L1_error)

#line 1177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_t_7 = ((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *)__pyx_v_peak)->neutral_mass;

#line 1177 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_self->__pyx_base.current_mean = __pyx_t_7;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1178
 *             if self.current_mean == 0 and self.total_weight == 0:
 *                 self.current_mean = (<DeconvolutedPeak?>peak).neutral_mass
 *                 self.total_weight = 1             # <<<<<<<<<<<<<<
 *             else:
 *                 return
 */

#line 1178 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_v_self->__pyx_base.total_weight = 1.0;

      /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1176
 *     cpdef add(self, PeakBase peak):
 *         if peak.intensity == 0:
 *             if self.current_mean == 0 and self.total_weight == 0:             # <<<<<<<<<<<<<<
 *                 self.current_mean = (<DeconvolutedPeak?>peak).neutral_mass
 *                 self.total_weight = 1
 */

#line 1176 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L4;

#line 1176 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1180
 *                 self.total_weight = 1
 *             else:
 *                 return             # <<<<<<<<<<<<<<
 * 
 *         agg = (self.total_weight * self.current_mean) + \
 */

#line 1180 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    /*else*/ {

#line 1180 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __Pyx_XDECREF(__pyx_r);

#line 1180 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 1180 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
      goto __pyx_L0;

#line 1180 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    }

#line 1180 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_L4:;

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1175
 *     @cython.cdivision(True)
 *     cpdef add(self, PeakBase peak):
 *         if peak.intensity == 0:             # <<<<<<<<<<<<<<
 *             if self.current_mean == 0 and self.total_weight == 0:
 *                 self.current_mean = (<DeconvolutedPeak?>peak).neutral_mass
 */

#line 1175 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1183
 * 
 *         agg = (self.total_weight * self.current_mean) + \
 *             ((<DeconvolutedPeak?>peak).neutral_mass * peak.intensity)             # <<<<<<<<<<<<<<
 *         self.total_weight += peak.intensity
 *         self.current_count += 1
 */

#line 1183 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (!(likely(__Pyx_TypeTest(((PyObject *)__pyx_v_peak), __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak)))) __PYX_ERR(0, 1183, __pyx_L1_error)

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1182
 *                 return
 * 
 *         agg = (self.total_weight * self.current_mean) + \             # <<<<<<<<<<<<<<
 *             ((<DeconvolutedPeak?>peak).neutral_mass * peak.intensity)
 *         self.total_weight += peak.intensity
 */

#line 1182 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_agg = ((__pyx_v_self->__pyx_base.total_weight * __pyx_v_self->__pyx_base.current_mean) + (((struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak *)__pyx_v_peak)->neutral_mass * __pyx_v_peak->intensity));

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1184
 *         agg = (self.total_weight * self.current_mean) + \
 *             ((<DeconvolutedPeak?>peak).neutral_mass * peak.intensity)
 *         self.total_weight += peak.intensity             # <<<<<<<<<<<<<<
 *         self.current_count += 1
 * 
 */

#line 1184 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->__pyx_base.total_weight = (__pyx_v_self->__pyx_base.total_weight + __pyx_v_peak->intensity);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1185
 *             ((<DeconvolutedPeak?>peak).neutral_mass * peak.intensity)
 *         self.total_weight += peak.intensity
 *         self.current_count += 1             # <<<<<<<<<<<<<<
 * 
 *         if self.total_weight != 0:
 */

#line 1185 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_self->__pyx_base.current_count = (__pyx_v_self->__pyx_base.current_count + 1);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1187
 *         self.current_count += 1
 * 
 *         if self.total_weight != 0:             # <<<<<<<<<<<<<<
 *             self.current_mean = agg / (self.total_weight)
 *         else:
 */

#line 1187 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_5 = ((__pyx_v_self->__pyx_base.total_weight != 0.0) != 0);

#line 1187 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (__pyx_t_5) {

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1188
 * 
 *         if self.total_weight != 0:
 *             self.current_mean = agg / (self.total_weight)             # <<<<<<<<<<<<<<
 *         else:
 *             print("NaN produced in add()")
 */

#line 1188 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    __pyx_v_self->__pyx_base.current_mean = (__pyx_v_agg / __pyx_v_self->__pyx_base.total_weight);

    /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1187
 *         self.current_count += 1
 * 
 *         if self.total_weight != 0:             # <<<<<<<<<<<<<<
 *             self.current_mean = agg / (self.total_weight)
 *         else:
 */

#line 1187 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    goto __pyx_L7;

#line 1187 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1190
 *             self.current_mean = agg / (self.total_weight)
 *         else:
 *             print("NaN produced in add()")             # <<<<<<<<<<<<<<
 *         return self
 */

#line 1190 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /*else*/ {

#line 1190 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
    if (__Pyx_PrintOne(0, __pyx_kp_s_NaN_produced_in_add) < 0) __PYX_ERR(0, 1190, __pyx_L1_error)

#line 1190 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  }

#line 1190 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L7:;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1191
 *         else:
 *             print("NaN produced in add()")
 *         return self             # <<<<<<<<<<<<<<
 */

#line 1191 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 1191 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(((PyObject *)__pyx_v_self));

#line 1191 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = ((PyObject *)__pyx_v_self);

#line 1191 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1174
 *     @cython.nonecheck(False)
 *     @cython.cdivision(True)
 *     cpdef add(self, PeakBase peak):             # <<<<<<<<<<<<<<
 *         if peak.intensity == 0:
 *             if self.current_mean == 0 and self.total_weight == 0:
 */

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_2);

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_3);

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_4);

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverageNeutralMass.add", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = 0;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
/* Python wrapper */

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_1add(PyObject *__pyx_v_self, PyObject *__pyx_v_peak); /*proto*/

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_add[] = "RunningWeightedAverageNeutralMass.add(self, PeakBase peak)";

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_1add(PyObject *__pyx_v_self, PyObject *__pyx_v_peak) {

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = 0;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("add (wrapper)", 0);

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_peak), __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase, 1, "peak", 0))) __PYX_ERR(0, 1174, __pyx_L1_error)
  __pyx_r = 
#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_add(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *)__pyx_v_self), ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)__pyx_v_peak));

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_add(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *__pyx_v_self, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *__pyx_v_peak) {

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_r = NULL;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyDeclarations

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyObject *__pyx_t_1 = NULL;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_lineno = 0;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  const char *__pyx_filename = NULL;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  int __pyx_clineno = 0;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannySetupContext("add", 0);

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_r);

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_add(__pyx_v_self, __pyx_v_peak, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1174, __pyx_L1_error)

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = __pyx_t_1;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  goto __pyx_L0;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"


#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  /* function exit code */

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L1_error:;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XDECREF(__pyx_t_1);

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverageNeutralMass.add", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_r = NULL;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_L0:;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGIVEREF(__pyx_r);

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_RefNannyFinishContext();

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  return __pyx_r;

#line 1174 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_2__reduce_cython__[] = "RunningWeightedAverageNeutralMass.__reduce_cython__(self)";

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_2__reduce_cython__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *)__pyx_v_self));

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_2__reduce_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *__pyx_v_self) {
  PyObject *__pyx_v_state = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v__dict = 0
#line 1 "(tree fragment)"
;
  int __pyx_v_use_setstate
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_2 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_3 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_5;

#line 1 "(tree fragment)"
  int __pyx_t_6;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.current_count, self.current_mean, self.total_weight)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */

#line 5 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->__pyx_base.current_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.current_mean); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_2);

#line 5 "(tree fragment)"
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.total_weight); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_3);

#line 5 "(tree fragment)"
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_2);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);

#line 5 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_3);

#line 5 "(tree fragment)"
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);

#line 5 "(tree fragment)"
  __pyx_t_1 = 0;

#line 5 "(tree fragment)"
  __pyx_t_2 = 0;

#line 5 "(tree fragment)"
  __pyx_t_3 = 0;

#line 5 "(tree fragment)"
  __pyx_v_state = ((PyObject*)__pyx_t_4);

#line 5 "(tree fragment)"
  __pyx_t_4 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.current_count, self.current_mean, self.total_weight)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */

#line 6 "(tree fragment)"
  __pyx_t_4 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 6 "(tree fragment)"
  __pyx_v__dict = __pyx_t_4;

#line 6 "(tree fragment)"
  __pyx_t_4 = 0;

  /* "(tree fragment)":7
 *     state = (self.current_count, self.current_mean, self.total_weight)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
  __pyx_t_5 = (__pyx_v__dict != Py_None);

#line 7 "(tree fragment)"
  __pyx_t_6 = (__pyx_t_5 != 0);

#line 7 "(tree fragment)"
  if (__pyx_t_6) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */

#line 8 "(tree fragment)"
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 8 "(tree fragment)"
    __Pyx_INCREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v__dict);

#line 8 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v__dict);

#line 8 "(tree fragment)"
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)

#line 8 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 8 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 8 "(tree fragment)"
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));

#line 8 "(tree fragment)"
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */

#line 9 "(tree fragment)"
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.current_count, self.current_mean, self.total_weight)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */

#line 7 "(tree fragment)"
    goto __pyx_L3;

#line 7 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_RunningWeightedAverageNeutralMass, (type(self), 0x69a8b2a, None), state
 */

#line 11 "(tree fragment)"
  /*else*/ {

#line 11 "(tree fragment)"
    __pyx_v_use_setstate = 0;

#line 11 "(tree fragment)"
  }

#line 11 "(tree fragment)"
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_RunningWeightedAverageNeutralMass, (type(self), 0x69a8b2a, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  __pyx_t_6 = (__pyx_v_use_setstate != 0);

#line 12 "(tree fragment)"
  if (__pyx_t_6) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_RunningWeightedAverageNeutralMass, (type(self), 0x69a8b2a, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_RunningWeightedAverageNeutralMass, (type(self), 0x69a8b2a, state)
 */

#line 13 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 13 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_RunningWeightedAv_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 13 "(tree fragment)"
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 13 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_110791466);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_110791466);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_110791466);

#line 13 "(tree fragment)"
    __Pyx_INCREF(Py_None);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(Py_None);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 2, Py_None);

#line 13 "(tree fragment)"
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_3);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_4);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);

#line 13 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 13 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 13 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state);

#line 13 "(tree fragment)"
    __pyx_t_3 = 0;

#line 13 "(tree fragment)"
    __pyx_t_4 = 0;

#line 13 "(tree fragment)"
    __pyx_r = __pyx_t_2;

#line 13 "(tree fragment)"
    __pyx_t_2 = 0;

#line 13 "(tree fragment)"
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_RunningWeightedAverageNeutralMass, (type(self), 0x69a8b2a, None), state
 *     else:
 */

#line 12 "(tree fragment)"
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_RunningWeightedAverageNeutralMass, (type(self), 0x69a8b2a, None), state
 *     else:
 *         return __pyx_unpickle_RunningWeightedAverageNeutralMass, (type(self), 0x69a8b2a, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_RunningWeightedAverageNeutralMass__set_state(self, __pyx_state)
 */

#line 15 "(tree fragment)"
  /*else*/ {

#line 15 "(tree fragment)"
    __Pyx_XDECREF(__pyx_r);

#line 15 "(tree fragment)"
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_RunningWeightedAv_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_2);

#line 15 "(tree fragment)"
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 15 "(tree fragment)"
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_int_110791466);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_int_110791466);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_110791466);

#line 15 "(tree fragment)"
    __Pyx_INCREF(__pyx_v_state);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_v_state);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state);

#line 15 "(tree fragment)"
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)

#line 15 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_3);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_2);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);

#line 15 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_t_4);

#line 15 "(tree fragment)"
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);

#line 15 "(tree fragment)"
    __pyx_t_2 = 0;

#line 15 "(tree fragment)"
    __pyx_t_4 = 0;

#line 15 "(tree fragment)"
    __pyx_r = __pyx_t_3;

#line 15 "(tree fragment)"
    __pyx_t_3 = 0;

#line 15 "(tree fragment)"
    goto __pyx_L0;

#line 15 "(tree fragment)"
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_2);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_3);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverageNeutralMass.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v_state);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v__dict);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_RunningWeightedAverageNeutralMass, (type(self), 0x69a8b2a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_RunningWeightedAverageNeutralMass__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
/* Python wrapper */

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/

#line 16 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_4__setstate_cython__[] = "RunningWeightedAverageNeutralMass.__setstate_cython__(self, __pyx_state)";

#line 16 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = 
#line 16 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_4__setstate_cython__(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_4__setstate_cython__(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *__pyx_v_self, PyObject *__pyx_v___pyx_state) {

#line 16 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 16 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 16 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 16 "(tree fragment)"
  int __pyx_lineno = 0;

#line 16 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 16 "(tree fragment)"
  int __pyx_clineno = 0;

#line 16 "(tree fragment)"
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_RunningWeightedAverageNeutralMass, (type(self), 0x69a8b2a, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_RunningWeightedAverageNeutralMass__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */

#line 17 "(tree fragment)"
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __pyx_t_1 = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_unpickle_RunningWeightedAverageNeutralMass__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)

#line 17 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 17 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_RunningWeightedAverageNeutralMass, (type(self), 0x69a8b2a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_RunningWeightedAverageNeutralMass__set_state(self, __pyx_state)
 */

#line 16 "(tree fragment)"


#line 16 "(tree fragment)"
  /* function exit code */

#line 16 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 16 "(tree fragment)"
  goto __pyx_L0;

#line 16 "(tree fragment)"
  __pyx_L1_error:;

#line 16 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 16 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverageNeutralMass.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 16 "(tree fragment)"
  __pyx_r = NULL;

#line 16 "(tree fragment)"
  __pyx_L0:;

#line 16 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 16 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 16 "(tree fragment)"
  return __pyx_r;

#line 16 "(tree fragment)"
}

/* "(tree fragment)":1
 * def __pyx_unpickle_FeatureBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_1__pyx_unpickle_FeatureBase(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_unpickle_FeatureBase[] = "__pyx_unpickle_FeatureBase(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_11feature_map_12lcms_feature_1__pyx_unpickle_FeatureBase = 
#line 1 "(tree fragment)"
{"__pyx_unpickle_FeatureBase", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_1__pyx_unpickle_FeatureBase, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_unpickle_FeatureBase};

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_1__pyx_unpickle_FeatureBase(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0
#line 1 "(tree fragment)"
;
  long __pyx_v___pyx_checksum
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_state = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_FeatureBase (wrapper)", 0);

#line 1 "(tree fragment)"
  {

#line 1 "(tree fragment)"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};

#line 1 "(tree fragment)"
    PyObject* values[3] = {0,0,0};

#line 1 "(tree fragment)"
    if (unlikely(__pyx_kwds)) {

#line 1 "(tree fragment)"
      Py_ssize_t kw_args;

#line 1 "(tree fragment)"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 1 "(tree fragment)"
      switch (pos_args) {
        case  3: 
#line 1 "(tree fragment)"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 1 "(tree fragment)"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 1 "(tree fragment)"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  0: break;
        default: 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      kw_args = PyDict_Size(__pyx_kwds);

#line 1 "(tree fragment)"
      switch (pos_args) {

#line 1 "(tree fragment)"
        case  0:

#line 1 "(tree fragment)"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  1:

#line 1 "(tree fragment)"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FeatureBase", 1, 3, 3, 1); 
#line 1 "(tree fragment)"
__PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  2:

#line 1 "(tree fragment)"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FeatureBase", 1, 3, 3, 2); 
#line 1 "(tree fragment)"
__PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      if (unlikely(kw_args > 0)) {

#line 1 "(tree fragment)"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_FeatureBase") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 1 "(tree fragment)"
      goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
    } else {

#line 1 "(tree fragment)"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
    }

#line 1 "(tree fragment)"
    __pyx_v___pyx_type = values[0];

#line 1 "(tree fragment)"
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
    __pyx_v___pyx_state = values[2];

#line 1 "(tree fragment)"
  }

#line 1 "(tree fragment)"
  goto __pyx_L4_argument_unpacking_done;

#line 1 "(tree fragment)"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FeatureBase", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 1 "(tree fragment)"
__PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
  __pyx_L3_error:;

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.__pyx_unpickle_FeatureBase", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return NULL;

#line 1 "(tree fragment)"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_unpickle_FeatureBase(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_unpickle_FeatureBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_result = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_2;

#line 1 "(tree fragment)"
  int __pyx_t_3;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_FeatureBase", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9e63e75, 0x870ba3c, 0x5312e73):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x9e63e75, 0x870ba3c, 0x5312e73) = (_end_time, _mz, _start_time, nodes))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__8, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 4 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9e63e75, 0x870ba3c, 0x5312e73):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x9e63e75, 0x870ba3c, 0x5312e73) = (_end_time, _mz, _start_time, nodes))" % __pyx_checksum)
 *     __pyx_result = FeatureBase.__new__(__pyx_type)
 */

#line 5 "(tree fragment)"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __pyx_v___pyx_PickleError = __pyx_t_1;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x9e63e75, 0x870ba3c, 0x5312e73):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x9e63e75, 0x870ba3c, 0x5312e73) = (_end_time, _mz, _start_time, nodes))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = FeatureBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */

#line 6 "(tree fragment)"
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 6 "(tree fragment)"
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; 
#line 6 "(tree fragment)"
__pyx_t_6 = NULL;

#line 6 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {

#line 6 "(tree fragment)"
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);

#line 6 "(tree fragment)"
      if (likely(__pyx_t_6)) {

#line 6 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 6 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_6);

#line 6 "(tree fragment)"
        __Pyx_INCREF(function);

#line 6 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_1, function);

#line 6 "(tree fragment)"
      }

#line 6 "(tree fragment)"
    }

#line 6 "(tree fragment)"
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 6 "(tree fragment)"
    if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 6 "(tree fragment)"
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9e63e75, 0x870ba3c, 0x5312e73):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x9e63e75, 0x870ba3c, 0x5312e73) = (_end_time, _mz, _start_time, nodes))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x9e63e75, 0x870ba3c, 0x5312e73) = (_end_time, _mz, _start_time, nodes))" % __pyx_checksum)
 *     __pyx_result = FeatureBase.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FeatureBase__set_state(<FeatureBase> __pyx_result, __pyx_state)
 */

#line 7 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 7 "(tree fragment)"
  __pyx_t_5 = NULL;

#line 7 "(tree fragment)"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {

#line 7 "(tree fragment)"
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);

#line 7 "(tree fragment)"
    if (likely(__pyx_t_5)) {

#line 7 "(tree fragment)"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 7 "(tree fragment)"
      __Pyx_INCREF(__pyx_t_5);

#line 7 "(tree fragment)"
      __Pyx_INCREF(function);

#line 7 "(tree fragment)"
      __Pyx_DECREF_SET(__pyx_t_1, function);

#line 7 "(tree fragment)"
    }

#line 7 "(tree fragment)"
  }

#line 7 "(tree fragment)"
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);

#line 7 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 7 "(tree fragment)"
  if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 7 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 7 "(tree fragment)"
  __pyx_v___pyx_result = __pyx_t_4;

#line 7 "(tree fragment)"
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x9e63e75, 0x870ba3c, 0x5312e73) = (_end_time, _mz, _start_time, nodes))" % __pyx_checksum)
 *     __pyx_result = FeatureBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_FeatureBase__set_state(<FeatureBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);

#line 8 "(tree fragment)"
  __pyx_t_2 = (__pyx_t_3 != 0);

#line 8 "(tree fragment)"
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = FeatureBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FeatureBase__set_state(<FeatureBase> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_FeatureBase__set_state(FeatureBase __pyx_result, tuple __pyx_state):
 */

#line 9 "(tree fragment)"
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __pyx_t_4 = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_unpickle_FeatureBase__set_state(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 9 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x9e63e75, 0x870ba3c, 0x5312e73) = (_end_time, _mz, _start_time, nodes))" % __pyx_checksum)
 *     __pyx_result = FeatureBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_FeatureBase__set_state(<FeatureBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FeatureBase__set_state(<FeatureBase> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_FeatureBase__set_state(FeatureBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._end_time = __pyx_state[0]; __pyx_result._mz = __pyx_state[1]; __pyx_result._start_time = __pyx_state[2]; __pyx_result.nodes = __pyx_state[3]
 */

#line 10 "(tree fragment)"
  __Pyx_XDECREF(__pyx_r);

#line 10 "(tree fragment)"
  __Pyx_INCREF(__pyx_v___pyx_result);

#line 10 "(tree fragment)"
  __pyx_r = __pyx_v___pyx_result;

#line 10 "(tree fragment)"
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_FeatureBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.__pyx_unpickle_FeatureBase", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_result);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":11
 *         __pyx_unpickle_FeatureBase__set_state(<FeatureBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_FeatureBase__set_state(FeatureBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._end_time = __pyx_state[0]; __pyx_result._mz = __pyx_state[1]; __pyx_result._start_time = __pyx_state[2]; __pyx_result.nodes = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_unpickle_FeatureBase__set_state(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {

#line 11 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 11 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 11 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 11 "(tree fragment)"
  double __pyx_t_2;

#line 11 "(tree fragment)"
  int __pyx_t_3;

#line 11 "(tree fragment)"
  Py_ssize_t __pyx_t_4;

#line 11 "(tree fragment)"
  int __pyx_t_5;

#line 11 "(tree fragment)"
  int __pyx_t_6;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_7 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_8 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_9 = NULL;

#line 11 "(tree fragment)"
  int __pyx_lineno = 0;

#line 11 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 11 "(tree fragment)"
  int __pyx_clineno = 0;

#line 11 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_FeatureBase__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_FeatureBase__set_state(FeatureBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._end_time = __pyx_state[0]; __pyx_result._mz = __pyx_state[1]; __pyx_result._start_time = __pyx_state[2]; __pyx_result.nodes = __pyx_state[3]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->_end_time = __pyx_t_2;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->_mz = __pyx_t_2;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->_start_time = __pyx_t_2;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList))))) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->nodes);

#line 12 "(tree fragment)"
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->nodes));

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->nodes = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_FeatureBase__set_state(FeatureBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._end_time = __pyx_state[0]; __pyx_result._mz = __pyx_state[1]; __pyx_result._start_time = __pyx_state[2]; __pyx_result.nodes = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */

#line 13 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 13 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 13 "(tree fragment)"
    __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_4 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_5 = ((__pyx_t_4 > 4) != 0);

#line 13 "(tree fragment)"
  if (__pyx_t_5) {

#line 13 "(tree fragment)"
  } else {

#line 13 "(tree fragment)"
    __pyx_t_3 = __pyx_t_5;

#line 13 "(tree fragment)"
    goto __pyx_L4_bool_binop_done;

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_6 = (__pyx_t_5 != 0);

#line 13 "(tree fragment)"
  __pyx_t_3 = __pyx_t_6;

#line 13 "(tree fragment)"
  __pyx_L4_bool_binop_done:;

#line 13 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result._end_time = __pyx_state[0]; __pyx_result._mz = __pyx_state[1]; __pyx_result._start_time = __pyx_state[2]; __pyx_result.nodes = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[4])             # <<<<<<<<<<<<<<
 */

#line 14 "(tree fragment)"
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_7);

#line 14 "(tree fragment)"
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_8);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 14 "(tree fragment)"
    if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 14 "(tree fragment)"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 14 "(tree fragment)"
      __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_7);

#line 14 "(tree fragment)"
    __pyx_t_9 = NULL;

#line 14 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {

#line 14 "(tree fragment)"
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);

#line 14 "(tree fragment)"
      if (likely(__pyx_t_9)) {

#line 14 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);

#line 14 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_9);

#line 14 "(tree fragment)"
        __Pyx_INCREF(function);

#line 14 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_8, function);

#line 14 "(tree fragment)"
      }

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);

#line 14 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 14 "(tree fragment)"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_FeatureBase__set_state(FeatureBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._end_time = __pyx_state[0]; __pyx_result._mz = __pyx_state[1]; __pyx_result._start_time = __pyx_state[2]; __pyx_result.nodes = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */

#line 13 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_FeatureBase__set_state(<FeatureBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_FeatureBase__set_state(FeatureBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._end_time = __pyx_state[0]; __pyx_result._mz = __pyx_state[1]; __pyx_result._start_time = __pyx_state[2]; __pyx_result.nodes = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
  /* function exit code */

#line 11 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 11 "(tree fragment)"
  goto __pyx_L0;

#line 11 "(tree fragment)"
  __pyx_L1_error:;

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_7);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_8);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_9);

#line 11 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.__pyx_unpickle_FeatureBase__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 11 "(tree fragment)"
  __pyx_r = 0;

#line 11 "(tree fragment)"
  __pyx_L0:;

#line 11 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 11 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 11 "(tree fragment)"
  return __pyx_r;

#line 11 "(tree fragment)"
}

/* "(tree fragment)":1
 * def __pyx_unpickle_EmptyFeature(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_3__pyx_unpickle_EmptyFeature(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_2__pyx_unpickle_EmptyFeature[] = "__pyx_unpickle_EmptyFeature(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_11feature_map_12lcms_feature_3__pyx_unpickle_EmptyFeature = 
#line 1 "(tree fragment)"
{"__pyx_unpickle_EmptyFeature", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_3__pyx_unpickle_EmptyFeature, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_2__pyx_unpickle_EmptyFeature};

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_3__pyx_unpickle_EmptyFeature(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0
#line 1 "(tree fragment)"
;
  long __pyx_v___pyx_checksum
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_state = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_EmptyFeature (wrapper)", 0);

#line 1 "(tree fragment)"
  {

#line 1 "(tree fragment)"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};

#line 1 "(tree fragment)"
    PyObject* values[3] = {0,0,0};

#line 1 "(tree fragment)"
    if (unlikely(__pyx_kwds)) {

#line 1 "(tree fragment)"
      Py_ssize_t kw_args;

#line 1 "(tree fragment)"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 1 "(tree fragment)"
      switch (pos_args) {
        case  3: 
#line 1 "(tree fragment)"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 1 "(tree fragment)"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 1 "(tree fragment)"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  0: break;
        default: 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      kw_args = PyDict_Size(__pyx_kwds);

#line 1 "(tree fragment)"
      switch (pos_args) {

#line 1 "(tree fragment)"
        case  0:

#line 1 "(tree fragment)"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  1:

#line 1 "(tree fragment)"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_EmptyFeature", 1, 3, 3, 1); 
#line 1 "(tree fragment)"
__PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  2:

#line 1 "(tree fragment)"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_EmptyFeature", 1, 3, 3, 2); 
#line 1 "(tree fragment)"
__PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      if (unlikely(kw_args > 0)) {

#line 1 "(tree fragment)"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_EmptyFeature") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 1 "(tree fragment)"
      goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
    } else {

#line 1 "(tree fragment)"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
    }

#line 1 "(tree fragment)"
    __pyx_v___pyx_type = values[0];

#line 1 "(tree fragment)"
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
    __pyx_v___pyx_state = values[2];

#line 1 "(tree fragment)"
  }

#line 1 "(tree fragment)"
  goto __pyx_L4_argument_unpacking_done;

#line 1 "(tree fragment)"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_EmptyFeature", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 1 "(tree fragment)"
__PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
  __pyx_L3_error:;

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.__pyx_unpickle_EmptyFeature", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return NULL;

#line 1 "(tree fragment)"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_2__pyx_unpickle_EmptyFeature(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_2__pyx_unpickle_EmptyFeature(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_result = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_2;

#line 1 "(tree fragment)"
  int __pyx_t_3;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_EmptyFeature", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9e63e75, 0x870ba3c, 0x5312e73):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x9e63e75, 0x870ba3c, 0x5312e73) = (_end_time, _mz, _start_time, nodes))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__8, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 4 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9e63e75, 0x870ba3c, 0x5312e73):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x9e63e75, 0x870ba3c, 0x5312e73) = (_end_time, _mz, _start_time, nodes))" % __pyx_checksum)
 *     __pyx_result = EmptyFeature.__new__(__pyx_type)
 */

#line 5 "(tree fragment)"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __pyx_v___pyx_PickleError = __pyx_t_1;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x9e63e75, 0x870ba3c, 0x5312e73):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x9e63e75, 0x870ba3c, 0x5312e73) = (_end_time, _mz, _start_time, nodes))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = EmptyFeature.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */

#line 6 "(tree fragment)"
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 6 "(tree fragment)"
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; 
#line 6 "(tree fragment)"
__pyx_t_6 = NULL;

#line 6 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {

#line 6 "(tree fragment)"
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);

#line 6 "(tree fragment)"
      if (likely(__pyx_t_6)) {

#line 6 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 6 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_6);

#line 6 "(tree fragment)"
        __Pyx_INCREF(function);

#line 6 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_1, function);

#line 6 "(tree fragment)"
      }

#line 6 "(tree fragment)"
    }

#line 6 "(tree fragment)"
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 6 "(tree fragment)"
    if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 6 "(tree fragment)"
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x9e63e75, 0x870ba3c, 0x5312e73):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x9e63e75, 0x870ba3c, 0x5312e73) = (_end_time, _mz, _start_time, nodes))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x9e63e75, 0x870ba3c, 0x5312e73) = (_end_time, _mz, _start_time, nodes))" % __pyx_checksum)
 *     __pyx_result = EmptyFeature.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_EmptyFeature__set_state(<EmptyFeature> __pyx_result, __pyx_state)
 */

#line 7 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 7 "(tree fragment)"
  __pyx_t_5 = NULL;

#line 7 "(tree fragment)"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {

#line 7 "(tree fragment)"
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);

#line 7 "(tree fragment)"
    if (likely(__pyx_t_5)) {

#line 7 "(tree fragment)"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 7 "(tree fragment)"
      __Pyx_INCREF(__pyx_t_5);

#line 7 "(tree fragment)"
      __Pyx_INCREF(function);

#line 7 "(tree fragment)"
      __Pyx_DECREF_SET(__pyx_t_1, function);

#line 7 "(tree fragment)"
    }

#line 7 "(tree fragment)"
  }

#line 7 "(tree fragment)"
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);

#line 7 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 7 "(tree fragment)"
  if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 7 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 7 "(tree fragment)"
  __pyx_v___pyx_result = __pyx_t_4;

#line 7 "(tree fragment)"
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x9e63e75, 0x870ba3c, 0x5312e73) = (_end_time, _mz, _start_time, nodes))" % __pyx_checksum)
 *     __pyx_result = EmptyFeature.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_EmptyFeature__set_state(<EmptyFeature> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);

#line 8 "(tree fragment)"
  __pyx_t_2 = (__pyx_t_3 != 0);

#line 8 "(tree fragment)"
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = EmptyFeature.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_EmptyFeature__set_state(<EmptyFeature> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_EmptyFeature__set_state(EmptyFeature __pyx_result, tuple __pyx_state):
 */

#line 9 "(tree fragment)"
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __pyx_t_4 = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_unpickle_EmptyFeature__set_state(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 9 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x9e63e75, 0x870ba3c, 0x5312e73) = (_end_time, _mz, _start_time, nodes))" % __pyx_checksum)
 *     __pyx_result = EmptyFeature.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_EmptyFeature__set_state(<EmptyFeature> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_EmptyFeature__set_state(<EmptyFeature> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_EmptyFeature__set_state(EmptyFeature __pyx_result, tuple __pyx_state):
 *     __pyx_result._end_time = __pyx_state[0]; __pyx_result._mz = __pyx_state[1]; __pyx_result._start_time = __pyx_state[2]; __pyx_result.nodes = __pyx_state[3]
 */

#line 10 "(tree fragment)"
  __Pyx_XDECREF(__pyx_r);

#line 10 "(tree fragment)"
  __Pyx_INCREF(__pyx_v___pyx_result);

#line 10 "(tree fragment)"
  __pyx_r = __pyx_v___pyx_result;

#line 10 "(tree fragment)"
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_EmptyFeature(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.__pyx_unpickle_EmptyFeature", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_result);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":11
 *         __pyx_unpickle_EmptyFeature__set_state(<EmptyFeature> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_EmptyFeature__set_state(EmptyFeature __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._end_time = __pyx_state[0]; __pyx_result._mz = __pyx_state[1]; __pyx_result._start_time = __pyx_state[2]; __pyx_result.nodes = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_unpickle_EmptyFeature__set_state(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {

#line 11 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 11 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 11 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 11 "(tree fragment)"
  double __pyx_t_2;

#line 11 "(tree fragment)"
  int __pyx_t_3;

#line 11 "(tree fragment)"
  Py_ssize_t __pyx_t_4;

#line 11 "(tree fragment)"
  int __pyx_t_5;

#line 11 "(tree fragment)"
  int __pyx_t_6;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_7 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_8 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_9 = NULL;

#line 11 "(tree fragment)"
  int __pyx_lineno = 0;

#line 11 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 11 "(tree fragment)"
  int __pyx_clineno = 0;

#line 11 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_EmptyFeature__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_EmptyFeature__set_state(EmptyFeature __pyx_result, tuple __pyx_state):
 *     __pyx_result._end_time = __pyx_state[0]; __pyx_result._mz = __pyx_state[1]; __pyx_result._start_time = __pyx_state[2]; __pyx_result.nodes = __pyx_state[3]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->__pyx_base._end_time = __pyx_t_2;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->__pyx_base._mz = __pyx_t_2;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->__pyx_base._start_time = __pyx_t_2;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList))))) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.nodes);

#line 12 "(tree fragment)"
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->__pyx_base.nodes));

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->__pyx_base.nodes = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_EmptyFeature__set_state(EmptyFeature __pyx_result, tuple __pyx_state):
 *     __pyx_result._end_time = __pyx_state[0]; __pyx_result._mz = __pyx_state[1]; __pyx_result._start_time = __pyx_state[2]; __pyx_result.nodes = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */

#line 13 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 13 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 13 "(tree fragment)"
    __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_4 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_5 = ((__pyx_t_4 > 4) != 0);

#line 13 "(tree fragment)"
  if (__pyx_t_5) {

#line 13 "(tree fragment)"
  } else {

#line 13 "(tree fragment)"
    __pyx_t_3 = __pyx_t_5;

#line 13 "(tree fragment)"
    goto __pyx_L4_bool_binop_done;

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_6 = (__pyx_t_5 != 0);

#line 13 "(tree fragment)"
  __pyx_t_3 = __pyx_t_6;

#line 13 "(tree fragment)"
  __pyx_L4_bool_binop_done:;

#line 13 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result._end_time = __pyx_state[0]; __pyx_result._mz = __pyx_state[1]; __pyx_result._start_time = __pyx_state[2]; __pyx_result.nodes = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[4])             # <<<<<<<<<<<<<<
 */

#line 14 "(tree fragment)"
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_7);

#line 14 "(tree fragment)"
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_8);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 14 "(tree fragment)"
    if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 14 "(tree fragment)"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 14 "(tree fragment)"
      __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_7);

#line 14 "(tree fragment)"
    __pyx_t_9 = NULL;

#line 14 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {

#line 14 "(tree fragment)"
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);

#line 14 "(tree fragment)"
      if (likely(__pyx_t_9)) {

#line 14 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);

#line 14 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_9);

#line 14 "(tree fragment)"
        __Pyx_INCREF(function);

#line 14 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_8, function);

#line 14 "(tree fragment)"
      }

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);

#line 14 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

#line 14 "(tree fragment)"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_EmptyFeature__set_state(EmptyFeature __pyx_result, tuple __pyx_state):
 *     __pyx_result._end_time = __pyx_state[0]; __pyx_result._mz = __pyx_state[1]; __pyx_result._start_time = __pyx_state[2]; __pyx_result.nodes = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */

#line 13 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_EmptyFeature__set_state(<EmptyFeature> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_EmptyFeature__set_state(EmptyFeature __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._end_time = __pyx_state[0]; __pyx_result._mz = __pyx_state[1]; __pyx_result._start_time = __pyx_state[2]; __pyx_result.nodes = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
  /* function exit code */

#line 11 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 11 "(tree fragment)"
  goto __pyx_L0;

#line 11 "(tree fragment)"
  __pyx_L1_error:;

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_7);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_8);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_9);

#line 11 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.__pyx_unpickle_EmptyFeature__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 11 "(tree fragment)"
  __pyx_r = 0;

#line 11 "(tree fragment)"
  __pyx_L0:;

#line 11 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 11 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 11 "(tree fragment)"
  return __pyx_r;

#line 11 "(tree fragment)"
}

/* "(tree fragment)":1
 * def __pyx_unpickle_RunningWeightedAverage(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_5__pyx_unpickle_RunningWeightedAverage(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_4__pyx_unpickle_RunningWeightedAverage[] = "__pyx_unpickle_RunningWeightedAverage(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_11feature_map_12lcms_feature_5__pyx_unpickle_RunningWeightedAverage = 
#line 1 "(tree fragment)"
{"__pyx_unpickle_RunningWeightedAverage", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_5__pyx_unpickle_RunningWeightedAverage, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_4__pyx_unpickle_RunningWeightedAverage};

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_5__pyx_unpickle_RunningWeightedAverage(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0
#line 1 "(tree fragment)"
;
  long __pyx_v___pyx_checksum
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_state = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_RunningWeightedAverage (wrapper)", 0);

#line 1 "(tree fragment)"
  {

#line 1 "(tree fragment)"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};

#line 1 "(tree fragment)"
    PyObject* values[3] = {0,0,0};

#line 1 "(tree fragment)"
    if (unlikely(__pyx_kwds)) {

#line 1 "(tree fragment)"
      Py_ssize_t kw_args;

#line 1 "(tree fragment)"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 1 "(tree fragment)"
      switch (pos_args) {
        case  3: 
#line 1 "(tree fragment)"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 1 "(tree fragment)"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 1 "(tree fragment)"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  0: break;
        default: 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      kw_args = PyDict_Size(__pyx_kwds);

#line 1 "(tree fragment)"
      switch (pos_args) {

#line 1 "(tree fragment)"
        case  0:

#line 1 "(tree fragment)"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  1:

#line 1 "(tree fragment)"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_RunningWeightedAverage", 1, 3, 3, 1); 
#line 1 "(tree fragment)"
__PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  2:

#line 1 "(tree fragment)"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_RunningWeightedAverage", 1, 3, 3, 2); 
#line 1 "(tree fragment)"
__PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      if (unlikely(kw_args > 0)) {

#line 1 "(tree fragment)"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_RunningWeightedAverage") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 1 "(tree fragment)"
      goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
    } else {

#line 1 "(tree fragment)"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
    }

#line 1 "(tree fragment)"
    __pyx_v___pyx_type = values[0];

#line 1 "(tree fragment)"
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
    __pyx_v___pyx_state = values[2];

#line 1 "(tree fragment)"
  }

#line 1 "(tree fragment)"
  goto __pyx_L4_argument_unpacking_done;

#line 1 "(tree fragment)"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_RunningWeightedAverage", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 1 "(tree fragment)"
__PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
  __pyx_L3_error:;

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.__pyx_unpickle_RunningWeightedAverage", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return NULL;

#line 1 "(tree fragment)"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_4__pyx_unpickle_RunningWeightedAverage(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_4__pyx_unpickle_RunningWeightedAverage(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_result = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_2;

#line 1 "(tree fragment)"
  int __pyx_t_3;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_RunningWeightedAverage", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x69a8b2a, 0xbe93219, 0x16980be):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x69a8b2a, 0xbe93219, 0x16980be) = (current_count, current_mean, total_weight))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__9, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 4 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x69a8b2a, 0xbe93219, 0x16980be):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x69a8b2a, 0xbe93219, 0x16980be) = (current_count, current_mean, total_weight))" % __pyx_checksum)
 *     __pyx_result = RunningWeightedAverage.__new__(__pyx_type)
 */

#line 5 "(tree fragment)"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __pyx_v___pyx_PickleError = __pyx_t_1;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x69a8b2a, 0xbe93219, 0x16980be):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x69a8b2a, 0xbe93219, 0x16980be) = (current_count, current_mean, total_weight))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = RunningWeightedAverage.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */

#line 6 "(tree fragment)"
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 6 "(tree fragment)"
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; 
#line 6 "(tree fragment)"
__pyx_t_6 = NULL;

#line 6 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {

#line 6 "(tree fragment)"
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);

#line 6 "(tree fragment)"
      if (likely(__pyx_t_6)) {

#line 6 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 6 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_6);

#line 6 "(tree fragment)"
        __Pyx_INCREF(function);

#line 6 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_1, function);

#line 6 "(tree fragment)"
      }

#line 6 "(tree fragment)"
    }

#line 6 "(tree fragment)"
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 6 "(tree fragment)"
    if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 6 "(tree fragment)"
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x69a8b2a, 0xbe93219, 0x16980be):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x69a8b2a, 0xbe93219, 0x16980be) = (current_count, current_mean, total_weight))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x69a8b2a, 0xbe93219, 0x16980be) = (current_count, current_mean, total_weight))" % __pyx_checksum)
 *     __pyx_result = RunningWeightedAverage.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_RunningWeightedAverage__set_state(<RunningWeightedAverage> __pyx_result, __pyx_state)
 */

#line 7 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 7 "(tree fragment)"
  __pyx_t_5 = NULL;

#line 7 "(tree fragment)"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {

#line 7 "(tree fragment)"
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);

#line 7 "(tree fragment)"
    if (likely(__pyx_t_5)) {

#line 7 "(tree fragment)"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 7 "(tree fragment)"
      __Pyx_INCREF(__pyx_t_5);

#line 7 "(tree fragment)"
      __Pyx_INCREF(function);

#line 7 "(tree fragment)"
      __Pyx_DECREF_SET(__pyx_t_1, function);

#line 7 "(tree fragment)"
    }

#line 7 "(tree fragment)"
  }

#line 7 "(tree fragment)"
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);

#line 7 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 7 "(tree fragment)"
  if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 7 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 7 "(tree fragment)"
  __pyx_v___pyx_result = __pyx_t_4;

#line 7 "(tree fragment)"
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x69a8b2a, 0xbe93219, 0x16980be) = (current_count, current_mean, total_weight))" % __pyx_checksum)
 *     __pyx_result = RunningWeightedAverage.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_RunningWeightedAverage__set_state(<RunningWeightedAverage> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);

#line 8 "(tree fragment)"
  __pyx_t_2 = (__pyx_t_3 != 0);

#line 8 "(tree fragment)"
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = RunningWeightedAverage.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_RunningWeightedAverage__set_state(<RunningWeightedAverage> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_RunningWeightedAverage__set_state(RunningWeightedAverage __pyx_result, tuple __pyx_state):
 */

#line 9 "(tree fragment)"
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __pyx_t_4 = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_unpickle_RunningWeightedAverage__set_state(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 9 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x69a8b2a, 0xbe93219, 0x16980be) = (current_count, current_mean, total_weight))" % __pyx_checksum)
 *     __pyx_result = RunningWeightedAverage.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_RunningWeightedAverage__set_state(<RunningWeightedAverage> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_RunningWeightedAverage__set_state(<RunningWeightedAverage> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_RunningWeightedAverage__set_state(RunningWeightedAverage __pyx_result, tuple __pyx_state):
 *     __pyx_result.current_count = __pyx_state[0]; __pyx_result.current_mean = __pyx_state[1]; __pyx_result.total_weight = __pyx_state[2]
 */

#line 10 "(tree fragment)"
  __Pyx_XDECREF(__pyx_r);

#line 10 "(tree fragment)"
  __Pyx_INCREF(__pyx_v___pyx_result);

#line 10 "(tree fragment)"
  __pyx_r = __pyx_v___pyx_result;

#line 10 "(tree fragment)"
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_RunningWeightedAverage(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.__pyx_unpickle_RunningWeightedAverage", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_result);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":11
 *         __pyx_unpickle_RunningWeightedAverage__set_state(<RunningWeightedAverage> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_RunningWeightedAverage__set_state(RunningWeightedAverage __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.current_count = __pyx_state[0]; __pyx_result.current_mean = __pyx_state[1]; __pyx_result.total_weight = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_unpickle_RunningWeightedAverage__set_state(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {

#line 11 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 11 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 11 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 11 "(tree fragment)"
  size_t __pyx_t_2;

#line 11 "(tree fragment)"
  double __pyx_t_3;

#line 11 "(tree fragment)"
  int __pyx_t_4;

#line 11 "(tree fragment)"
  Py_ssize_t __pyx_t_5;

#line 11 "(tree fragment)"
  int __pyx_t_6;

#line 11 "(tree fragment)"
  int __pyx_t_7;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_8 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_9 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_10 = NULL;

#line 11 "(tree fragment)"
  int __pyx_lineno = 0;

#line 11 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 11 "(tree fragment)"
  int __pyx_clineno = 0;

#line 11 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_RunningWeightedAverage__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_RunningWeightedAverage__set_state(RunningWeightedAverage __pyx_result, tuple __pyx_state):
 *     __pyx_result.current_count = __pyx_state[0]; __pyx_result.current_mean = __pyx_state[1]; __pyx_result.total_weight = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_2 = __Pyx_PyInt_As_size_t(__pyx_t_1); if (unlikely((__pyx_t_2 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->current_count = __pyx_t_2;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->current_mean = __pyx_t_3;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->total_weight = __pyx_t_3;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_RunningWeightedAverage__set_state(RunningWeightedAverage __pyx_result, tuple __pyx_state):
 *     __pyx_result.current_count = __pyx_state[0]; __pyx_result.current_mean = __pyx_state[1]; __pyx_result.total_weight = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */

#line 13 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 13 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 13 "(tree fragment)"
    __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_5 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_6 = ((__pyx_t_5 > 3) != 0);

#line 13 "(tree fragment)"
  if (__pyx_t_6) {

#line 13 "(tree fragment)"
  } else {

#line 13 "(tree fragment)"
    __pyx_t_4 = __pyx_t_6;

#line 13 "(tree fragment)"
    goto __pyx_L4_bool_binop_done;

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_7 = (__pyx_t_6 != 0);

#line 13 "(tree fragment)"
  __pyx_t_4 = __pyx_t_7;

#line 13 "(tree fragment)"
  __pyx_L4_bool_binop_done:;

#line 13 "(tree fragment)"
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.current_count = __pyx_state[0]; __pyx_result.current_mean = __pyx_state[1]; __pyx_result.total_weight = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */

#line 14 "(tree fragment)"
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_8);

#line 14 "(tree fragment)"
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_9);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 14 "(tree fragment)"
    if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 14 "(tree fragment)"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 14 "(tree fragment)"
      __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_8);

#line 14 "(tree fragment)"
    __pyx_t_10 = NULL;

#line 14 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {

#line 14 "(tree fragment)"
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);

#line 14 "(tree fragment)"
      if (likely(__pyx_t_10)) {

#line 14 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);

#line 14 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_10);

#line 14 "(tree fragment)"
        __Pyx_INCREF(function);

#line 14 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_9, function);

#line 14 "(tree fragment)"
      }

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_8);

#line 14 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 14 "(tree fragment)"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_RunningWeightedAverage__set_state(RunningWeightedAverage __pyx_result, tuple __pyx_state):
 *     __pyx_result.current_count = __pyx_state[0]; __pyx_result.current_mean = __pyx_state[1]; __pyx_result.total_weight = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */

#line 13 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_RunningWeightedAverage__set_state(<RunningWeightedAverage> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_RunningWeightedAverage__set_state(RunningWeightedAverage __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.current_count = __pyx_state[0]; __pyx_result.current_mean = __pyx_state[1]; __pyx_result.total_weight = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
  /* function exit code */

#line 11 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 11 "(tree fragment)"
  goto __pyx_L0;

#line 11 "(tree fragment)"
  __pyx_L1_error:;

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_8);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_9);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_10);

#line 11 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.__pyx_unpickle_RunningWeightedAverage__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 11 "(tree fragment)"
  __pyx_r = 0;

#line 11 "(tree fragment)"
  __pyx_L0:;

#line 11 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 11 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 11 "(tree fragment)"
  return __pyx_r;

#line 11 "(tree fragment)"
}

/* "(tree fragment)":1
 * def __pyx_unpickle_RunningWeightedAverageNeutralMass(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
/* Python wrapper */

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_7__pyx_unpickle_RunningWeightedAverageNeutralMass(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/

#line 1 "(tree fragment)"
static char __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_6__pyx_unpickle_RunningWeightedAverageNeutralMass[] = "__pyx_unpickle_RunningWeightedAverageNeutralMass(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_12ms_deisotope_2_c_11feature_map_12lcms_feature_7__pyx_unpickle_RunningWeightedAverageNeutralMass = 
#line 1 "(tree fragment)"
{"__pyx_unpickle_RunningWeightedAverageNeutralMass", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_7__pyx_unpickle_RunningWeightedAverageNeutralMass, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_6__pyx_unpickle_RunningWeightedAverageNeutralMass};

#line 1 "(tree fragment)"
static PyObject *__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_7__pyx_unpickle_RunningWeightedAverageNeutralMass(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0
#line 1 "(tree fragment)"
;
  long __pyx_v___pyx_checksum
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_state = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_RunningWeightedAverageNeutralMass (wrapper)", 0);

#line 1 "(tree fragment)"
  {

#line 1 "(tree fragment)"
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};

#line 1 "(tree fragment)"
    PyObject* values[3] = {0,0,0};

#line 1 "(tree fragment)"
    if (unlikely(__pyx_kwds)) {

#line 1 "(tree fragment)"
      Py_ssize_t kw_args;

#line 1 "(tree fragment)"
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);

#line 1 "(tree fragment)"
      switch (pos_args) {
        case  3: 
#line 1 "(tree fragment)"
values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  2: 
#line 1 "(tree fragment)"
values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;
        case  1: 
#line 1 "(tree fragment)"
values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  0: break;
        default: 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      kw_args = PyDict_Size(__pyx_kwds);

#line 1 "(tree fragment)"
      switch (pos_args) {

#line 1 "(tree fragment)"
        case  0:

#line 1 "(tree fragment)"
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else 
#line 1 "(tree fragment)"
goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  1:

#line 1 "(tree fragment)"
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_RunningWeightedAverageNeutralMass", 1, 3, 3, 1); 
#line 1 "(tree fragment)"
__PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
        CYTHON_FALLTHROUGH;

#line 1 "(tree fragment)"
        case  2:

#line 1 "(tree fragment)"
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;

#line 1 "(tree fragment)"
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_RunningWeightedAverageNeutralMass", 1, 3, 3, 2); 
#line 1 "(tree fragment)"
__PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
        }

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
      if (unlikely(kw_args > 0)) {

#line 1 "(tree fragment)"
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_RunningWeightedAverageNeutralMass") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
      }

#line 1 "(tree fragment)"
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {

#line 1 "(tree fragment)"
      goto __pyx_L5_argtuple_error;

#line 1 "(tree fragment)"
    } else {

#line 1 "(tree fragment)"
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);

#line 1 "(tree fragment)"
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);

#line 1 "(tree fragment)"
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);

#line 1 "(tree fragment)"
    }

#line 1 "(tree fragment)"
    __pyx_v___pyx_type = values[0];

#line 1 "(tree fragment)"
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
    __pyx_v___pyx_state = values[2];

#line 1 "(tree fragment)"
  }

#line 1 "(tree fragment)"
  goto __pyx_L4_argument_unpacking_done;

#line 1 "(tree fragment)"
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_RunningWeightedAverageNeutralMass", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); 
#line 1 "(tree fragment)"
__PYX_ERR(1, 1, __pyx_L3_error)

#line 1 "(tree fragment)"
  __pyx_L3_error:;

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.__pyx_unpickle_RunningWeightedAverageNeutralMass", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return NULL;

#line 1 "(tree fragment)"
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = 
#line 1 "(tree fragment)"
__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_6__pyx_unpickle_RunningWeightedAverageNeutralMass(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
static PyObject *__pyx_pf_12ms_deisotope_2_c_11feature_map_12lcms_feature_6__pyx_unpickle_RunningWeightedAverageNeutralMass(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0
#line 1 "(tree fragment)"
;
  PyObject *__pyx_v___pyx_result = 0
#line 1 "(tree fragment)"
;

#line 1 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 1 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 1 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 1 "(tree fragment)"
  int __pyx_t_2;

#line 1 "(tree fragment)"
  int __pyx_t_3;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_4 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_5 = NULL;

#line 1 "(tree fragment)"
  PyObject *__pyx_t_6 = NULL;

#line 1 "(tree fragment)"
  int __pyx_lineno = 0;

#line 1 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 1 "(tree fragment)"
  int __pyx_clineno = 0;

#line 1 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_RunningWeightedAverageNeutralMass", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x69a8b2a, 0xbe93219, 0x16980be):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x69a8b2a, 0xbe93219, 0x16980be) = (current_count, current_mean, total_weight))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 4 "(tree fragment)"
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__9, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 4 "(tree fragment)"
  __pyx_t_3 = (__pyx_t_2 != 0);

#line 4 "(tree fragment)"
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x69a8b2a, 0xbe93219, 0x16980be):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x69a8b2a, 0xbe93219, 0x16980be) = (current_count, current_mean, total_weight))" % __pyx_checksum)
 *     __pyx_result = RunningWeightedAverageNeutralMass.__new__(__pyx_type)
 */

#line 5 "(tree fragment)"
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __Pyx_GIVEREF(__pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);

#line 5 "(tree fragment)"
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)

#line 5 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __Pyx_INCREF(__pyx_t_1);

#line 5 "(tree fragment)"
    __pyx_v___pyx_PickleError = __pyx_t_1;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x69a8b2a, 0xbe93219, 0x16980be):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x69a8b2a, 0xbe93219, 0x16980be) = (current_count, current_mean, total_weight))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = RunningWeightedAverageNeutralMass.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */

#line 6 "(tree fragment)"
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 6 "(tree fragment)"
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; 
#line 6 "(tree fragment)"
__pyx_t_6 = NULL;

#line 6 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {

#line 6 "(tree fragment)"
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);

#line 6 "(tree fragment)"
      if (likely(__pyx_t_6)) {

#line 6 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 6 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_6);

#line 6 "(tree fragment)"
        __Pyx_INCREF(function);

#line 6 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_1, function);

#line 6 "(tree fragment)"
      }

#line 6 "(tree fragment)"
    }

#line 6 "(tree fragment)"
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);

#line 6 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 6 "(tree fragment)"
    if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)

#line 6 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 6 "(tree fragment)"
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);

#line 6 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

#line 6 "(tree fragment)"
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x69a8b2a, 0xbe93219, 0x16980be):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x69a8b2a, 0xbe93219, 0x16980be) = (current_count, current_mean, total_weight))" % __pyx_checksum)
 */

#line 4 "(tree fragment)"
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x69a8b2a, 0xbe93219, 0x16980be) = (current_count, current_mean, total_weight))" % __pyx_checksum)
 *     __pyx_result = RunningWeightedAverageNeutralMass.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_RunningWeightedAverageNeutralMass__set_state(<RunningWeightedAverageNeutralMass> __pyx_result, __pyx_state)
 */

#line 7 "(tree fragment)"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 7 "(tree fragment)"
  __pyx_t_5 = NULL;

#line 7 "(tree fragment)"
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {

#line 7 "(tree fragment)"
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);

#line 7 "(tree fragment)"
    if (likely(__pyx_t_5)) {

#line 7 "(tree fragment)"
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);

#line 7 "(tree fragment)"
      __Pyx_INCREF(__pyx_t_5);

#line 7 "(tree fragment)"
      __Pyx_INCREF(function);

#line 7 "(tree fragment)"
      __Pyx_DECREF_SET(__pyx_t_1, function);

#line 7 "(tree fragment)"
    }

#line 7 "(tree fragment)"
  }

#line 7 "(tree fragment)"
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);

#line 7 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

#line 7 "(tree fragment)"
  if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 7, __pyx_L1_error)

#line 7 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_4);

#line 7 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 7 "(tree fragment)"
  __pyx_v___pyx_result = __pyx_t_4;

#line 7 "(tree fragment)"
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x69a8b2a, 0xbe93219, 0x16980be) = (current_count, current_mean, total_weight))" % __pyx_checksum)
 *     __pyx_result = RunningWeightedAverageNeutralMass.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_RunningWeightedAverageNeutralMass__set_state(<RunningWeightedAverageNeutralMass> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);

#line 8 "(tree fragment)"
  __pyx_t_2 = (__pyx_t_3 != 0);

#line 8 "(tree fragment)"
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = RunningWeightedAverageNeutralMass.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_RunningWeightedAverageNeutralMass__set_state(<RunningWeightedAverageNeutralMass> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_RunningWeightedAverageNeutralMass__set_state(RunningWeightedAverageNeutralMass __pyx_result, tuple __pyx_state):
 */

#line 9 "(tree fragment)"
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __pyx_t_4 = __pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_unpickle_RunningWeightedAverageNeutralMass__set_state(((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 9, __pyx_L1_error)

#line 9 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_4);

#line 9 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x69a8b2a, 0xbe93219, 0x16980be) = (current_count, current_mean, total_weight))" % __pyx_checksum)
 *     __pyx_result = RunningWeightedAverageNeutralMass.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_RunningWeightedAverageNeutralMass__set_state(<RunningWeightedAverageNeutralMass> __pyx_result, __pyx_state)
 *     return __pyx_result
 */

#line 8 "(tree fragment)"
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_RunningWeightedAverageNeutralMass__set_state(<RunningWeightedAverageNeutralMass> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_RunningWeightedAverageNeutralMass__set_state(RunningWeightedAverageNeutralMass __pyx_result, tuple __pyx_state):
 *     __pyx_result.current_count = __pyx_state[0]; __pyx_result.current_mean = __pyx_state[1]; __pyx_result.total_weight = __pyx_state[2]
 */

#line 10 "(tree fragment)"
  __Pyx_XDECREF(__pyx_r);

#line 10 "(tree fragment)"
  __Pyx_INCREF(__pyx_v___pyx_result);

#line 10 "(tree fragment)"
  __pyx_r = __pyx_v___pyx_result;

#line 10 "(tree fragment)"
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_RunningWeightedAverageNeutralMass(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"


#line 1 "(tree fragment)"
  /* function exit code */

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_4);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_5);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_6);

#line 1 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.__pyx_unpickle_RunningWeightedAverageNeutralMass", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1 "(tree fragment)"
  __pyx_r = NULL;

#line 1 "(tree fragment)"
  __pyx_L0:;

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);

#line 1 "(tree fragment)"
  __Pyx_XDECREF(__pyx_v___pyx_result);

#line 1 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return __pyx_r;

#line 1 "(tree fragment)"
}

/* "(tree fragment)":11
 *         __pyx_unpickle_RunningWeightedAverageNeutralMass__set_state(<RunningWeightedAverageNeutralMass> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_RunningWeightedAverageNeutralMass__set_state(RunningWeightedAverageNeutralMass __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.current_count = __pyx_state[0]; __pyx_result.current_mean = __pyx_state[1]; __pyx_result.total_weight = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
static PyObject *__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_unpickle_RunningWeightedAverageNeutralMass__set_state(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {

#line 11 "(tree fragment)"
  PyObject *__pyx_r = NULL;

#line 11 "(tree fragment)"
  __Pyx_RefNannyDeclarations

#line 11 "(tree fragment)"
  PyObject *__pyx_t_1 = NULL;

#line 11 "(tree fragment)"
  size_t __pyx_t_2;

#line 11 "(tree fragment)"
  double __pyx_t_3;

#line 11 "(tree fragment)"
  int __pyx_t_4;

#line 11 "(tree fragment)"
  Py_ssize_t __pyx_t_5;

#line 11 "(tree fragment)"
  int __pyx_t_6;

#line 11 "(tree fragment)"
  int __pyx_t_7;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_8 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_9 = NULL;

#line 11 "(tree fragment)"
  PyObject *__pyx_t_10 = NULL;

#line 11 "(tree fragment)"
  int __pyx_lineno = 0;

#line 11 "(tree fragment)"
  const char *__pyx_filename = NULL;

#line 11 "(tree fragment)"
  int __pyx_clineno = 0;

#line 11 "(tree fragment)"
  __Pyx_RefNannySetupContext("__pyx_unpickle_RunningWeightedAverageNeutralMass__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_RunningWeightedAverageNeutralMass__set_state(RunningWeightedAverageNeutralMass __pyx_result, tuple __pyx_state):
 *     __pyx_result.current_count = __pyx_state[0]; __pyx_result.current_mean = __pyx_state[1]; __pyx_result.total_weight = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_2 = __Pyx_PyInt_As_size_t(__pyx_t_1); if (unlikely((__pyx_t_2 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->__pyx_base.current_count = __pyx_t_2;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->__pyx_base.current_mean = __pyx_t_3;

#line 12 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 12 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 12 "(tree fragment)"
    __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  }

#line 12 "(tree fragment)"
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_1);

#line 12 "(tree fragment)"
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)

#line 12 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 12 "(tree fragment)"
  __pyx_v___pyx_result->__pyx_base.total_weight = __pyx_t_3;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_RunningWeightedAverageNeutralMass__set_state(RunningWeightedAverageNeutralMass __pyx_result, tuple __pyx_state):
 *     __pyx_result.current_count = __pyx_state[0]; __pyx_result.current_mean = __pyx_state[1]; __pyx_result.total_weight = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */

#line 13 "(tree fragment)"
  if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 13 "(tree fragment)"
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");

#line 13 "(tree fragment)"
    __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_5 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_6 = ((__pyx_t_5 > 3) != 0);

#line 13 "(tree fragment)"
  if (__pyx_t_6) {

#line 13 "(tree fragment)"
  } else {

#line 13 "(tree fragment)"
    __pyx_t_4 = __pyx_t_6;

#line 13 "(tree fragment)"
    goto __pyx_L4_bool_binop_done;

#line 13 "(tree fragment)"
  }

#line 13 "(tree fragment)"
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)

#line 13 "(tree fragment)"
  __pyx_t_7 = (__pyx_t_6 != 0);

#line 13 "(tree fragment)"
  __pyx_t_4 = __pyx_t_7;

#line 13 "(tree fragment)"
  __pyx_L4_bool_binop_done:;

#line 13 "(tree fragment)"
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.current_count = __pyx_state[0]; __pyx_result.current_mean = __pyx_state[1]; __pyx_result.total_weight = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */

#line 14 "(tree fragment)"
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_8);

#line 14 "(tree fragment)"
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_9);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 14 "(tree fragment)"
    if (unlikely(__pyx_v___pyx_state == Py_None)) {

#line 14 "(tree fragment)"
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");

#line 14 "(tree fragment)"
      __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_8);

#line 14 "(tree fragment)"
    __pyx_t_10 = NULL;

#line 14 "(tree fragment)"
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {

#line 14 "(tree fragment)"
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);

#line 14 "(tree fragment)"
      if (likely(__pyx_t_10)) {

#line 14 "(tree fragment)"
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);

#line 14 "(tree fragment)"
        __Pyx_INCREF(__pyx_t_10);

#line 14 "(tree fragment)"
        __Pyx_INCREF(function);

#line 14 "(tree fragment)"
        __Pyx_DECREF_SET(__pyx_t_9, function);

#line 14 "(tree fragment)"
      }

#line 14 "(tree fragment)"
    }

#line 14 "(tree fragment)"
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_8);

#line 14 "(tree fragment)"
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 14 "(tree fragment)"
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)

#line 14 "(tree fragment)"
    __Pyx_GOTREF(__pyx_t_1);

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

#line 14 "(tree fragment)"
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_RunningWeightedAverageNeutralMass__set_state(RunningWeightedAverageNeutralMass __pyx_result, tuple __pyx_state):
 *     __pyx_result.current_count = __pyx_state[0]; __pyx_result.current_mean = __pyx_state[1]; __pyx_result.total_weight = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */

#line 13 "(tree fragment)"
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_RunningWeightedAverageNeutralMass__set_state(<RunningWeightedAverageNeutralMass> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_RunningWeightedAverageNeutralMass__set_state(RunningWeightedAverageNeutralMass __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.current_count = __pyx_state[0]; __pyx_result.current_mean = __pyx_state[1]; __pyx_result.total_weight = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"


#line 11 "(tree fragment)"
  /* function exit code */

#line 11 "(tree fragment)"
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 11 "(tree fragment)"
  goto __pyx_L0;

#line 11 "(tree fragment)"
  __pyx_L1_error:;

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_1);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_8);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_9);

#line 11 "(tree fragment)"
  __Pyx_XDECREF(__pyx_t_10);

#line 11 "(tree fragment)"
  __Pyx_AddTraceback("ms_deisotope._c.feature_map.lcms_feature.__pyx_unpickle_RunningWeightedAverageNeutralMass__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 11 "(tree fragment)"
  __pyx_r = 0;

#line 11 "(tree fragment)"
  __pyx_L0:;

#line 11 "(tree fragment)"
  __Pyx_XGIVEREF(__pyx_r);

#line 11 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 11 "(tree fragment)"
  return __pyx_r;

#line 11 "(tree fragment)"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":734
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":735
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 735, __pyx_L1_error)

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 735 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":734
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 734 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":737
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":738
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 738, __pyx_L1_error)

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 738 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":737
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 737 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":740
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":741
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 741, __pyx_L1_error)

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 741 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":740
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 740 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":743
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":744
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 744, __pyx_L1_error)

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 744 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":743
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 743 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":746
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":747
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 747, __pyx_L1_error)

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_t_1);

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = __pyx_t_1;

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = 0;

#line 747 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":746
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 746 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":749
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_1;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":750
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */

#line 750 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);

#line 750 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (__pyx_t_1) {

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":751
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */

#line 751 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_r);

#line 751 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));

#line 751 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);

#line 751 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L0;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":750
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */

#line 750 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":753
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /*else*/ {

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_r);

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_INCREF(__pyx_empty_tuple);

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_r = __pyx_empty_tuple;

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L0;

#line 753 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":749
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 749 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":928
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":929
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */

#line 929 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_INCREF(__pyx_v_base);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":930
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */

#line 930 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":928
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 928 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":932
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base
#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_r = NULL;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_1;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":933
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */

#line 933 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":934
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */

#line 934 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);

#line 934 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (__pyx_t_1) {

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":935
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */

#line 935 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_r);

#line 935 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);

#line 935 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L0;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":934
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */

#line 934 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":936
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */

#line 936 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_r);

#line 936 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_INCREF(((PyObject *)__pyx_v_base));

#line 936 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = ((PyObject *)__pyx_v_base);

#line 936 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":932
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XGIVEREF(__pyx_r);

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 932 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":940
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_2 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_3 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_4;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_5 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_6 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_7 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_8 = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":941
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_declare

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_assign

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_1);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_2);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_3);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    /*try:*/ {

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":942
 * cdef inline int import_array() except -1:
 *     try:
 *         __pyx_import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */

#line 942 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(3, 942, __pyx_L3_error)

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":941
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L8_try_end;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L3_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":943
 *     try:
 *         __pyx_import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (__pyx_t_4) {

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(3, 943, __pyx_L5_except_error)

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_5);

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_6);

#line 943 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":944
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 944, __pyx_L5_except_error)

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_8);

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __PYX_ERR(3, 944, __pyx_L5_except_error)

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L5_except_error;

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L5_except_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":941
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_1);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L1_error;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L8_try_end:;

#line 941 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":940
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_5);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_6);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_7);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_8);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = -1;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 940 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":946
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_2 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_3 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_4;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_5 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_6 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_7 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_8 = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":947
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_declare

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_assign

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_1);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_2);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_3);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    /*try:*/ {

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":948
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */

#line 948 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(3, 948, __pyx_L3_error)

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":947
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L8_try_end;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L3_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":949
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (__pyx_t_4) {

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(3, 949, __pyx_L5_except_error)

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_5);

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_6);

#line 949 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":950
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 950, __pyx_L5_except_error)

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_8);

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __PYX_ERR(3, 950, __pyx_L5_except_error)

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L5_except_error;

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L5_except_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":947
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_1);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L1_error;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L8_try_end:;

#line 947 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":946
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_5);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_6);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_7);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_8);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = -1;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 946 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":952
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_1 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_2 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_3 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_t_4;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_5 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_6 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_7 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *__pyx_t_8 = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_lineno = 0;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  const char *__pyx_filename = NULL;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_clineno = 0;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":953
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_declare

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_PyThreadState_assign

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_1);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_2);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGOTREF(__pyx_t_3);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    /*try:*/ {

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":954
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */

#line 954 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(3, 954, __pyx_L3_error)

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":953
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L8_try_end;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L3_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":955
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (__pyx_t_4) {

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(3, 955, __pyx_L5_except_error)

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_5);

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_6);

#line 955 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":956
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef extern from *:
 */

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 956, __pyx_L5_except_error)

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_GOTREF(__pyx_t_8);

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __PYX_ERR(3, 956, __pyx_L5_except_error)

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L5_except_error;

#line 956 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L5_except_error:;

    /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":953
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_1);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_2);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_XGIVEREF(__pyx_t_3);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    goto __pyx_L1_error;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_L8_try_end:;

#line 953 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":952
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = 0;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_5);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_6);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_7);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_8);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = -1;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 952 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":966
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_is_timedelta64_object(PyObject *__pyx_v_obj) {

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("is_timedelta64_object", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":978
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyTimedeltaArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 978 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyTimedeltaArrType_Type));

#line 978 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":966
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 966 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":981
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE int __pyx_f_5numpy_is_datetime64_object(PyObject *__pyx_v_obj) {

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int __pyx_r;

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyDeclarations

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannySetupContext("is_datetime64_object", 0);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":993
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyDatetimeArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 993 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyDatetimeArrType_Type));

#line 993 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":981
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 981 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":996
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE npy_datetime __pyx_f_5numpy_get_datetime64_value(PyObject *__pyx_v_obj) {

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  npy_datetime __pyx_r;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1003
 *     also needed.  That can be found using `get_datetime64_unit`.
 *     """
 *     return (<PyDatetimeScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1003 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = ((PyDatetimeScalarObject *)__pyx_v_obj)->obval;

#line 1003 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":996
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 996 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1006
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE npy_timedelta __pyx_f_5numpy_get_timedelta64_value(PyObject *__pyx_v_obj) {

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  npy_timedelta __pyx_r;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1010
 *     returns the int64 value underlying scalar numpy timedelta64 object
 *     """
 *     return (<PyTimedeltaScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 1010 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = ((PyTimedeltaScalarObject *)__pyx_v_obj)->obval;

#line 1010 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1006
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 1006 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

/* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1013
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static CYTHON_INLINE NPY_DATETIMEUNIT __pyx_f_5numpy_get_datetime64_unit(PyObject *__pyx_v_obj) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  NPY_DATETIMEUNIT __pyx_r;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1017
 *     returns the unit part of the dtype for a numpy datetime64 object.
 *     """
 *     return <NPY_DATETIMEUNIT>(<PyDatetimeScalarObject*>obj).obmeta.base             # <<<<<<<<<<<<<<
 */

#line 1017 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_r = ((NPY_DATETIMEUNIT)((PyDatetimeScalarObject *)__pyx_v_obj)->obmeta.base);

#line 1017 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1013
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /* function exit code */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_r;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *p;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    o = (*t->tp_alloc)(t, 0);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  } else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (unlikely(!o)) return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->__pyx_vtab = __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->roots = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->_node_id_hash = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList(PyObject *o) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_USE_TP_FINALIZE

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject_GC_UnTrack(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->roots);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->_node_id_hash);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  (*Py_TYPE(o)->tp_free)(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList(PyObject *o, visitproc v, void *a) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->roots) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(p->roots, a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->_node_id_hash) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(p->_node_id_hash, a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList(PyObject *o) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject* tmp;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  tmp = ((PyObject*)p->roots);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->roots = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_XDECREF(tmp);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  tmp = ((PyObject*)p->_node_id_hash);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->_node_id_hash = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_XDECREF(tmp);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_sq_item_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList(PyObject *o, Py_ssize_t i) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *r;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_DECREF(x);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return r;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_node_id_hash(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_12node_id_hash_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_roots(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5roots_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_roots(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5roots_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5roots_5__del__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList__node_id_hash(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_13_node_id_hash_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList__node_id_hash(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_13_node_id_hash_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_13_node_id_hash_5__del__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList[] = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"_make_node", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_3_make_node, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_2_make_node},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"find_time", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_5find_time, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_4find_time},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"_build_node_id_hash", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_7_build_node_id_hash, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_6_build_node_id_hash},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"insert_node", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_9insert_node, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_8insert_node},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"insert", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_11insert, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_10insert},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"extend", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_13extend, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_12extend},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"__reduce__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_21__reduce__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_20__reduce__},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"clone", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_23clone, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_22clone},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"unspool", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_25unspool, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_24unspool},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"common_nodes", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_27common_nodes, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_26common_nodes},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {0, 0, 0, 0}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList[] = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"node_id_hash", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_node_id_hash, 0, (char *)0, 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"roots", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_roots, __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_roots, (char *)"roots: list", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"_node_id_hash", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList__node_id_hash, __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList__node_id_hash, (char *)"_node_id_hash: set", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {0, 0, 0, 0, 0}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PySequenceMethods __pyx_tp_as_sequence_LCMSFeatureTreeList = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_17__len__, /*sq_length*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_concat*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_repeat*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_sq_item_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList, /*sq_item*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_slice*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_ass_item*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_ass_slice*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_contains*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_inplace_concat*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_inplace_repeat*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyMappingMethods __pyx_tp_as_mapping_LCMSFeatureTreeList = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_17__len__, /*mp_length*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_15__getitem__, /*mp_subscript*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*mp_ass_subscript*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyVarObject_HEAD_INIT(0, 0)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  "ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList", /*tp_name*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList), /*tp_basicsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_itemsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList, /*tp_dealloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX < 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall_offset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION < 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_compare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION >= 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_async*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_29__repr__, /*tp_repr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_number*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  &__pyx_tp_as_sequence_LCMSFeatureTreeList, /*tp_as_sequence*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  &__pyx_tp_as_mapping_LCMSFeatureTreeList, /*tp_as_mapping*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_hash*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_call*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_str*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_buffer*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  "LCMSFeatureTreeList(roots=None)", /*tp_doc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList, /*tp_traverse*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList, /*tp_clear*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_richcompare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklistoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_19__iter__, /*tp_iter*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_iternext*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_methods_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList, /*tp_methods*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_members*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_getsets_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList, /*tp_getset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_base*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dict*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_get*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_set*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dictoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_1__init__, /*tp_init*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_alloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList, /*tp_new*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_free*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_is_gc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_bases*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_mro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_cache*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_subclasses*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklist*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_del*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_version_tag*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030400a1

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_finalize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_pypy_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *p;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    o = (*t->tp_alloc)(t, 0);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  } else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (unlikely(!o)) return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->__pyx_vtab = __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->members = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->_most_abundant_member = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->node_id = Py_None; Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase(PyObject *o) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_USE_TP_FINALIZE

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject_GC_UnTrack(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->members);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->_most_abundant_member);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->node_id);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  (*Py_TYPE(o)->tp_free)(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase(PyObject *o, visitproc v, void *a) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->members) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(p->members, a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->_most_abundant_member) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(((PyObject *)p->_most_abundant_member), a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->node_id) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(p->node_id, a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase(PyObject *o) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject* tmp;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  tmp = ((PyObject*)p->members);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->members = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_XDECREF(tmp);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  tmp = ((PyObject*)p->_most_abundant_member);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->_most_abundant_member = ((struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_XDECREF(tmp);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  tmp = ((PyObject*)p->node_id);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->node_id = Py_None; Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_XDECREF(tmp);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_peaks(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_5peaks_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_time(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_4time_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_time(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_4time_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return -1;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_members(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7members_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_members(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7members_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7members_5__del__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase__most_abundant_member(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_21_most_abundant_member_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase__most_abundant_member(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_21_most_abundant_member_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_21_most_abundant_member_5__del__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_node_id(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7node_id_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_node_id(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7node_id_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7node_id_5__del__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase[] = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"clone", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_1clone, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_clone},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"_unspool_strip_children", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_3_unspool_strip_children, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_2_unspool_strip_children},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"__getstate__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_6__getstate__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_5__getstate__},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"__setstate__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_8__setstate__, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_7__setstate__},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"__reduce__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_10__reduce__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_9__reduce__},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"add", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_12add, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_11add},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"_total_intensity_members", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_14_total_intensity_members, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_13_total_intensity_members},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"max_intensity", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_16max_intensity, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_15max_intensity},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"total_intensity", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_18total_intensity, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_17total_intensity},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"_recalculate", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_20_recalculate, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_19_recalculate},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"_calculate_most_abundant_member", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_22_calculate_most_abundant_member, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_21_calculate_most_abundant_member},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {0, 0, 0, 0}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase[] = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"peaks", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_peaks, 0, (char *)0, 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"time", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_time, __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_time, (char *)"time: 'double'", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"members", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_members, __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_members, (char *)"members: list", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"_most_abundant_member", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase__most_abundant_member, __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase__most_abundant_member, (char *)"_most_abundant_member: ms_peak_picker._c.peak_set.PeakBase", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"node_id", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_node_id, __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_node_id, (char *)"node_id: object", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {0, 0, 0, 0, 0}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyVarObject_HEAD_INIT(0, 0)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  "ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNodeBase", /*tp_name*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase), /*tp_basicsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_itemsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase, /*tp_dealloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX < 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall_offset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION < 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_compare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION >= 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_async*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_repr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_number*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_sequence*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_mapping*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_24__hash__, /*tp_hash*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_call*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_str*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_buffer*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_doc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase, /*tp_traverse*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase, /*tp_clear*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_richcompare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklistoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_iter*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_iternext*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_methods_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase, /*tp_methods*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_members*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_getsets_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase, /*tp_getset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_base*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dict*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_get*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_set*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dictoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_init*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_alloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase, /*tp_new*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_free*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_is_gc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_bases*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_mro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_cache*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_subclasses*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklist*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_del*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_version_tag*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030400a1

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_finalize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_pypy_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *p;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *o = __pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase(t, a, k);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (unlikely(!o)) return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *)o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase*)__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_mz(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_2mz_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode__mz(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_3_mz_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode__mz(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_3_mz_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return -1;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode[] = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"_recalculate", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_3_recalculate, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_2_recalculate},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"_eq", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_5_eq, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_4_eq},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"_ne", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_7_ne, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_6_ne},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {0, 0, 0, 0}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode[] = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"mz", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_mz, 0, (char *)0, 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"_mz", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode__mz, __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode__mz, (char *)"_mz: 'double'", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {0, 0, 0, 0, 0}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyVarObject_HEAD_INIT(0, 0)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  "ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeNode", /*tp_name*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode), /*tp_basicsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_itemsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase, /*tp_dealloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX < 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall_offset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION < 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_compare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION >= 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_async*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_11__repr__, /*tp_repr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_number*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_sequence*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_mapping*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_hash*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_call*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_str*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_buffer*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  "LCMSFeatureTreeNode(time, members=None)", /*tp_doc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase, /*tp_traverse*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase, /*tp_clear*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_9__richcmp__, /*tp_richcompare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklistoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_iter*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_iternext*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_methods_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode, /*tp_methods*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_members*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_getsets_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode, /*tp_getset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_base*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dict*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_get*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_set*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dictoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_1__init__, /*tp_init*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_alloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode, /*tp_new*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_free*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_is_gc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_bases*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_mro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_cache*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_subclasses*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklist*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_del*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_version_tag*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030400a1

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_finalize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_pypy_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *__pyx_freelist_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase[10000000];

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase = 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *p;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase)) & ((t->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)) == 0))) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    o = (PyObject*)__pyx_freelist_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase[--__pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase];

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    memset(o, 0, sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase));

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    (void) PyObject_INIT(o, t);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyObject_GC_Track(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  } else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      o = (*t->tp_alloc)(t, 0);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    } else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (unlikely(!o)) return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->__pyx_vtab = __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->nodes = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase(PyObject *o) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_USE_TP_FINALIZE

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject_GC_UnTrack(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->nodes);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase < 10000000) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase)) & ((Py_TYPE(o)->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)) == 0))) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_freelist_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase[__pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase++] = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  } else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    (*Py_TYPE(o)->tp_free)(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase(PyObject *o, visitproc v, void *a) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->nodes) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(((PyObject *)p->nodes), a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase(PyObject *o) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject* tmp;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  tmp = ((PyObject*)p->nodes);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->nodes = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_XDECREF(tmp);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_nodes(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_5nodes_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_nodes(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_5nodes_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_5nodes_5__del__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase__mz(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_3_mz_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase__mz(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_3_mz_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return -1;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase__start_time(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_11_start_time_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase__start_time(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_11_start_time_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return -1;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase__end_time(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_9_end_time_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase__end_time(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_9_end_time_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return -1;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase[] = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"find_time", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_3find_time, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_2find_time},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_9__reduce_cython__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_8__reduce_cython__},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_11__setstate_cython__, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_10__setstate_cython__},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {0, 0, 0, 0}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase[] = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"nodes", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_nodes, __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_nodes, (char *)"nodes: ms_deisotope._c.feature_map.lcms_feature.LCMSFeatureTreeList", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"_mz", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase__mz, __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase__mz, (char *)"_mz: 'double'", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"_start_time", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase__start_time, __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase__start_time, (char *)"_start_time: 'double'", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"_end_time", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase__end_time, __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase__end_time, (char *)"_end_time: 'double'", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {0, 0, 0, 0, 0}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PySequenceMethods __pyx_tp_as_sequence_FeatureBase = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_5__len__, /*sq_length*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_concat*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_repeat*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_item*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_slice*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_ass_item*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_ass_slice*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_contains*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_inplace_concat*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_inplace_repeat*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyMappingMethods __pyx_tp_as_mapping_FeatureBase = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_5__len__, /*mp_length*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*mp_subscript*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*mp_ass_subscript*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyVarObject_HEAD_INIT(0, 0)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  "ms_deisotope._c.feature_map.lcms_feature.FeatureBase", /*tp_name*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase), /*tp_basicsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_itemsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase, /*tp_dealloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX < 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall_offset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION < 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_compare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION >= 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_async*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_repr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_number*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  &__pyx_tp_as_sequence_FeatureBase, /*tp_as_sequence*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  &__pyx_tp_as_mapping_FeatureBase, /*tp_as_mapping*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_hash*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_call*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_str*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_buffer*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  "FeatureBase(nodes)", /*tp_doc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase, /*tp_traverse*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase, /*tp_clear*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_richcompare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklistoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_7__iter__, /*tp_iter*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_iternext*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_methods_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase, /*tp_methods*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_members*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_getsets_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase, /*tp_getset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_base*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dict*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_get*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_set*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dictoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_1__init__, /*tp_init*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_alloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase, /*tp_new*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_free*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_is_gc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_bases*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_mro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_cache*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_subclasses*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklist*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_del*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_version_tag*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030400a1

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_finalize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_pypy_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *p;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *o = __pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase(t, a, k);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (unlikely(!o)) return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase*)__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->_times = Py_None; Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->_peaks = Py_None; Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->adducts = Py_None; Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->used_as_adduct = Py_None; Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->created_at = Py_None; Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->feature_id = Py_None; Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->_peak_averager = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature(PyObject *o) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_USE_TP_FINALIZE

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject_GC_UnTrack(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->_times);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->_peaks);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->adducts);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->used_as_adduct);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->created_at);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->feature_id);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->_peak_averager);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject_GC_Track(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature(PyObject *o, visitproc v, void *a) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  e = __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase(o, v, a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->_times) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(p->_times, a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->_peaks) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(p->_peaks, a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->adducts) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(p->adducts, a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->used_as_adduct) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(p->used_as_adduct, a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->created_at) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(p->created_at, a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->feature_id) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(p->feature_id, a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->_peak_averager) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(((PyObject *)p->_peak_averager), a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature(PyObject *o) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject* tmp;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  tmp = ((PyObject*)p->_times);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->_times = Py_None; Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_XDECREF(tmp);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  tmp = ((PyObject*)p->_peaks);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->_peaks = Py_None; Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_XDECREF(tmp);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  tmp = ((PyObject*)p->adducts);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->adducts = Py_None; Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_XDECREF(tmp);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  tmp = ((PyObject*)p->used_as_adduct);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->used_as_adduct = Py_None; Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_XDECREF(tmp);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  tmp = ((PyObject*)p->created_at);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->created_at = Py_None; Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_XDECREF(tmp);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  tmp = ((PyObject*)p->feature_id);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->feature_id = Py_None; Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_XDECREF(tmp);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  tmp = ((PyObject*)p->_peak_averager);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->_peak_averager = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_XDECREF(tmp);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_sq_item_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature(PyObject *o, Py_ssize_t i) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *r;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_DECREF(x);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return r;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_total_signal(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_12total_signal_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_intensity(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_9intensity_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_neutral_mass(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_12neutral_mass_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_mz(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_2mz_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_times(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_5times_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_peaks(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_5peaks_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_start_time(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10start_time_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_end_time(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_8end_time_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_apex_time(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_9apex_time_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__total_intensity(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_16_total_intensity_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__total_intensity(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_16_total_intensity_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return -1;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__last_mz(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_8_last_mz_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__last_mz(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_8_last_mz_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return -1;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__times(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_times_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__times(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_times_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_times_5__del__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__peaks(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_peaks_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__peaks(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_peaks_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_peaks_5__del__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_adducts(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_7adducts_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_adducts(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_7adducts_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_7adducts_5__del__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_used_as_adduct(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14used_as_adduct_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_used_as_adduct(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14used_as_adduct_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14used_as_adduct_5__del__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_created_at(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10created_at_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_created_at(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10created_at_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10created_at_5__del__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_feature_id(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10feature_id_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_feature_id(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10feature_id_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10feature_id_5__del__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__peak_averager(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14_peak_averager_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__peak_averager(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14_peak_averager_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14_peak_averager_5__del__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature[] = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"_initialize_averager", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_3_initialize_averager, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_2_initialize_averager},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"_update_from_averager", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_5_update_from_averager, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_4_update_from_averager},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"_reaverage_and_update", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_7_reaverage_and_update, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_6_reaverage_and_update},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"invalidate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_9invalidate, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_8invalidate},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"_invalidate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_11_invalidate, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_10_invalidate},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"overlaps_in_time", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_13overlaps_in_time, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_12overlaps_in_time},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"spans_in_time", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_15spans_in_time, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_14spans_in_time},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"as_arrays", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_17as_arrays, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_16as_arrays},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"_copy_chunk", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_21_copy_chunk, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_20_copy_chunk},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"split_at", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_23split_at, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_22split_at},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"between_times", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_25between_times, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_24between_times},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"split_sparse", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_27split_sparse, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_26split_sparse},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"truncate_before", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_29truncate_before, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_28truncate_before},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"truncate_after", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_31truncate_after, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_30truncate_after},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"clone", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_33clone, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_32clone},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"insert_node", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_35insert_node, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_34insert_node},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"insert", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_37insert, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_36insert},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"merge", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_39merge, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_38merge},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"_eq", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_41_eq, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_40_eq},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"_ne", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_43_ne, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_42_ne},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"__reduce__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_47__reduce__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_46__reduce__},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"max_intensity", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_53max_intensity, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_52max_intensity},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {0, 0, 0, 0}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature[] = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"total_signal", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_total_signal, 0, (char *)0, 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"intensity", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_intensity, 0, (char *)0, 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"neutral_mass", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_neutral_mass, 0, (char *)0, 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"mz", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_mz, 0, (char *)0, 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"times", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_times, 0, (char *)0, 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"peaks", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_peaks, 0, (char *)0, 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"start_time", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_start_time, 0, (char *)0, 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"end_time", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_end_time, 0, (char *)0, 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"apex_time", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_apex_time, 0, (char *)0, 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"_total_intensity", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__total_intensity, __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__total_intensity, (char *)"_total_intensity: 'double'", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"_last_mz", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__last_mz, __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__last_mz, (char *)"_last_mz: 'double'", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"_times", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__times, __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__times, (char *)"_times: object", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"_peaks", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__peaks, __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__peaks, (char *)"_peaks: object", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"adducts", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_adducts, __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_adducts, (char *)"adducts: object", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"used_as_adduct", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_used_as_adduct, __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_used_as_adduct, (char *)"used_as_adduct: object", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"created_at", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_created_at, __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_created_at, (char *)"created_at: object", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"feature_id", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_feature_id, __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_feature_id, (char *)"feature_id: object", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"_peak_averager", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__peak_averager, __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__peak_averager, (char *)"_peak_averager: ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverage", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {0, 0, 0, 0, 0}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PySequenceMethods __pyx_tp_as_sequence_LCMSFeature = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_COMPILING_IN_PYPY

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_5__len__, /*sq_length*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #else

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_length*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_concat*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_repeat*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_sq_item_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature, /*sq_item*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_slice*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_ass_item*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_ass_slice*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_contains*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_inplace_concat*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_inplace_repeat*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyMappingMethods __pyx_tp_as_mapping_LCMSFeature = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_COMPILING_IN_PYPY

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_5__len__, /*mp_length*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #else

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*mp_length*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_51__getitem__, /*mp_subscript*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*mp_ass_subscript*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyVarObject_HEAD_INIT(0, 0)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  "ms_deisotope._c.feature_map.lcms_feature.LCMSFeature", /*tp_name*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature), /*tp_basicsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_itemsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature, /*tp_dealloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX < 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall_offset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION < 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_compare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION >= 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_async*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_19__repr__, /*tp_repr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_number*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  &__pyx_tp_as_sequence_LCMSFeature, /*tp_as_sequence*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  &__pyx_tp_as_mapping_LCMSFeature, /*tp_as_mapping*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_49__hash__, /*tp_hash*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_call*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_str*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_buffer*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  "LCMSFeature(nodes=None, adducts=None, used_as_adduct=None, feature_id=None)", /*tp_doc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature, /*tp_traverse*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature, /*tp_clear*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_45__richcmp__, /*tp_richcompare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklistoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_COMPILING_IN_PYPY

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_7__iter__, /*tp_iter*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #else

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_iter*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_iternext*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_methods_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature, /*tp_methods*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_members*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_getsets_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature, /*tp_getset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_base*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dict*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_get*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_set*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dictoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_1__init__, /*tp_init*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_alloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature, /*tp_new*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_free*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_is_gc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_bases*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_mro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_cache*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_subclasses*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklist*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_del*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_version_tag*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030400a1

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_finalize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_pypy_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *p;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *o = __pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase(t, a, k);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (unlikely(!o)) return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *)o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase*)__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_sq_item_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature(PyObject *o, Py_ssize_t i) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *r;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_DECREF(x);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return r;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_mz(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_2mz_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_start_time(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_10start_time_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_end_time(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_8end_time_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature[] = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_7__reduce_cython__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_6__reduce_cython__},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_9__setstate_cython__, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_8__setstate_cython__},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {0, 0, 0, 0}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature[] = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"mz", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_mz, 0, (char *)0, 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"start_time", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_start_time, 0, (char *)0, 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"end_time", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_end_time, 0, (char *)0, 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {0, 0, 0, 0, 0}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PySequenceMethods __pyx_tp_as_sequence_EmptyFeature = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_3__len__, /*sq_length*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_concat*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_repeat*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_sq_item_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature, /*sq_item*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_slice*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_ass_item*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_ass_slice*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_contains*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_inplace_concat*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*sq_inplace_repeat*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyMappingMethods __pyx_tp_as_mapping_EmptyFeature = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_3__len__, /*mp_length*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_5__getitem__, /*mp_subscript*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*mp_ass_subscript*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyVarObject_HEAD_INIT(0, 0)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  "ms_deisotope._c.feature_map.lcms_feature.EmptyFeature", /*tp_name*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature), /*tp_basicsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_itemsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase, /*tp_dealloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX < 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall_offset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION < 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_compare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION >= 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_async*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_repr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_number*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  &__pyx_tp_as_sequence_EmptyFeature, /*tp_as_sequence*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  &__pyx_tp_as_mapping_EmptyFeature, /*tp_as_mapping*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_hash*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_call*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_str*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_buffer*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  "EmptyFeature(mz)", /*tp_doc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase, /*tp_traverse*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase, /*tp_clear*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_richcompare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklistoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_COMPILING_IN_PYPY

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_7__iter__, /*tp_iter*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #else

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_iter*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_iternext*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_methods_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature, /*tp_methods*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_members*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_getsets_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature, /*tp_getset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_base*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dict*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_get*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_set*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dictoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_1__init__, /*tp_init*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_alloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature, /*tp_new*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_free*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_is_gc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_bases*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_mro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_cache*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_subclasses*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklist*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_del*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_version_tag*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030400a1

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_finalize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_pypy_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *p;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    o = (*t->tp_alloc)(t, 0);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  } else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (unlikely(!o)) return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->__pyx_vtab = __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->features = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->real_features = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator(PyObject *o) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_USE_TP_FINALIZE

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject_GC_UnTrack(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyObject *etype, *eval, *etb;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyErr_Fetch(&etype, &eval, &etb);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_3__dealloc__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyErr_Restore(etype, eval, etb);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->features);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->real_features);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  (*Py_TYPE(o)->tp_free)(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator(PyObject *o, visitproc v, void *a) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->features) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(p->features, a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->real_features) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(p->real_features, a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator(PyObject *o) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject* tmp;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  tmp = ((PyObject*)p->features);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->features = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_XDECREF(tmp);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  tmp = ((PyObject*)p->real_features);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->real_features = ((PyObject*)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_XDECREF(tmp);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_current_time(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_12current_time_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_features(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8features_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_features(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8features_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8features_5__del__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_real_features(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_13real_features_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_real_features(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_13real_features_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_13real_features_5__del__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_start_time(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_10start_time_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_start_time(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_10start_time_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return -1;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_end_time(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8end_time_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_end_time(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8end_time_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return -1;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_last_time_seen(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_14last_time_seen_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_last_time_seen(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_14last_time_seen_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return -1;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_specialmethod___pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_19__next__(PyObject *self, CYTHON_UNUSED PyObject *arg) {return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_19__next__(self);}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator[] = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"init_indices", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_5init_indices, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_4init_indices},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"get_next_time", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_7get_next_time, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_6get_next_time},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"has_more", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_9has_more, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_8has_more},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"get_index_list", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_11get_index_list, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_10get_index_list},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"get_peaks_for_time", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_13get_peaks_for_time, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_12get_peaks_for_time},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"get_current_time", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_15get_current_time, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_14get_current_time},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"get_next_value", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_17get_next_value, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_16get_next_value},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"__next__", (PyCFunction)__pyx_specialmethod___pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_19__next__, METH_NOARGS|METH_COEXIST, 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_25__reduce_cython__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_24__reduce_cython__},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_27__setstate_cython__, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_26__setstate_cython__},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {0, 0, 0, 0}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator[] = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"current_time", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_current_time, 0, (char *)0, 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"features", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_features, __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_features, (char *)"features: list", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"real_features", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_real_features, __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_real_features, (char *)"real_features: list", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"start_time", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_start_time, __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_start_time, (char *)"start_time: 'double'", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"end_time", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_end_time, __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_end_time, (char *)"end_time: 'double'", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"last_time_seen", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_last_time_seen, __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_last_time_seen, (char *)"last_time_seen: 'double'", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {0, 0, 0, 0, 0}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyVarObject_HEAD_INIT(0, 0)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  "ms_deisotope._c.feature_map.lcms_feature.FeatureSetIterator", /*tp_name*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator), /*tp_basicsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_itemsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator, /*tp_dealloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX < 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall_offset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION < 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_compare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION >= 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_async*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_23__repr__, /*tp_repr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_number*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_sequence*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_mapping*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_hash*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_call*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_str*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_buffer*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  "FeatureSetIterator(features, start_time=None, end_time=None)", /*tp_doc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator, /*tp_traverse*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator, /*tp_clear*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_richcompare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklistoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_21__iter__, /*tp_iter*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_19__next__, /*tp_iternext*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_methods_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator, /*tp_methods*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_members*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_getsets_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator, /*tp_getset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_base*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dict*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_get*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_set*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dictoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_1__init__, /*tp_init*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_alloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator, /*tp_new*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_free*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_is_gc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_bases*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_mro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_cache*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_subclasses*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklist*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_del*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_version_tag*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030400a1

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_finalize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_pypy_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *p;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    o = (*t->tp_alloc)(t, 0);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  } else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (unlikely(!o)) return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *)o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->__pyx_vtab = __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage(PyObject *o) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_USE_TP_FINALIZE

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (PyObject_CallFinalizerFromDealloc(o)) return;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  (*Py_TYPE(o)->tp_free)(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_current_mean(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_12current_mean_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_current_mean(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_12current_mean_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return -1;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_current_count(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_13current_count_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_current_count(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_13current_count_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return -1;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_total_weight(PyObject *o, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_12total_weight_1__get__(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_total_weight(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (v) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_12total_weight_3__set__(o, v);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyErr_SetString(PyExc_NotImplementedError, "__del__");

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    return -1;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage[] = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"_initialize", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_1_initialize, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage__initialize},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"reset", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_5reset, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_4reset},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"add", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_7add, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_6add},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"feed_from_feature", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_9feed_from_feature, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_8feed_from_feature},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"update", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_11update, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_10update},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_15__reduce_cython__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_14__reduce_cython__},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_17__setstate_cython__, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_16__setstate_cython__},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {0, 0, 0, 0}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct PyGetSetDef __pyx_getsets_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage[] = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"current_mean", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_current_mean, __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_current_mean, (char *)"current_mean: 'double'", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"current_count", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_current_count, __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_current_count, (char *)"current_count: 'size_t'", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {(char *)"total_weight", __pyx_getprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_total_weight, __pyx_setprop_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_total_weight, (char *)"total_weight: 'double'", 0},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {0, 0, 0, 0, 0}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyVarObject_HEAD_INIT(0, 0)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  "ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverage", /*tp_name*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage), /*tp_basicsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_itemsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage, /*tp_dealloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX < 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall_offset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION < 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_compare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION >= 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_async*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_13__repr__, /*tp_repr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_number*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_sequence*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_mapping*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_hash*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_call*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_str*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_buffer*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  "RunningWeightedAverage(iterable=None)", /*tp_doc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_traverse*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_clear*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_richcompare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklistoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_iter*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_iternext*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_methods_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage, /*tp_methods*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_members*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_getsets_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage, /*tp_getset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_base*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dict*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_get*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_set*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dictoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_3__init__, /*tp_init*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_alloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage, /*tp_new*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_free*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_is_gc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_bases*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_mro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_cache*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_subclasses*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklist*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_del*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_version_tag*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030400a1

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_finalize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_pypy_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass(PyTypeObject *t, PyObject *a, PyObject *k) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *p;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *o = __pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage(t, a, k);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (unlikely(!o)) return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *)o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage*)__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyMethodDef __pyx_methods_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass[] = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"add", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_1add, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_add},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_3__reduce_cython__, METH_NOARGS, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_2__reduce_cython__},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_5__setstate_cython__, METH_O, __pyx_doc_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_4__setstate_cython__},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {0, 0, 0, 0}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyVarObject_HEAD_INIT(0, 0)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  "ms_deisotope._c.feature_map.lcms_feature.RunningWeightedAverageNeutralMass", /*tp_name*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass), /*tp_basicsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_itemsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage, /*tp_dealloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX < 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall_offset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION < 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_compare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION >= 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_async*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_COMPILING_IN_PYPY

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_13__repr__, /*tp_repr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #else

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_repr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_number*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_sequence*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_mapping*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_hash*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_call*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_str*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_buffer*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_doc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_traverse*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_clear*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_richcompare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklistoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_iter*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_iternext*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_methods_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass, /*tp_methods*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_members*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_base*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dict*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_get*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_set*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dictoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_COMPILING_IN_PYPY

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_pw_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_3__init__, /*tp_init*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #else

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_init*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_alloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass, /*tp_new*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_free*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_is_gc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_bases*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_mro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_cache*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_subclasses*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklist*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_del*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_version_tag*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030400a1

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_finalize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_pypy_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone *__pyx_freelist_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone[8];

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone = 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone)))) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    o = (PyObject*)__pyx_freelist_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone[--__pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone];

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    memset(o, 0, sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone));

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    (void) PyObject_INIT(o, t);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyObject_GC_Track(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  } else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    o = (*t->tp_alloc)(t, 0);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (unlikely(!o)) return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone(PyObject *o) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject_GC_UnTrack(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->__pyx_v_deep);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->__pyx_v_self);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone)))) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_freelist_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone[__pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone++] = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone *)o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  } else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    (*Py_TYPE(o)->tp_free)(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone(PyObject *o, visitproc v, void *a) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->__pyx_v_deep) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(p->__pyx_v_deep, a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->__pyx_v_self) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone(PyObject *o) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject* tmp;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  tmp = ((PyObject*)p->__pyx_v_deep);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->__pyx_v_deep = Py_None; Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_XDECREF(tmp);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  tmp = ((PyObject*)p->__pyx_v_self);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->__pyx_v_self = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_XDECREF(tmp);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyVarObject_HEAD_INIT(0, 0)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  "ms_deisotope._c.feature_map.lcms_feature.__pyx_scope_struct__clone", /*tp_name*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone), /*tp_basicsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_itemsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone, /*tp_dealloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX < 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall_offset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION < 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_compare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION >= 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_async*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_repr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_number*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_sequence*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_mapping*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_hash*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_call*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_str*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_buffer*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_doc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone, /*tp_traverse*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone, /*tp_clear*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_richcompare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklistoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_iter*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_iternext*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_methods*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_members*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_base*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dict*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_get*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_set*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dictoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_init*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_alloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone, /*tp_new*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_free*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_is_gc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_bases*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_mro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_cache*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_subclasses*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklist*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_del*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_version_tag*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030400a1

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_finalize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_pypy_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr *__pyx_freelist_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr[8];

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr = 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr)))) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    o = (PyObject*)__pyx_freelist_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr[--__pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr];

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    memset(o, 0, sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr));

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    (void) PyObject_INIT(o, t);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyObject_GC_Track(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  } else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    o = (*t->tp_alloc)(t, 0);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (unlikely(!o)) return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr(PyObject *o) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject_GC_UnTrack(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->__pyx_outer_scope);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->__pyx_v_node);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->__pyx_t_0);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr)))) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_freelist_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr[__pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr++] = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr *)o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  } else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    (*Py_TYPE(o)->tp_free)(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr(PyObject *o, visitproc v, void *a) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->__pyx_outer_scope) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->__pyx_v_node) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(p->__pyx_v_node, a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->__pyx_t_0) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(p->__pyx_t_0, a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyVarObject_HEAD_INIT(0, 0)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  "ms_deisotope._c.feature_map.lcms_feature.__pyx_scope_struct_1_genexpr", /*tp_name*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr), /*tp_basicsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_itemsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr, /*tp_dealloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX < 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall_offset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION < 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_compare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION >= 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_async*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_repr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_number*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_sequence*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_mapping*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_hash*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_call*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_str*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_buffer*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_doc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr, /*tp_traverse*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_clear*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_richcompare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklistoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_iter*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_iternext*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_methods*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_members*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_base*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dict*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_get*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_set*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dictoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_init*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_alloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr, /*tp_new*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_free*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_is_gc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_bases*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_mro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_cache*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_subclasses*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklist*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_del*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_version_tag*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030400a1

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_finalize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_pypy_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children *__pyx_freelist_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children[8];

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children = 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children)))) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    o = (PyObject*)__pyx_freelist_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children[--__pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children];

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    memset(o, 0, sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children));

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    (void) PyObject_INIT(o, t);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyObject_GC_Track(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  } else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    o = (*t->tp_alloc)(t, 0);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (unlikely(!o)) return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children(PyObject *o) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject_GC_UnTrack(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->__pyx_v_node);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->__pyx_v_self);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children)))) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_freelist_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children[__pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children++] = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children *)o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  } else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    (*Py_TYPE(o)->tp_free)(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children(PyObject *o, visitproc v, void *a) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->__pyx_v_node) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(p->__pyx_v_node, a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->__pyx_v_self) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyVarObject_HEAD_INIT(0, 0)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  "ms_deisotope._c.feature_map.lcms_feature.__pyx_scope_struct_2__unspool_strip_children", /*tp_name*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children), /*tp_basicsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_itemsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children, /*tp_dealloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX < 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall_offset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION < 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_compare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION >= 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_async*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_repr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_number*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_sequence*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_mapping*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_hash*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_call*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_str*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_buffer*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_doc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children, /*tp_traverse*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_clear*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_richcompare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklistoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_iter*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_iternext*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_methods*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_members*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_base*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dict*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_get*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_set*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dictoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_init*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_alloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children, /*tp_new*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_free*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_is_gc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_bases*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_mro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_cache*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_subclasses*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklist*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_del*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_version_tag*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030400a1

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_finalize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_pypy_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__ *__pyx_freelist_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__[8];

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__ = 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__)))) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    o = (PyObject*)__pyx_freelist_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__[--__pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__];

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    memset(o, 0, sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__));

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    (void) PyObject_INIT(o, t);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyObject_GC_Track(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  } else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    o = (*t->tp_alloc)(t, 0);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (unlikely(!o)) return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__(PyObject *o) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__ *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__ *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject_GC_UnTrack(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->__pyx_v_self);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__)))) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_freelist_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__[__pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__++] = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__ *)o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  } else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    (*Py_TYPE(o)->tp_free)(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__(PyObject *o, visitproc v, void *a) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__ *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__ *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->__pyx_v_self) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__(PyObject *o) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject* tmp;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__ *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__ *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  tmp = ((PyObject*)p->__pyx_v_self);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  p->__pyx_v_self = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *)Py_None); Py_INCREF(Py_None);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_XDECREF(tmp);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__ = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyVarObject_HEAD_INIT(0, 0)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  "ms_deisotope._c.feature_map.lcms_feature.__pyx_scope_struct_3___get__", /*tp_name*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__), /*tp_basicsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_itemsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__, /*tp_dealloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX < 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall_offset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION < 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_compare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION >= 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_async*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_repr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_number*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_sequence*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_mapping*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_hash*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_call*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_str*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_buffer*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_doc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__, /*tp_traverse*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_clear_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__, /*tp_clear*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_richcompare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklistoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_iter*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_iternext*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_methods*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_members*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_base*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dict*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_get*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_set*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dictoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_init*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_alloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__, /*tp_new*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_free*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_is_gc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_bases*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_mro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_cache*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_subclasses*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklist*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_del*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_version_tag*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030400a1

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_finalize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_pypy_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr *__pyx_freelist_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr[8];

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr = 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject *__pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject *o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr)))) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    o = (PyObject*)__pyx_freelist_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr[--__pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr];

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    memset(o, 0, sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr));

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    (void) PyObject_INIT(o, t);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyObject_GC_Track(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  } else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    o = (*t->tp_alloc)(t, 0);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (unlikely(!o)) return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static void __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr(PyObject *o) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyObject_GC_UnTrack(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->__pyx_outer_scope);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->__pyx_v_node);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_CLEAR(p->__pyx_t_0);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr)))) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_freelist_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr[__pyx_freecount_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr++] = ((struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr *)o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  } else {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    (*Py_TYPE(o)->tp_free)(o);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr(PyObject *o, visitproc v, void *a) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  int e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr *p = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr *)o;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->__pyx_outer_scope) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->__pyx_v_node) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(p->__pyx_v_node, a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (p->__pyx_t_0) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    e = (*v)(p->__pyx_t_0, a); if (e) return e;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return 0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyTypeObject __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  PyVarObject_HEAD_INIT(0, 0)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  "ms_deisotope._c.feature_map.lcms_feature.__pyx_scope_struct_4_genexpr", /*tp_name*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  sizeof(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr), /*tp_basicsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_itemsize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_dealloc_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr, /*tp_dealloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX < 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall_offset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION < 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_compare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_MAJOR_VERSION >= 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_async*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_repr*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_number*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_sequence*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_mapping*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_hash*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_call*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_str*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_setattro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_as_buffer*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_doc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_traverse_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr, /*tp_traverse*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_clear*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_richcompare*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklistoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_iter*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_iternext*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_methods*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_members*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_getset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_base*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dict*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_get*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_descr_set*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_dictoffset*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_init*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_alloc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tp_new_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr, /*tp_new*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_free*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_is_gc*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_bases*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_mro*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_cache*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_subclasses*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_weaklist*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_del*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_version_tag*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030400a1

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_finalize*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_vectorcall*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_print*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  0, /*tp_pypy_flags*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyMethodDef __pyx_methods[] = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {0, 0, 0, 0}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
#if PY_MAJOR_VERSION >= 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
#if CYTHON_PEP489_MULTI_PHASE_INIT

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static int __pyx_pymod_exec_lcms_feature(PyObject* module); /*proto*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static PyModuleDef_Slot __pyx_moduledef_slots[] = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {Py_mod_create, (void*)__pyx_pymod_create},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {Py_mod_exec, (void*)__pyx_pymod_exec_lcms_feature},

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  {0, NULL}

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
#endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
static struct PyModuleDef __pyx_moduledef = {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyModuleDef_HEAD_INIT,

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    "lcms_feature",

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    0, /* m_doc */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_PEP489_MULTI_PHASE_INIT

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    0, /* m_size */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #else

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    -1, /* m_size */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_methods /* m_methods */,

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_PEP489_MULTI_PHASE_INIT

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    __pyx_moduledef_slots, /* m_slots */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #else

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    NULL, /* m_reload */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    NULL, /* m_traverse */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    NULL, /* m_clear */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    NULL /* m_free */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
};

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
#endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_, __pyx_k_, sizeof(__pyx_k_), 0, 0, 1, 0},
  {&__pyx_kp_s_Duplicate_Node_s, __pyx_k_Duplicate_Node_s, sizeof(__pyx_k_Duplicate_Node_s), 0, 0, 1, 0},
  {&__pyx_n_s_EmptyFeature, __pyx_k_EmptyFeature, sizeof(__pyx_k_EmptyFeature), 0, 0, 1, 1},
  {&__pyx_kp_s_Empty_Feature_is_not_of_type_Emp, __pyx_k_Empty_Feature_is_not_of_type_Emp, sizeof(__pyx_k_Empty_Feature_is_not_of_type_Emp), 0, 0, 1, 0},
  {&__pyx_kp_s_Failed_to_allocate_index_list_fo, __pyx_k_Failed_to_allocate_index_list_fo, sizeof(__pyx_k_Failed_to_allocate_index_list_fo), 0, 0, 1, 0},
  {&__pyx_n_s_FeatureBase, __pyx_k_FeatureBase, sizeof(__pyx_k_FeatureBase), 0, 0, 1, 1},
  {&__pyx_n_s_FeatureSetIterator, __pyx_k_FeatureSetIterator, sizeof(__pyx_k_FeatureSetIterator), 0, 0, 1, 1},
  {&__pyx_kp_s_FeatureSetIterator_d_d_0_3f_0_3f, __pyx_k_FeatureSetIterator_d_d_0_3f_0_3f, sizeof(__pyx_k_FeatureSetIterator_d_d_0_3f_0_3f), 0, 0, 1, 0},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_k_Incompatible_checksums_0x_x_vs_0, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_k_Incompatible_checksums_0x_x_vs_0_2, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_2), 0, 0, 1, 0},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_n_s_LCMSFeature, __pyx_k_LCMSFeature, sizeof(__pyx_k_LCMSFeature), 0, 0, 1, 1},
  {&__pyx_n_s_LCMSFeatureTreeList, __pyx_k_LCMSFeatureTreeList, sizeof(__pyx_k_LCMSFeatureTreeList), 0, 0, 1, 1},
  {&__pyx_kp_s_LCMSFeatureTreeList_d_nodes_0_2f, __pyx_k_LCMSFeatureTreeList_d_nodes_0_2f, sizeof(__pyx_k_LCMSFeatureTreeList_d_nodes_0_2f), 0, 0, 1, 0},
  {&__pyx_n_s_LCMSFeatureTreeNode, __pyx_k_LCMSFeatureTreeNode, sizeof(__pyx_k_LCMSFeatureTreeNode), 0, 0, 1, 1},
  {&__pyx_n_s_LCMSFeatureTreeNodeBase, __pyx_k_LCMSFeatureTreeNodeBase, sizeof(__pyx_k_LCMSFeatureTreeNodeBase), 0, 0, 1, 1},
  {&__pyx_n_s_LCMSFeatureTreeNodeBase__unspool, __pyx_k_LCMSFeatureTreeNodeBase__unspool, sizeof(__pyx_k_LCMSFeatureTreeNodeBase__unspool), 0, 0, 1, 1},
  {&__pyx_n_s_LCMSFeature___get___locals_genex, __pyx_k_LCMSFeature___get___locals_genex, sizeof(__pyx_k_LCMSFeature___get___locals_genex), 0, 0, 1, 1},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_kp_s_NaN_produced_in_add, __pyx_k_NaN_produced_in_add, sizeof(__pyx_k_NaN_produced_in_add), 0, 0, 1, 0},
  {&__pyx_n_s_NotImplemented, __pyx_k_NotImplemented, sizeof(__pyx_k_NotImplemented), 0, 0, 1, 1},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_RunningWeightedAverage, __pyx_k_RunningWeightedAverage, sizeof(__pyx_k_RunningWeightedAverage), 0, 0, 1, 1},
  {&__pyx_n_s_RunningWeightedAverageNeutralMas, __pyx_k_RunningWeightedAverageNeutralMas, sizeof(__pyx_k_RunningWeightedAverageNeutralMas), 0, 0, 1, 1},
  {&__pyx_n_s_StopIteration, __pyx_k_StopIteration, sizeof(__pyx_k_StopIteration), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_add, __pyx_k_add, sizeof(__pyx_k_add), 0, 0, 1, 1},
  {&__pyx_n_s_adducts, __pyx_k_adducts, sizeof(__pyx_k_adducts), 0, 0, 1, 1},
  {&__pyx_n_s_argmax, __pyx_k_argmax, sizeof(__pyx_k_argmax), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_array, __pyx_k_array, sizeof(__pyx_k_array), 0, 0, 1, 1},
  {&__pyx_n_s_as_arrays, __pyx_k_as_arrays, sizeof(__pyx_k_as_arrays), 0, 0, 1, 1},
  {&__pyx_kp_s_best_mz_0, __pyx_k_best_mz_0, sizeof(__pyx_k_best_mz_0), 0, 0, 1, 0},
  {&__pyx_n_s_build_node_id_hash, __pyx_k_build_node_id_hash, sizeof(__pyx_k_build_node_id_hash), 0, 0, 1, 1},
  {&__pyx_n_s_calculate_most_abundant_member, __pyx_k_calculate_most_abundant_member, sizeof(__pyx_k_calculate_most_abundant_member), 0, 0, 1, 1},
  {&__pyx_n_s_children, __pyx_k_children, sizeof(__pyx_k_children), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_clone, __pyx_k_clone, sizeof(__pyx_k_clone), 0, 0, 1, 1},
  {&__pyx_n_s_clone_locals_genexpr, __pyx_k_clone_locals_genexpr, sizeof(__pyx_k_clone_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_cls, __pyx_k_cls, sizeof(__pyx_k_cls), 0, 0, 1, 1},
  {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
  {&__pyx_n_s_copy_chunk, __pyx_k_copy_chunk, sizeof(__pyx_k_copy_chunk), 0, 0, 1, 1},
  {&__pyx_n_s_created_at, __pyx_k_created_at, sizeof(__pyx_k_created_at), 0, 0, 1, 1},
  {&__pyx_n_s_deep, __pyx_k_deep, sizeof(__pyx_k_deep), 0, 0, 1, 1},
  {&__pyx_n_s_delta_rt, __pyx_k_delta_rt, sizeof(__pyx_k_delta_rt), 0, 0, 1, 1},
  {&__pyx_n_s_deque, __pyx_k_deque, sizeof(__pyx_k_deque), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {&__pyx_n_s_end_time, __pyx_k_end_time, sizeof(__pyx_k_end_time), 0, 0, 1, 1},
  {&__pyx_n_s_eq, __pyx_k_eq, sizeof(__pyx_k_eq), 0, 0, 1, 1},
  {&__pyx_n_s_feature_id, __pyx_k_feature_id, sizeof(__pyx_k_feature_id), 0, 0, 1, 1},
  {&__pyx_n_s_features, __pyx_k_features, sizeof(__pyx_k_features), 0, 0, 1, 1},
  {&__pyx_n_s_feed_from_feature, __pyx_k_feed_from_feature, sizeof(__pyx_k_feed_from_feature), 0, 0, 1, 1},
  {&__pyx_n_s_file, __pyx_k_file, sizeof(__pyx_k_file), 0, 0, 1, 1},
  {&__pyx_n_s_find_time, __pyx_k_find_time, sizeof(__pyx_k_find_time), 0, 0, 1, 1},
  {&__pyx_n_s_float64, __pyx_k_float64, sizeof(__pyx_k_float64), 0, 0, 1, 1},
  {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_get_current_time, __pyx_k_get_current_time, sizeof(__pyx_k_get_current_time), 0, 0, 1, 1},
  {&__pyx_n_s_get_next_time, __pyx_k_get_next_time, sizeof(__pyx_k_get_next_time), 0, 0, 1, 1},
  {&__pyx_n_s_get_next_value, __pyx_k_get_next_value, sizeof(__pyx_k_get_next_value), 0, 0, 1, 1},
  {&__pyx_n_s_get_peaks_for_time, __pyx_k_get_peaks_for_time, sizeof(__pyx_k_get_peaks_for_time), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_has_more, __pyx_k_has_more, sizeof(__pyx_k_has_more), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_inf, __pyx_k_inf, sizeof(__pyx_k_inf), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_n_s_init_indices, __pyx_k_init_indices, sizeof(__pyx_k_init_indices), 0, 0, 1, 1},
  {&__pyx_n_s_initialize, __pyx_k_initialize, sizeof(__pyx_k_initialize), 0, 0, 1, 1},
  {&__pyx_n_s_initialize_averager, __pyx_k_initialize_averager, sizeof(__pyx_k_initialize_averager), 0, 0, 1, 1},
  {&__pyx_n_s_insert, __pyx_k_insert, sizeof(__pyx_k_insert), 0, 0, 1, 1},
  {&__pyx_n_s_insert_node, __pyx_k_insert_node, sizeof(__pyx_k_insert_node), 0, 0, 1, 1},
  {&__pyx_n_s_invalidate, __pyx_k_invalidate, sizeof(__pyx_k_invalidate), 0, 0, 1, 1},
  {&__pyx_n_s_iterable, __pyx_k_iterable, sizeof(__pyx_k_iterable), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_make_node, __pyx_k_make_node, sizeof(__pyx_k_make_node), 0, 0, 1, 1},
  {&__pyx_n_s_max_intensity, __pyx_k_max_intensity, sizeof(__pyx_k_max_intensity), 0, 0, 1, 1},
  {&__pyx_n_s_members, __pyx_k_members, sizeof(__pyx_k_members), 0, 0, 1, 1},
  {&__pyx_n_s_merge, __pyx_k_merge, sizeof(__pyx_k_merge), 0, 0, 1, 1},
  {&__pyx_n_s_ms_deisotope__c_feature_map_lcms, __pyx_k_ms_deisotope__c_feature_map_lcms, sizeof(__pyx_k_ms_deisotope__c_feature_map_lcms), 0, 0, 1, 1},
  {&__pyx_n_s_ms_deisotope_utils, __pyx_k_ms_deisotope_utils, sizeof(__pyx_k_ms_deisotope_utils), 0, 0, 1, 1},
  {&__pyx_n_s_mz, __pyx_k_mz, sizeof(__pyx_k_mz), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_ne, __pyx_k_ne, sizeof(__pyx_k_ne), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_n_s_node, __pyx_k_node, sizeof(__pyx_k_node), 0, 0, 1, 1},
  {&__pyx_n_s_node_id, __pyx_k_node_id, sizeof(__pyx_k_node_id), 0, 0, 1, 1},
  {&__pyx_n_s_node_id_hash, __pyx_k_node_id_hash, sizeof(__pyx_k_node_id_hash), 0, 0, 1, 1},
  {&__pyx_n_s_nodes, __pyx_k_nodes, sizeof(__pyx_k_nodes), 0, 0, 1, 1},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_s_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 0, 1, 0},
  {&__pyx_kp_s_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 0, 1, 0},
  {&__pyx_n_s_overlaps_in_time, __pyx_k_overlaps_in_time, sizeof(__pyx_k_overlaps_in_time), 0, 0, 1, 1},
  {&__pyx_n_s_peak, __pyx_k_peak, sizeof(__pyx_k_peak), 0, 0, 1, 1},
  {&__pyx_n_s_peaks, __pyx_k_peaks, sizeof(__pyx_k_peaks), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_pop, __pyx_k_pop, sizeof(__pyx_k_pop), 0, 0, 1, 1},
  {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_EmptyFeature, __pyx_k_pyx_unpickle_EmptyFeature, sizeof(__pyx_k_pyx_unpickle_EmptyFeature), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_FeatureBase, __pyx_k_pyx_unpickle_FeatureBase, sizeof(__pyx_k_pyx_unpickle_FeatureBase), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_RunningWeightedAv, __pyx_k_pyx_unpickle_RunningWeightedAv, sizeof(__pyx_k_pyx_unpickle_RunningWeightedAv), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_RunningWeightedAv_2, __pyx_k_pyx_unpickle_RunningWeightedAv_2, sizeof(__pyx_k_pyx_unpickle_RunningWeightedAv_2), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_random, __pyx_k_random, sizeof(__pyx_k_random), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reaverage, __pyx_k_reaverage, sizeof(__pyx_k_reaverage), 0, 0, 1, 1},
  {&__pyx_n_s_reaverage_and_update, __pyx_k_reaverage_and_update, sizeof(__pyx_k_reaverage_and_update), 0, 0, 1, 1},
  {&__pyx_n_s_recalculate, __pyx_k_recalculate, sizeof(__pyx_k_recalculate), 0, 0, 1, 1},
  {&__pyx_n_s_recalculate_2, __pyx_k_recalculate_2, sizeof(__pyx_k_recalculate_2), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_reset, __pyx_k_reset, sizeof(__pyx_k_reset), 0, 0, 1, 1},
  {&__pyx_n_s_roots, __pyx_k_roots, sizeof(__pyx_k_roots), 0, 0, 1, 1},
  {&__pyx_kp_s_s_0_4f_0_2f_0_2f, __pyx_k_s_0_4f_0_2f_0_2f, sizeof(__pyx_k_s_0_4f_0_2f_0_2f), 0, 0, 1, 0},
  {&__pyx_kp_s_s_f_0_4f_s_d, __pyx_k_s_f_0_4f_s_d, sizeof(__pyx_k_s_f_0_4f_s_d), 0, 0, 1, 0},
  {&__pyx_kp_s_s_r_d, __pyx_k_s_r_d, sizeof(__pyx_k_s_r_d), 0, 0, 1, 0},
  {&__pyx_n_s_sample, __pyx_k_sample, sizeof(__pyx_k_sample), 0, 0, 1, 1},
  {&__pyx_n_s_sample_2, __pyx_k_sample_2, sizeof(__pyx_k_sample_2), 0, 0, 1, 1},
  {&__pyx_kp_s_self_index_list_cannot_be_conver, __pyx_k_self_index_list_cannot_be_conver, sizeof(__pyx_k_self_index_list_cannot_be_conver), 0, 0, 1, 0},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_spans_in_time, __pyx_k_spans_in_time, sizeof(__pyx_k_spans_in_time), 0, 0, 1, 1},
  {&__pyx_n_s_split_sparse, __pyx_k_split_sparse, sizeof(__pyx_k_split_sparse), 0, 0, 1, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_start_time, __pyx_k_start_time, sizeof(__pyx_k_start_time), 0, 0, 1, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_time, __pyx_k_time, sizeof(__pyx_k_time), 0, 0, 1, 1},
  {&__pyx_n_s_total_intensity, __pyx_k_total_intensity, sizeof(__pyx_k_total_intensity), 0, 0, 1, 1},
  {&__pyx_n_s_total_intensity_members, __pyx_k_total_intensity_members, sizeof(__pyx_k_total_intensity_members), 0, 0, 1, 1},
  {&__pyx_n_s_total_signal, __pyx_k_total_signal, sizeof(__pyx_k_total_signal), 0, 0, 1, 1},
  {&__pyx_n_s_uid, __pyx_k_uid, sizeof(__pyx_k_uid), 0, 0, 1, 1},
  {&__pyx_n_s_uid_2, __pyx_k_uid_2, sizeof(__pyx_k_uid_2), 0, 0, 1, 1},
  {&__pyx_n_s_unspool, __pyx_k_unspool, sizeof(__pyx_k_unspool), 0, 0, 1, 1},
  {&__pyx_n_s_unspool_strip_children, __pyx_k_unspool_strip_children, sizeof(__pyx_k_unspool_strip_children), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_update_from_averager, __pyx_k_update_from_averager, sizeof(__pyx_k_update_from_averager), 0, 0, 1, 1},
  {&__pyx_n_s_used_as_adduct, __pyx_k_used_as_adduct, sizeof(__pyx_k_used_as_adduct), 0, 0, 1, 1},
  {&__pyx_n_s_zeros, __pyx_k_zeros, sizeof(__pyx_k_zeros), 0, 0, 1, 1},
  {&__pyx_n_s_zip, __pyx_k_zip, sizeof(__pyx_k_zip), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 67, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 154, __pyx_L1_error)
  __pyx_builtin_NotImplemented = __Pyx_GetBuiltinName(__pyx_n_s_NotImplemented); if (!__pyx_builtin_NotImplemented) __PYX_ERR(0, 324, __pyx_L1_error)
  __pyx_builtin_zip = __Pyx_GetBuiltinName(__pyx_n_s_zip); if (!__pyx_builtin_zip) __PYX_ERR(0, 751, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(0, 817, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(0, 838, __pyx_L1_error)
  __pyx_builtin_StopIteration = __Pyx_GetBuiltinName(__pyx_n_s_StopIteration); if (!__pyx_builtin_StopIteration) __PYX_ERR(0, 1064, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(3, 944, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":403
 *         best_mz = self._peak_averager.current_mean
 *         if best_mz < 0:
 *             raise ValueError("best_mz < 0")             # <<<<<<<<<<<<<<
 *         self._last_mz = self._mz = best_mz
 * 
 */

#line 403 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_best_mz_0); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 403, __pyx_L1_error)

#line 403 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_tuple__2);

#line 403 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":838
 *         self.index_list = <size_t*>malloc(sizeof(size_t) * n)
 *         if self.index_list == NULL:
 *             raise MemoryError("Failed to allocate index list for LCMS Feature Iterator")             # <<<<<<<<<<<<<<
 *         self.start_time = 0
 *         self.end_time = INF
 */

#line 838 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_Failed_to_allocate_index_list_fo); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 838, __pyx_L1_error)

#line 838 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_tuple__4);

#line 838 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1030
 *                 else:
 *                     if not isinstance(feature, EmptyFeature):
 *                         raise ValueError("Empty Feature is not of type EmptyFeature")             # <<<<<<<<<<<<<<
 *                     node = None
 *                 if node is not None:
 */

#line 1030 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_Empty_Feature_is_not_of_type_Emp); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 1030, __pyx_L1_error)

#line 1030 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_tuple__5);

#line 1030 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.index_list cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.index_list cannot be converted to a Python object for pickling")
 */

#line 2 "(tree fragment)"
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_self_index_list_cannot_be_conver); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(1, 2, __pyx_L1_error)

#line 2 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__6);

#line 2 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "(tree fragment)":4
 *     raise TypeError("self.index_list cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.index_list cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */

#line 4 "(tree fragment)"
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_self_index_list_cannot_be_conver); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__7);

#line 4 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__7);

#line 4 "(tree fragment)"
  __pyx_tuple__8 = PyTuple_Pack(3, __pyx_int_166084213, __pyx_int_141605436, __pyx_int_87109235); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__8);

#line 4 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__8);

#line 4 "(tree fragment)"
  __pyx_tuple__9 = PyTuple_Pack(3, __pyx_int_110791466, __pyx_int_199832089, __pyx_int_23691454); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(1, 4, __pyx_L1_error)

#line 4 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__9);

#line 4 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":944
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(3, 944, __pyx_L1_error)

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_tuple__10);

#line 944 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":950
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(3, 950, __pyx_L1_error)

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GOTREF(__pyx_tuple__11);

#line 950 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "(tree fragment)":1
 * def __pyx_unpickle_FeatureBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"
  __pyx_tuple__12 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(1, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__12);

#line 1 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__12);

#line 1 "(tree fragment)"
  __pyx_codeobj__13 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__12, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_FeatureBase, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__13)) __PYX_ERR(1, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __pyx_tuple__14 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(1, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__14);

#line 1 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__14);

#line 1 "(tree fragment)"
  __pyx_codeobj__15 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__14, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_EmptyFeature, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__15)) __PYX_ERR(1, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __pyx_tuple__16 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(1, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__16);

#line 1 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__16);

#line 1 "(tree fragment)"
  __pyx_codeobj__17 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__16, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_RunningWeightedAv, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__17)) __PYX_ERR(1, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __pyx_tuple__18 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(1, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_tuple__18);

#line 1 "(tree fragment)"
  __Pyx_GIVEREF(__pyx_tuple__18);

#line 1 "(tree fragment)"
  __pyx_codeobj__19 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_RunningWeightedAv_2, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__19)) __PYX_ERR(1, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return 0;

#line 1 "(tree fragment)"
  __pyx_L1_error:;

#line 1 "(tree fragment)"
  __Pyx_RefNannyFinishContext();

#line 1 "(tree fragment)"
  return -1;

#line 1 "(tree fragment)"
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  __pyx_umethod_PyList_Type_pop.type = (PyObject*)&PyList_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_1eneg_4 = PyFloat_FromDouble(1e-4); if (unlikely(!__pyx_float_1eneg_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_23691454 = PyInt_FromLong(23691454L); if (unlikely(!__pyx_int_23691454)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_87109235 = PyInt_FromLong(87109235L); if (unlikely(!__pyx_int_87109235)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_110791466 = PyInt_FromLong(110791466L); if (unlikely(!__pyx_int_110791466)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_141605436 = PyInt_FromLong(141605436L); if (unlikely(!__pyx_int_141605436)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_166084213 = PyInt_FromLong(166084213L); if (unlikely(!__pyx_int_166084213)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_199832089 = PyInt_FromLong(199832089L); if (unlikely(!__pyx_int_199832089)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_12ms_deisotope_2_c_11feature_map_12lcms_feature__zeros = Py_None; Py_INCREF(Py_None);
  __pyx_v_12ms_deisotope_2_c_11feature_map_12lcms_feature_uid = Py_None; Py_INCREF(Py_None);
  __pyx_v_12ms_deisotope_2_c_11feature_map_12lcms_feature_sample = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList = &__pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList._create = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *(*)(PyObject *))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList__create;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList._invalidate = (void (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList__invalidate;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList._make_node = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *, double, PyObject *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList__make_node;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList.find_time = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *, double, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_find_time;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList._find_time = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *, double, size_t *))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList__find_time;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList.getitem = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *, size_t))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_getitem;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList.get_size = (size_t (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList *))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeList_get_size;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList) < 0) __PYX_ERR(0, 31, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList) < 0) __PYX_ERR(0, 31, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_LCMSFeatureTreeList, (PyObject *)&__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList) < 0) __PYX_ERR(0, 31, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList = &__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeList;
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase = &__pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase._total_intensity_members = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase__total_intensity_members;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase.max_intensity = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_max_intensity;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase.total_intensity = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_total_intensity;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase._find_most_abundant_member = (struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase__find_most_abundant_member;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase._calculate_most_abundant_member = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase__calculate_most_abundant_member;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase._recalculate = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase__recalculate;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase.getitem = (struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, size_t))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_getitem;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase.get_members_size = (size_t (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_23LCMSFeatureTreeNodeBase_get_members_size;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase) < 0) __PYX_ERR(0, 175, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase) < 0) __PYX_ERR(0, 175, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_LCMSFeatureTreeNodeBase, (PyObject *)&__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase) < 0) __PYX_ERR(0, 175, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase = &__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase;
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode = &__pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode.__pyx_base = *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode.__pyx_base._recalculate = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode__recalculate;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode._eq = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode__eq;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode._ne = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode__ne;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode.getpeak = (struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, size_t))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_getpeak;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode.get_mz = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_19LCMSFeatureTreeNode_get_mz;
  __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode.tp_base = __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNodeBase;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode) < 0) __PYX_ERR(0, 282, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode) < 0) __PYX_ERR(0, 282, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_LCMSFeatureTreeNode, (PyObject *)&__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode) < 0) __PYX_ERR(0, 282, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode = &__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode;
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase = &__pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase.get_mz = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_get_mz;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase.get_neutral_mass = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_get_neutral_mass;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase.get_start_time = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_get_start_time;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase.get_end_time = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_get_end_time;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase.get_size = (size_t (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_get_size;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase.find_time = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *, double, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase_find_time;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase._find_time = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *, double, size_t *))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11FeatureBase__find_time;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase) < 0) __PYX_ERR(0, 337, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase) < 0) __PYX_ERR(0, 337, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_FeatureBase, (PyObject *)&__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase) < 0) __PYX_ERR(0, 337, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase) < 0) __PYX_ERR(0, 337, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase = &__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase;
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature = &__pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature.__pyx_base = *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature.__pyx_base.get_mz = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_get_mz;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature.__pyx_base.get_start_time = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_get_start_time;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature.__pyx_base.get_end_time = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_get_end_time;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature._eq = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__eq;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature._ne = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__ne;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature.overlaps_in_time = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_overlaps_in_time;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature.spans_in_time = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, double, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_spans_in_time;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature.getitem = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, size_t))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_getitem;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature._initialize_averager = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__initialize_averager;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature._update_from_averager = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__update_from_averager;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature._reaverage_and_update = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__reaverage_and_update;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature.insert_node = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeatureTreeNode *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_insert_node;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature.insert = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, double, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_insert;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature._invalidate = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__invalidate *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature__invalidate;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature.split_sparse = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_split_sparse *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_split_sparse;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature.clone = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch, struct __pyx_opt_args_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_clone *__pyx_optional_args))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_clone;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature.max_intensity = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_11LCMSFeature_max_intensity;
  __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature.tp_base = __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature) < 0) __PYX_ERR(0, 373, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature) < 0) __PYX_ERR(0, 373, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_LCMSFeature, (PyObject *)&__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature) < 0) __PYX_ERR(0, 373, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature = &__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature;
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature = &__pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature.__pyx_base = *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature.__pyx_base.get_mz = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature_get_mz;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature._create = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature *(*)(double))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_12EmptyFeature__create;
  __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature.tp_base = __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature) < 0) __PYX_ERR(0, 790, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature) < 0) __PYX_ERR(0, 790, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_EmptyFeature, (PyObject *)&__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature) < 0) __PYX_ERR(0, 790, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature) < 0) __PYX_ERR(0, 790, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature = &__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_EmptyFeature;
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator = &__pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator._create = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *(*)(PyObject *))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator__create;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator._create_with_threshold = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *(*)(PyObject *, PyObject *, double))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator__create_with_threshold;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator._initialize = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, PyObject *))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator__initialize;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator.init_indices = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_init_indices;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator.get_next_time = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_get_next_time;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator.get_current_time = (double (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_get_current_time;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator.has_more = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_has_more;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator.get_peaks_for_time = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, double, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_get_peaks_for_time;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator.get_next_value = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_get_next_value;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator.get_size = (size_t (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_get_size;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator.getitem = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureBase *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator *, size_t))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_18FeatureSetIterator_getitem;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator) < 0) __PYX_ERR(0, 828, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator) < 0) __PYX_ERR(0, 828, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_FeatureSetIterator, (PyObject *)&__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator) < 0) __PYX_ERR(0, 828, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator) < 0) __PYX_ERR(0, 828, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator = &__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_FeatureSetIterator;
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage = &__pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage.reset = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_reset;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage._initialize = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage__initialize;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage.add = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_add;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage.update = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_update;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage.feed_from_feature = (int (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_LCMSFeature *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage_feed_from_feature;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage._create = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *(*)(PyObject *))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage__create;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage._update = (void (*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, PyObject *))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_22RunningWeightedAverage__update;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage) < 0) __PYX_ERR(0, 1075, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage) < 0) __PYX_ERR(0, 1075, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_RunningWeightedAverage, (PyObject *)&__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage) < 0) __PYX_ERR(0, 1075, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage) < 0) __PYX_ERR(0, 1075, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage = &__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage;
  __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass = &__pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass.__pyx_base = *__pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass.__pyx_base.add = (PyObject *(*)(struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *, struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase *, int __pyx_skip_dispatch))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass_add;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass.__pyx_base._create = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage *(*)(PyObject *))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass__create__pyx_wrap_1;
  __pyx_vtable_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass._create = (struct __pyx_obj_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass *(*)(PyObject *))__pyx_f_12ms_deisotope_2_c_11feature_map_12lcms_feature_33RunningWeightedAverageNeutralMass__create;
  __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass.tp_base = __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverage;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass) < 0) __PYX_ERR(0, 1157, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass.tp_dict, __pyx_vtabptr_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass) < 0) __PYX_ERR(0, 1157, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_RunningWeightedAverageNeutralMas, (PyObject *)&__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass) < 0) __PYX_ERR(0, 1157, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass) < 0) __PYX_ERR(0, 1157, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass = &__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature_RunningWeightedAverageNeutralMass;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone) < 0) __PYX_ERR(0, 144, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone = &__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct__clone;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr) < 0) __PYX_ERR(0, 145, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr = &__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_1_genexpr;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children) < 0) __PYX_ERR(0, 196, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children = &__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_2__unspool_strip_children;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__) < 0) __PYX_ERR(0, 476, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__ = &__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_3___get__;
  if (PyType_Ready(&__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr) < 0) __PYX_ERR(0, 478, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr.tp_dictoffset && __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr = &__pyx_type_12ms_deisotope_2_c_11feature_map_12lcms_feature___pyx_scope_struct_4_genexpr;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("ms_peak_picker._c.peak_set"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "PeakBase", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase) __PYX_ERR(4, 2, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakBase*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakBase->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakBase)) __PYX_ERR(4, 2, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "FittedPeak", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_FittedPeak), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak) __PYX_ERR(4, 12, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_FittedPeak*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_FittedPeak->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_FittedPeak)) __PYX_ERR(4, 12, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "PeakSet", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSet), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet) __PYX_ERR(4, 29, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSet*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSet->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSet)) __PYX_ERR(4, 29, __pyx_L1_error)
  __pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_set", "PeakSetIndexed", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed) __PYX_ERR(4, 83, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_8peak_set_PeakSetIndexed)) __PYX_ERR(4, 83, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(5, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(6, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_7complex_complex) __PYX_ERR(7, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.composition"); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7brainpy_2_c_11composition_PyComposition = __Pyx_ImportType(__pyx_t_1, "brainpy._c.composition", "PyComposition", sizeof(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7brainpy_2_c_11composition_PyComposition) __PYX_ERR(8, 123, __pyx_L1_error)
  __pyx_vtabptr_7brainpy_2_c_11composition_PyComposition = (struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition*)__Pyx_GetVtable(__pyx_ptype_7brainpy_2_c_11composition_PyComposition->tp_dict); if (unlikely(!__pyx_vtabptr_7brainpy_2_c_11composition_PyComposition)) __PYX_ERR(8, 123, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.isotopic_distribution"); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak = __Pyx_ImportType(__pyx_t_1, "brainpy._c.isotopic_distribution", "TheoreticalPeak", sizeof(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak) __PYX_ERR(9, 63, __pyx_L1_error)
  __pyx_vtabptr_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak = (struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak*)__Pyx_GetVtable(__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak->tp_dict); if (unlikely(!__pyx_vtabptr_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak)) __PYX_ERR(9, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(3, 199, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(3, 222, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(3, 226, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(3, 238, __pyx_L1_error)
  __pyx_ptype_5numpy_generic = __Pyx_ImportType(__pyx_t_1, "numpy", "generic", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_generic) __PYX_ERR(3, 770, __pyx_L1_error)
  __pyx_ptype_5numpy_number = __Pyx_ImportType(__pyx_t_1, "numpy", "number", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_number) __PYX_ERR(3, 772, __pyx_L1_error)
  __pyx_ptype_5numpy_integer = __Pyx_ImportType(__pyx_t_1, "numpy", "integer", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_integer) __PYX_ERR(3, 774, __pyx_L1_error)
  __pyx_ptype_5numpy_signedinteger = __Pyx_ImportType(__pyx_t_1, "numpy", "signedinteger", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_signedinteger) __PYX_ERR(3, 776, __pyx_L1_error)
  __pyx_ptype_5numpy_unsignedinteger = __Pyx_ImportType(__pyx_t_1, "numpy", "unsignedinteger", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_unsignedinteger) __PYX_ERR(3, 778, __pyx_L1_error)
  __pyx_ptype_5numpy_inexact = __Pyx_ImportType(__pyx_t_1, "numpy", "inexact", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_inexact) __PYX_ERR(3, 780, __pyx_L1_error)
  __pyx_ptype_5numpy_floating = __Pyx_ImportType(__pyx_t_1, "numpy", "floating", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_floating) __PYX_ERR(3, 782, __pyx_L1_error)
  __pyx_ptype_5numpy_complexfloating = __Pyx_ImportType(__pyx_t_1, "numpy", "complexfloating", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_complexfloating) __PYX_ERR(3, 784, __pyx_L1_error)
  __pyx_ptype_5numpy_flexible = __Pyx_ImportType(__pyx_t_1, "numpy", "flexible", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_flexible) __PYX_ERR(3, 786, __pyx_L1_error)
  __pyx_ptype_5numpy_character = __Pyx_ImportType(__pyx_t_1, "numpy", "character", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_character) __PYX_ERR(3, 788, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(3, 826, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_peak_picker._c.peak_index"); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex = __Pyx_ImportType(__pyx_t_1, "ms_peak_picker._c.peak_index", "PeakIndex", sizeof(struct __pyx_obj_14ms_peak_picker_2_c_10peak_index_PeakIndex), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex) __PYX_ERR(10, 5, __pyx_L1_error)
  __pyx_vtabptr_14ms_peak_picker_2_c_10peak_index_PeakIndex = (struct __pyx_vtabstruct_14ms_peak_picker_2_c_10peak_index_PeakIndex*)__Pyx_GetVtable(__pyx_ptype_14ms_peak_picker_2_c_10peak_index_PeakIndex->tp_dict); if (unlikely(!__pyx_vtabptr_14ms_peak_picker_2_c_10peak_index_PeakIndex)) __PYX_ERR(10, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.averagine"); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.averagine", "Averagine", sizeof(struct __pyx_obj_12ms_deisotope_2_c_9averagine_Averagine), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine) __PYX_ERR(11, 12, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_9averagine_Averagine = (struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_Averagine*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_9averagine_Averagine->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_9averagine_Averagine)) __PYX_ERR(11, 12, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.averagine", "TheoreticalIsotopicPattern", sizeof(struct __pyx_obj_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern) __PYX_ERR(11, 23, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern = (struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_9averagine_TheoreticalIsotopicPattern)) __PYX_ERR(11, 23, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.averagine", "AveragineCache", sizeof(struct __pyx_obj_12ms_deisotope_2_c_9averagine_AveragineCache), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache) __PYX_ERR(11, 79, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_9averagine_AveragineCache = (struct __pyx_vtabstruct_12ms_deisotope_2_c_9averagine_AveragineCache*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_9averagine_AveragineCache->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_9averagine_AveragineCache)) __PYX_ERR(11, 79, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.deconvoluter_base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.deconvoluter_base", "DeconvoluterBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase) __PYX_ERR(12, 25, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_DeconvoluterBase)) __PYX_ERR(12, 25, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.deconvoluter_base", "AveragineDeconvoluterBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase) __PYX_ERR(12, 57, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_AveragineDeconvoluterBase)) __PYX_ERR(12, 57, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.deconvoluter_base", "MultiAveragineDeconvoluterBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase) __PYX_ERR(12, 69, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_MultiAveragineDeconvoluterBase)) __PYX_ERR(12, 69, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.deconvoluter_base", "ChargeIterator", sizeof(struct __pyx_obj_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator) __PYX_ERR(12, 91, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator = (struct __pyx_vtabstruct_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_17deconvoluter_base_ChargeIterator)) __PYX_ERR(12, 91, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.scoring"); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "IsotopicFitRecord", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitRecord), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord) __PYX_ERR(13, 16, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitRecord = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitRecord*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitRecord->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitRecord)) __PYX_ERR(13, 16, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "FitSelectorBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FitSelectorBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase) __PYX_ERR(13, 38, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_FitSelectorBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FitSelectorBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_FitSelectorBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FitSelectorBase)) __PYX_ERR(13, 38, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "IsotopicFitterBase", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_IsotopicFitterBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase) __PYX_ERR(13, 50, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_IsotopicFitterBase*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_IsotopicFitterBase->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_IsotopicFitterBase)) __PYX_ERR(13, 50, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "LeastSquaresFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_LeastSquaresFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter) __PYX_ERR(13, 62, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_LeastSquaresFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_LeastSquaresFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_LeastSquaresFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_LeastSquaresFitter)) __PYX_ERR(13, 62, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "GTestFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_GTestFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter) __PYX_ERR(13, 66, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_GTestFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_GTestFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_GTestFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_GTestFitter)) __PYX_ERR(13, 66, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "ScaledGTestFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledGTestFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter) __PYX_ERR(13, 69, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledGTestFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledGTestFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledGTestFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledGTestFitter)) __PYX_ERR(13, 69, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "MSDeconVFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_MSDeconVFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter) __PYX_ERR(13, 72, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_MSDeconVFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_MSDeconVFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_MSDeconVFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_MSDeconVFitter)) __PYX_ERR(13, 72, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "PenalizedMSDeconVFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter) __PYX_ERR(13, 78, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_PenalizedMSDeconVFitter)) __PYX_ERR(13, 78, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "FunctionScorer", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_FunctionScorer), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer) __PYX_ERR(13, 84, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_FunctionScorer = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_FunctionScorer*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_FunctionScorer->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_FunctionScorer)) __PYX_ERR(13, 84, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "InterferenceDetection", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_InterferenceDetection), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection) __PYX_ERR(13, 89, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_InterferenceDetection = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_InterferenceDetection*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_InterferenceDetection->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_InterferenceDetection)) __PYX_ERR(13, 89, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "DistinctPatternFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DistinctPatternFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter) __PYX_ERR(13, 95, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_DistinctPatternFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DistinctPatternFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_DistinctPatternFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DistinctPatternFitter)) __PYX_ERR(13, 95, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "ScaledPenalizedMSDeconvFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter) __PYX_ERR(13, 102, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_ScaledPenalizedMSDeconvFitter)) __PYX_ERR(13, 102, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.scoring", "DotProductFitter", sizeof(struct __pyx_obj_12ms_deisotope_2_c_7scoring_DotProductFitter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter) __PYX_ERR(13, 112, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_7scoring_DotProductFitter = (struct __pyx_vtabstruct_12ms_deisotope_2_c_7scoring_DotProductFitter*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_7scoring_DotProductFitter->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_7scoring_DotProductFitter)) __PYX_ERR(13, 112, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.peak_set"); if (unlikely(!__pyx_t_1)) __PYX_ERR(14, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12ms_deisotope_2_c_8peak_set__Index = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "_Index", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__Index), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set__Index) __PYX_ERR(14, 5, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set__Index = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__Index*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set__Index->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set__Index)) __PYX_ERR(14, 5, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "EnvelopePair", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_EnvelopePair), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair) __PYX_ERR(14, 14, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_EnvelopePair = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_EnvelopePair*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_EnvelopePair->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_EnvelopePair)) __PYX_ERR(14, 14, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "Envelope", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_Envelope), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope) __PYX_ERR(14, 25, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_Envelope = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_Envelope*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_Envelope->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_Envelope)) __PYX_ERR(14, 25, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "DeconvolutedPeak", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak) __PYX_ERR(14, 40, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeak)) __PYX_ERR(14, 40, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "IonMobilityDeconvolutedPeak", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak) __PYX_ERR(14, 66, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_IonMobilityDeconvolutedPeak)) __PYX_ERR(14, 66, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "DeconvolutedPeakSolution", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution) __PYX_ERR(14, 71, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSolution)) __PYX_ERR(14, 71, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "DeconvolutedPeakSet", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet) __PYX_ERR(14, 75, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSet)) __PYX_ERR(14, 75, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "DeconvolutedPeakSetIndexed", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed) __PYX_ERR(14, 104, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set_DeconvolutedPeakSetIndexed)) __PYX_ERR(14, 104, __pyx_L1_error)
  __pyx_ptype_12ms_deisotope_2_c_8peak_set__CPeakSet = __Pyx_ImportType(__pyx_t_1, "ms_deisotope._c.peak_set", "_CPeakSet", sizeof(struct __pyx_obj_12ms_deisotope_2_c_8peak_set__CPeakSet), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12ms_deisotope_2_c_8peak_set__CPeakSet) __PYX_ERR(14, 161, __pyx_L1_error)
  __pyx_vtabptr_12ms_deisotope_2_c_8peak_set__CPeakSet = (struct __pyx_vtabstruct_12ms_deisotope_2_c_8peak_set__CPeakSet*)__Pyx_GetVtable(__pyx_ptype_12ms_deisotope_2_c_8peak_set__CPeakSet->tp_dict); if (unlikely(!__pyx_vtabptr_12ms_deisotope_2_c_8peak_set__CPeakSet)) __PYX_ERR(14, 161, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __pyx_t_1 = PyImport_ImportModule("ms_peak_picker._c.peak_set"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "INTERVAL_INDEX_SIZE", (void **)&__pyx_vp_14ms_peak_picker_2_c_8peak_set_INTERVAL_INDEX_SIZE, "size_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.composition"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "nist_mass", (void **)&__pyx_vp_7brainpy_2_c_11composition_nist_mass, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "PROTON", (void **)&__pyx_vp_7brainpy_2_c_11composition_PROTON, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "_ElementTable", (void **)&__pyx_vp_7brainpy_2_c_11composition__ElementTable, "struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("brainpy._c.isotopic_constants"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "DEFAULT_ISOTOPIC_CONSTANTS_SIZE", (void **)&__pyx_vp_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE, "size_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.averagine"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "PROTON", (void **)&__pyx_vp_12ms_deisotope_2_c_9averagine_PROTON, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.scoring"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "INFINITY", (void **)&__pyx_vp_12ms_deisotope_2_c_7scoring_INFINITY, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ms_deisotope._c.peak_set"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "INTERVAL_INDEX_SIZE", (void **)&__pyx_vp_12ms_deisotope_2_c_8peak_set_INTERVAL_INDEX_SIZE, "size_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initlcms_feature(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initlcms_feature(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_lcms_feature(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_lcms_feature(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_lcms_feature(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  double __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'lcms_feature' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_lcms_feature(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("lcms_feature", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_ms_deisotope___c__feature_map__lcms_feature) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "ms_deisotope._c.feature_map.lcms_feature")) {
      if (unlikely(PyDict_SetItemString(modules, "ms_deisotope._c.feature_map.lcms_feature", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_variable_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":5
 * cimport cython
 * 
 * from ms_deisotope.utils import uid as _uid             # <<<<<<<<<<<<<<
 * 
 * from collections import deque
 */

#line 5 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)

#line 5 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 5 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_n_s_uid);

#line 5 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_n_s_uid);

#line 5 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_uid);

#line 5 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_ms_deisotope_utils, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)

#line 5 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 5 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_uid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)

#line 5 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 5 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_uid_2, __pyx_t_1) < 0) __PYX_ERR(0, 5, __pyx_L1_error)

#line 5 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 5 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":7
 * from ms_deisotope.utils import uid as _uid
 * 
 * from collections import deque             # <<<<<<<<<<<<<<
 * from random import sample as _sample
 * 
 */

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7, __pyx_L1_error)

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_n_s_deque);

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_n_s_deque);

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_deque);

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_collections, __pyx_t_2, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_deque); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7, __pyx_L1_error)

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_deque, __pyx_t_2) < 0) __PYX_ERR(0, 7, __pyx_L1_error)

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 7 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":8
 * 
 * from collections import deque
 * from random import sample as _sample             # <<<<<<<<<<<<<<
 * 
 * from libc.stdlib cimport malloc, free
 */

#line 8 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 8, __pyx_L1_error)

#line 8 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 8 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_INCREF(__pyx_n_s_sample);

#line 8 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_n_s_sample);

#line 8 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_sample);

#line 8 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_random, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 8, __pyx_L1_error)

#line 8 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 8 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 8 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_sample); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 8, __pyx_L1_error)

#line 8 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 8 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sample_2, __pyx_t_1) < 0) __PYX_ERR(0, 8, __pyx_L1_error)

#line 8 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 8 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":20
 * 
 * cimport numpy as np
 * import numpy as np             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 20 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_numpy, 0, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 20, __pyx_L1_error)

#line 20 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 20 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_2) < 0) __PYX_ERR(0, 20, __pyx_L1_error)

#line 20 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":23
 * 
 * 
 * np.import_array()             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 23 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_3 = __pyx_f_5numpy_import_array(); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 23, __pyx_L1_error)

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":26
 * 
 * 
 * cdef object _zeros = np.zeros             # <<<<<<<<<<<<<<
 * 
 * cdef double INF = float("inf")
 */

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 26, __pyx_L1_error)

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 26, __pyx_L1_error)

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGOTREF(__pyx_v_12ms_deisotope_2_c_11feature_map_12lcms_feature__zeros);

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF_SET(__pyx_v_12ms_deisotope_2_c_11feature_map_12lcms_feature__zeros, __pyx_t_1);

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 26 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":28
 * cdef object _zeros = np.zeros
 * 
 * cdef double INF = float("inf")             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_4 = __Pyx_PyObject_AsDouble(__pyx_n_s_inf); if (unlikely(__pyx_t_4 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(0, 28, __pyx_L1_error)

#line 28 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_v_12ms_deisotope_2_c_11feature_map_12lcms_feature_INF = __pyx_t_4;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":171
 * 
 * 
 * cdef object uid = _uid             # <<<<<<<<<<<<<<
 * cdef object sample = _sample
 * 
 */

#line 171 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_uid_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 171, __pyx_L1_error)

#line 171 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 171 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGOTREF(__pyx_v_12ms_deisotope_2_c_11feature_map_12lcms_feature_uid);

#line 171 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF_SET(__pyx_v_12ms_deisotope_2_c_11feature_map_12lcms_feature_uid, __pyx_t_1);

#line 171 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 171 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":172
 * 
 * cdef object uid = _uid
 * cdef object sample = _sample             # <<<<<<<<<<<<<<
 * 
 * 
 */

#line 172 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_sample_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 172, __pyx_L1_error)

#line 172 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 172 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_XGOTREF(__pyx_v_12ms_deisotope_2_c_11feature_map_12lcms_feature_sample);

#line 172 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF_SET(__pyx_v_12ms_deisotope_2_c_11feature_map_12lcms_feature_sample, __pyx_t_1);

#line 172 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_1);

#line 172 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_1 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":554
 *         return self.get_start_time() <= time <= self.get_end_time()
 * 
 *     def as_arrays(self, dtype=np.float64):             # <<<<<<<<<<<<<<
 *         """Convert this object into a pair of time and intensity arrays, summing
 *         all peaks at the same time point.
 */

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 554, __pyx_L1_error)

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_1);

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_float64); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 554, __pyx_L1_error)

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_k__3 = __pyx_t_2;

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GIVEREF(__pyx_t_2);

#line 554 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_FeatureBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_11feature_map_12lcms_feature_1__pyx_unpickle_FeatureBase, NULL, __pyx_n_s_ms_deisotope__c_feature_map_lcms); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_2);

#line 1 "(tree fragment)"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_FeatureBase, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_FeatureBase__set_state(<FeatureBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_FeatureBase__set_state(FeatureBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._end_time = __pyx_state[0]; __pyx_result._mz = __pyx_state[1]; __pyx_result._start_time = __pyx_state[2]; __pyx_result.nodes = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_11feature_map_12lcms_feature_3__pyx_unpickle_EmptyFeature, NULL, __pyx_n_s_ms_deisotope__c_feature_map_lcms); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)

#line 11 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_2);

#line 11 "(tree fragment)"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_EmptyFeature, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)

#line 11 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_RunningWeightedAverage(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

#line 1 "(tree fragment)"
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_11feature_map_12lcms_feature_5__pyx_unpickle_RunningWeightedAverage, NULL, __pyx_n_s_ms_deisotope__c_feature_map_lcms); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_2);

#line 1 "(tree fragment)"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_RunningWeightedAv, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)

#line 1 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_RunningWeightedAverage__set_state(<RunningWeightedAverage> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_RunningWeightedAverage__set_state(RunningWeightedAverage __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.current_count = __pyx_state[0]; __pyx_result.current_mean = __pyx_state[1]; __pyx_result.total_weight = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

#line 11 "(tree fragment)"
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_12ms_deisotope_2_c_11feature_map_12lcms_feature_7__pyx_unpickle_RunningWeightedAverageNeutralMass, NULL, __pyx_n_s_ms_deisotope__c_feature_map_lcms); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)

#line 11 "(tree fragment)"
  __Pyx_GOTREF(__pyx_t_2);

#line 11 "(tree fragment)"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_RunningWeightedAv_2, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)

#line 11 "(tree fragment)"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "ms_deisotope/_c/feature_map/lcms_feature.pyx":1
 * # cython: embedsignature=True             # <<<<<<<<<<<<<<
 * 
 * cimport cython
 */

#line 1 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)

#line 1 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_GOTREF(__pyx_t_2);

#line 1 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)

#line 1 "ms_deisotope/_c/feature_map/lcms_feature.pyx"
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd":1013
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  /*--- Wrapped vars code ---*/

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"


#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  goto __pyx_L0;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L1_error:;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_1);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_XDECREF(__pyx_t_2);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  if (__pyx_m) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    if (__pyx_d) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
      __Pyx_AddTraceback("init ms_deisotope._c.feature_map.lcms_feature", __pyx_clineno, __pyx_lineno, __pyx_filename);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    Py_CLEAR(__pyx_m);

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  } else if (!PyErr_Occurred()) {

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
    PyErr_SetString(PyExc_ImportError, "init ms_deisotope._c.feature_map.lcms_feature");

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  }

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __pyx_L0:;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  __Pyx_RefNannyFinishContext();

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #if CYTHON_PEP489_MULTI_PHASE_INIT

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return (__pyx_m != NULL) ? 0 : -1;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #elif PY_MAJOR_VERSION >= 3

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return __pyx_m;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #else

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  return;

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
  #endif

#line 1013 "../../../hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/numpy/__init__.pxd"
}

#line 1 "cython_utility"

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#if defined(__Pyx_CyFunction_USED) && defined(NDEBUG)
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* DivInt[long] */
static CYTHON_INLINE long __Pyx_div_long(long a, long b) {
    long q = a / b;
    long r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* None */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod0 */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* pop */
static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L) {
    if (Py_TYPE(L) == &PySet_Type) {
        return PySet_Pop(L);
    }
    return __Pyx_PyObject_CallMethod0(L, __pyx_n_s_pop);
}
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L) {
    if (likely(PyList_GET_SIZE(L) > (((PyListObject*)L)->allocated >> 1))) {
        __Pyx_SET_SIZE(L, Py_SIZE(L) - 1);
        return PyList_GET_ITEM(L, PyList_GET_SIZE(L));
    }
    return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyList_Type_pop, L);
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* IsLittleEndian */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void)
{
  union {
    uint32_t u32;
    uint8_t u8[4];
  } S;
  S.u32 = 0x01020304;
  return S.u8[0] == 4;
}

/* BufferFormatCheck */
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t <= '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1)
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case '?': return "'bool'";
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparseable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    #ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    #endif
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case '?': case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count;
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break;
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue;
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static PyObject *
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number, ndim;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return NULL;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
    ndim = ctx->head->field->type->ndim;
    while (*ts && *ts != ')') {
        switch (*ts) {
            case ' ': case '\f': case '\r': case '\n': case '\t': case '\v':  continue;
            default:  break;
        }
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return NULL;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i])
            return PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
        if (*ts != ',' && *ts != ')')
            return PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim)
        return PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return NULL;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return Py_None;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        return ts;
      case ' ':
      case '\r':
      case '\n':
        ++ts;
        break;
      case '<':
        if (!__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T':
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}':
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }
        CYTHON_FALLTHROUGH;
      case '?': case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 'p':
        if ((ctx->enc_type == *ts) && (got_Z == ctx->is_complex) &&
            (ctx->enc_packmode == ctx->new_packmode) && (!ctx->is_valid_array)) {
          ctx->enc_count += ctx->new_count;
          ctx->new_count = 1;
          got_Z = 0;
          ++ts;
          break;
        }
        CYTHON_FALLTHROUGH;
      case 's':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->enc_count = ctx->new_count;
        ctx->enc_packmode = ctx->new_packmode;
        ctx->enc_type = *ts;
        ctx->is_complex = got_Z;
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (!__pyx_buffmt_parse_array(ctx, &ts)) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}

/* BufferGetAndValidate */
  static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info) {
  if (unlikely(info->buf == NULL)) return;
  if (info->suboffsets == __Pyx_minusones) info->suboffsets = NULL;
  __Pyx_ReleaseBuffer(info);
}
static void __Pyx_ZeroBuffer(Py_buffer* buf) {
  buf->buf = NULL;
  buf->obj = NULL;
  buf->strides = __Pyx_zeros;
  buf->shape = __Pyx_zeros;
  buf->suboffsets = __Pyx_minusones;
}
static int __Pyx__GetBufferAndValidate(
        Py_buffer* buf, PyObject* obj,  __Pyx_TypeInfo* dtype, int flags,
        int nd, int cast, __Pyx_BufFmt_StackElem* stack)
{
  buf->buf = NULL;
  if (unlikely(__Pyx_GetBuffer(obj, buf, flags) == -1)) {
    __Pyx_ZeroBuffer(buf);
    return -1;
  }
  if (unlikely(buf->ndim != nd)) {
    PyErr_Format(PyExc_ValueError,
                 "Buffer has wrong number of dimensions (expected %d, got %d)",
                 nd, buf->ndim);
    goto fail;
  }
  if (!cast) {
    __Pyx_BufFmt_Context ctx;
    __Pyx_BufFmt_Init(&ctx, stack, dtype);
    if (!__Pyx_BufFmt_CheckString(&ctx, buf->format)) goto fail;
  }
  if (unlikely((size_t)buf->itemsize != dtype->size)) {
    PyErr_Format(PyExc_ValueError,
      "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "d byte%s) does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "d byte%s)",
      buf->itemsize, (buf->itemsize > 1) ? "s" : "",
      dtype->name, (Py_ssize_t)dtype->size, (dtype->size > 1) ? "s" : "");
    goto fail;
  }
  if (buf->suboffsets == NULL) buf->suboffsets = __Pyx_minusones;
  return 0;
fail:;
  __Pyx_SafeReleaseBuffer(buf);
  return -1;
}

/* BufferFallbackError */
  static void __Pyx_RaiseBufferFallbackError(void) {
  PyErr_SetString(PyExc_ValueError,
     "Buffer acquisition failed on assignment; and then reacquiring the old buffer failed too!");
}

/* BufferIndexError */
  static void __Pyx_RaiseBufferIndexError(int axis) {
  PyErr_Format(PyExc_IndexError,
     "Out of bounds on buffer access (axis %d)", axis);
}

/* PyObjectSetAttrStr */
  #if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* RaiseNoneIterError */
  static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* PyIntBinop */
  #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* SliceObject */
  static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* py_abs */
  #if CYTHON_USE_PYLONG_INTERNALS
static PyObject *__Pyx_PyLong_AbsNeg(PyObject *n) {
    if (likely(Py_SIZE(n) == -1)) {
        return PyLong_FromLong(((PyLongObject*)n)->ob_digit[0]);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    {
        PyObject *copy = _PyLong_Copy((PyLongObject*)n);
        if (likely(copy)) {
            __Pyx_SET_SIZE(copy, -Py_SIZE(copy));
        }
        return copy;
    }
#else
    return PyNumber_Negative(n);
#endif
}
#endif

/* Import */
  static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
  static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* HasAttr */
  static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* PyObject_GenericGetAttrNoDict */
  #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
  #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
  static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyObjectGetAttrStrNoError */
  static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
  static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
  #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
  static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* pyobject_as_double */
  static double __Pyx__PyObject_AsDouble(PyObject* obj) {
    PyObject* float_value;
#if !CYTHON_USE_TYPE_SLOTS
    float_value = PyNumber_Float(obj);  if ((0)) goto bad;
#else
    PyNumberMethods *nb = Py_TYPE(obj)->tp_as_number;
    if (likely(nb) && likely(nb->nb_float)) {
        float_value = nb->nb_float(obj);
        if (likely(float_value) && unlikely(!PyFloat_Check(float_value))) {
            PyErr_Format(PyExc_TypeError,
                "__float__ returned non-float (type %.200s)",
                Py_TYPE(float_value)->tp_name);
            Py_DECREF(float_value);
            goto bad;
        }
    } else if (PyUnicode_CheckExact(obj) || PyBytes_CheckExact(obj)) {
#if PY_MAJOR_VERSION >= 3
        float_value = PyFloat_FromString(obj);
#else
        float_value = PyFloat_FromString(obj, 0);
#endif
    } else {
        PyObject* args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        PyTuple_SET_ITEM(args, 0, obj);
        float_value = PyObject_Call((PyObject*)&PyFloat_Type, args, 0);
        PyTuple_SET_ITEM(args, 0, 0);
        Py_DECREF(args);
    }
#endif
    if (likely(float_value)) {
        double value = PyFloat_AS_DOUBLE(float_value);
        Py_DECREF(float_value);
        return value;
    }
bad:
    return (double)-1;
}

/* CLineInTraceback */
  #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
  static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
  #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
  #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
    PyErr_Format(PyExc_TypeError, "'%.200s' does not have the buffer interface", Py_TYPE(obj)->tp_name);
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
    if ((0)) {}
    view->obj = NULL;
    Py_DECREF(obj);
}
#endif


  /* Print */
  #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION < 3
static PyObject *__Pyx_GetStdout(void) {
    PyObject *f = PySys_GetObject((char *)"stdout");
    if (!f) {
        PyErr_SetString(PyExc_RuntimeError, "lost sys.stdout");
    }
    return f;
}
static int __Pyx_Print(PyObject* f, PyObject *arg_tuple, int newline) {
    int i;
    if (!f) {
        if (!(f = __Pyx_GetStdout()))
            return -1;
    }
    Py_INCREF(f);
    for (i=0; i < PyTuple_GET_SIZE(arg_tuple); i++) {
        PyObject* v;
        if (PyFile_SoftSpace(f, 1)) {
            if (PyFile_WriteString(" ", f) < 0)
                goto error;
        }
        v = PyTuple_GET_ITEM(arg_tuple, i);
        if (PyFile_WriteObject(v, f, Py_PRINT_RAW) < 0)
            goto error;
        if (PyString_Check(v)) {
            char *s = PyString_AsString(v);
            Py_ssize_t len = PyString_Size(v);
            if (len > 0) {
                switch (s[len-1]) {
                    case ' ': break;
                    case '\f': case '\r': case '\n': case '\t': case '\v':
                        PyFile_SoftSpace(f, 0);
                        break;
                    default:  break;
                }
            }
        }
    }
    if (newline) {
        if (PyFile_WriteString("\n", f) < 0)
            goto error;
        PyFile_SoftSpace(f, 0);
    }
    Py_DECREF(f);
    return 0;
error:
    Py_DECREF(f);
    return -1;
}
#else
static int __Pyx_Print(PyObject* stream, PyObject *arg_tuple, int newline) {
    PyObject* kwargs = 0;
    PyObject* result = 0;
    PyObject* end_string;
    if (unlikely(!__pyx_print)) {
        __pyx_print = PyObject_GetAttr(__pyx_b, __pyx_n_s_print);
        if (!__pyx_print)
            return -1;
    }
    if (stream) {
        kwargs = PyDict_New();
        if (unlikely(!kwargs))
            return -1;
        if (unlikely(PyDict_SetItem(kwargs, __pyx_n_s_file, stream) < 0))
            goto bad;
        if (!newline) {
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                goto bad;
            if (PyDict_SetItem(kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                goto bad;
            }
            Py_DECREF(end_string);
        }
    } else if (!newline) {
        if (unlikely(!__pyx_print_kwargs)) {
            __pyx_print_kwargs = PyDict_New();
            if (unlikely(!__pyx_print_kwargs))
                return -1;
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                return -1;
            if (PyDict_SetItem(__pyx_print_kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                return -1;
            }
            Py_DECREF(end_string);
        }
        kwargs = __pyx_print_kwargs;
    }
    result = PyObject_Call(__pyx_print, arg_tuple, kwargs);
    if (unlikely(kwargs) && (kwargs != __pyx_print_kwargs))
        Py_DECREF(kwargs);
    if (!result)
        return -1;
    Py_DECREF(result);
    return 0;
bad:
    if (kwargs != __pyx_print_kwargs)
        Py_XDECREF(kwargs);
    return -1;
}
#endif

/* Declarations */
  #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
  #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntFromPy */
  static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* PrintOne */
  #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION < 3
static int __Pyx_PrintOne(PyObject* f, PyObject *o) {
    if (!f) {
        if (!(f = __Pyx_GetStdout()))
            return -1;
    }
    Py_INCREF(f);
    if (PyFile_SoftSpace(f, 0)) {
        if (PyFile_WriteString(" ", f) < 0)
            goto error;
    }
    if (PyFile_WriteObject(o, f, Py_PRINT_RAW) < 0)
        goto error;
    if (PyFile_WriteString("\n", f) < 0)
        goto error;
    Py_DECREF(f);
    return 0;
error:
    Py_DECREF(f);
    return -1;
    /* the line below is just to avoid C compiler
     * warnings about unused functions */
    return __Pyx_Print(f, NULL, 0);
}
#else
static int __Pyx_PrintOne(PyObject* stream, PyObject *o) {
    int res;
    PyObject* arg_tuple = PyTuple_Pack(1, o);
    if (unlikely(!arg_tuple))
        return -1;
    res = __Pyx_Print(stream, arg_tuple, 1);
    Py_DECREF(arg_tuple);
    return res;
}
#endif

/* FastTypeChecks */
  #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* FetchCommonType */
  static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* SwapException */
  #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* PyObjectCallMethod1 */
  static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* CoroutineBase */
  #include <structmember.h>
#include <frameobject.h>
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            assert(f->f_back == NULL);
            #if PY_VERSION_HEX >= 0x030B00A1
            f->f_back = PyThreadState_GetFrame(tstate);
            #else
            Py_XINCREF(tstate->frame);
            f->f_back = tstate->frame;
            #endif
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
            _PyGen_SetStopIterationValue(result);
        }
        Py_CLEAR(result);
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (Py_REFCNT(self) > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(Py_REFCNT(self) > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = Py_REFCNT(self);
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(Py_TYPE(self)) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *frame = self->gi_frame;
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
  static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
  #ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
  static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {(char *) "gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
  static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* VoidPtrImport */
  #ifndef __PYX_HAVE_RT_ImportVoidPtr
#define __PYX_HAVE_RT_ImportVoidPtr
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, name);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C variable %.200s",
                PyModule_GetName(module), name);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, desc);
        goto bad;
    }
    *p = PyCObject_AsVoidPtr(cobj);}
#endif
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
  static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
